Package: reproducible
Type: Package
Title: Enhance Reproducibility of R Code
Description: A collection of high-level, machine- and OS-independent tools
    for making reproducible and reusable content in R.
    The two workhorse functions are 'Cache()' and 'prepInputs()'.
    'Cache()' allows for nested caching, is robust to environments and objects
    with environments (like functions), and deals with some classes of 
    file-backed R objects e.g., from 'terra' and 'raster' packages. 
    Both functions have been developed to be foundational components of data
    retrieval and processing in continuous workflow situations. In both functions,
    efforts are made to make the first and subsequent calls of functions have 
    the same result, but faster at subsequent times by way of checksums
    and digesting. Several features are still under development, including
    cloud storage of cached objects allowing for sharing between users. Several
    advanced options are available, see '?reproducibleOptions()'.
SystemRequirements: 'unrar' (Linux/macOS) or '7-Zip' (Windows) to work
        with '.rar' files.
URL: https://reproducible.predictiveecology.org,
        https://github.com/PredictiveEcology/reproducible
Date: 2026-05-17
Version: 3.1.1
Authors@R: 
    c(person(given = "Eliot J B",
             family = "McIntire",
             role = c("aut", "cre"),
             email = "eliot.mcintire@canada.ca",
             comment = c(ORCID = "0000-0002-6914-8316")),
      person(given = "Alex M",
             family = "Chubaty",
             role = c("aut"),
             email = "achubaty@for-cast.ca",
             comment = c(ORCID = "0000-0001-7146-8135")),
      person(given = "Tati",
             family = "Micheletti",
             role = "ctb",
             email = "tati.micheletti@gmail.com",
             comment = c(ORCID = "0000-0003-4838-8342")),
      person(given = "Ceres",
             family = "Barros",
             role = "ctb",
             email = "ceres.barros@ubc.ca",
             comment = c(ORCID = "0000-0003-4036-977X")),
      person(given = "Ian",
             family = "Eddy",
             role = "ctb",
             email = "ian.eddy@nrcan-rncan.gc.ca",
             comment = c(ORCID = "0000-0001-7397-2116")),
      person(given = "His Majesty the King in Right of Canada, as represented by the Minister of Natural Resources Canada",
             role = "cph"))
Depends: R (>= 4.3)
Imports: cli, data.table (>= 1.10.4), digest, filelock (>= 1.0.3),
        fpCompare, fs, lobstr, methods, stats, tools, utils
Suggests: archive, covr, curl (>= 7.0.0), DBI, future, geodata, glue,
        googledrive, httr, httr2 (>= 1.2.1), knitr, parallel,
        parallelly, qs2, raster (>= 3.5-15), RCurl (>= 1.95-4.8),
        rlang, rmarkdown, RSQLite, R.utils, rvest, sf, sp (>= 1.4-2),
        terra (>= 1.7-20), testthat, withr
Encoding: UTF-8
Language: en-CA
License: GPL-3
VignetteBuilder: knitr, rmarkdown
BugReports: https://github.com/PredictiveEcology/reproducible/issues
ByteCompile: yes
Config/roxygen2/version: 8.0.0
Collate: 'DBI.R' 'messages.R' 'GPT2.R' 'cache-helpers.R'
        'cache-internals.R' 'robustDigest.R' 'cache.R' 'cacheGeo.R'
        'checksums.R' 'cloud.R' 'convertPaths.R' 'copy.R' 'download.R'
        'downloadTileAndUpload.R' 'exportedMethods.R' 'gis.R'
        'helpers.R' 'listNamed.R' 'objectSize.R' 'options.R'
        'packages.R' 'paths.R' 'pipe.R' 'postProcess.R'
        'postProcessTo.R' 'preProcess.R' 'prepInputs.R'
        'prepInputsCOG.R' 'reproducible-deprecated.R'
        'reproducible-package.R' 'search.R' 'showCacheEtc.R'
        'spatialObjects-class.R' 'terra-migration.R' 'zzz.R'
NeedsCompilation: no
Packaged: 2026-05-18 04:44:29 UTC; emcintir
Author: Eliot J B McIntire [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6914-8316>),
  Alex M Chubaty [aut] (ORCID: <https://orcid.org/0000-0001-7146-8135>),
  Tati Micheletti [ctb] (ORCID: <https://orcid.org/0000-0003-4838-8342>),
  Ceres Barros [ctb] (ORCID: <https://orcid.org/0000-0003-4036-977X>),
  Ian Eddy [ctb] (ORCID: <https://orcid.org/0000-0001-7397-2116>),
  His Majesty the King in Right of Canada, as represented by the Minister
    of Natural Resources Canada [cph]
Maintainer: Eliot J B McIntire <eliot.mcintire@canada.ca>
Repository: CRAN
Date/Publication: 2026-05-18 07:30:08 UTC
Built: R 4.6.0; ; 2026-05-18 10:04:20 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(.unwrap,PackedSpatExtent2)
S3method(.unwrap,PackedSpatVector)
S3method(.unwrap,PackedSpatVector2)
S3method(.unwrap,data.table)
S3method(.unwrap,default)
S3method(.unwrap,environment)
S3method(.unwrap,list)
S3method(.wrap,default)
S3method(.wrap,environment)
S3method(.wrap,list)
S3method(asPath,character)
S3method(asPath,null)
S3method(assessDataType,default)
S3method(cropInputs,default)
S3method(fixErrors,default)
S3method(makeMemoisable,data.table)
S3method(makeMemoisable,default)
S3method(maskInputs,default)
S3method(objSize,"function")
S3method(objSize,Path)
S3method(objSize,default)
S3method(objSize,environment)
S3method(objSize,list)
S3method(postProcess,default)
S3method(postProcess,list)
S3method(projectInputs,default)
S3method(unmakeMemoisable,default)
S3method(updateFilenameSlots,default)
S3method(updateFilenameSlots,environment)
S3method(updateFilenameSlots,list)
S3method(usesPointer,default)
S3method(usesPointer,environment)
S3method(usesPointer,list)
S3method(writeOutputs,default)
export(.addChangedAttr)
export(.addTagsRepo)
export(.addTagsToOutput)
export(.cacheMessage)
export(.cacheMessageObjectToRetrieve)
export(.checkCacheRepo)
export(.file.move)
export(.formalsNotInCurrentDots)
export(.grepSysCalls)
export(.isCRSany)
export(.isGridded)
export(.isMemoised)
export(.isSF)
export(.isSpat)
export(.isSpatialAny)
export(.isVector)
export(.objSizeWithTry)
export(.orderDotsUnderscoreFirst)
export(.preDigestByClass)
export(.prefix)
export(.prepareOutput)
export(.requireNamespace)
export(.robustDigest)
export(.setSubAttrInList)
export(.sortDotsUnderscoreFirst)
export(.suffix)
export(.tagsByClass)
export(.unwrap)
export(.updateTagsRepo)
export(.whereInStack)
export(.wrap)
export(Cache)
export(CacheDBFile)
export(CacheDBTableName)
export(CacheDigest)
export(CacheGeo)
export(CacheIsACache)
export(CacheStorageDir)
export(CacheStoredFile)
export(CacheV2)
export(Checksums)
export(Copy)
export(Filenames)
export(asPath)
export(assessDataType)
export(basename2)
export(cacheId)
export(cc)
export(checkAndMakeCloudFolderID)
export(checkPath)
export(clearCache)
export(compareNA)
export(convertPaths)
export(convertRasterPaths)
export(copyFile)
export(createCache)
export(cropInputs)
export(cropTo)
export(dataType2)
export(detectActiveCores)
export(downloadFile)
export(extractFromCache)
export(fixErrors)
export(fixErrorsIn)
export(gdalMask)
export(gdalProject)
export(gdalResample)
export(getRelative)
export(internetExists)
export(isUpdated)
export(linkOrCopy)
export(listNamed)
export(loadFile)
export(loadFromCache)
export(makeMemoisable)
export(makeRelative)
export(maskInputs)
export(maskTo)
export(maxFn)
export(messageCache)
export(messageColoured)
export(messageDF)
export(messagePreProcess)
export(messagePrepInputs)
export(messageQuestion)
export(minFn)
export(movedCache)
export(nlayers2)
export(normPath)
export(normPathRel)
export(numCoresToUse)
export(objSize)
export(objSizeSession)
export(paddedFloatToChar)
export(postProcess)
export(postProcessTerra)
export(postProcessTo)
export(preProcess)
export(preProcessParams)
export(prepInputs)
export(prepInputsCOG)
export(prepInputsWithTiles)
export(prepopulateCacheAsync)
export(projectInputs)
export(projectTo)
export(purgeChecksums)
export(rasterRead)
export(remapFilenames)
export(reproducibleOptions)
export(retry)
export(rmFromCache)
export(searchFull)
export(searchFullEx)
export(set.randomseed)
export(showCache)
export(studyAreaName)
export(tempdir2)
export(tempfile2)
export(unmakeMemoisable)
export(updateFilenameSlots)
export(urlExists)
export(usesPointer)
export(values2)
export(writeFuture)
export(writeOutputs)
export(writeTo)
exportClasses(Path)
exportMethods(.addChangedAttr)
exportMethods(.addTagsToOutput)
exportMethods(.cacheMessage)
exportMethods(.checkCacheRepo)
exportMethods(.preDigestByClass)
exportMethods(.prepareOutput)
exportMethods(.robustDigest)
exportMethods(.tagsByClass)
exportMethods(Filenames)
exportMethods(checkPath)
exportMethods(clearCache)
exportMethods(keepCache)
exportMethods(mergeCache)
exportMethods(normPath)
exportMethods(showCache)
exportMethods(studyAreaName)
import(methods)
importFrom(data.table,":=")
importFrom(data.table,.N)
importFrom(data.table,.SD)
importFrom(data.table,as.data.table)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,fread)
importFrom(data.table,is.data.table)
importFrom(data.table,rbindlist)
importFrom(data.table,set)
importFrom(data.table,setDF)
importFrom(data.table,setDT)
importFrom(data.table,setcolorder)
importFrom(data.table,setindex)
importFrom(data.table,setkeyv)
importFrom(data.table,setnames)
importFrom(data.table,setorderv)
importFrom(digest,digest)
importFrom(fpCompare,"%==%")
importFrom(fs,is_absolute_path)
importFrom(fs,path_abs)
importFrom(fs,path_expand)
importFrom(fs,path_join)
importFrom(fs,path_norm)
importFrom(lobstr,obj_size)
importFrom(methods,"slot<-")
importFrom(methods,.S4methods)
importFrom(methods,formalArgs)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,selectMethod)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,slot)
importFrom(methods,slotNames)
importFrom(stats,na.omit)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(utils,capture.output)
importFrom(utils,download.file)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,isS3stdGeneric)
importFrom(utils,methods)
importFrom(utils,modifyList)
importFrom(utils,object.size)
importFrom(utils,packageVersion)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,untar)
importFrom(utils,unzip)
importFrom(utils,write.table)

# reproducible

<!-- badges: start -->
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/reproducible)](https://cran.r-project.org/package=reproducible)
[![CRAN Downloads](https://cranlogs.r-pkg.org/badges/grand-total/reproducible)](https://cran.r-project.org/package=reproducible)
[![R build status](https://github.com/PredictiveEcology/reproducible/workflows/R-CMD-check/badge.svg)](https://github.com/PredictiveEcology/reproducible/actions)
[![Codecov test coverage](https://codecov.io/gh/PredictiveEcology/reproducible/branch/master/graph/badge.svg)](https://app.codecov.io/gh/PredictiveEcology/reproducible?branch=master)
<!-- badges: end -->

A set of tools for R that enhance reproducibility for data analytics and forecasting.
This package aims at making high-level, robust, machine and OS independent tools for making deeply reproducible and reusable content in R.
The suggested package `geodata` can be installed from the repository 
(<https://PredictiveEcology.r-universe.dev>).

## News

See updates from latest [CRAN](https://cran.r-project.org/package=reproducible) and [development](https://github.com/PredictiveEcology/reproducible/blob/development/NEWS.md) versions.
Note that versions 2.0.0 and later are not compatible with previous versions.
The current version can be much faster and creates smaller repository files (each with specific options set using `Suggests` packages) and allows for different (e.g., `RPostgres` backends for the database[^1] -- not the saved files, however; these are still saved locally).

[^1]: <https://github.com/PredictiveEcology/SpaDES/wiki/Using-alternate-database-backends-for-Cache>

# Reproducible workflows

A reproducible workflow is a series of code steps (e.g., in a script) that, when run, produce the same output from the same inputs every time.
The big challenge with such a workflow is that many steps are so time consuming that a scientist tends to _not_ re-run each step every time. After many months of work, it is often unclear if the code will actually function from the start.
Is the original dataset still there? Have the packages that were used been updated? Are some of the steps missing because there was some "point and clicking"? 

The best way to maintain reproducibility is to have all the code re-run _all the time_.
That way, errors are detected early and can be fixed.
The challenge is how to make all the steps fast enough that it becomes convenient to re-run everything from scratch each time.

## `Cache`

Caching is the principle tool to achieve this reproducible work-flow.
There are many existing tools that support some notion of _caching_.
The main tool here, `Cache`, can be nested hierarchically, becoming very powerful for the data science developer who is regularly working at many levels of an analysis.

```
rnorm(1) # give a random number
Cache(rnorm, 1) # generates a random number
Cache(rnorm, 1) # recovers the previous random number because call is identical
```

## `prepInputs`

A common data problem is starting from a raw (spatial) dataset and getting it into shape for an analysis.
Often, copies of a dataset are haphazardly placed in *ad hoc* local file systems.
This makes it particularly difficult to share the workflow. The solution to this is use a canonical location (e.g., cloud storage, permalink to original data provider, etc.) and use tools that are smart enough to download only once.

Get a geospatial dataset. It will be checksummed (locally), meaning if the file is already in place locally, it will not download it again. 

```
# Using dlFun -- a custom download function -- passed to preProcess
test1 <- prepInputs(targetFile = "GADM_2.8_LUX_adm0.rds", # must specify currently
                    dlFun = "raster::getData", name = "GADM", country = "LUX", level = 0,
                    path = dPath)
```

## `Cache` with `prepInputs`

Putting these tools together allows for very rich data flows.
For example, with `prepInputs` and using the `fun` argument or passing a `studyArea`, a raw dataset can be downloaded, loaded into R, and post processed -- all potentially very time consuming steps resulting in a clean, often much smaller dataset.
Wrapping all these with a `Cache` can make it very quick.

```
test1 <- Cache(prepInputs, targetFile = "GADM_2.8_LUX_adm0.rds", # must specify currently
                    dlFun = "raster::getData", name = "GADM", country = "LUX", level = 0,
                    path = dPath)
```                    

See vignettes and help files for many more real-world examples.

## Installation

### Current release (on CRAN)

[![R build status](https://github.com/PredictiveEcology/reproducible/workflows/R-CMD-check/badge.svg?branch=master)](https://github.com/PredictiveEcology/reproducible/actions)
[![Codecov test coverage](https://codecov.io/gh/PredictiveEcology/reproducible/branch/master/graph/badge.svg)](https://app.codecov.io/gh/PredictiveEcology/reproducible?branch=master)

**Install from CRAN:**

```r
install.packages("reproducible")
```

**Install from GitHub:**
    
```r
#install.packages("devtools")
library("devtools")
install_github("PredictiveEcology/reproducible", dependencies = TRUE) 
```

### Development version

[![R build status](https://github.com/PredictiveEcology/reproducible/workflows/R-CMD-check/badge.svg?branch=development)](https://github.com/PredictiveEcology/reproducible/actions)
[![Codecov test coverage](https://codecov.io/gh/PredictiveEcology/reproducible/branch/development/graph/badge.svg)](https://app.codecov.io/gh/PredictiveEcology/reproducible?branch=development)

**Install from GitHub:**

```r
#install.packages("devtools")
library("devtools")
install_github("PredictiveEcology/reproducible", ref = "development", dependencies = TRUE) 
```

## Contributions

Please see `CONTRIBUTING.md` for information on how to contribute to this project.

data.table::setDTthreads(2)
tmpDir <- tempdir()
opts <- options(reproducible.cachePath = tmpDir)

# Usage -- All below are equivalent; even where args are missing or provided,
#   Cache evaluates using default values, if these are specified in formals(FUN)
a <- list()
b <- list(fun = rnorm)
bbb <- 1
ee <- new.env(parent = emptyenv())
ee$qq <- bbb

a[[1]] <- Cache(rnorm(1)) # no evaluation prior to Cache
a[[2]] <- Cache(rnorm, 1) # no evaluation prior to Cache
a[[3]] <- Cache(do.call, rnorm, list(1))
a[[4]] <- Cache(do.call(rnorm, list(1)))
a[[5]] <- Cache(do.call(b$fun, list(1)))
a[[6]] <- Cache(do.call, b$fun, list(1))
a[[7]] <- Cache(b$fun, 1)
a[[8]] <- Cache(b$fun(1))
a[[10]] <- Cache(quote(rnorm(1)))
a[[11]] <- Cache(stats::rnorm(1))
a[[12]] <- Cache(stats::rnorm, 1)
a[[13]] <- Cache(rnorm(1, 0, get("bbb", inherits = FALSE)))
a[[14]] <- Cache(rnorm(1, 0, get("qq", inherits = FALSE, envir = ee)))
a[[15]] <- Cache(rnorm(1, bbb - bbb, get("bbb", inherits = FALSE)))
a[[16]] <- Cache(rnorm(sd = 1, 0, n = get("bbb", inherits = FALSE))) # change order
a[[17]] <- Cache(rnorm(1, sd = get("ee", inherits = FALSE)$qq), mean = 0)

# with base pipe -- this is put in quotes ('') because R version 4.0 can't understand this
#  if you are using R >= 4.1 or R >= 4.2 if using the _ placeholder,
#  then you can just use pipe normally
usingPipe1 <- "b$fun(1) |> Cache()"  # base pipe

# For long pipe, need to wrap sequence in { }, or else only last step is cached
usingPipe2 <-
  '{"bbb" |>
      parse(text = _) |>
      eval() |>
      rnorm()} |>
    Cache()'
a[[9]] <- eval(parse(text = usingPipe1)) # recovers cached copy
a[[18]] <- eval(parse(text = usingPipe2)) # recovers cached copy

length(unique(a)) == 1 #  all same

### Pipe -- have to use { } or else only final function is Cached
b1a <- 'sample(1e5, 1) |> rnorm() |> Cache()'
b1b <- 'sample(1e5, 1) |> rnorm() |> Cache()'
b2a <- '{sample(1e5, 1) |> rnorm()} |> Cache()'
b2b <- '{sample(1e5, 1) |> rnorm()} |> Cache()'
b1a <- eval(parse(text = b1a))
b1b <- eval(parse(text = b1b))
b2a <- eval(parse(text = b2a))
b2b <- eval(parse(text = b2b))
all.equal(b1a, b1b) # Not TRUE because the sample is run first
all.equal(b2a, b2b) # TRUE because of {  }, sample is not run

#########################
# Advanced examples
#########################

# .cacheExtra -- add something to digest
Cache(rnorm(1), .cacheExtra = "sfessee11") # adds something other than fn args
Cache(rnorm(1), .cacheExtra = "nothing") # even though fn is same, the extra is different

# omitArgs -- remove something from digest (kind of the opposite of .cacheExtra)
Cache(rnorm(2, sd = 1), omitArgs = "sd") # removes one or more args from cache digest
Cache(rnorm(2, sd = 2), omitArgs = "sd") # b/c sd is not used, this is same as previous

# cacheId -- force the use of a digest -- can give undesired consequences
Cache(rnorm(3), cacheId = "k323431232") # sets the cacheId for this call
Cache(runif(14), cacheId = "k323431232") # recovers same as above, i.e, rnorm(3)

# Turn off Caching session-wide
opts <- options(reproducible.useCache = FALSE)
Cache(rnorm(3)) # doesn't cache
options(opts)

# showSimilar can help with debugging why a Cache call isn't picking up a cached copy
Cache(rnorm(4), showSimilar = TRUE) # shows that the argument `n` is different

###############################################
# devMode -- enables cache database to stay
#            small even when developing code
###############################################
opt <- options("reproducible.useCache" = "devMode")
clearCache(tmpDir, ask = FALSE)
centralTendency <- function(x) {
  mean(x)
}
funnyData <- c(1, 1, 1, 1, 10)
uniqueUserTags <- c("thisIsUnique", "reallyUnique")
ranNumsB <- Cache(centralTendency, funnyData, cachePath = tmpDir,
                  userTags = uniqueUserTags) # sets new value to Cache
showCache(tmpDir) # 1 unique cacheId -- cacheId is 71cd24ec3b0d0cac

# During development, we often redefine function internals
centralTendency <- function(x) {
  median(x)
}
# When we rerun, we don't want to keep the "old" cache because the function will
#   never again be defined that way. Here, because of userTags being the same,
#   it will replace the entry in the Cache, effetively overwriting it, even though
#   it has a different cacheId
ranNumsD <- Cache(centralTendency, funnyData, cachePath = tmpDir, userTags = uniqueUserTags)
showCache(tmpDir) # 1 unique artifact -- cacheId is 632cd06f30e111be

# If it finds it by cacheID, doesn't matter what the userTags are
ranNumsD <- Cache(centralTendency, funnyData, cachePath = tmpDir, userTags = "thisIsUnique")
options(opt)

#########################################
# For more in depth uses, see vignette
if (interactive())
  browseVignettes(package = "reproducible")

## ----setup, include = FALSE---------------------------------------------------
SuggestedPkgsNeeded <- c("terra")
hasSuggests <- all(sapply(SuggestedPkgsNeeded, require, character.only = TRUE, quietly = TRUE))
useSuggests <- !(tolower(Sys.getenv("_R_CHECK_DEPENDS_ONLY_")) == "true")

knitr::opts_chunk$set(eval = hasSuggests && useSuggests)

## ----function-level, echo=TRUE------------------------------------------------
library(reproducible)
library(data.table)

tmpDir <- file.path(tempfile(), "reproducible_examples", "Cache")
dir.create(tmpDir, recursive = TRUE)

# Source raster with a complete LCC definition
ras <- terra::rast(terra::ext(0, 300, 0, 300), vals = 1:9e4, res = 1)
terra::crs(ras) <- "+proj=lcc +lat_1=60 +lat_2=70 +lat_0=50 +lon_0=-100 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs"

# Target CRS in PROJ form (no EPSG lookup)
newCRS <- "+proj=longlat +datum=WGS84 +no_defs"

# Derive target extent from source extent (no registry lookup)
target_ext <- terra::project(terra::ext(ras), from = terra::crs(ras), to = newCRS)

# Build template with chosen resolution; assign CRS
tmplate <- terra::rast(target_ext, resolution = 0.00001)
terra::crs(tmplate) <- newCRS

# No Cache
system.time(map1 <- terra::project(ras, tmplate, method = "near"))

# Try with memoise for this example -- for many simple cases, memoising will not be faster
opts <- options("reproducible.useMemoise" = TRUE)
# With Cache -- a little slower the first time because saving to disk
system.time({
  suppressWarnings({
    map1 <- terra::project(ras, tmplate, method = "near") |> 
      Cache(cachePath = tmpDir)
  })
})

# faster the second time; improvement depends on size of object and time to run function
system.time({
  map2 <- terra::project(ras, tmplate, method = "near") |> 
    Cache(cachePath = tmpDir)
})

options(opts)

all.equal(map1, map2, check.attributes = FALSE) # TRUE

## -----------------------------------------------------------------------------
try(clearCache(tmpDir, ask = FALSE), silent = TRUE) # just to make sure it is clear

ranNumsA <- rnorm(10, 16) |> Cache(cachePath = tmpDir)

# All same
ranNumsB <- rnorm(10, 16) |> Cache(cachePath = tmpDir) # recovers cached copy
ranNumsD1 <- Cache(quote(rnorm(n = 10, 16))) |> Cache(cachePath = tmpDir) # recovers cached copy
ranNumsD2 <- Cache(rnorm(n = 10, 16)) |> Cache(cachePath = tmpDir) # recovers cached copy
# pipe
ranNumsD3 <- rnorm(n = 10, 16) |> Cache(cachePath = tmpDir) # recovers cached copy

# Any minor change makes it different
ranNumsE <- rnorm(10, 6) |> Cache(cachePath = tmpDir) # different

## ----tags---------------------------------------------------------------------
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- runif(4) |> Cache(cachePath = tmpDir, userTags = "objectName:b")

showCache(tmpDir, userTags = c("objectName"))
showCache(tmpDir, userTags = c("^a$")) # regular expression ... "a" exactly
showCache(tmpDir, userTags = c("runif")) # show only cached objects made during runif call

clearCache(tmpDir, userTags = c("runif"), ask = FALSE) # remove only cached objects made during runif call
showCache(tmpDir) # all

clearCache(tmpDir, ask = FALSE)

## ----accessed-tag-------------------------------------------------------------
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- runif(4) |> Cache(cachePath = tmpDir, userTags = "objectName:b")

# access it again, from Cache
Sys.sleep(1)
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
wholeCache <- showCache(tmpDir)

# keep only items accessed "recently" (i.e., only objectName:a)
onlyRecentlyAccessed <- showCache(tmpDir, userTags = max(wholeCache[tagKey == "accessed"]$tagValue))

# inverse join with 2 data.tables ... using: a[!b]
# i.e., return all of wholeCache that was not recently accessed
#   Note: the two different ways to access -- old way with "artifact" will be deprecated
toRemove <- unique(wholeCache[!onlyRecentlyAccessed, on = "cacheId"], by = "cacheId")$cacheId
clearCache(tmpDir, toRemove, ask = FALSE) # remove ones not recently accessed
showCache(tmpDir) # still has more recently accessed

## ----keepCache----------------------------------------------------------------
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- Cache(runif(4)) |> Cache(cachePath = tmpDir, userTags = "objectName:b")

# keep only those cached items from the last 24 hours
oneDay <- 60 * 60 * 24
keepCache(tmpDir, after = Sys.time() - oneDay, ask = FALSE)

# Keep all Cache items created with an rnorm() call
keepCache(tmpDir, userTags = "rnorm", ask = FALSE)
showCache(tmpDir)

# Remove all Cache items that happened within a rnorm() call
clearCache(tmpDir, userTags = "rnorm", ask = FALSE)
showCache(tmpDir) ## empty

# Also, can set a time before caching happens and remove based on this
#  --> a useful, simple way to control Cache
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
startTime <- Sys.time()
Sys.sleep(1)
ranNumsB <- rnorm(5) |> Cache(cachePath = tmpDir, userTags = "objectName:b")
keepCache(tmpDir, after = startTime, ask = FALSE) # keep only those newer than startTime

clearCache(tmpDir, ask = FALSE)

## ----searching-within-cache---------------------------------------------------
# default userTags is "and" matching; for "or" matching use |
ranNumsA <- runif(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:b")

# show all objects (runif and rnorm in this case)
showCache(tmpDir)

# show objects that are both runif and rnorm
# (i.e., none in this case, because objecs are either or, not both)
showCache(tmpDir, userTags = c("runif", "rnorm")) ## empty

# show objects that are either runif or rnorm ("or" search)
showCache(tmpDir, userTags = "runif|rnorm")

# keep only objects that are either runif or rnorm ("or" search)
keepCache(tmpDir, userTags = "runif|rnorm", ask = FALSE)

clearCache(tmpDir, ask = FALSE)

## ----expensive-computations---------------------------------------------------
ras <- terra::rast(terra::ext(0, 5, 0, 5),
  res = 1,
  vals = sample(1:5, replace = TRUE, size = 25),
  crs = "+proj=lcc +lat_1=48 +lat_2=33 +lon_0=-100 +ellps=WGS84"
)

rasCRS <- terra::crs(ras)
# Build an explicit target raster (same CRS, coarser resolution).
# Avoids the `terra::project(x, char_crs, res = N)` shorthand which
# recurses internally and forwards an unrecognized `wopt` set on
# recent terra versions, producing
# `[write] unknown option(s): xscale,yscale`.
rasTarget <- terra::rast(terra::ext(ras), crs = rasCRS, resolution = 5)

# A slow operation, like GIS operation
notCached <- suppressWarnings(
  # project raster generates warnings when run non-interactively
  terra::project(ras, rasTarget)
)

cached <- suppressWarnings(
  # project raster generates warnings when run non-interactively
  # using quote works also
  terra::project(ras, rasTarget) |> Cache(cachePath = tmpDir)
)

# second time is much faster
reRun <- suppressWarnings(
  # project raster generates warnings when run non-interactively
  terra::project(ras, rasTarget) |> Cache(cachePath = tmpDir)
)

# recovered cached version is same as non-cached version
all.equal(notCached, reRun, check.attributes = FALSE) ## TRUE

## ----nested-------------------------------------------------------------------
##########################
## Nested Caching
# Make 2 functions
inner <- function(mean) {
  d <- 1
  rnorm(n = 3, mean = mean)
}
outer <- function(n) {
  inner(0.1) |> Cache(cachePath = tmpdir2)
}

# make 2 different cache paths
tmpdir1 <- file.path(tempfile(), "first")
tmpdir2 <- file.path(tempfile(), "second")

# Run the Cache ... notOlderThan propagates to all 3 Cache calls,
#   but cachePath is tmpdir1 in top level Cache and all nested
#   Cache calls, unless individually overridden ... here inner
#   uses tmpdir2 repository
outer(n = 2) |> Cache(cachePath = tmpdir1)

showCache(tmpdir1) # 2 function calls
showCache(tmpdir2) # 1 function call

# userTags get appended
# all items have the outer tag propagate, plus inner ones only have inner ones
clearCache(tmpdir1, ask = FALSE)
outerTag <- "outerTag"
innerTag <- "innerTag"
inner <- function(mean) {
  d <- 1
  rnorm(n = 3, mean = mean) |> Cache(notOlderThan = Sys.time() - 1e5, userTags = innerTag)
}
outer <- function(n) {
  inner(0.1) |> Cache()
}
aa <- Cache(outer, n = 2) |> Cache(cachePath = tmpdir1, userTags = outerTag)
showCache(tmpdir1) # rnorm function has outerTag and innerTag, inner and outer only have outerTag

## ----selective-cacheId--------------------------------------------------------
### cacheId
set.seed(1)
rnorm(1) |> Cache(cachePath = tmpdir1)
# manually look at output attribute which shows cacheId: 422bae4ed2f770cc
rnorm(1) |> Cache(cachePath = tmpdir1, cacheId = "422bae4ed2f770cc") # same value
# override even with different inputs:
rnorm(2) |> Cache(cachePath = tmpdir1, cacheId = "422bae4ed2f770cc")

## ----manual-cache-------------------------------------------------------------
# As of reproducible version 1.0, there is a new backend directly using DBI
mapHash <- unique(showCache(tmpDir, userTags = "project")$cacheId)
map <- loadFromCache(mapHash[1], cachePath = tmpDir)
terra::plot(map)

## ----cleanup------------------------------------------------------------------
## cleanup
unlink(dirname(tmpDir), recursive = TRUE)

if (requireNamespace("terra", quietly = TRUE)) {
  ## LOG1S
  rasOrig <- terra::rast(ncols = 10, nrows = 10)
  ras <- rasOrig
  ras[] <- rep(c(0,1),50)
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- rep(c(0,1),50)
  assessDataType(ras)

  ras[] <- rep(c(TRUE,FALSE),50)
  assessDataType(ras)

  ras[] <- c(NA, NA, rep(c(0,1),49))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- c(0, NaN, rep(c(0,1),49))
  assessDataType(ras)


  ## INT1S
  ras[] <- -1:98
  assessDataType(ras)

  ras[] <- c(NA, -1:97)
  assessDataType(ras)

  ## INT1U
  ras <- rasOrig
  ras[] <- 1:100
  assessDataType(ras)

  ras[] <- c(NA, 2:100)
  assessDataType(ras)

  ## INT2U
  ras <- rasOrig
  ras[] <- round(runif(100, min = 64000, max = 65000))
  assessDataType(ras)

  ## INT2S
  ras <- rasOrig
  ras[] <- round(runif(100, min = -32767, max = 32767))
  assessDataType(ras)

  ras[54] <- NA
  assessDataType(ras)

  ## INT4U
  ras <- rasOrig
  ras[] <- round(runif(100, min = 0, max = 500000000))
  assessDataType(ras)

  ras[14] <- NA
  assessDataType(ras)

  ## INT4S
  ras <- rasOrig
  ras[] <- round(runif(100, min = -200000000, max = 200000000))
  assessDataType(ras)

  ras[14] <- NA
  assessDataType(ras)

  ## FLT4S
  ras <- rasOrig
  ras[] <- runif(100, min = -10, max = 87)
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = -3.4e+26, max = 3.4e+28))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = 3.4e+26, max = 3.4e+28))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = -3.4e+26, max = -1))
  assessDataType(ras)

  ## FLT8S
  ras <- rasOrig
  ras[] <- c(-Inf, 1, rep(c(0,1),49))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- c(Inf, 1, rep(c(0,1),49))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = -1.7e+30, max = 1.7e+308))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = 1.7e+30, max = 1.7e+308))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = -1.7e+308, max = -1))
  assessDataType(ras)

  # 2 layer with different types LOG1S and FLT8S
  ras <- rasOrig
  ras[] <- rep(c(0,1),50)
  ras1 <- rasOrig
  ras1[] <- round(runif(100, min = -1.7e+308, max = -1))
  sta <- c(ras, ras1)
  assessDataType(sta)

}

## normalize file paths
paths <- list("./aaa/zzz",
              "./aaa/zzz/",
              ".//aaa//zzz",
              ".//aaa//zzz/",
              ".\\\\aaa\\\\zzz",
              ".\\\\aaa\\\\zzz\\\\",
              file.path(".", "aaa", "zzz"))

checked <- normPath(paths)
length(unique(checked)) ## 1; all of the above are equivalent

## check to see if a path exists
tmpdir <- file.path(tempdir(), "example_checkPath")

dir.exists(tmpdir) ## FALSE
tryCatch(checkPath(tmpdir, create = FALSE), error = function(e) FALSE) ## FALSE

checkPath(tmpdir, create = TRUE)
dir.exists(tmpdir) ## TRUE

unlink(tmpdir, recursive = TRUE)

if (require("terra", quietly = TRUE)) {
  # prepare dummy data -- 3 SpatRasters, 2 SpatVectors
  # need 2 SpatRaster
  rf <- system.file("ex/elev.tif", package = "terra")
  elev1 <- terra::rast(rf)

  # a polygon vector
  f <- system.file("ex/lux.shp", package = "terra")
  vOrig <- terra::vect(f)
  v <- vOrig[1:2, ]

  # utm <- terra::crs("epsg:23028") # $wkt
  utm <- "+proj=utm +zone=28 +datum=WGS84 +units=m +no_defs"
  vInUTM <- terra::project(vOrig, utm)
  vAsRasInLongLat <- terra::rast(vOrig, resolution = 0.008333333)
  res100 <- 100
  rInUTM <- terra::rast(vInUTM, resolution = res100, vals = 1)
  # crop, reproject, mask, crop a raster with a vector in a different projection
  #  --> gives message about not enough information
  t1 <- postProcessTo(elev1, to = vInUTM)
  # crop, reproject, mask a raster to a different projection, then mask
  t2a <- postProcessTo(elev1, to = vAsRasInLongLat, maskTo = vInUTM)
  t3a <- postProcessTo(elev1, to = rInUTM, maskTo = vInUTM)

}

if (requireNamespace("terra", quietly = TRUE) &&
    requireNamespace("withr", quietly = TRUE)) {
  library(reproducible)
  withr::local_dir(withr::local_tempdir())
  withr::local_options(reproducible.inputPaths = NULL)
  # od <- setwd(tempdir2())
  # download a (spatial) file from remote url (which often is an archive) load into R
  # need 3 files for this example; 1 from remote, 2 local
  dPath <- file.path(tempdir2())
  remoteTifUrl <- "https://github.com/rspatial/terra/raw/master/inst/ex/elev.tif"

  localFileLuxSm <- system.file("ex/luxSmall.shp", package = "reproducible")
  localFileLux <- system.file("ex/lux.shp", package = "terra")

  # 1 step for each layer
  # 1st step -- get study area
  studyArea <- prepInputs(localFileLuxSm, fun = "terra::vect") # default is sf::st_read

  \donttest{
    # 2nd step: make the input data layer like the studyArea map
    # Requires internet -- so using try just in case, and kept out of the
    # timed examples (CRAN does not run \donttest)
    elevForStudy <- try(prepInputs(url = remoteTifUrl, to = studyArea, res = 250,
                                   destinationPath = dPath, useCache = FALSE))

    # Alternate way, one step at a time. Must know each of these steps, and perform for each layer
    dir.create(dPath, recursive = TRUE, showWarnings = FALSE)
    file.copy(localFileLuxSm, file.path(dPath, basename(localFileLuxSm)))
    studyArea2 <- terra::vect(localFileLuxSm)
    if (!all(terra::is.valid(studyArea2))) studyArea2 <- terra::makeValid(studyArea2)
    tf <- tempfile(fileext = ".tif")
    download.file(url = remoteTifUrl, destfile = tf, mode = "wb", quiet = TRUE)
    Checksums(dPath, write = TRUE, files = tf)
    elevOrig <- terra::rast(tf)
    studyAreaCrs <- terra::crs(studyArea)
    # Build an explicit target raster (same CRS as studyArea, res = 250) and
    # project to it. Avoids the recursive `terra::project(x, char_crs, res = N)`
    # shorthand that recent terra (~1.9) regressed with
    # `[write] unknown option(s): xscale,yscale`.
    elevTarget <- terra::project(terra::rast(elevOrig), studyAreaCrs)
    elevTarget <- terra::rast(terra::ext(elevTarget),
                              crs = terra::crs(elevTarget),
                              resolution = 250)
    elevForStudy2 <- terra::project(elevOrig, elevTarget) |>
      terra::mask(studyArea2) |>
      terra::crop(studyArea2)

    isTRUE(all.equal(elevForStudy, elevForStudy2)) # TRUE!
  }

  # sf class
  if (requireNamespace("sf", quietly = TRUE)) {
    studyAreaSmall <- prepInputs(localFileLuxSm, fun = "sf::st_read")
    studyAreas <- list()
    studyAreas[["orig"]] <- prepInputs(localFileLux)
    studyAreas[["reprojected"]] <- projectTo(studyAreas[["orig"]], studyAreaSmall)
    studyAreas[["cropped"]] <- suppressWarnings(cropTo(studyAreas[["orig"]], studyAreaSmall))
    studyAreas[["masked"]] <- suppressWarnings(maskTo(studyAreas[["orig"]], studyAreaSmall))
  }

  # SpatVector-- note: doesn't matter what class the "to" object is, only the "from"
  studyAreas <- list()
  studyAreaSmall <- prepInputs(localFileLuxSm)
  studyAreas[["orig"]] <- prepInputs(localFileLux)
  studyAreas[["reprojected"]] <- projectTo(studyAreas[["orig"]], studyAreaSmall)
  studyAreas[["cropped"]] <- suppressWarnings(cropTo(studyAreas[["orig"]], studyAreaSmall))
  studyAreas[["masked"]] <- suppressWarnings(maskTo(studyAreas[["orig"]], studyAreaSmall))
  if (interactive()) {
    par(mfrow = c(2,2));
    out <- lapply(studyAreas, function(x) terra::plot(x))
  }

  withr::deferred_run()
  # setwd(od)
}

#' Transform a GIS dataset so it has the properties (extent, projection, mask) of another
#'
#' This function provides a single step to achieve the GIS operations
#' "pre-crop-with-buffer-to-speed-up-projection", "project",
#' "post-projection-crop", "mask" and possibly "write".
#' It uses primarily the `terra` package internally
#' (with some minor functions from `sf`)
#' in an attempt to be as efficient as possible, except if all inputs are `sf` objects.
#' (in which case `sf` is used). Currently, this function is tested
#' with `sf`, `SpatVector`, `SpatRaster`, `Raster*` and `Spatial*` objects passed
#' to `from`, and the same plus `SpatExtent`, and `crs` passed to `to` or the
#' relevant `*to` functions.
#' For this function, Gridded means a `Raster*` class object from `raster` or
#' a `SpatRaster` class object from `terra`.
#' Vector means a `Spatial*` class object from `sp`, a `sf` class object
#' from `sf`, or a `SpatVector` class object from `terra`.
#' This function is also used internally with the deprecated family [postProcess()],
#' `*Inputs`, such as [cropInputs()].
#'
#' @details
#' `postProcessTo` is a wrapper around (an initial "wide" crop for speed)
#' `cropTo(needBuffer = TRUE)`, `projectTo`,
#' `cropTo` (the actual crop for precision), `maskTo`, `writeTo`.
#'  Users can call each of these individually.
#'
#' `postProcessTerra` is the early name of this function that is now `postProcessTo`.
#'
#' @section Use Cases:
#'
#' The table below shows what will result from passing different classes to `from`
#' and `to`:
#'
#' \tabular{lll}{
#'   **`from`**\tab **`to`**   \tab **`from`** will have:                             \cr
#'   `Gridded` \tab `Gridded`  \tab the extent, projection, origin, resolution
#'                                  and masking where there are `NA` from the `to`    \cr
#'   `Gridded` \tab `Vector`   \tab the projection, origin, and mask from `to`, and
#'                                  extent will be a round number of pixels that
#'                                  fit within the extent of `to`. Resolution will
#'                                  be the same as `from`.  See section
#'                                  below about `projectTo`. \cr
#'   `Vector` \tab `Vector`    \tab the projection, origin, extent and mask from `to` \cr
#' }
#'
#'
#' If one or more of the `*To` arguments are supplied, these will
#' override individual components of `to`. If `to` is omitted or `NULL`,
#' then only the `*To` arguments that are used will be performed. In all cases,
#' setting a `*To` argument to `NA` will prevent that step from happening.
#'
#' @section `projectTo`:
#' Since these functions use the gis capabilities of `sf` and `terra`, they will only
#' be able to do things that those functions can do. One key caution, which is
#' stated clearly in `?terra::project` is that projection of a raster (i.e., gridded)
#' object should always be with another gridded object. If the user chooses to
#' supply a `projectTo` that is a vector object for a `from` that is gridded,
#' there may be unexpected failures due e.g., to extents not overlapping during
#' the `maskTo` stage.
#'
#' @section Backwards compatibility with `postProcess`:
#'
#' \subsection{`rasterToMatch` and `studyArea`:}{
#'
#'   If these are supplied, `postProcessTo` will use them instead
#'   of `to`. If only `rasterToMatch` is supplied, it will be assigned to
#'   `to`. If only `studyArea` is supplied, it will be used for `cropTo`
#'   and `maskTo`; it will only be used for `projectTo` if `useSAcrs = TRUE`.
#'   If both `rasterToMatch` and `studyArea` are supplied,
#'   `studyArea` will only be applied to `maskTo` (unless `maskWithRTM = TRUE`),
#'    and, optionally, to `projectTo` (if `useSAcrs = TRUE`); everything else
#'    will be from `rasterToMatch`.
#'
#' }
#'
#' \subsection{`targetCRS`, `filename2`, `useSAcrs`, `maskWithRTM`:}{
#'
#'   `targetCRS` if supplied will be assigned to `projectTo`. `filename2` will
#'   be assigned to `writeTo`. If `useSAcrs` is set, then the `studyArea`
#'   will be assigned to `projectTo`. If `maskWithRTM` is used, then the
#'   `rasterToMath` will be assigned to `maskTo`. All of these will override
#'   any existing values for these arguments.
#'
#' }
#'
#'  See also [postProcess()] documentation section on
#'  *Backwards compatibility with `rasterToMatch` and/or `studyArea`* for further
#'  detail.
#'
#' @section Cropping:
#' If `cropTo` is not `NA`, `postProcessTo` does cropping twice, both the first and last steps.
#' It does it first for speed, as cropping is a very fast algorithm. This will quickly remove
#' a bunch of pixels that are not necessary. But, to not create bias, this first crop is padded
#' by  `2 * res(from)[1]`), so that edge cells still have a complete set of neighbours.
#' The second crop is at the end, after projecting and masking. After the projection step,
#' the crop is no longer tight. Under some conditions, masking will effectively mask and crop in
#' one step, but under some conditions, this is not true, and the mask leaves padded NAs out to
#' the extent of the `from` (as it is after crop, project, mask). Thus the second
#' crop removes all NA cells so they are tight to the mask.
#'
#' @return
#' An object of the same class as `from`, but potentially cropped (via [cropTo()]),
#' projected (via [projectTo()]), masked (via [maskTo()]), and written to disk
#' (via [writeTo()]).
#'
#' @param from A Gridded or Vector dataset on which to do one or more of:
#'   crop, project, mask, and write
#' @param to A Gridded or Vector dataset which is the object
#'   whose metadata will be the target for cropping, projecting, and masking of `from`.
#' @param cropTo Optional Gridded or Vector dataset which,
#'   if supplied, will supply the extent with which to crop `from`. To omit
#'   cropping completely, set this to `NA`. If supplied, this will override `to`
#'   for the cropping step. Defaults to `NULL`, which means use `to`
#' @param projectTo Optional Gridded or Vector dataset, or `crs` object (e.g., sf::st_crs).
#'   If Gridded it will supply
#'   the `crs`, `extent`, `res`, and `origin`
#'   to project the `from` to. If Vector, it will provide the `crs` only.
#'   The resolution and extent will be taken from `res(from)` (i.e. `ncol(from)*nrow(from)`).
#'   If a Vector, the extent of the `projectTo` is not used (unless it is also passed to `cropTo`.
#'   To omit projecting, set this to `NA`.
#'   If supplied, this will override `to` for the projecting step.
#'   Defaults to `NULL`, which means use `to`.
#'   **Attention.** Conflicts may arise with when `projectTo` is a Vector/CRS object with a
#'   distinct CRS from `to`. Because `to` is used for masking *after* `from` is re-projected using
#'   `projectTo`, the extents of `to` and `from` may no longer overlap (as in *align*)
#'   perfectly leading to failure during
#'   the masking step. We  recommend passing a raster templates to `projectTo` whose extent and CRS
#'   are both compatible with the object used later for masking (either `to` or `maskTo`).
#' @param maskTo Optional Gridded or Vector dataset which,
#'   if supplied, will supply the extent with which to mask `from`.
#'   If Gridded, it will mask with the `NA` values on the `maskTo`;
#'   if Vector, it will mask on the `terra::aggregate(maskTo)`.
#'   To omit masking completely, set this to `NA`.
#'   If supplied, this will override `to` for the masking step.
#'   Defaults to `NULL`, which means use `to`
#' @param writeTo Optional character string of a filename to use `writeRaster` to save the final
#'   object. Default is `NULL`, which means there is no `writeRaster`
#' @param overwrite Logical. Used if `writeTo` is not `NULL`; also if `terra` determines
#'   that the object requires writing to disk during a `crop`, `mask` or `project` call
#'   e.g., because it is too large.
#' @param ... Arguments passed to `terra::mask` (for `maskTo`), `terra::project` (for `projectTo`)
#'   or `terra::writeRaster` (for `writeTo`) and not used for `cropTo`, as well `postProcess`'s
#'   `rasterToMatch` and `studyArea` arguments (see below). Commonly used arguments might be
#'   `method`, `touches`, and `datatype`. If `filename` is passed, it will be ignored; use
#'   `writeTo = `. If `reproducible.gdalwarp = TRUE`, then these will be passed to the
#'   `gdal*` functions. See them for details.
#' @inheritParams Cache
#'
#' @details
#'
#' This function is meant to replace [postProcess()] with the more efficient
#' and faster `terra` functions.
#'
#' @export
#' @seealso [maskTo()], [cropTo()], [projectTo()], [writeTo()], and [fixErrorsIn()].
#' Also the functions that
#' call `sf::gdal_utils(...)` directly: [gdalProject()], [gdalResample()], [gdalMask()]
#' @rdname postProcessTo
#' @example inst/examples/example_postProcessTo.R
#'
postProcessTo <- function(from, to,
                          cropTo = NULL, projectTo = NULL, maskTo = NULL, writeTo = NULL,
                          overwrite = TRUE, verbose = getOption("reproducible.verbose"),
                          ...) {
  st <- Sys.time()
  remapOldArgs(...) # converts studyArea, rasterToMatch, filename2, useSAcrs, targetCRS


  # Deal with combinations of to and *To
  if (missing(to)) to <- NULL
  if (is.null(to)) {
    if (missing(cropTo)) cropTo <- NULL
    if (missing(maskTo)) maskTo <- NULL
    if (missing(projectTo)) projectTo <- NULL
  } else {
    if (isRaster(to)) {
      .requireNamespace("terra", stopOnFALSE = TRUE)
      to <- terra::rast(to)
    }
    # case where all *To are NULL --> use to as in the arg defaults
    if (is.null(cropTo)) cropTo <- to
    if (is.null(maskTo)) maskTo <- to
    if (is.null(projectTo)) projectTo <- to
  }

  needPostProcessing <- !all(is.null(to), is.null(cropTo), is.null(maskTo),
                             is.null(projectTo), is.null(writeTo))
  if (needPostProcessing) {
    if (.isGridded(from)) {
      if (getOption("reproducible.leaveOnDisk", TRUE)) {
        co <- capture.output(origMemFrac <- terra::terraOptions()$memfrac)
        if (identical(origMemFrac, 0.5)) { # 0.5 is the default in `terra` on Dec 15, 2025
          terra::terraOptions(memfrac = 0)
          on.exit(terra::terraOptions(memfrac = origMemFrac))
        }
      }
    }

    messagePreProcess("Running `postProcessTo`", verbose = verbose, verboseLevel = 0)
    .message$IndentUpdate()
    if (isTRUE(is.character(from))) {
      fe <- fileExt(from)
      if (fe %in% "shp") {
        shpFilename <- gdalTransform(from, cropTo, projectTo, maskTo, writeTo)
        return(shpFilename)
      }
    }
    fromOrig <- from # may need it later
    # ASSERTION STEP
    postProcessToAssertions(from, to, cropTo, maskTo, projectTo)

    # Get the original class of from so that it can be recovered
    origFromClass <- is(from)
    isRasterHere <- any(origFromClass == "Raster")
    isRasterLayerHere <- any(origFromClass == "RasterLayer")
    isStackHere <- any(origFromClass == "RasterStack")
    isBrickHere <- any(origFromClass == "RasterBrick")
    # isSFHere <- any(origFromClass == "sf")
    isSpatialHere <- any(startsWith(origFromClass, "Spatial"))
    isSpatRasterHere <- any(origFromClass == "SpatRaster")
    isVectorNonTerra <- .isVector(from) && !.isSpat(from)

    # converting sf to terra then cropping is slower than cropping then converting to terra
    #    so if both are vector datasets, and sf format, crop first
    from <- cropSF(from, cropTo)

    if (isRasterHere) {
      fromCRS <- terra::crs(from)
      from <- terra::rast(from)
      if (!nzchar(terra::crs(from))) {
        terra::crs(from) <- fromCRS
      } # $input
    } else if (isSpatialHere) {
      osFrom <- object.size(from)
      lg <- osFrom > 5e8
      if (lg) {
        st <- Sys.time()
        messagePreProcess("`from` is large, converting to terra object will take some time ...",
                          verbose = verbose
        )
      }
      from <- suppressWarningsSpecific(terra::vect(from), shldBeChar)
      if (lg) {
        messagePreProcess("done in ", format(difftime(Sys.time(), st),
                                             units = "secs", digits = 3
        ),
        verbose = verbose
        )
      }
    }

    couldDoGDAL <- .isGridded(from) && .isVector(maskTo) && .isGridded(projectTo)
    stillNeed <- TRUE

    if (isTRUE(getOption("reproducible.gdalwarp", FALSE)) && couldDoGDAL) {
      messagePreProcess("option(reproducible.gdalwarp = TRUE) is now defunct; passing to postProcessTo...")
      # stillNeed <- FALSE
      # #############################################################
      # # project resample mask sequence ################################
      # #############################################################
      # messagePreProcess("using sf::gdal_utils('warp') because options(\"reproducible.gdalwarp\" = TRUE) ...", appendLF = TRUE, verbose = verbose)
      # st <- Sys.time()
      #
      # withCallingHandlers(
      #   tryCatch({
      #     from <- gdalProject(fromRas = from, toRas = projectTo, verbose = verbose, ...)
      #     from <- gdalResample(fromRas = from, toRas = projectTo, verbose = verbose, ...)
      #     if (.isGridded(maskTo)) { # won't be used at the moment because couldDoGDAL = FALSE for gridded
      #       from <- maskTo(from = from, maskTo = maskTo, verbose = verbose, ...)
      #     } else {
      #       from <- gdalMask(fromRas = from, maskToVect = maskTo, writeTo = writeTo, verbose = verbose, ...)
      #     }
      #   }, error = function(e) {
      #     stillNeed <<- TRUE
      #     couldDoGDAL <<- FALSE
      #     message("Attempted to use gdal* functions, but errors occured; trying without gdal*...")
      #   }),
      #   warning = function(w) {
      #     if (any(grepl("transformer options does not support option NUM_THREADS", w$message)))
      #       invokeRestart("muffleWarning")
      #   })
      #   # from <- setMinMax(from)

    } # else {

    if (stillNeed) {
      if (couldDoGDAL)
        message("Try setting options('reproducible.gdalwarp' = TRUE) to use a different, possibly faster, algorithm")
      #############################################################
      # crop project mask sequence ################################
      #############################################################
      # Basically, when both layers are vector, it appears to sometimes be lossy to do first
      #   cropTo --> i.e., projecting cropTo to from's crs, then crop, then proceed was making
      #   errors and slivers
      if (!(isPolygons(from) && isPolygons(projectTo) && identical(cropTo, projectTo)))
        from <- cropTo(from, cropTo, needBuffer = TRUE, verbose = verbose, ..., overwrite = overwrite) # crop first for speed
      from <- projectTo(from, projectTo, verbose = verbose, ..., overwrite = overwrite) # need to project with edges intact
      from <- maskTo(from, maskTo, verbose = verbose, ..., overwrite = overwrite)
      from <- cropTo(from, cropTo, needBuffer = FALSE, verbose = verbose, ..., overwrite = overwrite) # need to recrop to trim excess pixels in new projection

      # Put this message near the end so doesn't get lost
      if (is.naSpatial(cropTo) && .isVector(maskTo)) {
        messagePreProcess("** cropTo is NA, but maskTo is a Vector dataset; ",
                          verbose = verbose
        )
        messagePreProcess("this has the effect of cropping anyway",
                          verbose = verbose
        )
      }
      if (.isSF(from))
        from <- keepOrigGeom(from, fromOrig)

      # WRITE STEP
      from <- writeTo(
        from, writeTo, overwrite, isStackHere, isBrickHere, isRasterHere, isSpatRasterHere,
        ...
      )

    }


    # REVERT TO ORIGINAL INPUT CLASS
    from <- revertClass(from, origFromClass = origFromClass)
    .message$IndentRevert()
    messagePreProcess("postProcessTo ", gsub("^\b", "", messagePrefixDoneIn),
                      format(difftime(Sys.time(), st), units = "secs", digits = 3),
                      verbose = verbose)
    # messagePreProcess("postProcessTo done in ", format(difftime(Sys.time(), st),
    #                                                    units = "secs", digits = 3
    # ),
    # verbose = verbose
    # )
  }
  from
}

#' @export
#' @rdname postProcessTo
postProcessTerra <- postProcessTo

#' Some spatial helper functions
#'
#' @param x A spatial object.
#' @export
#' @rdname isSpatial
#' @aliases isSpatial
#' @details
#' `.isGridded` returns `TRUE` if the object is a `SpatRaster` or `Raster`
#'
#' `.isVector` returns `TRUE` if the object is `SpatVector`, `spatial` or `sf`
#'
#' `.isSF` returns `TRUE` if the object is `sf` or `sfc`
#'
#' `.isSpat` returns `TRUE` if the object is `SpatVector` or `SpatRaster`
#'
#' `.isSpatialAny` returns `TRUE` if the object returns `TRUE` for `.isGridded` or
#'   `.isVector`
#'
#'
#' @return Logical.
.isGridded <- function(x) .isSpatRaster(x) || inherits(x, "Raster")

#' @export
#' @rdname isSpatial
.isVector <- function(x) .isSpatVector(x) || inherits(x, "Spatial") || .isSF(x)

#' @export
#' @rdname isSpatial
.isSF <- function(x) inherits(x, "sf") || inherits(x, "sfc")

#' @export
#' @rdname isSpatial
.isSpat <- function(x) .isSpatRaster(x) || .isSpatVector(x)

#' @export
#' @rdname isSpatial
.isSpatialAny <- function(x) .isGridded(x) || .isVector(x)

#' @export
#' @rdname isSpatial
.isCRSany <- function(x) isCRSSF(x) || isCRScharacter(x) || isCRSTerra(x)

.isSpatVector <- function(x) inherits(x, "SpatVector")
.isSpatRaster <- function(x) inherits(x, "SpatRaster")

isSpatial <- function(x) inherits(x, "Spatial")
isSpat2 <- function(origClass) any(origClass %in% c("SpatVector", "SpatRaster"))
isRaster <- function(x) inherits(x, "Raster")
isCRSSF <- function(x) inherits(x, "crs")
isCRScharacter <- function(x) is.character(x) && (grepl("DATUM", x) || grepl("+proj", x) || grepl("epsg:", x))
isCRSTerra <- function(x) inherits(x, "CRS") || (is.character(x) && startsWith(x, "GEOGCRS"))

#' Fix common errors in GIS layers, using `terra`
#'
#' Currently, this only tests for validity of a SpatVect file, then if there is a problem,
#' it will run `terra::makeValid`
#' @export
#' @param x The SpatStat or SpatVect object to try to fix.
#' @param error The error message, e.g., coming from `try(...)`
#' @inheritParams Cache
#' @param fromFnName The function name that produced the error, e.g., `maskTo`
#'
#' @return
#' An object of the same class as `x`, but with some errors fixed via `terra::makeValid()`
#'
fixErrorsIn <- function(x, error = NULL, verbose = getOption("reproducible.verbose"), fromFnName = "") {
  if (.isVector(x)) {
    os <- 0
    if (!is.null(error)) {
      messageDeclareError(error, fromFnName, verbose)
      os <- objSize(x)
      if (os > 1e9) {
        messagePreProcess("... this may take a long time because the object is large (",
                          format(os), ")",
                          verbose = verbose
        )
      }
    }
    if (.isSF(x)) {
      .requireNamespace("sf", stopOnFALSE = TRUE)
      xValids <- sf::st_is_valid(x)
      if (any(!xValids)) {
        if (os > 1e9) {
          messagePreProcess("... found invalid components ... running sf::st_make_valid",
                            verbose = verbose
          )
        }

        x <- sf::st_make_valid(x)
      }
    } else {
      if (!.isSpat(x)) {
        origClass <- class(x)
        isSp <- isSpatial(x)
        x <- terra::vect(x)
      }
      if (os > 1e9 && isTRUE(getOption("reproducible.useCache"))) {
        messagePreProcess("... Caching the fixErrorTerra call on this large object", verbose = verbose)
        x <- Cache(makeVal(x), .functionName = "make.valid")
      } else {
        x <- makeVal(x)
      }
      if (exists("origClass", inherits = FALSE)) {
        x <- revertClass(x, # isSpatial = isSp,
                         origFromClass = origClass)
      }
    }
  }
  x
}

makeVal <- function(x) {
  xValids <- terra::is.valid(x)

  if (any(!xValids)) {
    x <- terra::makeValid(x)
  }

  x
  #
  # whValid <- which(xValids)
  # se <- seq(NROW(x))
  # if (length(whValid)) {
  #   whInValid <- se[-whValid]
  # } else {
  #   whInValid <- se
  # }
  #
  # if (any(!xValids)) {
  #   xGood <- terra::makeValid(x[whInValid])
  #   if (length(whValid)) {
  #     r <- rbind(x[whValid, ], xGood[, ])
  #     x <- r[order(c(whValid, whInValid)),]
  #   } else {
  #     x <- xGood
  #   }
  #
  # }
}


#' @export
#' @rdname postProcessTo
maskTo <- function(from, maskTo, # touches = FALSE,
                   overwrite = FALSE,
                   verbose = getOption("reproducible.verbose"), ...) {
  remapOldArgs(...) # converts studyArea, rasterToMatch, filename2, useSAcrs, targetCRS

  if (!is.null(maskTo) && !extntNA(from) && !extntNA(maskTo)) {
    if (!is.naSpatial(maskTo)) {
      omit <- FALSE
      origFromClass <- class(from)
      if (isRaster(maskTo)) {
        maskTo <- terra::rast(maskTo)
      }
      # if (.isGridded(maskTo) && .isVector(from)) {
      #   omit <- TRUE
      # }
      if (!.isSpatialAny(maskTo)) {
        if (is.na(maskTo) || .isCRSany(maskTo)) omit <- TRUE
      }

      if (!omit) {
        if (.isSF(from) || .isSF(projectTo)) {
          .requireNamespace("sf", stopOnFALSE = TRUE)
        }
        if (isSpatial(from)) {
          from <- sf::st_as_sf(from)
        }

        # New to deal with case where `maskTo` is a SpatRaster
        if (.isGridded(maskTo)) {
          maskToTmp <- !is.na(maskTo)[[1]] # the [[1]] is in case it is a multilayer stack; take first

          # Memory safe alternative to maskToTmp[maskToTmp[] == 0] <- NA
          maskToTmp <- terra::subst(
            x = maskToTmp,
            from = 0,
            to = NA
          )
          maskTo <- terra::as.polygons(maskToTmp)
        }

        if (.isSF(from)) {
          if (!.isSF(maskTo)) {
            maskTo <- sf::st_as_sf(maskTo)
          }
        }
        if (.isSpat(from) && .isVector(from)) {
          if (!.isSpat(maskTo)) {
            maskTo <- terra::vect(maskTo)
          }
        }
        if (!.isSpat(from) && !.isSF(from)) {
          if (.isVector(from)) {
            from <- terra::vect(from)
          } else {
            from <- terra::rast(from)
          }
        }

        if (.requireNamespace("sf")) {
          sameCRS <- sf::st_crs(from) == sf::st_crs(maskTo)
        } else {
          sameCRS <- terra::same.crs(from, maskTo)
        }

        if (!sameCRS) {
          withCallingHandlers({
            isSFHere <- .isSF(maskTo)
            maskTo2 <- maskTo
            attempt <- 1
            while (attempt <= 2) {
              if (.isGridded(maskTo2)) {
                maskTo3 <- terra::project(maskTo2, from, overwrite = overwrite)
              } else {
                if (.isSF(maskTo2)) {
                  maskTo3 <- sf::st_transform(maskTo2, sf::st_crs(from))
                } else {
                  if (isSpatial(maskTo2)) {
                    maskTo2 <- terra::vect(maskTo2)
                  }
                  maskTo3 <- terra::project(maskTo2, from)
                }
              }
              attempt <- attempt + 2
            }
          }, warning = function(w) {
            if (any(grepl(warningCertificateGrep, w$message))) {
              if (!isSFHere) {
                maskTo2 <<- convertToSFwMessage(w, maskTo2)
                attempt <<- 0
              }
              invokeRestart("muffleWarning")
            }
          })
          if (attempt == 4)
            message("... converting to sf object worked to deal with ", warningCertificateGrep)
          maskTo <- maskTo3
        }
        messagePreProcess("masking...", appendLF = FALSE, verbose = verbose)
        st <- Sys.time()

        # There are 2 tries; first is for `maskTo`, second is for `from`, rather than fix both in one step, which may be unnecessary
        maskAttempts <- 0
        env <- environment()

        attempt <- 1
        triedFrom <- NA
        while (attempt <= 2) {
          fromInt <- try(
            {
              if (.isVector(maskTo)) {
                if (length(maskTo) > 1) {
                  if (.isSF(maskTo)) {
                    maskTo <- sf::st_union(maskTo)
                  } else {
                    maskTo <- terra::aggregate(maskTo)
                  }
                }
              }

              if (.isVector(from)) {
                if (.isSF(from)) {
                  sf::st_intersection(from, maskTo)
                } else {
                  maskTo <- fixErrorsIn(maskTo)
                  terra::intersect(from, maskTo)
                }
              } else {
                if (.isGridded(maskTo)) {
                  if (terra::ext(from) > terra::ext(maskTo)) {
                    from <- terra::crop(from, maskTo)
                  }
                  if (terra::ext(maskTo) > terra::ext(from)) {
                    maskTo <- terra::crop(maskTo, from)
                  }
                  terra::mask(from, maskTo, overwrite = overwrite)
                } else {
                  if (.isSF(maskTo) || isSpatial(maskTo)) {
                    maskTo <- terra::vect(maskTo) # alternative is stars, and that is not Suggests
                  }

                  dotArgs <- intersect(...names(), c(writeRasterArgs, maskArgs))
                  if (length(dotArgs)) {
                    dotArgs <- list(...)[dotArgs]
                  }
                  ll <- append(list(from, maskTo, overwrite = overwrite), dotArgs)
                  do.call(terra::mask, ll)

                  # terra::mask(from, maskTo, touches = touches, overwrite = overwrite)
                }
              }
            },
            silent = TRUE
          )
          if (inherits(fromInt, "try-error")) {
            if (attempt == 1) {
              whichFailed <- grepl("geom 0|Loop 0", fromInt)
              if (isTRUE(whichFailed) && !(triedFrom %in% TRUE)) { # don't try same one again
                from <- fixErrorsIn(from, error = fromInt, fromFnName = "maskTo", verbose = verbose)
                triedFrom <- TRUE
              } else {
                maskTo <- fixErrorsIn(maskTo, error = fromInt, fromFnName = "maskTo", verbose = verbose)
                triedFrom <- FALSE
              }
            } else {
              stop(fromInt)
            }
          } else {
            if (attempt > 1) {
              messagePreProcess("...fixed!", verbose = verbose, verboseLevel = 1, appendLF = FALSE)
            }
            break
          }
          attempt <- attempt + 1
        }

        from <- fromInt
        messagePreProcess(messagePrefixDoneIn,
                          format(difftime(Sys.time(), st), units = "secs", digits = 3),
                          verbose = verbose
        )
        from <- revertClass(from, origFromClass = origFromClass)
      }
    }
  }
  from
}

#' @export
#' @rdname postProcessTo
projectTo <- function(from, projectTo, overwrite = FALSE,
                      verbose = getOption("reproducible.verbose"), ...) {
  remapOldArgs(...) # converts studyArea, rasterToMatch, filename2, useSAcrs, targetCRS

  hasMethod <- which(...names() %in% "method")
  method <- if (length(hasMethod)) {
    method <- assessDataTypeOuter(from, ...elt(hasMethod))
  } else {
    NULL
  }

  if (!is.null(projectTo) && !extntNA(from) && !extntNA(projectTo)) {
    origFromClass <- is(from)
    if (!is.naSpatial(projectTo)) {
      if (isRaster(projectTo)) {
        projectTo <- terra::rast(projectTo)
      }

      projectToOrig <- projectTo # keep for below
      sameProj <- try(terra::same.crs(projectTo, from), silent = TRUE)
      if (inherits(sameProj, "try-error")) {
        .requireNamespace("sf", stopOnFALSE = TRUE)
        sameCRS <- sf::st_crs(from) == sf::st_crs(maskTo)
      }

      isProjectToVecOrCRS <- .isCRSany(projectTo) || (.isVector(projectTo))
      sameRes <- if (.isVector(from) || isProjectToVecOrCRS) {
        TRUE
      } else {
        all(terra::res(projectTo) == terra::res(from))
      }

      # if (sameProj && sameRes) {
      #   messagePreProcess("projection of from is same as projectTo, not projecting",
      #                     verbose = verbose)
      # } else {
      if (.isSF(from) || .isSF(projectTo)) {
        .requireNamespace("sf", stopOnFALSE = TRUE)
      }
      messagePreProcess("projecting...",
                        appendLF = FALSE,
                        verbose = verbose
      )
      st <- Sys.time()
      if (isProjectToVecOrCRS && (.isSF(projectTo) || isSpatial(projectTo))) {
        projectToTmp <- sf::st_as_sfc(sf::st_bbox(from))
        if (.isVector(projectTo)) {
          projectTo <- sf::st_crs(projectTo)
        }
        projectToTmp <- sf::st_transform(projectToTmp, projectTo)
        projectTo <- terra::vect(projectToTmp)
      }

      if (.isVector(projectTo)) {
        if (.isGridded(from)) {
          if (!.isSpat(projectTo)) {
            projectTo <- terra::vect(projectTo)
          }

          messagePreProcess("", verbose = verbose)
          messagePreProcess("projectTo is a Vector dataset, which does not define all metadata required. ",
                            verbose = verbose
          )
          if (!terra::is.lonlat(from)) {
            # if (sf::st_crs("epsg:4326") != sf::st_crs(from)) {
            newRes <- terra::res(from)
            messagePreProcess("Using original resolution (", paste(newRes, collapse = "x"), "m) ",
                              verbose = verbose
            )
            projectTo <- terra::rast(projectTo, resolution = newRes)
          } else {
            projectTo <- terra::crs(projectTo)
          }

          messagePreProcess("Using the origin and extent from `ext(from)` in the projection from `crs(projectTo)`.",
                            verbose = verbose
          )
          messagePreProcess("If this is not correct, create a template gridded object and pass that to `projectTo`...",
                            verbose = verbose
          )
          messagePreProcess("",
                            appendLF = FALSE,
                            verbose = verbose
          )
        } else {
          projectTo <- terra::crs(projectTo)
          # projectTo <- sf::st_crs(projectTo)$wkt
        }
      }

      # For some reason, terra::project can't handle a crs class
      if (isCRSSF(projectTo))
        projectTo <- projectTo[["wkt"]]

      # Since we only use the crs when projectTo is a Vector, no need to "fixErrorsIn"
      from <- if (.isVector(from)) {
        isSpatial <- isSpatial(from)
        if (isSpatial) {
          from <- suppressWarningsSpecific(terra::vect(from), shldBeChar)
        }
        withCallingHandlers({
          attempt <- 1
          while (attempt <= 2) {
            isSFHere <- .isSF(from)
            if (isSFHere) {
              if (.isGridded(projectTo)) {
                projectTo <- sf::st_crs(projectTo)
              }
              from13 <- sf::st_transform(from, projectTo)
            } else {
              ll <- list(from, projectTo)
              ll <- addDotArgs(ll, terra::project, class(from), method, ...)
              from13 <- do.call(terra::project, ll)
            }
            attempt <- attempt + 2
          }
        }, warning = function(w) {
          if (any(grepl(warningCertificateGrep, w$message))) {
            if (!isSFHere) {
              from <<- convertToSFwMessage(w, from)
              attempt <<- 0
            }
            invokeRestart("muffleWarning")
          }
        })
        from <- from13
        from <- fixErrorsIn(from) # sometimes `project` makes invalid
        if (attempt == 4)
          message("... converting to sf object worked to deal with ", warningCertificateGrep)


        if (isSpatial) from <- as(from, "Spatial")
        from
      } else {
        dotArgs <- intersect(...names(), c(writeRasterArgs, projectArgs))
        if (length(dotArgs)) {
          dotArgs <- list(...)[dotArgs]
        }
        sameGeom <- if (.isSpat(from) && .isSpat(projectTo) ||
                        (isRaster(from) || isRaster(projectTo))) {
          terra::compareGeom(from, projectTo, stopOnError = FALSE)
        } else {
          FALSE
        }
        if (!isTRUE(sameGeom)) {
          ll <- list(from, projectTo)
          ll <- append(ll, list(overwrite = overwrite))
          ll <- addDotArgs(ll, terra::project, class(from), method, ...)
          # Workaround for terra (>= ~1.9) regression: the shorthand
          # `terra::project(x, char_crs, res = N)` recurses internally and
          # forwards an unrecognized `wopt` set, producing
          # `[write] unknown option(s): xscale,yscale`. When `projectTo` is
          # a CRS (string or `crs` object) and `res = N` was passed via `...`,
          # build the target raster ourselves: project a header-only
          # SpatRaster to compute the output extent in the target CRS, then
          # construct a new raster from (extent, crs, resolution).
          # Using `terra::rast(SpatRaster, resolution = ...)` is ambiguous
          # against the existing nrow/ncol; explicit ext+crs+resolution is
          # the documented constructor that always honors the resolution.
          if (.isCRSany(projectTo) && "res" %in% names(ll)) {
            res0 <- ll$res
            ll$res <- NULL
            proj0 <- terra::project(terra::rast(from), projectTo)
            # Snap extent to a multiple of res0 BEFORE constructing the
            # target. Otherwise terra::rast(extent, resolution = ...) keeps
            # the extent and adjusts resolution to fit exact ncol/nrow,
            # producing res like 249.99999 — which fails strict-equality
            # tests downstream.
            e <- terra::ext(proj0)
            r <- if (length(res0) == 1L) c(res0, res0) else res0
            xmin <- floor(terra::xmin(e) / r[1L]) * r[1L]
            xmax <- ceiling(terra::xmax(e) / r[1L]) * r[1L]
            ymin <- floor(terra::ymin(e) / r[2L]) * r[2L]
            ymax <- ceiling(terra::ymax(e) / r[2L]) * r[2L]
            ll[[2L]] <- terra::rast(
              xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax,
              crs = terra::crs(proj0),
              resolution = res0
            )
          }
          do.call(terra::project, ll)
        } else {
          from
        }
      }
      messagePreProcess(messagePrefixDoneIn, format(difftime(Sys.time(), st), units = "secs", digits = 3),
                        verbose = verbose
      )
    }
    # }
    from <- revertClass(from, origFromClass = origFromClass)
  }
  from
}

#' @param needBuffer Logical. Defaults to `FALSE`, meaning nothing is done out
#'   of the ordinary. If `TRUE`, then a buffer around the cropTo, so that if a reprojection
#'   has to happen on the `cropTo` prior to using it as a crop layer, then a buffer
#'   of 1.5 * res(cropTo) will occur prior, so that no edges are cut off.
#' @rdname postProcessTo
#' @export
cropTo <- function(from, cropTo = NULL, needBuffer = FALSE, overwrite = FALSE,
                   verbose = getOption("reproducible.verbose"), ...) {
  remapOldArgs(...) # converts studyArea, rasterToMatch, filename2, useSAcrs, targetCRS

  if (!is.null(cropTo) && !extntNA(from) && !extntNA(cropTo)) {
    if (.isSF(from) || .isSF(cropTo)) {
      .requireNamespace("sf", stopOnFALSE = TRUE)
    }
    omit <- FALSE
    origFromClass <- is(from)

    if (isRaster(cropTo)) {
      cropToCRS <- sf::st_crs(cropTo)
      cropTo <- terra::rast(cropTo)
      if (!nzchar(terra::crs(cropTo))) {
        terra::crs(cropTo) <- cropToCRS$input
      }
    }

    if (!.isSpatialAny(cropTo)) {
      if (is.na(cropTo) || .isCRSany(cropTo)) omit <- TRUE
    }

    if (!omit) {
      if (isSpatial(cropTo)) {
        cropTo <- terra::vect(cropTo)
      }
      if (isSpatial(from)) {
        from <- terra::vect(from)
      }

      messagePreProcess("cropping...",
                        appendLF = FALSE,
                        verbose = verbose
      )
      st <- Sys.time()

      if (.requireNamespace("sf")) {
        ext <- sf::st_as_sfc(sf::st_bbox(cropTo)) # create extent as an object; keeps crs correctly
        sameCRS <- sf::st_crs(from) == sf::st_crs(ext) # This is sf way of comparing CRS -- raster::compareCRS doesn't work for newer CRS
      } else {
        ext <- terra::ext(cropTo) # create extent as an object; keeps crs correctly
        sameCRS <- terra::same.crs(from, cropTo)
      }

      if (!sameCRS) {

        withCallingHandlers({
          attempt <- 1
          doneWarningAlready <- 0
          while (attempt <= 2) {
            isSFHere <- .isSF(cropTo)

            if (.isVector(cropTo) && !.isSpat(cropTo)) {
              cropToInFromCRS <- sf::st_transform(sf::st_as_sf(cropTo), sf::st_crs(from))
              ext <- sf::st_as_sfc(sf::st_bbox(cropToInFromCRS)) # create extent as an object; keeps crs correctly
            } else {
              terraCRSFrom <- terra::crs(from)
              if (packageVersion("terra") <= "1.5.21") { # an older terra issue; may not be precise version
                if (length(slotNames(terraCRSFrom)) > 0) {
                  terraCRSFrom <- terraCRSFrom@projargs
                }
              }
              if (.isVector(cropTo) && .isGridded(from)) {
                if (!.isSpat(cropTo)) {
                  if (!.isSF(cropTo)) { # e.g., sp
                    if (.requireNamespace("sf", stopOnFALSE = TRUE)) {
                      cropTo <- sf::st_as_sf(cropTo)
                    }
                  }
                  a <- sf::st_convex_hull(cropTo)
                  convH <- terra::vect(a)
                } else {
                  convH <- terra::convHull(cropTo)
                }

                cropToInFromCRS <- terra::project(convH, terraCRSFrom)
              } else {
                # cropToVec <- terra::as.polygons(terra::ext(cropTo), crs = terra::crs(cropTo))
                cropToInFromCRS <- terra::project(cropTo, terraCRSFrom)
              }
              ext <- terra::ext(cropToInFromCRS) # create extent as an object; keeps crs correctly
            }
            attempt <- attempt + 2
          }

        }, warning = function(w) {
          if (any(grepl(warningCertificateGrep, w$message)) && doneWarningAlready == 0) {
            if (!isSFHere) {
              cropTo <<- convertToSFwMessage(w, cropTo)
              attempt <<- 0
              doneWarningAlready <<- 1
            }
            invokeRestart("muffleWarning")
          }
        })
        if (attempt == 4)
          message("... converting to sf object worked to deal with ", warningCertificateGrep)

      }
      if (.isVector(from) && !.isSF(from)) {
        ext <- terra::vect(ext)
      }

      # This is only needed if crop happens before a projection... need to cells beyond edges so projection is accurate
      if (needBuffer) {
        # if (.isGridded(from) || .isGridded(cropTo)) {
          if (.isGridded(from)) {
            res <- terra::res(from)
          } else if (.isGridded(cropTo)) {
            res <- terra::res(cropTo)
          }
          if (!.isSpat(ext)) {
            ext <- terra::vect(ext)
            terra::crs(ext) <- terra::crs(from)
          }
          extTmp <- terra::ext(ext)
          extFrom <- terra::ext(from)
          extOrder <- c("xmin", "ymin", "xmax", "ymax")
          extNum <- extFrom[][extOrder]

          if (!exists("res", inherits = FALSE) || isTRUE(suppressWarnings(terra::is.lonlat(ext)))) {
            # No gridded object (both vectors) OR lonlat CRS: scale buffer with extent size, cap at extFrom.
            # Scaling ensures large polygons at high latitudes get sufficient buffer (original fix from da48ec76).
            # Capping ensures we skip the pre-crop when the buffer reaches the raster edge (no-op crop).
            ranges <- c(abs(terra::xmin(extTmp) - terra::xmax(extTmp)),
                        abs(terra::ymin(extTmp) - terra::ymax(extTmp)))
            extendBy <- min(0.2, max(0.05, (max(ranges) - 20)/20 * 0.5))
            extendBy <- max(ranges) * extendBy
            extTmp2 <- terra::extend(extTmp, extendBy)
            exts <- c(
              xmin = max(terra::xmin(extTmp2), terra::xmin(extFrom)),
              ymin = max(terra::ymin(extTmp2), terra::ymin(extFrom)),
              xmax = min(terra::xmax(extTmp2), terra::xmax(extFrom)),
              ymax = min(terra::ymax(extTmp2), terra::ymax(extFrom))
            )
            ext <- if (packageVersion("terra") <= "1.5-21") {
              terra::ext(exts)
            } else {
              terra::ext(xy = TRUE, exts)
            }
          } else {
            ext <- terra::extend(extTmp, res[1] * 15)
          }

          exts <- ext[][extOrder]
          # This won't work if the the ext is tight with the from i.e., if they are the same;
          #   test and skip cropping with needBuffer =TRUE if it is too tight
          if (isTRUE(any(extNum == exts)) && .isGridded(from))
            return(from)
        # }
      }

      attempt <- 1
      while (attempt <= 2) {
        if (.isGridded(from)) {
          if (!.isSpat(from)) { # terra::crop can handle Raster but only if ext is `extent`
            .requireNamespace("raster", stopOnFALSE = TRUE)
            ext <- raster::extent(ext[])
          }
          fromInt <- try(terra::crop(from, ext, overwrite = overwrite), silent = TRUE)
        } else {
          if (.isSF(from)) {
            fromInt <- try(sf::st_crop(from, ext), silent = TRUE)
          } else {
            fromInt <- try(terra::crop(from, ext), silent = TRUE)
          }
        }

        wasError <- inherits(fromInt, "try-error")
        noOverlap <- NROW(fromInt) == 0
        if (noOverlap || wasError) {
          fail <- FALSE
          if (wasError) {
            if (grepl("extents do not overlap", fromInt)) {
              fail <- TRUE
            }
          }
          if (NROW(fromInt) == 0) { # likely don't overlap
            messagePreProcess("It looks like the cropping results in no data ",
                              "(do not overlap or no intersection)",
                              verbose = verbose
            )
            fail <- FALSE
          }
          if (fail) {
            stop(fromInt)
          }
        }



        if (inherits(fromInt, "try-error")) {
          if (attempt == 1) {
            from <- fixErrorsIn(from, error = fromInt, fromFnName = "cropTo", verbose = verbose)
          } else {
            stop(fromInt)
          }
        } else {
          if (attempt > 1) {
            messagePreProcess("...fixed!",
                              verbose = verbose, verboseLevel = 1,
                              appendLF = FALSE
            )
          }
          break
        }
        attempt <- attempt + 1
      }
      from <- fromInt
      messagePreProcess(messagePrefixDoneIn, format(difftime(Sys.time(), st), units = "secs", digits = 3),
                        verbose = verbose
      )
    }
    from <- revertClass(from, origFromClass = origFromClass)
  }
  from
}

#' @export
#' @rdname postProcessTo
#' @param isStack,isBrick,isRaster,isSpatRaster Logical. Default `NULL`. Used to convert `from`
#'   back to these classes prior to writing, if provided.
#'
writeTo <- function(from, writeTo, overwrite = getOption("reproducible.overwrite"),
                    isStack = NULL, isBrick = NULL, isRaster = NULL,
                    isSpatRaster = NULL,
                    verbose = getOption("reproducible.verbose"), ...) {
  remapOldArgs(...) # converts studyArea, rasterToMatch, filename2, useSAcrs, targetCRS
  if (!missing(writeTo)) {
    if (!is.null(writeTo)) {
      dPath <- which(...names() %in% "destinationPath")
      destinationPath <- if (length(dPath)) {
        destinationPath <- ...elt(dPath)
      } else {
        getOption("reproducible.destinationPath", ".")
      }
      if (all(isAbsolutePath(writeTo))) {
        destinationPath <- dirname(writeTo)
        writeTo <- basename(writeTo)
      }

      writeTo <- determineFilename(writeTo, destinationPath = destinationPath, verbose = verbose)

      hasDatatype <- which(...names() %in% "datatype")
      datatype <- if (length(hasDatatype)) ...elt(hasDatatype) else NULL

      if (isTRUE(isStack)) from <- raster::stack(from)
      if (isTRUE(isBrick)) from <- raster::brick(from)

      writeDone <- FALSE

      if (!any(is.na(writeTo))) {
        .requireNamespace("terra", stopOnFALSE = TRUE)
        messagePreProcess("writing...",
                          appendLF = FALSE,
                          verbose = verbose
        )
        st <- Sys.time()

        if (is.null(isSpatRaster)) isSpatRaster <- .isSpat(from) && .isGridded(from)
        if (is.null(isRaster)) isRaster <- inherits(from, "Raster")

        if (any(file.exists(writeTo))) {
          if (isFALSE(overwrite)) {
            stop(writeTo, " already exists and `overwrite = FALSE`; please set `overwrite = TRUE` and run again.")
          }
          unlink(writeTo, force = TRUE, recursive = TRUE)
        }
        if (isSpatRaster || .isVector(from)) {
          ## trying to prevent write failure and subsequent overwrite error with terra::writeRaster
          # if (any(file.exists(writeTo))) {
          #   if (isFALSE(overwrite)) {
          #     stop(writeTo, " already exists and `overwrite = FALSE`; please set `overwrite = TRUE` and run again.")
          #   }
          #   unlink(writeTo, force = TRUE, recursive = TRUE)
          # }
          if (isSpatRaster) {
            ## if the file still exists it's probably already "loaded"
            ## and `terra` can't overwrite it even if `overwrite = TRUE`
            ## this can happen when multiple modules touch the same object
            if (!any(file.exists(writeTo))) {
              from <- terra::writeRaster(from,
                                         filename = writeTo, overwrite = FALSE,
                                         datatype = datatype
              )
              writeDone <- TRUE
            } else {
              stop("File can't be unlinked for overwrite.")
            }
          } else {
            if (.isSF(from)) {
              written <- sf::st_write(from, dsn = writeTo)
            } else {
              written <- terra::writeVector(from, filename = writeTo, overwrite = FALSE)
            }
            writeDone <- TRUE
          }
        } else if (isRaster) {
          nlyrsFrom <- nlayers2(from)
          if (nlyrsFrom == 1 || length(writeTo) == 1) {
            from <- terra::writeRaster(from,
                                       filename = writeTo, overwrite = overwrite,
                                       datatype = datatype
            )
          } else {
            outs <- lapply(seq(nlyrsFrom), function(ind) {
              out <- terra::writeRaster(from[[ind]],
                                        filename = writeTo[ind], overwrite = overwrite,
                                        datatype = datatype
              )
            })
            from <- raster::stack(outs)
          }
          writeDone <- TRUE
        } else {
          fe <- .fileExtsKnown()
          whType <- fe$extension %in% tools::file_ext(writeTo)
          if (any(whType)) {
            eval(parse(text = fe$saveFun[whType]))(from, writeTo)
          } else {
            messagePreProcess("... nothing written; object not a known object type to write.",
                              verbose = verbose
            )
          }
        }
        if (isTRUE(writeDone)) {
          messagePreProcess(messagePrefixDoneIn, format(difftime(Sys.time(), st), units = "secs", digits = 3),
                            verbose = verbose
          )
        } else {
          messagePreProcess("", verbose = verbose) # need to "end" the line
        }
      }
    }
  }

  from
}

postProcessToAssertions <- function(from, to, cropTo, maskTo, projectTo,
                                    verbose = getOption("reproducible.verbose")) {
  # sometimes there are quosures
  for (y in ls()) {
    ll <- get(y, inherits = FALSE)
    if (inherits(ll, "quosure")) {
      .requireNamespace("rlang", stopOnFALSE = TRUE)
      assign(y, rlang::eval_tidy(ll))
    }
  }

  if (.isSpat(from) || .isSpat(to)) {
    if (!requireNamespace("terra", quietly = TRUE)) {
      stop("Need terra and sf: install.packages(c('terra', 'sf'))")
    }
  }
  if (.isSF(from) || .isSF(to)) {
    if (!requireNamespace("sf", quietly = TRUE)) stop("Need sf: install.packages('sf')")
  }

  if (!(.isSpatialAny(from))) stop("from must be a Spat* or sf*")

  if (!missing(to)) {
    if (!is.null(to)) {
      if (!.isSpatialAny(to) && !.isCRSany(to)) stop("to must be a ", .message$Spatial$anySpatialClass)
      # if (.isVector(from))
      #   if (!.isVector(to) && !.isCRSany(to)) {
      #     # as long as maskTo and projectTo are supplied, then it is OK
      #     if (!.isVector(maskTo) && !.isVector(projectTo))
      #       stop("if from is a Vector object, to must also be a Vector or crs object")
      #   }
    }
  }

  if (!missing(cropTo)) {
    if (!is.naSpatial(cropTo)) {
      if (!is.null(cropTo)) {
        if (!.isSpatialAny(cropTo) && !.isCRSany(cropTo)) {
          stop("cropTo must be a ", .message$Spatial$anySpatialClass)
        }
        # apparently, cropTo can be a gridded object no matter what
        # if (.isVector(from)) if (!.isVector(cropTo) && !.isCRSany(cropTo))
        #   stop("if from is a Vector object, cropTo must also be a Vector object")
      }
    }
  }
  if (!missing(maskTo)) {
    if (!is.naSpatial(maskTo)) {
      if (!is.null(maskTo)) {
        if (!.isSpatialAny(maskTo) && !.isCRSany(maskTo)) {
          stop("maskTo must be a ", .message$Spatial$anySpatialClass)
        }
        # if (.isVector(from)) if (!.isVector(maskTo) && !.isCRSany(maskTo))
        #   stop("if from is a Vector object, maskTo must also be a Vector object")
      }
    }
  }
  if (!missing(projectTo)) {
    if (!is.naSpatial(projectTo)) {
      if (!is.null(projectTo)) {
        if (isCRScharacter(projectTo)) {
          projectTo <- try(silent = TRUE, sf::st_crs(projectTo))
        }
        if (!.isCRSany(projectTo)) {
          if (!.isSpatialAny(projectTo)) {
            stop("projectTo must be a ", .message$Spatial$anySpatialClass)
          }
          # if (.isVector(from)) if (!.isVector(projectTo))
          # stop("if from is a Vector object, projectTo must also be a Vector object")
        }
      }
    }
  }

  return(invisible())
}

is.naSpatial <- function(x) {
  isna <- FALSE
  if (!is.null(x)) {
    if (!.isSpatialAny(x)) {
      if (all(is.na(x))) isna <- TRUE
    }
  }
  isna
}

cropSF <- function(from, cropToVect, verbose = getOption("reproducible.verbose")) {
  st <- Sys.time()
  if (.isSF(from) && (.isSF(cropToVect) || isSpatial(cropToVect))) {
    .requireNamespace("sf", stopOnFALSE = TRUE)
    messagePreProcess("pre-cropping because `from` is sf and cropTo is sf/Spatial*",
                      verbose = verbose
    )
    attempt <- 1
    while (attempt <= 2) {
      from2 <- try(
        sf::st_crop(from, sf::st_transform(
          sf::st_as_sfc(sf::st_bbox(cropToVect)),
          sf::st_crs(from)
        )),
        silent = TRUE
      )
      if (inherits(from2, "try-error")) {
        if (attempt == 1) {
          messageDeclareError(error = from2, fromFnName = "cropSF", verbose)
          from <- fixErrors(from, useCache = FALSE, verbose = verbose - 1)
        } else {
          stop(from2)
        }
      } else {
        if (attempt > 1) {
          messagePreProcess("...fixed!",
                            verbose = verbose, verboseLevel = 1,
                            appendLF = FALSE
          )
        }
        break
      }

      attempt <- attempt + 1
    }

    if (extntNA(from2)) {
      messagePreProcess("resulting extent is NA, probably because objects don't overlap",
                        verbose = verbose
      )
    }
    if (!inherits(from2, "try-error")) {
      from <- from2
    }

    messagePreProcess("done in ", format(difftime(Sys.time(), st),
                                         units = "secs", digits = 3
    ),
    verbose = verbose
    )
  }
  from
}

shldBeChar <- "should be a character value"

revertClass <- function(from, origFromClass = NULL) {
  # if (!isSpat2(origFromClass)) {
  if (!is.null(origFromClass)) {
    # overrides all others!
    isStackHere <- any(origFromClass == "RasterStack")
    isBrickHere <- any(origFromClass == "RasterBrick")
    isRasterLayerHere <- any(origFromClass == "RasterLayer")
    isSFHere <- any(origFromClass == "sf")
    isSpatialHere <- any(startsWith(origFromClass, "Spatial"))
    isSV <- any(origFromClass == "SpatVector")

    if (isSV && !.isSpatVector(from)) {
      from <- terra::vect(from)
    } else if (isStackHere && !inherits(from, "RasterStack")) {
      from <- raster::stack(from) # coming out of writeRaster, becomes brick
    } else if (isBrickHere && !inherits(from, "RasterBrick")) {
      from <- raster::brick(from) # coming out of writeRaster, becomes brick
    } else if (isRasterLayerHere && !inherits(from, "RasterLayer")) {
      from <- raster::raster(from) # coming out of writeRaster, becomes brick
    } else if (isSFHere || isSpatialHere) {
      .requireNamespace("sf", stopOnFALSE = TRUE)
      from <- sf::st_as_sf(from)
      if (isSpatialHere) {
        from <- sf::as_Spatial(from)
      }

    }
  }
  from
}

messageDeclareError <- function(error, fromFnName, verbose = getOption("reproducible.verbose")) {
  errWOWordError <- gsub("Error {0,1}: ", "", error)
  messagePreProcess("", fromFnName, " resulted in following error: \n    - ", errWOWordError, "    --> attempting to fix",
                    appendLF = FALSE, verbose = verbose, verboseLevel = 1
  )
}

#' @importFrom stats na.omit
remapOldArgs <- function(..., fn = sys.function(sys.parent()), envir = parent.frame(),
                         verbose = getOption("reproducible.verbose")) {
  forms <- formals(fn)
  dots <- list(...)

  if (!is.null(dots$filename2)) {
    stop("filename2 arg is deprecated - please pass 'writeTo' instead (see `reproducible::writeTo`")
  }

  dots <- dots[!vapply(dots, is.null, FUN.VALUE = logical(1))]

  ret <- list()

  # First, what is supplied has to be one of the .remappings
  oldUsed <- intersect(names(dots), names(.remappings))
  oldUsed <- oldUsed[na.omit(match(names(.remappings), oldUsed))] # order provided in .remappings; not user

  # Second, what is supplied
  remap <- .remappings[oldUsed]

  newHereAll <- intersect(
    unname(unlist(remap)),
    names(forms)
  )

  if (length(newHereAll)) {
    # remove iterative duplication e.g., cropTo should only come from rtm if both rtm and sa supplied
    newHere <- character()
    Map(re = rev(remap), nam = names(rev(remap)), function(re, nam) {
      newOnes <- setdiff(re, newHere)
      newHere <<- c(newHere, newOnes)
      remap[[nam]] <<- newOnes
    })

    Map(
      elem = oldUsed, newHere = remap,
      function(elem, newHere) {
        if (length(elem)) {
          mes <- paste(newHere, collapse = ", ")
          messagePreProcess(elem, " is supplied (deprecated); assigning it to ", mes,
                            verbose = verbose - 1
          )
          lapply(newHere, function(nh) ret[nh] <<- list(dots[[elem]]))
        }
      }
    )
    if (isTRUE(dots$useSAcrs)) {
      ret$projectTo <- dots$studyArea
    }

    if (isFALSE(dots$useSAcrs)) {
      ret$projectTo <- NULL
    }

    if (isTRUE(dots$maskWithRTM)) {
      ret$maskTo <- dots$rasterToMatch
    }

    # ret <- ret[!vapply(ret, is.null, FUN.VALUE = logical(1))]
    list2env(ret, envir)
  }
  invisible(ret)
}


.remappings <- list(
  studyArea = c("cropTo", "maskTo"),
  rasterToMatch = c("cropTo", "projectTo", "to"),
  filename2 = "writeTo",
  targetCRS = "projectTo",
  useSAcrs = "projectTo"
)


assessDataTypeOuter <- function(from, method) {
  if (.isGridded(from)) {
    if (is.null(method)) {
      method <- assessDataType(from, type = "projectRaster")
    }
    whMethodBilin <- method == "bilinear"

    isInt <- if (inherits(from, "Raster")) {
      if (nlayers2(from) == 1) is.integer(from[]) else apply(from[], 2, is.integer)
    } else {
      terra::is.int(from)
    }

    if (isTRUE(any(whMethodBilin & isInt))) {
      warning("method is bilinear, but the data are integer; please confirm this is correct")
    }
    whMethodNGB <- method == "ngb"
    if (any(whMethodNGB)) {
      method[whMethodNGB] <- "near"
    }
  }
  method
}

## TODO: don't hardcode these, use code to get them; formals() doesn't capture other args
writeRasterArgs <- c(
  ## terra::writeRaster args
  "x", "filename", "overwrite",
  "datatype", "filetype", "gdal", "tempdir", "progress", "memfrac", "memmax", "names", "NAflag",
  "scale", "offset", "verbose", "steps", "todisk", "wopt",

  ## other args
  "memmin", "ncopies", "progressbar", "tolerance" ## TODO: where are these from? terra options?
) |> unique()

projectArgs <- c("x", "y", "method", "mask", "align", "res", "origin", "threads", "filename",
                 "use_gdal", "by_util")

maskArgs <- c("x", "mask", "inverse", "maskvalues", "updatevalue", "touches", "filename")

extntNA <- function(x) {
  out <- if (.isSF(x)) {
    sf::st_bbox(x)
  } else {
    if (.isSpat(x) || isRaster(x)) {
      terra::ext(x)
    } else {
      FALSE
    }
  }
  out <- anyNA(as.numeric(out[]))
  return(out)
}

warningCertificateGrep <- "CertGetCertificateChain trust error CERT_TRUST_IS_PARTIAL_CHAIN"

convertToSFwMessage <- function(w, obj) {
  w$message <- paste(w$message, "\n ... attempting to use `sf` instead")
  message(w$message)
  obj <- sf::st_as_sf(obj)
  obj
}


# extManyPoints <- function(ras, res) {
#   if (missing(res)) {
#     if (isRaster(ras))
#       res <- res(ras)
#     else
#       res <- c(250, 250)
#   }
#
#   xmn <- xmin(ras)
#   xmx <- xmax(ras)
#   ymn <- ymin(ras)
#   ymx <- ymax(ras)
#   xs <- seq(xmn, xmx, by = res[1])
#   ys <- seq(ymn, ymx, by = res[2])
#   rbind(cbind(xs, ymn),
#         cbind(xmx, ys),
#         cbind(rev(xs), ymx),
#         cbind(xmn, rev(ys)),
#         cbind(xmn, ymn)
#   )
# }

# isPoints <- function(geom) {
#   isGeomType(geom, "points")
# }
#
isPolygons <- function(geom) {
  isGeomType(geom, "polygons")
}

# isLines <- function(geom) {
#   isGeomType(geom, "lines")
# }

isGeomType <- function(geom, type) {
  out <- FALSE
  if (.isVector(geom)) {
    out <- if (.isSpat(geom)) {
      if (type == "points")
        terra::is.points(geom)
      else if (type == "polygons")
        terra::is.polygons(geom)
      else if (type == "lines")
        terra::is.lines(geom)
    } else {
      if (type == "points")
        inherits(sf::st_geometry(geom), "sfc_POINT")
      else if (type == "polygons")
        inherits(sf::st_geometry(geom), "sfc_POLYGON")
      else if (type == "lines")
        inherits(sf::st_geometry(geom), "sfc_LINE")
      else {
        warning("geom is not simple point, polygon or line geometry; returning class")
        inherits(sf::st_geometry(geom))
      }
    }
  }
  return(out)
}





#' 3-Step postProcess sequence for SpatRasters using `gdalwarp`
#'
#' DEFUNCT: Please use the `postProcessTo` functions.
#'
#' `gdalProject` is a thin wrapper around `sf::gdal_utils('gdalwarp', ...)` with specific options
#' set, notably, `-r` to `method` (in the ...), `-t_srs` to the crs of the `toRas`,
#' `-te` to the extent of the `toRas`, `-te_srs` to the `crs` of the `toRas`,
#' `-dstnodata = NA`, and `-overwrite`.
#'
#' @details
#' These three functions are used within `postProcessTo`, in the sequence:
#' `gdalProject`, `gdalResample` and `gdalMask`, when `from` and `projectTo` are `SpatRaster` and
#' `maskTo` is a `SpatVector`, but only if `options(reproducible.gdalwarp = TRUE)` is set.
#'
#' This sequence is a slightly different order than the sequence when `gdalwarp = FALSE` or
#' the arguments do not match the above. This sequence was determined to be faster and
#' more accurate than any other sequence, including running all three steps in one
#' `gdalwarp` call (which `gdalwarp` can do). Using one-step `gdalwarp` resulted in
#' very coarse pixelation when converting from a coarse resolution to fine resolution, which
#' visually was inappropriate in test cases.
#'
#' @export
#' @example inst/examples/example_postProcessTo.R
#' @rdname gdalwarpFns
#' @aliases gdalProject
#' @param ... For `gdalProject`, this can be `method`. For `gdalMask` can be `destinationPath` and `touches`.
#'   For all `gdal*`, this can also be and `datatype`.
#' @inheritParams gdalResample
#' @inheritParams postProcessTo
#' @seealso [gdalResample()], and [gdalMask()] and the overarching [postProcessTo()]
gdalProject <- function(fromRas, toRas, filenameDest, verbose = getOption("reproducible.verbose"), ...) {

  .Defunct("postProcessTo")
  # if (!requireNamespace("sf", quietly = TRUE) && !requireNamespace("terra", quietly = TRUE))
  #   stop("Can't use gdalProject without sf and terra")
  #
  # messagePreProcess("running gdalProject ...", appendLF = FALSE, verbose = verbose)
  # st <- Sys.time()
  #
  # hasMethod <- which(...names() %in% "method")
  # method <- if (length(hasMethod)) {
  #   method <- assessDataTypeOuter(fromRas, ...elt(hasMethod))
  # } else {
  #   NULL
  # }
  # if (is.null(method))
  #   method <- "near"
  #
  # fns <- unique(Filenames(fromRas))
  # if (length(fns) == 1 && isTRUE(nzchar(fns))) {
  #   fnSource <- fns
  # } else {
  #   fnSource <- tempfile(fileext = ".tif")
  #   terra::writeRaster(fromRas, filename = fnSource)
  #   on.exit(unlink(fnSource))
  # }
  #
  # if (missing(filenameDest))
  #   filenameDest <- tempfile(fileext = ".tif")
  #
  # tf4 <- tempfile(fileext = ".prj")
  # on.exit(unlink(tf4), add = TRUE)
  # cat(sf::st_crs(toRas)$wkt, file = tf4)
  #
  # threads <- detectThreads()
  #
  # opts <- c(
  #   "-t_srs", tf4,
  #   "-r", method,
  #   "-te", as.vector(terra::ext(toRas))[c(1, 3, 2, 4)],
  #   "-te_srs", tf4,
  #   "-wo", paste0("NUM_THREADS=", threads),
  #   # "-srcnodata", "NA",
  #   "-dstnodata", "NA", # THERE IS AN ARTIFACT THAT OCCURS
  #   "-overwrite"
  # )
  #
  # opts <- addDataType(opts, fromRas[[1]], ...)
  # opts <- updateDstNoData(opts, fromRas)
  #
  # tried <- retry(retries = 2,
  #                sf::gdal_utils(
  #                  util = "warp",
  #                  source = fnSource,
  #                  destination = filenameDest,
  #                  options = opts))
  #
  # out <- terra::rast(filenameDest)
  # messagePreProcess(messagePrefixDoneIn,
  #                   format(difftime(Sys.time(), st), units = "secs", digits = 3),
  #                   verbose = verbose)
  #
  # out
}



#' @description
#' `gdalResample` is a thin wrapper around `sf::gdal_utils('gdalwarp', ...)` with specific options
#' set, notably, `"-r", "near"`, `-te`, `-te_srs`, `tr`, `-dstnodata = NA`, `-overwrite`.
#'
#'
#' @export
#' @param fromRas see `from` argument from [postProcessTo()], but can only be a  `SpatRaster`.
#' @param toRas see `to` argument from [postProcessTo()], but can only be a  `SpatRaster`.
#' @param filenameDest A filename with an appropriate extension (e.g., `.tif`) for
#'   `gdal` to write the output to. Since this function is conceived to be part of a
#'   chain, and not the final step, this function does not use `writeTo`, which is
#'   reserved for the final step in the chain.
#' @inheritParams postProcessTo
#' @rdname gdalwarpFns
#' @aliases gdalResample
gdalResample <- function(fromRas, toRas, filenameDest, verbose = getOption("reproducible.verbose"), ...) {
  .Defunct("projectTo")

  # if (!requireNamespace("sf") && !requireNamespace("terra"))
  #   stop("Can't use gdalResample without sf and terra")
  #
  # messagePreProcess("running gdalResample ...", appendLF = FALSE, verbose = verbose)
  # st <- Sys.time()
  #
  # hasMethod <- which(...names() %in% "method")
  # method <- if (length(hasMethod)) {
  #   method <- assessDataTypeOuter(fromRas, ...elt(hasMethod))
  # } else {
  #   NULL
  # }
  # if (is.null(method))
  #   method <- "near"
  #
  # fns <- unique(Filenames(fromRas))
  # if (length(fns) ==1 && isTRUE(nzchar(fns))) {
  #   fnSource <- fns
  # } else {
  #   fnSource <- tempfile(fileext = ".tif")
  #   terra::writeRaster(fromRas, filename = fnSource)
  #   on.exit(unlink(fnSource))
  # }
  #
  # if (missing(filenameDest))
  #   filenameDest <- tempfile(fileext = ".tif")
  #
  # tf4 <- tempfile(fileext = ".prj")
  # cat(sf::st_crs(toRas)$wkt, file = tf4)
  #
  #
  # opts <- c(
  #   "-r", method,
  #   "-te", c(terra::xmin(toRas), terra::ymin(toRas),
  #            terra::xmin(toRas) + (terra::ncol(toRas) ) * terra::res(toRas)[1],
  #            terra::ymin(toRas) + (terra::nrow(toRas) ) * terra::res(toRas)[2]),
  #   "-te_srs", tf4, # 3347, 3348, 3978, 3979
  #   "-tr", terra::res(toRas),
  #   "-dstnodata", "NA",
  #   # "-tap",
  #   "-overwrite"
  # )
  #
  # opts <- addDataType(opts, fromRas[[1]], ...)
  # opts <- updateDstNoData(opts, fromRas)
  #
  # tried <- retry(retries = 2,
  #                sf::gdal_utils(
  #                  util = "warp",
  #                  source = fnSource,
  #                  destination = filenameDest,
  #                  options = opts)
  # )
  #
  # out <- terra::rast(filenameDest)
  # messagePreProcess(messagePrefixDoneIn,
  #                   format(difftime(Sys.time(), st), units = "secs", digits = 3),
  #                   verbose = verbose)
  # out
}


#' @description
#' `gdalMask` is a thin wrapper around `sf::gdal_utils('gdalwarp', ...)` with specific options
#' set, notably, `-cutline`, `-dstnodata = NA`, and `-overwrite`.
#'
#' @export
#' @param fromRas see `from` argument from [postProcessTo()], but can only be a  `SpatRaster`.
#' @param maskToVect see `maskTo` argument from [maskTo()], but can only be a `SpatVector`
#' @inheritParams postProcessTo
#' @rdname gdalwarpFns
#' @aliases gdalMask
gdalMask <- function(fromRas, maskToVect, writeTo = NULL, verbose = getOption("reproducible.verbose"), ...) {
  .Defunct("maskTo")

  # if (!requireNamespace("sf") && !requireNamespace("terra"))
  #   stop("Can't use gdalMask without sf and terra")
  #
  # messagePreProcess("running gdalMask ...", appendLF = FALSE, verbose = verbose)
  # st <- Sys.time()
  #
  # fns <- unique(Filenames(fromRas))
  #
  # if (length(fns) ==1 && isTRUE(nzchar(fns))) {
  #   fnSource <- fns
  # } else {
  #   fnSource <- tempfile(fileext = ".tif")
  #   terra::writeRaster(fromRas, filename = fnSource)
  #   on.exit(unlink(fnSource))
  # }
  #
  # tf3 <- tempfile(fileext = ".shp")
  # on.exit(unlink(tf3), add = TRUE)
  # if (.isGridded(maskToVect)) { # not used by default because postProcessTo will return couldDoGDAL = FALSE
  #   if (!.isSpatRaster(maskToVect)) {
  #     maskToVect <- terra::rast(maskToVect)
  #   }
  #   maskToVect <- terra::as.polygons(maskToVect, values = FALSE)
  # }
  # if (.isSF(maskToVect)) {
  #   shp <- sf::st_transform(maskToVect, terra::crs(fromRas))
  #   sf::st_write(shp, dsn = tf3)
  # } else {
  #   shp <- terra::project(maskToVect, terra::crs(fromRas))
  #   terra::writeVector(shp, filename = tf3)
  # }
  #
  # dPath <- which(...names() %in% "destinationPath")
  # destinationPath <- if (length(dPath)) {
  #   destinationPath <- ...elt(dPath)
  # } else {
  #   getOption("reproducible.destinationPath", ".")
  # }
  #
  # if (is.null(writeTo))
  #   writeTo <- tempfile(fileext = ".tif")
  #
  # writeTo <- determineFilename(writeTo, destinationPath = destinationPath, verbose = verbose)
  #
  # opts <- c(
  #   "-cutline", tf3,
  #   "-dstnodata", "NA",
  #   "-overwrite"
  # )
  # if (!isFALSE(list(...)$touches)) # default is TRUE, like terra::mask
  #   opts <- c(opts, "-wo", "CUTLINE_ALL_TOUCHED=TRUE")
  #
  # opts <- addDataType(opts, fromRas[[1]], ...)
  # opts <- updateDstNoData(opts, fromRas)
  #
  # tried <- retry(retries = 2,
  #                sf::gdal_utils(
  #                  util = "warp",
  #                  source = fnSource,
  #                  destination = writeTo,
  #                  options = opts))
  #
  # out <- terra::rast(writeTo)
  # messagePreProcess(messagePrefixDoneIn,
  #                   format(difftime(Sys.time(), st), units = "secs", digits = 3),
  #                   verbose = verbose)
  # out
}

messagePrefixDoneIn <- "\bdone! took:  "


#' Keep original geometries of `sf` objects
#'
#' When intersections occur, what was originally 2 polygons features can become
#' `LINESTRING` and/or `POINT` and any `COLLECTIONS` or `MULTI-` versions of these.
#' This function evaluates what the original geometry was and drops any newly created
#' *different* geometries. For example, if a `POLYGON` becomes a `COLLECTION` of
#' `MULTIPOLYGON`, `POLYGON` and `POINT` geometries, the `POINT` geometries will
#' be dropped. This function is used internally in [postProcessTo()].
#' @param newObj The new, derived `sf` object
#' @param origObj The previous, object whose geometries should be used.
#' @return The original `newObj`, but with only the type of geometry that entered
#' into the function.
#'
keepOrigGeom <- function(newObj, origObj) {
  from2Geom <- sort(unique(sf::st_geometry_type(newObj)))
  fromGeom <- sort(unique(sf::st_geometry_type(origObj)))
  if (!identical(from2Geom, fromGeom)) {
    possTypes <- c("POINT", "LINESTRING", "POLYGON")
    hasTypes <- vapply(possTypes, function(pt) isTRUE(any(grepl(pt, fromGeom))), FUN.VALUE = logical(1))
    fromGeomSimple <- names(hasTypes)[hasTypes]

    has2Types <- vapply(possTypes, function(pt) isTRUE(any(grepl(pt, from2Geom))), FUN.VALUE = logical(1))
    from2GeomSimple <- names(has2Types)[has2Types]

    # hasTypes2 <- sapply(possTypes, function(pt) any(grepl(pt, from2Geom)))#, FUN.VALUE = logical(1))
    # from2GeomSimple <- names(hasTypes2)[hasTypes2]

    # isSameTypeAsFromGeom <- apply(do.call(rbind, lapply(fromGeom, function(fg) grepl(fg, from2Geom))), 2, all)
    if (!all(from2GeomSimple %in% fromGeomSimple)) {
      # hasMulti <- grepl("MULTI", from2Geom) & isSameTypeAsFromGeom
      newObj <- sf::st_collection_extract(newObj, type = as.character(fromGeomSimple))
    }
  }
  newObj
}

detectThreads <- function(threads = getOption("reproducible.gdalwarpThreads", 2)) {
  isNotNumThreads <- !is.numeric(threads)
  lenNumThreadsNot1 <- length(threads) > 1
  isNAThreads <- all(is.na(threads))
  isNULLThreads <- is.null(threads)
  if (isNULLThreads) {
    threads <- 1L
  } else {
    if (isNotNumThreads || isNAThreads || lenNumThreadsNot1) {
      if (isNotNumThreads)
        threads <- 2L
      if (lenNumThreadsNot1)
        threads <- threads[1]
      if (isNAThreads)
        threads <- 1L
    } else {
      if (threads < 1) {
        threads <- 1L
      } else {
        .requireNamespace("parallel", stopOnFALSE = TRUE)
        detCors <- parallel::detectCores()
        if (threads > detCors)
          threads <- detCors - 1
      }
    }
  }

  threads
}

addDataType <- function(opts, fromRas, ...) {
  hasDatatype <- which(...names() %in% "datatype")
  datatype <- if (length(hasDatatype)) ...elt(hasDatatype) else {
    assessDataType(fromRas, type = "GDAL")
  }
  if (!is.null(datatype)) {
    datatype <- switchDataTypes(datatype, type = "GDAL")
    opts <- c(opts, "-ot", datatype)
  }
  opts <- unname(opts)
  opts
}

gdalTransform <- function(from, cropTo, projectTo, maskTo, writeTo, verbose) {
  messagePreProcess("running gdalTransform ...", appendLF = FALSE, verbose = verbose)
  st <- Sys.time()
  tf4 <- tempfile(fileext = ".prj")
  on.exit(unlink(tf4), add = TRUE)
  wkt1 <- sf::st_crs(projectTo)$wkt
  cat(wkt1, file = tf4)
  tf <- tempfile(fileext = ".shp")
  tf2 <- tempfile(fileext = ".shp")
  # tf3 <- tempfile(fileext = ".shp")
  if (!sf::st_crs(projectTo) == sf::st_crs(maskTo)) {
    stop("maskTo and projectTo must have the same crs")
  }
  # prjFile <- dir(dirname(from), pattern = paste0(basename(tools::file_path_sans_ext(from)), ".prj"), full.names = TRUE)
  # maskToInFromCRS <- terra::project(maskTo, prjFile)
  # terra::writeVector(maskToInFromCRS, filename = tf3, overwrite = TRUE)
  # system.time(sf::gdal_utils(util = "vectortranslate", source = "C:/Eliot/GitHub/Edehzhie/modules/fireSense_dataPrepFit/data/NFDB_poly_20210707.shp",
  #                        destination = tf2, options =
  #                          c(# "-t_srs", tf4,
  #                            "-clipdst", tf3, "-overwrite"
  #                          )))
  # system.time(gdal_utils(util = "vectortranslate",
  #                        source = tf2,
  #                        destination = tf,
  #                        options =
  #                          c("-t_srs", tf4,
  #                            # "-clipdst", tf2,
  #                            "-overwrite"
  #                          )))
  #
  terra::writeVector(maskTo, filename = tf2)
  # system.time(
    sf::gdal_utils(util = "vectortranslate", source = "C:/Eliot/GitHub/Edehzhie/modules/fireSense_dataPrepFit/data/NFDB_poly_20210707.shp",
                             destination = tf, options =
                               c("-t_srs", tf4,
                                 "-clipdst", tf2, "-overwrite"
                               ))# )
  messagePreProcess(messagePrefixDoneIn,
                    format(difftime(Sys.time(), st), units = "secs", digits = 3),
                    verbose = verbose)
  tf
}

updateDstNoData <- function(opts, fromRas) {
  hasDashOT <- which(opts %in% "-ot")
  valForNoData <- MaxVals[[switchDataTypes(opts[hasDashOT + 1], "writeRaster")]]
  hasMM <- terra::hasMinMax(fromRas)
  if (!isTRUE(all(hasMM)))
    fromRas <- terra::setMinMax(fromRas)
  minmaxRas <- terra::minmax(fromRas)
  if (any(minmaxRas[2, ] >= valForNoData)) {
    valForNoData <- MinVals[[switchDataTypes(opts[hasDashOT + 1], "writeRaster")]]
    if (any(minmaxRas[1, ] <= valForNoData))
      valForNoData <- NA
  }
  valForNoData <- as.character(valForNoData)
  valForNoData
  hasDstNoData <- which(opts %in% "-dstnodata")
  va <- try(valForNoData)
  if (length(va) == 0) browser()
  if (is(va, "try-error")) browser()

  opts[hasDstNoData + 1] <- va
  opts
}

addDotArgsInner <- function(ll, fun, class, ...) {
  formsForProject <- formals4reproducible(fun, class)
  dotsElements <- intersect(names(formsForProject), ...names())
  if (length(dotsElements)) {
    mc <- match.call(expand.dots = FALSE)
    dots <- mc$...
    ll <- append(ll, dots[dotsElements])
  }
  ll
}


addDotArgs <- function(ll, fun, class, method, ...) {
  ll <- addDotArgsInner(ll, fun, class, ...)
  if (!is.null(method))
    ll <- modifyList(ll, list(method = method))
  ll
}

test_that("test Copy", {
  skip_if_not_installed("terra")
  testInit(c("terra", "data.table"), tmpFileExt = ".tif")

  ras <- terra::rast(terra::ext(0, 10, 0, 10), vals = 1)
  ras <- suppressWarningsSpecific(
    falseWarnings = proj6Warn,
    writeRaster(ras, filename = tmpfile, overwrite = TRUE)
  )
  # This will make hardlink
  ras2 <- suppressWarningsSpecific(Copy(ras, tmpdir), "NOT UPDATED FOR PROJ >= 6")

  expect_true(all.equal(ras2[], ras[]))
  expect_false(Filenames(ras2) == Filenames(ras))

  dt <- data.table(a = 1:2, b = rev(LETTERS[1:2]))
  tmpdir <- normPath(tempdir2("ras2"))
  checkPath(tmpdir, create = TRUE)
  on.exit(unlink(tmpdir, recursive = TRUE), add = TRUE)
  li <- list(dt = dt, ras = ras, ras2 = ras2)
  li2 <- Copy(li, tmpdir)

  # same content
  expect_true(all(unlist(lapply(seq_along(li), function(i) {
    if (.isSpatRaster(li[[i]])) {
      all.equal(values2(li[[i]]), values2(li2[[i]]))
    } else {
      all.equal(li[[i]], li2[[i]])
    }
  }))))

  # different filenames for Rasters
  expect_false(all(unlist(lapply(seq_along(li)[-1], function(i) {
    isTRUE(all.equal(Filenames(li[[i]]), Filenames(li2[[i]])))
  }))))

  # data.table
  setkeyv(li[[1]], "b")
  expect_false(isTRUE(all.equal(li[[1]], li2[[1]])))

  ### environments
  dt <- data.table(a = 1:2, b = rev(LETTERS[1:2]))
  li <- list(dt = dt, ras = ras, ras2 = ras2)
  li <- list2env(li, envir = new.env(parent = emptyenv()))

  tmpdir <- tempdir2("ras3")
  li2 <- Copy(li, tmpdir)

  expect_true(all(unlist(lapply(names(li), function(i) {
    if (.isSpatRaster(li[[i]])) {
      all.equal(li[[i]][], li2[[i]][])
    } else {
      all.equal(li[[i]], li2[[i]])
    }
  }))))

  # different filenames for Rasters
  expect_false(all(unlist(lapply(names(li)[-1], function(i) {
    isTRUE(all.equal(Filenames(li[[i]]), Filenames(li2[[i]])))
  }))))

  # data.table
  setkeyv(li[["dt"]], "b")
  expect_false(isTRUE(all.equal(li[["dt"]], li2[["dt"]])))

  ### Nested Environments
  dt <- data.table(a = 1:2, b = rev(LETTERS[1:2]))
  li <- list(dt = dt, ras = ras, ras2 = ras2)
  env1 <- new.env(parent = emptyenv())
  env2 <- new.env(parent = emptyenv())
  liEnv <- list2env(li, envir = env1)
  liEnv[["env"]] <- li

  tmpdir <- tempdir2("ras3")
  liEnv2 <- Copy(liEnv, tmpdir)

  expect_true(all(unlist(lapply(names(liEnv[["env"]]), function(i) {
    if (.isSpatRaster(li[[i]])) {
      all.equal(liEnv[["env"]][[i]][], liEnv2[["env"]][[i]][])
    } else {
      all.equal(liEnv[["env"]][[i]], liEnv2[["env"]][[i]])
    }
  }))))

  # different filenames for Rasters
  expect_false(all(unlist(lapply(names(liEnv[["env"]])[-1], function(i) {
    isTRUE(all.equal(Filenames(liEnv[["env"]][[i]]), Filenames(liEnv2[["env"]][[i]])))
  }))))

  # data.table
  setkeyv(liEnv[["env"]][["dt"]], "b")
  expect_false(isTRUE(all.equal(liEnv[["env"]][["dt"]], liEnv2[["env"]][["dt"]])))

  ###################
  # This is an aside on testing whether the hard link can be messed with by overwrite
  fn <- Filenames(ras)
  ras3 <- terra::rast(terra::ext(0, 10, 0, 10), vals = 2)
  ras3 <- suppressWarningsSpecific(
    writeRaster(ras3, filename = tmpfile, overwrite = TRUE),
    "NOT UPDATED FOR PROJ >= 6"
  ) # overwrite the original
  # The hardlink is not affected by "overwrite = TRUE" -- it is not by filename, but by file location
  expect_true(all(ras2[] == 1))
})

test_that("lightweight tests for code coverage", {
  skip_on_cran()

  testInit(
    c("sf", "terra"),
    opts = list("reproducible.overwrite" = TRUE, "reproducible.inputPaths" = NULL),
    needGoogleDriveAuth = TRUE
  )
  dPath <- checkPath(file.path(tempdir2()), create = TRUE)
  dPath2 <- checkPath(file.path(tempdir2()), create = TRUE)

  cloudFolderID <- "https://drive.google.com/drive/folders/1An8s2YLFPopQKr4BWK9o06fLSXx-Zggw"
  targetFile <- "fireSenseParams.rds"
  targetFile2 <- "fireSenseParams2.gpkg"
  localFileLux <- system.file("ex/lux.shp", package = "terra")

  # 1 step for each layer
  # 1st step -- get study area
  (full <- prepInputs(localFileLux, destinationPath = dPath)) |> capture.output() -> co # default is sf::st_read
  zoneA <- full[3:6, c("NAME_1", "AREA")]
  zoneB <- full[8, c("NAME_1", "AREA")] # not in A
  zoneC <- full[3, c("NAME_1", "AREA")] # yes in A
  zoneD <- full[7:8, c("NAME_1", "AREA")] # not in A, B or C
  zoneE <- full[3:5, c("NAME_1", "AREA")] # yes in A

  # This will be 1, 2 and 3 -- THIS IS THE INTERESTING ONE ... it will mean that a
  #  test below will have 2 different polygons the are "contains", so, result of
  #  Cache will be not just one polygon, but 2
  zoneF <- aggregate(
    full[, c("AREA")],
    by = list(NAME_1 = c(rep(1, 3), rep(2, NROW(full) - 3))),
    sum
  )
  zoneF <- zoneF[zoneF$NAME_1 == 1, ]

  # zoneF[, "AREA"] <- sf::st_area(zoneF)/1e6
  # 2nd step: re-write to disk as read/write is lossy; want all "from disk" for this ex.
  co <- capture.output({
    writeTo(zoneA, writeTo = "zoneA.shp", destinationPath = dPath)
    writeTo(zoneB, writeTo = "zoneB.shp", destinationPath = dPath)
    writeTo(zoneC, writeTo = "zoneC.shp", destinationPath = dPath)
    writeTo(zoneD, writeTo = "zoneD.shp", destinationPath = dPath)
    writeTo(zoneE, writeTo = "zoneE.shp", destinationPath = dPath)
    writeTo(zoneF, writeTo = "zoneF.shp", destinationPath = dPath)
    # Must re-read to get identical columns
    zoneA <- sf::st_read(file.path(dPath, "zoneA.shp"))
    zoneB <- sf::st_read(file.path(dPath, "zoneB.shp"))
    zoneC <- sf::st_read(file.path(dPath, "zoneC.shp"))
    zoneD <- sf::st_read(file.path(dPath, "zoneD.shp"))
    zoneE <- sf::st_read(file.path(dPath, "zoneE.shp"))
    zoneF <- sf::st_read(file.path(dPath, "zoneF.shp"))
  })

  # The function that is to be run. This example returns a data.frame because
  #    saving `sf` class objects with list-like columns does not work with
  #    many st_driver()
  fun <- function(domain, newField) {
    domain |>
      as.data.frame() |>
      cbind(params = I(lapply(seq_len(NROW(domain)), function(x) newField)))
  }

  # fun2 <- function(domain, newField) {
  #   domain |> as.data.frame() |>
  #     dplyr::mutate(params2 = list(list(a = seq_len(NROW(domain)),
  #                                   b = LETTERS[seq_len(NROW(domain))],
  #                                   d = TRUE)))
  # }
  fun3 <- function(domain, paramsVec) {
    cbind(domain, params = I(lapply(seq(NROW(domain)), function(x) paramsVec)))
  }

  # Run sequence -- A, B will add new entries in targetFile, C will not,
  #                 D will, E will not
  paramsVecList <- list(
    list(a = 1, b = 2, c = "D"),
    list(a = 2, b = 3, d = 4),
    list(a = 2, b = 3, e = 4),
    list(a = 2, b = 3, d = 4),
    list(a = 2, b = 3, d = 4),
    list(a = 2, b = 3, d = 4)
  )
  iter <- 0
  for (z in list(zoneA, zoneB, zoneC, zoneD, zoneE, zoneF)) {
    iter <- iter + 1
    if (identical(z, zoneA)) {
      # First iteration: targetFile doesn't exist yet, so different message
      mess <- "No remote targetFile exists"
    } else if (identical(z, zoneB) || identical(z, zoneD) || identical(z, zoneF)) {
      mess <- "Domain is not contained within the targetFile"
    }
    if (identical(z, zoneC) || identical(z, zoneE)) {
      mess <- "Spatial domain is contained within the url"
    }

    messCap <- capture_messages(
      out <- CacheGeo(
        targetFile = targetFile,
        domain = z,
        FUN = fun(domain, newField = I(list(list(a = 1, b = 1:2, c = "D")))),
        fun = fun, # pass whatever is needed into the function
        destinationPath = dPath,
        action = "update",
        verbose = 0
      )
    )
    expect_match(messCap, mess, all = FALSE)

    co <- capture.output({
      warns <- capture_warnings(expect_message(
        out2 <- CacheGeo(
          targetFile = targetFile2,
          domain = z,
          FUN = fun3(domain, paramsVec = paramsVecList[[iter]]),
          fun3 = fun3, # pass whatever is needed into the function
          paramsVecList = paramsVecList,
          iter = iter,
          destinationPath = dPath,
          action = "update"
        ),
        mess
      ))
    })

    if (NROW(warns)) {
      expect_match(warns, substr(.message$BecauseOfLossOfColumn(""), start = 1, 10), all = FALSE)
      expect_match(warns, "Dropping", all = FALSE)
    }
  }

  outSF <- sf::st_as_sf(out)

  skip_if_service_account_releaseVer_NotLinux()

  gls <- googledrive::drive_ls(cloudFolderID)
  alreadyThere <- gls$name %in% c(targetFile, targetFile2)

  if (any(alreadyThere)) {
    googledrive::drive_rm(gls$id[which(alreadyThere)])
  }
  on.exit({
    gls <- googledrive::drive_ls(cloudFolderID)
    googledrive::drive_rm(gls[gls$name %in% c(targetFile, targetFile2), ])
  })

  iter <- 0
  # the following will fail if not predictiveecology@gmail.com or eliotmcintire@gmail.com or the funky
  #   service account eliot-githubauthentication@genial-cycling-408722.iam.gserviceaccount.com if that
  #   has been added to the environment
  try(
    {
      for (z in list(zoneA, zoneB, zoneC, zoneD, zoneE, zoneF)) {
        iter <- iter + 1
        if (
          identical(z, zoneA) || identical(z, zoneB) || identical(z, zoneD) || identical(z, zoneF)
        ) {
          mess <- "Domain is not contained within the targetFile"
        }
        if (identical(z, zoneC) || identical(z, zoneE)) {
          mess <- "Spatial domain is contained within the url"
        }
        # With directory url

        out <- CacheGeo(
          targetFile = targetFile,
          domain = z,
          useCloud = TRUE,
          cloudFolderID = cloudFolderID,
          FUN = fun(domain, newField = I(list(list(a = 1, b = 1:2, c = "D")))),
          fun = fun, # pass whatever is needed into the function
          destinationPath = dPath2,
          action = "update",
          verbose = 0
        )

        co <- capture.output(
          warns <- capture_warnings(expect_message(
            out2 <- CacheGeo(
              targetFile = targetFile2,
              domain = z,
              useCloud = TRUE,
              cloudFolderID = cloudFolderID,
              FUN = fun3(domain, paramsVec = paramsVecList[[iter]]),
              fun3 = fun3, # pass whatever is needed into the function
              paramsVecList = paramsVecList,
              iter = iter,
              destinationPath = dPath,
              action = "update",
              verbose = 0
            ),
            mess
          ))
        )
        if (NROW(warns)) {
          expect_match(
            warns,
            substr(.message$BecauseOfLossOfColumn(""), start = 1, 10),
            all = FALSE
          )
          expect_match(warns, "Dropping", all = FALSE)
        }
      }
      outSFCloud <- sf::st_as_sf(out)
      expect_true(identical(outSFCloud, outSF))

      keeps <- sf::st_contains(outSF, outSF[1, 1], sparse = FALSE)

      polysWithParams <- outSF[keeps, ]

      expect_true(NROW(polysWithParams) == 2)

      smaller <- sf::st_as_sf(terra::buffer(terra::vect(polysWithParams[1, ]), width = -2000))

      plot(polysWithParams[2, 1], reset = FALSE)
      plot(polysWithParams[1, 1], add = TRUE, col = "red", reset = FALSE)
      smaller <- sf::st_as_sf(terra::buffer(terra::vect(polysWithParams[1, ]), width = -2000))
      plot(smaller[1, 1], add = TRUE, col = "green")

      out <- CacheGeo(
        targetFile = targetFile,
        domain = smaller,
        useCloud = TRUE,
        cloudFolderID = cloudFolderID,
        FUN = fun(domain, newField = I(list(list(a = 1, b = 1:2, c = "D")))),
        fun = fun, # pass whatever is needed into the function
        destinationPath = dPath2,
        action = "nothing"
      )
      outSFCloudSmaller <- sf::st_as_sf(out)
      expect_identical(as.data.frame(outSFCloudSmaller)[, "params"], out[, "params"])
    },
    silent = TRUE
  )
})

test_that("prepInputs correctly unzips large files", {
  skip_on_cran()
  skip_on_ci()
  skip_if_not(isTRUE(getOption("reproducible.runLargeFileTests")),
              "set reproducible.runLargeFileTests=TRUE to run (~30 GB extraction)")
  ## based on #145. extracted file is ~30 GB so this takes a long time to test!
  testInit("terra")
  # tmpdir <- "/mnt/d/temp" # need a drive that is large enough
  # if (!"emcintir" %in% Sys.info()["user"] || (!dir.exists(tmpdir)))
  #   skip("This requires a lot of drive space")
  targFile <- "CA_harvest_year_1985_2015.tif"
  on.exit(
    {
      unlink(file.path(tmpdir, targFile), recursive = TRUE)
    },
    add = TRUE
  )

  url <- "https://opendata.nfis.org/downloads/forest_change/CA_forest_harvest_mask_year_1985_2015.zip"
  withr::local_options(reproducible.tempPath = file.path(tmpdir, "ttt"))
  ff <- prepInputs(
    url = url,
    targetFile = targFile,
    destinationPath = asPath(tmpdir),
    fun = "terra::rast",
    userTags = c("objectName:forestHarvestMask", "goal:posthocGIS")
  )
  fout <- file.path(tmpdir, "CA_harvest_year_1985_2015.tif")
  expect_true(identical(normPath(Filenames(ff)), fout))
  expect_true(file.info(fout)[["size"]] > 28 * 1024^3)
})

test_that("Issue 181 geodatabase file", {
  skip_on_cran()
  skip_on_ci()
  skip_if_not(isTRUE(getOption("reproducible.runLargeFileTests")),
              "set reproducible.runLargeFileTests=TRUE to run (Google Drive download)")

  ## based on #145. extracted file is ~30 GB so this takes a long time to test!
  testInit(c("terra", "googledrive"), needGoogleDriveAuth = TRUE)
  rstLCC <- prepInputs(
    targetFile = "EOSD_Mosaic.gdb",
    archive = "EOSD_2000_2007_combined.zip",
    alsoExtract = "similar",
    url = "https://drive.google.com/file/d/1p66_P6dNdlrvAF3Mp99Xz9Bdz2lvfaQ7",
    destinationPath = tmpdir,
    filename2 = NULL,
    fun = NA) |>
    Cache(userTags = c(
      "outFun:Cache",
      "step:prepEOSD"
    )
  )
  expect_true(is(sf::st_read(rstLCC, layer = "EOSD_Mosaic_BWC_range_clip", quiet = TRUE), "sf"))
})

test_that("Issue 242 masking fail", {
  skip_on_cran()
  skip_on_ci()
  testInit("terra", needInternet = TRUE, verbose = FALSE)
  # skip_if_not(isInteractive(), "test #3: extracting large files should be run manually with devtools::test()")
  testInit(c("terra", "googledrive"), needGoogleDriveAuth = FALSE)
  studyArea <- vect(structure(
    c(
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1418427,
      -1415931, -1398140, -1393331, -1382857, -1371358, -1365393, -1350548,
      -1325832, -1323489, -1322689, -1321448, -1319605, -1318493, -1309216,
      -1307387, -1305766, -1305267, -1303130, -1295515, -1296061, -1296486,
      -1298156, -1299017, -1300951, -1300981, -1301682, -1302736, -1317683,
      -1320782, -1322070, -1322341, -1324137, -1335839, -1336905, -1346232,
      -1401188, -1422500, -1423243, -1423419, -1418427, 6990085, 6999544,
      7008667, 7021910, 7028787, 7029407, 7029780, 7035153, 7036928,
      7039810, 7050226, 7044431, 7035561, 7027977, 7027582, 7026342,
      7021796, 7006494, 7006125, 6994606, 6994507, 6984394, 6975963,
      6975848, 6968927, 6961588, 6957530, 6933733, 6933100, 6926385,
      6925727, 6929217, 6932967, 6947379, 6948395, 6954427, 6980171,
      6980300, 6987758, 6989296, 6990085, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    ),
    dim = c(41L, 5L),
    dimnames = list(NULL, c("geom", "part", "x", "y", "hole"))
  ))
  crs(studyArea) <- "PROJCRS[\"unknown\",\nBASEGEOGCRS[\"unknown\",\nDATUM[\"North American Datum 1983\",\nELLIPSOID[\"GRS 1980\",6378137,298.257222101,\nLENGTHUNIT[\"metre\",1]],\nID[\"EPSG\",6269]],\nPRIMEM[\"Greenwich\",0,\nANGLEUNIT[\"degree\",0.0174532925199433],\nID[\"EPSG\",8901]]],\nCONVERSION[\"unknown\",\nMETHOD[\"Lambert Conic Conformal (2SP)\",\nID[\"EPSG\",9802]],\nPARAMETER[\"Latitude of false origin\",0,\nANGLEUNIT[\"degree\",0.0174532925199433],\nID[\"EPSG\",8821]],\nPARAMETER[\"Longitude of false origin\",-95,\nANGLEUNIT[\"degree\",0.0174532925199433],\nID[\"EPSG\",8822]],\nPARAMETER[\"Latitude of 1st standard parallel\",49,\nANGLEUNIT[\"degree\",0.0174532925199433],\nID[\"EPSG\",8823]],\nPARAMETER[\"Latitude of 2nd standard parallel\",77,\nANGLEUNIT[\"degree\",0.0174532925199433],\nID[\"EPSG\",8824]],\nPARAMETER[\"Easting at false origin\",0,\nLENGTHUNIT[\"metre\",1],\nID[\"EPSG\",8826]],\nPARAMETER[\"Northing at false origin\",0,\nLENGTHUNIT[\"metre\",1],\nID[\"EPSG\",8827]]],\nCS[Cartesian,2],\nAXIS[\"(E)\",east,\nORDER[1],\nLENGTHUNIT[\"metre\",1,\nID[\"EPSG\",9001]]],\nAXIS[\"(N)\",north,\nORDER[2],\nLENGTHUNIT[\"metre\",1,\nID[\"EPSG\",9001]]]]"
  studyAreaRas <- rasterize(
    studyArea,
    rast(extent = ext(studyArea), crs = crs(studyArea), resolution = 250)
  )
  sppAbundURL <- "https://ftp.maps.canada.ca/pub/nrcan_rncan/Forests_Foret/canada-forests-attributes_attributs-forests-canada/2001-attributes_attributs-2001/NFI_MODIS250m_2001_kNN_Species_Pinu_Alb_v1.tif"
  a <- capture.output(
    sppAbundance <- prepInputs(
      targetFile = "NFI_MODIS250m_2001_kNN_Species_Pinu_Alb_v1.tif",
      url = sppAbundURL,
      destinationPath = tempdir2(),
      fun = "terra::rast",
      to = studyAreaRas,
      overwrite = TRUE
    )
  )
  expect_true(.isSpatRaster(sppAbundance))
  expect_true(all.equal(ext(studyAreaRas), ext(sppAbundance)))
})

templateURL <- "https://ftp.maps.canada.ca/pub/nrcan_rncan/Forests_Foret/SCANFI/v1/SCANFI_sps_douglasFir_SW_2020_v1.2.tif"

# Small polygon in BC in geographic coordinates
studyAreaBC <- function() {
  terra::vect(
    matrix(c(-122, 50.95, -121.95, 50.95, -121.95, 51, -122, 51, -122, 50.95), ncol = 2, byrow = TRUE),
    type = "polygons", crs = "EPSG:4326"
  )
}

test_that("prepInputsCOG returns NULL for invalid pre-conditions", {
  skip_if_not_installed("terra")

  expect_identical(prepInputsCOG(NULL,        cropTo = studyAreaBC()), "NULL")
  expect_identical(prepInputsCOG("not-a-url", cropTo = studyAreaBC()), "NULL")
  expect_identical(prepInputsCOG(templateURL),                         "NULL") # no to/cropTo/maskTo

  # Non-GeoTiff URLs (archives etc.) must short-circuit before any /vsicurl/ read.
  expect_identical(
    prepInputsCOG("https://geodata.ucdavis.edu/geodata/elv/CAN_elv_msk.zip",
                  cropTo = studyAreaBC()),
    "NULL"
  )
  expect_identical(
    prepInputsCOG("https://example.com/data.tar.gz", cropTo = studyAreaBC()),
    "NULL"
  )
  # Query/fragment after a .tif must still be recognized as a tif.
  expect_identical(
    prepInputsCOG("https://example.com/x.zip?foo=bar", cropTo = studyAreaBC()),
    "NULL"
  )
})

test_that("prepInputsCOG returns a windowed SpatRaster for the SCANFI tif", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE, verbose = FALSE)

  result <- skip_if_transient_stream_warnings(
    prepInputsCOG(templateURL, cropTo = studyAreaBC())
  )

  expect_true(.isSpatRaster(result))
  expect_gt(terra::ncell(result), 0L)
})

test_that("prepInputs COG fast-path works with cropTo/maskTo (stays in native CRS)", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE, verbose = FALSE)

  result <- skip_if_transient_stream_warnings(
    prepInputs(url = templateURL, cropTo = studyAreaBC(), maskTo = studyAreaBC(),
               destinationPath = tmpdir)
  )

  expect_true(.isSpatRaster(result))
  expect_gt(terra::ncell(result), 0L)
})

test_that("testing terra", {
  # if (interactive()) {
  testInit("terra",
    needGoogleDriveAuth = FALSE,
    opts = list(
      reproducible.useMemoise = FALSE,
      reproducible.cacheSaveFormat = .qs2Format,
      "rgdal_show_exportToProj4_warnings" = "none"
    )
  )
  withr::local_options(reproducible.cachePath = tmpCache)

  skip_if_not_installed("terra")
  f <- system.file("ex/elev.tif", package = "terra")
  tf <- tempfile(fileext = ".tif")
  tf1 <- tempfile(fileext = ".tif")
  tf2 <- tempfile(fileext = ".tif")
  tf3 <- tempfile(fileext = ".tif")
  tf4 <- tempfile(fileext = ".tif")
  tf5 <- tempfile(fileext = ".tif")
  tf6 <- tempfile(fileext = ".tif")
  tf7 <- tempfile(fileext = ".tif") # don't create it: testing writeTo
  file.copy(f, tf)
  file.copy(f, tf1)
  file.copy(f, tf2)
  file.copy(f, tf3)
  file.copy(f, tf4)
  file.copy(f, tf5)
  file.copy(f, tf6)
  r <- list(terra::rast(f), terra::rast(tf))
  r1 <- list(terra::rast(tf1), terra::rast(tf2))
  r2 <- list(terra::rast(tf3), terra::rast(tf4))
  rmem <- r2
  terra::values(rmem[[1]]) <- values2(rmem[[1]])
  terra::values(rmem[[2]]) <- values2(rmem[[2]])

  fn <- function(listOf) {
    listOf
  }

  # Test Cache of various nested and non nested SpatRaster
  # double nest
  b <- Cache(fn, list(r, r1), cachePath = tmpCache)
  expect_true(is(b, "list"))
  expect_true(is(b[[1]], "list"))
  expect_true(.isSpatRaster(b[[1]][[1]]))

  # Single nest
  b <- Cache(fn, r, cachePath = tmpCache)
  expect_true(is(b, "list"))
  expect_true(.isSpatRaster(b[[1]]))

  # mixed nest
  b <- Cache(fn, list(r[[1]], r1), cachePath = tmpCache)
  expect_true(is(b, "list"))
  expect_true(.isSpatRaster(b[[1]]))
  expect_true(.isSpatRaster(b[[2]][[1]]))

  # mix memory and disk
  b <- Cache(fn, list(r[[1]], r1, rmem), cachePath = tmpCache)
  expect_true(is(b, "list"))
  expect_true(.isSpatRaster(b[[1]]))
  expect_true(.isSpatRaster(b[[2]][[1]]))
  expect_true(terra::inMemory(b[[3]][[1]]))
  expect_true(!terra::inMemory(b[[2]][[1]]))
  expect_true(!terra::inMemory(b[[1]]))

  f <- system.file("ex/lux.shp", package = "terra")
  vOrig <- terra::vect(f)
  v <- vOrig[1:2, ]
  rf <- system.file("ex/elev.tif", package = "terra")
  xOrig <- terra::rast(rf)
  elevRas <- terra::deepcopy(xOrig)
  xCut <- terra::classify(xOrig, rcl = 5)
  xVect <- terra::as.polygons(xCut)
  xVect2 <- terra::deepcopy(xVect)

  y <- terra::deepcopy(elevRas)
  y[y > 200 & y < 300] <- NA_integer_
  terra::values(elevRas) <- rep(1L, terra::ncell(y))
  vRast <- terra::rast(v, resolution = 0.008333333)

  # SR, SR
  t1 <- postProcessTo(elevRas, y)
  expect_true(sum(is.na(t1[]) != is.na(y[])) == 0)

  t7 <- postProcessTo(elevRas, projectTo = y)
  expect_true(identical(t7, elevRas))

  t8 <- postProcessTo(elevRas, maskTo = y)
  expect_true(all.equal(t8, t1))

  t9 <- postProcessTo(elevRas, cropTo = vRast)
  expect_true(terra::ext(v) <= terra::ext(t9))

  # SR, SV
  t2 <- postProcessTo(elevRas, v)

  # No crop
  t3 <- postProcessTo(elevRas, maskTo = v)
  expect_true(terra::ext(t3) == terra::ext(elevRas))

  if (.requireNamespace("sf")) {
    vsf <- sf::st_as_sf(v)
    vOrigsf <- sf::st_as_sf(vOrig)
  }

  t4 <- postProcessTo(elevRas, cropTo = v, maskTo = v)
  expect_true(terra::ext(t4) == terra::ext(t2))

  t5 <- postProcessTo(elevRas, cropTo = v, maskTo = v, projectTo = v)
  expect_true(identical(t5[], t2[]))

  if (.requireNamespace("sf")) {
    t5sf <- postProcessTo(elevRas, cropTo = vsf, maskTo = vsf, projectTo = vsf)
    expect_true(identical(t5sf[], t2[]))
  }

  t6 <- terra::extract(elevRas, v, mean, na.rm = TRUE)
  expect_true(all(t6$elevation == 1))
  expect_true(NROW(t6) == 2)

  # Only writeTo
  expect_false(file.exists(tf7))
  t11a <- suppressWarnings({
    postProcessTo(elevRas, writeTo = tf7)
  }) ## WARNING: Discarded datum Unknown based on GRS80 ellipsoid in Proj4 definition
  expect_true(file.exists(tf2))
  expect_equivalent(elevRas, t11a)


  t10 <- postProcessTo(xVect, v)
  expect_true(terra::ext(t10) < terra::ext(xVect))

  #
  ## following #253
  # https://github.com/PredictiveEcology/reproducible/issues/253#issuecomment-1263562631
  tf1 <- tempfile(fileext = ".shp")
  tf2 <- tempfile(fileext = ".shp")
  t11 <- suppressWarnings({
    postProcessTo(xVect, v, writeTo = tf1)
  }) ## WARNING: Discarded datum Unknown based on GRS80 ellipsoid in Proj4 definition

  # Only writeTo
  t11a <- suppressWarnings({
    postProcessTo(xVect, writeTo = tf2)
  }) ## WARNING: Discarded datum Unknown based on GRS80 ellipsoid in Proj4 definition
  expect_true(file.exists(tf2))
  expect_true(identical(xVect, t11a))

  tw_t11 <- terra::wrap(t11)
  vv <- terra::vect(tf1)
  tw_vv <- terra::wrap(vv)
  expect_true(terra::same.crs(tw_vv@crs, tw_t11@crs))

  ## following #253 with different driver
  ## https://github.com/PredictiveEcology/reproducible/issues/253#issuecomment-1263562631
  tf1 <- tempfile(fileext = ".gpkg")
  t11 <- suppressWarnings({
    postProcessTo(xVect, v, writeTo = tf1)
  }) ## WARNING: GDAL Message 6: dataset does not support layer creation option ENCODING

  tw_t11 <- terra::wrap(t11)
  vv <- terra::vect(tf1)
  tw_vv <- terra::wrap(vv)
  expect_equivalent(tw_vv, tw_t11) ## TODO: not identical

  # Test fixErrorTerra
  v1 <- terra::simplifyGeom(v)
  gv1 <- terra::geom(v1)
  gv1[gv1[, "geom"] == 2, "geom"] <- 1
  # gv1[9,"y"] <- 51
  v2 <- terra::vect(gv1, "polygons")
  # plot(v2)
  # v2 <- is.valid(v2)

  terra::crs(v2) <- terra::crs(v)

  # Projection --> BAD BUG HERE ... CAN"T REPRODUCE ALWAYS --> use sf for testing Dec 9, 2022
  if (FALSE) {
    utm <- terra::crs("epsg:23028") # sf::st_crs("epsg:23028")$wkt
    # albers <- sf::st_crs("epsg:5070")$wkt
    vutm <- terra::project(v, utm)
  }

  if (.requireNamespace("sf")) {
    utmWKT <- paste0('PROJCRS["ED50 / UTM zone 28N",
    BASEGEOGCRS["ED50",
        DATUM["European Datum 1950",
            ELLIPSOID["International 1924",6378388,297,
                LENGTHUNIT["metre",1]]],
        PRIMEM["Greenwich",0,
            ANGLEUNIT["degree",0.0174532925199433]],
        ID["EPSG",4230]],
    CONVERSION["UTM zone 28N",
        METHOD["Transverse Mercator",
            ID["EPSG",9807]],
        PARAMETER["Latitude of natural origin",0,
            ANGLEUNIT["degree",0.0174532925199433],
            ID["EPSG",8801]],
        PARAMETER["Longitude of natural origin",-15,
            ANGLEUNIT["degree",0.0174532925199433],
            ID["EPSG",8802]],
        PARAMETER["Scale factor at natural origin",0.9996,
            SCALEUNIT["unity",1],
            ID["EPSG",8805]],
        PARAMETER["False easting",500000,
            LENGTHUNIT["metre",1],
            ID["EPSG",8806]],
        PARAMETER["False northing",0,
            LENGTHUNIT["metre",1],
            ID["EPSG",8807]]],
    CS[Cartesian,2],
        AXIS["(E)",east,
            ORDER[1],
            LENGTHUNIT["metre",1]],
        AXIS["(N)",north,
            ORDER[2],
            LENGTHUNIT["metre",1]],
    USAGE[
        SCOPE["Engineering survey, topographic mapping."],
        AREA["Europe - between 18°W and 12°W - Ireland offshore."],
        BBOX[48.43,-16.1,56.57,-12]],
    ID["EPSG",23028]]')
    # utm <- sf::st_crs("epsg:23028")#$wkt

    vsfutm <- sf::st_transform(vsf, utmWKT)
    vutm <- terra::vect(vsfutm)
    res100 <- 100
    rutm <- terra::rast(vutm, resolution = res100)
    rutm <- terra::rasterize(vutm, rutm, field = "NAME_2")

    vsfInUTMviaCRS <- postProcessTo(vsf, sf::st_crs(rutm))
    expect_true(is(vsfInUTMviaCRS, "sf"))
    expect_true(terra::same.crs(vsfInUTMviaCRS, rutm))

    # from is sf, to is SpatRast --> skip maskTo
    if (isWindows()) {
      vsfInUTMviaSpatRast <-
        suppressWarningsSpecific(
          falseWarnings = "attribute variables are assumed",
          postProcessTo(vOrigsf, rutm)
        )
      expect_true(is(vsfInUTMviaSpatRast, "sf"))
      expect_true(sf::st_crs(vsfInUTMviaSpatRast) == sf::st_crs(rutm))
      expect_true(isTRUE(all.equal(
        round(terra::ext(rutm), 6),
        round(terra::ext(vsfInUTMviaSpatRast), 6)
      )))
    }

    # Check for cases where `to` does not overlap with `from`
    ext1 <- terra::ext(vsf) + c(-2, 2, -2, 2)
    ext1SR <- terra::rast(ext1)
    if (.requireNamespace("raster")) {
      ext1Ra <- raster::raster(ext1SR)
      ext1Ex <- raster::extent(ext1Ra)
    }
    ext2 <- terra::vect(ext1)
    terra::crs(ext2) <- terra::crs(vsf)
    ext3 <- sf::st_as_sf(ext2)
    if (.requireNamespace("sf")) {
      expect_warning(expect_message(postProcessTo(vOrigsf, ext3))) # sf gives warning too
      expect_warning(expect_message(postProcessTo(terra::vect(vOrigsf), ext2)),
                     "no intersection")
      # expect_warning(expect_error(postProcessTo(vOrigsf, ext3))) # sf gives warning too
      # expect_error(postProcessTo(terra::vect(vOrigsf), ext2))
    }
    if (.requireNamespace("sf")) {
      if (.requireNamespace("sp")) {
        expect_error(postProcessTo(as(vOrigsf, "Spatial"), as(ext2, "Spatial")))
      }
      if (.requireNamespace("raster")) {
        expect_error(postProcessTo(as(vOrigsf, "Spatial"), ext1Ra))
      }
    }

    t11 <- postProcessTo(elevRas, vutm)
    expect_true(terra::same.crs(t11, vutm))

    # use raster dataset -- take the projectTo resolution, i.e., res100
    t13 <- postProcessTo(elevRas, rutm)
    expect_true(identical(terra::res(t13)[1], res100))
    expect_true(terra::same.crs(t13, vutm))

    # no projection
    t12 <- postProcessTo(elevRas, cropTo = vutm, maskTo = vutm)
    expect_false(terra::same.crs(t12, vutm))

    # projection with errors
    if (isWindows()) { # bug in older `terra` that is not going to be fixed here
      # utm <- terra::crs("epsg:23028") # This is same as above, but terra way
      v2 <- terra::makeValid(v2)
      vutmErrors <- terra::project(v2, utmWKT)

      # Switch from qs2 to rds with Cache
      if (requireNamespace(.qs2Format)) {
        opts <- options(reproducible.cacheSaveFormat = .qs2Format)
        t13a <- Cache(postProcessTo(xVect, vutmErrors))
        opts <- options(reproducible.cacheSaveFormat = .rdsFormat)
        t13a <- Cache(postProcessTo(xVect, vutmErrors))
        opts <- options(reproducible.cacheSaveFormat = .qs2Format)
        t13b <- Cache(postProcessTo(xVect, vutmErrors))
        expect_equal(t13a, t13b)
        # a <- try(ncol(t13a), silent = TRUE)
        # expect_false(is(a, "try-error"))
      }
    }

    # try NA to *To
    # Vectors
    vutmSF <- sf::st_as_sf(vutm)
    xVect2SF <- sf::st_as_sf(xVect2)
    t14 <- postProcessTo(xVect2, vutm, projectTo = NA)

    expect_true(terra::same.crs(t14, xVect2))
    expect_false(terra::same.crs(t14, vutm))

    if (isWindows()) { # bug in older `terra` that is not going to be fixed here
      suppressWarningsSpecific(
        falseWarnings = "attribute variables",
        t14SF <- postProcessTo(xVect2SF, vutmSF, projectTo = NA)
      )
      expect_true(terra::same.crs(t14SF, xVect2SF))
      expect_false(terra::same.crs(t14SF, vutmSF))
    }

    t15 <- postProcessTo(xVect2, vutm, maskTo = NA)

    expect_false(terra::same.crs(t15, xVect2))
    expect_true(terra::same.crs(t15, vutm))

    if (isWindows()) { # bug in older `terra` that is not going to be fixed here
      suppressWarningsSpecific(
        falseWarnings = "attribute variables",
        t15SF <- postProcessTo(xVect2SF, vutmSF, maskTo = NA)
      )
      expect_false(terra::same.crs(t15SF, xVect2SF))
      expect_true(terra::same.crs(t15SF, vutmSF))
    }

    t18 <- postProcessTo(xVect2, vutm, cropTo = NA)
    expect_false(terra::same.crs(t18, xVect2))
    expect_true(terra::same.crs(t18, vutm))

    suppressWarningsSpecific(
      falseWarnings = "attribute variables",
      t18SF <- postProcessTo(xVect2SF, vutmSF, cropTo = NA)
    )
    expect_false(terra::same.crs(t18SF, xVect2SF))
    expect_true(terra::same.crs(t18SF, vutmSF))

    # Rasters
    t16 <- postProcessTo(elevRas, rutm, cropTo = NA)
    expect_false(terra::same.crs(t16, elevRas))
    expect_true(terra::same.crs(t16, rutm))
    expect_true(terra::ext(t16) >= terra::ext(rutm))

    t17 <- postProcessTo(elevRas, rutm, projectTo = NA)
    expect_true(terra::same.crs(t17, elevRas))
    expect_false(terra::same.crs(t17, rutm))

    t19 <- postProcessTo(elevRas, rutm, maskTo = NA)
    expect_false(terra::same.crs(t19, elevRas))
    expect_true(terra::same.crs(t19, vutm))
    expect_true(sum(values2(t19), na.rm = TRUE) > sum(values2(t13), na.rm = TRUE))

    # Raster with Vector
    t16 <- postProcessTo(elevRas, vutm, cropTo = NA)
    expect_false(terra::same.crs(t16, elevRas))
    expect_true(terra::same.crs(t16, vutm))

    t16SF <- postProcessTo(elevRas, vutmSF, cropTo = NA)
    expect_false(terra::same.crs(t16SF, elevRas))
    expect_true(terra::same.crs(t16SF, vutmSF))

    t17 <- postProcessTo(elevRas, vutm, projectTo = NA)
    expect_true(terra::same.crs(t17, elevRas))
    expect_false(terra::same.crs(t17, vutm))

    t17SF <- postProcessTo(elevRas, vutmSF, projectTo = NA)
    expect_true(terra::same.crs(t17SF, elevRas))
    expect_false(terra::same.crs(t17SF, vutmSF))

    t19 <- postProcessTo(elevRas, vutm, maskTo = NA)
    expect_false(terra::same.crs(t19, elevRas))
    expect_true(terra::same.crs(t19, vutm))
    # expect_true(sum(values2(t19), na.rm = TRUE) > sum(values2(t13), na.rm = TRUE))

    t19SF <- postProcessTo(elevRas, vutmSF, maskTo = NA)
    expect_false(terra::same.crs(t19SF, elevRas))
    expect_true(terra::same.crs(t19SF, vutmSF))
    # expect_true(sum(values2(t19SF), na.rm = TRUE) > sum(values2(t13), na.rm = TRUE))

    t21 <- postProcessTo(elevRas, projectTo = vutm)
    t21SF <- postProcessTo(elevRas, projectTo = vutmSF)
    t20 <- postProcessTo(elevRas, projectTo = terra::crs(vutm))
    expect_true(all.equal(t20, t21))
    expect_true(all.equal(t20, t21SF))
    # This is now (Apr 24, 2023) because passing all to terra::project --> which keep square pixels
    expect_false(identical(terra::size(elevRas), terra::size(t20)))

    t20res250 <- postProcessTo(elevRas, projectTo = terra::crs(vutm), res = 250)
    expect_true(all(terra::res(t20res250) == 250))

    ## same projection change resolution only (will likely affect extent)
    y2 <- terra::rast(crs = terra::crs(y), resolution = 0.008333333 * 2, extent = terra::ext(y))
    y2 <- terra::setValues(y2, rep(1, terra::ncell(y2)))

    t22 <- postProcessTo(elevRas, to = y2, overwrite = TRUE) # not sure why need this; R devel on Winbuilder Nov 26, 2022
    expect_true(terra::same.crs(t22, elevRas))
    expect_true(terra::ext(t22) == terra::ext(y2)) ## "identical" may say FALSE (decimal plates?)
    expect_true(identical(terra::res(t22), terra::res(y2)))
    expect_false(identical(terra::res(t22), terra::res(elevRas)))

    vutmSF <- sf::st_as_sf(vutm)
    xVectSF <- sf::st_as_sf(xVect)
    ## It is a real warning about geometry stuff, but not relevant here
    err <- capture_error( # there is an error in R 4.0 Windows
      warn <- capture_warnings({
        t22 <- postProcessTo(xVectSF, vutmSF)
      })
    )
    #  }

    if (.requireNamespace("raster")) {
      ras1 <- raster::raster(tf5)
      ras2 <- raster::raster(tf6)
      ras1[] <- ras1[]
      ras2[] <- ras2[]
      r3 <- list(ras1, ras2) # this is for
      r3[[1]][] <- r3[[1]][]
      r3[[2]][] <- r3[[2]][] # bring to RAM

      # Raster & SpatVect
      ras1Small <- cropTo(ras1, t18)
      ras1SmallMasked <- maskTo(ras1, t18)
      # The warning is about some datum
      suppressWarnings(ras1SmallAll <- postProcessTo(ras1, t18))
      expect_true(is(ras1Small, "Raster"))
      expect_true(is(ras1SmallMasked, "Raster"))
      expect_true(is(ras1SmallAll, "Raster"))

      # Check that extents are similar
      expect_true(terra::ext(ras1SmallAll) < terra::extend(terra::ext(t18), terra::res(ras1SmallAll) * 2))

      # SpatRaster & Raster
      t20CroppedByRas <- cropTo(t20, ras1SmallAll)
      t20MaskedByRas <- maskTo(t20, ras1SmallAll)
      t20ProjectedByRas <- projectTo(t20, ras1SmallAll)
      t20AllByRas <- postProcessTo(t20, ras1SmallAll) # only does terra::rast 1x
      expect_true(.isSpatRaster(t20AllByRas))
      expect_true(.isSpatRaster(t20CroppedByRas))
      expect_true(.isSpatRaster(t20MaskedByRas))
      expect_true(.isSpatRaster(t20ProjectedByRas))
      expect_equal(terra::res(t20ProjectedByRas), terra::res(ras1SmallAll))
      expect_equal(terra::res(t20AllByRas), terra::res(ras1SmallAll))
      expect_equal(terra::ext(t20AllByRas), terra::ext(ras1SmallAll))
      expect_equal(terra::ext(t20ProjectedByRas), terra::ext(ras1SmallAll))
      expect_true(
        sum(!is.na(values2(ras1SmallAll))) <
          sum(!is.na(values2(t20ProjectedByRas)))
      )
      expect_true(
        sum(!is.na(values2(ras1SmallAll))) ==
          sum(!is.na(values2(t20AllByRas)))
      )

      # The below was slightly off because RasterLayer was not exactly same as t20 proj
      spatRas1SmallAll <- projectTo(terra::rast(ras1SmallAll), t20)
      expect_equal( #  these are off b/c of projection probably
        sum(!is.na(values2(spatRas1SmallAll))),
        sum(!is.na(values2(t20MaskedByRas)))
      )

      if (FALSE) {
        dev.off()
        terra::plot(ras1SmallAll)
        terra::plot(t18, add = TRUE)
        terra::plot(t20AllByRas)
        terra::plot(t20CroppedByRas)
        terra::plot(t20MaskedByRas)
        terra::plot(t20ProjectedByRas)
      }

      w <- terra::vect("POLYGON ((0 -5, 10 0, 10 -10, 0 -5))")
      w1 <- terra::vect("POLYGON ((0 -5, 10 0, 10 -10, 4 -2, 0 -5))")
      w1a <- terra::vect("POLYGON ((20 15, 30 20, 30 10, 24 18, 20 15))")
      w2 <- rbind(w, w1, w, w1a, w)
      w3 <- fixErrorsIn(w2)
      expect_true(!all(terra::is.valid(w2)))
      expect_true(all(terra::is.valid(w3)))
    }
  }
})

utils::globalVariables(c(
  "goe", "goc", "gauth_path"
))

#' A wrapper around a set of downloading functions
#'
#' Currently, this only deals with `googledrive::drive_download`,
#' and [utils::download.file()]. In general, this is not intended for use by a
#' user.
#'
#' @inheritParams prepInputs
#' @inheritParams preProcess
#' @inheritParams prepInputs
#' @inheritParams extractFromArchive
#' @param overwrite Logical. If `TRUE` then the download will overwrite an existing file
#'   if it exists.
#' @param dlFun Optional "download function" name, such as `"raster::getData"`, which does
#'              custom downloading, in addition to loading into R. Still experimental.
#' @param ... Passed to `dlFun`. Still experimental. Can be e.g., `type` for google docs.
#' @param checksumFile A character string indicating the absolute path to the `CHECKSUMS.txt`
#'                     file.
#' @inheritParams loadFromCache
#' @inheritParams Cache
#' @author Eliot McIntire
#' @return
#' This function is called for its side effects, which will be a downloaded file
#' (`targetFile`), placed in `destinationPath`. This file will be checksummed, and
#' that checksum will be appended to the `checksumFile`.
#'
#' @export
#' @include checksums.R
downloadFile <- function(archive, targetFile, neededFiles,
                         destinationPath = getOption("reproducible.destinationPath", "."), quick,
                         checksumFile, dlFun = NULL,
                         checkSums, url, needChecksums, preDigest,
                         overwrite = getOption("reproducible.overwrite", TRUE),
                         alsoExtract = "similar",
                         verbose = getOption("reproducible.verbose", 1),
                         purge = FALSE, .tempPath, .callingEnv,
                         ...) {
  dots <- list(...)
  # if (is.null(dots$.callingEnv)) {
  #   .callingEnv <- parent.frame()
  # } else {
  #   .callingEnv <- dots$.callingEnv
  #   dots$.callingEnv <- NULL
  # }

  # browser(expr = exists("._downloadFile_1"))
  if (missing(.tempPath)) {
    .tempPath <- tempdir2(rndstr(1, 6))
    on.exit(unlink(.tempPath, recursive = TRUE), add = TRUE)
  }
  if (missing(targetFile)) {
    targetFile <- NULL
  }

  if (!is.null(url) || !is.null(dlFun)) {
    missingNeededFiles <- missingFiles(neededFiles, checkSums, destinationPath)

    if (missingNeededFiles) { # needed may be missing, but maybe can skip download b/c archive exists
      if (!is.null(archive)) {
        localArchivesExist <- file.exists(archive)
        if (any(localArchivesExist)) {
          filesInLocalArchives <- unique(unlist(lapply(archive, .listFilesInArchive)))
          neededFilesRel <- makeRelative(neededFiles, destinationPath)
          haveAll <- if (isNULLorNA(neededFiles) || length(neededFiles) == 0) FALSE else all(neededFilesRel %in% filesInLocalArchives)
          if (haveAll) { # local archive has all files needed
            extractedFromArchive <- extractFromArchive(
              archive = archive[localArchivesExist],
              destinationPath = destinationPath,
              neededFiles = neededFiles, checkSums = checkSums,
              needChecksums = needChecksums,
              checkSumFilePath = checksumFile,
              quick = quick,
              .tempPath = .tempPath
            )
            checkSums <- if (!file.exists(checksumFile) || is.null(neededFiles) || length(neededFiles) == 0) {
              needChecksums <- 1
              .emptyChecksumsResult
            } else {
              Checksums(
                files = neededFiles,
                checksumFile = checksumFile,
                path = destinationPath,
                quickCheck = quick,
                write = FALSE,
                verbose = verbose
              )
            }

            # Check again, post extract ... If FALSE now, then it got it from local, already existing archive
            missingNeededFiles <- missingFiles(neededFiles, checkSums, destinationPath)
            if (!missingNeededFiles) {
              archive <- archive[localArchivesExist]
            }
          } else {
            messagePreProcess("Have local archive, ", archive, ", but its files are not listed in the CHECKSUMS.txt file.", verbose = verbose)
            messagePreProcess("\nRedownloading to start from file at url...", verbose = verbose)
          }
        }
      }
    }

    if (missingNeededFiles) {
      if (needChecksums == 0) needChecksums <- 2 # use binary addition -- 1 is new file, 2 is append
    }

    if (missingNeededFiles) {
      fileToDownload <- if (is.null(archive[1])) {
        neededFiles
      } else {
        result <- checkSums[checkSums$expectedFile %in% basename(archive[1]), ]$result
        missingArchive <- !isTRUE(result == "OK")
        if (missingArchive) {
          archive[1]
        } else {
          NA # means nothing to download because the archive is already in hand
        }
      }

      # The download step
      failed <- 1
      numTries <- 2

      while (failed > 0 && failed <= numTries) {
        messOrig <- character()
        warns <- character()
        withCallingHandlers({
          downloadResults <- tryCatch(
            downloadRemote(
              url = url,
              archive = archive, # both url and fileToDownload must be NULL to skip downloading
              targetFile = targetFile,
              fileToDownload = fileToDownload,
              messSkipDownload = .message$SkipDownload,
              checkSums = checkSums,
              dlFun = dlFun,
              destinationPath = destinationPath,
              overwrite = overwrite,
              needChecksums = needChecksums,
              preDigest = preDigest,
              alsoExtract = alsoExtract,
              verbose = verbose,
              .tempPath = .tempPath,
              .callingEnv = .callingEnv,
              ...
            )
            , error = function(e) {
              .downloadErrorFn(e)
            }

          )
        },
        warnings = function(w) {
          warns <<- w$message
        },
        message = function(m) {
          messOrig <<- c(messOrig, m$message)
        })
        if (isTRUE(isDirectory(url, mustExist = FALSE)) && !is(downloadResults, "try-error")) {
          fileToDownload <- downloadResults$destFile
          neededFiles <- downloadResults$destFile
        }

        if (is(downloadResults, "try-error")) {
          if (any(grepl("is required but not yet installed", messOrig)))
            failed <- numTries + 2

          downloadResults <- dlErrorHandling(failed, downloadResults, warns, messOrig, numTries, url,
                                             fileToDownload, destinationPath, targetFile, checksumFile,
                                             verbose)

          failed <- failed + 1
        } else {
          # This is so that we essentially treat it as a file, not an object, which means
          #   the second time we try this call, we can access the file locally, without needed to download
          if (is(downloadResults$out, "Spatial")) downloadResults$out <- NULL # TODO This appears to be a bug
          failed <- 0
        }
      }

      if (file.exists(checksumFile)) {
        # This is case where we didn't know what file to download, and only now
        if (is.null(fileToDownload) ||
            tryCatch(isTRUE(is.na(fileToDownload)), warning = function(x) FALSE)) {
          # do we know
          fileToDownload <- downloadResults$destFile
        }
        if (!is.null(fileToDownload)) {
          if ((length(readLines(checksumFile)) > 0)) {
            checkSums <-
              Checksums(
                files = fileToDownload,
                checksumFile = checksumFile,
                path = destinationPath,
                quickCheck = quick,
                write = FALSE,
                verbose = verbose - 1
              )
            isOK <- checkSums[checkSums$expectedFile %in% basename(fileToDownload) |
                                checkSums$actualFile %in% basename(fileToDownload), ]$result
            isOK <- isOK[!is.na(isOK)] == "OK"
            if (length(isOK) > 0) { # This is length 0 if there are no entries in the Checksums
              if (!isTRUE(all(isOK))) {
                if (purge > 0) {
                  # This is case where we didn't know what file to download, and only now
                  # do we know
                  checkSums <- .purge(
                    checkSums = checkSums,
                    purge = purge,
                    url = fileToDownload
                  )
                  downloadResults$needChecksums <- 2
                } else {
                  tf <- tryCatch(
                    makeRelative(targetFile, destinationPath) %in% fileToDownload,
                    error = function(x) {
                      FALSE
                    }
                  )
                  af <- tryCatch(
                    basename2(archive) %in% fileToDownload,
                    error = function(x) {
                      FALSE
                    }
                  )

                  sc <- sys.calls()
                  piCall <- grep("^prepInputs", sc, value = TRUE)
                  purgeTry <- if (length(piCall)) {
                    gsub(piCall,
                         pattern = ")$",
                         replacement = paste0(", purge = 7)")
                    )
                  } else {
                    ""
                  }
                  stop(
                    "\nDownloaded version of ",
                    normPath(fileToDownload),
                    " from url: ",
                    url,
                    " did not match expected file (checksums failed). There are several options:\n",
                    " 1) This may be an intermittent internet problem -- try to rerun this ",
                    "current function call.\n",
                    " 2) The local copy of the file may have been changed or corrupted -- run:\n",
                    "      file.remove('",
                    normPath(fileToDownload),
                    "')\n",
                    "      then rerun this current function call.\n",
                    if (!is.null(.getDestinationPathShared())) {
                      obj <- dir(.getDestinationPathShared(), full.names = TRUE, pattern = basename(fileToDownload))
                      if (length(obj)) {
                        paste0(" 2b) The copy of the file in getOption('reproducible.destinationPathShared')",
                               " may have been changed or corrupted -- run:\n",
                               "      file.remove(c('",
                               paste(normPath(obj), collapse = "', '"),
                               "'))\n",
                               "      then rerun this current function call.\n")
                      }

                    },
                    " 3) The download is correct, and the Checksums should be rewritten for this file:\n",
                    "      --> rerun this current function call, specifying 'purge = 7' possibly\n",
                    purgeTry,
                    " 4) manually run \nreproducible::purgeChecksums('", checksumFile, "', \n               fileToRemove = '", fileToDownload, "')",
                    "      ",
                    call. = FALSE
                  )
                }
              } else if (isTRUE(all(isOK))) {
                downloadResults$needChecksums <- 0
              }
            }
          }
        }
      } # checksum file doesn't exist
    } else {
      # not missing any files to download
      fileAlreadyDownloaded <- if (is.null(archive[1])) {
        expectedFile <- checkSums[compareNA(checkSums$result, "OK"), ]$expectedFile

        archivePossibly <- setdiff(expectedFile, neededFiles)
        archivePossibly <- .isArchive(archivePossibly)
        if (!is.null(archivePossibly)) {
          archivePossibly
        } else {
          neededFiles
        }
      } else {
        archive
      }

      downloadResults <- list(
        needChecksums = needChecksums,
        destFile = makeAbsolute(fileAlreadyDownloaded, destinationPath)
      )
      if (is.null(targetFile)) {
        messagePreProcess("Skipping download because all needed files are listed in ",
                          "CHECKSUMS.txt file and are present.",
                          " If this is not correct, rerun prepInputs with purge = TRUE",
                          verbose = verbose
        )
      } else {
        if (exists("extractedFromArchive", inherits = FALSE)) {
          messagePreProcess("Skipping download: All requested files extracted from local archive:\n    ",
                            archive,
                            verbose = verbose
          )
        } else {
          messagePreProcess("Skipping download. All requested files already present", verbose = verbose)
        }
      }
    }
    archiveReturn <- if (is.null(archive)) {
      .isArchive(downloadResults$destFile)
    } else {
      if (!file.exists(archive)) {
        if (length(.isArchive(downloadResults$destFile))) {
          hardLinkOrCopy(downloadResults$destFile, archive, verbose = verbose)
        }
      }
      archive
    }

    ## This was commented out because of LandWeb -- removed b/c of this case:
    ##  have local archive, but not yet have the targetFile
    # if (!is.null(downloadResults$destFile))
    #   neededFiles <- unique(basename(c(downloadResults$destFile, neededFiles)))
  } else {
    downloadResults <- list(needChecksums = needChecksums, destFile = NULL)
    archiveReturn <- archive
  }
  list(
    needChecksums = downloadResults$needChecksums, archive = archiveReturn,
    neededFiles = neededFiles,
    downloaded = downloadResults$destFile, checkSums = checkSums, object = downloadResults$out
  )
}

#' Download file from Google Drive
#'
#' @param url  The url (link) to the file.
#'
#' @author Eliot McIntire and Alex Chubaty
#' @keywords internal
#' @inheritParams preProcess
#' @param ... Not used here. Only used to allow other arguments to other fns to not fail.
#'
dlGoogle <- function(url, archive = NULL, targetFile = NULL,
                     checkSums, messSkipDownload, destinationPath, type = NULL,
                     overwrite, needChecksums, verbose = getOption("reproducible.verbose", 1),
                     team_drive = NULL, ...) {
  .requireNamespace("googledrive", stopOnFALSE = TRUE)

  if (missing(destinationPath)) {
    destinationPath <- tempdir2(rndstr(1, 6))
  }
  downloadFilename <- assessGoogle(
    url = url, archive = archive,
    targetFile = targetFile,
    destinationPath = destinationPath,
    verbose = verbose,
    team_drive = team_drive
  )

  destFile <- file.path(destinationPath, basename2(downloadFilename))
  if (!isTRUE(checkSums[checkSums$expectedFile == basename(destFile), ]$result == "OK")) {
    messagePreProcess("Downloading from Google Drive.", verbose = verbose)
    fs <- attr(archive, "fileSize")
    if (is.null(fs)) {
      fs <- attr(downloadFilename, "fileSize")
      if (is.null(fs)) {
        fs <- attr(assessGoogle(url, verbose = verbose, team_drive = team_drive), "fileSize")
      }
    }
    if (!is.null(fs)) {
      if (!is.numeric(fs))
        fs <- as.numeric(fs)
      class(fs) <- "object_size"
    }
    isLargeFile <- ifelse(is.null(fs), FALSE, fs > 1e6)

    # download_with_speed(url, local_path = destFile)
    if (.requireNamespace("httr2", stopOnFALSE = FALSE)) {
      downloadCall <-
        quote(
          download_resumable_httr2(
            url, local_path = destFile,
            gdriveDetails = list(id = googledrive::as_id(url),
                                 drive_resource = attr(downloadFilename, "drive_resource"))))
      # downloadCall <- quote(drive_downloadWProgress(url, local_path = destFile))
    } else {
      downloadCall <- quote(
        googledrive::drive_download(
          googledrive::as_id(url),
          path = destFile,
          type = type,
          overwrite = overwrite, verbose = TRUE)
      )
    }

    if (!isWindows() && requireNamespace("future", quietly = TRUE) && isLargeFile &&
        !isFALSE(getOption("reproducible.futurePlan"))) {
      messagePreProcess("Downloading a large file in background using future", verbose = verbose)
      message("Make sure to set\noptions(gargle_oauth_email = 'youremail@somewhere.edu')\n, and possibly ",
              "\noptions(gargle_oauth_cache = 'localPathToCache')")
      fp <- future::plan()
      if (!is(fp, getOption("reproducible.futurePlan"))) {
        fpNew <- getOption("reproducible.futurePlan")
        future::plan(fpNew, workers = 1)
        on.exit({
          future::plan(fp)
        })
      }
      b <- future::future({
        options(gargle_oauth_cache = goc,
                gargle_oauth_email = goe)
      },
      globals = list(

      ))
      a <- future::future(
        {
          # Re-authenticate inside the future worker. Prefer a service-account
          # JSON when one is configured (CI / GOOGLEDRIVE_AUTH path), otherwise
          # fall back to the user-account email + cache used by the parent.
          # Interactive auth (no email, no path) would silently hang the
          # worker — pass the JSON path through globals when available.
          if (nzchar(gauth_path) && file.exists(gauth_path)) {
            try(googledrive::drive_auth(path = gauth_path), silent = TRUE)
          } else {
            googledrive::drive_auth(email = goe, cache = goc)
          }
          retry(retries = 2, downloadCall)
        },
        globals = list(
          # Guard against NULL: getOption("gargle_oauth_cache") defaults to
          # NULL when unset, and normalizePath(NULL) errors with the cryptic
          # `path.expand(path) : invalid 'path' argument` inside the future
          # worker. Pass NULL through — gargle handles an absent cache.
          goc = local({
            v <- getOption("gargle_oauth_cache")
            if (is.null(v) || !nzchar(v)) v else normalizePath(v, mustWork = FALSE)
          }),
          goe = getOption("gargle_oauth_email"),
          gauth_path = Sys.getenv("GOOGLEDRIVE_AUTH", ""),
          downloadCall = downloadCall,
          downloadFilename = downloadFilename,
          download_resumable_httr2 = download_resumable_httr2,
          drive_download = googledrive::drive_download,
          as_id = googledrive::as_id,
          retry = retry,
          # drive_deauth = googledrive::drive_deauth,
          url = url,
          type = type,
          overwrite = overwrite,
          destFile = destFile
        )
      )
      cat("\n")
      notResolved <- TRUE
      while (notResolved) {
        Sys.sleep(0.05)
        notResolved <- !future::resolved(a)
        fsActual <- file.size(destFile)
        class(fsActual) <- "object_size"
        if (!is.na(fsActual)) {
          cat(
            format(fsActual, units = "auto"), "of", format(fs, units = "auto"),
            "downloaded         \r"
          )
        }
      }
      cat("\nDone!\n")
    } else {
      useGoogleDrive <- TRUE
      if (isTRUE(getOption("reproducible.useGdown", FALSE))) {
        messForGdownIsTRUE <- "options('reproducible.useGdown') is TRUE"
        gdown <- "gdown"
        if (nchar(Sys.which(gdown))) {
          gdownCall <- paste0(gdown, " ", googledrive::as_id(url), " -O '", destFile, "'")
          messagePreProcess("Using gdown to get files from GoogleDrive because ", messForGdownIsTRUE)

          b <- try(system(gdownCall))
          if (!is(b, "try-error")) {# likely because of authentication
            messagePreProcess(messForGdownIsTRUE, ", but the attempt failed; possibly a private url?\n",
                              url, "\nUsing googledrive package")
            useGoogleDrive <- FALSE
          }
        } else {
          messagePreProcess(messForGdownIsTRUE,
                            ", but gdown is not available at the cmd line; skipping")
        }
      }

      if (isTRUE(useGoogleDrive)) {
        a <- retry(downloadCall, retries = 2)
      }
    }
  } else {
    messagePreProcess(messSkipDownload, verbose = verbose)
    needChecksums <- 0
  }

  return(list(destFile = destFile, needChecksums = needChecksums))
}

#' Download file from generic source url
#'
#' @param url  The url (link) to the file.
#'
#' @author Eliot McIntire and Alex Chubaty
#' @keywords internal
#' @importFrom utils download.file
#' @inheritParams preProcess
dlGeneric <- function(url, destinationPath, verbose = getOption("reproducible.verbose", 1)) {
  if (missing(destinationPath)) {
    destinationPath <- tempdir2(rndstr(1, 6))
  }

  bn <- basename2(url)
  bn <- gsub("\\?|\\&", "_", bn) # causes errors with ? and maybe &
  destFile <- file.path(destinationPath, bn)

  # if (suppressWarnings(httr::http_error(url))) ## TODO: http_error is throwing warnings
  #   stop("Can not access url ", url)

  messagePreProcess("Downloading ", url, " ...", verbose = verbose)

  needDwnFl <- TRUE # this will try download.file if no httr2 or httr2 fails
  # R version 4.1.3 doesn't have httr2 that can do these steps; httr2 is too old, I believe

  if (.requireNamespace("httr2") && .requireNamespace("curl")) {
    for (i in 1:2) {
      totalTimeout <- getOption("reproducible.timeout", 1200)
      req <- httr2::request(url) |>
        httr2::req_timeout(totalTimeout) |>
        httr2::req_options(connecttimeout = totalTimeout)
      if (i == 1) # only try on first run through, in case this is the cause of failure; which it is on some sites
        req <- req |> httr2::req_user_agent(getOption("reproducible.useragent"))
      if (verbose > 0) {
        # req_progress is not in the binary httr2 available for R version 4.1.3; fails on CRAN checks
        # Also wrap in tryCatch: cli's app$styles can be NULL/NA in parallel/non-interactive contexts
        req <- tryCatch({
          reqProgress <- get("req_progress", envir = asNamespace("httr2"))
          req |> reqProgress()
        }, error = function(e) req)
      }

      resp <- req |> httr2::req_url_query() |>
        httr2::req_perform(path = destFile)
      a <- httr2::resp_body_string(resp)
      isRjcted <- grepl("Request Rejected", a)
      if (!isTRUE(any(isRjcted)) && !httr2::resp_is_error(resp)) {
        needDwnFl <- FALSE
        break
      }
    }
  } else {
    messagePreProcess("If downloads fail; please install httr2 and try again")
  }

  if (needDwnFl) {
    out <- try(download.file(url, destfile = destFile))
    if (is(out, "try-error")) {
      unlink(destFile)
      stop(.txtDownloadFailedFn("httr2"))
      # stop("Download failed; try rerunning after installing 'httr2' package.")
    }
  }

  list(destFile = destFile)
}

#' Download a remote file
#'
#' @inheritParams prepInputs
#' @inheritParams preProcess
#' @param needChecksums Logical indicating whether to generate checksums. ## TODO: add overwrite arg to the function?
#' @param messSkipDownload The character string text to pass to messaging if download skipped
#' @param checkSums TODO
#' @param fileToDownload TODO
#' @inheritParams loadFromCache
#' @inheritParams prepInputs
#' @inheritParams preProcess
#'
downloadRemote <- function(url, archive, targetFile, checkSums, dlFun = NULL,
                           fileToDownload, messSkipDownload,
                           destinationPath, overwrite, needChecksums, .tempPath, preDigest,
                           alsoExtract = "similar",
                           verbose = getOption("reproducible.verbose", 1), .callingEnv = parent.frame(),
                           ...) {
  dots <- list(...)
  # if (is.null(dots$.callingEnv)) {
  #   .callingEnv <- parent.frame()
  # } else {
  #   .callingEnv <- dots$.callingEnv
  #   dots$.callingEnv <- NULL
  # }

  noTargetFile <- is.null(targetFile) || length(targetFile) == 0
  if (missing(.tempPath)) {
    .tempPath <- tempdir2(rndstr(1, 6))
    on.exit(
      {
        unlink(.tempPath, recursive = TRUE)
      },
      add = TRUE
    )
  }

  if (!is.null(url) || !is.null(dlFun)) { # if no url, no download
    # if (!is.null(fileToDownload)  ) { # don't need to download because no url --- but need a case
    if (!isTRUE(tryCatch(is.na(fileToDownload), warning = function(x) FALSE))) {
      messagePreProcess("...downloading...", verbose = verbose)

      ## NA means archive already in hand
      out <- NULL

      if (!is.null(dlFun)) {
        dlFunName <- dlFun
        dlFunPoss <- try(.extractFunction(dlFun, envir = list2env(list(...))), silent = TRUE)
        if (is(dlFunPoss, "try-error"))
          dlFunPoss <- get0(dlFun, envir = .callingEnv)
        dlFun <- dlFunPoss
        fun <- if (is(dlFun, "call")) {
          CacheMatchedCall <- match.call(call = dlFun)
          .fnCleanup(dlFun, callingFun = "downloadRemote", CacheMatchedCall = CacheMatchedCall)
        } else {
          NULL
        }
        forms <- .argsToRemove
        overlappingForms <- fun$formalArgs[fun$formalArgs %in% forms]
        overlappingForms <- grep("\\.\\.\\.", overlappingForms, invert = TRUE, value = TRUE)

        # remove arguments that are in .argsToRemove, i.e., the sequence
        args <- if (length(overlappingForms)) {
          append(list(...), mget(overlappingForms))
        } else {
          list(...)
        }
        args <- args[!names(args) %in% forms]
        if (noTargetFile) {
          # Must mirror the `recursive = TRUE` snapshot below; otherwise files
          # that were already in subdirectories of destinationPath (e.g. files
          # extracted there by an earlier prepInputs() call into the same
          # `reproducible.inputPaths` stash) are absent from the "before" set
          # and the setdiff() at the post-dlFun snapshot incorrectly classifies
          # them as newly created. They then propagate as `downloadResults$destFile`
          # and trip "already exists at <stash path>" in the desiredPath check.
          fileInfo <- file.info(dir(destinationPath, recursive = TRUE,
                                    full.names = TRUE))
        }

        if (is.call(dlFun)) {
          out <- try(eval(dlFun, envir = .callingEnv), silent = TRUE)
          if (is(out, "try-error")) {
            sfs <- sys.frames()
            for (i in seq_along(sfs)) {
              env1 <- new.env(parent = sys.frame(-i))
              list2env(args, env1)
              out <- try(eval(dlFun, envir = env1), silent = TRUE)
              if (is.function(out)) { # in the previous "call", it may have just returned an unevaluated function
                dlFun <- out
              }
              if (!is(out, "try-error")) {
                break
              }
            }
          }
        }

        if (!is.call(dlFun) && !is.null(dlFun)) {
          out <- runDlFun(args, dlFun)
          # argsOrig <- args
          # formsDlFun <- formalArgs(dlFun)
          # argsKeep <- intersect(formsDlFun, names(args))
          # args <- args[argsKeep]
          # for (iii in 1:2) {
          #   out <- try(do.call(dlFun, args = args), silent = TRUE)
          #   if (!is(out, "try-error")) {
          #     break
          #   }
          #   args <- argsOrig
          # }

        }

        needSave <- !is.null(out) # TRUE
        if (noTargetFile) {
          # recursive gets rid of directories
          fileInfoAfter <- file.info(dir(destinationPath, recursive = TRUE, full.names = TRUE))
          possibleTargetFile <- setdiff(rownames(fileInfoAfter), rownames(fileInfo))

          possibleTargetFile <- makeAbsolute(possibleTargetFile, destinationPath)

          if (length(possibleTargetFile)) {
            destFile <- targetFile <- possibleTargetFile
            needSave <- FALSE
          } else {
            destFile <- normPath(file.path(destinationPath, basename(tempfile(fileext = ".rds"))))
          }
        } else {
          destFile <- makeAbsolute(targetFile, destinationPath)
          # destFile <- normPath(file.path(destinationPath, targetFile))
        }

        # some functions will load the object, not just download them, since we may not know
        #   where the function actually downloaded the file, we save it as an RDS file
        if (needSave) {
          if (!file.exists(destFile)) {
            out2 <- .wrap(out, preDigest = preDigest)
            saveRDS(out2, file = destFile)
          }
        }
        downloadResults <- list(out = out, destFile = normPath(destFile), needChecksums = 2)
      }

      if (is.null(out) && !is.null(url)) { # if url is NULL and out is NULL; means dlFun did all the work
        isGID <- isGoogleDriveURL(url) || isGoogleID(url)
        if (any(isGID, grepl("d.+.google.com", url))) {
          if (!requireNamespace("googledrive", quietly = TRUE)) {
            stop(.message$RequireNamespaceFn("googledrive", "to use google drive files"))
          }

          teamDrive <- getTeamDrive(dots)

          if (isGoogleDriveDirectory(url)) {
            drive_files <- googledrive::drive_ls(googledrive::as_id(url))
            if (!is.null(alsoExtract) && length(alsoExtract) > 0) {
              fileIndex <- seq_len(NROW(drive_files))
              if (length(alsoExtract) > 1)
                fileIndex <- sapply(alsoExtract, function(ae) grep(pattern = ae, drive_files$name)) |>
                  as.vector()
              else {
                if (!identical("all", alsoExtract))
                  fileIndex <- grep(pattern = alsoExtract, drive_files$name)
              }
              drive_files <- drive_files[fileIndex, ]
            }

            existingFiles <- drive_files$name %in% dir(destinationPath)
            if (any(existingFiles)) {
              messagePreProcess("Local version of files exists")
              if (isFALSE(overwrite)) {
                drive_files <- drive_files[!existingFiles, ]
                messagePreProcess("Overwrite is FALSE; only getting new ones:\n",
                                  paste0(drive_files$name, collapse = "\n"))

              }
            }

            ids <- drive_files$id
            downloadResults <- lapply(ids, function(ids)
              dlGoogle(
                url = ids, archive = archive, # targetFile = targetFile,
                checkSums = checkSums, messSkipDownload = messSkipDownload, destinationPath = .tempPath,
                overwrite = overwrite, needChecksums = needChecksums, verbose = verbose,
                team_drive = teamDrive, ...
              )
            )
            if (length(downloadResults)) {
              downloadResults <- list(destFile = vapply(downloadResults, function(x) x$destFile, FUN.VALUE = character(1)),
                                      needChecksums = max(vapply(downloadResults, function(x) x$needChecksums, FUN.VALUE = numeric(1))))
            } else {
              downloadResults <- list(destFile = character(), needChecksums = 0)
            }

          } else {
            downloadResults <- dlGoogle(
              url = url, archive = archive, targetFile = targetFile,
              checkSums = checkSums, messSkipDownload = messSkipDownload, destinationPath = .tempPath,
              overwrite = overwrite, needChecksums = needChecksums, verbose = verbose,
              team_drive = teamDrive, ...
            )
          }
        } else if (isTRUE(grepl("dl.dropbox.com", url))) {
          stop("Dropbox downloading is currently not supported")
        } else if (isTRUE(grepl("onedrive.live.com", url))) {
          stop("Onedrive downloading is currently not supported")
        } else {
          if (isTRUE(isDirectory(url, mustExist = FALSE))) { # a folder
            if (.requireNamespace("httr") && .requireNamespace("curl")) {
              list_files <- curl::new_handle()
              curl::handle_setopt(list_files, ftp_use_epsv = TRUE, dirlistonly = TRUE)
              con <- curl::curl(url = url, "r", handle = list_files)
              on.exit(close(con), add = TRUE)
              filenames <- readLines(con)
              # This is from NFI example
              filenames <- grep("href", filenames, value = TRUE)
              filenames <- grep("\\[PARENTDIR\\]|\\[ICO\\]", filenames, value = TRUE, invert = TRUE)
              filenames2 <- gsub(".+<a href=\"(.+)\">.+/a>.+", "\\1", filenames)
              # This was from mexico example from Steve
              # filenames3 <- gsub(".+<a.+\">(.+)</a>.+", "\\1", filenames)
              # rm http tags, plus the two files Description and Parent Directory that are in a directory
              filenames <- grep("<|>|Description|Parent Directory", filenames2, value = TRUE, invert = TRUE)
              if (isTRUE(nzchar(alsoExtract))) {
                if (grepl("^sim", alsoExtract)) {
                  theGrep <- filePathSansExt(targetFile)
                } else if (grepl("none", alsoExtract)) {
                  theGrep <- paste0("^", targetFile, "$")
                } else {
                  theGrep <- paste(alsoExtract, collapse = "|")
                }
                filenames <- grep(theGrep, filenames, value = TRUE)
              }
              # now that we have filenames; need to checksum
              urls <- file.path(url, filenames)

              checkSums <- runChecksums(destinationPath, checkSumFilePath = destinationPath, filenames, verbose)
              checkSums <- checkSums$checkSums[expectedFile %in% filenames]
              checkSums <- checkSums[data.table(expectedFile = basename2(filenames)), on = "expectedFile"]
              missingNeededFiles <- missingFiles(filenames, checkSums, destinationPath)
              stillNeed <- !checkSums$result %in% "OK"

              downloadResults <- list(destFile = character())
              if (missingNeededFiles) {
                stillNeedFile <- match(basename2(urls), checkSums$expectedFile[stillNeed])
                messagePrepInputs("url was supplied as a directory; downloading\n",
                                         paste(urls[stillNeed], collapse = "\n"),
                                  verbose = verbose)

                downloadResults <- vapply(urls[stillNeedFile], function(url)
                  dlGeneric(url, destinationPath = .tempPath, verbose = verbose) |> unlist(),
                  FUN.VALUE = character(1))
                # named list of local filenames; named with urls
                downloadResults <- list(destFile = downloadResults)
              }
              if (any(!stillNeed)) {
                filenamesAlreadyHave <- makeAbsolute(checkSums$expectedFile[stillNeed %in% FALSE], destinationPath)
                alreadyHave <- match(checkSums$expectedFile[stillNeed %in% FALSE], basename2(urls))
                names(filenamesAlreadyHave) <-  urls[alreadyHave]
                # downloadResults$destFile <- c(downloadResults$destFile, filenamesAlreadyHave)
              }

            } else {
              stop("url is a directory; need to install.packages(c('httr', 'curl'))")
            }
          } else {

            downloadResults <- dlGeneric(url = url, destinationPath = .tempPath, verbose = verbose)
          }
          downloadResults$needChecksums <- needChecksums
        }
      }
      # if destinationPath is tempdir, then don't copy and remove

      testFTD <- length(fileToDownload) > 0
      if (isTRUE(testFTD)) testFTD <- isTRUE(all(!downloadResults$destFile %in% fileToDownload))

      # Don't use .tempPath directly because of non-google approaches too

      if (!(identical(
        unique(dirname(normPath(downloadResults$destFile))),
        normPath(as.character(destinationPath))
      )) || testFTD) {
        # basename2 is OK because the destFile will be flat; it is just archive extraction that needs to allow nesting
        desiredPath <- makeAbsolute(basename2(downloadResults$destFile), destinationPath)
        desiredPathExists <- file.exists(desiredPath)
        if (any(desiredPathExists) && !isTRUE(overwrite)) {
          stopMess <- paste(desiredPath, " already exists and overwrite = FALSE; would you like to overwrite anyway? Y or N:  ")
          if (interactive()) {
            interactiveRes <- readline(stopMess)
            if (startsWith(tolower(interactiveRes), "y")) {
              overwrite <- TRUE
            }
          }
          if (!identical(overwrite, TRUE)) {
            stop(targetFile, " already exists at ", desiredPath, ". Use overwrite = TRUE?")
          }
        }

        # Try hard link first -- the only type that R deeply recognizes
        # if that fails, fall back to copying the file.
        # NOTE: never use symlink because the original will be deleted.
        result <- hardLinkOrCopy(downloadResults$destFile, desiredPath, verbose = verbose - 3)

        # result <- suppressWarningsSpecific(
        #   file.link(downloadResults$destFile, desiredPath),
        #   falseWarnings = "already exists|Invalid cross-device")
        # # result <- suppressWarnings(
        # #   file.link(downloadResults$destFile, desiredPath)
        # # )
        #
        # if (isFALSE(result)) {
        #   result <- file.copy(downloadResults$destFile, desiredPath)
        # }

        tmpFile <- makeRelative(downloadResults$destFile, dirname(downloadResults$destFile))
        downloadResults$destFile <- makeAbsolute(tmpFile, destinationPath)
        # downloadResults$destFile <- file.path(destinationPath, basename(downloadResults$destFile))
      }
      # }
    } else {
      messagePreProcess(messSkipDownload, verbose = verbose)
      downloadResults <- list(needChecksums = 0, destFile = NULL)
    }
  } else {
    messagePreProcess("No downloading; no url", verbose = verbose)
  }
  # clean up from "directory" downloads
  if (exists("filenamesAlreadyHave", inherits = FALSE)) {
    downloadResults$destFile <- c(downloadResults$destFile, filenamesAlreadyHave)
  }
  downloadResults
}

missingFiles <- function(files, checkSums, destinationPath) {
  filesBasename <- makeRelative(files, destinationPath)
  if (is.null(files)) {
    result <- unique(checkSums$result)
  } else {
    result <- checkSums[checkSums$expectedFile %in% filesBasename, ]$result
  }
  if (length(result) == 0) result <- NA

  (!(all(compareNA(result, "OK")) && all(filesBasename %in% checkSums$expectedFile)) ||
      is.null(files))
}

assessGoogle <- function(url, archive = NULL, targetFile = NULL,
                         destinationPath = getOption("reproducible.destinationPath", "."),
                         verbose = getOption("reproducible.verbose", 1),
                         team_drive = NULL) {
  if (!requireNamespace("googledrive", quietly = TRUE)) {
    stop(.message$RequireNamespaceFn("googledrive", "to use google drive files"))
  }
  if (.isRstudioServer()) {
    .requireNamespace("httr", stopOnFALSE = TRUE)
    opts <- options(httr_oob_default = TRUE)
    on.exit(options(opts))
  }

  # Cache the drive_get / drive_ls result indefinitely. The Cache key
  # includes the URL/ID, so each distinct file pays one API hit ever per
  # cachePath; subsequent calls (in-memory or on-disk) are near-instant.
  # Without this, .guessAtFile (and therefore the very first phase of
  # pp_resolve_files) hits the Google Drive API on every call, ~1-2 s of
  # latency that the sidecar fast-path further down the pipeline cannot
  # recover.
  #
  # Do NOT wrap the inner call in `quote(...)`: Cache's digest treats a
  # `quote(...)` expression as opaque text and never resolves the `url`
  # symbol against the calling frame, so every URL collides on the same
  # cache key and the first-cached fileAttr is returned for every later
  # URL. retry() captures its `expr` argument with substitute() and works
  # the same way whether or not it's wrapped in quote().
  # if (is.null(archive) || is.na(archive)) {
  if (isTRUE(isDirectory(url, FALSE))) {
    fileAttr <- Cache(
      retry(retries = 1, googledrive::drive_ls(googledrive::as_id(url),
                                               shared_drive = team_drive)),
      verbose = FALSE
    )
  } else {
    if (packageVersion("googledrive") < "2.0.0") {
      fileAttr <- Cache(
        retry(retries = 1, googledrive::drive_get(googledrive::as_id(url),
                                                  team_drive = team_drive)),
        verbose = FALSE
      )
    } else {
      fileAttr <- Cache(
        retry(retries = 1, googledrive::drive_get(googledrive::as_id(url),
                                                  shared_drive = team_drive)),
        verbose = FALSE
      )
    }
  }

  fileSize <- sapply(fileAttr$drive_resource, function(x) x$size)
  if (!is.null(unlist(fileSize))) {
    messageAboutFilesize(fileSize, verbose)
    # fileSize <- as.numeric(fileSize)
    # len <- length(fileSize)
    # if (len > 1)
    #   fileSize <- sum(fileSize)
    # class(fileSize) <- "object_size"
    # Fils <- singularPlural(c("File", "Files"), v = len)
    # isAre <- isAre(v = len)
    # messagePreProcess(Fils, " on Google Drive ", isAre, " ", format(fileSize, units = "auto"),
    #                   verbose = verbose
    # )
  }
  archive <- .isArchive(fileAttr$name)
  if (is.null(archive)) {
    if (is.null(targetFile)) {
      # make the guess
      targetFile <- fileAttr$name
    }
    downloadFilename <- targetFile # override if the targetFile is not an archive
  } else {
    archive <- file.path(destinationPath, basename2(archive))
    downloadFilename <- archive
  }
  attr(downloadFilename, "drive_resource") <- fileAttr$drive_resource
  # } else {
  #   downloadFilename <- archive
  # }
  if (exists("fileSize", inherits = FALSE)) {
    attr(downloadFilename, "fileSize") <- fileSize
  }

  return(downloadFilename)
}

.isRstudioServer <- function() {
  isRstudioServer <- FALSE

  if (isTRUE("tools:rstudio" %in% search())) { ## running in Rstudio
    rsAPIFn <- get(".rs.api.versionInfo", as.environment("tools:rstudio"))
    versionInfo <- rsAPIFn()
    if (!is.null(versionInfo)) {
      isRstudioServer <- identical("server", versionInfo$mode)
    }
  }
  isRstudioServer
}

SSL_REVOKE_BEST_EFFORT <- function(envir = parent.frame(1)) {
  # Take from https://github.com/rstudio/rstudio/issues/10163#issuecomment-1193316767 #
  prevCurlVal <- Sys.getenv("R_LIBCURL_SSL_REVOKE_BEST_EFFORT")
  Sys.setenv(R_LIBCURL_SSL_REVOKE_BEST_EFFORT = TRUE)
  on.exit2({#withr::defer({
    if (nzchar(prevCurlVal))
      Sys.setenv(R_LIBCURL_SSL_REVOKE_BEST_EFFORT = prevCurlVal)
    else
      Sys.unsetenv("R_LIBCURL_SSL_REVOKE_BEST_EFFORT")
  }, envir = envir)
}

on.exit2 <- function(expr, envir = sys.frame(-2), add = TRUE, after = TRUE) {
  funExpr <- as.call(list(function() expr))
  do.call(base::on.exit, list(funExpr, add, after), envir = envir)
}

dlErrorHandling <- function(failed, downloadResults, warns, messOrig, numTries, url,
                            fileToDownload, destinationPath, targetFile, checksumFile, verbose) {
  if (isTRUE(grepl(paste("already exists", .txtDownloadFailedFn(".+"), sep = "|"), downloadResults))) {
    stop(downloadResults)
  }

  if (isTRUE(grepl("already exists", downloadResults))) {
    stop(downloadResults)
  }

  SSLwarns <- grepl(.txtUnableToAccessIndex, warns)
  SSLwarns2 <- grepl("SSL peer certificate or SSH remote key was not OK", messOrig)
  if (any(SSLwarns) || any(SSLwarns2)) {
    SSL_REVOKE_BEST_EFFORT()
  }

  if (failed >= numTries) {
    isGID <- all(grepl("^[A-Za-z0-9_-]{33}$", url), # Has 33 characters as letters, numbers or - or _
                 !grepl("\\.[^\\.]+$", url)) # doesn't have an extension
    if (isGID) {
      urlMessage <- googledriveIDtoHumanURL(url)
      # urlMessage <- paste0("https://drive.google.com/file/d/", url)
    } else {
      urlMessage <- url
    }
    messCommon <- paste0(
      "Download of ", url, " failed. This may be a permissions issue. ",
      "Please check the url and permissions are correct.\n",
      "If the url is correct, it is possible that manually downloading it will work. ",
      "To try this, with your browser, go to\n",
      urlMessage, ",\n ... then download it manually, give it this name: '", fileToDownload,
      "', and place file here: ", destinationPath
    )
    if (isInteractive() && getOption("reproducible.interactiveOnDownloadFail", TRUE)) {
      mess <- paste0(
        messCommon,
        ".\n ------- \nIf you have completed a manual download, press 'y' to continue; otherwise press any other key to stop now. ",
        "\n(To prevent this behaviour in the future, set options('reproducible.interactiveOnDownloadFail' = FALSE)  )"
      )
      if (failed == numTries + 2) {
        stop(paste(messOrig, collapse = "\n"))
      } else {
        messagePreProcess(mess, verbose = verbose + 1)
      }
      resultOfPrompt <- .readline("Type y if you have attempted a manual download and put it in the correct place: ")
      resultOfPrompt <- tolower(resultOfPrompt)
      if (!identical(resultOfPrompt, "y")) {
        stop(downloadResults, "\n", messOrig, "\nDownload failed")
      }
      downloadResults <- list(
        destFile = file.path(destinationPath, targetFile),
        needChecksums = 2
      )
    } else {
      message(downloadResults)
      stop(
        downloadResults, "\n", messOrig, "\n", messCommon, ".\n-------------------\n",
        "If manual download was successful, you will likely also need to run Checksums",
        " manually after you download the file with this command: ",
        "reproducible:::appendChecksumsTable(checkSumFilePath = '", checksumFile, "', filesToChecksum = '", targetFile,
        "', destinationPath = '", dirname(checksumFile), "', append = TRUE)"
      )
    }
  } else {
    if (failed > 1) Sys.sleep(0.5) else SSL_REVOKE_BEST_EFFORT() # uses withr::defer to remove it after this test
  }

  # # ELIOT removed this as httr is being deprecated --> the above chunk should work
  # # if (any(grepl("SSL peer certificate or SSH remote key was not OK", messOrig))) {
  # #   # THIS IS A MAJOR WORK AROUND FOR SSL ISSUES IN SOME WORK ENVIRONMENTS. NOT ADVERTISED.
  # #   # https://stackoverflow.com/questions/46331066/quantmod-ssl-unable-to-get-local-issuer-certificate-in-r
  # #   if (isFALSE(as.logical(Sys.getenv("REPRODUCIBLE_SSL_VERIFYPEER")))) {
  # #     .requireNamespace("httr", stopOnFALSE = TRUE)
  # #     message(
  # #       "Temporarily setting ssl_verifypeer to FALSE because ",
  # #       "'SSL peer certificate or SSH remote key was not OK'"
  # #     )
  # #     sslOrig <- httr::set_config(httr::config(ssl_verifypeer = FALSE))
  # #     on.exit(httr::set_config(sslOrig), add = TRUE)
  # #   }
  # # }
  #
  # # if (any(grepl("is required but not yet installed", messOrig))) {
  # #   failed <- numTries + 2
  # # }
  # if (failed >= numTries) {
  #   isGID <- all(grepl("^[A-Za-z0-9_-]{33}$", url), # Has 33 characters as letters, numbers or - or _
  #                !grepl("\\.[^\\.]+$", url)) # doesn't have an extension
  #   if (isGID) {
  #     urlMessage <- paste0("https://drive.google.com/file/d/", url)
  #   } else {
  #     urlMessage <- url
  #   }
  #   messCommon <- paste0(
  #     "Download of ", url, " failed. This may be a permissions issue. ",
  #     "Please check the url and permissions are correct.\n",
  #     "If the url is correct, it is possible that manually downloading it will work. ",
  #     "To try this, with your browser, go to\n",
  #     urlMessage, ",\n ... then download it manually, give it this name: '", fileToDownload,
  #     "', and place file here: ", destinationPath
  #   )
  #   if (isInteractive() && getOption("reproducible.interactiveOnDownloadFail", TRUE)) {
  #     mess <- paste0(
  #       messCommon,
  #       ".\n ------- \nIf you have completed a manual download, press 'y' to continue; otherwise press any other key to stop now. ",
  #       "\n(To prevent this behaviour in the future, set options('reproducible.interactiveOnDownloadFail' = FALSE)  )"
  #     )
  #     if (failed == numTries + 2) {
  #       stop(paste(messOrig, collapse = "\n"))
  #     } else {
  #       messagePreProcess(mess, verbose = verbose + 1)
  #     }
  #     resultOfPrompt <- .readline("Type y if you have attempted a manual download and put it in the correct place: ")
  #     resultOfPrompt <- tolower(resultOfPrompt)
  #     if (!identical(resultOfPrompt, "y")) {
  #       stop(downloadResults, "\n", messOrig, "\nDownload failed")
  #     }
  #     downloadResults <- list(
  #       destFile = file.path(destinationPath, targetFile),
  #       needChecksums = 2
  #     )
  #   } else {
  #     message(downloadResults)
  #     stop(
  #       downloadResults, "\n", messOrig, "\n", messCommon, ".\n-------------------\n",
  #       "If manual download was successful, you will likely also need to run Checksums",
  #       " manually after you download the file with this command: ",
  #       "reproducible:::appendChecksumsTable(checkSumFilePath = '", checksumFile, "', filesToChecksum = '", targetFile,
  #       "', destinationPath = '", dirname(checksumFile), "', append = TRUE)"
  #     )
  #   }
  # } else {
  #   Sys.sleep(0.5)
  # }
  downloadResults
}

.downloadErrorFn <- function(xxxx) {
  try(stop(xxxx))
}

runDlFun <- function(args, dlFun) {
  argsOrig <- args
  formsDlFun <- formalArgs(dlFun)
  argsKeep <- intersect(formsDlFun, names(args))
  args <- args[argsKeep]
  for (iii in 1:2) {
    out <- try(do.call(dlFun, args = args), silent = TRUE)
    if (!is(out, "try-error")) {
      break
    }
    args <- argsOrig
  }
  out
}

#' Purge the checksums of a single file
#'
#' This is a manual way of achieving `prepInputs(..., purge = 7)`, useful in cases
#' where `prepInputs` is not called directly by the user, so it would be difficult
#' to set `purge = 7`.
#' @inheritParams downloadFile
#' @param fileToRemove The filename to remove from the `checksumFile`
#'
#' @export
#' @return NULL. Run for its side effect, namely, and file removed from the \file{CHECKSUMS.txt}
#'   file.
purgeChecksums <- function(checksumFile, fileToRemove) {
  dt <- data.table::fread(checksumFile)
  toPurge <- dt[file %in% fileToRemove]
  dtNew <- dt[!toPurge, on = c("file", "checksum")]
  data.table::fwrite(dtNew, file = checksumFile)
}

download_resumable_httr2 <- function(file_name, local_path, gdriveDetails, fileSize = NULL,
                                     verbose = getOption("reproducible.verbose")) {
  .requireNamespace("googledrive", stopOnFALSE = TRUE)

  ## Normalize path to avoid issues with ~
  local_path_expanded <- normalizePath(local_path, mustWork = FALSE)

  if (missing(gdriveDetails)) {
    isGD <- isGoogleDriveURL(file_name) || inherits(file_name, "drive_id")

    completed <- FALSE
    if (isGD) {
      gdriveDetails <- googledrive::drive_get(file_name)
    }
  } else {
    isGD <- TRUE
  }

  if (isGD) {
    file_id <- gdriveDetails$id
    fileSize <- as.numeric(gdriveDetails$drive_resource[[1]]$size)
    file_name <- googledriveIDtoDownloadURL(file_id)
    bearer <- .get_fresh_gd_bearer()  # refresh token via gargle before use
  } else {
    if (is.null(fileSize)) {
      fileSize <- getRemoteFileSize(isGD, url)
    }
  }

  if ( (isGD &&  (.Platform$OS.type == "windows")) || nzchar(Sys.which("curl")) %in% FALSE ||
      fileSize < 1e9) { # i.e., < 1GB can just use the simpler httr2 progress
    ## Google Drive download using httr2 (no resume support)
    ## Retry once on 401: the token may have expired mid-session; refresh and retry.
    con <- file(local_path_expanded, open = "wb")
    on.exit(try(close(con), silent = TRUE), add = TRUE)

    for (.attempt in 1:2) {
      req <- httr2::request(file_name)
      if (isGD) req <- req |> httr2::req_auth_bearer_token(bearer)
      req <- tryCatch(req |> httr2::req_progress(), error = function(e) req)

      err <- tryCatch({
        resp <- httr2::req_perform(req)
        body <- httr2::resp_body_raw(resp)
        writeBin(body, con)
        completed <- TRUE
        NULL
      }, error = function(e) e)

      if (is.null(err)) break  # success

      is401 <- inherits(err, "httr2_http_401") ||
        grepl("401", conditionMessage(err), fixed = TRUE)
      if (.attempt < 2L && is401 && isGD) {
        messagePreProcess("Google Drive token expired (HTTP 401); refreshing and retrying...",
                          verbose = verbose)
        bearer <- .get_fresh_gd_bearer(force_refresh = TRUE)
      } else {
        stop("Google Drive download failed: ", conditionMessage(err))
      }
    }

  } else {
    if (.Platform$OS.type != "windows" && nzchar(Sys.which("curl"))) {
      # Use download.file with curl on Linux/macOS
      method <- "curl"
      if (!isGD) {
        extra_args <- "-C -"
        messagePreProcess("Using 'curl' with resume support on Linux/macOS.", verbose = verbose)
      }

      for (.attempt in 1:2) {
        if (isGD)
          extra_args <- paste("-L -H", shQuote(paste("Authorization: Bearer", bearer)))
        err <- tryCatch({
          utils::download.file(
            url = file_name,
            destfile = local_path_expanded,
            method = method,
            quiet = verbose < 1,
            extra = extra_args
          )
          completed <- TRUE
          NULL
        }, error = function(e) e)

        if (is.null(err)) break

        is401 <- grepl("401", conditionMessage(err), fixed = TRUE)
        if (.attempt < 2L && is401 && isGD) {
          messagePreProcess("Google Drive token expired (HTTP 401); refreshing and retrying...",
                            verbose = verbose)
          bearer <- .get_fresh_gd_bearer(force_refresh = TRUE)
        } else {
          stop("Non-Google Drive download failed: ", conditionMessage(err))
        }
      }

    } # else {
      # # Use httr2 for non-Google Drive downloads on Windows or if curl is unavailable
      # downloaded_bytes <- if (file.exists(local_path_expanded)) file.info(local_path_expanded)$size else 0
      #
      # # Try to get total size
      # head_resp <- try(httr::HEAD(file_name), silent = TRUE)
      # total_size <- if (inherits(head_resp, "response")) {
      #   as.numeric(httr::headers(head_resp)[["content-length"]])
      # } else {
      #   NA
      # }
      #
      # req <- httr2::request(file_name)
      #
      # if (!is.na(total_size) && total_size > downloaded_bytes) {
      #   req <- req |>
      #     httr2::req_headers(Range = paste0("bytes=", downloaded_bytes, "-"))
      # }
      #
      # req <- req |> httr2::req_progress()
      #
      # con <- file(local_path_expanded, open = if (downloaded_bytes > 0) "ab" else "wb")
      # on.exit(try(close(con), silent = TRUE), add = TRUE)
      #
      # tryCatch({
      #   resp <- httr2::req_perform(req)
      #   body <- httr2::resp_body_raw(resp)
      #   writeBin(body, con)
      #   completed <- TRUE
      #   # message("Non-Google Drive download completed using httr2.")
      # }, error = function(e) {
      #   stop("Download failed: ", e$message)
      # })
    # }
  }

  if (isTRUE(completed)) {
    messagePreProcess("Download of " , local_path, " complete",  verbose = verbose)
  }
}

messageAboutFilesize <- function(fileSize, verbose, msgMiddle = " on Google Drive ") {
  fileSize <- as.numeric(fileSize)
  len <- length(fileSize)
  if (len > 1)
    fileSize <- sum(fileSize)
  class(fileSize) <- "object_size"
  Fils <- singularPlural(c("File", "Files"), v = len)
  isAre <- isAre(v = len)
  messagePreProcess(Fils, msgMiddle, isAre, " ", format(fileSize, units = "auto"),
                    verbose = verbose
  )
}

googledriveIDtoDownloadURL <- function(id) {
  paste0("https://www.googleapis.com/drive/v3/files/", id, "?alt=media")
}

## Fetch a Google Drive bearer token.  When force_refresh = TRUE, trigger
## gargle's auto-refresh by making a lightweight drive_user() API call through
## httr — this is the safe way to refresh without calling Token2.0$refresh()
## directly (which writes to the gargle cache and can fail on read-only mounts).
.get_fresh_gd_bearer <- function(force_refresh = FALSE) {
  if (!googledrive::drive_has_token())
    stop("no googledrive token discovered; run drive_auth() to authenticate.")
  if (force_refresh)
    tryCatch(googledrive::drive_user(), error = function(e) NULL)
  googledrive::drive_token()$auth_token$credentials$access_token
}

googledriveIDtoHumanURL <- function(id) {
  paste0("https://drive.google.com/file/d/", id)
}


getRemoteFileSize <- function(isGD, url) {
  if (isGD) {
    file <- googledrive::drive_get(url)
    file_id <- file$id
    download_url <- googledriveIDtoDownloadURL(file_id)
    total_size <- as.numeric(file$drive_resource[[1]]$size)
  } else {
    download_url <- url
    head_resp <- httr::HEAD(download_url)
    total_size <- as.numeric(httr::headers(head_resp)[["content-length"]])
  }
  total_size
}

test_that("dlGeneric works", {
  skip_on_cran()
  skip_if_not_installed("httr")
  testInit(needInternet = TRUE)
  url <- "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip"
  noisyOutput <- capture.output({
    res <- skip_on_transient_http(dlGeneric(url, tempdir2(rndstr(1, 6))))
  })
  expect_true(file.exists(res$destFile))
  unlink(res$destFile)
})

test_that("prepInputs reads Google Drive spreadsheets", {
  skip_on_cran()
  skip_if_not_installed("googledrive")
  skip_if_service_account_releaseVer_NotLinux()
  testInit(needGoogleDriveAuth = TRUE)
  userDist <- skip_on_transient_http(prepInputs(
    url = "https://docs.google.com/spreadsheets/d/1fOikb83aOuLlFYIn6pjmC7Jydjcy77TH",
    targetFile = "userDist.csv",
    destinationPath = tempdir(),
    type = "spreadsheet",
    fun = "data.table::fread"
  ))
  expect_true(is(userDist, "data.table"))
})

#' `reproducible` options
#'
#' These provide top-level, powerful settings for a comprehensive
#' reproducible workflow. To see defaults, run `reproducibleOptions()`.
#' See Details below.
#'
#' @export
#' @return
#' This function returns a list of all the options that the `reproducible` package
#' sets and uses. See below for details of each.
#'
#' @details
#'
#' Below are options that can be set with `options("reproducible.xxx" = newValue)`,
#' where `xxx` is one of the values below, and `newValue` is a new value to
#' give the option. Sometimes these options can be placed in the user's `.Rprofile`
#' file so they persist between sessions.
#'
#' The following options are likely of interest to most users:
#' \describe{
#'   \item{`ask`}{
#'     Default: `TRUE`. Used in [clearCache()] and [keepCache()].
#'   }
#'   \item{`cacheChaining`}{
#'     Default: `FALSE`. Used in [Cache()] in the `.cacheChaining` argument.
#'   }
#'   \item{`cachePath`}{
#'     Default: `.reproducibleTempCacheDir`. Used in [Cache()] and many others.
#'     The default path for repositories if not passed as an argument.
#'   }
#'   \item{`cacheSaveFormat`}{
#'     Default: `"rds"`. What save format to use; currently, `"qs"` (which will use
#'     `qs2` package as of `reproducible` version ">= 2.1.3"), `"qs2"`, or `"rds"`.
#'   }
#'   \item{`cacheSpeed`}{
#'     Default `"slow"`. One of `"slow"` or `"fast"` (1 or 2).
#'     `"slow"` uses `digest::digest` internally, which is transferable across operating
#'     systems, but much slower than `digest::digest(algo = "spooky)`.
#'     So, if all caching is happening on a single machine, `"fast"` would be a good setting.
#'   }
#'   \item{`checkRemoteHash`}{
#'     Default: `FALSE`. Used in [preProcess()] / [prepInputs()]. Controls whether
#'     `pp_remote_hash_check` re-contacts the remote source (e.g. Google Drive,
#'     HTTP HEAD) when a `.hash` sidecar from a previous successful match
#'     already exists in `destinationPath`. With the default (`FALSE`), the
#'     sidecar is trusted and the remote check is skipped — typically saving
#'     1–2 s per file when the cluster cache is warm. Set to `TRUE` to force a
#'     remote round-trip on every call (the pre-3.0.0.9050 behaviour); use
#'     this if the upstream file may change and you need to detect that.
#'     Removing the `<file>_*.hash` sidecar also forces a re-check.
#'   }
#'   \item{`conn`}{
#'     Default: `NULL`. Sets a specific connection to a database, e.g.,
#'     `dbConnect(drv = RSQLite::SQLite())` or `dbConnect(drv = RPostgres::Postgres()`.
#'     For remote database servers, setting one connection may be far faster than using
#'     `drv` which must make a new connection every time.
#'   }
#'   \item{`destinationPath`}{
#'     Default: `NULL`. Used in [prepInputs()] and [preProcess()].
#'     Can be set globally here.
#'   }
#'   \item{`drv`}{
#'     Default: `RSQLite::SQLite()`. Sets the default driver for the backend database system.
#'     Only tested with `RSQLite::SQLite()` and `RPostgres::Postgres()`.
#'   }
#'   \item{dryRun}{
#'     Default: `FALSE`.
#'   }
#'   \item{`futurePlan`}{
#'     Default: `FALSE`. On Linux OSes, `Cache` and `cloudCache` have some
#'     functionality that uses the `future` package.
#'     Default is to not use these, as they are experimental.
#'     They may, however, be very effective in speeding up some things, specifically,
#'     uploading cached elements via `googledrive` in `cloudCache`.
#'   }
#'   \item{`gdalwarp`}{
#'     **Deprecated — do not use.** Default: `FALSE`. This option previously
#'     switched `postProcessTo` to use `sf::gdal_utils("warp")` for a specific
#'     combination of raster/vector inputs. It is no longer needed: current
#'     versions of `terra` handle this case well and produce equivalent results
#'     without the GDAL detour. The option is retained only for backwards
#'     compatibility and will be removed in a future release.
#'   }
#'   \item{`gdalwarpThreads`}{
#'     **Deprecated — do not use** (see `gdalwarp` above). Default: `2`.
#'     Previously set `-wo NUM_THREADS=` for `gdalProject`.
#'   }
#'   \item{`destinationPathShared`}{
#'     Default: `NULL`. Used in [prepInputs()] and [preProcess()].
#'     If set to a path, this will cause these functions to save their downloaded and preprocessed
#'     file to this location, with a hardlink (via `file.link`) to the file created in the
#'     `destinationPath`.
#'     This can be used so that individual projects that use common data sets can maintain
#'     modularity (by placing downloaded objects in their `destinationPath`, but also minimize
#'     re-downloading the same (perhaps large) file over and over for each project.
#'     Because the files are hardlinks, there is no extra space taken up by the apparently
#'     duplicated files.
#'
#'     **Note:** the previous name for this option was `reproducible.inputPaths`; the old
#'     name is still accepted and will continue to work, but `reproducible.destinationPathShared` is
#'     preferred going forward (it matches the [prepInputs()] naming family).
#'   }
#'   \item{`destinationPathSharedRecursive`}{
#'     Default: `FALSE`. Used in [prepInputs()] and [preProcess()].
#'     Should `reproducible.destinationPathShared` be searched recursively for existence of a file?
#'
#'     **Note:** the previous name for this option was `reproducible.inputPathsRecursive`;
#'     the old name is still accepted but `reproducible.destinationPathSharedRecursive` is preferred.
#'   }
#'   \item{`inputPaths`}{
#'     **Deprecated** — use `reproducible.destinationPathShared` instead.
#'     Retained for backwards compatibility; if set and `reproducible.destinationPathShared` is `NULL`,
#'     the value of `reproducible.inputPaths` is used automatically.
#'   }
#'   \item{`inputPathsRecursive`}{
#'     **Deprecated** — use `reproducible.destinationPathSharedRecursive` instead.
#'     Retained for backwards compatibility.
#'   }
#'   \item{`leaveOnDisk`}{
#'     Default: `TRUE`. Used in [postProcess()].
#'     When there is a `SpatRaster` object, should `postProcess` force any file-backed object,
#'     to use the file-based, memory-safe tools within `terra` (by temporarily setting
#'     `terraOption(memfrac = 0)`. Alternatively, if this is set to `FALSE`,
#'     then `postProcess` will let `terra` decide on its own based on its internal
#'     cues (largely based on `memfrac`, `maxmem` `terraOptions`). This will be ignored,
#'     however, if the user has set the `terraOptions` away from its default of `0.5`. The default
#'     increases predictability of whether the returned object is on disk or in memory.
#'   }
#'   \item{`memoisePersist`}{
#'     Default: `FALSE`. Used in [Cache()].
#'     Should the memoised copy of the Cache objects persist even if `reproducible` reloads
#'     e.g., via `devtools::load_all`? This is mostly useful for developers of
#'     `reproducible`. If `TRUE`, a object named `paste0(".reproducibleMemoise_", cachePath)`
#'     will be placed in the `.GlobalEnv`, i.e., one for each `cachePath`.
#'   }
#'   \item{`nThreads`}{
#'     Default: `1`. The number of threads to use for reading/writing cache files.
#'   }
#'   \item{`objSize`}{
#'     Default: `TRUE`. Logical. If `TRUE`, then object sizes will be included in
#'     the cache database. Simplying calculating object size of large objects can
#'     be time consuming, so setting this to `FALSE` will make caching up to 10%
#'     faster, depending on the objects.
#'   }
#'   \item{`overwrite`}{
#'     Default: `FALSE`. Used in [prepInputs()], [preProcess()],
#'     [downloadFile()], and [postProcess()].
#'   }
#'   \item{`quick`}{
#'     Default: `FALSE`. Used in [Cache()]. This will cause `Cache` to use
#'     `file.size(file)` instead of the `digest::digest(file)`.
#'     Less robust to changes, but faster. *NOTE: this will only affect objects on disk*.
#'   }
#'   \item{`rasterRead`}{
#'     Used during `prepInputs` when reading `.tif`, `.grd`, and `.asc` files.
#'     Default: `terra::rast`. Can be `raster::raster` for backwards compatibility.
#'     Can be set using environment variable `R_REPRODUCIBLE_RASTER_READ`.
#'   }
#'   \item{`shapefileRead`}{
#'     Default `NULL`. Used during `prepInputs` when reading a `.shp` file.
#'     If `NULL`, it will use `sf::st_read` if `sf` package is available; otherwise,
#'     it will use `raster::shapefile`
#'   }
#'   \item{`showSimilar`}{
#'     Default `FALSE`. Passed to `Cache`.
#'   }
#'   \item{`testCharacterAsFile`}{
#'     Default `FALSE`. The behaviour of `.robustDigest` on `character` vectors prior to
#'     `reproducible == 2.1.2` was that the function would test for whether they were
#'     filenames by using `file.exists`. If it was a filename, then it would digest
#'     the file content. In cases of a character vector or a data.frame of "filenames",
#'     this could cause long hanging of the R system as it tries to digest the file
#'     contents of potentially many files. This behaviour is not transparent to a user.
#'     Now the default is to **not** digest the file content of a `character` vector
#'     even if they are filenames. To force file content digesting, then convert to
#'     either `asPath` or `fs::as_fs_path`. Or set this option to `TRUE` and the previous
#'     behaviour will return, where it tries to guess whether a character vector
#'     is filenames or not, and if it is, then digest the file content.
#'   }
#'   \item{`timeout`}{
#'     Default `1200`. Used in `preProcess` when downloading occurs. If a user has `R.utils`
#'     package installed, `R.utils::withTimeout(  , timeout = getOption("reproducible.timeout"))`
#'     will be wrapped around the download so that it will timeout (and error) after this many
#'     seconds.
#'   }
#'   \item{`useCache`}{
#'     Default: `TRUE`. Used in [Cache()]. If `FALSE`, then the entire
#'     `Cache` machinery is skipped and the functions are run as if there was no Cache occurring.
#'     Can also take 2 other values: `'overwrite'` and `'devMode'`.
#'     `'overwrite'` will cause no recovery of objects from the cache repository, only new
#'     ones will be created. If the hash is identical to a previous one, then this will overwrite
#'     the previous one.
#'     `'devMode'` will function as normally `Cache` except it will use the
#'     `userTags` to determine if a previous function has been run. If the `userTags`
#'     are identical, but the digest value is different, the old value will be deleted from the
#'     cache repository and this new value will be added.
#'     This addresses a common situation during the development stage: functions are changing
#'     frequently, so any entry in the cache repository will be stale following changes to
#'     functions, i.e., they will likely never be relevant again.
#'     This will therefore keep the cache repository clean of stale objects.
#'     If there is ambiguity in the `userTags`, i.e., they do not uniquely identify a single
#'     entry in the `cachePath`, then this option will default back to the non-dev-mode
#'     behaviour to avoid deleting objects.
#'     This, therefore, is most useful if the user is using unique values for `userTags`.
#'   }
#'   \item{`reproducible.useCacheV3`}{
#'     Default: `TRUE`. If this is set to `FALSE`, it will use the old `Cache` source
#'     code. This will only be available for a short period before it is deleted
#'     from the package. See also `reproducible.digestV3`. It is not guaranteed to
#'     be identical to using a previous version of `reproducible (<3.0)`.
#'   }
#'   \item{`useCloud`}{
#'     Default `FALSE`. Passed to `Cache`.
#'   }
#'   \item{`useDBI`}{
#'     Default: `TRUE` if \pkg{DBI} is available.
#'     Default value can be overridden by setting environment variable `R_REPRODUCIBLE_USE_DBI`.
#'     As of version 0.3, the backend is now \pkg{DBI} instead of \pkg{archivist}.
#'   }
#'   \item{`useGdown`}{
#'     Default: `FALSE`. If a user provides a Google Drive url to `preProcess`/`prepInputs`,
#'     `reproducible` will use the `googledrive` package. This works reliably in most cases.
#'     However, for large files on unstable internet connections, it will stall and
#'     stop the download with no error. If a user is finding this behaviour, they can
#'     install the `gdown` package, making sure it is available on the PATH. This call
#'     to `gdown` will only work for files that do not need authentication. If authentication
#'     is needed, `dlGoogle` will fall back to `googledrive::drive_download`, even
#'     if this option is `TRUE`, with a message.
#'     .
#'   }
#'   \item{`useMemoise`}{
#'     Default: `FALSE`. Used in [Cache()]. If `TRUE`, recovery of cached
#'     elements from the `cachePath` will use `memoise::memoise`.
#'     This means that the 2nd time running a function will be much faster than the first
#'     in a session (which either will create a new cache entry to disk or read a cached
#'     entry from disk).
#'     *NOTE: memoised values are removed when the R session is restarted*.
#'     **This option will use more RAM** and so may need to be turned off if RAM is limiting.
#'     `clearCache` of any sort will cause all memoising to be 'forgotten' (`memoise::forget`).
#'   }
#'   \item{`useNewDigestAlgorithm`}{
#'     Default: `1`. Option 1 is the version that has existed for sometime.
#'     There is now an option `2` which is substantially faster.
#'     It will, however, create Caches that are not compatible with previous ones.
#'     Options `1` and `2` are not compatible with the earlier `0`.
#'     `1` and `2` will make `Cache` less sensitive to minor but irrelevant changes
#'     (like changing the order of arguments) and will work successfully across operating systems
#'     (especially relevant for the new `cloudCache` function.
#'   }
#'   \item{`useTerra`}{
#'     Default: `FALSE`. The GIS operations in postProcess, by default use primarily
#'     the Raster package. The newer terra package does similar operations, but usually
#'     faster. A user can now set this option to `TRUE` and `prepInputs`
#'     and several components of `postProcess` will use `terra` internally.
#'   }
#'   \item{`verbose`}{
#'     Default: `FALSE`. If set to `TRUE` then every `Cache` call will show a
#'     summary of the objects being cached, their `object.size` and the time it took to digest
#'     them and also the time it took to run the call and save the call to the cache repository or
#'     load the cached copy from the repository.
#'     This may help diagnosing some problems that may occur.
#'   }
#'   \item{`digestV3`}{
#'     Default: `TRUE`. This uses a digest approach that includes the names of
#'     list elements and several other tweaks that were created for `reproducible 3.x`.
#'     Set this to `FALSE` to use *some of* the previous cache digesting to
#'     achieve some backwards compatibility with the digest algorithms of `reproducible (<3.x)`.
#'     It will not be possible to get it exact for all classes of objects, particularly
#'     those with file-backing.
#'   }
#'
#' }
#'
#' @section Advanced:
#' The following options are likely not needed by a user.
#' \describe{
#'   \item{`cloudChecksumsFilename`}{
#'     Default: `file.path(dirname(.reproducibleTempCacheDir()), "checksums.rds")`.
#'     Used as an experimental argument in [Cache()]
#'   }
#'   \item{`length`}{
#'     Default: `Inf`. Used in [Cache()], specifically to the internal
#'     calls to [CacheDigest()]. This is passed to `digest::digest`.
#'     Mostly this would be changed from default `Inf` if the digesting is taking too long.
#'     Use this with caution, as some objects will have *many* `NA` values in their first
#'     *many* elements
#'   }
#'   \item{`useragent`}{
#'     Default: `"https://github.com/PredictiveEcology/reproducible"`.
#'     User agent for downloads using this package.
#'   }
#' }
reproducibleOptions <- function() {
  list( # nolint
    reproducible.ask = TRUE,
    reproducible.cacheChaining = FALSE,
    reproducible.checkRemoteHash = FALSE,
    reproducible.cachePath = file.path(tempdir(), "reproducible", "cache"),
    reproducible.cacheSaveFormat = .rdsFormat,
    reproducible.cacheSpeed = "slow",
    reproducible.conn = NULL,
    reproducible.destinationPath = NULL,
    reproducible.drv = NULL, # RSQLite::SQLite(),
    reproducible.dryRun = FALSE,
    reproducible.futurePlan = FALSE, # future::plan("multisession"), #memoise
    reproducible.gdalwarp = FALSE,
    reproducible.gdalwarpThreads = 2L,
    reproducible.inputPath = file.path(tempdir(), "reproducible", "input"),
    reproducible.destinationPathShared = NULL,
    reproducible.destinationPathSharedRecursive = FALSE,
    reproducible.inputPaths = NULL,           # deprecated alias for reproducible.destinationPathShared
    reproducible.inputPathsRecursive = FALSE, # deprecated alias for reproducible.destinationPathSharedRecursive
    reproducible.leaveOnDisk = TRUE,
    reproducible.length = Inf,
    reproducible.memoisePersist = FALSE,
    reproducible.messageColourPrepInputs = "cyan",
    reproducible.messageColourCache = "blue",
    reproducible.messageColourQuestion = "green",
    reproducible.messageColourFunction = "red",
    reproducible.nThreads = 1,
    reproducible.objSize = TRUE,
    reproducible.overwrite = FALSE,
    reproducible.quick = FALSE,
    reproducible.rasterRead = getEnv("R_REPRODUCIBLE_RASTER_READ",
      default = "terra::rast",
      allowed = c("terra::rast", "raster::raster")
    ),
    reproducible.shapefileRead = "sf::st_read",
    reproducible.showSimilar = FALSE,
    reproducible.showSimilarDepth = 3,
    reproducible.tempPath = file.path(tempdir(), "reproducible"),
    reproducible.testCharacterAsFile = FALSE,
    reproducible.timeout = 1200,
    reproducible.useCOG = TRUE,
    reproducible.useCache = TRUE, # override Cache function
    reproducible.useCacheV3 = TRUE, # override Cache function
    reproducible.useCloud = FALSE, #
    reproducible.useDBI = {
      getEnv("R_REPRODUCIBLE_USE_DBI",
      default = {
        useDBI(getOption("reproducible.useDBI", NULL),  # a user may have set it before this runs; keep setting
                       verbose = interactive() - (useDBI() + 1), default = FALSE)
        }, # `FALSE` is useMultipleDBFiles now
      allowed = c("true", "false")
    ) |> as.logical()},
    reproducible.useGdown = FALSE,
    reproducible.useMemoise = FALSE, # memoise
    reproducible.useragent = "https://github.com/PredictiveEcology/reproducible",
    reproducible.verbose = 1,
    reproducible.digestV3 = TRUE
  )
}

getEnv <- function(envvar, default = NULL, allowed = NULL) {
  if (nzchar(Sys.getenv(envvar))) {
    val <- Sys.getenv(envvar)

    if (!val %in% allowed) {
      val <- default
    }
  } else {
    val <- default
  }

  return(val)
}

#' The `reproducible` package environments
#'
#' Environment used internally to store internal package objects and methods.
#'
#' - `.message` is specifically for messages and message-generating functions;
#' - `.pkgEnv` is for general use within the package;
#' - `.reproEnv` is used for `Cache`-related objects;
#'
#' @keywords internal
#' @rdname pkgEnv
.pkgEnv <- new.env(parent = emptyenv())
.pkgEnv$testCacheCounter <- 1L

.onLoad <- function(libname, pkgname) {
  ## set options using the approach used by devtools
  opts <- options()
  opts.reproducible <- reproducibleOptions()
  toset <- !(names(opts.reproducible) %in% names(opts))
  if (any(toset)) options(opts.reproducible[toset])
  .pkgEnv$SysInfo <- Sys.info() # record once at loading; repeatedly calling Sys.info is a waste

  .pkgEnv$runningOnMac <- isMac()
  ## Note: showCache async pre-population is no longer fired here.
  ## .onLoad runs before the user's real cachePath is set (e.g. by
  ## SpaDES.project::setupProject()), so spawning here would target the
  ## default tempdir() cachePath -- wasted work. Instead, the spawn fires
  ## lazily on the first Cache() / showCache() call against any given
  ## cachePath, where we know the path is the one the caller actually wants.

  invisible()
}

#' @importFrom utils packageVersion
.onAttach <- function(libname, pkgname) {
  if (isInteractive()) {
    packageStartupMessage(
      "Using reproducible version ",
      utils::packageVersion("reproducible"), ".",
      # "\n  'reproducible' has changed the default database backend.", # Not true yet
      " See ?reproducibleOptions for many possible settings.")
  }
}

.onUnload <- function(libpath) {
  ## unset reproducible options on unload
  cp <- getOption("reproducible.cachePath")
  if (!is.null(cp) && .Platform$OS.type != "windows") {
    # mccollect is fork-based and not available on Windows
    suppressMessages(parallel::mccollect())
  }
  
  o <- options()
  o[startsWith(names(o), prefix = "reproducible.")] <- NULL
  options(o)
  
}

.reproducibleTempPath <- function() getOption("reproducible.tempPath") # file.path(tempdir(), "reproducible")
.reproducibleTempCacheDir <- function() file.path(getOption("reproducible.tempPath"), "cache")
.reproducibleTempInputDir <- function() file.path(tempdir(), "reproducible", "inputs")

.argsToRemove <- unique(c(
  names(formals(prepInputs)),
  names(formals(cropInputs)),
  names(formals(cropTo)),
  names(formals(maskTo)),
  names(formals(projectTo)),
  names(formals(postProcessTo)),
  names(formals(fixErrors)),
  names(formals(fixErrorsIn)),
  # names(formals(raster::writeRaster)),
  # names(formals(raster::projectRaster)),
  names(formals(determineFilename)),
  names(formals(writeOutputs)),
  names(formals(writeTo)),
  unlist(lapply(
    methods("postProcess"),
    function(x) names(formals(x))
  ))
))

#' Change the absolute path of a file
#'
#' `convertPaths` is simply a wrapper around `gsub` for changing the
#' first part of a path.
#' `convertRasterPaths` is useful for changing the path to a file-backed
#' raster (e.g., after copying the file to a new location).
#'
#' @param x             For `convertPaths`, a character vector of file paths.
#'                      For `convertRasterPaths`, a disk-backed `RasterLayer`
#'                      object, or a list of such rasters.
#' @param patterns      Character vector containing a pattern to match (see `?gsub`).
#' @param replacements  Character vector of the same length of `patterns`
#'                      containing replacement text (see `?gsub`).
#'
#' @author Eliot McIntire and Alex Chubaty
#' @export
#' @rdname convertPaths
#' @return
#' A normalized path with the `patterns` replaced by `replacements`. Or a list of such
#' objects if `x` was a list.
#'
#' @examples
#' filenames <- c("/home/user1/Documents/file.txt", "/Users/user1/Documents/file.txt")
#' oldPaths <- dirname(filenames)
#' newPaths <- c("/home/user2/Desktop", "/Users/user2/Desktop")
#' convertPaths(filenames, oldPaths, newPaths)
#'
convertPaths <- function(x, patterns, replacements) {
  stopifnot(is(x, "character"))
  stopifnot(length(patterns) == length(replacements))
  patterns <- normPath(patterns)
  replacements <- normPath(replacements)
  x <- normPath(x)
  for (i in seq_along(patterns)) {
    x <- gsub(x = x, pattern = patterns[i], replacement = replacements[i])
  }
  normPath(x)
}

#' @author Eliot McIntire and Alex Chubaty
#' @export
#' @rdname convertPaths
convertRasterPaths <- function(x, patterns, replacements) {
  .requireNamespace("raster", stopOnFALSE = TRUE)

  if (is.list(x)) {
    x <- lapply(x, convertRasterPaths, patterns, replacements)
  } else if (!is.null(x)) {
    if (is.character(x)) {
      if (length(x) > 1) {
        x <- lapply(x, convertRasterPaths, patterns, replacements)
      } else {
        x <- raster::raster(x)
      }
    }

    x@file@name <- convertPaths(raster::filename(x), patterns, replacements)
  }
  x # handles null case
}

#' Return the filename(s) from a `Raster*` object
#'
#' This is mostly just a wrapper around `filename` from the `raster` package, except that
#' instead of returning an empty string for a `RasterStack` object, it will return a vector of
#' length >1 for `RasterStack`.
#'
#' @param obj A `Raster*` object (i.e., `RasterLayer`, `RasterStack`, `RasterBrick`)
#' @param allowMultiple Logical. If `TRUE`, the default, then all relevant
#'   filenames will be returned, i.e., in cases such as `.grd` where multiple files
#'   are required. If `FALSE`, then only the first file will be returned,
#'   e.g., `filename.grd`, in the case of default Raster format in R.
#' @param returnList Default `FALSE`. If `FALSE`, then return format will be a
#' character vector. When `TRUE`, list or environment objects will return a list
#' of character strings or vectors. When returned as a character vector, then
#' the names of objects with >1 filename associated with them will be given a numeric
#' suffix, which means the name in the returned vector does not match the object in
#' the list or environment. When returned as a list, their names are preserved.
#'
#' @author Eliot McIntire
#' @details
#' New methods can be made for this generic.
#'
#' @return
#' A character vector of filenames that are part of the objects passed to `obj`.
#' This returns `NULL` is the object is not file-backed or does not have a method
#' to recover the file-backed filename.
#' @export
#' @rdname Filenames
setGeneric("Filenames", function(obj, allowMultiple = TRUE, returnList = FALSE) {
  standardGeneric("Filenames")
})

#' @export
#' @rdname Filenames
setMethod(
  "Filenames",
  signature = "ANY",
  definition = function(obj, allowMultiple, returnList = FALSE) {
    if (inherits(obj, "RasterStack")) {
      fns <- unlist(lapply(seq_along(names(obj)), function(index) {
        Filenames(obj[[index]], allowMultiple = allowMultiple, returnList = returnList)
      }))

      dups <- duplicated(fns)
      if (any(dups)) {
        theNames <- names(fns)
        fns <- fns[!dups]
        names(fns) <- theNames[!dups]
      }
    } else if (inherits(obj, "RasterLayer")) {
      fns <- raster::filename(obj)
      if (exists("._Filenames_1")) browser()
      if (length(fns) == 0 || all(nchar(fns) == 0)) {
        fns <- NULL
      }
      fns <- allowMultipleFNs(allowMultiple, fns)
    } else if (.isSpatRaster(obj)) {
      if (!requireNamespace("terra", quietly = TRUE)) {
        stop("Please install terra package")
      }
      fns <- terra::sources(obj)
      if (any(fileExt(fns) %in% "vrt"))
        fns <- c(fns, terra::vrt_tiles(obj))

      fns <- allowMultipleFNs(allowMultiple, fns)
    } else {
      fns <- NULL
    }
    ## DON"T BE TEMPTED TO RM NZ i.e., MEMORY LAYERS -- NEED TO KNOW THAT THERE WERE SOME
    normPath(fns)
  }
)

#' @export
#' @rdname Filenames
setMethod(
  "Filenames",
  signature = "environment",
  definition = function(obj, allowMultiple = TRUE, returnList = FALSE) {
    rasterFilename <- Filenames(as.list(obj), allowMultiple = allowMultiple,
                                returnList = returnList)
    rasterFilenameDups <- lapply(rasterFilename, duplicated)

    if (any(unlist(rasterFilenameDups))) {
      rasterFilename <- rasterFilename[!unlist(rasterFilenameDups)]
    }
    return(rasterFilename)
  }
)

#' @export
#' @rdname Filenames
setMethod(
  "Filenames",
  signature = "list",
  definition = function(obj, allowMultiple = TRUE, returnList = FALSE) {
    ## convert a list to an environment -- this is to align it with a simList and environment
    if (is.null(names(obj))) {
      names(obj) <- as.character(seq(obj))
    }
    ll <- lapply(obj, function(o) Filenames(o, allowMultiple = allowMultiple,
                                            returnList = returnList))
    if (!isTRUE(returnList))
      ll <- unlist(ll)
    return(ll)
    # Filenames(as.environment(obj), allowMultiple = allowMultiple)
  }
)

#' @export
#' @rdname Filenames
setMethod(
  "Filenames",
  signature = "data.table",
  definition = function(obj, allowMultiple = TRUE, returnList = FALSE) {
    isCacheDB <- all(c(.cacheTableHashColName(), .cacheTableTagColName()) %in% colnames(obj))
    fromDsk <- NULL
    if (isCacheDB) {
      isFromDsk <- extractFromCache(obj, elem = "fromDisk") %in% "TRUE"
      if (any(isFromDsk)) {
        fromDsk <- extractFromCache(obj, elem = "origFilename")
      }
    }
    fromDsk
  }
)

#' @export
#' @rdname Filenames
setMethod(
  "Filenames",
  signature = "Path",
  definition = function(obj, allowMultiple = TRUE, returnList = FALSE) {
    obj <- allowMultipleFNs(allowMultiple, obj)
    # tags <- attr(obj, "tags")
    # if (!is.null(tags)) {
    #   tags1 <- parseTags(tags)
    #   if (allowMultiple)
    #   out <- tags1$tagValue[tags1$tagKey == "whichFiles"]
    # } else {
    #   out <- obj
    # }
    obj
  }
)


allowMultipleFNs <- function(allowMultiple, fns) {
  if (!is.null(fns)) {
    if (isTRUE(allowMultiple)) {
      anyGrd <- endsWith(fns, suffix = "grd")
      anyGri <- endsWith(fns, suffix = "gri")
      if (any(anyGrd) && !any(anyGri)) {
        nonGrd <- if (any(!anyGrd)) fns[!anyGrd] else NULL
        multiFns <- sort(c(fns[anyGrd], gsub("grd$", "gri", fns[anyGrd])))
        fnsNew <- c(nonGrd, multiFns)
        fns <- fnsNew[order(match(
          filePathSansExt(basename(fnsNew)),
          filePathSansExt(basename(fns))
        ))]
      }
    } else {
      fns <- grep("\\.gri$", fns, value = TRUE, invert = TRUE)
    }
  }
  fns

}

#' Get min or maximum value of a (Spat)Raster
#'
#' During the transition from raster to terra, some functions are not drop in
#' replacements, such as `minValue` and `maxValue` became `terra::minmax`. This
#' helper allows one function to be used, which calls the correct max or min
#' function, depending on whether the object is a `Raster` or `SpatRaster`.
#'
#' @param x A `Raster` or `SpatRaster` object.
#'
#' @return
#' A vector (not matrix as in `terra::minmax`) with the minimum or maximum
#' value on the `Raster` or `SpatRaster`, one value per layer.
#'
#' @export
#' @rdname terra-migration
#'
#' @examples
#' if (requireNamespace("terra", quietly = TRUE)) {
#'   ras <- terra::rast(terra::ext(0, 10, 0, 10), vals = 1:100)
#'   maxFn(ras)
#'   minFn(ras)
#' }
minFn <- function(x) {
  minmaxFn(x, "min")
}

#' @export
#' @rdname terra-migration
maxFn <- function(x) {
  minmaxFn(x, "max")
}

#' @importFrom utils head tail
minmaxFn <- function(x, which = "max") {
  out <- NULL
  if (is(x, "Raster")) {
    .requireNamespace("raster", stopOnFALSE = TRUE)
    fn <- get(paste0(which, "Value"), envir = asNamespace("raster"))
    out <- fn(x)
  } else {
    .requireNamespace("terra", stopOnFALSE = TRUE)
    fn <- ifelse(identical(which, "max"), "tail", "head")
    fn <- getFromNamespace(fn, ns = "utils")
    if (!terra::hasMinMax(x))
      x <- terra::setMinMax(x)
    out <- fn(terra::minmax(x), 1)[1, ]
  }
  if (is.null(out)) {
    stop("To use maxFn or minFn, you need either terra or raster package installed")
  }

  out
}

#' @export
#' @param ... Passed to the functions in `raster` or `terra`, as needed.
#' @rdname terra-migration
dataType2 <- function(x, ...) {
  if (is(x, "Raster")) {
    raster::dataType(x)
  } else {
    terra::datatype(x, ...)
  }
}

#' @export
#' @rdname terra-migration
nlayers2 <- function(x) {
  if (is(x, "Raster")) {
    raster::nlayers(x)
  } else {
    terra::nlyr(x)
  }
}


#' @export
#' @rdname terra-migration
values2 <- function(x, ...) {
  if (is(x, "Raster")) {
    raster::values(x, ...)
  } else {
    if (.isSpatRaster(x)) {
      terra::values(x, ..., mat = nlayers2(x) > 1)
    } else {
      terra::values(x, ...)
    }
  }
}

verboseCacheMessage <- function(preDigest, functionName,
                             startHashTime, modifiedDots, quick,
                             verbose = getOption("reproducible.verbose", 1),
                             verboseLevel = 1) {
  if (verbose >= verboseLevel) {
    preDigestUnlist <- .unlistToCharacter(preDigest, 4)
    verboseDF0(verbose, functionName, startHashTime)
    # verboseDF <- data.frame(
    #   functionName = functionName,
    #   component = "Hashing",
    #   elapsedTime = as.numeric(difftime(endHashTime, startHashTime, units = "secs")),
    #   units = "secs",
    #   stringsAsFactors = FALSE
    # )

    # hashObjectSize <- unlist(lapply(modifiedDots, objSize, recursive = FALSE, quick = TRUE))
    hashObjectSize <- if (getOption("reproducible.objSize", TRUE)) {
      unlist(lapply(modifiedDots, function(x) {
        unname(attr(objSize(x, quick = FALSE), "objSize"))
      }))
    } else {
      Map(modifiedDots, function(x) NA)
    }

    lengths <- unlist(lapply(preDigestUnlist, function(x) length(unlist(x))))
    hashDetails <- data.frame(
      objectNames = rep(names(preDigestUnlist), lengths),
      hashElements = names(unlist(preDigestUnlist)),
      hash = unname(unlist(preDigestUnlist)),
      stringsAsFactors = FALSE
    )
    preDigestUnlistNames <- unlist(lapply(
      strsplit(names(unlist(preDigestUnlist)), split = "\\."), # nolint
      function(x) paste0(tail(x, 2), collapse = ".")
    ))
    hashObjectSizeNames <- unlist(lapply(
      strsplit(names(hashObjectSize), split = "\\$"),
      function(x) paste0(tail(x, 2), collapse = ".")
    ))
    hashObjectSizeNames <- gsub("\\.y", replacement = "", hashObjectSizeNames)
    hashObjectSizeNames <- unlist(lapply(
      strsplit(hashObjectSizeNames, split = "\\."),
      function(x) paste0(tail(x, 2), collapse = ".")
    ))
    hashDetails$objSize <- NA
    hashDetails$objSize[preDigestUnlistNames %in% hashObjectSizeNames] <-
      hashObjectSize[hashObjectSizeNames %in% preDigestUnlistNames]

    if (exists("hashDetails", envir = .reproEnv, inherits = FALSE)) {
      .reproEnv$hashDetails <- rbind(.reproEnv$hashDetails, hashDetails)
    } else {
      .reproEnv$hashDetails <- hashDetails
      on.exit(
        {
          assign("hashDetailsAll", .reproEnv$hashDetails, envir = .reproEnv)
          messageDF(.reproEnv$hashDetails, colour = "blue", verbose = verbose, verboseLevel = verboseLevel)
          messageCache("The hashing details are available from .reproEnv$hashDetailsAll",
                       verbose = verbose, verboseLevel = verboseLevel
          )
          rm("hashDetails", envir = .reproEnv)
        },
        add = TRUE
      )
    }
  }
}

utils::globalVariables(c("X", "Y"))

checkNameHasGeom <- function(existingObj) {
  hasGeomNamedCol <- names(existingObj) %in% "geom"
  if (any(hasGeomNamedCol)) {
    browser()
    names(existingObj)[hasGeomNamedCol] <- "geometry"
  }
  existingObj
}

extractPolygonIfWithin <- function(domain, existingObjSF, bufferOK, existingObj, verbose = TRUE) {
  wh <- sf::st_within(domain, existingObjSF, sparse = FALSE)
  if (isTRUE(all(wh %in% FALSE)) && isTRUE(bufferOK)) {
    diffs <- mapply(minmax = list(c("xmin", "xmax"), c("ymin", "ymax")), function(minmax)
      round(abs(diff(sf::st_bbox(existingObjSF)[minmax])), 0))
    buff <- diffs * 0.025
    meanBuff <- mean(buff)
    meanBuffKm <- round(meanBuff/1e3, 1)
    message("domain is not within existing object; trying a 2.5% (", meanBuffKm, "km) buffer")
    existingObjSF_wider <- sf::st_buffer(existingObjSF, dist = meanBuff)
    wh <- sf::st_within(domain, existingObjSF_wider, sparse = FALSE)
  }
  wh1 <- apply(wh, 1, any)
  wh2 <- apply(wh, 2, any) # This is "are the several domains inside the several existingObjSF"
  # length of existingObjSF
  domainExisted <- all(wh1)
  if (domainExisted) {

    # THIS IS "PULL OUT INDIVIDUAL SF POLYGON FROM THE MANY"
    if (isTRUE(verbose)) {
      if (isTRUE(bufferOK)) # the message will be previously given
        message("domain is within the buffered object; returning the existing parameters")
      else
        message(.message$cacheGeoDomainContained)
    }
    existingObj <- existingObj[wh2, ]
    existingObjSF <- existingObjSF[wh2, ]
  }
  if (all(wh1 %in% FALSE)) {
    existingObj <- NULL
    existingObjSF <- NULL
  }
  list(existingObj = existingObj, existingObjSF = existingObjSF, domainExisted = domainExisted)
}


#  from https://github.com/zhukovyuri/SUNGEO/blob/master/R/update_bbox.R
update_bbox <- function(sfobj){
  # Manually calculate bounds from coordinates
  new_bb <- data.table::as.data.table(sf::st_coordinates(sfobj))[,c(min(X),min(Y),max(X),max(Y))]
  # Rename columns
  names(new_bb) <- c("xmin", "ymin", "xmax", "ymax")
  # Change object class
  attr(new_bb, "class") <- "bbox"
  # Assign to bbox slot of sfobj
  attr(sf::st_geometry(sfobj), "bbox") <- new_bb

  return(sfobj)
}

#' Cache-like function for spatial domains
#'
#' \if{html}{\figure{lifecycle-experimental.svg}{options: alt="experimental"}}
#'
#' This function is a combination of `Cache` and `prepInputs` but for spatial
#' domains. This differs from `Cache` in that the current function call doesn't
#' have to have an identical function call previously run. Instead, it needs
#' to have had a previous function call where the `domain` being passes is
#' *within* the geographic limits of the `targetFile`.
#' This is similar to a geospatial operation on a remote GIS server, with 2 differences:
#' 1. This downloads the object first before doing the GIS locally, and 2. it will
#' optionally upload an updated object if the geographic area did not yet exist.
#'
#' @details
#' This has a very specific use case: assess whether an existing `sf` polygon
#' or multipolygon object (local or remote) covers the spatial
#' area of a `domain` of interest. If it does, then return only that
#' part of the `sf` object that completely covers the `domain`.
#' If it does not, then run `FUN`. It is expected that `FUN` will produce an `sf`
#' polygon or multipolygon class object. The result of `FUN` will then be
#' appended to the `sf` object as a new entry (feature) or it will replace
#' the existing "same extent" entry in the `sf` object.
#'
#'
#' @param targetFile The (optional) local file (or path to file) name for a `sf`
#'   object or `data.frame` that can be coerced to a `sf` object (i.e., has a `geometry`
#'   column). If `cloudFolderID` is specified, then this will be the name of the
#'   file stored and/or accessed in that cloud folder.
#' @param domain An sf polygon object that is the spatial area of interest. If `NULL`,
#'   then this will return the whole object in `targetFile`.
#' @param FUN A function call that will be called if there is the `domain` is
#'   not already contained within the `sf` object at `targetFile`. This function
#'   call MUST return either a `sf` class object or a `data.frame` class object
#'   that has a geometry column (which can then be converted to `sf` with `sf::st_as_sf`)
#' @param cloudFolderID If this is specified, then it must be either 1) a Google Drive
#'   url to a folder where the `targetFile` will be read from or written to, or
#'   2) a `googledrive` id or 3) an absolute path to a (possibly non-existent yet)
#'   folder on your Google drive.
#' @param useCloud A logical.
#' @param bufferOK A logical. If `TRUE`, then after testing whether the domain is
#'   within the `targetFile` spatial object, and if it returns `FALSE`, then the function
#'   will create a larger object, buffered by 2.5% of the extent of the object. If
#'   `FALSE`, then it will be strict about whether the `domain` is within the `targetFile`.
#' @param ... All named objects that are needed for FUN, including the function itself,
#'   if it is not in a package.
#'
#' @inheritParams prepInputs
#' @inheritParams Cache
#' @param action A character string, with one of c("nothing", "update",
#'   "replace", "append"). Partial matching is used ("n" is sufficient).
#'   `nothing` will prevent any updating of the `targetFile`,
#'   i.e., "read only". `append` will add the spatial elements in domain to
#'   `targetFile` (and writing it back to disk). `update` will do the same as
#'   `append`, but will also remove any identical geometries before appending.
#'   `replace` does nothing currently.
#'
#' @return Returns an object that results from `FUN`, which will possibly be a subset
#' of a larger spatial object that is specified with `targetFile`.
#'
#' @export
#' @examples
#' \donttest{
#'
#' if (requireNamespace("sf", quietly = TRUE) &&
#'     requireNamespace("terra", quietly = TRUE)) {
#'   dPath <- checkPath(file.path(tempdir2()), create = TRUE)
#'   localFileLux <- system.file("ex/lux.shp", package = "terra")
#'
#'   # 1 step for each layer
#'   # 1st step -- get study area
#'   full <- prepInputs(localFileLux, destinationPath = dPath) # default is sf::st_read
#'   zoneA <- full[3:6, ]
#'   zoneB <- full[8, ] # not in A
#'   zoneC <- full[3, ] # yes in A
#'   zoneD <- full[7:8, ] # not in A, B or C
#'   zoneE <- full[3:5, ] # yes in A
#'   # 2nd step: re-write to disk as read/write is lossy; want all "from disk" for this ex.
#'   writeTo(zoneA, writeTo = "zoneA.shp", destinationPath = dPath)
#'   writeTo(zoneB, writeTo = "zoneB.shp", destinationPath = dPath)
#'   writeTo(zoneC, writeTo = "zoneC.shp", destinationPath = dPath)
#'   writeTo(zoneD, writeTo = "zoneD.shp", destinationPath = dPath)
#'   writeTo(zoneE, writeTo = "zoneE.shp", destinationPath = dPath)
#'   # Must re-read to get identical columns
#'   zoneA <- sf::st_read(file.path(dPath, "zoneA.shp"))
#'   zoneB <- sf::st_read(file.path(dPath, "zoneB.shp"))
#'   zoneC <- sf::st_read(file.path(dPath, "zoneC.shp"))
#'   zoneD <- sf::st_read(file.path(dPath, "zoneD.shp"))
#'   zoneE <- sf::st_read(file.path(dPath, "zoneE.shp"))
#'
#'   # The function that is to be run. This example returns a data.frame because
#'   #    saving `sf` class objects with list-like columns does not work with
#'   #    many st_driver()
#'   fun <- function(domain, newField) {
#'     domain |>
#'       as.data.frame() |>
#'       cbind(params = I(lapply(seq_len(NROW(domain)), function(x) newField)))
#'   }
#'
#'   # Run sequence -- A, B will add new entries in targetFile, C will not,
#'   #                 D will, E will not
#'   for (z in list(zoneA, zoneB, zoneC, zoneD, zoneE)) {
#'     out <- CacheGeo(
#'       targetFile = "fireSenseParams.rds",
#'       domain = z,
#'       FUN = fun(domain, newField = I(list(list(a = 1, b = 1:2, c = "D")))),
#'       fun = fun, # pass whatever is needed into the function
#'       destinationPath = dPath,
#'       action = "update"
#'       # , cloudFolderID = "cachedObjects" # to upload/download from cloud
#'     )
#'   }
#' }
#' }
CacheGeo <- function(targetFile = NULL,
                     domain,
                     FUN,
                     destinationPath = getOption("reproducible.destinationPath", "."),
                     useCloud = getOption("reproducible.useCloud", FALSE),
                     cloudFolderID = NULL,
                     purge = FALSE, useCache = getOption("reproducible.useCache"),
                     overwrite = getOption("reproducible.overwrite"),
                     action = c("nothing", "update", "replace", "append"),
                     bufferOK = FALSE, verbose = getOption("reproducible.verbose"),
                     ...) {
  objExisted <- TRUE
  if (is.null(targetFile)) {
    objExisted <- FALSE
  }

  if (!missing(targetFile)) {
    if (!isAbsolutePath(targetFile)) {
      targetFileWithDP <- file.path(destinationPath, targetFile)
    }
  } else {
    stop("Either targetFile must be supplied")
  }
  if (!file.exists(targetFileWithDP)) {
    objExisted <- FALSE
  }
  domainExisted <- objExisted # !missing(domain) # objExisted
  urlThisTargetFile <- NULL

  alreadyOnRemote <- FALSE
  cacheExtra <- NULL

  if (isTRUE(useCloud) || !is.null(cloudFolderID)) {
    .requireNamespace("googledrive", stopOnFALSE = TRUE)
    objsInGD <- googledrive::drive_ls(cloudFolderID)
    if (!(nchar(cloudFolderID) == 33 || grepl("https://drive", cloudFolderID))) {
      googledrive::with_drive_quiet(folderExists <- googledrive::drive_get(cloudFolderID))
      if (NROW(folderExists)) {
        cloudFolderID <- googledrive::as_id(folderExists$id)
      }
    } else {
      cloudFolderID <- googledrive::as_id(cloudFolderID)
      folderExists <- googledrive::drive_get(cloudFolderID)
    }
    folderExists <- NROW(folderExists) > 0

    if (!folderExists) {
      cloudFolderID <- googledrive::drive_mkdir(cloudFolderID)
    }
    objID <- objsInGD[objsInGD$name %in% basename2(targetFile), ]

    if (NROW(objID)) urlThisTargetFile <- objID$drive_resource[[1]]$webViewLink

    objExisted <- if (NROW(objID)) TRUE else FALSE
    if (is.null(targetFile)) {
      targetFile <- objID$name
      targetFileWithDP <- file.path(destinationPath, objID$name)
    }
    if (file.exists(targetFileWithDP) && NROW(objID)) {
      md5Checksum <- digest::digest(file = targetFileWithDP)
      alreadyOnRemote <- identical(objID$drive_resource[[1]]$md5Checksum, md5Checksum)
      cacheExtra <- objID$drive_resource[[1]]$md5Checksum
    }
  }

  if (isTRUE(objExisted)) {

    aa <- quote(prepInputs(
      targetFile = asPath(targetFile),
      url = urlThisTargetFile,
      destinationPath = destinationPath, # domain = domain,
      useCache = FALSE,
      purge = purge, # It isn't relevant if the file is different than the Checksums
      overwrite = overwrite
    ))
    for (attempt in 1:2) {
      # There were cases where the Cache recovered, but the file was not there.
      existingObj <- eval(aa) |>
        Cache(.cacheExtra = cacheExtra, .functionName = paste0("prepInputs_", basename(targetFile))) # cacheExtra is the md5Checksum on GDrive
      if (file.exists(targetFileWithDP))
        break
      else
        clearCache(cacheId = cacheId(existingObj), ask = FALSE)
    }

    existingObjOrig <- existingObj

    cn <- colnames(existingObj)
    # Assumption that data.frame should be data.table
    if (is(existingObj, "list")) {
      existingObj <- lapply(existingObj, function(x) if (is.data.frame(x[[1]])) I(list(as.data.table(x[[1]]))) else x) |>
        as.data.table(fill = TRUE) |> as.data.frame()
    } else {
      if (!is(existingObj, "data.frame"))
        existingObj <- as.data.frame(existingObj)
    }
    colnames(existingObj) <- cn

    newCRS <- unique(existingObj[["crs"]])[[1]]
    if (length(newCRS) > 1)
      stop("The file on googledrive has more than one crs; currently this is not allowed.",
           " You can try rebuilding the data.frame with a single crs, assigned as a list column ",
           "in the data.frame")
    existingObjSF <- if (is(existingObj, "sf")) existingObj else sf::st_as_sf(existingObj)
    existingObjSF <- existingObjSF[!sf::st_is_empty(existingObjSF),] #for some reason some have empty geometries...

    existingObjSF <- update_bbox(existingObjSF)
    if (!is.null(newCRS))
      suppressWarnings(sf::st_crs(existingObjSF) <- newCRS)
    existingObjSFOrig <- existingObjSF
    if (!missing(domain)) {
      #must be sf for st_within
      if (!inherits(domain, "sf")) domain <- sf::st_as_sf(domain)
      #must have same crs for st_within
      existingObjSF <- sf::st_transform(existingObjSF, sf::st_crs(domain))

      outs <- extractPolygonIfWithin(domain, existingObjSF, bufferOK, existingObj)
      list2env(outs, envir = environment()) # existingObjSF, existingObj, domainExisted
    } else {
      domainExisted <- TRUE
      message("Spatial domain is missing; returning entire spatial domain")
    }
  }
  msgActionIsNothing <- "action was 'nothing'; nothing done"
  if ( (isFALSE(objExisted) || isFALSE(domainExisted) ) && !missing(FUN)) {
    if (isFALSE(objExisted)) {
      message(.message$cacheGeoNoRemoteExists)
    } else {
      message(.message$cacheGeoDomainNotContained)
    }
    FUNcaptured <- substitute(FUN)
    env <- environment()
    list2env(list(...), envir = env) # need the ... to be "here"
    newObj <- try(eval(FUNcaptured, envir = env), silent = TRUE)
    newObjSF <- if (is(newObj, "sf")) newObj else sf::st_as_sf(newObj)
    if (isFALSE(objExisted)) {
      existingObj <- newObj
    }
  } else {
    if (isFALSE(objExisted)) {
      message("targetFile (", targetFile,") does not exist")
      existingObj <- NULL
    }
    if (isFALSE(domainExisted))
      message("domain does not exist in targetFile (", targetFile, ")")
    if (missing(FUN) && isFALSE(domainExisted)) {
      if (!action %in% "nothing")
        message("FUN is missing; no evaluation possible")
      if (action %in% "nothing")
        message(msgActionIsNothing)
    }



  }
  if (isFALSE(domainExisted)) {
    if (isTRUE(objExisted)) {
      if (any(grepl("^a|^u", action[1], ignore.case = TRUE))) {
        # .gpkg seems to change geometry to "geom"
        existingObjSF <- checkNameHasGeom(existingObjSF)
        if (!any(is(sf::st_geometry(newObjSF), "sfc_MULTIPOLYGON"))) {
          newObjSF <- sf::st_cast(newObjSF, "MULTIPOLYGON")
        }
        # newObj <- sf::st_cast(newObj, "MULTIPOLYGON")
        if (!is(sf::st_geometry(newObj$geometry), "MULTIPOLYGON"))
          newObj$geometry <- sf::st_cast(newObj$geometry, "MULTIPOLYGON")
        if (!is(sf::st_geometry(existingObjOrig$geometry), "MULTIPOLYGON"))
          existingObjOrig$geometry <- sf::st_cast(existingObjOrig$geometry, "MULTIPOLYGON")

        # THE APPEND LINE
        existingObj <- as.data.frame(rbindlist(
          list(as.data.table(existingObjOrig), as.data.table(newObj)), fill = TRUE, use.names = TRUE))
        existingObjSF <- sf::st_as_sf(existingObj)

        if (sf::st_crs(domain) != sf::st_crs(existingObjSF)) {
          domain <- sf::st_transform(domain, sf::st_crs(existingObjSF))
        }
        outs <- extractPolygonIfWithin(domain, existingObjSF, bufferOK, existingObj)
        list2env(outs) # existingObjSF, existingObj, domainExisted

        # existingObj <- rbind(existingObj, newObj)
        if (any(duplicated(existingObj))) {
          existingObj <- unique(existingObj)
        }
      }
    }
    if (!any(grepl("^n", action[1], ignore.case = TRUE))) {
      if (!isAbsolutePath(targetFileWithDP)) {
        targetFileWithDP <- file.path(destinationPath, targetFile)
      }
      # if (is(existingObj, "sf")) existingObj <- as.data.frame(existingObj)

      # Put it in order
      if (!is.null(existingObj[["polygonID"]])) {
        # if it has 2 sets of dots, like "4.2.1"
        if (isTRUE(any(grepl("\\..+\\.", existingObj$polygonID)))) {
          polygonIDforSorting <- as.package_version(existingObj$polygonID)
        } else {
          suppressWarnings(polygonIDforSorting <- as.numeric(gsub("(\\..)\\.", "\\1", existingObj$polygonID)))
          if (isTRUE(any(is.na(polygonIDforSorting))))
            polygonIDforSorting <- existingObj$polygonID
        }
        ord <- order(polygonIDforSorting)
        existingObj <- existingObj[ord, ]
        rownames(existingObj) <- seq_len(NROW(existingObj))
      }
      ## end of putting it in order

      for (attempt in 1:2) {
        if (identical("rds", fs::path_ext(targetFileWithDP)) || identical(attempt, 2L))  {
          saveRDS(existingObj, file = targetFileWithDP)
          break
        } else {
          warns <- character()
          withCallingHandlers(
            writeTo(existingObj, writeTo = targetFileWithDP,
                    overwrite = TRUE, append = TRUE),
            warning = function(w) {
              warns <<- w$message
            }
          )
          if (!any(grepl("Dropping column", warns))) break
          targetFileWithDP <- gsub(fileExt(targetFileWithDP), "rds", targetFileWithDP)
          warning(.message$BecauseOfLossOfColumn(targetFileWithDP))

        }
      }
    } else {
      if (!missing(FUN)) {
        message("The spatial domain is new, and should be added, but\n")
        message(msgActionIsNothing)
      }
    }
  }
  # Cloud
  if (isTRUE(useCloud) || !is.null(cloudFolderID)) {
    if (!isAbsolutePath(targetFile)) targetFileWithDP <- file.path(destinationPath, targetFile)
    if (NROW(objID)) {
      md5Checksum <- digest::digest(file = targetFileWithDP)
      alreadyOnRemote <- identical(objID$drive_resource[[1]]$md5Checksum, md5Checksum)
    }
    if (!alreadyOnRemote) {
      if (!any(grepl("^n", action[1], ignore.case = TRUE))) {
        if (!dir.exists(tempdir())) {
          dir.create(tempdir(), recursive = TRUE)
        }
        out <- googledrive::drive_put(
          media = targetFileWithDP,
          path = googledrive::as_id(cloudFolderID)
        )
        attr(existingObj, "id") <- out
      } else {
        msgSkippingUpload <- "skipping googledrive upload;"
        message(msgSkippingUpload)
        message(msgActionIsNothing)
      }

    }
  }

  if (exists("existingObjOrig", inherits = FALSE)) {
    messageColoured("To get the full object from googledrive, which looks like this:\n")
    useOrig <- (NROW(existingObj) < NROW(existingObjOrig))
    objLooksLike <- if (useOrig) existingObjOrig else existingObj
    if (verbose > 0) {
      cat(cli::col_yellow(capture.output(objLooksLike)), sep = "\n")
      messageColoured("... run the following:\n")
      enn <- environment()
      ll <- lapply(aa, function(x) eval(x, envir = enn))
      coo <- capture.output(as.call(append(list(quote(prepInputs)), ll[-1])))#, sep = "\n")
      cat(cli::col_yellow(coo), sep = "\n")
    }
  }


  if (!exists("existingObjSF")) {
    if (is.null(existingObj)) {
      existingObjSF <- existingObj
    } else {
      existingObjSF <- sf::st_as_sf(existingObj)
    }
  }


  existingObjSF
}

#' Wrapper around `lobstr::obj_size`
#'
#' This function attempts to estimate the real object size of an object. If the object
#' has pass-by-reference semantics, it may not estimate the object size well without
#' a specific method developed. For the case of `terra` class objects, this will
#' be accurate (both RAM and file size), but only if it is not passed inside
#' a list or environment. To get an accurate size of these, they should be passed
#' individually.
#'
#' @return
#' This will return the result from `lobstr::obj_size`, i.e., a `lobstr_bytes`
#' which is a `numeric`. If `quick = FALSE`, it will also have an attribute,
#' "objSize", which will
#' be a list with each element being the `objSize` of the individual elements of `x`.
#' This is particularly useful if `x` is a `list` or `environment`.
#' However, because of the potential for shared memory, the sum of the individual
#' elements will generally not equal the value returned from this function.
#'
#' @param x An object
#' @param quick Logical. If `FALSE`, then an attribute, "objSize" will be added to
#'   the returned value, with each of the elements' object size returned also.
#' @param recursive Logical. If `TRUE`, then, in addition to evaluating the whole object,
#'   it will also return the recursive sizes of the elements of a list or environment.
#' @param ...  Additional arguments (currently unused), enables backwards compatible use.
#'
#' @export
#' @rdname objSize
#'
#' @examples
#' library(utils)
#'
#' foo <- new.env()
#' foo$b <- 1:10
#' foo$d <- 1:10
#'
#' objSize(foo) # all the elements in the environment
#' utils::object.size(foo) # different - only measuring the environment as an object
#'
#' utils::object.size(prepInputs) # only the function, without its enclosing environment
#' objSize(prepInputs) # the function, plus its enclosing environment
#'
#' os1 <- utils::object.size(as.environment("package:reproducible"))
#' (os1) # very small -- just the environment container
#'
#' @details
#' For functions, a user can include the enclosing environment as described
#' <https://www.r-bloggers.com/2015/03/using-closures-as-objects-in-r/> and
#' <http://adv-r.had.co.nz/memory.html>.
#' It is not entirely clear which estimate is better.
#' However, if the enclosing environment is the `.GlobalEnv`, it will
#' not be included even though `enclosingEnvs = TRUE`.
#'
#' @export
objSize <- function(x, quick = FALSE, recursive = FALSE, ...) {
  UseMethod("objSize", x)
}

#' @export
#' @importFrom lobstr obj_size
objSize.default <- function(x, quick = FALSE, ...) {
  FNs <- Filenames(x)
  if (!is.null(FNs) && length(FNs) && !isTRUE(quick)) {
    if (any(nzchar(FNs))) {
      FNs <- asPath(FNs)
      out2 <- objSize(FNs, quick = FALSE)
    }
  }
  if (.isSpat(x)) {
    if (.requireNamespace("terra")) {
      if (inherits(x, "SpatVector")) { # too slow for large SpatVectors
        x <- list(terra::geom(x), terra::values(x))
      } # approximate
      else {
        if (!isTRUE(quick))
          x <- list(terra::wrap(x))
      }
    }
  }
  out <- object.size(x)
  # out <- .objSizeWithTry(x)  # out <- try(obj_size(x), silent = TRUE) # Error in obj_size_(dots, env, size_node(), size_vector()) :
  #                                        # bad binding access
  if (exists("out2", inherits = FALSE)) {
    out <- sum(out, out2)
  }
  class(out) <- "lobstr_bytes"

  if (!quick) {
    attr(out, "objSize") <- list(obj = out)
  }
  out
}

#' @export
#' @importFrom lobstr obj_size
objSize.list <- function(x, quick = FALSE, recursive = FALSE, ...) {
  os <- .objSizeWithTry(x)
  # os <- try(obj_size(x), silent = TRUE) # need to get overall object size; not just elements;
  # but this doesn't work for e.g., terra

  if (!quick && (isTRUE(recursive) || (is.numeric(recursive) && recursive > 0))) {
    out <- Map(x = x, function(x) objSize(x, quick = quick, recursive = recursive - 1))
    os2 <- sum(unlist(out))
    os <- max(os2, os)
    class(os) <- "lobstr_bytes"
    attr(os, "objSize") <- out
  }

  os
}

#' @export
#' @importFrom lobstr obj_size
objSize.Path <- function(x, quick = FALSE, ...) {
  if (quick) {
    os <- .objSizeWithTry(x, FALSE)
  } else {
    os <- file.size(x)
  }

  if (!quick) {
    attr(os, "objSize") <- os
  }
  class(os) <- "lobstr_bytes"

  os
}

#' @export
#' @importFrom lobstr obj_size
objSize.environment <- function(x, quick = FALSE, recursive = FALSE, ...) {
  os <- .objSizeWithTry(x)
  if (!quick) {
    out <- lapply(as.list(x, all.names = TRUE), objSize, quick = quick)
    attr(os, "objSize") <- out
  }
  os
}

#' @export
#' @importFrom lobstr obj_size
objSize.function <- function(x, quick = FALSE, ...) {
  os <- .objSizeWithTry(x)
  if (!quick)
    attr(os, "objSize") <- os
  os
}


#' @param sumLevel Numeric, indicating at which depth in the list of objects should the
#'   object sizes be summed (summarized). Default is `Inf`, meaning no sums. Currently,
#'   the only option other than Inf is 1: `objSizeSession(1)`,
#'   which gives the size of each package.
#' @param enclosingEnvs Logical indicating whether to include enclosing environments.
#'                      Default `TRUE`.
#' @param .prevEnvirs For internal account keeping to identify and prevent duplicate counting
#'
#'
#' @details `objSizeSession` will give the size of the whole session, including loaded packages.
#' Because of the difficulties in calculating the object size of `base`
#' and `methods` packages and `Autoloads`, these are omitted.
#'
#' @export
#' @rdname objSize
objSizeSession <- function(sumLevel = Inf, enclosingEnvs = TRUE, .prevEnvirs = list()) {
  .Defunct("Please use lobstr::obj_size instead")
}


#' `lobstr::obj_size` with a `try` to address issue #72
#'
#' It is not clear why, but it appears that running `lobstr::obj_size` again, after
#' a bad binding error, it will work.
#' @export
#' @importFrom lobstr obj_size
#' @inheritParams objSize
#' @param useTry Logical. If `TRUE`, the default, then it will use `try`. Can optionally
#'   avoid this if set to `FALSE`. The `try` takes sufficient extra compute time
#'   that it is worth avoiding it if possible.
#' @return The size of an object, using `lobstr::obj_size` or `object.size` if the
#'   first fails
.objSizeWithTry <- function(x, useTry = TRUE) {
  for (i in 1:2) {
    if (isTRUE(useTry)) {
      out <- try(obj_size(x), silent = TRUE)
    } else {
      out <- obj_size(x)
    }

    if (inherits(out, "try-error")) {
      if (identical(i, 1L)) next
    } else {
      break
    }
    out <- try(length(serialize(x, NULL)), silent = TRUE)
    if (inherits(out, "try-error")) {
      out <- object.size(x)
    }
    class(out) <- "lobstr_bytes"
  }
  out
}

#' The `reproducible` package
#'
#' @description
#' \if{html}{\figure{lifecycle-maturing.svg}{options: alt="maturing"}}
#'
#' This package aims at making
#' high-level, robust, machine and OS independent tools for making deeply
#' reproducible and reusable content in R. The core user functions are `Cache`
#' and `prepInputs`. Each of these is built around many core and edge cases
#' required to have reproducible code of arbitrary complexity.
#'
#' @section Main Tools:
#'
#' There are many elements within the reproducible package.
#' However, there are currently two main ones that are critical for reproducible research.
#' The key element for reproducible research is that the code must always return the same content
#' every time it is run, but it must be vastly faster the 2nd, 3rd, 4th etc, time it is run.
#' That way, the entire code sequence for a project of arbitrary size can be run
#' *from the start* every time.
#'
#' \describe{
#'   \item{[Cache()]:}{A robust wrapper for any function, including those with environments,
#'   disk-backed storage (currently on `Raster`) class), operating-system independent,
#'   whose first time called will execute the function, second time will compare the inputs to a
#'   database of entries, and recover the first result if inputs are identical.
#'   If `options("reproducible.useMemoise" = TRUE)`, the second time will be very fast as it
#'   will recover the answer from RAM.}
#'   \item{[prepInputs()]for other specifics for other classes.:}{ Download, or load objects, and possibly post-process them.
#'   The main advantage to using this over more direct routes is that it will automatically build
#'   checksums tables, use `Cache` internally where helpful, and possibly run a variety of
#'   post-processing actions.
#'   This means this function can also itself be cached for even more speed.
#'   This allows all project data to be stored in custom cloud locations or in their original online
#'   data repositories, `without altering code` between the first, second, third, etc., times
#'   the code is run.}
#' }
#'
#' @section Package options:
#'
#' See [reproducibleOptions()] for a complete description of package
#' [options()] to configure behaviour.
#'
#' @import methods
"_PACKAGE"

#' Low-level functions to create and work with a cache
#'
#' **These are intended for advanced use only.**
#'
#' @param cachePath A path describing the directory in which to create
#'   the database file(s)
#'
#' @inheritParams Cache
#'
#' @param drv A driver, passed to `dbConnect`
#'
#' @param force Logical. Should it create a cache in the `cachePath`,
#'   even if it already exists, overwriting.
#'
#' @details
#' - `createCache()` will create a Cache folder structure and necessary files, based on
#' the particular `drv` or `conn` provided;
#'
#' @return
#' - `createCache()` returns `NULL` (invisibly) and intended to be called for side effects;
#'
#' @export
#' @importFrom data.table data.table
#' @rdname CacheHelpers
#'
#' @examples
#' data.table::setDTthreads(2)
#' newCache <- tempdir2()
#' createCache(newCache)
#'
#' out <- Cache(rnorm(1), cachePath = newCache)
#' cacheId <- gsub("cacheId:", "", attr(out, "tags"))
#' loadFromCache(newCache, cacheId = cacheId)
#'
#' rmFromCache(newCache, cacheId = cacheId)
#'
#' # clean up
#' unlink(newCache, recursive = TRUE)
#'
createCache <- function(cachePath = getOption("reproducible.cachePath"),
                        drv = getDrv(getOption("reproducible.drv", NULL)),
                        conn = getOption("reproducible.conn", NULL), force = FALSE,
                        verbose = getOption("reproducible.verbose")) {
  alreadyExists <- CacheIsACache(cachePath, drv = drv, conn = conn, create = TRUE, verbose = verbose)
  if (alreadyExists && force == FALSE) {
    messageCache("Cache already exists at ", cachePath, " and force = FALSE. Not creating new cache.",
      verbose = verbose
    )
    return(invisible(cachePath))
  }

  checkPath(cachePath, create = TRUE)
  checkPath(CacheStorageDir(cachePath), create = TRUE)
  if (useDBI()) {
    .createCache(cachePath = cachePath, drv = drv, conn = conn)
  }

  invisible(NULL)
}

#' @keywords internal
.createCache <- function(cachePath, drv, conn) {
  if (is.null(conn)) {
    conn <- dbConnectAll(drv, cachePath = cachePath)
    on.exit(DBI::dbDisconnect(conn))
  }
  dt <- .emptyCacheTable

  # Some tough to find cases where stalls on dbWriteTable -- this *may* prevent some
  a <- retry(
    retries = 250, exponentialDecayBase = 1.01,
    quote(DBI::dbListTables(conn))
  )

  if (isTRUE(!CacheDBTableName(cachePath, drv) %in% a)) {
    # retry(retries = 5, exponentialDecayBase = 1.5, quote(
    try(DBI::dbWriteTable(conn, CacheDBTableName(cachePath, drv), dt,
      overwrite = FALSE,
      field.types = c(
        cacheId = "text", tagKey = "text",
        tagValue = "text", createdDate = "text"
      )
    ), silent = TRUE)
  }
  # )
}

#' Save an object to Cache
#'
#' This is not expected to be used by a user as it requires that the `cacheId` be
#' calculated in exactly the same as it calculated inside `Cache`
#' (which requires `match.call` to match arguments with their names, among other things).
#'
#' @inheritParams Cache
#'
#' @param cacheId The hash string representing the result of `.robustDigest`
#'
#' @param obj The R object to save to the cache
#'
#' @param linkToCacheId Optional. If a `cacheId` is provided here, then a `file.link`
#'   will be made to the file with that `cacheId` name in the cache repo.
#'   This is used when identical outputs exist in the cache. This will save disk space.
#'
#' @return
#' This is used for its side effects, namely, it will add the object to the cache and
#' cache database.
saveToCache <- function(cachePath = getOption("reproducible.cachePath"),
                        cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                        drv = getDrv(getOption("reproducible.drv", NULL)),
                        conn = getOption("reproducible.conn", NULL), obj, userTags, cacheId,
                        linkToCacheId = NULL,
                        verbose = getOption("reproducible.verbose")) {

  # saveToCache can be coming from a few places, not just Cache
  if (cacheSaveFormat %in% c(.qsFormat))
    cacheSaveFormat <- getOption("reproducible.qsFormat", .qs2Format)

  if (useDBI()) {
    if (is.null(conn)) {
      conn <- dbConnectAll(drv, cachePath = cachePath)
      on.exit(DBI::dbDisconnect(conn))
    }
    if (is.list(conn))
      conn <- conn[[cachePath]]
  }

  if (missing(userTags)) userTags <- otherFunctions
  if (length(userTags) == 0) userTags <- otherFunctions
  if (NCOL(userTags) > 1) {
    tagKey <- userTags$tagKey
    tagValue <- userTags$tagValue
  } else {
    tagKey <- sub(userTags, pattern = ":.*$", replacement = "")
    tagValue <- sub(userTags, pattern = "^[^:]*:", replacement = "")
  }

  fts <- CacheStoredFile(cachePath, cacheId, obj = obj, cacheSaveFormat = cacheSaveFormat) # this includes the extra files

  # TRY link first, if there is a linkToCacheId, but some cases will fail; not sure what these cases are
  if (!is.null(linkToCacheId)) {
    ftL <- CacheStoredFile(cachePath, linkToCacheId, obj = obj, cacheSaveFormat = cacheSaveFormat,
                           readOnly = TRUE)
    ftLfs <- file.size(ftL)
    out <- if (isTRUE(all(ftLfs > 0))) {# means corrupted if file.size is 0
      suppressWarnings({
        try(file.link(from = ftL, to = fts), silent = TRUE)
      })
    } else {
      # maybe could be deleting those files here because they are corrupted;
      #  but should happen whenever they are needed
      FALSE
    }

    if (is(out, "try-error") || !all((out %in% TRUE))) {
      linkToCacheId <- NULL
    } else {
      .message$FileLinkUsed(ftL, fts, verbose)
      # messageCache("  (A file with identical properties already exists in the Cache: ", basename(ftL), "; ")
      # messageCache("    The newly added (", basename(fts), ") is a file.link to that file)",
      #   verbose = verbose
      # )
    }
    fs <- file.size(fts)
  }

  # Save to db file first, then storage file
  dt <- metadataDT(cacheId, tagKey, tagValue)

  # dt <- data.table(
  #   "cacheId" = cacheId, "tagKey" = tagKey,
  #   "tagValue" = tagValue, "createdDate" = as.character(Sys.time())
  # )
  if (!useDBI()) {
    dtFile <- saveDBFileSingle(dt = dt, cachePath, cacheId, cacheSaveFormat = cacheSaveFormat)
  } else {
    # fl <- "/home/emcintir/tmp/usingDBI.rds"
    # usingDBI <- if (file.exists(fl)) readRDS(fl) else 1
    # usingDBI <- ifelse(length(usingDBI) < 1, 1, usingDBI  + 1)
    # saveRDS(usingDBI, file = fl)
    a <- retry(retries = 250, exponentialDecayBase = 1.01, quote(
      DBI::dbAppendTable(conn, CacheDBTableName(cachePath, drv), dt)
    ))
  }

  if (is.null(linkToCacheId)) {
    fs <- saveFilesInCacheFolder(cachePath = cachePath, obj, fts, cacheId = cacheId,
                                 cacheSaveFormat = cacheSaveFormat)
  }
  if (isTRUE(getOption("reproducible.useMemoise"))) {
    obj <- .unwrap(obj, cachePath, cacheId, drv, conn) # This takes time, but whether it happens now or later, same
    obj2 <- makeMemoisable(obj)
    assign(cacheId, obj2, envir = memoiseEnv(cachePath))
  }

  fsChar <- as.character(fs)

  tagKeyHasFS <- tagKey %in% "file.size"
  if (isFALSE(any(tagKeyHasFS))) {
    tagKey <- c(tagKey, "file.size")
    tagValue <- c(tagValue, fsChar)
  } else {
    tagValue[tagKeyHasFS] <- fsChar
  }

  # Compare the file size with the object size -- to test for "captured environments"
  #  There is a buffer of 4x, plus file sizes are smaller than binary size with qs defaults
  #  So effectively, it is like 6x buffer to try to avoid false positives.
  if (isTRUE(sum(fs) > 1e4)) {
    whichOS <- which(tagKey == "object.size")
    if (length(whichOS)) {
      objSize <- if (identical(unname(tagValue[whichOS]), "NA")) NA else as.numeric(tagValue[whichOS])
      fsBig <- (objSize * 4) < fs

      if (isTRUE(fsBig)) {
        messageCache("Object with cacheId ", cacheId, " appears to have a much larger size ",
                     "on disk than in memory. ",
                     "This usually means that the object has captured an environment with ",
                     "many objects due to how a function or a formula is defined. ",
                     "Usually, a solution involves using quote and eval around the formulas ",
                     "and defining functions in a package or otherwise clean space, ",
                     "i.e., not inside another function.\n",
                     "See http://adv-r.had.co.nz/memory.html#gc and 'capturing environments'.",
                     verbose = verbose
        )
      }
    }
  }

  return(obj)
}

#' @inheritParams CacheStoredFile
#'
#' @param fullCacheTableForObj The result of `showCache`, but subsetted for only
#'   the `cacheId` being loaded or selected
#'
#' @param .dotsFromCache Optional. Used internally.
#'
#' @param .functionName Optional. Used for messaging when this function is called from `Cache`
#'
#' @param preDigest The list of `preDigest` that comes from `CacheDigest` of an object
#'
#' @details
#' - `loadFromCache()` retrieves a single object from the cache, given its `cacheId`;
#'
#' @return
#' - `loadFromCache()` returns the object from the cache that has the particular `cacheId`;
#'
#' @export
#' @rdname CacheHelpers
loadFromCache <- function(cachePath = getOption("reproducible.cachePath"),
                          cacheId, preDigest,
                          fullCacheTableForObj = NULL,
                          cacheSaveFormat = getOption("reproducible.cacheSaveFormat", .rdsFormat),
                          .functionName = NULL, .dotsFromCache = NULL,
                          drv = getDrv(getOption("reproducible.drv", NULL)),
                          conn = getOption("reproducible.conn", NULL),
                          verbose = getOption("reproducible.verbose")) {
  if (verbose > 3) {
    startLoadTime <- Sys.time()
  }

  if (length(cacheId) > 1) {
    cacheId <- unique(cacheId)
  }

  isMemoised <- .isMemoised(cacheId, cachePath = cachePath)
  # isMemoised <- NA
  # if (isTRUE(getOption("reproducible.useMemoise"))) {
  #   isMemoised <- exists(cacheId, envir = memoiseEnv(cachePath))
  if (isTRUE(isMemoised)) {
    obj <- get(cacheId, envir = memoiseEnv(cachePath))
    obj <- unmakeMemoisable(obj)
  }
  # }
  if (cacheSaveFormat %in% c(.qsFormat))
    cacheSaveFormat <- getOption("reproducible.qsFormat", .qs2Format)

  if (!isTRUE(isMemoised)) {
    # Put this in a loop -- try the cacheSaveFormat that the user requested, but switch back if can't do it
    for (i in 1:2) {
      f <- CacheStoredFile(cachePath, cacheId, cacheSaveFormat, readOnly = TRUE)
      f <- unique(f) # It is OK if there is a vector of unique cacheIds e.g., loadFromCache(showCache(userTags = "hi")$cacheId)

      # First test if it is correct cacheSaveFormat
      obj <- loadFromCacheSwitchFormat(f, verbose, cachePath, fullCacheTableForObj, cacheId, preDigest, drv, conn)
      if (!is.null(obj))
        return(obj)
      # if (!all(file.exists(f))) {
      #   sameCacheID <- dir(dirname(f), pattern = filePathSansExt(basename(f)))
      #   if (!useDBI() || length(sameCacheID) > 1) {
      #     sameCacheID <- onlyStorageFiles(sameCacheID)
      #   }
      #
      #   if (length(sameCacheID)) {
      #     # if (!identical(.whereInStack("sim"), .GlobalEnv)) {
      #     #   cacheSaveFormat <- setdiff(c(.rdsFormat, .qsFormat), cacheSaveFormat)
      #     #   message("User tried to change options('reproducible.cacheSaveFormat') for an ",
      #     #           "existing cache, while using a simList. ",
      #     #           "This currently does not work. Keeping the ",
      #     #           "option at: ", cacheSaveFormat)
      #     #   next
      #     # }
      #
      #     messageCache(.message$changingFormat(prevFile = sameCacheID, newFile = f),
      #                  verbose = verbose)
      #     #messageCache("     (Changing cacheSaveFormat of Cache entry from ", fileExt(sameCacheID), " to ",
      #     #             fileExt(f), ")",
      #     obj <- loadFromCache(
      #       cachePath = cachePath, fullCacheTableForObj = fullCacheTableForObj,
      #       cacheId = cacheId,
      #       cacheSaveFormat = fileExt(sameCacheID),
      #       preDigest = preDigest,
      #       verbose = verbose
      #     )
      #
      #     obj2 <- .wrap(obj, cachePath = cachePath, drv = drv, conn = conn)
      #     fs <- saveToCache(
      #       obj = obj2, cachePath = cachePath, drv = drv, conn = conn,
      #       cacheId = cacheId
      #     )
      #     rmFromCache(
      #       cachePath = cachePath, cacheId = cacheId, drv = drv, conn = conn,
      #       cacheSaveFormat = fileExt(sameCacheID)
      #     )
      #     return(obj)
      #   }
      # }
      # Need exclusive lock

      obj <- loadFile(f, cacheId = cacheId, cachePath = cachePath,
                      drv = drv, conn = conn, verbose = verbose)#, cacheSaveFormat = cacheSaveFormat)
      obj <- .unwrap(obj,
                     cachePath = cachePath,
                     cacheId = cacheId,
                     drv = drv, conn = conn
      )
      break # if you got this far, then break out of the for i loop
    }
  }

  # Class-specific message
  useMemoise <- if (getOption("reproducible.useMemoise") %in% TRUE) TRUE else NA
  fromMemoise <- isMemoised && useMemoise
  loadFromMgs <- .cacheMessage(obj, .functionName, fromMemoise = fromMemoise, verbose = verbose)

  # bug that affects Caching of functions that have an argument called "objects": PR#403
  if ("object" %in% names(.dotsFromCache))
    .dotsFromCache <- .dotsFromCache[setdiff(names(.dotsFromCache), "object")]

  obj <- do.call(.prepareOutput, args = append(list(object = obj, cachePath), .dotsFromCache))

  if (isTRUE(useMemoise) && !isTRUE(isMemoised)) {
  # if (isTRUE(getOption("reproducible.useMemoise")) && !isTRUE(isMemoised)) {
    obj2 <- makeMemoisable(obj)
    assign(cacheId, obj2, envir = memoiseEnv(cachePath))
  }

  if (verbose > 3) {
    endLoadTime <- Sys.time()
    verboseDF <- data.frame(
      functionName = .functionName,
      component = gsub("(.+)(ed)(.+) result from.+$", "\\1ing\\3", loadFromMgs),
      elapsedTime = as.numeric(difftime(endLoadTime, startLoadTime, units = "secs")),
      units = "secs",
      stringsAsFactors = FALSE
    )

    if (exists("verboseTiming", envir = .reproEnv)) {
      .reproEnv$verboseTiming <- rbind(.reproEnv$verboseTiming, verboseDF)
    }
  }

  obj
}

#' @param sc a cache tags `data.table` object
#' @param elem character string specifying a `tagKey` value to match
#' @param ifNot character (or NULL) specifying the return value to use if `elem` not matched
#'
#' @details
#' - `extractFromCache()` retrieves a single `tagValue` from the cache based on
#' the `tagKey` of `elem`;
#'
#' @return
#' - `extractFromCache()` returns the `tagValue` from the cache corresponding to `elem` if found,
#' otherwise the value of `ifNot`;
#'
#' @export
#' @rdname CacheHelpers
extractFromCache <- function(sc, elem, ifNot = NULL) {
  rowNum <- sc[["tagKey"]] %in% elem
  elemExtracted <- if (any(rowNum)) {
    sc[["tagValue"]][rowNum]
  } else {
    ifNot
  }
  elemExtracted
}

#' @details
#' - `rmFromCache()` removes one or more items from the cache, and updates the cache
#' database files.
#'
#' @return
#' - `rmFromCache()` returns `NULL` (invisibly) and is intended to be called for side effects;
#'
#' @export
#' @inheritParams loadFile
#' @rdname CacheHelpers
rmFromCache <- function(cachePath = getOption("reproducible.cachePath"),
                        cacheId, drv = getDrv(getOption("reproducible.drv", NULL)),
                        conn = getOption("reproducible.conn", NULL),
                        cacheSaveFormat = getOption("reproducible.cacheSaveFormat", .rdsFormat),
                        verbose = getOption("reproducible.verbose"), ...) {
  # backwards compatibility
  if (!is.null(list(...)$format))
    cacheSaveFormat <- list(...)$format

  if (useDBI()) {
    if (is.null(conn)) {
      conn <- dbConnectAll(drv, cachePath = cachePath, create = FALSE)
      on.exit(DBI::dbDisconnect(conn))
    }
    # from https://cran.r-project.org/web/packages/DBI/vignettes/spec.html
    query <- glue::glue_sql(
      "DELETE FROM {DBI::SQL(glue::double_quote(dbTabName))} WHERE \"cacheId\" IN ({cacheId*})",
      dbTabName = CacheDBTableName(cachePath, drv),
      cacheId = cacheId,
      .con = conn
    )
    res <- DBI::dbSendQuery(conn, query)

    if (FALSE) { # this is the "unsafe" version
      query <- paste0("DELETE FROM \"", CacheDBTableName(cachePath, drv), "\" WHERE \"cacheId\" = $1")
      res <- DBI::dbSendStatement(conn, query)
      DBI::dbBind(res, list(cacheId))
    }

    DBI::dbClearResult(res)
  } else {
    dtFile <- CacheDBFileSingle(cachePath = cachePath, cacheId = cacheId, cacheSaveFormat = cacheSaveFormat)
    unlink(dtFile)
  }
  unlink(CacheStoredFile(cachePath, cacheId = cacheId, cacheSaveFormat = cacheSaveFormat, readOnly = TRUE))
}

dbConnectAll <- function(drv = getDrv(getOption("reproducible.drv", NULL)),
                         cachePath = getOption("reproducible.cachePath"),
                         conn = getOption("reproducible.conn", NULL), create = TRUE,
                         verbose = getOption("reproducible.verbose")) {
  args <- list(drv = drv)
  if (is(drv, "SQLiteDriver")) {
    args <- append(args, list(
      dbname = CacheDBFile(cachePath, drv = drv, conn = conn),
      synchronous = NULL
    ))
  }
  conn <- try(do.call(DBI::dbConnect, args), silent = TRUE)
  if (is(conn, "try-error")) {
    messageCache("There is no Cache at this location", verbose = verbose)
    return(invisible(NULL))
  }
  conn
}

.emptyCacheTable <- data.table::data.table(
  cacheId = character(), tagKey = character(),
  tagValue = character(), createdDate = character()
)



#' Add a Tag to a Cached Object in the Repository
#'
#' This hidden function appends a single tag (key-value pair) to the metadata
#' of a cached object identified by its `cacheId`. Tags can be stored either in
#' a database (via DBI) or in a file-based cache system.
#'
#' @param cacheId `character(1)`
#'   The unique identifier of the cached object. Must be of length 1.
#' @param cachePath `character(1)`
#'   Path to the cache directory. Defaults to `getOption("reproducible.cachePath")`.
#' @param tagKey `character(1)`
#'   The key for the tag. Defaults to `"accessed"` if not provided.
#' @param tagValue `character(1)`
#'   The value for the tag. Defaults to the current system time if not provided.
#' @param cacheSaveFormat `character(1)`
#'   Format used for saving cache files. Defaults to `getOption("reproducible.cacheSaveFormat")`.
#' @param drv A DBI driver object. Defaults to `getDrv(getOption("reproducible.drv", NULL))`.
#' @param conn A DBI connection object. If `NULL`, a new connection is created internally.
#' @param verbose `integer` or `logical`. If `> 0` or `TRUE`, print informational messages.
#'   Defaults to `getOption("reproducible.verbose")`.
#'
#' @details
#' This function is primarily used internally by the `reproducible` package to
#' maintain metadata about cached objects. It supports both database-backed and
#' file-based caching systems.
#'
#' @return `NULL` (invisibly). The function is called for its side effects.
#'
#' @rdname addTags
#' @examples
#' \dontrun{
#' a <- Cache(rnorm(1))
#' .addTagsRepo(cacheId = gsub("cacheId:", "", attr(a, "tags")),
#'              tagKey = "status", tagValue = "processed")
#' showCache() # last entry is the above line
#' }
#'
#' @export
.addTagsRepo <- function(cacheId, cachePath = getOption("reproducible.cachePath"),
                         tagKey = character(), tagValue = character(),
                         cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                         drv = getDrv(getOption("reproducible.drv", NULL)),
                         conn = getOption("reproducible.conn", NULL),
                         verbose = getOption("reproducible.verbose")) {
  if (length(cacheId) > 0) {
    if (length(cacheId) > 1) stop(".addTagsRepo can only handle appending 1 tag at a time")
    curTime <- as.character(Sys.time())
    if (length(tagKey) < length(cacheId)) {
      tagKey <- "accessed"
    }
    if (length(tagValue) < length(cacheId)) {
      tagValue <- curTime
    }

    if (useDBI()) {
      if (is.null(conn)) {
        conn <- dbConnectAll(drv, cachePath = cachePath, create = FALSE)
        on.exit(DBI::dbDisconnect(conn))
      }

      # This is what the next code pair of lines does
      # dt <- data.table("cacheId" = cacheId, "tagKey" = "accessed",
      #                 "tagValue" = as.character(Sys.time()),
      #                 "createdDate" = as.character(Sys.time()))
      #
      # retry(quote(dbAppendTable(conn, CacheDBTableName(cachePath, drv), dt), retries = 15))
      rs <- retry(retries = 250, exponentialDecayBase = 1.01, quote(
        DBI::dbSendStatement(
          conn,
          paste0(
            "insert into \"", CacheDBTableName(cachePath, drv), "\"",
            " (\"cacheId\", \"tagKey\", \"tagValue\", \"createdDate\") values ",
            "('", cacheId,
            "', '", tagKey, "', '", tagValue, "', '", curTime, "')"
          )
        )
      ))

      DBI::dbClearResult(rs)
    } else {
      dt <- list(
        "cacheId" = cacheId, "tagKey" = tagKey,
        "tagValue" = tagValue,
        "createdDate" = as.character(Sys.time())
      )
      # dt <- data.table(
      #   "cacheId" = cacheId, "tagKey" = tagKey,
      #   "tagValue" = tagValue,
      #   "createdDate" = as.character(Sys.time())
      # )
      dtFile <- CacheDBFileSingle(cachePath = cachePath, cacheId = cacheId, cacheSaveFormat = cacheSaveFormat)
      dt2 <- loadFile(dtFile,
                      cacheId = cacheId, cachePath = cachePath, # in case it needs swapCacheFormat
                      drv = drv, conn = conn, verbose = verbose)
      dt <- rbindlist(list(dt2, dt), fill = TRUE)
      saveFilesInCacheFolder(dt, dtFile, cachePath = cachePath, cacheId = cacheId,
                             cacheSaveFormat = cacheSaveFormat)
    }
  }
}


#' Update or Add a Tag for a Cached Object
#'
#' Updates the value of an existing tag for a cached object identified by its
#' `cacheId`. If the tag does not exist and `add = TRUE`, the tag will be added.
#' This function supports both database-backed and file-based cache systems.
#'
#' @param cacheId `character(1)`
#'   Unique identifier of the cached object. Must be of length 1.
#' @param cachePath `character(1)`
#'   Path to the cache directory. Defaults to `getOption("reproducible.cachePath")`.
#' @param tagKey `character(1)`
#'   The key for the tag. Must be supplied.
#' @param tagValue `character(1)`
#'   The new value for the tag. Must be supplied.
#' @param add `logical(1)`
#'   If `TRUE`, adds the tag if it does not exist. Defaults to `TRUE`.
#' @param cacheSaveFormat `character(1)`
#'   Format used for saving cache files. Defaults to `getOption("reproducible.cacheSaveFormat")`.
#' @param drv A DBI driver object. Defaults to `getDrv(getOption("reproducible.drv", NULL))`.
#' @param conn A DBI connection object. If `NULL`, a new connection is created internally.
#'
#' @details
#' - If `useDBI()` returns `TRUE`, the tag update is performed in the database table.
#' - If no rows are affected and `add = TRUE`, the tag is inserted using `.addTagsRepo()`.
#' - For file-based caches, the function modifies the tag in the corresponding metadata file.
#'
#' @return `NULL` (invisibly). Called for its side effects.
#'
#' @seealso [`.addTagsRepo()`] for adding tags without updating.
#'
#' @rdname addTags
#' @examples
#' \dontrun{
#' a <- Cache(rnorm(1))
#' # Update an existing tag
#' .updateTagsRepo(cacheId = gsub("cacheId:", "", attr(a, "tags")),
#'              tagKey = "status", tagValue = "second")
#'
#' # Add a tag if it doesn't exist
#' .updateTagsRepo(cacheId = gsub("cacheId:", "", attr(a, "tags")),
#'              tagKey = "status", tagValue = "new", add = TRUE)
#' }
#'
#' @export
.updateTagsRepo <- function(cacheId, cachePath = getOption("reproducible.cachePath"),
                            tagKey = character(), tagValue = character(),
                            add = TRUE,
                            cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                            drv = getDrv(getOption("reproducible.drv", NULL)),
                            conn = getOption("reproducible.conn", NULL),
                            verbose = getOption("reproducible.verbose")) {
  if (length(cacheId) > 0) {
    curTime <- as.character(Sys.time())
    if (length(tagKey) < length(cacheId)) {
      warning("tagKey and/or tagValue must both be supplied for .updateTagsRepo.")
      return(invisible())
    }
    if (length(cacheId) > 1) stop(".updateTagsRepo can only handle updating 1 tag at a time")
    if (useDBI()) {
      if (is.null(conn)) {
        conn <- dbConnectAll(drv, cachePath = cachePath, create = FALSE)
        on.exit(DBI::dbDisconnect(conn))
      }

      # This is what the next code pair of lines does
      # dt <- data.table("cacheId" = cacheId, "tagKey" = "accessed",
      #                 "tagValue" = as.character(Sys.time()),
      #                 "createdDate" = as.character(Sys.time()))
      #
      # retry(quote(dbAppendTable(conn, CacheDBTableName(cachePath, drv), dt), retries = 15))
      rs <- # retry(retries = 250, exponentialDecayBase = 1.01, quote(
        DBI::dbSendStatement(
          conn,
          paste0(
            "update \"", CacheDBTableName(cachePath, drv), "\"",
            " set \"tagValue\" = '", tagValue, "' where ",
            " \"cacheId\" = '", cacheId, "'", " AND \"tagKey\" = '", tagKey, "'"
          )
        )
      # ))
      affectedAnyRows <- DBI::dbGetRowsAffected(rs) > 0
      DBI::dbClearResult(rs)
      if (!affectedAnyRows) {
        if (isTRUE(add)) {
          .addTagsRepo(cacheId, cachePath, tagKey, tagValue, cacheSaveFormat = cacheSaveFormat,
                       drv = drv, conn = conn)
        }
      }
    } else {
      dt <- data.table(
        "cacheId" = cacheId, "tagKey" = tagKey,
        "tagValue" = tagValue,
        "createdDate" = as.character(Sys.time())
      )
      dtFile <- CacheDBFileSingle(cachePath = cachePath, cacheId = cacheId, cacheSaveFormat = cacheSaveFormat)
      dt3 <- loadFile(dtFile,
                      cacheId = cacheId, cachePath = cachePath, # in case it needs swapCacheFormat
                      drv = drv, conn = conn, verbose = verbose)
      tk <- tagKey
      alreadyThere <- sum(dt3$tagKey == tk & dt3$cacheId == cacheId)
      if (add && alreadyThere == 0) {
        dt3 <- rbindlist(list(dt3, dt), fill = TRUE)
      } else {
        set(dt3, which(dt3$tagKey == tk & dt3$cacheId == cacheId), "tagValue", dt$tagValue)
        # dt3[tagKey == tk & cacheId == cacheId, tagValue := dt$tagValue]
      }
      saveFilesInCacheFolder(dt3, dtFile, cachePath = cachePath, cacheId = cacheId,
                             cacheSaveFormat = cacheSaveFormat)
    }
  }
}



.cacheTagsSecondGroup <- c("class", "object.size", "fromDisk", "resultHash", "elapsedTimeFirstRun")
.cacheTagsFirstGroup <- c("function", "userTags", "accessed", "inCloud", "elapsedTimeDigest", "preDigest")

.cacheTagsDefault <- c(.cacheTagsFirstGroup, .cacheTagsSecondGroup)

.cacheNumDefaultTags <- function() {
  length(setdiff(.cacheTagsDefault, .ignoreTagKeys())) # currently 7
}

.ignoreTagKeys <- function() {
  c("preDigest", otherFunctions, "accessed", "elapsedTimeLoad", "elapsedTimeFirstRun",
    "fromDisk", "origRaster", "cacheRaster", "userTags",
    "cacheId")
}

.cacheTableHashColName <- function() {
  "cacheId"
}

.cacheTableTagColName <- function(option = NULL) {
  "tagValue"
}

#' @inheritParams Cache
#'
#' @inheritParams createCache
#'
#' @return
#' - `CacheDBFile()` returns the name of the database file for a given Cache,
#' when `useDBI() == FALSE`, or `NULL` if `TRUE`;
#' - `CacheDBFiles()` (i.e,. plural) returns the name of all the database files for
#' a given Cache when `useDBI() == TRUE`, or `NULL` if `FALSE`;
#' - `CacheStoredFile()` returns the file path to the file with the specified hash value,
#' This can be loaded to memory with e.g., [loadFile()].;
#'
#' @export
#' @rdname CacheHelpers
#'
#' @examples
#' data.table::setDTthreads(2)
#' newCache <- tempdir2()
#'
#' # Given the drv and conn, creates the minimum infrastructure for a cache
#' createCache(newCache)
#'
#' CacheDBFile(newCache) # identifies the database file
#' CacheStorageDir(newCache) # identifies the directory where cached objects are stored
#'
#' out <- Cache(rnorm(1), cachePath = newCache)
#' cacheId <- gsub("cacheId:", "", attr(out, "tags"))
#' CacheStoredFile(newCache, cacheId = cacheId)
#'
#' # The name of the table inside the SQL database
#' CacheDBTableName(newCache)
#'
#' CacheIsACache(newCache) # returns TRUE
#'
#' # clean up
#' unlink(newCache, recursive = TRUE)
CacheDBFile <- function(cachePath = getOption("reproducible.cachePath"),
                        drv = getDrv(getOption("reproducible.drv", NULL)),
                        conn = getOption("reproducible.conn", NULL)) {
  type <- gsub("Driver", "", class(drv))

  if (useDBI()) {
    if (!is.null(conn)) {
      type <- gsub("Connection", "", class(conn))
    }
    #   }

    if (grepl(type, "SQLite")) {
      file.path(cachePath, "cache.db")
    } else {
      file.path(cachePath, "cache.txt")
    }
  } else {
    file.path(cachePath, "multifileDB.txt")
  }
}

#' @return
#' - `CacheStorageDir()` returns the name of the directory where cached objects are stored;
#'
#' @export
#' @rdname CacheHelpers
CacheStorageDir <- function(cachePath = getOption("reproducible.cachePath")) {
  file.path(cachePath, "cacheOutputs")
}

#' @param obj The optional object that is of interest; it may have an attribute "saveRawFile"
#'   that would be important.
#'
#' @param cacheId The cacheId or otherwise digested hash value, as character string.
#'
#' @param cacheSaveFormat The text string representing the file extension used normally by
#'   different save formats; currently only `"rds"` or `"qs"` (which now uses `qs2` package.
#'   Defaults to `getOption("reproducible.cacheSaveFormat", "rds")`
#' @param readOnly Logical. Only relevant during transition from `qs` to `qs2`.
#'   Essentially, during transition, `qs` objects can be read, but not saved.
#'   If `TRUE` then the `CacheStoredFile` can return a `.qs` file; if `FALSE`,
#'   then this will not be able to return `qs`; instead it will return `qs2`
#'   files.
#'
#' @return
#' - `CacheStoredFile` returns the file path to the file with the specified hash value;
#'
#' @export
#' @rdname CacheHelpers
CacheStoredFile <- function(cachePath = getOption("reproducible.cachePath"), cacheId,
                            cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                            obj = NULL, readOnly = FALSE) {
  # if (is.null(cacheSaveFormat)) cacheSaveFormat <- getOption("reproducible.cacheSaveFormat", .rdsFormat)
  if (missing(cacheId)) cacheId <- NULL
  if (any(cacheSaveFormat %in% "check")) {
    cacheSaveFormat <- formatCheck(cachePath, cacheId, cacheSaveFormat = cacheSaveFormat)
  }
  csf <- cacheSaveFormat
  # qs <- grep(.qsFormat, .cacheSaveFormats, value = TRUE, ignore.case = TRUE)
  # rds <- grep(.rdsFormat, .cacheSaveFormats, value = TRUE, ignore.case = TRUE)
  csExtension <- if (isTRUE(any(c(.qsFormat, .qs2Format) %in% csf))) {
    if (isTRUE(readOnly)) {
      csf
    } else {
      qsForm <- getOption("reproducible.qsFormat", .qs2Format)
      qsForm
    }
  } else if (.rdsFormat %in% csf) {
    # convert from qs to qs2
    csf
  } else {
    if (is.character(cacheSaveFormat)) {
      cacheSaveFormat
    } else {
      .rdsFormat
    }
  }

  filename <- if (is.null(cacheId)) NULL else paste(cacheId, csExtension, sep = ".")

  if (length(cacheId) > 1) {
    filename <- vapply(filename, nextNumericName, FUN.VALUE = character(1))
    for (i in seq(filename[-1]) + 1) {
      filename[i] <- basename2(nextNumericName(filename[i - 1]))
    }
  }

  fns <- Filenames(obj, allowMultiple = TRUE)
  if (length(fns)) {
    dirs <- fs::is_dir(fns)
    if (isTRUE(any(dirs)))
      fns <- fns[dirs %in% FALSE]
  }
  fnsExtras <- basename2(fns)
  fnsExtras <- fnsExtras[nzchar(fnsExtras)]
  fnsExtras <- filenameInCacheWPrefix(fnsExtras, cacheId)
  file.path(CacheStorageDir(cachePath), c(filename, fnsExtras))
}

#' @return
#' - `CacheDBTableName()` returns the name of the table inside the SQL database, if that
#' is being used;
#'
#' @export
#' @rdname CacheHelpers
CacheDBTableName <- function(cachePath = getOption("reproducible.cachePath"),
                             drv = getDrv(getOption("reproducible.drv", NULL))) {
  if (!is(cachePath, "Path")) {
    cachePath <- asPath(cachePath, nParentDirs = 2)
  }
  if (useDBI()) {
    toGo <- attr(cachePath, "nParentDirs")
    cachePathTmp <- normPath(cachePath)
    newPath <- basename2(cachePathTmp)
    while (toGo > 1) {
      toGo <- toGo - 1
      cachePathTmp <- dirname(cachePathTmp)
      newPath <- paste(basename2(cachePathTmp), newPath, sep = "_")
    }
  } else {
    newPath <- "dt"
  }
  # SQLite can't handle numbers as initial character of a table name
  if (grepl("^[[:digit:]]", newPath)) {
    newPath <- paste0("_", newPath)
  }
  return(newPath)
}

#' @param create Logical. Currently only affects non \pkg{RSQLite} default drivers.
#'        If `TRUE` and there is no Cache database, the function will create one.
#'
#' @return
#' - `CacheIsACache()` returns a logical indicating whether the `cachePath` is currently
#' a `reproducible` cache database;
#'
#' @export
#' @rdname CacheHelpers
CacheIsACache <- function(cachePath = getOption("reproducible.cachePath"), create = FALSE,
                          drv = getDrv(getOption("reproducible.drv", NULL)),
                          conn = getOption("reproducible.conn", NULL),
                          verbose = getOption("reproducible.verbose")) {

  checkPath(cachePath, create = TRUE)
  if (useDBI()) {
    if (is.null(conn)) {
      conn <- dbConnectAll(drv, cachePath = cachePath)
      on.exit(DBI::dbDisconnect(conn))
    }
    type <- gsub("Connection", "", class(conn))
  }

  ret <- all(basename2(c(CacheDBFile(cachePath, drv, conn), CacheStorageDir(cachePath))) %in%
    list.files(cachePath))

  ## Need to check even if ret is TRUE because we may be in the process of changing
  convertDBbackendIfIncorrect(cachePath, drv, conn, verbose = verbose)

  needCreate <- FALSE
  if (useDBI()) {
    if (ret) {
      tablesInDB <- retry(
        retries = 250, exponentialDecayBase = 1.01,
        quote(DBI::dbListTables(conn))
      )
      tableShouldBe <- CacheDBTableName(cachePath)
      if (length(tablesInDB) == 1) {
        if (!any(tablesInDB %in% tableShouldBe) && grepl(type, "SQLite")) {
          warning(paste0(
            "The table in the Cache repo does not match the cachePath. ",
            "If this is because of a moved repository (i.e., files ",
            "copied), then it is being updated automatically. ",
            "If not, cache is in an error state. ",
            "You may need to delete the Cache"
          ))
          movedCache(cachePath, # old = tablesInDB,
            drv = drv, conn = conn
          )
        }
      }
      ret <- ret && any(grepl(tableShouldBe, tablesInDB))
    }

    if (isFALSE(ret) && isTRUE(create)) {
      if (grepl(type, "Pq")) {
        needCreate <- TRUE
      }
    }
  } else { # This is for DBI = FALSE
    if (isTRUE(create)) {
      needCreate <- TRUE
    }
  }
  if (isTRUE(needCreate)) {
    file.create(CacheDBFile(cachePath, drv = drv, conn = conn))
  }

  return(ret)
}

#' Deal with moved cache issues
#'
#' If a user manually copies a complete Cache folder (including the db file and rasters folder),
#' there are issues that must be addressed, depending on the Cache backend used.
#' If using DBI (e.g., RSQLite or Postgres), the db table must be renamed. Run
#' this function after a manual copy of a cache folder. See examples for one way to do that.
#'
#' @param  new Either the path of the new `cachePath` where the cache was moved or copied to, or
#'   the new DB Table Name
#' @param  old Optional, if there is only one table in the `new` cache path.
#'   Either the path of the previous `cachePath` where the cache was moved or copied from, or
#'   the old DB Table Name
#' @inheritParams Cache
#' @export
#' @details
#' When the backend database for a `reproducinle` cache is an SQL database, the files
#' on disk cannot be copied manually to a new location because they contain internal
#' tables. Because `reproducible` gives the main table a name based on the `cachePath`
#' path, calls to `Cache` will attempt to call this internally if it detects a
#' name mismatch.
#' @return
#' `movedCache` does not return anything; it is called for its side effects.
#'
#' @examples
#' data.table::setDTthreads(2)
#' tmpdir <- "tmpdir"
#' tmpCache <- "tmpCache"
#' tmpCacheDir <- normalizePath(file.path(tempdir(), tmpCache), mustWork = FALSE)
#' tmpdirPath <- normalizePath(file.path(tempdir(), tmpdir), mustWork = FALSE)
#' bb <- Cache(rnorm, 1, cachePath = tmpCacheDir)
#'
#' # Copy all files from tmpCache to tmpdir
#' froms <- normalizePath(dir(tmpCacheDir, recursive = TRUE, full.names = TRUE),
#'   mustWork = FALSE
#' )
#' dir.create(file.path(tmpdirPath, "rasters"), recursive = TRUE, showWarnings = FALSE)
#' dir.create(file.path(tmpdirPath, "cacheOutputs"), recursive = TRUE, showWarnings = FALSE)
#' file.copy(
#'   from = froms, overwrite = TRUE,
#'   to = gsub(tmpCache, tmpdir, froms)
#' )
#'
#' # Can use 'movedCache' to update the database table, though will generally
#' #   happen automatically, with message indicating so
#' movedCache(new = tmpdirPath, old = tmpCacheDir)
#' bb <- Cache(rnorm, 1, cachePath = tmpdirPath) # should recover the previous call
#'
movedCache <- function(new, old, drv = getDrv(getOption("reproducible.drv", NULL)),
                       conn = getOption("reproducible.conn", NULL),
                       verbose = getOption("reproducible.verbose")) {
  if (useDBI()) {
    if (is.null(conn)) {
      conn <- dbConnectAll(drv, cachePath = new)
      on.exit(DBI::dbDisconnect(conn))
    }

    tables <- DBI::dbListTables(conn)
    if (missing(old)) {
      if (length(tables) == 1) {
        messageCache("Assuming old database table is ", tables,
          verbose = verbose
        )
      } else {
        dbname <- try(conn@dbname, silent = TRUE)
        if (is(dbname, "try-error")) {
          dbname <- "conn"
        }
        stop("old not provided and there are more than one database table in ", )
      }
      old <- tables
      oldTable <- old
    } else {
      oldTable <- CacheDBTableName(old, drv = drv)
    }

    if (!any(tables == oldTable)) {
      stop("The 'old' table name does not appear inside the path to the 'new'")
    }
    newTable <- CacheDBTableName(new, drv = drv)

    qry <- glue::glue_sql("ALTER TABLE {`old`} RENAME TO {`new`}",
      old = oldTable,
      new = newTable,
      .con = conn
    )
    res <- retry(retries = 15, exponentialDecayBase = 1.01, quote(DBI::dbSendQuery(conn, qry)))
    # dbFetch(res)
    out <- DBI::dbClearResult(res)
  }
  return(invisible())
}


#' Load a file from the cache
#'
#' @param file `character(1)` specifying the path to the file to load.
#' @param cacheId `character(1)` the unique cache identifier of the object.
#' @param cachePath `character(1)` path to the cache directory.
#' @param drv A DBI driver object (e.g., `RSQLite::SQLite()`). Used when
#'   the cache format needs to be swapped.
#' @param conn A DBI connection object. If `NULL`, a new connection is
#'   created internally as needed.
#' @param verbose `integer` or `logical`. If `> 0` or `TRUE`, print
#'   informational messages. Defaults to `getOption("reproducible.verbose")`.
#' @param ... Allows `format` for backward compatibility.
#' @return the object loaded from `file`
#'
#' @export
loadFile <- function(file, cacheId, cachePath, # in case it needs swapCacheFormat
                     drv, conn, verbose = getOption("reproducible.verbose"), ...) {
  if (!is.null(list(...)$format))
    cacheSaveFormat <- list(...)$format
  else {
    # if (is.null(cacheSaveFormat)) {
    fe <- fileExt(file)
    cacheSaveFormat <- fe
  # }
  }
  qsFormats <- c(.qsFormat, .qs2Format)
  isQsAny <- cacheSaveFormat %in% qsFormats
  # isQs2 <- cacheSaveFormat %in% .qs2Format
  # isQsAny <- isQs | isQs2

  if (isQsAny) {
    csfs <- unique(c(cacheSaveFormat, qsFormats))
    for (csf in csfs) {# put in order by fileextension; but could be wrong
      if (csf %in% c(.qs2Format))
        .requireNamespace(.qs2Format, stopOnFALSE = TRUE)
      else
        .requireNamespace(.qsFormat, stopOnFALSE = TRUE)
      # obj <- qs2::qs_read(file[isQsAny], nthreads = getOption("reproducible.nThreads", 1))
      # if (FALSE) {
      funRead <- .fileExtsKnown()$fun[.fileExtsKnown()$extension == csf]
      funRead <- eval(parse(text = funRead))
        # obj <- funRead(file = file[isQsAny], nthreads = getOption("reproducible.nThreads", 1))
        # obj <- funRead(file = file[isQsAny], nthreads = getOption("reproducible.nThreads", 1))
      obj <- try(funRead(file = file[isQsAny], nthreads = getOption("reproducible.nThreads", 1)), silent = TRUE)
        # obj <- try2(funRead(file = file[isQsAny], nthreads = getOption("reproducible.nThreads", 1)), silent = TRUE)
      # }
        if (!is(obj, "try-error")) {
        if (!identical(csf, fe)) {
          if (missing(cachePath))
            cachePath <- dirname(dirname(file))
          newFile <- gsub(paste0(fe, "$"), csf, file)
          if (missing(cacheId))
            cacheId <- filePathSansExt(filePathSansExt(basename(file)))
          
          sameCacheID <- checkSameCacheId(newFile)
          swapCacheFileFormat(wrappedObj = obj, cachePath = cachePath,
                              cacheId = cacheId, sameCacheID = sameCacheID, newFile = newFile,
                              drv = drv, conn = conn, verbose = verbose)
          # saveFilesInCacheFolder(obj, ,
          #                        cachePath = dirname(dirname(file)),
          #                        cacheId = )
        }
        break
      } else {
        # if (csf %in% tail(csfs, 1))
        stop(obj) # this will be caught by most outer calls to loadFile
      }
    }
    # obj <- qs2::qs_read(file = file[isQsAny], nthreads = getOption("reproducible.nThreads", 1))
  } else {
    suppressWarningsSpecific(falseWarnings = "\\'package:stats\\' may not be available when loading",
                             obj <- readRDS(file = file[!isQsAny])
    )
  }

  obj
}

saveFilesInCacheFolder <- function(obj, fts, cachePath, cacheId,
                                   cacheSaveFormat = getOption("reproducible.cacheSaveFormat")) {

  if (missing(fts)) {
    fts <- CacheStoredFile(cachePath, cacheId = cacheId, obj = obj, cacheSaveFormat = cacheSaveFormat) # adds prefix
  }

  fsOther <- numeric()
  if (length(fts) > 1) {
    ftsOther <- fts[-1]
    fnsExtras <- Filenames(obj, allowMultiple = TRUE)

    fnsExtrasBase <- basename(fnsExtras)
    ftsOtherBase <- gsub(paste0(cacheId, "_"), "", basename(ftsOther))
    fnsExtras <- fnsExtras[match(ftsOtherBase, fnsExtrasBase)]

    # ftsOther <- filenameInCacheWPrefix(ftsOther, cacheId, relative = FALSE) # already done in CacheStoredFile
    # ftsOther <- .prefix(ftsOther, prefixCacheId(cacheId)) # makes it unique in the cache

    hardLinkOrCopy(fnsExtras, ftsOther, verbose = -2)
    fsOther <- sum(file.size(ftsOther))
    fts <- fts[1]
  }
  if (cacheSaveFormat %in% c(.qsFormat, .qs2Format)) {
    .requireNamespace(.qs2Format, stopOnFALSE = TRUE)
    for (attempt in 1) {
      # During transition from qs to qs2; the user should not be able to save in qs, so
      #  use a hidden option; otherwise developers can still use it for testing
      formatToUse <- getOption("reproducible.qsFormat", .qs2Format) # force qs2
      if (formatToUse %in% .qsFormat)
        .requireNamespace(.qsFormat, stopOnFALSE = TRUE)
      funSave <- .fileExtsKnown()[["saveFun"]][.fileExtsKnown()[["extension"]] == formatToUse]
      funSave <- eval(parse(text = funSave))
      fs <- funSave(obj,
                      file = fts,
                      nthreads = getOption("reproducible.nThreads", 1)
      )
      # if (is(fs, "try-error")) browser()
      # fs <- file.size(fts)
    }
  } else {
    suppressWarningsSpecific(falseWarnings = "\\'package:stats\\' may not be available when loading",
                             saveRDS(obj, file = fts)
    )
  }
  fs <- sum(file.size(fts))
  fs <- sum(fs, fsOther)


  fs
}

CacheDBFileSingle <- function(cachePath, cacheId,
                              cacheSaveFormat = getOption("reproducible.cacheSaveFormat")) {
  fullSuff <- CacheDBFileSingleExt(cacheSaveFormat = cacheSaveFormat)
  if (any(cacheSaveFormat %in% "check")) {
    cacheSaveFormat <- formatCheck(cachePath,
                                   paste0(cacheId, gsub("\\.$*", "", suffixMultipleDBFiles())),
                                   cacheSaveFormat = cacheSaveFormat)
    if (!is.null(cacheSaveFormat)) {
      fullSuff <- CacheDBFileSingleExt(cacheSaveFormat)
    }
  }
  out <- file.path(CacheStorageDir(cachePath), paste0(cacheId, fullSuff))
  out
}

CacheDBFileSingleExt <- function(cacheSaveFormat = getOption("reproducible.cacheSaveFormat")) {
  paste0(suffixMultipleDBFiles(), cacheSaveFormat)
}

suffixMultipleDBFiles <- function() {
  ".dbFile."
}

suffixLockFile <- function() ".lock"

onlyStorageFiles <- function(files, cacheId) {
  # files2 <- grep(gsub("\\.", "\\\\.", paste0(suffixMultipleDBFiles(), "|", suffixLockFile(), "$")),
  #   files,
  #   invert = TRUE, value = TRUE
  # )
  shouldBeFiles <- unname(sapply(.cacheSaveFormats, function(form)
    basename(CacheStoredFile(cacheId = cacheId, cacheSaveFormat = form, readOnly = TRUE))))
  shouldBeFiles <- paste(shouldBeFiles, collapse = "|")
  grep(shouldBeFiles, files, value = TRUE)
}

formatCheck <- function(cachePath, cacheId, cacheSaveFormat = getOption("reproducible.cacheSaveFormat")) {

  for (ci in .cacheSaveFormats) {
    ff <- CacheStoredFile(cachePath, cacheId, cacheSaveFormat = ci, readOnly = TRUE)
    if (file.exists(ff)) {
      newFormat <- ci
      break
    }
  }
  if (exists("newFormat", inherits = FALSE)) {
    cacheSaveFormat <- newFormat
  } else if (cacheSaveFormat == "check") { # means there was no file; possibly deleted inadvertently
    cacheSaveFormat <- getOption("reproducible.cacheSaveFormat") #getOption("reproducible.cacheSaveFormat")
  }

  # altFile <- dir(dirname(CacheStoredFile(cachePath, cacheId)), pattern = cacheId)
  # altFile <- onlyStorageFiles(altFile, cacheId)
  # if (length(altFile)) {
  #   cacheSaveFormat <- tools::file_ext(altFile)
  # } else if (cacheSaveFormat == "check") {
  #   cacheSaveFormat <- getOption("reproducible.cacheSaveFormat")
  # }
  cacheSaveFormat
}

getDrv <- function(drv = NULL) {
  if (useDBI()) {
    if (is.null(drv)) {
      if (!requireNamespace("RSQLite", quietly = TRUE)) {
        stop("Need RSQLite package when using DBI; install.packages('RSQLite')")
      }
      drv <- RSQLite::SQLite()
    }
  } else {
    drv <- NULL
  }
  drv
}

saveDBFileSingle <- function(dt, cachePath, cacheId,
                             cacheSaveFormat = getOption("reproducible.cacheSaveFormat")) {
  dtFile <- CacheDBFileSingle(cachePath = cachePath, cacheId = cacheId, cacheSaveFormat = cacheSaveFormat)
  saveFilesInCacheFolder(dt, dtFile, cachePath = cachePath, cacheId = cacheId,
                         cacheSaveFormat = cacheSaveFormat)
  dtFile
}

convertDBbackendIfIncorrect <- function(cachePath, drv, conn,
                                        cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                                        verbose = getOption("reproducible.verbose")) {
  origDrv <- getDrv(drv)
  origDBI <- useDBI(verbose = -1)
  newDBI <- useDBI(!origDBI, verbose = -1) # switch to the other
  if (!identical(newDBI, origDBI)) { # if they are same, then DBI is not installed; not point proceeding
    on.exit(suppressMessages(useDBI(origDBI, verbose = -1)))
    drv <- getDrv(drv) # This will return the DBI driver, if it is installed, regardless of drv
    DBFileWrong <- CacheDBFile(cachePath, drv, conn)
    if (file.exists(DBFileWrong)) {
      sc <- showCache(cachePath, drv = drv, conn = conn, verbose = -2)
      if (NROW(sc)) {
        messageCache("This cache repository previously was using a ",
          messConvert()[[as.character(useDBI())]], ".\n",
          "User has requested to change this using ",
          "e.g., `useDBI(", useDBI(), ")`. Converting now ...",
          verbose = verbose, verboseLevel = 1
        )
        if (isTRUE(origDBI)) { # using DBI --> convert all data to a DBI database
          suppressMessages(useDBI(origDBI))
          .createCache(cachePath, drv = origDrv, conn = conn)
          Map(tv = sc$tagValue, tk = sc$tagKey, oh = sc$cacheId, function(tv, tk, oh) {
            .addTagsRepo(
              cacheId = oh, cachePath = cachePath,
              tagKey = tk, tagValue = tv,
              cacheSaveFormat = cacheSaveFormat, drv = origDrv, conn = conn
            )
          })
          unlink(CacheDBFiles(cachePath))
        } else { # using multifile DB --> convert all data to multi-file backend
          singles <- split(sc, by = "cacheId")
          Map(dt = singles, ci = names(singles), function(dt, ci) {
            saveDBFileSingle(dt, cachePath = cachePath, cacheId = ci, cacheSaveFormat = cacheSaveFormat)
          })
        }
        messageCache("... Done!", verbose = verbose, verboseLevel = 1)
      }
      unlink(DBFileWrong)
    }
  }
}

messConvert <- function() {
  list(
    `TRUE` = c("multi-file backend"),
    `FALSE` = c("DBI backend")
  )
}

CacheDBFiles <- function(cachePath = getOption("reproducible.cachePath")) {
  ext <- CacheDBFileSingleExt()
  dtFiles <- dir(CacheStorageDir(cachePath), pattern = ext, full.names = TRUE)
  dtFiles
}

memoiseEnv <- function(cachePath, envir = .GlobalEnv) {
  memPersist <- isTRUE(getOption("reproducible.memoisePersist", NULL))
  if (memPersist) {
    obj <- paste0(".reproducibleMemoise_", cachePath)
    if (!exists(obj, envir = envir))
      assign(obj, new.env(parent = emptyenv()), envir = envir)
    memEnv <- get(obj, envir = envir, inherits = FALSE)
  } else {
    if (is.null(.pkgEnv[[cachePath]])) {
      .pkgEnv[[cachePath]] <- new.env(parent = emptyenv())
    }
    memEnv <- .pkgEnv[[cachePath]]
  }
  memEnv
}


otherFunctions <- "otherFunctions"

#' Evaluate whether a cacheId is memoised
#'
#' Intended for internal use. Exported so other packages can use this function.
#'
#' @inheritParams Cache
#' @return A logical, length 1 indicating whether the `cacheId` is memoised.
#'
#' @export
.isMemoised <- function(cacheId, cachePath = getOption("reproducible.cachePath")) {
  isMemoised <- NA
  if (isTRUE(getOption("reproducible.useMemoise"))) {
    isMemoised <- exists(cacheId, envir = memoiseEnv(cachePath))
  }
  isMemoised
}



metadataDT <- function(cacheId, tagKey, tagValue) {
  data.table(
    "cacheId" = cacheId, "tagKey" = as.character(tagKey),
    "tagValue" = tagValue, "createdDate" = as.character(Sys.time())
  )
}


loadFromCacheSwitchFormat <- function(f, verbose, cachePath, fullCacheTableForObj, cacheId, preDigest, drv, conn) {
  obj <- NULL
  if (!all(file.exists(f))) {
    sameCacheID <- checkSameCacheId(f)
    # sameCacheID <- dir(dirname(f), pattern = filePathSansExt(basename(f)))
    # if (!useDBI() || length(sameCacheID) > 1) {
    #   sameCacheID <- onlyStorageFiles(sameCacheID)
    # }

    if (length(sameCacheID)) {
      # if (!identical(.whereInStack("sim"), .GlobalEnv)) {
      #   cacheSaveFormat <- setdiff(c(.rdsFormat, .qsFormat), cacheSaveFormat)
      #   message("User tried to change options('reproducible.cacheSaveFormat') for an ",
      #           "existing cache, while using a simList. ",
      #           "This currently does not work. Keeping the ",
      #           "option at: ", cacheSaveFormat)
      #   next
      # }

      # messageCache(.message$changingFormat(prevFile = sameCacheID, newFile = f),
      #              verbose = verbose)
      # #messageCache("     (Changing cacheSaveFormat of Cache entry from ", fileExt(sameCacheID), " to ",
      #             fileExt(f), ")",
      obj <- loadFromCache(
        cachePath = cachePath, fullCacheTableForObj = fullCacheTableForObj,
        cacheId = cacheId,
        cacheSaveFormat = fileExt(sameCacheID),
        preDigest = preDigest,
        verbose = verbose
      )

      obj2 <- .wrap(obj, cachePath = cachePath, drv = drv, conn = conn, cacheId = cacheId)
      swapCacheFileFormat(wrappedObj = obj2, cachePath = cachePath, drv = drv, conn = conn,
                          cacheId = cacheId, sameCacheID = sameCacheID, newFile = f, verbose = verbose)
      # fs <- saveToCache(
      #   obj = obj2, cachePath = cachePath, drv = drv, conn = conn,
      #   cacheId = cacheId
      # )
      # rmFromCache(
      #   cachePath = cachePath, cacheId = cacheId, drv = drv, conn = conn,
      #   cacheSaveFormat = fileExt(sameCacheID)
      # )
    }
  }
  return(obj)
}

checkSameCacheId <- function(f) {
  cacheId <- filePathSansExt(basename(f))
  sameCacheID <- grep("\\.lock$", dir(dirname(f), pattern = cacheId), invert = TRUE, value = TRUE)
  sameCacheID <- grep(paste0(paste(.cacheSaveFormats, collapse = "|"), "$"), sameCacheID, value = TRUE)
  if (!useDBI() && length(sameCacheID) > 1) {
    sameCacheID <- onlyStorageFiles(sameCacheID, cacheId)
  }
  sameCacheID
}

swapCacheFileFormat <- function(wrappedObj, cachePath, drv, conn, cacheId, sameCacheID,
                                userTags, newFile, verbose) {
  messageCache(.message$changingFormat(prevFile = sameCacheID, newFile = newFile),
               verbose = verbose)

  fs <- saveToCache(
    obj = wrappedObj, cachePath = cachePath,
    userTags = userTags, drv = drv, conn = conn,
    cacheId = cacheId, cacheSaveFormat = fileExt(newFile)
  )
  rmFromCache(
    cachePath = cachePath, cacheId = cacheId, drv = drv, conn = conn,
    cacheSaveFormat = fileExt(sameCacheID)
  )
}


dbDisconnectAll <- function(conn) {
  if (useDBI()) {
    if (!is.null(conn)) {
      suppressWarnings({
      if (is(conn, "list"))
        lapply(conn, DBI::dbDisconnect)
      else
        DBI::dbDisconnect(conn)
      })
    }
  }
}


.cacheSaveFormats <- c("qs2", "rds", "qs")
.qs2Format <- grep("qs2$", .cacheSaveFormats, value = TRUE, ignore.case = TRUE)
.qsFormat <- grep("qs$", .cacheSaveFormats, value = TRUE, ignore.case = TRUE)
.rdsFormat <- grep("rds$", .cacheSaveFormats, value = TRUE, ignore.case = TRUE)

#' Does an object use a pointer?
#'
#' @param x an object
#'
#' @return logical
#'
#' @export
usesPointer <- function(x) {
  UseMethod("usesPointer", x)
}

#' @export
usesPointer.default <- function(x) {
  xState <- FALSE
  if (.isSpatRaster(x))
    if (requireNamespace("terra")) {
      xState <- any(terra::inMemory(x))
    }
  xState
}

#' @export
usesPointer.list <- function(x) {
  lapply(x, usesPointer)
}

#' @export
usesPointer.environment <- function(x) {
  x <- as.list(x, all.names = TRUE)
  usesPointer(x)
}



try2 <- function(..., silent = FALSE) {
  if (getOption("reproducible.useTry", TRUE)) {
    try(..., silent = silent)
  } else
    eval(...)
}

tryCatch2 <- function(..., silent = FALSE, error, finally) {
  if (getOption("reproducible.useTry", TRUE)) {
    tryCatch(..., silent = silent)
  } else
    eval(...)
}

test_that("test devMode", {
  testInit(opts = list(reproducible.useCache = "devMode"))

  clearCache(tmpCache, ask = FALSE)
  theTags <- "hiTest"
  centralTendency <- function(x) {
    mean(x)
  }
  funnyData <- c(1, 1, 1, 1, 10)
  uniqueUserTags <- c("thisIsUnique", "reallyUnique")
  ranNumsB <- Cache(centralTendency, funnyData,
    cachePath = tmpCache,
    userTags = uniqueUserTags
  ) # sets new value to Cache
  a <- showCache(tmpCache) # 1 unique artifact -- cacheId is 8be9cf2a072bdbb0515c5f0b3578f474
  expect_true(NROW(unique(a[[.cacheTableHashColName()]])) == 1)

  # During development, we often redefine function internals
  centralTendency <- function(x) {
    median(x)
  }
  # When we rerun, we don't want to keep the "old" cache because the function will
  #   never again be defined that way. Here, because of userTags being the same,
  #   it will replace the entry in the Cache, effetively overwriting it, even though
  #   it has a different cacheId
  ranNumsD <- Cache(centralTendency, funnyData, cachePath = tmpCache, userTags = uniqueUserTags) |>
    capture.output() -> oo
  a <- showCache(tmpCache) # 1 unique artifact -- cacheId is bb1195b40c8d37a60fd6004e5d526e6b
  expect_true(NROW(unique(a[[.cacheTableHashColName()]])) == 1)

  # If it finds it by cacheID, doesn't matter what the userTags are
  ranNumsD <- Cache(centralTendency, funnyData, cachePath = tmpCache, userTags = "thisIsUnique")
  a <- showCache(tmpCache) # 1 unique artifact -- cacheId is bb1195b40c8d37a60fd6004e5d526e6b
  expect_true(NROW(unique(a[[.cacheTableHashColName()]])) == 1)

  ###### If you don't use userTags -- it acts like normal
  ranNumsE <- Cache(centralTendency, 1:10, cachePath = tmpCache)
  centralTendency <- function(x) {
    sort(table(a))[1]
  }
  ranNumsF <- Cache(centralTendency, 1:10, cachePath = tmpCache)

  a <- showCache(tmpCache) # 1 unique artifact -- cacheId is bb1195b40c8d37a60fd6004e5d526e6b
  expect_true(NROW(unique(a[[.cacheTableHashColName()]])) == 3)

  centralTendency <- function(x) {
    median(x)
  }
  ranNumsG <- Cache(centralTendency, 1:11, cachePath = tmpCache, userTags = theTags)
  a <- showCache(tmpCache) # 1 unique artifact -- cacheId is bb1195b40c8d37a60fd6004e5d526e6b
  expect_true(NROW(unique(a[[.cacheTableHashColName()]])) == 4)

  centralTendency <- function(x) {
    sort(table(x))[1]
  }
  ranNumsH <- Cache(centralTendency, 1:11, cachePath = tmpCache, userTags = theTags) |>
    capture.output() -> oo

  a <- showCache(tmpCache) #
  expect_true(NROW(unique(a[[.cacheTableHashColName()]])) == 4)

  # Test multiple with same userTags, ie, not unambiguous
  opt <- options(reproducible.useCache = TRUE)
  ranNumsG <- Cache(centralTendency, 1:12, cachePath = tmpCache, userTags = theTags)
  options(opt)
  centralTendency <- function(x) median(x)
  mess <- capture_messages({
    ranNumsG <- Cache(centralTendency, 1:12, cachePath = tmpCache, userTags = theTags, verbose = 3)
  }) |>
    capture.output() -> oo
  if (!getOption("reproducible.useCacheV3") %in% TRUE)
    expect_true(any(grepl("not unique; defaulting", mess)))

  ### Test that vague userTags don't accidentally delete a non-similar entry
  clearCache(tmpCache, ask = FALSE)
  centralTendency <- function(x) {
    sort(table(x))[1]
  }
  ranNumsH <- Cache(centralTendency, 1:11, cachePath = tmpCache, userTags = theTags)
  ranNumsI <- Cache(rnorm, 15, cachePath = tmpCache, userTags = theTags)
  a <- showCache(tmpCache) # 2 unique artifacts because VERY different
  expect_true(length(unique(a[[.cacheTableHashColName()]])) == 2)
})

test_that("showCache survives a corrupt cache file (unknown input format)", {
  ## Regression for: when a cache directory contains a file whose extension
  ## says ".rds" but whose contents aren't valid RDS, loadFile()'s
  ## readRDS() throws "unknown input format". The recovery loop in
  ## showCache() expects loadFile() errors to be caught as `try-error`
  ## values; without the try() wrap, the error escaped and the entire
  ## showCache() call (and any Cache() that triggered showSimilar) blew up.
  skip_on_cran()

  tmpCache <- file.path(tempdir(), basename(tempfile("rcache_corrupt_")))
  dir.create(tmpCache, showWarnings = FALSE, recursive = TRUE)
  withr::local_options(reproducible.cachePath  = tmpCache,
                       reproducible.useMemoise = FALSE,
                       reproducible.verbose    = 0)

  ## Seed one valid cache entry so the dir actually has loadable content.
  Cache(rnorm, 1, cachePath = tmpCache, cacheId = "good_v1", useCloud = FALSE)

  ## Plant a corrupt ".dbFile.rds" file in the cache storage dir. showCache()
  ## scans dir() with pattern ".dbFile.qs2|.dbFile.rds|.dbFile.qs" and calls
  ## loadFile() on each match -- so the corrupt file must match that pattern
  ## to actually exercise the broken loadFile call site.
  storage <- reproducible::CacheStorageDir(tmpCache)
  corruptFile <- file.path(storage, "corruptCacheId_v1.dbFile.rds")
  writeBin(as.raw(c(0x00, 0x01, 0x02, 0x03)), corruptFile)
  expect_true(file.exists(corruptFile))

  ## Pre-condition: readRDS on this file errors with "unknown input format"
  expect_error(readRDS(corruptFile), regexp = "unknown input format")

  ## Core invariant: showCache() must NOT propagate the readRDS error. Before
  ## the fix, the unwrapped loadFile() call let "unknown input format" escape
  ## the lapplyFun() loop and crash any Cache()->showSimilar()->showCache()
  ## chain.
  out <- expect_no_error(
    suppressMessages(reproducible::showCache(tmpCache, verbose = -1))
  )
  expect_true(NROW(out) > 0,
              info = "Valid cache entries should still be returned alongside the bad one")
})

test_that("test Cache argument inheritance to inner functions", {
  testInit("terra",
           verbose = TRUE,
           opts = list(
             "reproducible.showSimilar" = FALSE,
             "reproducible.useMemoise" = FALSE
           )
  )
  withr::local_options(reproducible.cachePath = tmpdir)
  tmpDirFiles <- dir(tempdir())
  on.exit(
    {
      newOnes <- setdiff(tmpDirFiles, dir(tempdir()))
      unlink(newOnes, recursive = TRUE)
    },
    add = TRUE
  )

  outer <- function(n, not = NULL) {
    Cache(rnorm, n, notOlderThan = not)
  }

  mess <- capture_messages(Cache(outer, n = 2))

  expect_equal(sum(grepl(.message$NoCacheRepoSuppliedGrep, mess)), 2)
  clearCache(ask = FALSE, x = tmpdir)

  out <- capture_messages(Cache(outer, n = 2))
  expect_true(all(unlist(lapply(
    c(.message$NoCacheRepoSuppliedGrep, .message$NoCacheRepoSuppliedGrep),
    function(mess) any(grepl(mess, out))
  ))))

  # does Sys.time() propagate to outer ones
  out <- capture_messages(Cache(outer(n = 2, not = Sys.time() + 1), notOlderThan = Sys.time() + 1))
  expect_equal(sum(grepl(.message$NoCacheRepoSuppliedGrep, out)), 2)

  # does Sys.time() propagate to outer ones -- no message about cachePath being tempdir()
  mess <- capture_messages(Cache(outer(n = 2, not = Sys.time()), notOlderThan = Sys.time(), cachePath = tmpdir))
  expect_equal(sum(grepl(.message$NoCacheRepoSuppliedGrep, mess)), 1)

  # does cachePath propagate to outer ones -- no message about cachePath being tempdir()
  out <- capture_messages(Cache(outer, n = 2, cachePath = tmpdir))
  expect_true(length(out) == 2)
  expect_true(sum(cli::ansi_grepl(paste0(.message$LoadedCacheResult(), ".+outer call"), out)) == 1)

  # check that the rnorm inside "outer" returns cached value even if outer "outer" function is changed
  outer <- function(n) {
    a <- 1
    Cache(rnorm, n)
  }
  out <- capture_messages(Cache(outer, n = 2, cachePath = tmpdir))
  expect_true(length(out) == 4)
  msgGrep <- paste(paste0(.message$LoadedCacheResult(), ".+rnorm call"),
                   "There is no similar item in the cachePath",
                   sep = "|"
  )
  expect_true(sum(cli::ansi_grepl(msgGrep, out)) == 1)

  # Override with explicit argument
  outer <- function(n) {
    a <- 1
    Cache(rnorm, n, notOlderThan = Sys.time() + 1)
  }
  out <- capture_messages(Cache(outer, n = 2, cachePath = tmpdir))
  expect_equal(sum(grepl(.message$NoCacheRepoSuppliedGrep, out)), 1)

  # change the outer function, so no cache on that, & have notOlderThan on rnorm,
  #    so no Cache on that
  outer <- function(n) {
    b <- 1
    Cache(rnorm, n, notOlderThan = Sys.time() + 1)
  }
  out <- capture_messages(Cache(outer, n = 2, cachePath = tmpdir))
  expect_equal(sum(grepl(.message$NoCacheRepoSuppliedGrep, out)), 1)

  # expect_true(all(grepl("There is no similar item in the cachePath", out)))
  # Second time will get a cache on outer
  out <- capture_messages(Cache(outer, n = 2, cachePath = tmpdir))
  expect_true(length(out) == 2)
  expect_true(sum(cli::ansi_grepl(paste0(.message$LoadedCacheResult(), ".+outer call"), out)) == 1)

  # doubly nested
  inner <- function(mean, useCache = TRUE) {
    d <- 1
    Cache(rnorm, n = 3, mean = mean, useCache = useCache)
  }
  outer <- function(n, useCache = TRUE, ...) {
    Cache(inner, 0.1, useCache = useCache, ...)
  }
  out <- capture_messages(Cache(outer, n = 2, cachePath = tmpdir))

  outer <- function(n) {
    Cache(inner, 0.1, notOlderThan = Sys.time() - 1e4)
  }

  out <- capture_messages(Cache(outer, n = 2, cachePath = tmpdir, notOlderThan = Sys.time()))
  msgGrep <- paste(paste(.message$LoadedCacheResult(), "inner call"),
                   "There is no similar item in the cachePath",
                   sep = "|"
  )
  expect_true(sum(grepl(.message$NoCacheRepoSuppliedGrep, out)) == 1)

  expect_true(sum(cli::ansi_grepl(msgGrep, out)) == 1)

  outer <- function(n) {
    Cache(inner, 0.1, notOlderThan = Sys.time())
  }
  inner <- function(mean) {
    d <- 1
    Cache(rnorm, n = 3, mean = mean, notOlderThan = Sys.time() - 1e5)
  }

  out <- capture_messages(Cache(outer, n = 2, cachePath = tmpdir, notOlderThan = Sys.time()))
  msgGrep <- paste(paste(.message$LoadedCacheResult(), "rnorm call"),
                   "There is no similar item in the cachePath",
                   sep = "|"
  )
  expect_true(sum(cli::ansi_grepl(msgGrep, out)) == 1)

  # Check userTags -- all items have it
  clearCache(tmpdir, ask = FALSE)
  outerTag <- "howdie"
  aa <- Cache(outer, n = 2, cachePath = tmpdir, userTags = outerTag)
  bb <- showCache(tmpdir, userTags = outerTag)
  cc <- showCache(tmpdir)
  data.table::setorderv(bb, c(.cacheTableHashColName(), "tagKey", "tagValue"))
  data.table::setorderv(cc, c(.cacheTableHashColName(), "tagKey", "tagValue"))
  expect_true(identical(bb, cc))

  # Check userTags -- all items have the outer tag propagate, plus inner ones only have inner ones
  innerTag <- "notHowdie"
  inner <- function(mean) {
    d <- 1
    Cache(rnorm, n = 3, mean = mean, notOlderThan = Sys.time() - 1e5, userTags = innerTag)
  }

  clearCache(tmpdir, ask = FALSE)
  aa <- Cache(outer, n = 2, cachePath = tmpdir, userTags = outerTag)
  bb <- showCache(tmpdir, userTags = outerTag)
  cc <- showCache(tmpdir)
  onCols <- names(.emptyCacheTable)
  data.table::setorderv(cc, cols = onCols)
  data.table::setorderv(bb, cols = onCols)
  expect_true(identical(bb, cc))

  #
  bb <- showCache(tmpdir, userTags = "notHowdie")
  cc <- showCache(tmpdir)
  data.table::setorderv(cc, cols = onCols)
  data.table::setorderv(bb, cols = onCols)
  expect_false(identical(bb, cc))
  expect_true(length(unique(bb[[.cacheTableHashColName()]])) == 1)
  expect_true(length(unique(cc[[.cacheTableHashColName()]])) == 3)
})

test_that("test file-backed raster caching", {
  skip_on_cran()
  testInit("terra",
           tmpFileExt = c(".tif", ".grd"),
           opts = list(reproducible.useMemoise = FALSE,
                       reproducible.verbose = FALSE)
  )

  nOT <- Sys.time()

  randomPolyToDisk <- function(tmpfile) {
    r <- terra::rast(ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
    .writeRaster(r, tmpfile[1], overwrite = TRUE)
    r <- terra::rast(tmpfile[1])
    r
  }

  a <- randomPolyToDisk(tmpfile[1])
  # confirm that the raster has the given tmp filename
  expect_identical(normPath(tmpfile[1]),
                   normPath(Filenames(a))
  )

  # Using mock interactive function
  # https://www.mango-solutions.com/blog/testing-without-the-internet-using-mock-functions
  # https://github.com/r-lib/testthat/issues/734 to direct it to reproducible::isInteractive
  #   solves the error about not being in the testthat package

  savePreDigests <- c(FALSE)#, TRUE) #TODO --> tests aren't updated yetdoesn't work yet
  for (savePreDigest in savePreDigests) {
    withr::local_options(reproducible.savePreDigest = savePreDigest)

    val1 <- .cacheNumDefaultTags()
    val1WOSpatRaster <- val1
    val1 <- val1 + length(setdiff(gsub(":.*$", "", tagsSpatRaster()), .ignoreTagKeys())) # adding a userTag here... the +8 is the SpatRaster extras
    # if (isTRUE(savePreDigest))
    #   val1 <- val1 + .cacheNumDefaultTags() # the preDigest ones
    ik <- .ignoreTagKeys()

    # val1WOSpatRaster <- val1WOSpatRaster * (savePreDigest + 1)

    aa <- Cache(randomPolyToDisk, tmpfile[1], cachePath = tmpCache, userTags = "something2")

    expect_equal(NROW(showCache(tmpCache)[!tagKey %in% .ignoreTagKeys()]), val1)
    clearCache(tmpCache, userTags = "something$", ask = FALSE)
    expect_equal(NROW(showCache(tmpCache)[!tagKey %in% .ignoreTagKeys()]), val1)
    clearCache(tmpCache, userTags = "something2", ask = FALSE)
    expect_equal(NROW(showCache(tmpCache)), 0)
    expect_equal(NROW(dir(CacheStorageDir(tmpCache))), 0) # make sure the `.tif` file is also gone

    aa <- Cache(randomPolyToDisk, tmpfile[1], cachePath = tmpCache, userTags = "something2")
    expect_equal(NROW(showCache(tmpCache)[!tagKey %in% .ignoreTagKeys()]), val1)
    clearCache(tmpCache, userTags = c("something$", "testing$"), ask = FALSE)
    expect_equal(NROW(showCache(tmpCache)[!tagKey %in% .ignoreTagKeys()]), val1)
    clearCache(tmpCache, userTags = c("something2$", "testing$"), ask = FALSE)
    expect_equal(NROW(showCache(tmpCache)[!tagKey %in% .ignoreTagKeys()]), val1)

    clearCache(tmpCache, userTags = c("something2$", "randomPolyToDisk$"), ask = FALSE)
    expect_equal(NROW(showCache(tmpCache)), 0)

    aa <- Cache(randomPolyToDisk, tmpfile[1], cachePath = tmpCache, userTags = "something2")

    # confirm that the raster has the new filename in the cachePath
    if (is(aa, "Raster")) {
      expect_false(identical(
        strsplit(tmpfile[1], split = "[\\/]"),
        strsplit(file.path(
          tmpCache, "rasters",
          basename(tmpfile[1])
        ), split = "[\\/]")
      ))
      expect_true(any(cli::ansi_grepl(
        pattern = basename(tmpfile[1]),
        dir(file.path(tmpCache, "rasters"))
      )))
    } else {
      sc <- showCache(tmpCache)
      origFile <- sc[tagKey == "origFilename"]$cacheId
      hasFilenameInCache <- NROW(sc[tagKey %in% tagFilenamesInCache])
      expect_true(length(dir(CacheStorageDir(tmpCache), pattern = origFile)) ==
                    (1 + hasFilenameInCache + as.integer(!useDBI()) + 2 * savePreDigest + 1L))
      # expect_true(length(dir(CacheStorageDir(tmpCache), pattern = origFile)) == 1 + !useDBI())
    }

    clearCache(x = tmpCache)
    bb <- Cache(randomPolyToDisk, tmpfile[1],
                cachePath = tmpCache, userTags = "something2",
                quick = TRUE
    )
    # bb <- Cache(randomPolyToDisk, tmpfile[1], cachePath = tmpdir, userTags = "something2")
    # clearCache(x = tmpdir)
    clearCache(x = tmpdir)
    try(unlink(CacheDBFile(tmpdir)), silent = TRUE)
    try(unlink(CacheStorageDir(tmpdir), recursive = TRUE), silent = TRUE)
    froms <- normPath(dir(tmpCache, recursive = TRUE, full.names = TRUE))
    checkPath(file.path(tmpdir, "rasters"), create = TRUE)
    checkPath(file.path(tmpdir, "cacheOutputs"), create = TRUE)
    file.copy(
      from = froms, overwrite = TRUE,
      to = gsub(normPath(tmpCache), normPath(tmpdir), froms)
    )
    # movedCache(tmpdir)
    # ._prepareOutputs_1 <<- ._prepareOutputs_2 <<- ._getFromRepo <<- 1
    # Will silently update the filename of the RasterLayer, and recover it
    type <- gsub("Connection", "", class(getOption("reproducible.conn")))
    isSQLite <- cli::ansi_grepl(type, "NULL")
    if (!isSQLite) {
      warn1 <- capture_warnings(movedCache(tmpdir, old = tmpCache))
    }

    warn <- capture_warnings({
      bb <- Cache(randomPolyToDisk, tmpfile[1],
                  cachePath = tmpdir, userTags = "something2",
                  quick = TRUE
      )
    })

    expect_false(attr(bb, ".Cache")$newCache)
    expect_true(file.exists(Filenames(bb)))
    expect_silent(bb[] <- bb[])

    # Delete the old everything to make sure previous didn't succeed because old pointer
    clearCache(x = tmpCache)
    try(unlink(CacheDBFile(tmpCache)), silent = TRUE)
    try(unlink(CacheStorageDir(tmpCache), recursive = TRUE), silent = TRUE)

    # ._Cache_6 <<- 1
    bb <- Cache(randomPolyToDisk, tmpfile[1],
                cachePath = tmpdir, userTags = "something2",
                quick = TRUE
    )
    expect_false(attr(bb, ".Cache")$newCache)
    expect_true(file.exists(Filenames(bb)))
    expect_silent(bb[] <- bb[])

    clearCache(tmpCache)
    clearCache(tmpdir)
    cc <- Cache(randomPolyToDisk, tmpfile[2],
                cachePath = tmpCache, userTags = "something2",
                quick = TRUE
    )
    bb <- Cache(randomPolyToDisk, tmpfile[1],
                cachePath = tmpCache, userTags = "something2",
                quick = TRUE
    )
    try(movedCache(tmpdir, tmpCache), silent = TRUE)

    #
    bbS <- c(bb, cc)
    fn2 <- function(stk) {
      stk
    }
    out <- Cache(fn2, bbS, cachePath = tmpCache, userTags = "something2")
    froms <- normPath(dir(tmpCache, recursive = TRUE, full.names = TRUE))
    # checkPath(file.path(tmpdir, "rasters"), create = TRUE)
    checkPath(file.path(tmpdir, "cacheOutputs"), create = TRUE)
    file.copy(
      from = froms, overwrite = TRUE,
      to = gsub(normPath(tmpCache), normPath(tmpdir), froms)
    )
    if (!isSQLite) {
      DBI::dbRemoveTable(conn, CacheDBTableName(tmpdir))
    }
    movedCache(tmpdir, tmpCache)
    out <- Cache(fn2, bbS, cachePath = tmpdir, userTags = "something2")

    clearCache(tmpdir)
    clearCache(tmpCache)

    ### Test for 2 caching events with same file-backing name
    randomPolyToDisk2 <- function(tmpfile, rand) {
      r <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
      .writeRaster(r, tmpfile[1], overwrite = TRUE)
      r <- terra::rast(tmpfile[1])
      r
    }

    a <- Cache(randomPolyToDisk2, tmpfile[1], runif(1))
    b <- Cache(randomPolyToDisk2, tmpfile[1], runif(1))

    # changed behaviour as of reproducible 1.2.0.9020
    #  -- now Cache doesn't protect user from filename collisions if user makes them
    expect_true(identical(Filenames(a), Filenames(b)))

    # Caching a raster as an input works
    rasterTobinary <- function(raster) {
      ceiling(raster[] / (mean(raster[]) + 1))
    }
    nOT <- Sys.time() + 1

    for (i in 1:2) {
      strt <- Sys.time()
      assign(paste0("a", i), Cache(rasterTobinary, a, cachePath = tmpCache, notOlderThan = nOT))
      fin <- Sys.time()
      assign(paste0("b", i), fin - strt)
      nOT <- Sys.time() - 100
    }
    expect_true(all.equalWONewCache(a1, a2))
    expect_true(identical(attr(a1, ".Cache")$newCache, TRUE))
    expect_true(identical(attr(a2, ".Cache")$newCache, FALSE))

    clearCache(tmpCache, ask = FALSE)

    # Check that Caching of rasters saves them to tif file instead of rdata
    randomPolyToMemory <- function() {
      r <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
      # terra::datatype(r) <- "INT1U"
      r
    }

    bb <- Cache(randomPolyToMemory, cachePath = tmpdir)
    expect_true(Filenames(bb) == "")
    expect_true(inMemory(bb))

    bb <- Cache(randomPolyToMemory, cachePath = tmpdir)
    expect_true(NROW(showCache(tmpdir)[!tagKey %in% .ignoreTagKeys()]) == val1WOSpatRaster)

    # Test that factors are saved correctly
    randomPolyToFactorInMemory <- function() {
      r <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
      levels(r) <- data.frame(
        ID = 1:30, vals = sample(LETTERS[1:5], size = 30, replace = TRUE),
        vals2 = sample(1:7, size = 30, replace = TRUE)
      )
      # dataType(r) <- "INT1U"
      r
    }
    bb <- Cache(randomPolyToFactorInMemory, cachePath = tmpdir)
    expect_true(terra::is.factor(bb))
    # expect_equal(dataType2(bb), "INT1U") # irrelevant because on disk
    expect_true(terra::is.factor(bb))
    expect_true(is(terra::cats(bb)[[1]], "data.frame"))
    expect_true(NCOL(terra::cats(bb)[[1]]) == 3)
    expect_true(NROW(terra::cats(bb)[[1]]) == 30)

    randomPolyToFactorOnDisk <- function(tmpfile) {
      r <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
      levels(r) <- data.frame(
        ID = 1:30, vals = sample(LETTERS[1:5], size = 30, replace = TRUE),
        vals2 = sample(1:7, size = 30, replace = TRUE)
      )
      r <- .writeRaster(r, tmpfile, overwrite = TRUE, datatype = "INT1U")
      r
    }

    # bb1 has original tmp filename
    bb1 <- randomPolyToFactorOnDisk(tmpfile[2])
    # bb has new one, inside of cache repository, with same basename
    bb <- Cache(randomPolyToFactorOnDisk, tmpfile = tmpfile[2], cachePath = tmpdir)
    # changed behaviour as of reproducible 1.2.0.9020 -- now Cache doesn't protect user from filename collisions if user makes them
    expect_true(unique(dirname(normPath(Filenames(bb)))) != normPath(file.path(tmpdir, "rasters")))
    expect_true(identical(basename(Filenames(bb, allowMultiple = FALSE)), basename(tmpfile[2])))
    expect_identical(normPath(Filenames(bb, allowMultiple = FALSE)), normPath(tmpfile[2]))
    expect_identical(normPath(dirname(Filenames(bb1, allowMultiple = FALSE))), normPath(dirname(tmpfile[2])))
    expect_true(basename(Filenames(bb1, allowMultiple = FALSE)) == basename(tmpfile[2]))
    expect_true(dataType2(bb) == "INT1U")
    if (.requireNamespace("raster")) {
      expect_true(raster::is.factor(bb))
    }
    expect_true(is(terra::cats(bb)[[1]], "data.frame"))
    expect_true(NCOL(terra::cats(bb)[[1]]) == 3)
    expect_true(NROW(terra::cats(bb)[[1]]) == 30)

    clearCache(tmpdir, ask = FALSE)
  }


  # })
})

test_that("test memory backed raster robustDigest", {
  testInit("terra", tmpFileExt = c(".tif", ".tif"))
  set.seed(123)
  r1 <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
  r2 <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
  set.seed(123)
  r3 <- terra::rast(terra::ext(0, 20, 0, 5), vals = sample(1:30, size = 100, replace = TRUE))
  expect_false(identical(.robustDigest(r1), .robustDigest(r2)))
  # metadata same, content different
  expect_false(identical(.robustDigest(r1), .robustDigest(r3)))
  # metadata different, content same
  expect_false(identical(.robustDigest(r1), .robustDigest(r3)))
  expect_true(identical(r1[], r3[]))

  set.seed(123)
  r1 <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
  set.seed(123)
  r2 <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
  expect_true(identical(.robustDigest(r1), .robustDigest(r2)))

  r <- terra::rast(matrix(1:10, 2, 5))
  b <- c(r, r)
  dig <- .robustDigest(b)

  r1 <- terra::rast(matrix(1:10, 2, 5))
  b1 <- c(r1, r1)
  dig1 <- .robustDigest(b1)

  expect_identical(dig, dig1)

  b <- .writeRaster(b, filename = tmpfile[1], overwrite = TRUE)
  dig <- .robustDigest(b)

  r <- terra::rast(matrix(1:10, 2, 5))
  b <- c(r, r)
  bb1 <- .writeRaster(b, filename = tmpfile[1], overwrite = TRUE)
  dig1 <- .robustDigest(bb1)

  expect_identical(dig, dig1)

  # Stacks
  dimA <- 100
  r <- terra::rast(matrix(1:dimA, round(sqrt(dimA)), round(sqrt(dimA))))
  b <- c(r, r)
  dig <- .robustDigest(b)

  r1 <- terra::rast(matrix(1:dimA, round(sqrt(dimA)), round(sqrt(dimA))))
  b1 <- c(r1, r1)
  dig1 <- .robustDigest(b1)

  expect_identical(dig, dig1)

  r4 <- .writeRaster(r, filename = tmpfile[1], overwrite = TRUE)
  r5 <- .writeRaster(r, filename = tmpfile[2], overwrite = TRUE)
  b <- c(r4, r5)
  dig <- .robustDigest(b)

  r2 <- .writeRaster(r1, filename = tmpfile[1], overwrite = TRUE)
  r3 <- .writeRaster(r1, filename = tmpfile[2], overwrite = TRUE)
  b1 <- c(r2, r3)
  # b1 <- .writeRaster(b1, filename = tmpfile[1], overwrite = TRUE)
  dig1 <- .robustDigest(b1)

  expect_identical(dig, dig1)
})

test_that("test 'quick' argument", {
  testInit("terra",
           tmpFileExt = c(".tif", ".tif", ".tif"),
           opts = list(
             "reproducible.verbose" = 1,
             "reproducible.useMemoise" = TRUE,
             "reproducible.showSimilar" = FALSE
           )
  )

  ### Make raster using Cache
  set.seed(123)
  # unlink(tmpfile[1])
  r1 <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
  r1 <- .writeRaster(r1, filename = tmpfile[1], overwrite = TRUE)
  quickFun <- function(rasFile) {
    ras <- terra::rast(rasFile)
    ras[sample(ncell(ras), size = 1)]
  }
  fn <- Filenames(r1)
  thePath1 <- asPath(Filenames(r1))

  out1a <- Cache(quickFun, thePath1, cachePath = tmpdir)
  out1b <- Cache(quickFun, thePath1, cachePath = tmpdir, quick = TRUE)
  r1[4] <- r1[4] + 1 # brings to memory
  r1 <- .writeRaster(r1, filename = tmpfile[2], overwrite = TRUE)
  thePath <- asPath(Filenames(r1))
  mess1 <- capture_messages({
    out1c <- Cache(quickFun, thePath, cachePath = tmpdir, quick = TRUE)
  })

  expect_true(sum(cli::ansi_grepl(
    paste0(
      paste(.message$LoadedCache(.message$LoadedCacheResult(), "quickFun"), .message$AddingToMemoised), "|",
      .message$LoadedCache(.message$LoadedCacheResult("Memoised"), "quickFun")
    ),
    mess1
  )) == 0)
  mess2 <- capture_messages({
    out1c <- Cache(quickFun, thePath, cachePath = tmpdir, quick = FALSE)
  })
  expect_true(length(mess2) == 1) # because it is looking at the file contents

  # Using Raster directly -- not file
  quickFun <- function(ras) {
    ras[sample(ncell(ras), size = 1)]
  }

  out1a <- Cache(quickFun, r1, cachePath = tmpdir)
  out1b <- Cache(quickFun, r1, cachePath = tmpdir, quick = TRUE)
  r1[4] <- r1[4] + 1
  r1 <- .writeRaster(r1, filename = tmpfile[3], overwrite = TRUE)
  # bbbb <<- 1
  # on.exit(rm(bbbb, envir = .GlobalEnv), add = TRUE)
  mess1 <- capture_messages({
    out1c <- Cache(quickFun, r1, cachePath = tmpdir, quick = TRUE)
  })
  expect_true(sum(cli::ansi_grepl(
    paste0(
      paste(.message$LoadedCache(.message$LoadedCacheResult(), "quickFun"), .message$AddingToMemoised), "|",
      paste(.message$LoadedCacheResult("Memoised"), "quickFun call")
    ),
    mess1
  )) == 0)

  # mess3 <- capture_messages({ out1c <- Cache(quickFun, r1, cachePath = tmpdir, quick = FALSE) })
  # Should be "\033[34mSaved! Cache file: d3d8d40b1aeba01e.rds; fn: \033[31mquickFun\033[34m\033[39m\n"
  mess <- capture_messages({
    out1c <- Cache(quickFun, r1, cachePath = tmpdir, quick = FALSE)
  })
  expect_true(length(mess) == 1)
})

test_that("test date-based cache removal", {
  testInit("terra", tmpFileExt = ".pdf")

  a <- Cache(runif, 1, cachePath = tmpdir)
  a1 <- showCache(tmpdir)
  expect_true(NROW(a1) > 0)
  b <- clearCache(tmpdir, before = Sys.Date() - 1, ask = FALSE)
  expect_true(NROW(b) == 0)
  expect_equivalent(a1, showCache(tmpdir)) # now has an index on cacheId

  b <- clearCache(tmpdir, before = Sys.Date() + 1, ask = FALSE)
  expect_identical(data.table::setindex(b, NULL), data.table::setindex(a1, NULL))
})

test_that("test keepCache", {
  testInit("terra")
  Cache(rnorm, 10, cachePath = tmpdir)
  Cache(runif, 10, cachePath = tmpdir)

  # this next is round which is a primitive *with* methods -->
  # --> an exception see ?match.call which says "normally positional"
  # This failed previously because round dispatches .Primitive("round") which internally in the C
  # code now (R > 4.3.1) matches non-positionally e.g., round.POSIXt has 'units'
  Cache(round, runif(4), cachePath = tmpdir)

  expect_true(NROW(showCache(tmpdir)[!tagKey %in% .ignoreTagKeys()]) ==
                .cacheNumDefaultTags() * 3)
  expect_true(NROW(showCache(tmpdir, c("rnorm", "runif"))[!tagKey %in% .ignoreTagKeys()]) ==
                0) # and search
  expect_true(NROW(keepCache(tmpdir, "rnorm", ask = FALSE)[!tagKey %in% .ignoreTagKeys()]) == .cacheNumDefaultTags())

  # do it twice to make sure it can deal with repeats
  expect_true(NROW(keepCache(tmpdir, "rnorm", ask = FALSE)[!tagKey %in% .ignoreTagKeys()]) == .cacheNumDefaultTags())
  st <- Sys.time()
  Sys.sleep(1)
  Cache(sample, 10, cachePath = tmpdir)
  Cache(length, 10, cachePath = tmpdir)
  Cache(sum, runif(4), cachePath = tmpdir)
  expect_true(NROW(showCache(tmpdir, before = st)[!tagKey %in% .ignoreTagKeys()]) == .cacheNumDefaultTags())
  expect_true(NROW(keepCache(tmpdir, before = st, ask = FALSE)[!tagKey %in% .ignoreTagKeys()]) == .cacheNumDefaultTags())
  expect_true(NROW(showCache(tmpdir)[!tagKey %in% .ignoreTagKeys()]) == .cacheNumDefaultTags())

  ranNums <- Cache(runif, 4, cachePath = tmpdir, userTags = "objectName:a")
  ranNums <- Cache(rnorm, 4, cachePath = tmpdir, userTags = "objectName:a")
  showCache(tmpdir) # shows rnorm, runif and rnorm objects

  # shows nothing because object has both spades and rnorm
  expect_true(NROW(showCache(tmpdir, userTags = c("spades", "rnorm"))) == 0)

  # "or" search
  expect_true(length(unique(
    showCache(tmpdir, userTags = "spades|rnorm")[[.cacheTableHashColName()]]
  )) == 2)

  # keep all with spades or rnorm
  expect_true(length(unique(
    keepCache(tmpdir, userTags = "spades|rnorm", ask = FALSE)[[.cacheTableHashColName()]]
  )) == 2)

  # shows spades, runif and rnorm objects
  expect_true(length(unique(showCache(tmpdir)[[.cacheTableHashColName()]])) == 2)

  # This is from changes to round by luke tierney
  aa <- Sys.time()
  Cache(round, aa, units = "mins", cachePath = tmpdir)
})

test_that("test environments", {
  testInit("terra", tmpFileExt = ".pdf")

  # make several unique environments
  a <- new.env(parent = emptyenv())
  b <- new.env(parent = emptyenv())
  g <- new.env(parent = emptyenv())
  d <- new.env(parent = emptyenv())
  f <- new.env(parent = emptyenv())
  h <- new.env(parent = emptyenv())
  # give the same value to a in each
  a$a <- runif(10)
  b$a <- a$a
  g$a <- a$a
  # put an environment in a
  a$d <- d
  b$d <- f
  g$d <- h

  # Upper level -- all are same, because values are same, even though all are enviros
  shortFn <- function(a) sample(a$a)
  out <- Cache(shortFn, a = a, cachePath = tmpdir)
  out2 <- Cache(shortFn, a = b, cachePath = tmpdir)
  out3 <- Cache(shortFn, a = g, cachePath = tmpdir)
  attr(out2, ".Cache")$newCache <- TRUE
  expect_true(identical(attributes(out)["tags"], attributes(out2)["tags"]))
  expect_true(identical(attributes(out)["tags"], attributes(out3)["tags"]))

  # put 2 different values, 1 and 2 ... a's and g's envirs are same value, but
  #    different environment .. b's envir is different value
  a$d$a <- 1
  b$d$a <- 2
  g$d$a <- 1
  i <- as.list(a)
  i2 <- i
  i2$d <- as.list(i2$d)

  out <- Cache(shortFn, a = a, cachePath = tmpdir)
  out2 <- Cache(shortFn, a = b, cachePath = tmpdir)
  out3 <- Cache(shortFn, a = g, cachePath = tmpdir)
  out4 <- Cache(shortFn, a = i, cachePath = tmpdir)
  out5 <- Cache(shortFn, a = i2, cachePath = tmpdir)

  # test different values are different
  expect_false(identical(attributes(out), attributes(out2)))

  # test same values but different enviros are same
  expect_true(identical(attributes(out)["tags"], attributes(out3)["tags"]))

  # test environment is same as a list -- not sure what they should be
  expect_true(identical(attributes(out)["tags"], attributes(out4)["tags"]))

  # test environment is same as recursive list - they shouldn't be
  expect_true(identical(attributes(out)["tags"], attributes(out5)["tags"]))

  df <- data.frame(a = a$a, b = LETTERS[1:10])
  out6 <- Cache(shortFn, a = df, cachePath = tmpdir)
  out7 <- Cache(shortFn, a = df, cachePath = tmpdir)
  expect_true(identical(attributes(out6)["tags"], attributes(out7)["tags"])) # test data.frame
})

test_that("test asPath", {
  testInit("terra",
           tmpFileExt = "pdf",
           verbose = TRUE,
           opts = list(
             "reproducible.useMemoise" = TRUE,
             "reproducible.showSimilar" = FALSE
           )
  )
  unlink(dir(tmpdir, full.names = TRUE))

  obj <- sample(1e5, 10)
  # First -- has no filename.RData
  a1 <- capture_messages(Cache(saveRDS, obj, file = "filename.RData", cachePath = tmpdir))
  # Second -- has a filename.RData, and passing a character string,
  #           it tries to see if it is a file, if yes, it digests it
  a2 <- capture_messages(Cache(saveRDS, obj, file = asPath("filename.RData"), cachePath = tmpdir))
  # Third -- finally has all same as second time
  a3 <- capture_messages(Cache(saveRDS, obj, file = asPath("filename.RData"), cachePath = tmpdir))

  expect_equal(length(a1), 1)
  expect_equal(length(a2), 1)
  expect_true(sum(cli::ansi_grepl(paste(
    .message$LoadedCacheResult("Memoised"), "|",
    .message$LoadedCacheResult()
  ), a3)) == 1)

  unlink("filename.RData")
  try(clearCache(tmpdir, ask = FALSE), silent = TRUE)
  a1 <- capture_messages(Cache(saveRDS, obj,
                               file = asPath("filename.RData"),
                               quick = TRUE, cachePath = tmpdir
  ))
  a2 <- capture_messages(Cache(saveRDS, obj,
                               file = asPath("filename.RData"),
                               quick = TRUE, cachePath = tmpdir
  ))
  a3 <- capture_messages(Cache(saveRDS, obj,
                               file = asPath("filename.RData"),
                               quick = TRUE, cachePath = tmpdir
  ))
  expect_equal(length(a1), 1)
  expect_true(sum(cli::ansi_grepl(paste(
    .message$LoadedCacheResult("Memoised"), "|",
    .message$LoadedCacheResult()
  ), a2)) == 1)
  expect_true(sum(cli::ansi_grepl(paste(.message$LoadedCacheResult("Memoised"), "saveRDS call"), a3)) == 1)

  unlink("filename.RData")
  try(clearCache(tmpdir, ask = FALSE), silent = TRUE)
  a1 <- capture_messages(Cache(saveRDS, obj,
                               file = as("filename.RData", "Path"),
                               quick = TRUE, cachePath = tmpdir
  ))
  a2 <- capture_messages(Cache(saveRDS, obj,
                               file = as("filename.RData", "Path"),
                               quick = TRUE, cachePath = tmpdir
  ))
  a3 <- capture_messages(Cache(saveRDS, obj,
                               file = as("filename.RData", "Path"),
                               quick = TRUE, cachePath = tmpdir
  ))
  expect_equal(length(a1), 1)
  expect_true(sum(cli::ansi_grepl(paste(
    .message$LoadedCacheResult("Memoised"), "|",
    .message$LoadedCacheResult()
  ), a2)) == 1)
  expect_true(sum(cli::ansi_grepl(paste(.message$LoadedCacheResult("Memoised"), "saveRDS call"), a3)) == 1)
})

test_that("test wrong ways of calling Cache", {
  try(testInit(tmpFileExt = ".pdf"))

  # expect_error(Cache(sample(1), cachePath = tmpdir), "Can't understand")
  # expect_error(Cache(a <- sample(1), cachePath = tmpdir), "Can't understand")
  expect_true(1 == Cache(sample, 1, cachePath = tmpdir))
})

test_that("test quoted FUN in Cache", {
  testInit()

  A <- Cache(rnorm, 10, 16, cachePath = tmpdir) # nolint

  ## recover cached copies:
  B <- Cache(rnorm, 10, 16, cachePath = tmpdir) # nolint
  C <- Cache(quote(rnorm(n = 10, 16)), cachePath = tmpdir) # nolint

  expect_true(all.equalWONewCache(A, B))
  expect_true(all.equalWONewCache(A, C))
})

test_that("test mergeCache", {
  testInit("data.table", verbose = TRUE)

  a <- Cache(rnorm, 1, cachePath = tmpdir)
  b <- Cache(rnorm, 2, cachePath = tmpCache)

  aCache <- showCache(tmpdir)
  bCache <- showCache(tmpCache)

  d <- mergeCache(tmpCache, tmpdir)

  dCache <- showCache(d)
  abCache <- rbindlist(list(aCache, bCache))

  # Remove date and accessed time stamps
  dCache <- dCache[!tagKey %in% c("date", "accessed")]
  abCache <- abCache[!tagKey %in% c("date", "accessed")]

  # remove keys
  setkeyv(dCache, .cacheTableHashColName())
  setkeyv(abCache, .cacheTableHashColName())

  expect_true(all.equal(
    abCache[, list(
      get(.cacheTableHashColName()),
      tagKey, get(.cacheTableTagColName())
    )],
    dCache[, list(
      get(.cacheTableHashColName()),
      tagKey, get(.cacheTableTagColName())
    )]
  ))
  mess <- capture_messages({
    d1 <- mergeCache(tmpCache, tmpdir)
  })
  expect_true(any(cli::ansi_grepl("Skipping", mess)))
  expect_equivalent(showCache(d), showCache(d1))
})

test_that("test cache-helpers", {
  testInit(c("raster"), tmpFileExt = c(rep(".grd", 3), rep(".tif", 3)))
  # out <- reproducible::createCache(tmpCache)

  r1 <- raster::raster(raster::extent(0, 3, 0, 3), vals = 1)
  r2 <- raster::raster(raster::extent(0, 3, 0, 3), vals = 2)
  r3 <- raster::raster(raster::extent(0, 3, 0, 3), vals = 3)
  r2 <- writeRaster(r1, filename = tmpfile[2], overwrite = TRUE)
  r3 <- writeRaster(r1, filename = tmpfile[3], overwrite = TRUE)
  r2tif <- suppressWarningsSpecific(
    falseWarnings = proj6Warn,
    writeRaster(r1, filename = tmpfile[5], overwrite = TRUE)
  )
  r3tif <- suppressWarningsSpecific(
    falseWarnings = proj6Warn,
    writeRaster(r1, filename = tmpfile[6], overwrite = TRUE)
  )

  s1 <- raster::stack(r1, r1)
  s2 <- raster::stack(r1, r2)
  s3 <- raster::stack(r3, r2)
  s1 <- raster::stack(r1, r1)
  s2tif <- raster::stack(r1, r2tif)
  s3tif <- raster::stack(r3tif, r2tif)

  i <- 1
  for (rr in list(r1, r2, r3, r2tif, r3tif, s1, s2, s3, s2tif, s3tif)) {
    message(i)
    i <- i + 1

    out2 <- .prepareFileBackedRaster(rr, repoDir = tmpCache)
    test1 <- identical(out2, rr)
    test2 <- identical(Filenames(out2), Filenames(rr))
    test3 <- identical(
      Filenames(out2, allowMultiple = FALSE),
      Filenames(rr, allowMultiple = FALSE)
    )
    test4 <- identical(
      basename(Filenames(out2, allowMultiple = TRUE)),
      basename(Filenames(rr, allowMultiple = TRUE))
    )
    test5 <- identical(length(Filenames(out2)), length(Filenames(rr)))
    if (any(nchar(Filenames(out2)) > 0)) {
      expect_false(test1 && test2 && test3)
      expect_true(test4 && test5)
    } else {
      expect_true(test1 && test2 && test3 && test4 && test5)
    }
    unlink(Filenames(out2))
  }

  out2 <- .prepareFileBackedRaster(s2, repoDir = tmpCache)
  out3 <- .prepareFileBackedRaster(s2, repoDir = tmpCache)
  fn2 <- Filenames(out2)
  fn3 <- Filenames(out3)
  actualFiles <- nchar(fn2) > 0
  bnfn2 <- basename(fn2[actualFiles])
  bnfn3 <- basename(fn3[actualFiles])
  bnfn2 <- unique(filePathSansExt(bnfn2))
  bnfn3 <- unique(filePathSansExt(bnfn3))
  sameFileBase <- cli::ansi_grepl(pattern = bnfn2, x = bnfn3)
  expect_true(sameFileBase)

  unlink(Filenames(s2))
  expect_error(
    {
      out2 <- .prepareFileBackedRaster(s2, repoDir = tmpCache)
    },
    "most likely"
  )
})

test_that("test useCache = 'overwrite'", {
  testInit(verbose = TRUE)

  a <- Cache(rnorm, 1, useCache = "overwrite", cachePath = tmpCache)
  mess <- capture_messages({
    b <- Cache(rnorm, 1, useCache = "overwrite", cachePath = tmpCache)
  })
  expect_true(!identical(a, b))
  expect_true(any(cli::ansi_grepl(pattern = "Overwriting", mess)))

  clearCache(x = tmpCache, ask = FALSE)

  withr::deferred_clear()
  testInit(
    ask = FALSE, verbose = TRUE,
    opts = list("reproducible.useCache" = "overwrite")
  )

  a <- Cache(rnorm, 1, cachePath = tmpCache)
  mess <- capture_messages({
    b <- Cache(rnorm, 1, cachePath = tmpCache)
  })
  expect_true(!identical(a, b))
  expect_true(any(cli::ansi_grepl(pattern = "Overwriting", mess)))
})

test_that("test rm large non-file-backed rasters", {
  ## This is a large object test!
  skip_on_cran()
  if (!is.null(getOption("reproducible.conn", NULL))) {
    if (!cli::ansi_grepl("SQLite", class(getOption("reproducible.conn", NULL)))) {
      skip("This is not for non-SQLite")
    }
  }

  testInit(c(.qs2Format, "terra"), opts = list("reproducible.cacheSpeed" = "fast",
                                         "reproducible.cacheSaveFormat" = .qs2Format))

  ext <- terra::ext(0, 10000, 0, 10000)
  r <- Cache(terra::rast, ext,
             resolution = 1, vals = 1,
             cachePath = tmpdir, userTags = "first"
  )
  st1 <- system.time(clearCache(tmpdir, userTags = "first", ask = FALSE))
  expect_true(st1["user.self"] < 0.75) # This was > 2 seconds in old way
})

test_that("test cc", {
  skip_on_cran()

  testInit(verbose = TRUE)

  Cache(rnorm, 1, cachePath = tmpCache)
  Sys.sleep(1) # 0.2
  thisTime <- Sys.time()
  Sys.sleep(1) # 0.2
  Cache(rnorm, 2, cachePath = tmpCache)
  Cache(rnorm, 3, cachePath = tmpCache)
  Cache(rnorm, 4, cachePath = tmpCache)
  a <- showCache(x = tmpCache) # shows all 4 entries
  expect_true(length(unique(a[[.cacheTableHashColName()]])) == 4)

  # rmFC <<- 1
  cc(ask = FALSE, x = tmpCache)
  b <- showCache(x = tmpCache) # most recent is gone
  expect_true(length(unique(b[[.cacheTableHashColName()]])) == 3)

  # ._rmFromCache_1 <<- 1
  cc(thisTime, ask = FALSE, x = tmpCache)
  d <- showCache(x = tmpCache) # all those after this time gone, i.e., only 1 left
  expect_true(length(unique(d[[.cacheTableHashColName()]])) == 1)

  cc(ask = FALSE, x = tmpCache) # Cache is
  b1 <- showCache(x = tmpCache) # most recent is gone
  expect_true(length(unique(b1[[.cacheTableHashColName()]])) == 0)

  mess <- capture_messages(cc(ask = FALSE, x = tmpCache)) # Cache is already empty
  expect_true(any(cli::ansi_grepl("Cache already empty", mess)))
})

test_that("test .defaultUserTags", {
  testInit()

  b <- Cache(rnorm, 1, cachePath = tmpCache)
  sc <- showCache(tmpCache)
  actualTags <- sc$tagKey %in% .defaultUserTags
  anyNewTags <- any(!actualTags)
  if (isTRUE(anyNewTags)) stop("A new default userTag was added; please update .defaultUserTags")
  expect_false(anyNewTags)
})

test_that("test changing reproducible.cacheSaveFormat midstream", {

  formA <- c(.qs2Format, .rdsFormat, .qsFormat)
  forms <- expand.grid(A = formA, B = formA)
  forms <- forms[forms[["A"]]!=forms[["B"]],]
  forms <- forms[c(1, 2, 4),]

  for (ind in seq_len(NROW(forms))) {
    needed <- unname(unlist(t(forms[ind,]))[,1])
    neededpkgs <- setdiff(needed, "rds")
    have <- vapply(neededpkgs, function(need)  {
      requireNamespace(need, quietly = TRUE)
    }, logical(1))
    if (!all(have))
      next

    testInit(opts = list(
      reproducible.cacheSaveFormat = needed[1],
      reproducible.qsFormat = setdiff(needed, "rds")[1],
      reproducible.useMemoise = FALSE
    ))

    b <- Cache(rnorm, 1, cachePath = tmpdir)
    sc <- showCache(tmpdir)
    # ci <- unique(sc[[.cacheTableHashColName()]])
    withr::local_options(reproducible.cacheSaveFormat = needed[2],
                         reproducible.qsFormat = tail(setdiff(needed, "rds"), 1))
    # if (ind == 3)
    #   aaaa <<- 1; on.exit(rm(aaaa, envir = .GlobalEnv))
    # if (exists("aaaa", envir = .GlobalEnv)) browser()
    mess <- capture_messages({
      b <- Cache(rnorm, 1, cachePath = tmpdir)
    })
    expect_false(attr(b, ".Cache")$newCache)
    expect_true(sum(cli::ansi_grepl(paste0("Changing format of Cache entry from ",
                                           needed[1], " to ", needed[2]), mess)) == 1)

    withr::local_options(reproducible.cacheSaveFormat = needed[1],
                         reproducible.qsFormat = setdiff(needed, "rds")[1])
    mess <- capture_messages({
      b <- Cache(rnorm, 1, cachePath = tmpdir)
    })
    expect_false(attr(b, ".Cache")$newCache)
    expect_true(sum(cli::ansi_grepl(paste0("Changing format of Cache entry from ",
                                           needed[2], " to ", needed[1]), mess)) == 1)
  }
})

test_that("test file link with duplicate Cache", {
  testInit(verbose = TRUE, opts = list("reproducible.useMemoise" = FALSE))


  sam <- function(...) {
    sample(...)
  }

  sam1 <- function(...) {
    1
    sample(...)
  }

  N <- .objectSizeMinForBig/4

  set.seed(123)
  mess1 <- capture_messages({
    b <- Cache(sam, N, cachePath = tmpCache)
  })

  # Change in RSQLite 2.2.2 -- there is now a random number used in dbAppend,
  #   so this test no longer works after the second time -- running it a 3rd time
  #   is sufficient for the test. The point is, if it is an identical result,
  #   then there will be a file.link
  set.seed(123)
  mess2 <- capture_messages({
    d <- Cache(sample, N, cachePath = tmpCache)
  })

  set.seed(123)
  mess3 <- capture_messages({
    g <- Cache(sam1, N, cachePath = tmpCache)
  })

  expect_true(sum(cli::ansi_grepl("A file with identical", mess3)) == 1)

  set.seed(123)
  mess1 <- capture_messages({
    b <- Cache(sam, N, cachePath = tmpCache)
  })
  set.seed(123)
  # Because of RSQLite 2.2.2 this 2nd time is not considered identical -- need 3rd time
  mess2 <- capture_messages({
    d <- Cache(sample, N, cachePath = tmpCache)
  })
  clearCache(tmpCache, userTags = gsub("cacheId:", "", attr(b, "tags")))
  set.seed(123)
  mess2 <- capture_messages({
    d <- Cache(sam1, N, cachePath = tmpCache)
  })
  expect_true(any(cli::ansi_grepl(.message$LoadedCacheResult(), mess2)))
  expect_true(any(cli::ansi_grepl(.message$LoadedCacheResult(), mess1)))
  # There are intermittent "status 5" warnings on next line on Windows -- not relevant here
  warns <- capture_warnings({
    out1 <- try(system2("du", paste0("\"", tmpCache, "\""), stdout = TRUE), silent = TRUE)
  })
  # out1 <- try(system2("du", tmpCache, stdout = TRUE), silent = TRUE)
  if (!is(out1, "try-error")) {
    fs1 <- as.numeric(gsub("([[:digit:]]*).*", "\\1", out1))
  }

  # It must be same output, not same input
  clearCache(tmpCache)

  # N <- .objectSizeMinForBig/8 # if it gets bigger, then "Saving large object" will be a message

  set.seed(123)
  mess1 <- capture_messages({
    b <- Cache(sam, N, cachePath = tmpCache)
  })
  set.seed(1234)
  mess2 <- capture_messages({
    d <- Cache(sample, N, cachePath = tmpCache)
  })
  # Different inputs AND different output -- so no cache recovery and no file link
  expect_match(regexp = gsub("\\!", ".", .message$SavedTxt), mess2, all = FALSE)
  out2 <- try(system2("du", tmpCache, stdout = TRUE), silent = TRUE)
  if (!is(out2, "try-error")) {
    fs2 <- as.numeric(gsub("([[:digit:]]*).*", "\\1", out2))
    expect_true(all(fs1 * 1.5 < fs2)) # macos needed this to be smaller than 1.9; I think b/c mac uses 8kb block size
  }
  # Test if the `try` works if the file.link is not to a meaningful file
  cacheIds <- unique(showCache(tmpCache)[[.cacheTableHashColName()]])
  unlink(dir(CacheStorageDir(tmpCache), pattern = cacheIds[1], full.names = TRUE))
  unlink(dir(CacheStorageDir(tmpCache), pattern = cacheIds[2], full.names = TRUE))
  set.seed(1234)

  warn <- capture_warnings({
    d1 <- Cache(sam1, N, cachePath = tmpCache)
  })
  expect_true(length(warn) == 0) ## TODO: sometimes not true?
})

test_that("test .object arg for list in Cache", {
  testInit()
  withr::local_options(reproducible.cachePath = tmpdir)
  l <- list(a = 1, b = 2, f = 3)
  out1 <- Cache(unlist, l, .objects = "a")
  out2 <- Cache(unlist, l, .objects = "b")
  out3 <- Cache(unlist, l)
  out4 <- Cache(unlist, l, .objects = "c") # not in list, so cache empty list
  out5 <- Cache(unlist, l, .objects = "d")
  out6 <- Cache(unlist, l, .objects = c("a", "b", "f"))
  expect_false(identical(out1, out2))
  expect_false(identical(out1, out3))
  expect_false(identical(out2, out3))
  expect_true(sum(out4 - out5) == 0)
  expect_true(sum(out3 - out6) == 0)

  l <- list(a = 1, b = 2, f = 3, g = 4) # change list
  out7 <- Cache(unlist, l, .objects = c("a", "b", "f")) # subset should still be same as prev whole list
  expect_true(sum(out3 - out7) == 0)

  out1 <- .robustDigest(l, .objects = "a")
  out2 <- .robustDigest(l, .objects = "b")
  out3 <- .robustDigest(l)
  expect_false(identical(out1, out2))
  expect_false(identical(out1, out3))
  expect_false(identical(out2, out3))
})

test_that("quick arg in Cache as character", {
  skip_on_cran()
  testInit("terra", verbose = TRUE, tmpFileExt = c(.rdsFormat, "tif"))

  tf <- tmpfile[[1]]
  tf2 <- tmpfile[[2]]

  messes <- list()
  quicks <- rep(list(FALSE, FALSE, "file", "file", TRUE, TRUE), 2)
  rasRan <- c(rep(TRUE, 6), rep(FALSE, 6))

  ##     quicks ranRas
  ## 1   FALSE   TRUE
  ## 2   FALSE   TRUE
  ## 3    file   TRUE
  ## 4    file   TRUE
  ## 5    TRUE   TRUE
  ## 6    TRUE   TRUE
  ## 7   FALSE  FALSE
  ## 8   FALSE  FALSE
  ## 9    file  FALSE
  ## 10   file  FALSE
  ## 11   TRUE  FALSE
  ## 12   TRUE  FALSE

  for (i in seq(quicks)) {
    vals <- if (rasRan[i]) sample(1:100) else 1:100
    ranRas <- terra::rast(terra::ext(0, 10, 0, 10), vals = vals)
    ranRas <- suppressWarningsSpecific(
      falseWarnings = proj6Warn,
      writeRaster(ranRas, filename = asPath(tf2), overwrite = TRUE)
    )
    a <- sample(1e7, 1)
    saveRDS(a, file = tf)

    # new copy
    messes[[i]] <- capture_messages(Cache(saveRDS, ranRas,
                                          file = asPath(tf), cachePath = tmpCache,
                                          quick = quicks[[i]]
    ))
  }

  expect_true(length(messes[[6]]) > 0) # undesirable: quick = TRUE -- raster has changed
  expect_true(length(messes[[8]]) == 1) # undesirable: quick = FALSE -- raster & file not changed

  ## Desired: 9 not cache, 10 cached. But 9 is picking up cache from 4.
  expect_true(length(messes[[9]]) == 1) # undesirable: quick = 'file' -- raster & file changed
  expect_true(length(messes[[10]]) > 0) # undesirable: quick = 'file -- raster & file not changed
  expect_equal(sum(unlist(lapply(messes, function(x) length(x) > 1))), 4L)
})

test_that("List of Rasters", {
  testInit("terra", tmpFileExt = c("tif", "tif"))

  listOfRas <- lapply(1:2, function(x) {
    vals <- sample(1:100)
    ranRas <- terra::rast(terra::ext(0, 10, 0, 10), vals = vals)
    ranRas <- suppressWarningsSpecific(
      falseWarnings = proj6Warn,
      writeRaster(ranRas, filename = tmpfile[[x]], overwrite = TRUE)
    )
  })

  writeRasterList <- function(rasList) {
    lapply(rasList, function(ras) {
      filename <- paste0(Filenames(ras), rndstr(1, 6), ".tif")
      ras[] <- ras[]
      ranRas <- suppressWarningsSpecific(
        falseWarnings = proj6Warn,
        writeRaster(ras, filename = filename, overwrite = FALSE)
      )
    })
  }
  a <- Cache(writeRasterList, listOfRas, cachePath = tmpCache)
  b <- Cache(writeRasterList, listOfRas, cachePath = tmpCache)
  expect_false(isTRUE(attr(b, ".Cache")$newCache))
  expect_true(isTRUE(attr(a, ".Cache")$newCache))
})

test_that("Cache the dots; .cacheExtra", {
  testInit()

  fn1 <- function(a, b, ...) {
    out <- fn2(...)
    return(out)
  }

  fn2 <- function(d, e = 0, f = 1) {
    rnorm(d, e, f)
  }

  suppressMessages({
    out1 <- Cache(fn1, a = 1, b = 2, d = 1, cachePath = tmpCache)
  })
  suppressMessages({
    out2 <- Cache(fn1, a = 1, b = 2, d = 2, cachePath = tmpCache)
  })
  expect_true(!identical(out1, out2))


  fn3 <- function(a, b) {
    out <- fn2(d = 1)
  }

  suppressMessages({
    out3 <- Cache(fn3, a = 1, b = 2, .cacheExtra = "12342", cachePath = tmpCache)
  })
  suppressMessages({
    out4 <- Cache(fn3, a = 1, b = 2, .cacheExtra = "123422", cachePath = tmpCache)
  })
  expect_true(!identical(out3, out4))

  # These are now the same because the .cacheExtra is the same and the arg is ignored
  suppressMessages({
    out5 <- Cache(mean, 6, omitArgs = "x", .cacheExtra = "234", cachePath = tmpCache)
  })
  suppressMessages({
    out6 <- Cache(mean, 7, omitArgs = "x", .cacheExtra = "234", cachePath = tmpCache)
  })
  expect_true(out6 - 6 == 0) # takes first one
  attr(out5, ".Cache") <- NULL
  attr(out6, ".Cache") <- NULL
  expect_equal(out5, out6, ignore_attr = TRUE )# the attributes will be different because one is a recovery of the other
})

test_that("change to new capturing of FUN & base pipe", {
  testInit(opts = list(reproducible.verbose = 5))

  Nrand2 <- Nrand <- 1e6
  mess0 <- capture_messages({
    out0 <- Cache(rnorm(1, 2, round(mean(runif(Nrand, 1, 1.1)))), cachePath = tmpCache)
  })

  mess1 <- capture_messages({
    out1 <- Cache(do.call(rnorm, list(1, 2, sd = round(mean(runif(Nrand2, 1, 1.1))))),
                  cachePath = tmpCache)
  })

  # NO LONGER THE SAME CALL AS ABOVE
  f1 <- paste("
      {runif(1e6, 1, 1.1) |>
        mean() |>
        round() |>
        rnorm(1, 2, sd = _)} |> # _ Only works with R >= 4.2.0
        Cache(cachePath = tmpCache)
    ")
  mess2 <- capture_messages({
    out2 <- eval(parse(text = f1))
  })
  f2 <- paste("out3 <- {runif(Nrand, 1, 1.1) |>
        mean() |>
        round() |>
        rnorm(1, 2, sd = _)} |> # _ Only works with R >= 4.2.0
        # (function(xx) rnorm(1, 2, sd = xx))() |>
        Cache(cachePath = tmpCache)
    ")
  mess3 <- capture_messages(
    eval(parse(text = f2))
  )
  expect_true(attr(out0, ".Cache")$newCache)
  expect_false(attr(out1, ".Cache")$newCache)
  expect_true(attr(out2, ".Cache")$newCache)
  expect_false(attr(out3, ".Cache")$newCache)

  expect_true(length(grep("\\<sd\\>", mess0)) == 1) # digests just the 1
  expect_true(length(grep("\\<sd\\>", mess1)) == 1) # digests just the 1
  expect_true(length(grep("\\<sd\\>", strsplit(mess2[[1]], ":")[[1]])) == 1) # digests each element
  expect_true(length(grep("\\<sd\\>", strsplit(mess3[[1]], ":")[[1]])) == 1) # digests each element

  clearCache(tmpCache)
  for (i in 1:3) Cache(rnorm, i, cachePath = tmpCache)
  expect_true(length(unique(showCache(tmpCache)$cacheId)) == 3)
  # This would make sense it if only generates one Cache entry... i.e., do not evaluate the sample

  clearCache(tmpCache)


  for (i in 1:3) {
    sss <- paste("sample(100000000, ", i, ") |>  # creates 1 random number
      rnorm(1, 2, sd = _) |>  # passed to sd of rnorm
      Cache(cachePath = tmpCache)")
    eval(parse(text = sss))
  }
  sc <- data.table::copy(showCache(tmpCache))
  expect_true(length(unique(sc$cacheId)) == 3)

  # This, different way; not evaluating `sample`; so this is same as prev
  for (i in 1:3) Cache(rnorm(1, 2, sample(10000000, i)), cachePath = tmpCache)
  sc1 <- showCache(tmpCache)
  expect_true(length(unique(sc1$cacheId)) == 6)
  expect_true(NROW(sc1) > NROW(sc))


  # Try with squiggly braces
  out0 <- Cache(rnorm(1, 2, round(mean(runif(Nrand, 1, 1.1)))), cachePath = tmpCache)
  f1 <- paste("
      {runif(Nrand, 1, 1.1) |>
        mean() |>
        round() |>
        rnorm(1, 2, sd = _)} |> # _ Only works with R >= 4.2.0
        Cache(cachePath = tmpCache)
    ")
  mn <- 1
  st3 <- system.time({
    out2 <- eval(parse(text = f1))
  })
  st4 <- system.time({
    out3 <- Cache(
      {
        rnorm(1, 2, round(mean(runif(Nrand, 1, 1.1))))
      },
      cachePath = tmpCache
    )
  })
  # can pass a variable, but not a function
  st5 <- system.time({
    out3 <- Cache(
      {
        rnorm(1, 2, round(mean(runif(Nrand, mn, 1.1))))
      },
      cachePath = tmpCache
    )
  })
  f1 <- paste("
      { a <- runif(Nrand, 1, 1.1)
        b <- mean(a)
        d <- round(b)
        rnorm(1, 2, sd = d)} |> # _ Only works with R >= 4.2.0
        Cache(cachePath = tmpCache)
    ")
  err <- capture_error({
    out2 <- eval(parse(text = f1))
  })
  expect_true(is(err, "simpleError"))

  # Test for new `round` in R > 4.3.1 with ... i.e., a primitive with method dispatch
  f1 <- paste("
      {runif(1e6, 1, 1.1) |>
        mean() |>
        round()} |> # _ Only works with R >= 4.2.0
        Cache(cachePath = tmpCache)
    ")
  expect_no_error({
    mess2 <- capture_messages({
      out2 <- eval(parse(text = f1))
    })
  })
})

test_that("test cache with new approach to match.call", {
  testInit(opts = list("reproducible.verbose" = -2))
  withr::local_options(reproducible.cachePath = tmpdir)

  b <- list(fun = rnorm)
  a <- list()
  clearCache(ask = FALSE)

  bbb <- 1
  ee <- new.env(parent = emptyenv())
  ee$qq <- bbb
  a[[1]] <- Cache(rnorm(1)) # no evaluation prior to Cache
  a[[2]] <- Cache(rnorm, 1) # no evaluation prior to Cache
  a[[3]] <- Cache(do.call, rnorm, list(1))
  a[[4]] <- Cache(do.call(rnorm, list(1)))
  a[[5]] <- Cache(do.call(b$fun, list(1)))
  a[[6]] <- Cache(do.call, b$fun, list(1))
  a[[7]] <- Cache(b$fun, 1)
  a[[8]] <- Cache(b$fun(1))
  a[[9]] <- Cache(quote(rnorm(1)))
  a[[10]] <- Cache(stats::rnorm(1))
  a[[11]] <- Cache(stats::rnorm, 1)
  a[[12]] <- Cache(rnorm(1, 0, get("bbb", inherits = FALSE)))
  a[[13]] <- Cache(rnorm(1, 0, get("qq", inherits = FALSE, envir = ee)))
  a[[14]] <- Cache(rnorm(1, bbb - bbb, get("bbb", inherits = FALSE)))
  a[[15]] <- Cache(rnorm(sd = 1, 0, n = get("bbb", inherits = FALSE))) # change order
  a[[16]] <- Cache(rnorm(1, sd = get("ee", inherits = FALSE)$qq), mean = 0)
  a[[17]] <- eval(parse(text = "b$fun(1) |> Cache()"))
  ss <- '{"bbb" |>
      parse(text = _) |>
      eval() |>
      rnorm()} |>
    Cache()'
  a[[18]] <- eval(parse(text = ss))
  expect_identical(1L, length(unique(unlist(a))))
  # expect_true(identical(attr(a[[1]], ".Cache")$newCache, TRUE))
  # for (i in 2:NROW(a)) {
  #   expect_true(identical(attr(a[[i]], ".Cache")$newCache, FALSE))
  # }

  for (fun in list(.robustDigest, print)) {
    clearCache(ask = FALSE)
    b <- list(fun = fun)
    a <- list()
    lala <- capture.output(a[[1]] <- Cache(fun(1)))
    a[[2]] <- Cache(fun, 1)
    a[[3]] <- Cache(do.call, fun, list(1))
    a[[4]] <- Cache(do.call(fun, list(1)))
    a[[5]] <- Cache(do.call(b$fun, list(1)))
    a[[6]] <- Cache(do.call, b$fun, list(1))
    a[[7]] <- Cache(b$fun, 1)
    a[[8]] <- Cache(b$fun(1))
    a[[9]] <- Cache(quote(fun(1)))
    # expect_true(identical(attr(a[[1]], ".Cache")$newCache, TRUE))

    a[[9]] <- eval(parse(text = "b$fun(1) |> Cache()"))

    expect_identical(1L, length(unique(unlist(a))))

    # for (i in 2:NROW(a)) {
    #   test <- identical(attr(a[[i]], ".Cache")$newCache, FALSE)
    #   if (isFALSE(test)) browser()
    #   expect_true(test)
    # }
  }

  if (.requireNamespace("terra")) {
    m <- matrix(1:4, nrow = 2)
    clearCache(ask = FALSE)
    a <- list()
    lala <- capture.output(a[[1]] <- Cache(terra::rast(m, digits = 4)))
    a[[2]] <- Cache(terra::rast, m, digits = 4)
    a[[3]] <- Cache(do.call, terra::rast, list(m, digits = 4))
    a[[4]] <- Cache(do.call(terra::rast, list(m, digits = 4)))
    a[[5]] <- Cache(quote(terra::rast(m, digits = 4)))
    dig <- .robustDigest(a)
    # dig <- dig[!nzchar(names(dig))]
    expect_identical(1L, length(unique(unlist(dig))))
    # expect_true(identical(attr(a[[1]], ".Cache")$newCache, TRUE))
    # for (i in 2:NROW(a)) {
    #   test <- identical(attr(a[[i]], ".Cache")$newCache, FALSE)
    #   if (isFALSE(test)) browser()
    #   expect_true(test)
    # }
  }

  # This tries to do a method that is not actually exported from a package; the generic (sf::st_make_valid) is
  if (.requireNamespace("sf")) {
    clearCache(ask = FALSE)
    b <- list(fun = fun)
    a <- list()

    p1 <- sf::st_as_sfc("POLYGON((0 0, 0 10, 10 0, 10 10, 0 0))")
    a[[1]] <- Cache(sf::st_make_valid(p1)) # not
    a[[2]] <- Cache(sf::st_make_valid, p1) # not
    a[[3]] <- Cache(quote(sf::st_make_valid(p1))) # not

    # The warning is intermittent and related to whether sf was built for exactly this R version
    if (requireNamespace("sf", quietly = TRUE)) {
      warns <- capture_warnings(library(sf))
      a[[4]] <- Cache(st_make_valid(p1)) # not
      ff <- sf::st_make_valid
      a[[5]] <- Cache(ff(p1))

      dig <- .robustDigest(a)
      # dig <- dig[!nzchar(names(dig))]
      expect_identical(1L, length(unique(dig)))

    }
    # expect_true(identical(attr(a[[1]], ".Cache")$newCache, TRUE))
    # for (i in 2:length(a)) {
    #   test <- identical(attr(a[[i]], ".Cache")$newCache, FALSE)
    #   if (isFALSE(test)) browser()
    #   expect_true(test)
    # }
  }
})

test_that("test cache; new approach to match.call, postProcess", {
  skip_if_not_installed("DBI") # sf needs DBI
  testInit(c("terra", "sf"),
           tmpFileExt = c(".tif", ".tif"),
  )
  withr::local_options(reproducible.cachePath = tmpdir,
                       "rasterTmpDir" = tempdir2(rndstr(1, 6)),
                       "reproducible.inputPaths" = NULL,
                       "reproducible.overwrite" = TRUE)

  # Add a study area to Crop and Mask to
  # Create a "study area"
  coords <- structure(c(-122.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
                      .Dim = c(5L, 2L)
  )
  coords2 <- structure(c(-115.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
                       .Dim = c(5L, 2L)
  )

  StudyArea <- terra::vect(coords, "polygons")
  terra::crs(StudyArea) <- crsToUse

  # Sr1 <- Polygon(coords)
  # Srs1 <- Polygons(list(Sr1), "s1")
  # StudyArea <- SpatialPolygons(list(Srs1), 1L)
  # crs(StudyArea) <- crsToUse

  # Sr1 <- Polygon(coords2)
  # Srs1 <- Polygons(list(Sr1), "s1")
  # StudyArea2 <- SpatialPolygons(list(Srs1), 1L)
  # crs(StudyArea2) <- crsToUse

  StudyArea2 <- terra::vect(coords2, "polygons")
  terra::crs(StudyArea2) <- crsToUse

  nonLatLongProj <- paste(
    "+proj=lcc +lat_1=49 +lat_2=77 +lat_0=0 +lon_0=-95",
    "+x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs"
  )
  nc <- sf::st_as_sf(StudyArea) # system.file("shape/nc.shp", package="sf"))
  nc1 <- sf::st_transform(nc, nonLatLongProj)
  ncSmall <- sf::st_as_sf(StudyArea2)
  ncSmall <- sf::st_transform(ncSmall, nonLatLongProj)
  ncSmall <- sf::st_buffer(ncSmall, dist = -10000)

  # S3 classes
  b1 <- Cache(postProcess(nc1, studyArea = ncSmall, filename2 = NULL))
  expect_true(identical(attr(b1, ".Cache")$newCache, TRUE))
  b2 <- Cache(postProcess(nc1, studyArea = ncSmall, filename2 = NULL))
  expect_true(identical(attr(b2, ".Cache")$newCache, FALSE))
  b3 <- Cache(postProcess, nc1, studyArea = ncSmall, filename2 = NULL)
  expect_true(identical(attr(b3, ".Cache")$newCache, FALSE))

  # S4 classes
  c1 <- Cache(.robustDigest(data.frame(a = 1)))
  c2 <- Cache(.robustDigest(data.frame(a = 1)))
  c3 <- Cache(.robustDigest(data.frame(a = 1)), verbose = -2) # add a non-relevant arg
  c4 <- Cache(.robustDigest, data.frame(a = 1))
  c5 <- Cache(.robustDigest, data.frame(a = 1), verbose = -2)

  expect_true(identical(attr(c1, ".Cache")$newCache, TRUE))
  expect_true(identical(attr(c2, ".Cache")$newCache, FALSE))
  expect_true(identical(attr(c3, ".Cache")$newCache, FALSE))
  expect_true(identical(attr(c4, ".Cache")$newCache, FALSE))
  expect_true(identical(attr(c5, ".Cache")$newCache, FALSE))

  # S4 classes
  d1 <- Cache(Checksums(tmpdir, write = FALSE))
  d2 <- Cache(Checksums(tmpdir, write = FALSE))
  expect_true(identical(attr(c1, ".Cache")$newCache, TRUE))
  expect_true(identical(attr(c2, ".Cache")$newCache, FALSE))
})

test_that("test cache; SpatRaster attributes", {
  testInit(c("terra", "sf"),
           tmpFileExt = c(".tif", ".tif"), needInternet = TRUE
  )
  withr::local_options(reproducible.cachePath = tmpdir,
                       "rasterTmpDir" = tempdir2(rndstr(1, 6)),
                       "reproducible.inputPaths" = NULL,
                       "reproducible.overwrite" = TRUE)

  dPath <- file.path(tmpdir, "inputs")

  targetFile <- "rasterTest.tif"
  url <- "https://github.com/tati-micheletti/host/raw/master/data/rasterTest.tif"

  testFun <- function(url, targetFile) {
    ras <- prepInputs(
      url = url,
      targetFile = targetFile
    )
    pixIDs <- which(as.vector(ras[]) == 1)
    attr(ras, "pixIDs") <- pixIDs
    ras
  }

  (ras <- Cache(testFun,
               url = url,
               targetFile = targetFile
  )) |> capture.output() -> co
  expect_true(is.integer(attr(x = ras, "pixIDs")))

  ## re-run. attributes still there?
  ras <- Cache(testFun,
               url = url,
               targetFile = targetFile
  )
  expect_true(is.integer(attr(x = ras, "pixIDs")))
})

test_that("Issue 316 - writeOutputs in a non getwd dir", {
  testInit(c("terra"), tmpFileExt = c(".tif", ".tif"))

  cPath <- file.path(tmpdir, "cache")

  withr::local_options(
    "reproducible.cachePath" = cPath,
    "reproducible.destinationPath" = tmpdir
  )

  cacheTags <- "test"
  studyAreaName <- "test"

  f <- system.file("ex/elev.tif", package = "terra")
  rasterToMatch <- terra::rast(f)

  RTMvals <- as.vector(rasterToMatch[])
  rasterToMatch[!is.na(RTMvals)] <- 1

  rasterToMatchLarge <- list()
  for (i in 1:2) {
    rasterToMatchLarge[[i]] <- Cache(writeOutputs, rasterToMatch,
                                     writeTo = .suffix(
                                       file.path(tmpdir, "rasterToMatchLarge.tif"),
                                       paste0("_", studyAreaName)
                                     ), datatype = "INT2U",
                                     overwrite = TRUE, userTags = c(cacheTags, "rasterToMatchLarge"),
                                     quick = "filename2",
                                     omitArgs = c("userTags")
    )
  }

  expect_true(terra::all.equal(rasterToMatchLarge[[1]][], rasterToMatchLarge[[2]][]))
})

test_that("test useDBI TRUE <--> FALSE", {
  testInit(verbose = TRUE)
  on.exit(
    {
      useDBI(orig)
    },
    add = TRUE
  )
  withr::local_options(reproducible.cachePath = tmpdir)
  orig <- useDBI()
  useDBI(TRUE)
  d <- b <- a <- list()
  b[[1]] <- Cache(rnorm(1))
  b[[2]] <- Cache(rnorm(2))
  b[[3]] <- Cache(runif(3))
  useDBI(FALSE)
  a[[1]] <- Cache(rnorm(1))
  a[[2]] <- Cache(rnorm(2))
  a[[3]] <- Cache(runif(3))
  useDBI(TRUE)
  d[[1]] <- Cache(rnorm(1))
  d[[2]] <- Cache(rnorm(2))
  d[[3]] <- Cache(runif(3))
  lapply(a, function(aa) expect_false(attr(aa, ".Cache")$newCache))
  lapply(b, function(aa) expect_true(attr(aa, ".Cache")$newCache))
  lapply(d, function(aa) expect_false(attr(aa, ".Cache")$newCache))
})

test_that("terra files were creating file.link", {
  testInit("terra",
           tmpFileExt = c(".tif", ".tif"),
           opts = list(reproducible.useMemoise = FALSE)
  )
  rasts <- lapply(1:2, function(x)
    ras1 <- terra::rast(nrows = 1e3, ncols = 1e3, vals = sample(1e6),
                        resolution = 1, xmin = 0, xmax = 1000, ymin = 0, ymax = 1000)
  )
  func <- function(ras, fn) {
    ras <- Cache(writeRaster(ras, filename = fn), quick = c("fn"))
    ras
  }
  mess <- capture_messages(Map(f = func, fn = tmpfile, ras = rasts))
  expect_false(any(cli::ansi_grepl("file.link", mess)))

})

test_that("pass NA to userTags", {
  testInit(verbose = FALSE)
  expect_no_error({
    a <- Cache(rnorm(1), userTags = c("NA", "hi"))
  })
})

test_that("multifile cache saving", {
  skip_on_cran()
  testInit("terra",
           tmpFileExt = c(".tif", ".tif"),
           opts = list(reproducible.useMemoise = FALSE)
  )

  nOT <- Sys.time()

  randomPolyToDisk2 <- function(tmpfiles) {
    r <- terra::rast(ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
    r2 <- terra::rast(ext(0, 10, 0, 10), vals = sample(1:30, size = 100, replace = TRUE))
    .writeRaster(r, tmpfiles[1], overwrite = TRUE)
    .writeRaster(r, tmpfiles[2], overwrite = TRUE)
    r <- c(terra::rast(tmpfiles[1]), terra::rast(tmpfiles[2]))
    r
  }
  a <- Cache(randomPolyToDisk2(tmpfile), quick = "tmpfiles")
  b <- Cache(randomPolyToDisk2(tmpfile), quick = "tmpfiles")
  expect_false(attr(b, ".Cache")$newCache)
  expect_true(attr(a, ".Cache")$newCache)

  fns <- basename(.prefix(Filenames(a),  prefixCacheId(cacheId(a))))
  expect_true(all(fns %in% dir(CacheStorageDir())))
  expect_false(all(fns %in% dir(CacheStorageDir(), full.names = TRUE)))

  # expect_true(all(basename(Filenames(a)) %in% dir(CacheStorageDir())))
  # expect_false(all(Filenames(a) %in% dir(CacheStorageDir(), full.names = TRUE)))

})

test_that("test omitArgs = 'x', #400", {
  skip_if_not_installed("terra")
  testInit("terra")

  wkt <- c("MULTIPOLYGON ( ((40 40, 20 45, 45 30, 40 40)),
((20 35, 10 30, 10 10, 30 5, 45 20, 20 35),(30 20, 20 15, 20 25, 30 20)))",
           "POLYGON ((0 -5, 10 0, 10 -10, 0 -5))")
  w <- terra::vect(wkt)

  ## fails - ommitting x arg
  testthat::expect_no_error(Cache(centroids,
                                  x = w,
                                  omitArgs = c("x")))

})

test_that("cacheId = 'previous'", {
  testInit()
  withr::local_options(reproducible.cachePath = tmpdir)

  fnName <- "rnorm_this_one"
  a <- rnorm(1) |> Cache(.functionName = fnName)
  b <- rnorm(3) |> Cache(.functionName = fnName)
  d <- rnorm(2) |> Cache(.functionName = fnName, cacheId = "previous")
  e <- rnorm(2) |> Cache(.functionName = fnName)
  expect_true(all.equalWONewCache(b, d))
  expect_false(isTRUE(all.equalWONewCache(e, d)))

  # cacheId = "previous" returns normal if there is no previous
  fnName <- "rnorm_this_second"
  d <- rnorm(4) |> Cache(.functionName = fnName, cacheId = "previous")
  expect_true(unlist(attr(d, ".Cache")))
  e <- rnorm(4) |> Cache(.functionName = fnName, cacheId = "previous")
  expect_false(unlist(attr(e, ".Cache")))
})

test_that("cacheId = 'customName'", {
  testInit()
  withr::local_options(reproducible.cachePath = tmpdir)

  fnName <- "rnorm_this_one"
  d <- rnorm(2) |> Cache(.functionName = fnName, cacheId = "myCacheObj")
  e <- rnorm(3) |> Cache(.functionName = fnName, cacheId = "myCacheObj")
  expect_true(all.equalWONewCache(e, d))

  rudbi <- getOption("reproducible.useDBI")


  newDBI <- setdiff(c(TRUE, FALSE), rudbi)
  withr::local_options(reproducible.useDBI = newDBI)

  f <- rnorm(4) |> Cache(.functionName = fnName, cacheId = "myCacheObj")
  g <- rnorm(5) |> Cache(.functionName = fnName, cacheId = "myCacheObj")
  expect_true(all.equalWONewCache(e, d))
  expect_true(all.equalWONewCache(e, f))
  expect_true(all.equalWONewCache(e, g))
})

test_that("simple userTags", {
  testInit()
  withr::local_options(reproducible.cachePath = tmpdir)
  ut1 <- c("b:d", "a", "free")
  ut2 <- c("b:d", "q", "free2")
  ut3 <- "a"
  ut4 <- c("b:d", "rr", "Free2")
  ut5 <- c("b:d", "ss", "free2")
  ut6 <- c("b:d", "ttt", "free")

  funs <- list(rnorm, runif, sample, `+`, `+`, rnorm)
  ab <- Cache(funs[[1]], 2, userTags = ut1)
  ac <- Cache(runif(2e6), userTags = ut2)
  ad <- Cache(sample(1), userTags = ut3)
  a <- sample(1e5)
  aa <- Cache(a + 1, userTags = ut4)
  bb <- Cache(a + 2 - 1, userTags = ut5)
  expect_equivalent(aa, bb)

  bbb <- Cache(rnorm(2), userTags = ut6, cachePath = c(tmpdir, tempdir()))
  expect_equivalent(ab, bbb)
  sc <- showCache(userTags = "rnorm") # it was "funs[[1]]" not "rnorm"
  expect_identical(0L, length(unique(showCache(userTags = "rnorm")$cacheId)))
  sc1 <- showCache(userTags = "runif")
  sc2 <- showCache(userTags = "sample")
  sc1Tags <- vapply(strsplit(ut2, split = ":"), tail, 1, FUN.VALUE = character(1))
  sc2Tags <- vapply(strsplit(ut3, split = ":"), tail, 1, FUN.VALUE = character(1))
  expect_in(sc1Tags, sc1$tagValue)
  expect_in(sc2Tags, sc2$tagValue)



})

test_that("lightweight tests for code coverage", {
  skip_on_cran()
  out <- testInit(verbose = TRUE)
  withr::local_options(reproducible.cachePath = tmpdir)

  expect_error(Cache(), "requires")
  fn <- expect_error
  fn(Cache(compareRasterFileLength = TRUE, rnorm(1)), regexp = "compareRasterFileLength")
  fn(Cache(sideEffect = TRUE, rnorm(1)), regexp = "sideEffect")

})

test_that("test future", {
  skip_on_cran()
  skip_on_ci()
  # skip_if_not_installed("future")

  .onLinux <- .Platform$OS.type == "unix" && unname(Sys.info()["sysname"]) == "Linux"
  # if (.onLinux) {
  testInit(c("terra", "future"),
           verbose = TRUE, tmpFileExt = ".rds",
           opts = list(
             "future.supportsMulticore.unstable" = "quiet",
             "reproducible.futurePlan" = "multicore"
           )
  )

  # There is now a warning with future package
  a <- list()
  (aa <- system.time({
    for (i in c(1:3)) a[[i]] <- Cache(cachePath = tmpCache, rnorm, 1e6 + i)
  }))
  sca <- showCache(tmpCache)
  expect_true(length(unique(sca[[.cacheTableHashColName()]])) == 3)

  try(unlink(tmpCache, recursive = TRUE))
  b <- list()
  (bb <- system.time({
    for (i in 1:3) b[[i]] <- Cache(cachePath = tmpCache, rnorm(1e6 + i))
  }))
  bb <- showCache(tmpCache)
  expect_true(length(unique(bb[[.cacheTableHashColName()]])) == 3)

  # Test the speed of rerunning same line
  d <- list()
  (dd <- system.time({
    for (i in 1:3) d[[i]] <- Cache(cachePath = tmpCache, rnorm(1e6 + i))
  }))
  # windows is slower to spawn -- reduced this multiplier to *2
  expect_true((dd[[3]] * 2) < aa[[3]])
  for (i in 1:3) {
    expect_true(identical(attr(d[[i]], ".Cache")$newCache, FALSE))
  }
  # }
})


test_that("test pre-creating conn", {
  if (!useDBI()) skip("Only relevant for DBI backend")
  testInit("terra", ask = FALSE, tmpFileExt = c(".tif", ".tif"))
  on.exit({
    DBI::dbDisconnect(conn)
  })

  conn <- dbConnectAll(cachePath = tmpdir, conn = NULL)
  ra <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:100))
  rb <- terra::rast(terra::ext(0, 10, 0, 10), vals = sample(1:100))
  r1 <- Cache(.writeRaster, ra, filename = tmpfile[1], overwrite = TRUE, cachePath = tmpCache)
  r2 <- Cache(.writeRaster, rb,
              filename = tmpfile[2], overwrite = TRUE, cachePath = tmpdir,
              conn = conn
  )
  expect_true(file.exists(Filenames(r1)))
  expect_true(file.exists(Filenames(r2)))
  expect_false(grepl(basename(dirname(Filenames(r1))), "rasters")) # changed behaviour as of reproducible 1.2.0.9020
  expect_false(grepl(basename(dirname(Filenames(r2))), "rasters")) # changed behaviour as of reproducible 1.2.0.9020
})

test_that("test defunct arguments", {
  skip_if_not(getOption("reproducible.useCacheV3"))
  testInit()

  # Manual testing
  outr <- function(makeCopy = TRUE) rnorm(1)
  a <- Cache(outr, makeCopy = FALSE)
  expect_is(a, "numeric")
  expect_error(Cache(rnorm, 1, makeCopy = FALSE)) # defunct argument
  expect_error(Cache(rnorm, 1, sideEffects = FALSE)) # defunct argument

  for (d in .defunctCacheArgs) {
    fn <- eval(parse(text = paste0("function(a, ", d," = TRUE) rnorm(1)")))
    nn <- list(1, TRUE)
    names(nn) <- c("a", d)
    bb <- Cache(do.call, fn, nn)
    expect_is(bb, "numeric")

    fn <- eval(parse(text = paste0("function(a) rnorm(1)")))
    bb <- parse(text = paste0("list(Cache, fn, ", d," = TRUE, a = 2)"))
    fn <- eval(bb)
    b <- as.call(fn)
    expect_error(eval(b))
  }

})

test_that("test messaging for useMemoise = TRUE", {
  testInit()
  withr::local_options(reproducible.cachePath = tmpCache,
                       reproducible.useMemoise = TRUE)

  mess1 <- capture_messages(
    Cache(rnorm(1))
  )
  # should add to memoise on first save
  expect_match_noSlashN(object = mess1, .message$AddingToMemoised, all = FALSE)

  # should recover from memoised
  mess2 <- capture_messages(
    Cache(rnorm(1))
  )
  expect_match(object = mess2, .message$LoadedCacheResult("Memoised"), all = FALSE)
  # and not add to memoised
  expect_false(any(grepl(mess2, pattern = gsub("\\)", "", gsub("\\(", "", .message$AddingToMemoised)))))
  rm(list = ls(memoiseEnv(tmpCache)), envir = memoiseEnv(tmpCache))
  mess3 <- capture_messages(
    Cache(rnorm(1))
  )
  # should add to memoise on first load if it isn't there from the first save
  expect_match_noSlashN(object = mess3, .message$AddingToMemoised, all = FALSE)

})

test_that("cacheId is same as calculated", {
  testInit()
  withr::local_options(reproducible.cachePath = tmpCache)
  mess1 <- capture_messages(a <- Cache(rnorm, 1))

  # manually look at output attribute which shows cacheId: ca275879d5116967
  manualCacheId <- "ca275879d5116967"
  mess2 <- capture_messages(b <- Cache(rnorm, 1, cacheId = manualCacheId))
  mess2 <- gsub("\n ", "", cli::ansi_strip(mess2)) # there may be "\n " if window width too small
  expect_match(mess2, .message$cacheIdNotAssessed(manualCacheId), all = FALSE)
  expect_equivalent(a, b)
})

test_that("Cache with do.call and variables as list", {
  testInit()

  a <- list()
  a[[1]] <- Cache(do.call(rnorm, list(n = 1, mean = 2)))

  Args <- list(n = 1, mean = 2)
  a[[2]] <- Cache(do.call(rnorm, Args))

  b <- 2
  Args <- list(n = 1, mean = b)
  a[[3]] <- Cache(do.call(rnorm, Args))

  b <- function() 2
  Args <- list(n = 1, mean = b())
  a[[4]] <- Cache(do.call(rnorm, Args))

  # previously, Args would just be unevaluated so 2, 3, 4 would be same, but different than 1
  expect_equivalent(length(unique(as.numeric(a))), 1L)

})

test_that("ensure default tags are correct", {
  testInit()

  # This test is necessary because it uses these in a very specific order; CANNOT CHANGE THESE
  #  without being fully aware of changes
  # cacheTagsSecondGroup <- c("class", "object.size", "fromDisk", "resultHash", "elapsedTimeFirstRun")
  # cacheTagsFirstGroup <- c("function", "userTags", "accessed", "inCloud", "elapsedTimeDigest", "preDigest")
  cacheTagsFirstGroup <- .cacheTagsFirstGroup
  cacheTagsSecondGroup <- .cacheTagsSecondGroup

  expect_identical(.cacheTagsFirstGroup, cacheTagsFirstGroup)
  expect_identical(.cacheTagsSecondGroup, cacheTagsSecondGroup)

  missingFromDefault <- setdiff(.cacheTagsSecondGroup, .cacheTagsDefault)
  expect_true(length(missingFromDefault) == 0)

  missingFromDefault <- setdiff(.cacheTagsFirstGroup, .cacheTagsDefault)
  expect_true(length(missingFromDefault) == 0)

})

test_that("cacheChaining", {
  testInit()
  withr::local_options(reproducible.verbose = TRUE,
                       reproducible.useMemoise = FALSE)
  samWMean <- function(x, size, other) {
    sample(x, size) |> mean()
  }
  arb <- list()
  args <- c(3,5)
  N <- 1e6
  df <- list(a = list(TRUE, NULL, TRUE), b = list(NULL, NULL, NULL))
  for (dfIndex in 1:2) {
    for (arg in args) { # These are 2 different functions (below); one is identical each time, the other is not
      clearCache(ask = F, verbose = FALSE)
      mess <- list()
      sc <- list()
      index <- 0L

      for (index in 1:3) {
        # i <- c(TRUE, FALSE, TRUE)[index]
        #if (index == 3 && arg == args[2])
        i <- df[[dfIndex]][[index]]
        # index <- index + 1L
        withr::local_seed(123)
        iChar <- as.character(index)
        withr::local_options(reproducible.cacheChaining = i)
        fn1 <- function(x) {
          a <- sample(N) |> Cache()
          b <- samWMean(a, size = length(a) * 0.9, other = x) |> Cache()
          d <- samWMean(a, size = length(a) * 0.8, other = x) |> Cache()
          c(mean(a), b, d)
        }
        if (arg == args[1]) {
          fn2 <- function(y) {
            fn1(2)
          }
        } else {
          fn2 <- function(y) {
            fn1(sample(1e6, size = 1))
          }
        }
        arb[[iChar]] <- list()
        if (index < 3) { # don't clear the 3rd one so that we can test that cacheChaining doesn't need a new entry in the Cache
          clearCache(ask = FALSE)
        }
        mess[[iChar]] <- capture_messages({
          arb[[iChar]][[1]] <- fn2() # a --> calculate & slow; b --> no digest, but still calculate & slow; d --> no digest, still calculate & slow
          arb[[iChar]][[2]] <- fn2()# a --> digest & return cache; b --> skip digest, return cache; d --> skip digest, return cache
        })
        sc[[iChar]] <- showCache(verbose = FALSE)[tagKey == "elapsedTimeDigest"]

      }

      # Should be the same pattern of saving/loading regardless of chainCaching
      #if (dfIndex == 2)
      #   browser()
      expect_equivalent(length(grep("Saved", mess$`2`)), arg)
      expect_equivalent(length(grep("Saved", mess$`1`)), arg)
      comp <- if (arg == args[1]) 0 else 4
      expect_equivalent(length(grep("Saved", mess$`3`)), comp) # the 2 that are `fn1(sample(1e6, size = 1))`


      # Should only show the messaging when cacheChaining is on
      if (dfIndex == 1 && (useDBI() %in% FALSE)) {
        expect_equivalent(length(grep("cacheChaining", mess$`1`)), 6)
        expect_equivalent(length(grep("Skipping digest", mess$`1`)), 4)
        expect_equivalent(length(grep("cacheChaining", mess$`2`)), 0)
        expect_equivalent(length(grep("Skipping digest", mess$`2`)), 0)
        expect_equivalent(length(grep("cacheChaining", mess$`3`)), 6)
        expect_equivalent(length(grep("Skipping digest", mess$`3`)), 4)
      }

      # Basically, 2 of the 3 MUST be faster to digest — unreliable because
      # sample(1e6) is fast to digest; to confirm, set N to 1e7 and uncomment:
      #   expect_true(sum(sc$`1`$tagValue < sc$`2`$tagValue) >= 2)

      # cacheChaining shouldn't change anything; they should be the same
      expect_equivalent(arb$`TRUE`, arb$`FALSE`)
    }
  }


  if (FALSE) { # for benchmarking
    # fn2 <- function(y) {
    #   fn1(2)
    # }
    # TF <- FALSE
    # withr::local_options(reproducible.verbose = FALSE)
    # mbs <- list()
    # for (TF in c(TRUE, FALSE)) {
    #   mbs[[as.character(TF)]] <- microbenchmark::microbenchmark(N1 = {options(reproducible.cacheChaining = TF); N <- 10^1; fn2(); fn2()},
    #                                         N2 = {options(reproducible.cacheChaining = TF); N <- 10^2; fn2(); fn2()},
    #                                         N3 = {options(reproducible.cacheChaining = TF); N <- 10^3; fn2(); fn2()},
    #                                         N4 = {options(reproducible.cacheChaining = TF); N <- 10^4; fn2(); fn2()},
    #                                         N5 = {options(reproducible.cacheChaining = TF); N <- 10^5; fn2(); fn2()},
    #                                         N6 = {options(reproducible.cacheChaining = TF); N <- 10^6; fn2(); fn2()},
    #                                         N7 = {options(reproducible.cacheChaining = TF); N <- 10^7; fn2(); fn2()},
    #                                         N7.5 = {options(reproducible.cacheChaining = TF); N <- 10^7.5; fn2(); fn2()},
    #                                         times = 4)
    # }
  }
})

test_that("Cache with weird dots", {
  testInit()
  a <- as.data.frame(list(1, 2), F, F, a = 2, rr = 23) |> Cache()
  b <- as.data.frame(list(1, 2), F, F, a = 2, rr = 23) |> Cache()
  d <- as.data.frame(list(1, 2), F, F, a = 3, rr = 23) |> Cache()
  expect_true(attr(a, ".Cache")$newCache)
  expect_false(attr(b, ".Cache")$newCache)
  expect_true(attr(d, ".Cache")$newCache)

})

test_that("Cache works when `...` is the first formal (#466)", {
  testInit()

  myFun <- function(..., arg1, arg2) {
    rnorm(arg1, arg2, ...)
  }
  myFun2 <- function(arg1, arg2, ...) {
    rnorm(arg1, arg2, ...)
  }

  a <- myFun(arg1 = 100, arg2 = 2, sd = 10) |> Cache()
  b <- myFun(arg1 = 100, arg2 = 2, sd = 10) |> Cache()
  d <- myFun(arg1 = 100, arg2 = 2, sd = 99) |> Cache()
  expect_true(attr(a, ".Cache")$newCache)
  expect_false(attr(b, ".Cache")$newCache)
  expect_true(attr(d, ".Cache")$newCache)

  a2 <- myFun2(arg1 = 100, arg2 = 2, sd = 10) |> Cache()
  b2 <- myFun2(arg1 = 100, arg2 = 2, sd = 10) |> Cache()
  d2 <- myFun2(arg1 = 200, arg2 = 2, sd = 10) |> Cache()
  expect_true(attr(a2, ".Cache")$newCache)
  expect_false(attr(b2, ".Cache")$newCache)
  expect_true(attr(d2, ".Cache")$newCache)
})

test_that(".digest with empty and broken files", {
  testInit()
  tf <- tempfile()
  # tf2 <- tempfile(fileext = ".tif")
  file.create(tf)
  # file.create(tf2)
  a <- .digest(file = tf)
  file.remove(tf)
  expect_error(.digest(file = tf))
  # .digest(file = tf2)

  expect_true(is.character(a))
  expect_true(length(a) == 1)

})

test_that("test parallel collisions", {
  skip_on_cran() # testing multi-threaded things on CRAN
  # skip_on_os("mac")
  skip_if_not_installed("parallel")

  startTime <- Sys.time()

  testInit("parallel", tmpFileExt = c(".tif", ".grd", ".txt"))

  # make cluster -- note this works if cluster is FORK also, but for simplicity, using default
  #   which works on Linux, Mac, Windows
  N <- min(2L, parallel::detectCores())

  if (useDBI()) {
    if (!file.exists(CacheDBFile(tmpdir))) {
      createCache(tmpdir)
    }
  }

  # make function that will write to cache repository from with clusters
  numNew <- 20
  fun <- function(x, cachePath, numNew, keepLog) {
    require("reproducible")
    if (keepLog) {
      fn <- file.path("c:/Eliot/tmpCache/log", Sys.getpid())
      checkPath(dirname(fn), create = TRUE)
    }
    withCallingHandlers(
      Cache(rnorm, 10,
            sd = x %% numNew + 1, cachePath = cachePath,
            verbose = 4
      ), # The "overwrite" means delete, rewrite, delete, rewrite over and over
      message = function(m) {
        if (keepLog) {
          cat(m$message, file = fn, append = TRUE)
        }
      }
    )
  }
  cl <- try(makeCluster(N), silent = TRUE)
  if (!is(cl, "try-error")) {
    on.exit(parallel::stopCluster(cl), add = TRUE)

    keepLog <- FALSE
    if (isWindows() && Sys.info()["user"] == "emcintir") {
      # tmpdir <- 'c:/Eliot/tmpCache/'
      keepLog <- TRUE
    }

    clusterSetRNGStream(cl)
    parallel::clusterEvalQ(cl, {
      require("reproducible")
    })
    numToRun <- 140
    # if (interactive()) {
    #   print(tmpdir)
    # }

    # There is a 'creating Cache at the same time' problem -- haven't resolved
    #  Just make cache first and it seems fine
    clearCache(tmpdir) # If in c:/Eliot, then it may be reused
    Cache(rnorm, 1, cachePath = tmpdir)
    a <- try(
      clusterMap(
        cl = cl, fun, seq(numToRun), numNew = numNew, keepLog = keepLog,
        cachePath = tmpdir, .scheduling = "dynamic"
      ),
      silent = FALSE
    )
    if (!is(a, "try-error")) {
      expect_true(is.list(a))
      expect_true(length(a) == numToRun)
      news <- vapply(a, function(x) attr(x, ".Cache")$newCache, FUN.VALUE = logical(1))
      expect_equal(sum(news), numNew)
    }
    endTime <- Sys.time()
  }

})

test_that("Cache() lazily spawns the async showCache job for the cachePath used", {
  ## Regression for: .onLoad-only spawn missed the user's real cachePath
  ## (set later by setupProject), so showCache() ran a 60s sync scan on
  ## first call. The lazy spawn from Cache() now targets whichever
  ## cachePath the caller actually uses.
  skip_on_cran()
  if (.Platform$OS.type == "windows")
    skip("forking-based; not relevant on Windows")
  if (!requireNamespace("parallel", quietly = TRUE))
    skip("parallel not available")

  tmpCache <- file.path(tempdir(), basename(tempfile("rcache_lazy_")))
  dir.create(tmpCache, showWarnings = FALSE, recursive = TRUE)
  withr::local_options(reproducible.cachePath  = tmpCache,
                       reproducible.useMemoise = FALSE,
                       reproducible.verbose    = 0)

  ## Pre-condition: no spawn job exists yet for this cachePath
  pkgEnv <- reproducible:::memoiseEnv(cachePath = tmpCache)
  expect_false(exists("shownCache", envir = pkgEnv) &&
               !is.null(pkgEnv[["shownCache"]]$shownCache_jobs) &&
               exists(tmpCache, envir = pkgEnv[["shownCache"]]$shownCache_jobs,
                      inherits = FALSE),
               info = "Pre-condition: no async job before Cache() is called")

  ## Calling Cache() should kick off the spawn
  invisible(Cache(rnorm, 1, cachePath = tmpCache,
                  cacheId = "lazy_v1", useCloud = FALSE))

  ## Post-condition: a spawn job is now registered for this cachePath
  expect_true(exists("shownCache", envir = pkgEnv),
              info = "Cache() should have created the shownCache env")
  jobsEnv <- pkgEnv[["shownCache"]]$shownCache_jobs
  expect_true(is.environment(jobsEnv),
              info = "shownCache_jobs should be an environment")
  expect_true(exists(tmpCache, envir = jobsEnv, inherits = FALSE),
              info = "A spawn job should be registered for the cachePath")
})

test_that("prepopulateCacheAsync() is exported and idempotent", {
  skip_on_cran()
  if (.Platform$OS.type == "windows")
    skip("forking-based; not relevant on Windows")
  if (!requireNamespace("parallel", quietly = TRUE))
    skip("parallel not available")

  tmpCache <- file.path(tempdir(), basename(tempfile("rcache_prep_")))
  dir.create(tmpCache, showWarnings = FALSE, recursive = TRUE)

  ## Exported
  expect_true(exists("prepopulateCacheAsync",
                     envir = asNamespace("reproducible"),
                     inherits = FALSE))

  ## First call schedules a job
  reproducible::prepopulateCacheAsync(tmpCache)
  pkgEnv <- reproducible:::memoiseEnv(cachePath = tmpCache)
  jobsEnv <- pkgEnv[["shownCache"]]$shownCache_jobs
  expect_true(exists(tmpCache, envir = jobsEnv, inherits = FALSE))
  job1 <- get(tmpCache, envir = jobsEnv, inherits = FALSE)

  ## Second call reuses the same job (idempotent: spawn_showCache_async
  ## returns the existing job under overwrite = FALSE)
  reproducible::prepopulateCacheAsync(tmpCache)
  job2 <- get(tmpCache, envir = jobsEnv, inherits = FALSE)
  expect_identical(job1$pid, job2$pid)
})

test_that("prepopulateCacheAsync() is a no-op for invalid inputs", {
  if (.Platform$OS.type == "windows")
    skip("forking-based; not relevant on Windows")

  ## NULL / empty / non-character should silently no-op
  expect_silent(reproducible::prepopulateCacheAsync(NULL))
  expect_silent(reproducible::prepopulateCacheAsync(""))
  expect_silent(reproducible::prepopulateCacheAsync(123))
})

##########################
test_that("test multiple cachePath", {
  testInit()

  withr::local_options(
    "reproducible.cachePath" = c(tmpdir, tmpCache),
    reproducible.useMemoise = TRUE)

  for (memoise in c(TRUE, FALSE)) {
    options(reproducible.useMemoise = memoise)
    a <- Cache(rnorm, 1, cachePath = tmpCache)
    suppressMessages(aCache <- showCache(tmpCache))
    expect_true(length(unique(aCache[[.cacheTableHashColName()]])) == 1)

    b <- Cache(rnorm, 2, cachePath = tmpdir)
    suppressMessages(bCache <- showCache(tmpdir))
    expect_true(length(unique(bCache[[.cacheTableHashColName()]])) == 1)

    d <- Cache(rnorm, 1, cachePath = c(tmpdir, tmpCache))
    suppressMessages(dCache <- showCache(tmpCache))
    expect_true(length(unique(dCache[[.cacheTableHashColName()]])) == 1)

    f <- Cache(rnorm, 2, cachePath = c(tmpdir, tmpCache))
    suppressMessages(fCache <- showCache(tmpdir))
    expect_true(length(unique(fCache[[.cacheTableHashColName()]])) == 1)

    d <- Cache(rnorm, 1)
    suppressMessages(dCache <- showCache())
    expect_true(length(unique(dCache[[.cacheTableHashColName()]])) == 1)

    f <- Cache(rnorm, 2)
    suppressMessages({
      fCache <- showCache()
    })
    expect_true(length(unique(fCache[[.cacheTableHashColName()]])) == 1)

    clearCache(tmpdir)
    clearCache(tmpCache)
  }

})

utils::globalVariables("arg")

#' @param dryRun See [reproducibleOptions].
#'
#' @include messages.R
#' @export
#' @rdname Cache
Cache <- function(FUN, ..., dryRun = getOption("reproducible.dryRun", FALSE),
                  notOlderThan = NULL,
                  .objects = NULL, .cacheExtra = NULL, .functionName = NULL,
                  .cacheChaining = getOption("reproducible.cacheChaining", NULL),
                  outputObjects = NULL, # nolint
                  algo = "xxhash64",
                  cachePath = NULL,
                  length = getOption("reproducible.length", Inf),
                  userTags = c(),
                  omitArgs = NULL,
                  classOptions = list(),
                  debugCache = character(),
                  quick = getOption("reproducible.quick", FALSE),
                  verbose = getOption("reproducible.verbose", 1),
                  cacheId = NULL,
                  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                  useCache = getOption("reproducible.useCache", TRUE),
                  useCloud = getOption("reproducible.useCloud", FALSE),
                  cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
                  showSimilar = getOption("reproducible.showSimilar", FALSE),
                  drv = getOption("reproducible.drv", NULL),
                  conn = getOption("reproducible.conn", NULL)) {

  .callingEnv <- parent.frame()

  # Sets useDBI(TRUE) if a user has supplied a drv or conn
  optionsSetForCache(drv = drv, conn = conn)

  validateUseCloud(useCloud)

  ## Lazy showCache async pre-populate: idempotent, ~10us when already spawned.
  ## See R/showCacheEtc.R::.maybeSpawnShowCacheAsync. Targets the cachePath
  ## actually being used (not the default at .onLoad time).
  .maybeSpawnShowCacheAsync(cachePath)

  # Capture and match call so it can be manipulated
  callList <- matchCall2(sys.function(0), sys.call(0), envir = .callingEnv, FUN = FUN)

  # Check if this is a nested Cache call; this must be before skipCache because useCache may be numeric
  userTags <- setupCacheNesting(userTags, useCache) # get nested userTags

  # Skip Cache if user passes useCache = FALSE or 0 or nesting level is deeper than useCache
  useCache <- useCacheFromNested(useCache)
  if (isFALSE(useCache))
    return(skipCache(FUN, ..., usesDots = callList$usesDots, useCache = useCache,
                     functionName = format(callList$FUNorig), verbose = verbose, .callingEnv = .callingEnv))

  # Harmonize call so the different versions are all cannonical, now that useCache = FALSE is past
  callList <- harmonizeCall(callList, .callingEnv, .functionName)
  # Add .functionName to .pkgEnv userTags in case this becomes part of a nested Cache
  appendFunctionNameToNestedTags(userTags, callList$.functionName)
  # .pkgEnv$.reproEnv2$userTags

  # do the Digest
  times <- list()
  times$CacheDigestStart <- Sys.time()

  # Construct the full file path for the cache directory and possible file
  cachePaths <- getCacheRepos(cachePath, callList$new_call[-1], verbose = verbose)

  # Override keyFull$key if user has specified with cacheId
  if (!is.null(cacheId) && !is.na(cacheId)) {
    keyFull <- list()
    keyFull$key <- cacheIdOverride(cacheId, keyFull$key, callList$.functionName, verbose)
    if (is.null(keyFull$key))
      cacheId <- NULL
  }

  if (is.null(cacheId) || is.na(cacheId)) {
    cacheChainDetails <- cacheChainingSetup(.cacheChaining, callList, omitArgs, verbose)
    toDigest <- doDigestPrepare(callList$new_call, cacheChainDetails$omitArgs, .cacheExtra)
    keyFull <- try2(doDigest(toDigest, callList$.functionName, .objects,
                            length, algo, quick, classOptions, times$CacheDigestStart,
                            verbose = verbose))
    if (is(keyFull, "try-error")) {
      # This is the bit that indicates that one or more objects in the toDigest
      #   are corrupted and can't be digested. So, it is the inputs to the
      #   function that are corrupted: this can't self heal. Needs better user
      #   error message to give help.
      stopRcppError(toDigest, .objects, length, algo, quick, classOptions)
    }
    # update with cacheChain info
    keyFull <- cacheChainingStep(keyFull, callList, .cacheChaining, cacheChainDetails, cachePaths)

  }

  # If debugCache is "quick", short circuit after doDigest
  if (isTRUE(!is.na(pmatch(debugCache, "quick"))))
    return(list(hash = keyFull$preDigest, content = callList$func_call))

  CacheDBFileCheckAndCreate(cachePaths[[1]], drv, conn, verbose = verbose) # checks that we are using multiDBfile backend

  if (cloudWrite(useCloud)) {
    cloudFolderID <- checkAndMakeCloudFolderID(cloudFolderID, cachePaths[[1]], create = TRUE, verbose = verbose)
    gdriveLs <- retry(quote(driveLs(cloudFolderID, keyFull$key, cachePath = cachePaths[[1]], verbose = verbose)))
  }

  if (missing(dryRun)) dryRun <- getOption("reproducible.cacheDryRun", FALSE)

  if (cacheSaveFormat %in% c(.qsFormat))
    cacheSaveFormat <- getOption("reproducible.qsFormat", .qs2Format)

  # Memoise and return if it is there #
  if (!dryRun) {
    outputFromMemoise <- check_and_get_memoised_copy(keyFull, cachePaths, callList$.functionName,
                                                     callList$func, useCache, useCloud,
                                                     cloudFolderID, gdriveLs, full_call = callList$new_call,
                                                     outputObjects = outputObjects,
                                                     cacheSaveFormat = cacheSaveFormat,
                                                     .cacheChaining = .cacheChaining,
                                                     drv = drv, conn = conn, verbose = verbose)
    if (!identical2(.returnNothing, outputFromMemoise))
      return(outputFromMemoise)

    # After memoising fail, try files; need to check Cache dir and set lockfile
    locked <- lockFile(cachePaths[[1]], keyFull$key, verbose = verbose)

    if (useDBI()) {
      connOrig <- conn
      conn <- checkConns(cachePaths, conn)
      drv <- getDrv(getOption("reproducible.drv", NULL))
      for (cachePath in cachePaths)
        conn <- createConns(cachePath, conn, drv, verbose = verbose) # this will convert backend if it is wrong

      if (is.null(connOrig)) # don't disconnect if conn was user passed
        # if this is >1st cachePath, then the db will already be disconnected; suppressWarnings
        on.exit(dbDisconnectAll(conn), add = TRUE)
    }

    # Check if keyFull$key is on disk and return if it is there
    outputFromDisk <- check_and_get_cached_copy(keyFull, cachePaths, cache_file, callList$.functionName, callList$func,
                                                useCache, useCloud, cloudFolderID, gdriveLs,
                                                full_call = callList$new_call,
                                                outputObjects = outputObjects,
                                                cacheSaveFormat = cacheSaveFormat,
                                                .cacheChaining = .cacheChaining,
                                                drv, conn, verbose = verbose)

    if (!identical2(.returnNothing, outputFromDisk))
      return(outputFromDisk)

  }
  if (useDBI()) conn <- attr(outputFromDisk, ".Cache")$conn

  cache_file <- CacheStoredFile(cachePaths[[1]], keyFull$key) # now we know it is not in Cache; use 1st cachePath
  if (cloudReadOnly(useCloud)) {# now that it is established it isn't in cache locally
    cloudFolderID <- checkAndMakeCloudFolderID(cloudFolderID, cachePaths[[1]], create = TRUE, verbose = verbose)
    gdriveLs <- retry(quote(driveLs(cloudFolderID, keyFull$key, cachePath = cachePaths[[1]], verbose = verbose)))
  }

  if (cloudWriteOrRead(useCloud) && isTRUE(any(keyInGdriveLs(keyFull$key, gdriveLs)))) {
    newFileName <- gdriveLs$name[which(keyInGdriveLs(keyFull$key, gdriveLs))] # paste0(outputHash,".rda")
    shownCache <- cloudDownload(keyFull$key, newFileName, gdriveLs, cachePaths[[1]], cloudFolderID,
                                drv = drv, conn = conn, verbose = verbose)
    outputFromDisk <- check_and_get_cached_copy(keyFull, cachePaths, cache_file, callList$.functionName, callList$func,
                                                useCache, useCloud = FALSE, cloudFolderID, gdriveLs,
                                                full_call = callList$new_call,
                                                outputObjects = outputObjects,
                                                cacheSaveFormat = cacheSaveFormat,
                                                .cacheChaining = .cacheChaining,
                                                drv, conn, verbose = verbose)
    return(outputFromDisk)
  } # Derive some metadata prior to evaluation so "showSimilar" can have something to compare with

  times$EvaluateStart <- Sys.time()
  metadata <- metadata_define_preEval(keyFull, callList$.functionName, userTags,
                                      .objects, length, algo, quick, classOptions,
                                      times$EvaluateStart, times$CacheDigestStart)

  if (isTRUE(showSimilar) || isDevMode(useCache, userTags) || isTRUE(dryRun)) {
    if (dryRun) messageColoured(.txtDryRunTRUE, colour = "green")
    showSimilar(cachePaths[[1]], metadata, callList$.functionName, userTags, useCache,
                # cacheSaveFormat = cacheSaveFormat,
                drv = drv, conn = conn, verbose)
  }
  if (isTRUE(dryRun))
    return(invisible(NULL))

  # ## evaluate the call ## #
  outputFromEvaluate <- evalTheFunAndAddChanged(callList = callList, keyFull = keyFull,
                                                outputObjects = outputObjects, length = length,
                                                algo = algo, quick = quick, classOptions = classOptions,
                                                .callingEnv = .callingEnv,
                                                verbose = verbose, ...)

  # ## Save to Cache; including to Memoise location; including metadata ## #
  times$SaveStart <- Sys.time()
  elapsedTimeFUN <- difftime(times$SaveStart, times$EvaluateStart, units = "secs")

  # update metadata with other elements including elapsedTime for evaluation
  metadata <- metadata_define_postEval(metadata, keyFull$key, outputFromEvaluate,
                                       userTags, .objects, length, algo, quick,
                                       classOptions, elapsedTimeFUN)

  outputFromEvaluate <- doSaveToCache(outputFromEvaluate, metadata, cachePaths, callList = callList, # callList$func,
                                      .objects, length, algo, quick, classOptions,
                                      cache_file, userTags, # callList$.functionName,
                                      debugCache,
                                      keyFull, outputObjects = outputObjects,
                                      useCloud, cloudFolderID, gdriveLs,
                                      # func_call = callList$func_call,
                                      cacheSaveFormat = cacheSaveFormat, drv = drv, conn = conn,
                                      useMemoise = getOption("reproducible.useMemoise", FALSE),
                                      .cacheChaining = .cacheChaining,
                                      verbose = verbose,
                                      times$SaveStart, times$EvaluateStart)
  times$SaveEnd <- Sys.time()
  if (getOption("reproducible.savePreDigest", FALSE)) {
    keyFullPreDigest <- keyFull
    keyFullPreDigest$key <- paste0(.txtPreDigest, "_", keyFullPreDigest$key)
    times$SavePreDigestStart <- Sys.time()
    locked <- lockFile(cachePaths[[1]], keyFullPreDigest$key, verbose = verbose)

    toDigestOut <- doSaveToCache(toDigest, metadata, cachePaths, callList = callList, # callList$func,
                                 .objects, length, algo, quick, classOptions,
                                 cache_file, userTags, # callList$.functionName,
                                 debugCache,
                                 keyFullPreDigest, outputObjects = outputObjects,
                                 # func_call = callList$func_call,
                                 cacheSaveFormat = cacheSaveFormat,
                                 drv = drv, conn = conn,
                                 useCloud = FALSE, # not this preDigest one
                                 cloudFolderID = NULL, gdriveLs = NULL,# not this preDigest one
                                 useMemoise = FALSE, # not this preDigest one
                                 .cacheChaining = .cacheChaining,
                                 verbose = verbose,
                                 times$SavePreDigestStart, times$SaveStart)
    times$SaveEnd <- Sys.time()
  }
  verboseCacheDFAll(verbose, callList$.functionName, times)

  return(outputFromEvaluate)
}

#' @rdname Cache
cache2 <- Cache

#' Convert all ways of calling a function into canonical form, including defaults
#'
#' e.g., stats::rnorm(1) --> rnorm(n = 1, mean = 0, sd = 1)
#' @param call The full captured call as it was passed by user.
#' @param usesDots Logical. Whether the original `Cache` call used `...`
#' @param isSquiggly Logical. Whether there are curly braces e.g., as in a pipe sequence.
#' @param .callingEnv Environment. The environment from which `Cache` was called.
convertCallToCommonFormat <- function(call, usesDots, isSquiggly, .callingEnv) {

  if (requireNamespace("covr", quietly = TRUE) && covr::in_covr()) {
    strip_covr_wrappers <- function(expr) {
      while (is.call(expr) && identical(expr[[1]], as.name("{"))) expr <- expr[[length(expr)]]
      expr
    }
    call <- strip_covr_wrappers(call)
    if (length(call) >= 2L && is.language(call[[2]])) call[[2]] <- strip_covr_wrappers(call[[2]])
  }

  .functionName <- NULL
  # Check if the first argument is a function call
  func_full <- NULL

  func_call <- NULL
  if (is.call(call[[2]])) {

    func_call <- call[[2]]  # This is the actual function call (e.g., stats::rnorm)
    # Extract the function without the package prefix
    if (is.call(func_call[[1]]) && func_call[[1]][[1]] == quote(`::`)) {
      func <- func_full <- func_call[[1]]
      if (length(func_call) == 2)
        args <- func_call[[-1]]
      else
        args <- as.list(func_call)[-1]
      func_call <- as.call(c(func_call[[1]][[3]], args))
    } else {
      if (func_call[[1]] == quote(`::`) || func_call[[1]] == quote(`:::`)) {
        func_full <- func_call
        func <- func_call  # Package prefix, using FUN as name only
        args <- as.list(call[-(1:2)])
        func_call <- as.call(c(func_call[[3]], args))
      } else {
        if (isDollarOnlySqBr(func_call)) {
          func <- eval(func_call, envir = .callingEnv)  # No package prefix
          if (usesDots)
            func_call <- as.call(append(list(func), as.list(call[-(1:2)])))
        } else {
          # It is a complete call e.g., FUN = rnorm(1)
          func <- func_call[[1]]  # No package prefix
          if (isDollarOnlySqBr(func)) {
            func <- eval(func, envir = .callingEnv)  # No package prefix
          }
        }
        if (identical(func, quote(do.call))) {
          func_call <- undoDoCall(func_call, .callingEnv = .callingEnv)
          func <- func_call[[1]]  # Extract the function for do.call (e.g., rnorm)
        }
        args <- as.list(func_call)[-1]
      }
    }
  } else if (identical(call[[2]], quote(do.call))) {
    # Special handling for do.call to return the function unevaluated
    func <- call[[3]]  # Extract the function for do.call (e.g., rnorm)
    args <- eval(call[[4]], envir = .callingEnv)  # Evaluate the argument list
  } else {
    func <- call[[2]]  # This is the function (e.g., rnorm)
    args <- as.list(call[-(1:2)])  # These are the arguments (e.g., 1)
    func_call <- as.call(append(list(func), args))
    # Check for package prefix
    if (is.call(func) && func[[1]] == quote(`::`)) {
      func <- func[[3]]  # Get the actual function name (e.g., rnorm)
    }
  }

  if (is.call(func) || is.name(func)) {
    if (is.name(func))
      .functionName <- format(func)
    fun <- if (is.null(func_full)) func else func_full
    if (is.name(fun)) {
      infixes <- c("+", "-", "*", "/", "==", "!=", "<", ">", "<=", ">=", "&&", "||")
      areInfixes <- any(fun == infixes)
      if (!any(areInfixes)) {
        fun_chr <- as.character(fun)
        # Only parse if it's not a reserved word
        if (!(fun_chr %in% c("if", "function"))) {
          fun <- parse(text = fun_chr)
        }
        # else: leave fun as-is (symbol), so downstream logic can handle it
        # fun <- parse(text = fun)
      }
    }
    func <- eval(fun, envir = .callingEnv)
  }

  # deal with defunct arguments
  if (usesDots) { # any defunct argument will show up in the usesDots; need to keep them for defunct fn
    argsSupplied <- unique(names(call))[-1]
  } else {
    argsSupplied <- names(args)
  }
  defunct(setdiff(argsSupplied, formalArgs(func))) # pull the plug if args are defunct, and not used in FUN
  argsRm <- names(args) %in% setdiff(names(.formalsCache), names(formals(func)))
  if (any(argsRm %in% TRUE))
    args <- args[!argsRm %in% TRUE]

  # build new call from func and args; both must be correct by here
  new_call <- as.call(append(list(func), args))
  # This matches call on the FUN, not a duplicate of matchCall2
  matched_call <- match_call_primitive(func, new_call, expand.dots = TRUE, envir = .callingEnv)

  if (isSquiggly) {
    FUNcaptured <- recursiveEvalNamesOnly(matched_call, envir = .callingEnv) # deals with e.g., stats::rnorm, b$fun, b[[fun]]
    args <- as.list(FUNcaptured[-1])
  } else {
    args <- as.list(matched_call)[-1]
    args <- evaluate_args(args, envir = .callingEnv)
  }
  combined_args <- combine_clean_args(func, args, .objects = NULL, .callingEnv)

  # Check for arguments that are in both Cache and the FUN
  matched_call <- checkOverlappingArgs(call, combined_args, dotsCaptured = args,
                                       functionName = "outer", matched_call, whichCache = "cache2")

  if (is.null(func_call)) func_call <- new_call
  func_call2 <- as.call(c(func_call[[1]], args))
  attr(matched_call, ".Cache")$func_call <- func_call2
  attr(matched_call, ".Cache")$args_w_defaults <- combined_args
  attr(matched_call, ".Cache")$method <- func
  attr(matched_call, ".Cache")$.functionName <- .functionName

  return(matched_call)
}

evaluate_args <- function(args, envir) {
  lapply(args, function(arg) {
    if (is.call(arg)) {
      arg <- tryCatch(eval(arg, envir = envir), error = function(err) { # can't be tryCatch2 --> this must always be a tryCatch
        # If it's a call that cannot be evaluated, evaluate recursively
        fail <- "fail"
        newPossArgMinus1 <- tryCatch(evaluate_args(as.list(arg[-1]), envir), error = function(err) {
          fail
        })
        if (!identical(newPossArgMinus1, fail)) {
          arg <- as.call(c(arg, as.list(newPossArgMinus1[-1])))
        }
        arg
      })
    } else if (is.symbol(arg)) {
      # If it's a symbol, evaluate it in the specified environment
      arg <- eval(arg, envir)
    }
    return(arg)
  })
}

check_and_get_cached_copy <- function(detailed_key, cachePaths, cache_file, functionName,
                                      func, useCache, useCloud, cloudFolderID, gdriveLs,
                                      full_call, outputObjects,
                                      cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                                      .cacheChaining = getOption("reproducible.cacheChaining", FALSE),
                                      drv, conn, envir = parent.frame(), verbose) {
  cache_key <- detailed_key$key
  # Check if the result is already cached
  connOrig <- conn
  conns <- conn
  if (!is.null(conn) && !is.list(conns)) {
    conns <- list(conn)
    names(conns) <- cachePaths
  }

  for (cachePath in cachePaths) {
    cache_file <- CacheStoredFile(cachePath, cache_key, cacheSaveFormat = cacheSaveFormat, readOnly = TRUE)
    cacheFileExists <- file.exists(cache_file) # could be length >1
    if (useDBI()) {
      inReposPoss <- searchInRepos(cachePath,
                                   outputHash = cache_key,
                                   drv = drv, conn = conns[[cachePath]]
      )
      if (cachePath == cachePaths[[1]] || NROW(inReposPoss$isInRepo)) {
        # keep important parts if it is first one, or if it has the object in the cacheRepo
        # inRepos <- inReposPoss
        conn <- conns[[cachePath]] # keep it as a list so places where it needs the name work
        if (is.null(connOrig)) # don't disconnect if conn was user passed
          # if this is >1st cachePath, then the db will already be disconnected; suppressWarnings
          on.exit2(suppressWarnings(DBI::dbDisconnect(conn)), envir = envir)

        shownCache <- inReposPoss$fullCacheTableForObj
        if (NROW(inReposPoss$isInRepo)) {
          break
        }
        if (cachePath == tail(cachePaths, 1)) { # if it is the last or only cachePath, then end
          ret <- .returnNothing
          attr(ret, ".Cache")$conn <- conn
          return(invisible(ret))
        }

        # this disconnect won't happen if user passed just one conn because already returned/break from this loop
        DBI::dbDisconnect(conn) # try next cachePath -- disconnect previous;
      }
    } else {
      shownCache <- NULL
      if (isTRUE(cacheFileExists))
        break
    }
  }

  # Check if it was saved with other CacheSaveFormat
  changedSaveFormat <- FALSE
  if (sum(cacheFileExists) == 0) { # if it doesn't exist; could be changed backend or not in Cache
    sameCacheID <- checkSameCacheId(cache_file)
    if (length(sameCacheID) > 0) {
      changedSaveFormat <- TRUE
      cacheFileExists <- TRUE
      cache_file_orig <- cache_file
      cache_file <- file.path(dirname(cache_file), sameCacheID)
    }
  }

  if (sum(cacheFileExists)) {
    output <- loadFromDiskOrMemoise(fromMemoise = FALSE, useCache, useCloud,
                                    cloudFolderID = cloudFolderID, gdriveLs = gdriveLs,
                                    cachePath = cachePath,
                                    detailed_key, functionName, cache_file = cache_file,
                                    changedSaveFormat = changedSaveFormat, sameCacheID,
                                    cache_file_orig, func, shownCache = shownCache,
                                    full_call = full_call,
                                    outputObjects = outputObjects,
                                    cacheSaveFormat = cacheSaveFormat,
                                    .cacheChaining = .cacheChaining,
                                    drv = drv, conn = conn, verbose = verbose)
    return(output)

  }
  invisible(.returnNothing)
}

.returnNothing <- ".nothing"

combine_clean_args <- function(FUN, args, .objects, .callingEnv) {
  # has to be after match.call --> relies on name matched arguments
  defaults <- get_function_defaults(eval(FUN, .callingEnv))
  combined_args <- reorder_arguments(defaults, args)
  empties <- vapply(combined_args, function(ca) if (is.symbol(ca)) capture.output(ca) else "Normal", character(1))
  empties <- !nzchar(empties)
  if (isTRUE(any(empties)))
    combined_args <- combined_args[!empties]

  # Process the .objects argument using the helper function
  if (!is.null(.objects)) {
    combined_args <- filter_objects(combined_args, .objects)
  }

  combined_args
}

metadata_update <- function(outputToSave, metadata, cache_key) {
  userTagsExtra <- attr(outputToSave, "tags") # .wrap may have added tags
  userTagsExtra <- grep("cacheId:", userTagsExtra, invert = TRUE, value = TRUE) # don't add cacheId to tagKey
  if (!is.null(userTagsExtra) && length(userTagsExtra) > 0) {
    ut <- strsplitOnlySingleColon(userTagsExtra)
    # ut <- strsplit(userTagsExtra, split = ":")
    ll <- lapply(ut, tail, 1)
    names(ll) <- lapply(ut, head, 1)
    userTagsList <- ll
    metadata <- rbindlist(list(metadata, userTagsListToDT(cache_key, userTagsList)))
  }
  metadata
}

check_and_get_memoised_copy <- function(detailed_key, cachePaths, functionName, func,
                                        useCache, useCloud, cloudFolderID, gdriveLs,
                                        full_call, outputObjects,
                                        cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                                        .cacheChaining = getOption("reproducible.cacheChaining", FALSE),
                                        drv, conn, verbose) {
  cache_key <- detailed_key$key
  if (getOption("reproducible.useMemoise", FALSE)) {
    for (cachePath in cachePaths) {
      cache_key_in_memoiseEnv <- exists(cache_key, envir = memoiseEnv(cachePath), inherits = FALSE)
      if (isTRUE(cache_key_in_memoiseEnv))
        break
    }

    if (cache_key_in_memoiseEnv) {
      output <- loadFromDiskOrMemoise(fromMemoise = TRUE, useCache = useCache, useCloud = useCloud,
                                      cloudFolderID = cloudFolderID, gdriveLs = gdriveLs,
                                      cachePath = cachePath, detailed_key = detailed_key,
                                      functionName = functionName, func = func,
                                      full_call = full_call,
                                      changedSaveFormat = FALSE,
                                      outputObjects = outputObjects,
                                      cacheSaveFormat = cacheSaveFormat,
                                      .cacheChaining = .cacheChaining,
                                      drv = drv, conn = conn, verbose = verbose,
                                      )
      return(output)
    }
  } else {
    # If useMemoise gets turned off, it needs to be emptied or there will be stale entries.
    # Preserve the "shownCache" binding -- it holds the showCache memoised
    # data.table and the async-spawn job table; both are independent of
    # useMemoise and clearing them would defeat the lazy async pre-populate
    # mechanism (jobs would be re-spawned on every Cache() call).
    me <- memoiseEnv(cachePaths[[1]])
    le <- setdiff(ls(me), "shownCache")
    if (length(le))
      rm(list = le, envir = me)
  }
  return(invisible(.returnNothing))
}

# Helper function to filter arguments based on .objects
filter_objects <- function(evaluated_args, .objects) {
  list_or_env_arg <- NULL
  for (name in names(evaluated_args)) {
    if (is.list(evaluated_args[[name]]) || is.environment(evaluated_args[[name]])) {
      list_or_env_arg <- name
      break
    }
  }

  if (!is.null(list_or_env_arg)) {
    actual_list <- evaluated_args[[list_or_env_arg]]
    filtered_elements <- actual_list[.objects]
    filtered_list <- actual_list
    filtered_list[names(filtered_list) %in% .objects] <- filtered_elements
    filtered_list <- filtered_list[.objects]
    evaluated_args[[list_or_env_arg]] <- filtered_list
  }

  return(evaluated_args)
}

# Function to normalize the call to handle `do.call`
undoDoCall <- function(call, .callingEnv) {
  if (is.call(call) && all(as.character(call[[1]]) == "do.call")) {
    func <- call[[2]]
    args <- call[[3]]

    if (isTRUE(is.call(args)) && isTRUE(as.character(args[[1]]) == "list")) {
      args <- as.list(args[-1])
    }
    if (is.name(args))
      args <- recursiveEvalNamesOnly(args, envir = .callingEnv)

  } else {
    func <- call[[1]]
    args <- as.list(call[-1])
  }
  return(as.call2(func, args))
}

# Helper function to get function defaults
get_function_defaults <- function(func) {
  if (is.primitive(func)) {
    formals_list <- formals(args(list))
  } else {
    formals_list <- formals(func)
  }
  return(as.list(formals_list))
}

# Helper function to reorder arguments based on formal arguments, combining defaults and user args
reorder_arguments <- function(formals, args) {
  # Combine defaults and args: user args override defaults

  areDots <- names(args) %in% "..."
  if (any(areDots)) {
    args2 <- args
    args2[[which(areDots)]] <- NULL
    args <- append(args2, args[[which(areDots)]])
  }

  if (FALSE) {
    # areDots <- names(args) %in% "..."
    namesOfArgs <- names(args) %in% "..."
    areDots <- any(names(formals) %in% "...") || any(namesOfArgs)
    if (any(areDots)) {
      if (length(namesOfArgs)) {
        args2 <- args
        for (wh in which(namesOfArgs)) {
          args2[[wh]] <- NULL
          args <- append(args2, args[[which(namesOfArgs)]])
        }

      } else {
        # these are unnamed args in the dots
      }
    }


  }

  if (length(formals) == 1 && all(names(formals) %in% "...")) {
    # This is case of things like `list`, `file.path`
    ordered_args <- args
  } else {
    # This will remove unnamed elements; which isn't right
    combined_args <- modifyList(formals, args, keep.null = TRUE)
    emptyNams <- names(args) %in% ""
    if (any(emptyNams)) {
      combined_args <- append(combined_args, args[emptyNams])
    }
    areDots <- names(combined_args) %in% "..."
    if (any(areDots)) {

      # argPlaceInsert <- which(!names(args) %in% names(formals))

      # needArgs <- !names(args) %in% names(combined_args)
      combined_args[areDots] <- NULL
      ordered_args <- combined_args
      # combined_args <- append(combined_args, args[needArgs])
      # areDots2 <- names(formals) %in% "..."
      # whNotDots <- which(!areDots2)
      # whDots <- which(areDots2)
      # first <- if (whDots > 1) seq(whDots - 1) else numeric()
      # anySeconds <- !whDots > whNotDots
      # second <- if (any(anySeconds)) whNotDots[anySeconds] else numeric()
      # ordered_args <- c(args[argPlaceInsert], formals[second])
    } else {
      ordered_args <- combined_args[union(names(formals), names(combined_args))]
    }
  }
  # Preserve the order of the formals

  return(ordered_args)
}

match_call_primitive <- function(definition = sys.function(sys.parent()),
                                 call = sys.call(sys.parent()),
                                 expand.dots = TRUE,
                                 envir = parent.frame()) {
  # Check if the function is a primitive infix operator
  if (is.primitive(definition)) {
    # For infix operators like +, -, *, etc., they are not called in the standard way
    infixes <- c(`+`, `-`, `*`, `/`, `==`, `!=`, `<`, `>`, `<=`, `>=`, `&&`, `||`)
    areInfixes <- vapply(infixes, function(i) identical(i, definition), FUN.VALUE = logical(1))

    if (isTRUE(any(areInfixes))) {
      # Handle infix operators by keeping the call intact
      return(call)
    }

    # For other primitives, match as best as possible
    args <- as.list(call)[-1]  # remove the function name
    if (expand.dots) {
      args <- lapply(args, eval, envir = envir)
    }
    # Construct the matched call manually for primitive
    matched <- as.call(c(definition, args))
    return(matched)
  } else {
    # Non-primitive function: fall back to regular match.call
    return(base::match.call(definition = definition,
                            call = call,
                            expand.dots = expand.dots,
                            envir = envir))
  }
}


cache_Id_Identical <- function(metadata, cachePaths, cache_key,
                               cacheSaveFormat = getOption("reproducible.cacheSaveFormat")) {
  linkToCacheId <- NULL
  os <- metadata$tagValue[metadata$tagKey == "object.size"]

  skipPreDigest <- startsWith(cache_key, .txtPreDigest)

  if (!identical(os, "NA") && skipPreDigest %in% FALSE) {
    if (isTRUE(as.numeric(os) > .objectSizeMinForBig)) {
      for (cachePath in cachePaths) {
        allCache <- showCache(x = cachePath, verbose = -2)
        if (NROW(allCache)) {
          resultHash <- metadata$tagValue[metadata$tagKey == "resultHash"]
          alreadyExists <- allCache[allCache$tagKey == "resultHash" &
                                      allCache[[.cacheTableTagColName()]] %in% resultHash &
                                      allCache[[.cacheTableHashColName()]] != cache_key]
          if (NROW(alreadyExists)) {
            linkToCacheId <- alreadyExists[["cacheId"]][[1]]
          }
        }
      }
    }
  }
  if (!is.null(linkToCacheId))
    linkToCacheId <- CacheStoredFile(cachePath, linkToCacheId, cacheSaveFormat = cacheSaveFormat)
  linkToCacheId
}

metadata_define_preEval <- function(detailed_key, func_name, userTags,
                                    .objects, length, algo, quick, classOptions,
                                    timeEvaluateStart, timeCacheDigestStart) {

  elapsedTimeCacheDigest <- difftime(timeEvaluateStart, timeCacheDigestStart, units = "secs")

  useCloud <- FALSE

  df <- unlist(
    .unlistToCharacter(unname(detailed_key[-1]), getOption("reproducible.showSimilarDepth", 3))
  )
  pat <- "[[:digit:]]{1,5}$"
  didWeGainNumerics <- grep(names(df), pattern = pat)
  wouldBe <- gsub("", pattern = pat, names(df))
  dups <- which(duplicated(wouldBe))
  wasFirstsOfDups <- setdiff(didWeGainNumerics, dups)
  isTheDupAGainedNumeric <- wasFirstsOfDups %in% didWeGainNumerics
  if (any(isTheDupAGainedNumeric)) {
    changeThese <- c(wasFirstsOfDups, dups)
    names(df)[changeThese] <- wouldBe[changeThese]
  }

  tagKey <- paste0(names(df), ":", as.character(df))
  if (length(userTags)) {
    ut <- strsplitOnlySingleColon(userTags)
    # ut <- strsplit(userTags, split = ":")
    ll <- lapply(ut, tail, 1)
    strt <- lapply(ut, function(x) x[-length(x)])
    utagLabel <- rep("userTags", length(ll))#lapply(ut, head, 1)
    hasLabel <- lengths(strt) > 0
    if (any(hasLabel)) {
      utagLabel[hasLabel] <- sapply(strt[hasLabel], function(x) paste0(x, collapse = ":"))
    }
    names(ll) <- utagLabel
    userTags <- ll
  }
  userTagsList <- c(
    list(func_name) |> setNames(nm = .cacheTagsFirstGroup[1]),
    userTags,
    list(sysTimeForCacheToChar()) |> setNames(nm = .cacheTagsFirstGroup[3]),
    list(isTRUE(useCloud)) |> setNames(nm = .cacheTagsFirstGroup[4]),
    list(format(elapsedTimeCacheDigest, units = "secs")) |> setNames(nm = .cacheTagsFirstGroup[5]),
    list(tagKey) |> setNames(nm = .cacheTagsFirstGroup[6])
  )
  names(userTagsList)[1] <- "function"
  cache_key <- detailed_key$key
  metadata <- userTagsListToDT(cache_key, userTagsList)
  return(metadata)
}

metadata_define_postEval <- function(metadata, cacheId, outputToSave, userTags,
                                     .objects, length, algo, quick, classOptions,
                                     elapsedTimeFUN) {
  objSize <- NA
  if (getOption("reproducible.objSize", TRUE)) {
    hasPointer <- usesPointer(outputToSave)
    if (any(unlist(hasPointer))) {
      os <- objSize(outputToSave, recursive = TRUE)
    } else {
      os <- objSize(outputToSave)
    }
    objSize <- sum(os)
  }

  resultHash <- ""
  if (isTRUE(objSize > .objectSizeMinForBig)) {
    resultHash <- CacheDigest(outputToSave,
                              .objects = .objects,
                              length = length, algo = algo, quick = quick,
                              classOptions = classOptions, calledFrom = "Cache"
    )$outputHash
  }
  fns <- Filenames(outputToSave)
  # tagsFromDefaults <- .cacheTagsDefault
  # .cacheTagsSecondGroup <- c("class", "object.size", "fromDisk", "resultHash", "elapsedTimeFirstRun")

  userTagsList <- c(
    list(class(outputToSave)[1]) |> setNames(nm = .cacheTagsSecondGroup[1]),
    list(format(as.numeric(objSize))) |> setNames(nm = .cacheTagsSecondGroup[2]),
    list(isTRUE(any(nchar(fns) > 0))) |> setNames(nm = .cacheTagsSecondGroup[3]),
    list(resultHash) |> setNames(nm = .cacheTagsSecondGroup[4]),
    list(format(elapsedTimeFUN, units = "secs")) |> setNames(nm = .cacheTagsSecondGroup[5])
  )
  cache_key <- cacheId
  metadataNew <- userTagsListToDT(cache_key, userTagsList)
  metadata <- rbindlist(list(metadata, metadataNew))
  # attr(metadata, "tags")$objectSize <- objSize
  metadata
}

userTagsListToDT <- function(cache_key, userTagsList) {
  theChars <- vapply(userTagsList, function(x) is.character(x) | is.logical(x), logical(1))
  if (any(!theChars)) {
    for (tc in which(!theChars))
      userTagsList[[tc]] <- tryCatch2(format(userTagsList[[tc]]), error = function(u) as.character())
  }
  userTagsList <- utils::stack(userTagsList)
  metadataDT(cacheId = cache_key, tagKey = userTagsList$ind, tagValue = userTagsList$values)
}

as.call2 <- function(func, args) {
  as.call(c(as.name(deparse(func)), args))
}

skipCache <- function(FUN, ..., usesDots, functionName, useCache, verbose, .callingEnv) {
  .message$useCacheIsFALSE(.pkgEnv$.reproEnv2$nestLevel - 1, # original Cache counted differently; use -1 here
                           functionName = functionName, useCache = useCache, verbose = verbose)
  if (isTRUE(usesDots)) {
    FUN(...)
  } else {
    eval(FUN, envir = .callingEnv)
  }
}

clearCacheOverwrite <- function(cachePath, cache_key, functionName, drv, conn, verbose) {
  clearCache(x = cachePath, cacheId = cache_key, ask = FALSE, conn = conn, drv = drv, verbose = verbose - 1)
  .message$overwriting(functionName, type = "function", verbose)
}

sysTimeForCacheToChar <- function(digits = 5)
  format(Sys.time(), digits = digits)


setupCacheNesting <- function(userTags, useCache, envir = parent.frame(1)) {
  if (!exists(".reproEnv2", envir = .pkgEnv)) {
    .pkgEnv$.reproEnv2 <- new.env(parent = asNamespace("reproducible"))
    .pkgEnv$.reproEnv2$userTags <- userTags
    .pkgEnv$.reproEnv2$nestLevel <- 1
    .pkgEnv$.reproEnv2$useCache <- useCache
    on.exit2(rm(list = ".reproEnv2", envir = .pkgEnv), envir = envir)
  } else {
    userTagsOld <- .pkgEnv$.reproEnv2$userTags
    allUT1 <- c(userTagsOld, userTags)

    hasColon <- grepl(.txtGrepStrSplitSingleColon, allUT1, perl = TRUE)
    if (isTRUE(any(!hasColon)))
      allUT1[!hasColon] <- paste0("userTags:", allUT1[!hasColon])

    if (!is.null(allUT1)) {
      allUT2 <- allUT1[!duplicated(sapply(strsplitOnlySingleColon(allUT1), tail, 1))]

      splitted <- strsplitOnlySingleColon(allUT2)
      # firstPart <- sapply(strsplitOnlySingleColon(allUT2), function(x) x[[2]])
      # allUT2 <- allUT2[order(firstPart)]

      allUT2 <- sapply(
        reorder_by_first_element(splitted), function(x) paste0(x[[1]], ":", x[[2]])
      )

      userTags <- allUT2
      .pkgEnv$.reproEnv2$userTags <- userTags
      nestLevelOld <- .pkgEnv$.reproEnv2$nestLevel
      .pkgEnv$.reproEnv2$nestLevel <- nestLevelOld + 1
      on.exit2({
        .pkgEnv$.reproEnv2$nestLevel <- nestLevelOld
        .pkgEnv$.reproEnv2$userTags <- userTagsOld
      }, envir = envir)
    }
  }
  userTags
}

.addTagsRepoAccessedTime <- function(cache_key, cachePath = cachePath,
                                     cacheSaveFormat = getOption("reproducible.cacheSaveFormat")) {
  .addTagsRepo(cacheId = cache_key, tagKey = "accessed", tagValue = sysTimeForCacheToChar()
               , cacheSaveFormat = cacheSaveFormat, cachePath = cachePath)
}

callIsQuote <- function(call) {
  if (length(call$FUN) > 1) # just a function
    if (identical(call$FUN[[1]], quote(quote))) {
      call$FUN <- as.list(call$FUN)[[-1]] # unquote it
    }
  call
}

releaseLockFile <- function(locked) {
  filelock::unlock(locked)
  ## Do NOT delete the lock file: the fcntl lock is what protects the critical
  ## section, not the file's existence.  Deleting and recreating the file under
  ## concurrent load creates two bugs:
  ##   1. Workers that were blocked on fcntl(F_SETLKW) already have the old inode
  ##      open; a fresh caller that arrives after the delete creates a *new* inode
  ##      at the same path — both callers hold a "lock" on different inodes and
  ##      the critical section is no longer protected.
  ##   2. If a prior run was executed as root (or another user), a stale .lock file
  ##      with wrong ownership is left behind; the next caller gets EACCES at
  ##      open(O_RDWR|O_CREAT) — "Permission denied".
  ## Leaving the (empty) .lock file in place is safe and correct.
}

#' @importFrom stats runif
lockFile <- function(cachePath, cache_key,
                     envir   = parent.frame(),
                     verbose = getOption("reproducible.verbose")) {
  if (!useDBI()) {
    csd <- CacheStorageDir(cachePath)
    checkPath(csd, create = TRUE)

    lock_path <- file.path(csd, paste0(cache_key, suffixLockFile()))

    ## Three outcomes from filelock::lock:
    ##   NULL   — contention; sleep 2.5 s and retry
    ##   EMFILE — process near fd limit from other sources; gc + small sleep
    ##   EACCES — stale file owned by another user; remove and retry
    ##   other  — unexpected; re-throw immediately
    ##
    ## Note: PredictiveEcology/filelock >= 1.0.3.9001 fixes a bug in the
    ## upstream package where every failed non-blocking attempt leaked one fd
    ## (close()/CloseHandle() missing on the NULL return path in C).

    locked          <- NULL
    waiting         <- FALSE
    emfile_attempts <- 0L

    repeat {
      locked <- tryCatch(
        filelock::lock(lock_path, timeout = 0L),
        error = function(e) {
          msg <- conditionMessage(e)
          if (!grepl("Cannot open lock file", msg, fixed = TRUE)) stop(e)

          if (grepl("Too many open files", msg, fixed = TRUE)) {
            emfile_attempts <<- emfile_attempts + 1L
            if (emfile_attempts > 10L)
              stop("Persistent 'Too many open files' acquiring lock: ", lock_path,
                   "\nRaise ulimit -n or report a filelock fd-leak bug",
                   call. = FALSE)
            gc(FALSE)
            Sys.sleep(runif(1L, 0.1, 0.3) * emfile_attempts)
            return(NULL)
          }

          ## EACCES or similar — remove stale file and retry
          removed <- suppressWarnings(file.remove(lock_path))
          if (!isTRUE(removed))
            stop("Cannot open lock file and cannot remove it.\n",
                 "Manually delete (may need sudo): ", lock_path, "\n",
                 "Original error: ", msg, call. = FALSE)
          messageCache("Lock file not accessible; removed and retrying",
                       verbose = verbose + 1)
          dir.create(csd, showWarnings = FALSE, recursive = TRUE)
          return(NULL)
        }
      )

      if (!is.null(locked)) break

      if (!waiting) {
        waiting <- TRUE
        messageCache(
          "The cache file (", lock_path, ") is locked due to a concurrent process; waiting...",
          "\nTo diagnose the holding process (works on Linux/macOS):",
          "\n  system(\"fuser '", lock_path, "'\")",
          "\n  system(\"lsof '", lock_path, "'\")",
          "\nOn a network filesystem (NFS/CIFS), unlink() will NOT remove the file while",
          "\na process holds it open -- kill the holding process first, then the lock releases.",
          "\nIf no process is found (stale lock on a local filesystem), then delete the lockfile:",
          "\n  unlink('", lock_path, "', force = TRUE)",
          verbose = verbose + 2
        )
      }

      Sys.sleep(2.5)
    }

    if (waiting)
      messageCache("  ... ", lock_path, " released, continuing ... ", verbose = verbose + 2)

    # on.exit(filelock::unlock(locked), add = TRUE)
    #
    # # Try repeatedly, but with bounded waits and backoff
    # repeat {
    #   ## If you still want a time cap on the *attempt*, make it transient and reset:
    #   setTimeLimit(elapsed = 3, transient = TRUE)
    #   locked <- filelock::lock(lock_path, timeout = 250000)   # ~2.5 s wait, returns NULL on timeout
    #   setTimeLimit(elapsed = Inf, transient = TRUE)
    #
    #   if (!is.null(locked)) break  # acquired
    #
    #   if (isTRUE(first)) {
    #     first <- FALSE
    #     messageCache(
    #       "The cache file (", lock_path, ") is locked due to a concurrent process; waiting... ",
    #       "\nIf there is no concurrent process (i.e., no parallelism), delete that lockfile",
    #       verbose = verbose + 2
    #     )
    #   }
    #   Sys.sleep(0.25)  # backoff
    # }
    # Ensure release when the *outer* scope exits
    on.exit2(releaseLockFile(locked), envir = envir)
    locked
  }
}


#' @importFrom data.table setorderv setcolorder
showSimilar <- function(cachePath, metadata, .functionName, userTags, useCache,
                        # cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                        drv, conn, verbose) {
  devMode <- isDevMode(useCache, userTags)  # don't use devMode if no userTags
  shownCacheUserTags <- showCache(cachePath, Function = .functionName, userTags = userTags,
                          verbose = verbose - 2)
  shownCache <- showCache(cachePath, Function = .functionName, # userTags = userTags,
                          verbose = verbose - 2)
  # functionByDigest <- metadata[tagKey %in% "preDigest" & startsWith(tagValue, dotFunTxt)]$tagValue
  # shownCache <- shownCache[tagKey %in% "preDigest" & tagValue %in% functionByDigest]
  setorderv(shownCache, "createdDate", order = -1)
  # shownCache <- shownCache[tagKey != "outerFunction"] # doesn't matter what outerFunctions do, if all others are same
  # metadata <- metadata[tagKey != "outerFunction"]
  onKey <- c("tagKey", "tagValue")

  if (NROW(shownCache)) {
    userTagsMess <- if (!is.null(userTags)) {
      paste0(.message$BecauseOfA,
             "with user supplied tags: '",
             paste(userTags, collapse = ", "), "' "
      )
    }

    rmTagKeys <- "otherFunction|elapsedTime|accessed|module:|eventType:|eventTime:|outerFunction:"
    shownCache <- shownCache[grep(x = tagKey, rmTagKeys, invert = TRUE)]
    metadataSmall <- metadata[grep(x = tagKey, rmTagKeys, invert = TRUE)]
    # Can only compare on tagKeys that are *not yet* in the metadata; e.g., object.size may
    #   not be there, so don't know if it is different
    similarFull <- unique(shownCache[tagKey %in% unique(c(metadata$tagKey))], by = .dtFileMainCols)
    similarFullList <- split(similarFull, by = "cacheId")
    notInThisCall <- lapply(similarFullList, function(x) x[!metadataSmall, on = onKey])
    notInSC <- lapply(similarFullList, function(x) metadataSmall[!x, on = onKey])
    notInThisCall0 <- lapply(notInThisCall, function(x) x[grep("userTags", tagKey ), tagValue := paste0(tagKey, ":", tagValue)])
    notInSC0 <- lapply(notInSC, function(x) x[grep("userTags", tagKey ), tagValue := paste0(tagKey, ":", tagValue)])

    similar <- notInThisCall
    other <- logical()
    if (NROW(similar) == 0) {
      other <- vapply(strsplitOnlySingleColon(similarFull$tagValue, split = "\\:"),
                      function(x) ifelse(length(x) == 2, x[[2]], NA_character_), FUN.VALUE = character(1))
      otherLabels <- vapply(strsplitOnlySingleColon(similarFull$tagValue, split = "\\:"),
                      function(x) ifelse(length(x) == 2, x[[1]], NA_character_), FUN.VALUE = character(1))
      whOther <- other == "other"
      cacheIdOfSimilar <- unique(similarFull$cacheId)
      simFun <- list(funName = unique(shownCache$tagValue[shownCache$tagKey == "function"]))
      messageCache("Cache of ", .messageFunctionFn(simFun), " differs from", verbose = verbose)
      sameNames <- simFun$funName %in% .functionName
      fnTxt <- paste0(if (!is.null(.functionName))
        paste0("of '", .messageFunctionFn(.functionName), "' ") else "call ")

      if (!all(sameNames)) {
        fnTxt <- paste0("(whose function name(s) was/were '", .messageFunctionFn(paste(simFun$funName, collapse = "', '")), "')")
      }
      messageCache(paste0(.message$BecauseOfA, "the next closest cacheId(s) ",
                          paste(cacheIdOfSimilar, collapse = ", "), " ",
                          fnTxt, userTagsMess,
                          collapse = "\n"
      ), appendLF = TRUE, verbose = verbose)
      messageCache("...possible, unknown, differences in a nested list ",
                   "that is deeper than ", getOption("reproducible.showSimilarDepth", 3), " in ",
                   paste(collapse = ", ", as.character(otherLabels[whOther %in% TRUE])),
                   verbose = verbose
      )

    }

    # This is for dryRun: i.e., there is a cacheId, but no difference in metadata
    # isIdentical0 <- vapply(similar, function(x) NROW(x) == 0, FUN.VALUE = logical(1))
    isIdentical1 <- vapply(notInSC0, function(x) NROW(x) == 0, FUN.VALUE = logical(1))
    isIdentical2 <- vapply(notInThisCall0, function(x) NROW(x) == 0, FUN.VALUE = logical(1))
    isIdentical <- isIdentical1 & isIdentical2
    if (any(isIdentical)) {
      messageCache("Call is identical to ", paste(names(similar)[isIdentical], collapse = ", "),
                   " and would return that object")
      return(NULL)
    }

    if (NROW(similar)) {

      notInSCLen <- vapply(notInSC0, NROW, FUN.VALUE = integer(1))
      notInThisCallLen <- vapply(notInThisCall0, NROW, FUN.VALUE = integer(1))
      numSimilars <- length(notInSCLen)

      # First pass -- this will shrink probably down a lot
      diffs <- mapply(n = notInSCLen, m = notInThisCallLen, function(n, m) n + m, SIMPLIFY = TRUE)
      minNumDiffs <- min(diffs)
      smallestDiffs <- which(diffs == minNumDiffs)
      notInSC2 <- notInSC0[smallestDiffs]
      notInThisCall2 <- notInThisCall0[smallestDiffs]

      notInSC4 <- lapply(notInSC2, function(x) {
        x <- createSimilar(x, verbose = verbose, devMode = devMode, .functionName = .functionName)
        data.table::setnames(x, old = c(valInCacheTxt, cacheIdInCacheTxt),
                             new = c(valThisCallTxt, cacheIdThisCallTxt),
                             skip_absent = TRUE)
        })
      notInThisCall3 <- lapply(notInThisCall2, function(x) {
        ss <- createSimilar(x, verbose = verbose, devMode = devMode, .functionName = .functionName)
        if (isTRUE(any("lsStr" %in% colnames(ss))))
          set(ss, NULL, "lsStr", NULL)
        ss
        })

      simi <- Map(n = names(notInThisCall3), function(n) {
        if (NROW(notInThisCall3[[n]]) || NROW(notInSC4[[n]])) {
          a <- notInSC4[[n]][notInThisCall3[[n]], on = argTxt, allow.cartesian = TRUE]
          b <- notInThisCall3[[n]][notInSC4[[n]], on = argTxt, allow.cartesian = TRUE]
          d <- unique(rbindlist(list(a, b), fill = TRUE))
        } else {
          d <- data.table(notInSC4[[n]], valueInCache = NA, cacheIdInCache = NA)
        }

        # Convert .FUN to the actual function name; need 2 mechanisms because SpaDES.core manually
        #   places an entry with the actual name
        hasDotFun <- d[[argTxt]] %in% dotFunTxt
        if (any(hasDotFun)) {
          dups <- duplicated(d[[valThisCallTxt]])
          if (any(dups)) {
            # Remove .FUN if there is another one with "more info"
            theDupCI <- d[[valThisCallTxt]][dups]
            theDotFun <- d[[valThisCallTxt]] %in% theDupCI & hasDotFun
            d <- d[!theDotFun]
          } else {
            # case where it shows only ".FUN", with no duplication
            scHere <- shownCache[shownCache$cacheId %in% d[[cacheIdInCacheTxt]], ]# $tagKey %in% "function"
            funName <- scHere[["tagValue"]][scHere[["tagKey"]] %in% "function"]
            if (length(funName))
              d[[argTxt]] <- funName
          }

        }
        setcolorder(d, c(argTxt, cacheIdInCacheTxt, valInCacheTxt,
                         cacheIdThisCallTxt, valThisCallTxt))
        d
      })

      # Second pass -- this will be different if there were no new arguments; just arg value changes
      diffs <- mapply(x = simi, function(x) NROW(x), SIMPLIFY = TRUE)
      minNumDiffs <- min(diffs)
      smallestDiffs <- which(diffs == minNumDiffs)
      numSmallest <- length(smallestDiffs)
      simi <- simi[smallestDiffs]

      messageCache("There are ", numSimilars,
                   " calls with same fn (", .messageFunctionFn(.functionName), ") in the Cache repository.",
                   verbose = verbose * !devMode)
      if (identical(numSimilars, 1L)) {
        messageCache("It has ", minNumDiffs, " differences", verbose = verbose * !devMode)
      } else {
        messageCache("With fewest differences (", minNumDiffs, "), there ", isAre(v = numSmallest),
                     " ", numSmallest,
                     " similar calls in the Cache repository.", verbose = verbose * !devMode)
      }

      if (isDevMode(useCache, userTags)) {
        # Only replace entries that actually matched on userTags (not just function name)
        cacheIdsToClear <- intersect(unique(names(simi)), unique(shownCacheUserTags$cacheId))
        if (length(cacheIdsToClear)) {
          messageCache("------ devMode -------", verbose = verbose)
          messageCache("Previous call(s) exist in the cache with identical userTags (",
                       paste0(userTags, collapse = ", "), ")", verbose = verbose)
          messageCache("This call to cache will replace entry with cacheId(s): ",
                       paste0(simi[["cacheId"]], collapse = ", "), verbose = verbose)
          clearCache(cachePath, cacheId = cacheIdsToClear, ask = FALSE,  drv = drv, conn = conn, verbose = verbose - 2)
        }
      }
      nShow <- min(numSmallest, 5)
      messageCache("with different elements (", nShow, " most recent at top):", verbose = verbose)
      # don't add a prefix if there is no `sim` in the stack
      wis <- .whereInStack("sim")
      prefix <- if (identical(.GlobalEnv, wis) || is.null(wis)) "" else .message$NoPrefix
      messageCache(.message$dashes, prefix)
      keepers <- seq_len(nShow)
      lala <- Map(si = simi[keepers], nam = names(simi[keepers]), function(si, nam) {
        messageCache(paste0("Compared to cacheId: ", nam, prefix), verbose = verbose)
        if (verbose > 0) {
          oo <- capture.output(si)
          fn <- cliCol(getOption("reproducible.messageColourCache"))
          oo <- paddDFInitial(oo, rows = 1:2, .spaceTmpChar, colour = getOption("reproducible.messageColourCache"))
          messageColoured(paste0(paste(oo, collapse = "\n"), .message$NoPrefix),
                          colour = getOption("reproducible.messageColourCache"))
        }
        messageCache(.message$dashes, prefix)
      })

      messageCache("------ devMode -------", verbose = verbose * devMode)

    }
  } else {
    messageCache(.message$noSimilarCacheTxt(.functionName), verbose = verbose)
  }
}

CacheDBFileCheckAndCreate <- function(cachePath, drv = NULL, conn = NULL, verbose) {

  convertDBbackendIfIncorrect(cachePath, drv, conn, verbose = verbose - 1)

  dbfile <- CacheDBFile(cachePath, drv = drv, conn = conn)
  if (isTRUE(!file.exists(dbfile[1])))
    file.create(dbfile[1])
  if (!useDBI()) {
    oldDBFile <- file.path(cachePath, "cache.db")
    if (isTRUE(file.exists(oldDBFile)))
      file.remove(oldDBFile)
  }
  dbfile
}

convertCallWithSquigglyBraces <- function(call, usesDots) {
  if (length(call) == 2) {
    if (length(call[[-1]]) > 2)
      stop("Cache does not yet support multi-step caching unless using the pipe (|>)")
    call <- as.call(c(call[[1]], call[[-1]][[-1]]))
  } else if ((length(call) > 2) && isFALSE(usesDots)) {
    call <- as.call(c(call[[1]], FUN = as.list(call[-1])[[1]][[-1]], as.list(call[-1])[-1]))
  }
  call
}

wrapSaveToCache <- function(outputFromEvaluate, metadata, cache_key, cachePath, # userTags,
                            preDigest, .functionName, outputObjects,
                            cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                            drv, conn, verbose) {
  cacheIdIdentical <- cache_Id_Identical(metadata, cachePath, cache_key, cacheSaveFormat = cacheSaveFormat)
  linkToCacheId <- if (!is.null(cacheIdIdentical)) filePathSansExt(basename(cacheIdIdentical))  else NULL
  outputToSave <- .wrap(outputFromEvaluate, cachePath = cachePath, preDigest = preDigest,
                        outputObjects = outputObjects,
                        cacheId = cache_key, verbose = verbose)
  metadata <- metadata_update(outputToSave, metadata, cache_key) # .wrap may have added tags
  userTags <- paste0(metadata$tagKey, ":", metadata$tagValue)
  fs <- saveToCache(cachePath = cachePath, # drv = NULL, conn = NULL,
                    obj = outputToSave, verbose = verbose, # cache_file[1],
                    userTags = userTags, linkToCacheId = linkToCacheId,
                    cacheSaveFormat = cacheSaveFormat,
                    drv = drv, conn = conn,
                    cacheId = cache_key)
  .message$Saved(cachePath, cache_key, functionName = .functionName,
                 cacheSaveFormat = cacheSaveFormat, verbose = verbose)
  return(metadata)
}

doSaveToCache <- function(outputFromEvaluate, metadata, cachePaths, callList, # func,
                          .objects, length, algo, quick, classOptions,
                          cache_file, userTags, # .functionName,
                          debugCache,
                          detailed_key, # func_call,
                          outputObjects,
                          useCloud, cloudFolderID, gdriveLs,
                          cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                          drv, conn, useMemoise = getOption("reproducible.useMemoise", FALSE),
                          .cacheChaining = getOption("reproducible.cacheChaining", FALSE),
                          verbose, timeSaveStart, timeEvaluateStart) {
  # Can't save NULL with attributes
  if (is.null(outputFromEvaluate)) outputFromEvaluate <- "NULL"

  outputFromEvaluate <- addCacheAttr(outputFromEvaluate, .CacheIsNew = TRUE, detailed_key$key, callList$func)

  outputFromEvaluate <- cacheChainingPost(detailed_key, outputFromEvaluate,
                                            attr(callList$new_call, ".Cache")[cacheChainingOuterFunctionName],
                                            cachePaths[[1]], linkToCacheId = NULL, cacheSaveFormat,
                                          .cacheChaining = .cacheChaining, drv, conn, verbose = verbose)
  metadata <- wrapSaveToCache(outputFromEvaluate, metadata, detailed_key$key, cachePaths[[1]],
                              # userTags = paste0(metadata$tagKey, ":", metadata$tagValue),
                              outputObjects = outputObjects,
                              preDigest = detailed_key$preDigest, callList$.functionName,
                              cacheSaveFormat = cacheSaveFormat, drv, conn, verbose)

  # Memoize the outputFromEvaluate by saving it in RAM
  if (isTRUE(useMemoise)) {
    assign(detailed_key$key, outputFromEvaluate, envir = memoiseEnv(cachePaths[[1]]))
  }


  if (identical(outputFromEvaluate, "NULL")) outputFromEvaluate <- NULL

  if (isTRUE(!is.na(pmatch(debugCache, "complete"))))
    outputFromEvaluate <- .debugCache(outputFromEvaluate, detailed_key$preDigest, fullCall = callList$func_call)

  if (cloudWrite(useCloud)) {
    cloudUploadFromCache(detailed_key$key %in% filePathSansExt(gdriveLs[["name"]]), detailed_key$key,
                         cachePaths[[1]], cloudFolderID = cloudFolderID, outputFromEvaluate, verbose = verbose)
  }
  outputFromEvaluate

}




#' Remove `quote` and determine if call uses `...`
#'
#' Minor cleaning up of the `FUN` and `...` to be used subsequently. This does only very minor
#' things as it is run even if `useCache = FALSE`, i.e., even if the `Cache` is skipped.
#'
#' @inheritParams Cache
#' @inheritParams base::match.call
#' @param envir2 Environment. The environment where `matchCall2` was called.
#' @return A named list with `call` (the original call, without `quote`),
#' `FUNorig`, the original value passed by user to `FUN`, and `usesDots` which
#' is a logical indicating whether the `...` are used.
matchCall2 <- function(definition, call, envir, envir2 = parent.frame(), FUN) {
  if (missing(FUN)) {
    stop(.message$CacheRequiresFUNtxt())
  } else {
    FUNcaptured <- substitute(FUN, env = envir2)
    # This matches call for Cache
    call <- match.call(definition, call = call, expand.dots = TRUE, envir = envir)
    # call <- callIsQuote(call) # stip `quote`
    FUNorig <- call$FUN

    usesDots <- sum(!nzchar(names(call))) > 1 || sum(!names(call) %in% .namesCacheFormals) > 2
  }
  list(call = call, FUNorig = FUNorig, usesDots = usesDots, FUNcaptured = FUNcaptured)
}

#' Harmonize all forms of call
#'
#' This will convert all known (imagined) calls so that they have the same canonical
#' format i.e., `rnorm(n = 1, mean = 0, sd = 1)`
#'
#' @param callList A named list with elements `call`, `usesDots` and `FUNorig`
#' @param .callingEnv The calling environment where `Cache` was called from
#' @param .functionName A possible function name. If omitted, then it will be deduced
#'   from the `callList` and may be inaccurate.
#' @return A named list. We illustrate with the example `rnorm(1)`. The named
#' list will have the original `callList` (`call` (the original call, without `quote`),
#' `FUNorig`, the original value passed by user to `FUN`, and `usesDots` which
#' is a logical indicating whether the `...` are used), and appended with `new_call`
#' (the harmonized call, with the function and arguments evaluated, e.g.,
#' `(function (n, mean = 0, sd = 1) .Call(C_rnorm, n, mean, sd))(1)`), `func_call`, the same harmonized call
#' with neither function nor arguments not evaluated (e.g., `rnorm(1)`), `func` which
#' will be function or method definition
#' `function (n, mean = 0, sd = 1) .Call(C_rnorm, n, mean, sd)`,
#' and `.functionName`, which will be the function name as a character string (`rnorm`)
#' either directly passed from the user's `.functionName` or deduced from the `func_call`.
harmonizeCall <- function(callList, .callingEnv, .functionName = NULL) {
  callList$call <- callIsQuote(callList$call) # stip `quote`

  isSquiggly <- isSquigglyCall(callList$FUNorig)
  # isSquiggly <- is(callList$FUNorig, "{")
  if (isTRUE(isSquiggly))
    callList$call <- convertCallWithSquigglyBraces(callList$call, callList$usesDots)
  new_call <- convertCallToCommonFormat(callList$call, callList$usesDots, isSquiggly, .callingEnv) # evaluated arguments
  func_call <- attr(new_call, ".Cache")$func_call         # not evaluated arguments
  .functionNamePoss <- attr(new_call, ".Cache")$.functionName
  func <- as.list(new_call)[[1]]

  if (!is.null(.functionName)) {
    dotFnGrep <- "\\.functionName"
    hasDotFNLogical <- grepl(dotFnGrep, .functionName)
    hasDotFN <- isTRUE(any(hasDotFNLogical) )
    if (hasDotFN)
      .functionName <- gsub(dotFnGrep, .functionNamePoss, .functionName)
  }

  # Try to identify the .functionName; if can't just use the matched call callList$FUNorig
  if (is.null(.functionName)) {
    if (!is.null(.functionNamePoss))
      .functionName <- .functionNamePoss
    else
      .functionName <- getFunctionName2(func_call)# as.character(normalized_FUN[[1]])
  }
  if (!isTRUE(any(nzchar(.functionName)))) {
    .functionName <- format(callList$FUNorig)
  }
  append(callList, list(new_call = new_call, func_call = func_call,
                           func = func, .functionName = .functionName))
}

cacheIdOverride <- function(cacheId, key, .functionName, verbose) {
  if  (identical(cacheId, "previous")) {
    cacheId <- getPreviousEntryInCache(.functionName, cacheId, verbose)
  } else {
    shownCache <- cacheIdCheckInCache(cacheId, calculatedCacheId = key, .functionName, verbose)
    # if (NROW(shownCache) == 0)
    #   cacheId <- NULL
  }
  cacheId
}

useCacheFromNested <- function(useCache) {
  isNested <- isTRUE(.pkgEnv$.reproEnv2$nestLevel > 1)
  if (isNested && isTRUE(useCache))
    useCache <- .pkgEnv$.reproEnv2$useCache
  useCacheDueToNumeric <- (is.numeric(useCache) && isTRUE(useCache < .pkgEnv$.reproEnv2$nestLevel))
  if (is.character(useCache)) {
    if (any(!is.na(pmatch(table = useCache, c("over", "dev"))))) {
      return(useCache)
    }
  }
  !(isFALSE(useCache) || useCache == 0 || isTRUE(useCacheDueToNumeric))
}

loadFromDiskOrMemoise <- function(fromMemoise = FALSE, useCache,
                                  useCloud, cloudFolderID = NULL, gdriveLs,
                                  cachePath, detailed_key,
                                  functionName,
                                  cache_file = NULL, changedSaveFormat, sameCacheID,
                                  cache_file_orig, func, shownCache = NULL,
                                  full_call, outputObjects,
                                  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                                  .cacheChaining = getOption("reproducible.cacheChaining", FALSE),
                                  drv, conn, verbose) {

  cache_key <- detailed_key$key
  if (identical(useCache, "overwrite")) {
    clearCacheOverwrite(cachePath, cache_key, functionName, drv, conn, verbose)
    return(invisible(.returnNothing))
  } else {
    format <- if (missing(cache_file) || is.null(cache_file)) cacheSaveFormat else
      fileExt(cache_file)

    for (iii in 1:2) {
      fe <- CacheDBFileSingle(cachePath = cachePath, cacheId = cache_key,
                              cacheSaveFormat = cacheSaveFormat)
      if (useDBI()) {
        rerun <- FALSE
      } else {
        feReally <- file.exists(fe)
        if (any(feReally %in% FALSE)) {
          formatNew <- formatCheck(cachePath, cache_key, format)
          if (!identical(formatNew, cacheSaveFormat)) {
            cacheSaveFormat <- formatNew
            next
          }

        }
        # usually happens when user had memoise on before, then turned it off, then turned it back on
        danglingMemoise <- (isFALSE(any(feReally)) && fromMemoise)
        if (isTRUE(danglingMemoise)) rm(list = cache_key, envir = memoiseEnv(cachePath))
        rerun <- (!isTRUE(any(feReally)) && !fromMemoise) || danglingMemoise
        break
      }
    }

    cacheSaveFormatFail <- FALSE
    if (is.null(shownCache)) {
      # shownCache <- showCacheFast(cache_key, cachePath, dtFile = fe,
      #                                 # cacheSaveFormat = cacheSaveFormat,
      #                                 drv = drv, conn = conn)
      shownCache <- try(showCacheFast(cache_key, cachePath, dtFile = fe,
                                      # cacheSaveFormat = cacheSaveFormat,
                                      drv = drv, conn = conn),
                        silent = TRUE)
      if (is(shownCache, "try-error")) {
        if (isTRUE(any(grepl("format not detected", shownCache)))) {
          cacheSaveFormatFail <- TRUE
        } else { # e.g., change from qs to qs2
          cacheSaveFormatFail <- TRUE
        }
      }
    }

    if (isFALSE(cacheSaveFormatFail))
      .cacheMessageObjectToRetrieve(functionName, shownCache, cachePath,
                                    cacheId = cache_key, cacheSaveFormat = cacheSaveFormat, verbose = verbose)
    memoiseFail <- FALSE
    if (fromMemoise && !rerun) {
      # output <- get(cache_key, envir = memoiseEnv(cachePath))
      output <- .unwrap(get(cache_key, envir = memoiseEnv(cachePath)), cacheId = cache_key, cachePath = cachePath,
                        drv = drv, conn = conn)
      # need to update the individual files in file-backed objects from the cache; can't use memoise

      # Some objects, especially Rcpp objects can get stale; rerun if this is the case; the test with subsetting 1st element
      #   is not great, but I could not find a better one that will fail on those Rcpp fails. The problem
      #   is that the object exists, but it's inner structure is wrong
      outputTestIntegrity <- try(output[1], silent = TRUE) # This needs to be `try`, not `try2`
      fns <- try2(Filenames(output), silent = TRUE) # previous will only get some of the failures
      memoiseFail <- dealWithCacheRecoveryErrors(memoiseFail, outputTestIntegrity, fns, cache_key, cachePath, outputObjects)
      # if (isTRUE(is(outputTestIntegrity, "try-error")) || isTRUE(is(fns, "try-error"))) {
      #   failMsgs <- "external pointer.+not valid|NULL value passed as symbol address"
      #   if (isTRUE(any(grepl(failMsgs, outputTestIntegrity))) ||
      #       isTRUE(any(grepl(failMsgs, fns)))) {
      #     memoiseFail <- TRUE
      #     rm(list = cache_key, envir = memoiseEnv(cachePath))
      #     cache_file <- CacheStoredFile(cachePath, cache_key, readOnly = TRUE)
      #   }
      # } else {
      #   fns <- fns[nzchar(fns)]
      #   if (!is.null(fns) && length(fns) > 0) {
      #     fnsInOutputObjects <- intersect(names(fns), outputObjects)
      #     fns <- fns[fnsInOutputObjects]
      #     fnsExistBefore <- try2(file.exists(fns))
      #     fnsInCache <- file.path(CacheStorageDir(cachePath),
      #                             basename(.prefix(fns, prefixCacheId(cacheId = cache_key))))
      #     hardLinkOrCopy(fnsInCache, fns, overwrite = TRUE, verbose = FALSE)
      #     fnsExistAfter <- file.exists(fns)
      #     if (any(fnsExistAfter %in% FALSE) && isTRUE(any(fnsExistBefore != fnsExistAfter))) # this means that hardLinkOrCopy failed
      #       browser()
      #   }
      # }
    }

    if (!fromMemoise || rerun || memoiseFail || cacheSaveFormatFail) {
      obj <- if (!is.null(cache_file)) {
        # loadFile(cache_file, cacheSaveFormat = cacheSaveFormat,
        #              cacheId = cache_key, cachePath = cachePath, # in case it needs swapCacheFormat
        #              drv = drv, conn = conn, verbose = verbose)
        try(loadFile(cache_file, cacheSaveFormat = cacheSaveFormat,
                     cacheId = cache_key, cachePath = cachePath, # in case it needs swapCacheFormat
                     drv = drv, conn = conn, verbose = verbose), silent = TRUE)
      } else {
        rerun <- TRUE
      }

      if (isTRUE(changedSaveFormat)) {
        swapTry <- try(swapCacheFileFormat(
          wrappedObj = obj, cachePath = cachePath, drv = drv, conn = conn,
          cacheId = cache_key, sameCacheID = sameCacheID,
          userTags = paste0(shownCache$tagKey, ":", shownCache$tagValue),
          newFile = cache_file_orig, verbose = verbose), silent = TRUE)
        cacheSaveFormat <- fileExt(cache_file_orig) # setdiff(.cacheSaveFormats, cacheSaveFormat)
        # rerun <- TRUE
      }
      output <- try(.unwrap(obj, cachePath = cachePath, cacheId = cache_key))
      if (is(obj, "try-error") || rerun || is(output, "try-error")) {
        messageCache("It looks like the cache file is corrupt or was interrupted during write; deleting and recalculating")
        otherFiles2 <- dir(CacheStorageDir(cachePath), pattern = cache_key, full.names = TRUE)
        if (!is(shownCache, "try-error")) {
          if (!is.null(shownCache)) {
            otherFiles <- normPath(file.path(CacheStorageDir(cachePath),
                                             shownCache[tagKey == "filesToLoad"]$tagValue))
            otherFiles2 <- c(otherFiles, otherFiles2)
          }
        }
        rmFiles <- unique(c(cache_file, otherFiles2))
        unlink(rmFiles)
        return(.returnNothing)
      }

    }

    if (cloudWrite(useCloud)) {
      cloudUploadFromCache(cache_key %in% filePathSansExt(gdriveLs[["name"]]), cache_key,
                           cachePath, cloudFolderID = cloudFolderID, output, verbose = verbose)
    }

    .cacheMessage(object = output, functionName = functionName, fromMemoise = fromMemoise, verbose = verbose)

    if (getOption("reproducible.useMemoise", FALSE)) {
      cache_key_in_memoiseEnv <- exists(cache_key, envir = memoiseEnv(cachePath), inherits = FALSE)
      if (cache_key_in_memoiseEnv %in% FALSE) {
        # assign(cache_key, .unwrap(obj, cachePath = cachePath, cacheId = cache_key),
        #        envir = memoiseEnv(cachePath))
        assign(cache_key, obj, envir = memoiseEnv(cachePath)) # try without .unwrap in memoiseEnv
      }
    }

    if (!is.null(output))
      output <- addCacheAttr(output, .CacheIsNew = FALSE, outputHash = cache_key, func)

    .addTagsRepoAccessedTime(cache_key, cachePath = cachePath, cacheSaveFormat = cacheSaveFormat)
    attr(output, ".Cache")$newCache <- FALSE

    .dotsFromCache <- as.list(attr(full_call, ".Cache")$func_call)[-1]
    # # This allows for any class specific things
    if ("object" %in% names(.dotsFromCache))
      .dotsFromCache <- .dotsFromCache[setdiff(names(.dotsFromCache), "object")]

    output <- do.call(.prepareOutput, args = append(list(object = output, cachePath),
                                                    .dotsFromCache))

    output <- cacheChainingPost(detailed_key, output,
                                attr(full_call, ".Cache")[cacheChainingOuterFunctionName],
                                cachePath, linkToCacheId = NULL, cacheSaveFormat,
                                .cacheChaining = .cacheChaining, drv, conn, verbose = verbose)

    return(output)
  }
}

defunct <- function(argNames) {
  # argNames <- call)
  deps <- .defunctCacheArgs
  for (d in deps)
    if (d %in% argNames) {
      stop(.message$defunct(d), call. = FALSE)
    }
}


.defunctCacheArgs <- c("sideEffect", "makeCopy", "compareRasterFileLength",
                       "cacheRepo", "digestPathContent")


isDevMode <- function(useCache, userTags) {
  isTRUE(any(pmatch(table = useCache, "dev") %in% 1)) && !is.null(userTags)
}

## `useCloud` accepts: TRUE/FALSE/NULL, or one of "push"/"pull".
##   - TRUE  / "push": developer role -- bidirectional. Download on cloud hit;
##                     upload on miss.
##   - "pull"        : user role -- read-only. Download on cloud hit; never
##                     upload. If the local cache already has the object, the
##                     cloud is not consulted at all (the gdriveLs fetch is
##                     deferred until after the local check fails).
##   - FALSE / NULL  : cloud disabled.
## Legacy "^w"/"^r" prefix matching is retained for back-compat (e.g. "write",
## "read", "readOnly") since the contract is otherwise narrow.
cloudWrite <- function(useCloud) {
  isTRUE(useCloud) ||
    identical(useCloud, "push") ||
    isTRUE(any(grepl("^w", useCloud) %in% 1))
}

cloudWriteOrRead <- function(useCloud) {
  cloudWrite(useCloud) || cloudRead(useCloud)
}

cloudReadOnly <- function(useCloud) {
  identical(useCloud, "pull") ||
    isTRUE(any(grepl("^r", useCloud) %in% 1))
}

cloudRead <- function(useCloud) {
  cloudReadOnly(useCloud) || isTRUE(useCloud) || identical(useCloud, "push")
}

## Validate the `useCloud` argument and return it unchanged. Errors on a
## character value that is not "pull" or "push" (or a legacy ^w/^r prefix).
validateUseCloud <- function(useCloud) {
  if (is.null(useCloud) || isTRUE(useCloud) || isFALSE(useCloud))
    return(invisible(useCloud))
  if (is.character(useCloud) && length(useCloud) == 1L &&
      (useCloud %in% c("pull", "push") ||
       grepl("^[wr]", useCloud)))
    return(invisible(useCloud))
  stop("`useCloud` must be TRUE, FALSE, NULL, \"pull\", or \"push\"; got: ",
       deparse(useCloud), call. = FALSE)
}

keyInGdriveLs <- function(cache_key, gdriveLs) {
  grepl(paste0("^(", cache_key, ").+$"), gdriveLs[["name"]])
  #filePathSansExt(filePathSansExt(gdriveLs[["name"]])) %in%  # double filePathSansExt because of the .dbFile.rds
  #  cache_key
}

verboseCacheDFAll <- function(verbose, functionName, times) {
  verboseDF1(verbose, functionName, times$CacheDigestStart, times$EvaluateStart)
  verboseDF2(verbose, functionName, times$EvaluateStart, times$SaveStart)
  verboseDF3(verbose, functionName, times$CacheDigestStart, times$SaveEnd)
  .message$CacheTimings(verbose)
}

optionsSetForCache <- function(drv = NULL, conn = NULL, envir = parent.frame(1),
                                verbose = getOption("reproducible.verbose")) {
  if (!is.null(drv) || !is.null(conn)) {
    useDBI(TRUE, verbose = verbose)
  }
  if (isFALSE(useDBI())) {
    opts <- options(
      reproducible.useDBI = FALSE
    )
  }
  opt2 <- options(
    reproducible.useCacheV3 = TRUE
  )
}

identical2 <- function(a, b) {
  id <- identical(a, b)
  if (isTRUE(id))
    return(TRUE)
  isTRUE(all.equal(a, b, check.attributes = FALSE))
}



evalTheFunAndAddChanged <- function(callList, keyFull, outputObjects, length, algo, quick,
                                    classOptions, .callingEnv, verbose, ...) {
  outputFromEvaluate <- evalTheFun(callList$FUNcaptured, !callList$usesDots,
                                   matchedCall = callList$call, envir = .callingEnv,
                                   verbose = verbose, ...)

  # Because this has be run, it means that it has changed; add an attribute to say that
  outputFromEvaluate <- .addChangedAttr(outputFromEvaluate, keyFull$preDigest,
                                        origArguments = attr(callList$new_call, ".Cache")$args_w_defaults,
                                        .objects = outputObjects, length = length,
                                        algo = algo, quick = quick, classOptions = classOptions, ...
  )
  outputFromEvaluate
}



.dtFileMainCols <- c("cacheId", "tagKey", "tagValue", "createdDate")

doDigestPrepare <- function(new_call, omitArgs, .cacheExtra) {
  toDigest <- attr(new_call, ".Cache")$args_w_defaults # not evaluated arguments

  toDigest$.FUN <- attr(new_call, ".Cache")$method
  # Deal with omitArgs:
  # - TRUE  => drop every captured arg; digest is based on .FUN (the actual
  #            function value, body included, so source edits still bust the
  #            cache) plus .cacheExtra
  # - char  => drop the named args
  # - NULL  => default, no change
  if (isTRUE(omitArgs)) {
    toDigest <- toDigest[names(toDigest) %in% ".FUN"]
  } else if (is.character(omitArgs)) {
    if (any("FUN" %in% omitArgs))
      omitArgs <- c(dotFunTxt, omitArgs)
    toDigest[omitArgs] <- NULL
  }
  # Deal with .cacheExtra by adding it to the list of objects to digest
  if (!is.null(.cacheExtra))
    toDigest <- append(toDigest, list(.cacheExtra = .cacheExtra))
  toDigest
}




doDigest <- function(toDigest, .functionName, .objects, length, algo, quick,
                      classOptions, timeCacheDigestStart, verbose) {
  detailed_key <- CacheDigest(toDigest,
                              .functionName = .functionName,
                              .objects = .objects,
                              length = length, algo = algo, quick = quick,
                              classOptions = classOptions,
                              calledFrom = "Cache"
  )
  diTi <- difftime(Sys.time(), timeCacheDigestStart, units = "sec")
  if (diTi > 5) {
    messageCache("Object digesting for ", .messageFunctionFn(.functionName)," took: ", format(diTi, digits = 2))
  }
  verboseCacheMessage(detailed_key$preDigest, .functionName, timeCacheDigestStart, quick = quick,
                      modifiedDots = toDigest, verbose = verbose, verboseLevel = 3)

  names(detailed_key)[[1]] <- "key"
  detailed_key
}



appendFunctionNameToNestedTags <- function(userTags, functionName) {
  # allUT <- c(paste0("outerFunction:", functionName), userTags)
  # dups <- duplicated(sapply(strsplitOnlySingleColon(allUT), tail, 1))
  # allUT <- allUT[!dups] # only take after :
  # allUT <- sort(allUT)
  .pkgEnv$.reproEnv2$userTags <- c(.pkgEnv$.reproEnv2$userTags,
                                   paste0("outerFunction:", functionName))
  .pkgEnv$.reproEnv2$userTags <- .pkgEnv$.reproEnv2$userTags[!duplicated(.pkgEnv$.reproEnv2$userTags)]
}

.txtGrepStrSplitSingleColon <- "(?<!:):(?!:)"

strsplitOnlySingleColon <- function(x, ...) {
  strsplit(x, split = .txtGrepStrSplitSingleColon, perl = TRUE)
}


reorder_by_first_element <- function(x) {
  keys <- sapply(x, `[[`, 1)  # extract the first elements
  seen <- character()
  res <- list()

  for (i in seq_along(x)) {
    key <- keys[i]
    if (!key %in% seen) {
      # First time we see this key: append
      res[[length(res) + 1]] <- x[[i]]
      seen <- c(seen, key)
    } else {
      # Find last index where this key is already in res
      last_idx <- max(which(sapply(res, `[[`, 1) == key))
      res <- append(res, list(x[[i]]), after = last_idx)
    }
  }

  res
}



createSimilar <- function(similar, .functionName, verbose, devMode) {

  simi <- similar[, .N, by = "cacheId"][similar, on = "cacheId"]
  data.table::setorderv(simi, c("N", "createdDate"))
  numSimilars <- NROW(unique(similar$cacheId))
  # messageCache("There are ", numSimilars,
  #              " similar calls (same fn: ", .messageFunctionFn(.functionName), ") in the Cache repository.",
  #              verbose = verbose * !devMode)
  simi <- split(simi, by = "N") # take first element in split list
  if (length(simi)) {
    simi <- simi[[1]]
    # if (identical(numSimilars, 1L)) {
    #   messageCache("It has ", simi$N[[1]], " differences", verbose = verbose * !devMode)
    # } else {
    #   messageCache("With fewest differences (", simi$N[[1]], "), there are ",
    #                NROW(unique(simi$cacheId)),
    #                " similar calls in the Cache repository.", verbose = verbose * !devMode)
    # }
    twoCols <- strsplitOnlySingleColon(simi[["tagValue"]])
    lens <- lengths(twoCols)
    hasNoColon <- lens == 1
    if (isTRUE(any(hasNoColon))) {
      whNoColon <- which(hasNoColon)
      twoCols[whNoColon] <- lapply(whNoColon, function(x) c(simi[["tagKey"]][[x]], twoCols[[x]]))
    }

    args <- vapply(twoCols, function(x) x[[1]], FUN.VALUE = character(1))
    lens <- lengths(twoCols)
    vals <- rep("", length(twoCols))
    vals[lens > 1] <- vapply(twoCols[lens > 1], function(x) x[[2]], FUN.VALUE = character(1))
    set(simi, NULL, argTxt, args)
    set(simi, NULL, "value", vals)
    set(simi, NULL, c("N", "tagKey", "tagValue", "createdDate"), NULL)
    setcolorder(simi, c("cacheId", argTxt, "value"))
    setnames(simi, old = c("cacheId", "value"), new = c(cacheIdInCacheTxt, valInCacheTxt))
  } else {
    simi <- data.table(arg = character(), cacheIdInCache = character(), valueInCache = character())
  }
  simi
}


.txtNoPrefix <- "noPrefix"
.txtDryRunTRUE <- "dryRun = TRUE: "


stopRcppError <- function(toDigest, .objects, length, algo, quick, classOptions) {
  ooo <- Map(obj = names(toDigest), function(obj)
    try2(.robustDigest(toDigest[[obj]], .objects = .objects,
                      length, algo, quick, classOptions), silent = TRUE))
  ite <- Map(o = ooo, function(o) {
    is(o, "try-error")
  })
  ite <- ite[unlist(ite)]
  if (length(ite))
    stop(paste(names(ite), collapse = ", "), " ", isAre(ite), " corrupt. ",
         "This can usually be resolved by restarting the R session")
  else
    stop("One or more objects to be digested for Cache are corrupt. ",
         "This can usually be resolved by restarting the R session")
}



isSquigglyCall <- function(x) {
  is(x, "{")
}


cacheChainingSetup <- function(.cacheChaining, callList, omitArgs, verbose) {
  if (isTRUE(.cacheChaining %in% TRUE)) { #
    .cacheChaining <- sys.function(-2)
  }
  cfdigList <- details <- preDigests <- NULL
  messageCacheChainChanged <- FALSE

  if (useCacheChaining(.cacheChaining)) {
    bb <- attr(callList$new_call, ".Cache")
    hasCacheTags <- lapply(bb$args_w_defaults, function(y) attr(y, "tags")) |> unlist()
    cfdig <- .robustDigest(.cacheChaining)
    cfdigList <- list(cfdig) |> setNames(cacheChainingOuterFunctionName)
    if (length(hasCacheTags)) {
      if (is.null(.pkgEnv[["cacheChaining"]])) {
        .pkgEnv$cacheChaining <- new.env(parent = emptyenv())
      }
      details <- .pkgEnv$cacheChaining[[cfdigList[[1]]]]
      cids <- names(details)
      if (length(cids) == 0) { # not in the RAM stashing place "yet"; use normal Cache
        sc2 <- showCacheFast(cacheId = cfdigList[[1]])
        cids <- sc2[["tagValue"]][sc2$tagKey == "cacheChain"]
      }

      if (length(cids)) {
        # The function being assessed has to assess objects that were created within this same function;
        #   otherwise they could be from a Cache outside this function
        wasItInThisFn <- lapply(bb$args_w_defaults, function(y) attr(y, cacheChainingOuterFunctionName)) |> unlist()
        wasItInThisFn <- identical(wasItInThisFn[[1]], cfdigList[[1]])
        if (length(hasCacheTags) && wasItInThisFn) {
          onlyOneCid <- gsub("cacheId:", "", hasCacheTags)
          messageCache("Using cacheChaining ...", verbose = verbose)

          if (exists("sc2", inherits = FALSE)) {
            onlyOneCidReal <- sc2[["cacheId"]][sc2$tagValue %in% onlyOneCid]
            sc3 <- sc2[sc2[["cacheId"]] %in% onlyOneCidReal] # in case of duplicate entries
            sc4 <- sc3#[sc3$tagValue == onlyOneCid]
            preDigests <- Map(nam = names(hasCacheTags), function(nam)
              sc4[["tagValue"]][sc4[["tagKey"]] == nam])
            details <- as.list(sc2$tagValue[-1]) |> setNames(sc2$tagKey[-1]) |>
              list() |> setNames("preDigests") |>
              list() |> setNames(onlyOneCid) |>
              as.environment()
          } else {
            preDigests <- Map(nam = names(hasCacheTags), oocid = onlyOneCid, function(nam, oocid) {
              .pkgEnv$cacheChaining[[cfdigList[[1]]]][[oocid]]$preDigests[[nam]]
            })
          }

          omitArgs <- c(omitArgs, names(hasCacheTags))
          messageCache("Skipping digest of ", paste0(names(hasCacheTags), collapse = ", "), verbose = verbose)
          # .cacheExtra <- c(.cacheExtra, preDigests)
        } else {
          messageCacheChainChanged <- TRUE
          # messageCache("Using cacheChaining; but .cacheChaining has changed; adding to new chain")
        }
      } else {
        messageCacheChainChanged <- TRUE
        # messageCache("Using cacheChaining; but .cacheChaining has changed or ",
        #              "this is the first call in the .cacheChaining; starting a new chain")
      }
    } else {
      messageCacheChainChanged <- TRUE
    }
    attr(callList$new_call, ".Cache") <- append(attr(callList$new_call, ".Cache"), cfdigList)

  }
  if (isTRUE(messageCacheChainChanged))
    messageCache("Using cacheChaining; but enclosing function has changed or ",
                 "this is the first Cached call in the function where ", .messageFunctionFn(callList$.functionName),
                 " is being Cached; starting a new chain")
  list(.cacheChaining = .cacheChaining,
       preDigests = preDigests,
       omitArgs = omitArgs,
       callList = callList,
       details = details,
       cfdigList = cfdigList)
}

cacheChainingPost <- function(detailed_key, outputFromEvaluate, cacheChainingOuterFunction,
                              cachePath, linkToCacheId, cacheSaveFormat, .cacheChaining, drv, conn,
                              verbose = getOption("reproducible.verbose")) {
  if (!isTRUE(.cacheChaining %in% FALSE)) {

    dk <- detailed_key[["preDigest"]]
    if (!is.null(dk)) { # some have only `key` and no `preDigest`, e.g., Cache(.inputObjects(sim), .objects = objectsToEvaluateForCaching,
      if (!is.character(detailed_key$preDigest$.cacheExtra) &&
          !is.null(detailed_key$preDigest$.cacheExtra$cacheChainingOuterFunction)) {
        cacheChainingFnDigest <- detailed_key$preDigest$.cacheExtra$cacheChainingOuterFunction
      } else {
        cacheChainingFnDigest <- dk[[surroundingFunctionLabel]]
      }
      attr(outputFromEvaluate, cacheChainingOuterFunctionName) <- cacheChainingFnDigest

      if (is.null(.pkgEnv$cacheChaining[[cacheChainingFnDigest]])) {
        .pkgEnv$cacheChaining[[cacheChainingFnDigest]] <- new.env(parent = emptyenv())
      }

      dkSimple <- dk[-which(names(dk) == surroundingFunctionLabel)]
      if (any(names(dkSimple) %in% ".cacheExtra")) {
        if (is.null(names(dkSimple[[".cacheExtra"]]))) {
          names(dkSimple[[".cacheExtra"]]) <- as.character(seq_along(length(dkSimple[[".cacheExtra"]])))
        }
      }
      userTags <- paste0(names(unlist(dkSimple)), ":", paste0(detailed_key$key, ":", unlist(dkSimple)))
      fil <- CacheDBFileSingle(cachePath = cachePath, cacheId = cacheChainingFnDigest)
      needWrite <- TRUE
      if (file.exists(fil)) {
        # browser() # what should 'cacheId' be --> detailed_key?
        tmp <- loadFile(fil,
                        # cacheId = cacheId,
                        cachePath = cachePath, # in case it needs swapCacheFormat
                        drv = drv, conn = conn, verbose = verbose)
        userTags1 <- paste0(tmp$tagKey, ":", tmp$tagValue)
        userTags2 <- union(userTags, userTags1)
        if (identical(length(userTags2), length(userTags1))) {
          needWrite <- FALSE
        } else {
          userTags <- userTags2
        }

      }
      if (isTRUE(needWrite)) {
        # This adds or updates a new entry in the cache repository about the function itself
        fs <- saveToCache(cachePath = cachePath,
                          obj = NULL, verbose = verbose - 1, # cache_file[1],
                          userTags = userTags, linkToCacheId = linkToCacheId,
                          cacheSaveFormat = cacheSaveFormat,
                          drv = drv, conn = conn,
                          cacheId = cacheChainingFnDigest)

        assign(detailed_key$key,
               list(preDigests = detailed_key$preDigest) ,
               envir = .pkgEnv$cacheChaining[[cacheChainingFnDigest]])
      }
    }
  }
  return(outputFromEvaluate)
}

cacheChainingOuterFunctionName <- "cacheChainingOuterFunction"
cacheChainLabel <- "cacheChain_"
surroundingFunctionLabel <- "surroundingFunction"


useCacheChaining <- function(.cacheChaining) {
  first <- !is.null(.cacheChaining) && (is.function(.cacheChaining) || !.cacheChaining %in% FALSE)
  if (isTRUE(first)) {
    udbi <- useDBI(verbose = -2)
    if (udbi %in% TRUE) {
      if (is.null(.pkgEnv$cacheChainingMessage) ||
          isTRUE(difftime(Sys.time(), .pkgEnv$cacheChainingMessage) > 60*60)) {
        .pkgEnv$cacheChainingMessage <- Sys.time()
        message("cacheChaining will only work if not using DBI cache backend; ",
                "\nset `options(reproducible.cacheChaining = FALSE)` to remove this message",
                "\nor set `useDBI(FALSE)` ... this message will be shown at package startup and every hour")
      }
      first <- FALSE
    }
  }
  first
}


cacheChainingStep <- function(keyFull, callList, .cacheChaining, cacheChainDetails, cachePaths) {
  if (!isTRUE(.cacheChaining %in% FALSE)) {

    alreadyCachedArgs <- lapply(attr(callList$new_call, ".Cache")$args_w_defaults,
                                function(x) attr(x, "tags")) |> unlist()
    if (!is.null(alreadyCachedArgs)) {
      alreadyCachedTags <- paste0(cacheChainLabel, names(alreadyCachedArgs))#, ":",
      newBits <- Map(act = alreadyCachedTags, aca = alreadyCachedArgs, function(act, aca) {
        unname(gsub("cacheId:", "", aca))
      })
      keyFull[["preDigest"]] <- modifyList(keyFull[["preDigest"]], newBits)
    }
    .cacheChaining <- if (missing(cacheChainDetails)) .cacheChaining else cacheChainDetails$.cacheChaining
    if (!is.function(.cacheChaining))
      .cacheChaining <- sys.function(-2)
    keyFull[["preDigest"]][[surroundingFunctionLabel]] <- .robustDigest(.cacheChaining)
    if (useCacheChaining(.cacheChaining) && !is.null(cacheChainDetails$omitArgs)) {
      ccd <- lapply(cacheChainDetails$details, function(x) x$preDigests)
      cacheDigestDetails <- rbindlist(ccd, idcol = "obj", use.names = TRUE, fill = TRUE)

      sc <- showCacheFast(cacheId = cacheChainDetails$cfdigList[[1]])
      if (!is.null(sc)) {
        sc <- setDT(sc)
        sss <- strsplit(sc$tagValue, ":")
        set(sc, NULL, "cacheId2", vapply(sss, function(x) x[[1]], character(1)))
        set(sc, NULL, "tagValue", vapply(sss, function(x) x[[2]], character(1)))
        kf <- keyFull$preDigest[!names(keyFull$preDigest) %in% surroundingFunctionLabel]
        pre <- setDT(list(tagKey = names(unlist(kf)), tagValue = unname(unlist(kf))))

        # The next few steps are slower with data.table and are the bottlenecks when profiling
        # outs2 <- sc[pre, on = colnames(pre), nomatch = NA]
        outs <- setDT(merge(setDF(sc), setDF(pre), by = colnames(pre), all.y = T))
        if (anyNA(outs$cacheId)) {
          # if (length(outs$cacheId) == 0) {
          # not usable -- skip
        } else {
          neededNum <- length(unique(outs$tagKey))
          # lll <- outs[, .N, by = "cacheId2"]
          lll2 <- table(outs$cacheId2)
          # setorderv(lll, "cacheId2")
          # lll2 <- lll2[order(names(lll2))]
          # if (!(identical(lll$cacheId2, names(lll2)) && identical(lll$N, as.integer(unname(lll2))))) {
          # }
          hasAll <- lll2 == neededNum
          # hasAll <- lll$N == neededNum
          if (any(hasAll)) {
            cidToCheck <- names(lll2)[hasAll]
            # cidToCheck <- lll$cacheId2[hasAll]
            if (NROW(cidToCheck) > 1) {
              browser() # these are now wrong -- should be unnecessary -- there should not ever be >1 of these; so don't need to !keep
            }
            if (NROW(cidToCheck)) {
              if (!keyFull$key %in% cidToCheck) { # no override needed
                sc <- setDT(sc)
                rrr <- sc[sc$cacheId2 %in% cidToCheck & startsWith(sc$tagKey, cacheChainLabel)]
                tv <- unique(rrr$tagValue)
                tk <- unique(rrr$tagKey)
                scCheck <- showCacheFast(tv, cachePaths[[1]])
                argName <- gsub(cacheChainLabel, "", tk)
                allGood <- any(startsWith(scCheck$tagValue, paste0(argName, ":")))
                if (allGood) {
                  cacheIdOverrideFromChaining <- showCacheFast(cacheId = cidToCheck, cachePaths[[1]])
                  keyFull$key <- cacheIdOverrideFromChaining$cacheId[[1]]
                }
              }
            }
          }
        }
      }
    }
  }
  return(keyFull)
}


inCacheTxt <- "InCache"
thisCallTxt <- "ThisCall"
argTxt <- "arg"
valInCacheTxt <- paste0("value", inCacheTxt)
cacheIdInCacheTxt <- paste0("cacheId", inCacheTxt)
cacheIdThisCallTxt <- paste0("cacheIdOf", thisCallTxt)
valThisCallTxt <- paste0("value", thisCallTxt)


dealWithCacheRecoveryErrors <- function(memoiseFail, outputTestIntegrity, fns, cache_key, cachePath, outputObjects) {
  if (isTRUE(is(outputTestIntegrity, "try-error")) || isTRUE(is(fns, "try-error"))) {
    failMsgs <- "external pointer.+not valid|NULL value passed as symbol address"
    if (isTRUE(any(grepl(failMsgs, outputTestIntegrity))) ||
        isTRUE(any(grepl(failMsgs, fns)))) {
      memoiseFail <- TRUE
      rm(list = cache_key, envir = memoiseEnv(cachePath))
      cache_file <- CacheStoredFile(cachePath, cache_key, readOnly = TRUE)
    }
  } else {
    fns <- fns[nzchar(fns)]
    if (!is.null(fns) && length(fns) > 0) {
      fnsInOutputObjects <- intersect(names(fns), outputObjects)
      fns <- fns[fnsInOutputObjects]
      fnsExistBefore <- try2(file.exists(fns))
      fnsInCache <- file.path(CacheStorageDir(cachePath),
                              basename(.prefix(fns, prefixCacheId(cacheId = cache_key))))
      hardLinkOrCopy(fnsInCache, fns, overwrite = TRUE, verbose = FALSE)
      fnsExistAfter <- file.exists(fns)
      if (any(fnsExistAfter %in% FALSE) && isTRUE(any(fnsExistBefore != fnsExistAfter))) # this means that hardLinkOrCopy failed
        browser()
    }
  }
  memoiseFail
}

test_that("omitArgs = TRUE drops every captured arg from the digest", {
  ## Two Cache() calls with different inputs and `omitArgs = TRUE` plus the
  ## same `.cacheExtra` should resolve to the same cache key, so the second
  ## call hits the cache populated by the first.
  skip_on_cran()

  tmpCache <- file.path(tempdir(), basename(tempfile("rcache_omitTRUE_")))
  dir.create(tmpCache, showWarnings = FALSE, recursive = TRUE)
  withr::local_options(reproducible.cachePath  = tmpCache,
                       reproducible.useMemoise = FALSE,
                       reproducible.verbose    = 0)

  ## Sentinel counter to detect re-evaluation
  e <- new.env()
  e$count <- 0L
  f <- function(x) { e$count <- e$count + 1L; x * 2 }

  out1 <- Cache(f, x = 1, cachePath = tmpCache,
                omitArgs    = TRUE,
                .cacheExtra = "vTag1",
                useCloud    = FALSE)
  out2 <- Cache(f, x = 999, cachePath = tmpCache,
                omitArgs    = TRUE,
                .cacheExtra = "vTag1",
                useCloud    = FALSE)

  expect_equal(e$count, 1L,
               info = "omitArgs = TRUE should make x irrelevant -> 2nd call is a cache hit")
  ## Compare values only; the .Cache attribute differs between miss/hit
  expect_equal(as.numeric(out1), as.numeric(out2),
               info = "Cache hit returns the value computed on the 1st call (x = 1)")

  ## Different .cacheExtra is the only thing that should drive a miss now
  out3 <- Cache(f, x = 1, cachePath = tmpCache,
                omitArgs    = TRUE,
                .cacheExtra = "vTag2",
                useCloud    = FALSE)
  expect_equal(e$count, 2L,
               info = "Changing .cacheExtra invalidates the entry under omitArgs = TRUE")
})

test_that("omitArgs = TRUE still busts the cache when FUN's body changes", {
  ## .FUN is the function VALUE (not just its name), so editing the body
  ## should produce a new digest even though omitArgs = TRUE drops all the
  ## captured args. This is the safety net that prevents stale results when
  ## a developer changes their function under a fixed .cacheExtra.
  skip_on_cran()

  tmpCache <- file.path(tempdir(), basename(tempfile("rcache_omitTRUE_body_")))
  dir.create(tmpCache, showWarnings = FALSE, recursive = TRUE)
  withr::local_options(reproducible.cachePath  = tmpCache,
                       reproducible.useMemoise = FALSE,
                       reproducible.verbose    = 0)

  f <- function(x) x * 2
  v1 <- as.numeric(Cache(f, x = 1, cachePath = tmpCache,
                         omitArgs = TRUE, .cacheExtra = "vTag",
                         useCloud = FALSE))

  ## Edit the function body; same .cacheExtra; expect a new value (cache miss)
  f <- function(x) x * 100
  v2 <- as.numeric(Cache(f, x = 1, cachePath = tmpCache,
                         omitArgs = TRUE, .cacheExtra = "vTag",
                         useCloud = FALSE))
  expect_false(isTRUE(all.equal(v1, v2)),
               info = "Editing FUN's body should bust the cache even under omitArgs = TRUE")
})

test_that("doDigestPrepare(omitArgs = TRUE) keeps only .FUN (and .cacheExtra if set)", {
  ## Direct unit test of the digest-prep helper. With omitArgs = TRUE we
  ## should retain only the special .FUN slot; .cacheExtra is appended after
  ## the omit pass.
  fakeCall <- structure(
    quote(f(a = 1, b = 2)),
    .Cache = list(
      args_w_defaults = list(a = 1, b = 2),
      method          = "f"
    )
  )

  d1 <- reproducible:::doDigestPrepare(fakeCall, omitArgs = TRUE,
                                       .cacheExtra = NULL)
  expect_named(d1, ".FUN")
  expect_equal(d1$.FUN, "f")

  d2 <- reproducible:::doDigestPrepare(fakeCall, omitArgs = TRUE,
                                       .cacheExtra = "v1")
  expect_setequal(names(d2), c(".FUN", ".cacheExtra"))
  expect_equal(d2$.cacheExtra, "v1")

  ## Back-compat: character vector still works
  d3 <- reproducible:::doDigestPrepare(fakeCall, omitArgs = "a",
                                       .cacheExtra = NULL)
  expect_named(d3, c("b", ".FUN"))
})

##########################
test_that("test reproducible.verbose", {
  testInit(verbose = 4, tmpFileExt = ".rds")

  warn <- capture_warnings(Cache(rnorm, 1, cachePath = tmpdir))
  expect_is(.reproEnv$cacheTimings, "data.frame") ##
  colVals <- c("Hashing", "Running rnorm", "Saving to cachePath", "Whole Cache call")
  colNams <- c("functionName", "component", "elapsedTime", "units")
  expect_in(.reproEnv$cacheTimings$component, colVals)
  expect_in(names(.reproEnv$cacheTimings), colNams)

  # Test Path class objects
  a <- sample(1e4)

  saveRDS(a, file = tmpfile)
  out1 <- Cache(readRDS, tmpfile, cachePath = tmpdir)
  nam <- grep("hashDetails", names(.reproEnv), value = TRUE)
  out1Details <- .reproEnv[[nam]]
  ap <- asPath(tmpfile)
  out2 <- Cache(readRDS, ap, cachePath = tmpdir)
  out2Details <- .reproEnv[[nam]]

  # should be vastly larger when actual file, rather than just filename
  expect_true((20 * out1Details$objSize[1]) < out2Details$objSize[1]) ##
})

## Unit tests for the cloud-mode helpers and the validateUseCloud gate. These
## need no Google Drive auth -- they test the small predicates that decide
## whether a cloud listing/download/upload happens for a given useCloud value.

test_that("cloudWrite recognises TRUE and \"push\"", {
  expect_true(reproducible:::cloudWrite(TRUE))
  expect_true(reproducible:::cloudWrite("push"))
  ## legacy ^w prefix retained
  expect_true(reproducible:::cloudWrite("write"))
})

test_that("cloudWrite returns FALSE for FALSE / NULL / \"pull\"", {
  expect_false(reproducible:::cloudWrite(FALSE))
  expect_false(reproducible:::cloudWrite(NULL))
  expect_false(reproducible:::cloudWrite("pull"))
})

test_that("cloudReadOnly recognises \"pull\" only", {
  expect_true(reproducible:::cloudReadOnly("pull"))
  ## legacy ^r prefix retained
  expect_true(reproducible:::cloudReadOnly("read"))
  expect_true(reproducible:::cloudReadOnly("readOnly"))

  expect_false(reproducible:::cloudReadOnly(TRUE))
  expect_false(reproducible:::cloudReadOnly(FALSE))
  expect_false(reproducible:::cloudReadOnly(NULL))
  expect_false(reproducible:::cloudReadOnly("push"))
})

test_that("cloudRead is TRUE for TRUE / \"push\" / \"pull\"; FALSE for FALSE/NULL", {
  expect_true(reproducible:::cloudRead(TRUE))
  expect_true(reproducible:::cloudRead("push"))
  expect_true(reproducible:::cloudRead("pull"))

  expect_false(reproducible:::cloudRead(FALSE))
  expect_false(reproducible:::cloudRead(NULL))
})

test_that("cloudWriteOrRead is TRUE iff cloud is active", {
  expect_true(reproducible:::cloudWriteOrRead(TRUE))
  expect_true(reproducible:::cloudWriteOrRead("push"))
  expect_true(reproducible:::cloudWriteOrRead("pull"))

  expect_false(reproducible:::cloudWriteOrRead(FALSE))
  expect_false(reproducible:::cloudWriteOrRead(NULL))
})

test_that("validateUseCloud accepts the documented forms and rejects others", {
  for (good in list(TRUE, FALSE, NULL, "pull", "push",
                    "write", "read", "readOnly")) {
    expect_silent(reproducible:::validateUseCloud(good))
  }
  for (bad in list("nonsense", "PULL", c("pull", "push"), 1L, NA)) {
    expect_error(reproducible:::validateUseCloud(bad),
                 regexp = "must be TRUE, FALSE, NULL")
  }
})

test_that("Cache() rejects an unknown character useCloud at the front door", {
  ## Construct a tiny Cache() call with a bad useCloud; should error before
  ## any digest / cloud lookup work happens.
  tmpCache <- file.path(tempdir(), basename(tempfile("rcache_")))
  dir.create(tmpCache, showWarnings = FALSE, recursive = TRUE)
  withr::local_options(reproducible.cachePath = tmpCache,
                       reproducible.useMemoise = FALSE)
  expect_error(
    Cache(rnorm, 1, cachePath = tmpCache, useCloud = "yolo"),
    regexp = "must be TRUE, FALSE, NULL"
  )
})

test_that("Cache(useCloud=\"pull\") with a local hit does not consult Google Drive", {
  ## Verify the local-hit short-circuit: when there's already a local cache
  ## entry, a "pull" call must not trigger any googledrive call. We monkey-
  ## patch reproducible:::driveLs to throw if invoked, then run a Cache call
  ## that should be served entirely from local disk.
  tmpCache <- file.path(tempdir(), basename(tempfile("rcache_pull_")))
  dir.create(tmpCache, showWarnings = FALSE, recursive = TRUE)
  withr::local_options(reproducible.cachePath = tmpCache,
                       reproducible.useMemoise = FALSE,
                       reproducible.verbose = 0)

  ## Seed local cache with a deterministic value
  local_cacheId <- "pullModeLocalHit_v1"
  Cache(rnorm, 1, cachePath = tmpCache, cacheId = local_cacheId, useCloud = FALSE)

  ## Sentinel: any call to driveLs should fail loudly
  ns <- asNamespace("reproducible")
  origDriveLs <- ns$driveLs
  poison <- function(...) stop("driveLs should not be called for a pull-mode local hit")
  unlockBinding("driveLs", ns)
  assign("driveLs", poison, envir = ns)
  withr::defer({
    unlockBinding("driveLs", ns)
    assign("driveLs", origDriveLs, envir = ns)
    lockBinding("driveLs", ns)
  })

  ## The second call (same cacheId) must hit local cache without touching driveLs
  expect_silent({
    out <- Cache(rnorm, 1, cachePath = tmpCache, cacheId = local_cacheId,
                 useCloud = "pull",
                 cloudFolderID = "irrelevant",
                 verbose = 0)
  })
  expect_true(is.numeric(out))
})

#' Alternative to `prepInputs` that can use Spatial Tiles stored locally or on Google Drive
#'
#' Downloads, processes and optionally uploads a `SpatRaster` object through a tiling intermediary.
#' If the original `url` is for a very large object, but `to` is a relatively small subset
#' of the area represented by the spatial file at `url`, then this function will
#' potentially by-pass the download of the large file at `url` and instead only download
#' the minimum number of tiles necessary to cover the `to` area. When `doUploads` is
#' TRUE, then this function will potentially create and upload the tiles to `tileFolder`,
#' prior to returning the spatial object, `postProcess`ed to `to`. This function supports
#' both Google Drive and HTTP(S) URLs.
#'
#'
#' @param targetFile Character. Name of the target file to be downloaded or processed.
#'   If missing, it will be inferred from the URL or Google Drive metadata.
#' @param url Character. URL to the full dataset (Google Drive or HTTP/S).
#' @param destinationPath Character. Path to the directory where files will be downloaded and processed.
#' @param to A spatial object (e.g., `SpatRaster`, `SpatVector`, `sf`, or `Spatial*`) defining the area of interest.
#' @param tilesFolder A local file path to put tiles. If this is an absolute path, then
#'   that will be used; if it is a relative path, then it will be
#'   `file.path(destinationPath, tilesFolder)`
#' @param urlTiles Character. URL to the tile source (e.g., Google Drive folder or HTTP/S endpoint). Default is `getOption("reproducible.prepInputsUrlTiles", NULL)`.
#' @param doUploads Logical. Whether to upload processed tiles.
#'   Default is `getOption("reproducible.prepInputsDoUploads", FALSE)`.
#' @param tileGrid Either length 3 character string, such as "CAN", to be sent to `geodata::gadm(...)`
#'   or an actual `SpatVector` object with a grid of polygons
#' @param numTiles Integer. Number of tiles to generate. Optional.
#' @param plot.grid Logical. Whether to plot the tile grid and area of interest. Default is `FALSE`.
#' @param purge Logical or Integer. `0/FALSE` (default) keeps existing `CHECKSUMS.txt` file and
#'   `prepInputs` will write or append to it. `1/TRUE` will deleted the entire `CHECKSUMS.txt` file.
#' @param verbose Logical or numeric. Controls verbosity of messages. Default is `getOption("reproducible.verbose")`.
#' @param ... Either `maskTo`, `cropTo` (which will be used if `to` is not supplied, or
#'   arguments passed to `writeRaster`, e.g., `datatype` (used when writing tiles).
#'
#' @return A single, merged `SpatRaster` object `postProcess`ed to the area of interest (`to`),
#' composed of the necessary tiles.
#' If the post-processed file already exists locally, it will be returned directly.
#'
#' @details
#' This function can be triggered *inside* `prepInputs`
#' if the `to` is supplied and both `url` and `urlTiles` are supplied. **NOTE**:
#' `urlTiles` can be supplied using the
#' `option(reproducible.prepInputsUrlTiles = someGoogleDriveFolderURL`), so the original
#' `prepInputs` function call can remain unaffected.
#'
#' This function also uses a different checksumming procedure compared to the normal `prepInputs`.
#' This function will assess the remote url for a hash. If that hash exists, then
#' it will compare it to a local file with `targetFile` name, suffixed with `.hash`. If the
#' two hashes differ (remote and local), then it will be redownloaded; otherwise the local
#' one will be returned.
#'
#' This function is useful for working with large spatial datasets, but where the user
#' only requires a "relatively small" section of that dataset. This function will
#' potentially bypass the full download and download only the tiles that are necessary
#' for the `to`.
#' It handles downloading only the required tiles based on spatial intersection
#' with the target area, and supports resumable downloads from Google Drive or HTTP/S sources.
#'
#' If `targetFile` is missing, the function attempts to infer it from the URL
#' using the `Content-Disposition` header or the basename of the URL.
#' For Google Drive URLs, it uses the file metadata.
#'
#' @seealso [googledrive::drive_get()], [terra::rast()], [terra::crop()], [terra::merge()]
#'
#' @examples
#'
#' if (FALSE) {
#'   to <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(-123.3656, 48.4284)), crs = 4326))
#'   result <- prepInputsWithTiles(
#'     url = "https://example.com/data.tif",
#'     destinationPath = tempdir(),
#'     to = to,
#'     urlTiles = "https://example.com/tiles/",
#'     tileGrid = "CAN"
#'   )
#' }
#'
#' @export
prepInputsWithTiles <- function(targetFile, url, destinationPath,
                                to,
                                tilesFolder = file.path(getOption("reproducible.inputPath"), "tiles"),
                                urlTiles = getOption("reproducible.prepInputsUrlTiles", NULL),
                                doUploads = getOption("reproducible.prepInputsDoUploads", FALSE),
                                tileGrid = "CAN",
                                numTiles = NULL,
                                plot.grid = FALSE,
                                purge = FALSE,
                                verbose = getOption("reproducible.verbose"), ...) {

  st <- Sys.time()

  env <- environment()

  # deal with `to` first, to identify the tiles, then rest can be Cached easily, even
  #  if the to changes slightly
  maskToCropTo <- c("maskTo", "cropTo")
  whMaskToCropTo <- match(maskToCropTo, ...names())
  if ( (missing(to) && anyNA(whMaskToCropTo)) || is.null(urlTiles)) {
    messagePreProcess(
      "prepInputsWithTiles must have `urlTiles` and `url` plus a `to`, `cropTo` or ",
      "`maskTo` spatial object",
      verbose = verbose)
    return("NULL")
  }
  if (missing(to)) # take the first of maskTo or cropTo, which are in the ...
    if (!anyNA(whMaskToCropTo)) to <- ...elt(whMaskToCropTo[1])
  if (.isSpatRaster(to))
    to <- boundaryPolygon(to)
  dig <- .robustDigest(to)

  datatype <- "FLT4S"
  dtype <- list(...)$datatype
  if (!is.null(dtype))
    datatype <- dtype

  # Preview intersecting tile IDs
  url <- gsub("(?<!:)//+", "/", url, perl = TRUE) # removes double // except in http://
  isGDid <- isGoogleID(url)
  isGDurl <- isGoogleDriveURL(url)
  if (isGDid) {
    url <- googledriveIDtoHumanURL(url)
    isGDurl <- TRUE
  }

  remoteMetadata <- getRemoteMetadata(targetFile, isGDurl, url)
  remoteHashFile <- makeRemoteHashFile(url, destinationPath,
                                       remoteMetadata$targetFile, remoteMetadata$remoteHash)
  if (!is.null(.isArchive(remoteMetadata$targetFile)))  {
    messagePreProcess(
      "prepInputsWithTiles does not work with archives yet",
      verbose = verbose)
    return("NULL")
  }

  if (is.null(remoteMetadata$targetFile)) {
    stop("Please supply `targetFile` or a url from which `targetFile` can be extracted from")
  }

  targetFileFullPath <- file.path(destinationPath, remoteMetadata$targetFile)
  purge <- checkHaveCorrectHashedVersion(targetFileFullPath, remoteHashFile, remoteMetadata, purge, verbose)
  messagePreProcess("Preparing ", .messageFunctionFn(targetFileFullPath), verbose = verbose)
  targetFilePostProcessedFullPath <- .suffix(targetFileFullPath, dig)

  if (isTRUE(purge)) {
    purgeLocals(targetFilePostProcessedFullPath, targetFileFullPath, remoteHashFile, verbose)
  }

  if (file.exists(targetFilePostProcessedFullPath) && doUploads < 1) {
    messagePreProcess("Correct post processed file exists (",
                             .messageFunctionFn(targetFilePostProcessedFullPath),
                             ");\nreturning it now...", verbose = verbose)
    messagePreProcess("prepInputsWithTiles ", gsub("^\b", "", messagePrefixDoneIn),
                      format(difftime(Sys.time(), st), units = "secs", digits = 3),
                      verbose = verbose)
    return(terra::rast(targetFilePostProcessedFullPath))
  }

  tilesFolderFullPath <- file.path(tilesFolder, filePathSansExt(remoteMetadata$targetFile))
  if (fs::is_absolute_path(tilesFolder) %in% FALSE) {
    tilesFolderFullPath <- file.path(destinationPath, tilesFolderFullPath)
  }
  dirTilesFolder <- dir(tilesFolderFullPath, recursive = TRUE, all.files = TRUE)

  if (isTRUE(purge) && length(dirTilesFolder)) {
    dirTilesFolder <- purgeLocalTiles(tilesFolderFullPath, verbose)
  }

  # Need to get target object crs targetObjCRS; first try local file, then local tile,
  #     then gdrive tile, then full remote file
  targetObjCRS <- getTargetCRS(targetFileFullPath, dirTilesFolder, tilesFolderFullPath, remoteMetadata$targetFile,
                               destinationPath = destinationPath,
                           url, urlTiles, remoteMetadata$fileSize, remoteMetadata$remoteHash,
                           remoteAlgorithm = remoteMetadata$remoteAlgorithm,
                           purge, doUploads, verbose)
  # need to rerun because there may have been a rm in previous line
  dirTilesFolder <- dir(tilesFolderFullPath, recursive = TRUE, all.files = TRUE)

  noTiles <- FALSE

  to_inTileGrid <- postProcessTo(to, to = targetObjCRS, verbose = verbose - 2)
  tileGridAndArea <- makeAndPlotTileGrid(tileGrid, numTiles, targetObjCRS,
                                         plot.grid, to = to_inTileGrid, verbose)

  # Find intersecting tiles
  all_tile_names <- sort(makeTileNames(tileGridAndArea$tileGrid$tile_id))

  intersecting_tiles <- terra::intersect(tileGridAndArea$tileGrid, terra::ext(to_inTileGrid))
  needed_tile_names <- makeTileNames(intersecting_tiles$tile_id)
  needed_tile_names <- sort(needed_tile_names)

  missingTilesLocal <- setdiff(needed_tile_names, dirTilesFolder)

  missingTilesLocalAll <- setdiff(all_tile_names, dirTilesFolder)
  tilesToGet <- missingTilesLocal
  haveLocalTiles <- FALSE
    messagePreProcess("Need to load/get these tiles:\n", verbose = verbose) # use message because of line wrap
    if (verbose > 0) {
      messagePreProcess(.messageFunctionFn(paste(needed_tile_names, collapse =  ", ")), verbose = verbose)
  }
  haveAllNeededTiles <- if (doUploads > 0) length(missingTilesLocalAll) == 0 else TRUE

  if (length(missingTilesLocal) == 0) {# && (haveAllNeededTiles)) {
    messagePreProcess(
      "All needed tiles are available locally. Proceeding to load them",
      verbose = verbose)
    haveLocalTiles <- TRUE
  } else {
    messagePreProcess(
      "Tiles are missing locally in:\n",.messageFunctionFn(tilesFolderFullPath),
      "\nWill try to download these:\n", verbose = verbose)
    messagePreProcess(.messageFunctionFn(paste(missingTilesLocal, collapse = ", ")), verbose = verbose)
    messagePreProcess(paste0("... from urlTiles (",.messageFunctionFn(urlTiles),")"), verbose = verbose)
  }

  for (ii in 1:2) { # try twice in case a local tile is corrupt; if yes, delete it, redownload, reload
    if (haveLocalTiles %in% FALSE || doUploads > 0) {
      needed_tile_names <- downloadMakeAndUploadTiles(url, urlTiles, remoteMetadata$targetFile, targetFileFullPath,
                                                      needed_tile_names, tilesToGet, all_tile_names, haveLocalTiles,
                                                      tilesFolderFullPath, tileGridAndArea$tileGrid, tileGridAndArea$numTiles,
                                                      to_inTileGrid, doUploads, datatype, verbose)
    }
    tile_rasters <- rastTiles(needed_tile_names, tilesFolderFullPath)
    if (any(sapply(tile_rasters, is.null))) {
      missingTilesLocal <- setdiff(needed_tile_names, dir(tilesFolderFullPath))
      if (length(missingTilesLocal))
        haveLocalTiles <- FALSE
    } else {
      break
    }
  }
  noData <- FALSE

  if (file.exists(targetFilePostProcessedFullPath)) {
    messagePreProcess("Correct post processed file exists (",
                             .messageFunctionFn(targetFilePostProcessedFullPath),
                      ");\n returning it now...", verbose = verbose)
    return(terra::rast(targetFilePostProcessedFullPath))
  }

  if (noTiles %in% FALSE) {
    rfull <- sprcMosaicRast(url, tile_rasters, to_inTileGrid, targetFilePostProcessedFullPath,
                            remoteMetadata$fileSize, needed_tile_names, tilesFolderFullPath,
                            noData, datatype, verbose)
  }
  messagePreProcess("prepInputsWithTiles ", gsub("^\b", "", messagePrefixDoneIn),
                    format(difftime(Sys.time(), st), units = "secs", digits = 3),
                    verbose = verbose)

  rfull
}

tile_raster_write_auto <- function(raster_path, out_dir, tileGrid, all_tile_names, nx = 10, ny = 5,
                                   datatype = NULL,
                                   verbose = getOption("reproducible.verbose")) {
  r <- terra::rast(raster_path)

  ext <- terra::ext(r)
  x_breaks <- seq(ext[1], ext[2], length.out = nx + 1)
  y_breaks <- seq(ext[3], ext[4], length.out = ny + 1)

  dir.create(out_dir, showWarnings = FALSE, recursive = TRUE)

  # Build tile specs
  tile_specs <- list()
  tile_id <- 1

  for (i in 1:nx) {
    for (j in 1:ny) {
      tile_ext <- terra::ext(x_breaks[i], x_breaks[i + 1], y_breaks[j], y_breaks[j + 1])
      tile_path <- file.path(out_dir, paste0(all_tile_names[tile_id]))
      #                       paste0("tile_", sprintf("%02d", tile_id),
      #                              ".tif"))
      tile_specs[[tile_id]] <- list(ext = tile_ext, path = tile_path)
      tile_id <- tile_id + 1
    }
  }

  # Worker function
  process_tile <- function(spec, datatype) {
    if (!file.exists(spec$path)) {
      tile <- terra::crop(r, spec$ext)
      # isAllNA <- terra::allNA(tile)[1] %in% TRUE
      # if (isAllNA %in% FALSE) {

        terra::writeRaster(tile, spec$path, datatype = datatype,
                           overwrite = FALSE,
                           gdal = c("COMPRESS=LZW", "TILED=YES"))
        return(paste("Saved:", spec$path))
      # }
    } else {
      return(paste("Skipped (already exists):", spec$path))
    }
  }

  messagePreProcess("Creating tiles ...", verbose = verbose)

  # Choose parallel or sequential based on OS
  if (.Platform$OS.type == "unix" && requireNamespace("parallel")) {
    numCoresToUse <- min(getOption("mc.cores"), numCoresToUse(max = length(tile_specs)))
    results <- parallel::mclapply(
      tile_specs, process_tile,
      mc.cores = numCoresToUse, datatype = datatype)
  } else {
    results <- lapply(tile_specs, process_tile, datatype = datatype)
  }

  # Print results
  for (msg in results[!sapply(results, is.null)]) messagePreProcess(msg, verbose = verbose)
  messagePreProcess("Tiling complete.", verbose = verbose)
}

extract_drive_id <- function(url) {
  # Try to match folder ID
  folder_match <- sub(".*?/folders/([a-zA-Z0-9_-]+).*", "\\1", url)
  # Try to match file ID if folder match didn't change the string
  if (identical(folder_match, url)) {
    file_match <- sub(".*?/file/d/([a-zA-Z0-9_-]+).*", "\\1", url)
    return(file_match)
  }
  return(folder_match)
}

upload_tiles_to_drive_url_parallel <- function(local_dir, drive_folder_url, thisFilename,
                                               verbose = getOption("reproducible.verbose")) {
  stopifnot(requireNamespace("googledrive", quietly = TRUE))

  # Extract parent folder ID from URL
  parent_id <- extract_drive_id(drive_folder_url)

  # Create subfolder named after original raster filename
  subfolder_name <- basename(tools::file_path_sans_ext(thisFilename))
  subfolder <- googledrive::with_drive_quiet(
    googledrive::drive_find(q = paste0("name = '", subfolder_name, "' and '", parent_id, "' in parents")))

  if (nrow(subfolder) == 0) {
    subfolder <- googledrive::with_drive_quiet(googledrive::drive_mkdir(subfolder_name,
                                                                        path = googledrive::as_id(parent_id)))
    messagePreProcess("Created subfolder: ", .messageFunctionFn(subfolder_name), verbose = verbose)
  } else {
    messagePreProcess("Found existing subfolder: ", .messageFunctionFn(subfolder_name), verbose = verbose)
  }

  # List local .tif files
  tif_files <- dir(local_dir, pattern = "\\.tif$", full.names = TRUE)

  # Get existing files in Drive subfolder
  existingAll <- googledrive::with_drive_quiet(googledrive::drive_ls(subfolder$id))
  existing_names <- existingAll$name

  # Upload helper
  upload_one <- function(file_path) {
    file_name <- basename(file_path)
    if (!(file_name %in% existing_names)) {
      googledrive::drive_upload(file_path, path = googledrive::as_id(subfolder$id))
      return(paste("Uploaded:", file_name))
    } else {
      return(paste("Skipped (already exists):", file_name))
    }
  }

  # Upload in parallel on Linux/macOS, sequential on Windows
  if (.Platform$OS.type == "unix" && requireNamespace("parallel")) {
    numCoresToUse <- min(getOption("mc.cores"), numCoresToUse(max = 7))
    # numCoresToUse <- numCoresToUse(max = 7) # more than 7 on a fast internet connection
                         # tends to be slower; but this will depend on connection speed
    results <- parallel::mclapply(
      tif_files, upload_one,
      mc.cores = numCoresToUse)
  } else {
    results <- lapply(tif_files, upload_one)
  }

  # Print results
  for (msg in results) messagePreProcess(msg, verbose = verbose)
  messagePreProcess("Upload complete.", verbose = verbose)
}

makeTileGrid <- function(ext, crs, numTiles) {
  stopifnot(
    requireNamespace("sf", quietly = TRUE),
    requireNamespace("terra", quietly = TRUE)
  )

  if (missing(crs)) crs <- proj4stringSCANFI

  # ext <- terra::ext(c(xmin = -2341500, xmax = 3010500, ymin = 5863500, ymax = 9436500))
  areaV <- terra::as.polygons(ext, crs = crs)
  areaGrid <- sf::st_make_grid(sf::st_as_sfc(sf::st_as_sf(areaV)), n = numTiles) |>
    terra::vect()
  m <- t(matrix(seq(prod(numTiles)), nrow = numTiles[[2]], byrow = FALSE))
  areaGrid[["tile_id"]] <- makePaddedNamesForTiles(as.character(m))
  areaGrid
}

proj4stringSCANFI <- "+proj=lcc +lat_0=0 +lon_0=-95 +lat_1=49 +lat_2=77 +x_0=0 +y_0=0 +datum=NAD83 +units=m +no_defs"

makeTileNames <- function(tileIds) {
  paddedTileNumbers <- makePaddedNamesForTiles(tileIds)
  paste0("tile_", paddedTileNumbers, ".tif")
}

rastTiles <- function(tiles, tilesFolderFullPath) {
  tile_rasters <- Map(x = tiles, function(x) {
    a <- try(terra::rast(file.path(tilesFolderFullPath, x)), silent = TRUE)
    if (is(a, "try-error")) {
      a <- rmRastIfTryError(a, tilesFolderFullPath, x)
    }
    a
    })
}


build_lambert_proj4 <- function(ext_obj) {
  if (!inherits(ext_obj, "SpatExtent")) stop("Input must be a terra::ext object")

  # Extract geographic bounds
  lon_min <- ext_obj[1]
  lon_max <- ext_obj[2]
  lat_min <- ext_obj[3]
  lat_max <- ext_obj[4]

  # Central meridian and latitude of origin
  lon_0 <- (lon_min + lon_max) / 2
  lat_0 <- (lat_min + lat_max) / 2

  # Standard parallels: 1/6 from top and bottom
  lat_1 <- lat_min + (lat_max - lat_min) / 6
  lat_2 <- lat_max - (lat_max - lat_min) / 6

  # Build PROJ.4 string
  proj4string <- paste(
    "+proj=lcc",
    paste0("+lat_1=", round(lat_1, 6)),
    paste0("+lat_2=", round(lat_2, 6)),
    paste0("+lat_0=", round(lat_0, 6)),
    paste0("+lon_0=", round(lon_0, 6)),
    "+x_0=0 +y_0=0",
    "+datum=WGS84 +units=m +no_defs"
  )

  return(proj4string)
}

best_square_grid <- function(m, n, min_tiles = 1, max_tiles = 1000) {
  best_diff <- Inf
  best_grid <- NULL

  for (tiles in seq(min_tiles, max_tiles)) {
    for (nx in 1:tiles) {
      ny <- ceiling(tiles / nx)
      tile_w <- m / nx
      tile_h <- n / ny
      aspect_ratio <- tile_w / tile_h
      diff <- abs(log(aspect_ratio))  # closer to 0 = more square

      if (diff < best_diff) {
        best_diff <- diff
        best_grid <- list(
          nx = nx,
          ny = ny,
          tile_width = tile_w,
          tile_height = tile_h,
          total_tiles = nx * ny,
          aspect_ratio = round(aspect_ratio, 3)
        )
      }
    }
  }

  return(best_grid)
}


makeTileGridFromGADMcode <- function(tileGrid, numTiles = NULL, crs) {
  g <- geodata::gadm(tileGrid, resolution = 2) |> Cache()
  if (is.null(g) || (is.character(g) && isTRUE(g == "NULL"))) {
    # most likely geodata server is down
    tileExt <- terra::ext(c(xmin = -2342000, xmax = 3011000, ymin = 5860000, ymax = 9436000))
    tilePoly2 <- tileExt
  } else {

    tilePoly <- {terra::aggregate(g)} |> Cache()

    if (grepl("CAN", substr(tileGrid, 1, 3), ignore.case = TRUE)) {
      vals <- terra::ext(tilePoly)[]
      vals[["ymax"]] <- 70
      tilePoly <- terra::crop(tilePoly, terra::ext(vals))
    }
    if (missing(crs))
      crs <- build_lambert_proj4(terra::ext(tilePoly))
    tilePoly2 <- postProcess(tilePoly, to = crs) |> Cache()
    vals <- terra::ext(tilePoly2)[]
    tileExt <- c(xmin = floor(vals[["xmin"]]/1e3) * 1e3,
                 xmax = ceiling(vals[["xmax"]]/1e3) * 1e3,
                 ymin = floor(vals[["ymin"]]/1e3) * 1e3,
                 ymax = ceiling(vals[["ymax"]]/1e3) * 1e3)
    tileExt <- terra::ext(tileExt)

  }
  if (is.null(numTiles)) {
    bsg <- best_square_grid(m = tileExt[][["xmax"]] - tileExt[][["xmin"]],
                            n = tileExt[][["ymax"]] - tileExt[][["ymin"]]
                            , min_tiles = 100, max_tiles = 200)
    numTiles <- unlist(bsg[c("nx", "ny")])
  }
  tg <- makeTileGrid(tileExt, crs = crs, numTiles = numTiles)
  list(tileGrid = tg, numTiles = numTiles, area = tilePoly2)
}

makePaddedNamesForTiles <- function(tileIds) {
  ncharNeeded <- max(nchar(tileIds))
  sprintf(paste0("%0", ncharNeeded,"d"), as.integer(tileIds))
}

lsExistingTilesOnGoogleDrive <- function(urlTiles, targetFile) {
  urlTilesID <- googledrive::as_id(extract_drive_id(urlTiles))
  tile_folder_onGoogleDrive <- googledrive::drive_get(urlTilesID)
  # targetFile <- "alnu_rub.tif"

  # List all files in the folder
  existing_tiles <- googledrive::with_drive_quiet(googledrive::drive_ls(tile_folder_onGoogleDrive))
  whFolders <- sapply(seq(NROW(existing_tiles)), function(x)
    isGoogleDriveDirectoryFromTibble(existing_tiles[x,]))
  hasSubfolder <- grep(filePathSansExt(targetFile), existing_tiles$name[whFolders])
  if (length(hasSubfolder)) {
    tile_subfolder <- existing_tiles[hasSubfolder, ]$id
    existing_tiles <- googledrive::with_drive_quiet(googledrive::drive_ls(tile_subfolder))
  } else {
    existing_tiles <- NULL
  }
  existing_tiles
}

crsFromLocalTile <- function(tilesFolderFullPath, dirTilesFolder) {
  for (iii in 1:3) { # try a few in case there is a corrupt one
    theFile <- file.path(tilesFolderFullPath, dirTilesFolder[iii])
    targetObjCRS <- tryRastThenGetCRS(theFile)
    if (!is.null(targetObjCRS))
      break
  }
  targetObjCRS
}

crsFromGoogleDriveTile <- function(tilesFolderFullPath, existing_tiles, fileSize, verbose = getOption("reproducible.verbose")) {
  ogwd <- getwd()
  if (dir.exists(tilesFolderFullPath) %in% FALSE)
    dir.create(tilesFolderFullPath, recursive = TRUE, showWarnings = FALSE)
  setwd(tilesFolderFullPath)
  on.exit(setwd(ogwd))
  download_resumable_httr2(existing_tiles$id[1], existing_tiles$name[1],
                           gdriveDetails = existing_tiles[1, ], fileSize, verbose = verbose - 1)
  targetObjCRS <- tryRastThenGetCRS(file.path(tilesFolderFullPath, existing_tiles$name[1]))
  setwd(ogwd)
  targetObjCRS
}

crsFromLocalFile <- function(targetFileFullPath, targetObjCRS) {
  tryRastThenGetCRS(targetFileFullPath)
  # targetObj <- try(terra::rast(targetFileFullPath))
  # if (is(targetObj, "try-error")) {
  #   # unlink(targetFileFullPath, force = TRUE)
  #   message("File appears to be corrupt; deleting it and trying local tiles, then remotes")
  # } else {
  #   targetObjCRS <- terra::crs(targetObj)
  # }
}

getTargetCRS <- function(targetFileFullPath, dirTilesFolder, tilesFolderFullPath,
                         targetFile, destinationPath,
                         url, urlTiles, fileSize, remoteHash,
                         remoteAlgorithm = .classifyRemoteHashAlgo(remoteHash),
                         purge, doUploads, verbose) {

  targetObjCRS <- NULL # don't know it yet
  if (file.exists(targetFileFullPath)) {
    targetObjCRS <- crsFromLocalFile(targetFileFullPath, targetObjCRS)
  }
  # need to get the targetObjCRS to know what the tiles will look like
  if (is.null(targetObjCRS)) {
    targetObjCRS <- crsFromLocalOrGDTiles(targetObjCRS, dirTilesFolder, tilesFolderFullPath,
                                          urlTiles,
                                          targetFile, purge, doUploads, fileSize, verbose)
  }
  if (is.null(targetObjCRS)) {
    # still doesn't have it
    messagePreProcess("Downloading full file (", .messageFunctionFn(targetFile),") from\n", url, verbose = verbose)
    if (!exists("fileSize", inherits = FALSE))
      messageAboutFilesize(fileSize, verbose = verbose)
    download_resumable_httr2(url, targetFileFullPath, fileSize = fileSize)

    # rfull <- terra::rast(targetFileFullPath)
    targetObjCRS <- terra::crs(terra::rast(targetFileFullPath))
  }
  makeRemoteHashFile(url, destinationPath, targetFile, remoteHash,
                     algorithm = remoteAlgorithm, write = TRUE)
  targetObjCRS
}

plotGridAndArea <- function(tileGrid, theArea, to) {
  a <- terra::centroids(tileGrid)
  terra::plot(tileGrid)
  terra::text(a, labels = a$tile_id, col = "blue", cex = 1.2)
  if (is(theArea, "SpatExtent")) {
    theArea <- terra::as.polygons(theArea, crs = to)
  }
  tilePolyTG <- terra::project(theArea, tileGrid)
  terra::plot(tilePolyTG, add = TRUE)
  toForPlot <- if (!any(terra::compareGeom(tileGrid, to))) postProcess(to, terra::crs(tileGrid)) else to
  terra::plot(toForPlot, add = TRUE, col = "red")
}

getTilesFromGoogleDrive <- function(tilesToGet, existing_tiles, tilesFolderFullPath) {
  whGet <- match(tilesToGet, existing_tiles$name)
  tileIDSToGet <- existing_tiles[whGet, ]
  ogwd <- getwd()
  if (dir.exists(tilesFolderFullPath) %in% FALSE)
    dir.create(tilesFolderFullPath, recursive = TRUE, showWarnings = FALSE)
  setwd(tilesFolderFullPath)
  on.exit(setwd(ogwd))
  by(tileIDSToGet, seq_len(NROW(tileIDSToGet)), function(i) {
    download_resumable_httr2(i$id, i$name, gdriveDetails = i,
                             fileSize = as.numeric(i$drive_resource[[1]]$size))
  })
  haveLocalTiles <- TRUE
  setwd(ogwd)
  haveLocalTiles
}

downloadMakeAndUploadTiles <- function(url, urlTiles, targetFile, targetFileFullPath,
                                       needed_tile_names, tilesToGet, all_tile_names, haveLocalTiles,
                                       tilesFolderFullPath, tileGrid, numTiles,
                                       to_inTileGrid, doUploads, datatype, verbose) {
  if (!requireNamespace("terra")) {
    stop("Please install.packages('terra')")
  }

  existing_tiles <- lsExistingTilesOnGoogleDrive(urlTiles, targetFile)

  available_tile_names_onGoogleDrive <- existing_tiles$name

  # Determine which tiles are missing
  missingTilesOnRemote <- setdiff(needed_tile_names, available_tile_names_onGoogleDrive)
  # tilesToGet <- intersect(needed_tile_names, available_tile_names_onGoogleDrive)

  haveRemoteTiles <- all(all_tile_names %in% existing_tiles$name)
  # Preview decision
  needUploads <- TRUE
  doTileDownload <- FALSE
  missingTilesRemoteAll <- setdiff(all_tile_names, existing_tiles$name)

  tilesFullOnRemote <- TRUE
  if (doUploads > 1) tilesFullOnRemote <- length(missingTilesRemoteAll) == 0

  if (length(missingTilesOnRemote) == 0) {
    doTileDownload <- haveLocalTiles %in% FALSE
    messagePreProcess("All needed tiles are available on Google Drive.  ",
                      verbose = verbose)
    needUploads <- tilesFullOnRemote %in% FALSE
    if (doTileDownload) {
      messagePreProcess("Proceeding to download only the needed tiles...", verbose = verbose)
    } else {
      messagePreProcess("Nothing to download", verbose = verbose)
    }
  } else {
    messagePreProcess("Some tiles are missing on Google Drive:")
    missingOnes <- if (doUploads > 1) missingTilesRemoteAll else missingTilesOnRemote
    if (verbose > 0) message(paste(missingOnes, collapse = ", "))
  }

  if (needUploads && length(missingTilesOnRemote) == 0) {
    messagePreProcess("Some 'unneeded' tiles are missing, but doUploads is TRUE and local tiles exist: ",
                      "uploading: ", verbose = verbose)
    messagePreProcess(.messageFunctionFn(paste(missingTilesRemoteAll, collapse = ", ")), verbose = verbose)
  }

  if (haveLocalTiles %in% FALSE && doTileDownload %in% TRUE) {
    haveLocalTiles <- getTilesFromGoogleDrive(tilesToGet, existing_tiles, tilesFolderFullPath)
  }

  if (needUploads %in% TRUE || (doUploads > 0 && haveRemoteTiles %in% FALSE)) {
    fe <- file.exists(targetFileFullPath)
    if (fe %in% FALSE)
      download_resumable_httr2(url, targetFileFullPath)


    if (haveLocalTiles %in% FALSE || (doUploads > 0 && needUploads))
      tile_raster_write_auto(targetFileFullPath, tilesFolderFullPath, tileGrid,
                             all_tile_names = all_tile_names, datatype = datatype,
                             nx = numTiles[[1]], ny = numTiles[[2]],
                             verbose = verbose)
    if (needUploads %in% FALSE && doUploads > 0)
      messagePreProcess("Nothing to upload", verbose = verbose)

    upload_tiles_to_drive_url_parallel(tilesFolderFullPath, urlTiles, targetFileFullPath,
                                       verbose = verbose)
    tile_paths <- dir(tilesFolderFullPath, pattern = "\\.tif$")
    saExt <- terra::ext(to_inTileGrid)

    # Filter tiles that intersect the study area
    keep_idx <- vapply(tile_paths, function(path) {
      tile_ext <- terra::ext(terra::rast(file.path(tilesFolderFullPath, path)))

      # bounding box overlap (x then y)
      !(tile_ext[1] > saExt[2] || tile_ext[2] < saExt[1] ||  # x no-overlap
          tile_ext[3] > saExt[4] || tile_ext[4] < saExt[3])    # y no-overlap
    }, logical(1))

    intersecting_tiles2 <- tile_paths[keep_idx]

    # intersecting_tiles2 <- purrr::keep(tile_paths, function(path) {
    #   tile_ext <- terra::ext(terra::rast(file.path(tilesFolderFullPath, path)))
    #
    #   # Check for bounding box overlap
    #   !(tile_ext[1] > saExt[2] || tile_ext[2] < saExt[1] ||  # x overlap
    #       tile_ext[3] > saExt[4] || tile_ext[4] < saExt[3])    # y overlap
    # })
    if (!identical(needed_tile_names, intersecting_tiles2)) {
      messagePreProcess("`to` does not overlap with any tiles on file at:\n",
                        .messageFunctionFn(url), verbose = verbose)
      # the intersecting_tiles2 from the newly created need to be the same as the
      # expected from the grid
    }

  }
  needed_tile_names
}


messageAboutFilesizeCompare <- function(fileSize, needed_tile_names,
                                        targetFilePostProcessedFullPath,  tilesFolderFullPath,
                                        verbose) {
  # fileSize <- file$drive_resource[[1]]$size
  messageAboutFilesize(fileSize, verbose = verbose, msgMiddle = " on remote url ")
  fsLocal <- file.size(targetFilePostProcessedFullPath)
  dd1 <- dir(tilesFolderFullPath)
  dd2 <- dir(tilesFolderFullPath, full.names = TRUE)
  tilesUsed <- dd2[match(needed_tile_names, dd1)]
  messageAboutFilesize(file.size(tilesUsed), verbose = verbose, msgMiddle = " on local drive using tiles ")
}



tryRastThenGetCRS <- function(targetFileFullPath) {
  targetObj <- try(terra::rast(targetFileFullPath))
  if (is(targetObj, "try-error")) {
    rmRastIfTryError(targetObj, dirname(targetFileFullPath), basename(targetFileFullPath))
    # unlink(targetFileFullPath, force = TRUE)
    targetObjCRS <- NULL
    # message("File (", targetFileFullPath, ") appears to be corrupt")#; deleting it and trying local tiles, then remotes")
  } else {
    targetObjCRS <- terra::crs(targetObj)
  }
  targetObjCRS
}



#' Estimate Number of CPU Cores to Use for Parallel Processing
#'
#' This function estimates the number of CPU cores that can be safely used for
#' parallel processing, taking into account a minimum threshold, the total
#' number of physical cores, and currently active threads.
#'
#' @param min An integer specifying the minimum number of cores to use. Default is `2`.
#' @param max An integer specifying the maximum number of cores available,
#'   typically the number of physical cores. Default is
#'   `max(1L, getOption("Ncpus", 1L), parallel::detectCores() - 1, logical = FALSE, na.rm = TRUE)`.
#'
#' @return An integer representing the number of cores that can be used for
#'   parallel tasks, ensuring at least `min` cores are used, while subtracting
#'   one for the current process and an estimate of actively used threads (via
#'   `detectActiveCores()`).
#'
#' @examples
#' if (FALSE) {
#'   numCoresToUse()
#'   numCoresToUse(min = 4)
#' }
#'
#' @note This function depends on `detectActiveCores()` and is not supported on
#'   Windows systems.
#'
#' @export
#' @seealso [detectActiveCores()]
numCoresToUse <- function(min = 2, max = NULL) {
  if (.requireNamespace("parallelly")) {
    nctu <- max(min, min(max, parallelly::freeCores()))
    return(nctu)
  }
  # if (is.null(.pkgEnv$detectedCores)) {
  #   ## see <https://parallelly.futureverse.org/#availablecores-vs-paralleldetectcores>
  #   .pkgEnv$detectedCores <- max(1L, getOption("Ncpus", 1L), parallel::detectCores() - 1,
  #                                logical = FALSE, na.rm = TRUE)
  # }
  # dc <- .pkgEnv$detectedCores
  # if (is.null(max)) {
  #   max <- dc
  # }
  # max <- min(dc -  # total
  #              1 - # remove one for the current process
  #              detectActiveCores(), # estimate actively used ones
  #            max)
  # max(min, max)
}

# Classify a remote-supplied hash string into a content-hash algorithm or
# "etag-opaque" when no positive trust is possible. Google Drive ETag-shaped
# strings are forced to "md5" via the isGDurl override at the call site.
#
# Heuristic:
#   ^[0-9a-f]{32}$ -> md5   (most common content-hash ETag; Google Drive)
#   ^[0-9a-f]{40}$ -> sha1
#   ^[0-9a-f]{64}$ -> sha256
#   anything else  -> "etag-opaque" (weak ETags, server-derived, unknown)
.classifyRemoteHashAlgo <- function(hash, isGDurl = FALSE) {
  if (isTRUE(isGDurl)) return("md5")
  if (is.null(hash) || is.na(hash) || !is.character(hash) || !nzchar(hash))
    return("etag-opaque")
  hl <- tolower(hash)
  if (grepl("^[0-9a-f]{32}$", hl)) return("md5")
  if (grepl("^[0-9a-f]{40}$", hl)) return("sha1")
  if (grepl("^[0-9a-f]{64}$", hl)) return("sha256")
  "etag-opaque"
}

# Parse the contents of a `.hash` sidecar file. Format is `<algo>:<hash>`
# (one line). For backward compatibility with legacy single-hash sidecars
# written before this format change, fall back to inferring `algo` from the
# hash length (32/40/64 hex => md5/sha1/sha256).
#
# Returns list(algorithm = ..., hash = ...) or NULL on read error.
.parseRemoteHashFile <- function(remoteHashFile) {
  if (!file.exists(remoteHashFile)) return(NULL)
  txt <- try(readLines(remoteHashFile, warn = FALSE), silent = TRUE)
  if (is(txt, "try-error") || !length(txt)) return(NULL)
  line <- txt[[1L]]
  if (grepl(":", line, fixed = TRUE)) {
    parts <- strsplit(line, ":", fixed = TRUE)[[1L]]
    if (length(parts) >= 2L)
      return(list(algorithm = parts[[1L]],
                  hash      = paste(parts[-1L], collapse = ":")))
  }
  # Legacy single-hash sidecar: infer algorithm from hash length.
  list(algorithm = .classifyRemoteHashAlgo(line), hash = line)
}

makeRemoteHashFile <- function(url, destinationPath, targetFile, remoteHash,
                               algorithm = NULL, write = FALSE) {
  url_no_protocol <- sub("^https?://", "", url)
  # Replace all slashes with underscores
  urlWithUnderscores <- gsub("/", "_", file.path(basename(targetFile), dirname(url_no_protocol)))
  # Hide the sidecar with a leading "." so it doesn't show up in dir() listings
  # (e.g. test patterns like `dir(tmpdir, pattern = "Shapefile")` would otherwise
  # match `Shapefile1.zip_drive.google.com_..._.hash` and inflate file counts).
  remoteHashFile <- file.path(destinationPath, paste0(".", urlWithUnderscores, ".hash"))
  if (isTRUE(write) && !file.exists(remoteHashFile)) {
    if (is.null(algorithm) || is.na(algorithm) || !nzchar(algorithm)) {
      # Legacy callers: write hash-only line.
      writeLines(remoteHash, remoteHashFile)
    } else {
      writeLines(paste0(algorithm, ":", remoteHash), remoteHashFile)
    }
  }
  return(remoteHashFile)
}



checkHaveCorrectHashedVersion <- function(targetFile, remoteHashFile, remoteMetadata, purge, verbose) {
  haveCorrectVersion <- FALSE
  askAboutPurge <- FALSE
  fe <- file.exists(remoteHashFile)
  # But still could be incomplete
  if (isTRUE(fe)) {
    parsed <- .parseRemoteHashFile(remoteHashFile)
    haveCorrectVersion <- !is.null(parsed) &&
      identical(parsed$hash, remoteMetadata$remoteHash)
    if (haveCorrectVersion %in% FALSE) {
      askAboutPurge <- TRUE
    } else {
      if (file.exists(targetFile)) {
        if (!identical(file.size(targetFile), as.numeric(remoteMetadata$fileSize) )) {
          askAboutPurge <- TRUE
        }
      }
    }
  }
  if (isTRUE(askAboutPurge)) {
    message("The local version is not the version that matches the remote version")
    message("Do you want to purge all local data and redownload? Y or N")
    yorn <- readline(" ")
    yorn <- substr(tolower(yorn), 1, 1)
    if (identical("y", yorn))
      purge <- TRUE

  }
  if (!purge %in% TRUE)
    messagePreProcess("Local files match the current remote file version; proceeding",
                      verbose = verbose)

  purge
}

getRemoteMetadata <- function(targetFile, isGDurl, url) {
  # browser()
  if (missing(isGDurl))
    isGDurl <- isGoogleDriveURL(url) || isGoogleID(url)
  if (isGDurl) {
    # Cache indefinitely — same rationale as assessGoogle().
    file <- googledrive::drive_get(url) |>
      Cache(verbose = FALSE)
    fileSize <- file$drive_resource[[1]]$size
    remoteHash <- file$drive_resource[[1]]$md5Checksum
    if (missing(targetFile)) {
      targetFile <- file$name
    }
    timestampOnline <- file$drive_resource[[1]]$modifiedTime
  }

  if (missing(targetFile)) {
    response <- httr2::request(url) |> httr2::req_method("HEAD") |> httr2::req_perform()
    remoteHash <- httr2::resp_headers(response)[["etag"]] |>
      gsub(pattern = "^\"|\"", replacement = "")

    content_disposition <- httr2::resp_header(response, "content-disposition")
    fileSize <- httr2::resp_header(response, "content-length") |> as.numeric()
    timestampOnline <- httr2::resp_header(response, "Date")
    if (isTRUE(!(is.na(content_disposition)))) {
      targetFile <- sub('.*filename="([^"]+)".*', '\\1', content_disposition)
    } else {
      # Fallback: extract from URL
      targetFile <- basename(url)
    }
  }
  remoteAlgorithm <- .classifyRemoteHashAlgo(remoteHash, isGDurl = isGDurl)
  list(targetFile = targetFile, fileSize = fileSize, remoteHash = remoteHash,
       remoteAlgorithm = remoteAlgorithm, timestampOnline = timestampOnline)
}

sprcMosaicRast <- function(url, tile_rasters, to_inTileGrid, targetFilePostProcessedFullPath,
                           fileSize, needed_tile_names, tilesFolderFullPath, noData,
                           datatype, verbose) {
  allNull <- all(sapply(tile_rasters, is.null))
  if (allNull %in% FALSE) {
    anyNull <- any(sapply(tile_rasters, is.null))
    if (anyNull) {
      stop("For unknown reasons, the tiles are not available")
    }
    mosaic_raster <- terra::sprc(tile_rasters)
    # mosaic_raster <- terra::vrt(mosaic_raster)
    intersects <- terra::intersect(terra::ext(mosaic_raster), terra::ext(to_inTileGrid))
    if (!is.null(intersects)) {
      messagePrepInputs("cropping ... ", verbose = verbose)
      st1 <- Sys.time()
      final <- terra::crop(mosaic_raster, to_inTileGrid)
      messagePreProcess("  ", gsub("^\b", "", messagePrefixDoneIn),
                        format(difftime(Sys.time(), st1), units = "secs", digits = 3),
                        verbose = verbose)

      st3 <- Sys.time()
      messagePrepInputs("merging tiles in ", .messageFunctionFn(tilesFolderFullPath), " ...", verbose = verbose)
      rfull <- terra::merge(final)
      messagePreProcess("  ", gsub("^\b", "", messagePrefixDoneIn),
                        format(difftime(Sys.time(), st3), units = "secs", digits = 3),
                        verbose = verbose)

      st2 <- Sys.time()
      # messagePrepInputs("writing ", .messageFunctionFn(targetFilePostProcessedFullPath), " ...", verbose = verbose)
      # rfull <- terra::writeRaster(merged, filename = targetFilePostProcessedFullPath,
      #                             datatype = datatype,
      #                             overwrite = TRUE)
      messagePreProcess("  ", gsub("^\b", "", messagePrefixDoneIn),
                        format(difftime(Sys.time(), st2), units = "secs", digits = 3),
                        verbose = verbose)

      if (exists("fileSize", inherits = FALSE)) {
        messageAboutFilesizeCompare(fileSize, needed_tile_names,
                                    targetFilePostProcessedFullPath,  tilesFolderFullPath,
                                    verbose)
      }
    } else {
      noData <- TRUE
    }
  } else {
    noData <- TRUE
  }
  if (isTRUE(noData)) {
    messagePreProcess("The dataset at \n", url, "\ndoes not have data that overlaps with `to`",
                      verbose = verbose)
    rfull <- NULL
  }
  rfull
}

purgeLocals <- function(targetFilePostProcessedFullPath, targetFileFullPath, remoteHashFile, verbose) {
  messagePreProcess("purge = TRUE; purging local targetFile", verbose = verbose)
  if (file.exists(targetFilePostProcessedFullPath))
    unlink(targetFilePostProcessedFullPath)
  if (file.exists(targetFileFullPath))
    unlink(targetFileFullPath)
  if (file.exists(remoteHashFile))
    unlink(remoteHashFile)
}

purgeLocalTiles <- function(tilesFolderFullPath, verbose) {
  messagePreProcess("purge = TRUE; purging local tiles", verbose = verbose)
  dirTilesFolder2 <- dir(tilesFolderFullPath, recursive = TRUE, all.files = TRUE, full.names = TRUE)
  unlink(dirTilesFolder2)
  dirTilesFolder <- NULL
  dirTilesFolder
}

purgeGoogleTiles <- function(urlTiles, targetFile, verbose) {
  messagePreProcess("purging GoogleDrive tiles...", verbose = verbose)
  folderID <- googledrive::drive_ls(googledrive::as_id(extract_drive_id(urlTiles)),
                                    pattern = filePathSansExt(targetFile))
  googledrive::drive_rm(folderID)
  existing_tiles <- NULL
  existing_tiles
}


rmRastIfTryError <- function(obj, tilesFolderFullPath, x) {
  if (any(grepl("cannot open this file as a SpatRaster", obj))) {
    fn <- file.path(tilesFolderFullPath, x)
    message("Tile ", fn, " appears to be corrupt; deleting and redownloading")
    unlink(fn)
    obj <- NULL
  }
  obj
}



crsFromLocalOrGDTiles <- function(targetObjCRS, dirTilesFolder, tilesFolderFullPath,
                                  urlTiles, targetFile, purge, doUploads, fileSize, verbose) {
  existing_tiles <- NULL
  for (i in 1:2) { # try local file, then googledrive, then back to local after googledrive download
    if (length(dirTilesFolder))  {
      targetObjCRS <- crsFromLocalTile(tilesFolderFullPath, dirTilesFolder)
      if (!is.null(targetObjCRS)) break
    }
    if (is.null(targetObjCRS) && is.null(existing_tiles)) {
      existing_tiles <- lsExistingTilesOnGoogleDrive(urlTiles, targetFile)
      if (!is.null(existing_tiles) && NROW(existing_tiles) > 0) {
        if (isTRUE(purge) && doUploads > 1) {
          existing_tiles <- purgeGoogleTiles(urlTiles, targetFile, verbose)
        } else {
          targetObjCRS <- crsFromGoogleDriveTile(tilesFolderFullPath, existing_tiles, fileSize, verbose = verbose)
        }
      }
    }
    if (is.null(targetObjCRS)) {
      dirTilesFolder <- dir(tilesFolderFullPath, recursive = TRUE, all.files = TRUE)
      if ( (is.null(existing_tiles) || NROW(existing_tiles) == 0) &&
           length(dirTilesFolder) == 0)
        break
    } else {
      break
    }
  }
  targetObjCRS
}

makeAndPlotTileGrid <- function(tileGrid, numTiles, targetObjCRS, plot.grid, to, verbose) {
  if (is.character(tileGrid)) {
    tg <- makeTileGridFromGADMcode(tileGrid, numTiles, crs = targetObjCRS) |>
      Cache(verbose = verbose - 1)
    tileGrid <- tg$tileGrid
    numTiles <- tg$numTiles
    theArea <- tg$area
  } else {
    stop("tileGrid must be a character string")
  }
  # if (missing(theArea)) {
  #   theArea <- terra::ext(tileGrid)
  # }
  if (isTRUE(plot.grid) && !missing(to)) {
    plotGridAndArea(tileGrid, theArea, to)
  }
  list(tileGrid = tileGrid, numTiles = numTiles)
}


# Get resolution
boundaryPolygon <- function(r) {
  res_x <- terra::res(r)[1]
  res_y <- terra::res(r)[2]

  # Get raster extent
  ext <- terra::ext(r)

  # Generate coordinates of pixel corners along the boundary
  # Top edge (left to right)
  top <- cbind(seq(ext[1], ext[2] - res_x, by = res_x), rep(ext[4], ncol(r)))

  # Right edge (top to bottom)
  right <- cbind(rep(ext[2], nrow(r)), seq(ext[4], ext[3] + res_y, by = -res_y))

  # Bottom edge (right to left)
  bottom <- cbind(seq(ext[2], ext[1] + res_x, by = -res_x), rep(ext[3], ncol(r)))

  # Left edge (bottom to top)
  left <- cbind(rep(ext[1], nrow(r)), seq(ext[3], ext[4] - res_y, by = res_y))

  # Combine all edges into one closed polygon
  boundary_coords <- rbind(top, right, bottom, left, top[1, , drop = FALSE])

  # Create polygon
  terra::vect(list(boundary_coords), type = "polygons", crs = terra::crs(r))
}

# Tests for `reproducible.destinationPathShared` — the read-write shared local cache
# used by prepInputs() / preProcess() to avoid re-downloads across projects.
#
# Test matrix lives in dev/destinationPathShared-design.md (§5). Case IDs (P1, H1, ...)
# in this file map 1:1 to that matrix. Tests for behaviour that has not been
# implemented in this branch have been removed (rather than left as `skip()`
# stubs that misleadingly pad the count) — see `dev/destinationPathShared-design.md` §4
# for the rolling status of each design item.
#
# Note: the option is now named `reproducible.destinationPathShared` (matching the
# `prepInputs` naming family); `reproducible.inputPaths` remains as a
# backwards-compatible alias. The intermediate `reproducible.dataPath` name
# briefly used in this branch was dropped before release — case IDs that
# referenced it have been retired.
#
# Conventions:
#   - No real network. Fixtures use file:// URLs pointing at on-disk files.
#   - CHECKSUMS.txt is hand-written into the shared dir to exercise the
#     existing checksum-driven local-source path.
#   - withr::local_options() isolates option state per test_that() block.
#   - Internal helpers reached via getFromNamespace() so tests load on
#     covr / R CMD check / source-loaded sessions.

# ---------------------------------------------------------------------------
# Fixture helpers
# ---------------------------------------------------------------------------

# Build a tiny on-disk fixture with a known content/hash. Creates the dir
# if it doesn't exist. Returns list(path, hash, size, url).
makeFixture <- function(dir, name = "foo.csv", content = "a,b\n1,2\n3,4\n") {
  if (!dir.exists(dir)) dir.create(dir, recursive = TRUE)
  p <- file.path(dir, name)
  writeBin(charToRaw(content), p)        # avoid line-ending mangling
  list(
    path = normPath(p),
    hash = digest::digest(file = p, algo = "xxhash64"),
    size = file.info(p)$size,
    url  = toFileUrl(p)
  )
}

# Write a CHECKSUMS.txt covering a single fixture into a directory.
# Format matches what `Checksums(write = TRUE)` produces.
writeChecksumsFor <- function(dir, fixture, name = basename(fixture$path)) {
  csf <- file.path(dir, "CHECKSUMS.txt")
  hdr <- '"file" "checksum" "filesize" "algorithm"'
  row <- sprintf('"%s" "%s" "%d" "xxhash64"', name, fixture$hash, fixture$size)
  writeLines(c(hdr, row), csf)
  csf
}

# Inode of a file, as a string. POSIX-only; returns NA on failure or
# non-POSIX OS. Kept as character (not integer) because inode numbers can
# exceed R's .Machine$integer.max (2^31-1) on large filesystems, where
# as.integer() would silently coerce to NA. We only ever test equality.
# GNU stat (Linux) uses `-c %i`; BSD stat (macOS) uses `-f %i`.
inoOf <- function(p) {
  if (.Platform$OS.type != "unix") return(NA_character_)
  args <- if (Sys.info()[["sysname"]] == "Darwin") c("-f", "%i") else c("-c", "%i")
  out <- suppressWarnings(
    system2("stat", c(args, shQuote(p)), stdout = TRUE, stderr = FALSE)
  )
  if (length(out) == 0L || !grepl("^[0-9]+$", out[[1L]])) return(NA_character_)
  out[[1L]]
}

# Two paths refer to the same physical file (hardlinked) iff inodes match
# AND both look like real files. Returns NA when we can't tell (Windows).
sameInode <- function(p1, p2) {
  i1 <- inoOf(p1); i2 <- inoOf(p2)
  if (is.na(i1) || is.na(i2)) return(NA)
  identical(i1, i2)
}

getInternalOrNull <- function(name) {
  tryCatch(getFromNamespace(name, "reproducible"), error = function(e) NULL)
}
`%||%` <- function(a, b) if (!is.null(a)) a else b

# Build a portable file:// URL. RFC 8089 says file:///<path>, with empty
# host. On POSIX, normPath returns "/home/x" → "file:///home/x".
# On Windows, normPath returns "C:\\path" or "C:/path" → "file:///C:/path".
toFileUrl <- function(p) {
  p <- gsub("\\\\", "/", normPath(p))
  if (substr(p, 1L, 1L) != "/") p <- paste0("/", p)   # Windows drive letter
  paste0("file://", p)
}

getDestinationPathSharedFn <- function() {
  getInternalOrNull(".getDestinationPathShared") %||% getInternalOrNull(".getDataPath")
}
getDestinationPathSharedRecursiveFn <- function() {
  getInternalOrNull(".getDestinationPathSharedRecursive") %||%
    getInternalOrNull(".getDataPathRecursive")
}

# Build a minimal pp_* ctx matching what .pp_make_ctx produces, with just the
# fields pp_remote_hash_check reads/writes. Keeps tests self-contained without
# invoking the full preProcess() pipeline.
makePPCtx <- function(url, destinationPath, neededFiles = NULL, archive = NULL) {
  identifyCS <- getFromNamespace("identifyCHECKSUMStxtFile", "reproducible")
  emptyCS    <- getFromNamespace(".emptyChecksumsResult", "reproducible")
  list(
    url = url, archive = archive, neededFiles = neededFiles,
    destinationPath = destinationPath,
    checkSumFilePath = identifyCS(destinationPath),
    checkSums = emptyCS, needChecksums = 0L,
    skipDownload = FALSE, hashVerified = character(),
    remoteMetadata = NULL, verbose = 0L
  )
}

readSidecar <- function(dir, fixture, url) {
  mk <- getFromNamespace("makeRemoteHashFile", "reproducible")
  p  <- mk(url, dir, basename(fixture$path), "")
  if (!file.exists(p)) return(NULL)
  list(path = p, contents = readLines(p, warn = FALSE))
}

readChecksumsRows <- function(dir) {
  csf <- file.path(dir, "CHECKSUMS.txt")
  if (!file.exists(csf) || file.size(csf) == 0L) return(NULL)
  read.table(csf, header = TRUE, stringsAsFactors = FALSE)
}


# ===========================================================================
# §5.1  Option plumbing (no I/O)
# ===========================================================================

test_that("P1: option unset → getter returns NULL", {
  testInit()
  withr::local_options(list(
    reproducible.destinationPathShared = NULL,
    reproducible.inputPaths   = NULL
  ))
  fn <- getDestinationPathSharedFn()
  skip_if(is.null(fn), "no destinationPathShared getter available")
  expect_null(fn())
})

test_that("P2: destinationPathShared = '/x' → '/x'", {
  testInit()
  fn <- getInternalOrNull(".getDestinationPathShared")
  skip_if(is.null(fn), "no destinationPathShared getter available")
  withr::local_options(list(reproducible.destinationPathShared = "/x"))
  expect_identical(fn(), "/x")
})

test_that("P3: vector value preserved", {
  testInit()
  fn <- getInternalOrNull(".getDestinationPathShared")
  skip_if(is.null(fn), "no destinationPathShared getter available")
  withr::local_options(list(reproducible.destinationPathShared = c("/a", "/b")))
  expect_identical(fn(), c("/a", "/b"))
})

test_that("P5: only inputPaths set → returned + deprecation message", {
  testInit()
  fn <- getDestinationPathSharedFn()
  skip_if(is.null(fn), "no getter available")
  withr::local_options(list(
    reproducible.destinationPathShared = NULL,
    reproducible.inputPaths   = "/x"
  ))
  # Deprecation messaging is one-shot per session; reset the registry so
  # this test sees the message regardless of test ordering.
  pkgEnv <- getInternalOrNull(".pkgEnv")
  if (!is.null(pkgEnv)) pkgEnv$.deprecMsgEmitted <- character()
  msgs <- testthat::capture_messages(out <- fn())
  expect_identical(out, "/x")
  expect_true(any(grepl("deprecated", msgs)))
})

test_that("P6: destinationPathShared and inputPaths both set → destinationPathShared wins, no deprecation", {
  testInit()
  fn <- getInternalOrNull(".getDestinationPathShared")
  skip_if(is.null(fn), "no destinationPathShared getter available")
  withr::local_options(list(
    reproducible.destinationPathShared = "/new",
    reproducible.inputPaths   = "/old"
  ))
  msgs <- testthat::capture_messages(out <- fn())
  expect_identical(out, "/new")
  expect_false(any(grepl("deprecated", msgs)))
})

test_that("P11: recursive option unset → FALSE", {
  testInit()
  fn <- getDestinationPathSharedRecursiveFn()
  skip_if(is.null(fn), "no recursive getter available")
  withr::local_options(list(
    reproducible.destinationPathSharedRecursive = NULL,
    reproducible.inputPathsRecursive   = NULL
  ))
  expect_false(fn())
})

test_that("P12: destinationPathSharedRecursive = TRUE → TRUE", {
  testInit()
  skip_if_not(
    !is.null(getInternalOrNull(".getDestinationPathSharedRecursive")),
    "recursive option not yet wired (§4 step 5)"
  )
  withr::local_options(list(reproducible.destinationPathSharedRecursive = TRUE))
  expect_true(getInternalOrNull(".getDestinationPathSharedRecursive")())
})

test_that("P13: only inputPathsRecursive = TRUE → TRUE + deprecation", {
  testInit()
  fn <- getDestinationPathSharedRecursiveFn()
  skip_if(is.null(fn), "no recursive getter available")
  withr::local_options(list(
    reproducible.destinationPathSharedRecursive = NULL,
    reproducible.inputPathsRecursive   = TRUE
  ))
  # Reset one-shot deprecation registry — see P5 comment.
  pkgEnv <- getInternalOrNull(".pkgEnv")
  if (!is.null(pkgEnv)) pkgEnv$.deprecMsgEmitted <- character()
  msgs <- testthat::capture_messages(out <- fn())
  expect_true(out)
  expect_true(any(grepl("deprecated", msgs)))
})


# ===========================================================================
# §5.2  Resolution — happy paths
# ===========================================================================
#
# Strategy: file:// URL fixtures + a hand-written CHECKSUMS.txt in the shared
# dir, so the *current* code path (.checkLocalSources) recognizes the file
# and hardlinks it into destinationPath. Inode comparison on POSIX is the
# strongest assertion that no copy/download happened.

test_that("H1: file in destinationPathShared with matching CHECKSUMS → hardlinked, no copy", {
  testInit("digest")
  shared <- normPath(file.path(tmpdir, "shared"))
  dest   <- normPath(file.path(tmpdir, "dest"))
  src    <- normPath(file.path(tmpdir, "src"))
  for (d in c(shared, dest, src)) dir.create(d, recursive = TRUE)

  fixSrc    <- makeFixture(src,    "foo.csv")          # the "remote"
  fixShared <- makeFixture(shared, "foo.csv")          # same content → same hash
  writeChecksumsFor(shared, fixShared)

  withr::local_options(list(
    reproducible.inputPaths          = shared,
    reproducible.inputPathsRecursive = FALSE
  ))

  capture.output(
    preProcess(url = fixSrc$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1),
    type = "output"
  )

  expect_true(file.exists(file.path(dest, "foo.csv")))
  skip_on_os("windows")
  expect_true(isTRUE(sameInode(file.path(dest, "foo.csv"), fixShared$path)),
               info = "destinationPath copy should share inode with destinationPathShared (hardlink)")
})

test_that("H3: file in subdir, recursive=FALSE → not found, falls through", {
  testInit("digest")
  shared <- normPath(file.path(tmpdir, "shared"))
  sub    <- normPath(file.path(shared, "sub"))
  dest   <- normPath(file.path(tmpdir, "dest"))
  src    <- normPath(file.path(tmpdir, "src"))
  for (d in c(shared, sub, dest, src)) dir.create(d, recursive = TRUE)

  fixSrc    <- makeFixture(src, "foo.csv")
  fixShared <- makeFixture(sub, "foo.csv")
  writeChecksumsFor(sub, fixShared)

  withr::local_options(list(
    reproducible.inputPaths          = shared,    # ← search top-level only
    reproducible.inputPathsRecursive = FALSE
  ))

  capture.output(
    preProcess(url = fixSrc$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1),
    type = "output"
  )

  expect_true(file.exists(file.path(dest, "foo.csv")))
  skip_on_os("windows")
  expect_false(isTRUE(sameInode(file.path(dest, "foo.csv"), fixShared$path)),
               info = "non-recursive search must not link from subdir")
})

test_that("H5: two destinationPathShared entries; file only in second → found", {
  testInit("digest")
  s1   <- normPath(file.path(tmpdir, "s1"))
  s2   <- normPath(file.path(tmpdir, "s2"))
  dest <- normPath(file.path(tmpdir, "dest"))
  src  <- normPath(file.path(tmpdir, "src"))
  for (d in c(s1, s2, dest, src)) dir.create(d, recursive = TRUE)

  fixSrc    <- makeFixture(src, "foo.csv")
  fixShared <- makeFixture(s2,  "foo.csv")
  writeChecksumsFor(s2, fixShared)

  withr::local_options(list(reproducible.inputPaths = c(s1, s2)))

  capture.output(
    preProcess(url = fixSrc$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1),
    type = "output"
  )

  skip_on_os("windows")
  expect_true(isTRUE(sameInode(file.path(dest, "foo.csv"), fixShared$path)))
})

test_that("H6: file in both destinationPathShared entries → first wins", {
  testInit("digest")
  s1   <- normPath(file.path(tmpdir, "s1"))
  s2   <- normPath(file.path(tmpdir, "s2"))
  dest <- normPath(file.path(tmpdir, "dest"))
  src  <- normPath(file.path(tmpdir, "src"))
  for (d in c(s1, s2, dest, src)) dir.create(d, recursive = TRUE)

  fixSrc <- makeFixture(src, "foo.csv")
  fix1   <- makeFixture(s1,  "foo.csv")
  fix2   <- makeFixture(s2,  "foo.csv")
  writeChecksumsFor(s1, fix1)
  writeChecksumsFor(s2, fix2)

  withr::local_options(list(reproducible.inputPaths = c(s1, s2)))

  capture.output(
    preProcess(url = fixSrc$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1),
    type = "output"
  )

  skip_on_os("windows")
  expect_true(isTRUE(sameInode(file.path(dest, "foo.csv"), fix1$path)),
               info = "deterministic order: first destinationPathShared entry wins")
  expect_false(isTRUE(sameInode(file.path(dest, "foo.csv"), fix2$path)))
})

test_that("H7: file in destinationPath wins; destinationPathShared not consulted", {
  testInit("digest")
  shared <- normPath(file.path(tmpdir, "shared"))
  dest   <- normPath(file.path(tmpdir, "dest"))
  src    <- normPath(file.path(tmpdir, "src"))
  for (d in c(shared, dest, src)) dir.create(d, recursive = TRUE)

  fixSrc    <- makeFixture(src,    "foo.csv")
  fixDest   <- makeFixture(dest,   "foo.csv")
  fixShared <- makeFixture(shared, "foo.csv")
  writeChecksumsFor(dest,   fixDest)
  writeChecksumsFor(shared, fixShared)

  withr::local_options(list(reproducible.inputPaths = shared))

  origIno <- inoOf(fixDest$path)

  capture.output(
    preProcess(url = fixSrc$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1),
    type = "output"
  )

  skip_on_os("windows")
  expect_equal(inoOf(file.path(dest, "foo.csv")), origIno,
               info = "destinationPath copy untouched; not relinked from shared")
})

# ===========================================================================
# §5.3  Resolution — sad paths
# ===========================================================================

test_that("S3: destinationPathShared path doesn't exist → falls through", {
  testInit("digest")
  shared <- file.path(tmpdir, "doesnotexist")
  dest   <- normPath(file.path(tmpdir, "dest"))
  src    <- normPath(file.path(tmpdir, "src"))
  for (d in c(dest, src)) dir.create(d, recursive = TRUE)
  fixSrc <- makeFixture(src, "foo.csv")

  withr::local_options(list(reproducible.inputPaths = shared))

  expect_no_error(capture.output(
    preProcess(url = fixSrc$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1),
    type = "output"
  ))
  expect_true(file.exists(file.path(dest, "foo.csv")))
})

test_that("S7: HEAD network error → falls back to download path", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.tif")
  url <- "https://example.com/path/data.tif"

  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = fx$path)

  out <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) stop("simulated network error"),
    { pp(ctx) }
  )
  # Network failure must not poison state — skipDownload stays FALSE so
  # pp_download proceeds normally, and no sidecar/CHECKSUMS row is written.
  expect_false(isTRUE(out$skipDownload))
  expect_false(fx$path %in% out$hashVerified)
})


# ===========================================================================
# §5.4  Sidecar lifecycle and CHECKSUMS interaction
# ===========================================================================

test_that("C4: sidecar with missing adjacent file → ignored", {
  testInit()
  td  <- withr::local_tempdir()
  url <- "https://example.com/data.tif"

  # Write an orphan sidecar (no adjacent file).
  mk <- getFromNamespace("makeRemoteHashFile", "reproducible")
  sc <- mk(url, td, "data.tif", paste0(strrep("a", 32L)),
           algorithm = "md5", write = TRUE)
  expect_true(file.exists(sc))

  # pp_remote_hash_check returns ctx unchanged when localFile resolves to
  # NULL (no on-disk file to verify), regardless of orphan sidecars.
  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = file.path(td, "data.tif"))
  out <- pp(ctx)
  expect_false(isTRUE(out$skipDownload))
})
test_that("C8: legacy single-line .hash sidecar parsed via length heuristic", {
  parse <- getFromNamespace(".parseRemoteHashFile", "reproducible")
  td <- withr::local_tempdir()
  legacy <- file.path(td, "legacy.hash")
  writeLines(strrep("a", 32L), legacy)        # md5-shaped, no `:` prefix
  out <- parse(legacy)
  expect_identical(out$algorithm, "md5")
  expect_identical(out$hash, strrep("a", 32L))
})
test_that("C9: legacy CHECKSUMS row with multi-algorithm coexists with new row", {
  testInit()
  td <- withr::local_tempdir()
  fx <- makeFixture(td, "shared.dat")
  csf <- writeChecksumsFor(td, fx)            # xxhash64 row only

  upsert <- getFromNamespace(".upsertChecksumsRow", "reproducible")
  upsert(csf, file = basename(fx$path),
         hash = digest::digest(file = fx$path, algo = "md5"),
         filesize = fx$size, algorithm = "md5")

  rows <- read.table(csf, header = TRUE, stringsAsFactors = FALSE)
  algos <- sort(rows$algorithm[rows$file == basename(fx$path)])
  expect_identical(algos, c("md5", "xxhash64"))
})
test_that("C10: malformed sidecar contents → treated as missing", {
  parse <- getFromNamespace(".parseRemoteHashFile", "reproducible")
  td <- withr::local_tempdir()

  # Empty file → NULL
  empty <- file.path(td, "empty.hash")
  file.create(empty)
  expect_null(parse(empty))

  # Non-existent file → NULL
  expect_null(parse(file.path(td, "does-not-exist.hash")))

  # Garbage one-line → still parses (algorithm classified as etag-opaque),
  # but the hash won't match any remote so downstream comparison fails.
  garbage <- file.path(td, "garbage.hash")
  writeLines("not-a-real-hash-string", garbage)
  out <- parse(garbage)
  expect_identical(out$algorithm, "etag-opaque")
})


# ===========================================================================
# §5.5  targetFile inference
# ===========================================================================

test_that("T1: existing sidecar → no HEAD request fired", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.tif")
  url <- "https://example.com/path/data.tif"

  # Pre-write the sidecar; pp_remote_hash_check should fast-path on existence.
  mk <- getFromNamespace("makeRemoteHashFile", "reproducible")
  mk(url, td, basename(fx$path),
     digest::digest(file = fx$path, algo = "md5"),
     algorithm = "md5", write = TRUE)

  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = fx$path)

  callCount <- 0L
  out <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) {
      callCount <<- callCount + 1L
      stop("HEAD should not be called when sidecar exists")
    },
    { pp(ctx) }
  )
  expect_identical(callCount, 0L)
  expect_true(isTRUE(out$skipDownload))
})
test_that("T2: no sidecar → HEAD fires; size+hash match → sidecar written", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.tif")
  url <- "https://example.com/path/data.tif"

  fakeMeta <- list(
    targetFile = "data.tif",
    fileSize = as.character(fx$size),
    remoteHash = digest::digest(file = fx$path, algo = "md5"),
    remoteAlgorithm = "md5", timestampOnline = "now"
  )
  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = fx$path)

  callCount <- 0L
  out <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) { callCount <<- callCount + 1L; fakeMeta },
    { pp(ctx) }
  )
  expect_identical(callCount, 1L)
  expect_true(isTRUE(out$skipDownload))
  expect_false(is.null(readSidecar(td, fx, url)))
})
test_that("T6: HEAD network error → silent fall-through", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.tif")
  url <- "https://example.com/path/data.tif"

  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = fx$path)

  out <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) stop("simulated HEAD failure"),
    { pp(ctx) }
  )
  # Silent fall-through: no skipDownload, no sidecar, no error.
  expect_false(isTRUE(out$skipDownload))
  expect_null(readSidecar(td, fx, url))
})

test_that("T7: targetFile NULL + url NULL (local-only) → no regression", {
  testInit("digest")
  dest <- normPath(file.path(tmpdir, "dest"))
  src  <- normPath(file.path(tmpdir, "src"))
  for (d in c(dest, src)) dir.create(d, recursive = TRUE)
  fix <- makeFixture(src, "foo.csv")

  ## dlFun-only call: copy the fixture into destinationPath via dlFun.
  dlFun <- quote(file.copy(SRC, DEST, overwrite = TRUE))
  ## simplest: just use a real local URL — covers the "no crash" intent.
  capture.output(
    preProcess(url = fix$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1),
    type = "output"
  )
  expect_true(file.exists(file.path(dest, "foo.csv")))
})


# ===========================================================================
# §5.6  remoteHash matching
# ===========================================================================

test_that("R1–R7: remoteHash-driven candidate validation", {
  # Cluster of behaviours, each verified inline:
  #   R1 size-match + hash-match → accept
  #   R2 size-match + hash-mismatch → reject (download)
  #   R3 size-mismatch (regardless of hash) → reject (download)
  #   R4 opaque ETag → reject (no positive trust possible)
  #   R5 missing remote hash → reject
  #   R6 sidecar pre-exists with matching contents under opt-in re-check
  #   R7 sidecar pre-exists with NON-matching contents → re-verify
  testInit()

  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")

  ## R1: size + hash match → skipDownload
  td1 <- withr::local_tempdir()
  fx1 <- makeFixture(td1, "r1.tif")
  out1 <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) list(
      targetFile = "r1.tif", fileSize = as.character(fx1$size),
      remoteHash = digest::digest(file = fx1$path, algo = "md5"),
      remoteAlgorithm = "md5", timestampOnline = "now"),
    { pp(makePPCtx("https://e.com/r1.tif", td1, neededFiles = fx1$path)) }
  )
  expect_true(isTRUE(out1$skipDownload))

  ## R2: size match, hash mismatch → reject
  td2 <- withr::local_tempdir()
  fx2 <- makeFixture(td2, "r2.tif")
  out2 <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) list(
      targetFile = "r2.tif", fileSize = as.character(fx2$size),
      remoteHash = digest::digest("DIFFERENT", algo = "md5", serialize = FALSE),
      remoteAlgorithm = "md5", timestampOnline = "now"),
    { pp(makePPCtx("https://e.com/r2.tif", td2, neededFiles = fx2$path)) }
  )
  expect_false(isTRUE(out2$skipDownload))

  ## R3: size mismatch → reject
  td3 <- withr::local_tempdir()
  fx3 <- makeFixture(td3, "r3.tif")
  out3 <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) list(
      targetFile = "r3.tif", fileSize = as.character(fx3$size + 1L),
      remoteHash = digest::digest(file = fx3$path, algo = "md5"),
      remoteAlgorithm = "md5", timestampOnline = "now"),
    { pp(makePPCtx("https://e.com/r3.tif", td3, neededFiles = fx3$path)) }
  )
  expect_false(isTRUE(out3$skipDownload))

  ## R4: opaque ETag → reject
  td4 <- withr::local_tempdir()
  fx4 <- makeFixture(td4, "r4.tif")
  out4 <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) list(
      targetFile = "r4.tif", fileSize = as.character(fx4$size),
      remoteHash = "W/abc-12-34", remoteAlgorithm = "etag-opaque",
      timestampOnline = "now"),
    { pp(makePPCtx("https://e.com/r4.tif", td4, neededFiles = fx4$path)) }
  )
  expect_false(isTRUE(out4$skipDownload))

  ## R5: empty remote hash → reject (treated as opaque)
  td5 <- withr::local_tempdir()
  fx5 <- makeFixture(td5, "r5.tif")
  out5 <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) list(
      targetFile = "r5.tif", fileSize = as.character(fx5$size),
      remoteHash = "", remoteAlgorithm = "etag-opaque",
      timestampOnline = "now"),
    { pp(makePPCtx("https://e.com/r5.tif", td5, neededFiles = fx5$path)) }
  )
  expect_false(isTRUE(out5$skipDownload))

  ## R6: sidecar with matching contents under opt-in re-check
  td6 <- withr::local_tempdir()
  fx6 <- makeFixture(td6, "r6.tif")
  url6 <- "https://e.com/r6.tif"
  hash6 <- digest::digest(file = fx6$path, algo = "md5")
  mk <- getFromNamespace("makeRemoteHashFile", "reproducible")
  mk(url6, td6, basename(fx6$path), hash6, algorithm = "md5", write = TRUE)

  withr::local_options(reproducible.checkRemoteHash = TRUE)
  out6 <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) list(
      targetFile = "r6.tif", fileSize = as.character(fx6$size),
      remoteHash = hash6, remoteAlgorithm = "md5", timestampOnline = "now"),
    { pp(makePPCtx(url6, td6, neededFiles = fx6$path)) }
  )
  expect_true(isTRUE(out6$skipDownload))
})


# ===========================================================================
# §5.7  Backward compatibility
# ===========================================================================

test_that("B1: legacy reproducible.inputPaths still works (regression guard)", {
  testInit("digest")
  shared <- normPath(file.path(tmpdir, "shared"))
  dest   <- normPath(file.path(tmpdir, "dest"))
  src    <- normPath(file.path(tmpdir, "src"))
  for (d in c(shared, dest, src)) dir.create(d, recursive = TRUE)

  fixSrc    <- makeFixture(src,    "foo.csv")
  fixShared <- makeFixture(shared, "foo.csv")
  writeChecksumsFor(shared, fixShared)

  withr::local_options(list(reproducible.inputPaths = shared))

  capture.output(
    preProcess(url = fixSrc$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1),
    type = "output"
  )

  skip_on_os("windows")
  expect_true(isTRUE(sameInode(file.path(dest, "foo.csv"), fixShared$path)))
})

# B2 / B3 deleted: covered by P5 (inputPaths-only deprecation chain) and P6
# (destinationPathShared wins over inputPaths). The third option name `dataPath` was
# removed before the rename ever shipped — `destinationPathShared` is canonical and
# `inputPaths` is the only legacy alias.


# ===========================================================================
# §5.8  Integration with prepInputs/Cache
# ===========================================================================

test_that("I1+I2: Cache(prepInputs(...)) — cold uses destinationPathShared; warm is cache hit", {
  # Cache(...) -> doSaveToCache -> wrapSaveToCache -> .wrap.default needs
  # terra to wrap raster-shaped outputs. Without terra (nosuggests=true /
  # _R_CHECK_DEPENDS_ONLY_=true matrix entry), it aborts with "Please
  # install terra package". Skip then; full-Suggests entries still cover.
  skip_if_not_installed("terra")
  testInit("digest")
  shared <- normPath(file.path(tmpdir, "shared"))
  dest   <- normPath(file.path(tmpdir, "dest"))
  src    <- normPath(file.path(tmpdir, "src"))
  cache  <- normPath(file.path(tmpdir, "cache"))
  for (d in c(shared, dest, src, cache)) dir.create(d, recursive = TRUE)

  fixSrc    <- makeFixture(src,    "foo.csv")
  fixShared <- makeFixture(shared, "foo.csv")
  writeChecksumsFor(shared, fixShared)

  withr::local_options(list(
    reproducible.inputPaths = shared,
    reproducible.cachePath  = cache
  ))

  doIt <- function() {
    preProcess(url = fixSrc$url, targetFile = "foo.csv",
               destinationPath = dest, fun = NA, verbose = -1)
  }
  capture.output(Cache(doIt), type = "output")
  capture.output(Cache(doIt), type = "output")

  expect_true(file.exists(file.path(dest, "foo.csv")))
  skip_on_os("windows")
  expect_true(isTRUE(sameInode(file.path(dest, "foo.csv"), fixShared$path)))
})


# ===========================================================================
# §5.10  Performance smoke
# ===========================================================================

test_that("Q1: 10k unrelated files in shared, recursive=FALSE → fast lookup", {
  skip_on_cran()
  testInit("digest")
  if (!isTRUE(getOption("reproducible.runLargeFileTests", FALSE)))
    skip("set reproducible.runLargeFileTests=TRUE to run")

  shared <- normPath(file.path(tmpdir, "shared"))
  dest   <- normPath(file.path(tmpdir, "dest"))
  src    <- normPath(file.path(tmpdir, "src"))
  for (d in c(shared, dest, src)) dir.create(d, recursive = TRUE)
  for (i in seq_len(10000))
    file.create(file.path(shared, sprintf("noise_%05d.bin", i)))
  fixSrc    <- makeFixture(src,    "foo.csv")
  fixShared <- makeFixture(shared, "foo.csv")
  writeChecksumsFor(shared, fixShared)

  withr::local_options(list(reproducible.inputPaths = shared))

  t <- system.time(
    capture.output(
      preProcess(url = fixSrc$url, targetFile = "foo.csv",
                 destinationPath = dest, fun = NA, verbose = -1),
      type = "output"
    )
  )
  # lookup over 10k flat files should be subsecond on typical disk;
  # 5 s is a generous CI ceiling.
  expect_lt(t[["elapsed"]], 5)
})

test_that("Q3: hot destinationPath, repeated calls → no HEAD after first", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.tif")
  url <- "https://example.com/path/data.tif"
  fakeMeta <- list(
    targetFile = "data.tif", fileSize = as.character(fx$size),
    remoteHash = digest::digest(file = fx$path, algo = "md5"),
    remoteAlgorithm = "md5", timestampOnline = "now"
  )
  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")

  callCount <- 0L
  testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) { callCount <<- callCount + 1L; fakeMeta },
    { pp(makePPCtx(url, td, neededFiles = fx$path)) }
  )
  expect_identical(callCount, 1L)

  testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) {
      callCount <<- callCount + 1L
      stop("HEAD should not fire on hot calls")
    },
    { for (i in 1:5) pp(makePPCtx(url, td, neededFiles = fx$path)) }
  )
  expect_identical(callCount, 1L)
})

test_that("Q5: classify md5/sha1/sha256/etag-opaque by hash shape", {
  classify <- getFromNamespace(".classifyRemoteHashAlgo", "reproducible")
  expect_identical(classify(strrep("a", 32L)), "md5")
  expect_identical(classify(strrep("b", 40L)), "sha1")
  expect_identical(classify(strrep("c", 64L)), "sha256")
  expect_identical(classify("anything", isGDurl = TRUE), "md5")
  expect_identical(classify("W/abc-12-34"), "etag-opaque")
})



# ===========================================================================
# §5.14  Remote-hash content verification (size as negative-only signal)
# ---------------------------------------------------------------------------
# These tests exercise the algorithm:
#   1. remote.size != local.size            -> download
#   2. remote algorithm == "etag-opaque"    -> download
#   3. size match + hash match              -> sidecar + CHECKSUMS row, skip
#   4. size match + hash mismatch (.tif     -> download (the size-poison case
#      false-positive scenario)                that the previous code missed)
# ===========================================================================

# Helpers (makePPCtx / readSidecar / readChecksumsRows) hoisted to the top
# of this file so they're visible to earlier test_that() blocks too.

# .classifyRemoteHashAlgo --------------------------------------------------

test_that("classify: md5/sha1/sha256/opaque + Google Drive override", {
  classify <- getFromNamespace(".classifyRemoteHashAlgo", "reproducible")
  expect_identical(classify(strrep("a", 32L)), "md5")
  expect_identical(classify(strrep("0", 40L)), "sha1")
  expect_identical(classify(strrep("f", 64L)), "sha256")
  expect_identical(classify("W/abc-12-34"), "etag-opaque")
  expect_identical(classify(""), "etag-opaque")
  expect_identical(classify(NA_character_), "etag-opaque")
  expect_identical(classify(NULL), "etag-opaque")
  # Google Drive override forces md5 regardless of shape
  expect_identical(classify("anything", isGDurl = TRUE), "md5")
})

# .parseRemoteHashFile (legacy + new format) -------------------------------

test_that("parse sidecar: legacy single-line vs new <algo>:<hash>", {
  parse <- getFromNamespace(".parseRemoteHashFile", "reproducible")
  td <- withr::local_tempdir()

  # New format
  pNew <- file.path(td, "new.hash")
  writeLines("md5:abcdef0123456789abcdef0123456789", pNew)
  out <- parse(pNew)
  expect_identical(out$algorithm, "md5")
  expect_identical(out$hash, "abcdef0123456789abcdef0123456789")

  # Legacy single-line, md5-shaped
  pLegacyMd5 <- file.path(td, "legacymd5.hash")
  writeLines(strrep("a", 32L), pLegacyMd5)
  out2 <- parse(pLegacyMd5)
  expect_identical(out2$algorithm, "md5")
  expect_identical(out2$hash, strrep("a", 32L))

  # Legacy single-line, sha1-shaped
  pLegacySha1 <- file.path(td, "legacysha1.hash")
  writeLines(strrep("b", 40L), pLegacySha1)
  out3 <- parse(pLegacySha1)
  expect_identical(out3$algorithm, "sha1")

  # Missing file
  expect_null(parse(file.path(td, "does-not-exist.hash")))
})

# pp_remote_hash_check: md5 match -> sidecar + CHECKSUMS row + skipDownload --

test_that("pp_remote_hash_check: md5 match writes md5 sidecar + row, skips DL", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.tif")
  url <- "https://example.com/path/data.tif"

  fakeMeta <- list(
    targetFile = "data.tif",
    fileSize = as.character(fx$size),
    remoteHash = digest::digest(file = fx$path, algo = "md5"),
    remoteAlgorithm = "md5",
    timestampOnline = "now"
  )
  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = fx$path)

  out <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) fakeMeta,
    { pp(ctx) }
  )

  expect_true(isTRUE(out$skipDownload))
  expect_true(fx$path %in% out$hashVerified)

  sc <- readSidecar(td, fx, url)
  expect_false(is.null(sc))
  expect_match(sc$contents, "^md5:")
  expect_identical(sub("^md5:", "", sc$contents), fakeMeta$remoteHash)

  rows <- readChecksumsRows(td)
  expect_false(is.null(rows))
  hit <- which(rows$algorithm == "md5" & basename(rows$file) == "data.tif")
  expect_length(hit, 1L)
  expect_identical(rows$checksum[hit], fakeMeta$remoteHash)
})

# pp_remote_hash_check: sha1 match -> sha1 sidecar + sha1 row ---------------

test_that("pp_remote_hash_check: sha1 match writes sha1 sidecar + row", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.bin", content = "sha1-content\n")
  url <- "https://example.com/sha1/data.bin"

  fakeMeta <- list(
    targetFile = "data.bin",
    fileSize = as.character(fx$size),
    remoteHash = digest::digest(file = fx$path, algo = "sha1"),
    remoteAlgorithm = "sha1",
    timestampOnline = "now"
  )
  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = fx$path)

  out <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) fakeMeta,
    { pp(ctx) }
  )

  expect_true(isTRUE(out$skipDownload))
  sc <- readSidecar(td, fx, url)
  expect_match(sc$contents, "^sha1:")

  rows <- readChecksumsRows(td)
  hit <- which(rows$algorithm == "sha1" & basename(rows$file) == "data.bin")
  expect_length(hit, 1L)
})

# pp_remote_hash_check: opaque ETag -> no positive trust -> download path ---

test_that("pp_remote_hash_check: opaque ETag falls through to download", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.bin")
  url <- "https://example.com/opaque/data.bin"

  fakeMeta <- list(
    targetFile = "data.bin",
    fileSize = as.character(fx$size),
    remoteHash = "W/some-weak-etag",
    remoteAlgorithm = "etag-opaque",
    timestampOnline = "now"
  )
  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = fx$path)

  out <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) fakeMeta,
    { pp(ctx) }
  )

  expect_false(isTRUE(out$skipDownload))
  expect_false(fx$path %in% out$hashVerified)
  expect_null(readSidecar(td, fx, url))
})

# pp_remote_hash_check: size match + hash MISMATCH (.tif poison case) ------

test_that("pp_remote_hash_check: size match + hash mismatch -> download", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.tif", content = "local-content-A\n")
  url <- "https://example.com/poison/data.tif"

  # Remote has the same size but completely different content.
  fakeMeta <- list(
    targetFile = "data.tif",
    fileSize = as.character(fx$size),
    remoteHash = digest::digest("DIFFERENT-CONTENT", algo = "md5",
                                serialize = FALSE),
    remoteAlgorithm = "md5",
    timestampOnline = "now"
  )
  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")
  ctx <- makePPCtx(url, td, neededFiles = fx$path)

  out <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) fakeMeta,
    { pp(ctx) }
  )

  # Size matched but hash didn't — must NOT skip download, must NOT write
  # sidecar or CHECKSUMS row (the previous size-only-fallback code would have
  # incorrectly accepted this case).
  expect_false(isTRUE(out$skipDownload))
  expect_false(fx$path %in% out$hashVerified)
  expect_null(readSidecar(td, fx, url))
})

# pp_remote_hash_check: second call hits sidecar fast-path (no HEAD) -------

test_that("pp_remote_hash_check: 2nd call uses sidecar fast-path (no HEAD)", {
  testInit()
  td  <- withr::local_tempdir()
  fx  <- makeFixture(td, "data.tif")
  url <- "https://example.com/path/data.tif"

  fakeMeta <- list(
    targetFile = "data.tif",
    fileSize = as.character(fx$size),
    remoteHash = digest::digest(file = fx$path, algo = "md5"),
    remoteAlgorithm = "md5",
    timestampOnline = "now"
  )
  pp <- getFromNamespace("pp_remote_hash_check", "reproducible")

  # First call: HEAD invoked, writes sidecar.
  callCount <- 0L
  ctx1 <- makePPCtx(url, td, neededFiles = fx$path)
  testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) { callCount <<- callCount + 1L; fakeMeta },
    { pp(ctx1) }
  )
  expect_identical(callCount, 1L)
  expect_false(is.null(readSidecar(td, fx, url)))

  # Second call: must NOT call getRemoteMetadata (sidecar fast-path).
  ctx2 <- makePPCtx(url, td, neededFiles = fx$path)
  out2 <- testthat::with_mocked_bindings(
    getRemoteMetadata = function(...) {
      callCount <<- callCount + 1L
      stop("getRemoteMetadata should not be called when sidecar exists")
    },
    { pp(ctx2) }
  )
  expect_identical(callCount, 1L)             # unchanged
  expect_true(isTRUE(out2$skipDownload))
})

# Checksums(): multi-row-per-file (xxhash64 + md5 coexist) -----------------

test_that("Checksums: multiple algorithms per file coexist; reads pick algo", {
  testInit()
  td <- withr::local_tempdir()
  fx <- makeFixture(td, "shared.dat")

  csf <- file.path(td, "CHECKSUMS.txt")
  hdr <- '"file" "checksum" "filesize" "algorithm"'
  rowX <- sprintf('"%s" "%s" "%d" "xxhash64"',
                  basename(fx$path), fx$hash, fx$size)
  md5  <- digest::digest(file = fx$path, algo = "md5")
  rowM <- sprintf('"%s" "%s" "%d" "md5"',
                  basename(fx$path), md5, fx$size)
  writeLines(c(hdr, rowX, rowM), csf)

  # Default algo (xxhash64): expect OK row matching the xxhash64 entry.
  resX <- Checksums(path = td, write = FALSE, files = basename(fx$path),
                    checksumFile = csf, verbose = 0)
  hits <- resX[!is.na(resX$result) & resX$result == "OK", ]
  expect_true(NROW(hits) >= 1L)
  expect_true(any(basename(hits$expectedFile) == basename(fx$path)))

  # Explicit md5: expect the md5 row to match.
  resM <- Checksums(path = td, write = FALSE, files = basename(fx$path),
                    checksumFile = csf, verbose = 0, algo = "md5")
  hitsM <- resM[!is.na(resM$result) & resM$result == "OK", ]
  expect_true(NROW(hitsM) >= 1L)
})

# .upsertChecksumsRow preserves rows for other algorithms ------------------

test_that("upsertChecksumsRow: preserves other-algorithm rows for same file", {
  testInit()
  td <- withr::local_tempdir()
  fx <- makeFixture(td, "shared.dat")

  csf <- writeChecksumsFor(td, fx)            # xxhash64 row
  upsert <- getFromNamespace(".upsertChecksumsRow", "reproducible")
  md5 <- digest::digest(file = fx$path, algo = "md5")

  upsert(csf, file = basename(fx$path), hash = md5,
         filesize = fx$size, algorithm = "md5")

  rows <- read.table(csf, header = TRUE, stringsAsFactors = FALSE)
  algos <- sort(rows$algorithm[rows$file == basename(fx$path)])
  expect_identical(algos, c("md5", "xxhash64"))
})


# ===========================================================================
# §5.6  Stale sidecar with locally-deleted file
#
# User scenario: a previous prepInputs() wrote a .hash sidecar in
# destinationPath and the file is later removed by hand. The sidecar is left
# behind. Two sub-cases matter:
#
#   C5: a fresh copy of the file is still in destinationPathShared. The next
#       prepInputs() should re-materialise the file (hardlink) from shared
#       *without* a network round-trip — the sidecar's recorded hash is the
#       authority that lets us trust the shared copy.
#
#   C6: the file is gone everywhere (no shared copy). The next prepInputs()
#       must fall through to a real download. The stale sidecar's recorded
#       hash is now meaningless — we cannot trust it because there is no
#       on-disk file to verify it against.
#
# Snapshots (rather than regex pattern matching) lock down the user-visible
# messages so a refactor of the message text can't silently break the
# documented contract — `testthat::snapshot_accept()` makes the intent
# explicit when wording is updated on purpose.
# ===========================================================================

# Scrub volatile tmpdir paths so the snapshot is stable across runs, and
# normalize continuation-line wrapping so the snapshot doesn't break when
# messagePreProcess reflows the same message into a different number of
# lines (terminal width, indent style, etc.).
.snapTransform <- function(lines) {
  ## testthat is called once per captured Message/Code block. We collapse the
  ## entire block into ONE line: `messagePreProcess` wraps its output to
  ## `getOption("width")`, and the line boundaries (and indentation of
  ## continuation lines) shift between environments — for the same logical
  ## message, an 80-col tty produces a different line count than a 120-col
  ## tty. Erasing both newlines and run-of-whitespace inside the block
  ## removes that variability while keeping the semantics ("Running
  ## preProcess Preparing: foo.csv ...downloading... Downloading <file-url>
  ## ... Appending checksums to ...") intact.
  joined <- paste(as.character(lines), collapse = " ")
  joined <- gsub("/tmp/Rtmp[A-Za-z0-9]+", "<tmpdir>", joined)
  joined <- gsub("file://[^ ]+", "<file-url>", joined)
  joined <- gsub("file[A-Za-z0-9]{6,}", "<tmp>", joined)
  joined <- gsub("[[:space:]]+", " ", joined)
  joined <- sub("^[[:space:]]+", "", joined)
  joined <- sub("[[:space:]]+$", "", joined)
  joined
}

## Plant a remote-hash sidecar that `.findRemoteHashSidecars` will discover
## by its `.<filename>_*.hash` prefix. We bypass `makeRemoteHashFile()` here
## on purpose: that helper derives the suffix from the source URL, which
## under R CMD check can produce paths whose `file(con, "w")` fails for
## reasons unrelated to what these tests are exercising. The lookup logic
## only inspects the prefix/suffix and the file contents, so any matching
## name works.
.plantSidecar <- function(destDir, filename, algorithm, hash, tag = "synthetic") {
  stopifnot(dir.exists(destDir))
  sc <- file.path(destDir, paste0(".", filename, "_", tag, ".hash"))
  writeLines(paste0(algorithm, ":", hash), sc)
  sc
}

test_that("C5: deleted local file + stale sidecar + file in destinationPathShared → relink, no download", {
  ## Under nosuggests R CMD check (`_R_CHECK_DEPENDS_ONLY_=true`), reproducible
  ## emits a one-time "install httr2 and try again" hint via `preProcess()`
  ## that we don't expect in the snapshot. The test is exercising the
  ## prepInputs sidecar logic, which already depends on Suggests packages
  ## (digest, terra). Easier to skip the whole pair under nosuggests than
  ## to scrub the hint from .snapTransform.
  skip_if_not_installed("httr2")
  testInit("digest")
  ## Force a wide print width so testthat's snapshot of the `Code` block
  ## (the deparsed expression) doesn't reflow between an 80-col local TTY
  ## and a 200-col CI runner — `expect_snapshot()`'s `transform` argument
  ## only touches captured OUTPUT, not the rendered code line.
  withr::local_options(width = 200)
  shared <- normPath(file.path(tmpdir, "shared"))
  dest   <- normPath(file.path(tmpdir, "dest"))
  src    <- normPath(file.path(tmpdir, "src"))
  for (d in c(shared, dest, src)) dir.create(d, recursive = TRUE)

  fixSrc    <- makeFixture(src,    "foo.csv")
  fixShared <- makeFixture(shared, "foo.csv")             # same content/hash
  writeChecksumsFor(shared, fixShared)

  # Plant a sidecar that records the (correct) hash, but leave dest/foo.csv
  # itself absent — the user manually deleted it.
  sc <- .plantSidecar(dest, "foo.csv", "xxhash64", fixShared$hash)
  expect_true(file.exists(sc))
  expect_false(file.exists(file.path(dest, "foo.csv")))

  withr::local_options(list(reproducible.inputPaths = shared))

  testthat::local_edition(3)
  expect_snapshot(
    {
      invisible(preProcess(url = fixSrc$url, targetFile = "foo.csv",
                           destinationPath = dest, fun = NA, verbose = 1))
    },
    transform = .snapTransform
  )

  # The destination should now have the file back, and (POSIX) it should
  # share an inode with the shared copy — i.e. it was hardlinked, not copied
  # via a fresh download.
  expect_true(file.exists(file.path(dest, "foo.csv")))
  skip_on_os("windows")
  expect_true(isTRUE(sameInode(file.path(dest, "foo.csv"), fixShared$path)),
              info = "C5: should hardlink from destinationPathShared, not re-download")
})

test_that("C6: deleted local file + stale sidecar + nothing in shared → download proceeds", {
  skip_if_not_installed("httr2")     # see C5 for rationale
  testInit("digest")
  withr::local_options(width = 200)  # see C5 for rationale
  dest <- normPath(file.path(tmpdir, "dest"))
  src  <- normPath(file.path(tmpdir, "src"))
  for (d in c(dest, src)) dir.create(d, recursive = TRUE)

  fixSrc <- makeFixture(src, "foo.csv")

  # Stale sidecar with a hash that has NOTHING to do with current content —
  # simulates the file having drifted (or just an old run) and being deleted.
  sc <- .plantSidecar(dest, "foo.csv", "xxhash64", strrep("d", 16L))
  expect_true(file.exists(sc))
  expect_false(file.exists(file.path(dest, "foo.csv")))

  withr::local_options(list(
    reproducible.inputPaths            = NULL,
    reproducible.destinationPathShared = NULL
  ))

  testthat::local_edition(3)
  expect_snapshot(
    {
      invisible(preProcess(url = fixSrc$url, targetFile = "foo.csv",
                           destinationPath = dest, fun = NA, verbose = 1))
    },
    transform = .snapTransform
  )

  # The file should now exist locally (downloaded from `file://src/foo.csv`).
  expect_true(file.exists(file.path(dest, "foo.csv")))
})

## This is an incomplete file; it will be slowly transitioned to have all messaging here
## Any new message should be written as a .messageGreps entry, then used in the functions
## with the mess*.
## All messages and message-generating functions are stored in the .message environment,
## to allow updating (i.e., `.message$PreProcessIndent`).

#' @keywords internal
#' @rdname pkgEnv
.message <- new.env()

.message$SkipDownload <- "Skipping download of url; local copy already exists and passes checksums"

.message$stopNeedArchive <- function(archive)
  paste0("Please install.packages('archive') to extract files from \n", archive)

.message$Greps <- list(
  studyArea_Spatial = "The \\'studyArea\\' provided is not a Spatial\\* object.",
  rasterToMatch_Raster = "The \\'rasterToMatch\\' provided is not a Raster\\* object.",
  anySpatialClass = "Raster\\*, Spat\\*, sf or Spatial object"
)

.message$PreProcessIndentOrig <- .message$PreProcessIndent <- ""

.message$CacheIndent <- "    "

.message$Spatial <- lapply(.message$Greps, gsub, pattern = "\\\\", replacement = "")

.message$cacheGeoDomainContained <- "Spatial domain is contained within the url; returning the object"
.message$cacheGeoDomainNotContained <- "Domain is not contained within the targetFile; running FUN"
.message$cacheGeoNoRemoteExists <- "No remote targetFile exists; starting to build a new one"

.message$LoadedCacheResult <- function(src = 1) {
  srcPoss <- c("Cached", "Memoised")
  if (is.numeric(src)) {
    src <- srcPoss[src]
  } else if (is.character(src)) {
    src <- srcPoss[grepl(src, srcPoss)]
  }
  paste0("Loaded! ", src, " result from previous")
}

.message$overwriting <- function(names, type = c("function", "userTags"), verbose) {
  messageCache("Overwriting Cache entry with ", type[1],": '", paste(names, collapse = ", "), "'",
               verbose = verbose
  )
}

.message$changingFormatTxt <- "Changing format of Cache entry from"

.message$changingFormat <- function(prevFile, newFile) {
  paste0("     (", .message$changingFormatTxt, " ", fileExt(prevFile), " to ",
  fileExt(newFile), ")")
}


.message$AddingToMemoised <- "(and added a memoised copy)"

.message$LoadedCache <- function(root, functionName) {
  paste0(root, " ", functionName, " call")
}

.message$BecauseOfA <- "...because of (a)"

.message$BecauseOfA <- "  "

.message$NoCachePathSupplied <- "No cachePath supplied"

.message$NoCacheRepoSuppliedGrep <- paste0(.message$NoCachePathSupplied, " and.+getOption\\('reproducible.cachePath'\\).+is.+inside")

.message$RequireNamespaceFn <- function(pkg, messageExtra = character(), minVersion = NULL) {
  mess <- paste0(
    pkg, if (!is.null(minVersion)) {
      paste0("(>=", minVersion, ")")
    }, " is required but not yet installed. Try: ",
    "install.packages('", pkg, "')"
  )
  if (length(messageExtra) > 0) {
    mess <- paste(mess, messageExtra)
  }
  mess
}

#' Use `message` with a consistent use of `verbose`
#'
#' This family has a consistent use of `verbose` allowing messages to be
#' turned on or off or verbosity increased or decreased throughout the family of
#' messaging in `reproducible`.
#'
#' - `messageDF` uses `message` to print a clean square data structure.
#' - `messageColoured` allows specific colours to be used.
#' - `messageQuestion` sets a high level for `verbose` so that the message always gets asked.
#'
#' @param df A data.frame, data.table, matrix
#' @param round An optional numeric to pass to `round`
#' @param colour Passed to `getFromNamespace(colour, ns = "cli")`,
#'   so any colour that `cli` can use
#' @param colnames Logical or `NULL`. If `TRUE`, then it will print
#'   column names even if there aren't any in the `df` (i.e., they will)
#'   be `V1` etc., `NULL` will print them if they exist, and `FALSE`
#'   which will omit them.
#' @param verboseLevel The numeric value for this `message*` call, equal or above
#'   which `verbose` must be. The higher this is set, the more unlikely the call
#'   will show a message.
#' @param indent An integer, indicating whether to indent each line
#' @inheritParams base::message
#'
#' @return
#' Used for side effects. This will produce a message of a structured `data.frame`.
#'
#' @inheritParams Cache
#'
#' @export
#' @importFrom data.table is.data.table as.data.table
#' @importFrom utils capture.output
#' @rdname messageColoured
messageDF <- function(df, round, colour = NULL, colnames = NULL, indent = NULL,
                      verbose = getOption("reproducible.verbose"), verboseLevel = 1,
                      appendLF = TRUE) {
  if (isTRUE(verboseLevel <= verbose)) {
    origColNames <- if (is.null(colnames) || isTRUE(colnames)) colnames(df) else NULL

    if (is.matrix(df)) {
      df <- as.data.frame(df)
    }
    if (!is.data.table(df)) {
      df <- as.data.table(df)
    }
    df <- Copy(df)
    skipColNames <- if (is.null(origColNames) && !isTRUE(colnames)) TRUE else FALSE
    if (!missing(round)) {
      isNum <- sapply(df, is.numeric)
      isNum <- colnames(df)[isNum]
      for (Col in isNum) {
        set(df, NULL, Col, round(df[[Col]], round))
      }
    }
    outMess <- capture.output(df)
    outMess <- paddDFInitial(outMess, rows = 1:2, .spaceTmpChar, colour)

    if (skipColNames) outMess <- outMess[-1]
    outMess <- .addSlashNToAllButFinalElement(outMess)
    messageColoured(outMess, indent = indent, hangingIndent = FALSE,
                    colour = colour, verbose = verbose,
                    verboseLevel = verboseLevel, appendLF = appendLF)
    # out <- lapply(outMess, function(x) {
    #   messageColoured(x,
    #     colour = colour, indent = indent, appendLF = appendLF, verbose = verbose,
    #     verboseLevel = verboseLevel
    #   )
    # })
  }
}

#' @export
#' @rdname messageColoured
messagePrepInputs <- function(..., appendLF = TRUE,
                              verbose = getOption("reproducible.verbose"),
                              verboseLevel = 1) {
  messageColoured(...,
                  colour = getOption("reproducible.messageColourPrepInputs"),
                  verboseLevel = verboseLevel, verbose = verbose, appendLF = appendLF
  )
}

#' @export
#' @rdname messageColoured
messagePreProcess <- function(..., appendLF = TRUE,
                              verbose = getOption("reproducible.verbose"),
                              verboseLevel = 1) {
  messageColoured(..., indent = .message$PreProcessIndent,
                  colour = getOption("reproducible.messageColourPrepInputs"),
                  verboseLevel = verboseLevel, verbose = verbose, appendLF = appendLF
  )
}

#' @export
#' @rdname messageColoured
messageCache <- function(..., colour = getOption("reproducible.messageColourCache"),
                         verbose = getOption("reproducible.verbose"), verboseLevel = 1,
                         appendLF = TRUE) {
  if (isTRUE(verboseLevel <= verbose)) { # this is duplicate of messageColoured; but will be faster
    needIndent <- any(grepl("\b", unlist(list(...))))
    indent <- if (isTRUE(!needIndent)) .message$PreProcessIndent else ""
    messageColoured(..., indent = indent, # .message$CacheIndent,
                    colour = colour, appendLF = appendLF,
                    verboseLevel = verboseLevel, verbose = verbose
    )
  }
}

#' @export
#' @rdname messageColoured
messageQuestion <- function(..., verboseLevel = 0, appendLF = TRUE) {
  # force this message to print
  messageColoured(...,
                  colour = getOption("reproducible.messageColourQuestion"),
                  verbose = 10, verboseLevel = verboseLevel, appendLF = appendLF
  )
}

#' @rdname messageColoured
.messageFunctionFn <- function(..., appendLF = TRUE, verbose = getOption("reproducible.verbose"),
                               verboseLevel = 1) {
  fn <- cliCol(getOption("reproducible.messageColourFunction"))
  fn(...)
}

#' @param colour Any colour that can be understood by `cli`
#' @param hangingIndent Logical. If there are `\n`, should there be a handing indent of 2 spaces.
#'   Default is `TRUE`
#' @param ... Any character vector, passed to `paste0(...)`
#'
#' @export
#' @importFrom utils getFromNamespace
#' @rdname messageColoured
messageColoured <- function(..., colour = NULL, indent = NULL, hangingIndent = TRUE,
                            verbose = getOption("reproducible.verbose", 1),
                            verboseLevel = 1, appendLF = TRUE) {

  if (isTRUE(verboseLevel <= verbose)) {

    # if (getOption("reproducible.useCli", TRUE)) {
      mess <- paste0(..., collapse = "")
      # if (grepl("cacheIdInCache", mess)) browser()
      indentNum <- indent
      if (!is.null(indent))
        if (is.character(indent))
          indentNum <- cli::ansi_nchar(indent)
      if (is.null(indent)) indentNum <- 0

      # browser()
      # mess <- ansi_trimws_pureR(mess, "both")
      mess <- cli::ansi_trimws(mess, which = c("both"))
      # if (!all(mess2 == mess)) browser()
      if (any(grepl(.spaceTmpChar, mess)))
        mess <- gsub(.spaceTmpChar, " ", mess)
      hasSlashN <- any(grepl("\n", mess)) # faster than gregexpr that needs to count in the string
      if (!hasSlashN && cli::ansi_nchar(mess) > cli::console_width())
        mess <- cli::ansi_strwrap(x = mess,
                                  indent = indentNum,
                                  exdent = indentNum + hangingIndent * 2,
                                  simplify = TRUE)
      mess <- .addSlashNToAllButFinalElement(mess)

      if (!is.null(colour)) {
        fn <- cliCol(colour)
        hasSlashN2 <- gregexpr("\n", mess)[[1]]
        if (sum(hasSlashN2 > 0)) {
          mess <- paste0(fn(unlist(strsplit(mess, split = "\n"))), collapse = "\n")
        } else {
          # fn <- get(paste0("col_", colour), envir = asNamespace('cli'))
          mess <- fn(mess) # add the colour
        }
      }

      message(mess)

    # } else {
    #
    #   needCli <- FALSE
    #   if (!is.null(colour)) {
    #     if (is.character(colour)) {
    #       needCli <- TRUE
    #     }
    #   }
    #   mess <- paste0(..., collapse = "")
    #   if (!is.null(indent)) {
    #     mess <- paste0(indent, mess)
    #   }
    #
    #   # do line wrap with hanging indent
    #   maxLineLngth <- getOption("width") - 10 # 10 is a "buffer" for Rstudio miscalculations
    #   chars <- nchar(mess)
    #   if (chars > maxLineLngth) {
    #     splitOnSlashN <- strsplit(mess, "\n")
    #     newMess <- lapply(splitOnSlashN, function(m) {
    #       anyOneLine <- any(nchar(m) > maxLineLngth)
    #       if (anyOneLine) {
    #         messSplit <- strsplit(mess, split = " ")
    #         remainingChars <- chars
    #         messBuild <- character()
    #         while (remainingChars > maxLineLngth) {
    #           whNewLine <- which(cumsum(nchar(messSplit[[1]]) + 1) >= maxLineLngth)[1] - 1
    #           if (anyNA(whNewLine)) browser()
    #
    #           keepInd <- 1:whNewLine
    #           newMess <- paste(messSplit[[1]][keepInd], collapse = " ")
    #           messBuild <- c(messBuild, newMess)
    #           if (is.null(indent)) {
    #             # if it starts with a space -- that is the indent that is needed
    #             if (startsWith(newMess, " ")) {
    #               indent <<- sub("^( +).+", "\\1", newMess)
    #               if (grepl("^ +\\.\\.\\.", newMess)) {
    #                 indent <<- paste0(indent, " ")
    #               }
    #             } else {
    #               indent <<- ""
    #             }
    #
    #           }
    #           messSplit[[1]] <- messSplit[[1]][-keepInd]
    #           remainingChars <- remainingChars - nchar(newMess) - 1
    #           hangingIndent <<- TRUE
    #         }
    #         newMess <- paste(messSplit[[1]], collapse = " ")
    #         m <- c(messBuild, newMess)
    #       }
    #       m
    #     })
    #     mess <- unlist(newMess)
    #     mess <- paste0(.addSlashNToAllButFinalElement(mess), collapse = "")
    #   }
    #   hi <- if (isTRUE(hangingIndent)) paste0(indent, .message$BecauseOfA) else indent
    #   if (any(grepl("\n", mess))) {
    #     mess <- gsub("\n *", paste0("\n", hi), mess)
    #   }
    #   if (any(grepl(.spaceTmpChar, mess)))
    #     mess <- gsub(.spaceTmpChar, " ", mess)
    #   if (needCli && requireNamespace("cli", quietly = TRUE)) {
    #     mess <- lapply(strsplit(mess, "\n"), function(m)
    #       paste0(cliCol(colour)(m)))[[1]]
    #     mess <- .addSlashNToAllButFinalElement(mess)
    #     message(mess, appendLF = appendLF)
    #   } else {
    #     if (needCli && !isTRUE(.pkgEnv$.checkedCli) && !.requireNamespace("cli")) {
    #       message("To add colours to messages, install.packages('cli')", appendLF = appendLF)
    #       .pkgEnv$.checkedCli <- TRUE
    #     }
    #     message(mess, appendLF = appendLF)
    #   }
    # }
  }

}

#' @keywords internal
.message$CacheSize <- function(x, artifacts = NULL, cacheTable,
                              verbose = getOption("reproducible.verbose")) {
  tagCol <- "tagValue"
  if (missing(cacheTable)) {
    a <- showCache(x, verbose = verbose - 1, sorted = FALSE)
  } else {
    a <- cacheTable
  }
  cn <- if (any(colnames(a) %in% "tag")) "tag" else "tagKey"

  nas <- a[[.cacheTableTagColName()]] %in% "NA" & a[[cn]] == "object.size"
  if (any(nas))
    a <- a[!nas]

  b <- a[a[[cn]] == "object.size", ]
  if (any(colnames(a) %in% "tag")) {
    fsTotal <- sum(as.numeric(unlist(lapply(strsplit(b[[cn]], split = ":"), function(x) x[[2]])))) / 4
  } else {
    fsTotal <- sum(as.numeric(b[[.cacheTableTagColName()]])) / 4
  }
  fsTotalRasters <- sum(file.size(dir(file.path(x, "rasters"), full.names = TRUE, recursive = TRUE)))
  fsTotal <- fsTotal + fsTotalRasters
  class(fsTotal) <- "object_size"
  preMessage1 <- "  Total (including Rasters): "

  b <- a[a[[.cacheTableHashColName()]] %in% artifacts &
           (a[[cn]] %in% "object.size"), ]
  if (cn == "tag") {
    fs <- sum(as.numeric(unlist(lapply(strsplit(b[[cn]], split = ":"), function(x) x[[2]])))) / 4
  } else {
    fs <- sum(as.numeric(b[[.cacheTableTagColName()]])) / 4
  }

  class(fs) <- "object_size"
  preMessage <- "  Selected objects (not including Rasters): "

  messageCache("Cache size: ", verbose = verbose)
  messageCache(preMessage1, format(fsTotal, "auto"), verbose = verbose)
  messageCache(preMessage, format(fs, "auto"), verbose = verbose)
}

.message$ObjToRetrieveFn <- function(funName) {
  paste0("Object to retrieve (fn: ", .messageFunctionFn(funName))
}

.message$IndentDefault <- 1

.message$FileLinkUsed <- function(ftL, fts, verbose) {
  messageCache("  (A file with identical properties already exists in the Cache: ", basename(ftL), "; ")
  messageCache("    The newly added (", basename(fts), ") is a file.link to that file)",
               verbose = verbose
  )
}


.cacheMessageObjectSize <- function(otsObjSize, isBig) {
  if (!anyNA(otsObjSize)) {
    class(otsObjSize) <- "object_size"
    osMess <- format(otsObjSize, units = "auto")[isBig]
  } else {
    osMess <- ""
  }
  osMess
}


.message$SavedTxt <- "Saved! Cache file: "

.message$Saved <- function(cachePath, outputHash, functionName,
                           cacheSaveFormat = getOption("reproducible.cacheSaveFormat"), verbose) {
  postMess <- ""
  if (isTRUE(getOption("reproducible.useMemoise")))
    postMess <- paste0(" ", .message$AddingToMemoised)
  messageCache(.message$SavedTxt,
               basename2(CacheStoredFile(cachePath = cachePath, cacheId = outputHash,
                                         cacheSaveFormat = cacheSaveFormat)),
               "; fn: ", .messageFunctionFn(functionName), postMess,
               verbose = verbose)
}


.message$useCacheIsFALSE <- function(nestLevel, functionName, useCache, verbose) {
  nestedLev <- max(0, as.numeric(nestLevel)) ## nestedLev >= 0
  spacing <- paste(collapse = "", rep("  ", nestedLev))
  messageCache(spacing, "useCache is ", useCache,
               "; skipping Cache on function ", functionName,
               if (nestedLev > -1) paste0(" (currently running nested Cache level ", nestedLev + 1, ")"),
               verbose = verbose
  )
}

.message$cacheIdSameTxt <- "cacheId is same as calculated hash"
.message$cacheIdNotSameTxt <- function(cacheId)
  paste0("cacheId is not same as calculated hash. Manually searching for cacheId:", cacheId)

.message$cacheIdNotAssessed <- function(cacheId)
  paste0("cacheId passed to override automatic digesting; using ", cacheId)

.message$SavingToCacheTxt <- function(isBig, userTags, functionName, cacheId, otsObjSize, osMess) {
  if (missing(otsObjSize))
    otsObjSize <- objectSizeGetFromUserTags(userTags)
  if (missing(isBig))
    isBig <- isTRUE(otsObjSize > .objectSizeMinForBig)
  if (missing(osMess))
    osMess <- .cacheMessageObjectSize(otsObjSize, isBig)

  c(
    "Saving ", "large "[isBig], "object (fn: ", .messageFunctionFn(functionName),
    ", cacheId: ", cacheId, ") to Cache", ": "[isBig],
    osMess
  )
}

.message$IndentUpdate <- function(nchar = .message$IndentDefault, envir = parent.frame(), ns = "reproducible") {
  val <- paste0(rep(" ", nchar), collapse = "")
  .message$PreProcessIndent <- paste0(.message$PreProcessIndent, val)
  on.exit2(# withr::defer(
    envir = envir,
    expr =
      {
        .message$PreProcessIndent <- gsub(paste0(val, "$"), "", .message$PreProcessIndent)
      }
  )
}

.message$IndentRevert <- function(nchar = .message$IndentDefault, envir = parent.frame(), ns = "reproducible") {
  val <- paste0(rep(" ", nchar), collapse = "")
  .message$PreProcessIndent <- gsub(paste0(val, "$"), "", .message$PreProcessIndent)
  # browser()

  # Base R equivalent of withr::deferred_clear(envir = envir)
  if (!is.null(attr(envir, "handlers"))) {
    attr(envir, "handlers") <- NULL   # or set to list(), see note below
  }

  # withr::deferred_clear(envir = envir)
}

.spaceTmpChar <- "spAcE"

.message$noSimilarCacheTxt <- function(functionName) {
  paste0("There is no similar item in the cachePath ",
         if (!is.null(functionName))
           paste0("of ",  .messageFunctionFn(functionName), " ") else "")
}

.message$defunct <- function(arg)
  paste0(arg, " is a defunct argument; please remove it from this Cache call")



.message$CacheTimings <- function(verbose) {
  if (verbose > 3) {
    assign("cacheTimings", .reproEnv$verboseTiming, envir = .reproEnv)
    messageDF(.reproEnv$verboseTiming, colour = "blue", verbose = verbose, verboseLevel = 3)
    messageCache("This object is also available from .reproEnv$cacheTimings",
                 verbose = verbose, verboseLevel = 3
    )
    if (exists("verboseTiming", envir = .reproEnv)) {
      rm("verboseTiming", envir = .reproEnv)
    }
  }
}

.message$CacheRequiresFUNtxt <- function() {
  "Cache requires the FUN argument"
}

messageStripColor <- function(o)
  gsub("\033.[[:digit:]]{2}[[:alpha:]]", "", o)

.txtUnableToAccessIndex <- "unable to access index"

cliCol <- function(col) {
  if (!startsWith(col, "col_"))
    col <- paste0("col_", col)
  getFromNamespace(col, asNamespace("cli"))
}
# cliCol <- function(col) {
#   if (!startsWith(col, "col_")) {
#     repNS <- asNamespace("reproducible")
#     if (exists(col, envir = repNS))
#       return(get(col, envir = repNS))
#     col <- paste0("col_", col)
#   }
#   getFromNamespace(col, asNamespace("cli"))
#
# }

paddDFInitial <- function(outMess, rows = 1:2, .spaceTmpChar, colour) {
  for (r in rows) {
    if (isTRUE(grepl(" +", outMess[r]))) {
      numInitialSpaces <- length(gregexpr(" ", strsplit(outMess[r], split = "\\S")[[1]][1])[[1]])
      # Has the "class" 2nd row
      outMess[r] <- gsub("^ +", paste(rep(.spaceTmpChar, numInitialSpaces), collapse = ""), outMess[r])
    }
  }
  outMess
}


._prepInputsMetadata <- "._prepInputsMetadata"


.txtDownloadFailedFn <- function(pkg)
  paste0("Download failed; try rerunning after installing ", pkg, " package.")

singularPlural <- function (singPlur, l, v) {
  if (!missing(l)) {
    out <- singPlur[(length(l) > 1) + 1]
  }
  if (!missing(v)) {
    out <- singPlur[(v > 1) + 1]
  }
  out
}

isAre <- function (l, v) {
  singularPlural(c("is", "are"), l, v)
}


.txtPreDigest <- "preDigest"

.message$BecauseOfLossOfColumn <- function(targetFileWithDP) {
  paste0("Because of loss of column, resaving the same object as an RDS ",
         "file at: \n", targetFileWithDP, "\nIt will have to be loaded with
                    `readRDS('",targetFileWithDP,"') |> sf::st_as_sf()`")
}

.message$dashes <- "----------------------"

.message$NoPrefix <- "._noPrefix"

# colr <- function(..., digit = 32) paste0("\033[", digit, "m", paste0(...), "\033[39m")
# purple <- function(...) colr(..., digit = "38;5;129m")
# black <- function(...) colr(..., digit = 30)
# green2 <- function(...) colr(..., digit = 38)
# cyan <- function(...) colr(..., digit = 29)
# red <- function(...) colr(..., digit = 31)
# green <- function(...) colr(..., digit = 32)
# yellow <- function(...) colr(..., digit = 33)
# blue <- function(...) colr(..., digit = 34)
# turquoise <- function(...) colr(..., digit = 36)
# greyLight <- function(...) colr(..., digit = 90)


# cli::ansi_trimws is too slow. This is a copilot rewrite
# ansi_trimws_pureR <- function(x, which = c("both", "left", "right")) {
#   if (!is.character(x)) x <- as.character(x)
#   which <- match.arg(which)
#
#   # Ensure UTF-8
#   x <- enc2utf8(x)
#   if (!length(x)) return(x)
#
#   # Regex to remove ANSI escape sequences
#   ansi_regex <- "\033\\[[0-9;]*[A-Za-z]"
#
#   # Strip ANSI for whitespace detection
#   xs <- gsub(ansi_regex, "", x, perl = TRUE)
#   nxs <- nchar(xs)
#
#   # Compute left/right trims
#   sl <- if (which %in% c("both", "left")) nxs - nchar(sub("^\\s+", "", xs)) else integer(length(x))
#   rl <- if (which %in% c("both", "right")) nxs - nchar(sub("\\s+$", "", xs)) else integer(length(x))
#
#   # Vectorized substring
#   if (any(sl > 0L | rl > 0L)) {
#     start <- 1L + sl
#     end <- nchar(x) - rl
#     x <- substring(x, start, end)
#   }
#
#   x
# }

utils::globalVariables(c(
  ".", "artifact", "createdDate", "deeperThan3", "differs", "fun", "hash",
  "i.hash", "iden", "N", "tag", "tagKey", "tagValue"
))

.reproEnv <- new.env(parent = asNamespace("reproducible"))

#' Saves a wide variety function call outputs to disk and optionally RAM, for recovery later
#'
#' @description
#' \if{html}{\figure{lifecycle-maturing.svg}{options: alt="maturing"}}
#'
#' A function that can be used to wrap around other functions to cache function calls
#' for later use. This is normally most effective when the function to cache is
#' slow to run, yet the inputs and outputs are small. The benefit of caching, therefore,
#' will decline when the computational time of the "first" function call is fast and/or
#' the argument values and return objects are large. The default setting (and first
#' call to Cache) will always save to disk. The 2nd call to the same function will return
#' from disk, unless `options("reproducible.useMemoise" = TRUE)`, then the 2nd time
#' will recover the object from RAM and is normally much faster (at the expense of RAM use).
#'
#' @details
#'
#' There are other similar functions in the R universe.
#' This version of Cache has been used as part of a robust continuous workflow approach.
#'  As a result, we have tested it with many "non-standard" R objects (e.g., `RasterLayer`,
#' `Spat*` objects) and environments (which are always unique, so do not cache readily).
#'
#' This version of the `Cache` function accommodates those four special,
#' though quite common, cases by:
#' \enumerate{
#'   \item converting any environments into list equivalents;
#'   \item identifying the dispatched S4 method (including those made through
#'         inheritance) before hashing so the correct method is being cached;
#'   \item by hashing the linked file, rather than the raster object.
#'         Currently, only file-backed `Raster*` or `Spat*` objects are digested
#'         (e.g., not `ff` objects, or any other R object where the data
#'         are on disk instead of in RAM);
#'   \item Uses [digest::digest()]
#'         This is used for file-backed objects as well.
#'   \item Cache will save arguments passed by user in a hidden environment. Any
#'         nested Cache functions will use arguments in this order: 1) actual arguments
#'         passed at each Cache call; 2) any inherited arguments from an outer Cache
#'         call; 3) the default values of the Cache function. See section on *Nested Caching*.
#' }
#'
#' `Cache` will add a tag to the entry in the cache database called `accessed`,
#' which will assign the time that it was accessed, either read or write.
#' That way, cached items can be shown (using `showCache`) or removed (using
#' `clearCache`) selectively, based on their access dates, rather than only
#' by their creation dates. See example in [clearCache()].
#'
#' @section Nested Caching:
#' Commonly, Caching is nested, i.e., an outer function is wrapped in a `Cache`
#' function call, and one or more inner functions are also wrapped in a `Cache`
#' function call. A user *can* always specify arguments in every Cache function
#' call, but this can get tedious and can be prone to errors. The normal way that
#' \R handles arguments is it takes the user passed arguments if any, and
#' default arguments for all those that have no user passed arguments. We have inserted
#' a middle step. The order or precedence for any given `Cache` function call is
#' 1. user arguments, 2. inherited arguments, 3. default arguments. At this time,
#' the top level `Cache` arguments will propagate to all inner functions unless
#' each individual `Cache` call has other arguments specified, i.e., "middle"
#' nested `Cache` function calls don't propagate their arguments to further "inner"
#' `Cache` function calls.  See example.
#'
#' `userTags` is unique of all arguments: its values will be appended to the
#' inherited `userTags`.
#'
#' @section quick:
#' The `quick` argument is attempting to sort out an ambiguity with character strings:
#' are they file paths or are they simply character strings. When `quick = TRUE`,
#' `Cache` will treat these as character strings; when `quick = FALSE`,
#' they will be attempted to be treated as file paths first; if there is no file, then
#' it will revert to treating them as character strings. If user passes a
#' character vector to this, then this will behave like `omitArgs`:
#' `quick = "file"` will treat the argument `"file"` as character string.
#'
#' The most often encountered situation where this ambiguity matters is in arguments about
#' filenames: is the filename an input pointing to an object whose content we want to
#' assess (e.g., a file-backed raster), or an output (as in saveRDS) and it should not
#' be assessed. If only run once, the output file won't exist, so it will be treated
#' as a character string. However, once the function has been run once, the output file
#' will exist, and `Cache(...)` will assess it, which is incorrect. In these cases,
#' the user is advised to use `quick = "TheOutputFilenameArgument"` to
#' specify the argument whose content on disk should not be assessed, but whose
#' character string should be assessed (distinguishing it from `omitArgs =
#' "TheOutputFilenameArgument"`, which will not assess the file content nor the
#' character string).
#'
#' This is relevant for objects of class `character`, `Path` and
#' `Raster` currently. For class `character`, it is ambiguous whether
#' this represents a character string or a vector of file paths. If it is known
#' that character strings should not be treated as paths, then `quick =
#' TRUE` is appropriate, with no loss of information. If it is file or
#' directory, then it will digest the file content, or `basename(object)`.
#' For class `Path` objects, the file's metadata (i.e., filename and file
#' size) will be hashed instead of the file contents if `quick = TRUE`. If
#' set to `FALSE` (default), the contents of the file(s) are hashed. If
#' `quick = TRUE`, `length` is ignored. `Raster` objects are
#' treated as paths, if they are file-backed.
#'
#' @section Caching Speed:
#' Caching speed may become a critical aspect of a final product. For example,
#' if the final product is a shiny app, rerunning the entire project may need
#' to take less then a few seconds at most.
#' There are 3 arguments that affect `Cache` speed: `quick`, `length`, and `algo`.
#' `quick` is passed to `.robustDigest`, which currently
#' only affects `Path` and `Raster*` class objects.
#' In both cases, `quick` means that little or no disk-based information will be assessed.
#'
#'
#' @section Filepaths:
#' If a function has a path argument, there is some ambiguity about what should be
#' done. Possibilities include:
#' \enumerate{
#'   \item hash the string as is (this will be very system specific, meaning a
#'         `Cache` call will not work if copied between systems or directories);
#'   \item hash the `basename(path)`;
#'   \item hash the contents of the file.
#' }
#' If paths are passed in as is (i.e,. character string), the result will not be predictable.
#' Instead, one should use the wrapper function `asPath(path)`, which sets the
#' class of the string to a `Path`, and one should decide whether one wants
#' to digest the content of the file (using `quick = FALSE`),
#' or just the filename (`(quick = TRUE)`). See examples.
#'
#' @section Stochasticity or randomness:
#' In general, it is expected that caching will only be used when randomness is not
#' desired, e.g., `Cache(rnorm(1))` is unlikely to be useful in many cases. However,
#' `Cache` captures the call that is passed to it, leaving all functions unevaluated.
#' As a result `Cache(glm, x ~ y, rnorm(1))` will not work as a means of forcing
#' a new evaluation each time, as the `rnorm(1)` is not evaluated before the call
#' is assessed against the cache database. To force a new call each time, evaluate
#' the randomness prior to the Cache call, e.g., `ran = rnorm(1)` then pass this
#' to `.cacheExtra`, e.g., `Cache(glm, x ~ y, .cacheExtra = ran)`
#'
#' @section `drv` and `conn`:
#' By default, `drv` uses an SQLite database. This can be sufficient for most cases.
#' However, if a user has dozens or more cores making requests to the Cache database,
#' it may be insufficient. A user can set up a different database backend, e.g.,
#' PostgreSQL that can handle multiple simultaneous read-write situations. See
#' \url{https://github.com/PredictiveEcology/SpaDES/wiki/Using-alternate-database-backends-for-Cache}.
#'
#'
#' @section `useCache`:
#' Logical or numeric. If `FALSE` or `0`, then the entire Caching
#' mechanism is bypassed and the
#' function is evaluated as if it was not being Cached. Default is
#' `getOption("reproducible.useCache")`), which is `TRUE` by default,
#' meaning use the Cache mechanism. This may be useful to turn all Caching on or
#' off in very complex scripts and nested functions. Increasing levels of numeric
#' values will cause deeper levels of Caching to occur (though this may not
#' work as expected in all cases). The following is no longer supported:
#' Currently, only implemented
#' in `postProcess`: to do both caching of inner `cropInputs`, `projectInputs`
#' and `maskInputs`, and caching of outer `postProcess`, use
#' `useCache = 2`; to skip the inner sequence of 3 functions, use `useCache = 1`.
#' For large objects, this may prevent many duplicated save to disk events.
#'
#' If `useCache = "overwrite"`
#' (which can be set with `options("reproducible.useCache" =
#' "overwrite")`), then the function invoke the caching mechanism but will purge
#' any entry that is matched, and it will be replaced with the results of the
#' current call.
#'
#' If `useCache = "devMode"`: The point of this mode is to facilitate using the Cache when
#' functions and datasets are continually in flux, and old Cache entries are
#' likely stale very often. In `devMode`, the cache mechanism will work as
#' normal if the Cache call is the first time for a function OR if it
#' successfully finds a copy in the cache based on the normal Cache mechanism.
#' It *differs* from the normal Cache if the Cache call does *not* find a copy
#' in the `cachePath`, but it does find an entry that matches based on
#' `userTags`. In this case, it will delete the old entry in the `cachePath`
#' (identified based on matching `userTags`), then continue with normal `Cache`.
#' For this to work correctly, `userTags` must be unique for each function call.
#' This should be used with caution as it is still experimental. Currently, if
#' `userTags` are not unique to a single entry in the cachePath, it will
#' default to the behaviour of `useCache = TRUE` with a message. This means
#' that `"devMode"` is most useful if used from the start of a project.
#'
#' @section `useCloud`:
#' This is experimental and there are many conditions under which this is known
#' to not work correctly. This is a way to store all or some of the local Cache in the cloud.
#' Currently, the only cloud option is Google Drive, via \pkg{googledrive}.
#' For this to work, the user must be or be able to be authenticated
#' with `googledrive::drive_auth`. The principle behind this
#' `useCloud` is that it will be a full or partial mirror of a local Cache.
#' It is not intended to be used independently from a local Cache. To share
#' objects that are in the Cloud with another person, it requires 2 steps. 1)
#' share the `cloudFolderID$id`, which can be retrieved by
#' `getOption("reproducible.cloudFolderID")$id` after at least one Cache
#' call has been made. 2) The other user must then set their  `cacheFolderID` in a
#' `Cache\(..., reproducible.cloudFolderID = \"the ID here\"\)` call or
#' set their option manually
#' `options\(\"reproducible.cloudFolderID\" = \"the ID here\"\)`.
#'
#' If `TRUE`, then this Cache call will download
#'   (if local copy doesn't exist, but cloud copy does exist), upload
#'   (local copy does or doesn't exist and
#'   cloud copy doesn't exist), or
#'   will not download nor upload if object exists in both. If `TRUE` will be at
#'   least 1 second slower than setting this to `FALSE`, and likely even slower as the
#'   cloud folder gets large. If a user wishes to keep "high-level" control, set this to
#'   `getOption("reproducible.useCloud", FALSE)` or
#'   `getOption("reproducible.useCloud", TRUE)` (if the default behaviour should
#'   be `FALSE` or `TRUE`, respectively) so it can be turned on and off with
#'   this option. NOTE: *This argument will not be passed into inner/nested Cache calls.*)
#'
#' Two character values are also accepted, intended for separating developer
#' and user roles when sharing a cloud-cache folder:
#'
#' - `"push"` is equivalent to `TRUE` (developer role) -- bidirectional;
#'   downloads on a cloud hit, uploads on a miss.
#' - `"pull"` is read-only (user role) -- downloads on a cloud hit, but never
#'   uploads. If the local cache already has the object, the cloud is not
#'   consulted at all (the Google Drive listing is deferred until after the
#'   local lookup fails). When neither local nor cloud has the object, the
#'   call falls back to a normal local-only Cache run.
#'
#' @section Object attributes:
#' Users should be cautioned that object attributes may not be preserved, especially
#' in the case of objects that are file-backed, such as `Raster` or `SpatRaster` objects.
#' If a user needs to keep attributes, they may need to manually re-attach them to
#' the object after recovery. With the example of `SpatRaster` objects, saving
#' to disk requires `terra::wrap` if it is a memory-backed object. When running
#' `terra::unwrap` on this object, any attributes that a user had added are lost.
#'
#' @section `sideEffect`:
#' This feature is now deprecated. Do not use as it is ignored.
#'
#'
#'
#' @note As indicated above, several objects require pre-treatment before
#' caching will work as expected. The function `.robustDigest` accommodates this.
#' It is an S4 generic, meaning that developers can produce their own methods for
#' different classes of objects. Currently, there are methods for several types
#' of classes. See [.robustDigest()].
#'
#' @include cache-helpers.R
#' @include robustDigest.R
#'
#' @param FUN Either a function (e.g., `rnorm`), a function call (e.g., `rnorm(1)`),
#'             or an unevaluated function call (e.g., using `quote()`).
#'
#' @param ... Arguments passed to `FUN`, if `FUN` is not an expression.
#'
#' @param .objects Character vector of objects to be digested. This is only applicable
#'                if there is a list, environment (or similar) with named objects
#'                within it. Only this/these objects will be considered for caching,
#'                i.e., only use a subset of
#'                the list, environment or similar objects. In the case of nested list-type
#'                objects, this will only be applied outermost first.
#'
#' @param .cacheExtra A an arbitrary R object that will be included in the `CacheDigest`,
#'       but otherwise not passed into the `FUN`. If the user supplies a named list, then
#'       `Cache` will report which individual elements of `.cacheExtra` have changed
#'       when `options("reproducible.showSimilar" = TRUE)`. This can allow a user
#'       more control and understanding for debugging.
#'
#' @param .cacheChaining A logical or a the name of a function. If `TRUE`, then
#'   the current `Cache` call will evaluate the function "outside" the `Cache` call
#'   (via `sys.function(-1)`) and
#'   attach the `digest` of that outer function to the entry for this `Cache` call. This
#'   will then be used by any subsequent `Cache` call within the same function.
#'   If the outer function is unchanged, and there is one or more objects that had
#'   been returned by a previous `Cache` call,
#'   then those objects will not be digested; rather their `cacheId` tag will be used
#'   in place of a new `digest`. This *should* cause no change in Caching outcomes,
#'   and it should be faster in cases where there are several `Cache` calls within
#'   the same function. If `FALSE` (current default), then this feature is
#'   not used. If set to `NULL` (i.e., unset, the current default), then it will
#'   not use cache chaining, but it will attach more information to the Cache entries
#'   for each `cacheId`, as well as new entries for `"surroundingFunction"` digest,
#'   so that if a user switches to `.cacheChaining = TRUE`, then it will be able
#'   to begin using cache chaining without needing to rerun the calls again. Can be set by an `option`.
#' @param .functionName A an arbitrary character string that provides a name that is different
#'       than the actual function name (e.g., "rnorm") which will be used for messaging. This
#'       can be useful when the actual function is not helpful for a user, such as `do.call`.
#'
#' @param outputObjects Optional character vector indicating which objects to
#'                      return. This is only relevant for list, environment (or similar) objects
#'
#' @param algo The digest algorithm to use. Default `xxhash64` (see [digest::digest()] for others).
#'
#' @param cacheRepo Same as `cachePath`, but kept for backwards compatibility.
#'
#' @param cachePath A repository used for storing cached objects.
#'                  This is optional if `Cache` is used inside a SpaDES module.
#' @param length Numeric. If the element passed to Cache is a `Path` class
#'        object (from e.g., `asPath(filename)`) or it is a `Raster` with
#'        file-backing, then this will be
#'        passed to `digest::digest`, essentially limiting the number of bytes
#'        to digest (for speed). This will only be used if `quick = FALSE`.
#'        Default is `getOption("reproducible.length")`, which is set to `Inf`.
#'
#' @param compareRasterFileLength Being deprecated; use `length`.
#'
#' @param omitArgs Optional. A character vector of argument names in `FUN` to
#'   omit from the cache digest, or `TRUE` to omit *every* captured argument
#'   (the digest is then based on `FUN` itself -- including its body, so a
#'   meaningful edit to the function source still busts the cache -- and on
#'   `.cacheExtra`). Useful when the developer wants the cache to be
#'   insensitive to the function's inputs and pin freshness via `.cacheExtra`
#'   instead.
#'
#' @param classOptions Optional list. This will pass into `.robustDigest` for
#'        specific classes. Should be options that the `.robustDigest` knows what
#'        to do with.
#'
#' @param debugCache Character or Logical. Either `"complete"` or `"quick"` (uses
#'        partial matching, so "c" or "q" work). `TRUE` is equivalent to `"complete"`.
#'        If `"complete"`, then the returned object from the Cache
#'        function will have two attributes, `debugCache1` and `debugCache2`,
#'        which are the entire `list(...)` and that same object, but after all
#'        `.robustDigest` calls, at the moment that it is digested using
#'        `digest`, respectively. This `attr(mySimOut, "debugCache2")`
#'        can then be compared to a subsequent call and individual items within
#'        the object `attr(mySimOut, "debugCache1")` can be compared.
#'        If `"quick"`, then it will return the same two objects directly,
#'        without evalutating the `FUN(...)`.
#'
#' @param makeCopy Now deprecated. Ignored if used.
#'
#' @param userTags A character vector with descriptions of the Cache function call. These
#'   will be added to the Cache so that this entry in the Cache can be found using
#'   `userTags` e.g., via [showCache()].
#'
#' @param notOlderThan A time. Load an object from the Cache if it was created after this.
#'
#' @param quick Logical or character. If `TRUE`,
#'        no disk-based information will be assessed, i.e., only
#'        memory content. See Details section about `quick` in [Cache()].
#'
#' @param verbose Numeric, -1 silent (where possible), 0 being very quiet,
#'        1 showing more messaging, 2 being more messaging, etc.
#'        Default is 1. Above 3 will output much more information about the internals of
#'        Caching, which may help diagnose Caching challenges. Can set globally with an
#'        option, e.g., `options('reproducible.verbose' = 0) to reduce to minimal`
#'
#' @param cacheId Character string. If passed, this will override the calculated hash
#'        of the inputs, and return the result from this `cacheId` in the `cachePath`.
#'        Setting this is equivalent to manually saving the output of this function, i.e.,
#'        the object will be on disk, and will be recovered in subsequent
#'        This may help in some particularly finicky situations
#'        where `Cache` is not correctly detecting unchanged inputs. This will guarantee
#'        the object will be identical each time; this may be useful in operational code.
#'
#' @param useCache Logical, numeric or `"overwrite"` or `"devMode"`. See details.
#'
#' @param useCloud Logical (`TRUE` / `FALSE` / `NULL`) or one of `"pull"` /
#'   `"push"`. See Details.
#' @param cacheSaveFormat Character string: currently either `qs` or `rds`. Defaults to
#'    `getOption("reproducible.cacheSaveFormat")`. `qs` may be faster but appears to have
#'    narrower range of conditions that work; `rds` is safer, and may be slower.
#'
#' @param cloudFolderID A googledrive dribble of a folder, e.g., using `drive_mkdir()`.
#'   If left as `NULL`, the function will create a cloud folder with name from last
#'   two folder levels of the `cachePath` path, :
#'   `paste0(basename(dirname(cachePath)), "_", basename(cachePath))`.
#'   This `cloudFolderID` will be added to `options("reproducible.cloudFolderID")`,
#'   but this will not persist across sessions. If this is a character string, it will
#'   treat this as a folder name to create or use on GoogleDrive.
#'
#' @param showSimilar A logical or numeric. Useful for debugging.
#'        If `TRUE` or `1`, then if the Cache
#'        does not find an identical archive in the `cachePath`, it will report (via message)
#'        the next most recent similar archive, and indicate which argument(s) is/are different.
#'        If a number larger than `1`, then it will report the N most recent similar archived
#'        objects.
#'
#' @param drv If using a database backend, `drv` must be an object that
#'   inherits from `DBIDriver` (e.g., `RSQLite::SQLite`).
#' @param conn an optional `DBIConnection` object, as returned by `dbConnect()`.
#'
#' @return Returns the value of the
#' function call or the cached version (i.e., the result from a previous call
#' to this same cached function with identical arguments).
#'
#' @seealso [showCache()], [clearCache()], [keepCache()],
#'   [CacheDigest()] to determine the digest of a given function or expression,
#'   as used internally within `Cache`, [movedCache()], [.robustDigest()], and
#'   for more advanced uses there are several helper functions,
#'   e.g., [rmFromCache()], [CacheStorageDir()]
#'
#' @author Eliot McIntire
#' @importFrom digest digest
#' @importFrom data.table setDT := setkeyv .N .SD
#' @importFrom utils object.size tail
#' @importFrom methods formalArgs
#' @export
#' @rdname Cache
#'
#' @example inst/examples/example_Cache.R
#'
CacheV2 <-
  function(FUN, ..., notOlderThan = NULL,
           .objects = NULL, .cacheExtra = NULL, .functionName = NULL,
           outputObjects = NULL, # nolint
           algo = "xxhash64", cacheRepo = NULL,
           cachePath = NULL,
           length = getOption("reproducible.length", Inf),
           compareRasterFileLength, userTags = c(),
           omitArgs = NULL,
           classOptions = list(), debugCache = character(),
           # sideEffect = FALSE,
           makeCopy = FALSE,
           quick = getOption("reproducible.quick", FALSE),
           verbose = getOption("reproducible.verbose", 1), cacheId = NULL,
           useCache = getOption("reproducible.useCache", TRUE),
           useCloud = FALSE,
           cloudFolderID = NULL,
           showSimilar = getOption("reproducible.showSimilar", FALSE),
           drv = getDrv(getOption("reproducible.drv", NULL)),
           conn = getOption("reproducible.conn", NULL)) {

    .Defunct("Cache")
  }

#' @keywords internal
.formalsCache <- formals(Cache)[-(1:2)]

#' @keywords internal
.formalscache2 <- formals(cache2)[-(1:2)]

#' @keywords internal
.formalsCache[c("compareRasterFileLength", "digestPathContent")] <- NULL

#' @keywords internal
.namesCacheFormals <- names(.formalsCache)[]

#' @keywords internal
.namescache2Formals <- names(.formalscache2)[]

#' @keywords internal
.namesPostProcessFormals <- function() {
  c(
    "x", "filename1", "writeTo", "studyArea", "rasterToMatch",
    "overwrite", "useSAcrs", "useCache", "verbose"
  )
}


#' @keywords internal
.namesCacheFormalsSendToBoth <- intersect("verbose", names(.formalsCache)[])


#' @keywords internal
.unlistToCharacter <- function(l, max.level = 1) {
  if (max.level > 0) {
    lapply(l, function(l1) {
      if (is.character(l1)) {
        l1
      } else {
        if (is.list(l1)) {
          .unlistToCharacter(l1, max.level = max.level - 1)
        } else {
          "not list"
        }
      }
    })
  } else {
    if (is.list(l)) {
      unlist(l)
    } else {
      "not list2"
    }
    # unlist(l)
  }
}



#' Write to cache repository, using `future::future`
#'
#' This will be used internally if `options("reproducible.futurePlan" = TRUE)`.
#' This is still experimental.
#'
#' @param written Integer. If zero or positive then it needs to be written still.
#'                Should be 0 to start.
#' @param outputToSave The R object to save to repository
#' @param cachePath The file path of the repository
#' @param userTags Character string of tags to attach to this `outputToSave` in
#'                 the `CacheRepo`
#'
#' @export
#' @inheritParams Cache
#' @inheritParams saveToCache
#' @return
#' Run for its side effect.
#' This will add the `objectToSave` to the cache located at `cachePath`,
#' using `cacheId` as its id, while
#' updating the database entry. It will do this using the future package, so it is
#' written in a future.
writeFuture <- function(written, outputToSave, cachePath, userTags,
                        drv = getDrv(getOption("reproducible.drv", NULL)),
                        conn = getOption("reproducible.conn", NULL),
                        cacheId, linkToCacheId = NULL,
                        verbose = getOption("reproducible.verbose")) {
  counter <- 0
  if (!CacheIsACache(cachePath = cachePath, drv = drv, conn = conn, verbose = verbose)) {
    stop("That cachePath does not exist")
  }

  if (missing(cacheId)) {
    cacheId <- .robustDigest(outputToSave)
  }
  output <- saveToCache(
    cachePath = cachePath, drv = drv, userTags = userTags,
    conn = conn, obj = outputToSave, cacheId = cacheId,
    linkToCacheId = linkToCacheId
  )
  saved <- cacheId

  return(saved)
}


isDollarSqBrPkgColon <- function(args) {
  ret <- FALSE
  if (length(args) == 3 || length(args) == 1) { # i.e., only possible if it is just b$fun or stats::runif, not stats::runif(1) or b$fun(1)
    ret <- isTRUE(any(try(grepl("^\\$|\\[|\\:\\:", args)[1], silent = TRUE)))
  }
  ret
}

isPkgColon <- function(args) {
  ret <- FALSE
  if (length(args) == 3) { # i.e., only possible if it is just b$fun or stats::runif, not stats::runif(1) or b$fun(1)
    ret <- isTRUE(any(try(grepl("\\:\\:", args)[1], silent = TRUE)))
  }
  ret
}

isDollarOnlySqBr <- function(args) {
  ret <- FALSE
  if (length(args) == 3) { # i.e., only possible if it is just b$fun or stats::runif, not stats::runif(1) or b$fun(1)
    ret <- isTRUE(any(try(grepl("^\\$|\\[", args)[1], silent = TRUE)))
  }
  ret
}

recursiveEvalNamesOnly <- function(args, envir = parent.frame(), outer = TRUE, recursive = TRUE) {

  needsEvaling <- (length(args) > 1) || (length(args) == 1 && is.call(args)) # second case is fun() i.e., no args
  if (isTRUE(needsEvaling)) {
    if (is.call(args[[1]])) { # e.g., a$fun, stats::runif
      args[[1]] <- eval(args[[1]], envir)
    }

    isStandAlone <- FALSE
    if (length(args) == 3) { # e.g., status::runif or fun(1, 2); second case requires subsequent efforts
      if (!is.function(args[[1]])) { # this removes fun(1, 2) case
        isStandAlone <- isDollarSqBrPkgColon(args[[1]])
      }
    } else if (length(args[[1]]) == 3) {
      isStandAlone <- isDollarSqBrPkgColon(args[[1]])
    }

    if (identical(quote(`function`), args[[1]])) # if it is function definition, then leave the inside unevaluated
      isStandAlone <- TRUE

    if (identical(as.name("<-"), args[[1]])) {
      args <- as.list(args[-(1:2)])[[1]]
    }

    if (identical(quote(parse), args[[1]])) {
      args <- eval(args)
    }

    if (!isTRUE(recursive)) {
      isStandAlone <- TRUE
    }

    if (!any(isStandAlone)) {
      out <- lapply(args, function(xxxx) {
        if (is.name(xxxx)) {
          # exists(xxxx, envir = envir, inherits = FALSE)
          if (exists(xxxx, envir)) { # looks like variables that are in ... in the `envir` are not found; would need .whereInStack
            evd <- try(eval(xxxx, envir), silent = TRUE)
            isPrim <- is.primitive(evd)
            if (isPrim) {
              eval(xxxx)
            } else {
              isQuo <- is(evd, "quosure")
              if (isQuo) {
                evd <- rlang::eval_tidy(evd)
              }
              if (is(evd, "list")) {
                evd <- recursiveEvalNamesOnly(evd, envir, outer = FALSE)
              }
              evd
            }
          } else {
            ret <- xxxx
            ret
          }
        } else {
          if (is.call(xxxx)) {
            if (identical(quote(eval), xxxx[[1]])) { # basically "eval" should be evaluated
              message(
                "There is an `eval` call in a chain of calls for Cache; ",
                "\n  eval is evaluated before Cache which may be undesired. ",
                "\n  Perhaps use `do.call` if the evaluation should not occur prior to Cache"
              )
              ret <- eval(xxxx, envir = envir)
            } else {
              ret <- recursiveEvalNamesOnly(xxxx, envir, outer = FALSE)
            }
          } else {
            ret <- xxxx
          }
          ret
        }
      })



      args <- as.call(out)
      # args <- if (isTRUE(outer)) try(as.call(out)) else out
      if (is.function(args[[1]])) {
        args <- match_call_primitive(args[[1]], args, expand.dots = FALSE, envir = envir)
        args[[1]] <- getMethodAll(args, envir)
      }
    } else {
      # paths$inputPath comes here to be evaluated to its path
      args <- eval(args, envir)
    }
  } else {
    if (length(args) == 1 && is.name(args)) {
      args <- eval(args, envir)
    }
  }
  args
}


matchCall <- function(FUNcaptured, envir = parent.frame(), fnName) {
  if (length(FUNcaptured) > 1) {
    FUN <- FUNcaptured[[1]]
    args <- as.list(FUNcaptured[-1])
    if (is.call(FUN)) FUN <- eval(FUN, envir)
    if (is.function(FUN)) {
      forms <- if (is.primitive(FUN)) formals(args(FUN)) else formals(FUN)
      if (length(forms) == 0) {
        mc <- list(FUN)
      } else {
        if (is.primitive(FUN)) {
          # Must test for "non-normal non-positional matching", like round and round.POSIXt, ... see ?match.call
          #  can't tell a priori that a primitive doesn't have methods, so must test first.
          #  These will always be in base, so can just get0 directly, which is WAY faster than any other way
          nonPrimMeth <- NULL
          if (!is.null(fnName)) {
            cls <- is(args[[1]])
            # use for loop, so can break out if a method is found quickly
            for (classHere in cls) {
              nonPrimMeth <- get0(paste0(fnName, ".", classHere))
              if (!is.null(nonPrimMeth)) break
            }
          }
          if (length(nonPrimMeth)) {
            args2 <- formals(nonPrimMeth)
          } else {
            args2 <- forms
          }
          args2[seq(args)] <- args
          args2 <- args2[seq_along(args)] # chop off any trailing args
          mc <- append(list(FUN), args2)
        } else {
          mc <- match.call(FUN, FUNcaptured)
        }
      }
    } else {
      mc <- FUNcaptured
    }
  } else {
    mc <- list(FUNcaptured)
  }
  mc
}

#' @importFrom methods .S4methods
#' @importFrom utils getFromNamespace
getMethodAll <- function(FUNcaptured, callingEnv) {
  FUN <- FUNcaptured[[1]]
  if (!is.function(FUN))
    FUN <- tryCatch(eval(FUN, envir = callingEnv),
                    error = function(FU) eval(parse(text = FUN), envir = callingEnv))
  if (isS4(FUN)) {
    functionName <- FUN@generic
    # Not easy to selectMethod -- can't have trailing "ANY" -- see ?selectMethod last
    #  paragraph of "Using findMethod()" which says:
    # "Notice also that the length of the signature must be what the corresponding
    #  package used. If thisPkg had only methods for one argument, only length-1
    # signatures will match (no trailing "ANY"), even if another currently loaded
    # package had signatures with more arguments.
    numArgsInSig <- try(
      {
        suppressWarnings({
          info <- attr(methods::.S4methods(functionName), "info") # from hadley/sloop package s3_method_generic
          # info <- attr(utils::methods(functionName), "info")# from hadley/sloop package s3_method_generic
        })
        max(unlist(lapply(strsplit(rownames(info), split = ","), length)) - 1)
      },
      silent = TRUE
    )
    matchOn <- FUN@signature[seq(numArgsInSig)]

    argsClassesList <- lapply(FUNcaptured, function(x) class(x))
    # argsClasses <- unlist(argsClassesList)#[1]))
    argsClasses <- unlist(unname(argsClassesList[names(argsClassesList) %in% matchOn]))
    missingArgs <- matchOn[!(matchOn %in% names(argsClassesList))]

    missings <- rep("missing", length(missingArgs))
    names(missings) <- missingArgs
    argsClasses <- c(argsClasses, missings)

    argClassesAreCall <- argsClasses %in% "call" # maybe wasn't evaluated enough to know what it is; force eval
    if (any(argClassesAreCall)) {
      argsClasses <- "ANY"
      #whAreCall <- names(argsClasses[argClassesAreCall])
      #argsClasses <- Map(wac = whAreCall, function(wac) is(eval(FUNcaptured[[wac]], envir = callingEnv)))
    } else {
      FUN <- selectMethod(functionName, signature = argsClasses)
    }
    updatedFUN <- TRUE
  } else {
    isS3 <- isS3stdGeneric(FUN)
    if (!is.null(names(isS3))) {
      fnNameInitAlt <- names(isS3)
    }
    if (isS3) {
      updatedFUN <- TRUE
      classes <- is(FUNcaptured[[2]])
      for (cla in classes) {
        FUNposs <- utils::getS3method(fnNameInitAlt, cla, optional = TRUE) # S3 matches on 1st arg: FUNcaptured[[2]]
        if (!is.null(FUNposs)) {
          break
        }
      }

      # if generic fn was not exported, then getS3method won't find it above; try directly in NS
      if (is.null(FUNposs)) {
        envNam <- environmentName(environment(FUN))
        FUNpossGen <- get0(fnNameInitAlt, envir = asNamespace(envNam))
        for (cla in classes) {
          possMeth <- paste0(fnNameInitAlt, ".", cla)
          FUNposs <- try(getFromNamespace(possMeth, ns = envNam), silent = TRUE)
          if (!is(FUNposs, "try-error")) {
            break
          } else {
            FUNposs <- NULL
          }
        }
        if (is.null(FUNposs)) {
          FUNposs <- FUNpossGen
        }
      }

      if (is.null(FUNposs)) {
        FUNposs <- get0(fnNameInitAlt, envir = callingEnv)
        if (is.null(FUNposs) || isS4(FUNposs)) { # there are cases e.g., print that are both S4 & S3; this forces S3
          FUNposs <- get0(paste0(fnNameInitAlt, ".default"), envir = callingEnv)
        }
      }
      FUN <- FUNposs
    }
  }
  FUN
}

formals2 <- function(FUNcaptured) {
  modifiedDots <- as.list(FUNcaptured[-1])
  FUN <- FUNcaptured[[1]]
  modifiedDots <- formals3(FUN, modifiedDots, removeNulls = TRUE)
  modifiedDots
}


formals3 <- function(FUN, modifiedDots = list(), removeNulls = FALSE) {
  forms1 <- formals(FUN) # primitives don't have formals
  if (!is.null(forms1)) {
    forms1 <- forms1[setdiff(names(forms1), "...")]
    if (NROW(forms1)) {
      defaults <- setdiff(names(forms1), names(modifiedDots))
      if (removeNulls) {
        theNulls <- unlist(lapply(forms1[defaults], is.null))
        if (any(theNulls)) {
          defaults <- defaults[!theNulls]
        }
      }

      if (NROW(defaults)) { # there may be some arguments that are not specified

        # get the values of args that are eg. coming from options
        forms1[defaults] <- lapply(forms1[defaults], function(xxx) {
          yyy <- "default"
          if (length(xxx) > 0) {
            if (length(xxx) == 1) {
              if (isTRUE(nchar(xxx) == 0)) {
                yyy <- NULL
              }
            }
          }
          if (!is.null(yyy)) {
            # Some are used by other args, yet are undefined in the args ... because "missing"
            # ex is seq.default() # by is (from - to)/(length.out - 1), but length.out is NULL in args
            # so need try
            yyy <- try(eval(xxx, envir = modifiedDots), silent = TRUE)
            if (is(yyy, "try-error")) {
              yyy <- NULL
            }
          }
          yyy
        })
      }

      # Have to get rid of NULL because CacheDigest
      if (removeNulls) {
        forms1 <- forms1[!unlist(lapply(forms1, is.null))]
      }
      modifiedDots <- modifyList(forms1, modifiedDots)
      forms <- names(forms1)
    }
  }
  modifiedDots
}

# This is taken from Rdpack::S4formals
formals4reproducible <- function (fun, ...) {
  if (!is(fun, "MethodDefinition"))
    fun <- getMethod(fun, ...)
  fff <- fun@.Data
  funbody <- body(fff)
  if (length(funbody) == 3 && identical(funbody[[1]], as.name("{")) &&
      length(funbody[[2]]) == 3 && identical(funbody[[c(2,
                                                        1)]], as.name("<-")) && identical(funbody[[c(2, 2)]],
                                                                                          as.name(".local")) && is.function(funbody[[c(2, 3)]])) {
    formals(funbody[[c(2, 3)]])
  }
  else {
    formals(fff)
  }
}

getFunctionName2 <- function(mc) {
  if (length(mc) > 1) {
    if (identical(as.name("<-"), mc[[1]])) {
      mc <- mc[-(1:2)]
    }
    coloncolon <- .grepSysCalls(list(mc), "^\\$|\\[|\\:\\:")
    coloncoloncolon <- .grepSysCalls(list(mc), "^\\$|\\[|\\:\\:\\:")
    if (length(coloncolon)) { # stats::runif -- has to be first one, not some argument in middle
      if (length(coloncolon) && length(mc) != 3 || length(coloncoloncolon)) { # stats::runif

        #if (any(grepl("^\\$|\\[|\\:\\:", mc)[1])) { # stats::runif -- has to be first one, not some argument in middle
        #  if (any(grepl("^\\$|\\[|\\:\\:", mc[[1]])) && length(mc) != 3) { # stats::runif
        fnNameInit <- deparse(mc[[1]])
      } else {
        fnNameInit <- deparse(mc)
      }
    } else {
      fnNameInit <- deparse(as.list(mc[[1]])[[1]]) # fun() and fun could both be here in first slot
    }
  } else {
    fnNameInit <- deparse(mc)
  }
  fnNameInit
}

#' @importFrom utils modifyList isS3stdGeneric methods
.fnCleanup <- function(FUN, ..., callingFun, FUNcaptured = NULL, CacheMatchedCall,
                       .functionName = NULL, callingEnv = parent.frame(2), .fnCleanup,
                       omitArgs = "") {
  if (is.null(FUNcaptured)) {
    FUNcaptured <- substitute(FUN)
  }

  FUNcapturedOrig <- FUNcaptured

  whCharName <- is.function(FUNcaptured) # this is bad; it means that it was not captured. Happens when user
  #  erroneously does do.call(Cache, args)
  if (all(whCharName %in% TRUE)) {
    stop(
      "It looks like Cache is called incorrectly, possibly something like do.call(Cache, args); \n",
      "Cache should be the outermost function. See examples for correct ways to use Cache"
    )
  }
  # Remove `quote`
  isQuoted <- any(grepl("^quote", FUNcaptured)[1]) # won't work for complicated quote
  if (isQuoted) {
    FUNcaptured <- FUNcaptured[[2]]
  }

  dotsCaptured <- substitute(list(...))
  dotsCaptured <- as.list(dotsCaptured[-1]) # need to remove the `list` on the inside of the substitute

  # Backward compatibility; has no effect now
  userTagsOtherFunctions <- NULL

  if (isDollarSqBrPkgColon(FUNcaptured)) {
    if (isPkgColonFn(FUNcaptured)) {
      FUNcaptured <- eval(FUNcaptured, envir = callingEnv)
    } else if (isPkgColon(FUNcaptured)) { # this is TRUE ONLY if it is *just* b$fun or stats::runif, i.e., not b$fun(1)
      FUNcaptured[[1]] <- eval(FUNcaptured[[1]], envir = callingEnv)
    } else if (isDollarOnlySqBr(FUNcaptured)) {
      FUNcaptured <- eval(FUNcaptured, envir = callingEnv)
    }
  }

  if (length(FUNcaptured) > 1) { # this will cover the cases where previous misses, e.g.,
    if (isDollarSqBrPkgColon(FUNcaptured[[1]])) { # this is TRUE ONLY if it is *just* b$fun(1), stats::runif(1)
      FUNcaptured[[1]] <- eval(FUNcaptured[[1]], envir = callingEnv)
    }
  }

  if (!is.call(FUNcaptured)) { # isDollarSqBrPkgColon(FUNcaptured)) { # turn the rnorm, 1, 2 into rnorm(1, 2)
    FUNcaptured <- as.call(append(list(FUNcaptured), dotsCaptured))
  }

  whCharName <- unlist(lapply(FUNcaptured, function(x) is.call(x) || is.name(x) || is.function(x) || is.character(x)))
  isDoCall <- if (any(whCharName)) {
    isTRUE(grepl("^do\\.call", FUNcaptured[whCharName])[[1]]) ||
      identical(do.call, FUNcaptured[[1]])
  } else {
    FALSE
  }
  needRmList <- FALSE
  fnNameInit <- NULL
  if (isDoCall) {
    mc <- match.call(do.call, FUNcaptured)
    fnNameInit <- deparse(mc$what)
    if (length(mc$args) > 1) {
      argsForWhat <- mc$args[-1]
    } else {
      needRmList <- TRUE
      argsForWhat <- mc$args # mc$args will be a list; needs to be evaluated to be unlisted; do below
    }
    FUNcaptured <- try(as.call(append(list(mc$what), as.list(argsForWhat))))
  }

  isSquiggly <- FALSE
  if (!is.function(FUNcaptured[[1]])) { # e.g., just the name, such as rnorm --> convert to the actual function code
    if (is(FUNcaptured[[1]], "(")) {
      fnNameInit <- "headless"
    }
    FUNcaptured[[1]] <- eval(FUNcaptured[[1]], envir = callingEnv)
  }

  if (length(FUNcaptured) > 1) isSquiggly <- identical(`{`, FUNcaptured[[1]])

  if (isSquiggly) {
    # Get rid of squiggly
    FUNcaptured <- as.list(FUNcaptured[-1]) # [[1]] ... if it has many calls... pipe will be just one; but others will be more
    if (length(FUNcaptured) > 1) {
      stop("Cache can only handle curly braces if all internal code uses base pipe |>; see examples")
    }
    FUNcaptured <- FUNcaptured[[1]]
    FUNcapturedNamesEvaled <- recursiveEvalNamesOnly(FUNcaptured, envir = callingEnv) # deals with e.g., stats::rnorm, b$fun, b[[fun]]
    mc1 <- matchCall(FUNcaptured, envir = callingEnv, fnName = fnNameInit)
    if (is.null(fnNameInit)) {
      fnNameInit <- getFunctionName2(mc1[[1]])
    }
    FUNcapturedNamesEvaled <- matchCall(FUNcapturedNamesEvaled, envir = callingEnv, fnName = fnNameInit)
  } else {
    if (is.null(fnNameInit)) {
      fnNameInit <- getFunctionName2(FUNcapturedOrig)
    }
    if (length(FUNcaptured) > 1) {
      # The next line works for any object that is NOT in a ..., because the
      #   object never shows up in the environment; it is passed through
      # mced <- names(CacheMatchedCall)

      # if (!is.null(unlist(argsToKeep))) {
      FUNcapturedList <- as.list(FUNcaptured[-1])
      nams <- names(FUNcapturedList)
      if (is.null(nams))
        nams <- sapply(seq_along(FUNcapturedList), function(x) paste0(sample(LETTERS, 14), collapse = ""))
      FUNcapturedArgs <- Map(
        ee = FUNcapturedList, nam = nams, function(ee, nam) {

            out <- try(eval(ee, envir = callingEnv), silent = TRUE)
            if (is(out, "try-error")) {
              if (identical(as.name("..."), ee)) {
                out <- "..."
              } else {
                env2 <- try(if (isDollarSqBrPkgColon(ee)) {
                  .whereInStack(ee[[2]])
                } else {
                  .whereInStack(ee)
                }, silent = TRUE)
                if (is(env2, "try-error")) {
                  out <- try(paste(format(ee$destinationPath), collapse = " "), silent = TRUE)
                  if (is(out, "try-error"))
                    stop(env2)
                } else {
                  out <- try(eval(ee, envir = env2), silent = TRUE)
                  if (is(out, "try-error")) {
                    out <- as.character(parse(text = ee))
                  }
                }
              }
            }
          # }

          out
        }) # may be slow as it is evaluating the args
      if (needRmList) { # it has one too many list elements # not sure about the length(out) == 1
        FUNcapturedArgs <- FUNcapturedArgs[[1]]
      }
      # }

      FUNcapturedNamesEvaled <- as.call(append(list(FUNcaptured[[1]]), FUNcapturedArgs))
      FUNcapturedNamesEvaled <- matchCall(FUNcapturedNamesEvaled, callingEnv, fnName = fnNameInit)
                             } else { # this is a function called with no arguments
                               FUNcapturedNamesEvaled <- FUNcaptured
    }
  }



  # Now FUNcaptured will always have at least 1 element, because it is a call


  FUN <- FUNcapturedNamesEvaled[[1]] # This will be wrong if a fn has no args
  if (is.call(FUN)) { # This will only happen if there are no args to FUN e.g., fun()... anything else is a name fun(1)
    FUN <- FUN[[1]]
    FUNcapturedNamesEvaled[[1]] <- FUN
  }

  fnDetails <- list(
    functionName = fnNameInit,
    .FUN = FUN,
    nestLevel = 1
  )

  modifiedDots <- as.list(FUNcapturedNamesEvaled[-1]) # this is prior to filling in with defaults
  if (is.function(FUN)) {
    FUN <- getMethodAll(FUNcapturedNamesEvaled, callingEnv)
    forms <- if (is.primitive(FUN)) formals(args(FUN)) else formals(FUN)
    FUNcapturedNamesEvaled[[1]] <- FUN # may be same if it was a primitive or just a function
    fnDetails$.FUN <- FUN

    if (!is.primitive(FUN) && (length(forms) > 0)) {
      modifiedDots <- formals2(FUNcapturedNamesEvaled) # this gets default values for methods;
    }
  } else {
    # This comes from `CacheDigest(something$something)`
    FUNcapturedNamesEvaled <- append(list(NULL), FUNcaptured) # the first arg is supposed to be a function below; put NULL as placeholder
    forms <- names(FUNcapturedNamesEvaled[-1])
  }

  FUNcapturedNamesEvaled <- checkOverlappingArgs(CacheMatchedCall, forms, dotsCaptured,
                                                 functionName = fnDetails$functionName, FUNcapturedNamesEvaled)

  # # Check for args that are passed to both Cache and the FUN -- if any overlap; pass to both
  # possibleOverlap <- names(formals(args(Cache)))
  # possibleOverlap <- intersect(names(CacheMatchedCall), possibleOverlap)
  # actualOverlap <- intersect(names(forms), possibleOverlap)
  # if (length(actualOverlap) && !identical(list(), dotsCaptured)) { # e.g., useCache, verbose; but if not in dots, then OK because were separate already
  #   message(
  #     "The following arguments are arguments for both Cache and ", fnDetails$functionName, ":\n",
  #     paste0(actualOverlap, collapse = ", "),
  #     "\n...passing to both. If more control is needed, pass as a call, e.g., ",
  #     "Cache(", fnDetails$functionName, "(...))"
  #   )
  #   overlappingArgsAsList <- as.list(CacheMatchedCall)[actualOverlap]
  #   FUNcapturedNamesEvaled <- as.call(append(as.list(FUNcapturedNamesEvaled), overlappingArgsAsList))
  # }

  if (!is.null(.functionName)) {
    fnDetails$functionName <- .functionName
  }

  return(append(fnDetails, list(
    FUN = FUN, matchedCall = FUNcapturedNamesEvaled,
    modifiedDots = modifiedDots, # isDoCall = isDoCall,
    formalArgs = forms,
    userTags = userTagsOtherFunctions
  )))
}


#' Set subattributes within a list by reference
#'
#' Sets only a single element within a list attribute.
#' @param object An arbitrary object
#' @param attr The attribute name (that is a list object) to change
#' @param subAttr The list element name to change
#' @param value The new value
#'
#' @return
#' This sets or updates the `subAttr` element of a list that is located at
#' `attr(object, attr)`, with the `value`. This, therefore, updates a sub-element
#'  of a list attribute and returns that same object with the updated attribute.
#'
#' @export
#' @rdname setSubAttrInList
.setSubAttrInList <- function(object, attr, subAttr, value) {
  .CacheAttr <- attr(object, attr)
  if (is.null(.CacheAttr)) .CacheAttr <- list()
  .CacheAttr[[subAttr]] <- value
  attr(object, attr) <- .CacheAttr
  object
}

#' The exact digest function that `Cache` uses
#'
#' This can be used by a user to pre-test their arguments before running
#' `Cache`, for example to determine whether there is a cached copy.
#'
#'
#' @param ... passed to `.robustDigest`.
#' @param objsToDigest A list of all the objects (e.g., arguments) to be digested
#' @param calledFrom a Character string, length 1, with the function to
#'    compare with. Default is "Cache". All other values may not produce
#'    robust CacheDigest results.
#'
#' @inheritParams Cache
#'
#' @return
#' A list of length 2 with the `outputHash`, which is the digest
#' that Cache uses for `cacheId` and also `preDigest`, which is
#' the digest of each sub-element in `objsToDigest`.
#'
#' @export
#'
#' @examples
#' data.table::setDTthreads(2)
#' a <- Cache(rnorm, 1)
#'
#' # like with Cache, user can pass function and args in a few ways
#' CacheDigest(rnorm(1)) # shows same cacheId as previous line
#' CacheDigest(rnorm, 1) # shows same cacheId as previous line
#'
CacheDigest <- function(objsToDigest, ..., algo = "xxhash64", calledFrom = "CacheDigest",
                        .functionName = NULL, quick = FALSE) {
  FUNcaptured <- substitute(objsToDigest)
  # origFUN <- quote(objsToDigest)
  fromCache <- identical(FUNcaptured, as.name("toDigest"))
  dots <- list(...)
  forms <- .formalsNotInCurrentDots(.robustDigest, dots = dots)
  if (is(FUNcaptured, "call") || # as in rnorm(1) ... but also list(outputToSave) needs to be avoided
    (NROW(dots) > 0 && # if not an function with call, then it has to have something there
      # ... so not "just" an object in objsToDigest
      (NROW(forms) > 1 || is.null(forms)))) { # can be CacheDigest(rnorm, 1)
    fnDetails <- .fnCleanup(
      FUN = objsToDigest, callingFun = "Cache", ..., FUNcaptured = FUNcaptured,
      .functionName = .functionName, CacheMatchedCall = match.call(CacheDigest)
    )
    modifiedDots <- fnDetails$modifiedDots
    modifiedDots$.FUN <- fnDetails$.FUN
    objsToDigest <- modifiedDots
  }

  if (!is(objsToDigest, "list")) {
    objsToDigest <- list(objsToDigest)
  }

  if (identical("Cache", calledFrom)) {
    namesOTD <- names(objsToDigest)
    lengthChars <- nchar(namesOTD)
    if (!any(namesOTD %in% "FUN")) {
      zeroLength <- which(lengthChars == 0)
      alreadyHasDotFun <- dotFunTxt %in% namesOTD
      if (sum(zeroLength) > 0 && !alreadyHasDotFun) {
        names(objsToDigest)[zeroLength[1]] <- dotFunTxt
      }
    }
  }

  # need to omit arguments that are in Cache function call
  defaults <- names(objsToDigest) %in% .defaultCacheOmitArgs
  if (sum(defaults)) {
    objsToDigest[defaults] <- NULL
  }

  if (is.character(quick) || isTRUE(quick)) {
    quickObjs <- if (isTRUE(quick)) {
      rep(TRUE, length(objsToDigest))
    } else {
      if (is.null(names(objsToDigest))) {
         rep(FALSE, length(objsToDigest))
      } else {
        names(objsToDigest) %in% quick
      }

    }
    objsToDigestQuick <- objsToDigest[quickObjs]
    objsToDigest <- objsToDigest[!quickObjs]
    preDigestQuick <- .robustDigest(objsToDigestQuick, algo = algo, quick = TRUE, ...)
    # preDigestQuick <- lapply(objsToDigestQuick, function(x) {
    #   # remove the "newCache" attribute, which is irrelevant for digest
    #   if (!is.null(attr(x, ".Cache")$newCache)) {
    #     x <- .setSubAttrInList(x, ".Cache", "newCache", NULL)
    #     if (!identical(attr(x, ".Cache")$newCache, NULL)) stop("attributes are not correct 1")
    #   }
    #   .robustDigest(x, algo = algo, quick = TRUE, ...)
    # })
  }

  # if (!is(objsToDigest, "list"))
  preDigest <- .robustDigest(objsToDigest, algo = algo, quick = FALSE, ...)
  # preDigest <- Map(x = objsToDigest, i = seq_along(objsToDigest), function(x, i) {
  #   # remove the "newCache" attribute, which is irrelevant for digest
  #   if (!is.null(attr(x, ".Cache")$newCache)) {
  #     x <- .setSubAttrInList(x, ".Cache", "newCache", NULL)
  #     if (!identical(attr(x, ".Cache")$newCache, NULL)) stop("attributes are not correct 1")
  #   }
  #   withCallingHandlers({
  #     .robustDigest(x, algo = algo, quick = FALSE, ...)
  #   }, error = function(e) {
  #     nam <- names(objsToDigest)
  #     if (!is.null(nam))
  #       messageCache("Error occurred during .robustDigest of ", nam[i], " in ", .functionName)
  #   })
  # })


  # if (!isTRUE(all.equal(.orderDotsUnderscoreFirst(preDigest), .orderDotsUnderscoreFirst(preDigest2[names(preDigest)]))))
  if (is.character(quick) || isTRUE(quick)) {
    preDigest <- append(preDigest, preDigestQuick)
  }

  # preDigest <- .robustDigest(preDigest) # add the ._list
  # preDigest[["._list"]] <- NULL # don't need this for CacheDigest

  # don't unname -- Eliot Jan 13, 2025 -- this keeps the outputHash
  if (getOption("reproducible.digestV3", TRUE)) {
    res <- .doDigest(preDigest, algo = algo, ...)
  } else {
    res <- .robustDigest(unname(sort(unlist(preDigest))), algo = algo, quick = TRUE, ...)
    # res <- .robustDigest(.sortDotsUnderscoreFirst(unlist(preDigest)), algo = algo, quick = TRUE, ...)
  }
  list(outputHash = res, preDigest = preDigest)
}




#' @keywords internal
.defaultCacheOmitArgs <- c(
  "useCloud", "checksumsFileID", "cloudFolderID",
  "notOlderThan", ".objects", "outputObjects", "algo", "cachePath",
  "length", "compareRasterFileLength", "userTags", "digestPathContent",
  "omitArgs", "classOptions", "debugCache", "sideEffect", "makeCopy",
  "quick", "verbose", "cacheId", "useCache", "showSimilar", "cl"
)



verboseDF0 <- function(verbose, functionName, startHashTime, endTime) {
  if (verbose > 3) {
    if (missing(endTime))
      endTime <- Sys.time()
    verboseDF <- data.frame(
      functionName = functionName,
      component = "Hashing",
      elapsedTime = as.numeric(difftime(endTime, startHashTime, units = "secs")),
      units = "secs",
      stringsAsFactors = FALSE
    )
    verboseAppendOrCreateDF(verboseDF)
  }
  # if (exists("verboseTiming", envir = .reproEnv, inherits = FALSE)) {
  #   verboseDF$functionName <- paste0("  ", verboseDF$functionName)
  #   .reproEnv$verboseTiming <- rbind(.reproEnv$verboseTiming, verboseDF)
  # } else {
  #   .reproEnv$verboseTiming <- verboseDF
  # }
}

#' @keywords internal
verboseDF1 <- function(verbose, functionName, startRunTime, endTime) {
  if (verbose > 3) {
    if (missing(endTime))
      endTime <- Sys.time()
    verboseDF <- data.frame(
      functionName = functionName,
      component = paste("Running", functionName),
      elapsedTime = as.numeric(difftime(endTime, startRunTime, units = "secs")),
      units = "secs",
      stringsAsFactors = FALSE
    )

    if (exists("verboseTiming", envir = .reproEnv)) {
      .reproEnv$verboseTiming <- rbind(.reproEnv$verboseTiming, verboseDF)
    }
  }
}

#' @keywords internal
verboseDF2 <- function(verbose, functionName, startSaveTime, endTime) {
  if (verbose > 3) {
    if (missing(endTime))
      endTime <- Sys.time()
    verboseDF <-
      data.frame(
        functionName = functionName,
        component = "Saving to cachePath",
        elapsedTime = as.numeric(difftime(endTime, startSaveTime, units = "secs")),
        units = "secs",
        stringsAsFactors = FALSE
      )

    if (exists("verboseTiming", envir = .reproEnv)) {
      .reproEnv$verboseTiming <- rbind(.reproEnv$verboseTiming, verboseDF)
    }
  }
}


#' @keywords internal
verboseDF3 <- function(verbose, functionName, startCacheTime, endTime) {
  if (verbose > 3) {
    if (missing(endTime))
      endTime <- Sys.time()
    verboseDF <- data.frame(
      functionName = functionName,
      component = "Whole Cache call",
      elapsedTime = as.numeric(difftime(endTime, startCacheTime,
        units = "secs"
      )),
      units = "secs",
      stringsAsFactors = FALSE
    )

    if (exists("verboseTiming", envir = .reproEnv)) {
      .reproEnv$verboseTiming <- rbind(.reproEnv$verboseTiming, verboseDF)
    }
  }
}

#' @keywords internal

getCacheRepos <- function(cachePath, modifiedDots, verbose = getOption("reproducible.verbose", 1)) {
  if (is.null(cachePath)) {
    cachePath <- .checkCacheRepo(modifiedDots, create = TRUE, verbose = verbose)
  } else {
    if (any(!dir.exists(unlist(cachePath))))
      cachePath <- lapply(cachePath, function(repo) {
        if (!dir.exists(repo))
          repo <- checkPath(repo, create = TRUE)
        repo
      })
  }
  return(cachePath)
}


cloudFolderFromCacheRepo <- function(cachePath) {
  paste0(basename2(dirname(cachePath)), "_", basename2(cachePath))
}

.defaultUserTags <- c(
  "function", "class", "object.size", "accessed", "inCloud", "fromDisk",
  otherFunctions, "preDigest", "file.size", "cacheId",
  "elapsedTimeDigest", "elapsedTimeFirstRun", "resultHash", "elapsedTimeLoad"
)

.defaultOtherFunctionsOmit <- c(
  "(test_", "with_reporter", "force", "Restart", "with_mock",
  "eval", "::", "\\$", "\\.\\.", "standardGeneric",
  "Cache", "tryCatch", "doTryCatch", "withCallingHandlers",
  "FUN", "capture", "withVisible)"
)

isPkgColonFn <- function(x) {
  identical(x[[1]], quote(`::`))
}

evalTheFun <- function(FUNcaptured, isCapturedFUN, matchedCall, envir = parent.frame(),
                       verbose = getOption("reproducible.verbose"), ...) {
  .message$IndentUpdate()
  withCallingHandlers(
    {
      out <- eval(FUNcaptured, envir = envir)
      if (is.function(out)) { # if is wasn't "captured", then it is just a function, so now use it on the ...
        out <- out(...)
      }
    },
    warning = function(w) {
      asch <- format(w$call[[1]])
      warning("In ", format(matchedCall), ": ", w$message, call. = FALSE)
      invokeRestart("muffleWarning")
      #    }
    }
  )

  out
}

searchInRepos <- function(cachePaths, outputHash, drv, conn, verbose = getOption("reproducible.verbose")) {
  dbTabNam <- NULL
  tries <- 1
  while (tries <= length(cachePaths)) {
    repo <- cachePaths[[tries]]
    if (useDBI()) {
      if (is.list(conn))
        conn <- conn[[cachePaths[1]]]
      dbTabNam <- CacheDBTableName(repo, drv = drv)

      isInRepo <- getHashFromDB(tries, conn, drv, repo, dbTabNam, outputHash)
      # if (tries > 1) {
      #   DBI::dbDisconnect(conn)
      #   conn <- dbConnectAll(drv, cachePath = repo)
      # }
      # qry <- glue::glue_sql("SELECT * FROM {DBI::SQL(glue::double_quote(dbTabName))} where \"cacheId\" = ({outputHash})",
      #   dbTabName = dbTabNam,
      #   outputHash = outputHash,
      #   .con = conn
      # )
      # res <- retry(
      #   retries = 15, exponentialDecayBase = 1.01,
      #   quote(DBI::dbSendQuery(conn, qry))
      # )
      # isInRepo <- setDT(DBI::dbFetch(res))
      # DBI::dbClearResult(res)
    } else {
      # The next line will find it whether it is qs2, rds or other; this is necessary for "change cacheSaveFormat"
      csf <- CacheStoredFile(cachePath = repo, cacheId = outputHash, cacheSaveFormat = "check")

      if (all(file.exists(csf))) {
        dtFile <- CacheDBFileSingle(cachePath = repo, cacheId = outputHash)

        if (!file.exists(dtFile)) { # check first for wrong rds vs qs2
          dtFile <- CacheDBFileSingle(cachePath = repo, cacheId = outputHash, cacheSaveFormat = "check")
          fe <- file.exists(dtFile)
          if (isTRUE(!(fe))) { # still doesn't == means it is broken state
            warning(
              "The Cache file exists for ", outputHash, ", but there is no database entry for it; removing ",
              "the file and rerunning the call"
            )
            unlink(csf)
            dtFile <- NULL
          } else if (length(fe) > 1) { # has both the qs2 and rds dbFile
            browser()
          }
        }

        isInRepo <- if (!is.null(dtFile)) {
          loadFile(dtFile,
                   cacheId = outputHash, cachePath = repo, # in case it needs swapCacheFormat
                   drv = drv, conn = conn, verbose = verbose)
        } else {
          NULL
        }
      } else {
        isInRepo <- data.table::copy(.emptyCacheTable)
      }
    }
    fullCacheTableForObj <- isInRepo
    if (NROW(isInRepo) > 1) isInRepo <- isInRepo[NROW(isInRepo), ]
    if (NROW(isInRepo) > 0) {
      # browser(expr = exists("._Cache_4"))
      cachePath <- repo
      break
    }
    tries <- tries + 1
  }
  list(
    isInRepo = isInRepo, dbTabName = dbTabNam, fullCacheTableForObj = fullCacheTableForObj,
    cachePath = repo
  )
}




#' Search for objects in the call stack
#'
#' Normally, this is only used in special, advanced uses. The standard approach
#' to getting an object from an environment in the call stack is to explicitly
#' pass it into the function.
#'
#' @param obj Character string. The object name to search.
#' @param startingEnv An environment to start searching in.
#'
#' @return The environment in which the object exists. It will return the
#' first environment it finds, searching outwards from where the function is used.
#' @export
.whereInStack <- function(obj, startingEnv = parent.frame()) {
  foundStarting <- FALSE
  snf <- sys.nframe()
  for (i in 1:snf) {
    testEnv <- sys.frame(-i)
    if (!foundStarting) {
      if (identical(testEnv, startingEnv)) {
        foundStarting <- TRUE
      } else {
        next
      }
    }
    fn <- if (R.version$minor < "1.0" && R.version$major <= "4") { # faster than any other approach
      get0(as.character(parse(text = obj)), testEnv, inherits = FALSE)
    } else {
      get0(obj, testEnv, inherits = FALSE) # much faster; only works R >= 4.1
    }
    if (!is.null(fn)) {
      break
    }
  }
  if (identical(testEnv, .GlobalEnv) && identical(i, snf))
    testEnv <- NULL
  return(testEnv)
}

browserCond <- function(expr) {
  any(startsWith(ls(.GlobalEnv), expr))
}

spatVectorNamesForCache <- c("x", "type", "atts", "crs")


addCacheAttr <- function(output, .CacheIsNew, outputHash, FUN) {
  output <- .setSubAttrInList(output, ".Cache", "newCache", .CacheIsNew)
  attr(output, "tags") <- paste0("cacheId:", outputHash)
  attr(output, callInCache) <- ""
  if (!identical(attr(output, ".Cache")$newCache, .CacheIsNew)) {
    stop("attributes are not correct 3")
  }
  if (!identical(attr(output, callInCache), "")) {
    stop("attributes are not correct 4")
  }
  if (!identical(attr(output, "tags"), paste0("cacheId:", outputHash))) {
    stop("attributes are not correct 5")
  }

  if (isS4(FUN)) {
    attr(output, "function") <- FUN@generic
    if (!identical(attr(output, "function"), FUN@generic)) {
      stop("There is an unknown error 03")
    }
  }
  output
}


.objectSizeMinForBig <- 5e6

cacheIdCheckInCache <- function(cacheId, calculatedCacheId, .functionName,
                                verbose) {
  sc <- NULL
  if (!is.null(cacheId)) {
    if  (identical(cacheId, "previous")) {
      sc <- getPreviousEntryInCache(.functionName, cacheId, verbose)
    } else {
      outputHashManual <- cacheId
      sc <- list(1)
      if (identical(outputHashManual, calculatedCacheId)) {
        messageCache(.message$cacheIdSameTxt, verbose = verbose)
      } else {
        if (!is.null(calculatedCacheId)) {
          messageCache(.message$cacheIdNotSameTxt(cacheId), verbose = verbose)
        } else {
          messageCache(.message$cacheIdNotAssessed(cacheId), verbose = verbose)
        }
      }
      attr(sc, "cacheId") <- cacheId
      if (NROW(sc) == 0)
        sc <- NULL

    }
  }

  sc

}


checkOverlappingArgs <- function(CacheMatchedCall, forms, dotsCaptured, functionName,
                                 FUNcapturedNamesEvaled, whichCache = "Cache") {
  # Check for args that are passed to both Cache and the FUN -- if any overlap; pass to both
  possibleOverlap <- if (identical(whichCache, "Cache")) .namesCacheFormals else .namescache2Formals # names(formals(args(Cache)))
  if (!is.call(CacheMatchedCall[["FUN"]])) {
    possibleOverlap <- intersect(names(CacheMatchedCall), possibleOverlap)
    actualOverlap <- intersect(names(forms), possibleOverlap)
    if (length(actualOverlap) && !identical(list(), dotsCaptured)) { # e.g., useCache, verbose; but if not in dots, then OK because were separate already
      message(
        "The following arguments are arguments for both Cache and ", functionName, ":\n",
        paste0(actualOverlap, collapse = ", "),
        "\n...passing to both. If more control is needed, pass as a call, e.g., ",
        "Cache(", functionName, "(...))"
      )
      overlappingArgsAsList <- as.list(CacheMatchedCall)[actualOverlap]
      FUNcapturedNamesEvaled <- as.call(append(as.list(FUNcapturedNamesEvaled), overlappingArgsAsList))
    }
  }
  FUNcapturedNamesEvaled
}



verboseAppendOrCreateDF <- function(verboseDF) {
  if (exists("verboseTiming", envir = .reproEnv, inherits = FALSE)) {
    verboseDF$functionName <- paste0("  ", verboseDF$functionName)
    .reproEnv$verboseTiming <- rbind(.reproEnv$verboseTiming, verboseDF)
  } else {
    .reproEnv$verboseTiming <- verboseDF
  }
}



checkConns <- function(cachePaths, conn) {
  conns <- list()
  if (!is.null(conn)) { # if the conn was passed by user
    if (!is.list(conn)) {
      conn <- list(conn)
    }
    if (!identical(length(cachePaths), length(conn))) {
      stop("conn and cachePath are both provided, but are different lengths which is not allowed")
    }
    names(conn) <- cachePaths
    conns <- conn
  }
}


createConns <- function(cachePath, conns, drv,
                        verbose = getOption("reproducible.verbose")) {
  if (useDBI()) {
    drv <- getDrv(drv)
    if (is.null(conns[[cachePath]])) {
      conns[[cachePath]] <- dbConnectAll(drv, cachePath = cachePath)
      RSQLite::dbClearResult(RSQLite::dbSendQuery(conns[[cachePath]], "PRAGMA busy_timeout=5000;"))
      RSQLite::dbClearResult(RSQLite::dbSendQuery(conns[[cachePath]], "PRAGMA journal_mode=WAL;"))
    }
  }

  isIntactRepo <- CacheIsACache(
    cachePath = cachePath, drv = drv, create = TRUE,
    conn = conns[[cachePath]], verbose = verbose
  )
  if (any(!isIntactRepo)) {
    ret <- createCache(cachePath,
                       drv = drv, conn = conns[[cachePath]],
                       force = isIntactRepo
    )
  }
  conns
}

getHashFromDB <- function(tries, conn, drv, repo, dbTabNam, outputHash) {
  if (tries > 1) {
    DBI::dbDisconnect(conn)
    conn <- dbConnectAll(drv, cachePath = repo)
  }
  qry <- glue::glue_sql("SELECT * FROM {DBI::SQL(glue::double_quote(dbTabName))} where \"cacheId\" = ({outputHash})",
                        dbTabName = dbTabNam,
                        outputHash = outputHash,
                        .con = conn
  )
  res <- retry(
    retries = 15, exponentialDecayBase = 1.01,
    quote(DBI::dbSendQuery(conn, qry))
  )
  isInRepo <- setDT(DBI::dbFetch(res))
  DBI::dbClearResult(res)
  isInRepo
}

getPreviousEntryInCache <- function(.functionName, verbose, data.table, setorderv, tagKey, cacheId) {
  sc <- showCache(fun = .functionName, verbose = -2)
  if (NROW(sc)) {
    messageCache("cacheId is 'previous' meaning it will recover the most recent ",
                 "cache item (accessed) that matches on .functionName: ",
                 .messageFunctionFn(.functionName), "\nPlease ensure ",
                 "the function name is precise enough for this behaviour", verbose = verbose)
    outputHashNew <- data.table::setorderv(sc[tagKey == "accessed"], "tagValue", order = -1L)
    outputHash <- outputHashNew$cacheId[1]
    sc <- sc[cacheId %in% outputHash, ]
    attr(sc, "cacheId") <- outputHash
    # sc <- showCacheFast(cacheId = outputHash)
  } else {
    sc <- NULL
  }
}

callInCache <- "callInCache"
dotFunTxt <- ".FUN"

################################################################################
#' Convert numeric to character with padding
#'
#' This will pad floating point numbers, right or left. For integers, either class
#' integer or functionally integer (e.g., 1.0), it will not pad right of the decimal.
#' For more specific control or to get exact padding right and left of decimal,
#' try the `stringi` package. It will also not do any rounding. See examples.
#'
#' @param x numeric. Number to be converted to character with padding
#'
#' @param padL numeric. Desired number of digits on left side of decimal.
#'              If not enough, `pad` will be used to pad.
#'
#' @param padR numeric. Desired number of digits on right side of decimal.
#'              If not enough, `pad` will be used to pad.
#'
#' @param pad character to use as padding (`nchar(pad) == 1` must be `TRUE`).
#'
#' @return Character string representing the filename.
#'
#' @author Eliot McIntire and Alex Chubaty
#' @export
#' @rdname paddedFloatToChar
#' @importFrom fpCompare %==%
#'
#' @examples
#' paddedFloatToChar(1.25)
#' paddedFloatToChar(1.25, padL = 3, padR = 5)
#' paddedFloatToChar(1.25, padL = 3, padR = 1) # no rounding, so keeps 2 right of decimal
paddedFloatToChar <- function(x, padL = ceiling(log10(x + 1)), padR = 3, pad = "0") {
  xf <- x %% 1
  numDecimals <- nchar(gsub("(.*)(\\.)|([0]*$)", "", xf))
  newPadR <- ifelse(xf %==% 0, 0, pmax(numDecimals, padR))
  xFCEnd <- sprintf(paste0("%0", padL + newPadR + 1 * (newPadR > 0), ".", newPadR, "f"), x)
  return(xFCEnd)
}

#' Add a prefix or suffix to the basename part of a file path
#'
#' Prepend (or postpend) a filename with a prefix (or suffix).
#' If the directory name of the file cannot be ascertained from its path,
#' it is assumed to be in the current working directory.
#'
#' @param f       A character string giving the name/path of a file.
#' @param prefix  A character string to prepend to the filename.
#' @param suffix  A character string to postpend to the filename.
#'
#' @author Jean Marchal and Alex Chubaty
#' @export
#' @return
#' A character string or vector with the prefix pre-pended or suffix post-pended
#' on the `basename` of the `f`, before the file extension.
#' @rdname prefix
#'
#' @examples
#' # file's full path is specified (i.e., dirname is known)
#' myFile <- file.path("~/data", "file.tif")
#' .prefix(myFile, "small_") ## "/home/username/data/small_file.tif"
#' .suffix(myFile, "_cropped") ## "/home/username/data/myFile_cropped.shp"
#'
#' # file's full path is not specified
#' .prefix("myFile.shp", "small") ## "./small_myFile.shp"
#' .suffix("myFile.shp", "_cropped") ## "./myFile_cropped.shp"
#'
.prefix <- function(f, prefix = "") {
  file.path(dirname(f), paste0(prefix, basename(f)))
}

#' @export
#' @name suffix
#' @rdname prefix
.suffix <- function(f, suffix = "") {
  file.path(dirname(f), paste0(
    filePathSansExt(basename(f)), suffix,
    ".", fileExt(f)
  ))
}

#' Get a unique name for a given study area
#'
#' Digest a spatial object to get a unique character string (hash) of the study area.
#' Use `.suffix()` to append the hash to a filename,
#' e.g., when using `filename2` in `prepInputs`.
#'
#' @param studyArea Spatial object.
#' @param ... Other arguments (not currently used)
#'
#' @export
#' @return
#' A character string using the `.robustDigest` of the `studyArea`. This is only intended
#' for use with spatial objects.
#' @examples
#' studyAreaName("Ontario")
setGeneric("studyAreaName", function(studyArea, ...) {
  standardGeneric("studyAreaName")
})

# @export
# @rdname studyAreaName
# setMethod(
#   "studyAreaName",
#   signature = "SpatialPolygonsDataFrame",
#   definition = function(studyArea, ...) {
#     studyArea <- studyArea[, -c(1:ncol(studyArea))]
#     studyArea <- as(studyArea, "SpatialPolygons")
#     studyAreaName(studyArea, ...)
# })

#' @export
#' @rdname studyAreaName
setMethod(
  "studyAreaName",
  signature = "character",
  definition = function(studyArea, ...) {
    sort(studyArea) ## ensure consistent hash for same subset of study area names
    .robustDigest(studyArea, algo = "xxhash64") ## TODO: use `...` to pass `algo`
  }
)

#' @export
#' @rdname studyAreaName
setMethod(
  "studyAreaName",
  signature = "ANY",
  definition = function(studyArea, ...) {
    if (inherits(studyArea, "sf")) {
      .requireNamespace("sf", stopOnFALSE = TRUE)
      studyArea <- sf::st_geometry(studyArea)
    } else if (inherits(studyArea, "SpatialPolygonsDataFrame")) {
      studyArea <- studyArea[, -c(seq_len(ncol(studyArea)))]
      studyArea <- as(studyArea, "SpatialPolygons")
      studyAreaName(studyArea, ...)
    } else if (!(inherits(studyArea, "Spatial") || inherits(studyArea, "sfc") ||
                 inherits(studyArea, "SpatVector") || is.character(studyArea))) {
      stop("studyAreaName expects a spatialClasses object (or character vector)")
    }
    .robustDigest(studyArea, algo = "xxhash64") ## TODO: use `...` to pass `algo`
  }
)

#' Identify which formals to a function are not in the current `...`
#'
#' Advanced use.
#'
#' @keywords internal
#' @export
#' @return
#' A list of the formals of the `fun` that are missing from the `...` or `dots`.
#'
#' @param fun A function
#' @param ... The ... from inside a function. Will be ignored if `dots` is
#'        provided explicitly.
#' @param dots Optional. If this is provided via say `dots = list(...)`,
#'             then this will cause the `...` to be ignored.
#' @param formalNames Optional character vector. If provided then it will override the `fun`
.formalsNotInCurrentDots <- function(fun, ..., dots, formalNames, signature = character()) {
  if (is.character(fun)) {
    fun <- get(fun, mode = "function", envir = parent.frame())
  }

  if (missing(formalNames)) {
    if (isS4(fun)) {
      forms <- methodFormals(fun, signature = signature, envir = parent.frame())
      formalNames <- names(forms)
    } else {
      formalNames <- names(formals(fun))
    }
  }

  if (!missing(dots)) {
    out <- names(dots)[!(names(dots) %in% formalNames)]
  } else {
    out <- ...names()[!(...names() %in% formalNames)]
  }
  out
}

#' @keywords internal
rndstr <- function(n = 1, len = 8) {
  unlist(lapply(character(n), function(x) {
    x <- paste0(sample(c(0:9, letters, LETTERS),
                       size = len,
                       replace = TRUE
    ), collapse = "")
  }))
}

#' Alternative to `interactive()` for unit testing
#'
#' This is a suggestion from
#' <https://github.com/MangoTheCat/blog-with-mock/blob/master/Blogpost1.Rmd>
#' as a way to test interactive code in unit tests. Basically, in the unit tests,
#' we use `testthat::with_mock`, and inside that we redefine `isInteractive`
#' just for the test. In all other times, this returns the same things as
#' `interactive()`.
#' @keywords internal
isInteractive <- function() interactive()

#' A version of `base::basename` that is `NULL` resistant
#'
#' @return
#' `NULL` if x is `NULL`, otherwise, as `basename`.
#'
#' @param x A character vector of paths
#' @export
#' @return Same as [base::basename()]
#'
basename2 <- function(x) {
  if (is.null(x)) {
    NULL
  } else {
    basename(x)
  }
}

#' A wrapper around `try` that retries on failure
#'
#' This is useful for functions that are "flaky", such as `curl`, which may fail for unknown
#' reasons that do not persist.
#'
#' @details
#' Based on <https://github.com/jennybc/googlesheets/issues/219#issuecomment-195218525>.
#'
#' @param expr     An expression to run, i.e., `rnorm(1)`, similar to what is passed to `try`
#' @param retries  Numeric. The maximum number of retries.
#' @param envir    The environment in which to evaluate the quoted expression, default
#'   to `parent.frame(1)`
#' @param exponentialDecayBase Numeric > 1.0. The delay between
#'   successive retries will be `runif(1, min = 0, max = exponentialDecayBase ^ i - 1)`
#'   where `i` is the retry number (i.e., follows `seq_len(retries)`)
#' @param silent   Logical indicating whether to `try` silently.
#' @param exprBetween Another expression that should be run after a failed attempt
#'   of the `expr`. This should return a named list, where the names indicate the object names
#'   to update in the main expr, and the return value is the new value. (previous versions allowed
#'   a non-list return, but where the final line had to be an assignment operator,
#'   specifying what object (that is used in `expr`) will be updated prior to running
#'   the `expr` again. For backwards compatibility, this still works).
#' @param messageFn A function for messaging to console. Defaults to `message`
#'
#' @return
#' As with `try`, so the successfully returned `return()` from the `expr` or a `try-error`.
#'
#' @export
retry <- function(expr, envir = parent.frame(), retries = 5,
                  exponentialDecayBase = 1.3, silent = TRUE,
                  exprBetween = NULL, messageFn = message) {
  if (exponentialDecayBase < 1) {
    stop("exponentialDecayBase must be equal to or greater than 1")
  }
  hasRutils <- .requireNamespace("R.utils", stopOnFALSE = FALSE, messageStart = "")

  for (i in seq_len(retries)) {
    exprSub <- substitute(expr) # Have to deal with case where expr is already quoted
    # if (!(is.call(expr) || is.name(expr))) warning("expr is not a quoted expression")

    if ( hasRutils) {
      # wrap the expr with R.utils::withTimeout
      expr2 <- append(append(list(R.utils::withTimeout), exprSub),
                      list(timeout = getOption("reproducible.timeout", 1200), onTimeout = "error"))
      exprSub <- as.call(expr2)
    }

    result <- try(silent = silent,
                  expr = withCallingHandlers({
                    res <- eval(exprSub, envir = envir)
                    if (is.call(res))
                      if (is.call(expr))
                        res <- eval(res, envir = envir)
                    res
                  },
                  error = function(e) {
                    if (!hasRutils) {
                      message("If the download stalls/stalled, please interrupt this function ",
                              "then install R.utils, then rerun this prepInputs/preProcess. This ",
                              "function will then use `R.utils::withTimeout`, which will cause an error ",
                              "sooner")
                    }
                  })
    )

    if (inherits(result, "try-error")) {
      if (!is.null(exprBetween)) {
        finalPart <- length(format(exprBetween))

        # The expression is different if it is 1 line vs >1 line
        exprBetweenTail <- if (finalPart > 1) {
          finalPart <- length(exprBetween)
          exprBetween[[finalPart]]
        } else {
          exprBetween
        }
        # if (!identical(as.character(exprBetweenTail)[[1]], "<-"))
        #   stop("exprBetween must have an assignment operator <- with a object on",
        #        "the LHS that is used on the RHS of expr ")
        objName <- as.character(exprBetweenTail[[2]])
        result1 <-
          try(expr = eval(exprBetween, envir = envir), silent = silent)
        if (is.list(result1)) {
          if (!is.null(names(result1))) {
            objName <- names(result1)
          } else {
            stop("The return object from exprBetween must be a named list, with names being objects to overwrite")
          }
        } else {
          result1 <- list(result1)
          names(result1) <- objName
        }

        lapply(objName, function(objNam) assign(objNam, result1[[objNam]], envir = envir))
      }
      backoff <- sample(1:1000 / 1000, size = 1) * (exponentialDecayBase^i - 1)
      if (backoff > 3) {
        messageFn("Waiting for ", round(backoff, 1), " seconds to retry; the attempt is failing")
      }
      Sys.sleep(backoff)
    } else {
      if (exists("result1", inherits = FALSE)) {
        messageFn("    ...fixed!")
      }
      break
    }
  }

  if (inherits(result, "try-error")) {
    stop(result, "\nFailed after ", retries, " attempts.")
  } else {
    return(result)
  }
}

#' @keywords internal
isCI <- function() {
  as.logical(Sys.getenv("CI"))
}

#' Test whether system is Windows
#'
#' This is used so that unit tests can override this using `testthat::with_mock`.
#' @keywords internal
isWindows <- function() identical(.Platform$OS.type, "windows")

#' @keywords internal
isMac <- function() {
  Sys.info()[["sysname"]] |>
    tolower() |>
    identical("darwin")
}

#' Provide standard messaging for missing package dependencies
#'
#' This provides a standard message format for missing packages, e.g.,
#' detected via `requireNamespace`.
#'
#' @export
#'
#' @return
#' A logical or stop if the namespace is not available to be loaded.
#'
#' @param pkg Character string indicating name of package required
#' @param minVersion Character string indicating minimum version of package
#'   that is needed
#' @param messageStart A character string with a prefix of message to provide
#' @param stopOnFALSE Logical. If `TRUE`, this function will create an
#'   error (i.e., `stop`) if the function returns `FALSE`; otherwise
#'   it simply returns `FALSE`
.requireNamespace <- function(pkg = "methods", minVersion = NULL,
                              stopOnFALSE = FALSE,
                              messageStart = NULL) {
  need <- FALSE
  if (!requireNamespace(pkg, quietly = TRUE)) {
    need <- TRUE
  } else {
    if (!is.null(minVersion)) {
      if (isTRUE(packageVersion(pkg) < minVersion)) {
        need <- TRUE
      }
    }
  }

  if (need) { # separate these so it is faster
    if (isTRUE(stopOnFALSE)) {
      stop(.message$RequireNamespaceFn(pkg, messageExtra = messageStart, minVersion = minVersion))
    }
  }
  !need
}

## TODO: why not use the original funs below? why copy them?
## This is directly from tools::file_path_sans_ext
filePathSansExt <- function(x) {
  sub("([^.]+)\\.[[:alnum:]]+$", "\\1", x)
}

## This is directly from tools::file_ext
fileExt <- function(x) {
  pos <- regexpr("\\.([[:alnum:]]+)$", x)
  ifelse(pos > -1L, substring(x, pos + 1L), "")
}

isDirectory <- function(pathnames, mustExist = TRUE) {
  keep <- is.character(pathnames)
  if (length(pathnames) == 0) {
    return(logical())
  }
  if (isFALSE(keep)) stop("pathnames must be character")
  origPn <- pathnames
  if (isTRUE(mustExist)) {
    pathnames <- normPath(pathnames[keep])
    id <- dir.exists(pathnames)
    id[id] <- file.info(pathnames[id])$isdir
  } else {
    if (isGoogleID(pathnames) || isGoogleDriveURL(pathnames)) {
      id <- isGoogleDriveDirectory(pathnames)
    } else {
      id <- grepl("/$|\\\\$", pathnames)
    }
  }
  names(id) <- origPn
  id
}

isGoogleDriveDirectory <- function(url) {
  isFold <- grepl("folders", url)
  if (isFold %in% FALSE && is(url, "drive_id")) {
    # second check if it is just a google id
    driveMeta <- googledrive::drive_get(url)
    isFold <- isGoogleDriveDirectoryFromTibble(driveMeta)
  }
  isFold
}

isGoogleDriveDirectoryFromTibble <- function(tib) {
  if (tib$drive_resource[[1]]$mimeType == "application/vnd.google-apps.folder") {
    isFold <- TRUE
  } else {
    isFold <- FALSE
  }
  isFold
}

isFile <- function(pathnames) {
  keep <- is.character(pathnames)
  if (isFALSE(keep)) stop("pathnames must be character")
  origPn <- pathnames
  pathnames <- normPath(pathnames[keep])
  iF <- file.exists(pathnames)
  iF[iF] <- !file.info(pathnames[iF])$isdir
  names(iF) <- origPn
  iF
}

methodFormals <- function(fun, signature = character(), envir = parent.frame()) {
  if (is.character(fun)) {
    fun <- get(fun, mode = "function", envir = envir)
  }

  fdef <- getGeneric(fun)
  method <- selectMethod(fdef, signature)
  genFormals <- base::formals(fdef)
  b <- body(method)
  if (is(b, "{") && is(b[[2]], "<-") && identical(b[[2]][[2]], as.name(".local"))) {
    local <- eval(b[[2]][[3]])
    if (is.function(local)) {
      return(formals(local))
    }
    warning("Expected a .local assignment to be a function. Corrupted method?")
  }
  genFormals
}

.fileExtsKnown <- function() {
  if (requireNamespace("sf", quietly = TRUE) && requireNamespace("DBI", quietly = TRUE)) {
    shpFile <- getOption("reproducible.shapefileRead", "sf::st_read")
  } else {
    shpFile <- "terra::vect"
    if (identical(getOption("reproducible.shapefileRead"), "sf::st_read")) {
      options("reproducible.shapefileRead" = shpFile)
    } # can't use sf::st_read
  }
  griddedFile <- getOption("reproducible.rasterRead", "terra::rast")
  griddedFileSave <- ""
  shpFileSave <- ""
  if (griddedFile %in% "terra::rast") {
    griddedFileSave <- "terra::writeRaster"
  }
  if (griddedFile %in% "raster::raster") {
    griddedFileSave <- "terra::writeRaster"
  }
  if (shpFile %in% "sf::st_read") {
    shpFileSave <- "sf::st_write"
  }

  if (shpFile %in% "terra::vect") {
    shpFileSave <- "terra::writeVector"
  }

  df <- data.frame(
    rbind(
      c(.rdsFormat, "base::readRDS", "base::saveRDS", "binary"),
      c(.qsFormat, "qs::qread", "qs::qsave", .qsFormat),
      c(.qs2Format, "qs2::qs_read", "qs2::qs_save", .qs2Format),
      cbind(
        c("asc", "grd", "tif"), griddedFile, griddedFileSave,
        rasterType(rasterRead = griddedFile)
      ),
      cbind(
        c("shp", "gdb"), shpFile, shpFileSave,
        vectorType(vectorRead = shpFile)
      )
    )
  )

  sfCanRead <- c("bna","csv", "e00", "gdb", "geojson", "gml", "gmt", "gpkg", "gps",
    "gtm", "gxt", "jml", "map", "mdb", "nc", "ods", "osm", "pbf", "shp", "sqlite",
    "vdv")#, "xls", "xlsx" )
  sfFun <- shpFile
  sfSaveFun <- shpFileSave
  type = "sf"
  colnames(df) <- c("extension", "fun", "saveFun", "type")
  df2 <- data.frame(sfCanRead, sfFun, sfSaveFun, type)
  colnames(df2) <- c("extension", "fun", "saveFun", "type")
  df <- rbind(df, df2)

  colnames(df) <- c("extension", "fun", "saveFun", "type")
  df
}

#' A helper to `getOption("reproducible.rasterRead")`
#' @export
#' @param ... Passed to the function parsed and evaluated from
#'   `getOption("reproducible.rasterRead")`
#'
#' @return
#' A function, that will be the evaluated, parsed character
#' string, e.g., `eval(parse(text = "terra::rast"))`
rasterRead <- function(...) {
  eval(parse(text = getOption("reproducible.rasterRead")))(...)
}

rasterType <- function(nlayers = 1,
                       rasterRead = getOption("reproducible.rasterRead", "terra::rast")) {
  # This does not have to fail if `raster` is not installed because this function may be
  #    called with .fileExtsKnown(), even if no gridded object is being used
  if (is.character(rasterRead)) {
    if (identical("terra::rast", rasterRead)) {
      if (.requireNamespace("terra")) rasterRead <- eval(parse(text = rasterRead)) else ""
    }
    if (identical("raster::raster", rasterRead) ||
        identical("raster::stack", rasterRead) ||
        identical("raster::brick", rasterRead)) {
      rasterRead <- if (.requireNamespace("raster")) eval(parse(text = rasterRead)) else ""
    } else {
      ""
    }
  }

  if (!is.character(rasterRead)) {
    rasterRead <- if (identical(rasterRead, terra::rast)) {
      "SpatRaster"
    } else if (nlayers == 1) "RasterLayer" else "RasterStack"
  }
  rasterRead
}

vectorType <- function(vectorRead = getOption("reproducible.shapefileRead", "sf::st_read")) {
  needRasterPkg <- FALSE
  vectorReadSubs <- substitute(vectorRead)
  if (any(grepl("^shapefile$", vectorReadSubs))) {
    .requireNamespace("raster", stopOnFALSE = TRUE)
    needRasterPkg <- TRUE
  }
  if (is.character(vectorRead)) {
    if (endsWith(suffix = "shapefile", vectorRead)) {
      if (.requireNamespace("raster", stopOnFALSE = TRUE)) {
        needRasterPkg <- TRUE
      }
    }
    vectorRead <- if (.requireNamespace("terra") || .requireNamespace("sf") || .requireNamespace("sp")) {
      eval(parse(text = vectorRead))
    } else {
      ""
    }
  }
  if (!is.character(vectorRead)) {
    isTerra <- try(.requireNamespace("terra"), silent = TRUE)
    isTerra2 <- if (isTRUE(isTerra)) identical(vectorRead, terra::vect) else FALSE
    isSF <- try(.requireNamespace("sf"), silent = TRUE)
    isSF2 <- if (isTRUE(isSF)) identical(vectorRead, sf::st_read) else FALSE

    vectorRead <- if (isTerra2) {
      if (isTerra %in% FALSE) .requireNamespace("terra", stopOnFALSE = TRUE)
      "SpatVector"
    } else if (needRasterPkg) {
      .requireNamespace("raster", stopOnFALSE = TRUE)
      "SpatialPolygons"
    } else if (isTRUE(isSF2)) {
      if (isSF %in% FALSE) .requireNamespace("sf", stopOnFALSE = TRUE)
      "sf"
    } else {
      stop("No vector read options available (")
    }
  }
  vectorRead
}

#' Set seed with a random value using Sys.time()
#'
#' This will set a random seed.
#' @export
#' @param set.seed Logical. If `TRUE`, the default, then the function will call
#' `set.seed` internally with the new random seed.
#' @details
#' This function uses 6 decimal places of `Sys.time()`, i.e., microseconds. Due to
#' integer limits, it also truncates at 1000 seconds, so there is a possibility that
#' this will be non-unique after 1000 seconds (at the microsecond level). In
#' tests, this showed no duplicates after 1e7 draws in a loop, as expected.
#'
#' @note
#' This function does not appear to be as reliable on R <= 4.1.3
#'
#' @return
#' This will return the new seed invisibly. However, this is also called for
#' its side effects, which is a new seed set using `set.seed`
set.randomseed <- function(set.seed = TRUE) {
  digits <- 9
  newSeed <- as.numeric(Sys.time()) * 10^(digits - 3) # microseconds
  newSeed <- as.integer(round(newSeed, -digits) - newSeed)
  if (isTRUE(set.seed)) {
    set.seed(newSeed)
  }
  return(invisible(newSeed))
}

# This used to be in helper-allEqual.R --> one of the tests couldn't find it
getDataFn <- function(...) {
  if (requireNamespace("geodata", quietly = TRUE)) {
    suppressWarningsSpecific(
      {
        geodata::gadm(...)
      },
      falseWarnings = "getData will be removed in a future version of raster"
    )
  } else {
    stop("dependency package 'geodata' is not installed.\n",
         "Try `install.packages('geodata', repos = 'https://predictiveecology.r-universe.dev/')`")
  }
}

milliseconds <- function(time = Sys.time()) {
  tt <- as.numeric(time)
  rnd <- round(tt, -5)
  (tt - rnd) * 1000
}

layerNamesDelimiter <- "_%%_"

#' Checks for existed of a url or the internet using <https://CRAN.R-project.org>
#'
#' A lightweight function that may be less reliable than more purpose built solutions
#' such as checking a specific web page using `RCurl::url.exists`. However, this is
#' slightly faster and is sufficient for many uses.
#'
#' @return Logical, `TRUE` if internet site exists, `FALSE` otherwise
#'
#' @export
#' @name internetExists
#' @rdname internetExists
internetExists <- function() {
  # urlExists("https://CRAN.R-project.org")
  urlExists("https://www.google.com")
}

#' @rdname internetExists
#' @export
#' @name urlExists
#' @param url A url of the form `https://...` to test for existence.
#' @return Logical, `TRUE` if internet site exists, `FALSE` otherwise.
urlExists <- function(url) {
  con <- url(url)
  on.exit(try(close(con), silent = TRUE), add = TRUE)
  a <- try(suppressWarnings(readLines(con, n = 1)), silent = TRUE)
  !is(a, "try-error")
}


.addSlashNToAllButFinalElement <- function(mess) {
  if (length(mess) > 1)
    mess[1:(length(mess)-1)] <- paste0(mess[1:(length(mess)-1)], "\n")
  mess
}

prefixCacheId <- function(cacheId) {
  if (is.null(cacheId))
    character()
  else
    paste0(cacheId, "_")
}

#' Extract the cache id of an object
#'
#' Any object that was returned from the Cache or was calculated as part of a
#' Cache call will have an attribute, `tags` and an entry with `cacheId:` prefix.
#' This is a lightweight helper to extract that `cacheId`.
#'
#' @param obj Any R object
#'
#' @return The `cacheId` if this was part of a `Cache` call. Otherwise `NULL`
#'
#' @export
cacheId <- function(obj) {
  whHasCacheId <- which(grepl("cacheId:", attr(obj, "tags")))
  if (any(whHasCacheId))
    gsub("cacheId:", "", attr(obj, "tags")[whHasCacheId])
  else
    NULL
}

#' Count Active Threads Based on CPU Usage
#'
#' This function counts the number of active system processes (threads) that
#' match a given pattern and exceed a specified minimum CPU usage threshold. It
#' works on Unix-like systems (e.g., Linux, macOS) and does not support Windows.
#'
#' @param pattern A character string used to filter process lines. Only
#'   processes whose command line matches this pattern will be considered.
#'   Default is `""` (matches all).
#' @param minCPU A numeric value specifying the minimum CPU usage (in percent)
#'   for a process to be considered active. Default is `50`.
#'
#' @return An integer representing the number of active threads matching the
#'   pattern and exceeding the CPU usage threshold. Returns `NULL` with a
#'   message if run on Windows.
#'
#' @examples
#' \dontrun{
#'   detectActiveCores(pattern = "R", minCPU = 30)
#' }
#'
#' @note This function uses the `ps -ef` system command and regular expressions
#'   to parse CPU usage. It may not be portable across all Unix variants.
#'
#' @export
detectActiveCores <- function(pattern = "", minCPU = 50) {
  if (!identical(.Platform$OS.type, "windows")) {
    a0 <- system("ps -ef", intern = TRUE)[-1]
    a4 <- grep(pattern, a0, value = TRUE)
    a5 <- gsub("^.*[[:digit:]]* [[:digit:]]* ([[:digit:]]{1,3}) .*$",
               "\\1", a4)
    sum(as.numeric(a5) > minCPU)
  } else {
    message("Does not work on Windows")
  }
}

# ---------------------------------------------------------------------------
# Backwards-compatible accessors for the destinationPathShared / inputPaths option pair.
# The canonical names are reproducible.destinationPathShared and
# reproducible.destinationPathSharedRecursive (matching the prepInputs naming family).
# The old names (reproducible.inputPaths, reproducible.inputPathsRecursive) are
# still accepted: if the new name is NULL but the old name is set, the old value
# is used and a one-time deprecation message is emitted.
# ---------------------------------------------------------------------------

# One-shot per-session deprecation messaging. `.pkgEnv$.deprecMsgEmitted`
# accumulates the option names whose deprecation message has already fired,
# so subsequent calls in the same R session stay silent regardless of how
# many times the getter is invoked from inside the preProcess pipeline.
.deprecMsgOnce <- function(optName, replacementName) {
  emitted <- .pkgEnv$.deprecMsgEmitted
  if (is.null(emitted)) emitted <- character()
  if (optName %in% emitted) return(invisible())
  message("Option '", optName, "' is deprecated; ",
          "please use '", replacementName, "' instead.")
  .pkgEnv$.deprecMsgEmitted <- c(emitted, optName)
  invisible()
}

#' @keywords internal
.getDestinationPathShared <- function() {
  newVal <- getOption("reproducible.destinationPathShared", NULL)
  if (!is.null(newVal)) return(newVal)
  oldVal <- getOption("reproducible.inputPaths", NULL)
  if (!is.null(oldVal))
    .deprecMsgOnce("reproducible.inputPaths",
                   "reproducible.destinationPathShared")
  oldVal
}

#' @keywords internal
.getDestinationPathSharedRecursive <- function() {
  newVal <- getOption("reproducible.destinationPathSharedRecursive", NULL)
  if (!is.null(newVal)) return(newVal)
  oldVal <- getOption("reproducible.inputPathsRecursive", NULL)
  if (!is.null(oldVal))
    .deprecMsgOnce("reproducible.inputPathsRecursive",
                   "reproducible.destinationPathSharedRecursive")
  # Default is FALSE when neither is set
  if (is.null(oldVal)) FALSE else oldVal
}

# Deprecated wrappers — to be removed in a future release; preserved only so
# any out-of-tree code that referenced the previous (branch-only) names still
# resolves until callers can be updated.
#' @keywords internal
.getDataPath <- function() .getDestinationPathShared()
#' @keywords internal
.getDataPathRecursive <- function() .getDestinationPathSharedRecursive()

test_that("prepInputsUrlTiles", {
  skip_on_cran()
  skip_on_ci()

  testInit(needGoogleDriveAuth = TRUE,
           c("terra", "googledrive"),

           opts = list(
             rasterTmpDir = tempdir2(rndstr(1, 6)),
             reproducible.inputPaths = NULL,
             reproducible.overwrite = TRUE,
             reproducible.useMemoise = FALSE,
             reproducible.rasterRead = "terra::rast"
           ),
           needInternet = TRUE
  )

  withr::local_options(reproducible.cachePath = tmpdir,
                       reproducible.inputPath = tmpdir,
                       mc.cores = 2L)# used by tiles
  outerDriveFolder <- "1KuBraAYnBpyxl3Nf0udc05fQlTPds2xY"
  skip_if_service_account_releaseVer_NotLinux()



  urlForTiles <- try(googledrive::drive_ls(googledrive::as_id(outerDriveFolder)))
  urlForTiles <- googledrive::drive_mkdir(
    name = basename(tempfile(pattern = "urlForTiles_")),
    path = googledrive::as_id(outerDriveFolder))
  urlForTiles <- urlForTiles$id
  withr::local_options(reproducible.prepInputsUrlTiles = urlForTiles)
  fn <- "reproducible_testUrlTiles_test4Tiles.tif"
  ext <- c(xmin = 307000, xmax = 363000, ymin = 1306000, ymax = 1394000)
  ext <- unname(ext)
  b <- terra::rast(terra::ext(ext, xy = FALSE), resolution = 100, vals = 1)
  terra::crs(b) <- "epsg:3978"
  extLrg <- terra::extend(b, 1e2)
  terra::crs(extLrg) <- "epsg:3978"
  extLrg <- terra::writeRaster(x = extLrg, filename = fn, overwrite = TRUE)
  d <- googledrive::drive_upload(fn, path = googledrive::as_id(urlForTiles))

  mess1 <- capture_messages(
    warns <- capture_warnings(
      a1 <- prepInputs(url = d$id, to = b, doUploads = TRUE, numTiles = c(2,2))
    )
  )
  expect_is(a1, "SpatRaster")
  withr::local_options(reproducible.prepInputsUrlTiles = NULL)
  mess2 <- capture_messages(a2 <- prepInputs(url = d$id, to = b, useCache = FALSE))
  cat(mess2, file = "~/tmp/mess2.txt")
  cat(mess1, file = "~/tmp/mess1.txt")
  a1b <- .wrap(a1)
  a2b <- .wrap(a2)
  expect_is(a2, "SpatRaster")
  expect_is(a1, "SpatRaster")

  testthat::expect_equivalent(a1, a2)

  # if (FALSE) {
  withr::local_options(reproducible.prepInputsUrlTiles = urlForTiles)
  fn <- "reproducible_testUrlTiles_test1Tile.tif"
  ext <- c(ymin = 1306000, ymax = 1394000, xmin = 307000, xmax = 363000)
  ext <- unname(ext)
  b <- terra::rast(terra::ext(ext, xy = FALSE), resolution = 100, vals = 1)
  terra::crs(b) <- "epsg:3978"
  extLrg <- terra::extend(b, 1e2)
  terra::crs(extLrg) <- "epsg:3978"
  extLrg <- terra::writeRaster(x = extLrg, filename = fn, overwrite = TRUE)
  d <- googledrive::drive_upload(fn, path = googledrive::as_id(urlForTiles))
  b1 <- prepInputs(url = d$id, to = b, doUploads = TRUE, numTiles = c(2,2))
  withr::local_options(reproducible.prepInputsUrlTiles = NULL)
  b2 <- prepInputs(url = d$id, to = b)
  testthat::expect_equivalent(b1, b2)
  save(a1, a2, b1, b2, file = "~/tmp/objs.rda")
  gls <- googledrive::drive_ls(urlForTiles)

  # clean up
  # googledrive::drive_rm(gls)
  googledrive::drive_rm(urlForTiles)

  # }
})

test_that("prepInputs doesn't work (part 1)", {
  skip_on_cran()
  skip_on_ci()

  testInit(
    "terra",
    opts = list(
      rasterTmpDir = tempdir2(rndstr(1, 6)),
      reproducible.inputPaths = NULL,
      reproducible.overwrite = TRUE,
      reproducible.useMemoise = FALSE,
      reproducible.showSimilar = TRUE
    ),
    needInternet = TRUE
  )

  withr::local_options(reproducible.cachePath = tmpdir)

  globalNoisy <- capture.output({
    ## Add a study area to Crop and Mask to
    coords <- structure(
      c(-122.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
      .Dim = c(5L, 2L)
    )
    StudyArea <- terra::vect(coords, "polygons")
    terra::crs(StudyArea) <- crsToUse

    dPath <- file.path(tmpdir, "ecozones")

    url <- "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip"

    mess <- capture_messages({
      shpEcozone <- prepInputs(destinationPath = dPath, url = url)
    })
    expect_true(any(grepl(mess, pattern = "ecozone_shp[.]zip")))
    expect_true(any(grepl(mess, pattern = "Appending")))
    expect_true(is(shpEcozone, vectorType()))

    # test sf::st_read vs "sf::st_read" -- sf::st_read didn't work before Oc 29, 2024
    if (.requireNamespace("sf")) {
      out <- prepInputs(
        targetFile = "Ecozones/ecozones.shp",
        destinationPath = dPath,
        fun = sf::st_read
      )
      expect_is(out, "sf")
    }

    # Robust to partial file deletions:
    unlink(dir(dPath, full.names = TRUE)[1:3])
    expect_error(terra::vect(file.path(dPath, "ecozone_shp.zip")))
    rm(shpEcozone)
    shpEcozone1 <- prepInputs(destinationPath = dPath, url = url)
    expect_true(is(shpEcozone1, vectorType()))
    unlink(dPath, recursive = TRUE)

    ### url, targetFile, alsoExtract # # #g
    # Once this is done, can be more precise in operational code:
    #  specify targetFile, alsoExtract, and fun, wrap with Cache
    ecozoneFilename <- file.path(dPath, "Ecozones/ecozones.shp")
    ## fmt: skip
    ecozoneFiles <- c(
      "ecozones.dbf", "ecozones.prj", "ecozones.sbn",
      "ecozones.sbx", "ecozones.shp", "ecozones.shx"
    )
    shpEcozone2 <- prepInputs(
      targetFile = ecozoneFilename,
      url = "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip",
      alsoExtract = ecozoneFiles,
      destinationPath = dPath
    )

    if (.requireNamespace("sf")) {
      expect_true(is(shpEcozone2, "sf"))
      testObj <- if (!is(shpEcozone1, "sf")) as(shpEcozone1, "sf") else shpEcozone1
    }

    # As of Jan 2022 -- these objects are very different; character encoding of accents, numbers interpreted as character
    # expect_equivalent(testObj, shpEcozone2) # different attribute newCache

    ### url, targetFile, alsoExtract -- with Cache
    # specify targetFile, alsoExtract, and fun, wrap with Cache -- it is wrong b/c no subfolder
    ecozoneFilename <- file.path(dPath, "ecozones.shp")
    # Note, you don't need to "alsoExtract" the archive... if the archive is not there, but the
    #   targetFile is there, it will not redownload the archive.

    unlink(dirname(ecozoneFilename), recursive = TRUE)
    # Test useCache = FALSE -- doesn't error and has no "loading from cache" or "loading from memoised"
    warn <- suppressWarningsSpecific(falseWarnings = "attribute variables are assumed to be spatially constant", {
      mess <- capture_messages({
        shpEcozoneSm <- Cache(
          prepInputs(
            url = "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip",
            targetFile = reproducible::asPath(ecozoneFilename),
            alsoExtract = reproducible::asPath(ecozoneFiles),
            studyArea = StudyArea,
            destinationPath = dPath,
            writeTo = "EcozoneFile.shp",
            useCache = FALSE
          ),
          quick = "destinationPath"
        )
      })
    })
    expect_false(all(grepl("loading", mess)))

    # Test useCache -- doesn't error and loads from cache
    mess <- capture_messages({
      warn <- suppressWarningsSpecific(falseWarnings = "attribute variables are assumed to be spatially constant", {
        shpEcozoneSm <- Cache(
          prepInputs(
            url = "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip",
            targetFile = reproducible::asPath(ecozoneFilename),
            alsoExtract = reproducible::asPath(ecozoneFiles),
            studyArea = StudyArea,
            destinationPath = dPath,
            writeTo = "EcozoneFile.shp",
            useCache = TRUE # with useTerra = TRUE, this is only for loading, not postProcess
          ),
          quick = "destinationPath"
        )
      })
    })

    expect_true(any(grepl(.message$LoadedCacheResult(), mess)))

    ##  archive
    ## don't pass url -- use local copy of archive only
    ## use purge = TRUE to rm checksums file, rewrite it here
    shpEcozone <- prepInputs(
      destinationPath = dPath,
      archive = file.path(dPath, "ecozone_shp.zip"),
      purge = TRUE
    )
    expect_true(is(shpEcozone, vectorType()))

    ### archive, alsoExtract char
    shpEcozone <- prepInputs(
      destinationPath = dPath,
      archive = file.path(dPath, "ecozone_shp.zip"),
      ## fmt: skip
      alsoExtract = c(
        "ecozones.dbf", "ecozones.prj", "ecozones.sbn",
        "ecozones.sbx", "ecozones.shp", "ecozones.shx"
      )
    )
    expect_true(is(shpEcozone, vectorType()))

    rm(shpEcozone)
    expect_false(exists("shpEcozone", inherits = FALSE))

    ### url, alsoExtract, archive
    ## try again with url - should *not* download, even though checksums came from the
    ##   prepInputs that had locally generated -- confirming that checksums with a manually copied file will work
    ##   instead of forcing prepInputs to get the file.
    shpEcozone <- prepInputs(
      destinationPath = dPath,
      url = "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip",
      archive = file.path(dPath, "ecozone_shp.zip"),
      ## fmt: skip
      alsoExtract = c(
        "ecozones.dbf", "ecozones.prj", "ecozones.sbn",
        "ecozones.sbx", "ecozones.shp", "ecozones.shx"
      )
    )
    expect_true(is(shpEcozone, vectorType()))

    ## stops if deprecated arguments used
    expect_error(prepInputs(
      destinationPath = dPath,
      url = "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip",
      archive = file.path(dPath, "ecozone_shp.zip"),
      studyArea = StudyArea,
      filename2 = "use_writeTo_instead.shp"
    ))
  })
})

test_that("interactive prepInputs", {
  skip_on_cran()
  skip_on_ci()

  testInit(
    "terra",
    opts = list(
      rasterTmpDir = tempdir2(rndstr(1, 6)),
      reproducible.overwrite = TRUE,
      reproducible.inputPaths = NULL
    ),
    needGoogleDriveAuth = TRUE
  )

  globalNoisy <- capture.output({
    warns <- capture_warnings({
      test <- prepInputs(
        url = "https://drive.google.com/file/d/1BNsUiMqENJa0I8gzhO68K307ySPHbdGk/view?usp=sharing",
        destinationPath = tmpdir
      )
    })

    files <- dir(tmpdir, pattern = "FMA_Boundary")
    expect_true(length(files) == 9)
    expect_true(inherits(test, vectorType()))

    ### url, targetFile
    ## need authentication for this
    warns <- capture_warnings({
      test <- prepInputs(
        targetFile = "FMA_Boundary_Updated.shp",
        url = "https://drive.google.com/file/d/1BNsUiMqENJa0I8gzhO68K307ySPHbdGk",
        destinationPath = tmpdir
      )
    })

    ## There is a meaningless warning for this unit test -- ignore it :
    ## In rgdal::readOGR(dirname(x), fn, stringsAsFactors = stringsAsFactors,  :
    ##                  Z-dimension discarded
    expect_true(inherits(test, vectorType()))

    ## From Bird/Tati project
    testInit(
      "terra",
      opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE),
      needGoogleDriveAuth = TRUE
    )
    birdSpecies <- c("BBWA", "YRWA")
    urls <- c(
      "https://drive.google.com/open?id=1CmzYNpxwWr82PoRSbHWG8yg2cC3hncfb",
      "https://drive.google.com/open?id=11Hxk0CcwJsoAnUgfrwbJhXBJNM5Xbd9e"
    )

    ### url, targetFile, archive
    outsideModule <- Map(
      x = birdSpecies,
      url = urls,
      MoreArgs = list(tmpdir = tmpdir),
      function(x, url, tmpdir) {
        ras <- prepInputs(
          targetFile = paste0(x, "_currmean.asc"),
          archive = paste0(x, "_current.zip"),
          # fun = "raster::raster",
          url = url,
          destinationPath = tmpdir,
          overwrite = TRUE
        )
      }
    )
    expect_true(inherits(outsideModule[[1]], rasterType()))
    expect_true(inherits(outsideModule[[2]], rasterType()))
    # expect_true(inherits(terra::crs(outsideModule[[2]]), "CRS"))
    if (requireNamespace("sf")) {
      expect_true(inherits(sf::st_crs(outsideModule[[1]]), "crs"))
    }
    expect_false(identical(outsideModule[[1]], outsideModule[[2]]))

    ## remove the .prj files -- test "similar"
    ### url, targetFile, archive, alsoExtract similar
    file.remove(grep(
      pattern = "asc|zip|CHECK",
      invert = TRUE,
      value = TRUE,
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))]
    ))

    outsideModule <- Map(
      x = birdSpecies,
      url = urls,
      MoreArgs = list(tmpdir = tmpdir),
      function(x, url, tmpdir, purge) {
        ras <- prepInputs(
          targetFile = paste0(x, "_currmean.asc"),
          archive = paste0(x, "_current.zip"),
          url = url,
          # fun = "raster::raster",
          alsoExtract = "similar",
          destinationPath = tmpdir,
          overwrite = TRUE
        )
      }
    )
    expect_true(inherits(outsideModule[[1]], rasterType()))
    expect_true(inherits(outsideModule[[2]], rasterType()))
    # expect_true(inherits(crs(outsideModule[[2]]), "CRS"))
    # expect_true(inherits(crs(outsideModule[[1]]), "CRS"))
    expect_true(!is.na(crs(outsideModule[[1]])))
    expect_false(identical(outsideModule[[1]], outsideModule[[2]]))

    ## remove the .prj files -- test "similar"
    file.remove(grep(
      pattern = "asc|zip|CHECK",
      invert = TRUE,
      value = TRUE,
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))]
    ))

    ### url, targetFile, archive, alsoExtract NA
    ## because alsoExtract is NA ... no other files are unzipped, so no .prj and so no CRS
    outsideModule <- Map(
      x = birdSpecies,
      url = urls,
      MoreArgs = list(tmpdir = tmpdir),
      function(x, url, tmpdir, purge) {
        ras <- prepInputs(
          targetFile = paste0(x, "_currmean.asc"),
          archive = paste0(x, "_current.zip"),
          url = url,
          alsoExtract = NULL,
          destinationPath = tmpdir,
          overwrite = TRUE
        )
      }
    )
    expect_true(inherits(outsideModule[[1]], rasterType()))
    expect_true(inherits(outsideModule[[2]], rasterType()))
    expect_false(identical(terra::crs(outsideModule[[1]]), "")) # now with subfolders & all files, has crs
    expect_false(identical(outsideModule[[1]], outsideModule[[2]]))
  })
})

test_that("preProcess doesn't work", {
  skip_on_cran()
  skip_on_ci()

  testInit(
    "terra",
    opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE),
    needGoogleDriveAuth = TRUE
  )

  cls <- rasterType()
  # cls <- .fileExtsKnown()[.fileExtsKnown()[, "extension"] == "tif", "type"]

  # Note urlShapefiles1Zip, urlShapefilesZip, and urlTif1 are in helper-allEqual.R
  globalNoisy <- capture.output({
    # # # # # Comment
    # # # url
    # # # # # Comment
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(url = urlTif1, destinationPath = tmpdir)
      })
    })
    runTest("1_2_7_10_13", cls, 1, mess,
            expectedMess = expectedMessage,
            filePattern = "DEM", tmpdir = tmpdir, test = test
    )

    # 2nd time # no targetFile, but since url is simple, can guess correctly
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(url = urlTif1, destinationPath = tmpdir)
      })
    })

    runTest("1_2_8_10", cls, 1, mess,
            expectedMess = expectedMessage,
            filePattern = "DEM", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # url is an archive on googledrive -- can get file.info from remote -- so can do checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(url = urlShapefiles1Zip, destinationPath = tmpdir)
      })
    })

    runTest("1_4_7_10_12_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", # the file name is actually Shapefile1...
            tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warning({
        test <- prepInputs(url = urlShapefiles1Zip, destinationPath = tmpdir)
      })
    })
    runTest("1_8_9_10", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # # # # # Comment
    # # # url, targetFile
    # # # # # Comment
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(url = urlTif1, targetFile = basename(urlTif1), destinationPath = tmpdir)
      })
    })
    runTest("1_2_7_13", cls, 1, mess,
            expectedMess = expectedMessage,
            filePattern = "DEM", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(url = urlTif1, targetFile = basename(urlTif1), destinationPath = tmpdir)
      })
    })
    runTest("1_2_8", cls, 1, mess,
            expectedMess = expectedMessage,
            filePattern = "DEM", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # url is an archive on googledrive --
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefiles1Zip,
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_7_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    ## 2nd time; can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefiles1Zip,
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_8_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # # # # # Comment
    # # # url, alsoExtract
    # # # # # Comment
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(url = urlTif1, alsoExtract = "DEM.tif", destinationPath = tmpdir)
      })
    })
    runTest("1_2_7_10_13", cls, 1, mess,
            expectedMess = expectedMessage,
            filePattern = "DEM", tmpdir = tmpdir, test = test
    )

    # 2nd time # can use checksums, even though don't have targetFile, b/c simple url
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(url = urlTif1, alsoExtract = "DEM.tif", destinationPath = tmpdir)
      })
    })
    runTest("1_2_8_10", cls, 1, mess,
            expectedMess = expectedMessage,
            filePattern = "DEM", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # url is an archive on googledrive --
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefiles1Zip,
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_4_7_10_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can't checksums because no targetfile
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefiles1Zip,
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_8_9_10", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE), recursive = TRUE)

    # # # # # Comment
    # # # url, archive
    # # # # # Comment
    # url is an archive on googledrive -- here, zip has 2 Shapefile filesets -- Shapefile1* and Shapefile2*
    #   should extract all
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          archive = "Shapefiles.zip",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_4_7_10_12_13", vectorType(), 9, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          archive = "Shapefiles.zip",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_8_9_10", vectorType(), 9, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # # # # # Comment
    # # # url, archive, targetFile
    # # # # # Comment
    # url is an archive on googledrive --
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefiles1Zip,
          archive = "Shapefiles1.zip",
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_7_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefiles1Zip,
          archive = "Shapefiles1.zip",
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_8_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # # # # # Comment
    # # # url, targetFile, alsoExtract                        # # # # #
    # # # # # Comment
    # url is an archive on googledrive --
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          targetFile = "Shapefile1.shp",
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_7_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          targetFile = "Shapefile1.shp",
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_8_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          targetFile = "Shapefile1.shp",
          alsoExtract = c("similar"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_7_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlTif1,
          targetFile = "DEM.tif",
          alsoExtract = c("DEM.tif"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_7_13", cls, 1, mess,
            expectedMess = expectedMessage,
            filePattern = "DEM", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # # # # # Comment
    # # # url, archive, alsoExtract               # # #
    # # # # # Comment
    # url is an archive on googledrive --
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          archive = "Shapefiles.zip",
          alsoExtract = "similar",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_4_7_10_12_13", vectorType(), 9, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          archive = "Shapefiles.zip",
          alsoExtract = "similar",
          destinationPath = tmpdir
        )
      })
    })

    runTest("1_8_9_10_12", vectorType(), 9, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    unlink(dir(tmpdir, full.names = TRUE))
    ## Without a `targetFile` and with an incomplete `alsoExtract` (no .shp),
    ## `fun` cannot be guessed. preProcess() (the download/extract layer) is
    ## fine with a NULL fun — but prepInputs() is the loader and must refuse
    ## a call it cannot fulfil. (Commit 67edb6e8 dropped the stop() from the
    ## preProcess layer; prepInputs reinstates it just before process().)
    expect_error({
      mess <- capture_messages({
        warns <- capture_warnings({
          test <- prepInputs(
            url = urlShapefilesZip,
            archive = "Shapefiles.zip",
            alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shx"),
            destinationPath = tmpdir
          )
        })
      })
    })

    unlink(dir(tmpdir, full.names = TRUE))

    # # # # # # Comment
    # # # url, targetFile, alsoExtract               # # #
    # # # # # Comment
    # url is an archive on googledrive --
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          alsoExtract = "similar",
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_7_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          alsoExtract = "similar",
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_8_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shx"),
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_7_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shx"),
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_8_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    unlink(dir(tmpdir, full.names = TRUE))

    # # # # # Comment
    # # # url, archive, targetFile, alsoExtract               # # #
    # # # # # Comment
    # url is an archive on googledrive --
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          archive = "Shapefiles.zip",
          alsoExtract = "similar",
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_7_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          url = urlShapefilesZip,
          archive = "Shapefiles.zip",
          alsoExtract = "similar",
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_8_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # # # # # Comment
    # # # archive
    # # # # # Comment
    # archive exists locally
    # remove all non archive files
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "\\.zip",
      invert = TRUE,
      value = TRUE
    ))

    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(archive = "Shapefiles.zip", destinationPath = tmpdir)
      })
    })

    runTest("1_4_9_10_13", vectorType(), 9, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(archive = "Shapefiles.zip", destinationPath = tmpdir)
      })
    })
    runTest("1_9_10", vectorType(), 9, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # # # # # Comment
    # # # archive, targetFile
    # # # # # Comment
    # archive exists locally
    # remove all non archive files
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "\\.zip",
      invert = TRUE,
      value = TRUE
    ))

    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_9_13", vectorType(), 9, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          targetFile = "Shapefile1.shp",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_9", vectorType(), 9, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # # # # # Comment
    # # # archive, targetFile, alsoExtract                    # # #
    # # # # # Comment
    # archive exists locally
    # remove all non archive files
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "\\.zip",
      invert = TRUE,
      value = TRUE
    ))
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          targetFile = "Shapefile1.shp",
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_9_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          targetFile = "Shapefile1.shp",
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "\\.zip",
      invert = TRUE,
      value = TRUE
    ))
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "CHECKSUMS.txt",
      value = TRUE
    ))
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          targetFile = "Shapefile1.shp",
          alsoExtract = "similar",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_9_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          targetFile = "Shapefile1.shp",
          alsoExtract = c("similar"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # # # # # Comment
    # # # targetFile
    # # # # # Comment
    file.remove(grep(dir(tmpdir, full.names = TRUE), pattern = "CHECKSUMS.txt", value = TRUE))
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(targetFile = "Shapefile1.shp", destinationPath = tmpdir)
      })
    })
    runTest("1_2", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(targetFile = "Shapefile1.shp", destinationPath = tmpdir)
      })
    })
    runTest("1_2", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # # # # # Comment
    # # # targetFile, alsoExtract
    # # # # # Comment
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "CHECKSUMS.txt",
      value = TRUE
    ))
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          targetFile = "Shapefile1.shp",
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          targetFile = "Shapefile1.shp",
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # # # # # Comment
    # # # alsoExtract -- previously failed b/c no information; now ok-- .guessAtTargetAndFun # # #
    # # # # # Comment
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "CHECKSUMS.txt",
      value = TRUE
    ))
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })

    # # # # # Comment
    # # # archive, alsoExtract
    # # # # # Comment
    # archive exists locally
    # remove all non archive files
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "\\.zip",
      invert = TRUE,
      value = TRUE
    ))
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "CHECKSUMS.txt",
      value = TRUE
    ))
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_4_9_10_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shp", "Shapefile1.shx"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_9_10", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # Try without .shp -- fail
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "\\.zip",
      invert = TRUE,
      value = TRUE
    ))
    expect_error({
      mess <- capture_messages({
        warns <- capture_warnings({
          test <- prepInputs(
            archive = "Shapefiles1.zip",
            alsoExtract = c("Shapefile1.dbf", "Shapefile1.prj", "Shapefile1.shx"),
            destinationPath = tmpdir
          )
        })
      })
    })

    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "\\.zip",
      invert = TRUE,
      value = TRUE
    ))
    file.remove(grep(
      dir(tmpdir, full.names = TRUE)[!isDirectory(dir(tmpdir))],
      pattern = "CHECKSUMS.txt",
      value = TRUE
    ))
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          targetFile = "Shapefile1.shp",
          alsoExtract = "similar",
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_4_9_13", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )

    # 2nd time # can checksums
    mess <- capture_messages({
      warns <- capture_warnings({
        test <- prepInputs(
          archive = "Shapefiles.zip",
          targetFile = "Shapefile1.shp",
          alsoExtract = c("similar"),
          destinationPath = tmpdir
        )
      })
    })
    runTest("1_2_9", vectorType(), 5, mess,
            expectedMess = expectedMessage,
            filePattern = "Shapefile", tmpdir = tmpdir, test = test
    )
  })
})

test_that(".resolveDlFunCaptured handles all dlFun forms without eager eval", {
  cap <- function(dlFun) reproducible:::.resolveDlFunCaptured(substitute(dlFun), dlFun)

  # Symbol -> resolves to the function value
  myFn <- function() 42
  expect_true(is.function(cap(myFn)))

  # NULL literal -> NULL
  expect_null(cap(NULL))

  # quote(fn(...)) -> unwraps to inner call
  q <- cap(quote(myFn(x = 1)))
  expect_true(is.call(q))
  expect_identical(deparse(q), "myFn(x = 1)")

  # pkg::fn(args) -> kept as a deferred call; side effect must NOT fire
  ran <- FALSE
  fakePkg <- list(fakeFn = function(...) { ran <<- TRUE; "value" })
  attach(fakePkg, name = "fakePkg", warn.conflicts = FALSE)
  on.exit(detach("fakePkg"), add = TRUE)
  d <- cap(fakePkg::fakeFn(country = "LUX"))
  expect_false(ran)
  expect_true(is.call(d))

  # bare fn(args) -> kept as a deferred call; side effect must NOT fire
  ran2 <- FALSE
  side <- function() { ran2 <<- TRUE; 7 }
  d2 <- cap(side())
  expect_false(ran2)
  expect_true(is.call(d2))

  # Control-flow expression -> evaluated (e.g., if/else)
  useFn <- TRUE
  expect_true(is.function(cap(if (useFn) myFn else NULL)))
})

test_that("prepInputs(dlFun = ...) ignores pre-existing files in destinationPath subdirs", {
  # Regression test for a bug where downloadRemote()'s noTargetFile branch
  # took the "before dlFun" snapshot of destinationPath with
  # `dir(destinationPath, full.names = TRUE)` (non-recursive) and the "after
  # dlFun" snapshot with `dir(..., recursive = TRUE, full.names = TRUE)`.
  # The setdiff() of those mismatched listings classified files that were
  # already present in subdirectories of destinationPath as "newly created
  # by dlFun". They then propagated as `downloadResults$destFile` and tripped
  # a spurious "already exists at ..." stop later in the function.
  #
  # Real-world trigger: a user sets `reproducible.inputPaths` to a shared
  # data stash, then calls `prepInputs(url = some_archive.zip, ...)` whose
  # archive payload extracts into a subdirectory of the stash. A subsequent
  # `prepInputs(dlFun = some_function(...))` for an unrelated dataset would
  # fail with an error mentioning the previous archive's stashed files.
  testInit(
    opts = list(
      reproducible.interactiveOnDownloadFail = FALSE,
      reproducible.inputPaths = NULL,
      reproducible.overwrite = FALSE
    )
  )

  destPath <- file.path(tmpdir, "destSubdir")
  dir.create(destPath, recursive = TRUE)
  preExistingDir <- file.path(destPath, "preExisting")
  dir.create(preExistingDir)
  # Two files whose basenames collide between the subdir and top-level.
  # The collision is what makes the bug surface as an "already exists"
  # stop; without it, the buggy code still mis-identifies the subdir file
  # as new but the desiredPath check happens to pass.
  writeLines("subdir-content",   file.path(preExistingDir, "junk.txt"))
  writeLines("toplevel-content", file.path(destPath, "junk.txt"))

  # dlFun returns a small in-memory object; no network, no side effects on
  # destinationPath. The bug manifests purely from the snapshot logic.
  res <- expect_no_error(
    prepInputs(
      destinationPath = destPath,
      dlFun = function() data.frame(x = 1:3)
    )
  )

  # Sanity: prepInputs returned the dlFun's value, and the pre-existing
  # files are untouched.
  expect_s3_class(res, "data.frame")
  expect_identical(res$x, 1:3)
  expect_identical(readLines(file.path(preExistingDir, "junk.txt")), "subdir-content")
  expect_identical(readLines(file.path(destPath, "junk.txt")), "toplevel-content")
})

test_that("prepInputs when fun = NA", {
  skip_on_cran()

  ## Probe the GADM host with a short timeout. The test below calls
  ## geodata::gadm(), which talks to geodata.ucdavis.edu — when that endpoint
  ## is slow/unreachable (SSL hiccups, DNS, packet drops) geodata's internal
  ## retry loop can stall the test for minutes. Bail out cleanly instead.
  skip_if_not(.requireNamespace("httr2"), "httr2 not available for probe")
  gadmReachable <- tryCatch({
    httr2::request("https://geodata.ucdavis.edu/") |>
      httr2::req_method("HEAD") |>
      httr2::req_timeout(5) |>
      httr2::req_error(is_error = function(resp) FALSE) |>
      httr2::req_perform()
    TRUE
  }, error = function(e) FALSE)
  skip_if_not(isTRUE(gadmReachable), "geodata.ucdavis.edu unreachable")

  testInit(
    c("sf", "terra"),
    opts = list(
      rasterTmpDir = tempdir2(rndstr(1, 6)),
      reproducible.interactiveOnDownloadFail = FALSE,
      reproducible.inputPaths = NULL,
      reproducible.overwrite = TRUE
    ),
    needGoogleDriveAuth = TRUE
  )

  globalNoisy <- capture.output({
    coords <- structure(c(6, 6.1, 6.2, 6.15, 6, 49.5, 49.7, 49.8, 49.6, 49.5), .Dim = c(5L, 2L))
    StudyArea <- terra::vect(coords, "polygons")
    terra::crs(StudyArea) <- crsToUse

    mess1 <- capture_messages({
      test1 <- try(silent = TRUE, {
        prepInputs(
          fun = NA,
          dlFun = getDataFn,
          name = "GADM",
          country = "LUX",
          level = 0,
          path = tmpdir,
          quiet = TRUE
        )
      })
    })
    if (!is(test1, "try-error") && !any(grepl("out of service", mess1))) {
      expect_true(is(test1, "SpatVector"))
      # test quoted version of `dlFun`
      mess3 <- capture_messages({
        test3 <- prepInputs(
          fun = NA,
          dlFun = quote(getDataFn(name = "GADM", country = "LUX", level = 0, path = tmpdir)),
          destinationPath = tmpdir
        )
      })
      expect_true(is(test3, "SpatVector"))

      if (.requireNamespace("sf")) {
        mess6 <- capture_messages({
          test6 <- prepInputs(
            # targetFile = targetFileLuxRDS,
            dlFun = quote({
              out <- getDataFn(name = "GADM", country = "LUX", level = 0, path = tmpdir)
              sf::st_as_sf(out)
            }),
            tmpdir = tmpdir
          )
        })

        expect_is(test6, "sf")
      }
    }
  })
})

test_that("load rdata in prepInputs", {
  testInit(
    "terra",
    tmpFileExt = "rda",
    opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE),
    needGoogleDriveAuth = TRUE
  )
  a <- 1
  b <- 2
  save(a, b, file = tmpfile)
  aa <- prepInputs(tmpfile, fun = "base::load")
  expect_true(identical(aa, list(a = a, b = b)))

  d <- new.env(parent = emptyenv())
  aa <- prepInputs(tmpfile, fun = "base::load", envir = d)
  expect_false(identical(aa, list(a = a, b = b))) # not in aa, because loaded to d
  expect_true(identical(as.list(d), list(a = a, b = b)))
})

test_that("assessDataType doesn't work", {
  testInit(
    "terra",
    opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE),
    needGoogleDriveAuth = TRUE
  )

  ## LOG1S
  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- c(0, NaN, rep(c(0, 1), 49))
  expect_true(assessDataType(ras) == "LOG1S")

  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- rep(c(0, 1), 50)
  expect_true(assessDataType(ras) == "LOG1S")

  ras[] <- rep(c(TRUE, FALSE), 50)
  expect_true(assessDataType(ras) == "LOG1S")

  ras[] <- c(NA, NA, rep(c(0, 1), 49))
  expect_true(assessDataType(ras) == "LOG1S")

  ## INT1S
  ras[] <- -1:98
  expect_true(assessDataType(ras) == "INT1S")

  ras[] <- c(NA, -1:97)
  expect_true(assessDataType(ras) == "INT1S")

  ## INT1U
  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- 1:100
  expect_true(assessDataType(ras) == "INT1U")

  ras[] <- c(NA, 2:100)
  expect_true(assessDataType(ras) == "INT1U")

  ## INT2U
  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = 64000, max = 65000))
  expect_true(assessDataType(ras) == "INT2U")

  ## INT2S
  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = -32767, max = 32767))
  expect_true(assessDataType(ras) == "INT2S")

  ras[54] <- NA
  expect_true(assessDataType(ras) == "INT2S")

  ## INT4U
  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = 0, max = 500000000))
  expect_true(assessDataType(ras) == "INT4U")

  ras[14] <- NA
  expect_true(assessDataType(ras) == "INT4U")

  ## INT4S
  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = -200000000, max = 200000000))
  expect_true(assessDataType(ras) == "INT4S")

  ras[14] <- NA
  expect_true(assessDataType(ras) == "INT4S")

  ## FLT4S
  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- runif(100, min = -10, max = 87)
  expect_true(assessDataType(ras) == "FLT4S")

  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = -3.4e+26, max = 3.4e+28))
  expect_true(assessDataType(ras) == "FLT4S")

  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = 3.4e+26, max = 3.4e+28))
  expect_true(assessDataType(ras) == "FLT4S")

  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = -3.4e+26, max = -1))
  expect_true(assessDataType(ras) == "FLT4S")

  ## FLT8S
  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = -1.7e+30, max = 1.7e+308))
  expect_true(assessDataType(ras) == "FLT8S")

  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = 1.7e+30, max = 1.7e+308))
  expect_true(assessDataType(ras) == "FLT8S")

  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- round(runif(100, min = -1.7e+308, max = -1))
  expect_true(assessDataType(ras) == "FLT8S")

  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- c(-Inf, 1, rep(c(0, 1), 49))
  expect_true(assessDataType(ras) == "FLT8S")

  ras <- terra::rast(ncols = 10, nrows = 10)
  ras[] <- c(Inf, 1, rep(c(0, 1), 49))
  expect_true(assessDataType(ras) == "FLT8S")
})

test_that("assessDataType for categorical rasters", {
  testInit(c("terra", "raster"))

  r <- terra::rast(terra::ext(c(0, 2, 0, 2)), vals = 1:4, resolution = 1)
  levels(r) <- data.frame(ID = 1:4, Lett = LETTERS[1:4])
  expect_identical(assessDataType(r), "INT1U")

  r <- raster::raster(raster::extent(c(0, 2, 0, 2)), vals = 1:4, resolution = 1)
  levels(r) <- data.frame(ID = 1:4, Lett = LETTERS[1:4])
  expect_identical(assessDataType(r), "INT1U")
})

test_that("lightweight tests for code coverage", {
  skip_on_cran()

  testInit(
    c("sf", "terra"),
    opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE),
    needGoogleDriveAuth = TRUE
  )

  url <- "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip"
  globalOutput <- capture.output({
    checkPath(tmpdir, create = TRUE)
    checkSums <- .emptyChecksumsResult
    checkSumFilePath <- file.path(tmpdir, "CHECKSUMS.txt")

    downloadFile(
      url = url,
      neededFiles = "ecozones.shp",
      checkSums = checkSums,
      archive = "ecozone_shp.zip",
      needChecksums = TRUE,
      quick = FALSE,
      destinationPath = tmpdir,
      checksumFile = checkSumFilePath
    )
    expect_true(file.exists(dir(tmpdir, pattern = "ecozone", full.names = TRUE)))

    # have local copy
    unzip("ecozone_shp.zip", exdir = tmpdir)
    expect_true(all(file.copy(dir(file.path(tmpdir, "Ecozones"), full.names = TRUE), tmpdir)))
    checkSums <- Checksums(path = tmpdir, write = TRUE)

    aMess <- capture_messages(downloadFile(
      url = url,
      neededFiles = "ecozones.shp",
      checkSums = checkSums,
      targetFile = "ecozones.shp",
      archive = NULL,
      needChecksums = TRUE,
      quick = FALSE,
      destinationPath = file.path(tmpdir, "Ecozones"),
      checksumFile = file.path(tmpdir, "CHECKSUMS.txt")
    ))

    if (!isMac()) {
      expect_true(any(grepl("Skipping download", aMess))) ## 2023-05-08: fails on macOS
    }

    filesForShp <- dir(file.path(tmpdir), pattern = "ecozones", full.names = TRUE)
    expect_true(all(file.copy(filesForShp, tmpCache)))
    # Need these in a test further down -- mostly just need the CRS
    filesForShp2 <- dir(file.path(tmpCache), pattern = "ecozones", full.names = TRUE)
    if (.requireNamespace("sf")) {
      shpFile <- sf::st_read(grep(filesForShp2, pattern = "\\.shp", value = TRUE))
    }
    # Test when wrong archive exists, wrong checkSums
    expect_true(file.remove(file.path(tmpdir, "ecozone_shp.zip")))
    expect_true(all(file.remove(filesForShp)))
    expect_true(file.create(file.path(tmpdir, "ecozone_shp.zip")))
    checkSums <- Checksums(path = tmpdir, write = TRUE)
    expect_true(file.remove(file.path(tmpdir, "ecozone_shp.zip")))
    checkSums <- Checksums(path = tmpdir)

    out <- try(
      silent = TRUE,
      downloadFile(
        url = url,
        ## fmt: skip
        neededFiles = c(
          "ecozones.dbf", "ecozones.prj", "ecozones.sbn",
          "ecozones.sbx", "ecozones.shp", "ecozones.shx"
        ),
        checkSums = checkSums,
        targetFile = "ecozones.shp",
        archive = "ecozone_shp.zip",
        needChecksums = TRUE,
        quick = FALSE,
        destinationPath = tmpdir,
        checksumFile = checkSumFilePath
      )
    )

    ## try to purge from the CHECKSUMS.txt
    toPurgeCode <- grep(
      "purgeChecksums|fileToRemove",
      capture.output(attr(out, "condition")),
      value = TRUE
    )
    toPurgeCode <- parse(text = gsub(">", "", toPurgeCode))
    checksumsFile <- dir(tmpdir, pattern = "CHECKSUMS.txt", full.names = TRUE)
    dtBefore <- data.table::fread(checksumsFile)
    eval(toPurgeCode)
    dtAfter <- data.table::fread(checksumsFile)
    expect_equivalent(NROW(dtBefore[!dtAfter, on = "file"]), 1L)

    ## 2023-05-08: does not error on macOS
    isErr <- is(out, "try-error")
    # if (isMac()) expect_false(isErr) else
    expect_true(isErr)

    ## postProcess.default
    b <- 1
    expect_no_error(postProcess(b))

    ## postProcess.list
    b <- list(1, 1)
    expect_no_error(postProcess(b))

    ras <- terra::rast(terra::ext(0, 10, 0, 10), resolution = 1, vals = 1:100)
    terra::crs(ras) <- crsToUse

    expect_error(postProcess(ras, studyArea = 1), .message$Greps$anySpatialClass)
    expect_error(postProcess(ras, rasterToMatch = 1), .message$Greps$anySpatialClass)

    ## cropInputs.default
    b <- 1
    a <- cropInputs(b)
    expect_true(identical(a, b))

    ras2 <- terra::rast(terra::ext(0, 5, 0, 5), resolution = 1, vals = 1:25)
    terra::crs(ras2) <- crsToUse
    a <- cropInputs(ras, extentToMatch = terra::ext(ras2), extentCRS = terra::crs(ras2))
    expect_true(.isSpatRaster(a))

    ras4 <- terra::rast(terra::ext(7, 11, 7, 11), resolution = 1, vals = 1:16)
    sp4 <- terra::vect(terra::ext(ras4))
    terra::crs(sp4) <- crsToUse
    # sp4 <- sf::st_as_sfc(sf::st_bbox(ras4))
    # sf::st_crs(sp4) <- crsToUse

    grepMessHere <- "extents do not overlap"
    expect_error(cropInputs(ras2, studyArea = sp4), grepMessHere)

    ras3 <- terra::rast(terra::ext(0, 5, 0, 5), resolution = 1, vals = 1:25)
    terra::crs(ras3) <- crsToUse

    ## Different crs
    ## Because studyArea is a Raster, then it doesn't work correctly
    a <- cropInputs(ras2, studyArea = ras3)
    expect_true(.isSpatRaster(a))
    expect_true(identical(terra::crs(a), terra::crs(ras2)))

    # Now rasterToMatch used -- internally reprojects it to x
    a <- cropInputs(ras2, rasterToMatch = ras3)
    expect_true(.isSpatRaster(a))
    expect_true(identical(terra::crs(a), terra::crs(ras2)))

    ## fixErrors.default
    b <- 1
    a <- fixErrors(b)
    expect_true(identical(a, b))

    ## projectInputs.Raster
    a <- projectInputs(ras2, rasterToMatch = ras3, method = "near")
    expect_true(.isSpatRaster(a))
    expect_true(identical(terra::crs(a), terra::crs(ras3)))

    a <- projectInputs(ras2, targetCRS = terra::crs(ras3), rasterToMatch = ras3, method = "near")
    expect_true(.isSpatRaster(a))
    expect_true(identical(terra::crs(a), terra::crs(ras3)))

    ## warns if bilinear is passed for reprojecting integer
    if (.requireNamespace("sf")) {
      expect_warning(projectInputs(ras2, targetCRS = terra::crs(shpFile), method = "bilinear"))
    }

    ## Works with no rasterToMatch
    a <- projectInputs(ras2, targetCRS = crs(ras3), method = "near")
    expect_true(identical(crs(a), crs(ras3)))

    # }
  })
  # sp::CRS("+proj=lcc +lat_1=49 +lat_2=77 +lat_0=0 +lon_0=-95 +x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs"))
})

test_that("lightweight tests 2 for code coverage", {
  skip_on_cran()

  testInit(
    "terra",
    opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE),
    needGoogleDriveAuth = TRUE
  )

  theZipFile <- tempfile(tmpdir = tmpdir, fileext = ".zip")
  theZipFile2 <- tempfile(tmpdir = tmpdir, fileext = ".zip")
  theZipFile3 <- tempfile(tmpdir = tmpdir, fileext = ".zip")
  theZipName <- file.path(tmpdir, "hi.zip")
  theZapFile <- tempfile(tmpdir = tmpdir, fileext = ".zap")
  theRDSFile <- tempfile(tmpdir = tmpdir, fileext = ".rds")
  a <- 1
  saveRDS(a, file = theRDSFile)

  withr::with_dir(dirname(theRDSFile), {
    utils::zip(zipfile = theZipFile, files = basename(theRDSFile), flags = "-q")
    utils::zip(zipfile = theZipFile2, files = basename(theZipFile), flags = "-q")
    utils::zip(zipfile = theZipFile3, files = basename(theZipFile2), flags = "-q")
  })

  expect_error(extractFromArchive(theZapFile), "Archives of type zap are not currently supported")
  oo <- capture.output(
    type = "message",
    expect_error(extractFromArchive(theZipName), "No archive exists with filename")
  )

  extractFromArchive(theZipFile, neededFiles = character())

  csfp <- file.path(tmpdir, "CHECKSUMS.txt")
  data.table::fwrite(.emptyChecksumsFileContent, file = csfp, sep = "\t")

  # check Checksums fn
  a <- extractFromArchive(
    theZipFile,
    neededFiles = character(),
    checkSumFilePath = csfp,
    destinationPath = tmpdir
  )
  expect_true(file.exists(a$filesExtracted))
  # check Checksums fn

  expect_error(
    oo <- capture.output(type = "message",
    suppressWarnings(extractFromArchive(
      theZipFile,
      neededFiles = character(),
      checkSumFilePath = theRDSFile,
      destinationPath = tmpdir
    )),
    "checkSumFilePath is not a CHECKSUMS.txt"
  ))

  # Doubley nested zips -- extract inner, inner
  a <- extractFromArchive(
    c(theZipFile2, theZipFile),
    neededFiles = character(),
    checkSumFilePath = csfp,
    destinationPath = tmpdir
  )
  expect_true(isTRUE(all(file.exists(a$filesExtracted))))

  # triply
  a <- extractFromArchive(
    theZipFile3,
    neededFiles = theRDSFile,
    checkSumFilePath = csfp,
    destinationPath = tmpdir,
    .tempPath = tempdir2()
  )
  expect_true(length(a$extractedArchives) == 3)
  expect_true(length(a$filesExtracted) == 3)
  expect_true(all(basename(a$filesExtracted) %in% basename(c(theZipFile, theZipFile2, theRDSFile))))
  expect_true(all(
    basename(a$extractedArchives) %in% basename(c(theZipFile, theZipFile2, theZipFile3))
  ))

  allZipsAndRDS <- c(theZipFile, theZipFile2, theZipFile3, theRDSFile)
  Checksums(tmpdir, write = TRUE, files = allZipsAndRDS, overwrite = TRUE)
  a <- extractFromArchive(
    theZipFile3,
    neededFiles = theRDSFile,
    checkSumFilePath = csfp,
    destinationPath = tmpdir,
    checkSums = Checksums(tmpdir, files = allZipsAndRDS)
  )
})

test_that("options inputPaths", {
  skip_on_cran()
  skip_if_not_installed("geodata")

  testInit(
    c("terra", "geodata"),
    opts = list(reproducible.inputPaths = NULL, reproducible.inputPathsRecursive = FALSE),
    needInternet = TRUE
  )
  # SSL_REVOKE_BEST_EFFORT() # uses withr::defer to remove it after this test
  f <- formals3(prepInputs)
  getDataFn <- getDataFn # not exported from reproducible; can access here, not in the dlFun

  withr::local_options("reproducible.inputPaths" = NULL)
  withr::local_options("reproducible.inputPathsRecursive" = FALSE)

  mess1 <- capture_messages({
    test0 <- try(getDataFn(path = tmpdir, country = "LUX", quiet = TRUE), silent = TRUE)
  })

  useGADM <- !is(test0, "try-error") && any(grepl("server seems|server is", mess1)) %in% FALSE # NROW(dir(tmpdir, recursive = TRUE)) > 0

  if (useGADM) {
    # noisyOutput <- capture.output(type = "message", {
    mess1 <- capture_messages({
      test1 <- try(prepInputs(
        destinationPath = tmpdir,
        # url = if (!useGADM) url2 else f$url,
        # targetFile = if (useGADM) theFile else f$targetFile,
        dlFun = getDataFn,
        name = "GADM",
        country = "LUX",
        level = 0,
        path = tmpdir,
        quiet = TRUE
      ))
    })
    # })
  }

  theFile <- if (useGADM) {
    targetFileLuxRDS
  } else {
    "rasterTest.tif"
  }
  url2 <- "https://github.com/tati-micheletti/host/raw/master/data/rasterTest.tif"

  mess1 <- capture_messages({
    test1 <- try(prepInputs(
      destinationPath = tmpdir,
      url = if (!useGADM) url2 else f$url,
      targetFile = if (useGADM) theFile else f$targetFile,
      dlFun = if (useGADM) getDataFn else NULL,
      name = if (useGADM) "GADM" else NULL,
      country = if (useGADM) "LUX" else NULL,
      level = if (useGADM) 0 else NULL,
      path = if (useGADM) tmpdir else NULL
    ))
  })

  # Use inputPaths -- should do a link to tmpCache (the destinationPath)
  withr::local_options("reproducible.inputPaths" = tmpdir)
  withr::local_options("reproducible.inputPathsRecursive" = FALSE)
  dlFun1 <- if (useGADM) getDataFn else NULL

  mess1 <- capture_messages({
    test1 <- prepInputs(
      url = if (!useGADM) url2 else f$url,
      targetFile = if (useGADM) theFile else f$targetFile,
      dlFun = dlFun1,
      name = if (useGADM) "GADM" else NULL,
      country = if (useGADM) "LUX" else NULL,
      level = if (useGADM) 0 else NULL,
      path = if (useGADM) tmpdir else NULL,
      destinationPath = tmpCache,
      getDataFn = dlFun1,
      verbose = 2
    )
  })
  expect_true(sum(grepl(paste0("Hardlinked", ".*:"), mess1)) == 1)

  ## Now two folders - file not in destinationPath, not in 1st inputPaths, but yes 2nd
  ##   should hardlink from 2nd IP to destinationPath, make sure CHECKSUMS.txt is correct in both
  withr::local_options("reproducible.inputPaths" = c(tmpdir, tmpCache))
  file.remove(file.path(tmpdir, theFile))
  tmpdir3 <- file.path(tmpCache, "test")
  mess1 <- capture_messages({
    test1 <- prepInputs(
      url = if (!useGADM) url2 else f$url,
      targetFile = if (useGADM) theFile else f$targetFile,
      dlFun = if (useGADM) getDataFn else NULL,
      name = if (useGADM) "GADM" else NULL,
      country = if (useGADM) "LUX" else NULL,
      level = if (useGADM) 0 else NULL,
      path = if (useGADM) tmpdir else NULL,
      destinationPath = tmpdir3,
      verbose = 2
    )
  })
  expect_true(sum(grepl(paste0(hardlinkOrSymlinkMessagePrefixForGrep), mess1)) == 1)
  ## because the targetFile could be absent or present, this may create a spurious message
  ##  that we don't need to test for
  mess1b <- grep("targetFile was not supplied", mess1, invert = TRUE, value = TRUE)
  expect_true(sum(grepl(paste0(tmpdir3), mess1b)) == 2)

  ## THIS NEXT ONE DOESN"T PASS ON GitHub Actions on WINDOWS, skip it;
  ##  should copy from 2nd directory (tmpCache) because it is removed in the lower
  ##  tmpdir directory & has a CHECKSUMS.txt

  skip_on_ci()
  url_2 <- if (!useGADM) url2 else f$url
  targetFile_2 <- if (useGADM) theFile else f$targetFile
  dlFun_2 <- if (useGADM) getDataFn else NULL
  name_2 <- if (useGADM) "GADM" else NULL
  country_2 <- if (useGADM) "LUX" else NULL
  level_2 <- if (useGADM) 0 else NULL
  path_2 <- if (useGADM) tmpdir else NULL

  withr::local_options("reproducible.inputPaths" = tmpdir)
  withr::local_options("reproducible.inputPathsRecursive" = TRUE)
  file.remove(file.path(tmpCache, theFile))
  tmpdir1 <- file.path(tmpCache, "test1")
  warns <- capture_warnings(
    mess1 <- capture_messages({
      test1 <- prepInputs(
        url = url_2,
        targetFile = targetFile_2,
        dlFun = dlFun_2,
        name = name_2,
        country = country_2,
        level = level_2,
        path = path_2,
        destinationPath = tmpdir1,
        verbose = 3
      )
    })
  )

  mess1 <- gsub("\n    ", " ", mess1) ## remove misc new lines
  expect_true(sum(grepl(paste0(hardlinkOrSymlinkMessagePrefixForGrep), mess1)) == 1)
  expect_true(sum(grepl(whPointsToMessForGrep, mess1)) == 1)
  expect_true(sum(grepl(paste0(file.path(tmpdir1, theFile), ".+which point.+"), mess1)) == 1)
  expect_true(sum(basename(dir(file.path(tmpdir), recursive = TRUE)) %in% theFile) == 3)

  ## Try download to inputPath, intercepting the destination, creating a link
  testInit(
    "terra",
    opts = list(reproducible.inputPaths = NULL, reproducible.inputPathsRecursive = FALSE)
  )

  withr::local_options("reproducible.inputPaths" = tmpdir)
  tmpdir2 <- file.path(tmpdir, rndstr(1, 5))
  url_2 = if (!useGADM) url2 else f$url
  targetFile_2 = if (useGADM) theFile else f$targetFile
  dlFun_2 = if (useGADM) getDataFn else NULL
  name_2 = if (useGADM) "GADM" else NULL
  country_2 = if (useGADM) "LUX" else NULL
  level_2 = if (useGADM) 0 else NULL
  path_2 = if (useGADM) tmpdir else NULL

  # noisyOutput <- capture.output(type = "message", {
  mess1 <- capture_messages({
    test1 <- prepInputs(
      url = url_2,
      targetFile = targetFile_2,
      dlFun = dlFun_2,
      name = name_2,
      country = country_2,
      level = level_2,
      path = path_2,
      destinationPath = tmpdir2,
      quiet = TRUE
    )
  })
  # })

  ## Must remove the link that happens during downloading to a .tempPath
  test10 <- grep(hardlinkOrSymlinkMessagePrefixForGrep, mess1, value = TRUE)
  test10 <- grep(tmpdir2, test10, invert = TRUE, value = TRUE)
  expect_true(length(test10) == (1)) #

  ## Have file in inputPath, not in destinationPath
  unlink(file.path(tmpdir2, theFile))
  expect_false(file.exists(file.path(tmpdir2, theFile))) # FALSE -- confirm previous line
  expect_true(file.exists(file.path(tmpdir, theFile))) # TRUE b/c is in getOption('reproducible.inputPaths')
  tmpdir2 <- file.path(tmpdir, rndstr(1, 5))
  url_2 = if (!useGADM) url2 else f$url
  targetFile_2 = if (useGADM) theFile else f$targetFile
  dlFun_2 = if (useGADM) getDataFn else NULL
  name_2 = if (useGADM) "GADM" else NULL
  country_2 = if (useGADM) "LUX" else NULL
  level_2 = if (useGADM) 0 else NULL
  path_2 = if (useGADM) tmpdir else NULL

  mess1 <- capture_messages({
    test1 <- prepInputs(
      url = url_2,
      targetFile = targetFile_2,
      dlFun = dlFun_2,
      name = name_2,
      country = country_2,
      level = level_2,
      path = path_2,
      destinationPath = tmpdir2,
      verbose = 3
    )
  })
  expect_true(sum(grepl(hardlinkOrSymlinkMessagePrefixForGrep, mess1)) == 1) # used a linked version
  expect_true(sum(grepl(paste0("Hardlinked.*"), mess1)) == 1) # it is now in tmpdir2, i.e., the destinationPath
  expect_true(sum(grepl(paste0(basename(tmpdir2)), mess1)) %in% 2:3) # it is now in tmpdir2, i.e., the destinationPath

  ## Have file in destinationPath, not in inputPath
  unlink(file.path(tmpdir, theFile))
  expect_false(file.exists(file.path(tmpdir, theFile))) # FALSE -- confirm previous line
  expect_true(file.exists(file.path(tmpdir2, theFile))) # TRUE b/c is in getOption('reproducible.inputPaths')
  url_2 = if (!useGADM) url2 else f$url
  targetFile_2 = if (useGADM) theFile else f$targetFile
  dlFun_2 = if (useGADM) getDataFn else NULL
  name_2 = if (useGADM) "GADM" else NULL
  country_2 = if (useGADM) "LUX" else NULL
  level_2 = if (useGADM) 0 else NULL
  path_2 = if (useGADM) tmpdir else NULL
  mess1 <- capture_messages({
    test1 <- prepInputs(
      url = url_2,
      targetFile = targetFile_2,
      dlFun = dlFun_2,
      name = name_2,
      country = country_2,
      level = level_2,
      path = path_2,
      overwrite = TRUE,
      destinationPath = tmpdir2,
      verbose = 2
    )
  })
  # expect_true(sum(grepl(hardlinkOrSymlinkMessagePrefixForGrep, mess1)) == 1) # used a linked version
  # expect_true(sum(grepl(paste0("Hardlinked.*",basename(tmpdir2)), mess1)) == 1) # it is now in tmpdir2, i.e., the destinationPath

  ## Try with inputPaths == destinationPath
  unlink(file.path(tmpdir, theFile))
  unlink(file.path(tmpdir2, theFile))
  expect_false(file.exists(file.path(tmpdir, theFile))) # FALSE -- confirm previous line
  expect_false(file.exists(file.path(tmpdir2, theFile))) # TRUE b/c is in getOption('reproducible.inputPaths')
  withr::local_options("reproducible.inputPaths" = tmpdir)
  url_2 = if (!useGADM) url2 else f$url
  targetFile_2 = if (useGADM) theFile else f$targetFile
  dlFun_2 = if (useGADM) getDataFn else NULL
  name_2 = if (useGADM) "GADM" else NULL
  country_2 = if (useGADM) "LUX" else NULL
  level_2 = if (useGADM) 0 else NULL
  path_2 = if (useGADM) tmpdir else NULL
  # noisyOutput <- capture.output(type = "message", {
  mess1 <- capture_messages({
    test1 <- prepInputs(
      url = url_2,
      targetFile = targetFile_2,
      dlFun = dlFun_2,
      name = name_2,
      country = country_2,
      level = level_2,
      path = path_2,
      destinationPath = tmpdir,
      verbose = 2
    )
  })
  # })
  objType <- if (useGADM) vectorType() else rasterType()
  expect_true(is(test1, objType) || is(test1, "SpatVector"))
  test11 <- grep(hardlinkOrSymlinkMessagePrefixForGrep, mess1, value = TRUE)
  test11 <- grep(tmpdir, test11, invert = TRUE)
  expect_true(length(test11) == 0) # no link made b/c identical dir
  expect_true(sum(grepl(paste0("Hardlinked.*", basename(tmpdir2)), mess1)) == 0) # no link made b/c identical dir
})

test_that("writeOutputs saves factor rasters with .grd class to preserve levels", {
  skip_on_cran()

  testInit(
    "terra",
    opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE),
    needGoogleDriveAuth = TRUE
  )
  a <- terra::rast(terra::ext(0, 2, 0, 2), resolution = 1, vals = c(1, 1, 2, 2))
  levels(a) <- data.frame(ID = 1:2, Factor = c("This", "That"))
  tifTmp <- tempfile(tmpdir = tmpdir, fileext = ".tif")
  file.create(tifTmp)
  tifTmp <- normPath(tifTmp)

  b1 <- suppressWarnings(terra::writeRaster(a, filename = tifTmp, overwrite = TRUE)) # the GDAL>6 issue
  b1a <- writeOutputs(a, writeTo = tifTmp)
  expect_equivalent(b1, b1a)
  expect_equivalent(b1[], b1a[])

  expect_true(identical(normPath(Filenames(b1)), normPath(tifTmp)))
})

test_that("rasters aren't properly resampled", {
  skip_on_cran()

  testInit(
    "terra",
    opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE),
    needGoogleDriveAuth = TRUE
  )
  a <- terra::rast(terra::ext(0, 20, 0, 20), resolution = 2, vals = as.integer(1:100 * 4))
  b <- terra::rast(terra::ext(0, 30, 0, 30), resolution = c(3, 3), vals = 1L:100L)
  crs(a) <- crsToUse
  crs(b) <- crsToUse

  tiftemp1 <- normPath(tempfile(tmpdir = tmpdir, fileext = ".tif"))
  tiftemp2 <- normPath(tempfile(tmpdir = tmpdir, fileext = ".tif"))

  suppressWarnings({
    a <- terra::writeRaster(a, filename = tiftemp1, datatype = "INT2U")
    b <- terra::writeRaster(b, filename = tiftemp2, datatype = "INT2U")
  }) ## TODO: temporary GDAL>6

  # Test bilinear --> but keeps integer if it is integer
  suppressWarnings({
    out2 <- prepInputs(
      targetFile = tiftemp1,
      rasterToMatch = terra::rast(tiftemp2),
      destinationPath = dirname(tiftemp1),
      method = "bilinear",
      datatype = "INT2S",
      writeTo = tempfile(tmpdir = tmpdir, fileext = ".tif")
    )
  }) # about "raster layer has integer values"

  if (!isWindows()) {
    expect_true(dataType2(out2) %in% c("INT2S")) # because of "bilinear", it can become negative

    rrr1 <- terra::rast(terra::ext(0, 20, 0, 20), resolution = 1, vals = runif(400, 0, 1))
    terra::crs(rrr1) <- crsToUse
    tiftemp3 <- tempfile(tmpdir = tmpdir, fileext = ".tif")
    tiftemp4 <- tempfile(tmpdir = tmpdir, fileext = ".tif")
    suppressWarningsSpecific(terra::writeRaster(rrr1, filename = tiftemp3), proj6Warn)

    out3 <- prepInputs(
      targetFile = tiftemp3,
      rasterToMatch = terra::rast(tiftemp2),
      destinationPath = dirname(tiftemp3),
      writeTo = tempfile(tmpdir = tmpdir, fileext = ".tif")
    )
    expect_true(dataType2(out3) == "FLT4S")

    # Test for raster::stack
    rasStack <- c(terra::rast(tiftemp3), terra::rast(tiftemp3))
    rasStack[] <- rasStack[]
    rasStack[131][1] <- 1.5
    tiftemp4 <- tempfile(tmpdir = tmpdir, fileext = ".tif")

    rasStack <- terra::writeRaster(rasStack, filename = tiftemp4)
    rm(rasStack)
    warns <- capture_warnings(
      # rasters aren't properly resampled ─────
      # partial argument match of 'ncol' to 'ncols'
      out3 <- prepInputs(
        targetFile = tiftemp4,
        rasterToMatch = terra::rast(tiftemp2),
        destinationPath = dirname(tiftemp3),
        writeTo = tempfile(tmpdir = tmpdir, fileext = ".tif")
      )
    )
    expect_true(is(out3, rasterType()))
    expect_true(identical(length(Filenames(out3)), 1L))

    if (.requireNamespace("raster")) {
      rasterStackFn <- "raster::stack"
      suppressWarningsSpecific(falseWarnings = "partial argument match", {
        out4 <- prepInputs(
          targetFile = tiftemp4,
          rasterToMatch = terra::rast(tiftemp2),
          destinationPath = dirname(tiftemp3),
          fun = rasterStackFn,
          writeTo = c(
            tempfile(tmpdir = tmpdir, fileext = ".grd"),
            tempfile(tmpdir = tmpdir, fileext = ".grd")
          )
        )
      })
      expect_true(is(out4, rasterType(nlayers = nlayers2(out4), rasterRead = rasterStackFn)))
      expect_true(identical(length(Filenames(out4, allowMultiple = TRUE)), 4L))

      # Test for raster::stack with 3 layers, different types of writeRaster file ext
      rasStack <- c(terra::rast(tiftemp3), terra::rast(tiftemp3), terra::rast(tiftemp3))
      rasStack[] <- rasStack[]
      rasStack[131][1] <- 1.5
      rasStack[131][2] <- 2.5
      tiftemp5 <- tempfile(tmpdir = tmpdir, fileext = ".tif")

      rasStack <- writeRaster(rasStack, filename = tiftemp5)
      rm(rasStack)
      suppressWarningsSpecific(falseWarnings = "partial argument match", {
        out5 <- prepInputs(
          targetFile = tiftemp5,
          rasterToMatch = terra::rast(tiftemp2),
          destinationPath = dirname(tiftemp3),
          fun = rasterStackFn,
          writeTo = c(
            tempfile(tmpdir = tmpdir, fileext = ".grd"),
            tempfile(tmpdir = tmpdir, fileext = ".grd"),
            tempfile(tmpdir = tmpdir, fileext = ".tif")
          )
        )
      })
      expect_true(is(out5, "RasterStack"))
      expect_true(identical(length(Filenames(out5, allowMultiple = TRUE)), 5L))

      suppressWarningsSpecific(falseWarnings = "partial argument match", {
        out4 <- prepInputs(
          targetFile = tiftemp4,
          rasterToMatch = terra::rast(tiftemp2),
          destinationPath = dirname(tiftemp3),
          fun = rasterStackFn,
          writeTo = c(
            tempfile(tmpdir = tmpdir, fileext = ".grd"),
            tempfile(tmpdir = tmpdir, fileext = ".grd")
          )
        )
      })
      expect_true(is(out4, rasterType(nlayers2(out4), rasterStackFn)))
      expect_true(identical(length(Filenames(out4)), 4L))
    }
  }
})

test_that("test prepInputs url when a directory", {
  skip_on_cran()
  skip_if_not_installed("httr")
  skip_if_not_installed("curl")

  testInit("terra", opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE))
  withr::local_options(destinationPath = tmpdir)

  globalOutput <- capture.output({
    url <- "http://forestales.ujed.mx/incendios2/cartografia/tematicos/combustibles_y_vegetacion/tipo_combustibles_serie_VI/"

    if (!urlExists(url)) {
      skip("Mexico url doesn't exist; skipping")
    }
    # Nothing specified
    a <- prepInputs(url = url, fun = "terra::rast")
    expect_is(a, "SpatRaster")
    files <- dir(getwd(), pattern = "comb")
    expect_true(length(files) == 8)

    unlink(dir(getwd(), recursive = TRUE, full.names = TRUE))
    a <- prepInputs(url = url, targetFile = "comb_290719.tif", fun = "terra::rast")
    expect_is(a, "SpatRaster")
    files <- dir(getwd(), pattern = "comb")
    expect_true(length(files) == 8)

    unlink(dir(getwd(), recursive = TRUE, full.names = TRUE))
    a <- prepInputs(
      url = url,
      targetFile = "comb_290719.tif",
      alsoExtract = FALSE,
      fun = "terra::rast"
    )
    expect_is(a, "SpatRaster")
    files <- dir(getwd(), pattern = "comb")
    expect_true(length(files) == 1)

    unlink(dir(getwd(), recursive = TRUE, full.names = TRUE))
    a <- prepInputs(url = url, fun = "terra::rast")
    expect_is(a, "SpatRaster")
    files <- dir(getwd(), pattern = "comb_290719")
    expect_true(length(files) == 7)
  })
})

test_that("test prepInputs url when a gdrive directory", {
  skip_on_cran()

  testInit(
    c("terra", "googledrive"),
    opts = list(reproducible.inputPaths = NULL, reproducible.overwrite = TRUE)
  )
  withr::local_options(destinationPath = tmpdir)
  skip_if_no_token()

  globalOutput <- capture.output({
    withr::local_dir(tmpdir)
    dPath <- "."
    url <- "https://drive.google.com/drive/u/3/folders/1q3aosWJ_THpgEaDzchvCWLMwT91pD9Fs"
    a <- prepInputs(
      url = url,
      fun = quote({
        tfp <- sort(targetFilePath)
        b <- terra::rast(tfp)
        names(b) <- basename(tfp)
        b
      }),
      destinationPath = dPath
    ) |>
      Cache()

    expect_is(a, "SpatRaster")
    expect_true(terra::nlyr(a) > 1)
  })
})

test_that("test prepInputs with zip file with hidden files", {
  testInit()
  skip_on_os("mac")
  withr::local_dir(tmpdir)
  a <- 1
  theFile <- "__MACOSX/._theFile.txt"
  checkPath(dirname(theFile), create = TRUE)
  cat(a, file = theFile)
  zipFilename <- "test.zip"
  zip(files = theFile, zipfile = zipFilename, flags = "-q")
  unlink(theFile, recursive = TRUE)
  expect_false(file.exists(file = theFile))
  b <- prepInputs(targetFile = theFile, archive = zipFilename, fun = NA)
  expect_true(file.exists(file = theFile))
})

test_that(".guessAtTargetAndFun ignores OS archive metadata when auto-picking", {
  testInit()
  # When no targetFile is specified, OS-injected metadata files must not be
  # auto-selected: macOS __MACOSX/*, ._ AppleDouble, .DS_Store; Windows Thumbs.db,
  # desktop.ini.
  filesExtracted <- c(
    "data/clean_NMC_20kmBuff.shp",
    "data/__MACOSX/._clean_NMC_20kmBuff.shp",
    "data/._clean_NMC_20kmBuff.shp",
    "data/.DS_Store",
    "data/Thumbs.db",
    "data/desktop.ini"
  )
  out <- .guessAtTargetAndFun(
    targetFilePath = NULL,
    filesExtracted = filesExtracted,
    fun = NULL,
    verbose = 0
  )
  expect_identical(out$targetFilePath, "data/clean_NMC_20kmBuff.shp")

  # If the user explicitly asks for an otherwise-filtered path, honor it.
  out2 <- .guessAtTargetAndFun(
    targetFilePath = "data/__MACOSX/._clean_NMC_20kmBuff.shp",
    filesExtracted = filesExtracted,
    fun = "readLines",
    verbose = 0
  )
  expect_identical(out2$targetFilePath, "data/__MACOSX/._clean_NMC_20kmBuff.shp")
})

test_that("testing prepInputs with deauthorized googledrive", {
  skip_on_cran()
  skip_on_ci()
  skip_if_not_installed("googledrive")

  testInit("terra", needGoogleDriveAuth = TRUE)
  withr::local_dir(tmpdir)

  testthat::with_mocked_bindings(
    isInteractive = function() {
      FALSE
    },
    {
      noisyOutput <- capture.output({
        warn <- capture_warnings({
          BCR6_VT <- skip_on_transient_http(prepInputs(
            alsoExtract = "similar",
            url = "https://drive.google.com/open?id=1sEiXKnAOCi-f1BF7b4kTg-6zFlGr0YOH",
            targetFile = "BCR6.shp",
            overwrite = TRUE
          ))
        })
      })
    }
  )
  expect_true(is(BCR6_VT, vectorType()))

  if (.requireNamespace("sf", stopOnFALSE = FALSE)) {
    theQuietVar <- TRUE # test for finding this
    co <- capture.output(NFDB_PT <-
      skip_on_transient_http(prepInputs(
        url = "http://cwfis.cfs.nrcan.gc.ca/downloads/nfdb/fire_pnt/current_version/NFDB_point.zip",
        overwrite = TRUE,
        fun = sf::st_read(targetFile, quiet = theQuietVar)
      ))
    )
    expect_is(NFDB_PT, "sf")
    expect_true(all(c("zip", "sbx", "shp", "xml", "shx", "sbn") %in%
      fileExt(dir(pattern = "NFDB_point"))))

    noisyOutput <- capture.output({
      warn <- capture_warnings({
        NFDB_PT_BCR6 <- Cache(postProcess, NFDB_PT, studyArea = BCR6_VT)
      })
    })
    if (!all(grepl("attribute variables are assumed to be spatially constant", warn))) {
      warnings(warn)
    }
  }
})

test_that("testing rebuildColors", {
  # ONLY RELEVANT FOR RASTER
  testInit(needGoogleDriveAuth = FALSE, "raster")

  x <- raster::raster(extent(0, 10, 0, 10), vals = runif(100, 0, 197))
  origColors <- list(origColors = character(0), origMinValue = 0, origMaxValue = 197.100006103516)
  expect_is(rebuildColors(x, origColors), "Raster")
})

test_that("prepInputs in a simple one double nested zip file, passing only destinationPath and url", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)

  url <- "https://github.com/tati-micheletti/host/raw/master/data/rasterNested.zip"
  noisyOutput <- capture.output({
    testZip <- reproducible::prepInputs(url = url, destinationPath = tmpdir)
  })
  expect_true(exists("testZip"))
})

test_that("prepInputs in a simple one double nested zip file, passing targetFile", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)

  url <- "https://github.com/tati-micheletti/host/raw/master/data/rasterNested.zip"
  noisyOutput <- capture.output({
    testZip2 <- reproducible::prepInputs(
      url = url,
      targetFile = "rasterTest.tif",
      destinationPath = tmpdir
    )
  })
  expect_true(exists("testZip2"))
})

test_that("prepInputs in a two files double nested zip file, passing only destinationPath and url", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)

  url <- "https://github.com/tati-micheletti/host/raw/master/data/multiFilesOutter.zip"
  noisyOutput <- capture.output({
    testZip3 <- reproducible::prepInputs(url = url, destinationPath = tmpdir)
  })
  expect_true(exists("testZip3"))
})

test_that("prepInputs in a two files double nested zip file, passing targetFile", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)

  url <- "https://github.com/tati-micheletti/host/raw/master/data/multiFilesOutter.zip"
  noisyOutput <- capture.output({
    testZip4 <- reproducible::prepInputs(
      url = url,
      targetFile = "rasterTest.tif",
      destinationPath = tmpdir
    )
  })
  expect_true(exists("testZip4"))
})

test_that(
  paste0(
    "prepInputs in a two files double nested zip file, with the wanted file in",
    "the second layer, and a shapefile in the first, not specifying the targetFile"
  ),
  {
    skip_on_cran()
    testInit("terra", needInternet = TRUE)

    # Gets the first file it finds, the shapefile; warns the user about it
    url <- "https://github.com/tati-micheletti/host/raw/master/data/multiFilesMultiLevels.zip"
    # The warning is about the .prj file missing, which is not relevant here -
    #   Capture it and do nothing with it
    noisyOutput <- capture.output({
      warn <- capture_warnings({
        testZip6 <- reproducible::prepInputs(url = url, destinationPath = tmpdir)
      })
    })
    expect_true(exists("testZip6"))
    expect_is(testZip6, vectorType())
  }
)

test_that(
  paste0(
    "prepInputs in a two files double nested zip file, with the wanted file in",
    "the second layer, and a shapefile in the first, specifying the targetFile"
  ),
  {
    skip_on_cran()
    testInit("terra", needInternet = TRUE)

    url <- "https://github.com/tati-micheletti/host/raw/master/data/multiFilesMultiLevels.zip"
    noisyOutput <- capture.output({
      testZip7 <- reproducible::prepInputs(
        url = url,
        targetFile = "rasterTest.tif",
        destinationPath = tmpdir
      )
    })
    expect_true(exists("testZip7"))
    expect_is(testZip7, rasterType())
  }
)

test_that(
  paste(
    "prepInputs in a two files double nested rar file,",
    "with the wanted file in the second layer, not specifying the targetFile"
  ),
  {
    skip_on_cran()
    skip_on_os("mac") ## TODO: deal with unrar for macOS #266

    testInit("terra", needInternet = TRUE)

    noisyOutput <- capture.output({
      testRar <- reproducible::prepInputs(
        url = "https://github.com/tati-micheletti/host/raw/master/data/nestedRarTxtFiles.rar",
        destinationPath = tmpdir
      )
    })
    expect_true(exists("testRar"))
    expect_is(testRar, rasterType())
    # }
  }
)

test_that(
  paste0(
    "prepInputs in a two files double nested rar file, with the wanted file in",
    " the second layer, specifying the targetFile"
  ),
  {
    skip_on_cran()
    skip_on_os("mac") ## TODO: deal with unrar for macOS #266

    testInit("terra", needInternet = TRUE)
    url <- "https://github.com/tati-micheletti/host/raw/master/data/nestedRarTxtFiles.rar"

    noisyOutput <- capture.output({
      testRar2 <- reproducible::prepInputs(
        url = url,
        targetFile = "rasterTOtestRAR.tif",
        destinationPath = tmpdir
      )
    })
    expect_true(exists("testRar2"))
    expect_is(testRar2, rasterType())
    # }
  }
)

test_that(
  paste0(
    "prepInputs in a two files double nested rar file, with the wanted file in",
    "the second layer, not specifying the targetFile, passing the main archive"
  ),
  {
    skip_on_cran()
    skip_on_os("mac") ## TODO: deal with unrar for macOS #266

    testInit("terra", needInternet = TRUE)

    url <- "https://github.com/tati-micheletti/host/raw/master/data/nestedRarTxtFiles.rar"

    noisyOutput <- capture.output({
      testRar3 <- reproducible::prepInputs(
        url = url,
        archive = "nestedRarTxtFiles.rar",
        targetFile = "rasterTOtestRAR.tif",
        destinationPath = tmpdir
      )
    })
    expect_true(exists("testRar3"))
    expect_is(testRar3, rasterType())
    # }
  }
)

test_that("prepInputs works with nested rar file inside internal rar folder", {
  skip_on_cran()
  skip_on_ci() ## TODO: skip for now b/c need additional unrar tools
  skip_on_os("mac") ## TODO: deal with unrar for macOS #266

  testInit("terra", needInternet = TRUE)

  url <- "https://github.com/tati-micheletti/host/raw/master/data/testRasterNested.rar"
  # extractSystemCallPath <- .archiveExtractBinary()
  # if (is.null(extractSystemCallPath)) {
  #   noisyOutput <- capture.output(
  #     expect_error({
  #       testRar4 <- reproducible::prepInputs(
  #         url = url,
  #         targetFile = "rasterTest.tif",
  #         destinationPath = tmpdir,
  #         useCache = FALSE
  #       )
  #     })
  #   )
  # } else {
  noisyOutput <- capture.output({
    testRar4 <- reproducible::prepInputs(
      url = url,
      targetFile = "rasterTest.tif",
      destinationPath = tmpdir,
      useCache = FALSE
    )
  })
  expect_true(exists("testRar4"))
  expect_is(testRar4, rasterType())
  # }
})

test_that("preProcess works for .tar files", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestTar
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(url = url, destinationPath = tmpdir)
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works for .zip when provided only url and destinationPath", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(url = url, destinationPath = tmpdir)
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works with only url", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(ras <- reproducible::preProcess(url = url))
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works when provides only archive", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    pre <- reproducible::preProcess(url = url, destinationPath = tmpdir)
  )
  testthat::expect_is(object = pre, class = "list")
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(
      archive = file.path(
        pre$destinationPath,
        list.files(pre$destinationPath)[grepl(
          x = list.files(pre$destinationPath),
          pattern = ".zip|.tar"
        )]
      )
    )
  )
  testthat::expect_is(object = ras, class = "list")
})

test_that("preProcess works when provides archive and destinationPath", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    pre <- reproducible::preProcess(url = url, destinationPath = tmpdir)
  )
  testthat::expect_is(object = pre, class = "list")
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(
      archive = file.path(
        pre$destinationPath,
        list.files(pre$destinationPath)[grepl(
          x = list.files(pre$destinationPath),
          pattern = ".zip|.tar"
        )]
      ),
      destinationPath = tmpdir
    )
  )
  testthat::expect_is(object = ras, class = "list")
})

test_that("preProcess works when provides only targetFile", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    pre <- reproducible::preProcess(url = url, destinationPath = tmpdir)
  )
  testthat::expect_is(object = pre, class = "list")
  noisyOutput <- capture.output(ras <- reproducible::preProcess(targetFile = pre$targetFilePath))
  testthat::expect_is(object = ras, class = "list")
})

test_that("preProcess works when provides targetfile and destinationPath", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    pre <- reproducible::preProcess(url = url, destinationPath = tmpdir)
  )
  testthat::expect_is(object = pre, class = "list")
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(targetFile = pre$targetFilePath, destinationPath = tmpdir)
  )
  testthat::expect_is(object = ras, class = "list")
})

test_that("preProcess works when provides url, archive, targetfile and destinationPath", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(
      url = url,
      targetFile = theRasterTestFilename(suff = "tif"),
      archive = theRasterTestFilename(suff = "zip"),
      destinationPath = tmpdir
    )
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works when provides url, archive, and destinationPath", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(
      url = url,
      archive = theRasterTestFilename(suff = "zip"),
      destinationPath = tmpdir
    )
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works when provides url, archive, and destinationPath and reproducible.inputPaths", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  withr::local_options("reproducible.inputPaths" = tmpdir)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(
      url = url,
      archive = theRasterTestFilename(suff = "zip"),
      destinationPath = tmpCache
    )
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works when provides url, targetfile and destinationPath", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(
      url = url,
      targetFile = theRasterTestFilename(suff = "tif"),
      destinationPath = tmpdir
    )
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works when provides url and destinationPath for a .rar file", {
  skip_on_cran()
  skip_on_os("mac") ## TODO: deal with unrar for macOS #266

  testInit("terra", needInternet = TRUE)
  # extractSystemCallPath <- try(.testForArchiveExtract(), silent = TRUE)
  url <- theRasterTestRar

  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(url = url, destinationPath = tmpdir)
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
  # }
  # }
})

test_that("preProcess works when provides url, targetfile and destinationPath for a .rar file", {
  skip_on_cran()
  skip_on_os("mac") ## TODO: deal with unrar for macOS #266

  testInit("terra", needInternet = TRUE)
  ## extractSystemCallPath <- try(.testForArchiveExtract(), silent = TRUE)
  url <- theRasterTestRar

  wd <- getwd()
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(
      url = url,
      targetFile = theRasterTestFilename(suff = "tif"),
      destinationPath = tmpdir
    )
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
  expect_equal(wd, getwd()) # Test that working directory is restored after unrar call
  # }
  # }
})

test_that("preProcess works when provides url, archive and destinationPath for a .rar file", {
  skip_on_cran()
  skip_on_os("mac") ## TODO: deal with unrar for macOS #266

  testInit("terra", needInternet = TRUE)
  # extractSystemCallPath <- try(.testForArchiveExtract(), silent = TRUE)
  url <- theRasterTestRar
  rasterTestRarFilename <- theRasterTestFilename(suff = "rar")

  # if (!is(extractSystemCallPath, "try-error")) {
  #   if (is.null(extractSystemCallPath)) {
  #   noisyOutput <- capture.output(
  #     expect_error({
  #       ras <- reproducible::preProcess(url = url, archive = rasterTestRarFilename, destinationPath = tmpdir)
  #     })
  #   )
  # } else {
  noisyOutput <- capture.output(
    ras <- reproducible::preProcess(
      url = url,
      archive = rasterTestRarFilename,
      destinationPath = tmpdir
    )
  )
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
  #   }
  # }
})

test_that("preProcess works, but gives a warning when supplying cacheTags", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(testthat::expect_message({
    ras <- reproducible::preProcess(
      url = url,
      destinationPath = tmpdir,
      cacheTags = "objectName::ras"
    )
  }))
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works, but gives a warning when supplying postProcessedFilename", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(testthat::expect_message({
    ras <- reproducible::preProcess(
      url = url,
      destinationPath = tmpdir,
      postProcessedFilename = "ras.tif"
    )
  }))
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works, but gives a warning when supplying rasterInterpMethod", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(testthat::expect_message({
    ras <- reproducible::preProcess(
      url = url,
      destinationPath = tmpdir,
      rasterInterpMethod = "near"
    )
  }))
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works, but gives a warning when supplying rasterDatatype", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(testthat::expect_message({
    ras <- reproducible::preProcess(url = url, destinationPath = tmpdir, rasterDatatype = "INT1U")
  }))
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("preProcess works, but gives a warning when supplying pkg", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- theRasterTestZip
  noisyOutput <- capture.output(testthat::expect_message({
    ras <- reproducible::preProcess(
      url = url,
      destinationPath = tmpdir,
      pkg = "utils",
      fun = "unzip"
    )
  }))
  testthat::expect_is(object = ras, class = "list")
  testthat::expect_true(file.exists(ras$targetFilePath))
})

test_that("message when files from archive are already present", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- "https://github.com/tati-micheletti/host/raw/master/data/rasterTest.zip"
  noisyOutput <- capture.output(
    ccc <- testthat::capture_output(
      ras <- reproducible::preProcess(
        url = url,
        targetFile = "rasterTest.tif",
        destinationPath = tmpdir
      )
    )
  )
  noisyOutput <- capture.output(testthat::expect_message({
    ras <- reproducible::preProcess(archive = "rasterTest.zip", destinationPath = tmpdir)
  }))
})

test_that("message when file is a shapefile", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- "https://github.com/tati-micheletti/host/raw/master/data/shapefileTest.zip"
  noisyOutput <- capture.output(
    ccc <- testthat::capture_output(testthat::expect_message({
      ras <- reproducible::preProcess(url = url, destinationPath = tmpdir)
    }))
  )
})

test_that("preProcess succeeds when targetFile extension is unknown", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- "https://github.com/tati-micheletti/host/raw/master/data/unknownTargetFile.zip"
  # preProcess does not load the object, so an unrecognised extension should
  # not error; funChar is left NULL and the file is still extracted.
  noisyOutput <- capture.output(
    ccc <- testthat::capture_output({
      out <- reproducible::preProcess(url = url, destinationPath = tmpdir)
    })
  )
  expect_null(out$funChar)
  expect_true(!is.null(out$targetFilePath))
})

test_that("When supplying two files without archive, when archive and files have different names", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- "https://github.com/tati-micheletti/host/raw/master/data/twoKnownFiles.zip"
  noisyOutput <- capture.output(
    ccc <- testthat::capture_output(testthat::expect_error({
      ras <- reproducible::preProcess(
        url = url,
        targetFile = c("rasterTest.tif", "shapefileTest.shp"),
        destinationPath = tmpdir
      )
    }))
  )
})

test_that("message when archive has two known files (raster and shapefile)", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url <- "https://github.com/tati-micheletti/host/raw/master/data/knownFiles.zip"
  noisyOutput <- capture.output(
    ccc <- testthat::capture_output(testthat::expect_error({
      ras <- reproducible::preProcess(
        url = url,
        archive = "knownFiles.zip",
        targetFile = c("knownFiles.tif", "knownFiles.shp"),
        destinationPath = tmpdir
      )
    }))
  )
})

test_that("message when extracting a file that is already present", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  url1 <- "https://github.com/tati-micheletti/host/raw/master/data/rasterTest.zip"
  noisyOutput <- capture.output({
    ccc <- testthat::capture_output({
      ras <- reproducible::preProcess(
        url = url1,
        targetFile = "rasterTest.tif",
        destinationPath = tmpdir
      )
    })
  })
  url2 <- "https://github.com/tati-micheletti/host/raw/master/data/shapefileTest.zip"
  noisyOutput <- capture.output({
    ccc <- testthat::capture_output({
      shp <- reproducible::preProcess(url = url2, destinationPath = tmpdir)
    })
  })
  url3 <- "https://github.com/tati-micheletti/host/raw/master/data/knownFiles.zip"
  noisyOutput <- capture.output({
    ccc <- testthat::capture_output({
      testthat::expect_message({
        fl <- reproducible::preProcess(url = url3, destinationPath = tmpdir)
      })
    })
  })
})

test_that("masking with larger extent obj", {
  skip_on_cran()
  skip_on_ci()

  testInit("terra", needGoogleDriveAuth = TRUE, needInternet = TRUE)
  smallRT <- prepInputs(url = "https://drive.google.com/open?id=1WhL-DxrByCbzAj8A7eRx3Y1FVujtGmtN")
  sam <- sample(terra::ncell(smallRT), size = terra::ncell(smallRT) / 2)
  smallRT[] <- NA
  smallRT[sam] <- 1L
  if (is(smallRT, "Raster")) {
    a <- raster::extent(smallRT)
  } else {
    a <- terra::ext(smallRT)
  }
  a <- terra::extend(a, -3e5) # make it small
  test <- rasterRead(a, resolution = 250, vals = 2)
  terra::crs(test) <- terra::crs(smallRT)
  b <- postProcess(x = test, rasterToMatch = smallRT, maskWithRTM = TRUE)
  expect_true(is(b, rasterType()))
})

test_that("just google id not url", {
  skip_on_cran()
  skip_on_ci()

  testInit("terra", needGoogleDriveAuth = TRUE, needInternet = TRUE)
  co <- capture.output(smallObj <- prepInputs(url = "1Bk4SPz8rx8zziIlg2Yp9ELZmdNZytLqb"))
  # depends on whether `sf` is installed
  expect_is(smallObj, c("SpatVector", "sf"))
})

test_that("Test of using future and progress indicator for lrg files on Google Drive", {
  skip_on_cran()
  skip_on_ci()
  skip_if_not_installed("future")
  skip_if_not_installed("googledrive")

  testInit(
    c("terra", "future"),
    needGoogleDriveAuth = TRUE,
    needInternet = TRUE,
    opts = list("reproducible.futurePlan" = "multisession")
  )
  noisyOutput <- capture.output({
    ccc <- testthat::capture_output({
      smallRT <- skip_on_transient_http(preProcess(
        url = "https://drive.google.com/open?id=1WhL-DxrByCbzAj8A7eRx3Y1FVujtGmtN"
      ))
    })
  })
  expect_true(is(smallRT, "list"))
})

test_that("lightweight tests for preProcess code coverage", {
  skip_on_cran()
  testInit(c("sf", "terra"))
  expect_true(is.data.frame(preProcessParams()))
  expect_true(is.data.frame(preProcessParams(1)))
  expect_true(is.data.frame(preProcessParams(2)))
  expect_true(is.data.frame(preProcessParams(3)))
  expect_true(is.data.frame(preProcessParams(4)))

  # test purge
  localFileLuxSm <- system.file("ex/luxSmall.shp", package = "reproducible")
  capture.output(la <- prepInputs(targetFile = localFileLuxSm, destinationPath = tmpdir))
  csf <- dir(pattern = "CHECKSUMS", path = tmpdir, full.names = TRUE)
  a <- file.info(csf)
  Sys.sleep(0.1)
  capture.output(la2 <- prepInputs(targetFile = localFileLuxSm, destinationPath = tmpdir))
  b <- file.info(csf)
  expect_true(milliseconds(b$mtime) == milliseconds(a$mtime))
  if (isWindows() && isInteractive()) {
    # apparently atime is not write on *nix-alikes
    expect_false(milliseconds(b$atime) == milliseconds(a$atime))
  }

  # purge will delete CHECKSUMS 7 -- written, read
  capture.output(
    la3 <- prepInputs(targetFile = localFileLuxSm, destinationPath = tmpdir, purge = 7)
  )
  d <- file.info(csf)
  if (isWindows()) {
    # linux doesn't do ctime
    expect_true(milliseconds(d$ctime) == milliseconds(a$ctime))
    expect_false(milliseconds(d$atime) == milliseconds(a$atime))
  }
  expect_false(milliseconds(d$mtime) == milliseconds(a$mtime))

  # purge will delete CHECKSUMS 1 -- deleted, written, read
  Sys.sleep(0.1)
  capture.output(
    la4 <- prepInputs(targetFile = localFileLuxSm, destinationPath = tmpdir, purge = 1)
  )
  e <- file.info(csf)
  # if (isWindows()) # windows doesn't release a file's ctime even when removed
  #   expect_false(milliseconds(e$ctime) == milliseconds(a$ctime))
  expect_false(milliseconds(e$mtime) == milliseconds(a$mtime))
  expect_false(milliseconds(e$atime) == milliseconds(a$atime))

  expect_null(.decodeMagicNumber("sdfddsffdfs.tetes"))
  expect_true(is.character(.decodeMagicNumber("Shapefile")))
  expect_true(is.character(.decodeMagicNumber("RAR")))
  expect_true(is.character(.decodeMagicNumber("tar")))
  expect_true(is.character(.decodeMagicNumber("TIFF")))
  expect_true(is.character(.decodeMagicNumber("Zip")))
})

test_that("large test for nested file structures in zips", {
  skip_on_cran()
  skip_on_ci()
  testInit(c("sf", "googledrive", "terra"), needInternet = TRUE, needGoogleDriveAuth = TRUE)
  climateDataURL <- "https://drive.google.com/file/d/1we9GqEVAORWLbHi3it66VnCcvLu85QIk"

  ## extracts flat files, overwriting and keeping only the last subdir's files
  files <- list(
    paste0("Alberta/Year_", 2015:2019, "M/Eref01.asc"),
    paste0("Alberta/Year_", 1991:2019, "M/Eref01.asc"),
    c(
      paste0("Alberta/Year_", 1991:2019, "M/CMD01.asc"),
      paste0("Alberta/Year_", 1991:2019, "M/Eref01.asc")
    )
  )
  lapply(files, function(fis) {
    res1 <- preProcess(
      url = climateDataURL,
      destinationPath = tmpdir,
      targetFile = "Alberta/Year_2020M/CMD01.asc",
      alsoExtract = fis
    )
    testLength <- length(fis) + 2
    expect_equal(NROW(res1$checkSums[checksum.x != "dir"]), testLength)
  })
})

test_that("more nested file structures in zip; alsoExtract NA", {
  skip_on_cran()
  testInit("terra")
  tmpdir <- withr::local_tempdir()
  withr::local_dir(tmpdir)
  zipName <- "ex.zip"
  system.time({
    ras <- lapply(1:2, function(x) {
      td <- tempdir2()
      terra::rast(terra::ext(0, 4, 0, 4), vals = sample(1:16), resolution = 1) |>
        terra::writeRaster(filename = file.path(td, basename(tempfile(fileext = ".tif"))))
    })
    setwd(dirname(dirname(Filenames(ras[[1]]))))
    fns1 <- Filenames(ras)
    # zip(zipName, files = file.path(basename(dirname(fns)), basename(fns)))
    ras <- lapply(1:2, function(x) {
      terra::rast(terra::ext(0, 4, 0, 4), vals = sample(1:16), resolution = 1) |>
        terra::writeRaster(filename = file.path(basename(tempfile(fileext = ".tif"))))
    })
    fns2 <- Filenames(ras)
    zip(
      zipName,
      files = c(file.path(basename(dirname(fns1)), basename(fns1)), basename(fns2)),
      flags = "-q"
    )
  })
  zipName2 <- file.path(tmpdir, zipName)

  # don't specify targetFile --> fail because many options
  linkOrCopy(zipName, zipName2)
  a <- prepInputs(archive = zipName, destinationPath = tmpdir)
  knownOtherFiles <- c("cache.db", "CHECKSUMS.txt", "ex.zip", "testCache")
  files <- dir(tmpdir, recursive = TRUE) # %in% dirname(.listFilesInArchive(zipName2))
  expect_true(all(.listFilesInArchive(zipName2) %in% files))
  unlink(
    file.path(tmpdir, grep("\\.", invert = TRUE, dirname(files), value = TRUE)),
    recursive = TRUE
  )
  files2 <- dir(pattern = ".tif", tmpdir, full.names = TRUE)
  unlink(files2)

  # alsoExtract = NULL --> the default --> extract all
  checkPath(dirname(zipName2), create = TRUE)
  linkOrCopy(zipName, zipName2)
  a <- prepInputs(
    archive = zipName,
    targetFile = grep("\\.tif", basename(files), value = TRUE)[1],
    destinationPath = tmpdir
  )
  files <- dir(tmpdir, recursive = TRUE) # %in% dirname(.listFilesInArchive(zipName2))
  expect_true(all(.listFilesInArchive(zipName2) %in% files))
  unlink(
    file.path(tmpdir, grep("\\.", invert = TRUE, dirname(files), value = TRUE)),
    recursive = TRUE
  )
  files2 <- dir(pattern = ".tif", tmpdir, full.names = TRUE)
  unlink(files2)

  # alsoExtract = NA --> extract only targetFile -- with wrong filename (missing subdir)
  checkPath(dirname(zipName2), create = TRUE)
  linkOrCopy(zipName, zipName2)
  filesPre <- dir(tmpdir, recursive = TRUE) # %in% dirname(.listFilesInArchive(zipName2))
  a <- prepInputs(
    archive = zipName,
    alsoExtract = NA,
    targetFile = grep("\\.tif", basename(files), value = TRUE)[1],
    destinationPath = tmpdir
  )
  files <- dir(tmpdir, recursive = TRUE) # %in% dirname(.listFilesInArchive(zipName2))
  expect_false(all(.listFilesInArchive(zipName2) %in% files))
  expect_true(sum(.listFilesInArchive(zipName2) %in% files) == 1)
  unlink(
    file.path(tmpdir, grep("\\.", invert = TRUE, dirname(files), value = TRUE)),
    recursive = TRUE
  )
  files2 <- dir(pattern = ".tif", tmpdir, full.names = TRUE)
  unlink(files2)

  # alsoExtract = a filename --> extract targetFile & 1 other -- with correct filename (with subdir)
  checkPath(dirname(zipName2), create = TRUE)
  linkOrCopy(zipName, zipName2)
  possFiles <- .listFilesInArchive(zipName2)
  filesWithSubDir <- grep("\\/", possFiles, value = TRUE, invert = FALSE)
  filesPre <- dir(tmpdir, recursive = TRUE) # %in% dirname(.listFilesInArchive(zipName2))
  a <- prepInputs(
    archive = zipName,
    alsoExtract = filesWithSubDir[2],
    targetFile = filesWithSubDir[1],
    destinationPath = tmpdir
  )
  files <- dir(tmpdir, recursive = TRUE) # no subdir on this file, so non-recursive should do it
  expect_false(all(.listFilesInArchive(zipName2) %in% files))
  expect_true(sum(.listFilesInArchive(zipName2) %in% files) == 2) #
  unlink(
    file.path(tmpdir, grep("\\.", invert = TRUE, dirname(files), value = TRUE)),
    recursive = TRUE
  )
  files2 <- dir(pattern = ".tif", tmpdir, full.names = TRUE)
  unlink(files2)

  # alsoExtract = NA --> extract only targetFile -- with correct filename (with subdir)
  checkPath(dirname(zipName2), create = TRUE)
  linkOrCopy(zipName, zipName2)
  possFiles <- .listFilesInArchive(zipName2)
  filesNoSubDir <- grep("\\/", possFiles, value = TRUE, invert = TRUE)
  filesPre <- dir(tmpdir, recursive = TRUE) # %in% dirname(.listFilesInArchive(zipName2))
  a <- prepInputs(
    archive = zipName,
    alsoExtract = NA,
    targetFile = filesNoSubDir[1],
    destinationPath = tmpdir
  )
  files <- dir(tmpdir, recursive = FALSE) # no subdir on this file, so non-recursive should do it
  expect_false(all(.listFilesInArchive(zipName2) %in% files))
  expect_true(sum(.listFilesInArchive(zipName2) %in% files) == 1)
})

test_that("PR#358 if dwnld already exists, was missing nested paths", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE, needGoogleDriveAuth = TRUE)
  url <- "https://drive.google.com/file/d/1S-4itShMXtwzGxjKPgsznpdTD2ydE9qn/"
  noisyOutput <- capture.output(
    ras <- preProcess(
      targetFile = "all_gp_site_info.csv",
      url = url,
      destinationPath = getOption("reproducible.destinationPath", file.path(tmpdir, "dPath")),
      overwrite = TRUE,
      fun = "data.table::fread",
      useCache = FALSE
    )
  )
  testthat::expect_true(file.exists(ras$targetFilePath))
  noisyOutput <- capture.output(
    ras <- preProcess(
      targetFile = "all_gp_site_info.csv",
      url = url,
      destinationPath = getOption("reproducible.destinationPath", file.path(tmpdir, "dPath")),
      overwrite = TRUE,
      fun = "data.table::fread",
      useCache = FALSE
    )
  )
  testthat::expect_true(file.exists(ras$targetFilePath))
})

all.equalWONewCache <- function(a, b) {
  attr(a, ".Cache")$newCache <- NULL
  attr(b, ".Cache")$newCache <- NULL
  all.equal(a, b)
}

skip_if_no_token <- function() {
  testthat::skip_if_not(googledrive::drive_has_token(), "No Drive token")
}

## NOTE: needs to be called after testInit("googledrive", needGoogleDriveAuth = TRUE)
skip_if_service_account <- function() {
  ## service accounts cannot upload to standard drive folders (no quota)
  testthat::skip_if_not(!grepl("gserviceaccount", googledrive::drive_user()$emailAddress),
                        message =  "Using service account")
}

## Run `expr` and convert transient upstream HTTP/network errors into a skip
## so flaky CDN behaviour (e.g. GitHub raw returning 5xx during a request)
## does not surface as a test FAIL. Non-transient errors propagate unchanged.
## Evaluated in the caller's frame so any assignments inside `expr` are
## visible to the rest of the test.
.transientNetworkPattern <- paste(
  "HTTP 5\\d\\d", "Bad Gateway", "Service Unavailable", "Gateway Timeout",
  "Could(?:n't| not) resolve host", "Empty reply from server",
  "Timeout was reached", "Operation timed out",
  "Connection reset by peer", "Connection refused",
  "TLS connect error", "SSL connect error",
  "Recv failure", "Resolving timed out",
  sep = "|"
)

skip_on_transient_http <- function(expr) {
  expr <- substitute(expr)
  pf <- parent.frame()
  result <- tryCatch(eval(expr, envir = pf), error = identity)
  if (!inherits(result, "error")) return(invisible(result))
  msg <- paste(conditionMessage(result), collapse = "\n")
  if (grepl(.transientNetworkPattern, msg, perl = TRUE)) {
    testthat::skip(paste0("transient upstream HTTP/network error: ",
                          substring(msg, 1L, 240L)))
  }
  stop(result)
}

## Run `expr` and skip if any GDAL streaming-failure warning is emitted
## (truncated TIFF tile reads, IReadBlock failures, etc., or HTTP 5xx mid-
## stream). Returns the result of `expr` so the caller can chain assertions
## on it. Non-transient warnings are re-emitted so they remain visible in
## test output instead of being silently muffled.
.transientStreamPattern <- paste(
  "TIFFFillTile", "TIFFReadEncodedTile", "IReadBlock failed",
  "GDAL error", "HTTP 5\\d\\d", "Bad Gateway",
  "Service Unavailable", "Gateway Timeout",
  sep = "|"
)

## Warnings we know are harmless installation/config noise from GDAL/PROJ —
## emitted by terra/sf on systems where the PROJ database isn't on the
## default search path, but don't affect the result of operations that
## don't actually need CRS transforms. Muffled silently (no test skip,
## no re-emit) so they stop polluting CI output.
.benignGDALPattern <- paste(
  "PROJ: file is not a database",
  sep = "|"
)

skip_if_transient_stream_warnings <- function(expr) {
  expr <- substitute(expr)
  pf <- parent.frame()
  warns <- character()
  result <- withCallingHandlers(
    eval(expr, envir = pf),
    warning = function(w) {
      warns <<- c(warns, conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )
  if (any(grepl(.transientStreamPattern, warns, perl = TRUE))) {
    testthat::skip(paste0("transient upstream streaming failure: ",
                          substring(paste(warns, collapse = "; "), 1L, 240L)))
  }
  warns <- warns[!grepl(.benignGDALPattern, warns, perl = TRUE)]
  for (w in warns) warning(w, call. = FALSE)
  invisible(result)
}

skip_if_service_account_releaseVer_NotLinux <- function() {
  ## Service accounts (e.g. eliot-githubauthentication@...gserviceaccount.com)
  ## have no Drive quota on user-owned folders, so they cannot complete the
  ## upload-and-roundtrip path these tests exercise. We gate those tests on:
  ##   (a) we're actually running unattended (CI or R CMD check non-interactive)
  ##       AND the token currently in use IS a service account, AND
  ##   (b) the runner is not the supported Linux/release combination.
  ##
  ## R_VERSION_LABEL is a CI-only signal the team's runners set to declare
  ## "I am a release R job". A local R CMD check won't have it set, and
  ## absence MUST NOT trigger a skip — only an explicit non-"release" value
  ## should. Without this guard, every local `R CMD check` against a stored
  ## service-account token skipped these tests for no good reason.

  on_ci    <- isTRUE(as.logical(Sys.getenv("CI", "false")))
  isAuto   <- !interactive() || on_ci
  isSA     <- isAuto && grepl(
                "gserviceaccount",
                googledrive::drive_user()$emailAddress)
  notLinux <- Sys.info()[["sysname"]] != "Linux"
  verLabel <- Sys.getenv("R_VERSION_LABEL")
  notRelease <- nzchar(verLabel) && verLabel != "release"

  skip <- isSA && (notLinux || notRelease)

  if (requireNamespace("covr", quietly = TRUE) && covr::in_covr()) {
    skip <- FALSE
  }
  testthat::skip_if(
    skip,
    "Service-account token + non-Linux or non-release R: skipping Drive upload tests"
  )
}

## puts tmpdir, tmpCache, tmpfile (can be vectorized with length >1 tmpFileExt),
##   optsAsk in this environment,
## loads and libraries indicated plus testthat,
## sets options("reproducible.ask" = FALSE) if ask = FALSE
## if `needInternet = TRUE`, it will only re-try every 30 seconds
testInit <- function(libraries = character(), ask = FALSE, verbose, tmpFileExt = "",
                     opts = NULL, needGoogleDriveAuth = FALSE, needInternet = FALSE,
                     envir = parent.frame(1)) {
  set.randomseed()

  pf <- parent.frame()
  if (isTRUE(needGoogleDriveAuth)) {
    libraries <- c(libraries, "googledrive")
    needInternet <- TRUE
  }

  if (isTRUE(needInternet)) {
    if (!is.null(.pkgEnv$.internetExists)) {
      if (difftime(Sys.time(), .pkgEnv$.internetExistsLastCheck) > 30) {
        .pkgEnv$.internetExists <- NULL
        .pkgEnv$.internetExistsLastCheck <- NULL
      }
    }
    if (is.null(.pkgEnv$.internetExists)) {
      for (i in 1:2) {
        .pkgEnv$.internetExists <- internetExists()
        if (isTRUE(.pkgEnv$.internetExists))
          break
        SSL_REVOKE_BEST_EFFORT(envir)
      }
      .pkgEnv$.internetExistsLastCheck <- Sys.time()
    }
    intExists <- .pkgEnv$.internetExists
    if (!intExists) skip("Need internet")
  }

  if (!requireNamespace("withr"))
    skip("Need withr")

  if (length(libraries)) {
    libraries <- unique(libraries)
    if (identical(getOption("reproducible.rasterRead"), "raster::raster")) {
      libraries <- unique(c(libraries, "raster"))
    }
    loadedAlready <- vapply(libraries, function(pkg) {
      any(grepl(paste0("package:", pkg), search()))
    }, FUN.VALUE = logical(1))
    libraries <- libraries[!loadedAlready]

    if (length(libraries)) {
      pkgsLoaded <- unlist(lapply(libraries, requireNamespace, quietly = TRUE))
      if (!all(pkgsLoaded)) {
        lapply(libraries[!pkgsLoaded], skip_if_not_installed)
      }
      suppressWarnings(lapply(libraries, withr::local_package, .local_envir = pf))
    }
  }

  skip_gauth <- identical(Sys.getenv("SKIP_GAUTH"), "true") # only set in setup.R for covr
  if (isTRUE(needGoogleDriveAuth)) {
    if (isTRUE(skip_gauth))
      skip("SKIP_GAUTH=true; skipping Google Drive tests")
    if (isNamespaceLoaded("googledrive"))
      if ((!googledrive::drive_has_token())) {
        if (!nzchar(Sys.getenv("GOOGLEDRIVE_AUTH"))) {
          Sys.setenv("GOOGLEDRIVE_AUTH" = "~/genial-cycling-408722-788552a3ecac.json")
        }
        gauthEnv <- Sys.getenv("GOOGLEDRIVE_AUTH")
        if (nzchar(gauthEnv)) {
          if (file.exists(gauthEnv)) {
            ## Service-account credentials can be revoked at Google's end
            ## (key rotated, SA disabled, etc.). drive_auth() converts the
            ## underlying HTTP 400 / invalid_grant into a generic "Can't get
            ## Google credentials" abort, which turns every Drive-using test
            ## into an ERROR instead of a SKIP. Swallow it so the
            ## skip_if_no_token() below cleanly skips instead.
            tryCatch(googledrive::drive_auth(path = gauthEnv),
                     error = function(e) invisible(NULL))
          }
        }
      }

    skip_if_no_token()
  }

  out <- list()

  withr::local_options("reproducible.ask" = ask, .local_envir = pf)
  ## Never block test runs waiting on `Type y if you have attempted a manual
  ## download...` — downloadFile() prompts when isInteractive() && this option
  ## is TRUE (its default). Override here for all tests that go through
  ## testInit(); individual tests can still re-enable via `opts`.
  withr::local_options(
    "reproducible.interactiveOnDownloadFail" = FALSE,
    .local_envir = pf
  )
  if (!missing(verbose)) {
    withr::local_options("reproducible.verbose" = verbose, .local_envir = pf)
  }
  if (!is.null(opts)) {
    withr::local_options(opts, .local_envir = pf)
  }
  tmpdir <- normPath(withr::local_tempdir(tmpdir = tempdir2(), .local_envir = pf))
  tmpCache <- normPath(withr::local_tempdir(tmpdir = tmpdir, .local_envir = pf))

  # can't figure out how to build and delete a temporary working directory with withr
  #  The test is `test-cluster.R` that fails if using
  # withr::local_dir(withr::local_tempdir(.local_envir = pf), .local_envir = pf)

  # BUT CRAN does not let you change the setwd during testing ... so, neither works
  # wd <- tempfile2() |> checkPath(create = TRUE)
  # od <- getwd()
  # withr::defer({setwd(od); unlink(wd, recursive = TRUE)}, envir = pf)
  # withr::local_dir(wd, .local_envir = pf)
  ###

  if (isTRUE(any(nzchar(tmpFileExt)))) {
    dotStart <- startsWith(tmpFileExt, ".")
    if (any(!dotStart)) {
      tmpFileExt[!dotStart] <- paste0(".", tmpFileExt)
    }
    out$tmpfile <- normPath(withr::local_tempfile(fileext = tmpFileExt))
  }
  withr::local_dir(tmpdir, .local_envir = pf)
  withr::defer({
    try(reproducible::clearCache(cachePath = tmpCache, ask = FALSE, verbose = -1))
    try(reproducible::clearCache(ask = FALSE, verbose = -1), silent = TRUE)
    try(unlink(tmpCache, recursive = TRUE))
  }, envir = pf)
  out <- append(out, list(tmpdir = tmpdir, tmpCache = tmpCache))
  list2env(out, envir = pf)
  # withr::defer({
  #   browser()
  # })

  return(out)
}


runTest <- function(prod, class, numFiles, mess, expectedMess, filePattern, tmpdir, test) {
  files <- dir(tmpdir, pattern = filePattern, full.names = TRUE)
  if (length(files) != numFiles) {
    ## Use message() so the dump survives capture.output({...}) wrappers in the
    ## test body (capture.output redirects stdout, not stderr).
    message(sprintf(
      "[runTest] pattern=%s expected=%d got=%d tmpdir=%s",
      filePattern, numFiles, length(files), tmpdir
    ))
    message("  matching:    ", paste(basename(files), collapse = ", "))
    others <- setdiff(dir(tmpdir), basename(files))
    if (length(others))
      message("  also in dir: ", paste(others, collapse = ", "))
  }
  expect_true(length(files) == numFiles)
  expect_true(inherits(test, class))
  # messagePrepInputs(mess)
  hasMessageNum <- paste(collapse = "_", which(unlist(
    lapply(strsplit(expectedMess, "\\|")[[1]], function(m) {
      any(grepl(m, mess))
    })
  )))

  isOK <- hasMessageNum == prod
  if (!isOK) {
    if (interactive()) {
      expe <- as.numeric(strsplit(prod, split = "_")[[1]])
      getting <- as.numeric(strsplit(hasMessageNum, split = "_")[[1]])

      expectedMessVec <- strsplit(expectedMess, split = "\\|")[[1]]
      expecting <- paste(collapse = ", ", expectedMessVec[setdiff(expe, getting)])
      if (length(expecting)) {
        cat("\nexpecting, but didn't get: ", expecting)
      }
      got <- paste(collapse = ", ", expectedMessVec[setdiff(getting, expe)])
      if (length(got)) {
        cat("\ngot, but didn't expect ", got, "\n")
      }
    }
  }
  expect_true(isOK) #
}

expectedMessageRaw <- c(
  "Running `preP", "Preparing:", "File downloaded",
  "From:.*Shapefile", "Checking local", "Finished checking",
  "Downloading", "Skipping download", "Skipping extractFrom",
  "targetFile was not.*ry",
  "Writing checksums.*you can specify targetFile",
  "No targetFile supplied. Checksumming", "Appending checksums", "although coordinates are longitude"
)
expectedMessage <- paste0(collapse = "|", expectedMessageRaw)

expectedMessagePostProcessRaw <- c(
  "cropping", "Checking for errors", "Found no errors",
  "intersecting", "masking", "although coordinates are longitude"
)
expectedMessagePostProcess <- paste0(collapse = "|", expectedMessagePostProcessRaw)

urlTif1 <- "https://raw.githubusercontent.com/PredictiveEcology/quickPlot/master/inst/maps/DEM.tif"
urlShapefiles1Zip <- "https://drive.google.com/file/d/1Bk4SPz8rx8zziIlg2Yp9ELZmdNZytLqb/view?usp=sharing"
urlShapefilesZip <- "https://drive.google.com/file/d/1z1x0oI5jUDJQosOXacI8xbzbR15HFi0W/view?usp=sharing"

targetFileLuxRDS <- "gadm36_LUX_0_sp.rds"

testRasterInCloud <- function(fileext, cloudFolderID, numRasterFiles, tmpdir,
                              type = c("Raster", "Stack", "Brick")) {
  .requireNamespace("googledrive", stopOnFALSE = TRUE, messageStart = "to use google drive files")

  # Second test .grd which has two files
  ####################################################
  # neither cloud or local exist -- should create local and upload to cloud
  ####################################################
  fn <- function(raster) {
    return(raster)
  }

  tempFile <- replicate(14, tempfile(tmpdir = tmpdir, fileext = fileext))

  mc <- match.call()
  r1Orig <- terra::rast(terra::ext(0, 200, 0, 200), vals = 1, resolution = 1)
  r1Orig <- terra::writeRaster(r1Orig, filename = tempFile[1], overwrite = TRUE)

  if (mc$type == "Stack") {
    r1Orig2 <- terra::writeRaster(r1Orig, filename = tempFile[2], overwrite = TRUE)
    r1Orig <- c(r1Orig, r1Orig2)
  } else if (mc$type == "Brick") {
    message("Brick is deprecated; not tested any more")
  }

  r1End <- Cache(fn, r1Orig, useCloud = TRUE, cloudFolderID = cloudFolderID)

  cloudFolderID1 <- cloudFolderID
  on.exit({
    clearCache(useCloud = TRUE, cloudFolderID = cloudFolderID1)
  })

  r1EndData <- r1End[]
  r1EndFilename <- Filenames(r1End)
  r1EndCacheAttr <- attr(r1End, ".Cache")$newCache

  # Clear local copy
  rm(r1End)
  clearCache()
  ####################################################
  # cloud copy exists only -- should download to local copy
  ####################################################
  r2Orig <- terra::rast(terra::ext(0, 200, 0, 200), vals = 1, resolution = 1)
  r2Orig <- terra::writeRaster(r2Orig, filename = tempFile[3], overwrite = TRUE)
  if (mc$type == "Stack") {
    r2Orig2 <- terra::writeRaster(r2Orig, filename = tempFile[4], overwrite = TRUE)
    r2Orig <- c(r2Orig, r2Orig2)
  } else if (mc$type == "Brick") {
    r1Orig2 <- r1Orig
    r1Orig <- c(r1Orig, r1Orig2)
    r1Orig <- terra::writeRaster(r1Orig, filename = tempFile[4], overwrite = TRUE)
  }

  # TODO for SpatRaster -- this returns the Path not SpatRaster
  r2End <- Cache(fn, r2Orig, useCloud = TRUE, cloudFolderID = cloudFolderID)
  cloudFolderID2 <- cloudFolderID
  on.exit({
    clearCache(useCloud = TRUE, cloudFolderID = cloudFolderID2)
  })

  expect_true(identical(unname(r1EndData), unname(r2End[])))
  expect_true(all.equal(r1EndFilename, as.character(Filenames(r2End)))) # this now has correct: only 1 downloaded copy exists
  expect_false(identical(Filenames(r2Orig), Filenames(r1Orig)))
  expect_true(r1EndCacheAttr == TRUE)
  expect_true(attr(r2End, ".Cache")$newCache == FALSE)
  filnames2End <- unique(
    dir(dirname(Filenames(r2End)),
        pattern = paste(collapse = "|", basename(filePathSansExt(Filenames(r2End))))
    )
  )
  filnames1End <- unique(
    dir(dirname(r1EndFilename),
        pattern = paste(collapse = "|", basename(filePathSansExt(r1EndFilename)))
    )
  )
  expect_true(NROW(filnames1End) == numRasterFiles) # both sets because of the _1 -- a bit of an artifact due to same folder
  expect_true(NROW(filnames2End) == numRasterFiles) # both sets because of the _1


  ####################################################
  # only local exists -- upload to cloud
  ####################################################
  clearCache(useCloud = TRUE, cloudFolderID = cloudFolderID)
  r1Orig <- terra::rast(terra::ext(0, 200, 0, 200), vals = 5, resolution = 1)
  r1Orig <- terra::writeRaster(r1Orig, filename = tempFile[5], overwrite = TRUE)
  if (mc$type == "Stack") {
    r1Orig2 <- terra::writeRaster(r1Orig, filename = tempFile[12], overwrite = TRUE)
    r1Orig <- c(r1Orig, r1Orig2)
  } else if (mc$type == "Brick") {
    r1Orig2 <- r1Orig
    r1Orig <- c(r1Orig, r1Orig2)
    r1Orig <- terra::writeRaster(r1Orig, filename = tempFile[12], overwrite = TRUE)
  }
  r1End <- Cache(fn, r1Orig, useCloud = FALSE, cloudFolderID = cloudFolderID)

  expect_true(attr(r1End, ".Cache")$newCache == TRUE) # new to local cache

  r4End <- Cache(fn, r1Orig, useCloud = TRUE, cloudFolderID = cloudFolderID)
  cloudFolderID3 <- cloudFolderID
  on.exit({
    clearCache(useCloud = TRUE, cloudFolderID = cloudFolderID3)
  })

  expect_true(attr(r4End, ".Cache")$newCache == FALSE) # new to local cache
  driveLs <- googledrive::drive_ls(cloudFolderID)
  data.table::setDT(driveLs)
  # expect_true(all(basename(Filenames(r4End)) %in% driveLs$name))
  # should have 2 files in cloud b/c of grd and gri
  # expect_true(sum(filePathSansExt(driveLs$name) %in% filePathSansExt(basename(Filenames(r4End)))) == numRasterFiles)
  # should have 1 file that matches in local and in cloud, based on cacheId
  suppressMessages(expect_true(NROW(unique(showCache(userTags = filePathSansExt(driveLs[endsWith(name, "rda")]$name)),
                                           by = .cacheTableHashColName()
  )) == 1))

  ####################################################
  # both cloud and local exist -- take local only -- no change to cloud
  ####################################################
  clearCache(useCloud = TRUE, cloudFolderID = cloudFolderID)
  r1Orig <- terra::rast(terra::ext(0, 200, 0, 200), vals = 5, resolution = 1)
  r1Orig <- terra::writeRaster(r1Orig, filename = tempFile[6], overwrite = TRUE)
  if (mc$type == "Stack") {
    r1Orig2 <- terra::writeRaster(r1Orig, filename = tempFile[13], overwrite = TRUE)
    r1Orig <- c(r1Orig, r1Orig2)
  } else if (mc$type == "Brick") {
    r1Orig2 <- r1Orig
    r1Orig <- c(r1Orig, r1Orig2)
    r1Orig <- terra::writeRaster(r1Orig, filename = tempFile[13], overwrite = TRUE)
  }
  r1End <- Cache(fn, r1Orig, useCloud = TRUE, cloudFolderID = cloudFolderID)
  on.exit({
    clearCache(useCloud = TRUE, cloudFolderID = cloudFolderID)
  })

  expect_true(attr(r1End, ".Cache")$newCache == TRUE) # new to local cache


  driveLsBefore <- googledrive::drive_ls(cloudFolderID)
  r5Orig <- terra::rast(terra::ext(0, 200, 0, 200), vals = 5, resolution = 1)
  r5Orig <- terra::writeRaster(r5Orig, filename = tempFile[9], overwrite = TRUE)
  if (mc$type == "Stack") {
    r5Orig2 <- terra::writeRaster(r5Orig, filename = tempFile[14], overwrite = TRUE)
    r5Orig <- c(r5Orig, r5Orig2)
  } else if (mc$type == "Brick") {
    r5Orig2 <- r5Orig
    r5Orig <- c(r5Orig, r5Orig2)
    r5Orig <- terra::writeRaster(r5Orig, filename = tempFile[14], overwrite = TRUE)
  }
  r5End <- Cache(fn, r5Orig, useCloud = TRUE, cloudFolderID = cloudFolderID)
  on.exit({
    clearCache(useCloud = TRUE, cloudFolderID = cloudFolderID)
  })
  expect_true(attr(r5End, ".Cache")$newCache == FALSE) # new to local cache
  driveLsAfter <- googledrive::drive_ls(cloudFolderID)
  expect_true(all.equal(driveLsAfter[, 1:2], driveLsBefore[, 1:2])) # There are differences deep in the drive_resources
  clearCache(useCloud = TRUE, cloudFolderID = cloudFolderID)
  driveLsEnd <- googledrive::drive_ls(cloudFolderID)
  expect_true(NROW(driveLsEnd) == 0)
}

fnCacheHelper1 <- function() {
  1
}

fnCacheHelper <- function(a, cacheRepo2) {
  Cache(fnCacheHelper1, cachePath = cacheRepo2, verbose = 2)
}

crsToUse <- "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84"

.writeRaster <- function(...) {
  .requireNamespace("terra", stopOnFALSE = TRUE)
  suppressWarningsSpecific(
    falseWarnings = "NOT UPDATED FOR PROJ",
    terra::writeRaster(...)
  )
}

theRasterTests <- "https://github.com/tati-micheletti/host/raw/master/data/"
theRasterTestFilename <- function(pre = "", suff = "") {
  paste0(pre, "rasterTest.", suff)
}
theRasterTestZip <- theRasterTestFilename(theRasterTests, "zip") # "https://github.com/tati-micheletti/host/raw/master/data/rasterTest.zip"
theRasterTestRar <- theRasterTestFilename(theRasterTests, "rar") # "https://github.com/tati-micheletti/host/raw/master/data/rasterTest.rar"
theRasterTestTar <- theRasterTestFilename(theRasterTests, "tar") # "https://github.com/tati-micheletti/host/raw/master/data/rasterTest.tar"


runTestsWithTimings <- function(nameOfOuterList = "ff", envir = parent.frame(), authorizeGoogle = FALSE) {
  if (isTRUE(authorizeGoogle)) {
    testInit(needGoogleDriveAuth = TRUE)
  }
  prepend <- "/home/emcintir/GitHub/reproducible/tests/testthat"
  testFiles <- dir(prepend, pattern = "^test-", full.names = TRUE)
  testFiles <- grep("large", testFiles, value = TRUE, invert = TRUE)
  rrrr <- get(nameOfOuterList, envir = envir)
  testFiles <- setdiff(testFiles, file.path(prepend, names(rrrr)))
  for (tf in testFiles) {
    messageDF(colour = "blue", basename(tf))
    a <- parse(tf, keep.source = TRUE)
    labels <- unlist(lapply(a, function(x) x[[2]]))
    # Sys.setenv("NOT_CRAN" = "false") # doesn't work
    dd <- Map(testLabel = labels, parsed = a, function(parsed, testLabel) {
      message(testLabel)
      skipOnCran <- any(grepl("skip_on_cran", parsed[[3]]))
      start <- Sys.time()
      try(eval(parsed))
      end <- Sys.time()
      b <- difftime(end, start)
      print(format(b))
      data.table(elapsed = as.numeric(b), skipOnCRAN = skipOnCran)
    })
    ee <- data.table::rbindlist(dd, idcol = "Label")
    ee <- setNames(list(ee), basename(tf))
    rrrr <- append(rrrr, ee)
    assign(nameOfOuterList, rrrr, envir = envir)

    testFiles <- testFiles[-1]
  }

  gg <- data.table::rbindlist(get(nameOfOuterList, envir = envir),
                              idcol = "TestFile"
  )
  gg[, TestFile := basename(TestFile)]
  gg
}

expect_match_noSlashN <- function(object, regexp, ...) {
  object <- gsub("  ", " ", gsub("\\n", "", messageStripColor(object)))
  expect_match(object, regexp, ...)

}

googleSetupForUseCloud <- function(cloudFolderID, tmpdir, tmpCache) {
  testsForPkgs <- "testsForPkgs"
  if (isTRUE(tryCatch(googledrive::drive_ls(testsForPkgs), error = function(e) TRUE))) {
    testsForPkgsDir <- retry(quote(googledrive::drive_mkdir(name = testsForPkgs)))
    on.exit2(googledrive::drive_rm(testsForPkgsDir))
  }
  on.exit2({
    try(googledrive::drive_rm(testsForPkgsDir), silent = TRUE)
    try(googledrive::drive_rm(cloudFolderID), silent = TRUE)
    try(googledrive::drive_rm(cloudFolderFromCacheRepo(tmpdir)), silent = TRUE)
    try(googledrive::drive_rm(cloudFolderFromCacheRepo(tmpCache)), silent = TRUE)
  })
}

test_that("symlinks work with cache, input, output paths", {
  skip_on_cran()
  # skip_on_os("windows")
  skip_if_not_installed("withr")

  testInit(c("sf", "terra"), opts = list(
    "rasterTmpDir" = tempdir2(rndstr(1, 6)),
    "reproducible.inputPaths" = NULL,
    "reproducible.overwrite" = TRUE,
    reproducible.showSimilar = TRUE
  ), tmpFileExt = (".tif"), needInternet = FALSE)

  currentDir <- tmpdir


  # This will only test "different drive" if /mnt/scratch or tempdir() are on different drives.
  #  For Windows, this "different drive" must be on the a local volume, not a network drive
  #  file.symlink, in theory, will work on Windows, but user needs to be an Administrator; so this is
  #  not testable here.
  linkedDir <- ifelse(
    dir.exists("/mnt/scratch"),
    file.path("/mnt/scratch", Sys.info()[["user"]], basename(tempfile())),
    tempfile()
  ) |>
    file.path("reproducible_test_symlinks") |>
    checkPath(create = TRUE)

  linkedCacheDir <- file.path(linkedDir, "cache") |>
    checkPath(create = TRUE)
  linkedInputDir <- file.path(linkedDir, "inputs") |>
    checkPath(create = TRUE)
  linkedOutputDir <- file.path(linkedDir, "outputs") |>
    checkPath(create = TRUE)

  localDirs <- list(
    cacheDir = file.path(currentDir, "cache"),
    inputDir = file.path(currentDir, "inputs"),
    outputDir = file.path(currentDir, "outputs")
  )

  # test initial state
  expect_identical(character(), unlist(lapply(localDirs, dir)))

  linkFn <- if (isWindows()) Sys.junction else file.symlink
  expect_true(linkFn(linkedCacheDir, localDirs$cacheDir))

  # On Windows on Github actions, one of these paths has Windows `shortPathName`
  #   for c:/Users/RUNNER~1, but second has long name; need normalizePath in addition to fs::
  expect_identical(normalizePath(mustWork = FALSE, winslash = "/", fs::as_fs_path(linkedCacheDir)),
                   normalizePath(mustWork = FALSE, winslash = "/", fs::link_path(localDirs$cacheDir)))

  expect_true(linkFn(linkedInputDir, localDirs$inputDir))
  expect_identical(normalizePath(mustWork = FALSE, winslash = "/", fs::as_fs_path(linkedInputDir)),
                   normalizePath(mustWork = FALSE, winslash = "/", fs::link_path(localDirs$inputDir)))

  expect_true(linkFn(linkedOutputDir, localDirs$outputDir))
  expect_identical(normalizePath(mustWork = FALSE, winslash = "/", fs::as_fs_path(linkedOutputDir)),
                   normalizePath(mustWork = FALSE, winslash = "/", fs::link_path(localDirs$outputDir)))

  withr::local_options("reproducible.cachePath" = asPath(localDirs$cacheDir))
  on.exit({
    unlink(linkedDir, recursive = TRUE)
  }, add = TRUE)

  ## CRS
  targetCRS <- paste("+proj=lcc +lat_1=49 +lat_2=77 +lat_0=0 +lon_0=-95",
                     "+x_0=0 +y_0=0 +units=m +no_defs +ellps=GRS80 +towgs84=0,0,0")

  luxFile <- system.file("ex/lux.shp", package = "terra")
  luxVect <- terra::vect(luxFile)
  luxVectSmall <- luxVect[1, ]
  r <- terra::rast(luxVect, ncols=75, nrows=100)
  luxRast <- terra::rasterize(luxVect, r, "NAME_2")
  luxRastFile <- writeRaster(filename = tmpfile, luxRast)

  silence <- capture.output(
    studyArea <- prepInputs(
      targetFile = luxFile,
      destinationPath = asPath(localDirs$inputDir),
      useCache = FALSE
    ) |> Cache()
  )

  # Test useCache = FALSE -- this shouldn't need to be tested here, but it was failing here
  #   when first detected
  filesInCache <- dir(CacheStorageDir())
  for (i in 1:2)
    filesInCache <- gsub("^(.+)\\..+$", "\\1", filesInCache)
  filesInCache <- unique(filesInCache)
  expect_identical(length(filesInCache), 1L)

  rass <- prepInputs(targetFile = Filenames(luxRastFile),
             to = luxVectSmall,
             writeTo = asPath(file.path(localDirs$outputDir, basename(tmpfile))),
             useCache = FALSE
  ) |> Cache()

  testTif <- prepInputs(url = urlTif1, destinationPath = asPath(localDirs$inputDir),
                        useCache = FALSE) |>
    Cache()







  # These are VERY large for automated testing; reworked all the functionality to
  #   use small files
  ## get studyArea (Ontario forest management units)
  # studyArea <- {
  #   prepInputs(
  #     url = "https://www.gisapplication.lrc.gov.on.ca/fmedatadownload/Packages/FORMGMT.zip",
  #     destinationPath = asPath(inputDir),
  #     targetFile = "FOREST_MANAGEMENT_UNIT.shp",
  #     alsoExtract = "similar",
  #     fun = "sf::st_read"
  #   ) |>
  #     sf::st_transform(crs = targetCRS) |>
  #     sf::st_union()
  # } |>
  #   Cache()
  #
  # ## prep LCC for studyArea and write to outputDir
  # browser()
  # LCC <- Cache(
  #   prepInputs,
  #   url = paste0(
  #     "https://ftp.maps.canada.ca/pub/nrcan_rncan/",
  #     "Land-cover_Couverture-du-sol/canada-landcover_canada-couverture-du-sol/",
  #     "CanadaLandcover2010.zip"
  #   ),
  #   destinationPath = asPath(inputDir),
  #   # to = studyArea,
  #   cropTo = studyArea,
  #   # method = "near",
  #   datatype = "INT2U",
  #   fun = "terra::rast",
  #   targetFile = asPath("CAN_LC_2010_CAL.tif"),
  #   writeTo = asPath(file.path(outputDir, "LCC_ON_FMU.tif"))
  # )

  ## check files exist in the correct places
  ### no files at top level
  expect_true(all(list.files(linkedDir) %in% c("cache", "inputs", "outputs")))

  ### cache files in 'cache/'
  expect_true(length(list.files(linkedCacheDir)) > 0)
  expect_true(length(list.files(linkedCacheDir)) < 16) # Test for `useCache = FALSE`; should be only 3 unique cacheIds
  expect_equal(length(unique(showCache()$cacheId)), 3L)

  ### downloaded files etc. in 'inputs/'
  # expect_true(file.exists(file.path(linkedInputDir, "CanadaLandcover2010.zip")))
  expect_true(file.exists(file.path(linkedInputDir, "CHECKSUMS.txt")))
  expect_true(file.exists(file.path(linkedInputDir, "DEM.tif")))

  ### output files in 'outputs/'
  expect_true(file.exists(file.path(linkedOutputDir, basename(tmpfile))))
  # expect_identical(terra::sources(LCC), file.path(linkedOutputDir, "LCC_ON_FMU.tif"))

  # test end state
  expect_false(identical(character(), unlist(lapply(localDirs, dir))))
  ## cleanup
  try(unlink(linkedDir, recursive = TRUE))
  expect_identical(character(), unlist(lapply(localDirs, dir)))

  # withr::deferred_run()
})

test_that("preProcess fails if user provides non-existing file", {
  skip_on_cran()
  testInit("terra", opts = list(reproducible.inputPaths = NULL,
                                reproducible.interactiveOnDownloadFail = TRUE), verbose = 2)
  testthat::with_mocked_bindings(
    isInteractive = function() {
      FALSE
    },
    {
      errMsg <- testthat::capture_error({
        co <- capture.output({
          co <- capture.output(
            type = "message", {
            reproducible::preProcess(
              url = "https://github.com/tati-micheletti/host/raw/master/data/rasterTest",
              destinationPath = tmpdir
            )
          })
        })
      })
    }#,
    #.env = "reproducible"
  )
  expect_true(grepl("manual download", errMsg))
  expect_true(grepl("appendChecksumsTable", errMsg))

  withr::local_options(reproducible.interactiveOnDownloadFail = FALSE)

  .downloadErrorFn = function(xxxx) {
    tryCatch(stop(xxxx), httr2_http_404 = function(cnd) NULL,
             error = function(xxxx) xxxx,
             silent = TRUE) # httr2 has a unique error; need to silence it
    try(stop(xxxx), silent = TRUE)
  }
  testthat::with_mocked_bindings(
      .downloadErrorFn = .downloadErrorFn,
      isInteractive = function() {
        FALSE
      },
      {
      errMsg <- testthat::capture_error({
        preProcess(
          url = "https://github.com/tati-micheletti/host/raw/master/data/rasterTest",
          destinationPath = tmpdir
        )
      })
    }
    #.env = "reproducible"
  )
  expect_true(grepl("manual download", errMsg))
  expect_true(grepl("appendChecksumsTable", errMsg))
  withr::deferred_run()

  # This sub-block exercises the interactive-prompt branch of dlErrorHandling
  # (`isInteractive() && getOption("reproducible.interactiveOnDownloadFail")`).
  # Line 27 left the option at FALSE for the preceding sub-block — re-enable
  # it here so the AND-gate actually fires the prompt path, otherwise the
  # mocked `.readline` is never reached and the error message lacks the
  # "Download failed" tail this assertion checks for.
  withr::local_options(reproducible.interactiveOnDownloadFail = TRUE)
  testthat::with_mocked_bindings(
    isInteractive = function() {
      TRUE
    },
    .readline = function(prompt) {
      "n"
    },
    {
      co <- capture.output({
        co <- capture.output(type = "message", {
          mess <- testthat::capture_messages({
            errMsg <- testthat::capture_error({
              reproducible::preProcess(
                url = "https://github.com/tati-micheletti/host/raw/master/data/rasterTest",
                destinationPath = tmpdir
              )
            })
          })
        })
      })
    }#,
    #.env = "reproducible"
  )
  expect_true(sum(grepl("Download failed", errMsg)) == 1)

  url <- "https://github.com/tati-micheletti/host/raw/master/data/rasterTest"
  withr::local_options("reproducible.interactiveOnDownloadFail" = TRUE)
  zipFilename <- tempfile2(fileext = ".zip")
  testthat::with_mocked_bindings(
    .downloadErrorFn = .downloadErrorFn,
    isInteractive = function() {
      TRUE
    },
    .readline = function(prompt) {
      theFile <- file.path(tmpdir, "rasterTestAA")
      write.table(theFile, file = theFile)
      origDir <- setwd(dirname(theFile))
      on.exit(setwd(origDir), add = TRUE)
      zip(zipfile = zipFilename, files = basename2(theFile), flags = "-q")
      file.copy(zipFilename, file.path(tmpdir, basename(url)))
      "y"
    },
    {
      co <- capture.output(type = "message", {
        co <- capture.output({
          mess <- testthat::capture_messages({
            errMsg <- testthat::capture_error({
              reproducible::preProcess(
                fun = NA,
                url = url,
                destinationPath = tmpdir
              )
            })
          })
        })
      })
    }#,
    # .env = "reproducible"
  )
  expect_true(sum(grepl("manual.+download", mess)) == 1) # manual download may be broken by \n
  expect_true(sum(grepl("To prevent", mess)) == 1)
  expect_true(file.exists(zipFilename))
  cs <- fread(file.path(tmpdir, "CHECKSUMS.txt"), header = TRUE)
  expect_true(NROW(cs) == 2 || NROW(cs) == 3) # TODO this may be detecting a bug == on GA it is 2, locally it is 3
  expect_true(all(grepl("rasterTest", cs$file)))
  withr::deferred_run()
})

test_that("preProcess fails if user provides a non .zip/.tar as archive", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  co <- capture.output({
    co <- capture.output(type = "message", {
      pre <- reproducible::preProcess(url = theRasterTestZip, destinationPath = tmpdir)
    })
  })
  testthat::expect_is(object = pre, class = "list")
  oo <- capture.output(type = "message",
  testthat::expect_error({
    ras <- reproducible::preProcess(archive = pre$targetFilePath)
  })
  )
})

test_that("preProcess fails if user provides non-existing file", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  co <- capture.output({
    co <- capture.output(type = "message", {
      pre <- reproducible::preProcess(url = theRasterTestZip, destinationPath = tmpdir)
    })
  })
  testthat::expect_is(object = pre, class = "list")
  testthat::expect_error({
    ras <- reproducible::preProcess(archive = "fileDoesNotExist.zip")
  })
})

test_that("preProcess does not load when fun cannot be guessed", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  co <- capture.output({
    co <- capture.output(type = "message", {
      pre <- reproducible::preProcess(url = theRasterTestZip, destinationPath = tmpdir)
    })
  })
  testthat::expect_is(object = pre, class = "list")
  # Passing a directory as targetFile: no recognised extension means fun
  # cannot be guessed. preProcess does not load the object, so this should
  # succeed quietly with funChar = NULL rather than erroring.
  co <- capture.output({
    co <- capture.output(type = "message", {
      ras <- reproducible::preProcess(targetFile = tmpdir)
    })
  })
  expect_null(ras$funChar)
})

## 2022-11-03 this no longer fails on Ubuntu 20.04
# test_that("preProcess fails if the .rar file is defective", {
#   skip_on_cran()
#   testInit("raster")
#   co <- capture.output({
#     co <- capture.output(type = "message", {
#       testthat::expect_error({
#         ras <- preProcess(url = theRasterTestRar, destinationPath = tmpdir)
#       })
#     })
#   })
# })



test_that("preProcess fails if relative destPath not '.'", {
  # this fails on CRAN version 2.0.2
  skip_on_cran()
  skip_on_ci()
  testInit("googledrive", needGoogleDriveAuth = TRUE)
  dPath <- basename(tmpdir)
  out <- try(prepInputs(
    url = "https://drive.google.com/file/d/1u7o2BzPZ2Bo7hNcC8nEctNpDmp7ce84m",
    fun = "data.table::fread",
    destinationPath = dPath,
    purge = 1,
    filename2 = "userGcM3.csv", overwrite = TRUE
  ))
  expect_s3_class(out, "data.table")
})

test_that("test miscellaneous unit tests cache-helpers", {
  skip_if_not_installed("sf")
  skip_if_not_installed("terra")

  testInit(opts = list(reproducible.useMemoise = TRUE))

  a <- 1
  mess <- capture_message(.cacheMessage(a, "test", TRUE))
  expect_true(any(grepl(.message$LoadedCacheResult("Memoised"), mess)))

  mess <- capture_message(.cacheMessage(a, "test", FALSE))
  ## TODO: what was the old expected behaviour here? message now includes "added memoised copy"
  # expect_false(any(grepl(paste0(.message$LoadedCacheResult(), ".*added"), mess)))

  mess <- capture_message(.cacheMessage(a, "test", NA))
  expect_true(any(grepl(.message$LoadedCacheResult(), mess)))
  expect_false(all(grepl("adding", mess)))

  # studyAreaName with sf and sfc
  if (requireNamespace("sf", quietly = TRUE)) {
    pol <- sf::st_sfc(sf::st_polygon(list(cbind(c(0, 3, 3, 0, 0), c(0, 0, 3, 3, 0)))))
    h <- sf::st_sf(r = 5, pol)
    expect_true(is(studyAreaName(pol), "character"))
    expect_true(is(studyAreaName(h), "character"))
  }

  # studyAreaName with SpatVector
  if (requireNamespace("terra", quietly = TRUE)) {
    v <- terra::vect(system.file("ex/lux.shp", package = "terra"))
    expect_true(is(studyAreaName(v), "character"))
  }

  # studyAreaName with SPDF/SP
  # coords <- structure(c(-122.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
  #                     .Dim = c(5L, 2L))
  # Sr1 <- Polygon(coords)
  # Srs1 <- Polygons(list(Sr1), "s1")
  # StudyArea <- SpatialPolygons(list(Srs1), 1L)

  coords <- structure(c(-122.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
    .Dim = c(5L, 2L)
  )
  StudyArea <- terra::vect(coords, "polygons")
  terra::crs(StudyArea) <- crsToUse

  df <- data.frame(a = 1, row.names = row.names(StudyArea))

  # SPDF <- SpatialPolygonsDataFrame(StudyArea, df, match.ID = TRUE)
  expect_true(is(studyAreaName(StudyArea), "character"))
  # expect_true(is(studyAreaName(SPDF), "character"))

  # studyAreaName with random object
  expect_error(studyAreaName(integer(0)))

  # .checkCacheRepo
  withr::local_options(reproducible.cachePath = .reproducibleTempCacheDir())
  mess <- capture_message(.checkCacheRepo(a))
  expect_true(any(grepl(.message$NoCacheRepoSuppliedGrep, mess)))

  withr::local_options(reproducible.cachePath = NULL,
                       "rasterTmpDir" = tempdir2(rndstr(1, 6)),
                       "reproducible.inputPaths" = NULL,
                       "reproducible.overwrite" = TRUE)
  mess <- capture_message(.checkCacheRepo(a))
  expect_true(any(grepl(paste0(.message$NoCachePathSupplied, ". Using"), mess)))

  ## nextNumericName
  b <- nextNumericName("test.pdf")
  b1 <- nextNumericName(b)

  ## expect_true(grepl("_2.pdf", b1)) ## TODO: this number is not consistently 2 or 3
  aMess <- capture_messages({
    a <- Cache(rnorm, 1, useCache = FALSE, cachePath = tmpCache)
  })
  bMess <- capture_messages({
    b <- Cache(rnorm, 1, useCache = FALSE, cachePath = tmpCache)
  })
  expect_false(identical(a, b))
  expect_true(grepl("skipping Cache", aMess))
  expect_true(grepl("skipping Cache", bMess))

  ## getOption("reproducible.useMemoise" = FALSE)
  opt22 <- options("reproducible.useMemoise" = FALSE)
  aMess <- capture_messages({
    a <- Cache(rnorm, 1, cachePath = tmpCache)
  })
  bMess <- capture_messages({
    a <- Cache(rnorm, 1, cachePath = tmpCache)
  })
  options(opt22)
  cMess <- capture_messages({
    a <- Cache(rnorm, 1, cachePath = tmpCache)
  })
  dMess <- capture_messages({
    a <- Cache(rnorm, 1, cachePath = tmpCache)
  })
  # expect_true(identical(aMess, bMess[1]))
  expect_false(any(grepl(.message$LoadedCacheResult("Memoised"), bMess)))
  expect_true(any(grepl(.message$LoadedCacheResult("Memoised"), dMess)))

  ## showSimilar
  try(clearCache(ask = FALSE, x = tmpCache), silent = TRUE)
  aMess <- capture_messages({
    a <- Cache(rnorm, n = 1, mean = 1, cachePath = tmpCache)
  })
  # lapply(letters[11], function(l) assign(paste(rep(l, 4), collapse = ""), 1, envir = .GlobalEnv))
  oo <- capture.output(
    bMess <- capture_messages({
      b <- Cache(rnorm, n = 2, mean = 1, sd = 3, showSimilar = TRUE, cachePath = tmpCache)
    }))

  if (!getOption("reproducible.useCacheV3") %in% TRUE) {
    expect_true(any(grepl("different n", bMess)))
    expect_true(any(grepl("different .+sd", bMess)))
    # expect_true(any(grepl("new argument.*sd", bMess)))
    expect_true(any(grepl("next closest cacheId", bMess)))
    cMess <- capture_messages({
      b <- Cache(rnorm, n = 3, mean = 1, sd = 3, showSimilar = TRUE, cachePath = tmpCache)
    })
    expect_true(any(grepl("different n", cMess)))
    expect_false(any(grepl("new argument.*sd", cMess)))
    cMessCacheId <- gsub(".*cacheId (.*)\x1b\\[.*", "\\1", grep("cacheId", cMess, value = TRUE))
    bMessCacheId <- gsub(".*cacheId (.*)\x1b\\[.*", "\\1", grep("cacheId", bMess, value = TRUE))
    expect_false(identical(cMessCacheId, bMessCacheId))

    dMess <- capture_messages({
      b <- Cache(rnorm, n = 4, mean = 1, sd = 4, showSimilar = TRUE, cachePath = tmpCache)
    })

    # There are 2 ways this may come out -- similarity to 1 of 2 alternatives above
    expect1 <- any(grepl("different n, sd", dMess))
    expect2 <- any(grepl("different n", dMess)) && any(grepl("new argument.*sd", dMess))
    expect_true(expect1 || (expect2))
    dMessCacheId <- gsub(".*cacheId (.*)\x1b\\[.*", "\\1", grep("cacheId", dMess, value = TRUE))
    bMessCacheId <- gsub(".*cacheId (.*)\x1b\\[.*", "\\1", grep("cacheId", bMess, value = TRUE))
    if (expect1) {
      expect_false(identical(dMessCacheId, bMessCacheId))
    } else {
      expect_true(identical(dMessCacheId, bMessCacheId))
    }
  }

  rcompletelynew <- rmultinom
  # Now check function is prefered over args
  clearCache(tmpCache, ask = FALSE)
  eMess <- capture_messages({
    b <- Cache(rbinom, 4, 5, prob = 0.6, showSimilar = TRUE, cachePath = tmpCache)
  })
  fMess <- capture_messages({
    b <- Cache(rmultinom, 4, 5, prob = 0.6, showSimilar = TRUE, cachePath = tmpCache)
  })
  gMess <- capture_messages({
    b <- Cache(rmultinom, 14, 15, prob = 0.8, showSimilar = TRUE, cachePath = tmpCache)
  }) |> capture.output() -> oo
  hMess <- capture_messages({
    b <- Cache(rbinom, 14, 15, prob = 0.8, showSimilar = TRUE, cachePath = tmpCache)
  }) |> capture.output() -> oo
  iMess <- capture_messages({
    b <- Cache(rcompletelynew, 12, 15, prob = 0.8, showSimilar = TRUE, cachePath = tmpCache)
  })
  expect_true(any(grepl("no similar item", eMess))) # shouldn't find b/c new
  expect_true(any(grepl("no similar item", fMess))) # shouldn't find b/c args are same
  if (!getOption("reproducible.useCacheV3") %in% TRUE) {
    expect_true(any(grepl("next closest.+rmultin", gMess))) # should only find rmultinom
    expect_true(any(grepl("next closest.+rbinom", hMess))) # should only find rbinom
    expect_true(sum(grepl(".+rcompletelynew|next closest.+rmultin", iMess)) == 3) # should notice different name, but still find
  }

  ### UserTags matching -- prefer similar if all userTags match
  rcompletelynew <- rnorm
  # Now check function is prefered over args
  clearCache(tmpCache, ask = FALSE)
  jMess <- capture_messages({
    bj <- Cache(rnorm, 1, 2, 3, showSimilar = TRUE, cachePath = tmpCache, userTags = c("Hi"))
  })
  expect_true(any(grepl("no similar item", jMess))) # shouldn't find b/c new

  kMess <- capture_messages({
    bk <- Cache(rnorm, 1, 3, 4, showSimilar = TRUE, cachePath = tmpCache, userTags = c("By")) # not similar
  })

  expect_false(any(grepl("no similar item", kMess))) # should find bj as similar (args differ: 1,2,3 vs 1,3,4)

  lMess <- capture_messages({
    bl <- Cache(rnorm, 1, 3, 4, showSimilar = TRUE, cachePath = tmpCache, userTags = c("Hi")) # same, recovered
  })
  expect_true(any(grepl("Loaded", lMess))) # should only find rmultinom

  mMess <- capture_messages({
    bm <- Cache(rnorm, 1, 2, 3, showSimilar = TRUE, cachePath = tmpCache, userTags = c("By")) # same recovered
  })
  expect_true(any(grepl("Loaded", mMess))) # should only find rmultinom

  nMess <- capture_messages({
    bn <- Cache(rnorm, 1, 2, 2, showSimilar = TRUE, cachePath = tmpCache, userTags = c("By")) # similar to kMess
  }) |> capture.output() -> oo
  if (!getOption("reproducible.useCacheV3") %in% TRUE) {
    nMess <- grep("^.+next closest cacheId\\(s\\) (.+) of .+$", nMess, value = TRUE)
    expect_true(grepl(
      x = attr(bm, "tags"),
      gsub("^.+next closest cacheId\\(s\\) (.+) of .+$", "\\1", nMess)
    )) ## find mMess (jMess) because it's the most recent

    oMess <- capture_messages({
      bo <- Cache(rnorm, 1, 2, 1, showSimilar = TRUE, cachePath = tmpCache) # similar to kMess
    })
    oMess <- grep("^.+next closest cacheId\\(s\\) (.+) of .+$", oMess, value = TRUE)
    expect_true(grepl(
      x = attr(bn, "tags"),
      gsub("^.+next closest cacheId\\(s\\) (.+) of .+$", "\\1", oMess) ## TODO: fix failing test
    )) ## find nMess (jMess) because it's the most recent
  }
})

test_that("test debugCache arg", {
  testInit(opts = list(reproducible.useMemoise = TRUE))

  ## debugCache -- "complete"
  thing <- 1
  aa <- Cache(rnorm, thing, debugCache = "complete", cachePath = tmpCache)
  if (!getOption("reproducible.useCacheV3") %in% TRUE) {
    expect_true(identical(thing, attr(aa, "debugCache1")[[1]]))
  } else {
    expect_true(identical(thing, attr(aa, "debugCache1")[[2]]))
  }
  expect_true(identical(.robustDigest(thing), attr(aa, "debugCache2")$n))
  expect_true(is.numeric(aa))

  ## debugCache -- "quick"
  aa <- Cache(rnorm, thing, debugCache = "quick", cachePath = tmpCache)
  expect_true(identical(.robustDigest(thing), aa$hash$n))
  if (!getOption("reproducible.useCacheV3") %in% TRUE) {
    expect_true(identical(thing, aa$content[[1]]))
  } else {
    expect_true(identical(thing, aa$content[[2]]))
  }

  ## .unlistToCharacter
  expect_true(grepl("not list", unlist(.unlistToCharacter(1, 1))))
  expect_true(grepl("not list2", unlist(.unlistToCharacter(1, 0))))

  ## writeFuture
  comp <- # if (useDBI())
    .robustDigest("sdf") # else
  # "dda1fbb70d256e6b3b696ef0176c63de"
  drvHere <- if (useDBI() && .requireNamespace("RSQLite")) RSQLite::SQLite() else NULL

  expect_true(identical(
    comp,
    writeFuture(1, "sdf",
      cachePath = tmpCache, userTags = "",
      drv = drvHere
    )
  ))
  expect_error(writeFuture(1, "sdf", cachePath = "sdfd", userTags = ""))

  try(silent = TRUE, clearCache(tmpCache, ask = FALSE))
  warn <- capture_warnings(
    bMess <- capture_output({
      aMess <- capture_messages({
        aa <- Cache(fnCacheHelper, 1, verbose = 4, cachePath = tmpCache, cacheRepo2 = tmpCache)
      })
    })
  )
  # warn has 'package:reproducible' may not be available when loading, not relevant
  expect_true(any(grepl("fnCacheHelper", aMess)))
  expect_true(any(grepl("The hashing details", aMess)))
})

test_that("test warnings from cached functions", {
  skip_if_not_installed("sf")

  testInit(libraries = c("sf"), opts = list(reproducible.useMemoise = FALSE))
  warn1 <- capture_warnings({
    b <- Cache(rbinom, 4, 5, prob = 6, cachePath = tmpCache)
  })

  fun <- function(n, size, prob) {
    rbinom(n, size, prob)
  }
  warn2 <- capture_warnings({
    d <- Cache(fun, 4, 5, 6, cachePath = tmpCache)
  })
  warnCompare <- "rbinom.+NAs produced"
  expect_true(grepl(warnCompare, warn1)) # includes the call because .call = FALSE, and call added manually in Cache
  expect_true(grepl("NAs produced", warn2))
  expect_false(grepl(warnCompare, warn2)) # this is false because the warning message doesn't include the call with normal warn
})

test_that("test cache-helpers with stacks", {
  skip_if_not_installed("raster")

  # THIS TEST CAN BE DELETED AFTER RASTER IS DEFUNCT
  testInit("raster")

  tmpfile <- tempfile(tmpdir = tmpdir, fileext = ".tif")
  tmpfile2 <- tempfile(tmpdir = tmpdir, fileext = ".tif")
  r <- raster(extent(0, 5, 0, 5), resolution = 1, vals = rep(1:2, length.out = 25))
  r1 <- raster(extent(0, 5, 0, 5), resolution = 1, vals = rep(1:2, length.out = 25))
  s <- raster::stack(r, r1)

  ## in memory
  b <- .prepareFileBackedRaster(s, tmpCache)
  is(b, "RasterStack")
  expect_true(length(list.files(file.path(tmpCache, "rasters"))) == 0)

  ## with 1 backups
  r <- .writeRaster(r, filename = tmpfile, overwrite = TRUE)
  s <- addLayer(r, r1)
  b <- .prepareFileBackedRaster(s, tmpCache)

  expect_true(all(basename(c(tmpfile)) %in% basename(list.files(tmpCache, recursive = TRUE))))
  expect_false(all(basename(c(tmpfile2)) %in% basename(list.files(tmpCache, recursive = TRUE))))

  ## with 2 backups
  r1 <- .writeRaster(r1, filename = tmpfile2, overwrite = TRUE)
  s <- addLayer(r, r1)
  b <- .prepareFileBackedRaster(s, tmpCache)
  expect_true(all(basename(c(tmpfile, tmpfile2)) %in% basename(list.files(tmpCache, recursive = TRUE))))

  ## removing entry from Cache
  fls <- grep(basename(tmpfile),
    list.files(tmpCache, recursive = TRUE, full.names = TRUE),
    value = TRUE
  )
  file.remove(fls)
  expect_false(all(basename(c(tmpfile, tmpfile2)) %in% basename(list.files(tmpCache, recursive = TRUE))))
  b <- .prepareFileBackedRaster(s, tmpCache)
  expect_true(all(basename(c(tmpfile, tmpfile2)) %in% basename(list.files(tmpCache, recursive = TRUE))))

  # Test deleted raster backed file
  file.remove(tmpfile2)
  expect_error(
    {
      b <- .prepareFileBackedRaster(s, tmpCache)
    },
    "The following file-backed rasters"
  )
})

test_that("test miscellaneous unit tests cache-helpers", {
  skip_if_not_installed("googledrive")

  testInit("googledrive")
  a <- Cache(rnorm, 1, cachePath = tmpCache)
  mess <- capture_messages(clearCache(cachePath = tmpCache))
  expect_true(any(grepl("x not specified, but cachePath is", mess)))
})

test_that("test Cache(useCloud=TRUE, ...)", {
  skip_on_cran()
  skip_on_ci()

  testInit(
    c("googledrive", "terra"),
    tmpFileExt = c(".tif", ".grd"),
    needGoogleDriveAuth = TRUE,
  )

  ## service accounts cannot upload to standard drive folders (no quota)
  skip_if_service_account_releaseVer_NotLinux()

  withr::local_options(
    reproducible.cachePath = file.path(tempdir(), rndstr(1, 7)),
    reproducible.ask = FALSE,
    reproducible.useMemoise = FALSE
  )

  clearCache(x = tmpCache)
  googleSetupForUseCloud(cloudFolderID, tmpdir, tmpCache)

  testsForPkgs <- "testsForPkgs"
  tryCatch(googledrive::drive_ls(testsForPkgs), error = function(x)
    googledrive::drive_mkdir(name = testsForPkgs))
  newDir <- retry(quote(googledrive::drive_mkdir(name = rndstr(1, 6), path = testsForPkgs)))
  cloudFolderID <- newDir

  # local absent, cloud absent
  mess1 <- capture_messages({
    a1 <- Cache(rnorm, 1, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = TRUE)
  })
  expect_true(any(grepl("Uploading", mess1)))

  # local present, cloud present
  mess2 <- capture_messages({
    a1 <- Cache(rnorm, 1, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = TRUE)
  })
  expect_true(any(grepl(.message$LoadedCacheResult(), mess2)))
  .message$LoadedCacheResult
  expect_false(all(grepl("uploaded", ignore.case = TRUE, mess2)))
  expect_false(all(grepl("download", mess2)))

  # local absent, cloud present
  clearCache(userTags = .robustDigest(1), x = tmpCache, useCloud = FALSE)
  mess3 <- capture_messages({
    a1 <- Cache(rnorm, 1, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = TRUE)
  })

  # it is actually both loaded & saved locally; so should be loadedCachedResult and saved
  if (!useDBI()) {
    expect_true(any(grepl(.message$LoadedCacheResult(), mess3)))
  } else {
    expect_false(any(grepl(.message$LoadedCacheResult(), mess3)))
  }
  expect_false(any(grepl("Uploaded", mess3)))
  expect_true(any(grepl("Downloading", mess3)))

  # local present, cloud absent
  clearCache(x = tmpCache, useCloud = TRUE, cloudFolderID = cloudFolderID)
  a1 <- Cache(rnorm, 2, cloudFolderID = cloudFolderID, cachePath = tmpCache)
  mess4 <- capture_messages({
    a2 <- Cache(rnorm, 2, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = TRUE)
  })

  expect_true(any(grepl(.message$LoadedCacheResult(), mess4)))
  expect_true(any(grepl("Uploading", mess4)))
  expect_false(any(grepl("Download", mess4)))

  # cloudFolderID missing
  reproducible::clearCache(x = tmpCache, useCloud = TRUE, cloudFolderID = cloudFolderID)

  withr::local_options("reproducible.cloudFolderID" = NULL)
  # on.exit(try(googledrive::drive_rm(cloudFolderFromCacheRepo(tmpdir))), add = TRUE)
  # Try two different cloud folders -- based on tmpdir and tmpCache
  warn5 <- capture_warnings({
    mess5 <- capture_messages({
      a2 <- Cache(rnorm, 3, cachePath = tmpdir, useCloud = TRUE)
    })
  })
  options("reproducible.cloudFolderID" = NULL)

  expect_true(any(grepl("Uploading", mess5)))
  expect_false(any(grepl("Download", mess5)))


  # on.exit(try(googledrive::drive_rm(cloudFolderFromCacheRepo(tmpCache))), add = TRUE)
  warn6 <- capture_warnings({
    mess6 <- capture_messages({
      a2 <- Cache(rnorm, 3, cachePath = tmpCache, useCloud = TRUE)
    })
  })

  expect_true(any(grepl("Uploading", mess6)))
  expect_false(any(grepl("Download", mess6)))
  expect_false(any(grepl(.message$LoadedCacheResult(), mess6)))
  expect_true(isTRUE(all.equal(length(warn6), 0)))

  # Clear all
  try(drive_rm(drive_ls(cloudFolderFromCacheRepo(tmpCache))), silent = TRUE)

  # will use getOption("reproducible.cloudFolderID") b/c not specified, which is not cloudFolderFromCacheRepo(tmpCache)
  # clearCache(x = tmpCache, useCloud = TRUE)
  # Switch to tmpCache only
  withr::local_options(reproducible.cloudFolderID = tmpCache)
  cloudFolderID <- getOption("reproducible.cloudFolderID") # currently based on tmpCache

  # Add 3 things to cloud and local -- then clear them all
  for (i in 1:3) {
    a1 <- Cache(rnorm, i, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = TRUE)
  }
  expect_silent({
    mess1 <- capture_messages(
      clearCache(x = tmpCache, useCloud = TRUE, cloudFolderID = cloudFolderID)
    )
  })

  expect_true(NROW(googledrive::drive_ls(path = cloudFolderFromCacheRepo(tmpCache))) == 0)

  # Add 3 things to local, only 2 to cloud -- clear them all, without an error
  for (i in 1:2) {
    a1 <- Cache(rnorm, i, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = TRUE)
  }
  a1 <- Cache(rnorm, 3, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = FALSE)
  expect_silent({
    mess2 <- capture_messages(
      clearCache(x = tmpCache, useCloud = TRUE, cloudFolderID = cloudFolderID)
    )
  })
  expect_true(NROW(googledrive::drive_ls(path = cloudFolderFromCacheRepo(tmpCache))) == 0)

  # Add 2 things to local and cloud -- clear only 1 of them, without an error
  Cache(rnorm, 1, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = TRUE)
  Cache(rnorm, 2, cloudFolderID = cloudFolderID, cachePath = tmpCache, useCloud = TRUE)
  expect_silent({
    mess2 <- capture_messages(
      clearCache(x = tmpCache, userTags = CacheDigest(rnorm(1))$outputHash, useCloud = TRUE, cloudFolderID = cloudFolderID)
    )
  })

  gdriveLs <- googledrive::drive_ls(path = cloudFolderFromCacheRepo(tmpCache))
  expect_true(NROW(gdriveLs) == 2)
  expect_true(all(grepl(unique(showCache(tmpCache)[[.cacheTableHashColName()]]), gdriveLs$name)))
})

test_that("test Cache(useCloud=TRUE, ...) with raster-backed objs -- tif and grd", {
  skip_on_cran()
  skip_on_ci()
  testInit(c("googledrive", "terra"),
    needGoogleDriveAuth = TRUE,
    opts = list(reproducible.ask = FALSE)
  )

  ## service accounts cannot upload to standard drive folders (no quota)
  skip_if_service_account_releaseVer_NotLinux()

  googleSetupForUseCloud(cloudFolderID, tmpdir, tmpCache)

  withr::local_options(reproducible.cachePath = tmpdir)

  # on.exit(
  #   {
  #     retry(quote(googledrive::drive_rm(googledrive::as_id(newDir$id))))
  #   },
  #   add = TRUE
  # )
  clearCache(x = tmpCache)
  clearCache(x = tmpdir)
  newDir <- retry(quote(googledrive::drive_mkdir(name = basename2(tmpdir), path = "testsForPkgs")))
  cloudFolderID <- newDir

  testRasterInCloud(".tif",
    cloudFolderID = cloudFolderID, numRasterFiles = 1, tmpdir = tmpdir,
    type = "Raster"
  )

  retry(quote(googledrive::drive_rm(googledrive::as_id(newDir$id))))
  clearCache(x = tmpdir)
  newDir <- retry(quote(googledrive::drive_mkdir(name = rndstr(1, 6), path = "testsForPkgs")))
  cloudFolderID <- newDir

  ## the 3 raster files include the .grd, .gri, and .grd.aux.xml
  testRasterInCloud(".grd",
    cloudFolderID = cloudFolderID, numRasterFiles = 3, tmpdir = tmpdir,
    type = "Raster"
  )
})

test_that("test Cache(useCloud=TRUE, ...) with raster-backed objs -- stack", {
  skip_on_cran()
  skip_on_ci()
  testInit(c("googledrive", "terra"),
    needGoogleDriveAuth = TRUE,
    opts = list(reproducible.ask = FALSE)
  )

  ## service accounts cannot upload to standard drive folders (no quota)
  skip_if_service_account_releaseVer_NotLinux()

  googleSetupForUseCloud(cloudFolderID, tmpdir, tmpCache)
  withr::local_options(reproducible.cachePath = tmpdir)
  clearCache(x = tmpCache)
  clearCache(x = tmpdir)
  newDir <- retry(quote(googledrive::drive_mkdir(name = basename2(tmpdir), path = "testsForPkgs")))
  cloudFolderID <- newDir

  testRasterInCloud(".tif",
    cloudFolderID = cloudFolderID, numRasterFiles = 2, tmpdir = tmpdir,
    type = "Stack"
  )
})

test_that("test Cache(useCloud=TRUE, ...) with raster-backed objs -- brick", {
  skip_on_cran()
  skip_on_ci()
  testInit(c("googledrive", "terra"),
    needGoogleDriveAuth = TRUE,
    opts = list(reproducible.ask = FALSE)
  )

  ## service accounts cannot upload to standard drive folders (no quota)
  skip_if_service_account_releaseVer_NotLinux()

  googleSetupForUseCloud(cloudFolderID, tmpdir, tmpCache)
  withr::local_options(reproducible.cachePath = tmpdir)
  clearCache(x = tmpCache)
  clearCache(x = tmpdir)
  newDir <- retry(quote(googledrive::drive_mkdir(name = tempdir2(), path = "testsForPkgs")))
  cloudFolderID <- newDir

  testRasterInCloud(".tif",
    cloudFolderID = cloudFolderID, numRasterFiles = 1, tmpdir = tmpdir,
    type = "Brick"
  )
})

test_that("prepInputs works with team drives", {
  skip_on_cran()
  skip_on_ci()

  testInit(needGoogleDriveAuth = TRUE, "googledrive")
  withr::local_options(
    reproducible.cachePath = file.path(tempdir(), rndstr(1, 7)),
    reproducible.ask = FALSE
  )

  googleSetupForUseCloud(cloudFolderID, tmpdir, tmpCache)

  # zipUrl <- "https://drive.google.com/file/d/1zRX2c55ebJbQtjijCErEfGxhsa7Ieph2" # Orig
  zipUrl <- "https://drive.google.com/file/d/1JpdvM8QiyCyNyQAvSaFU0rAY-1I3mcbp"

  # This will fail if it is hit too many times -- we don't want the test to report
  #  fail because of this
  withr::local_options("reproducible.interactiveOnDownloadFail" = FALSE)
  if (packageVersion("googledrive") < "2.0.0") {
    wb <- prepInputs(
      targetFile = "WB_BCR.shp", destinationPath = tmpdir, url = zipUrl,
      alsoExtract = "similar",
      team_drive = TRUE
    )
  } else {
    err <- capture_error(
      noisy <- capture.output(
        wb <- prepInputs(
          targetFile = "WB_BCR.shp", destinationPath = tmpdir, url = zipUrl,
          alsoExtract = "similar",
          shared_drive = TRUE
        )
      )
    )
    if (is.null(err)) {
      if (.requireNamespace("sf")) {
        expect_true(is(wb, "sf"))
      }
    }
  }
})

test_that("test data.table caching", {
  testInit("data.table")

  # make sure that .robustDigest does not include the cache-added attributes
  c1 <- Cache(rnorm, 11, cachePath = tmpCache)
  c1a <- .robustDigest(c1)

  c1 <- Cache(rnorm, 11, cachePath = tmpCache)
  c1b <- .robustDigest(c1)
  expect_true(identical(c1a, c1b)) # failed pre reproducible 0.2.4.9000

  # This was wrong on with .robustDigest with no data.frame method -- this fails previous version
  a <- list(data.table(a = LETTERS, b = letters, c = letters, d = letters, e = letters))
  b <- list(data.table(a = rep(LETTERS, 2), b = rep(letters, 2), c = rep(letters, 2), d = rep(letters, 2), e = rep(letters, 2)))
  aC <- CacheDigest(a)
  bC <- CacheDigest(b)
  expect_false(identical(aC, bC))
})

test_that("test ALTREP integers", {
  testInit(.qs2Format, opts = list(reproducible.cacheSaveFormat = .qs2Format,
                             reproducible.cacheSpeed = "fast"))

  for (i in .cacheSaveFormats) {
    if (.requireNamespace(i)) {
      for (s in c("slow", "fast")) {
      withr::local_options(reproducible.cacheSaveFormat = i,
                           reproducible.cacheSpeed = s)
      a <- 1991:20200
      aDig <- .robustDigest(a)
      tf <- tempfile(fileext = i);
      if (identical(i, .rdsFormat)) {
        saveRDS(a, file = tf);
        b <- readRDS(tf)
      } else if (i %in% c(.qs2Format, .qsFormat)) {
        fek <- .fileExtsKnown()
        funSave <- fek$saveFun[fek$extension == i]
        funSave <- eval(parse(text = funSave))
        funRead <- fek$fun[fek$extension == i]
        funRead <- eval(parse(text = funRead))
        funSave(a, file = tf);
        b <- funRead(tf)
      } else {

      }
      bDig <- .robustDigest(b)
      expect_true(identical(aDig, bDig))
      withr::deferred_run()
      }
    }

  }
})

test_that("Checksums read and written correctly", {
  testInit()
  sampleDir <- system.file("examples", package = "reproducible")
  sampleFiles <- list.files(sampleDir, pattern = "[.R]", full.names = TRUE)
  on.exit(unlink(dirname(tmpdir), recursive = TRUE), add = TRUE)

  expect_true(all(file.copy(sampleFiles, tmpdir)))

  csf <- file.path(tmpdir, "CHECKSUMS.txt")
  cnamesR <- c(
    "result", "expectedFile", "actualFile", "checksum.x", "checksum.y",
    "algorithm.x", "algorithm.y", "filesize.x", "filesize.y"
  )
  cnamesW <- c("file", "checksum", "filesize", "algorithm")

  # 1. read Checksums without CHECKSUMS.txt file
  expect_true(NROW(Checksums(tmpdir)) == 0)

  Checksums(dirname(csf), write = TRUE)
  txt <- Checksums(dirname(csf))
  txt <- txt[grepl("R", expectedFile)]
  data.table::setorderv(txt, "expectedFile")

  csums <- txt$checksum.x

  # 2. read Checksums with empty CHECKSUMS.txt file
  expect_true(file.create(csf))
  txt <- Checksums(tmpdir)
  expect_true(all(colnames(txt) == cnamesR))
  expect_equal(nrow(txt), 0)

  # 3. write Checksums without CHECKSUMS.txt
  expect_true(file.remove(csf))
  txt <- Checksums(dirname(csf), write = TRUE)
  txt <- txt[grepl("[.]R$", basename(expectedFile))]
  data.table::setorderv(txt, "expectedFile")
  expect_true(all(colnames(txt) == cnamesR))
  expect_equal(nrow(txt), NROW(dir(tmpdir, pattern = "[.]R$")))

  # expect_identical(sort(txt$expectedFile), sort(basename(sampleFiles)))
  # expect_true(all(sort(txt$expectedFile) == sort(basename(sampleFiles))))
  a <- basename2(sort(grep("\\.R$", txt[result == "OK", ]$expectedFile, value = TRUE))) # the [result == "OK",] is because some unknown reason on GA
  b <- sort(basename2(sampleFiles))
  expect_identical(a, b)
  # expect_identical(a[1], b[1])
  # expect_identical(a[2], b[2])
  # expect_identical(a[3], b[3])
  # expect_identical(a[4], b[4])
  # expect_identical(a[5], b[5])
  # expect_identical(a[6], b[6])

  # 4. read Checksums with non-empty, but incomplete CHECKSUMS.txt file
  out <- txt[, `:=`(
    file = expectedFile,
    checksum = checksum.x,
    algorithm = algorithm.x
  )][1:4]
  # out <- data.table::data.table(file = basename(sampleFiles[-1]),
  #                  checksum = csums[-1],
  #                  algorithm = c("xxhash64", "xxhash64", "xxhash64", "xxhash64"),
  #                  stringsAsFactors = FALSE)
  utils::write.table(out, csf, eol = "\n", col.names = TRUE, row.names = FALSE)

  txt <- Checksums(tmpdir, write = FALSE)
  txt <- txt[!is.na(result)]
  data.table::setorderv(txt, "expectedFile")
  expect_equal(nrow(txt), NROW(out))
  expect_true(all(txt$result == "OK"))

  # 5. make Checksums complete -- just append one line
  expect_true(all(colnames(txt) == cnamesR))
  txt <- Checksums(tmpdir, write = TRUE)
  txt <- Checksums(tmpdir)
  txt <- txt[grepl("[.]R$", expectedFile)][result == "OK", ]
  expect_equal(nrow(txt), NROW(dir(tmpdir, pattern = "[.]R$")))
  # expect_true(all(sort(txt$expectedFile) == sort(basename(sampleFiles))))
  a <- basename2(sort(grep("\\.R$", txt$expectedFile, value = TRUE)))
  # print(txt$expectedFile)
  expect_identical(a, b)

  # Now add a leading dot (./) to filenames; this must be removed
  csf <- dir(tmpdir, full.names = TRUE, pattern = "CHECKSUMS.txt")
  txt <- read.table(csf, header = TRUE) |> as.data.table()
  txt[, file := paste0("./", file)]
  write.table(txt, file = csf)
  doCS <- Checksums(tmpdir)
  doCS <- doCS[grepl("[.]R$", expectedFile)][result == "OK", ]
  expect_equal(nrow(doCS), NROW(dir(tmpdir, pattern = "[.]R$")))

})

test_that("prepInputs works with NULL archive + file without extension, but originally a .zip", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  skip_on_transient_http({
    noisyOutput <- capture.output({
      testthat::expect_message({
        ras <- reproducible::prepInputs(
          url = "https://github.com/tati-micheletti/host/raw/master/data/unknownExtension",
          alsoExtract = "similar",
          destinationPath = tempdir2(rndstr(1, 6))
        )
      })
    })
  })
  testthat::expect_is(object = ras, class = rasterType())
})

test_that("prepInputs WORKS if the file is not originally a .zip, but archive is provided (only extension matters)", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  skip_on_transient_http({
    noisyOutput <- capture.output(
      testthat::expect_message({
        ras <- reproducible::prepInputs(
          url = "https://github.com/tati-micheletti/host/raw/master/data/unknownTAR",
          alsoExtract = "similar", archive = "unknownTAR.tar",
          destinationPath = tempdir2(rndstr(1, 6))
        )
      })
    )
  })
  testthat::expect_is(object = ras, class = rasterType())
})

test_that("prepInputs WORKS if passing archive .zip", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  clearCache(x = normalizePath(getOption("reproducible.cachePath"), winslash = "/"))
  skip_on_transient_http({
    noisyOutput <- capture.output(
      testthat::expect_message({
        ras <- reproducible::prepInputs(
          url = "https://github.com/tati-micheletti/host/raw/master/data/unknownExtension",
          archive = "unknownExtension.zip",
          alsoExtract = "similar", destinationPath = tempdir2(rndstr(1, 6))
        )
      })
    )
  })
  testthat::expect_is(object = ras, class = rasterType())
})

test_that("prepInputs WORKS passing just targetFile that is NOT an archive", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  skip_on_transient_http({
    noisyOutput <- capture.output({
      testthat::expect_message({
        ras <- reproducible::prepInputs(
          url = "https://github.com/tati-micheletti/host/raw/master/data/unknownTIF",
          alsoExtract = "similar", targetFile = "unknownTIF.tif",
          destinationPath = tempdir2(rndstr(1, 6))
        )
      })
    })
  })
  testthat::expect_is(object = ras, class = rasterType())
})

test_that("prepInputs WORKS passing archive + targetFile", {
  skip_on_cran()
  testInit("terra", needInternet = TRUE)
  clearCache(x = normalizePath(getOption("reproducible.cachePath"), winslash = "/"))
  skip_on_transient_http({
    noisyOutput <- capture.output({
      testthat::expect_message({
        ras <- reproducible::prepInputs(
          url = "https://github.com/tati-micheletti/host/raw/master/data/unknownExtension",
          archive = "unknownExtension.zip", targetFile = "rasterTest.tif",
          alsoExtract = "similar",
          destinationPath = tempdir2(rndstr(1, 6))
        )
      })
    })
  })
  testthat::expect_is(object = ras, class = rasterType())
})

#' Create a list with names from object names
#'
#' This is a convenience wrapper around `a <- 1; newList <- list(a); names(newList) <- "a"`.
#'
#' @param ... Any elements to add to a list, as in `base::list`
#' @details
#' This will return a named list, where names are the object names, captured
#' internally in the function and assigned to the list. If a user manually supplies
#' names, these will be kept (i.e., not overwritten by the object name).
#' @export
#' @examples
#' a <- 1
#' b <- 2
#' d <- 3
#' (newList <- listNamed(a, b, dManual = d)) # "dManual" name kept
#'
#'
listNamed <- function(...) {
  dotsSUB <- as.list(substitute(list(...)))[-1]
  names <- vapply(dotsSUB, deparse, FUN.VALUE = character(1))
  ll <- list(...)
  existingNames <- names(dotsSUB)
  if (!is.null(existingNames)) {
    nonEmptyNames <- nzchar(existingNames)
    names[nonEmptyNames] <- existingNames[nonEmptyNames]
  }
  names(ll) <- names
  ll
}

#' Generic function to post process objects
#'
#' \if{html}{\figure{lifecycle-maturing.svg}{options: alt="maturing"}}
#'
#' @export
#' @param x  A GIS object of postProcessing,
#'           e.g., Spat* or sf*. This can be provided as a
#'           `rlang::quosure` or a normal R object.
#' @importFrom utils capture.output
#' @seealso `prepInputs`
#' @rdname postProcess
postProcess <- function(x, ...) {
  UseMethod("postProcess")
}



#' @export
#' @rdname postProcess
postProcess.list <- function(x, ...) {
  lapply(x, function(y) postProcess(y, ...))
}

#' Post processing for GIS objects
#'
#' The method for GIS objects (terra `Spat*` & sf classes) will
#' crop, reproject, and mask, in that order.
#' This is a wrapper for [cropTo()], [fixErrorsIn()],
#' [projectTo()], [maskTo()] and [writeTo()],
#' with a required amount of data manipulation between these calls so that the crs match.
#'
#' @section Post processing sequence:
#'
#'   If the `rasterToMatch` or `studyArea` are passed, then
#'   the following sequence will occur:
#'
#'   \enumerate{
#'     \item Fix errors [fixErrorsIn()]. Currently only errors fixed are for
#'            `SpatialPolygons` using `buffer(..., width = 0)`.
#'     \item Crop using [cropTo()]
#'     \item Project using [projectTo()]
#'     \item Mask using [maskTo()]
#'     \item Determine file name [determineFilename()]
#'     \item Write that file name to disk, optionally [writeTo()]
#'   }
#'
#'   NOTE: checksumming does not occur during the post-processing stage, as
#'   there are no file downloads. To achieve fast results, wrap
#'   `prepInputs` with `Cache`
#'
#'
#' @param ... Additional arguments passed to methods. For `spatialClasses`,
#'            these are: [cropTo()], [fixErrorsIn()],
#'            [projectTo()], [maskTo()],
#'            [determineFilename()], and [writeTo()].
#'            Each of these may also pass `...` into other functions, like
#'            [writeTo()].
#'            This might include potentially important arguments like `datatype`,
#'            `format`. Also passed to `terra::project`,
#'            with likely important arguments such as `method = "bilinear"`.
#'            See details.
#'
#'
#' @section Backwards compatibility with `rasterToMatch` and/or `studyArea` arguments:
#'
#' For backwards compatibility, `postProcess` will continue to allow passing
#'  `rasterToMatch` and/or `studyArea` arguments. Depending on which of these
#'  are passed, different things will happen to the `targetFile` located at `filename1`.
#'
#' See *Use cases* section in [postProcessTo()] for post processing behaviour with
#'   the new `from` and `to` arguments.
#'
#' \subsection{If `targetFile` is a raster (`Raster*`, or `SpatRaster`) object:}{
#'   \tabular{lccc}{
#'                  \tab `rasterToMatch`      \tab `studyArea`  \tab  Both               \cr
#'     `extent`     \tab Yes                  \tab   Yes        \tab `rasterToMatch`     \cr
#'     `resolution` \tab Yes                  \tab   No         \tab `rasterToMatch`     \cr
#'     `projection` \tab Yes                  \tab   No*        \tab `rasterToMatch`*    \cr
#'     `alignment`  \tab Yes                  \tab   No         \tab `rasterToMatch`     \cr
#'     `mask`       \tab No**                 \tab   Yes        \tab `studyArea`**       \cr
#'   }
#'  *Can be overridden with `useSAcrs`.
#'
#'  **Will mask with `NA`s from `rasterToMatch` if `maskWithRTM`.
#' }
#'
#' \subsection{If `targetFile` is a vector (`Spatial*`, `sf` or `SpatVector`) object:}{
#'   \tabular{lccc}{
#'                  \tab `rasterToMatch`      \tab `studyArea`  \tab Both               \cr
#'     `extent`     \tab Yes                  \tab   Yes        \tab `rasterToMatch`    \cr
#'     `resolution` \tab NA                   \tab   NA         \tab NA                 \cr
#'     `projection` \tab Yes                  \tab   No*        \tab `rasterToMatch`*   \cr
#'     `alignment`  \tab NA                   \tab   NA         \tab NA                 \cr
#'     `mask`       \tab No                   \tab   Yes        \tab `studyArea`        \cr
#'   }
#'  *Can be overridden with `useSAcrs`
#' }
#'
#'
#' @export
#' @example inst/examples/example_postProcess.R
#' @rdname postProcess
#' @return A GIS file (e.g., `RasterLayer`, `SpatRaster` etc.) that has been
#' appropriately cropped, reprojected, masked, depending on the inputs.
#'
postProcess.default <- function(x, ...) {
  if (inherits(x, "quosure")) {
    .requireNamespace("rlang", stopOnFALSE = TRUE)
    x <- rlang::eval_tidy(x)
  }
  return(postProcessTo(from = x, ...))
}





#' Crop a `Spatial*` or `Raster*` object
#'
#' This function is deprecated. Use `cropTo`. If used, all arguments will be passed
#' to `cropTo` anyway.
#'
#' @param x A `Spatial*`, `sf`, or `Raster*` object.
#'
#' @param studyArea `SpatialPolygons*` object used for masking and possibly cropping
#'                  if no `rasterToMatch` is provided.
#'                  If not in same CRS, then it will be `spTransform`ed to
#'                  CRS of `x` before masking. Currently, this function will not reproject the
#'                  `x`. Optional in `postProcess`.
#'
#' @param rasterToMatch Template `Raster*` object used for cropping (so extent should be
#'                      the extent of desired outcome) and reprojecting (including changing the
#'                      resolution and projection).
#'                      See details in [postProcess()].
#'
#' @param ... Passed to `[cropTo()]`
#'
#' @param useCache Logical, default `getOption("reproducible.useCache", FALSE)`, whether
#'                 `Cache` is used internally.
#'
#' @inheritParams projectInputs
#'
#' @author Eliot McIntire, Jean Marchal, Ian Eddy, and Tati Micheletti
#' @export
#' @importFrom methods is
#' @return A GIS file (e.g., RasterLayer, SpatRaster etc.) that has been
#' appropriately cropped.
#' @rdname deprecated
#' @name deprecated
cropInputs <- function(x, studyArea, rasterToMatch, verbose = getOption("reproducible.verbose", 1), ...) {
  UseMethod("cropInputs")
}

#' @export
#' @rdname deprecated
cropInputs.default <- function(x, ...) {
  cropTo(x, ...)
}


#' Do some minor error fixing
#'
#' `fixErrors` --> `fixErrosTerra`
#'
#' @param x A `SpatialPolygons*` or `sf` object.
#'
#' @param objectName Optional. This is only for messaging; if provided, then messages relayed
#'                   to user will mention this.
#'
#' @param attemptErrorFixes Will attempt to fix known errors. Currently only some failures
#'        for `SpatialPolygons*` are attempted.
#'        Notably with `terra::buffer(..., width = 0)`.
#'        Default `TRUE`, though this may not be the right action for all cases.
#' @param useCache Logical, default `getOption("reproducible.useCache", FALSE)`, whether
#'                 Cache is used on the internal `terra::buffer` command.
#' @param testValidity Logical. If `TRUE`, the a test for validity will happen
#'                 before actually running buffering (which is the solution in most
#'                 cases). However, sometimes it takes longer to test for validity
#'                 than just buffer without testing (there are no consequences of
#'                 buffering if everything is valid). If `FALSE`, then the
#'                 test will be skipped and the buffering will happen regardless.
#'                 If `NA`, then all testing and buffering will be skipped.
#' @param ... Passed to methods. None currently implemented.
#'
#' @export
#' @rdname deprecated
#' @return A GIS file (e.g., RasterLayer, SpatRaster etc.) that has been
#' attempted to be fixed, if it finds errors.
#' @keywords internal
#' @seealso [fixErrorsIn()], [postProcessTo()], [postProcess()]
#'
fixErrors <- function(x, objectName, attemptErrorFixes = TRUE,
                      useCache = getOption("reproducible.useCache", FALSE),
                      verbose = getOption("reproducible.verbose", 1),
                      testValidity = getOption("reproducible.testValidity", TRUE),
                      ...) {
  UseMethod("fixErrors")
}

#' @export
#' @keywords internal
#' @rdname deprecated
fixErrors.default <- function(x, objectName, attemptErrorFixes = TRUE,
                              useCache = getOption("reproducible.useCache", FALSE),
                              verbose = getOption("reproducible.verbose", 1),
                              testValidity = getOption("reproducible.testValidity", TRUE),
                              ...) {
  fixErrorsIn(x)
}


#' Project `Raster*` or `Spatial*` or `sf` objects
#'
#' Deprecated. Use [projectTo()].
#'
#' @param x A `Raster*`, `Spatial*` or `sf` object
#'
#' @param targetCRS The CRS of x at the end  of this function (i.e., the goal)
#'
#' @param ... Passed to [projectTo()].
#'
#' @param rasterToMatch Template `Raster*` object passed to the `to` argument of
#'                      [projectTo()], thus will changing the
#'                      resolution and projection of `x`.
#'                      See details in [postProcessTo()].
#'
#' @return A file of the same type as starting, but with projection (and possibly
#' other characteristics, including resolution, origin, extent if changed).
#'
#' @export
#' @inheritParams prepInputs
#' @rdname deprecated
#' @seealso [projectTo()]
#' @return A GIS file (e.g., RasterLayer, SpatRaster etc.) that has been
#' appropriately reprojected.
#'
projectInputs <- function(x, targetCRS, verbose = getOption("reproducible.verbose", 1), ...) {
  UseMethod("projectInputs")
}

#' @export
#' @rdname deprecated
projectInputs.default <- function(x, targetCRS, ...) {
  projectTo(x, projectTo = targetCRS, ...)
}





#' Mask module inputs
#'
#' `maskInputs` is deprecated. Use [maskTo()]
#'
#' @param x An object to do a geographic raster::mask/raster::intersect.
#'          See methods.
#' @param ... Passed to methods. None currently implemented.
#'
#' @inheritParams cropInputs
#'
#' @author Eliot McIntire and Jean Marchal
#' @export
#' @importFrom utils capture.output
#' @rdname deprecated
#' @return A GIS file (e.g., RasterLayer, SpatRaster etc.) that has been
#' appropriately masked.
#' @seealso [maskTo()], [postProcessTo()] for related examples
maskInputs <- function(x, studyArea, ...) {
  UseMethod("maskInputs")
}


#' @export
maskInputs.default <- function(x, studyArea, rasterToMatch = NULL, maskWithRTM = NULL,
                               verbose = getOption("reproducible.verbose", 1), ...) {
  if (!is.null(maskWithRTM) && !is.null(rasterToMatch))
    maskTo <- rasterToMatch
  else
    maskTo <- studyArea
  maskTo(x, maskTo = maskTo, verbose = verbose, ...)
}


#' Determine filename, either automatically or manually
#'
#' Determine the filename, given various combinations of inputs.
#'
#' @details
#' The post processing workflow, which includes this function,
#' addresses several scenarios, and depending on which scenario, there are
#' several file names at play. For example, `Raster` objects may have
#'   file-backed data, and so *possess a file name*, whereas `Spatial`
#'   objects do not. Also, if post processing is part of a [prepInputs()]
#'   workflow, there will always be a file downloaded. From the perspective of
#'   `postProcess`, these are the "inputs" or `filename1`.
#'   Similarly, there may or may not be a desire to write an
#'   object to disk after all post processing, `filename2`.
#'
#'   This subtlety means that there are two file names that may be at play:
#'   the "input" file name (`filename1`), and the "output" filename (`filename2`).
#'   When this is used within `postProcess`, it is straight forward.
#'
#'
#'   However, when `postProcess` is used within a `prepInputs` call,
#'   the `filename1` file is the file name of the downloaded file (usually
#'   automatically known following the downloading, and refered to as `targetFile`)
#'   and the `filename2` is the file name of the of post-processed file.
#'
#'   If `filename2` is `TRUE`, i.e., not an actual file name, then the cropped/masked
#'   raster will be written to disk with the original `filenam1/targetFile`
#'   name, with `prefix` prefixed to the basename(`targetFile`).
#'
#'   If `filename2` is a character string, it will be the path of the saved/written
#'   object e.g., passed to `writeOutput`. It will be tested whether it is an
#'   absolute or relative path and used as is if absolute or
#'   prepended with `destinationPath` if relative.
#'
#' @inheritParams projectInputs
#'
#' @param destinationPath Optional. If `filename2` is a relative file path, then this
#'                        will be the directory of the resulting absolute file path.
#'
#' @param prefix The character string to prepend to `filename1`, if `filename2`
#'               not provided.
#'
#' @include helpers.R
#'
#' @details
#'  If `filename2` is `logical`, then the output
#'  filename will be `prefix` prefixed to the basename(`filename1`).
#'  If a character string, it
#'  will be the path returned. It will be tested whether it is an
#'  absolute or relative path and used as is if absolute or prepended with
#'  `destinationPath` if provided, and if `filename2` is relative.
#'
#' @rdname determineFilename
#' @param filename1  Character strings giving the file paths of
#'                   the *input* object (`filename1`) `filename1`
#'                   is only used for messaging (i.e., the object itself is passed
#'                   in as `x`) and possibly naming of output (see details
#'                   and `filename2`).
#'
#' @param filename2   `filename2` is optional, and is either
#'                   NULL (no writing of outputs to disk), or several options
#'                   for writing the object to disk. If
#'                   `TRUE` (the default), it will give it a file name determined by
#'                   `.prefix(basename(filename1), prefix)`. If
#'                   a character string, it will use this as its file name. See
#'                   [determineFilename()].
#'
#' @inheritParams postProcess
determineFilename <- function(filename2 = NULL, filename1 = NULL,
                              destinationPath = getOption("reproducible.destinationPath", "."),
                              verbose = getOption("reproducible.verbose", 1),
                              prefix = "Small", ...) {
  if (!is.null(filename2)) {
    dots <- list(...)

    if (!is.null(dots$inputFilePath)) {
      stop("inputFilePath is being deprecated; use filename1")
    }

    if (!is.null(dots$postProcessedFilename)) {
      stop("postProcessedFilename is being deprecated; use filename2")
    }

    if (!is.null(dots$targetFilePath)) {
      stop(
        "targetFilePath is being deprecated from determineFilename:\n",
        "  use filename2 and filename1."
      )
    }

    if (!(is.logical(filename2) || is.character(filename2) || is.null(filename2))) {
      stop("filename2 must be logical or character string or NULL")
    }

    filename2 <- if (!identical(filename2, FALSE)) { # allow TRUE or path
      if (isTRUE(filename2)) {
        # 1. Take destinationPath, if it exists
        # 2. Take dirname of filename1, if it exists and is absolute path
        # 3. Take getwd()
        theDir <- destinationPath
        if (is.null(destinationPath)) {
          if (is.character(filename1)) {
            if (isAbsolutePath(filename1)) {
              theDir <- dirname(filename1)
              messagePrepInputs("filename2 is NULL; using dirname(filename1) as destinationPath",
                verbose = verbose
              )
            }
          }
        }
        filename3 <- normPath(tempfile(tmpdir = theDir, fileext = ".tif"))
        .prefix(filename3, prefix)
      } else {
        iap <- isAbsolutePath(filename2)
        if (all(iap)) {
          filename2
        } else {
          if (any(iap)) {
            stop("filename2 must be all relative or all absolute paths")
          }
          if (!is.null(destinationPath)) {
            file.path(destinationPath, basename(filename2))
          } else {
            filename2 # accept relative
          }
        }
      }
    } else {
      NULL
    }
    if (exists("tmpfile", inherits = FALSE)) {
      messagePrepInputs("Saving output to ", filename2, ".",
        "Specify filename1 or filename2 for more control, ",
        "or set filename2 to NULL to prevent saving to disk",
        verbose = verbose
      )
    }
  }
  filename2
}

#' Write module inputs on disk
#'
#' See [writeTo()]
#'
#' @param x  The object save to disk i.e., write outputs
#'
#' @param overwrite Logical. Should file being written overwrite an existing file if it exists.
#'
#' @param ... Passed into [writeTo()]
#'
#' @inheritParams prepInputs
#'
#' @author Eliot McIntire and Jean Marchal
#' @return A GIS file (e.g., SpatRaster etc.) that has been
#' appropriately written to disk. In the case of vector datasets, this will
#' be a side effect. In the case of gridded objects (Raster*, SpatRaster), the
#' object will have a file-backing.
#' @importFrom methods is
#' @export
#' @rdname deprecated
#'
#' @examples
#' if (requireNamespace("terra", quietly = TRUE)) {
#'   r <- terra::rast(terra::ext(0, 100, 0, 100), vals = 1:1e2)
#'
#'   tf <- tempfile(fileext = ".tif")
#'   writeOutputs(r, tf)
#' }
writeOutputs <- function(x, ..., # filename2,
                         overwrite = getOption("reproducible.overwrite", NULL),
                         verbose = getOption("reproducible.verbose", 1)) {
  UseMethod("writeOutputs")
}

#' @rdname deprecated
#' @export
writeOutputs.default <- function(x, ..., # filename2,
                                 overwrite = getOption("reproducible.overwrite", FALSE),
                                 verbose = getOption("reproducible.verbose", 1)) {
  writeTo(x, ..., overwrite = overwrite, verbose = verbose)
}



#' Assess the appropriate raster layer data type
#'
#' When writing raster-type objects to disk, a `datatype` can be specified. These
#' functions help identify what smallest `datatype` can be used.
#'
#' @param ras  The `RasterLayer` or `RasterStack` for which data type will be assessed.
#' @param type Character. `"writeRaster"` (default) or `"GDAL"` (defunct)
#'             to return the recommended
#'             data type for writing from the raster packages, respectively, or
#'             `"projectRaster"` to return recommended resampling type.
#'
#' @export
#' @return A character string indicating the data type of the spatial layer
#' (e.g., "INT2U"). See `terra::datatype()`
#' @rdname assessDataType
#'
#' @example inst/examples/example_assessDataType.R
assessDataType <- function(ras, type = "writeRaster") {
  UseMethod("assessDataType")
}

#' @export
#' @rdname assessDataType
assessDataType.default <- function(ras, type = "writeRaster") {
  ## using ras@data@... is faster, but won't work for @values in large rasters

  if (inherits(ras, "RasterStack")) {
    xs <- lapply(names(ras), FUN = function(x) {
      y <- assessDataType(ras = ras[[x]], type)
      return(y)
    })

    return(unlist(xs))
  }

  N <- 1e5

  if (nlayers2(ras) > 1) { # for RasterStack, RasterBrick, SpatRaster of nlyr > 1
    xs <- lapply(seq(names(ras)), FUN = function(x) { # can't use names(ras) directly because can't have same name 2x in lapply, but can in SpatRaster
      y <- assessDataType(ras = ras[[x]], type)
      return(y)
    })

    return(unlist(xs))
  }

  datatype <- NULL
  .requireNamespace("terra", stopOnFALSE = TRUE)
  if (terra::ncell(ras) > 1e8) { # for very large rasters, try a different way
    maxValCurrent <- maxFn(ras)
    ras <- setMinMaxIfNeeded(ras)
    # if (maxValCurrent != maxValue(ras))
    datatype <- dataType2(ras)
  } else {
    ras <- setMinMaxIfNeeded(ras)
  }

  if (is.null(datatype)) {
    if (terra::ncell(ras) > N) {
      rasVals <- tryCatch(suppressWarnings(sampRand(x = ras, size = N)),
        error = function(x) rep(NA_integer_, N)
      )
      if (is.factor(rasVals))
        rasVals <- as.integer(rasVals)
    } else {
      rasVals <- values2(ras)
    }
    minVal <- minFn(ras) # min(ras@data@min)
    maxVal <- maxFn(ras) # max(ras@data@max)
    signVal <- minVal < 0
    doubVal <- any(floor(rasVals) != rasVals, na.rm = TRUE) ## faster than any(x %% 1 != 0)
    datatype <- if (doubVal) {
      names(MinValsFlts)[min(which(minVal >= MinValsFlts & maxVal <= MaxValsFlts))]
    } else {
      ## only check for binary if there are no decimals and no signs
      logi <- all(!is.na(.bincode(rasVals[!is.na(rasVals)], c(-1, 1)))) ## range needs to include 0
      if (logi) {
        "LOG1S"
      } else {
        names(MinVals)[min(which(minVal >= unlist(MinVals) & maxVal <= unlist(MaxVals)))]
      }
    }
  }
  # convert datatype if needed
  datatype <- switchDataTypes(datatype, type = type)
  datatype
}




setMinMaxIfNeeded <- function(ras) {
  # special case where the colours already match the discrete values
  suppressWarnings(maxValCurrent <- maxFn(ras))
  needSetMinMax <- FALSE
  if (isTRUE(any(is.na(maxValCurrent)))) {
    needSetMinMax <- TRUE
  } else {
    # if the colors are set and are the same length of the integer sequence between min and max, don't override
    if (length(.getColors(ras)[[1]])) {
      if (!is.na(suppressWarnings(maxFn(ras))) && !is.na(suppressWarnings(minFn(ras)))) {
        if (length(.getColors(ras)[[1]]) == (maxFn(ras) - minFn(ras) + 1)) {
          return(ras)
        }
      }
    }
    possibleShortCut <- maxValCurrent %in% c(unlist(MaxVals), unlist(MaxVals) + 1)
    if (isTRUE(all(possibleShortCut))) {
      needSetMinMax <- TRUE
    }
  }
  if (isTRUE(needSetMinMax)) {
    .requireNamespace("terra", stopOnFALSE = TRUE)
    large <- if (nlayers2(ras) > 25 || terra::ncell(ras) > 1e7) TRUE else FALSE
    if (large) message("  Large ", class(ras), " detected; setting minimum and maximum may take time")
    suppressWarnings(ras <- terra::setMinMax(ras))
    if (large) message("  ... Done")
  }
  ras
}


roundTo6Dec <- function(x) {
  # check if integer
  if (all(x %% 1 != 0)) {
    # First test whether they are remotely close to each other
    rounded <- round(x, 6)
    if (!identical(x, rounded)) {
      x <- rounded
    }
  }
  x
}

#' @importFrom utils capture.output
suppressWarningsSpecific <- function(code, falseWarnings, verbose = getOption("reproducible.verbose", 1)) {
  warns <- list()
  suppressWarnings(withCallingHandlers(
    {
      yy <- eval(code)
    },
    warning = function(xx) {
      trueWarnings <- grep(falseWarnings, xx$message, invert = TRUE, value = TRUE)
      if (length(trueWarnings)) {
        warns <<- paste(trueWarnings, collapse = "\n  ")
      }
    },
    error = function(xx) stop(xx$message),
    message = function(xx) xx
  ))
  if (length(warns)) {
    lapply(warns, warning)
  }

  return(yy)
}

#' @importFrom utils capture.output
captureWarningsToAttr <- function(code, verbose = getOption("reproducible.verbose", 1)) {
  warns <- character()
  warn <- capture.output(
    type = "message",
    suppressWarnings(withCallingHandlers(
      {
        yy <- eval(code)
      },
      warning = function(xx) {
        warns <<- paste0("warn::", xx$message)
      }
    ))
  )
  trueWarnings <- grepl("warn::.*", warns)
  if (length(warn[!trueWarnings])) {
    lapply(warns[!trueWarnings], warning)
  }
  warns <- gsub("warn::", "", warns[trueWarnings])
  attr(yy, "warning") <- paste(warns, collapse = "\n")
  return(yy)
}

dtp <- list()
dtp[["LOG1S"]] <- 1
dtp[["INT1"]] <- 255 / 2
dtp[["INT2"]] <- 65534 / 2
dtp[["INT4"]] <- 4294967296 / 2
dtp[["FLT4"]] <- 3.4e+38
dtp[["FLT8"]] <- Inf
dtps <- c("LOG1S", "INT1U", "INT1S", "INT2U", "INT2S", "INT4U", "INT4S", "FLT4S", "FLT8S", "FLT8U")
names(dtps) <- dtps
datatypeVals <- lapply(dtps, function(namdtp) {
  d <- dtp[grep(substr(namdtp, 1, 4), names(dtp), value = TRUE)]
  div <- substr(namdtp, 5, 5)
  mult <- ifelse(div == "U", 2, 1)
  Max <- trunc(unlist(d) * mult)
  sign1 <- ifelse(div == "U", 0, -1)
  if (grepl("LOG", names(d))) {
    Min <- 0
  } else {
    Min <- Max * sign1
  }
  list(Min = Min, Max = Max)
})
MaxVals <- lapply(datatypeVals, function(x) unname(x$Max))
MinVals <- lapply(datatypeVals, function(x) unname(x$Min))
MinValsFlts <- MinVals[grep("FLT", names(MinVals), value = TRUE)]
MaxValsFlts <- MaxVals[grep("FLT", names(MinVals), value = TRUE)]
projNotWKT2warn <- "Using PROJ not WKT2"


switchDataTypes <- function(datatype, type) {

  gdalVersion <- "3.1"
  if (.requireNamespace("sf"))
    gdalVersion <- sf::sf_extSoftVersion()["GDAL"]

  if (missing(datatype))
    datatype <- "Float32"
  gdals <- list(
    LOG1S = "Byte",
    INT1S = "Int16", # added below if gdalversion ok
    INT2S = "Int16",
    INT4S = "Int32",
    INT8S = "Int64",
    INT1U = "Byte",
    INT2U = "UInt16",
    # INT4U = "UInt32", # added below if gdalversion ok
    # INT8U = "UInt64", # added below if gdalversion ok
    FLT4S = "Float32",
    FLT8S = "Float64"
  )

  gdalsOrig <- gdals

  if (gdalVersion >= as.numeric_version("3.5"))
    gdals <- append(gdals,
                    list(
                      INT4U = "UInt32",
                      INT8U = "UInt64"
                    ))
  if (gdalVersion >= as.numeric_version("3.7"))
    gdals[which(names(gdals) %in% "INT1S")] <- list(INT1S = "Int8")



  rast <- names(gdals)
  names(rast) <- gdals

  if (identical(type, "GDAL"))
    if (isTRUE(!datatype %in% names(gdals)))
      if (!datatype %in% unname(unlist(gdals))) {
        warning("datatype ", datatype, " is not an option with this version of gdal: ",
                gdalVersion, "\nSetting to ", tail(gdalsOrig, 1))
        datatype <- tail(gdalsOrig, 1)
      }

  gdals <- append(
    gdals,
    list(datatype)) # default is user-supplied -- which could be already a gdal-correct specification for example
  rast <- append(rast, list(datatype))

  datatype <- switch(type,
    GDAL = {
      do.call(switch, append(list(datatype), gdals))
    },
    projectRaster = {
      switch(datatype,
        Float32 = "bilinear",
        Float64 = "bilinear",
        datatype <- "ngb"
      )
    },
    writeRaster = {
      do.call(switch, append(list(datatype), rast))
    },
    stop("incorrect argument: type must be one of writeRaster, projectRaster, or GDAL")
  )
  return(datatype)
}


sampRand <- function(x, size, method, ...) {
  if (isRaster(x)) {
    .requireNamespace("raster")
    raster::sampleRandom(x, size = size, ...)
  } else {
    .requireNamespace("terra")
    out <- terra::spatSample(x, size = size, ...)
    out <- out[, 1] # it is returned as a df
  }
}

test_that("prepInputs doesn't work (part 3)", {
  skip_on_cran() # too long
  skip_if_not_installed("sf")
  testInit(c("terra", "sf"),
    tmpFileExt = c(".tif", ".tif", ".tif"),
    opts = list(
      "rasterTmpDir" = tempdir2(rndstr(1, 6)),
      "reproducible.inputPaths" = NULL,
      "reproducible.overwrite" = TRUE,
      "rgdal_show_exportToProj4_warnings" = "none"
    ) # https://gis.stackexchange.com/questions/390945/importing-raster-files-warning-and-extracting-covariates-error-with-raster-and
  )

  withr::local_options("reproducible.cachePath" = tmpdir)

  # Add a study area to Crop and Mask to
  # Create a "study area"

  coords <- structure(c(-122.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
    .Dim = c(5L, 2L)
  )
  coords2 <- structure(c(-115.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
    .Dim = c(5L, 2L)
  )

  StudyArea <- terra::vect(coords, "polygons")
  terra::crs(StudyArea) <- crsToUse
  StudyArea2 <- terra::vect(coords2, "polygons")
  terra::crs(StudyArea2) <- crsToUse

  # coords <- structure(c(-122.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
  #                     .Dim = c(5L, 2L))
  # coords2 <- structure(c(-115.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
  #                      .Dim = c(5L, 2L))
  # Sr1 <- Polygon(coords)
  # Srs1 <- Polygons(list(Sr1), "s1")
  # StudyArea <- SpatialPolygons(list(Srs1), 1L)
  # crs(StudyArea) <- crsToUse
  #
  # Sr1 <- Polygon(coords2)
  # Srs1 <- Polygons(list(Sr1), "s1")
  # StudyArea2 <- SpatialPolygons(list(Srs1), 1L)
  # crs(StudyArea2) <- crsToUse

  nonLatLongProj <- paste(
    "+proj=lcc +lat_1=49 +lat_2=77 +lat_0=0 +lon_0=-95",
    "+x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs"
  )
  nc <- sf::st_as_sf(StudyArea) # system.file("shape/nc.shp", package="sf"))
  nc1 <- sf::st_transform(nc, nonLatLongProj)
  ncSmall <- sf::st_as_sf(StudyArea2)
  ncSmall <- sf::st_transform(ncSmall, nonLatLongProj)
  ncSmall <- sf::st_buffer(ncSmall, dist = -10000)
  b <- postProcess(nc1, studyArea = ncSmall, filename2 = NULL)
  expect_true(is(b, "sf"))
  expect_equal(terra::ext(b), terra::ext(ncSmall))
  expect_true(sf::st_area(b) < sf::st_area(nc1))

  r <- suppressWarnings(terra::rast(nc1, resolution = 1000)) # TODO: temporary until raster crs fixes

  rB <- suppressWarnings(terra::rast(nc1, resolution = 4000)) # TODO: temporary until raster crs fixes
  rSmall <- suppressWarnings(terra::rast(ncSmall, resolution = 4000)) # TODO: temporary until raster crs fixes

  # Tests with RasterBrick
  r2 <- r1 <- rB
  r1[] <- runif(terra::ncell(rB))
  r2[] <- runif(terra::ncell(rB))

  b <- c(r1, r2)
  terra::crs(b) <- sf::st_crs(ncSmall)$input
  b1 <- postProcess(b, studyArea = ncSmall, useCache = FALSE)
  expect_true(.isSpatRaster(b1))

  s <- c(r1, r2)
  crs(s) <- crs(nonLatLongProj)
  s1 <- postProcess(s, studyArea = ncSmall, useCache = FALSE)
  expect_true(.isSpatRaster(s1))
  expect_equal(s1[], b1[], ignore_attr = TRUE)

  b <- writeRaster(b, filename = tmpfile[1], overwrite = TRUE)
  b1 <- postProcess(b, studyArea = ncSmall, useCache = FALSE, writeTo = tmpfile[2], overwrite = TRUE)
  expect_true(.isSpatRaster(b1))

  s1 <- postProcess(s, studyArea = ncSmall, useCache = FALSE, writeTo = tmpfile[2], overwrite = TRUE)
  expect_true(.isSpatRaster(s1))

  # Test datatype setting
  dt1 <- "INT2U"
  s <- writeRaster(s, filename = tmpfile[2], overwrite = TRUE)
  s1 <- postProcess(s,
    studyArea = ncSmall, useCache = FALSE, writeTo = tmpfile[1], overwrite = TRUE,
    datatype = dt1
  )
  expect_identical(terra::datatype(s1), rep(dt1, terra::nlyr(s)))

  # Test datatype setting
  dt1 <- c("INT2U", "INT4U")
  s <- writeRaster(s, filename = tmpfile[1], overwrite = TRUE)
  warns <- capture_error({
    s1 <- postProcess(s,
      studyArea = ncSmall, useCache = FALSE, writeTo = tmpfile[2], overwrite = TRUE,
      datatype = dt1
    )
  })
  expect_true(any(grepl("Expecting a single", warns)))

  dt1 <- "INT4U"
  b <- writeRaster(b, filename = tmpfile[2], overwrite = TRUE)
  b1 <- postProcess(b,
    studyArea = ncSmall, useCache = FALSE, writeTo = tmpfile[1], overwrite = TRUE,
    datatype = dt1
  )
  expect_identical(terra::datatype(b1), rep(dt1, terra::nlyr(b1)))

  # now raster with sf
  skip_if_not_installed("terra")
  r1 <- terra::rasterize(terra::vect(nc1), r)
  r2 <- postProcess(r1, studyArea = ncSmall, filename2 = NULL)
  expect_true(.isSpatRaster(r2))
  expect_true(terra::ncell(r2) < terra::ncell(r1))
  expect_true((terra::xmin(terra::ext(ncSmall)) - terra::xmin(r2)) < terra::res(r2)[1] * 2)
  expect_true((terra::ymin(terra::ext(ncSmall)) - terra::ymin(r2)) < terra::res(r2)[2] * 2)
  expect_true((terra::ymax(terra::ext(ncSmall)) - terra::ymax(r2)) > -(terra::res(r2)[2] * 2))
  expect_true((terra::xmax(terra::ext(ncSmall)) - terra::xmax(r2)) > -(terra::res(r2)[2] * 2))

  # postProcess
  expect_error(postProcess(1, to = r2), regexp = "from must be a")
  expect_error(postProcess(list(1, 1), to = r2), regexp = "from must be a")

  nc2 <- postProcess(nc1, studyArea = as(ncSmall, "sf"))
  expect_equal(st_area(nc2), st_area(ncSmall))

  # cropInputs
  expect_true(identical(1, cropInputs(1)))
  nonLatLongProj2 <- paste(
    "+proj=lcc +lat_1=51 +lat_2=77 +lat_0=0 +lon_0=-95",
    "+x_0=0 +y_0=0 +ellps=GRS80 +units=m +no_defs"
  )
  nc3 <- suppressWarningsSpecific(
    {
      sf::st_transform(nc1, CRSobj = nonLatLongProj2)
    },
    falseWarnings = "Discarded datum Unknown based on GRS80 ellipsoid in Proj4 definition|PROJ support is provided by the sf and terra packages among others"
  )
  nc4 <- cropInputs(nc3, studyArea = ncSmall)
  ncSmall2 <- suppressWarningsSpecific(
    {
      sf::st_transform(ncSmall, CRSobj = nonLatLongProj2)
    },
    falseWarnings = "Discarded datum Unknown based on GRS80 ellipsoid in Proj4 definition|PROJ support is provided by the sf and terra packages among others"
  )
  expect_true(isTRUE(all.equal(terra::ext(nc4), terra::ext(ncSmall2))))

  mess <- capture_error({
    nc4 <- cropInputs(nc3, studyArea = 1)
  })
  expect_s3_class(mess, "simpleError")

  ncSmallShifted <- ncSmall + 10000000
  ncSmallShifted <- st_as_sf(ncSmallShifted)
  st_crs(ncSmallShifted) <- st_crs(ncSmall)
  mess <- capture_messages(
    cropInputs(ncSmall, studyArea = ncSmallShifted)
  )
  expect_true(any(grepl("overlap", mess)))

  # cropInputs.sf
  nc3 <- st_transform(nc1, crs = nonLatLongProj2)
  nc4 <- cropInputs(nc3, studyArea = ncSmall)
  ncSmall2 <- st_transform(ncSmall, crs = nonLatLongProj2)
  expect_true(isTRUE(all.equal(terra::ext(nc4), terra::ext(ncSmall2))))

  # studyArea as spatial object
  nc5 <- cropInputs(nc3, studyArea = ncSmall)
  ncSmall2 <- st_transform(ncSmall, crs = nonLatLongProj2)
  expect_true(isTRUE(all.equal(terra::ext(nc5), terra::ext(ncSmall2))))
  expect_true(isTRUE(all.equal(terra::ext(nc5), terra::ext(nc4))))


  # rasterToMatch
  nc5 <- cropTo(nc3, cropTo = r)
  nc5Extent_r <- st_transform(nc5, crs = crs(r))
  expect_true(isTRUE(abs(terra::xmin(r) - sf::st_bbox(nc5Extent_r)["xmin"]) < terra::res(r)[1]))
  expect_true(isTRUE(abs(terra::ymin(r) - sf::st_bbox(nc5Extent_r)["ymin"]) < terra::res(r)[1]))
  expect_true(isTRUE(abs(terra::xmax(r) - sf::st_bbox(nc5Extent_r)["xmax"]) < terra::res(r)[1]))
  expect_true(isTRUE(abs(terra::ymax(r) - sf::st_bbox(nc5Extent_r)["ymax"]) < terra::res(r)[1]))

  ncSmallShifted <- ncSmall + 10000000
  ncSmallShifted <- st_as_sf(ncSmallShifted)
  st_crs(ncSmallShifted) <- st_crs(ncSmall)
  expect_message(
    regexp = "results in no data",
    aaa <- cropInputs(ncSmall, studyArea = ncSmallShifted)
  )
  expect_s3_class(aaa, "sf")
  # expect_true(NROW(out11) == 0)

  # LINEARRING Example
  p6 <- terra::vect("POLYGON ((0 60, 0 0, 60 0, 60 20, 100 20, 60 20, 60 60, 0 60))")
  p6a <- fixErrorsIn(p6)
  expect_true(terra::is.valid(p6a))
  expect_false(terra::is.valid(p6))
  # projectInputs pass through
  expect_error(projectInputs(x = 1), "argument .+ is missing")

  #using deprecated filename2 arg
  expect_error(cropInputs(ncSmall, studyArea = ncSmallShifted, filename2 = "use_WriteTo_Instead.shp"))

})

test_that("writeOutputs with non-matching writeTo", {
  testInit(c("terra"), tmpFileExt = c(".grd", ".tif"))

  r <- terra::rast(terra::ext(0, 10, 0, 10), vals = rnorm(100))
  r <- terra::writeRaster(r, filename = tmpfile[1], overwrite = TRUE)
  r[] <- r[]
  warn <- capture_warnings({
    r1 <- writeOutputs(r, writeTo = tmpfile[2])
  })
  r2 <- terra::rast(Filenames(r1))
  vals1 <- r2[]
  vals2 <- r1[]
  vals3 <- r[]
  expect_true(identical(vals1, vals2))
  expect_true(identical(vals1, vals3))
  expect_false(identical(normPath(Filenames(r)), normPath(Filenames(r1))))
  expect_true(identical(normPath(Filenames(r2)), normPath(Filenames(r1))))
})

test_that("cropInputs crops too closely when input projections are different", {
  skip_on_cran()
  # skip_on_ci() ## TODO: why is this failing on GHA but not locally??? (2022-11-04)

  testInit("terra", opts = list(
    "rasterTmpDir" = tempdir2(rndstr(1, 6)),
    "reproducible.overwrite" = TRUE,
    "reproducible.inputPaths" = NULL,
    "rgdal_show_exportToProj4_warnings" = "none" # https://gis.stackexchange.com/questions/390945/importing-raster-files-warning-and-extracting-covariates-error-with-raster-and
  ), needGoogleDriveAuth = TRUE)

  oo <- capture.output(orig <- terra::terraOptions()$memfrac)
  terra::terraOptions(memfrac = 0.6)
  on.exit(terra::terraOptions(memfrac = orig))

  ext2 <- terra::ext(c(
    xmin = -3229772.32501426,
    xmax = 3680227.67498574,
    ymin = -365833.605586135,
    ymax = 3454166.39441387
  ))
  x <- terra::rast(ext2,
    crs = paste(
      "+proj=laea +lat_0=45 +lon_0=-100 +x_0=0 +y_0=0",
      "+a=6370997 +b=6370997 +units=m +no_defs"
    ),
    resolution = c(10000, 10000)
  )
  x <- terra::setValues(x, 1)

  RTMext <- terra::ext(c(
    xmin = -1613500.00000023,
    xmax = -882500.000000228,
    ymin = 7904500,
    ymax = 9236000
  ))
  RTM <- terra::rast(RTMext,
    crs = paste(
      "+proj=lcc +lat_1=49 +lat_2=77 +lat_0=0 +lon_0=-95 +x_0=0 +y_0=0",
      "+ellps=GRS80 +units=m +no_defs"
    ),
    resolution = c(250, 250)
  )
  RTM <- setValues(RTM, 2)
  out <- postProcess(x = x, rasterToMatch = RTM, filename2 = NULL)
  expect_true(nrow(out[is.na(out[]) & is.na(RTM[])]) == 0)
})

test_that("maskInputs errors when x is Lat-Long", {
  skip_on_cran()
  skip_on_ci()
  testInit("sf", opts = list(
    "rasterTmpDir" = tempdir2(rndstr(1, 6)),
    "reproducible.overwrite" = TRUE,
    "reproducible.inputPaths" = NULL
  ), needGoogleDriveAuth = TRUE)

  i <- 0
  roads <- list()
  clearCache()

  smallSA <- terra::ext(c(
    -117.580383168455, -117.43120279669,
    61.0576330401172, 61.0937107807574
  ))

  x <- terra::rast(smallSA,
    crs = "+proj=longlat +ellps=GRS80 +no_defs",
    resolution = c(0.001, 0.001)
  )
  suppressWarnings(smallSA <- terra::vect(terra::ext(x), "polygons"))
  terra::crs(smallSA) <- terra::crs(x)

  noisyOutput <- capture.output(
    suppressWarningsSpecific(
      falseWarnings = "attribute variables|st_buffer does not correctly buffer longitude",
      roads1 <- prepInputs(
        targetFile = "miniRoad.shp",
        alsoExtract = "similar",
        url = "https://drive.google.com/file/d/1Z6ueq8yXtUPuPWoUcC7_l2p0_Uem34CC",
        studyArea = smallSA,
        useCache = FALSE,
        fun = "sf::st_read",
        destinationPath = tmpdir,
        writeTo = "miniRoad.shp"
      )
    )
  )
  # clearCache()
  noisyOutput <- capture.output(
    roads2 <- prepInputs(
      targetFile = "miniRoad.shp",
      alsoExtract = "similar",
      url = "https://drive.google.com/file/d/1Z6ueq8yXtUPuPWoUcC7_l2p0_Uem34CC",
      # studyArea = smallSA,
      useCache = FALSE,
      fun = "sf::st_read",
      destinationPath = tmpdir,
      filename2 = "miniRoads"
    )
  )
  attr(roads1, "tags") <- NULL

  # There are floating point issues with 32 bit vs 64 bit approaches. The following fails:
  # expect_true(all.equal(roads[[1]], roads[[2]], check.attributes = FALSE))

  # diffs <- sum(abs(unlist(lapply(sf::st_geometry(roads[[1]]), as.numeric)) -
  #                    unlist(lapply(sf::st_geometry(roads[[2]]), as.numeric))))
  # expect_true(diffs < 0.0001)
  expect_true(terra::compareGeom(terra::rast(terra::ext(roads1)), terra::rast(terra::ext(smallSA))))
  expect_error(terra::compareGeom(terra::rast(terra::ext(roads2)), terra::rast(terra::ext(smallSA))))
  expect_true(terra::ext(roads2) > terra::ext(roads1))
})

# test_that("prepInputs doesn't work (part 3)", {
#   skip("The Google Drive url is dead")
#   if (interactive()) {
#     testInit(needGoogleDriveAuth = TRUE)
#
#     # Tati's reprex
#     # tmpdir <- "/mnt/d/temp/Cache"
#     wd <- checkPath(file.path(tmpdir, "reprex"), create = TRUE)
#     ranges <- prepInputs(
#       url = "https://drive.google.com/file/d/1AfGfRjaDsdq3JqcsidGRo3N66OUjRJnn",
#       destinationPath = wd,
#       fun = "terra::vect"
#     )
#     LCC05 <- prepInputs(
#       url = "https://drive.google.com/file/d/1g9jr0VrQxqxGjZ4ckF6ZkSMP-zuYzHQC",
#       targetFile = "LCC2005_V1_4a.tif",
#       studyArea = ranges,
#       fun = "terra::rast",
#       destinationPath = wd
#     )
#     sumNonNAs <- sum(!is.na(!LCC05[]))
#
#     # These are suitably vague that they will capture the mask if it gets it right
#     expect_true(sumNonNAs < 38000000)
#     expect_true(sumNonNAs > 37000000)
#   }
# })

################################################################################
#' The `spatialClasses` class
#'
#' This class is defunct.
#'
#' @aliases spatialClasses
#' @name spatialClasses-class
#' @rdname spatialClasses-class
#' @author Eliot McIntire

utils::globalVariables(c(
  "actualFile", "algorithm", "checksum", "checksum.x", "checksum.y",
  "filesize", "filesize.x", "filesize.y",
  "i.algorithm", "i.checksum", "i.filesize", "result"
))

################################################################################
#' Calculate checksum
#'
#' Verify (and optionally write) checksums.
#' Checksums are computed using [.digest()], which is simply a
#' wrapper around `digest::digest`.
#'
#' @note In version 1.2.0 and earlier, two checksums per file were required
#' because of differences in the checksum hash values on Windows and Unix-like
#' platforms. Recent versions use a different (faster) algorithm and only require
#' one checksum value per file.
#' To update your \file{CHECKSUMS.txt} files using the new algorithm, see
#' <https://github.com/PredictiveEcology/SpaDES/issues/295#issuecomment-246513405>.
#'
#' @param path    Character string giving the directory path containing `CHECKSUMS.txt`
#'                file, or where it will be written if `checksumFile = TRUE`.
#'
#' @param write   Logical indicating whether to overwrite `CHECKSUMS.txt`.
#'                Default is `FALSE`, as users should not change this file.
#'                Module developers should write this file prior to distributing
#'                their module code, and update accordingly when the data change.
#'
#' @param quickCheck Logical. If `TRUE`, then this will only use file sizes,
#'                   rather than a digest::digest hash. This is generally faster,
#'                   but will be *much* less robust.
#'
#' @param checksumFile The filename of the checksums file to read or write to.
#'                     The default is \file{CHECKSUMS.txt} located at
#'                     `file.path(path, module, "data", checksumFile)`.
#'                     It is likely not a good idea to change this, and should
#'                     only be used in cases such as `Cache`, which can
#'                     evaluate if the `checksumFile` has changed.
#'
#' @param files An optional character string or vector of specific files to checksum.
#'              This may be very important if there are many files listed in a
#'              `CHECKSUMS.txt` file, but only a few are to be checksummed.
#'
#' @param ...     Passed to [digest::digest()] and [utils::write.table()].
#'                For `digest`, the notable argument is `algo`. For `write.table`,
#'                the notable argument is `append`.
#'
#' @inheritParams Cache
#' @return A `data.table` with columns: `result`, `expectedFile`,
#'         `actualFile`, `checksum.x`, `checksum.y`,
#'         `algorithm.x`, `algorithm.y`, `filesize.x`, `filesize.y`
#'         indicating the result of comparison between local file (`x`) and
#'         expectation based on the `CHECKSUMS.txt` file.
#'
#' @author Alex Chubaty
#' @export
#' @rdname Checksums
#'
#' @examples
#' \dontrun{
#' modulePath <- file.path(tempdir(), "myModulePath")
#' dir.create(modulePath, recursive = TRUE, showWarnings = FALSE)
#' moduleName <- "myModule"
#' cat("hi", file = file.path(modulePath, moduleName)) # put something there for this example
#'
#' ## verify checksums of all data files
#' Checksums(modulePath, files = moduleName)
#'
#' ## write new CHECKSUMS.txt file
#' Checksums(files = moduleName, modulePath, write = TRUE)
#' }
#'
setGeneric("Checksums", function(path, write, quickCheck = getOption("reproducible.quickCheck", FALSE),
                                 checksumFile = identifyCHECKSUMStxtFile(path),
                                 files = NULL, verbose = getOption("reproducible.verbose", 1),
                                 ...) {
  standardGeneric("Checksums")
})

#' @importFrom data.table setnames
#' @importFrom methods formalArgs
#' @importFrom stats na.omit
#' @importFrom utils read.table write.table
#' @rdname Checksums
setMethod(
  "Checksums",
  signature = c(
    path = "character", quickCheck = "ANY",
    write = "logical", files = "ANY"
  ),
  definition = function(path, write, quickCheck = getOption("reproducible.quickCheck", FALSE),
                        checksumFile,
                        files, verbose = getOption("reproducible.verbose", 1), ...) {
    defaultHashAlgo <- "xxhash64"
    defaultWriteHashAlgo <- "xxhash64"
    dots <- list(...)
    dotsWriteTable <- dots[names(dots) %in% formalArgs(write.table)]
    dots <- dots[names(dots) %in% formalArgs(digest::digest)]
    checkPath(path, create = write)

    if (!file.exists(checksumFile)) {
      writeChecksumsTable(.emptyChecksumsFileContent, checksumFile, dotsWriteTable)
    }

    if (is.null(files)) {
      files <- list.files(path, full.names = TRUE)
      files <- grep(files,
        pattern = makeRelative(checksumFile, path),
        value = TRUE, invert = TRUE
      )
      # files <- fs::path_norm(files)
    } else {
      isAbs <- isAbsolutePath(files)
      if (any(!isAbs)) {
        files[!isAbs] <- makeAbsolute(files[!isAbs], path)
      }
    }

    txt <- if (file.size(checksumFile) == 0) {
      .emptyChecksumsFileContent
    } else {
      read.table(checksumFile,
        header = TRUE,
        stringsAsFactors = FALSE
      )
    }
    txt <- as.data.table(lapply(txt, as.character))
    set(txt, NULL, "file", makeRelative(txt$file, path))
    if (is.null(txt$filesize)) txt$filesize <- rep("", NROW(txt))
    txtRead <- txt # keep a copy even if writing
    if (!(!write && file.info(checksumFile)$size > 0)) {
      txt <- data.frame(
        file = character(0), checksum = character(0),
        filesize = character(0), stringsAsFactors = FALSE
      )
    }

    if (is.null(dots$algo)) {
      if (NROW(files)) {
        if (write) {
          dots$algo <- defaultWriteHashAlgo
        } else {
          dots$algo <- defaultHashAlgo
        }
      } else {
        dots$algo <- character()
      }
    }

    stStart <- Sys.time()
    filesToCheck <- if (length(txt$file) & length(files)) {
      inTxt <- makeRelative(files, path) %in% makeRelative(txt$file, path)
      if (isTRUE(any(inTxt)))
        files <- files[inTxt]
      else {
        # might fail because it is listed in inputPaths; check there
        possPath <- .getDestinationPathShared()
        # can be length > 1
        if (!is.null(possPath)) {
          possPath <- normPath(possPath)
          if (!identical(possPath, path)) {
            for (pp in possPath) {
              inTxt <- makeRelative(files, path) %in%
                makeRelative(txt$file, pp)
              if (isTRUE(any(inTxt))) {
                files <- files[inTxt]
                break
              }
            }
          }
        }
      }
      files
    } else {
      files
    }

    if (length(filesToCheck) != length(files[!endsWith(files, "similar")])) {
      # Could be a case of user passing file path that is not with subdirectories; offer help
      justByBasename <- basename(txt$file) %in% basename(files)
      if (sum(justByBasename) == length(files)) {
        messagePreProcess(
          "Files found in CHECKSUMS.txt that match by basename; using these.\n",
          "  User should specify all files (e.g., targetFile, alsoExtract, archive)\n",
          "  with subfolders specified."
        )
        filesToCheck <- unique(c(filesToCheck, makeAbsolute(txt$file[justByBasename], path)))
      }
    }

    filesToCheck <- filesToCheck[file.exists(filesToCheck)] # remove non existing files
    # filesToCheck <- filesToCheck[!dir.exists(filesToCheck)] # remove directories # need to keep directories b/c e.g., gdb files need directories

    if (!is.null(txt$algorithm)) {
      if (!write) {
        dots$algo <- unique(txt[txt$file %in% makeRelative(filesToCheck, path), ][["algorithm"]])
        dots$algo <- dots$algo[!is.na(dots$algo)][1]
        # dots$algo <- na.omit(dots$algo)[1]
        if (is.na(dots$algo)) dots$algo <- defaultWriteHashAlgo
      }
      # Multi-algorithm support: a file may have multiple rows in CHECKSUMS.txt,
      # one per algorithm (e.g. xxhash64 written by pp_finalize, md5 written by
      # pp_remote_hash_check from a remote md5Checksum). Restrict the comparison
      # to rows matching the algorithm we're about to digest with — otherwise
      # the join below produces N rows per file (one per recorded algorithm)
      # and the dedup `.SD[1, ], by = expectedFile` step risks keeping a
      # spurious FAIL row where a different algo's recorded hash didn't match
      # the freshly-digested hash.
      if (NROW(txt) && length(dots$algo) && nzchar(dots$algo)) {
        txt <- txt[is.na(txt$algorithm) | txt$algorithm == dots$algo, , drop = FALSE]
      }
    } else {
      if (NROW(txt)) {
        txt$algorithm <- defaultWriteHashAlgo
      } else {
        txt$algorithm <- character()
      }
      if (NROW(txtRead)) {
        txtRead$algorithm <- defaultWriteHashAlgo
      } else {
        txtRead$algorithm <- character()
      }
    }

    if (is.null(txt$filesize)) {
      quickCheck <- FALSE
      messagePreProcess("Not possible to use quickCheck;\n ",
        "    CHECKSUMS.txt file does not have filesizes",
        sep = "", verbose = verbose
      )
    }
    checksums <- rep(list(rep("", length(filesToCheck))), 2)
    dirs <- dir.exists(filesToCheck)
    filesToCheckWODirs <- filesToCheck[!dirs]
      if (quickCheck | write) {
        checksums[[2]][!dirs] <- do.call(.digest,
                                         args = append(
          list(file = filesToCheckWODirs, quickCheck = TRUE),
          dots
        )
      )
    }

    if (!quickCheck | write) {
      checksums[[1]][!dirs] <- do.call(.digest,
        args = append(
          list(file = filesToCheckWODirs, quickCheck = FALSE),
          dots
        )
      )
    }
    if (any(dirs)) {
      checksums[[1]][dirs] <- "dir"
      checksums[[2]][dirs] <- 0
    }

    verboseTmp <- difftime(Sys.time(), stStart) > 8
      messagePreProcess("Finished checking local files.", sep = "", verbose = verbose - 1 + verboseTmp)

    filesToCheckRel <- makeRelative(filesToCheck, path)
    out <- if (length(filesToCheck)) {
      data.table(
        file = filesToCheckRel, checksum = checksums[[1]],
        filesize = checksums[[2]], algorithm = dots$algo, stringsAsFactors = FALSE
      )
    } else {
      data.table(
        file = character(0), checksum = character(0), filesize = character(0),
        algorithm = character(0), stringsAsFactors = FALSE
      )
    }

    out1 <- data.table::copy(out)
    if (write) {
      writeChecksumsTable(out1, checksumFile, dotsWriteTable)
      txt <- txtRead
      txt <- txt[out, on = colnames(out)]
      # txt1Old <- dplyr::right_join(txt1Old, out)
    }
    txt1 <- data.table::copy(txt)

    out[, actualFile := file]
    if (write) {
      out <- txt[out, on = "file"]
    } else {
      out <- out[txt, on = "file"]
    }
    setnames(out, "file", "expectedFile")
    if (quickCheck) {
      out[, result := ifelse(filesize != i.filesize, "FAIL", "OK")]
    } else {
      out[, result := ifelse(checksum != i.checksum, "FAIL", "OK")]
    }
    data.table::setorderv(out, "result", order = -1L, na.last = TRUE)
    out <- out[, .SD[1, ], by = "expectedFile"]
    out <- checksumsDirsOk(out)

    results.df <- out[, list(
      "result" = result,
      "expectedFile" = expectedFile,
      "actualFile" = actualFile,
      "checksum.x" = i.checksum,
      "checksum.y" = checksum,
      "algorithm.x" = i.algorithm,
      "algorithm.y" = algorithm,
      "filesize.x" = i.filesize,
      "filesize.y" = filesize
    )]

    return(invisible(results.df))
  }
)

#' @rdname Checksums
setMethod(
  "Checksums",
  signature = c(
    path = "character", quickCheck = "ANY",
    write = "missing", files = "ANY"
  ),
  definition = function(path, quickCheck = getOption("reproducible.quickCheck", FALSE), checksumFile,
                        files, verbose, ...) {
    Checksums(path,
      write = FALSE, quickCheck = quickCheck, checksumFile = checksumFile,
      files = files, verbose = verbose, ...
    )
  }
)

#' @keywords internal
writeChecksumsTable <- function(out, checksumFile, dots) {
  out <- out[.orderDotsUnderscoreFirst(out$file), ] ## sort by filename alphabetically

  do.call(write.table,
    args = append(
      list(
        x = out, file = checksumFile, eol = "\n",
        col.names = !isTRUE(dots$append),
        row.names = FALSE
      ),
      dots
    )
  )
}

# Upsert a single (file, algorithm) row into a CHECKSUMS.txt file using a
# precomputed hash. Used by pp_remote_hash_check after a confirmed remote
# match to record an md5/sha1/sha256 row without re-digesting via Checksums().
# Existing rows for OTHER algorithms recording the same file are preserved.
.upsertChecksumsRow <- function(checkSumFilePath, file, hash, filesize,
                                algorithm) {
  newRow <- data.table::data.table(
    file = as.character(file),
    checksum = as.character(hash),
    filesize = as.character(filesize),
    algorithm = as.character(algorithm)
  )

  existing <- if (file.exists(checkSumFilePath) &&
                    file.size(checkSumFilePath) > 0L) {
    cs <- try(read.table(checkSumFilePath, header = TRUE,
                          stringsAsFactors = FALSE), silent = TRUE)
    if (is(cs, "try-error")) NULL else data.table::as.data.table(cs)
  } else NULL

  combined <- if (is.null(existing) || !NROW(existing)) {
    newRow
  } else {
    if (is.null(existing$algorithm)) existing[, algorithm := NA_character_]
    if (is.null(existing$filesize)) existing[, filesize := NA_character_]
    existing[, file := as.character(file)]
    existing[, checksum := as.character(checksum)]
    existing[, filesize := as.character(filesize)]
    existing[, algorithm := as.character(algorithm)]

    keep <- !(existing$file %in% newRow$file &
                existing$algorithm %in% newRow$algorithm)
    rbindlist(list(existing[keep, ], newRow), fill = TRUE)
  }

  writeChecksumsTable(as.data.frame(combined), checkSumFilePath, dots = list())
  invisible(combined)
}

#' Calculate the hashes of multiple files
#'
#' Internal function. Wrapper for [digest::digest()] using `algo = xxhash64`.
#'
#' @param file  Character vector of file paths.
#' @param quickCheck Logical indicating whether to use a fast file size check as a heuristic
#'                   for determining changes to a file.
#' @param ...   Additional arguments to [digest::digest()].
#'
#' @return A character vector of hashes.
#'
#' @author Alex Chubaty
#' @importFrom digest digest
#' @keywords internal
#' @rdname digest
setGeneric(".digest", function(file, quickCheck = FALSE, ...) {
  standardGeneric(".digest")
})

#' @rdname digest
setMethod(
  ".digest",
  signature = c(file = "character"),
  definition = function(file, quickCheck = FALSE, algo = "xxhash64", ...) {
    fss <- file.size(file)
    if (quickCheck) {
      as.character(fss) # need as.character for empty case
    } else {
      # fss <- file.size(file)
      evalThis <- quote(as.character(
        unname(
          unlist(
            lapply(file, function(f) {
              digest::digest(object = f, file = TRUE, algo = algo, ...)
            })
          )
        )
      )) # need as.character for empty case # nolint
      nonZeroSize <- fss != 0

      if (isTRUE(any(nonZeroSize))) {
        # some zero length files fail digest::digest, others don't. Don't know the exact reason yet.
        # BUT don't use `try` for all/any digests as it is too slow to run all/any time
        fss2 <- try(eval(evalThis), silent = TRUE)
        if (is(fss2, "try-error")) {
          browser()
          fss2 <- character(length(file))
          file <- file[nonZeroSize]
          fss2[nonZeroSize] <- eval(evalThis)
        }
      } else {
        fss2 <- eval(evalThis)
      }

      fss2
    }
  }
)


checksumsDirsOk <- function(out) {
  cscols <- "checksum.x|i.checksum"
  if (any(grepl(cscols, colnames(out)))) {
    cscol <- grep(cscols, colnames(out), value = TRUE)[1]
    dirsHave <- unique(dirname(out[!get(cscol) %in% "dir" & result == "OK"]$expectedFile))
    dirsHave <- grep("\\.", dirsHave, value = TRUE, invert = TRUE)
    if (length(dirsHave)) {
      out[get(cscol) %in% "dir" & expectedFile %in% dirsHave, result := "OK"]
    }
  }
  out
}

#' @exportClass Path
#' @rdname Path-class
setClass("Path", slots = c(.Data = "character"), contains = "character", prototype = NA_character_)

#' Coerce a character string to a class "Path"
#'
#' Allows a user to specify that their character string is indeed a filepath.
#' Thus, methods that require only a filepath can be dispatched correctly.
#'
#' It is often difficult or impossible to know algorithmically whether a
#' character string corresponds to a valid filepath.
#' In the case where it is en existing file, `file.exists` can work.
#' But if it does not yet exist, e.g., for a `save`, it is difficult to know
#' whether it is a valid path before attempting to save to the path.
#'
#' This function can be used to remove any ambiguity about whether a character
#' string is a path. It is primarily useful for achieving repeatability with Caching.
#' Essentially, when Caching, arguments that are character strings should generally be
#' digested verbatim, i.e., it must be an exact copy for the Cache mechanism
#' to detect a candidate for recovery from the cache.
#' Paths, are different. While they are character strings, there are many ways to
#' write the same path. Examples of identical meaning, but different character strings are:
#' path expanding of `~` vs. not, double back slash vs. single forward slash,
#' relative path vs. absolute path.
#' All of these should be assessed for their actual file or directory location,
#' NOT their character string. By converting all character string that are actual
#' file or directory paths with this function, then `Cache` will correctly assess
#' the location, NOT the character string representation.
#'
#' @param obj A character string to convert to a `Path`.
#' @param nParentDirs A numeric indicating the number of parent directories starting
#'                    from basename(obj) = 0 to keep for the digest
#'
#' @export
#' @rdname Path-class
#' @return
#' A vector of class `Path`, which is similar to a character, but
#' has an attribute indicating how deep the Path should be
#' considered "digestible". In other words, most of the time, only some
#' component of an absolute path is relevant for evaluating its purpose in
#' a Cache situation. In general, this is usually equivalent to just the "relative" path
#'
#' @examples
#' tmpf <- tempfile(fileext = ".csv")
#' file.exists(tmpf) ## FALSE
#' tmpfPath <- asPath(tmpf)
#' is(tmpf, "Path") ## FALSE
#' is(tmpfPath, "Path") ## TRUE
#'
asPath <- function(obj, nParentDirs = 0) {
  UseMethod("asPath", obj)
}

#' @export
#' @importFrom methods is
#' @rdname Path-class
asPath.character <- function(obj, nParentDirs = 0) { # nolint
  class(obj) <- unique(c("Path", class(obj)))
  attr(obj, "nParentDirs") <- nParentDirs
  return(obj)
}

#' @export
#' @rdname Path-class
asPath.null <- function(obj, nParentDirs = 0) { # nolint
  return(NULL)
}

#' If using `as("string", "Path")`, there is no option to pass `nParentDirs`.
#' So, using `asPath` directly (e.g., `asPath("string", 0))`) is preferred.
#' @export
#' @importFrom methods new
#' @rdname Path-class
#' @name asPath
setAs(from = "character", to = "Path", function(from) {
  asPath(from, 0)
})


#' Copy a file using `robocopy` on Windows and `rsync` on Linux/macOS
#'
#' This is replacement for `file.copy`, but for one file at a time.
#' The additional feature is that it will use `robocopy` (on Windows) or
#' `rsync` on Linux or Mac, if they exist.
#' It will default back to `file.copy` if none of these exists.
#' If there is a possibility that the file already exists, then this function
#' should be very fast as it will do "update only", i.e., nothing.
#'
#' @param from The source file.
#'
#' @param to The new file.
#'
#' @param useRobocopy For Windows, this will use a system call to `robocopy`
#'        which appears to be much faster than the internal `file.copy` function.
#'        Uses `/MIR` flag. Default `TRUE`.
#'
#' @param overwrite Passed to `file.copy`
#'
#' @param delDestination Logical, whether the destination should have any files deleted,
#' if they don't exist in the source. This is `/purge` for robocopy and --delete for
#' rsync.
#'
#' @param create Passed to `checkPath`.
#'
#' @param silent Should a progress be printed.
#'
#' @author Eliot McIntire and Alex Chubaty
#' @rdname copyFile
#' @return
#' This function is called for its side effect, i.e., a file is copied `from` to `to`.
#'
#' @examples
#' tmpDirFrom <- file.path(tempdir(), "example_fileCopy_from")
#' tmpDirTo <- file.path(tempdir(), "example_fileCopy_to")
#' tmpFile1 <- tempfile("file1", tmpDirFrom, ".csv")
#' tmpFile2 <- tempfile("file2", tmpDirFrom, ".csv")
#' dir.create(tmpDirFrom, recursive = TRUE, showWarnings = FALSE)
#' dir.create(tmpDirTo, recursive = TRUE, showWarnings = FALSE)
#' f1 <- normalizePath(tmpFile1, mustWork = FALSE)
#' f2 <- normalizePath(tmpFile2, mustWork = FALSE)
#' t1 <- normalizePath(file.path(tmpDirTo, basename(tmpFile1)), mustWork = FALSE)
#' t2 <- normalizePath(file.path(tmpDirTo, basename(tmpFile2)), mustWork = FALSE)
#'
#' write.csv(data.frame(a = 1:10, b = runif(10), c = letters[1:10]), f1)
#' write.csv(data.frame(c = 11:20, d = runif(10), e = letters[11:20]), f2)
#' copyFile(c(f1, f2), c(t1, t2))
#' file.exists(t1) ## TRUE
#' file.exists(t2) ## TRUE
#' identical(read.csv(f1), read.csv(f2)) ## FALSE
#' identical(read.csv(f1), read.csv(t1)) ## TRUE
#' identical(read.csv(f2), read.csv(t2)) ## TRUE
#'
copySingleFile <- function(from = NULL, to = NULL, useRobocopy = TRUE,
                           overwrite = TRUE, delDestination = FALSE,
                           create = TRUE, silent = FALSE) {
  if (any(length(from) != 1, length(to) != 1)) stop("from and to must each be length 1")
  useFileCopy <- identical(dirname(from), dirname(to))

  lapply(unique(dirname(to)), checkPath, create = create)

  os <- tolower(Sys.info()[["sysname"]])
  .onLinux <- .Platform$OS.type == "unix" && unname(os) == "linux"
  if (!useFileCopy) {
    if (isWindows() && isTRUE(file.size(from) > 1e6)) {
      if (!isTRUE(unique(dir.exists(to)))) toDir <- dirname(to) # extract just the directory part
      robocopyBin <- tryCatch(Sys.which("robocopy"), warning = function(w) NA_character_)

      robocopy <- if (silent) {
        paste0(
          robocopyBin, " /purge"[delDestination], " /ETA /XJ /XO /NDL /NFL /NJH /NJS \"", # nolint
          normalizePath(dirname(from), mustWork = TRUE, winslash = "\\"), "\" \"",
          normalizePath(toDir, mustWork = FALSE, winslash = "\\"), "\" ",
          basename(from)
        )
      } else {
        paste0(
          robocopyBin, " /purge"[delDestination], " /ETA /XJ /XO \"", # nolint
          normalizePath(dirname(from), mustWork = TRUE, winslash = "\\"), "\" \"",
          normalizePath(toDir, mustWork = FALSE, winslash = "\\"), "\" ",
          basename(from)
        )
      }

      useFileCopy <- if (useRobocopy && !is.na(robocopyBin)) {
        suppressWarnings(tryCatch(system(robocopy, intern = TRUE), error = function(x) TRUE))
      } else {
        TRUE
      }
      if (isTRUE(any(grepl("ERROR", useFileCopy)))) {
        useFileCopy <- TRUE
      }

      # means that the file didn't get copied because it is actually the same directory
      if (any(!nzchar(useFileCopy))) {
        useFileCopy <- TRUE
      }
    } else if ((.onLinux)) { # nolint
      if (!identical(basename(from), basename(to))) {
        # rsync can't handle file renaming on copy
        useFileCopy <- TRUE
      } else {
        if (!dir.exists(to)) toDir <- dirname(to) # extract just the directory part
        rsyncBin <- tryCatch(Sys.which("rsync"), warning = function(w) NA_character_)
        opts <- if (silent) " -a " else " -avP "
        # rsync command can't handle spaces file/dir names -- must protect them
        toDir <- normalizePath(toDir, mustWork = FALSE)
        toDir <- gsub("\ ", "\\ ", toDir, fixed = TRUE)
        if (!dir.exists(toDir)) dir.create(toDir, recursive = TRUE, showWarnings = FALSE)
        from <- normalizePath(from, mustWork = TRUE)
        rsync <- paste0(
          rsyncBin, " ", opts, " --delete "[delDestination],
          "'", from, "' ", toDir, "/"
        )

        # THe warnings here are being caught; and use file.copy
        useFileCopy <- capture.output(system(rsync, intern = FALSE, ignore.stderr = TRUE, ignore.stdout = TRUE))
        filesCopied <- file.exists(file.path(toDir, basename(from)))
        useFileCopy <- any(!filesCopied)
      }
    } else {
      useFileCopy <- TRUE
    }
  }
  if (isTRUE(useFileCopy)) {
    lapply(unique(dirname(to)), checkPath, create = create)
    file.copy(from = from, to = to, overwrite = overwrite, recursive = FALSE)
  }

  return(invisible(to))
}

#' @export
#' @rdname copyFile
copyFile <- Vectorize(copySingleFile, vectorize.args = c("from", "to"))

#' @importFrom methods slotNames
#' @importFrom digest digest
.digestRasterLayer <- function(object, length, algo, quick) {
  .requireNamespace("raster", stopOnFALSE = TRUE)

  isRasterStack <- is(object, "RasterStack")
  if (!isRasterStack) {
    objList <- list(object)
  } else {
    objList <- object@layers
  }
  dig <- lapply(objList, function(object) {
    sn <- slotNames(object@data)
    sn <- sn[!(sn %in% c( # "min", "max", "haveminmax", "names", "isfactor",
      "dropped", "nlayers", "fromdisk", "inmemory"
      # "offset", "gain"
    ))]
    dataSlotsToDigest <- lapply(sn, function(s) slot(object@data, s))
    names(dataSlotsToDigest) <- sn
    theData <- .robustDigest(object@data@values)
    dataSlotsToDigest$values <- NULL
    dig <- .robustDigest(
      append(list(
        dim(object), raster::res(object), terra::crs(object),
        raster::extent(object), theData
      ), dataSlotsToDigest),
      length = length, quick = quick,
      algo = algo
    ) # don't include object@data -- these are volatile

    # Legend
    sn <- slotNames(object@legend)
    legendSlotsToDigest <- lapply(sn, function(s) slot(object@legend, s))
    dig2 <- .robustDigest(legendSlotsToDigest,
                          length = length, quick = quick,
                          algo = algo
    ) # don't include object@data -- these are volatile
    dig <- c(dig, dig2)

    sn <- slotNames(object@file)
    sn <- sn[!(sn %in% c("name"))]
    fileSlotsToDigest <- lapply(sn, function(s) slot(object@file, s))
    digFile <- .robustDigest(asPath(fileSlotsToDigest),
                             length = length, quick = quick,
                             algo = algo
    ) # don't include object@file -- these are volatile

    dig <- c(dig, digFile)
  })

  fns <- Filenames(object, allowMultiple = FALSE)
  if (length(fns[nchar(fns) > 0])) {
    # if the Raster is on disk, has the first length characters;
    isGrd <- endsWith(basename(fns), suffix = ".grd")
    if (isTRUE(any(isGrd))) {
      fns[isGrd] <- sub(fns[isGrd], pattern = ".grd$", replacement = ".gri")
    }
    # # there is no good reason to use depth = 0, 1, or 2 or more -- but I think 2 is *more* reliable
    dig2 <- .robustDigest(asPath(fns, 2), length = length, quick = quick, algo = algo)
    dig <- c(dig, unname(dig2))
  }

  dig <- .robustDigest(unlist(dig), length = length, quick = quick, algo = algo)
  dig
}

#' @details
#' `.sortDotsUnderscoreFirst`: This exists so Windows, Linux, and Mac machines can have
#' the same order after a sort. It will put dots and underscores first
#' (with the sort key based on their second character, see examples.
#' It also sorts lower case before upper case.
#'
#' @param obj  An arbitrary R object for which a `names` function
#'              returns a character vector.
#'
#' @return `.sortDotsUnderscoreFirst`: the same object as `obj`,
#'   but sorted with dots and underscores first,
#'   lower case before upper case.
#'
#' @author Eliot McIntire
#' @export
#' @rdname exportedMethods
#'
#' @examples
#' items <- c(A = "a", Z = "z", `.D` = ".d", `_C` = "_C")
#' .sortDotsUnderscoreFirst(items)
#'
#' # dots & underscore (using 2nd character), then all lower then all upper
#' items <- c(B = "Upper", b = "lower", A = "a", `.D` = ".d", `_C` = "_C")
#' .sortDotsUnderscoreFirst(items)
#'
#' # with a vector
#' .sortDotsUnderscoreFirst(c(".C", "_B", "A")) # _B is first
#'
.sortDotsUnderscoreFirst <- function(obj) {
  obj[.orderDotsUnderscoreFirst(obj)]
}

#' @export
#' @rdname exportedMethods
.orderDotsUnderscoreFirst <- function(obj) {
  if (!is.null(names(obj))) {
    namesObj <- names(obj)
  } else {
    namesObj <- obj
  }

  if (is.character(namesObj)) {
    namesObj <- gsub(namesObj, pattern = "\\.|_", replacement = "aa")
    allLower <- (tolower(namesObj) == namesObj) %in% TRUE
    namesObj[allLower] <- paste0("abALLLOWER", namesObj[allLower])

    onesChanged <- startsWith(namesObj, prefix = "a") %in% TRUE
    namesObj[!onesChanged] <- paste0("ZZZZZZZZZ", namesObj[!onesChanged])

    out <- order(namesObj)
  } else {
    out <- seq_along(obj)
  }
  out
}

#' Attach debug info to return for Cache
#'
#' Internal use only. Attaches an attribute to the output, usable for
#' debugging the Cache.
#'
#' @param obj  An arbitrary R object.
#' @param preDigest  A list of hashes.
#' @param ...  Dots passed from `Cache`
#' @param fullCall The original call to `Cache`
#'
#' @return The same object as `obj`, but with 2 attributes set.
#'
#' @author Eliot McIntire
#' @rdname debugCache
.debugCache <- function(obj, preDigest, ..., fullCall) {
  if (missing(fullCall))
    attr(obj, "debugCache1") <- list(...)
  else
    attr(obj, "debugCache1") <- fullCall
  attr(obj, "debugCache2") <- preDigest
  obj
}

#' @keywords internal

#' @keywords internal
nextNumericName <- function(string) {
  theExt <- fileExt(string)
  saveFilenameSansExt <- filePathSansExt(string)
  finalNumericPattern <- "_[[:digit:]]+$"
  allSimilarFilesInDir <- dir(dirname(saveFilenameSansExt), pattern = basename(saveFilenameSansExt))
  allSimilarFilesInDirSansExt <- if (length(allSimilarFilesInDir) == 0) {
    unique(saveFilenameSansExt)
  } else {
    unique(filePathSansExt(allSimilarFilesInDir))
  }
  alreadyHasNumeric <- grepl(allSimilarFilesInDirSansExt, pattern = finalNumericPattern)
  if (isTRUE(any(alreadyHasNumeric))) {
    splits <- strsplit(allSimilarFilesInDirSansExt[alreadyHasNumeric], split = "_")
    highestNumber <- max(unlist(lapply(splits, function(split) as.numeric(tail(split, 1)))),
                         na.rm = TRUE
    )
    preNumeric <- unique(unlist(lapply(splits, function(spl) paste(spl[-length(spl)], collapse = "_")))) # nolint
    ## keep rndstr in here (below), so that both streams keep same rnd number state
    out <- file.path(dirname(saveFilenameSansExt), paste0(preNumeric, "_", highestNumber + 1))
  } else {
    out <- paste0(saveFilenameSansExt, "_1")
  }

  paste0(out, ".", theExt)
}

list2envAttempts <- function(x, envir) {
  attempt <- try(list2env(x, envir), silent = TRUE) # must be try; i.e., must always be `try`
  output <- NULL
  if (is(attempt, "try-error")) {
    attempt <- try(list2env(x, envir@.xData), silent = TRUE)
    if (is(attempt, "try-error")) {
      output <- as.environment(x)
    }
  }
  output
}

#' Copy the file-backing of a file-backed Raster* object
#'
#' Rasters are sometimes file-based, so the normal save and copy and assign
#' mechanisms in R don't work for saving, copying and assigning.
#' This function creates an explicit file copy of the file that is backing the raster,
#' and changes the pointer (i.e., `filename(object)`) so that it is pointing
#' to the new file.
#'
#' @param obj The raster object to save to the repository.
#'
#' @param repoDir Character denoting an existing directory in which an artifact will be saved.
#'
#' @param overwrite Logical. Should the raster be saved to disk, overwriting existing file.
#'
#' @param ... Not used
#'
#' @return A raster object and its newly located file backing.
#'         Note that if this is a legitimate Cache repository, the new location
#'         will be a subdirectory called \file{rasters/} of \file{repoDir/}.
#'         If this is not a repository, the new location will be within `repoDir`.
#'
#' @author Eliot McIntire
#' @importFrom digest digest
#' @importFrom methods is selectMethod slot slot<-
#' @inheritParams Cache
#' @rdname prepareFileBackedRaster
.prepareFileBackedRaster <- function(obj, repoDir = NULL, overwrite = FALSE,
                                     drv = getDrv(getOption("reproducible.drv", NULL)),
                                     conn = getOption("reproducible.conn", NULL),
                                     verbose = getOption("reproducible.verbose"),
                                     ...) {
  fnsAll <- Filenames(obj)
  fnsShort <- Filenames(obj, FALSE)
  if (!all(nchar(fnsAll) == 0)) {
    repoDir <- checkPath(repoDir, create = TRUE)
    isRepo <- CacheIsACache(cachePath = repoDir, drv = drv, conn = conn, verbose = verbose)
    # thoseWithGRI <- endsWith(fnsAll, "gri")
    fns <- fnsAll
    FB <- nchar(fns) > 0
    ########################
    if (any(!file.exists(fns[FB]))) {
      FBshort <- nchar(fnsShort) > 0
      fnsOnly <- fnsShort[FBshort]
      badFileNames <- fnsOnly[!file.exists(fnsOnly)]

      trySaveFilename <- badFileNames
      if (any(grepl(basename(repoDir), badFileNames))) {
        # File is in wrong folder, usually the result of a copy of cache between 2 machines
        splittedFilenames <- strsplit(badFileNames, split = basename(repoDir))
        trySaveFilename <- if (length(splittedFilenames) == 1) {
          normalizePath(
            file.path(repoDir, splittedFilenames[[1]][[length(splittedFilenames[[1]])]]),
            winslash = "/", mustWork = FALSE
          )
        } else {
          splittedFilenames2 <- lapply(splittedFilenames, function(x) {
            ifelse(length(x), x[length(x)], "")
          })
          normalizePath(file.path(repoDir, splittedFilenames2), winslash = "/", mustWork = FALSE)
        }
      }
      if (any(!file.exists(trySaveFilename))) {
        stop(
          "The following file-backed rasters are supposed to be on disk ",
          "but appear to have been deleted:\n",
          paste("    ", badFileNames, collapse = "\n"),
          ". The most likely reason is that two functions had the same output ",
          "and one of them was removed with clearCache(...). ",
          "The best solution to this is never have two functions create the same ",
          "file-backed raster."
        )
      } else {
        obj <- updateFilenameSlots(obj, curFilenames = fnsOnly, trySaveFilename)
        fnsAll <- fns <- Filenames(obj)
      }
    }
    #################
    saveFilename <- fns
    bn <- basename(fns)
    bnFB <- bn[FB]
    saveFilename[FB] <- if (isRepo) {
      file.path(repoDir, "rasters"[isRepo], bnFB)
    } else {
      file.path(repoDir, bnFB)
    }
    dirForNewFiles <- unique(dirname(saveFilename[FB]))
    checkPath(dirForNewFiles, create = TRUE)
    saveFilename <- normPath(saveFilename)
    saveFilenamePreNumeric <- saveFilename
    exist <- file.exists(saveFilename)
    if (any(exist)) {
      saveFilename[exist] <- unlist(lapply(saveFilename[exist], nextNumericName))
    }
    FBAll <- nchar(fnsAll) > 0
    out <- hardLinkOrCopy(from = fnsAll[FBAll], to = saveFilename[FBAll])
    saveFilenamesToUpdateSlot <- saveFilename[basename(saveFilenamePreNumeric) %in%
      basename(fnsShort)]
    obj <- updateFilenameSlots(obj, fnsShort, saveFilenamesToUpdateSlot)
  }
  obj
}

withoutFinalNumeric <- function(string) {
  ext <- fileExt(string)
  string1 <- filePathSansExt(string)
  woNumeric <- gsub("^(.+)\\_[[:digit:]]+$", "\\1", string1)
  paste0(woNumeric, ".", ext)
}

setClass("PackedSpatExtent2")

wrapSpatVector <- function(obj) {

  # stopifnot(inherits(obj, "SpatVector"))
  if (FALSE) # old; stable but slow
    stWrap <- system.time(obj2 <- terra::wrap(obj))


  # stWrapSpecial <- system.time(
  obj <- list(
    geometry = terra::geom(obj),                      # matrix of coordinates
    attributes = as.data.frame(obj),           # attribute table
    crs = terra::crs(obj),                             # coordinate reference system
    extent = .wrap(terra::ext(obj)),                          # bounding box
    geom_type = terra::geomtype(obj),                  # "points", "lines", "polygons", etc.
    n_features = terra::nrow(obj),
    n_fields = terra::ncol(obj)
  )
  # )
  attr(obj, "class") <- c("PackedSpatVector2", attr(obj, "class"))

  # ss <- sample(1e8, 1)
  # env <- reproducible:::memoiseEnv(cachePath = getOption("reproducible.cachePath"))
  # if (!exists("wrapTimings", envir = env))
  #   env$wrapTimings <- new.env(parent = emptyenv())
  # assign(paste0("ss", ss), list(stWrap = stWrap, stWrapSpecial = stWrapSpecial), envir = env$wrapTimings)

  obj
}

unwrapSpatVector <- function(obj) {
  if (NROW(obj$geometry) == 0)
    obj$geom_type <- "points"
  sv <- terra::vect(obj$geometry, type = obj$geom_type, crs = obj$crs)

  # Attach attributes
  terra::values(sv) <- obj$attributes

  # obj <- terra::unwrap(obj)
  # obj$x <- cbind(obj$x$cols125[, 1:2, drop = FALSE], obj$x$cols34[, 1:2, drop = FALSE], obj$x$cols125[, 3, drop = FALSE])
  # do.call(terra::vect, obj)
  # obj
  sv

}

#' Has a cached object has been updated?
#'
#' @param x cached object
#'
#' @return logical
#'
#' @export
isUpdated <- function(x) {
  cond1 <- isTRUE(attr(x, ".Cache")[["newCache"]])
  cond2 <- length(attr(x, ".Cache")[["changed"]]) > 0

  cond1 || cond2
}

utils::globalVariables(c(
  "..onCol"
))

#' @param x A simList or a directory containing a valid Cache repository. Note:
#'   For compatibility with `Cache` argument, `cachePath` can also be
#'   used instead of `x`, though `x` will take precedence.
#' @param after A time (POSIX, character understandable by data.table).
#'                  Objects cached after this time will be shown or deleted.
#' @param before A time (POSIX, character understandable by data.table).
#'                   Objects cached before this time will be shown or deleted.
#' @param fun An optional character vector describing the function name to extract.
#'   Only functions with this/these functions will be returned.
#' @param cacheId An optional character vector describing the `cacheId`s to extract.
#'   Only entries with this/these `cacheId`s will be returned. If `useDBI(FALSE)`,
#'   this will also be dramatically faster than using `userTags`, for a large
#'   cache.
#'
#' @param ask Logical. If `FALSE`, then it will not ask to confirm deletions using
#'            `clearCache` or `keepCache`. Default is `TRUE`
#' @param ... Other arguments. Can be in the form of `tagKey = tagValue`, such as,
#'            `class = "numeric"` to find all entries that are numerics in the cache.
#'            Note: the special cases of `cacheId` and `fun` have their own
#'            named arguments in these functions.
#'            Also can be `regexp = xx`, where `xx` is `TRUE` if the user
#'            is passing a regular expression.
#'            Otherwise, `userTags` will need to be exact matches. Default is
#'            missing, which is the same as `TRUE`. If there are errors due
#'            to regular expression problem, try `FALSE`. For `cc`, it is
#'            passed to `clearCache`, e.g., `ask`, `userTags`. For `showCache`,
#'            it can also be `sorted = FALSE` to return the object unsorted.
#' @param userTags Character vector. If used, this will be used in place of the
#'                 `after` and `before`.
#'                 Specifying one or more `userTag` here will clear all
#'                 objects that match those tags.
#'                 Matching is via regular expression, meaning partial matches
#'                 will work unless strict beginning (`^`) and end (`$`) of string
#'                 characters are used.
#'                 Matching will be against any of the 3 columns returned by `showCache()`,
#'                 i.e., `artifact`, `tagValue` or `tagName`.
#'                 Also, if `length(userTags) > 1`, then matching is by `and`.
#'                 For `or` matching, use `|` in a single character string.
#'                 See examples.
#' @param useCloud Logical. If `TRUE`, then every object that is deleted locally will
#'    also be deleted in the `cloudFolderID`, if it is non-`NULL`
#'
#' @inheritParams Cache
#'
#' @details
#' If neither `after` or `before` are provided, nor `userTags`,
#' then all objects will be removed.
#' If both `after` and `before` are specified, then all objects between
#' `after` and `before` will be deleted.
#' If `userTags` is used, this will override `after` or `before`.
#'
#' @return Will clear all objects (or those that match `userTags`, or those
#' between `after` or `before`) from the repository located in
#' `cachePath`.
#' Invisibly returns a `data.table` of the removed items.
#'
#' @note If the cache is larger than 10MB, and clearCache is used, there will be
#' a message and a pause, if interactive, to prevent accidentally deleting of a
#' large cache repository.
#'
#' @export
#' @importFrom data.table setindex
#' @importFrom methods setGeneric setMethod
#' @importFrom utils object.size
#' @name showCache
#' @aliases clearCache
#' @rdname viewCache
#'
#' @examples
#' data.table::setDTthreads(2)
#'
#' tmpDir <- file.path(tempdir(), "reproducible_examples", "Cache")
#' try(clearCache(tmpDir, ask = FALSE), silent = TRUE) # just to make sure it is clear
#'
#' # Basic use
#' ranNumsA <- Cache(rnorm, 10, 16, cachePath = tmpDir)
#'
#' # All same
#' ranNumsB <- Cache(rnorm, 10, 16, cachePath = tmpDir) # recovers cached copy
#' ranNumsD <- Cache(quote(rnorm(n = 10, 16)), cachePath = tmpDir) # recovers cached copy
#'
#' # Any minor change makes it different
#' ranNumsE <- Cache(rnorm, 10, 6, cachePath = tmpDir) # different
#'
#' ## Example 1: basic cache use with tags
#' ranNumsA <- Cache(rnorm, 4, cachePath = tmpDir, userTags = "objectName:a")
#' ranNumsB <- Cache(runif, 4, cachePath = tmpDir, userTags = "objectName:b")
#' ranNumsC <- Cache(runif, 40, cachePath = tmpDir, userTags = "objectName:b")
#'
#' showCache(tmpDir, userTags = c("objectName"))
#' showCache(tmpDir, userTags = c("^a$")) # regular expression ... "a" exactly
#'
#' # Fine control of cache elements -- pick out only the large runif object, and remove it
#' cache1 <- showCache(tmpDir, userTags = c("runif")) # show only cached objects made during runif
#' toRemove <- cache1[tagKey == "object.size"][as.numeric(tagValue) > 700]$cacheId
#' clearCache(tmpDir, userTags = toRemove, ask = FALSE)
#' cacheAfter <- showCache(tmpDir, userTags = c("runif")) # Only the small one is left
#'
setGeneric("clearCache", function(x, userTags = character(), after = NULL, before = NULL,
                                  fun = NULL, cacheId = NULL,
                                  ask = getOption("reproducible.ask"),
                                  useCloud = FALSE,
                                  cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
                                  drv = getDrv(getOption("reproducible.drv", NULL)),
                                  conn = getOption("reproducible.conn", NULL),
                                  verbose = getOption("reproducible.verbose"),
                                  ...) {
  standardGeneric("clearCache")
})

#' @export
#' @rdname viewCache
setMethod(
  "clearCache",
  definition = function(x, userTags, after = NULL, before = NULL,
                        fun = NULL, cacheId = NULL,
                        ask, useCloud = FALSE,
                        cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
                        drv = getDrv(getOption("reproducible.drv", NULL)),
                        conn = getOption("reproducible.conn", NULL),
                        verbose = getOption("reproducible.verbose"),
                        ...) {
    # isn't clearing the raster backed file
    if (missing(x)) {
      x <- if (!is.null(list(...)$cachePath)) {
        messageCache("x not specified, but cachePath is; using ", list(...)$cachePath, verbose = verbose)
        list(...)$cachePath
      } else {
        messageCache("x not specified; using ", getOption("reproducible.cachePath")[1], verbose = verbose)
        x <- getOption("reproducible.cachePath")[1]
      }
    }

    dots <- list(...)
    hasNoOther <- (length(dots)) == 0 | is.null(dots[!names(dots) %in% sortedOrRegexp])

    # Check if no args -- faster to delete all then make new empty repo for large repos
    clearWholeCache <- all(missing(userTags), is.null(after), is.null(before),
                           is.null(fun), is.null(cacheId), isTRUE(hasNoOther))

    if (isTRUEorForce(useCloud) || !clearWholeCache) {
      if (isTRUEorForce(useCloud)) {
        .requireNamespace("googledrive", stopOnFALSE = TRUE, messageStart = "to use google drive files")
      }

      # browser(expr = exists("._clearCache_2"))
      # if (missing(after)) after <- NA # "1970-01-01"
      # if (missing(before)) before <- NA # Sys.time() + 1e5

      args <- append(
        list(x = x, after = after, before = before, userTags = userTags,
             fun = fun, cacheId = cacheId, sorted = FALSE, verbose = verbose),
        list(...)
      )

      objsDT <- do.call(showCache, args = args, quote = TRUE)
      if (isTRUE(useCloud) && NROW(objsDT) > 0 || identical(useCloud, "force")) {
        cacheIds <- unique(objsDT[[.cacheTableHashColName()]])
        fns <- Filenames(objsDT)
        rmFromCloudFolder(cloudFolderID, x, cacheIds, otherFiles = fns, verbose = verbose)
      }
    }

    # browser(expr = exists("rrrr"))
    # if (useDBI()) {
    if (!CacheIsACache(x, drv = drv, conn = conn)) {
      return(.emptyCacheTable)
    }
    # }

    if (clearWholeCache) {
      if (isInteractive()) {
        if (isTRUE(ask)) {
          cacheSize <- sum(file.size(dir(x, full.names = TRUE, recursive = TRUE)))
          class(cacheSize) <- "object_size"
          formattedCacheSize <- format(cacheSize, "auto")
          messageQuestion(
            "Your current cache size is ", formattedCacheSize, ".\n",
            " Are you sure you would like to delete it all? Y or N"
          )
          rl <- readline()
          if (!identical(toupper(rl), "Y")) {
            messageCache("Aborting clearCache", verbose = verbose)
            return(invisible())
          }
        }
      }

      unlink(CacheStorageDir(x), recursive = TRUE)

      if (useDBI()) {
        unlink(file.path(x, "rasters"), recursive = TRUE)
        unlink(CacheDBFile(x, drv = drv, conn = conn), recursive = TRUE, force = TRUE)
      }

      checkPath(x, create = TRUE)
      # if (useDBI()) {
      createCache(x, drv = drv, force = TRUE)
      # }
      if (isTRUE(getOption("reproducible.useMemoise"))) {
        objsInMemEnv <- ls(memoiseEnv(x))
        if (length(objsInMemEnv)) {
          rm(list = objsInMemEnv, envir = memoiseEnv(x))
        }
      }
      # memoise::forget(.loadFromLocalRepoMem)
      return(invisible())
    }

    if (isInteractive()) {
      objSz <- objsDT[tagKey == "object.size"][[.cacheTableTagColName()]]
      objSizes <- if ("NA" %in% objSz) NA else as.numeric(objSz)
      cacheSize <- sum(objSizes) / 4
    }

    if (NROW(objsDT)) {
      filesToRemove1 <- objsDT[grepl(pattern = "cacheRaster", tagKey)][[.cacheTableTagColName()]]
      filesToRemove2 <- objsDT[grepl(pattern = "origFilename|filesToLoad|filenamesInCache", tagKey)][[.cacheTableTagColName()]]
      filesToRemove2 <- normPath(file.path(CacheStorageDir(x), basename(filesToRemove2)))
      filesToRemove3 <- dir(CacheStorageDir(x), full.names = TRUE)

      # Way faster to gsub for the cacheId, rather than greps
      cacheIdsOfTheseFilenames <- gsub("^.*/([0-9a-zA-Z]+)\\..*$", "\\1", filesToRemove3)
      cacheIdsToRm <- unique(objsDT[["cacheId"]])
      indicesToRm <- which(cacheIdsOfTheseFilenames %in% cacheIdsToRm)
      filesToRemove4 <- filesToRemove3[indicesToRm]

      # grep can only handle so many -- do in groups
      # filesToRemove3 <- grep(paste(cacheIdsToRm, collapse = "|"), filesToRemove3, value = TRUE)
      # maxNumForGrep <- 20
      # sequen <- seq(cacheIdsToRm)
      # if (length(cacheIdsToRm) > maxNumForGrep) {
      #   groups <- cut(sequen, breaks = ceiling(length(cacheIdsToRm) / maxNumForGrep))
      #   groups <- split(sequen, groups)
      # } else {
      #   groups <- list(sequen)
      # }
      #
      # filesToRemove4 <- lapply(groups, function(g) {
      #   cis <- cacheIdsToRm[sequen[g]] |> unlist()
      #   grep(paste(cis, collapse = "|"), filesToRemove3, value = TRUE)
      # }) |> unlist() |> unname()

      filesToRemove <- unique(c(filesToRemove1, filesToRemove2, filesToRemove4))
      # filebackedInRepo <- objsDT[grepl(pattern = "fromDisk", tagKey) &
      #                           grepl(pattern = "TRUE", get(.cacheTableTagColName()))]
      #
      # rastersInRepo <- objsDT[grepl(pattern = "class", tagKey) &
      #                           grepl(pattern = "Raster", get(.cacheTableTagColName()))]
      # listsInRepo <-  objsDT[grepl(pattern = "class", tagKey) &
      #                          grepl(pattern = "list", get(.cacheTableTagColName()))]
      # hasARaster <- all(!is.na(rastersInRepo[[.cacheTableHashColName()]])) && NROW(rastersInRepo) > 0 # nolint
      # hasAList <- all(!is.na(rastersInRepo[[.cacheTableHashColName()]])) && NROW(listsInRepo) > 0 # nolint

      if (NROW(filesToRemove)) {
        # fileBackedRastersInRepo <- filebackedInRepo[[.cacheTableHashColName()]]# [rasterObjSizes < 1e5]
        # if (NROW(fileBackedRastersInRepo)) {
        filesToRemove <- unlist(filesToRemove)
        if (isInteractive()) {
          dirLs <- dir(unique(dirname(filesToRemove)), full.names = TRUE)
          filesToRemove <- dirLs[which(basename(dirLs)  %in% basename(filesToRemove))]
          # dirLs <- unlist(lapply(basename(filesToRemove), grep, dirLs, value = TRUE))
          # filesToRemove <- unique(c(filesToRemove, dirLs))
          cacheSize <- sum(file.size(filesToRemove))
        }
        # }
      }

      if (isInteractive() && isTRUE(!is.na(cacheSize))) {
        class(cacheSize) <- "object_size"
        formattedCacheSize <- format(cacheSize, "auto")
        if (isTRUE(ask)) {
          messageQuestion(
            "Your size of your selected objects (including file-backed objects) is ",
            formattedCacheSize, ".\n",
            " Are you sure you would like to delete it all? Y or N"
          )
          rl <- readline()
          if (!identical(toupper(rl), "Y")) {
            messageCache("Aborting clearCache", verbose = verbose)
            return(invisible())
          }
        }
      }

      # remove file-backed files
      if (NROW(filesToRemove) > 0) {
        unlink(filesToRemove)
      }

      objToGet <- unique(objsDT[[.cacheTableHashColName()]])
      if (useDBI()) {
        if (is.null(conn)) {
          conn <- dbConnectAll(drv, cachePath = x, create = FALSE)
          on.exit({
            DBI::dbDisconnect(conn)
          })
        }
      }
      rmFromCache(x, objToGet, conn = conn, drv = drv, verbose = verbose) # many = TRUE)
      if (isTRUE(getOption("reproducible.useMemoise"))) {
        exist <- vapply(objToGet, exists, envir = memoiseEnv(x), FUN.VALUE = logical(1))
        if (isTRUE(any(exist))) {
          suppressWarnings(rm(list = objToGet[exist], envir = memoiseEnv(x)))#.pkgEnv[[x]]))
        }
      }
    }
    try(setindex(objsDT, NULL), silent = TRUE)
    return(invisible(objsDT))
  }
)

#' @details
#' `cc(secs)` is just a shortcut for `clearCache(repo = currentRepo, after = secs)`,
#' i.e., to remove any cache entries touched in the last `secs` seconds. Since, `secs`
#' can be missing, this is also be a shorthand for "remove most recent entry from
#' the cache".
#'
#' @param secs Currently 3 options: the number of seconds to pass to `clearCache(after = secs)`,
#'     a `POSIXct` time e.g., from `Sys.time()`, or missing. If missing,
#'             the default, then it will delete the most recent entry in the Cache.
#'
#' @export
#' @rdname viewCache
#'
#' @examples
#' data.table::setDTthreads(2)
#' tmpDir <- file.path(tempdir(), "reproducible_examples", "Cache")
#' try(clearCache(tmpDir, ask = FALSE), silent = TRUE) # just to make sure it is clear
#'
#' Cache(rnorm, 1, cachePath = tmpDir)
#' thisTime <- Sys.time()
#' Cache(rnorm, 2, cachePath = tmpDir)
#' Cache(rnorm, 3, cachePath = tmpDir)
#' Cache(rnorm, 4, cachePath = tmpDir)
#' showCache(x = tmpDir) # shows all 4 entries
#' cc(ask = FALSE, x = tmpDir)
#' showCache(x = tmpDir) # most recent is gone
#' cc(thisTime, ask = FALSE, x = tmpDir)
#' showCache(x = tmpDir) # all those after thisTime gone, i.e., only 1 left
#' cc(ask = FALSE, x = tmpDir) # Cache is
#' cc(ask = FALSE, x = tmpDir) # Cache is already empty
cc <- function(secs, ..., verbose = getOption("reproducible.verbose")) {
  if (missing(secs)) {
    messageCache("No time provided; removing the most recent entry to the Cache",
                 verbose = verbose
    )
    suppressMessages({
      theCache <- reproducible::showCache(...)
    })
    if (NROW(theCache) > 0) {
      accessed <- data.table::setkey(theCache[tagKey == "accessed"], tagValue)
      clearCache(userTags = tail(accessed, 1)[[.cacheTableHashColName()]], ...)
    } else {
      messageCache("Cache already empty", verbose = verbose)
    }
  } else {
    if (is(secs, "POSIXct")) {
      reproducible::clearCache(after = secs, ...)
    } else {
      reproducible::clearCache(after = Sys.time() - secs, ...)
    }
  }
}

#' Examining and modifying the cache
#'
#' These are convenience wrappers around `DBI` package functions.
#' They allow the user a bit of control over what is being cached.
#'
#' \describe{
#'   \item{`clearCache`}{remove items from the cache based on their
#'                            `userTag` or `times` values.}
#'   \item{`keepCache`}{remove all cached items *except* those based on
#'                           certain `userTags` or `times` values.}
#'   \item{`showCache`}{display the contents of the cache.}
#' }
#'
#' @details
#' By default the return of `showCache` is sorted by `cacheId`. For convenience,
#' a user can optionally have it unsorted (passing `sorted = FALSE`),
#' which may be noticeably faster when
#' the cache is large (`> 1e4` entries).
#'
#' @inheritParams clearCache
#' @inheritParams Cache
#' @export
#' @importFrom data.table data.table set setkeyv
#' @rdname viewCache
#' @name showCache
#' @seealso [mergeCache()]. Many more examples in [Cache()].
#'
setGeneric("showCache", function(x, userTags = character(), after = NULL, before = NULL,
                                 fun = NULL, cacheId = NULL,
                                 # cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                                 drv = getDrv(getOption("reproducible.drv", NULL)),
                                 conn = getOption("reproducible.conn", NULL),
                                 verbose = getOption("reproducible.verbose"),
                                 ...) {
  standardGeneric("showCache")
})

#' @export
#' @rdname viewCache
setMethod(
  "showCache",
  definition = function(x, userTags, after = NULL, before = NULL, fun = NULL,
                        cacheId = NULL, # cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                        drv, conn, ...) {
    # browser(expr = exists("rrrr"))
    if (missing(x)) {
      messageCache("x not specified; using ", getOption("reproducible.cachePath")[1], verbose = verbose)
      x <- getOption("reproducible.cachePath")[1]
    }
    # browser(expr = exists("jjjj"))
    # if (useDBI()) {
    afterNA <- FALSE
    if (is.null(after)) {
      afterNA <- TRUE
      after <- NA
    }
    # "1970-01-01"
    beforeNA <- FALSE
    if (is.null(before)) {
      beforeNA <- TRUE
      before <- NA
    } # Sys.time() + 1e5
    # } else {
    #   if (is.null(after)) after <- "1970-01-01"
    #   if (is.null(before)) before <- Sys.time() + 1e5
    # }

    # not seeing userTags
    # Clear the futures that are resolved
    .onLinux <- .Platform$OS.type == "unix" && unname(Sys.info()["sysname"]) == "Linux" &&
      !isFALSE(getOption("reproducible.futurePlan"))
    if (.onLinux) {
      if (exists("futureEnv", envir = .reproEnv)) {
        hasFuture <- .requireNamespace("future",
                                       messageStart = "To use reproducible.futurePlan, "
        )
        if (hasFuture) {
          checkFutures(verbose)
        }
      }
    }

    if (!useDBI()) {
      pkgEnv <- memoiseEnv(cachePath = x)
      if (!exists("shownCache", envir = pkgEnv))
        pkgEnv[["shownCache"]] <- new.env()
      if (!exists(x, envir = pkgEnv[["shownCache"]]))
        pkgEnv[["shownCache"]][[x]] <- new.env()

      ## Note: do NOT lazy-spawn from here. The async fork itself runs
      ## showCache(); spawning again from inside the fork would recurse
      ## (the fork's pkgEnv has no job entry yet because spawn_showCache_async
      ## assigns the job AFTER mcparallel returns). Cache() is the lazy-spawn
      ## site; direct showCache() callers can use prepopulateCacheAsync()
      ## explicitly if they want a warm fork.

      # Non-blocking poll: if the async pre-load job has already finished,
      # harvest it; otherwise proceed synchronously.  Never block here — for
      # large caches the fork can take minutes, and blocking would defeat the
      # purpose of having the incremental-update mechanism below.
      collect_showCache_async(x, wait = FALSE, timeout = 0)
      scEnv <- pkgEnv[["shownCache"]][[x]]

      # periodically, a cache entry is corrupt; this while, tryCatch will remove the corrupt file and restart
      objsDT <- list()
      while(is(objsDT, "list")) {
        # filOutside <- character()
        objsDT <- tryCatch2(
          if (!is.null(cacheId)) {
            objsDT <- rbindlist(fill = TRUE, lapply(cacheId, function(fil) {
              showCacheFast(fil, cachePath = x,
                            drv = drv, conn = conn)
            }))
          } else {
            dd <- dir(CacheStorageDir(x),
                      pattern = paste(CacheDBFileSingleExt(cacheSaveFormat = .cacheSaveFormats), collapse = "|"),
                      full.names = TRUE
            )
            lapplyFun <- lapply
            curFileInfo <- file.info(dd) |> setDT(keep.rownames = "filename")

            # Compare only on stable, content-relevant columns.
            # Joining on all file.info columns (including atime) caused every
            # file to appear "new" on every call because loadFile() updates atime.
            stableCols <- c("filename", "mtime", "size")

            if (is.null(scEnv$FileInfo)) {
              newOnes <- curFileInfo
            } else {
              newOnes  <- curFileInfo[!scEnv$FileInfo,  on = stableCols]
              removeOnes <- scEnv$FileInfo[!curFileInfo, on = "filename"]
              if (NROW(removeOnes)) {
                scEnv$FileInfo <- scEnv$FileInfo[!removeOnes, on = "filename"]
                cis <- filePathSansExt(filePathSansExt(basename(removeOnes$filename)))
                scEnv$sc <- scEnv$sc[!cacheId %in% cis]
              }
            }
            dd <- newOnes[["filename"]]
            ddOrig <- dd
            # If a file's mtime changed (e.g. a cache hit rewrote the accessed
            # tag), it appears in newOnes even though it still exists.  Without
            # this purge, the old entry and the freshly-loaded one would both be
            # present in scEnv$sc, causing duplicate cacheIds and a cartesian
            # join error in the userTags filter below.
            if (!is.null(scEnv$sc) && length(dd) > 0) {
              cisToRemove <- filePathSansExt(filePathSansExt(basename(dd)))
              scEnv$sc <- scEnv$sc[!cacheId %in% cisToRemove]
            }
            scEnv$FileInfo <- curFileInfo

            keepDoing <- TRUE
            while(keepDoing) {
              allFilesLoaded <- lapplyFun(dd, function(fil) {

                ## Wrap loadFile in try(): readRDS / qs_read can throw on
                ## corrupt or wrong-format files (e.g. "unknown input format"
                ## from readRDS when the .rds extension lies). The recovery
                ## branch below expects out to be a try-error, so without
                ## the wrap the error escapes the loop and the whole
                ## showCache() call.
                out <- try(loadFile(fil,
                                    cachePath = x, # in case it needs swapCacheFormat
                                    drv = drv, conn = conn, verbose = verbose),
                           silent = TRUE)
                if (is(out, "try-error")) {
                  cacheId <- gsub(paste0(CacheDBFileSingleExt()), "",
                                  basename(fil))

                  fileEx <- fileExt(fil)
                  fileExs <- setdiff(.cacheSaveFormats, fileEx)
                  for (fe in fileExs) {
                    out <- try(loadFile(fil, format = fe,
                                        cacheId = cacheId, cachePath = x, # in case it needs swapCacheFormat
                                        drv = drv, conn = conn, verbose = verbose),
                               silent = TRUE)
                    if (!is(out, "try-error")) {
                      if (identical(getOption("reproducible.cacheSaveFormat"), .qsFormat))
                        optForUndo <- options("reproducible.qsFormat" = .qsFormat)
                      on.exit(options(optForUndo), add = TRUE)
                      saveFilesInCacheFolder(out, fts = fil, cachePath = x,
                                             cacheId = cacheId)

                      return(out)
                    }
                  }
                  # browser()
                  filesToRm <- dir(dirname(fil), pattern = cacheId, full.names = TRUE)
                  messageCache("The database file was corrupt; deleting Cache entry for ", cacheId,
                               verbose = getOption("reproducible.verbose"))
                  unlink(filesToRm)
                  # Return NULL (not the try-error) so rbindlist(fill = TRUE)
                  # below silently skips this entry. Letting the try-error
                  # escape made rbindlist error with "Item N is not a list",
                  # the catch handler didn't successfully retry, and the
                  # whole showCache() call returned NULL — which broke
                  # test-showCacheCorruptFile.R's NROW(out) > 0 invariant
                  # (valid sibling entries should still be returned).
                  return(NULL)
                }
                out
              })
              allFilesLoaded <- Filter(Negate(is.null), allFilesLoaded)


              ret <- tryCatch(

                rbindlist(fill = TRUE, allFilesLoaded), error = function(err) {
                if (any(grepl("Item .+ is not a", err$message))) {
                  toDel <- gsub("Item ([[:digit:]]+) of.+", "\\1", err$message) |> as.numeric()
                  unlink(dd[toDel])
                  dd <- dd[-toDel]
                }
                # browser()
                if (any(grepl("use qs::qread", err$message))) {
                  swapCacheFileFormat()
                }
                # next
              })
              #if (is(ret, "try-error"))
              #  browser()
              keepDoing <- FALSE
            }

            if (!is.null(scEnv$sc)) {
              ret <- rbindlist(list(scEnv$sc, ret))
            }
            scEnv$sc <- ret
            ret

          }# , error = function(e) {
          #   cacheId <- gsub(paste0(CacheDBFileSingleExt(), "|", cacheSaveFormat), "",
          #                   basename(file))
          #   filesToRm <- dir(dirname(file), pattern = cacheId, full.names = TRUE)
          #   messageCache("The database file was corrupt; deleting Cache entry for ", cacheId,
          #                   verbose = getOption("reproducible.verbose"))
          #   unlink(filesToRm)
          # }
        )
      }
      if (NROW(objsDT) == 0) {
        return(.emptyCacheTable)
      }
    } else {
      if (is.null(conn)) {
        conn <- dbConnectAll(drv, cachePath = x, create = FALSE)
        if (is.null(conn)) {
          return(.emptyCacheTable)
        }
        on.exit(DBI::dbDisconnect(conn), add = TRUE)
      }
      if (!CacheIsACache(x, drv = drv, conn = conn)) {
        return(.emptyCacheTable)
      }

      dbTabNam <- CacheDBTableName(x, drv = drv)
      # tab <- dbReadTable(conn, dbTabNam)
      res <- retry(retries = 250, exponentialDecayBase = 1.01, quote(
        DBI::dbSendQuery(conn, paste0("SELECT * FROM \"", dbTabNam, "\""))
      ))
      tab <- DBI::dbFetch(res)
      DBI::dbClearResult(res)
      if (is(tab, "try-error")) {
        objsDT <- .emptyCacheTable
      } else {
        objsDT <- setDT(tab)
      }
      if (!is.null(fun)) {

      }
    }

    onCol <- "cacheId"
    if (!is.null(cacheId)) {
      cacheIds <- cacheId
      objsDT <- objsDT[unique(objsDT[cacheId %in% cacheIds, ..onCol]), on = onCol]
    }
    if (!is.null(fun)) {
      objsDT <- objsDT[objsDT[tagKey %in% "function" & tagValue %in% fun, ..onCol], on = onCol]
    }
    dots <- list(...)

    dots <- dots[!names(dots) %in% sortedOrRegexp]
    if (length(dots)) {
      names(dots) <- gsub("^Function$", "function", names(dots)) # in case user uses Function instead of "function"
      Map(nam = names(dots), val = dots, function(nam, val) {
        objsDT <<- objsDT[objsDT[tagKey %in% nam & tagValue %in% val, ..onCol], on = onCol]
      })

    }
    sorted <- !isFALSE(list(...)$sorted) # NULL and TRUE are sorted
    if (isTRUE(sorted) && NROW(objsDT)) {
      data.table::setorderv(objsDT, onCol)
    }
    # }

    if (NROW(objsDT) > 0) {
      if (!afterNA || !beforeNA) {
        objsDT3 <- objsDT[tagKey == "accessed"]
        if (!beforeNA) {
          objsDT3 <- objsDT3[(tagValue <= before)]
        }
        if (!afterNA) {
          objsDT3 <- objsDT3[(tagValue >= after)]
        }
        objsDT <- objsDT[objsDT[[.cacheTableHashColName()]] %in%
                           unique(objsDT3[[.cacheTableHashColName()]])] # faster than data.table join
      }
      if (length(userTags) > 0) {
        if (isTRUE(list(...)$regexp) | is.null(list(...)$regexp)) {
          objsDTs <- list()
          for (ut in userTags) {
            if (grepl(":", ut)) {
              ut <- paste(strsplit(ut, ":")[[1]], collapse = "|")
            }
            objsDT2 <- objsDT[
              grepl(get(.cacheTableTagColName()), pattern = ut) |
                grepl(tagKey, pattern = ut) |
                grepl(get(.cacheTableHashColName()), pattern = ut)
            ]
            setkeyv(objsDT2, .cacheTableHashColName())
            shortDT <- unique(objsDT2, by = .cacheTableHashColName())[, get(.cacheTableHashColName())]
            # }
            objsDT <- if (NROW(shortDT)) objsDT[shortDT, on = .cacheTableHashColName()] else objsDT[0] # merge each userTags
          }
        } else {
          # if (useDBI()) {
          objsDT2 <- objsDT[cacheId %in% userTags | tagKey %in% userTags | tagValue %in% userTags]
          setkeyv(objsDT2, onCol)
          shortDT <- unique(objsDT2, by = onCol)[, cacheId]
          objsDT <- if (NROW(shortDT)) objsDT[shortDT, on = .cacheTableHashColName()] else objsDT[0] # merge each userTags
          # } else {
          #   objsDT2 <- objsDT[artifact %in% userTags | tagKey %in% userTags | tagValue %in% userTags]
          #   setkeyv(objsDT2, "artifact")
          #   shortDT <- unique(objsDT2, by = "artifact")[, artifact]
          #   objsDT <- if (NROW(shortDT)) objsDT[shortDT, on = .cacheTableHashColName()] else objsDT[0] # merge each userTags
          # }
        }
      }
    }
    .message$CacheSize(x,
                       artifacts = unique(objsDT[[.cacheTableHashColName()]]),
                       cacheTable = objsDT, verbose = verbose
    )
    return(objsDT)
  }
)

#' @rdname viewCache
setGeneric("keepCache", function(x, userTags = character(), after = NULL, before = NULL,
                                 ask = getOption("reproducible.ask"),
                                 drv = getDrv(getOption("reproducible.drv", NULL)),
                                 conn = getOption("reproducible.conn", NULL),
                                 verbose = getOption("reproducible.verbose"),
                                 ...) {
  standardGeneric("keepCache")
})

#' @export
#' @rdname viewCache
setMethod(
  "keepCache",
  definition = function(x, userTags, after, before, ask, drv, conn,
                        verbose = getOption("reproducible.verbose"),
                        ...) {
    if (missing(x)) {
      messageCache("x not specified; using ", getOption("reproducible.cachePath")[1], verbose = verbose)
      x <- getOption("reproducible.cachePath")[1]
    }
    args <- append(list(x = x, after = after, before = before, userTags = userTags),
                   modifyList(list(...), list(verbose = FALSE)))

    objsDTAll <- suppressMessages(showCache(x, verbose = FALSE, sorted = FALSE))
    objsDT <- do.call(showCache, args = args)
    keep <- unique(objsDT[[.cacheTableHashColName()]])

    eliminate <- unique(objsDTAll[[.cacheTableHashColName()]][
      !(objsDTAll[[.cacheTableHashColName()]] %in% keep)
    ])

    if (length(eliminate)) {
      clearCache(x, cacheId = eliminate, verbose = FALSE, regexp = FALSE, ask = ask)
    } else {
      messageCache("Nothing to remove; keeping all")
    }
    return(objsDT)
  }
)

#' Merge two cache repositories together
#'
#' \if{html}{\figure{lifecycle-experimental.svg}{options: alt="experimental"}}
#'
#' All the `cacheFrom` artifacts will be put into `cacheTo`
#' repository. All `userTags` will be copied verbatim, including
#' `accessed`, with 1 exception: `date` will be the
#' current `Sys.time()` at the time of merging. The
#' `createdDate` column will be similarly the current time
#' of merging.
#'
#' @param cacheTo The cache repository (character string of the file path)
#'                that will become larger, i.e., merge into this
#' @param cacheFrom The cache repository (character string of the file path)
#'                  from which all objects will be taken and copied from
#' @param drvTo The database driver for the `cacheTo`.
#' @param drvFrom The database driver for the `cacheFrom`
#' @param connTo The connection for the `cacheTo`. If not provided, then
#'   a new one will be made from `drvTo` and `cacheTo`
#' @param connFrom The database for the `cacheFrom`. If not provided, then
#'   a new one will be made from `drvFrom` and `cacheFrom`
#'
#' @return The character string of the path of `cacheTo`, i.e., not the
#' objects themselves.
#' @inheritParams Cache
setGeneric("mergeCache", function(cacheTo, cacheFrom,
                                  drvTo = getDrv(getOption("reproducible.drv", NULL)),
                                  drvFrom = getDrv(getOption("reproducible.drv", NULL)),
                                  connTo = NULL, connFrom = NULL,
                                  verbose = getOption("reproducible.verbose")) {
  standardGeneric("mergeCache")
})

#' @export
#' @rdname mergeCache
setMethod(
  "mergeCache",
  definition = function(cacheTo, cacheFrom, drvTo, drvFrom, connTo, connFrom,
                        verbose = getOption("reproducible.verbose")) {

    if (useDBI()) {
      if (is.null(connTo)) {
        connTo <- dbConnectAll(drvTo, cachePath = cacheTo)
        on.exit(DBI::dbDisconnect(connTo), add = TRUE)
      }

      if (is.null(connFrom)) {
        connFrom <- dbConnectAll(drvFrom, cachePath = cacheFrom)
        on.exit(DBI::dbDisconnect(connFrom), add = TRUE)
      }
    }

    suppressMessages({
      cacheFromList <- showCache(cacheFrom, drv = drvFrom, conn = connFrom, sorted = FALSE)
    })
    suppressMessages({
      cacheToList <- showCache(cacheTo, drv = drvTo, conn = connTo, sorted = FALSE)
    })

    artifacts <- unique(cacheFromList[[.cacheTableHashColName()]])
    objectList <- lapply(artifacts, function(artifact) {
      # browser(expr = exists("gggg"))

      if (!(artifact %in% cacheToList[[.cacheTableHashColName()]])) {
        outputToSave <- # if (useDBI()) {
          try(loadFromCache(
            cachePath = cacheFrom, fullCacheTableForObj = cacheToList,
            cacheId = artifact, verbose = verbose
          ))
        if (is(outputToSave, "try-error")) {
          messageCache("Continuing to load others", verbose = verbose)
          outputToSave <- NULL
        }

        ## Save it
        userTags <- cacheFromList[artifact, on = .cacheTableHashColName()][
          !tagKey %in% c("format", "name", "date"), list(tagKey, tagValue)
        ]
        outputToSave <- .wrap(outputToSave, cachePath = cacheTo, drv = drvTo, conn = connTo)
        output <- saveToCache(cacheTo,
                              userTags = userTags, obj = outputToSave, cacheId = artifact,
                              drv = drvTo, conn = connTo
        ) # nolint
        messageCache(artifact, " copied", verbose = verbose)
        outputToSave
      } else {
        messageCache("Skipping ", artifact, "; already in ", cacheTo, verbose = verbose)
      }
    })

    .message$CacheSize(cacheTo, cacheTable = showCache(cacheTo, sorted = FALSE), verbose = verbose)

    return(invisible(cacheTo))
  }
)


#' @keywords internal
#' @inheritParams Cache
checkFutures <- function(verbose = getOption("reproducible.verbose")) {
  # This takes a long time -- can't use it if
  resol1 <- FALSE
  count <- 0
  lsFutureEnv <- ls(.reproEnv$futureEnv)

  anyFutureWrite <- length(lsFutureEnv)

  if (anyFutureWrite > 0) {
    # objsInReproEnv <- ls(.reproEnv)
    # objsInReproEnv <- grep("^future|cloudCheckSums", objsInReproEnv, value = TRUE)
    while (any(!resol1)) {
      count <- count + 1
      # numSleeps <<- numSleeps+1
      if (count > 1) {
        Sys.sleep(0.001)
        if (count > 1e3) {
          messageCache("Future is not resolved after 1 second of waiting. Allowing to proceed.",
                       verbose = verbose
          )
          break
        }
      }
      resol <- future::resolved(.reproEnv$futureEnv)
      resol1 <- resol[!startsWith(names(resol), "cloudCheckSums")]
    }
    if (length(resol) > 0) {
      .reproEnv$futureEnv[[lsFutureEnv]] <- NULL
    }
  }
}

useDBI <- function(set = NULL, verbose = getOption("reproducible.verbose"), default = TRUE) {
  canSwitch <- TRUE
  if (!is.null(set)) {
    if (isTRUE(set)) {
      canSwitch <- .requireNamespace("RSQLite", stopOnFALSE = FALSE) &&
        .requireNamespace("DBI", stopOnFALSE = FALSE)
    }
    if (isTRUE(canSwitch)) {
      options("reproducible.useDBI" = set)
    }
  }
  ud <- getOption("reproducible.useDBI", default)
  if (isTRUE(ud)) {
    drv <- getOption("reproducible.drv")
    if (is.null(drv)) {
      canSwitch <- .requireNamespace("RSQLite", stopOnFALSE = FALSE) &&
        .requireNamespace("DBI", stopOnFALSE = FALSE)
      if (isFALSE(canSwitch)) {
        options("reproducible.useDBI" = FALSE)
        ud <- getOption("reproducible.useDBI")
      }
    }
  }

  if (isFALSE(canSwitch)) {
    messageColoured("User has requested to use DBI as the backend, but DBI and/or RSQLite not ",
                    "installed.",
                    verboseLevel = 1, verbose = verbose
    )
  }
  if (!is.null(set)) {
    messSet <- if (isTRUE(ud)) {
      "Using DBI backend."
    } else {
      "Using non-DBI backend."
    }
    if (verbose > -1) {
      messageColoured(messSet, verboseLevel = 0, verbose = verbose)
    }
  }

  ud
}

#' @inheritParams Cache
rmFromCloudFolder <- function(cloudFolderID, x, cacheIds, otherFiles,
                              verbose = getOption("reproducible.verbose")) {
  if (is.null(cloudFolderID)) {
    cloudFolderID <- checkAndMakeCloudFolderID(cloudFolderID, cachePath = x)
  }

  whEmpty <- !nzchar(otherFiles)
  if (any(whEmpty)) {
    otherFiles <- otherFiles[!whEmpty]
  }
  grepToSrch <- c(cacheIds, otherFiles)
  gdriveLs <- driveLs(cloudFolderID, pattern = paste(grepToSrch, collapse = "|"))
  isInCloud <- lapply(grepToSrch, function(ci) startsWith(prefix = ci, gdriveLs$name))
  isInCloud <- Reduce(rbind, isInCloud)
  if (!is.null(dim(isInCloud))) {
    isInCloud <- apply(isInCloud, 2, any)
  }

  if (any(isInCloud)) {
    toDelete <- gdriveLs[isInCloud, ]
  }
  if (any(isInCloud)) {
    retry(quote(googledrive::drive_rm(toDelete)))
  }

  return(invisible())
}


isTRUEorForce <- function(cond) {
  isTRUE(cond) || identical(cond, "force")
}

showCacheFast <- function(cacheId, cachePath = getOption("reproducible.cachePath"),
                          dtFile, strict = TRUE, # cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                          drv, conn, verbose = getOption("reproducible.verbose")) {

  if (missing(dtFile)) {
    # dtFile <- CacheDBFileSingle(cachePath, cacheId, cacheSaveFormat = "check")
    dtFile <- CacheDBFileSingle(cachePath, cacheId, cacheSaveFormat = "check")
    # dtFile <- dir(CacheStorageDir(cachePath), full.names = TRUE,
    #               pattern = paste0(cacheId, "\\", suffixMultipleDBFiles()))
  }
  fe <- file.exists(dtFile)
  sc <- NULL
  if (fe || isFALSE(strict)) {
    dtFile <- if (any(fe)) dtFile[fe][1] else character()
    if (length(dtFile)) {
      sc <- loadFile(dtFile,
                     cacheId = cacheId, cachePath = cachePath, # in case it needs swapCacheFormat
                     drv = drv, conn = conn, verbose = verbose) # , cacheSaveFormat = cacheSaveFormat)
    } else {
      sc <- showCache(cachePath, userTags = cacheId, drv = drv, conn = conn, verbose = FALSE)[cacheId %in% cacheId]
    }
  }
  sc[]
}

sortedOrRegexp <- c("sorted", "regexp", "ask")




# mcparallel is fork-based and not available on Windows
# pkgEnv <- reproducible:::pkgEnv()  # internal environment for package objects [3](https://rdrr.io/cran/reproducible/man/pkgEnv.html)
## Idempotent, all-guards-applied wrapper used by Cache() and showCache() to
## kick off a background showCache scan for `x` the first time the path is
## touched in a session. Cheap (~10us) when a job already exists -- safe to
## call from hot paths.
##
## Skipped silently on Windows (no fork), when parallel isn't available, or
## when `x` is NULL/non-character.
.maybeSpawnShowCacheAsync <- function(x = getOption("reproducible.cachePath")) {
  if (.Platform$OS.type == "windows") return(invisible(NULL))
  if (is.null(x) || !is.character(x) || !nzchar(x[[1L]])) return(invisible(NULL))
  if (!requireNamespace("parallel", quietly = TRUE)) return(invisible(NULL))
  ## spawn_showCache_async is itself idempotent via its overwrite=FALSE guard
  spawn_showCache_async(x[[1L]], silent = TRUE, overwrite = FALSE)
}

#' Pre-populate the in-memory `showCache` cache for a given `cachePath`
#'
#' Forks a background process that runs `showCache()` against `cachePath`;
#' subsequent `showCache()` / `Cache()->showSimilar()` calls in the same R
#' session can then harvest the result instead of re-scanning the cache
#' directory synchronously. Useful for very large caches (tens of thousands
#' of entries) where the cold first scan can take a minute or more.
#'
#' Idempotent: a second call with the same `cachePath` reuses the existing
#' job. Skipped silently on Windows (forking-based) and when the `parallel`
#' package isn't available.
#'
#' This helper is called automatically the first time `Cache()` or
#' `showCache()` is invoked against a given `cachePath`, so most users do
#' not need to call it explicitly. It is exported for workflows that want
#' to kick off the spawn early (e.g. inside `setupProject()`) so the fork
#' has more wall-clock time to complete before the first manual
#' `showCache()` call.
#'
#' @param cachePath A character path. Defaults to
#'   `getOption("reproducible.cachePath")`.
#' @return Invisibly returns the spawn job handle, or `NULL` if the spawn
#'   was skipped.
#' @export
prepopulateCacheAsync <- function(cachePath = getOption("reproducible.cachePath")) {
  invisible(.maybeSpawnShowCacheAsync(cachePath))
}

spawn_showCache_async <- function(
    x = getOption("reproducible.cachePath"),
    silent = TRUE,
    overwrite = FALSE
) {
  if (.Platform$OS.type == "windows") {
    return(NULL)
  }

  # Internal env for package state
  pkgEnv <- memoiseEnv(cachePath = x)
  if (!exists("shownCache", envir = pkgEnv))
    pkgEnv[["shownCache"]] <- new.env()
  if (!exists(x, envir = pkgEnv[["shownCache"]]))
    pkgEnv[["shownCache"]][[x]] <- new.env()

  if (is.null(pkgEnv[["shownCache"]]$shownCache_jobs)) pkgEnv[["shownCache"]]$shownCache_jobs <- new.env(parent = emptyenv())

  # If job exists and not overwriting, reuse it
  if (!overwrite && exists(x, envir = pkgEnv[["shownCache"]]$shownCache_jobs, inherits = FALSE)) {
    return(invisible(get(x, envir = pkgEnv[["shownCache"]]$shownCache_jobs, inherits = FALSE)))
  }

  # Capture the function objects in the *parent*.
  # This avoids any namespace loading inside the fork.
  ns <- asNamespace("reproducible")
  showCache_fun  <- get("showCache", envir = ns)
  memoiseEnv_fun <- get("memoiseEnv", envir = ns)

  # Build fork expression with injected function objects
  expr <- substitute({
    data.table::setDTthreads(1L)
    options(reproducible.nThreads = 1L)
    
    # Run slow call (side effects stay in child; we return the memoised object)
    SHOWCACHE(cp, drv = NULL, conn = NULL)
    
    # Harvest the memoised object created by showCache
    pkgEnv_child <- MEMOISEENV(cachePath = cp)
    sc <- pkgEnv_child[["shownCache"]][[cp]]

    # mcparallel/mccollect: NULL should not be returned (reserved as error signal) [1](https://www.r-bloggers.com/2023/06/dofuture-a-better-foreach-parallelization-operator-than-dopar/)
    if (is.null(sc)) {
      structure(list(error = "shownCache was NULL in child", cachePath = cp),
                class = "shownCache_error")
    } else {
      sc
    }
  }, list(
    cp = x,
    SHOWCACHE = showCache_fun,
    MEMOISEENV = memoiseEnv_fun
  ))

  job <- parallel::mcparallel(expr, name = paste0("showCache:", x), silent = silent)  # [1](https://www.r-bloggers.com/2023/06/dofuture-a-better-foreach-parallelization-operator-than-dopar/)
  assign(x, job, envir = pkgEnv[["shownCache"]]$shownCache_jobs)

  invisible(job)
}


collect_showCache_async <- function(
    x = getOption("reproducible.cachePath"),
    wait = FALSE,
    timeout = 10
) {
  if (.Platform$OS.type == "windows") {
    return(NULL)
    # stop("parallel::mccollect is not available on Windows (forking backend).")
  }

  pkgEnv <- memoiseEnv(cachePath = x)
  if (!exists("shownCache", envir = pkgEnv))
    pkgEnv[["shownCache"]] <- new.env()
  if (!exists(x, envir = pkgEnv[["shownCache"]]))
    pkgEnv[["shownCache"]][[x]] <- new.env()
  if (is.null(pkgEnv[["shownCache"]]$shownCache_jobs) || !exists(x, envir = pkgEnv[["shownCache"]]$shownCache_jobs, inherits = FALSE)) {
    return(invisible(NULL))  # nothing spawned for this cachePath
  }

  job <- get(x, envir = pkgEnv[["shownCache"]]$shownCache_jobs, inherits = FALSE)

  
  # The warning occurs if the pid has already be deleted e.g., manually
  suppressWarnings(
    # Poll or wait for results
    res_list <- parallel::mccollect(job, wait = wait, timeout = timeout)  # collect async results [1](https://www.rdocumentation.org/packages/parallel/versions/3.4.1/topics/mcparallel)[2](https://stat.ethz.ch/R-manual/R-devel/library/parallel/html/mcparallel.html)
  )

  # If still running, mccollect returns NULL (per docs)
  if (is.null(res_list)) {
    return(invisible(NULL))
  }

  # Extract result (single job => first element)
  sc <- res_list[[1]]

  # If child reported an internal NULL issue, propagate as an error
  if (inherits(sc, "shownCache_error")) {
    return(invisible(NULL))
  }

  # Install recovered shownCache object into main session memoiseEnv location.
  # The synchronous showCache() path expects pkgEnv[["shownCache"]][[x]] to be
  # an environment with $FileInfo and $sc slots; previously this assigned `sc`
  # directly at that key, replacing the env with an atomic data.table and then
  # crashing the next sync call at `is.null(scEnv$FileInfo)` with
  # "$ operator is invalid for atomic vectors".
  .installAsyncShownCache(memoiseEnv(cachePath = x), x, sc)

  # Clear job handle after successful install
  rm(list = x, envir = pkgEnv[["shownCache"]]$shownCache_jobs)

  invisible(sc)
}

## Install an async-collected showCache result into the per-cachePath env so
## the synchronous showCache() path can read it via `scEnv$sc` /
## `scEnv$FileInfo`. The async child harvests the inner env from its own
## pkgEnv (`pkgEnv_child[["shownCache"]][[cp]]`) and returns it, so the
## value we receive is *itself* the per-cachePath env -- not a data.table.
## Copy its bindings into the parent's env. If we receive a raw
## data.frame/data.table/list (older child shapes), store it at $sc.
## Anything else is treated as no-op (the env is left empty, which causes
## the sync path to fall through to its full disk scan).
.installAsyncShownCache <- function(pkgEnv_main, x, sc) {
  if (!is.environment(pkgEnv_main[["shownCache"]]))
    pkgEnv_main[["shownCache"]] <- new.env(parent = emptyenv())
  if (!is.environment(pkgEnv_main[["shownCache"]][[x]]))
    pkgEnv_main[["shownCache"]][[x]] <- new.env(parent = emptyenv())
  innerEnv <- pkgEnv_main[["shownCache"]][[x]]
  if (is.environment(sc)) {
    for (nm in ls(sc, all.names = TRUE)) {
      assign(nm, get(nm, envir = sc, inherits = FALSE), envir = innerEnv)
    }
  } else if (is.data.frame(sc) || data.table::is.data.table(sc) || is.list(sc)) {
    innerEnv$sc <- sc
  }
  invisible(NULL)
}

test_that(".installAsyncShownCache preserves the per-cachePath env shape", {
  ## Regression for the bug where collect_showCache_async() overwrote
  ## pkgEnv[["shownCache"]][[x]] with the raw `sc` data.table, causing the
  ## next synchronous showCache() call to crash at
  ## `is.null(scEnv$FileInfo)` with "$ operator is invalid for atomic vectors".
  pe <- new.env(parent = emptyenv())
  fakeSc <- data.frame(cacheId = "abc", tagKey = "k", tagValue = "v",
                       stringsAsFactors = FALSE)
  reproducible:::.installAsyncShownCache(pe, "/some/cache/path", fakeSc)

  outer <- pe[["shownCache"]]
  expect_true(is.environment(outer))

  scEnv <- outer[["/some/cache/path"]]
  expect_true(is.environment(scEnv))

  ## Sync path uses these two slots; sc must round-trip, FileInfo stays NULL
  ## so the sync path takes the "newOnes <- curFileInfo" branch on its first
  ## subsequent call.
  expect_identical(scEnv$sc, fakeSc)
  expect_null(scEnv$FileInfo)
})

test_that(".installAsyncShownCache is idempotent and updates $sc in place", {
  pe <- new.env(parent = emptyenv())
  reproducible:::.installAsyncShownCache(pe, "/p", data.frame(v = 1))
  scEnv1 <- pe[["shownCache"]][["/p"]]
  reproducible:::.installAsyncShownCache(pe, "/p", data.frame(v = 2))
  scEnv2 <- pe[["shownCache"]][["/p"]]

  ## Same env identity (not replaced); $sc updated to the new data
  expect_identical(scEnv1, scEnv2)
  expect_equal(scEnv2$sc$v, 2)
})

test_that(".installAsyncShownCache copies bindings from an env-shaped result", {
  ## Regression for: child returns the inner per-cachePath env (sc + FileInfo)
  ## and we previously stored that env directly at scEnv$sc, breaking
  ## rbindlist(list(scEnv$sc, ret)) downstream with
  ## "Item 1 of input is not a data.frame, data.table or list".
  pe <- new.env(parent = emptyenv())

  childEnv <- new.env(parent = emptyenv())
  childEnv$sc <- data.frame(cacheId = "abc", tagKey = "k", tagValue = "v",
                            stringsAsFactors = FALSE)
  childEnv$FileInfo <- data.frame(filename = "/tmp/foo.rds",
                                  mtime = Sys.time(), size = 1L,
                                  stringsAsFactors = FALSE)

  reproducible:::.installAsyncShownCache(pe, "/p", childEnv)

  scEnv <- pe[["shownCache"]][["/p"]]
  expect_true(is.environment(scEnv))
  expect_true(is.data.frame(scEnv$sc))
  expect_true(is.data.frame(scEnv$FileInfo))
  expect_identical(scEnv$sc$cacheId, "abc")
  ## scEnv$sc must be a data.frame so rbindlist downstream accepts it
  expect_silent(data.table::rbindlist(list(scEnv$sc, scEnv$sc), fill = TRUE))
})

test_that(".installAsyncShownCache leaves a clean empty env for unsupported inputs", {
  ## A bare atomic (string, numeric, NA) shouldn't poison scEnv. The contract
  ## with the sync path is: scEnv$sc must be NULL or a data.frame/data.table.
  pe <- new.env(parent = emptyenv())
  reproducible:::.installAsyncShownCache(pe, "/p", "some string")

  scEnv <- pe[["shownCache"]][["/p"]]
  expect_true(is.environment(scEnv))
  expect_null(scEnv$sc)
  expect_null(scEnv$FileInfo)
})

utils::globalVariables(c(
  "cacheId", "checksumsFilename", "checksumsID", "id"
))

#' Check for presence of `checkFolderID` (for `Cache(useCloud)`)
#'
#' Will check for presence of a `cloudFolderID` and make a new one
#' if one not present on Google Drive, with a warning.
#'
#' @inheritParams Cache
#' @param cloudFolderID The google folder ID where cloud caching will occur.
#' @param create Logical. If `TRUE`, then the `cloudFolderID` will be created.
#'     This should be used with caution as there are no checks for overwriting.
#'     See `googledrive::drive_mkdir`. Default `FALSE`.
#' @param overwrite Logical. Passed to `googledrive::drive_mkdir`.
#' @param team_drive Logical indicating whether to check team drives.
#'
#' @return
#' Returns the character string of the cloud folder ID created or reported
#' @export
checkAndMakeCloudFolderID <- function(cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
                                      cachePath = NULL,
                                      create = FALSE,
                                      overwrite = FALSE,
                                      verbose = getOption("reproducible.verbose", 1),
                                      team_drive = NULL) {
  .requireNamespace("googledrive", stopOnFALSE = TRUE, messageStart = "to use google drive files")

  if (!is(cloudFolderID, "dribble")) {
    isNullCFI <- is.null(cloudFolderID)
    if (isNullCFI) {
      if (is.null(cachePath)) {
        cachePath <- .checkCacheRepo(cachePath, verbose = verbose)
      }
      cloudFolderID <- cloudFolderFromCacheRepo(cachePath)
    }

    # This is an imperfect test for a google drive ID ... because of this, we try 2x,
    #   first with best guess, then if wrong, try the other branch of "if (isID)"
    stripHTTP <- gsub("http.+drive.google.com.+folders/(.{32,33})\\?*.+$", "\\1", cloudFolderID)
    if (isID(stripHTTP))
      cloudFolderID <- stripHTTP
    isID <- isTRUE(32 <= nchar(cloudFolderID) && nchar(cloudFolderID) <= 33)
    if (packageVersion("googledrive") < "2.0.0") {
      args <- list(temp_drive = team_drive)
    } else {
      args <- list(shared_drive = team_drive)
    }
    for (attempt in 1:2) {
      cfidTmp <- if (isID) googledrive::as_id(cloudFolderID) else {
        if (fs::is_absolute_path(cloudFolderID)) {
          cloudFolderID <- cloudFolderFromCacheRepo(cloudFolderID)
        }
        cloudFolderID
      }

      driveLs <- tryCatch(# suppressMessages( # will show: "The googledrive package is requesting access to your Google account."
        do.call(googledrive::drive_get, append(list(cfidTmp), args)# )
        ),
        error = function(e) {
          if (!is.null(e$parent)) {
            if (grepl("File not found", as.character(e$parent)) && attempt == 2) {
              stop(e)
            }
          }
        },
        silent = TRUE
      )
      if (is(driveLs, "dribble")) {
        break
      } else {
        isID <- !isID
      }
    }

    if (NROW(driveLs) == 0) {
      newcfid <- cloudFolderID
      if (isTRUE(create)) {
        if (isID) {
          if (is.null(cachePath)) {
            cachePath <- .checkCacheRepo(cachePath, verbose = verbose)
          }
          cloudFolderID <- cloudFolderFromCacheRepo(cachePath)
        }

        if (fs::is_absolute_path(cloudFolderID)) {
          cloudFolderID <- cloudFolderFromCacheRepo(cloudFolderID)
        }

        newcfid <- try(googledrive::drive_mkdir(cloudFolderID, path = NULL, overwrite = overwrite), silent = TRUE) # can handle a single string
      }
      if (is(newcfid, "try-error")) { # this happens if the folder exists
        a <- newcfid
        if (fs::is_absolute_path(cloudFolderID)) {
          cloudFolderID <- cloudFolderFromCacheRepo(cloudFolderID)
        }
        newcfid <- try(googledrive::drive_get(cloudFolderID))
        if (is(newcfid, "try-error"))
          stop(newcfid)
      }
      cloudFolderID <- newcfid

    } else {
      cloudFolderID <- driveLs
    }
    if (isNullCFI) {
      messageCache("Setting 'reproducible.cloudFolderID' option to be cloudFolder: ",
        ifelse(!is.null(names(cloudFolderID)), cloudFolderID$name, cloudFolderID),
        verbose = verbose
      )
    }
    optionValOrig <- options("reproducible.cloudFolderID" = cloudFolderID)
    on.exit2(options(optionValOrig), envir = parent.frame())
  }
  return(cloudFolderID)
}

driveLs <- function(cloudFolderID = NULL, pattern = NULL, cachePath = getOption("reproducible.cachePath"),
                    verbose = getOption("reproducible.verbose", 1),
                    team_drive = NULL) {
  .requireNamespace("googledrive",
    stopOnFALSE = TRUE,
    messageStart = "to use google drive files"
  )

  if (!is(cloudFolderID, "tbl")) {
    cloudFolderID <- checkAndMakeCloudFolderID(
      cloudFolderID = cloudFolderID, cachePath = cachePath, create = FALSE,
      team_drive = team_drive
    ) # only deals with NULL case
  }

  messageCache("Retrieving file list in cloud folder", verbose = verbose)
  gdriveLs <- retry(quote({
    googledrive::drive_ls(
      path = cloudFolderID, ## TODO: team drives needs a dribble
      pattern = paste0(collapse = "|", c(pattern))
    )
  }))
  if (is(gdriveLs, "try-error")) {
    fnf <- grepl("File not found", gdriveLs)
    if (!fnf) {
      gdriveLs <- retry(quote({
        googledrive::drive_ls(
          path = googledrive::as_id(cloudFolderID), ## TODO: team drives needs a dribble
          pattern = paste0(cloudFolderID, "|", pattern)
        )
      }))
    } else {
      stop("cloudFolderID not found on Gdrive\n", gdriveLs)
    }
  }
  gdriveLs
}


#' Download from cloud, if necessary
#'
#' Meant for internal use, as there are internal objects as arguments.
#'
#' @param newFileName The character string of the local filename that the downloaded object will have
#' @param outputHash The `cacheId` of the object to upload
#' @param gdriveLs The result of `googledrive::drive_ls(googledrive::as_id(cloudFolderID), pattern = "outputHash")`
#' @inheritParams Cache
cloudDownload <- function(outputHash, newFileName, gdriveLs, cachePath, cloudFolderID,
                          drv = getDrv(getOption("reproducible.drv", NULL)),
                          conn = getOption("reproducible.conn", NULL),
                          verbose = getOption("reproducible.verbose")) {
  .requireNamespace("googledrive",
    stopOnFALSE = TRUE,
    messageStart = "to use google drive files"
  )

  messageCache("Downloading cloud copy of ", newFileName, ", with cacheId: ", outputHash,
    verbose = verbose
  )
  isInCloud <- grepl(outputHash, gdriveLs$name)

  outs <- list()
  for (i in 1:1) { # was 1:2 when files were downloaded separately, when file-backed files didn't have cacheId
    localNewFilename <- file.path(tempdir2(), basename2(newFileName))
    outs <- append(outs, lapply(seq_along(isInCloud), function(ind) {
      retry(quote(googledrive::drive_download(
        file = googledrive::as_id(gdriveLs$id[ind]),
        path = localNewFilename[ind], # take first if there are duplicates
        overwrite = TRUE
      )))
    }))
    if (i %in% 1) {
      dtFile <- outs[[1]]$local_path # grep(CacheDBFileSingleExt(), outs$local_path, value = TRUE)
      dt <- loadFile(dtFile, cacheSaveFormat = fileExt(dtFile),
                     cacheId = outputHash, cachePath = cachePath, # in case it needs swapCacheFormat
                     drv = drv, conn = conn, verbose = verbose)
      fromDisk <- extractFromCache(dt, elem = "fromDisk") %in% "TRUE"
      if (all(!fromDisk)) break
      localFilenames <- Filenames(dt)
      # remapFilenames()
      # newFilename2 <- extractFromCache(dt, elem = "origFilename")
      # isInCloud <- seq(newFilename2)
      # gdriveLs <- googledrive::drive_ls(
      #   path = googledrive::as_id(cloudFolderID),
      #   pattern = paste(collapse = "|", newFilename2)
      # )
      # newFilename2 <- newFilename2[match(newFilename2, gdriveLs$name)]
    }
  }
  outs <- rbindlist(outs)

  if (!useDBI()) {
    dtFileInCache <- CacheDBFileSingle(cachePath, cacheId = outputHash)
    suppressMessages(hardLinkOrCopy(dtFile, dtFileInCache))
  }
  objFiles <- grep(CacheDBFileSingleExt(), outs$local_path, value = TRUE, invert = TRUE)
  # objFiles <- grep(paste0(".", formatCheck(cachePath, outputHash)), objFiles, value = TRUE)
  filenamesInCache <- file.path(CacheStorageDir(cachePath = cachePath), basename2(objFiles))
  hardLinkOrCopy(objFiles, to = filenamesInCache)

  if (useDBI()) { # with useDBI = FALSE, the dbFile is already there.
    Map(tv = dt$tagValue, tk = dt$tagKey, function(tv, tk) {
      .addTagsRepo(outputHash, cachePath, tagKey = tk, tagValue = tv, drv = drv, conn = conn)
    })
  }
  inReposPoss <- searchInRepos(
    cachePaths = cachePath, outputHash = outputHash,
    drv = drv, conn = conn
  )
  inReposPoss
}

#' Upload a file to cloud directly from local `cachePath`
#'
#' Meant for internal use, as there are internal objects as arguments.
#'
#' @param isInCloud     A logical indicating whether an `outputHash` is in the cloud already.
#' @param outputToSave  Only required if `any(rasters) == TRUE`.
#'                      This is the `Raster*` object.
#' @param rasters       A logical vector of length >= 1 indicating which elements in
#'                      `outputToSave` are `Raster*` objects.
#' @inheritParams cloudDownload
#'
#' @keywords internal
cloudUploadFromCache <- function(isInCloud, outputHash, cachePath, cloudFolderID,
                                 outputToSave, rasters,
                                 verbose = getOption("reproducible.verbose")) {
  .requireNamespace("googledrive",
    stopOnFALSE = TRUE,
    messageStart = "to use google drive files"
  )
  # browser(expr = exists("._cloudUploadFromCache_1"))

  if (!any(isInCloud)) {
    cacheIdFileName <- CacheStoredFile(cachePath, outputHash, "check", obj = outputToSave)
    if (useDBI()) {
      dt <- showCache(userTags = outputHash)
      td <- tempdir()
      useDBI(FALSE, verbose = -1)
      on.exit(useDBI(TRUE, verbose = -1))
      cacheDB <- CacheDBFileSingle(cachePath = td, outputHash) # put it in a temp location b/c don't want persistent
      on.exit(unlink(cacheDB), add = TRUE)
      if (!dir.exists(dirname(cacheDB))) {
        checkPath(dirname(cacheDB), create = TRUE)
        on.exit(unlink(dirname(cacheDB)), add = TRUE)
      }
      suppress <- saveFilesInCacheFolder(obj = dt, fts = cacheDB, cacheId = outputHash, cachePath = cachePath)
      useDBI(TRUE, verbose = -1)
    } else {
      cacheDB <- CacheDBFileSingle(cachePath, outputHash)
    }
    if (all(file.exists(cacheIdFileName))) {
      newFileName <- basename2(cacheIdFileName)

      cloudFolderID <- checkAndMakeCloudFolderID(cloudFolderID = cloudFolderID, cachePath = cachePath, create = TRUE)

      messageCache("Uploading new cached object -- file(s):\n", paste(newFileName, collapse = "\n"),
                   "\n ... with cacheId: ",
                   outputHash, " to cloud folder id: ", cloudFolderID$name, " or ", cloudFolderID$id,
                   verbose = verbose
      )
      du <- Map(med = cacheIdFileName, nam = newFileName, function(med, nam) {
        try(retry(quote(
          googledrive::drive_upload(
            media = med, path = googledrive::as_id(cloudFolderID),
            name = nam, overwrite = FALSE
          )
        )))
      })

      du2 <- try(retry(quote(googledrive::drive_upload(
        media = cacheDB,
        path = googledrive::as_id(cloudFolderID), name = basename2(cacheDB),
        overwrite = FALSE
      ))))
      if (is(du, "try-error")) {
        return(du)
      }
    } else {
      stop("File(s) to upload are not available")
    }
  }
  # cloudUploadRasterBackends(obj = outputToSave, cloudFolderID)
}

cloudUploadRasterBackends <- function(obj, cloudFolderID) {
  .requireNamespace("googledrive",
    stopOnFALSE = TRUE,
    messageStart = "to use google drive files"
  )

  # browser(expr = exists("._cloudUploadRasterBackends_1"))
  rasterFilename <- Filenames(obj)
  out <- NULL
  if (!is.null(unlist(rasterFilename)) && length(rasterFilename) > 0) {
    allRelevantFiles <- unique(rasterFilename)
    out <- lapply(allRelevantFiles, function(file) {
      try(retry(quote(googledrive::drive_upload(
        media = file, path = cloudFolderID,
        name = basename(file), overwrite = FALSE
      ))))
    })
  }
  return(invisible(out))
}

cloudDownloadRasterBackend <- function(output, cachePath, cloudFolderID,
                                       drv = getDrv(getOption("reproducible.drv", NULL)),
                                       conn = getOption("reproducible.conn", NULL)) {
  .requireNamespace("googledrive", stopOnFALSE = TRUE, messageStart = "to use google drive files")

  if (is(output, "Raster")) {
    rasterFilename <- Filenames(output)
    if (!is.null(unlist(rasterFilename)) && length(rasterFilename) > 0) {
      gdriveLs2 <- NULL
      cacheRepoRasterDir <- file.path(cachePath, "rasters")
      checkPath(cacheRepoRasterDir, create = TRUE)
      simpleFilenames <- unique(filePathSansExt(basename2(unlist(rasterFilename))))
      retry(quote({
        gdriveLs2 <- googledrive::drive_ls(
          path = as_id(cloudFolderID),
          pattern = paste(collapse = "|", simpleFilenames)
        )
      }))

      if (all(simpleFilenames %in% filePathSansExt(gdriveLs2$name))) {
        filenameMismatches <- unlist(lapply(seq_len(NROW(gdriveLs2)), function(idRowNum) {
          localNewFilename <- file.path(cacheRepoRasterDir, basename2(gdriveLs2$name[idRowNum]))
          filenameMismatch <- identical(localNewFilename, rasterFilename)
          retry(quote(googledrive::drive_download(
            file = gdriveLs2[idRowNum, ],
            path = localNewFilename, # take first if there are duplicates
            overwrite = TRUE
          )))
          return(filenameMismatch)
        }))
        if (any(filenameMismatches)) {
          fnM <- seq_along(filenameMismatches)
          if (is(output, "RasterStack")) {
            for (i in fnM[filenameMismatches]) {
              output@layers[[i]]@file@name <- file.path(cacheRepoRasterDir, basename2(rasterFilename)[i])
            }
          } else {
            output@filename <- file.path(cacheRepoRasterDir, basename2(rasterFilename))
          }
        }
      } else {
        warning(
          "Raster backed files are not available in googledrive; \n",
          "will proceed with rerunning code because cloud copy is incomplete"
        )
        output <- NULL
      }
    }
  }
  output
}

isOrHasRaster <- function(obj) {
  rasters <- if (is(obj, "environment")) {
    .requireNamespace("rlang", stopOnFALSE = TRUE)
    if (rlang::inherits_only(obj, "environment")) {
      lapply(mget(ls(obj), envir = obj), function(x) isOrHasRaster(x))
    } else {
      tryCatch(lapply(
        mget(ls(obj), envir = obj@.xData),
        function(x) isOrHasRaster(x)
      ), error = function(x) FALSE)
    }
  } else if (is.list(obj)) {
    lapply(obj, function(x) isOrHasRaster(x))
  } else {
    is(obj, "Raster") || .isSpatRaster(obj)
  }
  return(rasters)
}

isID <- function(x) {
  isTRUE(32 <= nchar(x) && nchar(x) <= 33)
}

test_that("test miscellaneous fns (part 1)", {
  # ONLY RELEVANT FOR RASTERS
  testInit("raster", tmpFileExt = c(".tif", ".grd"))

  expect_type(searchFullEx(), "list")
  expect_true(length(searchFullEx()) > length(search()))
  expect_true(length(searchFullEx()) == (3 + length(search())))

  expect_true(all(unlist(lapply(searchFull(simplify = FALSE), is.environment))))
  test <- lapply(searchFull(simplify = FALSE), attributes)
  test <- grep("withr_handler", value = TRUE, test, invert = TRUE)
  expect_true(all(is.character(unlist(test))))

  # NO LONGER RELIABLE TEST BECAUSE OF NEW REMOVAL OF PACKAGES fEB 24 2021
  # expect_true(sum(unlist(d1)) < sum(unlist(d)))

  # convertRasterPaths
  filenames <- normalizePath(c("/home/user1/Documents/file.txt", "/Users/user1/Documents/file.txt"),
    winslash = "/", mustWork = FALSE
  )
  oldPaths <- dirname(filenames)
  newPaths <- normalizePath(c("/home/user2/Desktop", "/Users/user2/Desktop"),
    winslash = "/", mustWork = FALSE
  )
  expect_true(grepl(newPaths[1], convertPaths(filenames, oldPaths, newPaths)[1]))

  r1 <- raster::raster(system.file("external/test.grd", package = "raster"))
  r2 <- raster::raster(system.file("external/rlogo.grd", package = "raster"))
  rasters <- list(r1, r2)
  oldPaths <- system.file("external", package = "raster")
  newPaths <- file.path("~/rasters")
  rasters <- convertRasterPaths(rasters, oldPaths, newPaths)

  ## spurious failures non-interactively when not sorting
  expect_true(identical(
    sort(unlist(lapply(rasters, raster::filename))),
    sort(normPath(file.path(newPaths, basename(unlist(lapply(list(r1, r2), raster::filename))))))
  ))

  r3 <- suppressWarnings(writeRaster(r1, tmpfile[1], overwrite = TRUE)) ## TODO: raster needs updating for crs stuff
  r4 <- suppressWarnings(convertRasterPaths(tmpfile[1], dirname(tmpfile[1]), newPaths)) ## TODO: raster needs updating for crs stuff

  expect_identical(
    normPath(file.path(newPaths, basename(filename(r4)))),
    normPath(Filenames(r4))
  )

  expect_silent({
    b <- retry(quote(rnorm(1)), retries = 1, silent = TRUE)
  })
  expect_error({
    b <- retry(quote(stop()), retries = 1, silent = TRUE)
  })

  expect_true(identical(NULL, basename2(NULL)))
  a <- .formalsNotInCurrentDots(rnorm, n = 1, b = 2)
  b <- .formalsNotInCurrentDots(rnorm, dots = list(n = 1, b = 2))
  expect_identical(a, b)
})

test_that("objSize and objSizeSession", {
  skip_on_cran()
  # objectSize
  a <- 1
  b <- tempfile()
  saveRDS(a, b)
  expect_true(is.numeric(objSize(asPath(b))))
  expect_true(is(objSize(asPath(b)), "lobstr_bytes"))
})

test_that("setting options works correctly", {
  testInit(verbose = 1, ask = TRUE)

  a <- reproducibleOptions()

  # The keep is during terra-migration
  keep <- setdiff(names(a), c(
    "reproducible.rasterRead",
    "reproducible.cachePath",
    "reproducible.overwrite", # This is a bug # TODO... something prior to this test is changing it
    "reproducible.useDBI",
    "reproducible.cacheSaveFormat",
    "reproducible.shapefileRead"
  ))
  a <- a[keep]

  a1 <- a[sapply(a, function(x) !is.null(x))]
  b <- options()
  # b$reproducible.verbose <- as.numeric(b$reproducible.verbose)
  bbb <- match(names(b), names(a1))
  # expect_true(identical(sort(names(a1)), sort(names(a1[na.omit(bbb)]))))
  expect_true(identical(sort(names(a1)), sort(names(a1[bbb[!is.na(bbb)]]))))
  # omit <- c(names(testInitOut$opts), names(testInitOut$optsAsk),
  #          "reproducible.inputPath", "reproducible.tempPath")
  b1 <- b[names(a1)]
  # b1 <- b1[!names(b1) %in% omit]
  a2 <- a1 # [!names(a1) %in% omit]
  #a1[names(a1) %in% "reproducible.useMemoise"] <- NULL # useMemoise may be TRUE during some tests
  expect_identical(b1, a2)
})

test_that("guessAtTargetAndFun works correctly", {
  testInit("terra")

  # expect_error(.guessAtTargetAndFun(fun = rnorm), "fun must be a")
  expect_message(
    .guessAtTargetAndFun(targetFilePath = NULL, filesExtracted = "", fun = "load"),
    "Don't know which file to load"
  )
  expect_message(
    .guessAtTargetAndFun(targetFilePath = NULL, filesExtracted = "hi.rds", fun = "readRDS"),
    "targetFile was not specified."
  )
  expect_message(
    .guessAtTargetAndFun(targetFilePath = NULL, filesExtracted = c("hi.rds", "hello.rds"), fun = "readRDS"),
    "More than one possible files to load"
  )
})

test_that("unrar is working as expected", {
  testInit("terra", tmpFileExt = c(".tif", ".grd"))

  cat("hi", file = tmpfile[1])
  cat("hi", file = tmpfile[2])
  rarPath <- file.path(tmpdir, "tmp.rar")
  file.create(tmpfile)
  out <- try(utils::zip(zipfile = rarPath, files = tmpfile, flags = "-q")) # this should only be relevant if system can unrar
  if (!is(out, "try-error")) {
    unrar <- .whichExtractFn(archive = rarPath, args = "")
    if (.requireNamespace("archive"))
      expect_true(identical(unrar$fun, archive::archive_extract))
    suppressWarnings(
      expect_error(.callArchiveExtractFn(unrar$fun, files = "", args = list(exdir = tmpCache)))
    )


  }
})

test_that("test miscellaneous fns (part 2)", {
  testInit("terra",
    tmpFileExt = c(".tif", ".grd"),
    needGoogleDriveAuth = TRUE,
    opts = list("reproducible.cloudFolderID" = NULL)
  )
  on.exit(
    {
      try(googledrive::drive_rm(googledrive::as_id(cloudFolderID)), silent = TRUE)
      try(googledrive::drive_rm(googledrive::as_id(tmpCloudFolderID)), silent = TRUE)
    },
    add = TRUE
  )

  ras <- terra::rast(terra::ext(0, 1, 0, 1), resolution = 1, vals = 1)
  ras <- terra::writeRaster(ras, filename = tmpfile[1], overwrite = TRUE)

  skip_if_service_account_releaseVer_NotLinux()

  gdriveLs1 <- data.frame(name = "GADM", id = "sdfsd", drive_resource = list(sdfsd = 1))
  tmpCloudFolderID <- checkAndMakeCloudFolderID(create = TRUE)
  gdriveLs <- driveLs(cloudFolderID = NULL, "sdfsdf")
  expect_true(NROW(gdriveLs) == 0)
  expect_true(is(checkAndMakeCloudFolderID("testy"), "dribble") ||
    is(checkAndMakeCloudFolderID("testy"), "character"))
  cloudFolderID <- checkAndMakeCloudFolderID("testy", create = TRUE)
  testthat::with_mocked_bindings(
    retry = function(..., retries = 1) TRUE,
    {
      if (useDBI()) {
        # Need to convert to cloudUpload from Cache
        mess1 <- capture_messages(
          capture_warnings( # about cache repo -- not the point here
            expect_error(
              cloudUploadFromCache( # outputToSave = ,
                isInCloud = FALSE, outputHash = "sdfsiodfja",
                # gdriveLs = gdriveLs1,
                cloudFolderID = cloudFolderID,
                cachePath = tmpCache
              )
            )
          )
        )
      }
    }
  )

  a <- cloudUploadRasterBackends(ras, cloudFolderID = cloudFolderID)
  mess1 <- capture_messages(expect_error(expect_warning({
    a <- cloudDownload(
      outputHash = "sdfsd", newFileName = "test.tif",
      gdriveLs = gdriveLs1, cloudFolderID = "testy", cachePath = tmpCache
    )
  })))
  expect_true(sum(grepl("Downloading cloud copy of test\\.tif", mess1)) == 1)
  testthat::with_mocked_bindings(
    retry = function(..., retries = 1) TRUE,
    {
      # cloudFolderID can't be meaningless "character", but retry is TRUE
      warns <- capture_warnings({
        err <- capture_error({
          cloudDownloadRasterBackend(output = ras, cachePath = tmpCache, cloudFolderID = "character")
        })
      })
      expect_true(is.null(err))
    }
  )

  a <- new.env(parent = emptyenv())
  a$a <- list(ras, ras)
  expect_true(all(unlist(isOrHasRaster(a))))
})

test_that("Filenames for environment", {
  testInit(c("terra"),
    tmpFileExt = c(".tif", ".grd", ".tif", ".tif", ".grd"),
    opts = list("reproducible.ask" = FALSE)
  )

  s <- new.env(parent = emptyenv())
  s$r <- terra::rast(terra::ext(0, 10, 0, 10), vals = 1, resolution = 1)
  s$r2 <- terra::rast(terra::ext(0, 10, 0, 10), vals = 1, resolution = 1)
  s$r <- suppressWarningsSpecific(
    terra::writeRaster(s$r, filename = tmpfile[1], overwrite = TRUE),
    "NOT UPDATED FOR PROJ >= 6"
  )
  s$r2 <- suppressWarningsSpecific(
    terra::writeRaster(s$r2, filename = tmpfile[3], overwrite = TRUE),
    "NOT UPDATED FOR PROJ >= 6"
  )
  s$s <- c(s$r, s$r2)
  s$b <- terra::writeRaster(s$s, filename = tmpfile[5], overwrite = TRUE)

  Fns <- Filenames(s)

  fnsGrd <- unlist(normPath(Filenames(s$b)))
  expect_true(identical(c(Fns[["b1"]], Fns[["b2"]]), fnsGrd))
  expect_true(identical(Fns[["r"]], normPath(Filenames(s$r))))
  expect_true(identical(Fns[["r2"]], normPath(Filenames(s$r2))))
  expect_true(identical(
    c(Fns[["s1"]], Fns[["s2"]]),
    sapply(seq_len(nlayers2(s$s)), function(rInd) normPath(Filenames(s$s[[rInd]])))
  ))

  FnsR <- Filenames(s$r)
  expect_true(identical(FnsR, normPath(Filenames(s$r))))

  FnsS <- Filenames(s$s)
  expect_true(identical(FnsS, sapply(
    seq_len(nlayers2(s$s)),
    function(rInd) normPath(Filenames(s$s[[rInd]]))
  )))

  FnsB <- Filenames(s$b)
  expect_true(identical(FnsB, fnsGrd))

  # Another stack with identical files
  rlogoFiles <- system.file("ex/test.grd", package = "terra")
  rlogoFiles <- c(rlogoFiles, gsub("grd$", "gri", rlogoFiles))
  secondSet <- file.path(tmpdir, c("one.grd", "one.gri"))
  res <- suppressWarnings(file.link(rlogoFiles, secondSet))
  if (all(res)) {
    b <- c(terra::rast(rlogoFiles[1]), terra::rast(secondSet[1]))
    expect_true(identical(sort(normPath(c(rlogoFiles, secondSet))), sort(Filenames(b))))
  }

  # Test duplicated filenames in same Stack
  b <- c(terra::rast(rlogoFiles[1]), terra::rast(rlogoFiles[1]))
  expect_true(identical(sort(normPath(c(rlogoFiles))), unique(sort(Filenames(b, allowMultiple = TRUE)))))

  rlogoFiles <- system.file("ex/test.grd", package = "terra")
  rlogoDir <- dirname(rlogoFiles)
  b <- terra::rast(rlogoFiles)
  rlogoFiles <- c(rlogoFiles, gsub("grd$", "gri", rlogoFiles))
  expect_true(identical(
    sort(normPath(dir(pattern = "test", rlogoDir, full.names = TRUE))),
    sort(Filenames(b))
  ))
})

test_that("test miscellaneous fns (part 3)", {
  testInit(opts = list(datatable.print.class = FALSE))

  x1 <- append(as.list(c(0, 1, -1, 10^(-(1:10)))), as.list(c(0L, 1L)))
  a <- lapply(x1, roundTo6Dec)

  # Keeps class
  expect_true(all(unlist(lapply(seq_along(x1), function(y) identical(class(x1[[y]]), class(a[[y]]))))))
  whBig <- which(x1 >= 10e-7)
  expect_true(identical(x1[whBig], a[whBig]))
  whSmall <- which(abs(unlist(x1)) < 10e-7 & unlist(x1) != 0)
  expect_false(all(unlist(lapply(whSmall, function(ws) identical(x1[[ws]], a[[ws]])))))
  whWhole <- which(unlist(x1) %% 1 != unlist(x1))
  expect_true(all(unlist(lapply(whWhole, function(ws) identical(x1[[ws]], a[[ws]])))))
  whZero <- which(unlist(x1) == 0)
  expect_true(all(unlist(lapply(whZero, function(ws) identical(x1[[ws]], a[[ws]])))))

  out <- capture_messages(messageDF(cbind(a = 1.1232), round = 2))
  out <- strsplit(out, "\n")[[1]] # the prev line is all on one line now (Dec 2023), with \n separating
  expect_true(is.character(out))
  expect_identical(length(out), 2L) ## TODO: only passes when run line by line interactively
  expect_true(is.numeric(as.numeric(gsub("\033.*", "", gsub(".*: ", "", out)[2]))))

  out <- capture_messages(messageDF(cbind(a = 1.1232), round = 2, colnames = FALSE))
  expect_true(is.character(out))
  expect_identical(length(out), 1L) ## TODO: only passes when run line by line interactively

  expect_true(is.numeric(as.numeric(gsub("\033.*", "", gsub(".*: ", "", out)))))


  out <- capture_messages(messageDF(1.1232, round = 2, colnames = TRUE))
  out <- strsplit(out, "\n")[[1]] # the prev line is all on one line now (Dec 2023), with \n separating
  expect_true(is.character(out))
  expect_identical(length(out), 2L) ## TODO: only passes when run line by line interactively
  expect_true(is.numeric(as.numeric(gsub("\033.*", "", gsub(".*: ", "", out)[2]))))
})

test_that("test set.randomseed", {
  testInit()

  N <- 1e4
  a <- integer(N)
  for (i in 1:N) {
    a[i] <- set.randomseed()
  }
  expect_false(any(duplicated(a)))
})

library(testthat)

# NOTE: ALL OPTIONS ARE BEING SET IN tests/testthat/setup.R

## Test matrix: 5 combinations of {raster, terra} × {USE_DBI, no DBI} +
## qs cache format. Running them sequentially in one R session was the
## documented behavior, but on ubuntu CI it triggered SIGTERM mass-kills
## (cumulative memory pressure from 5 test_check passes loading namespaces
## + raster/SF objects without releasing between iterations). Bisect on
## branch `bisect-ubuntu-fail` confirmed the SINGLE-pass full suite
## passes; only the 5-pass version wedges. Default to single pass; opt
## back into the full matrix via R_REPRO_TEST_FULL_MATRIX=true (e.g.
## locally or on a separate workflow) when you want the full coverage.
if (nzchar(Sys.getenv("R_REPRO_TEST_FULL_MATRIX")) &&
    as.logical(Sys.getenv("R_REPRO_TEST_FULL_MATRIX"))) {
  ## Full 5-combo matrix (opt-in)
  Sys.setenv(R_REPRODUCIBLE_RASTER_READ = "raster::raster")
  Sys.setenv(R_REPRODUCIBLE_USE_DBI = "false")
  test_check("reproducible")

  Sys.setenv(R_REPRODUCIBLE_RASTER_READ = "raster::raster")
  Sys.setenv(R_REPRODUCIBLE_USE_DBI = "true")
  test_check("reproducible")

  Sys.setenv(R_REPRODUCIBLE_RASTER_READ = "terra::rast") ## default
  Sys.setenv(R_REPRODUCIBLE_USE_DBI = "false") ## default
  test_check("reproducible")

  Sys.setenv(R_REPRODUCIBLE_RASTER_READ = "terra::rast")
  Sys.setenv(R_REPRODUCIBLE_USE_DBI = "true")
  test_check("reproducible")

  Sys.setenv(R_REPRODUCIBLE_CACHESAVEFORMAT = "qs") ## default
  Sys.setenv(R_REPRODUCIBLE_USE_DBI = "false") ## default
  test_check("reproducible")

} else {
  ## Single pass at defaults (terra + USE_DBI=FALSE + qs format).
  ## Covers ~95% of the test surface; alternate raster/DBI paths are
  ## still exercised by individual test files that explicitly opt in.
  test_check("reproducible")
}

library(data.table)

origDTthreads <- setDTthreads(2)

## Detect "this is a local dev run, not CI/CRAN" so we can flip a few
## opt-in switches on by default. CRAN/CI already skip via skip_on_cran() /
## skip_on_ci(); the runLargeFileTests gate is in addition to those for
## tests that need the 20–30 minute integration path.
onCI   <- isTRUE(as.logical(Sys.getenv("CI", "false")))
onCRAN <- !nzchar(Sys.getenv("NOT_CRAN")) &&
          !isTRUE(as.logical(Sys.getenv("NOT_CRAN", "false")))
wantMoreTests <- isInteractive() || Sys.info()[["user"]] %in% c("emcintir")

if (wantMoreTests) { # this is for covr::package_coverage
  Sys.setenv(NOT_CRAN = "true")
  # Sys.setenv(SKIP_GAUTH = "true")
}

opts <- options(
  ## Default: run the multi-minute large-file tests when this is a local
  ## (non-CI, non-CRAN) dev run. User can override either way:
  ##   options(reproducible.runLargeFileTests = TRUE  / FALSE)
  reproducible.runLargeFileTests = isTRUE(
    getOption("reproducible.runLargeFileTests", wantMoreTests && !onCI && !onCRAN)
  ),
  warnPartialMatchArgs = TRUE, # This gives false positives for `raster::stack`
  warnPartialMatchAttr = TRUE,
  warnPartialMatchDollar = TRUE,
  reproducible.useCacheV3 = !isFALSE(getOption("reproducible.useCacheV3"))#,
  #reproducible.useDBI = FALSE  # done TF  = c(691, 764.1), TT c(793,874), FF = c(875.5s, 876s), FT = c(1024.8,934)
)

# if (Sys.info()["user"] %in% "emcintir") {
#   opts2 <- options(gargle_oauth_email = "predictiveecology@gmail.com")
#   secretDir <- if (isWindows()) "C:/Eliot/.secret" else "~/.secret"
#   opts2 <- append(options(gargle_oauth_cache = secretDir), opts2)
#   if (requireNamespace("googledrive"))
#     googledrive::drive_auth()
#   opts <- append(opts, opts2)
# }


# uses eliot-githubauthentication@genial-cycling-408722.iam.gserviceaccount.com
# Whatever files are on googledrive must be shared with this google service account
if (isNamespaceLoaded("googledrive"))
  if ((!googledrive::drive_has_token())) {
    gauthEnv <- Sys.getenv("GOOGLEDRIVE_AUTH")
    if (nzchar(gauthEnv)) {
      if (file.exists(gauthEnv))
        ## see helper-allEqual.R: tolerate revoked / rotated service-account keys
        tryCatch(googledrive::drive_auth(path = gauthEnv),
                 error = function(e) invisible(NULL))
    }
  }


withr::defer(
  {
    if (wantMoreTests) {
      print(paste0("getOption('reproducible.cacheSaveFormat') = '", getOption("reproducible.cacheSaveFormat"), "'"))
      print(paste0("getOption('reproducible.useCacheV3') = ", getOption("reproducible.useCacheV3")))
      print(paste0("getOption('reproducible.rasterRead') = ", getOption("reproducible.rasterRead")))
      print(paste0("getOption('reproducible.runLargeFileTests') = ", getOption("reproducible.runLargeFileTests")))
      print(paste0("getOption('reproducible.useDBI') = ", getOption("reproducible.useDBI")))
      print(paste0("getOption('reproducible.useMemoise') = ", getOption("reproducible.useMemoise")))
      Sys.setenv(NOT_CRAN = "")
      Sys.setenv(SKIP_GAUTH = "")
    }
    options(opts)
    data.table::setDTthreads(origDTthreads)
    try(reproducible::clearCache(ask = FALSE, verbose = -1))
    try(unlink("CHECKSUMS.txt"), silent = TRUE) # comes from an unknown place
  },
  teardown_env()
)

if (wantMoreTests) {
  print(paste0("getOption('reproducible.cacheSaveFormat') = '", getOption("reproducible.cacheSaveFormat"), "'"))
  print(paste0("getOption('reproducible.useCacheV3') = ", getOption("reproducible.useCacheV3")))
  print(paste0("getOption('reproducible.rasterRead') = ", getOption("reproducible.rasterRead")))
  print(paste0("getOption('reproducible.runLargeFileTests') = ", getOption("reproducible.runLargeFileTests")))
  print(paste0("getOption('reproducible.useDBI') = ", getOption("reproducible.useDBI")))
  print(paste0("getOption('reproducible.useMemoise') = ", getOption("reproducible.useMemoise")))
}

rebuildColors <- function(x, origColors) {
  if (isTRUE(all(origColors$origMinValue != minFn(x)) || all(origColors$origMaxValue != maxFn(x)) ||
    !identical(.getColors(x)[[1]], origColors$origColors))) {
    colorSequences <- unlist(lapply(seq_along(origColors$origMinValue), function(ind) {
      origColors$origMinValue[ind]:origColors$origMaxValue[ind]
    }))
    if (isTRUE(length(origColors$origColors) == length(colorSequences))) {
      newSeq <- minFn(x):maxFn(x)
      oldSeq <- origColors$origMinValue:origColors$origMaxValue
      whFromOld <- match(newSeq, oldSeq)
      x@legend@colortable <- origColors$origColors[whFromOld]
    }
  }
  x
}

.getColors <- function(object) {
  if (.isSpatRaster(object)) {
    cols <- terra::coltab(object)
  } else {
    nams <- names(object)
    cols <- lapply(nams, function(x) {
      as.character(object[[x]]@legend@colortable)
    })
    names(cols) <- nams
  }
  return(cols)
}

#' Move a file to a new location -- Defunct -- use `hardLinkOrCopy`
#'
#' This will first try to `file.rename`, and if that fails, then it will
#' `file.copy` then `file.remove`.
#' @param from,to character vectors, containing file names or paths.
#' @param overwrite logical indicating whether to overwrite destination file if it exists.
#' @export
#' @return Logical indicating whether operation succeeded.
#'
.file.move <- function(from, to, overwrite = FALSE) {
  .Deprecated("hardLinkeOrCopy")
  hardLinkOrCopy(from, to, overwrite)
  file.remove(from)
}

#' Recursive copying of nested environments, and other "hard to copy" objects
#'
#' When copying environments and all the objects contained within them, there are
#' no copies made: it is a pass-by-reference operation. Sometimes, a deep copy is
#' needed, and sometimes, this must be recursive (i.e., environments inside
#' environments).
#'
#' @details
#' To create a new Copy method for a class that needs its own method, try something like
#' shown in example and put it in your package (or other R structure).
#'
#'
#' @param object  An R object (likely containing environments) or an environment.
#'
#' @param filebackedDir A directory to copy any files that are backing R objects,
#'                      currently only valid for `Raster` classes. Defaults
#'                      to `.reproducibleTempPath()`, which is unlikely to be very useful.
#'                      Can be `NULL`, which means that the file will not be
#'                      copied and could therefore cause a collision as the
#'                      pre-copied object and post-copied object would have the same
#'                      file backing them.
#'
#' @param ... Only used for custom Methods
#'
#' @author Eliot McIntire
#' @export
#' @importFrom data.table copy
#' @rdname Copy
#' @return
#' The same object as `object`, but with pass-by-reference class elements "deep" copied.
#' `reproducible` has methods for several classes.
#'
#' @seealso [.robustDigest()], [Filenames()]
#'
#' @examples
#' e <- new.env()
#' e$abc <- letters
#' e$one <- 1L
#' e$lst <- list(W = 1:10, X = runif(10), Y = rnorm(10), Z = LETTERS[1:10])
#' ls(e)
#'
#' # 'normal' copy
#' f <- e
#' ls(f)
#' f$one
#' f$one <- 2L
#' f$one
#' e$one ## uh oh, e has changed!
#'
#' # deep copy
#' e$one <- 1L
#' g <- Copy(e)
#' ls(g)
#' g$one
#' g$one <- 3L
#' g$one
#' f$one
#' e$one
#' ## To create a new deep copy method, use the following template
#' ## setMethod("Copy", signature = "the class", # where = specify here if not in a package,
#' ##           definition = function(object, filebackendDir, ...) {
#' ##           # write deep copy code here
#' ##           })
#'
setGeneric("Copy", function(object, ...) {
  standardGeneric("Copy")
})

#' @rdname Copy
#' @inheritParams Cache
setMethod(
  "Copy",
  signature(object = "ANY"),
  definition = function(object, filebackedDir,
                        drv = getDrv(getOption("reproducible.drv", NULL)),
                        conn = getOption("reproducible.conn", NULL),
                        verbose = getOption("reproducible.verbose"),
                        ...) {
    out <- object # many methods just do a pass through
    if (any(grepl("DBIConnection", is(object)))) {
      messageCache("Copy will not do a deep copy of a DBI connection object; no copy being made. ",
                   "This may have unexpected consequences...",
                   verbose = verbose
      )
    } else if (is(object, "proto")) { # don't want to import class for reproducible package; an edge case
      out <- get(class(object)[1])(object)
    } else if (inherits(object, "SQLiteConnection")) {
      con <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
      messageCache("Making a copy of the entire SQLite database: ", object@dbname,
                   "; this may not be desireable ...",
                   verbose = verbose
      )
      out <- RSQLite::sqliteCopyDatabase(object, con)
    } else if (!identical(is(object)[1], "environment") && is.environment(object)) {
      # keep this environment method here, as it will intercept "proto"
      #   and other environments that it shouldn't
      messageCache("Trying to do a deep copy (Copy) of object of class ", class(object),
                   ", which does not appear to be a normal environment. If it can be copied ",
                   "like a normal environment, ignore this message; otherwise, may need to create ",
                   "a Copy method for this class. See ?Copy",
                   verbose = verbose
      )
    } else if (is.environment(object)) {
      listVersion <- Copy(as.list(object, all.names = TRUE),
                          filebackedDir = filebackedDir,
                          drv = drv, conn = conn, verbose = verbose, ...
      )

      parentEnv <- parent.env(object)
      out <- new.env(parent = parentEnv)
      list2env(listVersion, envir = out)
      attr(out, "name") <- attr(object, "name")
    } else if (inherits(object, "Raster")) {
      if (any(nchar(Filenames(object)) > 0)) {
        if (missing(filebackedDir)) {
          filebackedDir <- tempdir2(rndstr(1, 11))
        }
        if (!is.null(filebackedDir)) {
          out <- .prepareFileBackedRaster(object, repoDir = filebackedDir, drv = drv, conn = conn,
                                          verbose = verbose)
        }
      }
    } else if (.isSpatRaster(object)) {
      fns <- Filenames(object, allowMultiple = FALSE)
      nz <- nzchar(fns)
      if (any(nz)) {
        fns <- fns[nz]
        fnsAll <- Filenames(object, allowMultiple = TRUE)
        hadNumeric <- grepl("_[:0-9:]+$", tools::file_path_sans_ext(fnsAll))
        needNewFn <- FALSE
        if (!missing(filebackedDir)) {
          if (!is.null(filebackedDir)) {
            needNewFn <- TRUE
          }
        }
        if (needNewFn) {
          if (!isAbsolutePath(filebackedDir)) {
            # relative
            filebackedDir <- file.path(getwd(), filebackedDir)
          }
          areAbs <- isAbsolutePath(fnsAll)
          fnsAllBase <- fnsAll
          if (any(areAbs)) {
            fnsAllBase[areAbs] <- reproducible::basename2(fnsAll[areAbs])
          }
          newFns <- file.path(filebackedDir, fnsAllBase)
        } else {
          newFns <- sapply(fnsAll, nextNumericName)
        }
        hasNumeric <- grepl("_[:0-9:]+$", tools::file_path_sans_ext(newFns))
        copyFile(fnsAll, newFns)
        # Copy may have given "nextNumericName"
        fnsBase <- tools::file_path_sans_ext(basename(fns))
        if (any(hadNumeric)) fnsBase <- gsub("_[:0-9:]+$", "", fnsBase)

        newFnsSingles <- newFns[Map(
          sn = fnsBase,
          fns1 = fns, function(sn, fns1) {
            which(startsWith(basename(newFns), sn) &
              endsWith(basename(newFns), tools::file_ext(fns1)))
          }
        ) |> unlist()]
        # newFnsSingles <- newFns[match(tools::file_path_sans_ext(basename(fns)), basename(newFns))]
        out <- terra::rast(newFnsSingles)
        if (length(nz) == 1) { # one file for all layers
          names(out) <- names(object)
        } else {
          names(out) <- names(object[[nz]])
        }


        # If there are layers that were in RAM; need to add them back, in correct order
        if (any(!nz)) {
          memoryLayers <- names(object)[!nz]
          out[[memoryLayers]] <- object[[memoryLayers]]
          out <- out[[match(names(object), names(out))]]
        }
      }
    }
    return(out)
  }
)

#' @rdname Copy
setMethod("Copy",
  signature(object = "data.table"),
  definition = function(object, ...) {
    data.table::copy(object)
  }
)

#' @rdname Copy
setMethod("Copy",
  signature(object = "list"),
  definition = function(object, ...) {
    lapply(object, function(x) {
      Copy(x, ...)
    })
  }
)

#' @rdname Copy
setMethod("Copy",
  signature(object = "refClass"),
  definition = function(object, ...) {
    if (exists("copy", envir = object)) {
      object$copy()
    } else {
      stop(
        "There is no method to copy this refClass object; ",
        "see developers of reproducible package"
      )
    }
  }
)

#' @rdname Copy
setMethod("Copy",
  signature(object = "data.frame"),
  definition = function(object, ...) {
    object
  }
)

#' Fast spatial subsetting of Cloud Optimized GeoTiff (COG) files
#'
#' An alternative fast-path inside [prepInputs()] for remote Cloud Optimized GeoTiffs.
#' When a URL points to a COG and the user has specified a spatial subsetting argument
#' (`to`, `cropTo`, or `maskTo`), this function reads only the spatial window of interest
#' via GDAL's `/vsicurl/` virtual filesystem — no full-file download is needed.
#' The returned `SpatRaster` is passed back to `prepInputs` where the normal
#' `postProcess` step (mask, reproject, write) completes the pipeline.
#'
#' This function is called automatically from inside [prepInputs()] when
#' `getOption("reproducible.useCOG")` is `TRUE` (the default). It can also be
#' called directly.
#'
#' @param url Character. An HTTP(S) URL pointing to a GeoTiff file.
#' @param verbose Numeric or Logical. Verbosity level.
#' @param ... Passed through; expected to contain at least one of `to`, `cropTo`,
#'   or `maskTo` (a spatial object defining the area of interest).
#'
#' @return A `SpatRaster` windowed to the bounding box of the `to`/`cropTo`/`maskTo`
#'   object (in the COG's own CRS), or the character string `"NULL"` if any
#'   pre-condition fails (not HTTP, no spatial arg, network error, empty window, etc.).
#'
#' @seealso [prepInputs()], [prepInputsWithTiles()]
#' @export
prepInputsCOG <- function(url,
                          verbose = getOption("reproducible.verbose", 1),
                          ...) {

  # ---- Pre-conditions: fast exits before any network access ----------------
  if (is.null(url) || !grepl("^https?://", url))
    return("NULL")

  # Strip query/fragment, then require a GeoTiff-style extension. This avoids
  # firing `/vsicurl/` reads against archives (.zip, .tar, .gz, ...) or other
  # non-raster URLs, which would emit a GDAL warning before failing.
  url_path <- sub("[?#].*$", "", url)
  if (!grepl("\\.(tif|tiff|cog|gtiff)$", url_path, ignore.case = TRUE))
    return("NULL")

  dots   <- list(...)
  to_obj <- if (!is.null(dots$to)) dots$to else if (!is.null(dots$cropTo)) dots$cropTo else dots$maskTo
  if (is.null(to_obj))
    return("NULL")

  if (!.requireNamespace("terra", stopOnFALSE = FALSE))
    return("NULL")

  if (!internetExists())
    return("NULL")

  vsicurl_path <- paste0("/vsicurl/", url)

  # ---- Open metadata-only remote connection --------------------------------
  r_meta <- tryCatch(
    terra::rast(vsicurl_path),
    error = function(e) NULL
  )
  if (is.null(r_meta))
    return("NULL")

  # ---- Reproject to_obj extent into COG's CRS for the window read ----------
  # Both `terra::crs()` and `postProcessTo()` can fail on a system where PROJ
  # is misconfigured (e.g. missing proj.db on macOS), in ways that range from
  # an empty CRS string to an unhandled C-level error. Wrap both, and bail
  # out to the regular `prepInputs` download path on any failure.
  cog_crs <- tryCatch(terra::crs(r_meta), error = function(e) NA_character_)
  if (is.na(cog_crs) || !nzchar(cog_crs))
    return("NULL")
  to_reproj <- tryCatch(
    postProcessTo(to_obj, to = cog_crs, verbose = verbose - 2),
    error = function(e) NULL
  )
  if (is.null(to_reproj))
    return("NULL")
  to_ext <- tryCatch(terra::ext(to_reproj), error = function(e) NULL)
  if (is.null(to_ext))
    return("NULL")

  # ---- Window-crop: GDAL fetches only tile blocks intersecting to_ext ------
  r_windowed <- tryCatch({
    terra::window(r_meta) <- to_ext
    terra::crop(r_meta, to_ext)
  }, error = function(e) NULL)

  if (is.null(r_windowed) || terra::ncell(r_windowed) == 0L)
    return("NULL")

  messagePreProcess(
    "prepInputsCOG: windowed remote read complete (tiled GeoTiff); full postProcess will follow",
    verbose = verbose
  )

  r_windowed
}

utils::globalVariables(c(
  ".SD", "..colsToKeep", "Account", "availableOnCRAN", "availableOnGitHub", "AvailableVersion",
  "Branch", "compareVersionAvail", "compareVersionAvailGH", "correctVersion", "correctVersionAvail",
  "correctVersionAvailGH", "fullGit", "githubPkgName", "inequality", "instPkgs", "isGH",
  "minVersion", "Package", "Repo", "RepoWBranch", "Version", "versionOnGH"
))


#' `NA`-aware comparison of two vectors
#'
#' Copied from
#' <http://www.cookbook-r.com/Manipulating_data/Comparing_vectors_or_factors_with_NA/>.
#' This function returns `TRUE` wherever elements are the same, including `NA`'s,
#' and `FALSE` everywhere else.
#'
#' @export
#' @param v1 A vector
#' @param v2 A vector
#' @return
#' A logical vector, indicating positions where two vectors are same or differ.
#'
#' @examples
#' a <- c(NA, 1, 2, NA)
#' b <- c(1, NA, 2, NA)
#' compareNA(a, b)
#'
compareNA <- function(v1, v2) {
  same <- (v1 == v2) | (is.na(v1) & is.na(v2))
  same[is.na(same)] <- FALSE
  return(same)
}


# internetExists <- function() {
#   if (requireNamespace("RCurl", quietly = TRUE)) {
#     out <- RCurl::url.exists("www.google.com")
#   } else {
#     messagePrepInputs(RCurlMess)
#     out <- urlExists("https://www.google.com")
#     # if (.Platform$OS.type == "windows") {
#     #   ipmessage <- system("ipconfig", intern = TRUE)
#     # } else {
#     #   ipmessage <- system("ifconfig", intern = TRUE)
#     # }
#     # validIP <- "((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)[.]){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
#     # linesWithIP <- grep(validIP, ipmessage)
#     # out <- if (length(linesWithIP)) {
#     #   any(linesWithIP) && all(!grepl("127.0.0.1", ipmessage[linesWithIP]))
#     # } else {
#     #   FALSE
#     # }
#     # out
#   }
#   return(out)
# }

RCurlMess <- paste0(
  "install.packages('RCurl') may give a more reliable detection ",
  "of internet connection"
)

# This allows mocking in the tests; do not delete
.readline <- function(prompt) {
  readline(prompt)
}

#' Search up the full scope for functions
#'
#' @description
#' This is like `base::search` but when used inside a function, it will
#' show the full scope (see figure in the section *Binding environments*
#' on <http://adv-r.had.co.nz/Environments.html>).
#' This full search path will be potentially much longer than
#' just `search()` (which always starts at `.GlobalEnv`).
#'
#' `searchFullEx` shows an example function that is inside this package
#' whose only function is to show the Scope of a package function.
#'
#' @param env The environment to start searching at. Default is
#'                calling environment, i.e., `parent.frame()`
#' @param simplify Logical. Should the output be simplified to character,
#'                 if possible (usually it is not possible because environments
#'                 don't always coerce correctly)
#' @return
#' A list of environments that is the actual search path, unlike `search()`
#' which only prints from `.GlobalEnv` up to `base` through user attached
#' packages.
#'
#' @export
#' @rdname search
#' @seealso [base::search()]
#'
#' @examples
#' seeScope <- function() {
#'   searchFull()
#' }
#' seeScope()
#' searchFull()
searchFull <- function(env = parent.frame(), simplify = TRUE) {
  envs <- list()
  counter <- 0
  while (!identical(parent.env(env), emptyenv())) {
    counter <- counter + 1
    envs[[counter]] <- parent.env(env)
    env <- envs[[counter]]
  }
  if (simplify) {
    lapply(envs, function(x) if (!is.null(attr(x, "name"))) environmentName(x) else x)
  } else {
    envs
  }
}

#' @details
#' `searchFullEx` can be used to show an example of the use of `searchFull`.
#'
#' @export
#' @examples
#' searchFullEx()
#' @rdname search
searchFullEx <- function() {
  searchFull()
}

utils::globalVariables(c(
  "datatype", "expectedFile", "method", "needBuffer", "objName",
  "rasterToMatch", "studyArea", "targetCRS", "to", "touches", "useSAcrs", "V1"
))

#' Download and optionally post-process files
#'
#' \if{html}{\figure{lifecycle-maturing.svg}{options: alt="maturing"}}
#'
#' This function can be used to prepare R objects from remote or local data sources.
#' The object of this function is to provide a reproducible version of
#' a series of commonly used steps for getting, loading, and processing data.
#' This function has two stages: Getting data (download, extracting from archives,
#' loading into R) and post-processing (for `Spatial*` and `Raster*`
#' objects, this is crop, reproject, mask/intersect).
#' To trigger the first stage, provide `url` or `archive`.
#' To trigger the second stage, provide `studyArea` or `rasterToMatch`.
#' See examples.
#'
#' @note This function is still experimental: use with caution.
#'
#' @section Stage 1 - Getting data:
#'
#' See [preProcess()] for combinations of arguments.
#'
#'   \enumerate{
#'     \item Download from the web via either `googledrive::drive_download()`,
#'     [utils::download.file()];
#'     \item Extract from archive using [unzip()] or [untar()];
#'     \item Load into R using `terra::rast`,
#'     `sf::st_read`, or any other function passed in with `fun`;
#'     \item Checksumming of all files during this process. This is put into a
#'     \file{CHECKSUMS.txt} file in the `destinationPath`, appending if it is
#'     already there, overwriting the entries for same files if entries already exist.
#'  }
#'
#' @section Stage 2 - Post processing:
#'
#'   This will be triggered if either `rasterToMatch` or `studyArea`
#'   is supplied.
#'
#'   \enumerate{
#'     \item Fix errors. Currently only errors fixed are for `SpatialPolygons`
#'     using `buffer(..., width = 0)`;
#'     \item Crop using [cropTo()];
#'     \item Project using [projectTo()];
#'     \item Mask using [maskTo()];
#'     \item write the file to disk via [writeTo()].
#'    }
#'
#'   NOTE: checksumming does not occur during the post-processing stage, as
#'   there are no file downloads. To achieve fast results, wrap
#'   `prepInputs` with `Cache`.
#'
#'   NOTE: `sf` objects are still very experimental.
#'
#' \subsection{postProcessing of `Spat*`, `sf`, `Raster*` and `Spatial*` objects:}{
#'
#'   The following has been DEPRECATED because there are a sufficient number of
#'   ambiguities that this has been changed in favour of `from` and the `*to` family.
#'   See [postProcessTo()].
#'
#'   DEPRECATED: If `rasterToMatch` or `studyArea` are used, then this will
#'   trigger several subsequent functions, specifically the sequence,
#'   *Crop, reproject, mask*, which appears to be a common sequence while
#'   preparing spatial data from diverse sources.
#'   See [postProcess()] documentation section on
#'   *Backwards compatibility with `rasterToMatch` and/or `studyArea` arguments*
#'   to understand various combinations of `rasterToMatch` and/or `studyArea`.
#'  }
#'
#'
#' @section `fun`:
#'
#'  `fun` offers the ability to pass any custom function with which to load
#'  the file obtained by `preProcess` into the session. There are two cases that are
#'  dealt with: when the `preProcess` downloads a file (including via `dlFun`),
#'  `fun` must deal with a file; and, when `preProcess` creates an R object
#'  (e.g., raster::getData returns an object), `fun` must deal with an object.
#'
#'  `fun` can be supplied in three ways: a function, a character string
#'   (i.e., a function name as a string), or an expression.
#'   If a character string or function, is should have the package name e.g.,
#'   `"terra::rast"` or as an actual function, e.g., `base::readRDS`.
#'   In these cases, it will evaluate this function call while passing `targetFile`
#'   as the first argument. These will only work in the simplest of cases.
#'
#'   When more precision is required, the full call can be written and where the
#'   filename can be referred to as `targetFile` if the function
#'   is loading a file. If `preProcess` returns an object, `fun` should be set to
#'   `fun = NA`.
#'
#'   If there is a custom function call, is not in a package, `prepInputs` may not find it. In such
#'   cases, simply pass the function as a named argument (with same name as function) to `prepInputs`.
#'   See examples.
#'   NOTE: passing `fun = NA` will skip loading object into R. Note this will essentially
#'   replicate the functionality of simply calling `preProcess` directly.
#'
#' @section `purge`:
#'
#' In options for control of purging the `CHECKSUMS.txt` file are:
#'
#'   \describe{
#'     \item{`0`}{keep file}
#'     \item{`1`}{delete file in `destinationPath`, all records of downloads need to be rebuilt}
#'     \item{`2`}{delete entry with same `targetFile`}
#'     \item{`4`}{delete entry with same `alsoExtract`}
#'     \item{`3`}{delete entry with same `archive`}
#'     \item{`5`}{delete entry with same `targetFile` & `alsoExtract`}
#'     \item{`6`}{delete entry with same `targetFile`, `alsoExtract` & `archive`}
#'     \item{`7`}{delete entry that same `targetFile`, `alsoExtract` & `archive` & `url`}
#'   }
#' will only remove entries in the `CHECKSUMS.txt` that are associated with
#'   `targetFile`, `alsoExtract` or `archive` When `prepInputs` is called,
#'   it will write or append to a (if already exists) `CHECKSUMS.txt` file.
#'   If the `CHECKSUMS.txt` is not correct, use this argument to remove it.
#'
#' @param targetFile Character string giving the filename (without relative or
#'   absolute path) to the eventual file
#'   (raster, shapefile, csv, etc.) after downloading and extracting from a zip
#'   or tar archive. This is the file *before* it is passed to
#'   `postProcess`. The internal checksumming does not checksum
#'   the file after it is `postProcess`ed (e.g., cropped/reprojected/masked).
#'   Using `Cache` around `prepInputs` will do a sufficient job in these cases.
#'   See table in [preProcess()].
#'
#' @param archive Optional character string giving the path of an archive
#'   containing `targetFile`, or a vector giving a set of nested archives
#'   (e.g., `c("xxx.tar", "inner.zip", "inner.rar")`). If there is/are (an) inner
#'   archive(s), but they are unknown, the function will try all until it finds
#'   the `targetFile`. See table in [preProcess()]. If it is `NA`,
#'   then it will *not* attempt to see it as an archive, even if it has archive-like
#'   file extension (e.g., `.zip`). This may be useful when an R function
#'   is expecting an archive directly.
#'
#' @param url Optional character string indicating the URL to download from.
#'   If not specified, then no download will be attempted. If not entry
#'   exists in the `CHECKSUMS.txt` (in `destinationPath`), an entry
#'   will be created or appended to. This `CHECKSUMS.txt` entry will be used
#'   in subsequent calls to
#'   `prepInputs` or `preProcess`, comparing the file on hand with the ad hoc
#'   `CHECKSUMS.txt`. See table in [preProcess()].
#'
#' @param alsoExtract Optional character string naming files other than
#'   `targetFile` that must be extracted from the `archive`. If
#'   `NULL`, the default, then it will extract all files. Other options:
#'   `"similar"` will extract all files with the same filename without
#'   file extension as `targetFile`. `NA` will extract nothing other
#'   than `targetFile`. A character string of specific file names will cause
#'   only those to be extracted. Each element may also be a regular expression:
#'   if an element does not match any archive member literally (by relative path
#'   or basename), it is passed to `grep()` against the archive's file list and
#'   all matching members are extracted. For example,
#'   `alsoExtract = "CMD_sm|CMD_sp"` extracts every file whose name contains
#'   `CMD_sm` or `CMD_sp`. See table in [preProcess()].
#'
#' @param destinationPath Character string of a directory in which to download
#'   and save the file that comes from `url` and is also where the function
#'   will look for `archive` or `targetFile`. NOTE (still experimental):
#'   To prevent repeated downloads in different locations, the user can also set
#'   `options("reproducible.destinationPathShared")` to one or more local file paths to
#'   search for the file before attempting to download. Default for that option is
#'   `NULL` meaning do not search locally. The previous name
#'   `options("reproducible.inputPaths")` is still accepted as a backwards-compatible
#'   alias.
#'
#' @param fun Optional. If specified, this will attempt to load whatever
#'   file was downloaded during `preProcess` via `dlFun`. This can be either a
#'   function (e.g., sf::st_read), character string (e.g., "base::load"),
#'   NA (for no loading, useful if `dlFun` already loaded the file) or
#'   if extra arguments are required
#'   in the function call, it must be a call naming
#'   `targetFile` (e.g., `sf::st_read(targetFile, quiet = TRUE)`)
#'   as the file path to the file to load. See details and examples below.
#'
#' @param quick Logical. This is passed internally to [Checksums()]
#'   (the quickCheck argument), and to
#'   [Cache()] (the quick argument). This results in faster, though
#'   less robust checking of inputs. See the respective functions.
#'
#' @param purge Logical or Integer. `0/FALSE` (default) keeps existing `CHECKSUMS.txt` file and
#'   `prepInputs` will write or append to it. `1/TRUE` will deleted the entire `CHECKSUMS.txt` file.
#'    Other options, see details.
#'
#' @param overwrite Logical. Passed to `writeTo` (possibly inside `postProcess`) and `postProcess`.
#'
#' @param ... Additional arguments passed to
#'   [postProcess()] and [reproducible::Cache()].
#'  Since `...` is passed to [postProcess()], these will
#'  `...` will also be passed into the inner
#'  functions, e.g., [cropInputs()]. Possibly useful other arguments include
#'  `dlFun` which is passed to `preProcess`. See details and examples.
#'
#' @param useCache Passed to `Cache` in various places.
#'   Defaults to `getOption("reproducible.useCache", 2L)` in `prepInputs`, and
#'   `getOption("reproducible.useCache", FALSE)` if calling any of the inner
#'   functions manually. For `prepInputs`, this mean it will use `Cache`
#'   only up to 2 nested levels, which includes `preProcess`. `postProcess` and
#'   its nested `*Input` functions (e.g., `cropInputs`, `projectInputs`,
#'   `maskInputs`) are no longer internally cached, as `terra` processing speeds
#'   mean internal caching is more time consuming. We recommend caching the full
#'   `prepInputs` call instead (e.g. `prepInputs(...) |> Cache()`).
#'
#' @param .tempPath Optional temporary path for internal file intermediate steps.
#'   Will be cleared `on.exit` from this function.
#'
#' @inheritParams Cache
#' @author Eliot McIntire, Jean Marchal, and Tati Micheletti
#' @export
#' @return
#' This is an omnibus function that will return an R object that will have resulted from
#' the running of [preProcess()] and [postProcess()] or [postProcessTo()]. Thus,
#' if it is a GIS object, it may have been cropped, reprojected, "fixed", masked, and
#' written to disk.
#'
#' @importFrom data.table data.table
#' @importFrom digest digest
#' @importFrom methods is
#' @importFrom utils methods modifyList
#' @include checksums.R download.R postProcess.R
#' @rdname prepInputs
#' @seealso [postProcessTo()], [downloadFile()], [extractFromArchive()],
#'          [postProcess()].
#' @examples
#' \donttest{
#' if (requireNamespace("terra", quietly = TRUE) &&
#'   requireNamespace("withr", quietly = TRUE)) {
#'   library(reproducible)
#'   withr::local_dir(withr::local_tempdir())
#'   # Make a dummy study area map -- user would supply this normally
#'   coords <- structure(c(-122.9, -116.1, -99.2, -106, -122.9, 59.9, 65.7, 63.6, 54.8, 59.9),
#'     .Dim = c(5L, 2L)
#'   )
#'   studyArea <- terra::vect(coords, "polygons")
#'   terra::crs(studyArea) <- "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84 +towgs84=0,0,0"
#'   # Make dummy "large" map that must be cropped to the study area
#'   outerSA <- terra::buffer(studyArea, 50000)
#'   terra::crs(outerSA) <- "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84 +towgs84=0,0,0"
#'   tf <- normPath(file.path(tempdir2(), "prepInputs2.shp"))
#'   terra::writeVector(outerSA, tf)
#'
#'   # run prepInputs -- load file, postProcess it to the studyArea
#'
#'   studyArea2 <- prepInputs(
#'     targetFile = tf, to = studyArea,
#'     fun = "terra::vect",
#'     destinationPath = tempdir2()
#'   ) |>
#'     suppressWarnings() # not relevant warning here
#'
#'   # clean up
#'   unlink("CHECKSUMS.txt")
#'
#'   ##########################################
#'   # Remote file using `url`
#'   ##########################################
#'   if (internetExists()) {
#'     data.table::setDTthreads(2)
#'     origDir <- getwd()
#'     # download a zip file from internet, unzip all files, load as shapefile, Cache the call
#'     # First time: don't know all files - prepInputs will guess, if download file is an archive,
#'     #   then extract all files, then if there is a .shp, it will load with sf::st_read
#'     dPath <- file.path(tempdir(), "ecozones")
#'     shpUrl <- "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip"
#'
#'     # Wrapped in a try because this particular url can be flaky
#'     shpEcozone <- try(prepInputs(
#'       destinationPath = dPath,
#'       url = shpUrl
#'     ))
#'     if (!is(shpEcozone, "try-error")) {
#'       # Robust to partial file deletions:
#'       unlink(dir(dPath, full.names = TRUE)[1:3])
#'       shpEcozone <- prepInputs(
#'         destinationPath = dPath,
#'         url = shpUrl
#'       )
#'       unlink(dPath, recursive = TRUE)
#'
#'       # Once this is done, can be more precise in operational code:
#'       #  specify targetFile, alsoExtract, and fun, wrap with Cache
#'       ecozoneFilename <- file.path(dPath, "ecozones.shp")
#'       ecozoneFiles <- c(
#'         "ecozones.dbf", "ecozones.prj",
#'         "ecozones.sbn", "ecozones.sbx", "ecozones.shp", "ecozones.shx"
#'       )
#'       shpEcozone <- prepInputs(
#'         targetFile = ecozoneFilename,
#'         url = shpUrl,
#'         fun = "terra::vect",
#'         alsoExtract = ecozoneFiles,
#'         destinationPath = dPath
#'       )
#'       unlink(dPath, recursive = TRUE)
#'
#'       # Add a study area to Crop and Mask to
#'       # Create a "study area"
#'       coords <- structure(c(-122.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
#'         .Dim = c(5L, 2L)
#'       )
#'       studyArea <- terra::vect(coords, "polygons")
#'       terra::crs(studyArea) <- "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84 +towgs84=0,0,0"
#'
#'       #  specify targetFile, alsoExtract, and fun, wrap with Cache
#'       ecozoneFilename <- file.path(dPath, "ecozones.shp")
#'       # Note, you don't need to "alsoExtract" the archive... if the archive is not there, but the
#'       #   targetFile is there, it will not redownload the archive.
#'       ecozoneFiles <- c(
#'         "ecozones.dbf", "ecozones.prj",
#'         "ecozones.sbn", "ecozones.sbx", "ecozones.shp", "ecozones.shx"
#'       )
#'       shpEcozoneSm <- Cache(prepInputs,
#'         url = shpUrl,
#'         targetFile = reproducible::asPath(ecozoneFilename),
#'         alsoExtract = reproducible::asPath(ecozoneFiles),
#'         studyArea = studyArea,
#'         fun = "terra::vect",
#'         destinationPath = dPath,
#'         writeTo = "EcozoneFile.shp"
#'       ) # passed to determineFilename
#'
#'       terra::plot(shpEcozone[, 1])
#'       terra::plot(shpEcozoneSm[, 1], add = TRUE, col = "red")
#'       unlink(dPath)
#'     }
#'   }
#'   withr::deferred_run()
#' }
#' }
#'
#' ## Using quoted dlFun and fun -- this is not intended to be run but used as a template
#' ## prepInputs(..., fun = customFun(x = targetFile), customFun = customFun)
#' ##   # or more complex
#' ##  test5 <- prepInputs(
#' ##   targetFile = targetFileLuxRDS,
#' ##   dlFun =
#' ##     getDataFn(name = "GADM", country = "LUX", level = 0) # preProcess keeps file from this!
#' ##   ,
#' ##   fun = {
#' ##     out <- readRDS(targetFile)
#' ##     sf::st_as_sf(out)}
#' ##  )
prepInputs <- function(targetFile = NULL, url = NULL, archive = NULL, alsoExtract = NULL,
                       destinationPath = getOption("reproducible.destinationPath", "."),
                       fun = NULL,
                       quick = getOption("reproducible.quick"),
                       overwrite = getOption("reproducible.overwrite", FALSE),
                       purge = FALSE,
                       useCache = getOption("reproducible.useCache", 2),
                       .tempPath,
                       verbose = getOption("reproducible.verbose", 1),
                       ...) {
  .callingEnv <- parent.frame()
  messagePreProcess("Running ", .messageFunctionFn("prepInputs"), verbose = verbose, verboseLevel = 0)
  .message$IndentUpdate()
  stStart <- Sys.time()
  if (missing(.tempPath)) {
    .tempPath <- tempdir2(rndstr(1, 6))
    on.exit(
      {
        unlink(.tempPath, recursive = TRUE)
      },
      add = TRUE
    )
  }
  funCaptured <- substitute(fun)
  ## When fun is a bare variable name (e.g. fun = myFun), substitute() captures
  ## the symbol rather than the value. Force the R promise directly: this resolves
  ## the symbol in the frame where the promise was created (the caller's frame),
  ## which works correctly even when prepInputs is invoked via Cache's call chain.
  if (is.name(funCaptured))
    funCaptured <- fun
  if (is.character(url) && length(url) == 0L) url <- NULL
  prepInputsAssertions(environment())

  runNormalPreProcess <- TRUE

  # COG fast-path: triggered when url is HTTP(S), a spatial subsetting arg is
  # present, and the remote file is confirmed as a Cloud Optimized GeoTiff.
  if (isTRUE(getOption("reproducible.useCOG", TRUE)) &&
      !is.null(url) && !isNULLorNA(url) &&
      grepl("^https?://", url) &&
      (!is.null(list(...)$to) || !is.null(list(...)$cropTo) || !is.null(list(...)$maskTo))) {
    x_cog <- prepInputsCOG(url = url, verbose = verbose, ...)
    if (!identical(x_cog, "NULL")) {
      x <- x_cog
      runNormalPreProcess <- FALSE
    }
  }

  rpiut <- getOption("reproducible.prepInputsUrlTiles")
  if (runNormalPreProcess &&
      !(isNULLorNA(rpiut) || rpiut %in% FALSE) && (
    !is.null(list(...)$to) || !is.null(list(...)$cropTo) || !is.null(list(...)$maskTo)
  )) {
    if (!is.null(url) && isGoogleDriveDirectory(rpiut)) {
      message("Using prepInputsWithTiles because `to` is supplied and \n",
              "options(reproducible.prepInputsUrlTiles) is set to a Google Drive folder")
      x <- prepInputsWithTiles(url = url, destinationPath = destinationPath, purge = purge,
                               ...)
      if (!identical(x, "NULL")) runNormalPreProcess <- FALSE
    }

  }

  if (runNormalPreProcess) {

    ##################################################################
    # preProcess
    ##################################################################
    out <- preProcess(
      targetFile = targetFile,
      url = url,
      archive = archive,
      alsoExtract = alsoExtract,
      destinationPath = destinationPath,
      fun = funCaptured,
      quick = quick,
      overwrite = overwrite,
      purge = purge,
      useCache = useCache,
      .tempPath = .tempPath,
      verbose = verbose,
      .callingEnv = .callingEnv,
      ...
    )

    ##################################################################
    # Load object to R
    ##################################################################
    # preProcess() returns out$fun = NULL when auto-guessing failed (e.g.
    # ambiguous archive contents, .shp missing from alsoExtract). The
    # download/extract layer is happy with that, but prepInputs() is the
    # loader and a NULL fun would silently produce an un-loaded result —
    # surprising and almost always a user error. If the caller explicitly
    # opted out of loading (`fun = NA`), respect that; otherwise stop early
    # with a message that points at the fix.
    if (is.null(out$fun) && !isTRUE(is.na(funCaptured))) {
      stop(
        "prepInputs cannot determine which function to use to load the result. ",
        "The file extension is ambiguous and no `fun` / `targetFile` was supplied. ",
        "Either specify `targetFile` and/or `fun`, or pass `fun = NA` to skip loading.",
        call. = FALSE
      )
    }
    x <- process(out,
                 funCaptured = funCaptured,
                 useCache = useCache, verbose = verbose, .callingEnv = .callingEnv, ...
    )
  }

  ##################################################################
  # postProcess
  ##################################################################
  needPostProcess <- ...names() %in% c(
    "studyArea", "rasterToMatch", "targetCRS", "to", "cropTo",
    "maskTo", "projectTo", "fixErrorsIn", "useSAcrs", "writeTo"
  )
  if (any(needPostProcess)) {
    TopoErrors <- list() # eventually to update a Google ID #TODO
    x <- withCallingHandlers(
      postProcessTo(from = x, ..., destinationPath = destinationPath, overwrite = overwrite),
      message = function(m) {
        hasTopoExcError <- grepl("TopologyException: Input geom 0 is invalid", m$message)
        if (any(hasTopoExcError)) {
          TopoErrors <<- append(TopoErrors, list(m$message))
        }
      }
    )
  }
  .message$IndentRevert()
  stFinal <- reportTime(stStart, mess = paste0(.messageFunctionFn("prepInputs"), " done; took "), minSeconds = 10)
  # if (getOption("reproducible.savePrepInputsState", FALSE))
  #   savePrepInputsState(url, archive, out, stFinal, sysCalls = sys.calls())
  return(x)
}

#' Extract files from archive
#'
#' Extract zip or tar archive files, possibly nested in other zip or tar archives.
#'
#' @param archive Character string giving the path of the archive
#' containing the `file` to be extracted. This path must exist or be `NULL`
#'
#' @param destinationPath Character string giving the path where `neededFiles` will be
#' extracted. Defaults to the archive directory.
#'
#' @param neededFiles Character string giving the name of the file(s) to be extracted.
#'
#' @param extractedArchives Used internally to track archives that have been extracted from.
#' @param filesExtracted Used internally to track files that have been extracted.
#' @param checkSums A checksums file, e.g., created by Checksums(..., write = TRUE)
#' @param needChecksums A numeric, with `0` indicating do not write a new checksums,
#'                      `1` write a new one,
#'                      `2` append new information to existing one.
#' @param checkSumFilePath The full path to the checksum.txt file
#' @param quick Passed to `Checksums`
#' @param ... Passed to `unzip` or `untar`, e.g., `overwrite`
#' @inheritParams prepInputs
#'
#' @return A character vector listing the paths of the extracted archives.
#'
#' @author Jean Marchal and Eliot McIntire
#'
extractFromArchive <- function(archive,
                               destinationPath = getOption("reproducible.destinationPath", dirname(archive)),
                               neededFiles = NULL, extractedArchives = NULL, checkSums = NULL,
                               needChecksums = 0, filesExtracted = character(),
                               checkSumFilePath = character(), quick = FALSE,
                               verbose = getOption("reproducible.verbose", 1),
                               .tempPath, ...) {
  if (!is.null(archive)) {
    if (!(any(c(knownArchiveExtensions) %in% fileExt(archive)))) {
      stop(
        "Archives of type ", fileExt(archive), " are not currently supported. ",
        "Try extracting manually then placing extracted files in ", destinationPath
      )
    }
  }
  if (!is.null(archive) && !is.null(neededFiles)) {
    neededFiles <- setdiff(neededFiles, archive)
  }
  # neededFiles <- setdiff(basename2(neededFiles), basename2(archive))
  if (length(neededFiles) == 0) neededFiles <- NULL
  result <- if (!is.null(neededFiles)) {
    checkSums[checkSums$expectedFile %in% makeRelative(neededFiles, destinationPath), ]$result
  } else {
    "NotOK"
  }
  extractedObjs <- list(filesExtracted = character())
  # needs to pass checkSums & have all neededFiles files
  neededFilesRel <- makeRelative(neededFiles, destinationPath)
  hasAllFiles <- if (NROW(checkSums)) {
    all(neededFilesRel %in% checkSums$expectedFile) # need basename2 for comparison with checkSums
  } else {
    FALSE
  }

  for (dontUse in dontUses) {
    tryC <- try({
      if (!(all(compareNA(result, "OK")) && hasAllFiles)) {
        if (!is.null(archive)) {
          if (!file.exists(archive[1])) {
            stop(
              "No archive exists with filename: ", archive[1],
              ". Please pass an archive name to a path that exists"
            )
          }
          args <- list(archive[1], exdir = destinationPath[1])
          funWArgs <- .whichExtractFn(archive[1], args, dontUse = dontUse)

          # need to deal with \\ vs. / and also needs to stay relative
          filesInArchive <- makeRelative(.listFilesInArchive(archive), destinationPath)
          if (is.null(neededFiles)) {
            neededFiles <- filesInArchive
          }

          neededFiles <- checkRelative(neededFiles, absolutePrefix = destinationPath, filesInArchive)
          neededFilesRel <- makeRelative(neededFiles, destinationPath) # neededFiles may have been changed
          neededFiles <- makeAbsolute(neededFiles, destinationPath)
          # need to re-Checksums because
          checkSums <- .checkSumsUpdate(
            destinationPath = destinationPath,
            newFilesToCheck = neededFiles,
            checkSums = checkSums,
            checkSumFilePath = checkSumFilePath
          )
          # Compute result AFTER the update so it reflects current disk state.
          # Previously computed before .checkSumsUpdate, so files just extracted
          # would show NROW(result) == 0, forcing a spurious re-extraction attempt.
          result <- if (NROW(checkSums)) {
            checkSums[checkSums$expectedFile %in% neededFilesRel, ]$result
          } else {
            logical(0)
          }

          # isOK will have "directories" so it will be longer than neededFiles
          isOK <- if (!is.null(checkSums)) {
            .compareChecksumsAndFilesAddDirs(checkSums, neededFiles, destinationPath)
          } else {
            FALSE
          }

          # recheck, now that we have the whole file list
          if (!(all(isOK)) || NROW(result) == 0) {
            # don't extract if we already have all files and they are fine

            # use binary addition -- 1 is new file, 2 is append
            if (needChecksums == 0) needChecksums <- 2
            filesInArchiveAbs <- makeAbsolute(filesInArchive, destinationPath)
            if (length(archive) > 1) {
              filesExtracted <- c(
                filesExtracted,
                .callArchiveExtractFn(funWArgs$fun, funWArgs$args,
                                      absolutePrefix = destinationPath, archive = archive,
                                      files = basename2(archive[2]), .tempPath = .tempPath
                )
              )
              # recursion, removing one archive
              extractedObjs <- extractFromArchive(
                archive[-1],
                destinationPath = destinationPath,
                neededFiles = neededFiles,
                extractedArchives = extractedArchives,
                checkSums = checkSums,
                quick = quick,
                needChecksums = needChecksums,
                checkSumFilePath = checkSumFilePath,
                filesExtracted = filesExtracted,
                verbose = verbose,
                .tempPath = .tempPath
              )
            } else if (any(neededFiles %in% filesInArchiveAbs) || is.null(neededFiles)) {
              possibleFolders <- dir.exists(filesInArchive)
              if (sum(possibleFolders)) {
                filesInArchive <- setdiff(filesInArchive, possibleFolders)
              }
              neededFilesRel <- if (is.null(neededFiles)) {
                NULL
              } else {
                if (!is.null(names(isOK))) {
                  names(isOK)[!isOK]
                } else {
                  makeRelative(neededFiles[!isOK], destinationPath)
                }
              }
              filesToExtractNow <- intersect(filesInArchive, neededFilesRel)
              dt <- data.table(files = filesToExtractNow)
              # extractingTheseFiles <- paste(filesToExtractNow, collapse = "\n")
              # extractingTheseFiles <- paste(basename2(filesInArchive[basename2(filesInArchive) %in%
              #                                                         neededFiles]), collapse = ", ")
              # if (!any(nzchar(filesToExtractNow)))
              #   extractingTheseFiles <- paste0("all files: ",
              #                                  paste(filesInArchive, collapse = "\n"))
              messagePreProcess("From:\n", archive[1], "  \n", "Extracting", verbose = verbose)
              messageDF(dt, indent = .message$PreProcessIndent, verbose = verbose, colour = getOption("reproducible.messageColourPrepInputs"))
              filesExtracted <- c(
                filesExtracted,
                .callArchiveExtractFn(funWArgs$fun,
                                      funWArgs$args,
                                      absolutePrefix = destinationPath,
                                      files = filesToExtractNow,
                                      archive = archive,
                                      .tempPath = .tempPath
                )
              )
            } else {
              # don't have a 2nd archive, and don't have our neededFiles file
              # isArchive <- grepl(fileExt(filesInArchive), pattern = "(zip|tar|rar)", ignore.case = TRUE)
              isArchive <- grepl(fileExt(filesInArchive),
                                 pattern = paste0("(", paste(knownArchiveExtensions, collapse = "|"), ")"), ignore.case = TRUE
              )

              if (any(isArchive)) {
                arch <- makeRelative(filesInArchive[isArchive], destinationPath)
                filesExtracted <- c(
                  filesExtracted,
                  .callArchiveExtractFn(funWArgs$fun, funWArgs$args,
                                        files = arch,
                                        absolutePrefix = destinationPath,
                                        archive = archive,
                                        .tempPath = .tempPath
                  )
                )
                filesExtracted <- unique(filesExtracted) # maybe unnecessary

                prevExtract <- lapply(makeAbsolute(arch, destinationPath), function(ap) {
                  extractFromArchive(
                    archive = ap, destinationPath = destinationPath,
                    neededFiles = neededFiles,
                    extractedArchives = extractedArchives,
                    filesExtracted = filesExtracted,
                    checkSums = checkSums,
                    needChecksums = needChecksums,
                    checkSumFilePath = checkSumFilePath,
                    quick = quick,
                    .tempPath = .tempPath
                  )
                })

                extractedArchives <- c(prevExtract[[1]]$extractedArchives, extractedArchives)
                filesExtracted <- unique(c(prevExtract[[1]]$filesExtracted, filesExtracted))
              }
            }
          } else {
            messagePreProcess("Skipping extractFromArchive: all files already present", verbose = verbose)
            filesExtracted <- checkSums[checkSums$expectedFile %in%
                                          makeRelative(filesInArchive, destinationPath), ]$expectedFile
            filesExtracted <- makeAbsolute(filesInArchive, destinationPath)
          }
        }
      } else {
        if (!is.null(archive)) { # if archive is null, it means there was no archive passed
          messagePreProcess("Skipping extractFromArchive: all needed ",
                            "files now present",
                            verbose = verbose
          )
        }
        filesExtracted <- setdiff(neededFiles, if (!is.null(archive)) makeRelative(archive, destinationPath))
      }
    })
    if (!is(tryC, "try-error")) {
      break
    }
    if (identical(dontUse, tail(dontUses, 1)[[1]])) {
      stop(tryC)
    }
  }

  list(
    extractedArchives = c(extractedArchives, archive),
    neededFiles = neededFiles, # these may have been corrected for user supplying incorrect basename path
    filesExtracted = unique(c(filesExtracted, extractedObjs$filesExtracted)),
    needChecksums = needChecksums,
    checkSums = checkSums
  )
}

#' Try to pick a file to load
#'
#' @keywords internal
#' @rdname guessAtTarget
#' @name guessAtTarget
#' @importFrom utils unzip untar
#' @importFrom stats setNames
#' @inheritParams postProcess
#' @param filesExtracted A character vector of all files that have been extracted (e.g.,
#'                       from an archive)
#' @param destinationPath Full path of the directory where the target file should be
#' @keywords internal
.guessAtTargetAndFun <- function(targetFilePath,
                                 destinationPath = getOption("reproducible.destinationPath", "."),
                                 filesExtracted, fun = NULL, verbose = getOption("reproducible.verbose", 1)) {
  # fun = NA means "don't load anything" — skip all guessing and messaging.
  # Guard is.na() with is.atomic + length-1 so a user-supplied language object
  # (e.g. `fun = sf::st_read(targetFile)`) doesn't trip "is.na() applied to
  # non-(list or vector) of type 'language'".
  if (is.atomic(fun) && length(fun) == 1L && isTRUE(is.na(fun)))
    return(list(targetFilePath = targetFilePath, fun = fun))
  if (all(!is.na(targetFilePath))) {
    # Drop OS-injected archive metadata from auto-discovered candidates so they aren't
    # picked when no targetFile is specified. Covers macOS (__MACOSX/*, ._* AppleDouble,
    # .DS_Store) and Windows (Thumbs.db, desktop.ini). If the user explicitly named such
    # a file via targetFilePath, it is preserved via the union below.
    if (length(filesExtracted)) {
      bn <- basename(filesExtracted)
      isOSMeta <- grepl("(^|/)__MACOSX(/|$)", filesExtracted) |
        startsWith(bn, "._") |
        bn %in% c(".DS_Store", "Thumbs.db", "desktop.ini")
      filesExtracted <- filesExtracted[!isOSMeta]
    }
    possibleFiles <- unique(c(targetFilePath, filesExtracted))
    whichPossFile <- possibleFiles %in% targetFilePath
    if (isTRUE(any(whichPossFile))) {
      possibleFiles <- possibleFiles[whichPossFile]
    }
    isShapefile <- FALSE
    isRaster <- FALSE
    isRDS <- FALSE
    fileExt <- fileExt(possibleFiles)
    feKnown <- .fileExtsKnown() # An object in helpers.R

    funPoss <- lapply(fileExt, function(fe) feKnown[startsWith(prefix = feKnown[[1]], fe), ])
    funPoss <- do.call(rbind, funPoss)

    if (NROW(funPoss) == 0 && length(possibleFiles)) {
      funPoss <- checkSFWebPage(funPoss, fileExt, feKnown, verbose)
    }

    if (length(funPoss)) {
      isShapefile <- fileExt %in% funPoss[funPoss[, "type"] == vectorType(), "extension"]
      isRaster <- fileExt %in% funPoss[funPoss[, "type"] == rasterType(), "extension"]
      isRDS <- fileExt %in% funPoss[funPoss[, "extension"] == .rdsFormat, "extension"]
      if (any(isShapefile)) {
        if (is.null(fun)) {
          if (requireNamespace("sf", quietly = TRUE)) {
            if (!isTRUE(grepl("st_read", fun))) {
              messagePreProcess(
                "Using sf::st_read on shapefile because sf package is available; to force old ",
                "behaviour with 'raster::shapefile' use fun = 'raster::shapefile' or ",
                "options('reproducible.shapefileRead' = 'raster::shapefile')"
              )
            }
          }
        }
      }
    }
    if (is.null(fun)) {
      fun <- unique(funPoss[, "fun"])
      if (length(fun) > 1) {
        if (sum(isRaster) > 0 && sum(isShapefile) > 0) {
          isRaster[isRaster] <- FALSE
          funPoss <- funPoss[funPoss$type == vectorType(), ]
          fun <- unique(funPoss[, "fun"])
          message("The archive has both a shapefile and a raster; selecting the shapefile. If this is incorrect, specify targetFile")
        } else {
          stop(
            "more than one file; can't guess at function to load with; ",
            "please supply 'fun' or 'targetFile' argument to reduce ambiguity"
          )
        }
      }
      # If we can't guess `fun`, leave it NULL. preProcess() does not load the
      # object, so this is fine when called directly. prepInputs()'s process()
      # already handles a NULL fun gracefully by returning targetFilePath.
      if (length(fun) == 0) fun <- NULL
    }
    if (is.null(targetFilePath) || length(targetFilePath) == 0) {
      secondPartOfMess <- if (any(isShapefile)) {
        c(
          " Trying ", fun, " on ", paste(possibleFiles[isShapefile], collapse = ", "), ".",
          " If that is not correct, please specify a different targetFile",
          " and/or fun."
        )
      } else if (is.null(fun)) {
        c(" Also, file extension does not unambiguously specify how it should be loaded. Please specify fun.")
      } else {
        c(
          " Trying ", fun, ".\n",
          " If that is not correct, please specify a targetFile",
          " and/or different fun. The current files in the destinationPath",
          " are: \n",
          paste(possibleFiles, collapse = "\n")
        )
      }
      messagePreProcess(c("targetFile was not specified.", secondPartOfMess), verbose = verbose)

      targetFilePath <- if (is.null(fun)) {
        NULL
      } else if (length(possibleFiles[isShapefile]) > 0) {
        possibleFiles[isShapefile]
      } else {
        if (any(isRaster)) {
          possibleFiles[isRaster]
        } else if (any(isRDS)) {
          possibleFiles[isRDS]
        } else {
          messagePreProcess("Don't know which file to load. Please specify targetFile.", verbose = verbose)
        }
      }
      if (length(targetFilePath) > 1) {
        messagePreProcess("More than one possible files to load:\n", verbose = verbose)
        if (length(targetFilePath) > 100) {
          filesForMess <- data.table(Extracted = targetFilePath)
          messageDF(filesForMess, indent = .message$PreProcessIndent, verbose = verbose)
        } else {
          filesForMess <- paste(targetFilePath, collapse = "\n")
          messagePreProcess(filesForMess)
        }
        messagePreProcess("Picking the last one. If not correct, specify a targetFile.", verbose = verbose)
        targetFilePath <- targetFilePath[length(targetFilePath)]
      }
    }
  }
  list(targetFilePath = targetFilePath, fun = fun)
}

#' @importFrom utils untar unzip
.whichExtractFn <- function(archive, args, dontUse = NULL) {
  out <- NULL
  if (!(is.null(archive))) {
    if (!is.na(archive)) {
      ext <- tolower(fileExt(archive))
      if (!ext %in% knownArchiveExtensions) {
        stop(
          "preProcess can only deal with archives with following extensions:\n",
          paste(knownArchiveExtensions, collapse = ", ")
        )
      }
      canUseArchive <- .requireNamespace("archive")
      mustUseArchive <- !(ext %in% knownInternalArchiveExtensions)
      useArchive <- (mustUseArchive && canUseArchive || canUseArchive)
      sevenZ <- ""
      unrar <- ""
      if (mustUseArchive && canUseArchive %in% FALSE) {
        sevenZ <- Sys.which("7z")
        unrar <- Sys.which("unrar")
        doStop <- TRUE
        if (ext %in% "rar") {
          if (nchar(sevenZ) || nzchar(unrar)) {
            mustUseArchive <- FALSE
            canUseArchive <- FALSE
            doStop <- FALSE
          }
        }

        if (isTRUE(doStop))
          stop(.message$stopNeedArchive(archive))
      }
      if (useArchive && .requireNamespace("archive") && !isTRUE(any(dontUse %in% "archive"))) {
        fun <- archive::archive_extract
      } else { # base R or system call functions
        if (ext == "zip" && !isTRUE(any(dontUse %in% "unzip"))) {
          fun <- unzip
          args <- c(args, list(junkpaths = FALSE))
        } else if (ext %in% c("tar", "tar.gz", "gz") && !isTRUE(any(dontUse %in% "untar"))) {
          fun <- untar
        } else { # system only
          if (ext == "rar" && !isTRUE(any(dontUse %in% "unrar")) && nzchar(unrar)) {
            fun <- "unrar"
          } else if (nzchar(sevenZ)) { # if (ext == "7z") {
            fun <- "7z"
          } else {
            stop(.message$stopNeedArchive(archive))
          }
        }
      }
      out <- list(fun = fun, args = args)
    }
  }
  return(out)
}

#' @keywords internal
#' @importFrom utils capture.output
.callArchiveExtractFn <- function(fun, args, files, overwrite = TRUE,
                                  absolutePrefix = getOption("reproducible.destinationPath", "."),
                                  archive = "",
                                  verbose = getOption("reproducible.verbose", 1), .tempPath) {
  argList <- list(files = files)
  argList$files <- makeRelative(argList$files, absolutePrefix)

  if (missing(.tempPath)) {
    .tempPath <- tempdir2(rndstr(1, 6))
    on.exit(
      {
        unlink(.tempPath, recursive = TRUE)
      },
      add = TRUE
    )
  }

  if (is.character(fun)) {
    if (!fun %in% knownSystemArchiveExtensions) {
      fun <- eval(fun)
    }
  }

  origExdir <- args$exdir
  if (!is.null(args$exdir)) {
    args$exdir <- .tempPath
  }

  if (is.character(fun)) {
    messagePreProcess(
      paste0("The archive cannot be opened using R tools. Trying a system call to ", fun),
      verbose = verbose)
    extractSystemCallPath <- .testForArchiveExtract(archive)
    if (grepl(x = extractSystemCallPath, pattern = "7z")) {
      prependPath <- if (isWindows()) {
        paste0("\"", extractSystemCallPath, "\"")
      } else {
        extractSystemCallPath
      }

      # This spits out a message on non-Windows about arguments that are ignored
      suppressMessages({
        output <- system(paste0(prependPath, " x -aoa -o\"", .tempPath, "\" \"",
                                args[[1]], "\""),
                         wait = TRUE,
                         ignore.stdout = FALSE,
                         ignore.stderr = FALSE,
                         invisible = TRUE,
                         show.output.on.console = FALSE, intern = TRUE
        )
      })
    } else {
      system(paste0("unrar x ", args[[1]], " ", .tempPath), wait = TRUE, ignore.stdout = TRUE)
    }
    # list of full paths of all extracted files!
    # extractedFiles <- list.files(path = .tempPath, recursive = TRUE, include.dirs = TRUE)
    # internalFolders <- extractedFiles[fileExt(extractedFiles) == ""]
    # extractedFiles <- setdiff(x = extractedFiles, y = internalFolders)
  } else {
    # Try the direct, then indirect
    isUnzip <- if (identical(unzip, fun)) TRUE else ("overwrite" %in% names(formals(fun)))
    argList <- if (isUnzip) {
      c(argList, overwrite = overwrite)
    } else {
      c(argList)
    }

    # Start the extracting, starting with `archive`
    worked <- FALSE
    if (.requireNamespace("archive", stopOnFALSE = FALSE)) {
      #system.time(
        extractedFiles <- archive::archive_extract(args[[1]], args$exdir, argList$files)
      #  )
      listOfFilesExtracted <- extractedFiles <- list.files(
        path = .tempPath,
        # list of full paths of all extracted files!
        recursive = TRUE,
        all.files = TRUE,
        include.dirs = TRUE
      )

      worked <- all(extractedFiles %in% listOfFilesExtracted)
    }

    if (!worked) {
      if (exists("listOfFilesExtracted", inherits = FALSE))
        rm(listOfFilesExtracted)

      opt <- options("warn")$warn
      on.exit(options(warn = opt), add = TRUE)
      options(warn = 1)
      tooBig <- file.size(args[[1]]) > 5e9
      worked <- FALSE
      if (isUnzip && !tooBig) {
        fattrs <- unzip(args[[1]], list = TRUE)
        ids <- which(fattrs[["Name"]] %in% argList$files)
        tooBig <- any(fattrs[ids, ]["Length"][[1]] >= 4294967295) ## files >= 4GB are truncated; see ?unzip
      }

      if (!tooBig) {
        messagePreProcess("Extracting with R's unzip ... ")
        stExtract <- # system.time(
          mess <- capture.output(
          {
            extractedFiles <- do.call(fun, c(args, argList))
          },
          type = "message"
        )# )
        worked <- if (isUnzip) {
          all(normPath(file.path(args$exdir, argList[[1]])) %in% normPath(extractedFiles))
        } else {
          isTRUE(extractedFiles == 0)
        }
      }
      if (!isTRUE(worked) || isTRUE(tooBig)) {
        unz <- Sys.which("unzip")
        sZip <- Sys.which("7z")

        if (!isTRUE(tooBig)) {
          messagePreProcess("File unzipping using R does not appear to have worked.",
                            " Trying a system call of unzip...",
                            verbose = verbose
          )
        } else {
          messPart1 <- "R's unzip utility cannot handle a zip file this size.\n"
          if (nchar(sZip) > 0) {
            messagePreProcess(messPart1, verbose = verbose)
          } else {
            stop(
              paste(
                messPart1,
                # "Install 7zip and add it to your PATH (see https://www.7-zip.org/)."
                "Try installing the archive package then rerunning: \ninstall.packages('archive')"
              )
            )
          }
        }

        if (file.exists(args[[1]])) {
          pathToFile <- normPath(args[[1]])
        } else {
          if (file.exists(file.path(args$exdir, args[[1]]))) {
            pathToFile <- normPath(file.path(args$exdir, args[[1]]))
          } else {
            warning(mess)
            stop(
              "prepInputs cannot find the file ", basename2(args[[1]]), ".",
              " The file might have been moved during unzipping or is corrupted."
            )
          }
        }
        if (nchar(sZip) > 0) {
          messagePreProcess("Using '7zip'")
          op <- setwd(.tempPath)
          on.exit(
            {
              setwd(op)
            },
            add = TRUE
          )
          lstFiles <- system(paste0(sZip, " l ", pathToFile), intern = TRUE, wait = TRUE)
          startAndEnd <- grep("-----------", lstFiles)
          if (diff(startAndEnd) > 1) {
            lstFiles <- lstFiles[(startAndEnd[1] + 1):(startAndEnd[2] - 1)]
          }
          needListFiles <- FALSE
          if (length(files)) {
            filesAreInArch <- filenamesFromArchiveLst(lstFiles)
            if (all(files %in% filesAreInArch)) {
              if (all(filesAreInArch %in% files))
                needListFiles <- FALSE
              else
                needListFiles <- TRUE
            } else {
              stop("Some files are not in the archive (", pathToFile, "). Specifically:\n",
                   paste(files[!files %in% filesAreInArch], collapse = "\n"))
            }
          }

          # filesAreInArch <- unlist(lapply(files, function(x) any(grepl(x, lstFiles))))
          arg22 <- paste0(" x ", pathToFile)
          if (needListFiles) {
            arg22 <- paste(arg22, paste(files, collapse = " "))
          }
          system2(sZip,
                  args = arg22,
                  wait = TRUE,
                  stdout = NULL
          )
        } else if (nchar(unz) > 0) {
          messagePreProcess("Using 'unzip'")
          system2(unz,
                  args = paste0(pathToFile, " -d ", .tempPath),
                  wait = TRUE,
                  stdout = NULL
          )
        } else {
          if (nchar(unz) == 0) {
            stop(
              "unzip command cannot be found.",
              " Please try reinstalling Rtools if on Windows, and/or add unzip to system path",
              " (e.g., see 'https://cran.r-project.org/bin/windows/Rtools/'.)"
            )
          }
          stop(
            "There was no way to unzip all files; try manually. The file is located at: \n",
            pathToFile
          )
        }
      }
    }
    if (!isUnzip) {
      extractedFiles <- files
    }
  }

  if (!exists("listOfFilesExtracted", inherits = FALSE))
    listOfFilesExtracted <- list.files(
      path = .tempPath,
      # list of full paths of all extracted files!
      all.files = TRUE,
      recursive = TRUE,
      include.dirs = TRUE
    )

  mess <- paste0("       ... Done extracting ", length(listOfFilesExtracted), " files")
  if (exists("stExtract", inherits = FALSE))
    mess <- paste0(mess, "; took ", format(as.difftime(stExtract[3], units = "secs"), units = "auto"))
  messagePreProcess(mess)

  from <- makeAbsolute(listOfFilesExtracted, .tempPath)
  on.exit(
    {
      if (any(file.exists(from))) {
        suppressWarnings(try(unlink(from), silent = TRUE))
      }
    },
    add = TRUE
  )

  args$exdir <- origExdir
  to <- file.path(args$exdir, listOfFilesExtracted)

  suppressWarnings({
    out <- hardLinkOrCopy(from, to, verbose = 0)
  })

  if (!isTRUE(all(file.exists(to)))) {
    stop(paste("Could not move listOfFilesExtracted from", .tempPath, "to", args$exdir))
  }
  listOfFilesExtracted <- to
  # unlink(.tempPath, recursive = TRUE) # don't delete it if it was not created here --> on.exit does this

  if (length(listOfFilesExtracted) == 0) {
    stop(
      "preProcess could not extract the files from the archive ", args[[1]], ".\n",
      "Please try to extract it manually to the destinationPath"
    )
  }
  return(listOfFilesExtracted)
}

#' @keywords internal

.isArchive <- function(filename) {
  if (!is.null(filename)) {
    filename <- if (length(filename)) {
      isArchive <- fileExt(filename) %in% knownArchiveExtensions
      if (any(isArchive)) {
        filename[isArchive]
      } else {
        NULL
      }
    } else {
      NULL
    }
  }
  return(filename)
}

#' @keywords internal
#' @importFrom utils capture.output
#' @importFrom data.table rbindlist as.data.table setDT setDF
appendChecksumsTable <- function(checkSumFilePath, filesToChecksum,
                                 destinationPath = getOption("reproducible.destinationPath", "."),
                                 append = TRUE, verbose = getOption("reproducible.verbose", 1)) {
  csf <- tempfile(fileext = ".TXT")
  areAbs <- isAbsolutePath(filesToChecksum)
  if (any(!areAbs)) {
    filesToChecksum[!areAbs] <- file.path(destinationPath, filesToChecksum[!areAbs])
  }
  capture.output(type = "message", {
    currentFiles <- Checksums(
      path = destinationPath, write = TRUE,
      files = filesToChecksum,
      checksumFile = csf,
      verbose = verbose
    )
  })

  rip <- .getDestinationPathShared()
  checkSumFilePaths <- if (!is.null(rip)) {
    unique(c(checkSumFilePath, file.path(rip, basename(checkSumFilePath))))
  } else {
    checkSumFilePath
  }

  for (checkSumFilePath in checkSumFilePaths) {
    appendChecksumsTableWithCS(append, checkSumFilePath, destinationPath, filesToChecksum,
                               currentFiles = currentFiles, verbose = verbose)
  }
  return(currentFiles)
}


#' List files in either a `.zip` or or `.tar` file
#'
#' Makes the outputs from`.tar``.zip` the same, which they aren't by default.
#'
#' @param archive A character string of a single file name to list files in.
#' @param dontUse A character string or vector of unzipping tool(s) to not use,
#'   e.g., one or more of: `c("archive", "unzip", "7z")`
#'
#' @return A character string of all files in the archive.
#'
#' @keywords internal
#' @rdname listFilesInArchive
.listFilesInArchive <- function(archive, dontUse = NULL) {
  needSystemCall <- (length(archive) > 0 && fileExt(archive[1]) %in% knownSystemArchiveExtensions)

  for (dontUse in dontUses) {
    tryC <- try({

      if (length(archive) > 0) {
        if (file.exists(archive[1])) {
          fsArch <- file.size(archive[1])
          if (fsArch <= 10) {
            # corrupted
            unlink(archive[1])
            message("archive (", archive[1], ") appears corrupted; deleting it; ",
                    "you may have to manually delete it and the copy in `reproducible.destinationPathShared` if using ")
            return(NULL)
          }
          needSystemCall <- needSystemCall || fsArch > 2e9
        }
      }

      if (needSystemCall) {
        extractSystemCallPath <- .testForArchiveExtract(archive)
        funWArgs <- list(fun = extractSystemCallPath)
      } else {
        funWArgs <- .whichExtractFn(archive[1], NULL, dontUse = dontUse)
      }

      filesInArchive <- NULL
      if (!is.null(funWArgs$fun)) {
        if (file.exists(archive[1])) {
          if (!needSystemCall) {
            if (.requireNamespace("archive") && identical(archive::archive_extract, funWArgs$fun)) {
              filesInArchive <- archive::archive(archive[1])
            } else {
              filesInArchive <- funWArgs$fun(archive[1], list = TRUE)
            }
            nams <- names(filesInArchive)
            if ("Name" %in% nams) {
              # for zips, rm directories (length = 0)
              filesInArchive <- filesInArchive[filesInArchive$Length != 0, ]$Name
            } else if ("path" %in% nams) {
              # from archive::archive
              # Some zip files (e.g. Deflate64) cause archive::archive() to
              # report size = 0 for every entry even though the archive is valid
              # and non-empty.  Detect this and keep all paths rather than
              # incorrectly filtering everything out.
              allZeroSize <- all(filesInArchive$size == 0L) && file.size(archive[1]) > 0
              filesInArchive <- if (allZeroSize) {
                filesInArchive$path
              } else {
                filesInArchive[filesInArchive$size != 0, ]$path
              }
            } else {
              # untar & archive::archive
              filesInArchive
            }
          } else {
            if (grepl(x = extractSystemCallPath, pattern = "7z")) {
              extractSystemCall <- paste0("\"", extractSystemCallPath, "\"", " l \"", path.expand(archive[1]), "\"")
              if (isWindows()) {
                filesOutput <- captureWarningsToAttr(
                  system(extractSystemCall, show.output.on.console = FALSE, intern = TRUE)
                )
                warn <- attr(filesOutput, "warning")
                attr(filesOutput, "warning") <- NULL
              } else {
                # On Linux/MacOS
                filesOutput <- captureWarningsToAttr(
                  system(extractSystemCall, intern = TRUE, ignore.stderr = TRUE)
                )
                warn <- attr(filesOutput, "warning")
                attr(filesOutput, "warning") <- NULL
              }
            } else {
              archiveExtractBinary <- .archiveExtractBinary()
              if (is.null(archiveExtractBinary)) {
                stop("unrar is not on this system; please install it")
              }
              filesOutput <- system(paste0("unrar l ", archive[1]), intern = TRUE)
            }
            if (exists("warn", inherits = FALSE) && isTRUE(any(grepl("had status 2", warn)))) {
              stop(warn)
            }
            if (isTRUE(any(grepl("(Can not open the file as archive)|(Errors: 1)", filesOutput)))) {
              stop("archive appears defective")
            }
            # filesInBetween <- grep(pattern = "----", filesOutput)
            # filesLines <- filesOutput[(min(filesInBetween) + 1):(max(filesInBetween) - 1)]
            filesInArchive <- filenamesFromArchiveLst(filesOutput)
            # filenamesFromArchiveLst <- function(filesLines) {
            #   filesInArchive <- unlist(lapply(X = seq_along(filesLines), FUN = function(line) {
            #     first5trimmed <- unlist(strsplit(filesLines[[line]], split = " +"))[-(1:5)]
            #     if (length(first5trimmed) > 1)
            #       first5trimmed <- paste(first5trimmed, collapse = " ")
            #     # first5trimmed <- unlist(strsplit(filesLines[[line]], split = "  "))
            #     return(first5trimmed)
            #   }))
            # }
            if (length(filesInArchive) == 0) {
              stop("preProcess could not find any files in the archive ", archive)
            }
          }
        }
      }
      if (isTRUE(any(grepl("\\\\", filesInArchive))))
        filesInArchive <- gsub("\\\\", "/", filesInArchive)
    })
    if (!is(tryC, "try-error")) {
      break
    }
    if (identical(dontUse, tail(dontUses, 1)[[1]])) {
      stop(tryC)
    }
  }
  return(filesInArchive)
}

.compareChecksumsAndFilesAddDirs <- function(checkSums, files, destinationPath) {
  isOK <- NULL
  if (!is.null(files)) {
    checkSumsDT <- data.table(checkSums)
    if (NCOL(checkSumsDT) == 0) {
      checkSumsDT <- Copy(.emptyChecksumsResult)
    }
    dirs <- makeRelative(dirname(files), destinationPath) # basename2(unique(dirname(files)))
    dirs <- dirs[nzchar(dirs)]
    filesDT <- data.table(files = unique(makeRelative(c(files, dirs), destinationPath)))
    isOKDT <- checkSumsDT[filesDT, on = c(expectedFile = "files")]
    isOKDT2 <- checkSumsDT[filesDT, on = c(actualFile = "files"), nomatch = NA]
    # fill in any OKs from "actualFile" into the isOKDT
    isOKDT[compareNA(isOKDT2$result, "OK"), "result"] <- "OK"
    if (!all(compareNA(isOKDT$result, "OK"))) {
      isOKDT <- checksumsDirsOk(isOKDT)
    }

    isOK <- compareNA(isOKDT$result, "OK")
    names(isOK) <- makeRelative(filesDT$files, destinationPath)
  }
  isOK
}

#' Tests if unrar or 7zip exist
#'
#' @return
#' unrar or 7zip path if exist, or NULL
#'
#' @author Tati Micheletti
#'
#' @keywords internal
#' @rdname archiveExtractBinary
#' @name archiveExtractBinary
.archiveExtractBinary <- function(verbose = getOption("reproducible.verbose", 1)) {
  possPrograms <- unique(unlist(lapply(c("7z", "unrar"), Sys.which)))
  extractSystemCallPath <- if (!all(possPrograms == "")) {
    possPrograms[nzchar(possPrograms)][1] # take first one if there are more than one
  } else {
    ""
  }
  if (!(isWindows() && !isMac())) { ## TODO: macOS ?? #266
    if (grepl("7z", extractSystemCallPath)) {
      SevenZrarExists <- system("apt -qq list p7zip-rar", intern = TRUE, ignore.stderr = TRUE)
      SevenZrarExists <- grepl(SevenZrarExists, pattern = "installed")
      if (isFALSE(SevenZrarExists)) {
        messagePreProcess("To extract .rar files, you will need p7zip-rar, not just p7zip-full. Try: \n",
                          "--------------------------\n",
                          "apt install p7zip-rar\n",
                          "--------------------------\n",
                          verbose = verbose
        )
      }
    }
  }

  if (identical(extractSystemCallPath, "")) {
    if (isWindows()) {
      extractSystemCallPath <- Sys.which("7z.exe")
      if (extractSystemCallPath == "") {
        messagePreProcess("prepInputs is looking for 'unrar' or '7z' in your system...", verbose = verbose)
        extractSystemCallPath <- list.files("C:/Program Files",
                                            pattern = "unrar.exe|7z.exe",
                                            recursive = TRUE,
                                            all.files = TRUE,
                                            full.names = TRUE
        )
        if (extractSystemCallPath == "" || length(extractSystemCallPath) == 0) {
          extractSystemCallPath <- list.files(dirname(Sys.getenv("SystemRoot")),
                                              pattern = "unrar.exe|7z.exe",
                                              recursive = TRUE,
                                              all.files = TRUE,
                                              full.names = TRUE
          )
          if (extractSystemCallPath == "" || length(extractSystemCallPath) == 0) {
            extractSystemCallPath <- NULL
            messagePreProcess(missingUnrarMess, verbose = verbose)
          } else {
            messagePreProcess("The extracting software was found in an unusual location: ",
                              extractSystemCallPath, ".",
                              "If you receive an error when extracting the archive, please install ",
                              "'7zip' or 'unrar' in 'Program Files' directory.",
                              verbose = verbose
            )
          }
        }
        extractSystemCallPath <- extractSystemCallPath[1]
      }
    } else {
      messagePreProcess(missingUnrarMess,
                        "Try installing with, e.g.,: \n",
                        "--------------------------\n",
                        "apt install p7zip p7zip-rar p7zip-full -y\n",
                        "yum install p7zip p7zip-plugins -y\n",
                        "--------------------------",
                        verbose = verbose
      )
    }
  }
  if (!exists("extractSystemCallPath", inherits = FALSE)) extractSystemCallPath <- NULL
  if (is.null(extractSystemCallPath) || !nzchar(extractSystemCallPath)) extractSystemCallPath <- NULL

  return(extractSystemCallPath)
}

#' Returns `unrar` path and creates a shortcut as `.systemArchivePath`.
#' Was not incorporated in previous function so it can be used in the tests.
#'
#' @return
#' path to `unrar` or `7-zip` if it exists, and assign it to `.systemArchivePath`.
#' Advise the user if not found..
#'
#' @author Tati Micheletti
#'
#' @keywords internal
#' @rdname testForArchiveExtract
#' @name testForArchiveExtract
.testForArchiveExtract <- function(archive = "") {
  sevenzName <- grep("7z", knownSystemArchiveExtensions, value = TRUE)

  extractSystemCallPath <- NULL
  if (tools::file_ext(archive) == sevenzName) {
    extractSystemCallPath <- Sys.which(sevenzName)
  }

  if (is.null(extractSystemCallPath)) {
    if (!is.null(.systemArchivePath)) {
      extractSystemCallPath <- .systemArchivePath
    } else {
      # Find the path to unrar and assign to a package-stored object
      usrTg <- paste(sample(x = LETTERS, size = 15), collapse = "")
      # Cache for project-level persistence
      extractSystemCallPath <- Cache(.archiveExtractBinary, userTags = usrTg)
      utils::assignInMyNamespace(".systemArchivePath", extractSystemCallPath) # assign in namespace for pkg
    }
    if (is.null(extractSystemCallPath)) {
      clearCache(userTags = usrTg, ask = FALSE)
      stop(
        "prepInputs did not find '7-Zip' nor 'unrar' installed.",
        " Please install it before running prepInputs for a '.rar' archive"
      )
    }
  }
  return(extractSystemCallPath)
}

#' The known path for unrar or 7z
#' @rdname unrarPath
#' @name unrarPath
.systemArchivePath <- NULL

missingUnrarMess <- "The archive is a 'rar' archive; your system does not have unrar or 7zip;\n"
proj6Warn <- "NOT UPDATED FOR PROJ"

sevenzName <- "7z"
knownArchivePkgExtensions <- c("zip", "tar", "tar.gz", "gz", "rar", sevenzName, "cab")
knownInternalArchiveExtensions <- c("zip", "tar", "tar.gz", "gz")
# knownSystemArchiveExtensions <- c("rar", sevenzName, "unrar")
knownSystemArchiveExtensions <- c("rar", sevenzName)
knownArchiveExtensions <- unique(c(knownInternalArchiveExtensions, knownSystemArchiveExtensions,
                            knownArchivePkgExtensions))


prepInputsAssertions <- function(env) {
  noisy <- nullOr(c("character", "logical"), c("alsoExtract"), env = env)
  noisy <- nullOr(c("character", "logical"), c("useCache", "archive"), env)
  noisy <- nullOr(c("numeric", "logical"), c("purge", "verbose"), env)
  noisy <- nullOr("character", c(
    "destinationPath", "targetFile", "url", # "archive",
    ".tempPath"
  ), env = env)
  noisy <- nullOr("logical", c("quick", "overwrite"), env = env)
}

nullOr <- function(clses, vals, env) {
  vapply(vals, function(val) {
    out <- is.null(env[[val]])
    if (out) { # pull out fast if NULL
      return(out)
    }
    out <- inherits(env[[val]], clses)
    if (isFALSE(out)) {
      stop(
        val, " must be of class", "(es)"[(length(clses) > 1)], " ",
        paste(clses, collapse = ", "),
        " or set to its default. It is currently ", class(env[[val]]), "."
      )
    }
    out
  }, logical(1))
}

is.nulls <- function(x) lapply(x, is.null)



#' @include messages.R
process <- function(out, funCaptured,
                    useCache = getOption("reproducible.useCache"),
                    verbose = getOption("reproducible.verbose"),
                    .callingEnv = parent.frame(),
                    ...) {
  theFun <- out$fun
  suppressWarnings({
    naFun <- all(is.na(theFun))
  })

  funChar <- if (is.character(out$funChar)) out$funChar else NULL

  out <- modifyList(out, list(...))

  argsFromPrepInputsFamily <- unique(c(
    .namesPostProcessFormals(), formalArgs(prepInputs), formalArgs(preProcess),
    "checkSums", "dots", "object"
  ))
  args <- NULL
  # keep the ones for theFun
  isAlreadyQuoted <- tryCatch(any(grepl("quote", theFun)), silent = TRUE,
                              error = function(e) FALSE)
  if (isAlreadyQuoted) {
    theFun <- eval(theFun, envir = out)
  }
  if (is.name(theFun))
    theFun <- eval(theFun, envir = .callingEnv)

  # need to differentiate sf::st_read from sf::st_read(targetFile, TRUE) -- both are calls, both length 3; both have pkgColon
  if (length(theFun) == 3 && isDollarSqBrPkgColon(theFun) && all(lengths(as.list(theFun)) == 1)) {
    theFun <- eval(theFun, envir = out)
  }
  if (naFun %in% FALSE && !is.call(theFun)) {
    formsForTheFun <- names(formals3(theFun))
    argsFromPrepInputsFamily <- setdiff(argsFromPrepInputsFamily, names(formals3(theFun)))
    argsPassingToTheFun <- out[!names(out) %in% argsFromPrepInputsFamily]
    args <- argsPassingToTheFun[names(argsPassingToTheFun) %in% formsForTheFun]
  }

  otherFiles <- out$checkSums[result == "OK"]
  .cacheExtra <- NULL
  if (NROW(otherFiles)) {
    .cacheExtra <- .robustDigest(sort(otherFiles$checksum.x))
  }
  out[["targetFile"]] <- out[["targetFilePath"]] # handle both

  if (!(naFun || is.null(theFun))) {

    x <- if (is.null(out$object)) {
      st <- Sys.time()
      messagePreProcess("Running `process` (i.e., loading file into R)", verbose = verbose, verboseLevel = 0)
      .message$IndentUpdate()
      if (!is.null(out$targetFilePath)) {
        if (!all(is.na(out$targetFilePath)))
          messagePreProcess("targetFile located at:\n", paste(out$targetFilePath, collapse = "\n"), verbose = verbose)
      }

      if (!isTRUE(is.na(out$targetFilePath)))
        messagePreProcess("Loading object into R", verbose = verbose)
      needRaster <- any(grepl("raster$|stack$|brick$", funCaptured)) ||
        any(grepl("raster$|stack$|brick$", funChar)) || any(grepl("raster", capture.output(show(theFun))))
      needTerra <- any(grepl("terra|rast$", funCaptured)) || any(grepl("terra|rast$", funChar)) ||
        any(grepl("terra", capture.output(show(theFun))))
      if (needRaster) {
        .requireNamespace("raster", stopOnFALSE = TRUE)
      }
      if ((needRaster || needTerra) && !is.call(theFun)) {
        ## Don't cache the reading of a raster
        ## -- normal reading of raster on disk is fast b/c only reads metadata
        outProcess <- try(do.call(theFun, append(list(asPath(out$targetFilePath)), args)))
        if (is(outProcess, "try-error")) browser()
      } else {
        if (identical(theFun, base::load)) {
          if (is.null(args$envir)) {
            messagePreProcess("Running base::load, returning objects as a list. Pass envir = anEnvir ",
                              "if you would like it loaded to a specific environment",
                              verbose = verbose
            )
            tmpEnv <- new.env(parent = emptyenv())
            returnAsList <- TRUE
          } else {
            tmpEnv <- args$envir
            args$envir <- NULL
            returnAsList <- FALSE
          }
          args2 <- append(list(file = out$targetFilePath, envir = tmpEnv), args)
          outProcess <- do.call(theFun, args2)
          if (returnAsList) {
            outProcess <- as.list(tmpEnv, all.names = TRUE)
          }
        } else {
          useCache2 <- useCache
          if (any(fileExt(out$targetFilePath) %in% c(.qsFormat, .qs2Format, .rdsFormat)) &&
              !isTRUE(getOption("reproducible.useMemoise"))) {
            useCache2 <- FALSE
            messagePreProcess("targetFile is already a binary; skipping Cache while loading")
          }

          withCallingHandlers(
            # theFun could have been a call to get the function, e.g., fun = getOption("reproducible.shapefileRead")
            # so need to try 2x, just to figure out the function
            for (i in 1:2) {
              if (is.call(theFun)) { # an actual call, not just captured function name
                # put `targetFilePath` in the first position -- allows quoted call to use first arg
                out <- append(
                  append(
                    list(targetFilePath = out[["targetFilePath"]]),
                    out[-which(names(out) == "targetFilePath")]
                  ),
                  args
                )
                # out[["targetFile"]] <- out[["targetFilePath"]] # handle both
                if (is.null(funChar)) funChar <- paste0(substr(format(theFun), start = 1, stop = 40), "...")
                outProcess <- Cache(eval(theFun, envir = out, enclos = .callingEnv),
                                    useCache = useCache2, .cacheExtra = .cacheExtra,
                                    .functionName = funChar, omitArgs = "enclos"
                )
              } else {
                args2 <- append(list(asPath(out$targetFilePath)), args)
                outProcess <- Cache(do.call, theFun, args2,
                                    useCache = useCache2, .cacheExtra = .cacheExtra,
                                    .functionName = funChar
                )
              }
              # theFun could have been a call to get the function, e.g., fun = getOption("reproducible.shapefileRead")
              #  If this was the case, then the above will have just evaluated that
              if (identical(1L, length(outProcess))) {
                if (isTRUE(is.character(outProcess))) {
                  possTheFun <- eval(parse(text = outProcess), envir = out, enclos = .callingEnv)
                  if (isTRUE(is.function(possTheFun))) {
                    theFun <- possTheFun
                    next
                  }
                }
              }
              break
            },
            message = function(m) {
              m$message <- grep(paste0(.message$NoCachePathSupplied, "|useCache is FALSE"), m$message, invert = TRUE, value = TRUE)
              if (length(m$message)) {
                mm <- gsub("(.*)\n$", "\\1", m$message)
                message(mm) # this MUST NOT CREATE INDENTING -- using 'message' here
              }
              tryInvokeRestart("muffleMessage")
            }
          )
          # outProcess
        }
      }
      .message$IndentRevert()
      stNext <- reportTime(st, mess = "`process` done; took ", minSeconds = 10)
      outProcess

    } else {
      if (is.null(theFun) || !is.call(theFun) || naFun) {
        x <- out$object
      } else {
        # x <- out$object
        env1 <- new.env()
        list2env(out, envir = env1)
        eval(theFun, envir = env1)
      }
    }
  } else {
    x <- if ((is.null(theFun) || is.na(theFun)) && !is.null(out$object)) {
      out$object
    } else {
      messagePreProcess("No loading of object into R; fun = ", theFun, "; returning the targetFilePath: ",
                        out$targetFilePath, verbose = verbose)
      out$targetFilePath
    }
  }
  x
}

removeDirs <- function(paths) {
  out <- strsplit(paths, "\\\\|/")
  lens <- lengths(out)
  la <- unlist(lapply(unique(lens), function(len) {
    table(sapply(out[lens >= len], function(xx) paste(xx[seq(len)], collapse = "/")))
  }))
  dirs <- names(la[la > 1])
  paths <- paths[!paths %in% dirs]

}


filenamesFromArchiveLst <- function(filesOutput) {
  filesInBetween <- grep(pattern = "----", filesOutput)
  filesLines <- if (length(filesInBetween) == 0)
    filesOutput
  else
    filesOutput[(min(filesInBetween) + 1):(max(filesInBetween) - 1)]

  filesInArchive <- unlist(lapply(X = seq_along(filesLines), FUN = function(line) {
    tail(unlist(strsplit(filesLines[[line]], split = " +")), 1)
    # first5trimmed <- unlist(strsplit(filesLines[[line]], split = " +"))[-(1:5)]
    # if (length(first5trimmed) > 1)
    #   first5trimmed <- paste(first5trimmed, collapse = " ")
    # # first5trimmed <- unlist(strsplit(filesLines[[line]], split = "  "))
    # return(first5trimmed)
  }))
  if (isTRUE(any(grepl("\\\\", filesInArchive))))
    filesInArchive <- gsub("\\\\", "/", filesInArchive)

  filesInArchive <- removeDirs(filesInArchive)

  filesInArchive
}



reportTime <- function(stStart, mess, minSeconds) {
  stNow <- Sys.time()
  dt1sec <- difftime(stNow, stStart, units = "secs")
  dt1auto <- difftime(stNow, stStart)
  messagePreProcess(mess, format(dt1auto, units = "auto"), verbose = dt1sec > minSeconds)
  stNow
}


readCheckSumFilePath <- function(checkSumFilePath, destinationPath, filesToChecksum) {
  # a checksums file already existed, need to keep some of it
  cs <- suppressWarnings(try(read.table(checkSumFilePath, header = TRUE), silent = TRUE))
  if (is(cs, "try-error")) {
    # meant that it was an empty CHECKSUMS.txt file -- rebuild it
    cs <- NULL # append <- FALSE
  }
  cs
}

extractFileNOTtoChecksum <- function(cs, destinationPath, filesToChecksum,
                                     algorithms = NULL) {
  setDT(cs)
  files <- makeRelative(filesToChecksum, destinationPath)
  if (is.null(algorithms) || all(is.na(algorithms)) ||
      is.null(cs$algorithm)) {
    # Backward-compat: drop rows for any file we're updating, regardless of
    # algorithm. Used when caller doesn't know which algorithm it's writing.
    out <- cs[!(makeRelative(file, destinationPath) %in% files)]
  } else {
    # Multi-algorithm: only drop rows where (file, algorithm) is being
    # upserted. Rows recording other algorithms for the same file are kept.
    out <- cs[!(makeRelative(file, destinationPath) %in% files &
                  algorithm %in% algorithms)]
  }
  setDF(out)
  out
}



appendChecksumsTableWithCS <- function(append, checkSumFilePath, destinationPath,
                                       filesToChecksum, currentFiles, verbose) {
  if (append) {
    cs <- readCheckSumFilePath(checkSumFilePath, destinationPath, filesToChecksum)
    if (is.null(cs)) {
      append <- FALSE
    } else {
      # a checksums file already existed, need to keep some of it.
      # Restrict the drop to (file, algorithm) pairs we're writing so other
      # algorithms recorded for the same file (e.g. an md5 row written by
      # pp_remote_hash_check) are preserved alongside the xxhash64 row this
      # call is producing.
      currAlgo <- if (!is.null(currentFiles$algorithm.x)) {
        unique(currentFiles$algorithm.x)
      } else NULL
      nonCurrentFiles <- extractFileNOTtoChecksum(
        cs, destinationPath, filesToChecksum, algorithms = currAlgo
      )
    }
  }

  doWrite <- TRUE
  if (append) { # a checksums file already existed, need to keep some of it
    messStart <- "Appending "
    messagePreProcess(messStart, "checksums to CHECKSUMS.txt. If you see this message repeatedly, ",
                      "you can specify targetFile (and optionally alsoExtract) so it knows ",
                      "what to look for.", verbose = verbose)

    currentFilesToRbind <- currentFilesToChecksumsTable(currentFiles, nonCurrentFiles, verbose = verbose)

    # Sometimes a checksums file doesn't have filesize
    if (!is.null(cs$filesize)) {
      if (!is.character(cs$filesize)) {
        cs$filesize <- as.character(cs$filesize)
      }
    }
    if (identical(cs, as.data.frame(currentFilesToRbind))) {
      doWrite <- FALSE
    }
  } else {
    currentFilesToRbind <- currentFilesToChecksumsTable(currentFiles, verbose = verbose)
  }

  if (doWrite) {
    writeChecksumsTable(as.data.frame(currentFilesToRbind), checkSumFilePath, dots = list())
  }
}


currentFilesToChecksumsTable <- function(currentFiles, nonCurrentFiles = NULL, verbose) {
  currentFilesToRbind <- data.table::as.data.table(currentFiles)
  keepCols <- c("expectedFile", "checksum.x", "algorithm.x", "filesize.x")
  currentFilesToRbind <- currentFilesToRbind[, keepCols, with = FALSE]
  data.table::setnames(currentFilesToRbind,
                       old = keepCols,
                       new = c("file", "checksum", "algorithm", "filesize")
  )
  currentFilesToRbind <- rbindlist(list(nonCurrentFiles, currentFilesToRbind), fill = TRUE)

  # Attempt to not change CHECKSUMS.txt file if nothing new occurred
  currentFilesToRbind <- unique(currentFilesToRbind)
  anyDuplicates <- duplicated(currentFilesToRbind)
  if (any(anyDuplicates)) {
    messagePreProcess("The current targetFile is not the same as the expected targetFile in the ",
                      "CHECKSUMS.txt; appending new entry in CHECKSUMS.txt. If this is not ",
                      "desired, please check files for discrepancies",
                      verbose = verbose
    )
  }
  currentFilesToRbind
}

checkSFWebPage <- function(funPoss, fileExt, feKnown, verbose) {
  if (.requireNamespace("rvest", stopOnFALSE = FALSE)) {
    sfURL <- "https://r-spatial.github.io/sf/articles/sf2.html#guessing-a-driver-for-output"
    tbls_ls <- try({rvest::read_html(sfURL) |>
        rvest::html_nodes("table")  |>
        rvest::html_table(fill = TRUE) } |>
          Cache(verbose = FALSE))
    if (!is(tbls_ls, "try-error")) {
      exts <- tbls_ls[[1]]$extension
      if (isTRUE(any(fileExt %in% exts))) {
        funPoss <- data.frame(fileExt, "sf::st_read", "sf::st_write", "sf") |>
          setNames(names(feKnown))
      }
    } else {
      messagePrepInputs("It looks like the sf article identifying which extensions",
                        "it can read has changed", "\nPlease contact reproducible",
                        "developers citing this message. Currently using:\n",
                        sfURL, verbose = verbose + 1)
    }

  } else {
    messagePrepInputs("`reproducible` does not know the file type passed.\n",
                      "(see reproducible:::.fileExtsKnown() for `prepInputs` known types)\n",
                      "Please run `install.packages('rvest')` to load a broader ",
                      "set of file types that R knows how to read. ",

                      verbose = verbose + 1)
  }
  funPoss
}

dontUses <- list(NULL, "archive", c("archive", "unzip"))

utils::globalVariables(c(
  "reproducible.inputPaths", "successfulCheckSumFilePath", "successfulDir",
  "destinationPathUser"
))

#' @param n Number of non-null arguments passed to `preProcess`.
#' E.g., passing `n = 1` returns combinations with only a single non-NULL parameter.
#' If `NULL` (default), all parameter combinations are returned.
#'
#' @export
#' @rdname preProcess
preProcessParams <- function(n = NULL) {
  p1 <- data.frame(
    url = c("char", "NULL", "NULL", "NULL"),
    targetFile = c("NULL", "char", "NULL", "NULL"),
    archive = c("NULL", "NULL", "char", "NULL"),
    alsoExtract = c("NULL", "NULL", "NULL", "char"),
    Result = c(
      "Download, extract all files if an archive, guess at 'targetFile', load into R.",
      "Load 'targetFile' into R.",
      "Extract all files, guess at 'targetFile', load into R.",
      "Guess at 'targetFile' from files in 'alsoExtract', load into R."
    ),
    FirstChecksum = c(
      "Write or append all new files.",
      "Write or append 'targetFile'.",
      "Write or append all new files.",
      "Write or append all new files."
    ),
    SecondChecksum = c(
      "Same as first; no 'targetFile'.*",
      "No downloading, so no checksums used.",
      "No downloading, so no checksums used.",
      "No downloading, so no checksums used."
    ),
    stringsAsFactors = FALSE
  )
  p2 <- data.frame(
    url = c("char", "char", "char", "NULL", "NULL", "NULL"),
    targetFile = c("char", "NULL", "NULL", "char", "char", "NULL"),
    archive = c("NULL", "char", "NULL", "NULL", "char", "char"),
    alsoExtract = c("NULL", "NULL", "char", "char", "NULL", "char"),
    Result = c(
      "Download, extract all files if an archive, load 'targetFile' into R.",
      "Download, extract all files, guess at 'targetFile', load into R.",
      "Download, extract only named files in 'alsoExtract', guess at 'targetFile', load into R.",
      "Load 'targetFile' into R.",
      "Extract all files, load 'targetFile' into R.",
      " Extract only named files in 'alsoExtract', guess at 'targetFile', load into R."
    ),
    FirstChecksum = c(
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files."
    ),
    SecondChecksum = c(
      "Use Checksums, skip downloading.",
      "Same as first; no 'targetFile'.*",
      "Same as first; no 'targetFile'.*",
      "No downloading, so no checksums used.",
      "No downloading, so no checksums used.",
      "No downloading, so no checksums used."
    ),
    stringsAsFactors = FALSE
  )
  p3 <- data.frame(
    url = c("char", "char", "char", "char", "char", "char", "NULL"),
    targetFile = c("char", "NULL", "NULL", "char", "char", "char", "char"),
    archive = c("char", "char", "char", "NULL", "NULL", "char", "char"),
    alsoExtract = c("NULL", "char", "'similar'", "char", "'similar", "NULL", "char"),
    Result = c(
      "Download, extract all files, load 'targetFile' into R.",
      "Download, extract files named in 'alsoExtract', guess at 'targetFile', load into R.",
      "Download, extract all files (can't understand \"similar\"), guess at 'targetFile', load into R.",
      "Download, if an archive, extract files named in 'targetFile' and 'alsoExtract', load 'targetFile' into R.",
      "Download, if an archive, extract files with same base as 'targetFile', load 'targetFile' into R.",
      "Download, extract all files from archive, load 'targetFile' into R.",
      "Extract  files named in 'alsoExtract' from archive, load 'targetFile' into R."
    ),
    FirstChecksum = c(
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files.",
      "Write or append all new files."
    ),
    SecondChecksum = c(
      "Use Checksums, skip downloading.",
      "Use Checksums, skip downloading.",
      "Same as first; no 'targetFile'.",
      "Use Checksums, skip downloading.",
      "Use Checksums, skip downloading.",
      "Use Checksums, skip downloading.",
      "No downloading, so no checksums used."
    ),
    stringsAsFactors = FALSE
  )
  p4 <- data.frame(
    url = c("char", "char"),
    targetFile = c("char", "char"),
    archive = c("char", "char"),
    alsoExtract = c("char", "'similar'"),
    Result = c(
      "Download, extract files named in 'targetFile' and 'alsoExtract', load 'targetFile' into R.",
      "Download, extract all files with same basename as 'targetFile', load 'targetFile' into R."
    ),
    FirstChecksum = c(
      "Write or append all new files.",
      "Write or append all new files."
    ),
    SecondChecksum = c(
      "Use Checksums, skip downloading.",
      "Use Checksums, skip downloading."
    ),
    stringsAsFactors = FALSE
  )

  if (is.null(n)) {
    rbind(p1, p2, p3, p4)
  } else if (n == 1) {
    p1
  } else if (n == 2) {
    p2
  } else if (n == 3) {
    p3
  } else if (n == 4) {
    p4
  }
}

#' Download, checksum, extract files
#'
#' This does downloading (via `downloadFile`), checksumming (`Checksums`),
#' and extracting from archives (`extractFromArchive`), plus cleaning up of input
#' arguments (e.g., paths, function names).
#' This is the first stage of three used in `prepInputs`.
#'
#' @return
#' A list with 5 elements: `checkSums` (the result of a `Checksums`
#' after downloading), `dots` (cleaned up `...`, including deprecated argument checks),
#' `fun` (the function to be used to load the `preProcess`ed object from disk),
#' and `targetFilePath` (the fully qualified path to the `targetFile`).
#'
#' @param .callingEnv The environment where the function was called from. Used to find
#'   objects, if necessary.
#' @section Combinations of `targetFile`, `url`, `archive`, `alsoExtract`:
#'
#'   Use `preProcessParams()` for a table describing various parameter combinations and their
#'   outcomes.
#'
#'  `*` If the `url` is a file on Google Drive, checksumming will work
#'  even without a `targetFile` specified because there is an initial attempt
#'  to get the remove file information (e.g., file name). With that, the connection
#'  between the `url` and the filename used in the \file{CHECKSUMS.txt} file can be made.
#'
#' @inheritParams prepInputs
#' @inheritParams downloadFile
#'
#' @author Eliot McIntire
#' @export
#' @importFrom data.table fread setDT
preProcess <- function(targetFile = NULL, url = NULL, archive = NULL, alsoExtract = NULL,
                       destinationPath = getOption("reproducible.destinationPath", "."),
                       fun = NULL, dlFun = NULL,
                       quick = getOption("reproducible.quick"),
                       overwrite = getOption("reproducible.overwrite", FALSE),
                       purge = FALSE,
                       verbose = getOption("reproducible.verbose", 1),
                       .tempPath, .callingEnv = parent.frame(),
                       ...) {
  st <- Sys.time()
  messagePreProcess("Running `preProcess`", verbose = verbose, verboseLevel = 0)
  .message$IndentUpdate()
  on.exit(.message$IndentRevert(), add = TRUE)

  if (missing(.tempPath)) {
    .tempPath <- tempdir2(rndstr(1, 6))
    on.exit(unlink(.tempPath, recursive = TRUE), add = TRUE)
  }

  # Capture dlFun before any other evaluation (must be in caller frame)
  dlFunCaptured <- substitute(dlFun)
  prepInputsAssertions(environment())

  # Three cases for the captured expression:
  #   (a) quote(fn(...))            -> unwrap with eval to get the inner call
  #   (b) fn(...) or pkg::fn(...)   -> keep as a deferred call object (the
  #                                    canonical dlFun usage; evaluated later
  #                                    inside downloadFile in the caller env)
  #   (c) anything else (symbol,    -> force the lazy promise so ctx stores the
  #       NULL, control-flow like   -- actual value (function/NULL/string), not
  #       `if (x) fn else NULL`)    -- an expression referencing caller-frame
  #                                    variables that may be out of scope later
  dlFunCaptured <- .resolveDlFunCaptured(dlFunCaptured, dlFun)

  dots <- list(...)
  fun  <- .checkFunInDots(fun = fun, dots = dots)
  dots <- .checkDeprecated(dots, verbose = verbose)

  ctx <- .pp_make_ctx(
    targetFile = targetFile, url = url, archive = archive, alsoExtract = alsoExtract,
    destinationPath = destinationPath, fun = fun, dlFunCaptured = dlFunCaptured,
    dots = dots, quick = quick, overwrite = overwrite, purge = purge, verbose = verbose,
    .tempPath = .tempPath, .callingEnv = .callingEnv
  )

  ctx <- pp_resolve_files(ctx)
  ctx <- pp_checksums_init(ctx)
  ctx <- pp_purge(ctx)
  ctx <- pp_resolve_needed_files(ctx)
  ctx <- pp_check_local_sources(ctx)
  ctx <- pp_remote_hash_check(ctx)
  ctx <- pp_download(ctx)
  ctx <- pp_extract(ctx)
  ctx <- pp_link_to_destination(ctx)
  out <- pp_finalize(ctx)

  reportTime(st, mess = "`preProcess` done; took ", minSeconds = 10)
  out
}

# ---------------------------------------------------------------------------
# Context constructor
# ---------------------------------------------------------------------------
.pp_make_ctx <- function(targetFile, url, archive, alsoExtract, destinationPath, fun,
                         dlFunCaptured, dots, quick, overwrite, purge, verbose,
                         .tempPath, .callingEnv) {
  if (is.null(destinationPath))
    destinationPath <- getOption("reproducible.destinationPath", ".")
  destinationPath <- normPath(destinationPath)

  # Treat a zero-length character `url` (e.g. `character()`) as no URL — the
  # downstream guards check `is.null(ctx$url)`, and `grepl()` on length-0 input
  # returns `logical(0)`, which crashes `if`.
  if (is.character(url) && length(url) == 0L) url <- NULL

  if (!is.null(archive) && any(is.na(archive)) && all(!is.character(archive)))
    archive <- as.character(archive)

  if (is.logical(alsoExtract))
    alsoExtract <- c("none", "all")[alsoExtract + 1L]

  list(
    url = url, fun = fun, dlFunCaptured = dlFunCaptured, dots = dots,
    quick = quick, overwrite = overwrite, purge = purge, verbose = verbose,
    .tempPath = .tempPath, .callingEnv = .callingEnv,
    targetFile = targetFile, targetFilePath = NULL,
    archive = archive, alsoExtract = alsoExtract,
    destinationPath = destinationPath, destinationPathUser = NULL,
    checkSumFilePath = identifyCHECKSUMStxtFile(destinationPath),
    checkSums = .emptyChecksumsResult, needChecksums = 0L,
    neededFiles = NULL, filesToChecksum = NULL, filesExtr = NULL,
    reproducible.inputPaths = NULL, successfulDir = NULL,
    successfulCheckSumFilePath = NULL,
    downloadResult = NULL, funChar = NULL,
    skipDownload = FALSE, remoteMetadata = NULL,
    hashVerified = character(),
    verboseCFS = verbose
  )
}

# ---------------------------------------------------------------------------
# Phase 1: Resolve targetFile / archive / alsoExtract from url
# ---------------------------------------------------------------------------
pp_resolve_files <- function(ctx) {
  targetFileGuess <- NULL
  if (is.null(ctx[["targetFile"]]) || is.null(ctx$archive)) {
    targetFileGuess <- .guessAtFile(
      url = ctx$url, archive = ctx$archive, targetFile = ctx[["targetFile"]],
      destinationPath = ctx$destinationPath, verbose = ctx$verbose,
      team_drive = getTeamDrive(ctx$dots)
    )
    if (is.null(ctx$archive))
      ctx$archive <- .isArchive(targetFileGuess)
  }

  ctx$targetFilePath <- getTargetFilePath(
    ctx[["targetFile"]], ctx$archive, targetFileGuess, ctx$verbose,
    ctx$destinationPath, ctx$alsoExtract, ctx$checkSumFilePath
  )
  ctx[["targetFile"]]  <- makeRelative(ctx$targetFilePath, ctx$destinationPath)
  ctx$alsoExtract <- guessAlsoExtract(ctx[["targetFile"]], ctx$alsoExtract, ctx$checkSumFilePath)

  if (!dir.exists(ctx$destinationPath)) {
    if (isFile(ctx$destinationPath)) stop("destinationPath must be a directory")
    checkPath(ctx$destinationPath, create = TRUE)
  }

  if (isTRUE(!is.na(ctx[["targetFile"]])))
    messagePreProcess("Preparing: ", ctx[["targetFile"]], verbose = ctx$verbose)

  ctx$archive <- setupArchive(ctx$archive, ctx$destinationPath)
  ctx
}

# ---------------------------------------------------------------------------
# Phase 2: Load CHECKSUMS.txt; handle reproducible.inputPaths redirect
# ---------------------------------------------------------------------------
pp_checksums_init <- function(ctx) {
  filesToCheck <- na.omit(unique(c(ctx$archive, ctx$targetFilePath, ctx$alsoExtract)))

  # Pre-verify via remote-hash sidecar files. A `<basename>_<urlEncoded>.hash`
  # file in destinationPath (or any reproducible.inputPaths dir) is written
  # only by a successful pp_remote_hash_check, so its presence implies the
  # file has been validated against the remote source. Skip the local
  # CHECKSUMS.txt-driven re-digest for those files — re-hashing a multi-GB
  # archive in this phase otherwise dominates wall time of an idempotent
  # preProcess() call (4 invocations: destinationPath + inputPaths × 2
  # CHECKSUMS.txt locations). User can `rm` the .hash file to force a full
  # local re-verification.
  filesPreVerified <- character()
  if (length(filesToCheck)) {
    sidecarDirs <- unique(c(ctx$destinationPath, .getDestinationPathShared()))
    hasSidecar <- vapply(filesToCheck, function(f) {
      length(.findRemoteHashSidecars(basename2(f), sidecarDirs)) > 0L &&
        file.exists(f)
    }, FUN.VALUE = logical(1L))
    if (any(hasSidecar)) {
      filesPreVerified <- filesToCheck[hasSidecar]
      filesToCheck     <- filesToCheck[!hasSidecar]
      ctx$hashVerified <- unique(c(ctx$hashVerified, filesPreVerified))
    }
  }

  inputPaths <- runChecksums(ctx$destinationPath, ctx$checkSumFilePath, filesToCheck, ctx$verbose)

  ctx$reproducible.inputPaths <- inputPaths$reproducible.inputPaths
  ctx$destinationPathUser     <- inputPaths$destinationPathUser
  ctx$destinationPath         <- inputPaths$destinationPath
  ctx$checkSums               <- inputPaths$checkSums
  ctx$checkSumFilePath        <- identifyCHECKSUMStxtFile(ctx$destinationPath)

  if (!is.null(inputPaths$destinationPathUser)) {
    ctx$targetFilePath <- makeAbsolute(
      makeRelative(ctx$targetFilePath, inputPaths$destinationPathUser),
      ctx$destinationPath
    )
    if (!is.null(ctx$archive)) {
      ctx$archive <- makeAbsolute(
        makeRelative(ctx$archive, inputPaths$destinationPathUser),
        ctx$destinationPath
      )
    }
  }

  if (is.null(ctx$checkSums) || is(ctx$checkSums, "try-error")) {
    ctx$needChecksums <- 1L
    ctx$checkSums     <- .emptyChecksumsResult
  }

  # Synthesise OK rows for pre-verified files so downstream phases
  # (.compareChecksumsAndFilesAddDirs, .checkLocalSources) see them as
  # already-OK without a digest pass. If a row with the same expectedFile
  # already exists (e.g. left over from a stale CHECKSUMS.txt comparison),
  # drop it — the synthetic row's "OK" is authoritative for downstream joins.
  if (length(filesPreVerified)) {
    rel <- makeRelative(filesPreVerified, ctx$destinationPath)
    sz  <- as.character(file.size(filesPreVerified))
    okRows <- data.table::data.table(
      result       = "OK",
      expectedFile = rel,
      actualFile   = rel,
      checksum.x   = NA_character_,
      checksum.y   = NA_character_,
      algorithm.x  = NA_character_,
      algorithm.y  = NA_character_,
      filesize.x   = sz,
      filesize.y   = sz
    )
    if (NROW(ctx$checkSums)) {
      ctx$checkSums <- ctx$checkSums[!ctx$checkSums$expectedFile %in% rel, ]
    }
    ctx$checkSums <- data.table::rbindlist(
      list(ctx$checkSums, okRows), fill = TRUE
    )
  }

  ctx
}

# ---------------------------------------------------------------------------
# Phase 3: Apply purge strategy
# ---------------------------------------------------------------------------
pp_purge <- function(ctx) {
  purge <- ctx$purge
  if (is.logical(purge)) purge <- as.integer(purge)

  if (purge == 1L) {
    unlink(ctx$checkSumFilePath)
    ctx$needChecksums <- 1L
  } else if (purge > 1L) {
    ctx$checkSums     <- .purge(
      checkSums = ctx$checkSums, purge = purge,
      targetFile = ctx[["targetFile"]], archive = ctx$archive,
      url = ctx$url, alsoExtract = ctx$alsoExtract,
      destinationPath = ctx$destinationPath
    )
    ctx$needChecksums <- 2L
  }
  ctx$purge <- purge
  ctx
}

# ---------------------------------------------------------------------------
# Phase 4: Build neededFiles; pre-download extraction attempt if archive local
# ---------------------------------------------------------------------------
pp_resolve_needed_files <- function(ctx) {
  archiveOut    <- dealWithArchive(
    ctx$archive, ctx$url, ctx[["targetFile"]], ctx$checkSums, ctx$alsoExtract,
    ctx$destinationPath, getTeamDrive(ctx$dots), ctx$verbose
  )
  ctx$checkSums <- archiveOut$checkSums
  ctx$archive   <- archiveOut$archive

  # Expand any regex patterns in alsoExtract while the archive is already on
  # disk.  Doing this BEFORE makeAbsolute() prevents fake absolute paths like
  # "/dest/CMD_sm|CMD_sp" from entering neededFiles.
  if (!isNULLorNA(ctx$alsoExtract) &&
      !isNULLorNA(ctx$archive) && any(file.exists(ctx$archive))) {
    fia <- makeRelative(.listFilesInArchive(ctx$archive), ctx$destinationPath)
    if (length(fia))
      ctx$alsoExtract <- .expandAlsoExtractPatterns(ctx$alsoExtract, fia)
  }

  neededFiles <- c(ctx[["targetFile"]], makeAbsolute(ctx$alsoExtract, ctx$destinationPath))
  if (is.null(neededFiles)) neededFiles <- makeAbsolute(ctx$archive)
  if (any(is.na(neededFiles)))  neededFiles <- na.omit(neededFiles)
  neededFiles <- grep("similar$", neededFiles, value = TRUE, invert = TRUE)
  neededFiles <- grep("none$",    neededFiles, value = TRUE, invert = TRUE)

  neededFilesOrig <- NULL
  if (is.null(ctx$alsoExtract)) {
    filesInsideArchive <- .listFilesInArchive(ctx$archive)
    neededFiles        <- checkRelative(neededFiles, ctx$destinationPath, filesInsideArchive)
    if (length(filesInsideArchive) > 0L)
      ctx$checkSums <- .checkSumsUpdate(
        ctx$destinationPath,
        makeAbsolute(filesInsideArchive, ctx$destinationPath),
        checkSums = ctx$checkSums
      )
    neededFiles <- unique(c(neededFiles, filesInsideArchive))
  } else {
    outFromSimilar  <- .checkForSimilar(
      neededFiles, ctx$alsoExtract, ctx$archive, ctx[["targetFile"]],
      destinationPath = ctx$destinationPath, ctx$checkSums,
      checkSumFilePath = ctx$checkSumFilePath, ctx$url, verbose = ctx$verboseCFS
    )
    ctx$verboseCFS  <- ctx$verbose - 1L
    neededFilesOrig <- unique(makeAbsolute(neededFiles, ctx$destinationPath))
    ctx$checkSums   <- outFromSimilar$checkSums
    neededFiles     <- outFromSimilar$neededFiles
  }

  neededFiles <- unique(makeAbsolute(neededFiles, ctx$destinationPath))
  neededFiles <- unique(c(neededFilesOrig, neededFiles))
  neededFiles <- grep("none$", neededFiles, value = TRUE, invert = TRUE)

  filesToChecksum <- if (is.null(ctx$archive) || isTRUE(is.na(ctx$archive))) NULL else ctx$archive
  filesToChecksum <- unique(c(filesToChecksum, neededFiles))

  isOK <- .compareChecksumsAndFilesAddDirs(ctx$checkSums, filesToChecksum, ctx$destinationPath)
  if (isTRUE(!all(isOK))) {
    results <- .tryExtractFromArchive(
      archive = if (isTRUE(is.na(ctx$archive))) NULL else ctx$archive,
      neededFiles = neededFiles, alsoExtract = ctx$alsoExtract,
      destinationPath = ctx$destinationPath, checkSums = ctx$checkSums,
      needChecksums = ctx$needChecksums, checkSumFilePath = ctx$checkSumFilePath,
      filesToChecksum = filesToChecksum, targetFilePath = ctx$targetFilePath,
      quick = ctx$quick, verbose = ctx$verbose, .tempPath = ctx$.tempPath
    )
    ctx$neededFiles    <- results$neededFiles
    ctx$checkSums      <- results$checkSums
    ctx$filesExtr      <- results$filesExtr
    ctx$targetFilePath <- results$targetFilePath
    filesToChecksum    <- results$filesToChecksum
    ctx$needChecksums  <- results$needChecksums

    if (ctx$needChecksums > 0L) {
      ctx$checkSums     <- appendChecksumsTable(
        checkSumFilePath = ctx$checkSumFilePath,
        filesToChecksum  = unique(filesToChecksum),
        destinationPath  = ctx$destinationPath,
        append           = results$needChecksums >= 2L
      )
      ctx$needChecksums <- 0L
    }
  } else {
    targetFilePoss <- makeRelative(ctx$targetFilePath, ctx$destinationPath)
    if (isTRUE(!targetFilePoss %in% names(isOK))) {
      wh <- grep(targetFilePoss, names(isOK))
      if (length(wh)) ctx$targetFilePath <- names(isOK)[wh]
    }
    ctx$neededFiles <- neededFiles
    ctx$filesExtr   <- setdiff(
      unique(c(filesToChecksum, neededFiles)),
      .isArchive(c(filesToChecksum, neededFiles))
    )
  }

  ctx$filesToChecksum <- filesToChecksum
  ctx
}

# ---------------------------------------------------------------------------
# Phase 5: Check reproducible.inputPaths for existing local copies
# ---------------------------------------------------------------------------
pp_check_local_sources <- function(ctx) {
  localChecks <- .checkLocalSources(
    ctx$neededFiles, checkSums = ctx$checkSums,
    checkSumFilePath = ctx$checkSumFilePath,
    otherPaths = ctx$reproducible.inputPaths,
    destinationPath = ctx$destinationPath,
    needChecksums = ctx$needChecksums, verbose = ctx$verbose
  )
  ctx$neededFiles                <- localChecks$neededFiles
  ctx$checkSums                  <- localChecks$checkSums
  ctx$needChecksums              <- localChecks$needChecksums
  ctx$successfulDir              <- localChecks$successfulDir
  ctx$successfulCheckSumFilePath <- localChecks$successfulCheckSumFilePath

  if (!is.null(ctx$reproducible.inputPaths)) {
    outs <- copyFromDPtoReproducibleIPs(
      ctx$targetFilePath, ctx$destinationPathUser, ctx$destinationPath,
      ctx$reproducible.inputPaths, ctx$neededFiles, ctx$archive,
      ctx$checkSums, checkSumFilePath = ctx$checkSumFilePath, ctx$verbose
    )
    ctx$neededFiles         <- outs$neededFiles
    ctx$targetFilePath      <- outs$targetFilePath
    ctx$destinationPath     <- outs$destinationPath
    ctx$destinationPathUser <- outs$destinationPathUser
    ctx$archive             <- outs$archive
  }
  ctx
}

# ---------------------------------------------------------------------------
# Phase 6: Remote hash check
# ---------------------------------------------------------------------------
# When a local copy already exists, decide whether it can stand in for a fresh
# download by consulting the remote source. The semantic rule:
#
#   filesize is a NEGATIVE-only signal. remote.size != local.size proves
#   the bytes differ → download. remote.size == local.size proves nothing
#   on its own (two GeoTIFFs of identical pixel count have identical byte
#   size with arbitrary content), so a positive trust decision REQUIRES a
#   content-hash comparison in the remote's algorithm.
#
# Algorithm:
#   1. If a `.hash` sidecar exists and `reproducible.checkRemoteHash = FALSE`
#      (the default), trust it and skip the network round-trip entirely.
#   2. Otherwise HEAD the remote (`getRemoteMetadata`) for size, hash, and
#      hash algorithm.
#   3. If `remote.size != local.size` → return; let `pp_download` proceed.
#   4. If `remote.algorithm == "etag-opaque"` (HTTP ETag we cannot reproduce
#      locally — no positive trust possible) → return; `pp_download` proceeds.
#   5. Digest local file with the remote's algorithm; if hashes differ →
#      return; `pp_download` proceeds.
#   6. On match: write `.hash` sidecar (`<algo>:<hash>`), upsert the
#      `(file, hash, size, algorithm)` row in CHECKSUMS.txt, mark
#      `skipDownload` and `hashVerified`.
#
# Any network/package error causes a silent fall-through to the normal
# download (best-effort optimisation, not a hard precondition).
pp_remote_hash_check <- function(ctx) {
  if (is.null(ctx$url)) return(ctx)
  # Skip file:// URLs only: they are local (no remote to compare against),
  # and makeRemoteHashFile's URL-to-filename mapping only strips ^https?://,
  # so file:// leaves a colon in the path and fails on Windows. For any
  # other scheme, attempt the metadata fetch — getRemoteMetadata is wrapped
  # in tryCatch below, so unsupported schemes (s3://, gs://, ...) silently
  # fall through to the normal download.
  if (grepl("^file://", ctx$url)) return(ctx)

  # Identify the local file that would be the download target
  localFile <- if (!is.null(ctx$archive) && !isTRUE(is.na(ctx$archive)) &&
                    file.exists(ctx$archive)) {
    ctx$archive
  } else {
    nf <- ctx$neededFiles
    if (!is.null(nf) && length(nf) > 0L) {
      nf <- nf[file.exists(nf)]
      if (length(nf)) nf[[1L]] else NULL
    }
  }
  if (is.null(localFile)) return(ctx)   # nothing local yet — let download proceed

  # ------------------------------------------------------------------
  # Step 1: sidecar fast-path (no network)
  # ------------------------------------------------------------------
  # A `.hash` sidecar from a previous successful match means the file has
  # already been validated against the remote source. By default we trust it
  # and skip the network round-trip entirely. To force a remote re-check on
  # every call (e.g. when the upstream file may change), set
  # `options(reproducible.checkRemoteHash = TRUE)`. Sidecar removal forces
  # re-verification.
  remoteHashFileLocal <- makeRemoteHashFile(
    ctx$url, ctx$destinationPath, basename(localFile), ""
  )
  haveSidecar <- file.exists(remoteHashFileLocal) &&
    isTRUE(suppressWarnings(file.size(remoteHashFileLocal)) > 0L)
  if (haveSidecar &&
      !isTRUE(getOption("reproducible.checkRemoteHash", FALSE))) {
    messagePreProcess(
      "Skipping download; local file matches remote version (cached): ",
      .messageFunctionFn(basename(localFile)), verbose = ctx$verbose
    )
    ctx$skipDownload <- TRUE
    ctx$hashVerified <- unique(c(ctx$hashVerified, localFile))
    if (is.null(ctx$archive) && !is.null(.isArchive(localFile)))
      ctx$archive <- localFile
    return(ctx)
  }

  # ------------------------------------------------------------------
  # Step 2: HEAD remote
  # ------------------------------------------------------------------
  remoteMetadata <- tryCatch(
    getRemoteMetadata(url = ctx$url),
    error = function(e) NULL
  )
  if (is.null(remoteMetadata)) return(ctx)

  # ------------------------------------------------------------------
  # Step 3: size fail-fast (negative-only)
  # ------------------------------------------------------------------
  remoteSize <- suppressWarnings(as.numeric(remoteMetadata$fileSize))
  localSize  <- file.size(localFile)
  if (!is.na(remoteSize) && !is.na(localSize) && remoteSize != localSize) {
    return(ctx)  # bytes definitely differ → let pp_download fetch
  }

  # ------------------------------------------------------------------
  # Step 4: opaque ETag → no positive trust possible → download
  # ------------------------------------------------------------------
  remoteAlgo <- remoteMetadata$remoteAlgorithm
  if (is.null(remoteAlgo) || identical(remoteAlgo, "etag-opaque") ||
      !nzchar(remoteAlgo)) {
    return(ctx)
  }

  # ------------------------------------------------------------------
  # Step 4b: existing sidecar with checkRemoteHash = TRUE — quick check.
  # If the sidecar's recorded hash already matches what the server now
  # advertises, skip re-digesting the local file.
  # ------------------------------------------------------------------
  if (haveSidecar) {
    parsedSidecar <- .parseRemoteHashFile(remoteHashFileLocal)
    if (!is.null(parsedSidecar) &&
        identical(parsedSidecar$hash, remoteMetadata$remoteHash)) {
      messagePreProcess(
        "Skipping download; local file matches remote version: ",
        .messageFunctionFn(basename(localFile)), verbose = ctx$verbose
      )
      ctx$skipDownload   <- TRUE
      ctx$remoteMetadata <- remoteMetadata
      ctx$hashVerified   <- unique(c(ctx$hashVerified, localFile))
      if (is.null(ctx$archive) && !is.null(.isArchive(localFile)))
        ctx$archive <- localFile
      return(ctx)
    }
  }

  # ------------------------------------------------------------------
  # Step 5: digest local file with remote algorithm and compare
  # ------------------------------------------------------------------
  localHash <- tryCatch(
    .digest(localFile, algo = remoteAlgo, quickCheck = FALSE),
    error = function(e) NULL
  )
  if (is.null(localHash) || !length(localHash) ||
      !identical(tolower(localHash[[1L]]),
                 tolower(as.character(remoteMetadata$remoteHash)))) {
    return(ctx)  # hash mismatch (or compute failed) → download
  }

  # ------------------------------------------------------------------
  # Step 6: confirmed match — persist sidecar + CHECKSUMS row, skip download
  # ------------------------------------------------------------------
  makeRemoteHashFile(
    ctx$url, ctx$destinationPath,
    remoteMetadata$targetFile, remoteMetadata$remoteHash,
    algorithm = remoteAlgo, write = TRUE
  )

  # Upsert (file, algorithm) row in CHECKSUMS.txt without re-digesting via
  # Checksums(write=TRUE) — we already have the hash. Preserves any other
  # algorithm rows recorded for this file (e.g. xxhash64 from a prior
  # pp_finalize).
  csfp <- ctx$checkSumFilePath
  if (is.null(csfp) || !nzchar(csfp))
    csfp <- identifyCHECKSUMStxtFile(ctx$destinationPath)
  tryCatch(
    .upsertChecksumsRow(
      checkSumFilePath = csfp,
      file      = makeRelative(localFile, ctx$destinationPath),
      hash      = remoteMetadata$remoteHash,
      filesize  = localSize,
      algorithm = remoteAlgo
    ),
    error = function(e) NULL
  )

  messagePreProcess(
    "Skipping download; local file matches remote version: ",
    .messageFunctionFn(basename(localFile)), verbose = ctx$verbose
  )
  ctx$skipDownload   <- TRUE
  ctx$remoteMetadata <- remoteMetadata
  ctx$hashVerified   <- unique(c(ctx$hashVerified, localFile))
  if (is.null(ctx$archive) && !is.null(.isArchive(localFile)))
    ctx$archive <- localFile

  ctx
}

# ---------------------------------------------------------------------------
# Phase 7: Download
# ---------------------------------------------------------------------------
pp_download <- function(ctx) {
  # ------------------------------------------------------------------
  # Fast path: pp_remote_hash_check confirmed local file == remote.
  # Skip downloadFile() entirely; synthesise a minimal result so the
  # .checkForSimilar / filesToChecksum logic below works unchanged.
  # ------------------------------------------------------------------
  if (isTRUE(ctx$skipDownload)) {
    downloadFileResult <- list(
      downloaded    = ctx$archive,
      archive       = ctx$archive,
      neededFiles   = ctx$neededFiles,
      checkSums     = ctx$checkSums,
      needChecksums = ctx$needChecksums,
      targetFilePath = NULL,
      out           = NULL
    )
  } else {
  # ------------------------------------------------------------------
  # Normal path: call downloadFile()
  # dlFun may be a call/language object (e.g. quote(getDataFn(...))).
  # do.call() EVALUATES language objects in its args list by calling eval() on
  # each element — so passing a call object through do.call executes it
  # immediately. Fix: capture dlFunCaptured as a local variable and call
  # downloadFile via a closure, so dlFun is passed as a lazy promise (variable
  # lookup) rather than being evaluated by do.call.
  dlFunCaptured <- ctx$dlFunCaptured
  dlArgs <- c(
    list(
      archive        = if (isTRUE(is.na(ctx$archive))) NULL else ctx$archive,
      targetFile     = ctx[["targetFile"]],
      neededFiles    = ctx$neededFiles,
      destinationPath = ctx$destinationPath,
      quick          = ctx$quick,
      checkSums      = ctx$checkSums,
      url            = ctx$url,
      checksumFile   = asPath(ctx$checkSumFilePath),
      needChecksums  = ctx$needChecksums,
      overwrite      = ctx$overwrite,
      purge          = ctx$purge,
      alsoExtract    = ctx$alsoExtract,
      verbose        = ctx$verbose,
      .tempPath      = ctx$.tempPath,
      .callingEnv    = ctx$.callingEnv
    ),
    ctx$dots
  )
  # The closure captures dlFunCaptured; do.call only sees non-language args.
  downloadFile_wrapper <- function(...) downloadFile(dlFun = dlFunCaptured, ...)
  downloadFileResult <- do.call(downloadFile_wrapper, dlArgs)
  } # end normal path

  if (!isTRUE(ctx$skipDownload))
    downloadFileResult <- .fixNoFileExtension(
      downloadFileResult = downloadFileResult,
      targetFile = ctx[["targetFile"]], archive = ctx$archive,
      destinationPath = ctx$destinationPath, verbose = ctx$verbose
    )

  if (!is.null(downloadFileResult$targetFilePath))
    ctx$targetFilePath <- makeAbsolute(downloadFileResult$neededFiles, ctx$destinationPath)
  ctx$checkSums     <- downloadFileResult$checkSums
  ctx$needChecksums <- downloadFileResult$needChecksums
  ctx$neededFiles   <- downloadFileResult$neededFiles

  if (!isTRUE(is.na(ctx$archive))) {
    if (identical(downloadFileResult$downloaded, downloadFileResult$archive))
      ctx$archive <- downloadFileResult$downloaded
    if (is.null(ctx$archive))
      ctx$archive <- downloadFileResult$archive
  }

  outFromSimilar  <- .checkForSimilar(
    neededFiles = ctx$neededFiles, alsoExtract = ctx$alsoExtract,
    archive = ctx$archive, targetFile = ctx[["targetFile"]],
    destinationPath = ctx$destinationPath, checkSums = ctx$checkSums,
    checkSumFilePath = ctx$checkSumFilePath, url = ctx$url,
    verbose = ctx$verboseCFS
  )
  ctx$verboseCFS  <- ctx$verbose - 1L
  ctx$neededFiles <- outFromSimilar$neededFiles
  ctx$checkSums   <- outFromSimilar$checkSums

  if (length(ctx$neededFiles) > 1L)
    ctx$alsoExtract <- setdiff(ctx$neededFiles, ctx[["targetFile"]])

  ctx$filesToChecksum <- if (isTRUE(is.na(ctx$archive)) || is.null(ctx$archive)) {
    downloadFileResult$downloaded
  } else {
    ctx$archive
  }
  ctx$downloadResult <- downloadFileResult
  ctx
}

# ---------------------------------------------------------------------------
# Phase 8: Extract from archive (single consolidated call)
# ---------------------------------------------------------------------------
pp_extract <- function(ctx) {
  # Expand any regex patterns in alsoExtract now that the archive is on disk
  # (covers the download path where expansion in pp_resolve_needed_files was
  # skipped because the archive didn't exist yet).  If the expansion changes
  # alsoExtract we also patch ctx$neededFiles: remove the fake absolute paths
  # that makeAbsolute() created from the unresolved pattern strings and replace
  # them with the real expanded paths.
  if (!isNULLorNA(ctx$alsoExtract) &&
      !isNULLorNA(ctx$archive) && any(file.exists(ctx$archive))) {
    fia <- makeRelative(.listFilesInArchive(ctx$archive), ctx$destinationPath)
    if (length(fia)) {
      oldAbs <- makeAbsolute(ctx$alsoExtract, ctx$destinationPath)
      ctx$alsoExtract <- .expandAlsoExtractPatterns(ctx$alsoExtract, fia)
      newAbs <- makeAbsolute(ctx$alsoExtract, ctx$destinationPath)
      if (!identical(oldAbs, newAbs))
        ctx$neededFiles <- unique(c(setdiff(ctx$neededFiles, oldAbs), newAbs))
    }
  }

  filesToChecksum <- unique(c(ctx$filesToChecksum, ctx$neededFiles))
  isOK <- .compareChecksumsAndFilesAddDirs(ctx$checkSums, filesToChecksum, ctx$destinationPath)

  if (isTRUE(!all(isOK))) {
    extracted <- .tryExtractFromArchive(
      archive = ctx$archive, neededFiles = ctx$neededFiles,
      alsoExtract = ctx$alsoExtract, destinationPath = ctx$destinationPath,
      checkSums = ctx$checkSums, needChecksums = ctx$needChecksums,
      checkSumFilePath = ctx$checkSumFilePath, filesToChecksum = filesToChecksum,
      targetFilePath = ctx$targetFilePath, quick = ctx$quick,
      verbose = ctx$verbose, .tempPath = ctx$.tempPath
    )
    ctx$neededFiles    <- extracted$neededFiles
    ctx$checkSums      <- extracted$checkSums
    ctx$filesExtr      <- extracted$filesExtr
    ctx$targetFilePath <- extracted$targetFilePath
    filesToChecksum    <- extracted$filesToChecksum
    ctx$needChecksums  <- extracted$needChecksums
  } else {
    if (!is.null(.isArchive(ctx$archive)))
      messagePreProcess("Skipping extractFromArchive attempt: no files missing",
                        verbose = ctx$verbose)
    if (!is.null(ctx$targetFilePath) && isTRUE(!is.na(ctx$targetFilePath))) {
      if (any(!makeAbsolute(ctx$targetFilePath, ctx$destinationPath) %in%
              makeAbsolute(ctx$neededFiles, ctx$destinationPath))) {
        if (!basename2(ctx$targetFilePath) %in%
            makeRelative(ctx$neededFiles, ctx$destinationPath)) {
          poss <- grep(basename2(ctx$targetFilePath), ctx$neededFiles, value = TRUE)
          if (length(ctx$targetFilePath) > 1L) ctx$targetFilePath <- poss
        }
      }
    }
    ctx$filesExtr <- setdiff(
      unique(c(filesToChecksum, ctx$neededFiles)),
      .isArchive(c(filesToChecksum, ctx$neededFiles))
    )
  }

  ctx$filesExtr       <- unique(c(ctx$filesExtr, filesToChecksum))
  ctx$filesToChecksum <- filesToChecksum

  # Handle nested archives
  if (any(fileExt(ctx$neededFiles) %in% c("zip", "tar", "rar")) &&
      !isTRUE(is.na(ctx$archive))) {
    nestedArchive <- makeAbsolute(
      ctx$neededFiles[fileExt(ctx$neededFiles) %in% c("zip", "tar", "rar")][1L],
      ctx$destinationPath
    )
    messagePreProcess(
      "There are still archives in the extracted files.",
      " preProcess will try to extract the files from ", basename2(nestedArchive), ".",
      " If this is incorrect, please supply archive.",
      verbose = ctx$verbose
    )
    nestedTargetFiles  <- .listFilesInArchive(archive = nestedArchive)
    outFromSimilar     <- .checkForSimilar(
      alsoExtract = ctx$alsoExtract, archive = nestedArchive,
      neededFiles = nestedTargetFiles, destinationPath = ctx$destinationPath,
      checkSums = ctx$checkSums, checkSumFilePath = ctx$checkSumFilePath,
      targetFile = ctx[["targetFile"]], verbose = ctx$verboseCFS
    )
    neededFilesNested  <- outFromSimilar$neededFiles
    ctx$checkSums      <- outFromSimilar$checkSums
    alsoExtractNested  <- if (length(neededFilesNested) > 1L) {
      setdiff(neededFilesNested, ctx[["targetFile"]])
    } else {
      ctx$alsoExtract
    }
    ftcNested <- if (is.null(ctx$archive)) {
      ctx$downloadResult$downloaded
    } else {
      basename2(ctx$archive)
    }
    extractedNested    <- .tryExtractFromArchive(
      archive = nestedArchive, neededFiles = neededFilesNested,
      alsoExtract = alsoExtractNested, destinationPath = ctx$destinationPath,
      checkSums = ctx$checkSums, needChecksums = ctx$needChecksums,
      checkSumFilePath = ctx$checkSumFilePath, filesToChecksum = ftcNested,
      targetFilePath = ctx$targetFilePath, quick = ctx$quick,
      verbose = ctx$verbose, .tempPath = ctx$.tempPath
    )
    feOrig   <- fs::path_rel(ctx$filesExtr, fs::path_common(ctx$filesExtr))
    absFiles <- fs::is_absolute_path(extractedNested$filesExtr)
    feNew    <- c(
      extractedNested$filesExtr[!absFiles],
      fs::path_rel(
        extractedNested$filesExtr[absFiles],
        fs::path_common(extractedNested$filesExtr[absFiles])
      )
    )
    ctx$filesExtr      <- c(setdiff(ctx$filesExtr[!feOrig %in% feNew], ctx$archive),
                            extractedNested$filesExtr)
    ctx$needChecksums  <- extractedNested$needChecksums
    ctx$checkSums      <- extractedNested$checkSums
    ctx$targetFilePath <- extractedNested$targetFilePath
  }
  ctx
}

# ---------------------------------------------------------------------------
# Phase 9: Link files back to user's destinationPath if inputPaths was used
# ---------------------------------------------------------------------------
pp_link_to_destination <- function(ctx) {
  if (is.null(ctx$reproducible.inputPaths)) return(ctx)

  if (!is.null(ctx$destinationPathUser)) {
    foundInInputPaths <- grepl(normPath(ctx$destinationPath), normPath(ctx$filesExtr))
    to <- ctx$targetFilePath
    if (isTRUE(any(foundInInputPaths))) {
      wh <- which(file.exists(ctx$filesExtr[foundInInputPaths]))
      if (length(wh)) {
        from <- ctx$filesExtr[wh]
        to   <- makeAbsolute(makeRelative(from, ctx$destinationPath), ctx$destinationPathUser)
        if (!isTRUE(all(from %in% to)))
          messagePreProcess("...using file(s) in getOption('reproducible.inputPaths')...",
                            verbose = ctx$verbose)
        hardLinkOrCopy(from, to, verbose = ctx$verbose)
        ctx$filesExtr[foundInInputPaths] <- to
      }
    }
    if (!is.null(ctx$targetFilePath) && !identical(to, ctx$targetFilePath)) {
      tmp <- to[basename(to) %in% basename(ctx$targetFilePath)]
      if (any(file.exists(tmp))) {
        ctx$targetFilePath <- tmp
      } else {
        ctx$targetFilePath <- makeAbsolute(
          makeRelative(ctx$targetFilePath, ctx$destinationPath),
          ctx$destinationPathUser
        )
      }
    }
    ctx$destinationPath <- ctx$destinationPathUser
  } else {
    foundInLocalPaths <- grepl(normPath(ctx$destinationPath), normPath(ctx$filesExtr))
    if (isTRUE(any(foundInLocalPaths))) {
      wh <- which(file.exists(ctx$filesExtr[foundInLocalPaths]))
      if (length(wh)) {
        from <- ctx$filesExtr[wh]
        to   <- from
        for (riP in ctx$reproducible.inputPaths) {
          to <- makeAbsolute(makeRelative(from, ctx$destinationPath), riP)
          if (all(from %in% to)) break
        }
        a            <- fread(ctx$checkSumFilePath)
        common       <- ctx$checkSums[ctx$checkSums$expectedFile %in% a$file]
        missingFiles <- common[!a, on = c("expectedFile" = "file", "checksum.x" = "checksum")]
        if (NROW(missingFiles)) {
          messagePreProcess("... linking to getOption('reproducible.destinationPathShared')...",
                            verbose = ctx$verbose)
          hardLinkOrCopy(from, to, verbose = ctx$verbose)
        } else {
          messagePreProcess("Skipping copy from destinationPathShared; all files present",
                            verbose = ctx$verbose)
        }
      }
    }
  }
  ctx
}

# ---------------------------------------------------------------------------
# Phase 10: Finalize — guess target/fun, write checksums, validate, return
# ---------------------------------------------------------------------------
pp_finalize <- function(ctx) {
  targetParams       <- .guessAtTargetAndFun(
    ctx$targetFilePath, ctx$destinationPath,
    filesExtracted = ctx$filesExtr, ctx$fun, verbose = ctx$verbose
  )
  ctx[["targetFile"]]     <- makeRelative(targetParams$targetFilePath, ctx$destinationPath)
  ctx$targetFilePath <- targetParams$targetFilePath
  ctx$funChar        <- targetParams$fun

  if (is.null(ctx$targetFilePath)) {
    ctx$targetFilePath <- if (!is.null(ctx$filesExtr)) ctx$filesExtr else ctx$downloadResult$downloaded
  }
  if (is.null(ctx[["targetFile"]]) && !is.null(ctx$targetFilePath))
    ctx[["targetFile"]] <- makeRelative(ctx$targetFilePath, ctx$destinationPath)

  fun <- .extractFunction(ctx$funChar)

  # Write checksums (replaces 3 scattered on.exit blocks)
  needChecksums <- ctx$needChecksums
  # Drop files that pp_remote_hash_check already validated against the remote
  # sidecar (.hash file) — re-hashing a multi-GB archive locally to populate
  # CHECKSUMS.txt is wasted work; the .hash file is already the authoritative
  # record for this run and any future run. Also drop NAs (which can land in
  # filesToChecksum from the skipDownload path where downloaded == archive == NA)
  # so they don't keep the "write" branch alive on otherwise-empty input.
  filesToChecksum <- ctx$filesToChecksum
  filesToChecksum <- filesToChecksum[!is.na(filesToChecksum)]
  if (length(ctx$hashVerified) && length(filesToChecksum)) {
    fileBases     <- basename2(filesToChecksum)
    verifiedBases <- basename2(ctx$hashVerified)
    filesToChecksum <- filesToChecksum[!fileBases %in% verifiedBases]
  }
  if (needChecksums > 0L && length(filesToChecksum) > 0L) {
    if (needChecksums == 3L) {
      if (identical(ctx$checkSumFilePath, ctx$successfulCheckSumFilePath))
        ctx$checkSumFilePath <- identifyCHECKSUMStxtFile(ctx$successfulDir)
    }
    csps <- ctx$destinationPath
    if (!is.null(ctx$reproducible.inputPaths))
      csps <- c(csps, ctx$reproducible.inputPaths)
    for (csp in csps) {
      csfp <- identifyCHECKSUMStxtFile(csp)
      ctx$checkSums <- appendChecksumsTable(
        checkSumFilePath = csfp,
        filesToChecksum  = basename2(unique(filesToChecksum)),
        destinationPath  = csp,
        append           = needChecksums >= 2L
      )
    }
    if (!is.null(ctx$reproducible.inputPaths) && needChecksums != 3L) {
      ctx$checkSums <- appendChecksumsTable(
        checkSumFilePath = identifyCHECKSUMStxtFile(ctx$reproducible.inputPaths[[1L]]),
        filesToChecksum  = unique(filesToChecksum),
        destinationPath  = ctx$destinationPath,
        append           = needChecksums == 2L,
        verbose          = ctx$verbose - 1L
      )
    }
  }

  if (isTRUE(isDirectory(ctx$url, mustExist = FALSE)) && is.null(ctx[["targetFile"]])) {
    messagePrepInputs(
      "url pointed to a directory, but no `targetFile` specified; using targetFilePath:\n",
      paste0(ctx$downloadResult$downloaded, collapse = "\n")
    )
    ctx$targetFilePath <- ctx$downloadResult$downloaded
  }

  failStop <- FALSE
  if (is.null(ctx$targetFilePath)) {
    failStop <- TRUE
  } else if (isTRUE(all(is.na(ctx$targetFilePath)))) {
    if (length(ctx$targetFilePath) > 1L) ctx$targetFilePath <- NA
  } else if (!isTRUE(all(file.exists(ctx$targetFilePath)))) {
    failStop <- TRUE
  }
  if (isTRUE(failStop))
    stop("targetFile appears to be misspecified at: ", ctx$targetFilePath, ". ",
         "Possibly, it does not exist in the specified archive, ",
         "or the file doesn't exist in destinationPath")

  # Only mark rows as "ArchiveOK" when there is actually an archive. With
  # archive = NA, makeRelative(NA, ...) is NA and `%in% NA` matches any
  # NA actualFile rows, producing a spurious assignment.
  if (!isNULLorNA(ctx$archive)) {
    archiveInChecksums <- ctx$checkSums$actualFile %in%
      makeRelative(ctx$archive, ctx$destinationPath)
    if (any(archiveInChecksums))
      ctx$checkSums[which(archiveInChecksums), result := "ArchiveOK"]
  }

  list(
    checkSums       = ctx$checkSums,
    dots            = ctx$dots,
    fun             = fun,
    funChar         = ctx$funChar,
    targetFilePath  = ctx$targetFilePath,
    destinationPath = ctx$destinationPath,
    object          = ctx$downloadResult$object
  )
}

#' Purge individual line items from checksums file
#'
#' @inheritParams downloadFile
#' @keywords internal
#' @rdname purge
.purge <- function(checkSums, purge, targetFile, archive, alsoExtract, url, destinationPath) {
  purgeChar <- as.character(purge)
  checkSums <- tryCatch(
    switch(purgeChar,
           "2" = checkSums[!(checkSums$expectedFile %in% makeRelative(targetFile, destinationPath)), ],
           "3" = checkSums[!(checkSums$expectedFile %in% makeRelative(archive, destinationPath)), ],
           "4" = checkSums[!(checkSums$expectedFile %in% makeRelative(alsoExtract, destinationPath)), ],
           "5" = checkSums[!(checkSums$expectedFile %in% makeRelative(unique(c(targetFile, alsoExtract)), destinationPath)), ], # nolint
           "6" = checkSums[!(checkSums$expectedFile %in%
                               makeRelative(unique(c(targetFile, alsoExtract, archive)), destinationPath)), ], # nolint
           "7" = checkSums[!(checkSums$expectedFile %in%
                               makeRelative(unique(c(targetFile, alsoExtract, archive, url)), destinationPath)), ] # nolint
    ),
    error = function(x) checkSums
  )
  checkSums
}

#' @keywords internal
.emptyChecksumsResult <- data.table::data.table(
  expectedFile = character(),
  actualFile = character(),
  result = character()
)

#' @keywords internal
.emptyChecksumsFileContent <- data.frame(
  file = character(),
  checksum = character(),
  filesize = character(),
  algorithm = character()
)

#' Decide how to materialize a captured `dlFun` expression
#'
#' `captured` is the result of `substitute(dlFun)`; `lazy` is the (still-promise)
#' formal `dlFun` argument. See preProcess() for the three cases this handles.
#' @keywords internal
#' @noRd
.resolveDlFunCaptured <- function(captured, lazy) {
  if (is.call(captured)) {
    head <- captured[[1L]]
    # quote(...) / bquote(...) — unwrap to the inner call
    if (is.symbol(head) && as.character(head) %in% c("quote", "bquote"))
      return(eval(captured))
    # pkg::fn(...) or pkg:::fn(...) — a namespaced function call to defer
    if (is.call(head) && is.symbol(head[[1L]]) &&
        as.character(head[[1L]]) %in% c("::", ":::"))
      return(captured)
    # fn(...) — a plain function call to defer, unless `fn` is a control-flow
    # / special-form name (in which case the expression must be evaluated now)
    controlFlow <- c("if", "for", "while", "repeat", "{", "(",
                     "&&", "||", "<-", "<<-", "=", "function", "~", "@",
                     "$", "[", "[[", "::", ":::")
    if (is.symbol(head) && !(as.character(head) %in% controlFlow))
      return(captured)
  }
  lazy  # symbol, literal, or control-flow call: force the promise
}

#' @keywords internal
#' @importFrom utils getFromNamespace
.extractFunction <- function(fun, envir = parent.frame()) {
  if (!is.null(fun)) {
    if (is.name(fun)) {
      possFun <- get0(fun, envir = envir)
      if (is.null(possFun)) {
        env <- .whereInStack(fun)
        fun <- get(fun, envir = env)
      }
    }
    if (is.call(fun)) {
      fun
    } else {
      suppressWarnings(isNAFun <- is.na(fun))
      if (!any(isNAFun)) {
        if (!is.function(fun)) {
          if (any(grepl("::", fun))) {
            fun2 <- strsplit(fun, "::")[[1]]
            pkg <- fun2[1]
            fun <- fun2[2]
            fun <- getFromNamespace(fun, pkg)
          } else {
            fun <- get(fun, envir)
          }
        }
      }
    }
  }
  fun
}

#' @keywords internal
.guessAtFile <- function(url, archive, targetFile, destinationPath,
                         verbose = getOption("reproducible.verbose", 1), team_drive = NULL) {
  # if (is.null(targetFile)) {
  guessedFile <- if (!is.null(url)) {
    if (isTRUE(isDirectory(url, FALSE))) {
      gf <- NULL
    } else {
      gf <- file.path(destinationPath, basename2(url))
    }

    # Test for just Google ID supplied
    isGID <- isGoogleID(url)
    # isGID <- all(grepl("^[A-Za-z0-9_-]{33}$", url), # Has 33 characters as letters, numbers or - or _
    #              !grepl("\\.[^\\.]+$", url))
    # doesn't have an extension
    if (any(isGoogleDriveURL(url), isGID)) {
      if (isGID) messagePreProcess("url seems to be a Google Drive ID", verbose = verbose, verboseLevel = 2)

      # if (grepl("drive.google.com", url)) {
      # ie <- isTRUE(internetExists())
      # if (ie) {
      gf <- assessGoogle(
        url = url, archive = archive, targetFile = targetFile,
        destinationPath = destinationPath, verbose = verbose, team_drive = NULL
      )
      gf <- makeAbsolute(gf, destinationPath)
      # }
    }
    gf
  } else {
    NULL
  }
  if (!is.null(guessedFile))
    guessedFile <- normPath(guessedFile)
  guessedFile
}


isGoogleID <- function(url) {
  all(grepl("^[A-Za-z0-9_-]{33}$", url), # Has 33 characters as letters, numbers or - or _
      !grepl("\\.[^\\.]+$", url)) # ||
    # grepl("drive.google.com", url)
}

isGoogleDriveURL <- function(url) {
  grepl("drive.google.com", url)
}
# COPIED FROM REQUIRE
# urlExists <- function(url) {
#   con <- url(url)
#   on.exit(try(close(con), silent = TRUE), add = TRUE)
#   for (i in 1:5) {
#     a <- try(suppressWarnings(readLines(con, n = 1)), silent = TRUE)
#     try(close(con), silent = TRUE)
#     ret <- if (is(a, "try-error")) FALSE else TRUE
#     if (isTRUE(ret)) {
#       break
#     } else {
#       Sys.sleep(0.1)
#     }
#   }
#   ret
# }


#' @keywords internal
.checkSumsUpdate <- function(destinationPath, newFilesToCheck, checkSums,
                             checkSumFilePath = NULL, verbose = getOption("reproducible.verbose", 1)) {
  if (!is.null(newFilesToCheck)) {
    if (is.null(checkSumFilePath) || length(checkSumFilePath) == 0) {
      checkSumFilePath <- identifyCHECKSUMStxtFile(destinationPath)
    }
    if (!file.exists(checkSumFilePath)) {
      checkSums
    } else {
      checkSums2 <- try(Checksums(
        path = destinationPath, write = FALSE,
        files = newFilesToCheck,
        checksumFile = checkSumFilePath,
        verbose = verbose - 1
      ), silent = TRUE)
      if (!is(checkSums2, "try-error")) {
        checkSums <- rbindlist(list(checkSums, checkSums2))
        data.table::setkey(checkSums, result)
        checkSums <- unique(checkSums, fromLast = TRUE, by = "expectedFile")
        checkSums <- rbindlist(list(
          checkSums[compareNA("OK", result)],
          checkSums[compareNA("FAIL", result)],
          checkSums[is.na(result)]
        ))
      } else {
        stop("checkSumFilePath is not a CHECKSUMS.txt file")
      }
    }
  }
  checkSums
}

#' @keywords internal
.similarFilesInCheckSums <- function(file, checkSums, alsoExtract) {
  if (NROW(checkSums)) {
    if (!missing(alsoExtract)) {
      file <- unique(c(file, alsoExtract))
    }
    anySimilarInCS <- vapply(filePathSansExt(file), function(fi) {
      res <- checkSums[grepl(paste0(fi, "\\."), checkSums$expectedFile), ]$result
      res <- na.omit(res)
      if (length(res) == 0) res <- "NotOK"
      isTRUE(all(compareNA("OK", res)))
    }, FUN.VALUE = logical(1))

    all(anySimilarInCS)
  } else {
    FALSE
  }
}

#' Expand regex patterns in `alsoExtract` against a known list of archive files.
#'
#' For each element of `alsoExtract` that is not a literal match in
#' `filesInArchive` (by relative path or basename), the element is treated as a
#' regular expression and `grep()`-ed against `filesInArchive`.  Matched names
#' replace the pattern; unmatched patterns are kept as-is so that downstream
#' code can emit a useful "file not found" error.  Special sentinel values
#' (`"similar"`, `"none"`, `NA`) are passed through unchanged.
#'
#' This lets users write e.g.
#' `alsoExtract = "CMD_sm|CMD_sp"` to select all archive members whose name
#' contains `CMD_sm` or `CMD_sp`.
#'
#' @param alsoExtract Character vector (or `NULL`).
#' @param filesInArchive Character vector of files inside the archive, as
#'   returned by `makeRelative(.listFilesInArchive(archive), destinationPath)`.
#' @return A character vector with pattern elements replaced by their matches.
#' @keywords internal
.expandAlsoExtractPatterns <- function(alsoExtract, filesInArchive) {
  if (is.null(alsoExtract) || length(alsoExtract) == 0L ||
      length(filesInArchive) == 0L) return(alsoExtract)
  out <- character(0L)
  for (pat in alsoExtract) {
    if (is.na(pat) || pat %in% c("similar", "none")) {
      out <- c(out, pat)
      next
    }
    # Literal match by relative path or basename → keep as-is
    if (pat %in% filesInArchive || pat %in% basename2(filesInArchive)) {
      out <- c(out, pat)
    } else {
      # Treat as regex pattern; grep against archive member names
      matched <- grep(pat, filesInArchive, value = TRUE)
      out <- c(out, if (length(matched)) matched else pat)
    }
  }
  out
}

#' @keywords internal
.checkForSimilar <- function(neededFiles, alsoExtract, archive, targetFile,
                             destinationPath, checkSums, checkSumFilePath,
                             url, verbose = getOption("reproducible.verbose", 1)) {
  # No archive => the alsoExtract / "files in archive" messaging below is not
  # meaningful; return inputs unchanged so callers (e.g. archive = NA) don't
  # see spurious "alsoExtract is unspecified" / "Checksumming all files in
  # archive" output.
  if (isNULLorNA(archive))
    return(list(neededFiles = neededFiles, checkSums = checkSums))
  lookForSimilar <- FALSE
  if (is.null(alsoExtract) || length(alsoExtract) == 0) {
    messagePreProcess("alsoExtract is unspecified; assuming that all files must be extracted",
                      verbose = verbose)
    lookForSimilar <- "all"
  } else {
    if (!all(is.na(alsoExtract))) {
      if ("similar" %in% basename2(alsoExtract)) {
        lookForSimilar <- TRUE
      }
    }
  }

  if (isTRUE(lookForSimilar) || ("all" %in% lookForSimilar && !is.null(archive))) {
    allFiles <- .listFilesInArchive(archive)
    archiveFilesInCS <- allFiles %in% checkSums$expectedFile
    rerunChecksums <- TRUE
    if (any(archiveFilesInCS)) {
      if (all(archiveFilesInCS)) {
        isOK <- checkSums[expectedFile %in% allFiles][, isOK := result %in% "OK"]#$result %in% "OK"
        if (isTRUE(all(isOK$isOK))) {
          rerunChecksums <- FALSE
        } else { # some not OK, but present
          allFiles <- isOK$expectedFile[isOK$isOK %in% FALSE]
        }
      } else { # some files in the archive are not yet in checkSums -- rerunChecksums on these
        allFiles <- allFiles[!archiveFilesInCS]
      }
    }
    if (rerunChecksums) {
      neededFiles <- checkRelative(neededFiles, destinationPath, allFiles)
      if (is.null(targetFile) || isTRUE(all(is.na(targetFile)))) {
        messagePreProcess("No targetFile supplied. ", messageChecksummingAllFiles,
                          verbose = verbose)
        neededFiles <- allFiles
      } else if ("all" %in% lookForSimilar) {
        messagePreProcess(messageChecksummingAllFiles, verbose = verbose)
        neededFiles <- allFiles
      } else {
        allOK <- .similarFilesInCheckSums(targetFile, checkSums, alsoExtract)
        if (!allOK) {
          filePatternToKeep <- gsub(basename2(targetFile),
                                    pattern = fileExt(basename2(targetFile)), replacement = ""
          )
          filesToGet <- grep(allFiles, pattern = filePatternToKeep, value = TRUE)
          neededFiles <- c(neededFiles, filesToGet)
        }
      }
      if (exists("filesToGet", inherits = FALSE)) {
        if (length(filesToGet) == 0) {
          rerunChecksums <- FALSE
        }
      }
      neededFiles <- unique(makeAbsolute(neededFiles, destinationPath))

      if (!is.null(neededFiles) && rerunChecksums) {
        checkSums <- .checkSumsUpdate(
          destinationPath = destinationPath, newFilesToCheck = neededFiles,
          checkSums = checkSums,
          checkSumFilePath = checkSumFilePath, verbose = verbose
        )
      }
    }
  }
  list(neededFiles = neededFiles, checkSums = checkSums)
}

#' @keywords internal
.checkLocalSources <- function(neededFiles, checkSumFilePath, checkSums, otherPaths, needChecksums,
                               destinationPath, verbose = getOption("reproducible.verbose", 1)) {
  foundInInputPaths <- character()
  successfulCheckSumFilePath <- character()
  successfulDir <- character()
  if (!is.null(neededFiles)) {
    filesInHand <- checkSums[compareNA(checkSums$result, "OK"), ]$expectedFile
    neededFilesRel <- makeRelative(neededFiles, destinationPath)
    if (!all(neededFilesRel %in% filesInHand)) {
      for (op in otherPaths) {
        recursively <- .getDestinationPathSharedRecursive()
        opFiles <- dir(op, recursive = recursively, full.names = TRUE)
        if (any(neededFilesRel %in% basename2(opFiles))) {
          isNeeded <- basename2(opFiles) %in% neededFilesRel
          dirNameOPFiles <- dirname(opFiles[isNeeded])

          # foundRecursively <- dirNameOPFiles != dirname(opFiles[isNeeded])
          # foundRecursively <- basename2(opFiles[isNeeded][foundRecursively])

          uniqueDirsOPFiles <- rev(unique(dirNameOPFiles))

          checkSumFilePathTry <- checkSumFilePath
          # check CHECKSUMS.txt files, first the one in destinationPath, then ones in inputPaths
          for (dirOPFiles in uniqueDirsOPFiles) {
            # for (i in seq(1 + length(uniqueDirsOPFiles))) {
            checkSumFilePathTry <- identifyCHECKSUMStxtFile(dirOPFiles)
            checkSumsInputPath <- suppressMessages(
              Checksums(
                path = dirOPFiles, write = FALSE,
                files = neededFilesRel,
                checksumFile = checkSumFilePathTry,
                verbose = verbose
              )
            )
            isOK <- checkSumsInputPath[checkSumsInputPath$expectedFile %in% neededFilesRel, ]$result
            if (length(isOK)) {
              if (all(compareNA(isOK, "OK"))) {
                needChecksums <- 3 # Basically this means that we *may* have to update
                #   checksums file in either destinationPath or
                #   options("reproducible.inputPaths")
                successfulCheckSumFilePath <- checkSumFilePathTry
                successfulDir <- unique(dirOPFiles)
                break
              }
            }

            checkSumFilePathTry <- identifyCHECKSUMStxtFile(dirOPFiles)
          }
          checkSumsIPOnlyNeeded <- checkSumsInputPath[compareNA(checkSumsInputPath$result, "OK"), ]
          filesInHandIP <- checkSumsIPOnlyNeeded$expectedFile
          filesInHandIPLogical <- neededFilesRel %in% filesInHandIP
          if (any(filesInHandIPLogical)) {
            outHLC <- hardLinkOrCopy(
              from = makeAbsolute(filesInHandIP, dirOPFiles),
              to = neededFiles[filesInHandIPLogical], verbose = verbose
            )
            checkSums <- rbindlist(list(checkSumsIPOnlyNeeded, checkSums))
            checkSums <- unique(checkSums, by = "expectedFile")
          }
          foundInInputPaths <- c(foundInInputPaths, filesInHandIP)
          if (isTRUE(all(filesInHandIPLogical))) {
            break
          }
        }
      }
    }
    # do a check here that destinationPath is already the destinationPathShared
    #   need to emulate the above behaviour
    reproducible.inputPaths <- .getDestinationPathShared()
    if (!is.null(reproducible.inputPaths)) {
      reproducible.inputPaths <- normPath(reproducible.inputPaths)
    }

    if (identical(destinationPath, reproducible.inputPaths)) {
      foundInInputPaths <- filesInHand
      successfulDir <- destinationPath
      successfulCheckSumFilePath <- checkSumFilePath
    }
  }
  list(
    checkSums = checkSums, needChecksums = needChecksums,
    # foundRecursively = foundRecursively,
    neededFiles = neededFiles,
    successfulCheckSumFilePath = successfulCheckSumFilePath,
    successfulDir = successfulDir,
    foundInInputPaths = makeAbsolute(foundInInputPaths, destinationPath)
  )
}

#' Hardlink, symlink, or copy a file
#'
#' Attempt first to make a hardlink. If that fails, try to make
#' a symlink (on non-windows systems and `symlink = TRUE`).
#' If that fails, copy the file.
#'
#' @note Use caution with files-backed objects (e.g., rasters). See examples.
#'
#' @param from,to  Character vectors, containing file names or paths.
#'                 `to` can alternatively be the path to a single existing directory.
#' @param symlink  Logical indicating whether to use symlink (instead of hardlink).
#'                 Default `FALSE`.
#' @inheritParams prepInputs
#' @seealso [file.link()], [file.symlink()], [file.copy()].
#'
#' @author Alex Chubaty and Eliot McIntire
#' @export
#' @return
#' This function is called for its side effects, which will be a `file.link` is that
#' is available or `file.copy` if not (e.g., the two directories are not on the
#' same physical disk).
#'
#' @examples
#'
#' tmpDir <- file.path(tempdir(), "symlink-test")
#' tmpDir <- normalizePath(tmpDir, winslash = "/", mustWork = FALSE)
#' dir.create(tmpDir)
#'
#' f0 <- file.path(tmpDir, "file0.csv")
#' write.csv(iris, f0)
#'
#' d1 <- file.path(tmpDir, "dir1")
#' dir.create(d1)
#' write.csv(iris, file.path(d1, "file1.csv"))
#'
#' d2 <- file.path(tmpDir, "dir2")
#' dir.create(d2)
#' f2 <- file.path(tmpDir, "file2.csv")
#'
#' ## create link to a file
#' linkOrCopy(f0, f2)
#' file.exists(f2) ## TRUE
#' identical(read.table(f0), read.table(f2)) ## TRUE
#'
#' ## deleting the link shouldn't delete the original file
#' unlink(f0)
#' file.exists(f0) ## FALSE
#' file.exists(f2) ## TRUE
#'
#' if (requireNamespace("terra", quietly = TRUE)) {
#'   ## using spatRasters and other file-backed objects
#'   f3a <- system.file("ex/test.grd", package = "terra")
#'   f3b <- system.file("ex/test.gri", package = "terra")
#'   r3a <- terra::rast(f3a)
#'   f4a <- file.path(tmpDir, "raster4.grd")
#'   f4b <- file.path(tmpDir, "raster4.gri")
#'   linkOrCopy(f3a, f4a) ## hardlink
#'   linkOrCopy(f3b, f4b) ## hardlink
#'   r4a <- terra::rast(f4a)
#'
#'   isTRUE(all.equal(r3a, r4a)) # TRUE
#'
#'   ## cleanup
#'   unlink(tmpDir, recursive = TRUE)
#' }
linkOrCopy <- function(from, to, symlink = TRUE, overwrite = TRUE,
                       verbose = getOption("reproducible.verbose", 1)) {
  existsLogical <- file.exists(from)
  existsTo <- file.exists(to)

  toCollapsed <- paste(to, collapse = "\n")
  fromCollapsed <- paste(from, collapse = "\n")
  result <- TRUE

  ## if the filename is the same, you can't copy from self to self
  if (!all(normPath(to) %in% normPath(from))) {
    if (any(existsLogical)) {
      toDirs1 <- unique(dirname(to))
      dirDoesntExist1 <- !dir.exists(toDirs1)

      ## some directories in `from` won't look like directories in the prev `to`
      ##  e.g., test/folder1/folder2
      ##     folder2 could be a file or a dir, and "dir.exists(to)" won't know which
      ##     because test/folder1 didn't exist
      ## So, identify the dirs in the `from`, and those ones will also be dirs in `to`
      fromDirs <- dir.exists(from)
      toDirs2 <- to[fromDirs]
      dirDoesntExist2 <- rep(TRUE, length(toDirs2))

      if (any(dirDoesntExist1) || any(dirDoesntExist2)) {
        needCreate <- unique(c(toDirs1[dirDoesntExist1], toDirs2[dirDoesntExist2]))
        if (any(is.na(needCreate))) {
          needCreate <- na.omit(needCreate)
        }
        lapply(needCreate, dir.create, recursive = TRUE)
      }
      isDir <- dir.exists(to)
      dups <- duplicated(from)

      if (any(existsTo)) {
        unlink(to[existsTo])
      }
      # Try hard link first -- the only type that R deeply recognizes
      result <- captureWarningsToAttr(
        file.link(from[!dups & !isDir], to[!dups & !isDir])
      )
      warns <- attr(result, "warning")
      attr(result, "warning") <- NULL

      if (isTRUE(all(result))) {
        messagePreProcess("Hardlinked ", hardlinkOrSymlinkMessagePrefix, ":", verbose = verbose)
        messagePreProcess("\n", toCollapsed, "\n",
                          whPointsToMess, "\n",
                          fromCollapsed,
                          verbose = verbose - 1
        )
        messagePreProcess(messageNoCopyMade, verbose = verbose)
      } else {
        if ((
          grepl("cannot link.+different disk drive", warns) ||
          grepl("Invalid cross-device link", warns)) && !isTRUE(symlink)) {
          messageColoured("An attempt was made to use hard links to make a quick pointer ",
                          "from one (set of) file(s) to another; \nthis is not possible because ",
                          "the files would be on different drives. Consider changing the paths\n",
                          "so that they will be on the same physical drive", colour = "red",
                          verbose = verbose)
          if (verbose > 0)
            message(warns)
        }
      }

      if (any(grepl("file already exists", warns))) {
        messagePreProcess("File named ", toCollapsed, " already exists; will try to use it/them", verbose = verbose)
        result <- TRUE
      }

      # On *nix types -- try symlink
      if (isFALSE(all(result)) && isTRUE(symlink)) {
        if (!isWindows()) {
          result <- suppressWarnings(file.symlink(from[!result], to[!result]))
          if (isTRUE(all(result))) {
            messagePreProcess("Symlinked", hardlinkOrSymlinkMessagePrefix, verbose = verbose)
            messagePreProcess("\n", toCollapsed, "\n",
                              whPointsToMess, "\n",
                              fromCollapsed,
                              verbose = verbose - 1
            )
            messagePreProcess(messageNoCopyMade, verbose = verbose)
          }
        }
      }

      if (isFALSE(all(result))) {
        len <- length(from[!result])
        # if (len < 50) fromCollapsed[!result] else c(head(fromCollapsed[!result]), tail(fromCollapsed[!result]))
        if (len < 50) {
          fromMess <- from[!result]
          toMess <- to[!result]
        } else {
          fromMess <- c(head(fromCollapsed[!result]), tail(fromCollapsed[!result]))
          toMess <- c(head(toCollapsed[!result], 24), "... (omitting many)", tail(toCollapsed[!result], 24))
        }
        result2 <- try(file.copy(from[!result], to[!result], overwrite = overwrite))
        if (is(result2, "try-error")) browser()

        toMessCollapsed <- paste(toMess, collapse = "\n")
        fromMessCollapsed <- paste(fromMess, collapse = "\n")

        messagePreProcess(singularPlural(c("Copy", "Copies"), l = fromMess), " of ", singularPlural(c("file: ", "files: "), l = fromMess),
                          "\n", fromMessCollapsed, ",\n",
                          singularPlural(c("was", "were"), l = fromMess)," created at:\n",
                          toMessCollapsed, verbose = verbose)
      }
    } else {
      messagePreProcess("File ", fromCollapsed, " does not exist. Not copying.", verbose = verbose)
      result <- FALSE
    }
  }
  return(result)
}

#' @keywords internal
.tryExtractFromArchive <- function(archive,
                                   neededFiles,
                                   filesToChecksum,
                                   alsoExtract,
                                   destinationPath,
                                   checkSums,
                                   needChecksums,
                                   checkSumFilePath,
                                   targetFilePath,
                                   quick, verbose = getOption("reproducible.verbose", 1),
                                   .tempPath) {
  if (missing(.tempPath)) {
    .tempPath <- tempdir2(rndstr(1, 6))
    on.exit(
      {
        unlink(.tempPath, recursive = TRUE)
      },
      add = TRUE
    )
  }
  alsoExtract <- grep("none$", alsoExtract, value = TRUE, invert = TRUE) # remove "none" from neededFiles; for extracting
  neededFiles <- c(neededFiles, if (!isNULLorNA(alsoExtract)) alsoExtract)
  neededFiles <- setdiff(neededFiles, "similar") # remove "similar" from neededFiles; for extracting
  neededFiles <- unique(makeAbsolute(neededFiles, destinationPath)) # unique is b/c neededFiles was absolute and alsoExtract was rel
  alsoExtract <- makeAbsolute(alsoExtract, destinationPath)

  filesExtr <- NULL
  if (!isNULLorNA(archive)) {
    # if (!is.na(archive)) {
    if (any(file.exists(archive))) {
      filesExtracted <- extractFromArchive(
        archive = archive,
        destinationPath = destinationPath,
        neededFiles = neededFiles,
        checkSums = checkSums,
        needChecksums = needChecksums,
        checkSumFilePath = checkSumFilePath,
        quick = quick,
        verbose = verbose,
        .tempPath = .tempPath
      )
      neededFiles <- filesExtracted$neededFiles # will have been potentially corrected if user supplied incorrect relative paths

      targetFilePath <- checkRelative(targetFilePath, destinationPath, neededFiles, verbose = verbose - 1)
      filesToChecksum <- checkRelative(filesToChecksum, destinationPath, neededFiles, verbose = verbose - 1)

      checkSums <- .checkSumsUpdate(
        destinationPath = destinationPath,
        newFilesToCheck = filesExtracted$filesExtracted,
        checkSums = filesExtracted$checkSums, verbose = verbose
      )

      # filesToChecksum may be wrong because of relative path without subfolder
      filesToChecksum <- unique(c(
        filesToChecksum, targetFilePath, # alsoExtract, # alsoExtract will be in filesExtracted$filesExtracted
        filesExtracted$filesExtracted
      ))

      needChecksums <- filesExtracted$needChecksums
      data.table::setDT(filesExtracted$checkSums)
      dontNeedChecksums <- if (NROW(filesExtracted$checkSums) > 0) {
        filesExtracted$checkSums[
          filesExtracted$checkSums$expectedFile %in%
            filesToChecksum & compareNA(result, "OK"),
          expectedFile
        ]
      } else {
        dontNeedChecksums <- character()
      }
      filesToChecksum <- setdiff(filesToChecksum, dontNeedChecksums)

      if (needChecksums > 0) {
        checkSums <- appendChecksumsTable(
          checkSumFilePath = checkSumFilePath,
          filesToChecksum = unique(filesToChecksum),
          destinationPath = destinationPath,
          append = needChecksums >= 2
        )
        needChecksums <- 0
      }

      ## targetFilePath might still be NULL, need destinationPath too
      filesExtr <- unique(c(
        filesToChecksum,
        if (is.null(filesExtracted$filesExtracted) ||
            length(filesExtracted$filesExtracted) == 0) {
          character()
        } else {
          filesExtracted$filesExtracted
        }
      ))
    }
    # }
  }
  if (!is.null(filesExtr)) {
    filesExtr <- unique(filesExtr)
  }
  list(
    filesToChecksum = filesToChecksum, filesExtr = filesExtr,
    needChecksums = needChecksums,
    targetFilePath = targetFilePath,
    neededFiles = neededFiles, checkSums = checkSums
  )
}

#' @keywords internal
.decodeMagicNumber <- function(magicNumberString) {
  fileExt <- if (any(grepl(pattern = "Zip", x = magicNumberString))) {
    ".zip"
  } else if (any(grepl(pattern = "RAR", x = magicNumberString))) {
    ".rar"
  } else if (any(grepl(pattern = "tar", x = magicNumberString))) {
    ".tar"
  } else if (any(grepl(pattern = "TIFF", x = magicNumberString))) {
    ".tif"
  } else if (any(grepl(pattern = "Shapefile", x = magicNumberString))) {
    ".shp"
  } else {
    NULL
  }
  return(fileExt)
}

#' @keywords internal
.guessFileExtension <- function(file) {
  if (isWindows()) {
    tryCatch(
      {
        possLocs <- c(
          "C:/cygwin/bin/file.exe",
          "C:\\cygwin64/bin/file.exe"
        )
        findFile <- file.exists(possLocs)
        if (any(findFile)) {
          fileLoc <- possLocs[findFile][1]
        }

        magicNumber <- captureWarningsToAttr(
          system(paste(fileLoc, file), intern = TRUE)
        )
        warn <- attr(magicNumber, "warning")
        attr(magicNumber, "warning") <- NULL

        if (length(warn) > 0) {
          splitted <- unlist(strsplit(x = file, split = ":/"))
          fileAdapted <- file.path(paste0("/mnt/", tolower(splitted[1])), splitted[2])
          magicNumber <- captureWarningsToAttr(
            shell(paste0("'file ", fileAdapted, "'"), "bash",
                  intern = TRUE,
                  wait = TRUE, translate = FALSE, mustWork = TRUE
            )
          )
          warn <- attr(magicNumber, "warning")
          attr(magicNumber, "warning") <- NULL
        }
        fileExt <- if (length(warn) == 0) {
          .decodeMagicNumber(magicNumberString = magicNumber)
        } else {
          NULL
        }
        return(fileExt)
      },
      error = function(e) {
        fileExt <- NULL
        return(fileExt)
      }
    )
  } else {
    magicNumber <- system(paste0("file ", file), wait = TRUE, intern = TRUE)
    fileExt <- .decodeMagicNumber(magicNumberString = magicNumber)
    return(fileExt)
  }
}

#' @keywords internal
.fixNoFileExtension <- function(downloadFileResult, targetFile, archive,
                                destinationPath, verbose = getOption("reproducible.verbose", 1)) {
  needFinalCopy <- TRUE
  if (!is.null(downloadFileResult$downloaded) &&
      identical(fileExt(downloadFileResult$downloaded), "")) {
    if (!is.null(targetFile) && !identical(fileExt(normPath(basename2(downloadFileResult$neededFiles))), "")) {
      if (is.null(archive)) {
        messagePreProcess(
          "Downloaded file has no extension: targetFile is provided, but archive is not.\n",
          " Downloaded file will be considered as the targetFile. If the downloaded file is an archive\n",
          " that contains the targetFile, please specify both archive and targetFile.",
          verbose = verbose
        )
        newFileWithExtension <- downloadFileResult$neededFiles
      } else {
        messagePreProcess(
          "Downloaded file has no extension: both targetFile and archive are provided.\n",
          " Downloaded file will be considered as the archive.",
          verbose = verbose
        )
        newFileWithExtension <- downloadFileResult$archive
      }
    } else {
      if (!is.null(archive)) {
        messagePreProcess(
          "Downloaded file has no extension: archive is provided. \n",
          " downloaded file will be considered as the archive.",
          verbose = verbose
        )
        downloadFileResult$neededFiles <- archive
        newFileWithExtension <- downloadFileResult$neededFiles
      } else {
        messagePreProcess(
          "Downloaded file has no extension: neither archive nor targetFile are provided. \n",
          "prepInputs will try accessing the file type.",
          verbose = verbose
        )
        fileExt <- .guessFileExtension(file = normPath(downloadFileResult$downloaded))
        if (is.null(fileExt)) {
          messagePreProcess("The file was not recognized by prepInputs. ",
                            "Will assume the file is an archive and add '.zip' extension. ",
                            "If this is incorrect or return error, please supply archive or targetFile",
                            verbose = verbose
          )
          fileExt <- ".zip"
        }
        newFileWithExtension <- paste0(downloadFileResult$neededFiles, fileExt)
        hardLinkOrCopy(downloadFileResult$neededFiles, newFileWithExtension, verbose = 0)
        needFinalCopy <- FALSE
        downloadFileResult$neededFiles <- makeAbsolute(.listFilesInArchive(newFileWithExtension), destinationPath)
        downloadFileResult$archive <- newFileWithExtension
        downloadFileResult$targetFilePath <- normPath(downloadFileResult$neededFiles)
      }
    }
    if (isTRUE(needFinalCopy)) {
      hardLinkOrCopy(
        verbose = 0,
        from = normPath(downloadFileResult$downloaded),
        to = normPath(newFileWithExtension)
      )
    }
    downloadFileResult$downloaded <- newFileWithExtension
  }
  downloadFileResult
}

moveAttributes <- function(source, receiving, attrs = NULL) {
  if (!is.null(receiving)) {
    sourceAttributes <- attributes(source)
    if (length(sourceAttributes) > 0) {
      if (!is.null(attrs)) {
        sourceAttributes <- attrs
      }

      for (i in length(sourceAttributes)) {
        attr(receiving, names(sourceAttributes)[i]) <- sourceAttributes[[i]]
      }
    }
  }
  receiving
}

.checkDeprecated <- function(dots, verbose = getOption("reproducible.verbose", 1)) {
  if (!is.null(dots$cacheTags)) {
    messagePreProcess("cacheTags is being deprecated;",
                      " use userTags which will pass directly to Cache.",
                      verbose = verbose
    )
    dots$userTags <- dots$cacheTags
    dots$cacheTags <- NULL
  }
  if (!is.null(dots$postProcessedFilename)) {
    messagePreProcess("postProcessedFilename is being deprecated;",
                      " use filename2, used in determineFilename.",
                      verbose = verbose
    )
    dots$filename2 <- dots$postProcessedFilename
    dots$postProcessedFilename <- NULL
  }
  if (!is.null(dots$writeCropped)) {
    messagePreProcess("writeCropped is being deprecated;",
                      " use filename2, used in determineFilename.",
                      verbose = verbose
    )
    dots$filename2 <- dots$writeCropped
    dots$writeCropped <- NULL
  }
  if (!is.null(dots$rasterInterpMethod)) {
    messagePreProcess("rasterInterpMethod is being deprecated;",
                      " use method which will pass directly to projectRaster.",
                      verbose = verbose
    )
    dots$method <- dots$rasterInterpMethod
    dots$rasterInterpMethod <- NULL
  }
  if (!is.null(dots$rasterDatatype)) {
    messagePreProcess("rasterDatatype is being deprecated;",
                      " use datatype which will pass directly to writeRaster.",
                      verbose = verbose
    )
    dots$datatype <- dots$rasterDatatype
    dots$rasterDatatype <- NULL
  }
  if (!is.null(dots$pkg)) {
    messagePreProcess("pkg is being deprecated;",
                      "name the package and function directly, if needed,\n",
                      "  e.g., 'pkg::fun'.",
                      verbose = verbose
    )
    dots$pkg <- NULL
  }

  dots
}

.checkFunInDots <- function(fun = NULL, dots) {
  if (is.null(fun)) {
    if (!is.null(dots$pkg)) {
      fun <- paste0(dots$pkg, "::", fun)
    }
  }
  fun
}

hardLinkOrCopy <- function(from, to, overwrite = FALSE, verbose = TRUE) {
  linkOrCopy(from, to, symlink = FALSE, verbose = verbose)
}

escapeRegexChars <- function(str, repl = c("(", ")")) {
  for (r in repl) {
    str <- gsub(paste0("\\", r, ""), paste0("\\\\", r), str)
  }
  str
}

hardlinkOrSymlinkMessagePrefix <- "version of file(s) created"
hardlinkOrSymlinkMessagePrefixForGrep <- escapeRegexChars(hardlinkOrSymlinkMessagePrefix)

messageNoCopyMade <- "... no copy/copies made."

whPointsToMess <- "which point(s) to"
whPointsToMessForGrep <- escapeRegexChars(whPointsToMess)

getTeamDrive <- function(dots) {
  if (requireNamespace("googledrive", quietly = TRUE)) {
    teamDrive <- if (packageVersion("googledrive") < "2.0.0") {
      dots[["team_drive"]]
    } else {
      dots[["shared_drive"]]
    }
  } else {
    teamDrive <- NULL
  }
}

getTargetFilePath <- function(targetFile, archive, fileGuess, verbose,
                              destinationPath, alsoExtract, checkSumFilePath) {
  if (is.null(targetFile)) {
    if ((is.null(archive) || is.na(archive)) && !is.null(fileGuess)) {
      messagePreProcess("targetFile was not supplied; guessed and will try ", fileGuess,
                        ". If this is incorrect, please supply targetFile",
                        verbose = verbose
      )
      targetFile <- makeRelative(fileGuess, destinationPath)
      targetFilePath <- makeAbsolute(targetFile, destinationPath)
    } else {

      # Case when archive is passed, and fileGuess exists
      # if ((!is.null(archive) || !is.na(archive)) && !is.null(fileGuess)) {
      #   messagePreProcess("archieve was supplied, but targetFile not; guessed and will try ", fileGuess,
      #                     ". If this is incorrect, please supply targetFile",
      #                     verbose = verbose
      #   )
      #   targetFile <- makeRelative(fileGuess, destinationPath)
      #   targetFilePath <- makeAbsolute(targetFile, destinationPath)
      # } else {
      targetFilePath <- NULL
      # }

    }
  } else {
    if (length(targetFile) > 1) {
      stop("targetFile should be only 1 file")
    }

    targetFilePath <- normPath(makeAbsolute(targetFile, destinationPath))
  }
  targetFilePath
}

guessAlsoExtract <- function(targetFile, alsoExtract, checkSumFilePath) {
  if (is.null(alsoExtract)) {
    if (file.exists(checkSumFilePath)) {
      if (file.size(checkSumFilePath) > 0) {
        # if alsoExtract is not specified, then try to find all files in CHECKSUMS.txt with
        # same base name, without extension
        if (!is.null(targetFile)) {
          checksumsTmp <- read.table(checkSumFilePath)
          alsoExtract <- grep(paste(collapse = "|", paste0(filePathSansExt(targetFile), "\\.")),
                              checksumsTmp$file,
                              value = TRUE
          )
          rm(checksumsTmp) # clean up
        }
      }
    }
  }
  if (any(is.na(alsoExtract))) {
    alsoExtract <- NA
  } else if (!is.null(alsoExtract)) { # must keep relative because user may not know what path is in archive
    if (isTRUE(all(is.na(alsoExtract)))) {
      alsoExtract <- character()
    }
  }

  alsoExtract
}

updateArchiveWithGuess <- function(archive, guess) {
  if (!is.null(guess)) {
    if (is.null(archive)) {
      archive <- .isArchive(guess)
    }
    if (isTRUE(!is.na(archive))) {
      archive <- moveAttributes(guess, archive)
    }
  }
  archive
}

setupArchive <- function(archive, destinationPath) {
  if (!is.null(archive)) {
    if (!is.na(archive)) {
      tmpArchive <- archive
      archive <- makeAbsolute(archive, destinationPath)
      # archive <- file.path(destinationPath, basename2(archive))
      # filesToCheck <- unique(c(filesToCheck, archive))
      archive <- moveAttributes(tmpArchive, archive)
    }
  }
  archive
}

runChecksums <- function(destinationPath, checkSumFilePath, filesToCheck, verbose) {
  reproducible.inputPaths <- .getDestinationPathShared()
  if (!is.null(reproducible.inputPaths)) {
    reproducible.inputPaths <- checkPath(reproducible.inputPaths, create = TRUE)
  }
  if (!is.null(reproducible.inputPaths)) {
    reproducible.inputPaths <- path.expand(reproducible.inputPaths)
  }

  destinationPathUser <- NULL
  # When filesToCheck is empty (e.g., dlFun-only call with no url/targetFile/
  # archive), do not consult reproducible.inputPaths: there is no canonical
  # filename to look up there, and Checksums() with files = NULL would match
  # arbitrary unrelated files in the stash.
  possDirs <- if (length(filesToCheck) == 0L) {
    destinationPath
  } else {
    unique(c(destinationPath, reproducible.inputPaths))
  }
  csfps <- vapply(possDirs, function(dp) identifyCHECKSUMStxtFile(dp), character(1))
  allDone <- FALSE
  for (dp in possDirs) {
    for (csfp in csfps) { # there can be a mismatch between checksums and file location
      # csfp <- identifyCHECKSUMStxtFile(dp)
      checkSumsTmp1 <- try(Checksums(
        path = dp, write = FALSE, checksumFile = csfp,
        files = makeRelative(filesToCheck, absoluteBase = destinationPath),
        verbose = verbose - 1
      ), silent = TRUE)
      checkSums <- NULL
      if (!is(checkSumsTmp1, "try-error")) {
        checkSums <- checkSumsTmp1
        if (!all(is.na(checkSums$result))) { # found something
          if (isTRUE(any(dp %in% reproducible.inputPaths))) {
            destinationPathUser <- destinationPath
            destinationPath <- dp
            on.exit(
              {
                destinationPath <- destinationPathUser
              },
              add = TRUE
            )
          }
          allDone <- TRUE
          break
        }
      }
    }
    if (isTRUE(allDone))
      break
  }
  list(
    reproducible.inputPaths = reproducible.inputPaths,
    destinationPathUser = destinationPathUser,
    destinationPath = destinationPath,
    checkSums = checkSums
  )
}

dealWithArchive <- function(archive, url, targetFile, checkSums, alsoExtract, destinationPath, teamDrive, verbose) {
  fileGuess <- NULL
  if (is.null(archive)) {
    if (!is.null(url)) {
      allOK <- .similarFilesInCheckSums(targetFile, checkSums, alsoExtract)

      if (!allOK) { # skip identification of archive if we have all files with same basename as targetFile
        # BUT if we don't have all files with identical root name (basename sans ext), then assess for
        #   an archive, either remotely, in the case of google or from the basename of url
        fileGuess <- .guessAtFile(
          url = url, archive = archive,
          targetFile = targetFile, destinationPath = destinationPath,
          verbose = verbose, team_drive = teamDrive
        )
        archive <- .isArchive(fileGuess)
        # The fileGuess MAY have a fileSize attribute, which can be attached to "archive"
        archive <- moveAttributes(fileGuess, receiving = archive)
        sourceAttributes <- attributes(fileGuess)
        if (length(sourceAttributes) > 0 && !is.null(archive)) {
          for (i in length(sourceAttributes)) {
            attr(archive, names(sourceAttributes)[i]) <- sourceAttributes[[i]]
          }
        }

        checkSums <- .checkSumsUpdate(
          destinationPath = destinationPath,
          newFilesToCheck = archive,
          checkSums = checkSums,
          verbose = verbose
        )
      }
    }
  }
  list(
    checkSums = checkSums,
    archive = archive,
    fileGuess = fileGuess
  )
}

#' Find any remote-hash sidecar files for a given basename across one or more
#' directories. Returns absolute paths of non-empty sidecars.
#' @keywords internal
#' @noRd
.findRemoteHashSidecars <- function(filename, dirs) {
  if (is.null(filename) || length(filename) != 1L ||
      is.na(filename) || !nzchar(filename)) return(character())
  # Sidecars are written as `.<basename>_<urlencoded>.hash` (hidden); use
  # all.files = TRUE so the dotfile form is discoverable.
  prefix <- paste0(".", filename, "_")
  out <- character()
  for (d in dirs) {
    if (!is.null(d) && nzchar(d) && dir.exists(d)) {
      all <- list.files(d, full.names = FALSE, all.files = TRUE)
      isSidecar <- startsWith(all, prefix) & endsWith(all, ".hash")
      sidecars <- file.path(d, all[isSidecar])
      if (length(sidecars))
        sidecars <- sidecars[file.size(sidecars) > 0L]
      out <- c(out, sidecars)
    }
  }
  out
}

isNULLorNA <- function(x) {
  out <- TRUE
  if (!is.null(x)) {
    if (!isTRUE(is.na(x))) {
      out <- FALSE
    }
  }
  out
}

identifyCHECKSUMStxtFile <- function(path) {
  file.path(path, "CHECKSUMS.txt")
}

linkOrCopyUpdateOnly <- function(from, to, verbose) {
  fifrom <- file.info(from)
  fito <- file.info(to)
  whNotSame <- fifrom$ctime != fito$ctime
  needCopy <- !whNotSame %in% FALSE
  if (any(needCopy)) {
    linkOrCopy(from[needCopy], to[needCopy],
               verbose = verbose - 1
    )
  }
}

messageChecksummingAllFiles <- "Checksumming all files in archive"



# may already have been changed above
copyFromDPtoReproducibleIPs <- function(targetFilePath, destinationPathUser, destinationPath,
                                        reproducible.inputPaths, neededFiles, archive, checkSums,
                                        checkSumFilePath, verbose) {
  outCheck <- if (!is.null(targetFilePath)) {
    !file.exists(targetFilePath)
  } else {
    TRUE
  } ## if NULL, it doesn't exist and we want to proceed
  if (isTRUE(outCheck)) { # skip if it already existed locally
    if (is.null(destinationPathUser)) {
      destinationPathUser <- destinationPath
    }
    on.exit(
      {
        destinationPath <- destinationPathUser
      },
      add = TRUE
    )

    if (!identical(destinationPath, reproducible.inputPaths)) {
      # CHANGE destinationPath FOR REMAINDER OF THIS FUNCTION
      neededFilesNew <- makeRelative(neededFiles, destinationPath)
      targetFilePathNew <- makeRelative(targetFilePath, destinationPath)
      destinationPathNew <- reproducible.inputPaths[1]
      archiveExistInDestDir <- if (!isNULLorNA(archive)) {
        file.exists(archive)
      } else {
        FALSE
      }
      existInDestDir <- if (!isNULLorNA(neededFiles)) {
        file.exists(neededFiles)
      } else {
        FALSE
      }
      filesToAppendToChecksums <- character()
      if (any(existInDestDir)) {
        from <- neededFiles[existInDestDir]
        to <- makeAbsolute(neededFilesNew[existInDestDir],
                           absoluteBase = destinationPathNew
        )
        linkOrCopyUpdateOnly(from, to, verbose = verbose)
        filesToAppendToChecksums <- c(filesToAppendToChecksums, to)
      }
      if (any(archiveExistInDestDir)) {
        from <- archive[archiveExistInDestDir]
        to <- makeAbsolute(makeRelative(archive[archiveExistInDestDir], destinationPath),
                           absoluteBase = destinationPathNew
        )
        linkOrCopyUpdateOnly(from, to, verbose = verbose)
        filesToAppendToChecksums <- c(filesToAppendToChecksums, to)
      }
      if (any(nzchar(filesToAppendToChecksums) %in% TRUE)) {
        appendChecksumsTableWithCS(append = TRUE,
                                   checkSumFilePath = file.path(destinationPathNew, basename(checkSumFilePath)),
                                   destinationPathNew,
                                   filesToChecksum = filesToAppendToChecksums, currentFiles = checkSums,
                                   verbose)

      }
      targetPath <- targetFilePathNew
      destinationPath <- destinationPathNew
      neededFiles <- neededFilesNew
    }

    if (isTRUE(any(grepl(archive, pattern = destinationPathUser)))) {
      # might have a "." as destinationPath -- messes with grepl
      patt <- if (grepl("^\\.", destinationPathUser)) {
        gsub("^\\.", "^\\\\.", destinationPathUser)
      } else {
        destinationPathUser
      }
      archive <- gsub(archive, pattern = patt, replacement = destinationPath)
    }
    targetFilePath <- makeAbsolute(targetFilePath, destinationPath)
    neededFiles <- makeAbsolute(neededFiles, destinationPath)
  }
  list(neededFiles = neededFiles, targetFilePath = targetFilePath,
       destinationPathUser = destinationPathUser, destinationPath = destinationPath,
       archive = archive)
}

utils::globalVariables(c(
  "..cnNotPaths", "..cnPaths"
))

################################################################################
#' Create reproducible digests of objects in R
#'
#' Not all aspects of R objects are captured by current hashing tools in R
#' (e.g. `digest::digest`, `knitr` caching, `archivist::cache`).
#' This is mostly because many objects have "transient"
#' (e.g., functions have environments), or "disk-backed" features.
#' Since the goal of using reproducibility is to have tools that are not session specific,
#' this function attempts to strip all session specific information so that the digest
#' works between sessions and operating systems.
#' It is tested under many conditions and object types, there are bound to be others that don't
#' work correctly.
#'
#' @section Classes:
#'
#' `Raster*` objects have the potential for disk-backed storage, thus, require more work.
#' Also, because `Raster*` can have a built-in representation for having their data content
#' located on disk, this format will be maintained if the raster already is file-backed,
#' i.e., to create `.tif` or `.grd` backed rasters, use `writeRaster` first,
#' then `Cache`.
#' The \file{.tif} or \file{.grd} will be copied to the \file{raster/} subdirectory of the
#' `cachePath`.
#' Their RAM representation (as an R object) will still be in the usual  \file{cacheOutputs/}
#' (or formerly \file{gallery/}) directory.
#' For `inMemory` raster objects, they will remain as binary `.RData` files.
#'
#' Functions (which are contained within environments) are
#' converted to a text representation via a call to `format(FUN)`.
#'
#' Objects contained within a list or environment are recursively hashed
#' using [digest::digest()], while removing all references to
#' environments.
#'
#' Character strings are first assessed with `dir.exists` and `file.exists`
#' to check for paths. If they are found to be paths, then the path is hashed with
#' only its filename via `basename(filename)`. If it is actually a path, we suggest
#' using `asPath(thePath)`
#'
#' @param object an object to digest.
#'
#' @param objects Optional character vector indicating which objects are to
#'                be considered while making digestible. This argument is not used
#'                in the default cases; the only known method that uses this
#'                in the default cases; the only known method that uses this
#'                argument is the `simList` class from `SpaDES.core`.
#'
#' @inheritParams Cache
#'
#' @return A hash i.e., digest of the object passed in.
#'
#' @author Eliot McIntire
#' @export
#' @importFrom digest digest
#' @keywords internal
#' @rdname robustDigest
#' @examples
#'
#' a <- 2
#' tmpfile1 <- tempfile()
#' tmpfile2 <- tempfile()
#' tmpfile3 <- tempfile(fileext = ".grd")
#' tmpfile4 <- tempfile(fileext = ".grd")
#' save(a, file = tmpfile1)
#' save(a, file = tmpfile2)
#'
#' # treats as character string, so 2 filenames are different
#' digest::digest(tmpfile1)
#' digest::digest(tmpfile2)
#'
#' # tests to see whether character string is representing a file
#' .robustDigest(tmpfile1)
#' .robustDigest(tmpfile2) # same
#'
#' # if you tell it that it is a path, then you can decide if you want it to be
#' #  treated as a character string or as a file path
#' .robustDigest(asPath(tmpfile1), quick = TRUE)
#' .robustDigest(asPath(tmpfile2), quick = TRUE) # different because using file info
#'
#' .robustDigest(asPath(tmpfile1), quick = FALSE)
#' .robustDigest(asPath(tmpfile2), quick = FALSE) # same because using file content
#'
#' # SpatRasters are have pointers
#' if (requireNamespace("terra", quietly = TRUE)) {
#'   r <- terra::rast(system.file("ex/elev.tif", package = "terra"))
#'   r3 <- terra::deepcopy(r)
#'   r1 <- terra::writeRaster(r, filename = tmpfile3)
#'
#'   digest::digest(r)
#'   digest::digest(r3) # different but should be same
#'   .robustDigest(r1)
#'   .robustDigest(r3) # same... data & metadata are the same
#'
#'   # note, this is not true for comparing memory and file-backed rasters
#'   .robustDigest(r)
#'   .robustDigest(r1) # different
#' }
#'
setGeneric(".robustDigest", function(object, .objects = NULL,
                                     length = getOption("reproducible.length", Inf),
                                     algo = "xxhash64",
                                     quick = getOption("reproducible.quick", FALSE),
                                     classOptions = list(), ...) {
  standardGeneric(".robustDigest")
})

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "ANY",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    # browser(expr = exists("._robustDigest_1"))
    if (inherits(object, "quosure")) { # can't get this class from rlang via importClass rlang quosure
      if (!requireNamespace("rlang")) stop("Please `install.packages('rlang')`")
      object <- rlang::eval_tidy(object)
    }

    if (inherits(object, "Spatial")) {
      object <- .removeCacheAtts(object)
      if (inherits(object, "SpatialPoints")) {
        forDig <- as.data.frame(object)
      } else {
        forDig <- object
      }
      # The following Rounding is necessary to make digest equal on linux and windows
      if (inherits(forDig, "SpatialPolygonsDataFrame")) {
        bbb <- unlist(lapply(as.data.frame(forDig), is.numeric))
        if (sum(bbb)) {
          bbbWh <- which(bbb)
          for (i in bbbWh) { # changed because may not have correct names, can be NAs, Eliot March 2019
            #  Error was: Error in round(in[[i]], 4) :
            # non-numeric argument to mathematical function
            forDig[[i]] <- round(forDig[[i]], 4)
          }
        }
      }
    } else if (inherits(object, "Raster")) {
      object <- .removeCacheAtts(object)

      dig <- suppressWarnings(
        .digestRasterLayer(object, length = length, algo = algo, quick = quick)
      )
      forDig <- unlist(dig)
    } else if (inherits(object, "cluster")) { # can't get this class from parallel via importClass parallel cluster
      forDig <- NULL
    } else if (.isSpatRaster(object)) {
      if (!requireNamespace("terra", quietly = TRUE)) {
        stop("Please install terra package")
      }
      terraSrcs <- Filenames(object)
      if (any(nchar(terraSrcs) > 0)) {
        out <- lapply(terraSrcs, function(x) {
          if (grepl("^NETCDF:", x)) {
            x <- sub("^NETCDF:\"", "", x)
            x <- sub("\":.*$", "", x)
          }
          .robustDigest(object = asPath(x), length = length, algo = algo, quick = quick)
        })

        dig <- .robustDigest(
          list(
            terra::nrow(object), terra::ncol(object), terra::nlyr(object),
            terra::res(object), terra::crs(object),
            as.vector(terra::ext(object)), # There is something weird with this pointer that doesn't cache consistently
            names(object)
          ),
          length = length, quick = quick,
          algo = algo, classOptions = classOptions
        ) # don't include object@data -- these are volatile
        forDig <- list(out, dig)
      } else {
        warns <- withCallingHandlers(forDig <- terra::wrap(object),
                                     warning = function(w) {
                                       if (isTRUE(any(grepl("raster has no values", w[[1]]))))
                                         invokeRestart("muffleWarning")
                                     })
      }
    } else if (inherits(object, "SpatVector")) {
      if (!requireNamespace("terra", quietly = TRUE)) {
        stop("Please install terra package")
      }
      forDig <- wrapSpatVector(object)
    } else if (inherits(object, "SpatExtent")) {
      forDig <- .wrap(object)
    } else if (inherits(object, "drive_id")) {
      if (.requireNamespace("googledrive")) {
        forDig <- try(googledrive::drive_get(object))
        if (inherits(forDig, "try-error")) {
          message("Detected that object is a googledrive id; can't access it online; ",
                  "evaluating only the url as character string")
          forDig <- object
        }
      }
    } else {
      forDig <- .removeCacheAtts(object)
    }
    out <- .doDigest(forDig, algo)
    return(out)
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "function",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    .robustDigestFormatOnly(object, algo = algo)
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "expression",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    .robustDigestFormatOnly(object, algo = algo)
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "language",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    .robustDigestFormatOnly(object, algo = algo)
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "call",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    .robustDigestFormatOnly(object, algo = algo)
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "character",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    object <- .removeCacheAtts(object)

    simpleDigest <- TRUE # default -- only as character vector
    if (inherits(object, "fs_path")) {
      return(.robustDigest(asPath(object)))
    }
    if (getOption("reproducible.testCharacterAsFile", FALSE)) {
      if (!quick) {
        # If a character string has nonASCII characters e.g., from french
        #  "Cordill\xe8re arctique", file.exists fails with "file name conversion problem" error
        howMany <- min(10, NROW(object))
        whCheck <- object[1:howMany]
        asc <- iconv(whCheck, "latin1", "ASCII")
        if (anyNA(asc)) {
          whCheck <- whCheck[!is.na(asc)]
        }

        if (any(unlist(lapply(whCheck, file.exists)))) { # only try first 10 elements
          simpleDigest <- FALSE
        }
      }
    }
    if (!simpleDigest) { # Old behaviour -- digest as file
      # object <- asPath(object)
      unlist(lapply(object, function(x) {
        if (dir.exists(x)) {
          .doDigest(basename(x), algo)
        } else if (file.exists(x)) {
          digest(file = x, length = length, algo = algo)
        } else {
          .doDigest(x, algo)
        }
      }))
    } else {
      .doDigest(object, algo = algo)
    }
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "Path",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    nParentDirs <- attr(object, "nParentDirs")
    object <- .removeCacheAtts(object)

    if (!quick) {
      lapply(object, function(x) {
        isExistentFile <- FALSE
        if (file.exists(x)) {
          if (!dir.exists(x)) {
            isExistentFile <- TRUE
          }
        }
        if (isExistentFile) {
          digest::digest(file = x, length = length, algo = algo)
        } else {
          # just do file basename as a character string, if file does not exist
          .doDigest(basenames3(x, nParentDirs), algo = algo)
        }
      })
    } else {
      .doDigest(basenames3(object, nParentDirs), algo = algo)
    }
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "environment",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    object <- .removeCacheAtts(object)
    if (is.null(classOptions[["prevEnvir"]])) {
      classOptions[["prevEnvir"]] <- list()
      doneAlready <- list(FALSE)
    } else {
      doneAlready <- lapply(classOptions[["prevEnvir"]], function(pe) identical(pe, object))
    }
    classOptions[["prevEnvir"]] <- unique(append(classOptions[["prevEnvir"]], object))

    if (!any(unlist(doneAlready))) {
      asList <- as.list(object, all.names = TRUE)
      da <- which(unlist(doneAlready))
      if (length(da)) {
        asList <- asList[-da]
      }
      rd <- .robustDigest(asList,
                          .objects = .objects,
                          length = length,
                          algo = algo, quick = quick, classOptions = classOptions
      )
    } else {
      rd <- NULL
    }
    return(rd)
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "list",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    object <- .removeCacheAtts(object)
    object <- rmDotObjects(object, .objects)
    .objects <- dotObjectsToNULL(object, .objects) # only use it once

    # if (!is.null(.objects)) {
    #   # This will get "only the top=level" list ... if it matches
    #   correctList <- intersect(.objects, names(object))
    #   if (length(correctList) > 0) {
    #     object <- object[.objects]
    #     .objects <<- NULL
    #   }
    # }

    objsSorted <- .sortDotsUnderscoreFirst(object)
    # addr <- list()
    # seen <- new.env(parent = emptyenv())
    # localAddrs <- Map(o = objsSorted, function(o) lobstr::obj_addr(o))

    inner <- Map(x = objsSorted, i = seq_along(objsSorted), # addr = localAddrs,
                 function(x, i) {#, addr) {

                   if (!is.null(attr(x, ".Cache")$newCache)) {
                     x <- .setSubAttrInList(x, ".Cache", "newCache", NULL)
                     if (!identical(attr(x, ".Cache")$newCache, NULL)) stop("attributes are not correct 1")
                   }

                   # if (exists(addr, envir = seen)) {
                   #   return(seen[[addr]])
                   # }

                   withCallingHandlers({

                     result <- .robustDigest(
                       object = x, .objects = .objects,
                       length = length,
                       algo = algo, quick = quick, classOptions = classOptions
                     )
                   }, error = function(e) {
                     nam <- names(objsSorted)
                     if (!is.null(nam)) {
                       messageCache("Error occurred during .robustDigest of ", nam[i])
                     }
                   })
                   # seen[[addr]] <- result

                   result

    })
    ## have to distinguish a list from an object not in a list
    # append(list(._list = .doDigest(inner)), inner)
    inner
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "data.frame",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    #  Need a specific method for data.frame or else it get "list" method, which is wrong
    object <- .removeCacheAtts(object)
    dig <- lapply(object, .robustDigest, algo = algo, quick = quick, classOptions = classOptions)

    # # This is forcing `quick` TRUE because if a column is a character of file names,
    # #   it will evaluate every file; a user should specify with asPath()
    # Paths <- vapply(object, is, "Path", FUN.VALUE = logical(1))
    # dig <- character(length(object))
    # whNotPaths <- which(!Paths)
    # whPaths <- which(Paths)
    # # Can't use whNotPaths because data.table wouyld need ..whNotPaths ... so calculate it in the data.table
    # cn <- colnames(object)
    # cnNotPaths <- cn[whNotPaths]
    # cnPaths <- cn[whPaths]
    # if (any(!Paths)) {
    #   if (is.data.table(object))
    #     dig[whNotPaths] <- lapply(object[, ..cnNotPaths], .robustDigest, algo = algo, quick = TRUE, classOptions = classOptions)
    #   else
    #     dig[whNotPaths] <- lapply(object[, cnNotPaths], .robustDigest, algo = algo, quick = TRUE, classOptions = classOptions)
    # }
    # if (any(Paths)) {
    #   if (is.data.table(object))
    #     dig[whPaths] <- lapply(object[, ..cnPaths], .robustDigest, algo = algo, quick = TRUE, classOptions = classOptions)
    #   else
    #     dig[whPaths] <- lapply(object[, cnPaths], .robustDigest, algo = algo, quick = TRUE, classOptions = classOptions)
    # }
    .robustDigest(unlist(dig), quick = TRUE, algo = algo, classOptions = classOptions)
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "numeric",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    #  Need a specific method for data.frame or else it get "list" method, which is wrong
    object <- .removeCacheAtts(object)
    # From ad hoc tests, 6 was the highest I could go to maintain consistent between Linux and Windows
    .doDigest(round(object, getOption("reproducible.digestDigits", 7)), algo = algo)
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "matrix",
  definition = function(object, .objects, length, algo, quick, classOptions) {
    #  Need a specific method for data.frame or else it get "list" method, which is wrong
    object <- .removeCacheAtts(object)
    dims <- dim(object)
    dim(object) <- NULL # need to get the separate numeric or integer, i.e., there is rounding
    out <- .robustDigest(list(dims, object), classOptions = classOptions, algo = algo)
    .doDigest(out, algo = algo)
    # From ad hoc tests, 6 was the highest I could go to maintain consistent between Linux and Windows
  }
)

#' @rdname robustDigest
#' @export
setMethod(
  ".robustDigest",
  signature = "integer",
  definition = function(object, .objects, length, algo, quick, classOptions,
                        cacheSaveFormat = getOption("reproducible.cacheSaveFormat")) {
    #  Need a specific method for data.frame or else it get "list" method, which is wrong
    object <- .removeCacheAtts(object)
    if (identical(cacheSaveFormat, .qs2Format) &&
        identical(getOption("reproducible.cacheSpeed"), "fast")) {
      os <- objSize(object)
      if (os == 680) {
        # Means it is ALTREP --> convert to non-ALTREP for qs only
        if (!is.factor(object))
          object <- as.integer(object + 0.0)
      }
      # qs doesn't save ALTREP yet for numerics

    }
    # From ad hoc tests, 7 was the highest I could go to maintain consistent between Linux and Windows
    .doDigest(object, algo = algo)
  }
)


basenames3 <- function(object, nParentDirs) {
  if (missing(nParentDirs)) {
    nParentDirs <- 0
  }
  object <- if (nParentDirs > 0) {
    obj <- object
    objOut <- basename(obj)
    for (i in seq_len(nParentDirs)) {
      obj <- dirname(obj)
      objOut <- file.path(basename(obj), objOut)
    }
    objOut
  } else {
    basename(object)
  }
  object
}

#' Remove attributes that are highly varying
#'
#' @param x Any arbitrary R object that could have attributes
.removeCacheAtts <- function(x) {
  if (!is.null(attr(x, "tags"))) {
    attr(x, "tags") <- NULL
  }
  if (!is.null(attr(x, ".Cache"))) {
    attr(x, ".Cache") <- NULL
  }
  if (!is.null(attr(x, callInCache))) {
    attr(x, callInCache) <- NULL
  }
  if (!is.null(attr(x, cacheChainingOuterFunctionName))) {
    attr(x, cacheChainingOuterFunctionName) <- NULL
  }

  x
}

.robustDigestFormatOnly <- function(object, .objects, length, algo, quick,
                                    classOptions) {
  object <- .removeCacheAtts(object)
  .doDigest(format(object), algo = algo)
}

.doDigest <- function(x, algo, length = Inf, file,
                      newAlgo = NULL,
                      cacheSpeed = getOption("reproducible.cacheSpeed", "slow"), ...) {
  # the ... is just a passthrough so this function doesn't fail if there are other args
  if (missing(algo)) algo <- formals(.robustDigest)$algo

  out <- if (!missing(file)) {
    digest::digest(file = x, algo = algo, length = length)
  } else {
    if (cacheSpeed == "fast") {
      cacheSpeed <- 2L
    } else if (cacheSpeed == "slow") {
      cacheSpeed <- 1L
    }
    if (cacheSpeed == 1) {
      if (length(x) == 1) {
        if (is.atomic(x))
          if (isTRUE(is.na(x)))
            x <- NA # make all NAs (NA_real_, NA, NA_character_ equal
      }
    } else if (cacheSpeed == 2) {
      algo <- "spooky"
    } else {
      stop("options('reproducible.cacheSpeed') must be 1, 2, 'slow' or 'fast'")
    }
    out <- digest::digest(x, algo = algo)
  }
  out
}

rmDotObjects <- function(object, .objects) {
  if (!is.null(.objects)) {
    # This will get "only the top=level" list ... if it matches
    correctList <- intersect(.objects, names(object))
    if (length(correctList) > 0) {
      object <- object[.objects]
      attr(object, ".objects") <- .returnNothing
    }
  }
  object
}


dotObjectsToNULL <- function(object, .objects) {
  if (identical(attr(object, ".objects"), .returnNothing))
    .objects <- NULL # only use it once
  .objects
}



# CacheAddressEnv <- function(envir = .GlobalEnv, create = FALSE, remove = FALSE) {
#   browser()
#   env <- NULL
#   cae <- "CacheAddressEnv"
#   if (isTRUE(remove)) {
#     rm(list = cae, envir = .pkgEnv)
#   } else {
#     if (exists(cae, envir = .pkgEnv)) {
#       env <- get(cae, envir = .pkgEnv, inherits = FALSE)
#     }
#     if (create %in% TRUE) {
#       obj <- paste0(".reproducibleCacheAddressEnv")
#       if (!exists(obj, envir = envir))
#         assign(obj, new.env(parent = emptyenv()), envir = envir)
#
#       if (is.null(.pkgEnv[[cae]]))
#         assign(cae, new.env(parent = emptyenv()), envir = .pkgEnv)
#
#       assign(obj, envir[[obj]], envir = .pkgEnv[[cae]])# <- env
#     }
#
#     if (exists(obj, envir = envir) && is.null(env)) {
#       env <- get(obj, envir = envir, inherits = FALSE)
#     }
#
#   }
#   env
# }
#
# reproducible.CacheAddressEnv <- "reproducible.CacheAddressEnv"
#
# memoiseEnv <- function(cachePath, envir = .GlobalEnv) {
#   memPersist <- isTRUE(getOption("reproducible.memoisePersist", NULL))
#   if (memPersist) {
#     obj <- paste0(".reproducibleMemoise_", cachePath)
#     if (!exists(obj, envir = envir))
#       assign(obj, new.env(parent = emptyenv()), envir = envir)
#     memEnv <- get(obj, envir = envir, inherits = FALSE)
#   } else {
#     if (is.null(.pkgEnv[[cachePath]])) {
#       .pkgEnv[[cachePath]] <- new.env(parent = emptyenv())
#     }
#     memEnv <- .pkgEnv[[cachePath]]
#   }
#   memEnv
# }

## non-exported attribute stuff -------------------------------------------------
attributesReassign <- function(atts, obj) {
  attsNames <- setdiff(names(atts), knownAtts)
  if (length(attsNames))
    for (att in attsNames) {
      if (is.null(attr(obj, att))) {
        attr(obj, att) <- atts[[att]]
      }
    }
  obj
}

knownAtts <- c("cpp", "class", "attributes", "values", "definition", "pnt", "ptr", "pntr")

## non-exported tag stuff -------------------------------------------------------

parseTags <- function(tags) {
  out <- strsplit(tags, ":")
  tags2 <- lapply(out, function(x) x[[1]])
  tags1 <- as.data.table(do.call(rbind, tags2))
  vals <- vapply(seq_len(NROW(tags1)), function(i) {
    gsub(paste0(tags1[[1]][i], ":"), "", tags[i])
  }, FUN.VALUE = character(1))
  set(tags1, NULL, "tagValue", vals)
  setnames(tags1, "V1", "tagKey")
  tags1
}

tagFilesToLoad <- "filesToLoad"
tagRelToWhere <- "relToWhere"
tagOrigFilename <- "origFilename"
tagOrigRelName <- "origRelName"
tagFilenamesInCache <- "filenamesInCache"

tagsSpatRaster <- function(obj = NULL, relToWhere = NULL, relName = NULL, cls = NULL,
                           whLayers = NULL, layerNams = NULL, obj2 = NULL,
                           filenamesInCache = NULL, cacheId = NULL) {
  fe <- if (is.null(obj)) NULL else tools::file_ext(obj)
  if (is.null(cacheId)) cacheId <- NULL
  c(
    attr(obj, "tags"),
    paste0(tagRelToWhere, ":", names(relToWhere)),
    paste0(tagOrigFilename, ":", basename2(obj)),
    paste0(tagOrigRelName, ":", relName),
    paste0("fromDisk:", TRUE),
    paste0("class:", cls),
    paste0("fileFormat:", fe),
    paste0("saveRawFile:", TRUE),
    paste0("loadFun:", "terra::rast"),
    paste0("whLayers:", whLayers),
    paste0("layerNames:", layerNams),
    paste0(tagFilesToLoad, ":", basename2(obj2)),
    paste0(tagFilenamesInCache, ":", filenamesInCache),
    paste0("cacheId:", cacheId)
  )
}

## non-exported path functions --------------------------------------------------

absoluteBase <- function(relToWhere, cachePath = getOption("reproducible.cachePath"), ...) {
  if (identical(relToWhere, "cachePath") && !is.null(cachePath)) {
    ab <- cachePath
  } else if (identical(relToWhere, "getwd")) {
    ab <- getwd()
  } else {
    possRelPaths <- modifyListPaths(cachePath, ...)
    if (relToWhere %in% names(possRelPaths)) {
      ab <- try(possRelPaths[[relToWhere]])
    } else {
      ab <- try(possRelPaths[[1]])
    }
    if (is(ab, "try-error")) browser()
  }

  ab
}

#' @importFrom utils modifyList
modifyListPaths <- function(cachePath = getOption("reproducible.cachePath"), ...) {
  possRelPaths <- list()
  if (!missing(cachePath))
    possRelPaths$cachePath <- cachePath
  dots <- list(...)
  if (length(dots)) {
    if (is(dots[[1]], "list")) {
      if (is.null(names(dots[[1]])))
        stop("wrapSpatRaster and unwrapSpatRaster require named list passed to dots")
      possRelPaths <- modifyList(dots[[1]], possRelPaths)
    }
  }
  possRelPaths <- append(possRelPaths, list(getwd = getwd()))
}

relativeToWhat <- function(file, cachePath = getOption("reproducible.cachePath"), ...) {
  possRelPaths <- modifyListPaths(cachePath, ...)

  foundAbs <- FALSE
  dirnameFile <- dirname(file)
  whSame <- rep(FALSE, length(file))
  pc <- rep("", length(file))
  for (nams in names(possRelPaths)) {
    pc[!whSame] <- mapply(fn = file[!whSame], function(fn)
      fs::path_common(c(dirname(fn), possRelPaths[[nams]]))
    )

    out <- sapply(possRelPaths[[nams]], fs::path_rel, path = pc) |> as.character()
    whSame <- pc == dirnameFile
    if (all(whSame)) {
      out <- list(out)
      names(out) <- nams
      foundAbs <- TRUE
      break
    }
  }
  if (isFALSE(foundAbs)) {
    for (nams in names(possRelPaths)) {
      out <- dirname(file)
      names(out) <- ""
      if (FALSE) { # this is for rebuilding relative against
        # poss <- fs::path_common(c(file, possRelPaths[nams]))
        # if (!identical(poss, possRelPaths[nams])) {
        #   fileRel <- makeRelative(file, poss)
        #   rel <- makeRelative(possRelPaths[nams], poss)
        #   relWithDots <- rep("..", length(strsplit(rel, "/|\\\\")[[1]]))
        #   poss <- file.path(paste(relWithDots, collapse = "/"), dirname(fileRel))
        #   out <- poss
        #   out <- list(out)
        #   names(out) <- nams
        #   foundAbs <- TRUE
        #   break
        # }
        # names(out) <- nams
      }
    }
  }

  out
}

## non-exported wrap functions --------------------------------------------------

wrapSpatRaster <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL, ...) {
  fns <- Filenames(obj, allowMultiple = FALSE)

  cls <- class(obj)
  fnsMulti <- Filenames(obj, allowMultiple = TRUE)
  obj2 <- asPath(Filenames(obj, allowMultiple = FALSE))
  nlyrsInFile <- as.integer(terra::nlyr(terra::rast(fns)))
  layerNams <- paste(names(obj), collapse = layerNamesDelimiter)

  ## A file-backed raster can:
  ## 1) not be using all the layers in the file; and
  ## 2) have layer names renamed.
  whLayers <- seq_along(names(obj))
  if (!identical(nlyrsInFile, length(names(obj)))) {
    rr <- terra::rast(fns);
    objDigs <- unlist(lapply(layerNams, function(ln) .robustDigest(obj[[ln]][])))
    digs <- character()
    whLayers <- integer()

    # don't need to go through all layers if the current file has only some; run through from start
    for (ln in seq_len(terra::nlyr(rr))) {
      digs[ln] <- .robustDigest(rr[[ln]][])
      if (digs[ln] %in% objDigs)
        whLayers <- c(ln, whLayers)
      if (all(digs %in% objDigs))
        break
    }
  }
  obj <- asPath(fnsMulti)

  relToWhere <- relativeToWhat(obj, cachePath, ...)
  # if ("" %in% names(relToWhere)) {
  #   # absBase <- browser()
  # }
  # absBase <- absoluteBase(names(relToWhere), cachePath, ...)
  relPath <- unname(unlist(relToWhere))
  relName <- file.path(relPath, basename2(obj))

  # Change the filename of the file-backing to include the cachdId as a prefix
  filenameInCache <- filenameInCacheWPrefix(obj, cacheId)
  # filenameInCache <- if (is.null(cacheId)) obj else .prefix(obj, prefixCacheId(cacheId))
  # filenameInCache <- basename2(filenameInCache)
  # if (!identical(filenameInCache, filenameInCache2)) browser()

  tags <- tagsSpatRaster(obj, relToWhere, relName, cls, whLayers, layerNams, obj2, filenameInCache,
                         cacheId)
  attr(obj, "tags") <- tags

  # c(
  #   attr(obj, "tags"),
  #   paste0(tagRelToWhere, ":", names(relToWhere)),
  #   paste0(tagOrigFilename, ":", basename2(obj)),
  #   paste0(tagOrigRelName, ":", relName),
  #   # paste0("origDirname:", dirname(obj)),
  #   paste0("fromDisk:", TRUE),
  #   paste0("class:", cls),
  #   paste0("fileFormat:", tools::file_ext(obj)),
  #   paste0("saveRawFile:", TRUE),
  #   paste0("loadFun:", "terra::rast"),
  #   paste0("whLayers:", whLayers),
  #   paste0("layerNames:", layerNams),
  #   paste0(tagFilesToLoad, ":", basename2(obj2))
  # )
  obj
}

unwrapSpatRaster <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL, ...) {
  fns <- Filenames(obj)
  if (isTRUE(any(nchar(fns) > 0))) {
    tags <- attr(obj, "tags")
    if (!is.null(tags)) {
      if (!is.null(cachePath)) {
        filenameInCache <- CacheStoredFile(cachePath,
                                           # cacheId = tools::file_path_sans_ext(basename(obj)),
                                           obj = obj, readOnly = TRUE
        )
        filenameInCache <- filenameInCacheWPrefix(filenameInCache, cacheId = cacheId, relative = FALSE)
        # filenameInCache <- .prefix(filenameInCache, prefixCacheId(cacheId))
        # if (!identical(filenameInCache, filenameInCache2)) browser()
        feObjs <- file.exists(obj)
        if (any(feObjs))
          unlink(obj[feObjs])
        # fnToLoad <- fns
        newFiles <- remapFilenames(fns, tags, cachePath, ...)
        fromFiles <- unlist(filenameInCache)
      } else {
        newFiles <- remapFilenames(tags = tags, cachePath = cachePath, ...)
        fromFiles <- unlist(fns) # fnToLoad <- newFiles$newName
      }
      hardLinkOrCopy(fromFiles, newFiles$newName, verbose = 0)
      # tags <- parseTags(tags)
      # origRelName <- extractFromCache(tags, tagOrigRelName)
      # origFilename <- extractFromCache(tags, tagOrigFilename) # tv[tk == tagOrigFilename]
      # relToWhere <- extractFromCache(tags, "relToWhere")
      # # possPaths <- modifyListPaths(cachePath, ...)
      # absBase <- absoluteBase(relToWhere, cachePath, ...)
      # newName <- file.path(absBase, origRelName)

      # if (FALSE) {
      #   isAbs <- isAbsolutePath(origRelName)
      #   if (any(isAbs) || is.null(cachePath)) { # means that it had a specific path, not just relative
      #     newName2 <- file.path(normPath(extractFromCache(tags, "origDirname")), origFilename)
      #   } else {
      #     newName2 <- file.path(cachePath, origRelName)
      #   }
      # }

      # if (!identical(newName, newName2)) browser()

      # whFiles <- newFiles$newName[match(basename(extractFromCache(tags, tagFilesToLoad)), origFilename)]

      # if (!is.null(cachePath)) {
      #   filenameInCache <- CacheStoredFile(cachePath,
      #                                      # cacheId = tools::file_path_sans_ext(basename(obj)),
      #                                      obj = obj
      #   )
      #   feObjs <- file.exists(obj)
      #   if (any(feObjs))
      #     unlink(obj[feObjs])
      #   fnToLoad <- fns
      #   hardLinkOrCopy(unlist(filenameInCache), fns, verbose = 0)
      # } else {
      #   fnToLoad <- newFiles$newName
      #   hardLinkOrCopy(unlist(fns), newFiles$newName, verbose = 0)
      # }

      obj <- eval(parse(text = extractFromCache(newFiles$tagsParsed, "loadFun")))(newFiles$whFiles)
      possNames <- strsplit(extractFromCache(newFiles$tagsParsed, "layerNames"), split = layerNamesDelimiter)[[1]]
      namsObjs <- names(obj)
      if (!identical(possNames, namsObjs)) {
        whLayers <- as.integer(extractFromCache(newFiles$tagsParsed, "whLayers"))
        if (length(whLayers) != length(namsObjs)) {
          obj <- obj[[whLayers]]
        }
      }

      # names can be wrong e.g., with "nextNumericName" ... habitatQuality_1 instead of habitatQuality.
      #  Should use the one without the `nextNumericName`
      names(obj) <- possNames
    }
  }
  obj
}

unwrapRaster <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId) {
  origFilenames <- if (is(obj, "Raster")) {
    Filenames(obj) # This is legacy piece which allows backwards compatible
  } else {
    obj$origRaster
  }

  filesExist <- file.exists(origFilenames)
  cacheFilenames <- Filenames(obj)
  filesExistInCache <- file.exists(cacheFilenames)
  if (any(!filesExistInCache)) {
    fileTails <- gsub("^.+(rasters.+)$", "\\1", cacheFilenames)
    correctFilenames <- file.path(cachePath, fileTails)
    filesExistInCache <- file.exists(correctFilenames)
    if (all(filesExistInCache)) {
      cacheFilenames <- correctFilenames
    } else {
      stop("File-backed raster files in the cache are corrupt for cacheId: ", cacheId)
    }
  }
  out <- hardLinkOrCopy(cacheFilenames[filesExistInCache],
                        origFilenames[filesExistInCache],
                        overwrite = TRUE
  )

  newOutput <- updateFilenameSlots(obj$cacheRaster,
                                   Filenames(obj, allowMultiple = FALSE),
                                   newFilenames = grep("\\.gri$", origFilenames, value = TRUE, invert = TRUE)
  )
  obj <- newOutput
  obj <- .setSubAttrInList(obj, ".Cache", "newCache", FALSE)
  obj
}

## misc non-exported stuff ------------------------------------------------------
grepStartsTwoDots <- "^\\.\\."

## exported generics and functions ----------------------------------------------

### .tagsByClass ----------------------------------------------------------------

#' Exported generics and methods
#'
#' There are a number of generics that are exported for other packages to use.
#' These are listed below. They are not intended for use by normal users; rather,
#' they are made available for package developers to build specific methods.
#'
#' @details
#' The methods should do as described below.
#'
#' `.tagsByClass` should return a character vector, with a single colon (":") dividing
#' two parts: the tag type and tag value, for the specific class.
#'
#' @return
#' `.tagsByClass` default method returns `NULL`.
#'
#'
#' @param object Any R object.
#'
#' @author Eliot McIntire
#' @export
#' @name ExportedMethods
#' @aliases .tagsByClass
#' @rdname exportedMethods
#' @examples
#' .tagsByClass(character())
#'
setGeneric(".tagsByClass", function(object) {
  standardGeneric(".tagsByClass")
})

#' @export
#' @rdname exportedMethods
setMethod(
  ".tagsByClass",
  signature = "ANY",
  definition = function(object) {
    NULL
  }
)

### .cacheMessage ---------------------------------------------------------------

#' @details
#' `.cacheMessage` should make a call to `message` that gives information about
#' the loaded cached object being returned.
#'
#' @return `.cacheMessage`: nothing; called for its messaging side effect, which,
#' by default, just edits the name of the function into a generic "loaded cached result"
#' message.
#'
#' @param object Any R object.
#' @param functionName A character string indicating the function name
#' @param fromMemoise Logical. If `TRUE`, the message will be about
#'        recovery from memoised copy
#' @inheritParams Cache
#'
#' @author Eliot McIntire
#' @export
#' @rdname exportedMethods
#' @examples
#' a <- 1
#' .cacheMessage(a, "mean")
#'
setGeneric(".cacheMessage", function(object, functionName,
                                     fromMemoise = getOption("reproducible.useMemoise", TRUE),
                                     verbose = getOption("reproducible.verbose", 1)) {
  standardGeneric(".cacheMessage")
})

#' @export
#' @rdname exportedMethods
setMethod(
  ".cacheMessage",
  signature = "ANY",
  definition = function(object, functionName, fromMemoise, verbose = getOption("reproducible.verbose", 1)) {
    postMess <- NULL
    whMessage <- .message$LoadedCacheResult()
    if (isTRUE(fromMemoise)) {
      whMessage <- .message$LoadedCacheResult(2)
    } else if (fromMemoise %in% FALSE) {
      if (isTRUE(getOption("reproducible.useMemoise")))
        postMess <- paste0(" ", .message$AddingToMemoised)
    }
    baseMess <- .message$LoadedCache(whMessage, .messageFunctionFn(functionName)) # ELIOT HERE
    if (!is.null(postMess))
      baseMess <- paste0(baseMess, postMess)
    messageCache(baseMess, verbose = verbose)
    return(invisible(whMessage))
  }
)

### .cacheMessageObjectToRetrieve -------------------------------------------------------------

#' @param fullCacheTableForObj The data.table entry from the Cache database for only
#' this `cacheId`, e.g., via `showCache()`.
#' @inheritParams Cache
#' @inheritParams .unwrap
#'
#' @details
#' `.objecxtToRetrieveMessage` is the messaging for recovering an object from Cache.
#'
#' @export
#' @rdname exportedMethods
.cacheMessageObjectToRetrieve <- function(functionName, fullCacheTableForObj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                                          cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
                                          verbose) {
  objSize <- as.numeric(tail(extractFromCache(fullCacheTableForObj, elem = "file.size"), 1))
  class(objSize) <- "object_size"
  bigFile <- isTRUE(objSize > 1e6)

  fileFormat <- unique(extractFromCache(fullCacheTableForObj, elem = "fileFormat")) # can have a single tif for many entries
  if (is.null(fileFormat))
    fileFormat <- cacheSaveFormat

  messageCache(.message$ObjToRetrieveFn(functionName), ", ",
               #             messageCache("...(Object to retrieve (fn: ", .messageFunctionFn(functionName), ", ",
               basename2(CacheStoredFile(cachePath, cacheId, cacheSaveFormat = fileFormat,
                                         readOnly = TRUE)),
               ")",
               if (bigFile) " is large: ",
               if (bigFile) format(objSize, units = "auto"),
               " ... ",
               verbose = verbose
  )
}

#' @details
#' `.addTagsToOutput` should add one or more attributes to an object, named either
#' `"tags"`, `"call"` or `"function"`. It may be wise to do a "deep" copy within this
#' method, but it may not be necessary.
#'
#' @inheritParams Cache
#'
#' @param object Any R object.
#'
#' @param FUN A function
#'
#' @param preDigestByClass A list, usually from `.preDigestByClass`
#'
#' @return `.addTagsToOutput`: The inputted object but with tags attached.
#'
#' @author Eliot McIntire
#' @export
#' @rdname exportedMethods
#'
setGeneric(".addTagsToOutput", function(object, outputObjects, FUN, preDigestByClass) { # nolint
  standardGeneric(".addTagsToOutput")
})

#' @export
#' @rdname exportedMethods
setMethod(
  ".addTagsToOutput",
  signature = "ANY",
  definition = function(object, outputObjects, FUN, preDigestByClass) { # nolint
    object
  }
)


### .preDigestByClass ---------------------------------------------------------------------------

#' `preDigestByClass`: A function called to do any miscellaneous things to
#' do before `.robustDigest` and after `FUN` call
#'
#' @param object Any R object.
#'
#' @return `.preDigestByClass`: A list with elements that have a difficult time
#'   being digested correctly, e.g., an S4 object with some elements removed and
#'   handled for digesting purposes. The default method for `preDigestByClass` and
#'   simply returns `NULL`.
#'
#' @author Eliot McIntire
#' @export
#' @rdname exportedMethods
#' @examples
#' a <- 1
#' .preDigestByClass(a) # returns NULL in the simple case here.
#'
setGeneric(".preDigestByClass", function(object) { # nolint
  standardGeneric(".preDigestByClass")
})

#' @export
#' @rdname exportedMethods
setMethod(
  ".preDigestByClass",
  signature = "ANY",
  definition = function(object) { # nolint
    NULL
  }
)

### .checkCacheRepo ---------------------------------------------------------------------------

#' `.checkCacheRepo`: normally, `checkPath` can be called directly, and if more
#' friendly, but that does not have class-specific methods.
#'
#' @param object An R object
#' @param create Logical. If TRUE, then it will create the path for cache.
#' @inheritParams Cache
#'
#' @return `.checkCacheRepo`: A character string with a path to a cache repository.
#'
#' @author Eliot McIntire
#' @export
#' @rdname exportedMethods
#' @examples
#' a <- normalizePath(file.path(tempdir(), "test"), mustWork = FALSE)
#' .checkCacheRepo(a, create = TRUE)
#'
setGeneric(".checkCacheRepo", function(object, create = FALSE,
                                       verbose = getOption("reproducible.verbose", 1)) {
  standardGeneric(".checkCacheRepo")
})

#' @export
#' @rdname exportedMethods
#' @include messages.R
setMethod(
  ".checkCacheRepo",
  signature = "ANY",
  definition = function(object, create, verbose = getOption("reproducible.verbose", 1)) {
    cachePath <- tryCatch(checkPath(object, create), error = function(x) {
      cachePath <- if (isTRUE(nzchar(getOption("reproducible.cachePath")[1]))) {
        tmpDir <- .reproducibleTempCacheDir()
        # Test whether the user has accepted the default. If yes, then give message.
        #  If no, then user is aware and doesn't need a message
        if (any(grepl(normPath(tmpDir), normPath(getOption("reproducible.cachePath")))) ||
            any(grepl(normPath(tempdir()), normPath(getOption("reproducible.cachePath"))))) {
          messageCache(.message$NoCachePathSupplied, " and getOption('reproducible.cachePath') is inside a temporary directory;\n",
                       "  this will not persist across R sessions.",
                       verbose = verbose
          )
        }
        getOption("reproducible.cachePath", tmpDir)
      } else {
        messageCache(.message$NoCachePathSupplied, ". Using ", .reproducibleTempCacheDir(), verbose = verbose)
        .reproducibleTempCacheDir()
      }
      checkPath(path = cachePath, create = create)
    })
  }
)

### .prepareOutput ----------------------------------------------------------------------------

#'  @details
#'  `.prepareOutput` allows a package developer to do whatever is needed to prepare
#'  the output on recovery from the cache repository. In other words,
#'  there are cases where the object being cached may have some modifications for
#'  writing to file/database that are not part of the object when returned in
#'  R. Example is a `NULL` entry. A class may need this to be stored on disk,
#'  but `NULL` does not store reliably on disk. So this would convert the
#'  `"NULL"` (character string, stored on disk) to a `NULL` returned to user.
#'  @inheritParams Cache
#'
#' @param object Any R object
#'
#' @return `.prepareOutput`: The object, modified
#'
#' @author Eliot McIntire
#' @export
#' @rdname exportedMethods
#' @examples
#' a <- 1
#' .prepareOutput(a) # does nothing
#'
#' b <- "NULL"
#' .prepareOutput(b) # converts to NULL
#'
#' if (requireNamespace("terra", quietly = TRUE)) {
#'   r <- terra::rast(terra::ext(0, 10, 0, 10), vals = 1:100)
#'
#'   # write to disk manually -- will be in tempdir()
#'   r <- terra::writeRaster(r, filename = tempfile(fileext = ".tif"))
#'
#'   # copy it to the cache repository
#'   r <- .prepareOutput(r, tempdir())
#' }
setGeneric(".prepareOutput", function(object, cachePath = getOption("reproducible.cachePath"), ...) {
  standardGeneric(".prepareOutput")
})

#' @export
#' @rdname exportedMethods
setMethod(
  ".prepareOutput",
  signature = "ANY",
  definition = function(object, cachePath = getOption("reproducible.cachePath"), ...) {
    if (is.character(object)) {
      if (length(object) == 1) {
        # need something to attach tags to if it is actually NULL
        if (identical(object, "NULL")) object <- NULL
      }
    }
    object
  }
)

### .addChangedAttr ---------------------------------------------------------------------------

#' @details
#' `.addChangedAttr` should return the same object, with this a very specific
#' attribute added: it must be named ".Cache", and have a sub-list element named
#' "changed", which must be a logical, `TRUE` or `FALSE`, to describe whether the object
#' has changed or not since last attempt to cache it. This is mostly useful when there
#' are only one or a few sub-elements of e.g., a large list, that are changed. `Cache`
#' will be able to only recover the changed parts, to reduce time required to
#' complete a call to `Cache`.
#'
#' @param object Any R object returned from a function
#' @param preDigest The full, element by element hash of the input arguments to that same function,
#' e.g., from `.robustDigest`
#' @param origArguments These are the actual arguments (i.e., the values, not the names) that
#'        were the source for `preDigest`
#' @param ... Anything passed to methods.
#'
#' @return `.addChangedAttr`: the object, with an attribute ".Cache" and sub-element, "changed",
#'   added set to either `TRUE` or `FALSE`
#'
#' @author Eliot McIntire
#' @export
#' @rdname exportedMethods
#' @examples
#' a <- 1
#' .addChangedAttr(a) # does nothing because default method is just a pass through
setGeneric(".addChangedAttr", function(object, preDigest, origArguments, ...) {
  standardGeneric(".addChangedAttr")
})

#' @export
#' @rdname exportedMethods
setMethod(
  ".addChangedAttr",
  signature = "ANY",
  definition = function(object, preDigest, origArguments, ...) {
    object
  }
)

#' Remap file names
#'
#' Update file path metadata for file-backed objects (e.g., `SpatRasters`).
#' Useful when moving saved objects between projects or machines.
#'
#' @param obj (optional) object whose file path metadata will be remapped
#' @param tags cache tags `data.table` object
#' @param cachePath character string specifying the path to the cache directory or `NULL`
#' @param ... Additional path arguments, passed to `absoluteBase()` and `modifyListPaths()`
#'
#' @export
#' @importFrom fs path_join path_norm
remapFilenames <- function(obj, tags, cachePath = getOption("reproducible.cachePath"), ...) {
  tags <- parseTags(tags)
  origFilename <- extractFromCache(tags, tagOrigFilename) # tv[tk == tagOrigFilename]

  if (missing(obj)) {
    origRelName <- extractFromCache(tags, tagOrigRelName) |> fs::path_norm()
    relToWhere <- extractFromCache(tags, "relToWhere")

    ## NOTE: extractFromCache() is looking for specific tags which may not exist if saved
    ## using earlier versions of the package, and cannot be restored.
    if (is.null(relToWhere) || length(relToWhere) == 0) {
      stop("remapFileNames() cannot restore objects saved using a previous version of 'reproducible'.")
    }

    possRelPaths <- modifyListPaths(cachePath, ...)
    relToWhereInPossRelPaths <- relToWhere %in% names(possRelPaths)
    if (isTRUE(any(relToWhereInPossRelPaths))) {
      absBase <- absoluteBase(relToWhere, cachePath, ...)
    } else {
      absBase <- possRelPaths[[1]]
      isOutside <- grepl(grepStartsTwoDots, origRelName)
      if (isTRUE(any(isOutside))) {
        ## means the relative path is "outside" of something;
        ## strip all ".." if relToWhere doesn't exist
        while (any(grepl(grepStartsTwoDots, origRelName))) {
          origRelName <- gsub(paste0(grepStartsTwoDots, "|(\\\\|/)"), "", origRelName)
        }
      }
    }
    newName <- vapply(origRelName, function(x) {
      if (fs::is_absolute_path(x)) {
        x
      } else {
        fs::path_join(c(absBase, x)) |> fs::path_norm()
      }
    }, character(1))
  } else {
    newName <- obj
  }

  whFiles <- newName[match(basename(extractFromCache(tags, tagFilesToLoad)), origFilename)]
  list(newName = newName, whFiles = whFiles, tagsParsed = tags)
}

#' @details
#' `updateFilenameSlots`: this exists because when copying file-backed rasters, the
#' usual mechanism of `writeRaster` can be very slow. This function allows
#' for a user to optionally create a hard link to the old file, give it a new
#' name, then update the filename slot(s) in the `Raster*` class object. This
#' can be 100s of times faster for large rasters.
#'
#' @return
#' `updateFilenameSlots`: The original object, but with its internal file pointer
#' updated to the `newFilenames`
#'
#' @export
#' @keywords internal
#' @param obj An object. This function only has useful methods for `Raster*`,
#'   with all other classes being simply a pass-through
#' @param curFilenames An optional character vector of filenames currently existing
#'   and that are pointed to in the obj. If omitted, will take from the `obj`
#'   using `Filenames(obj)`
#' @param newFilenames An optional character vector of filenames to use instead of
#'   the curFilenames. This can also be a single directory, in which case the
#'   renaming will be given:
#'   `file.path(newFilenames, basename(Filenames(obj, allowMultiple = FALSE)))`
#' @rdname exportedMethods
updateFilenameSlots <- function(obj, curFilenames, newFilenames, isStack = NULL) {
  UseMethod("updateFilenameSlots")
}

#' @rdname exportedMethods
#' @export
#' @keywords internal
updateFilenameSlots.default <- function(obj, curFilenames, newFilenames, isStack = NULL, ...) {
  if (inherits(obj, "Raster")) {
    if (missing(curFilenames)) {
      curFilenames <- Filenames(obj, allowMultiple = FALSE)
    }

    if (missing(newFilenames)) stop("newFilenames can't be missing: either new filenames or a single directory")
    # if newFilenames is a directory
    areDirs <- dir.exists(newFilenames)
    if (any(areDirs) && length(newFilenames) == 1) {
      newFilenames <- file.path(newFilenames, basename(curFilenames))
    }

    if (length(curFilenames) > 1) {
      for (i in seq_along(curFilenames)) {
        if (is.list(obj)) {
          slot(slot(obj[[i]], "file"), "name") <- newFilenames[i]
        } else {
          slot(slot(slot(obj, "layers")[[i]], "file"), "name") <- newFilenames[i]
        }
      }
    } else {
      if (is.null(isStack)) isStack <- is(obj, "RasterStack")
      if (!isStack) {
        slot(slot(obj, "file"), "name") <- newFilenames
      } else {
        if (length(newFilenames) == 1) {
          newFilenames <- rep(newFilenames, nlayers2(obj))
        }
        for (i in seq_len(nlayers2(obj))) {
          whFilename <- unique(match(
            withoutFinalNumeric(basename(newFilenames)),
            withoutFinalNumeric(basename(curFilenames))
          ))
          isNAwhFn <- is.na(whFilename)
          if (any(isNAwhFn)) {
            whFilename <- i
          }
          slot(slot(obj@layers[[i]], "file"), "name") <- newFilenames[whFilename]
        }
        # }
      }
    }
  }
  obj
}

#' @rdname exportedMethods
#' @export
#' @keywords internal
updateFilenameSlots.list <- function(obj, ...) {
  areRasters <- vapply(obj, is, "RasterLayer", FUN.VALUE = logical(1))
  if (all(areRasters)) {
    .requireNamespace("raster", stopOnFALSE = TRUE)
    # a separate option for list of RasterLayers because curFilename will be
    #   as long as all the filenames because there is a method for lists;
    #   passing this to updateFilaneSlots will fail if it is one RasterLayer
    #   at a time
    out <- updateFilenameSlots(raster::stack(obj), ...)
    out <- raster::unstack(out)
  } else {
    out <- lapply(obj, function(o) {
      updateFilenameSlots(o, ...)
    })
  }
  out
}

#' @rdname exportedMethods
#' @export
#' @keywords internal
updateFilenameSlots.environment <- function(obj, ...) {
  if (is.null(names(obj))) {
    names(obj) <- as.character(seq(obj))
  }
  lapply(obj, function(o) {
    updateFilenameSlots(as.list(o), ...)
  })
}

#' @details
#' `makeMemoiseable` and `unmakeMemoisable` methods are run during `Cache`. The
#' methods should address any parts that will not successfully work for memoising,
#' most notably, `data.table` or other pass-by-reference objects likely need to
#' be deep copied, so that the memoised version doesn't get changed after being
#' stashed in the RAM cache (i.e., memoised)
#'
#' @details
#' `makeMemoiseable` and `unmakeMemoisable` are, by default, just a pass through
#' for most class. `reproducible` only has a method for `data.table` class objects.
#'
#' @param x  An object to make memoisable.
#'           See individual methods in other packages.
#' @return The same object, but with any modifications, especially
#' dealing with saving of environments, which memoising doesn't handle
#' correctly in some cases.
#'
#' @export
#' @rdname exportedMethods
makeMemoisable <- function(x) {
  UseMethod("makeMemoisable")
}

#' @export
#' @rdname exportedMethods
makeMemoisable.default <- function(x) {
  x
}

#' @export
#' @rdname exportedMethods
makeMemoisable.data.table <- function(x) {
  data.table::copy(x)
}

#' @export
#' @rdname exportedMethods
unmakeMemoisable <- function(x) {
  UseMethod("unmakeMemoisable")
}

#' @export
#' @rdname exportedMethods
unmakeMemoisable.default <- function(x) {
  x
}

#' Grep system calls
#'
#' A faster way of grepping the system call stack than just
#' `grep(sys.calls(), pattern = "test")`
#'
#' @param sysCalls The return from `sys.calls()`
#' @param pattern Character, passed to grep
#' @return
#' Numeric vector, equivalent to return from `grep(sys.calls(), pattern = "test")`,
#' but faster if `sys.calls()` is very big.
#'
#' @keywords internal
#' @export
#' @rdname grepSysCalls
.grepSysCalls <- function(sysCalls, pattern) {
  scallsFirstElement <- lapply(sysCalls, function(x) x[1])
  grep(scallsFirstElement, pattern = pattern)
}

#' Deal with class for saving to and loading from Cache or Disk
#'
#' This generic and some methods will do whatever is required to prepare an object for
#' saving to disk (or RAM) via e.g., `saveRDS`. Some objects (e.g., `terra`'s `Spat*`)
#' cannot be saved without first wrapping them. Also, file-backed objects are similar.
#'
#' @param obj Any arbitrary R object.
#' @param ... Arguments passed to methods; default does not use anything in `...`.
#' @inheritParams Cache
#' @inheritParams loadFromCache
#'
#' @return Returns an object that can be saved to disk e.g., via `saveRDS`.
#'
#' @export
#' @rdname dotWrap
.wrap <- function(obj, cachePath = getOption("reproducible.cachePath"), preDigest,  drv = getDrv(getOption("reproducible.drv", NULL)),
                  conn = getOption("reproducible.conn", NULL),
                  verbose = getOption("reproducible.verbose"), outputObjects  = NULL,
                  cacheId = NULL, ...) {
  UseMethod(".wrap")
}

#' @export
#' @rdname dotWrap
.wrap.list <- function(obj, cachePath = getOption("reproducible.cachePath"), preDigest, drv = getDrv(getOption("reproducible.drv", NULL)),
                       conn = getOption("reproducible.conn", NULL),
                       verbose = getOption("reproducible.verbose"), outputObjects = NULL,
                       cacheId = NULL, ...) {

  if (!is.null(outputObjects)) {
    allObjs <- names(obj)
    nullify <- setdiff(allObjs, outputObjects)
    obj[nullify] <- NULL
  }

  attrsOrig <- attributes(obj)
  ## this is not calling the right method:
  # obj <- lapply(obj, .wrap, preDigest = preDigest, cachePath = cachePath, drv = drv,
                # conn = conn, verbose = verbose, ...)

  obj <- lapply(obj, function(objj, ...) {
    .wrap(objj, ...)
  }, preDigest = preDigest, cachePath = cachePath, drv = drv,
     conn = conn, verbose = verbose, cacheId = cacheId, ...)
  hasTagAttr <- lapply(obj, function(x) attr(x, "tags"))
  tagAttr <- unname(unlist(hasTagAttr)) # this removed name
  if (length(tagAttr)) {
    if (is.null(attrsOrig[["tags"]])) {
      newList <- tagAttr
    } else {
      newList <- try(c(attrsOrig[["tags"]], tagAttr))
    }
    attrsOrig[["tags"]] <- newList
  }
  if (!is.null(attrsOrig)) {
    for (tt in intersect(names(attrsOrig), c(".Cache", "tags", callInCache)))
      attr(obj, tt) <- attrsOrig[[tt]]
  }
  obj
}

#' @export
#' @rdname dotWrap
.wrap.environment <- function(obj, cachePath = getOption("reproducible.cachePath"), preDigest, drv = getDrv(getOption("reproducible.drv", NULL)),
                              conn = getOption("reproducible.conn", NULL),
                              verbose = getOption("reproducible.verbose"), outputObjects = NULL,
                              cacheId = NULL, ...) {
  if (!is.null(outputObjects)) {
    allObjs <- ls(obj)
    nullify <- setdiff(allObjs, outputObjects)
    rm(list = nullify, envir = obj)
  }

  if (length(ls(obj, all.names = TRUE)) > 0) {
    obj2 <- as.list(obj, all.names = TRUE)
    out <- .wrap(obj2, cachePath = cachePath, preDigest = preDigest, drv = drv,
                 conn = conn, verbose = verbose, outputObjects = outputObjects, cacheId = cacheId, ...)
    # obj <- Copy(obj)
    obj2 <- list2envAttempts(out, obj)
    if (!is.null(obj2)) obj <- obj2

  }
  obj
}

#' @export
#' @rdname dotWrap
#' @examples
#' # For SpatExtent
#' if (requireNamespace("terra")) {
#'   ex <- terra::ext(c(0, 2, 0, 3))
#'   exWrapped <- .wrap(ex)
#'   ex1 <- .unwrap(exWrapped)
#' }
#'
.wrap.default <- function(obj, cachePath = getOption("reproducible.cachePath"), preDigest, drv = getDrv(getOption("reproducible.drv", NULL)),
                          conn = getOption("reproducible.conn", NULL),
                          verbose = getOption("reproducible.verbose"), outputObjects = NULL,
                          cacheId = NULL, ...) {
  rasters <- is(obj, "Raster")
  atts <- attributes(obj)
  reassignAtts <- TRUE
  if (any(rasters)) {
    .requireNamespace("raster", stopOnFALSE = TRUE)
    objOrig <- obj
    obj <- .prepareFileBackedRaster(obj,
                                    repoDir = cachePath,
                                    overwrite = FALSE, drv = drv, conn = conn,
                                    verbose = verbose
    )
    isFromDisk <- raster::fromDisk(obj)

    # have to reset all these attributes on the rasters as they were undone in prev steps
    hasFromDisk <- if (!is.null(atts$tags)) startsWith(atts$tags, "fromDisk") else FALSE

    if (any(hasFromDisk)) {
      atts$tags[hasFromDisk] <- paste("fromDisk", sep = ":", isFromDisk)
    } else {
      atts$tags <- c(atts$tags, paste("fromDisk", sep = ":", isFromDisk))
    }

    attr(obj, "tags") <- atts$tags
    obj <- .setSubAttrInList(obj, ".Cache", "newCache", atts$.Cache$newCache)
    attr(obj, callInCache) <- atts$callInCache

    if (!identical(attr(obj, ".Cache")$newCache, atts$.Cache$newCache)) {
      stop("attributes are not correct 6")
    }
    if (!identical(attr(obj, callInCache), atts$callInCache)) {
      stop("attributes are not correct 7")
    }
    if (!identical(attr(obj, "tags"), atts$tags)) {
      stop("attributes are not correct 8")
    }

    if (!is.null(atts[["function"]])) {
      attr(obj, "function") <- atts[["function"]]
      if (!identical(attr(obj, "function"), atts[["function"]])) {
        stop("There is an unknown error 04")
      }
    }
    if (isFromDisk) {
      obj <- list(origRaster = Filenames(objOrig), cacheRaster = obj)
      attr(obj, "tags") <- c(
        attributes(obj$cacheRaster)$tags,
        paste0("origRaster:", obj$origRaster),
        paste0("cacheRaster:", Filenames(obj))
      )
    }
  }

  if (is(obj, "SpatVectorCollection")) {
    obj <- .wrap(as.list(obj), cacheId = cacheId)
    class(obj) <- "PackedSpatVectorCollection"
  }

  if (any(inherits(obj, c("SpatVector", "SpatRaster", "SpatExtent", "data.table")))) {
    if (!requireNamespace("terra", quietly = TRUE)) {
      stop("Please install terra package")
    }
    # attrs <- attr(obj, ".Cache")

    # next is for terra objects --> terra::wrap is ridiculously slow for SpatVector objects; use
    #   custom version in reproducible where here
    useWrap <- TRUE
    if (.isSpatRaster(obj)) {
      if (all(nzchar(Filenames(obj)))) {
        useWrap <- FALSE
        obj <- wrapSpatRaster(obj, cachePath, cacheId = cacheId, ...)
      }
    } else if (is(obj, "SpatVector") && !missing(cachePath)) {
      useWrap <- FALSE
      obj <- wrapSpatVector(obj)
    } else if (is(obj, "SpatExtent")) {
      obj <- list(
        xmin = terra::xmin(obj), xmax = terra::xmax(obj),
        ymin = terra::ymin(obj), ymax = terra::ymax(obj)
      )
      attr(obj, "class") <- "PackedSpatExtent2"
      useWrap <- FALSE
      reassignAtts <- FALSE
    }
    if (is(obj, "data.table")) {
      obj <- data.table::copy(obj)
      useWrap <- FALSE
    }

    if (useWrap) {
      messageCache("wrapping terra object for saving...", verboseLevel = 2, verbose = verbose)
      obj <- terra::wrap(obj)
    } # let method dispatch work

    # attr(obj, ".Cache") <- attrs

    messageCache("\b Done!", verboseLevel = 2, verbose = verbose)
  }

  # put attributes back on the potentially packed object
  if (isTRUE(reassignAtts))
    obj <- attributesReassign(atts, obj)

  obj
}

#' @export
#' @rdname dotWrap
.unwrap.default <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                            drv = getDrv(getOption("reproducible.drv", NULL)),
                            conn = getOption("reproducible.conn", NULL), ...) {
  atts <- attributes(obj)
  if (any(inherits(obj, "PackedSpatVectorCollection"))) {
    obj <- lapply(obj, .unwrap)
    obj <- terra::svc(obj)
  }
  if (any(inherits(obj, c("PackedSpatRaster")))) {
    if (!requireNamespace("terra")) stop("Please install.packages('terra')")
    if (any(inherits(obj, "PackedSpatRaster"))) {
      obj <- terra::rast(obj)
    }

  } else if (is(obj, "Path")) {
    obj <- unwrapSpatRaster(obj, cachePath, cacheId = cacheId, ...)
    # obj2 <- try(unwrapSpatRaster(obj, cachePath, ...))
    # if (is(obj2, "try-error")) {
    #   browser()
    #   clearCache(cacheId = cacheId, x = cachePath, ask = FALSE, drv = drv, conn = conn)
    #   messageCache("Removing ", cacheId, " from the Cache as the file-backing is missing or corrupt")
    #   obj <- .returnNothing
    # } else {
    #   obj <- obj2
    # }
  }
  # put attributes back on the potentially packed object
  obj <- attributesReassign(atts, obj)

  obj
}

#' @export
#' @param cacheId Used strictly for messaging. This should be the cacheId of the object being recovered.
#'   Default is `NULL`.
#' @rdname dotWrap
.unwrap <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                    drv = getDrv(getOption("reproducible.drv", NULL)),
                    conn = getOption("reproducible.conn", NULL), ...) {
  UseMethod(".unwrap")
}

#' @export
#' @rdname dotWrap
.unwrap.environment <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                                drv = getDrv(getOption("reproducible.drv", NULL)),
                                conn = getOption("reproducible.conn", NULL), ...) {
  # the as.list doesn't get everything. But with a simList, this is OK; rest will stay
  atts <- attributes(obj)
  # if (!is.null(obj$fireSense_dataPrepFit$.objects$studyAreaUnion)) browser()
  objList <- as.list(obj, all.names = TRUE) # don't overwrite everything, just the ones in the list part; but need .mods, .objects etc.

  outList <- .unwrap(objList, cachePath = cachePath, cacheId = cacheId, drv = drv, conn = conn, ...)
  output2 <- list2envAttempts(outList, obj) # don't return it if the list2env retured nothing (a normal environment situation; not simList)
  if (!is.null(output2)) obj <- output2

  # put attributes back on the potentially packed object
  obj <- attributesReassign(atts, obj)

  obj
}

#' @export
#' @rdname dotWrap
.unwrap.list <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                         drv = getDrv(getOption("reproducible.drv", NULL)),
                         conn = getOption("reproducible.conn", NULL), ...) {
  atts <- attributes(obj)
  anyNames <- names(obj)
  # isSpatVector <- if (is.null(anyNames) || length(anyNames) == 0) FALSE else {
  #   identical(spatVectorNamesForCache, anyNames)
  # }
  FAIL <- FALSE
  # if (isTRUE(isSpatVector)) {
  #   obj <- unwrapSpatVector(obj)
  # } else {
    isRaster <- isTRUE("cacheRaster" %in% names(obj))
    if (isRaster) {
      obj <- unwrapRaster(obj, cachePath, cacheId)
    } else {
      obj <- lapply(obj, function(out) {
        ret <- .unwrap(out, cachePath, cacheId = cacheId, drv, conn, ...)
        if (is.character(ret))
          if (identical(.returnNothing, ret))
            FAIL <<- TRUE
        ret
      })
    }
  # }
  # put attributes back on the potentially packed object
  obj <- attributesReassign(atts, obj)

  if (isTRUE(FAIL))
    obj <- .returnNothing
  obj
}

#' @export
#' @rdname dotWrap
.unwrap.PackedSpatExtent2 <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                                      drv = getDrv(getOption("reproducible.drv", NULL)),
                                      conn = getOption("reproducible.conn", NULL), ...) {
  atts <- attributes(obj)
  obj <- terra::ext(unlist(obj))
  # put attributes back on the potentially packed object
  obj <- attributesReassign(atts, obj)

  obj

}


#' @export
#' @rdname dotWrap
.unwrap.PackedSpatVector2 <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                            drv = getDrv(getOption("reproducible.drv", NULL)),
                            conn = getOption("reproducible.conn", NULL), ...) {
  atts <- attributes(obj)
  obj <- unwrapSpatVector(obj)
  # put attributes back on the potentially packed object
  obj <- attributesReassign(atts, obj)

  obj

}

#' @export
#' @rdname dotWrap
.unwrap.data.table <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                                     drv = getDrv(getOption("reproducible.drv", NULL)),
                                     conn = getOption("reproducible.conn", NULL), ...) {
  atts <- attributes(obj)
  obj <- data.table::copy(obj)
  # put attributes back on the potentially packed object
  obj <- attributesReassign(atts, obj)

  obj

}

#' @export
#' @rdname dotWrap
.unwrap.PackedSpatVector <- function(obj, cachePath = getOption("reproducible.cachePath"), cacheId = NULL,
                                      drv = getDrv(getOption("reproducible.drv", NULL)),
                                      conn = getOption("reproducible.conn", NULL), ...) {
  atts <- attributes(obj)
  obj <- terra::vect(obj)
  # put attributes back on the potentially packed object
  obj <- attributesReassign(atts, obj)

  obj

}

filenameInCacheWPrefix <- function(obj, cacheId = NULL, relative = TRUE) {
  # cacheId will be missing if it is in e.g., prepInputs without Cache
  if (!is.null(obj)) {
    filenameInCache <- if (is.null(cacheId)) obj else .prefix(obj, prefixCacheId(cacheId))
    if (isTRUE(relative))
      obj <- basename2(filenameInCache)
    else
      obj <- filenameInCache
  }
  obj
}

test_that("getRelativePaths works as expected", {
  testInit("fs")

  ## ensure multiple e.g. module paths work;
  ## based on usage in SpaDES.core::saveSimList()
  path <- list(cachePath = structure("/mnt/scratch/myProject/cache",
                                     class = c("fs_path", "character")),
               inputPath = structure("/mnt/projects/myProject/inputs",
                                     class = c("fs_path", "character")),
               modulePath = structure(c("/home/testUser/myProject/modules",
                                        "/home/testUser/myProject/modules/scfm/modules"),
                                      class = c("fs_path", "character")),
               outputPath = structure("/mnt/projects/myProject/outputs/runName_for_rep01",
                                      class = c("fs_path", "character")),
               rasterPath = structure("/mnt/scratch/myProject/raster",
                                      class = c("fs_path", "character")),
               scratchPath = structure("/mnt/scratch/myProject",
                                       class = c("fs_path", "character")),
               terraPath = structure("/mnt/scratch/myProject/terra",
                                     class = c("fs_path", "character")))

  relativeToPath <- "/home/testUser/myProject"

  corePaths <- c("modulePath", "cachePath", "inputPath", "outputPath")

  path[corePaths] <- getRelative(path[corePaths], relativeToPath)

  expect_identical(path$cachePath, "cache")
  expect_identical(path$inputPath, "inputs")
  expect_identical(path$outputPath, file.path("outputs", "runName_for_rep01"))
  expect_identical(path$modulePath, c("modules", file.path("modules", "scfm", "modules")))
})

test_that("relativeToWhat can handle multiple paths", {
  expect_no_error({
    res <- relativeToWhat(
      file = "/mnt/projects/HRV/BC_HRV/outputs/NRD_Quesnel_scfm_hrv_FRT_res125/rep01/speciesLayers_2011_NRD_Quesnel.tif",
      cachePath = NULL,
      paths = list(
        cachePath = "/mnt/scratch/achubaty/BC_HRV/cache",
        inputPath = "/mnt/projects/HRV/BC_HRV/inputs",
        modulePath = c("/home/achubaty/GitHub/BC_HRV/modules", "/home/achubaty/GitHub/BC_HRV/modules/scfm/modules"),
        outputPath = "/mnt/projects/HRV/BC_HRV/outputs/NRD_Quesnel_scfm_hrv_FRT_res125/rep01",
        rasterPath = "/mnt/scratch/achubaty/BC_HRV/raster",
        scratchPath = "/mnt/scratch/achubaty/BC_HRV",
        terraPath = "/mnt/scratch/achubaty/BC_HRV/terra"
      )
    )
  })
  expect_identical(res, list(outputPath = "."))
})

#' Normalize file paths
#'
#' Checks the specified path for formatting consistencies:
#' 1. use slash instead of backslash;
#' 2. do tilde etc. expansion;
#' 3. remove trailing slash.
#'
#' Additionally, `normPath()` attempts to create a absolute paths,
#' whereas `normPathRel()` maintains relative paths.
#'
#' ```
#' d> getwd()
#' [1] "/home/achubaty/Documents/GitHub/PredictiveEcology/reproducible"
#' d> normPathRel("potato/chips")
#' [1] "potato/chips"
#' d> normPath("potato/chips")
#' [1] "/home/achubaty/Documents/GitHub/PredictiveEcology/reproducible/potato/chips"
#' ```
#'
#' @param path A character vector of filepaths.
#'
#' @return Character vector of cleaned up filepaths.
#'
#' @export
#' @rdname normPath
#'
#' @example inst/examples/example_checkPath.R
#'
setGeneric("normPath", function(path) {
  standardGeneric("normPath")
})

#' @export
#' @rdname normPath
setMethod(
  "normPath",
  signature(path = "character"),
  definition = function(path) {
    if (length(path) > 0) {
      nonEmpty <- nzchar(path)
      if (any(nonEmpty)) {
        pathOrig <- path
        path <- pathOrig[nonEmpty]

        nas <- is.na(path)
        if (!all(nas)) {
          if (any(!nas)) {
            # fs has different behaviour than `normalizePath`. Need to replace all the differences:
            # Below are "differences"... i.e., I am not writing what normalizePath does, as it is not this:
            # 1. fs::path_norm doesn't expand
            # 2. fs::path_expand: doesn't expand R_USER env var --> must use fs::path_expand_r
            # 3. fs::??? Long windows paths that are shortened to 8 characters (e.g., on GitHub Actions),
            #    normalizePaths does this, can't find equivalent in fs
            path[!nas] <-
              normalizePath(path[!nas], winslash = "/", mustWork = FALSE)
            # fs::path_expand_r(fs::path_abs(path[!nas])) # faster than normalizePath on some machines
          }
          if (any(nas)) {
            path[nas] <- NA_character_
          }

          # Eliot changed this Sept 24, 2019 because weird failures with getwd()
          # in non-interactive testing
          path <- unlist(path)
          if (!is.null(path)) {
            nonNApath <- path[!nas]
            nonNApath <- gsub("\\\\", "//", nonNApath)
            nonNApath <- gsub("//", "/", nonNApath)
            hasDotStart <- startsWith(nonNApath, "./")
            if (isTRUE(any(hasDotStart))) {
              nonNApath[hasDotStart] <-
                gsub("^[.]/", paste0(getwd(), "/"), nonNApath[hasDotStart]) # is absolute b/c getwd()
            }
            nonNApath <- gsub("/$", "", nonNApath) # nolint

            # if the files or dirs don't exist, then there is a possibility on *nix-alikes that they will still
            #    not be absolute -- this is true with R 4.2.3, maybe was not previously
            if (!isWindows() && !all(nas) && any(hasDotStart %in% FALSE)) {
              areAbs <- isAbsolutePath(nonNApath[!hasDotStart]) # hasDotStart is already absolute; still this is slow
              if (any(areAbs %in% FALSE)) {
                nonNApath[!hasDotStart][areAbs %in% FALSE] <-
                  normalizePath(file.path(getwd(), nonNApath[!hasDotStart][areAbs %in% FALSE]),
                                winslash = "/", mustWork = FALSE
                  )
              }
            }
            path[!nas] <- nonNApath

            if (!all(nonEmpty)) {
              pathOrig[nonEmpty] <- path
              path <- pathOrig
            }
          }
        }
      }
    }
    return(path)
  }
)

#' @export
#' @rdname normPath
setMethod(
  "normPath",
  signature(path = "list"),
  definition = function(path) {
    return(normPath(unlist(path)))
  }
)

#' @export
#' @rdname normPath
setMethod(
  "normPath",
  signature(path = "NULL"),
  definition = function(path) {
    return(character(0))
  }
)

#' @export
#' @rdname normPath
setMethod(
  "normPath",
  signature(path = "missing"),
  definition = function() {
    return(character(0))
  }
)

#' @export
#' @rdname normPath
setMethod(
  "normPath",
  signature(path = "logical"),
  definition = function(path) {
    return(NA_character_)
  }
)

#' @importFrom fs path_expand path_norm
#' @export
#' @rdname normPath
normPathRel <- function(path) {
  if (missing(path) || is.null(path)) path <- character(0)

  path <- unlist(path)

  ## empty paths should not be normalized b/c returns the current directory
  path[nzchar(path)] <- path[nzchar(path)] |>
    fs::path_norm() |>
    fs::path_expand() # |>
  # normalizePath(winslash = "/", mustWork = FALSE)

  path
}

#' Check directory path
#'
#' Checks the specified path to a directory for formatting consistencies,
#' such as trailing slashes, etc.
#'
#' @note This will not work for paths to files.
#' To check for existence of files, use `file.exists()`.
#' To normalize a path to a file, use `normPath()` or `normalizePath()`.
#'
#' @param path A character string corresponding to a directory path.
#'
#' @param create A logical indicating whether the path should
#' be created if it does not exist. Default is `FALSE`.
#'
#' @return Character string denoting the cleaned up filepath.
#'
#' @seealso [file.exists()], [dir.create()], [normPath()]
#'
#' @export
#' @rdname checkPath
#'
#' @example inst/examples/example_checkPath.R
#'
setGeneric("checkPath", function(path, create) {
  standardGeneric("checkPath")
})

#' @export
#' @rdname checkPath
setMethod(
  "checkPath",
  signature(path = "character", create = "logical"),
  definition = function(path, create) {
    if (isTRUE(all(is.na(path)))) {
      stop("Invalid path: cannot be NA.")
    } else {
      path <- normPath(path) # this is necessary to cover Windows
      # double slash used on non-Windows
      dirsThatExist <- dir.exists(path)
      if (any(!dirsThatExist)) {
        isExistingFile <- file.exists(path)
        if (all(isExistingFile)) {
          messageCache(
            "That path is an existing file(s)",
            verboseLevel = 0,
            verbose = getOption("reproducible.verbose")
          )
        } else {
          if (create == TRUE) {
            lapply(path[!dirsThatExist[!isExistingFile]], function(pth) {
              dir.create(file.path(pth),
                         recursive = TRUE,
                         showWarnings = FALSE
              )
            })
          } else {
            stop(
              "Specified path, ", normPath(path), ", does not exist. Maybe set `create = TRUE?`"
            )
          }
        }
      }
      if (.pkgEnv$SysInfo[["sysname"]] == "Darwin") path <- normPath(path) # ensure path re-normalized after creation

      return(path)
    }
  }
)

#' @export
#' @rdname checkPath
setMethod(
  "checkPath",
  signature(path = "character", create = "missing"),
  definition = function(path) {
    return(checkPath(path, create = FALSE))
  }
)

#' @export
#' @rdname checkPath
setMethod(
  "checkPath",
  signature(path = "NULL", create = "ANY"),
  definition = function(path) {
    stop("Invalid path: cannot be NULL.")
  }
)

#' @export
#' @rdname checkPath
setMethod(
  "checkPath",
  signature(path = "missing", create = "ANY"),
  definition = function() {
    stop("Invalid path: no path specified.")
  }
)

#' @importFrom fs is_absolute_path
isAbsolutePath <- function(pathnames) {
  fs::is_absolute_path(pathnames)
}

#' @importFrom fs path_abs
makeAbsolute <- function(files, absoluteBase) {
  nas <- is.na(files)
  if (!all(nas)) {
    needNormPath <- rep(TRUE, length(files))
    if (length(files[!nas])) {
      areAbs <- isAbsolutePath(files[!nas])
      if (any(!areAbs)) {
        files[!nas][!areAbs] <- fs::path_abs(files[!nas][!areAbs], absoluteBase)
        needNormPath[!nas][!areAbs] <- FALSE
      }
    }
    if (any(needNormPath))
      files[needNormPath] <- normPath(files[needNormPath])
  } else {
    files
  }
  files
}

#' Relative paths
#'
#' Extracting relative file paths.
#'
#' - `getRelative()` searches `path` "from the right" (instead of "from the left")
#' and tries to reconstruct it relative to directory specified by `relativeToPath`.
#' This is useful when dealing with symlinked paths.
#'
#' - `makeRelative()` checks to see if `files` and `normPath(absoluteBase)` share a common path
#' (i.e., "from the left"), otherwise it returns `files`.
#'
#' @param path character vector or list specifying file paths
#'
#' @param relativeToPath directory against which `path` will be relativized.
#'
#' @examplesIf !identical(.Platform$OS.type, "windows")
#'
#' ## create a project directory (e.g., on a hard drive)
#' (tmp1 <- tempdir2("myProject", create = TRUE))
#'
#' ## create a cache directory elsewhere (e.g., on an SSD)
#' (tmp2 <- tempdir2("my_cache", create = TRUE))
#'
#' ## symlink the project cache directory to tmp2
#' ## files created here are actually stored in tmp2
#' prjCache <- file.path(tmp1, "cache")
#' file.symlink(tmp2, prjCache)
#'
#' ## create a dummy cache object file in the project cache dir
#' (tmpf <- tempfile("cache_", prjCache))
#' cat(rnorm(100), file = tmpf)
#' file.exists(tmpf)
#' normPath(tmpf) ## note the 'real' location (i.e., symlink resolved)
#'
#' getRelative(tmpf, prjCache) ## relative path
#' getRelative(tmpf, tmp2) ## relative path
#'
#' makeRelative(tmpf, tmp2) ## abs path; tmpf and normPath(tmp2) don't share common path
#' makeRelative(tmpf, prjCache) ## abs path; tmpf and normPath(tmp2) don't share common path
#' makeRelative(normPath(tmpf), prjCache) ## rel path; share common path when both normPath-ed
#'
#' unlink(tmp1, recursive = TRUE)
#' unlink(tmp2, recursive = TRUE)
#'
#' @importFrom fs is_absolute_path
#' @export
#' @rdname relativePaths
getRelative <- function(path, relativeToPath) {
  path <- normPathRel(path)
  relativeToPath <- normPathRel(relativeToPath)

  relPath <- vapply(path, function(p) {
    if (is_absolute_path(p)) {
      a <- unlist(strsplit(p, "/"))
      a <- a[nzchar(a)]

      b <- unlist(strsplit(relativeToPath, "/"))
      b <- b[nzchar(b)]

      id <- which(a %in% b)
      if (length(id) > 0) {
        ## assume most internal subdirectory is the matching one
        relPath <- do.call(file.path, as.list(a[(max(id) + 1):length(a)]))
      } else {
        relPath <- p
      }
    } else {
      relPath <- p
    }
  }, character(1))

  return(relPath)
}
getRelative <- Vectorize(getRelative, USE.NAMES = FALSE)

#' @param files character vector or list specifying file paths
#' @param absoluteBase base directory (as absolute path) to prepend to `files`
#'
#' @export
#' @rdname relativePaths
makeRelative <- function(files, absoluteBase) {
  # NOT VECTORIZED on absoluteBase
  isList <- is(files, "list")
  filesOrig <- files
  if (isList) {
    nams <- names(files)
    files <- unlist(files)
  }
  if (length(files)) {
    areAbs <- isAbsolutePath(files)
    if (any(areAbs)) {
      files[areAbs] <- normPath(files[areAbs])
      if (length(absoluteBase) < length(files))
        absoluteBase <- rep(absoluteBase, length.out = length(files))
      if (.pkgEnv$runningOnMac) {
        # on macos, normPath returns a different answer depending on whether
        #   the file(s) exists or not. Here, try first as with other OSs; then
        #   below, try again for those that are still absolute.
        absoluteBaseOrig <- absoluteBase
      }
      absoluteBase <- normPath(absoluteBase) # can be "." which means 'any character' in a grep
      files[areAbs] <- unlist(Map(ab = absoluteBase[areAbs], file = files[areAbs], function(ab, file)
        gsub(paste0("^", ab, "/{0,1}"), "", file)
      ))
      if (.pkgEnv$runningOnMac) {
        areAbs <- isAbsolutePath(files)
        if (any(areAbs)) {
          # if (exists("aaaa", envir = .GlobalEnv)) browser()
          files[areAbs] <- unlist(Map(ab = absoluteBaseOrig[areAbs],
                                      file = files[areAbs], function(ab, file)
                                        gsub(paste0("^", ab, "/{0,1}"), "", file)
          ))
        }

      }
      # if (length(files) > 1)
      #   if (exists("aaaa", envir = .GlobalEnv)) browser()
    }
  }
  if (length(files)) {
    hasStartingDot <- startsWith(files, ".")
    if (isTRUE(any(hasStartingDot)))
      files[hasStartingDot] <- fs::path_norm(files[hasStartingDot])
  }
  if (isList) {
    if (length(nams) == length(files)) {
      names(files) <- nams
    }
  }
  files
}

#' An alternative to `basename` and `dirname` when there are sub-folders
#'
#' This confirms that the `files` which may be absolute actually
#'   exist when compared `makeRelative(knownRelativeFiles, absolutePrefix)`.
#'   This is different than just using `basename` because it will include any
#'   sub-folder structure within the `knownRelativePaths`
#'
#' @param files A character vector of files to check to see if they are the same
#'   as `knownRelativeFiles`, once the `absolutePrefix` is removed
#'
#' @param absolutePrefix A directory to "remove" from `files` to compare
#'   to `knownRelativeFiles`
#'
#' @param knownRelativeFiles A character vector of relative filenames, that could
#'   have sub-folder structure.
#'
#' @inheritParams prepInputs
#'
checkRelative <- function(files, absolutePrefix, knownRelativeFiles,
                          verbose = getOption("reproducible.verbose")) {
  if (!is.null(knownRelativeFiles)) {
    neededFilesRel <- makeRelative(files, absolutePrefix)
    areAbs <- isAbsolutePath(knownRelativeFiles)
    if (any(areAbs)) {
      knownRelativeFiles[areAbs] <- makeRelative(knownRelativeFiles, absolutePrefix)
    }
    relativeNamesCorrect <- neededFilesRel %in% knownRelativeFiles

    if (!all(relativeNamesCorrect)) { # means user has asked for incorrect relative path

      #  must include directory names
      knownRelativeFiles <- unique(c(knownRelativeFiles, dirname(knownRelativeFiles)))
      knownRelativeFilesBase <- basename2(knownRelativeFiles)
      basenamesCorrect <- basename2(neededFilesRel) %in% knownRelativeFilesBase
      needUpdateRelNames <- basenamesCorrect & !relativeNamesCorrect # filename is correct, but not nesting structure
      if (any(needUpdateRelNames)) { # means basenames
        # needUpdateFromArchive <- !knownRelativeFilesBase %in% basename2(neededFilesRel)
        needUpdateFromArchive <- match(basename2(neededFilesRel)[needUpdateRelNames], knownRelativeFilesBase)
        files[needUpdateRelNames] <- makeAbsolute(knownRelativeFiles[needUpdateFromArchive], absolutePrefix)
        files <- unique(files)
        messagePrepInputs("User supplied files don't correctly specify the ",
                          "files in the archive (likely because of sub-folders); \n",
                          "using items in archive with same basenames. Renaming to these: \n",
                          paste(makeRelative(files[needUpdateRelNames], absoluteBase = absolutePrefix), collapse = "\n"),
                          verbose = verbose
        )
      }
    }
  }
  files
}

#' Make a temporary (sub-)directory
#'
#' Create a temporary subdirectory in `getOption("reproducible.tempPath")`.
#'
#' @param sub Character string, length 1. Can be a result of
#'   `file.path("smth", "smth2")` for nested temporary subdirectories.
#'   If the zero length character, then a random sub-directory will be created.
#' @param tempdir Optional character string where the temporary
#'   directory should be placed. Defaults to `getOption("reproducible.tempPath")`.
#' @param create Logical. Should the directory be created. Default `TRUE`.
#' @return
#' A character string of a path (that will be created if `create = TRUE`) in a
#' sub-directory of the `tempdir()`.
#'
#' @seealso [tempfile2]
#' @export
tempdir2 <- function(sub = "",
                     tempdir = getOption("reproducible.tempPath", .reproducibleTempPath()),
                     create = TRUE) {
  if (!nzchar(sub)) {
    sub <- rndstr(1)
  }
  np <- normPath(file.path(tempdir, sub))
  if (isTRUE(create)) {
    checkPath(np, create = TRUE)
  }
  np
}

#' Make a temporary file in a temporary (sub-)directory
#'
#' @param ... passed to `tempfile`, e.g., `fileext`
#'
#' @seealso [tempdir2]
#' @inheritParams tempdir2
#' @param ... passed to `tempfile`, e.g., `fileext`
#' @return
#' A character string of a path to a file in a
#' sub-directory of the `tempdir()`. This file will likely not exist yet.
#' @export
tempfile2 <- function(sub = "",
                      tempdir = getOption("reproducible.tempPath", .reproducibleTempPath()),
                      ...) {
  normPath(file.path(tempdir2(sub = sub, tempdir = tempdir), basename(tempfile(...))))
}

test_that("checkPath: normPath and normPathRel consistency", {
  # Use the following here instead of above because it fails on Mac without this.
  testInit()
  # tmpdir <- tempdir2("test_normPath")
  # tmpdir <- normalizePath(tmpdir, winslash = "/", mustWork = FALSE)

  withr::local_dir(tmpdir)
  # cwd <- getwd()
  # setwd(tmpdir)

  # don't use checkPath here because we are testing normPath!

  paths <- list(
    "./aaa/zzz",
    "./aaa/zzz/",
    ".//aaa//zzz",
    ".//aaa//zzz/",
    ".\\aaa\\zzz",
    ".\\aaa\\zzz\\",
    paste0(tmpdir, "/aaa/zzz"), # nolint
    paste0(tmpdir, "/aaa/zzz/"), # nolint
    file.path(tmpdir, "aaa", "zzz")
  )

  checked <- normPath(paths)
  expect_equal(length(unique(checked)), 1)

  checkedRel <- normPathRel(paths)
  if (identical(.Platform$OS.type, "windows")) {
    ## Windows create absolute paths
    expect_equal(length(unique(checkedRel)), 2)
  } else {
    ## non-existent paths kept relative on other platforms
    expect_equal(length(unique(checkedRel)), 2)
  }

  # These don't exist ... added May 10, 2023 after discovering that *nix-alikes don't make
  #   absolute paths with normalizePath when file or dir doesn't exist
  # Also empty path should return "", not be normalized
  pathsToCheck <- c("", "hi", "~/Hi", "/home/emcintir/testing", NA, "c:/Eliot")
  out <- normPath(pathsToCheck)
  expect_identical(sum(isAbsolutePath(out)), 4L)

  out <- isAbsolutePath(pathsToCheck)
  expect_identical(c(FALSE, FALSE, TRUE, TRUE, FALSE, TRUE), unname(out))

  outRel <- normPathRel(pathsToCheck)
  if (!identical(.Platform$OS.type, "windows")) {
    ## Windows create absolute paths;
    ## non-existent paths kept relative on other platforms
    expect_identical(isAbsolutePath(outRel), isAbsolutePath(pathsToCheck))
  }

  # extra checks for missing/NA/NULL
  expect_equal(normPath(), character(0))
  expect_equal(normPathRel(), character(0))

  expect_true(all(is.na(normPath(list(NA, NA_character_)))))
  expect_true(all(is.na(normPathRel(list(NA, NA_character_)))))

  expect_equal(normPath(NULL), character(0))
  expect_equal(normPathRel(NULL), character(0))
})

test_that("checkPath: checkPath consistency", {
  testInit()
  withr::local_dir(tmpdir)

  dir.create("aaa/zzz", recursive = TRUE, showWarnings = FALSE)
  paths <- list(
    "./aaa/zzz",
    "./aaa/zzz/",
    ".//aaa//zzz",
    ".//aaa//zzz/",
    ".\\aaa\\zzz",
    ".\\aaa\\zzz\\",
    paste0(tmpdir, "/aaa/zzz"), # nolint
    paste0(tmpdir, "/aaa/zzz/"), # nolint
    file.path(tmpdir, "aaa", "zzz")
  )

  checked <- lapply(paths, checkPath, create = FALSE)
  expect_equal(length(unique(checked)), 1)
  # unlink(tmpdir, recursive = TRUE)

  # extra checks for missing/NA/NULL
  expect_error(checkPath(), "Invalid path: no path specified.")
  expect_error(checkPath(NULL), "Invalid path: cannot be NULL.")
  expect_error(checkPath(NA_character_), "Invalid path: cannot be NA.")

  # Case where it is an existing file
  f1 <- tempfile()
  expect_true(file.create(f1)) ## TRUE
  expect_true(file.exists(f1)) ## TRUE

  withr::local_options("reproducible.verbose" = 1)
  expect_message(checkPath(f1), "is an existing file")
})

test_that("all exported functions have examples", {
  testInit()
  fns <- ls("package:reproducible")
  omit <- which(fns == "cache") ## cache is deprecated, so omit it

  ## for debugging only:
  tmpDir <- if (grepl("VIC-", Sys.info()["nodename"])) {
    checkPath("~/tmp/reproducible-examples", create = TRUE)
  } else {
    checkPath(file.path(tempdir(), "reproducible-examples"), create = TRUE)
  }
  tmpExFile <- file.path(tmpDir, "test-examples-out.txt")

  if (grepl("VIC-", Sys.info()["nodename"])) {
    cat("#START##############\n", file = tmpExFile, append = FALSE)
    # cat(fns[-omit], sep = "\n", file = tmpExFile, append = TRUE)
    # cat("#END##############\n", file = tmpExFile, append = TRUE)
  }

  manDir <- if (dir.exists("../../man")) {
    "../../man" ## if called during devtools::check()
  } else if (dir.exists("./man")) {
    "./man" ## if called during devtools::test()
  } else {
    system.file("man", package = "reproducible")
  }
  exFiles <- normalizePath(list.files(manDir, full.names = TRUE, pattern = "[.]Rd$"))


  # use for loop as it keeps control at top level
  # owd <- getwd()
  # tmpdir <- file.path(tmpDir, "test_Examples") |> checkPath(create = TRUE)
  # setwd(tmpdir)
  # on.exit({
  #   unlink(tmpdir, recursive = TRUE)
  #   setwd(owd)}
  #   , add = TRUE)
  if (grepl("VIC-", Sys.info()["nodename"])) { # for debugging only
    cat(paste("All files exist: ", isTRUE(all(file.exists(exFiles))), "\n"), file = tmpExFile, append = TRUE)
  }

  for (file in exFiles) {
    if (grepl("VIC-", Sys.info()["nodename"])) { # for debugging only
      cat(paste(file, " -- ", "\n"), file = tmpExFile, append = TRUE)
    }
    # for debugging only
    #if (isTRUE(grepl("emcintir|achubaty", Sys.info()[["user"]])))
      # print(file)
    co3 <- capture.output(type = "output",
                          co2 <- capture_messages(
                            co <- capture.output(type = "message",
                                                 test_example(file)
                            ))
    )
  }
})

---
title: "Introduction to Cache"
author:
  - "Eliot J. B. McIntire"
date: '`r strftime(Sys.Date(), "%B %d %Y")`'
output:
  rmarkdown::html_vignette:
    fig_width: 7
    number_sections: yes
    self_contained: yes
    toc: yes
vignette: >
  %\VignetteIndexEntry{Introduction to Cache}
  %\VignetteDepends{data.table, knitr, reproducible, terra}
  %\VignetteKeyword{Cache}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---


```{r setup, include = FALSE}
SuggestedPkgsNeeded <- c("terra")
hasSuggests <- all(sapply(SuggestedPkgsNeeded, require, character.only = TRUE, quietly = TRUE))
useSuggests <- !(tolower(Sys.getenv("_R_CHECK_DEPENDS_ONLY_")) == "true")

knitr::opts_chunk$set(eval = hasSuggests && useSuggests)
```


# Reproducible workflow

As part of a reproducible workflow, caching of function calls, code chunks, and other elements of a project is a critical component.
The objective of a reproducible workflow is is likely that an entire work flow from raw data to publication, decision support, report writing, presentation building etc., could be built and be reproducible anywhere, on any computer, operating system, with any starting conditions, on demand.
The `reproducible::Cache` function is built to work with any R function. 

## Differences with other approaches

`Cache` users `DBI` as a backend, with key functions, `dbReadTable`, `dbRemoveTable`, `dbSendQuery`, `dbSendStatement`, `dbCreateTable` and `dbAppendTable`. These can all be accessed via `Cache`, `showCache`, `clearCache`, and `keepCache`. It is optimized for speed of transactions, using `digest::digest` on objects and files.
The main function is superficially similar to `archivist::cache`, which uses `digest::digest` in all cases to determine whether the arguments are identical in subsequent iterations.
It also but does *many* things that make standard caching with `digest::digest` don't work reliably between systems.
For these, the function `.robustDigest` is introduced to make caching transferable between systems.
This is relevant for file paths, environments, parallel clusters, functions (which are contained within an environment), and many others (e.g., see `?.robustDigest` for methods).
`Cache` also adds important elements like automated tagging and the option to retrieve disk-cached values via stashed objects in memory using `memoise::memoise`.
This means that running `Cache` 1, 2, and 3 times on the same function will get progressively faster.
This can be extremely useful for web apps built with, say `shiny`.

## Function-level caching

Any function can be cached by wrapping `Cache` around the function call, or by using base pipe `|>`:

This will be a slight change to a function call, such as:
`terra::project(raster, crs = terra::crs(newRaster))`
to
`Cache(terra::project(raster, crs = terra::crs(newRaster)))`
or with the pipe, which may be more convenient as it is easy to add and remove caching in the code base:
`terra::project(raster, crs = terra::crs(newRaster)) |> Cache()`

This is particularly useful for expensive operations.

```{r function-level, echo=TRUE}
library(reproducible)
library(data.table)

tmpDir <- file.path(tempfile(), "reproducible_examples", "Cache")
dir.create(tmpDir, recursive = TRUE)

# Source raster with a complete LCC definition
ras <- terra::rast(terra::ext(0, 300, 0, 300), vals = 1:9e4, res = 1)
terra::crs(ras) <- "+proj=lcc +lat_1=60 +lat_2=70 +lat_0=50 +lon_0=-100 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs"

# Target CRS in PROJ form (no EPSG lookup)
newCRS <- "+proj=longlat +datum=WGS84 +no_defs"

# Derive target extent from source extent (no registry lookup)
target_ext <- terra::project(terra::ext(ras), from = terra::crs(ras), to = newCRS)

# Build template with chosen resolution; assign CRS
tmplate <- terra::rast(target_ext, resolution = 0.00001)
terra::crs(tmplate) <- newCRS

# No Cache
system.time(map1 <- terra::project(ras, tmplate, method = "near"))

# Try with memoise for this example -- for many simple cases, memoising will not be faster
opts <- options("reproducible.useMemoise" = TRUE)
# With Cache -- a little slower the first time because saving to disk
system.time({
  suppressWarnings({
    map1 <- terra::project(ras, tmplate, method = "near") |> 
      Cache(cachePath = tmpDir)
  })
})

# faster the second time; improvement depends on size of object and time to run function
system.time({
  map2 <- terra::project(ras, tmplate, method = "near") |> 
    Cache(cachePath = tmpDir)
})

options(opts)

all.equal(map1, map2, check.attributes = FALSE) # TRUE
```

## Caching examples

### Basic use

```{r}
try(clearCache(tmpDir, ask = FALSE), silent = TRUE) # just to make sure it is clear

ranNumsA <- rnorm(10, 16) |> Cache(cachePath = tmpDir)

# All same
ranNumsB <- rnorm(10, 16) |> Cache(cachePath = tmpDir) # recovers cached copy
ranNumsD1 <- Cache(quote(rnorm(n = 10, 16))) |> Cache(cachePath = tmpDir) # recovers cached copy
ranNumsD2 <- Cache(rnorm(n = 10, 16)) |> Cache(cachePath = tmpDir) # recovers cached copy
# pipe
ranNumsD3 <- rnorm(n = 10, 16) |> Cache(cachePath = tmpDir) # recovers cached copy

# Any minor change makes it different
ranNumsE <- rnorm(10, 6) |> Cache(cachePath = tmpDir) # different
```

#### Example 1: Basic cache use with tags

```{r tags}
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- runif(4) |> Cache(cachePath = tmpDir, userTags = "objectName:b")

showCache(tmpDir, userTags = c("objectName"))
showCache(tmpDir, userTags = c("^a$")) # regular expression ... "a" exactly
showCache(tmpDir, userTags = c("runif")) # show only cached objects made during runif call

clearCache(tmpDir, userTags = c("runif"), ask = FALSE) # remove only cached objects made during runif call
showCache(tmpDir) # all

clearCache(tmpDir, ask = FALSE)
```

#### Example 2: using the "accessed" tag

```{r accessed-tag}
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- runif(4) |> Cache(cachePath = tmpDir, userTags = "objectName:b")

# access it again, from Cache
Sys.sleep(1)
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
wholeCache <- showCache(tmpDir)

# keep only items accessed "recently" (i.e., only objectName:a)
onlyRecentlyAccessed <- showCache(tmpDir, userTags = max(wholeCache[tagKey == "accessed"]$tagValue))

# inverse join with 2 data.tables ... using: a[!b]
# i.e., return all of wholeCache that was not recently accessed
#   Note: the two different ways to access -- old way with "artifact" will be deprecated
toRemove <- unique(wholeCache[!onlyRecentlyAccessed, on = "cacheId"], by = "cacheId")$cacheId
clearCache(tmpDir, toRemove, ask = FALSE) # remove ones not recently accessed
showCache(tmpDir) # still has more recently accessed
```

#### Example 3: using keepCache

`keepCache` does the same as previous example, but more simply.

```{r keepCache}
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- Cache(runif(4)) |> Cache(cachePath = tmpDir, userTags = "objectName:b")

# keep only those cached items from the last 24 hours
oneDay <- 60 * 60 * 24
keepCache(tmpDir, after = Sys.time() - oneDay, ask = FALSE)

# Keep all Cache items created with an rnorm() call
keepCache(tmpDir, userTags = "rnorm", ask = FALSE)
showCache(tmpDir)

# Remove all Cache items that happened within a rnorm() call
clearCache(tmpDir, userTags = "rnorm", ask = FALSE)
showCache(tmpDir) ## empty

# Also, can set a time before caching happens and remove based on this
#  --> a useful, simple way to control Cache
ranNumsA <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
startTime <- Sys.time()
Sys.sleep(1)
ranNumsB <- rnorm(5) |> Cache(cachePath = tmpDir, userTags = "objectName:b")
keepCache(tmpDir, after = startTime, ask = FALSE) # keep only those newer than startTime

clearCache(tmpDir, ask = FALSE)
```

#### Example 4: searching for multiple objects in the cache

```{r searching-within-cache}
# default userTags is "and" matching; for "or" matching use |
ranNumsA <- runif(4) |> Cache(cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- rnorm(4) |> Cache(cachePath = tmpDir, userTags = "objectName:b")

# show all objects (runif and rnorm in this case)
showCache(tmpDir)

# show objects that are both runif and rnorm
# (i.e., none in this case, because objecs are either or, not both)
showCache(tmpDir, userTags = c("runif", "rnorm")) ## empty

# show objects that are either runif or rnorm ("or" search)
showCache(tmpDir, userTags = "runif|rnorm")

# keep only objects that are either runif or rnorm ("or" search)
keepCache(tmpDir, userTags = "runif|rnorm", ask = FALSE)

clearCache(tmpDir, ask = FALSE)
```

#### Example 5: using caching to speed up rerunning expensive computations

```{r expensive-computations}
ras <- terra::rast(terra::ext(0, 5, 0, 5),
  res = 1,
  vals = sample(1:5, replace = TRUE, size = 25),
  crs = "+proj=lcc +lat_1=48 +lat_2=33 +lon_0=-100 +ellps=WGS84"
)

rasCRS <- terra::crs(ras)
# Build an explicit target raster (same CRS, coarser resolution).
# Avoids the `terra::project(x, char_crs, res = N)` shorthand which
# recurses internally and forwards an unrecognized `wopt` set on
# recent terra versions, producing
# `[write] unknown option(s): xscale,yscale`.
rasTarget <- terra::rast(terra::ext(ras), crs = rasCRS, resolution = 5)

# A slow operation, like GIS operation
notCached <- suppressWarnings(
  # project raster generates warnings when run non-interactively
  terra::project(ras, rasTarget)
)

cached <- suppressWarnings(
  # project raster generates warnings when run non-interactively
  # using quote works also
  terra::project(ras, rasTarget) |> Cache(cachePath = tmpDir)
)

# second time is much faster
reRun <- suppressWarnings(
  # project raster generates warnings when run non-interactively
  terra::project(ras, rasTarget) |> Cache(cachePath = tmpDir)
)

# recovered cached version is same as non-cached version
all.equal(notCached, reRun, check.attributes = FALSE) ## TRUE
```

### Nested Caching

Nested caching, which is when Caching of a function occurs inside an outer function, which is itself cached.
This is a critical element to working within a reproducible work flow.
It is not enough during development to cache flat code chunks, as there will be many levels of "slow" functions.
Ideally, at all points in a development cycle, it should be possible to get to any line of code starting from the very initial steps, running through everything up to that point, in less than a few seconds.
If the workflow can be kept very fast like this, then there is a guarantee that it will work at any point.

```{r nested}
##########################
## Nested Caching
# Make 2 functions
inner <- function(mean) {
  d <- 1
  rnorm(n = 3, mean = mean)
}
outer <- function(n) {
  inner(0.1) |> Cache(cachePath = tmpdir2)
}

# make 2 different cache paths
tmpdir1 <- file.path(tempfile(), "first")
tmpdir2 <- file.path(tempfile(), "second")

# Run the Cache ... notOlderThan propagates to all 3 Cache calls,
#   but cachePath is tmpdir1 in top level Cache and all nested
#   Cache calls, unless individually overridden ... here inner
#   uses tmpdir2 repository
outer(n = 2) |> Cache(cachePath = tmpdir1)

showCache(tmpdir1) # 2 function calls
showCache(tmpdir2) # 1 function call

# userTags get appended
# all items have the outer tag propagate, plus inner ones only have inner ones
clearCache(tmpdir1, ask = FALSE)
outerTag <- "outerTag"
innerTag <- "innerTag"
inner <- function(mean) {
  d <- 1
  rnorm(n = 3, mean = mean) |> Cache(notOlderThan = Sys.time() - 1e5, userTags = innerTag)
}
outer <- function(n) {
  inner(0.1) |> Cache()
}
aa <- Cache(outer, n = 2) |> Cache(cachePath = tmpdir1, userTags = outerTag)
showCache(tmpdir1) # rnorm function has outerTag and innerTag, inner and outer only have outerTag
```

### cacheId

Sometimes, it is not absolutely desirable to maintain the work flow intact because changes that are irrelevant to the analysis, such as changing messages sent to a user, may be changed, without a desire to rerun functions.
The `cacheId` argument is for this.
Once a piece of code is run, then the `cacheId` can be manually extracted (it is reported at the end of a Cache call) and manually placed in the code, passed in as, say, `cacheId = "ad184ce64541972b50afd8e7b75f821b"`.

```{r selective-cacheId}
### cacheId
set.seed(1)
rnorm(1) |> Cache(cachePath = tmpdir1)
# manually look at output attribute which shows cacheId: 422bae4ed2f770cc
rnorm(1) |> Cache(cachePath = tmpdir1, cacheId = "422bae4ed2f770cc") # same value
# override even with different inputs:
rnorm(2) |> Cache(cachePath = tmpdir1, cacheId = "422bae4ed2f770cc")
```

## Working with the Cache manually

Since the cache is simply a `DBI` data table (of an SQLite database by default).
In addition, there are several helpers in the `reproducible` package, including `showCache`, `keepCache` and `clearCache` that may be useful.
Also, one can access cached items manually (rather than simply rerunning the same `Cache` function again).

```{r manual-cache}
# As of reproducible version 1.0, there is a new backend directly using DBI
mapHash <- unique(showCache(tmpDir, userTags = "project")$cacheId)
map <- loadFromCache(mapHash[1], cachePath = tmpDir)
terra::plot(map)
```

```{r cleanup}
## cleanup
unlink(dirname(tmpDir), recursive = TRUE)
```

# Alternative database backends

By default, caching relies on a sqlite database for it's backend.
While this works in many situations, there are some important limitations of using sqlite for caching, including 1) speed; 2) concurrent transactions; 3) sharing database across machines or projects.
Fortunately, `Cache` makes use of `DBI` package and thus supports several database backends, including mysql and postgresql.

See <https://github.com/PredictiveEcology/SpaDES/wiki/Using-alternate-database-backends-for-Cache> for further information on configuring these additional backends.

# Reproducible workflow

In general, we feel that a liberal use of `Cache` will make a re-usable and reproducible work flow.
`shiny` apps can be made, taking advantage of `Cache`. 
Indeed, much of the difficulty in managing data sets and saving them for future use, can be accommodated by caching.

# C5: deleted local file + stale sidecar + file in destinationPathShared → relink, no download

    Code
      invisible(preProcess(url = fixSrc$url, targetFile = "foo.csv", destinationPath = dest, fun = NA, verbose = 1))
    Message
      Running `preProcess`
      Preparing: foo.csv
      Skipping download. All requested files already present
      ...using file(s) in getOption('reproducible.inputPaths')...
      Hardlinked version of file(s) created:
      ... no copy/copies made.

# C6: deleted local file + stale sidecar + nothing in shared → download proceeds

    Code
      invisible(preProcess(url = fixSrc$url, targetFile = "foo.csv", destinationPath = dest, fun = NA, verbose = 1))
    Message
      Running `preProcess`
      Preparing: foo.csv
      ...downloading...
      Downloading <file-url> ...
      Appending checksums to CHECKSUMS.txt. If you see this message repeatedly, you can specify targetFile (and optionally alsoExtract) so it knows what to look for.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GPT2.R
\name{convertCallToCommonFormat}
\alias{convertCallToCommonFormat}
\title{Convert all ways of calling a function into canonical form, including defaults}
\usage{
convertCallToCommonFormat(call, usesDots, isSquiggly, .callingEnv)
}
\arguments{
\item{call}{The full captured call as it was passed by user.}

\item{usesDots}{Logical. Whether the original \code{Cache} call used \code{...}}

\item{isSquiggly}{Logical. Whether there are curly braces e.g., as in a pipe sequence.}

\item{.callingEnv}{Environment. The environment from which \code{Cache} was called.}
}
\description{
e.g., stats::rnorm(1) --> rnorm(n = 1, mean = 0, sd = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache-helpers.R
\docType{class}
\name{Path-class}
\alias{Path-class}
\alias{asPath}
\alias{asPath.character}
\alias{asPath.null}
\title{Coerce a character string to a class "Path"}
\usage{
asPath(obj, nParentDirs = 0)

\method{asPath}{character}(obj, nParentDirs = 0)

\method{asPath}{null}(obj, nParentDirs = 0)
}
\arguments{
\item{obj}{A character string to convert to a \code{Path}.}

\item{nParentDirs}{A numeric indicating the number of parent directories starting
from basename(obj) = 0 to keep for the digest}
}
\value{
A vector of class \code{Path}, which is similar to a character, but
has an attribute indicating how deep the Path should be
considered "digestible". In other words, most of the time, only some
component of an absolute path is relevant for evaluating its purpose in
a Cache situation. In general, this is usually equivalent to just the "relative" path
}
\description{
Allows a user to specify that their character string is indeed a filepath.
Thus, methods that require only a filepath can be dispatched correctly.
}
\details{
It is often difficult or impossible to know algorithmically whether a
character string corresponds to a valid filepath.
In the case where it is en existing file, \code{file.exists} can work.
But if it does not yet exist, e.g., for a \code{save}, it is difficult to know
whether it is a valid path before attempting to save to the path.

This function can be used to remove any ambiguity about whether a character
string is a path. It is primarily useful for achieving repeatability with Caching.
Essentially, when Caching, arguments that are character strings should generally be
digested verbatim, i.e., it must be an exact copy for the Cache mechanism
to detect a candidate for recovery from the cache.
Paths, are different. While they are character strings, there are many ways to
write the same path. Examples of identical meaning, but different character strings are:
path expanding of \code{~} vs. not, double back slash vs. single forward slash,
relative path vs. absolute path.
All of these should be assessed for their actual file or directory location,
NOT their character string. By converting all character string that are actual
file or directory paths with this function, then \code{Cache} will correctly assess
the location, NOT the character string representation.
}
\examples{
tmpf <- tempfile(fileext = ".csv")
file.exists(tmpf) ## FALSE
tmpfPath <- asPath(tmpf)
is(tmpf, "Path") ## FALSE
is(tmpfPath, "Path") ## TRUE

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{normPath}
\alias{normPath}
\alias{normPath,character-method}
\alias{normPath,list-method}
\alias{normPath,NULL-method}
\alias{normPath,missing-method}
\alias{normPath,logical-method}
\alias{normPathRel}
\title{Normalize file paths}
\usage{
normPath(path)

\S4method{normPath}{character}(path)

\S4method{normPath}{list}(path)

\S4method{normPath}{NULL}(path)

\S4method{normPath}{missing}()

\S4method{normPath}{logical}(path)

normPathRel(path)
}
\arguments{
\item{path}{A character vector of filepaths.}
}
\value{
Character vector of cleaned up filepaths.
}
\description{
Checks the specified path for formatting consistencies:
\enumerate{
\item use slash instead of backslash;
\item do tilde etc. expansion;
\item remove trailing slash.
}
}
\details{
Additionally, \code{normPath()} attempts to create a absolute paths,
whereas \code{normPathRel()} maintains relative paths.

\if{html}{\out{<div class="sourceCode">}}\preformatted{d> getwd()
[1] "/home/achubaty/Documents/GitHub/PredictiveEcology/reproducible"
d> normPathRel("potato/chips")
[1] "potato/chips"
d> normPath("potato/chips")
[1] "/home/achubaty/Documents/GitHub/PredictiveEcology/reproducible/potato/chips"
}\if{html}{\out{</div>}}
}
\examples{
## normalize file paths
paths <- list("./aaa/zzz",
              "./aaa/zzz/",
              ".//aaa//zzz",
              ".//aaa//zzz/",
              ".\\\\\\\\aaa\\\\\\\\zzz",
              ".\\\\\\\\aaa\\\\\\\\zzz\\\\\\\\",
              file.path(".", "aaa", "zzz"))

checked <- normPath(paths)
length(unique(checked)) ## 1; all of the above are equivalent

## check to see if a path exists
tmpdir <- file.path(tempdir(), "example_checkPath")

dir.exists(tmpdir) ## FALSE
tryCatch(checkPath(tmpdir, create = FALSE), error = function(e) FALSE) ## FALSE

checkPath(tmpdir, create = TRUE)
dir.exists(tmpdir) ## TRUE

unlink(tmpdir, recursive = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{prepInputs}
\alias{prepInputs}
\title{Download and optionally post-process files}
\usage{
prepInputs(
  targetFile = NULL,
  url = NULL,
  archive = NULL,
  alsoExtract = NULL,
  destinationPath = getOption("reproducible.destinationPath", "."),
  fun = NULL,
  quick = getOption("reproducible.quick"),
  overwrite = getOption("reproducible.overwrite", FALSE),
  purge = FALSE,
  useCache = getOption("reproducible.useCache", 2),
  .tempPath,
  verbose = getOption("reproducible.verbose", 1),
  ...
)
}
\arguments{
\item{targetFile}{Character string giving the filename (without relative or
absolute path) to the eventual file
(raster, shapefile, csv, etc.) after downloading and extracting from a zip
or tar archive. This is the file \emph{before} it is passed to
\code{postProcess}. The internal checksumming does not checksum
the file after it is \code{postProcess}ed (e.g., cropped/reprojected/masked).
Using \code{Cache} around \code{prepInputs} will do a sufficient job in these cases.
See table in \code{\link[=preProcess]{preProcess()}}.}

\item{url}{Optional character string indicating the URL to download from.
If not specified, then no download will be attempted. If not entry
exists in the \code{CHECKSUMS.txt} (in \code{destinationPath}), an entry
will be created or appended to. This \code{CHECKSUMS.txt} entry will be used
in subsequent calls to
\code{prepInputs} or \code{preProcess}, comparing the file on hand with the ad hoc
\code{CHECKSUMS.txt}. See table in \code{\link[=preProcess]{preProcess()}}.}

\item{archive}{Optional character string giving the path of an archive
containing \code{targetFile}, or a vector giving a set of nested archives
(e.g., \code{c("xxx.tar", "inner.zip", "inner.rar")}). If there is/are (an) inner
archive(s), but they are unknown, the function will try all until it finds
the \code{targetFile}. See table in \code{\link[=preProcess]{preProcess()}}. If it is \code{NA},
then it will \emph{not} attempt to see it as an archive, even if it has archive-like
file extension (e.g., \code{.zip}). This may be useful when an R function
is expecting an archive directly.}

\item{alsoExtract}{Optional character string naming files other than
\code{targetFile} that must be extracted from the \code{archive}. If
\code{NULL}, the default, then it will extract all files. Other options:
\code{"similar"} will extract all files with the same filename without
file extension as \code{targetFile}. \code{NA} will extract nothing other
than \code{targetFile}. A character string of specific file names will cause
only those to be extracted. Each element may also be a regular expression:
if an element does not match any archive member literally (by relative path
or basename), it is passed to \code{grep()} against the archive's file list and
all matching members are extracted. For example,
\code{alsoExtract = "CMD_sm|CMD_sp"} extracts every file whose name contains
\code{CMD_sm} or \code{CMD_sp}. See table in \code{\link[=preProcess]{preProcess()}}.}

\item{destinationPath}{Character string of a directory in which to download
and save the file that comes from \code{url} and is also where the function
will look for \code{archive} or \code{targetFile}. NOTE (still experimental):
To prevent repeated downloads in different locations, the user can also set
\code{options("reproducible.destinationPathShared")} to one or more local file paths to
search for the file before attempting to download. Default for that option is
\code{NULL} meaning do not search locally. The previous name
\code{options("reproducible.inputPaths")} is still accepted as a backwards-compatible
alias.}

\item{fun}{Optional. If specified, this will attempt to load whatever
file was downloaded during \code{preProcess} via \code{dlFun}. This can be either a
function (e.g., sf::st_read), character string (e.g., "base::load"),
NA (for no loading, useful if \code{dlFun} already loaded the file) or
if extra arguments are required
in the function call, it must be a call naming
\code{targetFile} (e.g., \code{sf::st_read(targetFile, quiet = TRUE)})
as the file path to the file to load. See details and examples below.}

\item{quick}{Logical. This is passed internally to \code{\link[=Checksums]{Checksums()}}
(the quickCheck argument), and to
\code{\link[=Cache]{Cache()}} (the quick argument). This results in faster, though
less robust checking of inputs. See the respective functions.}

\item{overwrite}{Logical. Passed to \code{writeTo} (possibly inside \code{postProcess}) and \code{postProcess}.}

\item{purge}{Logical or Integer. \code{0/FALSE} (default) keeps existing \code{CHECKSUMS.txt} file and
\code{prepInputs} will write or append to it. \code{1/TRUE} will deleted the entire \code{CHECKSUMS.txt} file.
Other options, see details.}

\item{useCache}{Passed to \code{Cache} in various places.
Defaults to \code{getOption("reproducible.useCache", 2L)} in \code{prepInputs}, and
\code{getOption("reproducible.useCache", FALSE)} if calling any of the inner
functions manually. For \code{prepInputs}, this mean it will use \code{Cache}
only up to 2 nested levels, which includes \code{preProcess}. \code{postProcess} and
its nested \verb{*Input} functions (e.g., \code{cropInputs}, \code{projectInputs},
\code{maskInputs}) are no longer internally cached, as \code{terra} processing speeds
mean internal caching is more time consuming. We recommend caching the full
\code{prepInputs} call instead (e.g. \code{prepInputs(...) |> Cache()}).}

\item{.tempPath}{Optional temporary path for internal file intermediate steps.
Will be cleared \code{on.exit} from this function.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{Additional arguments passed to
\code{\link[=postProcess]{postProcess()}} and \code{\link[=Cache]{Cache()}}.
Since \code{...} is passed to \code{\link[=postProcess]{postProcess()}}, these will
\code{...} will also be passed into the inner
functions, e.g., \code{\link[=cropInputs]{cropInputs()}}. Possibly useful other arguments include
\code{dlFun} which is passed to \code{preProcess}. See details and examples.}
}
\value{
This is an omnibus function that will return an R object that will have resulted from
the running of \code{\link[=preProcess]{preProcess()}} and \code{\link[=postProcess]{postProcess()}} or \code{\link[=postProcessTo]{postProcessTo()}}. Thus,
if it is a GIS object, it may have been cropped, reprojected, "fixed", masked, and
written to disk.
}
\description{
\if{html}{\figure{lifecycle-maturing.svg}{options: alt="maturing"}}
}
\details{
This function can be used to prepare R objects from remote or local data sources.
The object of this function is to provide a reproducible version of
a series of commonly used steps for getting, loading, and processing data.
This function has two stages: Getting data (download, extracting from archives,
loading into R) and post-processing (for \verb{Spatial*} and \verb{Raster*}
objects, this is crop, reproject, mask/intersect).
To trigger the first stage, provide \code{url} or \code{archive}.
To trigger the second stage, provide \code{studyArea} or \code{rasterToMatch}.
See examples.
}
\note{
This function is still experimental: use with caution.
}
\section{Stage 1 - Getting data}{


See \code{\link[=preProcess]{preProcess()}} for combinations of arguments.

\enumerate{
\item Download from the web via either \code{googledrive::drive_download()},
\code{\link[utils:download.file]{utils::download.file()}};
\item Extract from archive using \code{\link[=unzip]{unzip()}} or \code{\link[=untar]{untar()}};
\item Load into R using \code{terra::rast},
\code{sf::st_read}, or any other function passed in with \code{fun};
\item Checksumming of all files during this process. This is put into a
\file{CHECKSUMS.txt} file in the \code{destinationPath}, appending if it is
already there, overwriting the entries for same files if entries already exist.
}
}

\section{Stage 2 - Post processing}{


This will be triggered if either \code{rasterToMatch} or \code{studyArea}
is supplied.

\enumerate{
\item Fix errors. Currently only errors fixed are for \code{SpatialPolygons}
using \code{buffer(..., width = 0)};
\item Crop using \code{\link[=cropTo]{cropTo()}};
\item Project using \code{\link[=projectTo]{projectTo()}};
\item Mask using \code{\link[=maskTo]{maskTo()}};
\item write the file to disk via \code{\link[=writeTo]{writeTo()}}.
}

NOTE: checksumming does not occur during the post-processing stage, as
there are no file downloads. To achieve fast results, wrap
\code{prepInputs} with \code{Cache}.

NOTE: \code{sf} objects are still very experimental.

\subsection{postProcessing of \verb{Spat*}, \code{sf}, \verb{Raster*} and \verb{Spatial*} objects:}{

The following has been DEPRECATED because there are a sufficient number of
ambiguities that this has been changed in favour of \code{from} and the \verb{*to} family.
See \code{\link[=postProcessTo]{postProcessTo()}}.

DEPRECATED: If \code{rasterToMatch} or \code{studyArea} are used, then this will
trigger several subsequent functions, specifically the sequence,
\emph{Crop, reproject, mask}, which appears to be a common sequence while
preparing spatial data from diverse sources.
See \code{\link[=postProcess]{postProcess()}} documentation section on
\emph{Backwards compatibility with \code{rasterToMatch} and/or \code{studyArea} arguments}
to understand various combinations of \code{rasterToMatch} and/or \code{studyArea}.
}
}

\section{\code{fun}}{


\code{fun} offers the ability to pass any custom function with which to load
the file obtained by \code{preProcess} into the session. There are two cases that are
dealt with: when the \code{preProcess} downloads a file (including via \code{dlFun}),
\code{fun} must deal with a file; and, when \code{preProcess} creates an R object
(e.g., raster::getData returns an object), \code{fun} must deal with an object.

\code{fun} can be supplied in three ways: a function, a character string
(i.e., a function name as a string), or an expression.
If a character string or function, is should have the package name e.g.,
\code{"terra::rast"} or as an actual function, e.g., \code{base::readRDS}.
In these cases, it will evaluate this function call while passing \code{targetFile}
as the first argument. These will only work in the simplest of cases.

When more precision is required, the full call can be written and where the
filename can be referred to as \code{targetFile} if the function
is loading a file. If \code{preProcess} returns an object, \code{fun} should be set to
\code{fun = NA}.

If there is a custom function call, is not in a package, \code{prepInputs} may not find it. In such
cases, simply pass the function as a named argument (with same name as function) to \code{prepInputs}.
See examples.
NOTE: passing \code{fun = NA} will skip loading object into R. Note this will essentially
replicate the functionality of simply calling \code{preProcess} directly.
}

\section{\code{purge}}{


In options for control of purging the \code{CHECKSUMS.txt} file are:

\describe{
\item{\code{0}}{keep file}
\item{\code{1}}{delete file in \code{destinationPath}, all records of downloads need to be rebuilt}
\item{\code{2}}{delete entry with same \code{targetFile}}
\item{\code{4}}{delete entry with same \code{alsoExtract}}
\item{\code{3}}{delete entry with same \code{archive}}
\item{\code{5}}{delete entry with same \code{targetFile} & \code{alsoExtract}}
\item{\code{6}}{delete entry with same \code{targetFile}, \code{alsoExtract} & \code{archive}}
\item{\code{7}}{delete entry that same \code{targetFile}, \code{alsoExtract} & \code{archive} & \code{url}}
}
will only remove entries in the \code{CHECKSUMS.txt} that are associated with
\code{targetFile}, \code{alsoExtract} or \code{archive} When \code{prepInputs} is called,
it will write or append to a (if already exists) \code{CHECKSUMS.txt} file.
If the \code{CHECKSUMS.txt} is not correct, use this argument to remove it.
}

\examples{
\donttest{
if (requireNamespace("terra", quietly = TRUE) &&
  requireNamespace("withr", quietly = TRUE)) {
  library(reproducible)
  withr::local_dir(withr::local_tempdir())
  # Make a dummy study area map -- user would supply this normally
  coords <- structure(c(-122.9, -116.1, -99.2, -106, -122.9, 59.9, 65.7, 63.6, 54.8, 59.9),
    .Dim = c(5L, 2L)
  )
  studyArea <- terra::vect(coords, "polygons")
  terra::crs(studyArea) <- "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84 +towgs84=0,0,0"
  # Make dummy "large" map that must be cropped to the study area
  outerSA <- terra::buffer(studyArea, 50000)
  terra::crs(outerSA) <- "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84 +towgs84=0,0,0"
  tf <- normPath(file.path(tempdir2(), "prepInputs2.shp"))
  terra::writeVector(outerSA, tf)

  # run prepInputs -- load file, postProcess it to the studyArea

  studyArea2 <- prepInputs(
    targetFile = tf, to = studyArea,
    fun = "terra::vect",
    destinationPath = tempdir2()
  ) |>
    suppressWarnings() # not relevant warning here

  # clean up
  unlink("CHECKSUMS.txt")

  ##########################################
  # Remote file using `url`
  ##########################################
  if (internetExists()) {
    data.table::setDTthreads(2)
    origDir <- getwd()
    # download a zip file from internet, unzip all files, load as shapefile, Cache the call
    # First time: don't know all files - prepInputs will guess, if download file is an archive,
    #   then extract all files, then if there is a .shp, it will load with sf::st_read
    dPath <- file.path(tempdir(), "ecozones")
    shpUrl <- "http://sis.agr.gc.ca/cansis/nsdb/ecostrat/zone/ecozone_shp.zip"

    # Wrapped in a try because this particular url can be flaky
    shpEcozone <- try(prepInputs(
      destinationPath = dPath,
      url = shpUrl
    ))
    if (!is(shpEcozone, "try-error")) {
      # Robust to partial file deletions:
      unlink(dir(dPath, full.names = TRUE)[1:3])
      shpEcozone <- prepInputs(
        destinationPath = dPath,
        url = shpUrl
      )
      unlink(dPath, recursive = TRUE)

      # Once this is done, can be more precise in operational code:
      #  specify targetFile, alsoExtract, and fun, wrap with Cache
      ecozoneFilename <- file.path(dPath, "ecozones.shp")
      ecozoneFiles <- c(
        "ecozones.dbf", "ecozones.prj",
        "ecozones.sbn", "ecozones.sbx", "ecozones.shp", "ecozones.shx"
      )
      shpEcozone <- prepInputs(
        targetFile = ecozoneFilename,
        url = shpUrl,
        fun = "terra::vect",
        alsoExtract = ecozoneFiles,
        destinationPath = dPath
      )
      unlink(dPath, recursive = TRUE)

      # Add a study area to Crop and Mask to
      # Create a "study area"
      coords <- structure(c(-122.98, -116.1, -99.2, -106, -122.98, 59.9, 65.73, 63.58, 54.79, 59.9),
        .Dim = c(5L, 2L)
      )
      studyArea <- terra::vect(coords, "polygons")
      terra::crs(studyArea) <- "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84 +towgs84=0,0,0"

      #  specify targetFile, alsoExtract, and fun, wrap with Cache
      ecozoneFilename <- file.path(dPath, "ecozones.shp")
      # Note, you don't need to "alsoExtract" the archive... if the archive is not there, but the
      #   targetFile is there, it will not redownload the archive.
      ecozoneFiles <- c(
        "ecozones.dbf", "ecozones.prj",
        "ecozones.sbn", "ecozones.sbx", "ecozones.shp", "ecozones.shx"
      )
      shpEcozoneSm <- Cache(prepInputs,
        url = shpUrl,
        targetFile = reproducible::asPath(ecozoneFilename),
        alsoExtract = reproducible::asPath(ecozoneFiles),
        studyArea = studyArea,
        fun = "terra::vect",
        destinationPath = dPath,
        writeTo = "EcozoneFile.shp"
      ) # passed to determineFilename

      terra::plot(shpEcozone[, 1])
      terra::plot(shpEcozoneSm[, 1], add = TRUE, col = "red")
      unlink(dPath)
    }
  }
  withr::deferred_run()
}
}

## Using quoted dlFun and fun -- this is not intended to be run but used as a template
## prepInputs(..., fun = customFun(x = targetFile), customFun = customFun)
##   # or more complex
##  test5 <- prepInputs(
##   targetFile = targetFileLuxRDS,
##   dlFun =
##     getDataFn(name = "GADM", country = "LUX", level = 0) # preProcess keeps file from this!
##   ,
##   fun = {
##     out <- readRDS(targetFile)
##     sf::st_as_sf(out)}
##  )
}
\seealso{
\code{\link[=postProcessTo]{postProcessTo()}}, \code{\link[=downloadFile]{downloadFile()}}, \code{\link[=extractFromArchive]{extractFromArchive()}},
\code{\link[=postProcess]{postProcess()}}.
}
\author{
Eliot McIntire, Jean Marchal, and Tati Micheletti
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcess.R
\name{postProcess}
\alias{postProcess}
\alias{postProcess.list}
\alias{postProcess.default}
\title{Generic function to post process objects}
\usage{
postProcess(x, ...)

\method{postProcess}{list}(x, ...)

\method{postProcess}{default}(x, ...)
}
\arguments{
\item{x}{A GIS object of postProcessing,
e.g., Spat* or sf*. This can be provided as a
\code{rlang::quosure} or a normal R object.}

\item{...}{Additional arguments passed to methods. For \code{spatialClasses},
these are: \code{\link[=cropTo]{cropTo()}}, \code{\link[=fixErrorsIn]{fixErrorsIn()}},
\code{\link[=projectTo]{projectTo()}}, \code{\link[=maskTo]{maskTo()}},
\code{\link[=determineFilename]{determineFilename()}}, and \code{\link[=writeTo]{writeTo()}}.
Each of these may also pass \code{...} into other functions, like
\code{\link[=writeTo]{writeTo()}}.
This might include potentially important arguments like \code{datatype},
\code{format}. Also passed to \code{terra::project},
with likely important arguments such as \code{method = "bilinear"}.
See details.}
}
\value{
A GIS file (e.g., \code{RasterLayer}, \code{SpatRaster} etc.) that has been
appropriately cropped, reprojected, masked, depending on the inputs.
}
\description{
\if{html}{\figure{lifecycle-maturing.svg}{options: alt="maturing"}}

The method for GIS objects (terra \verb{Spat*} & sf classes) will
crop, reproject, and mask, in that order.
This is a wrapper for \code{\link[=cropTo]{cropTo()}}, \code{\link[=fixErrorsIn]{fixErrorsIn()}},
\code{\link[=projectTo]{projectTo()}}, \code{\link[=maskTo]{maskTo()}} and \code{\link[=writeTo]{writeTo()}},
with a required amount of data manipulation between these calls so that the crs match.
}
\section{Post processing sequence}{


If the \code{rasterToMatch} or \code{studyArea} are passed, then
the following sequence will occur:

\enumerate{
\item Fix errors \code{\link[=fixErrorsIn]{fixErrorsIn()}}. Currently only errors fixed are for
\code{SpatialPolygons} using \code{buffer(..., width = 0)}.
\item Crop using \code{\link[=cropTo]{cropTo()}}
\item Project using \code{\link[=projectTo]{projectTo()}}
\item Mask using \code{\link[=maskTo]{maskTo()}}
\item Determine file name \code{\link[=determineFilename]{determineFilename()}}
\item Write that file name to disk, optionally \code{\link[=writeTo]{writeTo()}}
}

NOTE: checksumming does not occur during the post-processing stage, as
there are no file downloads. To achieve fast results, wrap
\code{prepInputs} with \code{Cache}
}

\section{Backwards compatibility with \code{rasterToMatch} and/or \code{studyArea} arguments}{


For backwards compatibility, \code{postProcess} will continue to allow passing
\code{rasterToMatch} and/or \code{studyArea} arguments. Depending on which of these
are passed, different things will happen to the \code{targetFile} located at \code{filename1}.

See \emph{Use cases} section in \code{\link[=postProcessTo]{postProcessTo()}} for post processing behaviour with
the new \code{from} and \code{to} arguments.

\subsection{If \code{targetFile} is a raster (\verb{Raster*}, or \code{SpatRaster}) object:}{
\tabular{lccc}{
\tab \code{rasterToMatch}      \tab \code{studyArea}  \tab  Both               \cr
\code{extent}     \tab Yes                  \tab   Yes        \tab \code{rasterToMatch}     \cr
\code{resolution} \tab Yes                  \tab   No         \tab \code{rasterToMatch}     \cr
\code{projection} \tab Yes                  \tab   No*        \tab \code{rasterToMatch}*    \cr
\code{alignment}  \tab Yes                  \tab   No         \tab \code{rasterToMatch}     \cr
\code{mask}       \tab No**                 \tab   Yes        \tab \code{studyArea}**       \cr
}
*Can be overridden with \code{useSAcrs}.

**Will mask with \code{NA}s from \code{rasterToMatch} if \code{maskWithRTM}.
}

\subsection{If \code{targetFile} is a vector (\verb{Spatial*}, \code{sf} or \code{SpatVector}) object:}{
\tabular{lccc}{
\tab \code{rasterToMatch}      \tab \code{studyArea}  \tab Both               \cr
\code{extent}     \tab Yes                  \tab   Yes        \tab \code{rasterToMatch}    \cr
\code{resolution} \tab NA                   \tab   NA         \tab NA                 \cr
\code{projection} \tab Yes                  \tab   No*        \tab \code{rasterToMatch}*   \cr
\code{alignment}  \tab NA                   \tab   NA         \tab NA                 \cr
\code{mask}       \tab No                   \tab   Yes        \tab \code{studyArea}        \cr
}
*Can be overridden with \code{useSAcrs}
}
}

\examples{
if (requireNamespace("terra", quietly = TRUE) &&
    requireNamespace("withr", quietly = TRUE)) {
  library(reproducible)
  withr::local_dir(withr::local_tempdir())
  withr::local_options(reproducible.inputPaths = NULL)
  # od <- setwd(tempdir2())
  # download a (spatial) file from remote url (which often is an archive) load into R
  # need 3 files for this example; 1 from remote, 2 local
  dPath <- file.path(tempdir2())
  remoteTifUrl <- "https://github.com/rspatial/terra/raw/master/inst/ex/elev.tif"

  localFileLuxSm <- system.file("ex/luxSmall.shp", package = "reproducible")
  localFileLux <- system.file("ex/lux.shp", package = "terra")

  # 1 step for each layer
  # 1st step -- get study area
  studyArea <- prepInputs(localFileLuxSm, fun = "terra::vect") # default is sf::st_read

  \donttest{
    # 2nd step: make the input data layer like the studyArea map
    # Requires internet -- so using try just in case, and kept out of the
    # timed examples (CRAN does not run \donttest)
    elevForStudy <- try(prepInputs(url = remoteTifUrl, to = studyArea, res = 250,
                                   destinationPath = dPath, useCache = FALSE))

    # Alternate way, one step at a time. Must know each of these steps, and perform for each layer
    dir.create(dPath, recursive = TRUE, showWarnings = FALSE)
    file.copy(localFileLuxSm, file.path(dPath, basename(localFileLuxSm)))
    studyArea2 <- terra::vect(localFileLuxSm)
    if (!all(terra::is.valid(studyArea2))) studyArea2 <- terra::makeValid(studyArea2)
    tf <- tempfile(fileext = ".tif")
    download.file(url = remoteTifUrl, destfile = tf, mode = "wb", quiet = TRUE)
    Checksums(dPath, write = TRUE, files = tf)
    elevOrig <- terra::rast(tf)
    studyAreaCrs <- terra::crs(studyArea)
    # Build an explicit target raster (same CRS as studyArea, res = 250) and
    # project to it. Avoids the recursive `terra::project(x, char_crs, res = N)`
    # shorthand that recent terra (~1.9) regressed with
    # `[write] unknown option(s): xscale,yscale`.
    elevTarget <- terra::project(terra::rast(elevOrig), studyAreaCrs)
    elevTarget <- terra::rast(terra::ext(elevTarget),
                              crs = terra::crs(elevTarget),
                              resolution = 250)
    elevForStudy2 <- terra::project(elevOrig, elevTarget) |>
      terra::mask(studyArea2) |>
      terra::crop(studyArea2)

    isTRUE(all.equal(elevForStudy, elevForStudy2)) # TRUE!
  }

  # sf class
  if (requireNamespace("sf", quietly = TRUE)) {
    studyAreaSmall <- prepInputs(localFileLuxSm, fun = "sf::st_read")
    studyAreas <- list()
    studyAreas[["orig"]] <- prepInputs(localFileLux)
    studyAreas[["reprojected"]] <- projectTo(studyAreas[["orig"]], studyAreaSmall)
    studyAreas[["cropped"]] <- suppressWarnings(cropTo(studyAreas[["orig"]], studyAreaSmall))
    studyAreas[["masked"]] <- suppressWarnings(maskTo(studyAreas[["orig"]], studyAreaSmall))
  }

  # SpatVector-- note: doesn't matter what class the "to" object is, only the "from"
  studyAreas <- list()
  studyAreaSmall <- prepInputs(localFileLuxSm)
  studyAreas[["orig"]] <- prepInputs(localFileLux)
  studyAreas[["reprojected"]] <- projectTo(studyAreas[["orig"]], studyAreaSmall)
  studyAreas[["cropped"]] <- suppressWarnings(cropTo(studyAreas[["orig"]], studyAreaSmall))
  studyAreas[["masked"]] <- suppressWarnings(maskTo(studyAreas[["orig"]], studyAreaSmall))
  if (interactive()) {
    par(mfrow = c(2,2));
    out <- lapply(studyAreas, function(x) terra::plot(x))
  }

  withr::deferred_run()
  # setwd(od)
}
}
\seealso{
\code{prepInputs}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcess.R
\name{assessDataType}
\alias{assessDataType}
\alias{assessDataType.default}
\title{Assess the appropriate raster layer data type}
\usage{
assessDataType(ras, type = "writeRaster")

\method{assessDataType}{default}(ras, type = "writeRaster")
}
\arguments{
\item{ras}{The \code{RasterLayer} or \code{RasterStack} for which data type will be assessed.}

\item{type}{Character. \code{"writeRaster"} (default) or \code{"GDAL"} (defunct)
to return the recommended
data type for writing from the raster packages, respectively, or
\code{"projectRaster"} to return recommended resampling type.}
}
\value{
A character string indicating the data type of the spatial layer
(e.g., "INT2U"). See \code{terra::datatype()}
}
\description{
When writing raster-type objects to disk, a \code{datatype} can be specified. These
functions help identify what smallest \code{datatype} can be used.
}
\examples{
if (requireNamespace("terra", quietly = TRUE)) {
  ## LOG1S
  rasOrig <- terra::rast(ncols = 10, nrows = 10)
  ras <- rasOrig
  ras[] <- rep(c(0,1),50)
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- rep(c(0,1),50)
  assessDataType(ras)

  ras[] <- rep(c(TRUE,FALSE),50)
  assessDataType(ras)

  ras[] <- c(NA, NA, rep(c(0,1),49))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- c(0, NaN, rep(c(0,1),49))
  assessDataType(ras)


  ## INT1S
  ras[] <- -1:98
  assessDataType(ras)

  ras[] <- c(NA, -1:97)
  assessDataType(ras)

  ## INT1U
  ras <- rasOrig
  ras[] <- 1:100
  assessDataType(ras)

  ras[] <- c(NA, 2:100)
  assessDataType(ras)

  ## INT2U
  ras <- rasOrig
  ras[] <- round(runif(100, min = 64000, max = 65000))
  assessDataType(ras)

  ## INT2S
  ras <- rasOrig
  ras[] <- round(runif(100, min = -32767, max = 32767))
  assessDataType(ras)

  ras[54] <- NA
  assessDataType(ras)

  ## INT4U
  ras <- rasOrig
  ras[] <- round(runif(100, min = 0, max = 500000000))
  assessDataType(ras)

  ras[14] <- NA
  assessDataType(ras)

  ## INT4S
  ras <- rasOrig
  ras[] <- round(runif(100, min = -200000000, max = 200000000))
  assessDataType(ras)

  ras[14] <- NA
  assessDataType(ras)

  ## FLT4S
  ras <- rasOrig
  ras[] <- runif(100, min = -10, max = 87)
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = -3.4e+26, max = 3.4e+28))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = 3.4e+26, max = 3.4e+28))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = -3.4e+26, max = -1))
  assessDataType(ras)

  ## FLT8S
  ras <- rasOrig
  ras[] <- c(-Inf, 1, rep(c(0,1),49))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- c(Inf, 1, rep(c(0,1),49))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = -1.7e+30, max = 1.7e+308))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = 1.7e+30, max = 1.7e+308))
  assessDataType(ras)

  ras <- rasOrig
  ras[] <- round(runif(100, min = -1.7e+308, max = -1))
  assessDataType(ras)

  # 2 layer with different types LOG1S and FLT8S
  ras <- rasOrig
  ras[] <- rep(c(0,1),50)
  ras1 <- rasOrig
  ras1[] <- round(runif(100, min = -1.7e+308, max = -1))
  sta <- c(ras, ras1)
  assessDataType(sta)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{tempfile2}
\alias{tempfile2}
\title{Make a temporary file in a temporary (sub-)directory}
\usage{
tempfile2(
  sub = "",
  tempdir = getOption("reproducible.tempPath", .reproducibleTempPath()),
  ...
)
}
\arguments{
\item{sub}{Character string, length 1. Can be a result of
\code{file.path("smth", "smth2")} for nested temporary subdirectories.
If the zero length character, then a random sub-directory will be created.}

\item{tempdir}{Optional character string where the temporary
directory should be placed. Defaults to \code{getOption("reproducible.tempPath")}.}

\item{...}{passed to \code{tempfile}, e.g., \code{fileext}}
}
\value{
A character string of a path to a file in a
sub-directory of the \code{tempdir()}. This file will likely not exist yet.
}
\description{
Make a temporary file in a temporary (sub-)directory
}
\seealso{
\link{tempdir2}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/listNamed.R
\name{listNamed}
\alias{listNamed}
\title{Create a list with names from object names}
\usage{
listNamed(...)
}
\arguments{
\item{...}{Any elements to add to a list, as in \code{base::list}}
}
\description{
This is a convenience wrapper around \verb{a <- 1; newList <- list(a); names(newList) <- "a"}.
}
\details{
This will return a named list, where names are the object names, captured
internally in the function and assigned to the list. If a user manually supplies
names, these will be kept (i.e., not overwritten by the object name).
}
\examples{
a <- 1
b <- 2
d <- 3
(newList <- listNamed(a, b, dManual = d)) # "dManual" name kept


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{retry}
\alias{retry}
\title{A wrapper around \code{try} that retries on failure}
\usage{
retry(
  expr,
  envir = parent.frame(),
  retries = 5,
  exponentialDecayBase = 1.3,
  silent = TRUE,
  exprBetween = NULL,
  messageFn = message
)
}
\arguments{
\item{expr}{An expression to run, i.e., \code{rnorm(1)}, similar to what is passed to \code{try}}

\item{envir}{The environment in which to evaluate the quoted expression, default
to \code{parent.frame(1)}}

\item{retries}{Numeric. The maximum number of retries.}

\item{exponentialDecayBase}{Numeric > 1.0. The delay between
successive retries will be \code{runif(1, min = 0, max = exponentialDecayBase ^ i - 1)}
where \code{i} is the retry number (i.e., follows \code{seq_len(retries)})}

\item{silent}{Logical indicating whether to \code{try} silently.}

\item{exprBetween}{Another expression that should be run after a failed attempt
of the \code{expr}. This should return a named list, where the names indicate the object names
to update in the main expr, and the return value is the new value. (previous versions allowed
a non-list return, but where the final line had to be an assignment operator,
specifying what object (that is used in \code{expr}) will be updated prior to running
the \code{expr} again. For backwards compatibility, this still works).}

\item{messageFn}{A function for messaging to console. Defaults to \code{message}}
}
\value{
As with \code{try}, so the successfully returned \code{return()} from the \code{expr} or a \code{try-error}.
}
\description{
This is useful for functions that are "flaky", such as \code{curl}, which may fail for unknown
reasons that do not persist.
}
\details{
Based on \url{https://github.com/jennybc/googlesheets/issues/219#issuecomment-195218525}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{.whereInStack}
\alias{.whereInStack}
\title{Search for objects in the call stack}
\usage{
.whereInStack(obj, startingEnv = parent.frame())
}
\arguments{
\item{obj}{Character string. The object name to search.}

\item{startingEnv}{An environment to start searching in.}
}
\value{
The environment in which the object exists. It will return the
first environment it finds, searching outwards from where the function is used.
}
\description{
Normally, this is only used in special, advanced uses. The standard approach
to getting an object from an environment in the call stack is to explicitly
pass it into the function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportedMethods.R
\name{.grepSysCalls}
\alias{.grepSysCalls}
\title{Grep system calls}
\usage{
.grepSysCalls(sysCalls, pattern)
}
\arguments{
\item{sysCalls}{The return from \code{sys.calls()}}

\item{pattern}{Character, passed to grep}
}
\value{
Numeric vector, equivalent to return from \code{grep(sys.calls(), pattern = "test")},
but faster if \code{sys.calls()} is very big.
}
\description{
A faster way of grepping the system call stack than just
\code{grep(sys.calls(), pattern = "test")}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{checkRelative}
\alias{checkRelative}
\title{An alternative to \code{basename} and \code{dirname} when there are sub-folders}
\usage{
checkRelative(
  files,
  absolutePrefix,
  knownRelativeFiles,
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{files}{A character vector of files to check to see if they are the same
as \code{knownRelativeFiles}, once the \code{absolutePrefix} is removed}

\item{absolutePrefix}{A directory to "remove" from \code{files} to compare
to \code{knownRelativeFiles}}

\item{knownRelativeFiles}{A character vector of relative filenames, that could
have sub-folder structure.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\description{
This confirms that the \code{files} which may be absolute actually
exist when compared \code{makeRelative(knownRelativeFiles, absolutePrefix)}.
This is different than just using \code{basename} because it will include any
sub-folder structure within the \code{knownRelativePaths}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convertPaths.R
\name{Filenames}
\alias{Filenames}
\alias{Filenames,ANY-method}
\alias{Filenames,environment-method}
\alias{Filenames,list-method}
\alias{Filenames,data.table-method}
\alias{Filenames,Path-method}
\title{Return the filename(s) from a \verb{Raster*} object}
\usage{
Filenames(obj, allowMultiple = TRUE, returnList = FALSE)

\S4method{Filenames}{ANY}(obj, allowMultiple = TRUE, returnList = FALSE)

\S4method{Filenames}{environment}(obj, allowMultiple = TRUE, returnList = FALSE)

\S4method{Filenames}{list}(obj, allowMultiple = TRUE, returnList = FALSE)

\S4method{Filenames}{data.table}(obj, allowMultiple = TRUE, returnList = FALSE)

\S4method{Filenames}{Path}(obj, allowMultiple = TRUE, returnList = FALSE)
}
\arguments{
\item{obj}{A \verb{Raster*} object (i.e., \code{RasterLayer}, \code{RasterStack}, \code{RasterBrick})}

\item{allowMultiple}{Logical. If \code{TRUE}, the default, then all relevant
filenames will be returned, i.e., in cases such as \code{.grd} where multiple files
are required. If \code{FALSE}, then only the first file will be returned,
e.g., \code{filename.grd}, in the case of default Raster format in R.}

\item{returnList}{Default \code{FALSE}. If \code{FALSE}, then return format will be a
character vector. When \code{TRUE}, list or environment objects will return a list
of character strings or vectors. When returned as a character vector, then
the names of objects with >1 filename associated with them will be given a numeric
suffix, which means the name in the returned vector does not match the object in
the list or environment. When returned as a list, their names are preserved.}
}
\value{
A character vector of filenames that are part of the objects passed to \code{obj}.
This returns \code{NULL} is the object is not file-backed or does not have a method
to recover the file-backed filename.
}
\description{
This is mostly just a wrapper around \code{filename} from the \code{raster} package, except that
instead of returning an empty string for a \code{RasterStack} object, it will return a vector of
length >1 for \code{RasterStack}.
}
\details{
New methods can be made for this generic.
}
\author{
Eliot McIntire
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{purgeChecksums}
\alias{purgeChecksums}
\title{Purge the checksums of a single file}
\usage{
purgeChecksums(checksumFile, fileToRemove)
}
\arguments{
\item{checksumFile}{A character string indicating the absolute path to the \code{CHECKSUMS.txt}
file.}

\item{fileToRemove}{The filename to remove from the \code{checksumFile}}
}
\value{
NULL. Run for its side effect, namely, and file removed from the \file{CHECKSUMS.txt}
file.
}
\description{
This is a manual way of achieving \code{prepInputs(..., purge = 7)}, useful in cases
where \code{prepInputs} is not called directly by the user, so it would be difficult
to set \code{purge = 7}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/messages.R
\name{messageDF}
\alias{messageDF}
\alias{messagePrepInputs}
\alias{messagePreProcess}
\alias{messageCache}
\alias{messageQuestion}
\alias{.messageFunctionFn}
\alias{messageColoured}
\title{Use \code{message} with a consistent use of \code{verbose}}
\usage{
messageDF(
  df,
  round,
  colour = NULL,
  colnames = NULL,
  indent = NULL,
  verbose = getOption("reproducible.verbose"),
  verboseLevel = 1,
  appendLF = TRUE
)

messagePrepInputs(
  ...,
  appendLF = TRUE,
  verbose = getOption("reproducible.verbose"),
  verboseLevel = 1
)

messagePreProcess(
  ...,
  appendLF = TRUE,
  verbose = getOption("reproducible.verbose"),
  verboseLevel = 1
)

messageCache(
  ...,
  colour = getOption("reproducible.messageColourCache"),
  verbose = getOption("reproducible.verbose"),
  verboseLevel = 1,
  appendLF = TRUE
)

messageQuestion(..., verboseLevel = 0, appendLF = TRUE)

.messageFunctionFn(
  ...,
  appendLF = TRUE,
  verbose = getOption("reproducible.verbose"),
  verboseLevel = 1
)

messageColoured(
  ...,
  colour = NULL,
  indent = NULL,
  hangingIndent = TRUE,
  verbose = getOption("reproducible.verbose", 1),
  verboseLevel = 1,
  appendLF = TRUE
)
}
\arguments{
\item{df}{A data.frame, data.table, matrix}

\item{round}{An optional numeric to pass to \code{round}}

\item{colour}{Any colour that can be understood by \code{cli}}

\item{colnames}{Logical or \code{NULL}. If \code{TRUE}, then it will print
column names even if there aren't any in the \code{df} (i.e., they will)
be \code{V1} etc., \code{NULL} will print them if they exist, and \code{FALSE}
which will omit them.}

\item{indent}{An integer, indicating whether to indent each line}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{verboseLevel}{The numeric value for this \verb{message*} call, equal or above
which \code{verbose} must be. The higher this is set, the more unlikely the call
will show a message.}

\item{appendLF}{logical: should messages given as a character string
    have a newline appended?}

\item{...}{Any character vector, passed to \code{paste0(...)}}

\item{hangingIndent}{Logical. If there are \verb{\\n}, should there be a handing indent of 2 spaces.
Default is \code{TRUE}}
}
\value{
Used for side effects. This will produce a message of a structured \code{data.frame}.
}
\description{
This family has a consistent use of \code{verbose} allowing messages to be
turned on or off or verbosity increased or decreased throughout the family of
messaging in \code{reproducible}.
}
\details{
\itemize{
\item \code{messageDF} uses \code{message} to print a clean square data structure.
\item \code{messageColoured} allows specific colours to be used.
\item \code{messageQuestion} sets a high level for \code{verbose} so that the message always gets asked.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/copy.R
\name{.file.move}
\alias{.file.move}
\title{Move a file to a new location -- Defunct -- use \code{hardLinkOrCopy}}
\usage{
.file.move(from, to, overwrite = FALSE)
}
\arguments{
\item{from, to}{character vectors, containing file names or paths.}

\item{overwrite}{logical indicating whether to overwrite destination file if it exists.}
}
\value{
Logical indicating whether operation succeeded.
}
\description{
This will first try to \code{file.rename}, and if that fails, then it will
\code{file.copy} then \code{file.remove}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/robustDigest.R
\name{.robustDigest}
\alias{.robustDigest}
\alias{.robustDigest,ANY-method}
\alias{.robustDigest,function-method}
\alias{.robustDigest,expression-method}
\alias{.robustDigest,language-method}
\alias{.robustDigest,call-method}
\alias{.robustDigest,character-method}
\alias{.robustDigest,Path-method}
\alias{.robustDigest,environment-method}
\alias{.robustDigest,list-method}
\alias{.robustDigest,data.frame-method}
\alias{.robustDigest,numeric-method}
\alias{.robustDigest,matrix-method}
\alias{.robustDigest,integer-method}
\title{Create reproducible digests of objects in R}
\usage{
.robustDigest(
  object,
  .objects = NULL,
  length = getOption("reproducible.length", Inf),
  algo = "xxhash64",
  quick = getOption("reproducible.quick", FALSE),
  classOptions = list(),
  ...
)

\S4method{.robustDigest}{ANY}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{function}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{expression}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{language}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{call}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{character}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{Path}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{environment}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{list}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{data.frame}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{numeric}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{matrix}(object, .objects, length, algo, quick, classOptions)

\S4method{.robustDigest}{integer}(
  object,
  .objects,
  length,
  algo,
  quick,
  classOptions,
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat")
)
}
\arguments{
\item{object}{an object to digest.}

\item{.objects}{Character vector of objects to be digested. This is only applicable
if there is a list, environment (or similar) with named objects
within it. Only this/these objects will be considered for caching,
i.e., only use a subset of
the list, environment or similar objects. In the case of nested list-type
objects, this will only be applied outermost first.}

\item{length}{Numeric. If the element passed to Cache is a \code{Path} class
object (from e.g., \code{asPath(filename)}) or it is a \code{Raster} with
file-backing, then this will be
passed to \code{digest::digest}, essentially limiting the number of bytes
to digest (for speed). This will only be used if \code{quick = FALSE}.
Default is \code{getOption("reproducible.length")}, which is set to \code{Inf}.}

\item{algo}{The digest algorithm to use. Default \code{xxhash64} (see \code{\link[digest:digest]{digest::digest()}} for others).}

\item{quick}{Logical or character. If \code{TRUE},
no disk-based information will be assessed, i.e., only
memory content. See Details section about \code{quick} in \code{\link[=Cache]{Cache()}}.}

\item{classOptions}{Optional list. This will pass into \code{.robustDigest} for
specific classes. Should be options that the \code{.robustDigest} knows what
to do with.}

\item{...}{Arguments passed to \code{FUN}, if \code{FUN} is not an expression.}

\item{cacheSaveFormat}{Character string: currently either \code{qs} or \code{rds}. Defaults to
\code{getOption("reproducible.cacheSaveFormat")}. \code{qs} may be faster but appears to have
narrower range of conditions that work; \code{rds} is safer, and may be slower.}

\item{objects}{Optional character vector indicating which objects are to
be considered while making digestible. This argument is not used
in the default cases; the only known method that uses this
in the default cases; the only known method that uses this
argument is the \code{simList} class from \code{SpaDES.core}.}
}
\value{
A hash i.e., digest of the object passed in.
}
\description{
Not all aspects of R objects are captured by current hashing tools in R
(e.g. \code{digest::digest}, \code{knitr} caching, \code{archivist::cache}).
This is mostly because many objects have "transient"
(e.g., functions have environments), or "disk-backed" features.
Since the goal of using reproducibility is to have tools that are not session specific,
this function attempts to strip all session specific information so that the digest
works between sessions and operating systems.
It is tested under many conditions and object types, there are bound to be others that don't
work correctly.
}
\section{Classes}{


\verb{Raster*} objects have the potential for disk-backed storage, thus, require more work.
Also, because \verb{Raster*} can have a built-in representation for having their data content
located on disk, this format will be maintained if the raster already is file-backed,
i.e., to create \code{.tif} or \code{.grd} backed rasters, use \code{writeRaster} first,
then \code{Cache}.
The \file{.tif} or \file{.grd} will be copied to the \file{raster/} subdirectory of the
\code{cachePath}.
Their RAM representation (as an R object) will still be in the usual  \file{cacheOutputs/}
(or formerly \file{gallery/}) directory.
For \code{inMemory} raster objects, they will remain as binary \code{.RData} files.

Functions (which are contained within environments) are
converted to a text representation via a call to \code{format(FUN)}.

Objects contained within a list or environment are recursively hashed
using \code{\link[digest:digest]{digest::digest()}}, while removing all references to
environments.

Character strings are first assessed with \code{dir.exists} and \code{file.exists}
to check for paths. If they are found to be paths, then the path is hashed with
only its filename via \code{basename(filename)}. If it is actually a path, we suggest
using \code{asPath(thePath)}
}

\examples{

a <- 2
tmpfile1 <- tempfile()
tmpfile2 <- tempfile()
tmpfile3 <- tempfile(fileext = ".grd")
tmpfile4 <- tempfile(fileext = ".grd")
save(a, file = tmpfile1)
save(a, file = tmpfile2)

# treats as character string, so 2 filenames are different
digest::digest(tmpfile1)
digest::digest(tmpfile2)

# tests to see whether character string is representing a file
.robustDigest(tmpfile1)
.robustDigest(tmpfile2) # same

# if you tell it that it is a path, then you can decide if you want it to be
#  treated as a character string or as a file path
.robustDigest(asPath(tmpfile1), quick = TRUE)
.robustDigest(asPath(tmpfile2), quick = TRUE) # different because using file info

.robustDigest(asPath(tmpfile1), quick = FALSE)
.robustDigest(asPath(tmpfile2), quick = FALSE) # same because using file content

# SpatRasters are have pointers
if (requireNamespace("terra", quietly = TRUE)) {
  r <- terra::rast(system.file("ex/elev.tif", package = "terra"))
  r3 <- terra::deepcopy(r)
  r1 <- terra::writeRaster(r, filename = tmpfile3)

  digest::digest(r)
  digest::digest(r3) # different but should be same
  .robustDigest(r1)
  .robustDigest(r3) # same... data & metadata are the same

  # note, this is not true for comparing memory and file-backed rasters
  .robustDigest(r)
  .robustDigest(r1) # different
}

}
\author{
Eliot McIntire
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GPT2.R, R/cache.R
\name{Cache}
\alias{Cache}
\alias{cache2}
\alias{CacheV2}
\title{Saves a wide variety function call outputs to disk and optionally RAM, for recovery later}
\usage{
Cache(
  FUN,
  ...,
  dryRun = getOption("reproducible.dryRun", FALSE),
  notOlderThan = NULL,
  .objects = NULL,
  .cacheExtra = NULL,
  .functionName = NULL,
  .cacheChaining = getOption("reproducible.cacheChaining", NULL),
  outputObjects = NULL,
  algo = "xxhash64",
  cachePath = NULL,
  length = getOption("reproducible.length", Inf),
  userTags = c(),
  omitArgs = NULL,
  classOptions = list(),
  debugCache = character(),
  quick = getOption("reproducible.quick", FALSE),
  verbose = getOption("reproducible.verbose", 1),
  cacheId = NULL,
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
  useCache = getOption("reproducible.useCache", TRUE),
  useCloud = getOption("reproducible.useCloud", FALSE),
  cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
  showSimilar = getOption("reproducible.showSimilar", FALSE),
  drv = getOption("reproducible.drv", NULL),
  conn = getOption("reproducible.conn", NULL)
)

cache2(
  FUN,
  ...,
  dryRun = getOption("reproducible.dryRun", FALSE),
  notOlderThan = NULL,
  .objects = NULL,
  .cacheExtra = NULL,
  .functionName = NULL,
  .cacheChaining = getOption("reproducible.cacheChaining", NULL),
  outputObjects = NULL,
  algo = "xxhash64",
  cachePath = NULL,
  length = getOption("reproducible.length", Inf),
  userTags = c(),
  omitArgs = NULL,
  classOptions = list(),
  debugCache = character(),
  quick = getOption("reproducible.quick", FALSE),
  verbose = getOption("reproducible.verbose", 1),
  cacheId = NULL,
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
  useCache = getOption("reproducible.useCache", TRUE),
  useCloud = getOption("reproducible.useCloud", FALSE),
  cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
  showSimilar = getOption("reproducible.showSimilar", FALSE),
  drv = getOption("reproducible.drv", NULL),
  conn = getOption("reproducible.conn", NULL)
)

CacheV2(
  FUN,
  ...,
  notOlderThan = NULL,
  .objects = NULL,
  .cacheExtra = NULL,
  .functionName = NULL,
  outputObjects = NULL,
  algo = "xxhash64",
  cacheRepo = NULL,
  cachePath = NULL,
  length = getOption("reproducible.length", Inf),
  compareRasterFileLength,
  userTags = c(),
  omitArgs = NULL,
  classOptions = list(),
  debugCache = character(),
  makeCopy = FALSE,
  quick = getOption("reproducible.quick", FALSE),
  verbose = getOption("reproducible.verbose", 1),
  cacheId = NULL,
  useCache = getOption("reproducible.useCache", TRUE),
  useCloud = FALSE,
  cloudFolderID = NULL,
  showSimilar = getOption("reproducible.showSimilar", FALSE),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL)
)
}
\arguments{
\item{FUN}{Either a function (e.g., \code{rnorm}), a function call (e.g., \code{rnorm(1)}),
or an unevaluated function call (e.g., using \code{quote()}).}

\item{...}{Arguments passed to \code{FUN}, if \code{FUN} is not an expression.}

\item{dryRun}{See \link{reproducibleOptions}.}

\item{notOlderThan}{A time. Load an object from the Cache if it was created after this.}

\item{.objects}{Character vector of objects to be digested. This is only applicable
if there is a list, environment (or similar) with named objects
within it. Only this/these objects will be considered for caching,
i.e., only use a subset of
the list, environment or similar objects. In the case of nested list-type
objects, this will only be applied outermost first.}

\item{.cacheExtra}{A an arbitrary R object that will be included in the \code{CacheDigest},
but otherwise not passed into the \code{FUN}. If the user supplies a named list, then
\code{Cache} will report which individual elements of \code{.cacheExtra} have changed
when \code{options("reproducible.showSimilar" = TRUE)}. This can allow a user
more control and understanding for debugging.}

\item{.functionName}{A an arbitrary character string that provides a name that is different
than the actual function name (e.g., "rnorm") which will be used for messaging. This
can be useful when the actual function is not helpful for a user, such as \code{do.call}.}

\item{.cacheChaining}{A logical or a the name of a function. If \code{TRUE}, then
the current \code{Cache} call will evaluate the function "outside" the \code{Cache} call
(via \code{sys.function(-1)}) and
attach the \code{digest} of that outer function to the entry for this \code{Cache} call. This
will then be used by any subsequent \code{Cache} call within the same function.
If the outer function is unchanged, and there is one or more objects that had
been returned by a previous \code{Cache} call,
then those objects will not be digested; rather their \code{cacheId} tag will be used
in place of a new \code{digest}. This \emph{should} cause no change in Caching outcomes,
and it should be faster in cases where there are several \code{Cache} calls within
the same function. If \code{FALSE} (current default), then this feature is
not used. If set to \code{NULL} (i.e., unset, the current default), then it will
not use cache chaining, but it will attach more information to the Cache entries
for each \code{cacheId}, as well as new entries for \code{"surroundingFunction"} digest,
so that if a user switches to \code{.cacheChaining = TRUE}, then it will be able
to begin using cache chaining without needing to rerun the calls again. Can be set by an \code{option}.}

\item{outputObjects}{Optional character vector indicating which objects to
return. This is only relevant for list, environment (or similar) objects}

\item{algo}{The digest algorithm to use. Default \code{xxhash64} (see \code{\link[digest:digest]{digest::digest()}} for others).}

\item{cachePath}{A repository used for storing cached objects.
This is optional if \code{Cache} is used inside a SpaDES module.}

\item{length}{Numeric. If the element passed to Cache is a \code{Path} class
object (from e.g., \code{asPath(filename)}) or it is a \code{Raster} with
file-backing, then this will be
passed to \code{digest::digest}, essentially limiting the number of bytes
to digest (for speed). This will only be used if \code{quick = FALSE}.
Default is \code{getOption("reproducible.length")}, which is set to \code{Inf}.}

\item{userTags}{A character vector with descriptions of the Cache function call. These
will be added to the Cache so that this entry in the Cache can be found using
\code{userTags} e.g., via \code{\link[=showCache]{showCache()}}.}

\item{omitArgs}{Optional. A character vector of argument names in \code{FUN} to
omit from the cache digest, or \code{TRUE} to omit \emph{every} captured argument
(the digest is then based on \code{FUN} itself -- including its body, so a
meaningful edit to the function source still busts the cache -- and on
\code{.cacheExtra}). Useful when the developer wants the cache to be
insensitive to the function's inputs and pin freshness via \code{.cacheExtra}
instead.}

\item{classOptions}{Optional list. This will pass into \code{.robustDigest} for
specific classes. Should be options that the \code{.robustDigest} knows what
to do with.}

\item{debugCache}{Character or Logical. Either \code{"complete"} or \code{"quick"} (uses
partial matching, so "c" or "q" work). \code{TRUE} is equivalent to \code{"complete"}.
If \code{"complete"}, then the returned object from the Cache
function will have two attributes, \code{debugCache1} and \code{debugCache2},
which are the entire \code{list(...)} and that same object, but after all
\code{.robustDigest} calls, at the moment that it is digested using
\code{digest}, respectively. This \code{attr(mySimOut, "debugCache2")}
can then be compared to a subsequent call and individual items within
the object \code{attr(mySimOut, "debugCache1")} can be compared.
If \code{"quick"}, then it will return the same two objects directly,
without evalutating the \code{FUN(...)}.}

\item{quick}{Logical or character. If \code{TRUE},
no disk-based information will be assessed, i.e., only
memory content. See Details section about \code{quick} in \code{\link[=Cache]{Cache()}}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{cacheId}{Character string. If passed, this will override the calculated hash
of the inputs, and return the result from this \code{cacheId} in the \code{cachePath}.
Setting this is equivalent to manually saving the output of this function, i.e.,
the object will be on disk, and will be recovered in subsequent
This may help in some particularly finicky situations
where \code{Cache} is not correctly detecting unchanged inputs. This will guarantee
the object will be identical each time; this may be useful in operational code.}

\item{cacheSaveFormat}{Character string: currently either \code{qs} or \code{rds}. Defaults to
\code{getOption("reproducible.cacheSaveFormat")}. \code{qs} may be faster but appears to have
narrower range of conditions that work; \code{rds} is safer, and may be slower.}

\item{useCache}{Logical, numeric or \code{"overwrite"} or \code{"devMode"}. See details.}

\item{useCloud}{Logical (\code{TRUE} / \code{FALSE} / \code{NULL}) or one of \code{"pull"} /
\code{"push"}. See Details.}

\item{cloudFolderID}{A googledrive dribble of a folder, e.g., using \code{drive_mkdir()}.
If left as \code{NULL}, the function will create a cloud folder with name from last
two folder levels of the \code{cachePath} path, :
\code{paste0(basename(dirname(cachePath)), "_", basename(cachePath))}.
This \code{cloudFolderID} will be added to \code{options("reproducible.cloudFolderID")},
but this will not persist across sessions. If this is a character string, it will
treat this as a folder name to create or use on GoogleDrive.}

\item{showSimilar}{A logical or numeric. Useful for debugging.
If \code{TRUE} or \code{1}, then if the Cache
does not find an identical archive in the \code{cachePath}, it will report (via message)
the next most recent similar archive, and indicate which argument(s) is/are different.
If a number larger than \code{1}, then it will report the N most recent similar archived
objects.}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{cacheRepo}{Same as \code{cachePath}, but kept for backwards compatibility.}

\item{compareRasterFileLength}{Being deprecated; use \code{length}.}

\item{makeCopy}{Now deprecated. Ignored if used.}
}
\value{
Returns the value of the
function call or the cached version (i.e., the result from a previous call
to this same cached function with identical arguments).
}
\description{
\if{html}{\figure{lifecycle-maturing.svg}{options: alt="maturing"}}

A function that can be used to wrap around other functions to cache function calls
for later use. This is normally most effective when the function to cache is
slow to run, yet the inputs and outputs are small. The benefit of caching, therefore,
will decline when the computational time of the "first" function call is fast and/or
the argument values and return objects are large. The default setting (and first
call to Cache) will always save to disk. The 2nd call to the same function will return
from disk, unless \code{options("reproducible.useMemoise" = TRUE)}, then the 2nd time
will recover the object from RAM and is normally much faster (at the expense of RAM use).
}
\details{
There are other similar functions in the R universe.
This version of Cache has been used as part of a robust continuous workflow approach.
As a result, we have tested it with many "non-standard" R objects (e.g., \code{RasterLayer},
\verb{Spat*} objects) and environments (which are always unique, so do not cache readily).

This version of the \code{Cache} function accommodates those four special,
though quite common, cases by:
\enumerate{
\item converting any environments into list equivalents;
\item identifying the dispatched S4 method (including those made through
inheritance) before hashing so the correct method is being cached;
\item by hashing the linked file, rather than the raster object.
Currently, only file-backed \verb{Raster*} or \verb{Spat*} objects are digested
(e.g., not \code{ff} objects, or any other R object where the data
are on disk instead of in RAM);
\item Uses \code{\link[digest:digest]{digest::digest()}}
This is used for file-backed objects as well.
\item Cache will save arguments passed by user in a hidden environment. Any
nested Cache functions will use arguments in this order: 1) actual arguments
passed at each Cache call; 2) any inherited arguments from an outer Cache
call; 3) the default values of the Cache function. See section on \emph{Nested Caching}.
}

\code{Cache} will add a tag to the entry in the cache database called \code{accessed},
which will assign the time that it was accessed, either read or write.
That way, cached items can be shown (using \code{showCache}) or removed (using
\code{clearCache}) selectively, based on their access dates, rather than only
by their creation dates. See example in \code{\link[=clearCache]{clearCache()}}.
}
\note{
As indicated above, several objects require pre-treatment before
caching will work as expected. The function \code{.robustDigest} accommodates this.
It is an S4 generic, meaning that developers can produce their own methods for
different classes of objects. Currently, there are methods for several types
of classes. See \code{\link[=.robustDigest]{.robustDigest()}}.
}
\section{Nested Caching}{

Commonly, Caching is nested, i.e., an outer function is wrapped in a \code{Cache}
function call, and one or more inner functions are also wrapped in a \code{Cache}
function call. A user \emph{can} always specify arguments in every Cache function
call, but this can get tedious and can be prone to errors. The normal way that
\R handles arguments is it takes the user passed arguments if any, and
default arguments for all those that have no user passed arguments. We have inserted
a middle step. The order or precedence for any given \code{Cache} function call is
\enumerate{
\item user arguments, 2. inherited arguments, 3. default arguments. At this time,
the top level \code{Cache} arguments will propagate to all inner functions unless
each individual \code{Cache} call has other arguments specified, i.e., "middle"
nested \code{Cache} function calls don't propagate their arguments to further "inner"
\code{Cache} function calls.  See example.
}

\code{userTags} is unique of all arguments: its values will be appended to the
inherited \code{userTags}.
}

\section{quick}{

The \code{quick} argument is attempting to sort out an ambiguity with character strings:
are they file paths or are they simply character strings. When \code{quick = TRUE},
\code{Cache} will treat these as character strings; when \code{quick = FALSE},
they will be attempted to be treated as file paths first; if there is no file, then
it will revert to treating them as character strings. If user passes a
character vector to this, then this will behave like \code{omitArgs}:
\code{quick = "file"} will treat the argument \code{"file"} as character string.

The most often encountered situation where this ambiguity matters is in arguments about
filenames: is the filename an input pointing to an object whose content we want to
assess (e.g., a file-backed raster), or an output (as in saveRDS) and it should not
be assessed. If only run once, the output file won't exist, so it will be treated
as a character string. However, once the function has been run once, the output file
will exist, and \code{Cache(...)} will assess it, which is incorrect. In these cases,
the user is advised to use \code{quick = "TheOutputFilenameArgument"} to
specify the argument whose content on disk should not be assessed, but whose
character string should be assessed (distinguishing it from \code{omitArgs = "TheOutputFilenameArgument"}, which will not assess the file content nor the
character string).

This is relevant for objects of class \code{character}, \code{Path} and
\code{Raster} currently. For class \code{character}, it is ambiguous whether
this represents a character string or a vector of file paths. If it is known
that character strings should not be treated as paths, then \code{quick = TRUE} is appropriate, with no loss of information. If it is file or
directory, then it will digest the file content, or \code{basename(object)}.
For class \code{Path} objects, the file's metadata (i.e., filename and file
size) will be hashed instead of the file contents if \code{quick = TRUE}. If
set to \code{FALSE} (default), the contents of the file(s) are hashed. If
\code{quick = TRUE}, \code{length} is ignored. \code{Raster} objects are
treated as paths, if they are file-backed.
}

\section{Caching Speed}{

Caching speed may become a critical aspect of a final product. For example,
if the final product is a shiny app, rerunning the entire project may need
to take less then a few seconds at most.
There are 3 arguments that affect \code{Cache} speed: \code{quick}, \code{length}, and \code{algo}.
\code{quick} is passed to \code{.robustDigest}, which currently
only affects \code{Path} and \verb{Raster*} class objects.
In both cases, \code{quick} means that little or no disk-based information will be assessed.
}

\section{Filepaths}{

If a function has a path argument, there is some ambiguity about what should be
done. Possibilities include:
\enumerate{
\item hash the string as is (this will be very system specific, meaning a
\code{Cache} call will not work if copied between systems or directories);
\item hash the \code{basename(path)};
\item hash the contents of the file.
}
If paths are passed in as is (i.e,. character string), the result will not be predictable.
Instead, one should use the wrapper function \code{asPath(path)}, which sets the
class of the string to a \code{Path}, and one should decide whether one wants
to digest the content of the file (using \code{quick = FALSE}),
or just the filename (\code{(quick = TRUE)}). See examples.
}

\section{Stochasticity or randomness}{

In general, it is expected that caching will only be used when randomness is not
desired, e.g., \code{Cache(rnorm(1))} is unlikely to be useful in many cases. However,
\code{Cache} captures the call that is passed to it, leaving all functions unevaluated.
As a result \code{Cache(glm, x ~ y, rnorm(1))} will not work as a means of forcing
a new evaluation each time, as the \code{rnorm(1)} is not evaluated before the call
is assessed against the cache database. To force a new call each time, evaluate
the randomness prior to the Cache call, e.g., \code{ran = rnorm(1)} then pass this
to \code{.cacheExtra}, e.g., \code{Cache(glm, x ~ y, .cacheExtra = ran)}
}

\section{\code{drv} and \code{conn}}{

By default, \code{drv} uses an SQLite database. This can be sufficient for most cases.
However, if a user has dozens or more cores making requests to the Cache database,
it may be insufficient. A user can set up a different database backend, e.g.,
PostgreSQL that can handle multiple simultaneous read-write situations. See
\url{https://github.com/PredictiveEcology/SpaDES/wiki/Using-alternate-database-backends-for-Cache}.
}

\section{\code{useCache}}{

Logical or numeric. If \code{FALSE} or \code{0}, then the entire Caching
mechanism is bypassed and the
function is evaluated as if it was not being Cached. Default is
\code{getOption("reproducible.useCache")}), which is \code{TRUE} by default,
meaning use the Cache mechanism. This may be useful to turn all Caching on or
off in very complex scripts and nested functions. Increasing levels of numeric
values will cause deeper levels of Caching to occur (though this may not
work as expected in all cases). The following is no longer supported:
Currently, only implemented
in \code{postProcess}: to do both caching of inner \code{cropInputs}, \code{projectInputs}
and \code{maskInputs}, and caching of outer \code{postProcess}, use
\code{useCache = 2}; to skip the inner sequence of 3 functions, use \code{useCache = 1}.
For large objects, this may prevent many duplicated save to disk events.

If \code{useCache = "overwrite"}
(which can be set with \code{options("reproducible.useCache" = "overwrite")}), then the function invoke the caching mechanism but will purge
any entry that is matched, and it will be replaced with the results of the
current call.

If \code{useCache = "devMode"}: The point of this mode is to facilitate using the Cache when
functions and datasets are continually in flux, and old Cache entries are
likely stale very often. In \code{devMode}, the cache mechanism will work as
normal if the Cache call is the first time for a function OR if it
successfully finds a copy in the cache based on the normal Cache mechanism.
It \emph{differs} from the normal Cache if the Cache call does \emph{not} find a copy
in the \code{cachePath}, but it does find an entry that matches based on
\code{userTags}. In this case, it will delete the old entry in the \code{cachePath}
(identified based on matching \code{userTags}), then continue with normal \code{Cache}.
For this to work correctly, \code{userTags} must be unique for each function call.
This should be used with caution as it is still experimental. Currently, if
\code{userTags} are not unique to a single entry in the cachePath, it will
default to the behaviour of \code{useCache = TRUE} with a message. This means
that \code{"devMode"} is most useful if used from the start of a project.
}

\section{\code{useCloud}}{

This is experimental and there are many conditions under which this is known
to not work correctly. This is a way to store all or some of the local Cache in the cloud.
Currently, the only cloud option is Google Drive, via \pkg{googledrive}.
For this to work, the user must be or be able to be authenticated
with \code{googledrive::drive_auth}. The principle behind this
\code{useCloud} is that it will be a full or partial mirror of a local Cache.
It is not intended to be used independently from a local Cache. To share
objects that are in the Cloud with another person, it requires 2 steps. 1)
share the \code{cloudFolderID$id}, which can be retrieved by
\code{getOption("reproducible.cloudFolderID")$id} after at least one Cache
call has been made. 2) The other user must then set their  \code{cacheFolderID} in a
\verb{Cache\\(..., reproducible.cloudFolderID = \\"the ID here\\"\\)} call or
set their option manually
\verb{options\\(\\"reproducible.cloudFolderID\\" = \\"the ID here\\"\\)}.

If \code{TRUE}, then this Cache call will download
(if local copy doesn't exist, but cloud copy does exist), upload
(local copy does or doesn't exist and
cloud copy doesn't exist), or
will not download nor upload if object exists in both. If \code{TRUE} will be at
least 1 second slower than setting this to \code{FALSE}, and likely even slower as the
cloud folder gets large. If a user wishes to keep "high-level" control, set this to
\code{getOption("reproducible.useCloud", FALSE)} or
\code{getOption("reproducible.useCloud", TRUE)} (if the default behaviour should
be \code{FALSE} or \code{TRUE}, respectively) so it can be turned on and off with
this option. NOTE: \emph{This argument will not be passed into inner/nested Cache calls.})

Two character values are also accepted, intended for separating developer
and user roles when sharing a cloud-cache folder:
\itemize{
\item \code{"push"} is equivalent to \code{TRUE} (developer role) -- bidirectional;
downloads on a cloud hit, uploads on a miss.
\item \code{"pull"} is read-only (user role) -- downloads on a cloud hit, but never
uploads. If the local cache already has the object, the cloud is not
consulted at all (the Google Drive listing is deferred until after the
local lookup fails). When neither local nor cloud has the object, the
call falls back to a normal local-only Cache run.
}
}

\section{Object attributes}{

Users should be cautioned that object attributes may not be preserved, especially
in the case of objects that are file-backed, such as \code{Raster} or \code{SpatRaster} objects.
If a user needs to keep attributes, they may need to manually re-attach them to
the object after recovery. With the example of \code{SpatRaster} objects, saving
to disk requires \code{terra::wrap} if it is a memory-backed object. When running
\code{terra::unwrap} on this object, any attributes that a user had added are lost.
}

\section{\code{sideEffect}}{

This feature is now deprecated. Do not use as it is ignored.
}

\examples{
data.table::setDTthreads(2)
tmpDir <- tempdir()
opts <- options(reproducible.cachePath = tmpDir)

# Usage -- All below are equivalent; even where args are missing or provided,
#   Cache evaluates using default values, if these are specified in formals(FUN)
a <- list()
b <- list(fun = rnorm)
bbb <- 1
ee <- new.env(parent = emptyenv())
ee$qq <- bbb

a[[1]] <- Cache(rnorm(1)) # no evaluation prior to Cache
a[[2]] <- Cache(rnorm, 1) # no evaluation prior to Cache
a[[3]] <- Cache(do.call, rnorm, list(1))
a[[4]] <- Cache(do.call(rnorm, list(1)))
a[[5]] <- Cache(do.call(b$fun, list(1)))
a[[6]] <- Cache(do.call, b$fun, list(1))
a[[7]] <- Cache(b$fun, 1)
a[[8]] <- Cache(b$fun(1))
a[[10]] <- Cache(quote(rnorm(1)))
a[[11]] <- Cache(stats::rnorm(1))
a[[12]] <- Cache(stats::rnorm, 1)
a[[13]] <- Cache(rnorm(1, 0, get("bbb", inherits = FALSE)))
a[[14]] <- Cache(rnorm(1, 0, get("qq", inherits = FALSE, envir = ee)))
a[[15]] <- Cache(rnorm(1, bbb - bbb, get("bbb", inherits = FALSE)))
a[[16]] <- Cache(rnorm(sd = 1, 0, n = get("bbb", inherits = FALSE))) # change order
a[[17]] <- Cache(rnorm(1, sd = get("ee", inherits = FALSE)$qq), mean = 0)

# with base pipe -- this is put in quotes ('') because R version 4.0 can't understand this
#  if you are using R >= 4.1 or R >= 4.2 if using the _ placeholder,
#  then you can just use pipe normally
usingPipe1 <- "b$fun(1) |> Cache()"  # base pipe

# For long pipe, need to wrap sequence in { }, or else only last step is cached
usingPipe2 <-
  '{"bbb" |>
      parse(text = _) |>
      eval() |>
      rnorm()} |>
    Cache()'
a[[9]] <- eval(parse(text = usingPipe1)) # recovers cached copy
a[[18]] <- eval(parse(text = usingPipe2)) # recovers cached copy

length(unique(a)) == 1 #  all same

### Pipe -- have to use { } or else only final function is Cached
b1a <- 'sample(1e5, 1) |> rnorm() |> Cache()'
b1b <- 'sample(1e5, 1) |> rnorm() |> Cache()'
b2a <- '{sample(1e5, 1) |> rnorm()} |> Cache()'
b2b <- '{sample(1e5, 1) |> rnorm()} |> Cache()'
b1a <- eval(parse(text = b1a))
b1b <- eval(parse(text = b1b))
b2a <- eval(parse(text = b2a))
b2b <- eval(parse(text = b2b))
all.equal(b1a, b1b) # Not TRUE because the sample is run first
all.equal(b2a, b2b) # TRUE because of {  }, sample is not run

#########################
# Advanced examples
#########################

# .cacheExtra -- add something to digest
Cache(rnorm(1), .cacheExtra = "sfessee11") # adds something other than fn args
Cache(rnorm(1), .cacheExtra = "nothing") # even though fn is same, the extra is different

# omitArgs -- remove something from digest (kind of the opposite of .cacheExtra)
Cache(rnorm(2, sd = 1), omitArgs = "sd") # removes one or more args from cache digest
Cache(rnorm(2, sd = 2), omitArgs = "sd") # b/c sd is not used, this is same as previous

# cacheId -- force the use of a digest -- can give undesired consequences
Cache(rnorm(3), cacheId = "k323431232") # sets the cacheId for this call
Cache(runif(14), cacheId = "k323431232") # recovers same as above, i.e, rnorm(3)

# Turn off Caching session-wide
opts <- options(reproducible.useCache = FALSE)
Cache(rnorm(3)) # doesn't cache
options(opts)

# showSimilar can help with debugging why a Cache call isn't picking up a cached copy
Cache(rnorm(4), showSimilar = TRUE) # shows that the argument `n` is different

###############################################
# devMode -- enables cache database to stay
#            small even when developing code
###############################################
opt <- options("reproducible.useCache" = "devMode")
clearCache(tmpDir, ask = FALSE)
centralTendency <- function(x) {
  mean(x)
}
funnyData <- c(1, 1, 1, 1, 10)
uniqueUserTags <- c("thisIsUnique", "reallyUnique")
ranNumsB <- Cache(centralTendency, funnyData, cachePath = tmpDir,
                  userTags = uniqueUserTags) # sets new value to Cache
showCache(tmpDir) # 1 unique cacheId -- cacheId is 71cd24ec3b0d0cac

# During development, we often redefine function internals
centralTendency <- function(x) {
  median(x)
}
# When we rerun, we don't want to keep the "old" cache because the function will
#   never again be defined that way. Here, because of userTags being the same,
#   it will replace the entry in the Cache, effetively overwriting it, even though
#   it has a different cacheId
ranNumsD <- Cache(centralTendency, funnyData, cachePath = tmpDir, userTags = uniqueUserTags)
showCache(tmpDir) # 1 unique artifact -- cacheId is 632cd06f30e111be

# If it finds it by cacheID, doesn't matter what the userTags are
ranNumsD <- Cache(centralTendency, funnyData, cachePath = tmpDir, userTags = "thisIsUnique")
options(opt)

#########################################
# For more in depth uses, see vignette
if (interactive())
  browseVignettes(package = "reproducible")
}
\seealso{
\code{\link[=showCache]{showCache()}}, \code{\link[=clearCache]{clearCache()}}, \code{\link[=keepCache]{keepCache()}},
\code{\link[=CacheDigest]{CacheDigest()}} to determine the digest of a given function or expression,
as used internally within \code{Cache}, \code{\link[=movedCache]{movedCache()}}, \code{\link[=.robustDigest]{.robustDigest()}}, and
for more advanced uses there are several helper functions,
e.g., \code{\link[=rmFromCache]{rmFromCache()}}, \code{\link[=CacheStorageDir]{CacheStorageDir()}}
}
\author{
Eliot McIntire
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{CacheDigest}
\alias{CacheDigest}
\title{The exact digest function that \code{Cache} uses}
\usage{
CacheDigest(
  objsToDigest,
  ...,
  algo = "xxhash64",
  calledFrom = "CacheDigest",
  .functionName = NULL,
  quick = FALSE
)
}
\arguments{
\item{objsToDigest}{A list of all the objects (e.g., arguments) to be digested}

\item{...}{passed to \code{.robustDigest}.}

\item{algo}{The digest algorithm to use. Default \code{xxhash64} (see \code{\link[digest:digest]{digest::digest()}} for others).}

\item{calledFrom}{a Character string, length 1, with the function to
compare with. Default is "Cache". All other values may not produce
robust CacheDigest results.}

\item{.functionName}{A an arbitrary character string that provides a name that is different
than the actual function name (e.g., "rnorm") which will be used for messaging. This
can be useful when the actual function is not helpful for a user, such as \code{do.call}.}

\item{quick}{Logical or character. If \code{TRUE},
no disk-based information will be assessed, i.e., only
memory content. See Details section about \code{quick} in \code{\link[=Cache]{Cache()}}.}
}
\value{
A list of length 2 with the \code{outputHash}, which is the digest
that Cache uses for \code{cacheId} and also \code{preDigest}, which is
the digest of each sub-element in \code{objsToDigest}.
}
\description{
This can be used by a user to pre-test their arguments before running
\code{Cache}, for example to determine whether there is a cached copy.
}
\examples{
data.table::setDTthreads(2)
a <- Cache(rnorm, 1)

# like with Cache, user can pass function and args in a few ways
CacheDigest(rnorm(1)) # shows same cacheId as previous line
CacheDigest(rnorm, 1) # shows same cacheId as previous line

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache-helpers.R
\name{.debugCache}
\alias{.debugCache}
\title{Attach debug info to return for Cache}
\usage{
.debugCache(obj, preDigest, ..., fullCall)
}
\arguments{
\item{obj}{An arbitrary R object.}

\item{preDigest}{A list of hashes.}

\item{...}{Dots passed from \code{Cache}}

\item{fullCall}{The original call to \code{Cache}}
}
\value{
The same object as \code{obj}, but with 2 attributes set.
}
\description{
Internal use only. Attaches an attribute to the output, usable for
debugging the Cache.
}
\author{
Eliot McIntire
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{reproducibleOptions}
\alias{reproducibleOptions}
\title{\code{reproducible} options}
\usage{
reproducibleOptions()
}
\value{
This function returns a list of all the options that the \code{reproducible} package
sets and uses. See below for details of each.
}
\description{
These provide top-level, powerful settings for a comprehensive
reproducible workflow. To see defaults, run \code{reproducibleOptions()}.
See Details below.
}
\details{
Below are options that can be set with \code{options("reproducible.xxx" = newValue)},
where \code{xxx} is one of the values below, and \code{newValue} is a new value to
give the option. Sometimes these options can be placed in the user's \code{.Rprofile}
file so they persist between sessions.

The following options are likely of interest to most users:
\describe{
\item{\code{ask}}{
Default: \code{TRUE}. Used in \code{\link[=clearCache]{clearCache()}} and \code{\link[=keepCache]{keepCache()}}.
}
\item{\code{cacheChaining}}{
Default: \code{FALSE}. Used in \code{\link[=Cache]{Cache()}} in the \code{.cacheChaining} argument.
}
\item{\code{cachePath}}{
Default: \code{.reproducibleTempCacheDir}. Used in \code{\link[=Cache]{Cache()}} and many others.
The default path for repositories if not passed as an argument.
}
\item{\code{cacheSaveFormat}}{
Default: \code{"rds"}. What save format to use; currently, \code{"qs"} (which will use
\code{qs2} package as of \code{reproducible} version ">= 2.1.3"), \code{"qs2"}, or \code{"rds"}.
}
\item{\code{cacheSpeed}}{
Default \code{"slow"}. One of \code{"slow"} or \code{"fast"} (1 or 2).
\code{"slow"} uses \code{digest::digest} internally, which is transferable across operating
systems, but much slower than \verb{digest::digest(algo = "spooky)}.
So, if all caching is happening on a single machine, \code{"fast"} would be a good setting.
}
\item{\code{checkRemoteHash}}{
Default: \code{FALSE}. Used in \code{\link[=preProcess]{preProcess()}} / \code{\link[=prepInputs]{prepInputs()}}. Controls whether
\code{pp_remote_hash_check} re-contacts the remote source (e.g. Google Drive,
HTTP HEAD) when a \code{.hash} sidecar from a previous successful match
already exists in \code{destinationPath}. With the default (\code{FALSE}), the
sidecar is trusted and the remote check is skipped — typically saving
1–2 s per file when the cluster cache is warm. Set to \code{TRUE} to force a
remote round-trip on every call (the pre-3.0.0.9050 behaviour); use
this if the upstream file may change and you need to detect that.
Removing the \verb{<file>_*.hash} sidecar also forces a re-check.
}
\item{\code{conn}}{
Default: \code{NULL}. Sets a specific connection to a database, e.g.,
\code{dbConnect(drv = RSQLite::SQLite())} or \verb{dbConnect(drv = RPostgres::Postgres()}.
For remote database servers, setting one connection may be far faster than using
\code{drv} which must make a new connection every time.
}
\item{\code{destinationPath}}{
Default: \code{NULL}. Used in \code{\link[=prepInputs]{prepInputs()}} and \code{\link[=preProcess]{preProcess()}}.
Can be set globally here.
}
\item{\code{drv}}{
Default: \code{RSQLite::SQLite()}. Sets the default driver for the backend database system.
Only tested with \code{RSQLite::SQLite()} and \code{RPostgres::Postgres()}.
}
\item{dryRun}{
Default: \code{FALSE}.
}
\item{\code{futurePlan}}{
Default: \code{FALSE}. On Linux OSes, \code{Cache} and \code{cloudCache} have some
functionality that uses the \code{future} package.
Default is to not use these, as they are experimental.
They may, however, be very effective in speeding up some things, specifically,
uploading cached elements via \code{googledrive} in \code{cloudCache}.
}
\item{\code{gdalwarp}}{
\strong{Deprecated — do not use.} Default: \code{FALSE}. This option previously
switched \code{postProcessTo} to use \code{sf::gdal_utils("warp")} for a specific
combination of raster/vector inputs. It is no longer needed: current
versions of \code{terra} handle this case well and produce equivalent results
without the GDAL detour. The option is retained only for backwards
compatibility and will be removed in a future release.
}
\item{\code{gdalwarpThreads}}{
\strong{Deprecated — do not use} (see \code{gdalwarp} above). Default: \code{2}.
Previously set \verb{-wo NUM_THREADS=} for \code{gdalProject}.
}
\item{\code{destinationPathShared}}{
Default: \code{NULL}. Used in \code{\link[=prepInputs]{prepInputs()}} and \code{\link[=preProcess]{preProcess()}}.
If set to a path, this will cause these functions to save their downloaded and preprocessed
file to this location, with a hardlink (via \code{file.link}) to the file created in the
\code{destinationPath}.
This can be used so that individual projects that use common data sets can maintain
modularity (by placing downloaded objects in their \code{destinationPath}, but also minimize
re-downloading the same (perhaps large) file over and over for each project.
Because the files are hardlinks, there is no extra space taken up by the apparently
duplicated files.

\if{html}{\out{<div class="sourceCode">}}\preformatted{**Note:** the previous name for this option was `reproducible.inputPaths`; the old
name is still accepted and will continue to work, but `reproducible.destinationPathShared` is
preferred going forward (it matches the [prepInputs()] naming family).
}\if{html}{\out{</div>}}

}
\item{\code{destinationPathSharedRecursive}}{
Default: \code{FALSE}. Used in \code{\link[=prepInputs]{prepInputs()}} and \code{\link[=preProcess]{preProcess()}}.
Should \code{reproducible.destinationPathShared} be searched recursively for existence of a file?

\if{html}{\out{<div class="sourceCode">}}\preformatted{**Note:** the previous name for this option was `reproducible.inputPathsRecursive`;
the old name is still accepted but `reproducible.destinationPathSharedRecursive` is preferred.
}\if{html}{\out{</div>}}

}
\item{\code{inputPaths}}{
\strong{Deprecated} — use \code{reproducible.destinationPathShared} instead.
Retained for backwards compatibility; if set and \code{reproducible.destinationPathShared} is \code{NULL},
the value of \code{reproducible.inputPaths} is used automatically.
}
\item{\code{inputPathsRecursive}}{
\strong{Deprecated} — use \code{reproducible.destinationPathSharedRecursive} instead.
Retained for backwards compatibility.
}
\item{\code{leaveOnDisk}}{
Default: \code{TRUE}. Used in \code{\link[=postProcess]{postProcess()}}.
When there is a \code{SpatRaster} object, should \code{postProcess} force any file-backed object,
to use the file-based, memory-safe tools within \code{terra} (by temporarily setting
\code{terraOption(memfrac = 0)}. Alternatively, if this is set to \code{FALSE},
then \code{postProcess} will let \code{terra} decide on its own based on its internal
cues (largely based on \code{memfrac}, \code{maxmem} \code{terraOptions}). This will be ignored,
however, if the user has set the \code{terraOptions} away from its default of \code{0.5}. The default
increases predictability of whether the returned object is on disk or in memory.
}
\item{\code{memoisePersist}}{
Default: \code{FALSE}. Used in \code{\link[=Cache]{Cache()}}.
Should the memoised copy of the Cache objects persist even if \code{reproducible} reloads
e.g., via \code{devtools::load_all}? This is mostly useful for developers of
\code{reproducible}. If \code{TRUE}, a object named \code{paste0(".reproducibleMemoise_", cachePath)}
will be placed in the \code{.GlobalEnv}, i.e., one for each \code{cachePath}.
}
\item{\code{nThreads}}{
Default: \code{1}. The number of threads to use for reading/writing cache files.
}
\item{\code{objSize}}{
Default: \code{TRUE}. Logical. If \code{TRUE}, then object sizes will be included in
the cache database. Simplying calculating object size of large objects can
be time consuming, so setting this to \code{FALSE} will make caching up to 10\%
faster, depending on the objects.
}
\item{\code{overwrite}}{
Default: \code{FALSE}. Used in \code{\link[=prepInputs]{prepInputs()}}, \code{\link[=preProcess]{preProcess()}},
\code{\link[=downloadFile]{downloadFile()}}, and \code{\link[=postProcess]{postProcess()}}.
}
\item{\code{quick}}{
Default: \code{FALSE}. Used in \code{\link[=Cache]{Cache()}}. This will cause \code{Cache} to use
\code{file.size(file)} instead of the \code{digest::digest(file)}.
Less robust to changes, but faster. \emph{NOTE: this will only affect objects on disk}.
}
\item{\code{rasterRead}}{
Used during \code{prepInputs} when reading \code{.tif}, \code{.grd}, and \code{.asc} files.
Default: \code{terra::rast}. Can be \code{raster::raster} for backwards compatibility.
Can be set using environment variable \code{R_REPRODUCIBLE_RASTER_READ}.
}
\item{\code{shapefileRead}}{
Default \code{NULL}. Used during \code{prepInputs} when reading a \code{.shp} file.
If \code{NULL}, it will use \code{sf::st_read} if \code{sf} package is available; otherwise,
it will use \code{raster::shapefile}
}
\item{\code{showSimilar}}{
Default \code{FALSE}. Passed to \code{Cache}.
}
\item{\code{testCharacterAsFile}}{
Default \code{FALSE}. The behaviour of \code{.robustDigest} on \code{character} vectors prior to
\verb{reproducible == 2.1.2} was that the function would test for whether they were
filenames by using \code{file.exists}. If it was a filename, then it would digest
the file content. In cases of a character vector or a data.frame of "filenames",
this could cause long hanging of the R system as it tries to digest the file
contents of potentially many files. This behaviour is not transparent to a user.
Now the default is to \strong{not} digest the file content of a \code{character} vector
even if they are filenames. To force file content digesting, then convert to
either \code{asPath} or \code{fs::as_fs_path}. Or set this option to \code{TRUE} and the previous
behaviour will return, where it tries to guess whether a character vector
is filenames or not, and if it is, then digest the file content.
}
\item{\code{timeout}}{
Default \code{1200}. Used in \code{preProcess} when downloading occurs. If a user has \code{R.utils}
package installed, \code{R.utils::withTimeout(  , timeout = getOption("reproducible.timeout"))}
will be wrapped around the download so that it will timeout (and error) after this many
seconds.
}
\item{\code{useCache}}{
Default: \code{TRUE}. Used in \code{\link[=Cache]{Cache()}}. If \code{FALSE}, then the entire
\code{Cache} machinery is skipped and the functions are run as if there was no Cache occurring.
Can also take 2 other values: \code{'overwrite'} and \code{'devMode'}.
\code{'overwrite'} will cause no recovery of objects from the cache repository, only new
ones will be created. If the hash is identical to a previous one, then this will overwrite
the previous one.
\code{'devMode'} will function as normally \code{Cache} except it will use the
\code{userTags} to determine if a previous function has been run. If the \code{userTags}
are identical, but the digest value is different, the old value will be deleted from the
cache repository and this new value will be added.
This addresses a common situation during the development stage: functions are changing
frequently, so any entry in the cache repository will be stale following changes to
functions, i.e., they will likely never be relevant again.
This will therefore keep the cache repository clean of stale objects.
If there is ambiguity in the \code{userTags}, i.e., they do not uniquely identify a single
entry in the \code{cachePath}, then this option will default back to the non-dev-mode
behaviour to avoid deleting objects.
This, therefore, is most useful if the user is using unique values for \code{userTags}.
}
\item{\code{reproducible.useCacheV3}}{
Default: \code{TRUE}. If this is set to \code{FALSE}, it will use the old \code{Cache} source
code. This will only be available for a short period before it is deleted
from the package. See also \code{reproducible.digestV3}. It is not guaranteed to
be identical to using a previous version of \verb{reproducible (<3.0)}.
}
\item{\code{useCloud}}{
Default \code{FALSE}. Passed to \code{Cache}.
}
\item{\code{useDBI}}{
Default: \code{TRUE} if \pkg{DBI} is available.
Default value can be overridden by setting environment variable \code{R_REPRODUCIBLE_USE_DBI}.
As of version 0.3, the backend is now \pkg{DBI} instead of \pkg{archivist}.
}
\item{\code{useGdown}}{
Default: \code{FALSE}. If a user provides a Google Drive url to \code{preProcess}/\code{prepInputs},
\code{reproducible} will use the \code{googledrive} package. This works reliably in most cases.
However, for large files on unstable internet connections, it will stall and
stop the download with no error. If a user is finding this behaviour, they can
install the \code{gdown} package, making sure it is available on the PATH. This call
to \code{gdown} will only work for files that do not need authentication. If authentication
is needed, \code{dlGoogle} will fall back to \code{googledrive::drive_download}, even
if this option is \code{TRUE}, with a message.
.
}
\item{\code{useMemoise}}{
Default: \code{FALSE}. Used in \code{\link[=Cache]{Cache()}}. If \code{TRUE}, recovery of cached
elements from the \code{cachePath} will use \code{memoise::memoise}.
This means that the 2nd time running a function will be much faster than the first
in a session (which either will create a new cache entry to disk or read a cached
entry from disk).
\emph{NOTE: memoised values are removed when the R session is restarted}.
\strong{This option will use more RAM} and so may need to be turned off if RAM is limiting.
\code{clearCache} of any sort will cause all memoising to be 'forgotten' (\code{memoise::forget}).
}
\item{\code{useNewDigestAlgorithm}}{
Default: \code{1}. Option 1 is the version that has existed for sometime.
There is now an option \code{2} which is substantially faster.
It will, however, create Caches that are not compatible with previous ones.
Options \code{1} and \code{2} are not compatible with the earlier \code{0}.
\code{1} and \code{2} will make \code{Cache} less sensitive to minor but irrelevant changes
(like changing the order of arguments) and will work successfully across operating systems
(especially relevant for the new \code{cloudCache} function.
}
\item{\code{useTerra}}{
Default: \code{FALSE}. The GIS operations in postProcess, by default use primarily
the Raster package. The newer terra package does similar operations, but usually
faster. A user can now set this option to \code{TRUE} and \code{prepInputs}
and several components of \code{postProcess} will use \code{terra} internally.
}
\item{\code{verbose}}{
Default: \code{FALSE}. If set to \code{TRUE} then every \code{Cache} call will show a
summary of the objects being cached, their \code{object.size} and the time it took to digest
them and also the time it took to run the call and save the call to the cache repository or
load the cached copy from the repository.
This may help diagnosing some problems that may occur.
}
\item{\code{digestV3}}{
Default: \code{TRUE}. This uses a digest approach that includes the names of
list elements and several other tweaks that were created for \verb{reproducible 3.x}.
Set this to \code{FALSE} to use \emph{some of} the previous cache digesting to
achieve some backwards compatibility with the digest algorithms of \verb{reproducible (<3.x)}.
It will not be possible to get it exact for all classes of objects, particularly
those with file-backing.
}

}
}
\section{Advanced}{

The following options are likely not needed by a user.
\describe{
\item{\code{cloudChecksumsFilename}}{
Default: \code{file.path(dirname(.reproducibleTempCacheDir()), "checksums.rds")}.
Used as an experimental argument in \code{\link[=Cache]{Cache()}}
}
\item{\code{length}}{
Default: \code{Inf}. Used in \code{\link[=Cache]{Cache()}}, specifically to the internal
calls to \code{\link[=CacheDigest]{CacheDigest()}}. This is passed to \code{digest::digest}.
Mostly this would be changed from default \code{Inf} if the digesting is taking too long.
Use this with caution, as some objects will have \emph{many} \code{NA} values in their first
\emph{many} elements
}
\item{\code{useragent}}{
Default: \code{"https://github.com/PredictiveEcology/reproducible"}.
User agent for downloads using this package.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cloud.R
\name{checkAndMakeCloudFolderID}
\alias{checkAndMakeCloudFolderID}
\title{Check for presence of \code{checkFolderID} (for \code{Cache(useCloud)})}
\usage{
checkAndMakeCloudFolderID(
  cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
  cachePath = NULL,
  create = FALSE,
  overwrite = FALSE,
  verbose = getOption("reproducible.verbose", 1),
  team_drive = NULL
)
}
\arguments{
\item{cloudFolderID}{The google folder ID where cloud caching will occur.}

\item{cachePath}{A repository used for storing cached objects.
This is optional if \code{Cache} is used inside a SpaDES module.}

\item{create}{Logical. If \code{TRUE}, then the \code{cloudFolderID} will be created.
This should be used with caution as there are no checks for overwriting.
See \code{googledrive::drive_mkdir}. Default \code{FALSE}.}

\item{overwrite}{Logical. Passed to \code{googledrive::drive_mkdir}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{team_drive}{Logical indicating whether to check team drives.}
}
\value{
Returns the character string of the cloud folder ID created or reported
}
\description{
Will check for presence of a \code{cloudFolderID} and make a new one
if one not present on Google Drive, with a warning.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache-helpers.R
\name{copySingleFile}
\alias{copySingleFile}
\alias{copyFile}
\title{Copy a file using \code{robocopy} on Windows and \code{rsync} on Linux/macOS}
\usage{
copySingleFile(
  from = NULL,
  to = NULL,
  useRobocopy = TRUE,
  overwrite = TRUE,
  delDestination = FALSE,
  create = TRUE,
  silent = FALSE
)

copyFile(
  from = NULL,
  to = NULL,
  useRobocopy = TRUE,
  overwrite = TRUE,
  delDestination = FALSE,
  create = TRUE,
  silent = FALSE
)
}
\arguments{
\item{from}{The source file.}

\item{to}{The new file.}

\item{useRobocopy}{For Windows, this will use a system call to \code{robocopy}
which appears to be much faster than the internal \code{file.copy} function.
Uses \verb{/MIR} flag. Default \code{TRUE}.}

\item{overwrite}{Passed to \code{file.copy}}

\item{delDestination}{Logical, whether the destination should have any files deleted,
if they don't exist in the source. This is \verb{/purge} for robocopy and --delete for
rsync.}

\item{create}{Passed to \code{checkPath}.}

\item{silent}{Should a progress be printed.}
}
\value{
This function is called for its side effect, i.e., a file is copied \code{from} to \code{to}.
}
\description{
This is replacement for \code{file.copy}, but for one file at a time.
The additional feature is that it will use \code{robocopy} (on Windows) or
\code{rsync} on Linux or Mac, if they exist.
It will default back to \code{file.copy} if none of these exists.
If there is a possibility that the file already exists, then this function
should be very fast as it will do "update only", i.e., nothing.
}
\examples{
tmpDirFrom <- file.path(tempdir(), "example_fileCopy_from")
tmpDirTo <- file.path(tempdir(), "example_fileCopy_to")
tmpFile1 <- tempfile("file1", tmpDirFrom, ".csv")
tmpFile2 <- tempfile("file2", tmpDirFrom, ".csv")
dir.create(tmpDirFrom, recursive = TRUE, showWarnings = FALSE)
dir.create(tmpDirTo, recursive = TRUE, showWarnings = FALSE)
f1 <- normalizePath(tmpFile1, mustWork = FALSE)
f2 <- normalizePath(tmpFile2, mustWork = FALSE)
t1 <- normalizePath(file.path(tmpDirTo, basename(tmpFile1)), mustWork = FALSE)
t2 <- normalizePath(file.path(tmpDirTo, basename(tmpFile2)), mustWork = FALSE)

write.csv(data.frame(a = 1:10, b = runif(10), c = letters[1:10]), f1)
write.csv(data.frame(c = 11:20, d = runif(10), e = letters[11:20]), f2)
copyFile(c(f1, f2), c(t1, t2))
file.exists(t1) ## TRUE
file.exists(t2) ## TRUE
identical(read.csv(f1), read.csv(f2)) ## FALSE
identical(read.csv(f1), read.csv(t1)) ## TRUE
identical(read.csv(f2), read.csv(t2)) ## TRUE

}
\author{
Eliot McIntire and Alex Chubaty
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DBI.R
\name{.isMemoised}
\alias{.isMemoised}
\title{Evaluate whether a cacheId is memoised}
\usage{
.isMemoised(cacheId, cachePath = getOption("reproducible.cachePath"))
}
\arguments{
\item{cacheId}{Character string. If passed, this will override the calculated hash
of the inputs, and return the result from this \code{cacheId} in the \code{cachePath}.
Setting this is equivalent to manually saving the output of this function, i.e.,
the object will be on disk, and will be recovered in subsequent
This may help in some particularly finicky situations
where \code{Cache} is not correctly detecting unchanged inputs. This will guarantee
the object will be identical each time; this may be useful in operational code.}

\item{cachePath}{A repository used for storing cached objects.
This is optional if \code{Cache} is used inside a SpaDES module.}
}
\value{
A logical, length 1 indicating whether the \code{cacheId} is memoised.
}
\description{
Intended for internal use. Exported so other packages can use this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{cacheId}
\alias{cacheId}
\title{Extract the cache id of an object}
\usage{
cacheId(obj)
}
\arguments{
\item{obj}{Any R object}
}
\value{
The \code{cacheId} if this was part of a \code{Cache} call. Otherwise \code{NULL}
}
\description{
Any object that was returned from the Cache or was calculated as part of a
Cache call will have an attribute, \code{tags} and an entry with \verb{cacheId:} prefix.
This is a lightweight helper to extract that \code{cacheId}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{guessAtTarget}
\alias{guessAtTarget}
\alias{.guessAtTargetAndFun}
\title{Try to pick a file to load}
\usage{
.guessAtTargetAndFun(
  targetFilePath,
  destinationPath = getOption("reproducible.destinationPath", "."),
  filesExtracted,
  fun = NULL,
  verbose = getOption("reproducible.verbose", 1)
)
}
\arguments{
\item{destinationPath}{Full path of the directory where the target file should be}

\item{filesExtracted}{A character vector of all files that have been extracted (e.g.,
from an archive)}
}
\description{
Try to pick a file to load
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/downloadTileAndUpload.R
\name{numCoresToUse}
\alias{numCoresToUse}
\title{Estimate Number of CPU Cores to Use for Parallel Processing}
\usage{
numCoresToUse(min = 2, max = NULL)
}
\arguments{
\item{min}{An integer specifying the minimum number of cores to use. Default is \code{2}.}

\item{max}{An integer specifying the maximum number of cores available,
typically the number of physical cores. Default is
\code{max(1L, getOption("Ncpus", 1L), parallel::detectCores() - 1, logical = FALSE, na.rm = TRUE)}.}
}
\value{
An integer representing the number of cores that can be used for
parallel tasks, ensuring at least \code{min} cores are used, while subtracting
one for the current process and an estimate of actively used threads (via
\code{detectActiveCores()}).
}
\description{
This function estimates the number of CPU cores that can be safely used for
parallel processing, taking into account a minimum threshold, the total
number of physical cores, and currently active threads.
}
\note{
This function depends on \code{detectActiveCores()} and is not supported on
Windows systems.
}
\examples{
if (FALSE) {
  numCoresToUse()
  numCoresToUse(min = 4)
}

}
\seealso{
\code{\link[=detectActiveCores]{detectActiveCores()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cacheGeo.R
\name{CacheGeo}
\alias{CacheGeo}
\title{Cache-like function for spatial domains}
\usage{
CacheGeo(
  targetFile = NULL,
  domain,
  FUN,
  destinationPath = getOption("reproducible.destinationPath", "."),
  useCloud = getOption("reproducible.useCloud", FALSE),
  cloudFolderID = NULL,
  purge = FALSE,
  useCache = getOption("reproducible.useCache"),
  overwrite = getOption("reproducible.overwrite"),
  action = c("nothing", "update", "replace", "append"),
  bufferOK = FALSE,
  verbose = getOption("reproducible.verbose"),
  ...
)
}
\arguments{
\item{targetFile}{The (optional) local file (or path to file) name for a \code{sf}
object or \code{data.frame} that can be coerced to a \code{sf} object (i.e., has a \code{geometry}
column). If \code{cloudFolderID} is specified, then this will be the name of the
file stored and/or accessed in that cloud folder.}

\item{domain}{An sf polygon object that is the spatial area of interest. If \code{NULL},
then this will return the whole object in \code{targetFile}.}

\item{FUN}{A function call that will be called if there is the \code{domain} is
not already contained within the \code{sf} object at \code{targetFile}. This function
call MUST return either a \code{sf} class object or a \code{data.frame} class object
that has a geometry column (which can then be converted to \code{sf} with \code{sf::st_as_sf})}

\item{destinationPath}{Character string of a directory in which to download
and save the file that comes from \code{url} and is also where the function
will look for \code{archive} or \code{targetFile}. NOTE (still experimental):
To prevent repeated downloads in different locations, the user can also set
\code{options("reproducible.destinationPathShared")} to one or more local file paths to
search for the file before attempting to download. Default for that option is
\code{NULL} meaning do not search locally. The previous name
\code{options("reproducible.inputPaths")} is still accepted as a backwards-compatible
alias.}

\item{useCloud}{A logical.}

\item{cloudFolderID}{If this is specified, then it must be either 1) a Google Drive
url to a folder where the \code{targetFile} will be read from or written to, or
2) a \code{googledrive} id or 3) an absolute path to a (possibly non-existent yet)
folder on your Google drive.}

\item{purge}{Logical or Integer. \code{0/FALSE} (default) keeps existing \code{CHECKSUMS.txt} file and
\code{prepInputs} will write or append to it. \code{1/TRUE} will deleted the entire \code{CHECKSUMS.txt} file.
Other options, see details.}

\item{useCache}{Passed to \code{Cache} in various places.
Defaults to \code{getOption("reproducible.useCache", 2L)} in \code{prepInputs}, and
\code{getOption("reproducible.useCache", FALSE)} if calling any of the inner
functions manually. For \code{prepInputs}, this mean it will use \code{Cache}
only up to 2 nested levels, which includes \code{preProcess}. \code{postProcess} and
its nested \verb{*Input} functions (e.g., \code{cropInputs}, \code{projectInputs},
\code{maskInputs}) are no longer internally cached, as \code{terra} processing speeds
mean internal caching is more time consuming. We recommend caching the full
\code{prepInputs} call instead (e.g. \code{prepInputs(...) |> Cache()}).}

\item{overwrite}{Logical. Passed to \code{writeTo} (possibly inside \code{postProcess}) and \code{postProcess}.}

\item{action}{A character string, with one of c("nothing", "update",
"replace", "append"). Partial matching is used ("n" is sufficient).
\code{nothing} will prevent any updating of the \code{targetFile},
i.e., "read only". \code{append} will add the spatial elements in domain to
\code{targetFile} (and writing it back to disk). \code{update} will do the same as
\code{append}, but will also remove any identical geometries before appending.
\code{replace} does nothing currently.}

\item{bufferOK}{A logical. If \code{TRUE}, then after testing whether the domain is
within the \code{targetFile} spatial object, and if it returns \code{FALSE}, then the function
will create a larger object, buffered by 2.5\% of the extent of the object. If
\code{FALSE}, then it will be strict about whether the \code{domain} is within the \code{targetFile}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{All named objects that are needed for FUN, including the function itself,
if it is not in a package.}
}
\value{
Returns an object that results from \code{FUN}, which will possibly be a subset
of a larger spatial object that is specified with \code{targetFile}.
}
\description{
\if{html}{\figure{lifecycle-experimental.svg}{options: alt="experimental"}}
}
\details{
This function is a combination of \code{Cache} and \code{prepInputs} but for spatial
domains. This differs from \code{Cache} in that the current function call doesn't
have to have an identical function call previously run. Instead, it needs
to have had a previous function call where the \code{domain} being passes is
\emph{within} the geographic limits of the \code{targetFile}.
This is similar to a geospatial operation on a remote GIS server, with 2 differences:
\enumerate{
\item This downloads the object first before doing the GIS locally, and 2. it will
optionally upload an updated object if the geographic area did not yet exist.
}

This has a very specific use case: assess whether an existing \code{sf} polygon
or multipolygon object (local or remote) covers the spatial
area of a \code{domain} of interest. If it does, then return only that
part of the \code{sf} object that completely covers the \code{domain}.
If it does not, then run \code{FUN}. It is expected that \code{FUN} will produce an \code{sf}
polygon or multipolygon class object. The result of \code{FUN} will then be
appended to the \code{sf} object as a new entry (feature) or it will replace
the existing "same extent" entry in the \code{sf} object.
}
\examples{
\donttest{

if (requireNamespace("sf", quietly = TRUE) &&
    requireNamespace("terra", quietly = TRUE)) {
  dPath <- checkPath(file.path(tempdir2()), create = TRUE)
  localFileLux <- system.file("ex/lux.shp", package = "terra")

  # 1 step for each layer
  # 1st step -- get study area
  full <- prepInputs(localFileLux, destinationPath = dPath) # default is sf::st_read
  zoneA <- full[3:6, ]
  zoneB <- full[8, ] # not in A
  zoneC <- full[3, ] # yes in A
  zoneD <- full[7:8, ] # not in A, B or C
  zoneE <- full[3:5, ] # yes in A
  # 2nd step: re-write to disk as read/write is lossy; want all "from disk" for this ex.
  writeTo(zoneA, writeTo = "zoneA.shp", destinationPath = dPath)
  writeTo(zoneB, writeTo = "zoneB.shp", destinationPath = dPath)
  writeTo(zoneC, writeTo = "zoneC.shp", destinationPath = dPath)
  writeTo(zoneD, writeTo = "zoneD.shp", destinationPath = dPath)
  writeTo(zoneE, writeTo = "zoneE.shp", destinationPath = dPath)
  # Must re-read to get identical columns
  zoneA <- sf::st_read(file.path(dPath, "zoneA.shp"))
  zoneB <- sf::st_read(file.path(dPath, "zoneB.shp"))
  zoneC <- sf::st_read(file.path(dPath, "zoneC.shp"))
  zoneD <- sf::st_read(file.path(dPath, "zoneD.shp"))
  zoneE <- sf::st_read(file.path(dPath, "zoneE.shp"))

  # The function that is to be run. This example returns a data.frame because
  #    saving `sf` class objects with list-like columns does not work with
  #    many st_driver()
  fun <- function(domain, newField) {
    domain |>
      as.data.frame() |>
      cbind(params = I(lapply(seq_len(NROW(domain)), function(x) newField)))
  }

  # Run sequence -- A, B will add new entries in targetFile, C will not,
  #                 D will, E will not
  for (z in list(zoneA, zoneB, zoneC, zoneD, zoneE)) {
    out <- CacheGeo(
      targetFile = "fireSenseParams.rds",
      domain = z,
      FUN = fun(domain, newField = I(list(list(a = 1, b = 1:2, c = "D")))),
      fun = fun, # pass whatever is needed into the function
      destinationPath = dPath,
      action = "update"
      # , cloudFolderID = "cachedObjects" # to upload/download from cloud
    )
  }
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{getRelative}
\alias{getRelative}
\alias{makeRelative}
\title{Relative paths}
\usage{
getRelative(path, relativeToPath)

makeRelative(files, absoluteBase)
}
\arguments{
\item{path}{character vector or list specifying file paths}

\item{relativeToPath}{directory against which \code{path} will be relativized.}

\item{files}{character vector or list specifying file paths}

\item{absoluteBase}{base directory (as absolute path) to prepend to \code{files}}
}
\description{
Extracting relative file paths.
}
\details{
\itemize{
\item \code{getRelative()} searches \code{path} "from the right" (instead of "from the left")
and tries to reconstruct it relative to directory specified by \code{relativeToPath}.
This is useful when dealing with symlinked paths.
\item \code{makeRelative()} checks to see if \code{files} and \code{normPath(absoluteBase)} share a common path
(i.e., "from the left"), otherwise it returns \code{files}.
}
}
\examples{
\dontshow{if (!identical(.Platform$OS.type, "windows")) withAutoprint(\{ # examplesIf}

## create a project directory (e.g., on a hard drive)
(tmp1 <- tempdir2("myProject", create = TRUE))

## create a cache directory elsewhere (e.g., on an SSD)
(tmp2 <- tempdir2("my_cache", create = TRUE))

## symlink the project cache directory to tmp2
## files created here are actually stored in tmp2
prjCache <- file.path(tmp1, "cache")
file.symlink(tmp2, prjCache)

## create a dummy cache object file in the project cache dir
(tmpf <- tempfile("cache_", prjCache))
cat(rnorm(100), file = tmpf)
file.exists(tmpf)
normPath(tmpf) ## note the 'real' location (i.e., symlink resolved)

getRelative(tmpf, prjCache) ## relative path
getRelative(tmpf, tmp2) ## relative path

makeRelative(tmpf, tmp2) ## abs path; tmpf and normPath(tmp2) don't share common path
makeRelative(tmpf, prjCache) ## abs path; tmpf and normPath(tmp2) don't share common path
makeRelative(normPath(tmpf), prjCache) ## rel path; share common path when both normPath-ed

unlink(tmp1, recursive = TRUE)
unlink(tmp2, recursive = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcessTo.R
\name{postProcessTo}
\alias{postProcessTo}
\alias{postProcessTerra}
\alias{maskTo}
\alias{projectTo}
\alias{cropTo}
\alias{writeTo}
\title{Transform a GIS dataset so it has the properties (extent, projection, mask) of another}
\usage{
postProcessTo(
  from,
  to,
  cropTo = NULL,
  projectTo = NULL,
  maskTo = NULL,
  writeTo = NULL,
  overwrite = TRUE,
  verbose = getOption("reproducible.verbose"),
  ...
)

postProcessTerra(
  from,
  to,
  cropTo = NULL,
  projectTo = NULL,
  maskTo = NULL,
  writeTo = NULL,
  overwrite = TRUE,
  verbose = getOption("reproducible.verbose"),
  ...
)

maskTo(
  from,
  maskTo,
  overwrite = FALSE,
  verbose = getOption("reproducible.verbose"),
  ...
)

projectTo(
  from,
  projectTo,
  overwrite = FALSE,
  verbose = getOption("reproducible.verbose"),
  ...
)

cropTo(
  from,
  cropTo = NULL,
  needBuffer = FALSE,
  overwrite = FALSE,
  verbose = getOption("reproducible.verbose"),
  ...
)

writeTo(
  from,
  writeTo,
  overwrite = getOption("reproducible.overwrite"),
  isStack = NULL,
  isBrick = NULL,
  isRaster = NULL,
  isSpatRaster = NULL,
  verbose = getOption("reproducible.verbose"),
  ...
)
}
\arguments{
\item{from}{A Gridded or Vector dataset on which to do one or more of:
crop, project, mask, and write}

\item{to}{A Gridded or Vector dataset which is the object
whose metadata will be the target for cropping, projecting, and masking of \code{from}.}

\item{cropTo}{Optional Gridded or Vector dataset which,
if supplied, will supply the extent with which to crop \code{from}. To omit
cropping completely, set this to \code{NA}. If supplied, this will override \code{to}
for the cropping step. Defaults to \code{NULL}, which means use \code{to}}

\item{projectTo}{Optional Gridded or Vector dataset, or \code{crs} object (e.g., sf::st_crs).
If Gridded it will supply
the \code{crs}, \code{extent}, \code{res}, and \code{origin}
to project the \code{from} to. If Vector, it will provide the \code{crs} only.
The resolution and extent will be taken from \code{res(from)} (i.e. \code{ncol(from)*nrow(from)}).
If a Vector, the extent of the \code{projectTo} is not used (unless it is also passed to \code{cropTo}.
To omit projecting, set this to \code{NA}.
If supplied, this will override \code{to} for the projecting step.
Defaults to \code{NULL}, which means use \code{to}.
\strong{Attention.} Conflicts may arise with when \code{projectTo} is a Vector/CRS object with a
distinct CRS from \code{to}. Because \code{to} is used for masking \emph{after} \code{from} is re-projected using
\code{projectTo}, the extents of \code{to} and \code{from} may no longer overlap (as in \emph{align})
perfectly leading to failure during
the masking step. We  recommend passing a raster templates to \code{projectTo} whose extent and CRS
are both compatible with the object used later for masking (either \code{to} or \code{maskTo}).}

\item{maskTo}{Optional Gridded or Vector dataset which,
if supplied, will supply the extent with which to mask \code{from}.
If Gridded, it will mask with the \code{NA} values on the \code{maskTo};
if Vector, it will mask on the \code{terra::aggregate(maskTo)}.
To omit masking completely, set this to \code{NA}.
If supplied, this will override \code{to} for the masking step.
Defaults to \code{NULL}, which means use \code{to}}

\item{writeTo}{Optional character string of a filename to use \code{writeRaster} to save the final
object. Default is \code{NULL}, which means there is no \code{writeRaster}}

\item{overwrite}{Logical. Used if \code{writeTo} is not \code{NULL}; also if \code{terra} determines
that the object requires writing to disk during a \code{crop}, \code{mask} or \code{project} call
e.g., because it is too large.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{Arguments passed to \code{terra::mask} (for \code{maskTo}), \code{terra::project} (for \code{projectTo})
or \code{terra::writeRaster} (for \code{writeTo}) and not used for \code{cropTo}, as well \code{postProcess}'s
\code{rasterToMatch} and \code{studyArea} arguments (see below). Commonly used arguments might be
\code{method}, \code{touches}, and \code{datatype}. If \code{filename} is passed, it will be ignored; use
\verb{writeTo = }. If \code{reproducible.gdalwarp = TRUE}, then these will be passed to the
\verb{gdal*} functions. See them for details.}

\item{needBuffer}{Logical. Defaults to \code{FALSE}, meaning nothing is done out
of the ordinary. If \code{TRUE}, then a buffer around the cropTo, so that if a reprojection
has to happen on the \code{cropTo} prior to using it as a crop layer, then a buffer
of 1.5 * res(cropTo) will occur prior, so that no edges are cut off.}

\item{isStack, isBrick, isRaster, isSpatRaster}{Logical. Default \code{NULL}. Used to convert \code{from}
back to these classes prior to writing, if provided.}
}
\value{
An object of the same class as \code{from}, but potentially cropped (via \code{\link[=cropTo]{cropTo()}}),
projected (via \code{\link[=projectTo]{projectTo()}}), masked (via \code{\link[=maskTo]{maskTo()}}), and written to disk
(via \code{\link[=writeTo]{writeTo()}}).
}
\description{
This function provides a single step to achieve the GIS operations
"pre-crop-with-buffer-to-speed-up-projection", "project",
"post-projection-crop", "mask" and possibly "write".
It uses primarily the \code{terra} package internally
(with some minor functions from \code{sf})
in an attempt to be as efficient as possible, except if all inputs are \code{sf} objects.
(in which case \code{sf} is used). Currently, this function is tested
with \code{sf}, \code{SpatVector}, \code{SpatRaster}, \verb{Raster*} and \verb{Spatial*} objects passed
to \code{from}, and the same plus \code{SpatExtent}, and \code{crs} passed to \code{to} or the
relevant \verb{*to} functions.
For this function, Gridded means a \verb{Raster*} class object from \code{raster} or
a \code{SpatRaster} class object from \code{terra}.
Vector means a \verb{Spatial*} class object from \code{sp}, a \code{sf} class object
from \code{sf}, or a \code{SpatVector} class object from \code{terra}.
This function is also used internally with the deprecated family \code{\link[=postProcess]{postProcess()}},
\verb{*Inputs}, such as \code{\link[=cropInputs]{cropInputs()}}.
}
\details{
\code{postProcessTo} is a wrapper around (an initial "wide" crop for speed)
\code{cropTo(needBuffer = TRUE)}, \code{projectTo},
\code{cropTo} (the actual crop for precision), \code{maskTo}, \code{writeTo}.
Users can call each of these individually.

\code{postProcessTerra} is the early name of this function that is now \code{postProcessTo}.

This function is meant to replace \code{\link[=postProcess]{postProcess()}} with the more efficient
and faster \code{terra} functions.
}
\section{Use Cases}{


The table below shows what will result from passing different classes to \code{from}
and \code{to}:

\tabular{lll}{
\strong{\code{from}}\tab \strong{\code{to}}   \tab \strong{\code{from}} will have:                             \cr
\code{Gridded} \tab \code{Gridded}  \tab the extent, projection, origin, resolution
and masking where there are \code{NA} from the \code{to}    \cr
\code{Gridded} \tab \code{Vector}   \tab the projection, origin, and mask from \code{to}, and
extent will be a round number of pixels that
fit within the extent of \code{to}. Resolution will
be the same as \code{from}.  See section
below about \code{projectTo}. \cr
\code{Vector} \tab \code{Vector}    \tab the projection, origin, extent and mask from \code{to} \cr
}

If one or more of the \verb{*To} arguments are supplied, these will
override individual components of \code{to}. If \code{to} is omitted or \code{NULL},
then only the \verb{*To} arguments that are used will be performed. In all cases,
setting a \verb{*To} argument to \code{NA} will prevent that step from happening.
}

\section{\code{projectTo}}{

Since these functions use the gis capabilities of \code{sf} and \code{terra}, they will only
be able to do things that those functions can do. One key caution, which is
stated clearly in \code{?terra::project} is that projection of a raster (i.e., gridded)
object should always be with another gridded object. If the user chooses to
supply a \code{projectTo} that is a vector object for a \code{from} that is gridded,
there may be unexpected failures due e.g., to extents not overlapping during
the \code{maskTo} stage.
}

\section{Backwards compatibility with \code{postProcess}}{


\subsection{\code{rasterToMatch} and \code{studyArea}:}{

If these are supplied, \code{postProcessTo} will use them instead
of \code{to}. If only \code{rasterToMatch} is supplied, it will be assigned to
\code{to}. If only \code{studyArea} is supplied, it will be used for \code{cropTo}
and \code{maskTo}; it will only be used for \code{projectTo} if \code{useSAcrs = TRUE}.
If both \code{rasterToMatch} and \code{studyArea} are supplied,
\code{studyArea} will only be applied to \code{maskTo} (unless \code{maskWithRTM = TRUE}),
and, optionally, to \code{projectTo} (if \code{useSAcrs = TRUE}); everything else
will be from \code{rasterToMatch}.

}

\subsection{\code{targetCRS}, \code{filename2}, \code{useSAcrs}, \code{maskWithRTM}:}{

\code{targetCRS} if supplied will be assigned to \code{projectTo}. \code{filename2} will
be assigned to \code{writeTo}. If \code{useSAcrs} is set, then the \code{studyArea}
will be assigned to \code{projectTo}. If \code{maskWithRTM} is used, then the
\code{rasterToMath} will be assigned to \code{maskTo}. All of these will override
any existing values for these arguments.

}

See also \code{\link[=postProcess]{postProcess()}} documentation section on
\emph{Backwards compatibility with \code{rasterToMatch} and/or \code{studyArea}} for further
detail.
}

\section{Cropping}{

If \code{cropTo} is not \code{NA}, \code{postProcessTo} does cropping twice, both the first and last steps.
It does it first for speed, as cropping is a very fast algorithm. This will quickly remove
a bunch of pixels that are not necessary. But, to not create bias, this first crop is padded
by  \code{2 * res(from)[1]}), so that edge cells still have a complete set of neighbours.
The second crop is at the end, after projecting and masking. After the projection step,
the crop is no longer tight. Under some conditions, masking will effectively mask and crop in
one step, but under some conditions, this is not true, and the mask leaves padded NAs out to
the extent of the \code{from} (as it is after crop, project, mask). Thus the second
crop removes all NA cells so they are tight to the mask.
}

\examples{
if (require("terra", quietly = TRUE)) {
  # prepare dummy data -- 3 SpatRasters, 2 SpatVectors
  # need 2 SpatRaster
  rf <- system.file("ex/elev.tif", package = "terra")
  elev1 <- terra::rast(rf)

  # a polygon vector
  f <- system.file("ex/lux.shp", package = "terra")
  vOrig <- terra::vect(f)
  v <- vOrig[1:2, ]

  # utm <- terra::crs("epsg:23028") # $wkt
  utm <- "+proj=utm +zone=28 +datum=WGS84 +units=m +no_defs"
  vInUTM <- terra::project(vOrig, utm)
  vAsRasInLongLat <- terra::rast(vOrig, resolution = 0.008333333)
  res100 <- 100
  rInUTM <- terra::rast(vInUTM, resolution = res100, vals = 1)
  # crop, reproject, mask, crop a raster with a vector in a different projection
  #  --> gives message about not enough information
  t1 <- postProcessTo(elev1, to = vInUTM)
  # crop, reproject, mask a raster to a different projection, then mask
  t2a <- postProcessTo(elev1, to = vAsRasInLongLat, maskTo = vInUTM)
  t3a <- postProcessTo(elev1, to = rInUTM, maskTo = vInUTM)

}
}
\seealso{
\code{\link[=maskTo]{maskTo()}}, \code{\link[=cropTo]{cropTo()}}, \code{\link[=projectTo]{projectTo()}}, \code{\link[=writeTo]{writeTo()}}, and \code{\link[=fixErrorsIn]{fixErrorsIn()}}.
Also the functions that
call \code{sf::gdal_utils(...)} directly: \code{\link[=gdalProject]{gdalProject()}}, \code{\link[=gdalResample]{gdalResample()}}, \code{\link[=gdalMask]{gdalMask()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/showCacheEtc.R
\name{mergeCache}
\alias{mergeCache}
\alias{mergeCache,ANY-method}
\title{Merge two cache repositories together}
\usage{
mergeCache(
  cacheTo,
  cacheFrom,
  drvTo = getDrv(getOption("reproducible.drv", NULL)),
  drvFrom = getDrv(getOption("reproducible.drv", NULL)),
  connTo = NULL,
  connFrom = NULL,
  verbose = getOption("reproducible.verbose")
)

\S4method{mergeCache}{ANY}(
  cacheTo,
  cacheFrom,
  drvTo = getDrv(getOption("reproducible.drv", NULL)),
  drvFrom = getDrv(getOption("reproducible.drv", NULL)),
  connTo = NULL,
  connFrom = NULL,
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{cacheTo}{The cache repository (character string of the file path)
that will become larger, i.e., merge into this}

\item{cacheFrom}{The cache repository (character string of the file path)
from which all objects will be taken and copied from}

\item{drvTo}{The database driver for the \code{cacheTo}.}

\item{drvFrom}{The database driver for the \code{cacheFrom}}

\item{connTo}{The connection for the \code{cacheTo}. If not provided, then
a new one will be made from \code{drvTo} and \code{cacheTo}}

\item{connFrom}{The database for the \code{cacheFrom}. If not provided, then
a new one will be made from \code{drvFrom} and \code{cacheFrom}}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\value{
The character string of the path of \code{cacheTo}, i.e., not the
objects themselves.
}
\description{
\if{html}{\figure{lifecycle-experimental.svg}{options: alt="experimental"}}
}
\details{
All the \code{cacheFrom} artifacts will be put into \code{cacheTo}
repository. All \code{userTags} will be copied verbatim, including
\code{accessed}, with 1 exception: \code{date} will be the
current \code{Sys.time()} at the time of merging. The
\code{createdDate} column will be similarly the current time
of merging.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cloud.R
\name{cloudUploadFromCache}
\alias{cloudUploadFromCache}
\title{Upload a file to cloud directly from local \code{cachePath}}
\usage{
cloudUploadFromCache(
  isInCloud,
  outputHash,
  cachePath,
  cloudFolderID,
  outputToSave,
  rasters,
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{isInCloud}{A logical indicating whether an \code{outputHash} is in the cloud already.}

\item{outputHash}{The \code{cacheId} of the object to upload}

\item{cachePath}{A repository used for storing cached objects.
This is optional if \code{Cache} is used inside a SpaDES module.}

\item{cloudFolderID}{A googledrive dribble of a folder, e.g., using \code{drive_mkdir()}.
If left as \code{NULL}, the function will create a cloud folder with name from last
two folder levels of the \code{cachePath} path, :
\code{paste0(basename(dirname(cachePath)), "_", basename(cachePath))}.
This \code{cloudFolderID} will be added to \code{options("reproducible.cloudFolderID")},
but this will not persist across sessions. If this is a character string, it will
treat this as a folder name to create or use on GoogleDrive.}

\item{outputToSave}{Only required if \code{any(rasters) == TRUE}.
This is the \verb{Raster*} object.}

\item{rasters}{A logical vector of length >= 1 indicating which elements in
\code{outputToSave} are \verb{Raster*} objects.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\description{
Meant for internal use, as there are internal objects as arguments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{.setSubAttrInList}
\alias{.setSubAttrInList}
\title{Set subattributes within a list by reference}
\usage{
.setSubAttrInList(object, attr, subAttr, value)
}
\arguments{
\item{object}{An arbitrary object}

\item{attr}{The attribute name (that is a list object) to change}

\item{subAttr}{The list element name to change}

\item{value}{The new value}
}
\value{
This sets or updates the \code{subAttr} element of a list that is located at
\code{attr(object, attr)}, with the \code{value}. This, therefore, updates a sub-element
of a list attribute and returns that same object with the updated attribute.
}
\description{
Sets only a single element within a list attribute.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{dlGoogle}
\alias{dlGoogle}
\title{Download file from Google Drive}
\usage{
dlGoogle(
  url,
  archive = NULL,
  targetFile = NULL,
  checkSums,
  messSkipDownload,
  destinationPath,
  type = NULL,
  overwrite,
  needChecksums,
  verbose = getOption("reproducible.verbose", 1),
  team_drive = NULL,
  ...
)
}
\arguments{
\item{url}{The url (link) to the file.}

\item{archive}{Optional character string giving the path of an archive
containing \code{targetFile}, or a vector giving a set of nested archives
(e.g., \code{c("xxx.tar", "inner.zip", "inner.rar")}). If there is/are (an) inner
archive(s), but they are unknown, the function will try all until it finds
the \code{targetFile}. See table in \code{\link[=preProcess]{preProcess()}}. If it is \code{NA},
then it will \emph{not} attempt to see it as an archive, even if it has archive-like
file extension (e.g., \code{.zip}). This may be useful when an R function
is expecting an archive directly.}

\item{targetFile}{Character string giving the filename (without relative or
absolute path) to the eventual file
(raster, shapefile, csv, etc.) after downloading and extracting from a zip
or tar archive. This is the file \emph{before} it is passed to
\code{postProcess}. The internal checksumming does not checksum
the file after it is \code{postProcess}ed (e.g., cropped/reprojected/masked).
Using \code{Cache} around \code{prepInputs} will do a sufficient job in these cases.
See table in \code{\link[=preProcess]{preProcess()}}.}

\item{destinationPath}{Character string of a directory in which to download
and save the file that comes from \code{url} and is also where the function
will look for \code{archive} or \code{targetFile}. NOTE (still experimental):
To prevent repeated downloads in different locations, the user can also set
\code{options("reproducible.destinationPathShared")} to one or more local file paths to
search for the file before attempting to download. Default for that option is
\code{NULL} meaning do not search locally. The previous name
\code{options("reproducible.inputPaths")} is still accepted as a backwards-compatible
alias.}

\item{overwrite}{Logical. Passed to \code{writeTo} (possibly inside \code{postProcess}) and \code{postProcess}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{Not used here. Only used to allow other arguments to other fns to not fail.}
}
\description{
Download file from Google Drive
}
\author{
Eliot McIntire and Alex Chubaty
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convertPaths.R
\name{convertPaths}
\alias{convertPaths}
\alias{convertRasterPaths}
\title{Change the absolute path of a file}
\usage{
convertPaths(x, patterns, replacements)

convertRasterPaths(x, patterns, replacements)
}
\arguments{
\item{x}{For \code{convertPaths}, a character vector of file paths.
For \code{convertRasterPaths}, a disk-backed \code{RasterLayer}
object, or a list of such rasters.}

\item{patterns}{Character vector containing a pattern to match (see \code{?gsub}).}

\item{replacements}{Character vector of the same length of \code{patterns}
containing replacement text (see \code{?gsub}).}
}
\value{
A normalized path with the \code{patterns} replaced by \code{replacements}. Or a list of such
objects if \code{x} was a list.
}
\description{
\code{convertPaths} is simply a wrapper around \code{gsub} for changing the
first part of a path.
\code{convertRasterPaths} is useful for changing the path to a file-backed
raster (e.g., after copying the file to a new location).
}
\examples{
filenames <- c("/home/user1/Documents/file.txt", "/Users/user1/Documents/file.txt")
oldPaths <- dirname(filenames)
newPaths <- c("/home/user2/Desktop", "/Users/user2/Desktop")
convertPaths(filenames, oldPaths, newPaths)

}
\author{
Eliot McIntire and Alex Chubaty
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache-helpers.R
\name{isUpdated}
\alias{isUpdated}
\title{Has a cached object has been updated?}
\usage{
isUpdated(x)
}
\arguments{
\item{x}{cached object}
}
\value{
logical
}
\description{
Has a cached object has been updated?
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DBI.R
\name{saveToCache}
\alias{saveToCache}
\title{Save an object to Cache}
\usage{
saveToCache(
  cachePath = getOption("reproducible.cachePath"),
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  obj,
  userTags,
  cacheId,
  linkToCacheId = NULL,
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{cachePath}{A repository used for storing cached objects.
This is optional if \code{Cache} is used inside a SpaDES module.}

\item{cacheSaveFormat}{Character string: currently either \code{qs} or \code{rds}. Defaults to
\code{getOption("reproducible.cacheSaveFormat")}. \code{qs} may be faster but appears to have
narrower range of conditions that work; \code{rds} is safer, and may be slower.}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{obj}{The R object to save to the cache}

\item{userTags}{A character vector with descriptions of the Cache function call. These
will be added to the Cache so that this entry in the Cache can be found using
\code{userTags} e.g., via \code{\link[=showCache]{showCache()}}.}

\item{cacheId}{The hash string representing the result of \code{.robustDigest}}

\item{linkToCacheId}{Optional. If a \code{cacheId} is provided here, then a \code{file.link}
will be made to the file with that \code{cacheId} name in the cache repo.
This is used when identical outputs exist in the cache. This will save disk space.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\value{
This is used for its side effects, namely, it will add the object to the cache and
cache database.
}
\description{
This is not expected to be used by a user as it requires that the \code{cacheId} be
calculated in exactly the same as it calculated inside \code{Cache}
(which requires \code{match.call} to match arguments with their names, among other things).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputsCOG.R
\name{prepInputsCOG}
\alias{prepInputsCOG}
\title{Fast spatial subsetting of Cloud Optimized GeoTiff (COG) files}
\usage{
prepInputsCOG(url, verbose = getOption("reproducible.verbose", 1), ...)
}
\arguments{
\item{url}{Character. An HTTP(S) URL pointing to a GeoTiff file.}

\item{verbose}{Numeric or Logical. Verbosity level.}

\item{...}{Passed through; expected to contain at least one of \code{to}, \code{cropTo},
or \code{maskTo} (a spatial object defining the area of interest).}
}
\value{
A \code{SpatRaster} windowed to the bounding box of the \code{to}/\code{cropTo}/\code{maskTo}
object (in the COG's own CRS), or the character string \code{"NULL"} if any
pre-condition fails (not HTTP, no spatial arg, network error, empty window, etc.).
}
\description{
An alternative fast-path inside \code{\link[=prepInputs]{prepInputs()}} for remote Cloud Optimized GeoTiffs.
When a URL points to a COG and the user has specified a spatial subsetting argument
(\code{to}, \code{cropTo}, or \code{maskTo}), this function reads only the spatial window of interest
via GDAL's \verb{/vsicurl/} virtual filesystem — no full-file download is needed.
The returned \code{SpatRaster} is passed back to \code{prepInputs} where the normal
\code{postProcess} step (mask, reproject, write) completes the pipeline.
}
\details{
This function is called automatically from inside \code{\link[=prepInputs]{prepInputs()}} when
\code{getOption("reproducible.useCOG")} is \code{TRUE} (the default). It can also be
called directly.
}
\seealso{
\code{\link[=prepInputs]{prepInputs()}}, \code{\link[=prepInputsWithTiles]{prepInputsWithTiles()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/downloadTileAndUpload.R
\name{prepInputsWithTiles}
\alias{prepInputsWithTiles}
\title{Alternative to \code{prepInputs} that can use Spatial Tiles stored locally or on Google Drive}
\usage{
prepInputsWithTiles(
  targetFile,
  url,
  destinationPath,
  to,
  tilesFolder = file.path(getOption("reproducible.inputPath"), "tiles"),
  urlTiles = getOption("reproducible.prepInputsUrlTiles", NULL),
  doUploads = getOption("reproducible.prepInputsDoUploads", FALSE),
  tileGrid = "CAN",
  numTiles = NULL,
  plot.grid = FALSE,
  purge = FALSE,
  verbose = getOption("reproducible.verbose"),
  ...
)
}
\arguments{
\item{targetFile}{Character. Name of the target file to be downloaded or processed.
If missing, it will be inferred from the URL or Google Drive metadata.}

\item{url}{Character. URL to the full dataset (Google Drive or HTTP/S).}

\item{destinationPath}{Character. Path to the directory where files will be downloaded and processed.}

\item{to}{A spatial object (e.g., \code{SpatRaster}, \code{SpatVector}, \code{sf}, or \verb{Spatial*}) defining the area of interest.}

\item{tilesFolder}{A local file path to put tiles. If this is an absolute path, then
that will be used; if it is a relative path, then it will be
\code{file.path(destinationPath, tilesFolder)}}

\item{urlTiles}{Character. URL to the tile source (e.g., Google Drive folder or HTTP/S endpoint). Default is \code{getOption("reproducible.prepInputsUrlTiles", NULL)}.}

\item{doUploads}{Logical. Whether to upload processed tiles.
Default is \code{getOption("reproducible.prepInputsDoUploads", FALSE)}.}

\item{tileGrid}{Either length 3 character string, such as "CAN", to be sent to \code{geodata::gadm(...)}
or an actual \code{SpatVector} object with a grid of polygons}

\item{numTiles}{Integer. Number of tiles to generate. Optional.}

\item{plot.grid}{Logical. Whether to plot the tile grid and area of interest. Default is \code{FALSE}.}

\item{purge}{Logical or Integer. \code{0/FALSE} (default) keeps existing \code{CHECKSUMS.txt} file and
\code{prepInputs} will write or append to it. \code{1/TRUE} will deleted the entire \code{CHECKSUMS.txt} file.}

\item{verbose}{Logical or numeric. Controls verbosity of messages. Default is \code{getOption("reproducible.verbose")}.}

\item{...}{Either \code{maskTo}, \code{cropTo} (which will be used if \code{to} is not supplied, or
arguments passed to \code{writeRaster}, e.g., \code{datatype} (used when writing tiles).}
}
\value{
A single, merged \code{SpatRaster} object \code{postProcess}ed to the area of interest (\code{to}),
composed of the necessary tiles.
If the post-processed file already exists locally, it will be returned directly.
}
\description{
Downloads, processes and optionally uploads a \code{SpatRaster} object through a tiling intermediary.
If the original \code{url} is for a very large object, but \code{to} is a relatively small subset
of the area represented by the spatial file at \code{url}, then this function will
potentially by-pass the download of the large file at \code{url} and instead only download
the minimum number of tiles necessary to cover the \code{to} area. When \code{doUploads} is
TRUE, then this function will potentially create and upload the tiles to \code{tileFolder},
prior to returning the spatial object, \code{postProcess}ed to \code{to}. This function supports
both Google Drive and HTTP(S) URLs.
}
\details{
This function can be triggered \emph{inside} \code{prepInputs}
if the \code{to} is supplied and both \code{url} and \code{urlTiles} are supplied. \strong{NOTE}:
\code{urlTiles} can be supplied using the
\verb{option(reproducible.prepInputsUrlTiles = someGoogleDriveFolderURL}), so the original
\code{prepInputs} function call can remain unaffected.

This function also uses a different checksumming procedure compared to the normal \code{prepInputs}.
This function will assess the remote url for a hash. If that hash exists, then
it will compare it to a local file with \code{targetFile} name, suffixed with \code{.hash}. If the
two hashes differ (remote and local), then it will be redownloaded; otherwise the local
one will be returned.

This function is useful for working with large spatial datasets, but where the user
only requires a "relatively small" section of that dataset. This function will
potentially bypass the full download and download only the tiles that are necessary
for the \code{to}.
It handles downloading only the required tiles based on spatial intersection
with the target area, and supports resumable downloads from Google Drive or HTTP/S sources.

If \code{targetFile} is missing, the function attempts to infer it from the URL
using the \code{Content-Disposition} header or the basename of the URL.
For Google Drive URLs, it uses the file metadata.
}
\examples{

if (FALSE) {
  to <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(-123.3656, 48.4284)), crs = 4326))
  result <- prepInputsWithTiles(
    url = "https://example.com/data.tif",
    destinationPath = tempdir(),
    to = to,
    urlTiles = "https://example.com/tiles/",
    tileGrid = "CAN"
  )
}

}
\seealso{
\code{\link[googledrive:drive_get]{googledrive::drive_get()}}, \code{\link[terra:rast]{terra::rast()}}, \code{\link[terra:crop]{terra::crop()}}, \code{\link[terra:merge]{terra::merge()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/terra-migration.R
\name{minFn}
\alias{minFn}
\alias{maxFn}
\alias{dataType2}
\alias{nlayers2}
\alias{values2}
\title{Get min or maximum value of a (Spat)Raster}
\usage{
minFn(x)

maxFn(x)

dataType2(x, ...)

nlayers2(x)

values2(x, ...)
}
\arguments{
\item{x}{A \code{Raster} or \code{SpatRaster} object.}

\item{...}{Passed to the functions in \code{raster} or \code{terra}, as needed.}
}
\value{
A vector (not matrix as in \code{terra::minmax}) with the minimum or maximum
value on the \code{Raster} or \code{SpatRaster}, one value per layer.
}
\description{
During the transition from raster to terra, some functions are not drop in
replacements, such as \code{minValue} and \code{maxValue} became \code{terra::minmax}. This
helper allows one function to be used, which calls the correct max or min
function, depending on whether the object is a \code{Raster} or \code{SpatRaster}.
}
\examples{
if (requireNamespace("terra", quietly = TRUE)) {
  ras <- terra::rast(terra::ext(0, 10, 0, 10), vals = 1:100)
  maxFn(ras)
  minFn(ras)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/messages.R, R/zzz.R
\name{.message}
\alias{.message}
\alias{.pkgEnv}
\title{The \code{reproducible} package environments}
\usage{
.message

.pkgEnv
}
\description{
Environment used internally to store internal package objects and methods.
}
\details{
\itemize{
\item \code{.message} is specifically for messages and message-generating functions;
\item \code{.pkgEnv} is for general use within the package;
\item \code{.reproEnv} is used for \code{Cache}-related objects;
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportedMethods.R
\name{remapFilenames}
\alias{remapFilenames}
\title{Remap file names}
\usage{
remapFilenames(obj, tags, cachePath = getOption("reproducible.cachePath"), ...)
}
\arguments{
\item{obj}{(optional) object whose file path metadata will be remapped}

\item{tags}{cache tags \code{data.table} object}

\item{cachePath}{character string specifying the path to the cache directory or \code{NULL}}

\item{...}{Additional path arguments, passed to \code{absoluteBase()} and \code{modifyListPaths()}}
}
\description{
Update file path metadata for file-backed objects (e.g., \code{SpatRasters}).
Useful when moving saved objects between projects or machines.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{set.randomseed}
\alias{set.randomseed}
\title{Set seed with a random value using Sys.time()}
\usage{
set.randomseed(set.seed = TRUE)
}
\arguments{
\item{set.seed}{Logical. If \code{TRUE}, the default, then the function will call
\code{set.seed} internally with the new random seed.}
}
\value{
This will return the new seed invisibly. However, this is also called for
its side effects, which is a new seed set using \code{set.seed}
}
\description{
This will set a random seed.
}
\details{
This function uses 6 decimal places of \code{Sys.time()}, i.e., microseconds. Due to
integer limits, it also truncates at 1000 seconds, so there is a possibility that
this will be non-unique after 1000 seconds (at the microsecond level). In
tests, this showed no duplicates after 1e7 draws in a loop, as expected.
}
\note{
This function does not appear to be as reliable on R <= 4.1.3
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{isWindows}
\alias{isWindows}
\title{Test whether system is Windows}
\usage{
isWindows()
}
\description{
This is used so that unit tests can override this using \code{testthat::with_mock}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/objectSize.R
\name{objSize}
\alias{objSize}
\alias{objSizeSession}
\title{Wrapper around \code{lobstr::obj_size}}
\usage{
objSize(x, quick = FALSE, recursive = FALSE, ...)

objSizeSession(sumLevel = Inf, enclosingEnvs = TRUE, .prevEnvirs = list())
}
\arguments{
\item{x}{An object}

\item{quick}{Logical. If \code{FALSE}, then an attribute, "objSize" will be added to
the returned value, with each of the elements' object size returned also.}

\item{recursive}{Logical. If \code{TRUE}, then, in addition to evaluating the whole object,
it will also return the recursive sizes of the elements of a list or environment.}

\item{...}{Additional arguments (currently unused), enables backwards compatible use.}

\item{sumLevel}{Numeric, indicating at which depth in the list of objects should the
object sizes be summed (summarized). Default is \code{Inf}, meaning no sums. Currently,
the only option other than Inf is 1: \code{objSizeSession(1)},
which gives the size of each package.}

\item{enclosingEnvs}{Logical indicating whether to include enclosing environments.
Default \code{TRUE}.}

\item{.prevEnvirs}{For internal account keeping to identify and prevent duplicate counting}
}
\value{
This will return the result from \code{lobstr::obj_size}, i.e., a \code{lobstr_bytes}
which is a \code{numeric}. If \code{quick = FALSE}, it will also have an attribute,
"objSize", which will
be a list with each element being the \code{objSize} of the individual elements of \code{x}.
This is particularly useful if \code{x} is a \code{list} or \code{environment}.
However, because of the potential for shared memory, the sum of the individual
elements will generally not equal the value returned from this function.
}
\description{
This function attempts to estimate the real object size of an object. If the object
has pass-by-reference semantics, it may not estimate the object size well without
a specific method developed. For the case of \code{terra} class objects, this will
be accurate (both RAM and file size), but only if it is not passed inside
a list or environment. To get an accurate size of these, they should be passed
individually.
}
\details{
For functions, a user can include the enclosing environment as described
\url{https://www.r-bloggers.com/2015/03/using-closures-as-objects-in-r/} and
\url{http://adv-r.had.co.nz/memory.html}.
It is not entirely clear which estimate is better.
However, if the enclosing environment is the \code{.GlobalEnv}, it will
not be included even though \code{enclosingEnvs = TRUE}.

\code{objSizeSession} will give the size of the whole session, including loaded packages.
Because of the difficulties in calculating the object size of \code{base}
and \code{methods} packages and \code{Autoloads}, these are omitted.
}
\examples{
library(utils)

foo <- new.env()
foo$b <- 1:10
foo$d <- 1:10

objSize(foo) # all the elements in the environment
utils::object.size(foo) # different - only measuring the environment as an object

utils::object.size(prepInputs) # only the function, without its enclosing environment
objSize(prepInputs) # the function, plus its enclosing environment

os1 <- utils::object.size(as.environment("package:reproducible"))
(os1) # very small -- just the environment container

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{searchFull}
\alias{searchFull}
\alias{searchFullEx}
\title{Search up the full scope for functions}
\usage{
searchFull(env = parent.frame(), simplify = TRUE)

searchFullEx()
}
\arguments{
\item{env}{The environment to start searching at. Default is
calling environment, i.e., \code{parent.frame()}}

\item{simplify}{Logical. Should the output be simplified to character,
if possible (usually it is not possible because environments
don't always coerce correctly)}
}
\value{
A list of environments that is the actual search path, unlike \code{search()}
which only prints from \code{.GlobalEnv} up to \code{base} through user attached
packages.
}
\description{
This is like \code{base::search} but when used inside a function, it will
show the full scope (see figure in the section \emph{Binding environments}
on \url{http://adv-r.had.co.nz/Environments.html}).
This full search path will be potentially much longer than
just \code{search()} (which always starts at \code{.GlobalEnv}).

\code{searchFullEx} shows an example function that is inside this package
whose only function is to show the Scope of a package function.
}
\details{
\code{searchFullEx} can be used to show an example of the use of \code{searchFull}.
}
\examples{
seeScope <- function() {
  searchFull()
}
seeScope()
searchFull()
searchFullEx()
}
\seealso{
\code{\link[base:search]{base::search()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{isInteractive}
\alias{isInteractive}
\title{Alternative to \code{interactive()} for unit testing}
\usage{
isInteractive()
}
\description{
This is a suggestion from
\url{https://github.com/MangoTheCat/blog-with-mock/blob/master/Blogpost1.Rmd}
as a way to test interactive code in unit tests. Basically, in the unit tests,
we use \code{testthat::with_mock}, and inside that we redefine \code{isInteractive}
just for the test. In all other times, this returns the same things as
\code{interactive()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/objectSize.R
\name{.objSizeWithTry}
\alias{.objSizeWithTry}
\title{\code{lobstr::obj_size} with a \code{try} to address issue #72}
\usage{
.objSizeWithTry(x, useTry = TRUE)
}
\arguments{
\item{x}{An object}

\item{useTry}{Logical. If \code{TRUE}, the default, then it will use \code{try}. Can optionally
avoid this if set to \code{FALSE}. The \code{try} takes sufficient extra compute time
that it is worth avoiding it if possible.}
}
\value{
The size of an object, using \code{lobstr::obj_size} or \code{object.size} if the
first fails
}
\description{
It is not clear why, but it appears that running \code{lobstr::obj_size} again, after
a bad binding error, it will work.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{archiveExtractBinary}
\alias{archiveExtractBinary}
\alias{.archiveExtractBinary}
\title{Tests if unrar or 7zip exist}
\usage{
.archiveExtractBinary(verbose = getOption("reproducible.verbose", 1))
}
\value{
unrar or 7zip path if exist, or NULL
}
\description{
Tests if unrar or 7zip exist
}
\author{
Tati Micheletti
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{unrarPath}
\alias{unrarPath}
\alias{.systemArchivePath}
\title{The known path for unrar or 7z}
\usage{
.systemArchivePath
}
\description{
The known path for unrar or 7z
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DBI.R
\name{usesPointer}
\alias{usesPointer}
\title{Does an object use a pointer?}
\usage{
usesPointer(x)
}
\arguments{
\item{x}{an object}
}
\value{
logical
}
\description{
Does an object use a pointer?
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GPT2.R
\name{matchCall2}
\alias{matchCall2}
\title{Remove \code{quote} and determine if call uses \code{...}}
\usage{
matchCall2(definition, call, envir, envir2 = parent.frame(), FUN)
}
\arguments{
\item{definition}{a function, by default the function from which
    \code{match.call} is called.  See details.}

\item{call}{an unevaluated call to the function specified by
    \code{definition}, as generated by \code{\link[base]{call}}.}

\item{envir}{an environment, from which the \code{\dots} in \code{call}
    are retrieved, if any.}

\item{envir2}{Environment. The environment where \code{matchCall2} was called.}

\item{FUN}{Either a function (e.g., \code{rnorm}), a function call (e.g., \code{rnorm(1)}),
or an unevaluated function call (e.g., using \code{quote()}).}
}
\value{
A named list with \code{call} (the original call, without \code{quote}),
\code{FUNorig}, the original value passed by user to \code{FUN}, and \code{usesDots} which
is a logical indicating whether the \code{...} are used.
}
\description{
Minor cleaning up of the \code{FUN} and \code{...} to be used subsequently. This does only very minor
things as it is run even if \code{useCache = FALSE}, i.e., even if the \code{Cache} is skipped.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{writeFuture}
\alias{writeFuture}
\title{Write to cache repository, using \code{future::future}}
\usage{
writeFuture(
  written,
  outputToSave,
  cachePath,
  userTags,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  cacheId,
  linkToCacheId = NULL,
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{written}{Integer. If zero or positive then it needs to be written still.
Should be 0 to start.}

\item{outputToSave}{The R object to save to repository}

\item{cachePath}{The file path of the repository}

\item{userTags}{Character string of tags to attach to this \code{outputToSave} in
the \code{CacheRepo}}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{cacheId}{Character string. If passed, this will override the calculated hash
of the inputs, and return the result from this \code{cacheId} in the \code{cachePath}.
Setting this is equivalent to manually saving the output of this function, i.e.,
the object will be on disk, and will be recovered in subsequent
This may help in some particularly finicky situations
where \code{Cache} is not correctly detecting unchanged inputs. This will guarantee
the object will be identical each time; this may be useful in operational code.}

\item{linkToCacheId}{Optional. If a \code{cacheId} is provided here, then a \code{file.link}
will be made to the file with that \code{cacheId} name in the cache repo.
This is used when identical outputs exist in the cache. This will save disk space.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\value{
Run for its side effect.
This will add the \code{objectToSave} to the cache located at \code{cachePath},
using \code{cacheId} as its id, while
updating the database entry. It will do this using the future package, so it is
written in a future.
}
\description{
This will be used internally if \code{options("reproducible.futurePlan" = TRUE)}.
This is still experimental.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DBI.R
\name{movedCache}
\alias{movedCache}
\title{Deal with moved cache issues}
\usage{
movedCache(
  new,
  old,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{new}{Either the path of the new \code{cachePath} where the cache was moved or copied to, or
the new DB Table Name}

\item{old}{Optional, if there is only one table in the \code{new} cache path.
Either the path of the previous \code{cachePath} where the cache was moved or copied from, or
the old DB Table Name}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\value{
\code{movedCache} does not return anything; it is called for its side effects.
}
\description{
If a user manually copies a complete Cache folder (including the db file and rasters folder),
there are issues that must be addressed, depending on the Cache backend used.
If using DBI (e.g., RSQLite or Postgres), the db table must be renamed. Run
this function after a manual copy of a cache folder. See examples for one way to do that.
}
\details{
When the backend database for a \code{reproducinle} cache is an SQL database, the files
on disk cannot be copied manually to a new location because they contain internal
tables. Because \code{reproducible} gives the main table a name based on the \code{cachePath}
path, calls to \code{Cache} will attempt to call this internally if it detects a
name mismatch.
}
\examples{
data.table::setDTthreads(2)
tmpdir <- "tmpdir"
tmpCache <- "tmpCache"
tmpCacheDir <- normalizePath(file.path(tempdir(), tmpCache), mustWork = FALSE)
tmpdirPath <- normalizePath(file.path(tempdir(), tmpdir), mustWork = FALSE)
bb <- Cache(rnorm, 1, cachePath = tmpCacheDir)

# Copy all files from tmpCache to tmpdir
froms <- normalizePath(dir(tmpCacheDir, recursive = TRUE, full.names = TRUE),
  mustWork = FALSE
)
dir.create(file.path(tmpdirPath, "rasters"), recursive = TRUE, showWarnings = FALSE)
dir.create(file.path(tmpdirPath, "cacheOutputs"), recursive = TRUE, showWarnings = FALSE)
file.copy(
  from = froms, overwrite = TRUE,
  to = gsub(tmpCache, tmpdir, froms)
)

# Can use 'movedCache' to update the database table, though will generally
#   happen automatically, with message indicating so
movedCache(new = tmpdirPath, old = tmpCacheDir)
bb <- Cache(rnorm, 1, cachePath = tmpdirPath) # should recover the previous call

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DBI.R
\name{loadFile}
\alias{loadFile}
\title{Load a file from the cache}
\usage{
loadFile(
  file,
  cacheId,
  cachePath,
  drv,
  conn,
  verbose = getOption("reproducible.verbose"),
  ...
)
}
\arguments{
\item{file}{\code{character(1)} specifying the path to the file to load.}

\item{cacheId}{\code{character(1)} the unique cache identifier of the object.}

\item{cachePath}{\code{character(1)} path to the cache directory.}

\item{drv}{A DBI driver object (e.g., \code{RSQLite::SQLite()}). Used when
the cache format needs to be swapped.}

\item{conn}{A DBI connection object. If \code{NULL}, a new connection is
created internally as needed.}

\item{verbose}{\code{integer} or \code{logical}. If \verb{> 0} or \code{TRUE}, print
informational messages. Defaults to \code{getOption("reproducible.verbose")}.}

\item{...}{Allows \code{format} for backward compatibility.}
}
\value{
the object loaded from \code{file}
}
\description{
Load a file from the cache
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache-helpers.R, R/exportedMethods.R
\name{.sortDotsUnderscoreFirst}
\alias{.sortDotsUnderscoreFirst}
\alias{.orderDotsUnderscoreFirst}
\alias{ExportedMethods}
\alias{.tagsByClass}
\alias{.tagsByClass,ANY-method}
\alias{.cacheMessage}
\alias{.cacheMessage,ANY-method}
\alias{.cacheMessageObjectToRetrieve}
\alias{.addTagsToOutput}
\alias{.addTagsToOutput,ANY-method}
\alias{.preDigestByClass}
\alias{.preDigestByClass,ANY-method}
\alias{.checkCacheRepo}
\alias{.checkCacheRepo,ANY-method}
\alias{.prepareOutput}
\alias{.prepareOutput,ANY-method}
\alias{.addChangedAttr}
\alias{.addChangedAttr,ANY-method}
\alias{updateFilenameSlots}
\alias{updateFilenameSlots.default}
\alias{updateFilenameSlots.list}
\alias{updateFilenameSlots.environment}
\alias{makeMemoisable}
\alias{makeMemoisable.default}
\alias{makeMemoisable.data.table}
\alias{unmakeMemoisable}
\alias{unmakeMemoisable.default}
\title{Exported generics and methods}
\usage{
.sortDotsUnderscoreFirst(obj)

.orderDotsUnderscoreFirst(obj)

.tagsByClass(object)

\S4method{.tagsByClass}{ANY}(object)

.cacheMessage(
  object,
  functionName,
  fromMemoise = getOption("reproducible.useMemoise", TRUE),
  verbose = getOption("reproducible.verbose", 1)
)

\S4method{.cacheMessage}{ANY}(
  object,
  functionName,
  fromMemoise = getOption("reproducible.useMemoise", TRUE),
  verbose = getOption("reproducible.verbose", 1)
)

.cacheMessageObjectToRetrieve(
  functionName,
  fullCacheTableForObj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
  verbose
)

.addTagsToOutput(object, outputObjects, FUN, preDigestByClass)

\S4method{.addTagsToOutput}{ANY}(object, outputObjects, FUN, preDigestByClass)

.preDigestByClass(object)

\S4method{.preDigestByClass}{ANY}(object)

.checkCacheRepo(
  object,
  create = FALSE,
  verbose = getOption("reproducible.verbose", 1)
)

\S4method{.checkCacheRepo}{ANY}(
  object,
  create = FALSE,
  verbose = getOption("reproducible.verbose", 1)
)

.prepareOutput(object, cachePath = getOption("reproducible.cachePath"), ...)

\S4method{.prepareOutput}{ANY}(object, cachePath = getOption("reproducible.cachePath"), ...)

.addChangedAttr(object, preDigest, origArguments, ...)

\S4method{.addChangedAttr}{ANY}(object, preDigest, origArguments, ...)

updateFilenameSlots(obj, curFilenames, newFilenames, isStack = NULL)

\method{updateFilenameSlots}{default}(obj, curFilenames, newFilenames, isStack = NULL, ...)

\method{updateFilenameSlots}{list}(obj, ...)

\method{updateFilenameSlots}{environment}(obj, ...)

makeMemoisable(x)

\method{makeMemoisable}{default}(x)

\method{makeMemoisable}{data.table}(x)

unmakeMemoisable(x)

\method{unmakeMemoisable}{default}(x)
}
\arguments{
\item{obj}{An object. This function only has useful methods for \verb{Raster*},
with all other classes being simply a pass-through}

\item{object}{Any R object returned from a function}

\item{functionName}{A character string indicating the function name}

\item{fromMemoise}{Logical. If \code{TRUE}, the message will be about
recovery from memoised copy}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{fullCacheTableForObj}{The data.table entry from the Cache database for only
this \code{cacheId}, e.g., via \code{showCache()}.}

\item{cachePath}{A repository used for storing cached objects.
This is optional if \code{Cache} is used inside a SpaDES module.}

\item{cacheId}{Character string. If passed, this will override the calculated hash
of the inputs, and return the result from this \code{cacheId} in the \code{cachePath}.
Setting this is equivalent to manually saving the output of this function, i.e.,
the object will be on disk, and will be recovered in subsequent
This may help in some particularly finicky situations
where \code{Cache} is not correctly detecting unchanged inputs. This will guarantee
the object will be identical each time; this may be useful in operational code.}

\item{cacheSaveFormat}{Character string: currently either \code{qs} or \code{rds}. Defaults to
\code{getOption("reproducible.cacheSaveFormat")}. \code{qs} may be faster but appears to have
narrower range of conditions that work; \code{rds} is safer, and may be slower.}

\item{outputObjects}{Optional character vector indicating which objects to
return. This is only relevant for list, environment (or similar) objects}

\item{FUN}{A function}

\item{preDigestByClass}{A list, usually from \code{.preDigestByClass}}

\item{create}{Logical. If TRUE, then it will create the path for cache.}

\item{...}{Anything passed to methods.}

\item{preDigest}{The full, element by element hash of the input arguments to that same function,
e.g., from \code{.robustDigest}}

\item{origArguments}{These are the actual arguments (i.e., the values, not the names) that
were the source for \code{preDigest}}

\item{curFilenames}{An optional character vector of filenames currently existing
and that are pointed to in the obj. If omitted, will take from the \code{obj}
using \code{Filenames(obj)}}

\item{newFilenames}{An optional character vector of filenames to use instead of
the curFilenames. This can also be a single directory, in which case the
renaming will be given:
\code{file.path(newFilenames, basename(Filenames(obj, allowMultiple = FALSE)))}}

\item{x}{An object to make memoisable.
See individual methods in other packages.}
}
\value{
\code{.sortDotsUnderscoreFirst}: the same object as \code{obj},
but sorted with dots and underscores first,
lower case before upper case.

\code{.tagsByClass} default method returns \code{NULL}.

\code{.cacheMessage}: nothing; called for its messaging side effect, which,
by default, just edits the name of the function into a generic "loaded cached result"
message.

\code{.addTagsToOutput}: The inputted object but with tags attached.

\code{.preDigestByClass}: A list with elements that have a difficult time
being digested correctly, e.g., an S4 object with some elements removed and
handled for digesting purposes. The default method for \code{preDigestByClass} and
simply returns \code{NULL}.

\code{.checkCacheRepo}: A character string with a path to a cache repository.

\code{.prepareOutput}: The object, modified

\code{.addChangedAttr}: the object, with an attribute ".Cache" and sub-element, "changed",
added set to either \code{TRUE} or \code{FALSE}

\code{updateFilenameSlots}: The original object, but with its internal file pointer
updated to the \code{newFilenames}

The same object, but with any modifications, especially
dealing with saving of environments, which memoising doesn't handle
correctly in some cases.
}
\description{
There are a number of generics that are exported for other packages to use.
These are listed below. They are not intended for use by normal users; rather,
they are made available for package developers to build specific methods.
}
\details{
\code{.sortDotsUnderscoreFirst}: This exists so Windows, Linux, and Mac machines can have
the same order after a sort. It will put dots and underscores first
(with the sort key based on their second character, see examples.
It also sorts lower case before upper case.

The methods should do as described below.

\code{.tagsByClass} should return a character vector, with a single colon (":") dividing
two parts: the tag type and tag value, for the specific class.

\code{.cacheMessage} should make a call to \code{message} that gives information about
the loaded cached object being returned.

\code{.objecxtToRetrieveMessage} is the messaging for recovering an object from Cache.

\code{.addTagsToOutput} should add one or more attributes to an object, named either
\code{"tags"}, \code{"call"} or \code{"function"}. It may be wise to do a "deep" copy within this
method, but it may not be necessary.

\code{.addChangedAttr} should return the same object, with this a very specific
attribute added: it must be named ".Cache", and have a sub-list element named
"changed", which must be a logical, \code{TRUE} or \code{FALSE}, to describe whether the object
has changed or not since last attempt to cache it. This is mostly useful when there
are only one or a few sub-elements of e.g., a large list, that are changed. \code{Cache}
will be able to only recover the changed parts, to reduce time required to
complete a call to \code{Cache}.

\code{updateFilenameSlots}: this exists because when copying file-backed rasters, the
usual mechanism of \code{writeRaster} can be very slow. This function allows
for a user to optionally create a hard link to the old file, give it a new
name, then update the filename slot(s) in the \verb{Raster*} class object. This
can be 100s of times faster for large rasters.

\code{makeMemoiseable} and \code{unmakeMemoisable} methods are run during \code{Cache}. The
methods should address any parts that will not successfully work for memoising,
most notably, \code{data.table} or other pass-by-reference objects likely need to
be deep copied, so that the memoised version doesn't get changed after being
stashed in the RAM cache (i.e., memoised)

\code{makeMemoiseable} and \code{unmakeMemoisable} are, by default, just a pass through
for most class. \code{reproducible} only has a method for \code{data.table} class objects.
}
\examples{
items <- c(A = "a", Z = "z", `.D` = ".d", `_C` = "_C")
.sortDotsUnderscoreFirst(items)

# dots & underscore (using 2nd character), then all lower then all upper
items <- c(B = "Upper", b = "lower", A = "a", `.D` = ".d", `_C` = "_C")
.sortDotsUnderscoreFirst(items)

# with a vector
.sortDotsUnderscoreFirst(c(".C", "_B", "A")) # _B is first

.tagsByClass(character())

a <- 1
.cacheMessage(a, "mean")

a <- 1
.preDigestByClass(a) # returns NULL in the simple case here.

a <- normalizePath(file.path(tempdir(), "test"), mustWork = FALSE)
.checkCacheRepo(a, create = TRUE)

a <- 1
.prepareOutput(a) # does nothing

b <- "NULL"
.prepareOutput(b) # converts to NULL

if (requireNamespace("terra", quietly = TRUE)) {
  r <- terra::rast(terra::ext(0, 10, 0, 10), vals = 1:100)

  # write to disk manually -- will be in tempdir()
  r <- terra::writeRaster(r, filename = tempfile(fileext = ".tif"))

  # copy it to the cache repository
  r <- .prepareOutput(r, tempdir())
}
a <- 1
.addChangedAttr(a) # does nothing because default method is just a pass through
}
\author{
Eliot McIntire
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache-helpers.R
\name{.prepareFileBackedRaster}
\alias{.prepareFileBackedRaster}
\title{Copy the file-backing of a file-backed Raster* object}
\usage{
.prepareFileBackedRaster(
  obj,
  repoDir = NULL,
  overwrite = FALSE,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  ...
)
}
\arguments{
\item{obj}{The raster object to save to the repository.}

\item{repoDir}{Character denoting an existing directory in which an artifact will be saved.}

\item{overwrite}{Logical. Should the raster be saved to disk, overwriting existing file.}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{Not used}
}
\value{
A raster object and its newly located file backing.
Note that if this is a legitimate Cache repository, the new location
will be a subdirectory called \file{rasters/} of \file{repoDir/}.
If this is not a repository, the new location will be within \code{repoDir}.
}
\description{
Rasters are sometimes file-based, so the normal save and copy and assign
mechanisms in R don't work for saving, copying and assigning.
This function creates an explicit file copy of the file that is backing the raster,
and changes the pointer (i.e., \code{filename(object)}) so that it is pointing
to the new file.
}
\author{
Eliot McIntire
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preProcess.R
\name{.purge}
\alias{.purge}
\title{Purge individual line items from checksums file}
\usage{
.purge(
  checkSums,
  purge,
  targetFile,
  archive,
  alsoExtract,
  url,
  destinationPath
)
}
\arguments{
\item{checkSums}{A checksums file, e.g., created by Checksums(..., write = TRUE)}

\item{purge}{Logical or Integer. \code{0/FALSE} (default) keeps existing \code{CHECKSUMS.txt} file and
\code{prepInputs} will write or append to it. \code{1/TRUE} will deleted the entire \code{CHECKSUMS.txt} file.
Other options, see details.}

\item{targetFile}{Character string giving the filename (without relative or
absolute path) to the eventual file
(raster, shapefile, csv, etc.) after downloading and extracting from a zip
or tar archive. This is the file \emph{before} it is passed to
\code{postProcess}. The internal checksumming does not checksum
the file after it is \code{postProcess}ed (e.g., cropped/reprojected/masked).
Using \code{Cache} around \code{prepInputs} will do a sufficient job in these cases.
See table in \code{\link[=preProcess]{preProcess()}}.}

\item{archive}{Optional character string giving the path of an archive
containing \code{targetFile}, or a vector giving a set of nested archives
(e.g., \code{c("xxx.tar", "inner.zip", "inner.rar")}). If there is/are (an) inner
archive(s), but they are unknown, the function will try all until it finds
the \code{targetFile}. See table in \code{\link[=preProcess]{preProcess()}}. If it is \code{NA},
then it will \emph{not} attempt to see it as an archive, even if it has archive-like
file extension (e.g., \code{.zip}). This may be useful when an R function
is expecting an archive directly.}

\item{alsoExtract}{Optional character string naming files other than
\code{targetFile} that must be extracted from the \code{archive}. If
\code{NULL}, the default, then it will extract all files. Other options:
\code{"similar"} will extract all files with the same filename without
file extension as \code{targetFile}. \code{NA} will extract nothing other
than \code{targetFile}. A character string of specific file names will cause
only those to be extracted. Each element may also be a regular expression:
if an element does not match any archive member literally (by relative path
or basename), it is passed to \code{grep()} against the archive's file list and
all matching members are extracted. For example,
\code{alsoExtract = "CMD_sm|CMD_sp"} extracts every file whose name contains
\code{CMD_sm} or \code{CMD_sp}. See table in \code{\link[=preProcess]{preProcess()}}.}

\item{url}{Optional character string indicating the URL to download from.
If not specified, then no download will be attempted. If not entry
exists in the \code{CHECKSUMS.txt} (in \code{destinationPath}), an entry
will be created or appended to. This \code{CHECKSUMS.txt} entry will be used
in subsequent calls to
\code{prepInputs} or \code{preProcess}, comparing the file on hand with the ad hoc
\code{CHECKSUMS.txt}. See table in \code{\link[=preProcess]{preProcess()}}.}

\item{destinationPath}{Character string of a directory in which to download
and save the file that comes from \code{url} and is also where the function
will look for \code{archive} or \code{targetFile}. NOTE (still experimental):
To prevent repeated downloads in different locations, the user can also set
\code{options("reproducible.destinationPathShared")} to one or more local file paths to
search for the file before attempting to download. Default for that option is
\code{NULL} meaning do not search locally. The previous name
\code{options("reproducible.inputPaths")} is still accepted as a backwards-compatible
alias.}
}
\description{
Purge individual line items from checksums file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DBI.R
\name{createCache}
\alias{createCache}
\alias{loadFromCache}
\alias{extractFromCache}
\alias{rmFromCache}
\alias{CacheDBFile}
\alias{CacheStorageDir}
\alias{CacheStoredFile}
\alias{CacheDBTableName}
\alias{CacheIsACache}
\title{Low-level functions to create and work with a cache}
\usage{
createCache(
  cachePath = getOption("reproducible.cachePath"),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  force = FALSE,
  verbose = getOption("reproducible.verbose")
)

loadFromCache(
  cachePath = getOption("reproducible.cachePath"),
  cacheId,
  preDigest,
  fullCacheTableForObj = NULL,
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat", .rdsFormat),
  .functionName = NULL,
  .dotsFromCache = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose")
)

extractFromCache(sc, elem, ifNot = NULL)

rmFromCache(
  cachePath = getOption("reproducible.cachePath"),
  cacheId,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat", .rdsFormat),
  verbose = getOption("reproducible.verbose"),
  ...
)

CacheDBFile(
  cachePath = getOption("reproducible.cachePath"),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL)
)

CacheStorageDir(cachePath = getOption("reproducible.cachePath"))

CacheStoredFile(
  cachePath = getOption("reproducible.cachePath"),
  cacheId,
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
  obj = NULL,
  readOnly = FALSE
)

CacheDBTableName(
  cachePath = getOption("reproducible.cachePath"),
  drv = getDrv(getOption("reproducible.drv", NULL))
)

CacheIsACache(
  cachePath = getOption("reproducible.cachePath"),
  create = FALSE,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{cachePath}{A path describing the directory in which to create
the database file(s)}

\item{drv}{A driver, passed to \code{dbConnect}}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{force}{Logical. Should it create a cache in the \code{cachePath},
even if it already exists, overwriting.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{cacheId}{The cacheId or otherwise digested hash value, as character string.}

\item{preDigest}{The list of \code{preDigest} that comes from \code{CacheDigest} of an object}

\item{fullCacheTableForObj}{The result of \code{showCache}, but subsetted for only
the \code{cacheId} being loaded or selected}

\item{cacheSaveFormat}{The text string representing the file extension used normally by
different save formats; currently only \code{"rds"} or \code{"qs"} (which now uses \code{qs2} package.
Defaults to \code{getOption("reproducible.cacheSaveFormat", "rds")}}

\item{.functionName}{Optional. Used for messaging when this function is called from \code{Cache}}

\item{.dotsFromCache}{Optional. Used internally.}

\item{sc}{a cache tags \code{data.table} object}

\item{elem}{character string specifying a \code{tagKey} value to match}

\item{ifNot}{character (or NULL) specifying the return value to use if \code{elem} not matched}

\item{...}{Arguments passed to \code{FUN}, if \code{FUN} is not an expression.}

\item{obj}{The optional object that is of interest; it may have an attribute "saveRawFile"
that would be important.}

\item{readOnly}{Logical. Only relevant during transition from \code{qs} to \code{qs2}.
Essentially, during transition, \code{qs} objects can be read, but not saved.
If \code{TRUE} then the \code{CacheStoredFile} can return a \code{.qs} file; if \code{FALSE},
then this will not be able to return \code{qs}; instead it will return \code{qs2}
files.}

\item{create}{Logical. Currently only affects non \pkg{RSQLite} default drivers.
If \code{TRUE} and there is no Cache database, the function will create one.}
}
\value{
\itemize{
\item \code{createCache()} returns \code{NULL} (invisibly) and intended to be called for side effects;
}

\itemize{
\item \code{loadFromCache()} returns the object from the cache that has the particular \code{cacheId};
}

\itemize{
\item \code{extractFromCache()} returns the \code{tagValue} from the cache corresponding to \code{elem} if found,
otherwise the value of \code{ifNot};
}

\itemize{
\item \code{rmFromCache()} returns \code{NULL} (invisibly) and is intended to be called for side effects;
}

\itemize{
\item \code{CacheDBFile()} returns the name of the database file for a given Cache,
when \code{useDBI() == FALSE}, or \code{NULL} if \code{TRUE};
\item \code{CacheDBFiles()} (i.e,. plural) returns the name of all the database files for
a given Cache when \code{useDBI() == TRUE}, or \code{NULL} if \code{FALSE};
\item \code{CacheStoredFile()} returns the file path to the file with the specified hash value,
This can be loaded to memory with e.g., \code{\link[=loadFile]{loadFile()}}.;
}

\itemize{
\item \code{CacheStorageDir()} returns the name of the directory where cached objects are stored;
}

\itemize{
\item \code{CacheStoredFile} returns the file path to the file with the specified hash value;
}

\itemize{
\item \code{CacheDBTableName()} returns the name of the table inside the SQL database, if that
is being used;
}

\itemize{
\item \code{CacheIsACache()} returns a logical indicating whether the \code{cachePath} is currently
a \code{reproducible} cache database;
}
}
\description{
\strong{These are intended for advanced use only.}
}
\details{
\itemize{
\item \code{createCache()} will create a Cache folder structure and necessary files, based on
the particular \code{drv} or \code{conn} provided;
}

\itemize{
\item \code{loadFromCache()} retrieves a single object from the cache, given its \code{cacheId};
}

\itemize{
\item \code{extractFromCache()} retrieves a single \code{tagValue} from the cache based on
the \code{tagKey} of \code{elem};
}

\itemize{
\item \code{rmFromCache()} removes one or more items from the cache, and updates the cache
database files.
}
}
\examples{
data.table::setDTthreads(2)
newCache <- tempdir2()
createCache(newCache)

out <- Cache(rnorm(1), cachePath = newCache)
cacheId <- gsub("cacheId:", "", attr(out, "tags"))
loadFromCache(newCache, cacheId = cacheId)

rmFromCache(newCache, cacheId = cacheId)

# clean up
unlink(newCache, recursive = TRUE)

data.table::setDTthreads(2)
newCache <- tempdir2()

# Given the drv and conn, creates the minimum infrastructure for a cache
createCache(newCache)

CacheDBFile(newCache) # identifies the database file
CacheStorageDir(newCache) # identifies the directory where cached objects are stored

out <- Cache(rnorm(1), cachePath = newCache)
cacheId <- gsub("cacheId:", "", attr(out, "tags"))
CacheStoredFile(newCache, cacheId = cacheId)

# The name of the table inside the SQL database
CacheDBTableName(newCache)

CacheIsACache(newCache) # returns TRUE

# clean up
unlink(newCache, recursive = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcess.R
\name{determineFilename}
\alias{determineFilename}
\title{Determine filename, either automatically or manually}
\usage{
determineFilename(
  filename2 = NULL,
  filename1 = NULL,
  destinationPath = getOption("reproducible.destinationPath", "."),
  verbose = getOption("reproducible.verbose", 1),
  prefix = "Small",
  ...
)
}
\arguments{
\item{filename2}{\code{filename2} is optional, and is either
NULL (no writing of outputs to disk), or several options
for writing the object to disk. If
\code{TRUE} (the default), it will give it a file name determined by
\code{.prefix(basename(filename1), prefix)}. If
a character string, it will use this as its file name. See
\code{\link[=determineFilename]{determineFilename()}}.}

\item{filename1}{Character strings giving the file paths of
the \emph{input} object (\code{filename1}) \code{filename1}
is only used for messaging (i.e., the object itself is passed
in as \code{x}) and possibly naming of output (see details
and \code{filename2}).}

\item{destinationPath}{Optional. If \code{filename2} is a relative file path, then this
will be the directory of the resulting absolute file path.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{prefix}{The character string to prepend to \code{filename1}, if \code{filename2}
not provided.}

\item{...}{Passed into \code{\link[=writeTo]{writeTo()}}}
}
\description{
Determine the filename, given various combinations of inputs.
}
\details{
The post processing workflow, which includes this function,
addresses several scenarios, and depending on which scenario, there are
several file names at play. For example, \code{Raster} objects may have
file-backed data, and so \emph{possess a file name}, whereas \code{Spatial}
objects do not. Also, if post processing is part of a \code{\link[=prepInputs]{prepInputs()}}
workflow, there will always be a file downloaded. From the perspective of
\code{postProcess}, these are the "inputs" or \code{filename1}.
Similarly, there may or may not be a desire to write an
object to disk after all post processing, \code{filename2}.

This subtlety means that there are two file names that may be at play:
the "input" file name (\code{filename1}), and the "output" filename (\code{filename2}).
When this is used within \code{postProcess}, it is straight forward.

However, when \code{postProcess} is used within a \code{prepInputs} call,
the \code{filename1} file is the file name of the downloaded file (usually
automatically known following the downloading, and refered to as \code{targetFile})
and the \code{filename2} is the file name of the of post-processed file.

If \code{filename2} is \code{TRUE}, i.e., not an actual file name, then the cropped/masked
raster will be written to disk with the original \code{filenam1/targetFile}
name, with \code{prefix} prefixed to the basename(\code{targetFile}).

If \code{filename2} is a character string, it will be the path of the saved/written
object e.g., passed to \code{writeOutput}. It will be tested whether it is an
absolute or relative path and used as is if absolute or
prepended with \code{destinationPath} if relative.

If \code{filename2} is \code{logical}, then the output
filename will be \code{prefix} prefixed to the basename(\code{filename1}).
If a character string, it
will be the path returned. It will be tested whether it is an
absolute or relative path and used as is if absolute or prepended with
\code{destinationPath} if provided, and if \code{filename2} is relative.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GPT2.R
\name{harmonizeCall}
\alias{harmonizeCall}
\title{Harmonize all forms of call}
\usage{
harmonizeCall(callList, .callingEnv, .functionName = NULL)
}
\arguments{
\item{callList}{A named list with elements \code{call}, \code{usesDots} and \code{FUNorig}}

\item{.callingEnv}{The calling environment where \code{Cache} was called from}

\item{.functionName}{A possible function name. If omitted, then it will be deduced
from the \code{callList} and may be inaccurate.}
}
\value{
A named list. We illustrate with the example \code{rnorm(1)}. The named
list will have the original \code{callList} (\code{call} (the original call, without \code{quote}),
\code{FUNorig}, the original value passed by user to \code{FUN}, and \code{usesDots} which
is a logical indicating whether the \code{...} are used), and appended with \code{new_call}
(the harmonized call, with the function and arguments evaluated, e.g.,
\code{(function (n, mean = 0, sd = 1) .Call(C_rnorm, n, mean, sd))(1)}), \code{func_call}, the same harmonized call
with neither function nor arguments not evaluated (e.g., \code{rnorm(1)}), \code{func} which
will be function or method definition
\code{function (n, mean = 0, sd = 1) .Call(C_rnorm, n, mean, sd)},
and \code{.functionName}, which will be the function name as a character string (\code{rnorm})
either directly passed from the user's \code{.functionName} or deduced from the \code{func_call}.
}
\description{
This will convert all known (imagined) calls so that they have the same canonical
format i.e., \code{rnorm(n = 1, mean = 0, sd = 1)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{compareNA}
\alias{compareNA}
\title{\code{NA}-aware comparison of two vectors}
\usage{
compareNA(v1, v2)
}
\arguments{
\item{v1}{A vector}

\item{v2}{A vector}
}
\value{
A logical vector, indicating positions where two vectors are same or differ.
}
\description{
Copied from
\url{http://www.cookbook-r.com/Manipulating_data/Comparing_vectors_or_factors_with_NA/}.
This function returns \code{TRUE} wherever elements are the same, including \code{NA}'s,
and \code{FALSE} everywhere else.
}
\examples{
a <- c(NA, 1, 2, NA)
b <- c(1, NA, 2, NA)
compareNA(a, b)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checksums.R
\name{Checksums}
\alias{Checksums}
\alias{Checksums,character,logical-method}
\alias{Checksums,character,missing-method}
\title{Calculate checksum}
\usage{
Checksums(
  path,
  write,
  quickCheck = getOption("reproducible.quickCheck", FALSE),
  checksumFile = identifyCHECKSUMStxtFile(path),
  files = NULL,
  verbose = getOption("reproducible.verbose", 1),
  ...
)

\S4method{Checksums}{character,logical}(
  path,
  write,
  quickCheck = getOption("reproducible.quickCheck", FALSE),
  checksumFile = identifyCHECKSUMStxtFile(path),
  files = NULL,
  verbose = getOption("reproducible.verbose", 1),
  ...
)

\S4method{Checksums}{character,missing}(
  path,
  write,
  quickCheck = getOption("reproducible.quickCheck", FALSE),
  checksumFile = identifyCHECKSUMStxtFile(path),
  files = NULL,
  verbose = getOption("reproducible.verbose", 1),
  ...
)
}
\arguments{
\item{path}{Character string giving the directory path containing \code{CHECKSUMS.txt}
file, or where it will be written if \code{checksumFile = TRUE}.}

\item{write}{Logical indicating whether to overwrite \code{CHECKSUMS.txt}.
Default is \code{FALSE}, as users should not change this file.
Module developers should write this file prior to distributing
their module code, and update accordingly when the data change.}

\item{quickCheck}{Logical. If \code{TRUE}, then this will only use file sizes,
rather than a digest::digest hash. This is generally faster,
but will be \emph{much} less robust.}

\item{checksumFile}{The filename of the checksums file to read or write to.
The default is \file{CHECKSUMS.txt} located at
\code{file.path(path, module, "data", checksumFile)}.
It is likely not a good idea to change this, and should
only be used in cases such as \code{Cache}, which can
evaluate if the \code{checksumFile} has changed.}

\item{files}{An optional character string or vector of specific files to checksum.
This may be very important if there are many files listed in a
\code{CHECKSUMS.txt} file, but only a few are to be checksummed.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{Passed to \code{\link[digest:digest]{digest::digest()}} and \code{\link[utils:write.table]{utils::write.table()}}.
For \code{digest}, the notable argument is \code{algo}. For \code{write.table},
the notable argument is \code{append}.}
}
\value{
A \code{data.table} with columns: \code{result}, \code{expectedFile},
\code{actualFile}, \code{checksum.x}, \code{checksum.y},
\code{algorithm.x}, \code{algorithm.y}, \code{filesize.x}, \code{filesize.y}
indicating the result of comparison between local file (\code{x}) and
expectation based on the \code{CHECKSUMS.txt} file.
}
\description{
Verify (and optionally write) checksums.
Checksums are computed using \code{\link[=.digest]{.digest()}}, which is simply a
wrapper around \code{digest::digest}.
}
\note{
In version 1.2.0 and earlier, two checksums per file were required
because of differences in the checksum hash values on Windows and Unix-like
platforms. Recent versions use a different (faster) algorithm and only require
one checksum value per file.
To update your \file{CHECKSUMS.txt} files using the new algorithm, see
\url{https://github.com/PredictiveEcology/SpaDES/issues/295#issuecomment-246513405}.
}
\examples{
\dontrun{
modulePath <- file.path(tempdir(), "myModulePath")
dir.create(modulePath, recursive = TRUE, showWarnings = FALSE)
moduleName <- "myModule"
cat("hi", file = file.path(modulePath, moduleName)) # put something there for this example

## verify checksums of all data files
Checksums(modulePath, files = moduleName)

## write new CHECKSUMS.txt file
Checksums(files = moduleName, modulePath, write = TRUE)
}

}
\author{
Alex Chubaty
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{rasterRead}
\alias{rasterRead}
\title{A helper to \code{getOption("reproducible.rasterRead")}}
\usage{
rasterRead(...)
}
\arguments{
\item{...}{Passed to the function parsed and evaluated from
\code{getOption("reproducible.rasterRead")}}
}
\value{
A function, that will be the evaluated, parsed character
string, e.g., \code{eval(parse(text = "terra::rast"))}
}
\description{
A helper to \code{getOption("reproducible.rasterRead")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{testForArchiveExtract}
\alias{testForArchiveExtract}
\alias{.testForArchiveExtract}
\title{Returns \code{unrar} path and creates a shortcut as \code{.systemArchivePath}.
Was not incorporated in previous function so it can be used in the tests.}
\usage{
.testForArchiveExtract(archive = "")
}
\value{
path to \code{unrar} or \code{7-zip} if it exists, and assign it to \code{.systemArchivePath}.
Advise the user if not found..
}
\description{
Returns \code{unrar} path and creates a shortcut as \code{.systemArchivePath}.
Was not incorporated in previous function so it can be used in the tests.
}
\author{
Tati Micheletti
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{.listFilesInArchive}
\alias{.listFilesInArchive}
\title{List files in either a \code{.zip} or or \code{.tar} file}
\usage{
.listFilesInArchive(archive, dontUse = NULL)
}
\arguments{
\item{archive}{A character string of a single file name to list files in.}

\item{dontUse}{A character string or vector of unzipping tool(s) to not use,
e.g., one or more of: \code{c("archive", "unzip", "7z")}}
}
\value{
A character string of all files in the archive.
}
\description{
Makes the outputs from\verb{.tar``.zip} the same, which they aren't by default.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{internetExists}
\alias{internetExists}
\alias{urlExists}
\title{Checks for existed of a url or the internet using \url{https://CRAN.R-project.org}}
\usage{
internetExists()

urlExists(url)
}
\arguments{
\item{url}{A url of the form \verb{https://...} to test for existence.}
}
\value{
Logical, \code{TRUE} if internet site exists, \code{FALSE} otherwise

Logical, \code{TRUE} if internet site exists, \code{FALSE} otherwise.
}
\description{
A lightweight function that may be less reliable than more purpose built solutions
such as checking a specific web page using \code{RCurl::url.exists}. However, this is
slightly faster and is sufficient for many uses.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{.requireNamespace}
\alias{.requireNamespace}
\title{Provide standard messaging for missing package dependencies}
\usage{
.requireNamespace(
  pkg = "methods",
  minVersion = NULL,
  stopOnFALSE = FALSE,
  messageStart = NULL
)
}
\arguments{
\item{pkg}{Character string indicating name of package required}

\item{minVersion}{Character string indicating minimum version of package
that is needed}

\item{stopOnFALSE}{Logical. If \code{TRUE}, this function will create an
error (i.e., \code{stop}) if the function returns \code{FALSE}; otherwise
it simply returns \code{FALSE}}

\item{messageStart}{A character string with a prefix of message to provide}
}
\value{
A logical or stop if the namespace is not available to be loaded.
}
\description{
This provides a standard message format for missing packages, e.g.,
detected via \code{requireNamespace}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/showCacheEtc.R
\name{showCache}
\alias{showCache}
\alias{clearCache}
\alias{clearCache,ANY-method}
\alias{cc}
\alias{showCache,ANY-method}
\alias{keepCache}
\alias{keepCache,ANY-method}
\title{Examining and modifying the cache}
\usage{
clearCache(
  x,
  userTags = character(),
  after = NULL,
  before = NULL,
  fun = NULL,
  cacheId = NULL,
  ask = getOption("reproducible.ask"),
  useCloud = FALSE,
  cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  ...
)

\S4method{clearCache}{ANY}(
  x,
  userTags = character(),
  after = NULL,
  before = NULL,
  fun = NULL,
  cacheId = NULL,
  ask = getOption("reproducible.ask"),
  useCloud = FALSE,
  cloudFolderID = getOption("reproducible.cloudFolderID", NULL),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  ...
)

cc(secs, ..., verbose = getOption("reproducible.verbose"))

showCache(
  x,
  userTags = character(),
  after = NULL,
  before = NULL,
  fun = NULL,
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  ...
)

\S4method{showCache}{ANY}(
  x,
  userTags = character(),
  after = NULL,
  before = NULL,
  fun = NULL,
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  ...
)

keepCache(
  x,
  userTags = character(),
  after = NULL,
  before = NULL,
  ask = getOption("reproducible.ask"),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  ...
)

\S4method{keepCache}{ANY}(
  x,
  userTags = character(),
  after = NULL,
  before = NULL,
  ask = getOption("reproducible.ask"),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  ...
)
}
\arguments{
\item{x}{A simList or a directory containing a valid Cache repository. Note:
For compatibility with \code{Cache} argument, \code{cachePath} can also be
used instead of \code{x}, though \code{x} will take precedence.}

\item{userTags}{Character vector. If used, this will be used in place of the
\code{after} and \code{before}.
Specifying one or more \code{userTag} here will clear all
objects that match those tags.
Matching is via regular expression, meaning partial matches
will work unless strict beginning (\code{^}) and end (\code{$}) of string
characters are used.
Matching will be against any of the 3 columns returned by \code{showCache()},
i.e., \code{artifact}, \code{tagValue} or \code{tagName}.
Also, if \code{length(userTags) > 1}, then matching is by \code{and}.
For \code{or} matching, use \code{|} in a single character string.
See examples.}

\item{after}{A time (POSIX, character understandable by data.table).
Objects cached after this time will be shown or deleted.}

\item{before}{A time (POSIX, character understandable by data.table).
Objects cached before this time will be shown or deleted.}

\item{fun}{An optional character vector describing the function name to extract.
Only functions with this/these functions will be returned.}

\item{cacheId}{An optional character vector describing the \code{cacheId}s to extract.
Only entries with this/these \code{cacheId}s will be returned. If \code{useDBI(FALSE)},
this will also be dramatically faster than using \code{userTags}, for a large
cache.}

\item{ask}{Logical. If \code{FALSE}, then it will not ask to confirm deletions using
\code{clearCache} or \code{keepCache}. Default is \code{TRUE}}

\item{useCloud}{Logical. If \code{TRUE}, then every object that is deleted locally will
also be deleted in the \code{cloudFolderID}, if it is non-\code{NULL}}

\item{cloudFolderID}{A googledrive dribble of a folder, e.g., using \code{drive_mkdir()}.
If left as \code{NULL}, the function will create a cloud folder with name from last
two folder levels of the \code{cachePath} path, :
\code{paste0(basename(dirname(cachePath)), "_", basename(cachePath))}.
This \code{cloudFolderID} will be added to \code{options("reproducible.cloudFolderID")},
but this will not persist across sessions. If this is a character string, it will
treat this as a folder name to create or use on GoogleDrive.}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{Other arguments. Can be in the form of \code{tagKey = tagValue}, such as,
\code{class = "numeric"} to find all entries that are numerics in the cache.
Note: the special cases of \code{cacheId} and \code{fun} have their own
named arguments in these functions.
Also can be \code{regexp = xx}, where \code{xx} is \code{TRUE} if the user
is passing a regular expression.
Otherwise, \code{userTags} will need to be exact matches. Default is
missing, which is the same as \code{TRUE}. If there are errors due
to regular expression problem, try \code{FALSE}. For \code{cc}, it is
passed to \code{clearCache}, e.g., \code{ask}, \code{userTags}. For \code{showCache},
it can also be \code{sorted = FALSE} to return the object unsorted.}

\item{secs}{Currently 3 options: the number of seconds to pass to \code{clearCache(after = secs)},
a \code{POSIXct} time e.g., from \code{Sys.time()}, or missing. If missing,
the default, then it will delete the most recent entry in the Cache.}
}
\value{
Will clear all objects (or those that match \code{userTags}, or those
between \code{after} or \code{before}) from the repository located in
\code{cachePath}.
Invisibly returns a \code{data.table} of the removed items.
}
\description{
These are convenience wrappers around \code{DBI} package functions.
They allow the user a bit of control over what is being cached.
}
\details{
If neither \code{after} or \code{before} are provided, nor \code{userTags},
then all objects will be removed.
If both \code{after} and \code{before} are specified, then all objects between
\code{after} and \code{before} will be deleted.
If \code{userTags} is used, this will override \code{after} or \code{before}.

\code{cc(secs)} is just a shortcut for \code{clearCache(repo = currentRepo, after = secs)},
i.e., to remove any cache entries touched in the last \code{secs} seconds. Since, \code{secs}
can be missing, this is also be a shorthand for "remove most recent entry from
the cache".

\describe{
\item{\code{clearCache}}{remove items from the cache based on their
\code{userTag} or \code{times} values.}
\item{\code{keepCache}}{remove all cached items \emph{except} those based on
certain \code{userTags} or \code{times} values.}
\item{\code{showCache}}{display the contents of the cache.}
}

By default the return of \code{showCache} is sorted by \code{cacheId}. For convenience,
a user can optionally have it unsorted (passing \code{sorted = FALSE}),
which may be noticeably faster when
the cache is large (\verb{> 1e4} entries).
}
\note{
If the cache is larger than 10MB, and clearCache is used, there will be
a message and a pause, if interactive, to prevent accidentally deleting of a
large cache repository.
}
\examples{
data.table::setDTthreads(2)

tmpDir <- file.path(tempdir(), "reproducible_examples", "Cache")
try(clearCache(tmpDir, ask = FALSE), silent = TRUE) # just to make sure it is clear

# Basic use
ranNumsA <- Cache(rnorm, 10, 16, cachePath = tmpDir)

# All same
ranNumsB <- Cache(rnorm, 10, 16, cachePath = tmpDir) # recovers cached copy
ranNumsD <- Cache(quote(rnorm(n = 10, 16)), cachePath = tmpDir) # recovers cached copy

# Any minor change makes it different
ranNumsE <- Cache(rnorm, 10, 6, cachePath = tmpDir) # different

## Example 1: basic cache use with tags
ranNumsA <- Cache(rnorm, 4, cachePath = tmpDir, userTags = "objectName:a")
ranNumsB <- Cache(runif, 4, cachePath = tmpDir, userTags = "objectName:b")
ranNumsC <- Cache(runif, 40, cachePath = tmpDir, userTags = "objectName:b")

showCache(tmpDir, userTags = c("objectName"))
showCache(tmpDir, userTags = c("^a$")) # regular expression ... "a" exactly

# Fine control of cache elements -- pick out only the large runif object, and remove it
cache1 <- showCache(tmpDir, userTags = c("runif")) # show only cached objects made during runif
toRemove <- cache1[tagKey == "object.size"][as.numeric(tagValue) > 700]$cacheId
clearCache(tmpDir, userTags = toRemove, ask = FALSE)
cacheAfter <- showCache(tmpDir, userTags = c("runif")) # Only the small one is left

data.table::setDTthreads(2)
tmpDir <- file.path(tempdir(), "reproducible_examples", "Cache")
try(clearCache(tmpDir, ask = FALSE), silent = TRUE) # just to make sure it is clear

Cache(rnorm, 1, cachePath = tmpDir)
thisTime <- Sys.time()
Cache(rnorm, 2, cachePath = tmpDir)
Cache(rnorm, 3, cachePath = tmpDir)
Cache(rnorm, 4, cachePath = tmpDir)
showCache(x = tmpDir) # shows all 4 entries
cc(ask = FALSE, x = tmpDir)
showCache(x = tmpDir) # most recent is gone
cc(thisTime, ask = FALSE, x = tmpDir)
showCache(x = tmpDir) # all those after thisTime gone, i.e., only 1 left
cc(ask = FALSE, x = tmpDir) # Cache is
cc(ask = FALSE, x = tmpDir) # Cache is already empty
}
\seealso{
\code{\link[=mergeCache]{mergeCache()}}. Many more examples in \code{\link[=Cache]{Cache()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{studyAreaName}
\alias{studyAreaName}
\alias{studyAreaName,character-method}
\alias{studyAreaName,ANY-method}
\title{Get a unique name for a given study area}
\usage{
studyAreaName(studyArea, ...)

\S4method{studyAreaName}{character}(studyArea, ...)

\S4method{studyAreaName}{ANY}(studyArea, ...)
}
\arguments{
\item{studyArea}{Spatial object.}

\item{...}{Other arguments (not currently used)}
}
\value{
A character string using the \code{.robustDigest} of the \code{studyArea}. This is only intended
for use with spatial objects.
}
\description{
Digest a spatial object to get a unique character string (hash) of the study area.
Use \code{.suffix()} to append the hash to a filename,
e.g., when using \code{filename2} in \code{prepInputs}.
}
\examples{
studyAreaName("Ontario")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preProcess.R
\name{linkOrCopy}
\alias{linkOrCopy}
\title{Hardlink, symlink, or copy a file}
\usage{
linkOrCopy(
  from,
  to,
  symlink = TRUE,
  overwrite = TRUE,
  verbose = getOption("reproducible.verbose", 1)
)
}
\arguments{
\item{from, to}{Character vectors, containing file names or paths.
\code{to} can alternatively be the path to a single existing directory.}

\item{symlink}{Logical indicating whether to use symlink (instead of hardlink).
Default \code{FALSE}.}

\item{overwrite}{Logical. Passed to \code{writeTo} (possibly inside \code{postProcess}) and \code{postProcess}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\value{
This function is called for its side effects, which will be a \code{file.link} is that
is available or \code{file.copy} if not (e.g., the two directories are not on the
same physical disk).
}
\description{
Attempt first to make a hardlink. If that fails, try to make
a symlink (on non-windows systems and \code{symlink = TRUE}).
If that fails, copy the file.
}
\note{
Use caution with files-backed objects (e.g., rasters). See examples.
}
\examples{

tmpDir <- file.path(tempdir(), "symlink-test")
tmpDir <- normalizePath(tmpDir, winslash = "/", mustWork = FALSE)
dir.create(tmpDir)

f0 <- file.path(tmpDir, "file0.csv")
write.csv(iris, f0)

d1 <- file.path(tmpDir, "dir1")
dir.create(d1)
write.csv(iris, file.path(d1, "file1.csv"))

d2 <- file.path(tmpDir, "dir2")
dir.create(d2)
f2 <- file.path(tmpDir, "file2.csv")

## create link to a file
linkOrCopy(f0, f2)
file.exists(f2) ## TRUE
identical(read.table(f0), read.table(f2)) ## TRUE

## deleting the link shouldn't delete the original file
unlink(f0)
file.exists(f0) ## FALSE
file.exists(f2) ## TRUE

if (requireNamespace("terra", quietly = TRUE)) {
  ## using spatRasters and other file-backed objects
  f3a <- system.file("ex/test.grd", package = "terra")
  f3b <- system.file("ex/test.gri", package = "terra")
  r3a <- terra::rast(f3a)
  f4a <- file.path(tmpDir, "raster4.grd")
  f4b <- file.path(tmpDir, "raster4.gri")
  linkOrCopy(f3a, f4a) ## hardlink
  linkOrCopy(f3b, f4b) ## hardlink
  r4a <- terra::rast(f4a)

  isTRUE(all.equal(r3a, r4a)) # TRUE

  ## cleanup
  unlink(tmpDir, recursive = TRUE)
}
}
\seealso{
\code{\link[=file.link]{file.link()}}, \code{\link[=file.symlink]{file.symlink()}}, \code{\link[=file.copy]{file.copy()}}.
}
\author{
Alex Chubaty and Eliot McIntire
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preProcess.R
\name{.expandAlsoExtractPatterns}
\alias{.expandAlsoExtractPatterns}
\title{Expand regex patterns in \code{alsoExtract} against a known list of archive files.}
\usage{
.expandAlsoExtractPatterns(alsoExtract, filesInArchive)
}
\arguments{
\item{alsoExtract}{Character vector (or \code{NULL}).}

\item{filesInArchive}{Character vector of files inside the archive, as
returned by \code{makeRelative(.listFilesInArchive(archive), destinationPath)}.}
}
\value{
A character vector with pattern elements replaced by their matches.
}
\description{
For each element of \code{alsoExtract} that is not a literal match in
\code{filesInArchive} (by relative path or basename), the element is treated as a
regular expression and \code{grep()}-ed against \code{filesInArchive}.  Matched names
replace the pattern; unmatched patterns are kept as-is so that downstream
code can emit a useful "file not found" error.  Special sentinel values
(\code{"similar"}, \code{"none"}, \code{NA}) are passed through unchanged.
}
\details{
This lets users write e.g.
\code{alsoExtract = "CMD_sm|CMD_sp"} to select all archive members whose name
contains \code{CMD_sm} or \code{CMD_sp}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{dlGeneric}
\alias{dlGeneric}
\title{Download file from generic source url}
\usage{
dlGeneric(url, destinationPath, verbose = getOption("reproducible.verbose", 1))
}
\arguments{
\item{url}{The url (link) to the file.}

\item{destinationPath}{Character string of a directory in which to download
and save the file that comes from \code{url} and is also where the function
will look for \code{archive} or \code{targetFile}. NOTE (still experimental):
To prevent repeated downloads in different locations, the user can also set
\code{options("reproducible.destinationPathShared")} to one or more local file paths to
search for the file before attempting to download. Default for that option is
\code{NULL} meaning do not search locally. The previous name
\code{options("reproducible.inputPaths")} is still accepted as a backwards-compatible
alias.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\description{
Download file from generic source url
}
\author{
Eliot McIntire and Alex Chubaty
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/robustDigest.R
\name{.removeCacheAtts}
\alias{.removeCacheAtts}
\title{Remove attributes that are highly varying}
\usage{
.removeCacheAtts(x)
}
\arguments{
\item{x}{Any arbitrary R object that could have attributes}
}
\description{
Remove attributes that are highly varying
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reproducible-package.R
\docType{package}
\name{reproducible-package}
\alias{reproducible}
\alias{reproducible-package}
\title{The \code{reproducible} package}
\description{
\if{html}{\figure{lifecycle-maturing.svg}{options: alt="maturing"}}

This package aims at making
high-level, robust, machine and OS independent tools for making deeply
reproducible and reusable content in R. The core user functions are \code{Cache}
and \code{prepInputs}. Each of these is built around many core and edge cases
required to have reproducible code of arbitrary complexity.
}
\section{Main Tools}{


There are many elements within the reproducible package.
However, there are currently two main ones that are critical for reproducible research.
The key element for reproducible research is that the code must always return the same content
every time it is run, but it must be vastly faster the 2nd, 3rd, 4th etc, time it is run.
That way, the entire code sequence for a project of arbitrary size can be run
\emph{from the start} every time.

\describe{
\item{\code{\link[=Cache]{Cache()}}:}{A robust wrapper for any function, including those with environments,
disk-backed storage (currently on \code{Raster}) class), operating-system independent,
whose first time called will execute the function, second time will compare the inputs to a
database of entries, and recover the first result if inputs are identical.
If \code{options("reproducible.useMemoise" = TRUE)}, the second time will be very fast as it
will recover the answer from RAM.}
\item{\code{\link[=prepInputs]{prepInputs()}}for other specifics for other classes.:}{ Download, or load objects, and possibly post-process them.
The main advantage to using this over more direct routes is that it will automatically build
checksums tables, use \code{Cache} internally where helpful, and possibly run a variety of
post-processing actions.
This means this function can also itself be cached for even more speed.
This allows all project data to be stored in custom cloud locations or in their original online
data repositories, \verb{without altering code} between the first, second, third, etc., times
the code is run.}
}
}

\section{Package options}{


See \code{\link[=reproducibleOptions]{reproducibleOptions()}} for a complete description of package
\code{\link[=options]{options()}} to configure behaviour.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://reproducible.predictiveecology.org}
  \item \url{https://github.com/PredictiveEcology/reproducible}
  \item Report bugs at \url{https://github.com/PredictiveEcology/reproducible/issues}
}

}
\author{
\strong{Maintainer}: Eliot J B McIntire \email{eliot.mcintire@canada.ca} (\href{https://orcid.org/0000-0002-6914-8316}{ORCID})

Authors:
\itemize{
  \item Eliot J B McIntire \email{eliot.mcintire@canada.ca} (\href{https://orcid.org/0000-0002-6914-8316}{ORCID})
  \item Alex M Chubaty \email{achubaty@for-cast.ca} (\href{https://orcid.org/0000-0001-7146-8135}{ORCID})
}

Other contributors:
\itemize{
  \item Tati Micheletti \email{tati.micheletti@gmail.com} (\href{https://orcid.org/0000-0003-4838-8342}{ORCID}) [contributor]
  \item Ceres Barros \email{ceres.barros@ubc.ca} (\href{https://orcid.org/0000-0003-4036-977X}{ORCID}) [contributor]
  \item Ian Eddy \email{ian.eddy@nrcan-rncan.gc.ca} (\href{https://orcid.org/0000-0001-7397-2116}{ORCID}) [contributor]
  \item His Majesty the King in Right of Canada, as represented by the Minister of Natural Resources Canada [copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcessTo.R
\name{.isGridded}
\alias{.isGridded}
\alias{isSpatial}
\alias{.isVector}
\alias{.isSF}
\alias{.isSpat}
\alias{.isSpatialAny}
\alias{.isCRSany}
\title{Some spatial helper functions}
\usage{
.isGridded(x)

.isVector(x)

.isSF(x)

.isSpat(x)

.isSpatialAny(x)

.isCRSany(x)
}
\arguments{
\item{x}{A spatial object.}
}
\value{
Logical.
}
\description{
Some spatial helper functions
}
\details{
\code{.isGridded} returns \code{TRUE} if the object is a \code{SpatRaster} or \code{Raster}

\code{.isVector} returns \code{TRUE} if the object is \code{SpatVector}, \code{spatial} or \code{sf}

\code{.isSF} returns \code{TRUE} if the object is \code{sf} or \code{sfc}

\code{.isSpat} returns \code{TRUE} if the object is \code{SpatVector} or \code{SpatRaster}

\code{.isSpatialAny} returns \code{TRUE} if the object returns \code{TRUE} for \code{.isGridded} or
\code{.isVector}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{.formalsNotInCurrentDots}
\alias{.formalsNotInCurrentDots}
\title{Identify which formals to a function are not in the current \code{...}}
\usage{
.formalsNotInCurrentDots(fun, ..., dots, formalNames, signature = character())
}
\arguments{
\item{fun}{A function}

\item{...}{The ... from inside a function. Will be ignored if \code{dots} is
provided explicitly.}

\item{dots}{Optional. If this is provided via say \code{dots = list(...)},
then this will cause the \code{...} to be ignored.}

\item{formalNames}{Optional character vector. If provided then it will override the \code{fun}}
}
\value{
A list of the formals of the \code{fun} that are missing from the \code{...} or \code{dots}.
}
\description{
Advanced use.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DBI.R
\name{.addTagsRepo}
\alias{.addTagsRepo}
\alias{.updateTagsRepo}
\title{Add a Tag to a Cached Object in the Repository}
\usage{
.addTagsRepo(
  cacheId,
  cachePath = getOption("reproducible.cachePath"),
  tagKey = character(),
  tagValue = character(),
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose")
)

.updateTagsRepo(
  cacheId,
  cachePath = getOption("reproducible.cachePath"),
  tagKey = character(),
  tagValue = character(),
  add = TRUE,
  cacheSaveFormat = getOption("reproducible.cacheSaveFormat"),
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{cacheId}{\code{character(1)}
Unique identifier of the cached object. Must be of length 1.}

\item{cachePath}{\code{character(1)}
Path to the cache directory. Defaults to \code{getOption("reproducible.cachePath")}.}

\item{tagKey}{\code{character(1)}
The key for the tag. Must be supplied.}

\item{tagValue}{\code{character(1)}
The new value for the tag. Must be supplied.}

\item{cacheSaveFormat}{\code{character(1)}
Format used for saving cache files. Defaults to \code{getOption("reproducible.cacheSaveFormat")}.}

\item{drv}{A DBI driver object. Defaults to \code{getDrv(getOption("reproducible.drv", NULL))}.}

\item{conn}{A DBI connection object. If \code{NULL}, a new connection is created internally.}

\item{verbose}{\code{integer} or \code{logical}. If \verb{> 0} or \code{TRUE}, print informational messages.
Defaults to \code{getOption("reproducible.verbose")}.}

\item{add}{\code{logical(1)}
If \code{TRUE}, adds the tag if it does not exist. Defaults to \code{TRUE}.}
}
\value{
\code{NULL} (invisibly). The function is called for its side effects.

\code{NULL} (invisibly). Called for its side effects.
}
\description{
This hidden function appends a single tag (key-value pair) to the metadata
of a cached object identified by its \code{cacheId}. Tags can be stored either in
a database (via DBI) or in a file-based cache system.

Updates the value of an existing tag for a cached object identified by its
\code{cacheId}. If the tag does not exist and \code{add = TRUE}, the tag will be added.
This function supports both database-backed and file-based cache systems.
}
\details{
This function is primarily used internally by the \code{reproducible} package to
maintain metadata about cached objects. It supports both database-backed and
file-based caching systems.

\itemize{
\item If \code{useDBI()} returns \code{TRUE}, the tag update is performed in the database table.
\item If no rows are affected and \code{add = TRUE}, the tag is inserted using \code{.addTagsRepo()}.
\item For file-based caches, the function modifies the tag in the corresponding metadata file.
}
}
\examples{
\dontrun{
a <- Cache(rnorm(1))
.addTagsRepo(cacheId = gsub("cacheId:", "", attr(a, "tags")),
             tagKey = "status", tagValue = "processed")
showCache() # last entry is the above line
}

\dontrun{
a <- Cache(rnorm(1))
# Update an existing tag
.updateTagsRepo(cacheId = gsub("cacheId:", "", attr(a, "tags")),
             tagKey = "status", tagValue = "second")

# Add a tag if it doesn't exist
.updateTagsRepo(cacheId = gsub("cacheId:", "", attr(a, "tags")),
             tagKey = "status", tagValue = "new", add = TRUE)
}

}
\seealso{
\code{\link[=.addTagsRepo]{.addTagsRepo()}} for adding tags without updating.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/copy.R
\name{Copy}
\alias{Copy}
\alias{Copy,ANY-method}
\alias{Copy,data.table-method}
\alias{Copy,list-method}
\alias{Copy,refClass-method}
\alias{Copy,data.frame-method}
\title{Recursive copying of nested environments, and other "hard to copy" objects}
\usage{
Copy(object, ...)

\S4method{Copy}{ANY}(
  object,
  filebackedDir,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  ...
)

\S4method{Copy}{data.table}(object, ...)

\S4method{Copy}{list}(object, ...)

\S4method{Copy}{refClass}(object, ...)

\S4method{Copy}{data.frame}(object, ...)
}
\arguments{
\item{object}{An R object (likely containing environments) or an environment.}

\item{...}{Only used for custom Methods}

\item{filebackedDir}{A directory to copy any files that are backing R objects,
currently only valid for \code{Raster} classes. Defaults
to \code{.reproducibleTempPath()}, which is unlikely to be very useful.
Can be \code{NULL}, which means that the file will not be
copied and could therefore cause a collision as the
pre-copied object and post-copied object would have the same
file backing them.}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\value{
The same object as \code{object}, but with pass-by-reference class elements "deep" copied.
\code{reproducible} has methods for several classes.
}
\description{
When copying environments and all the objects contained within them, there are
no copies made: it is a pass-by-reference operation. Sometimes, a deep copy is
needed, and sometimes, this must be recursive (i.e., environments inside
environments).
}
\details{
To create a new Copy method for a class that needs its own method, try something like
shown in example and put it in your package (or other R structure).
}
\examples{
e <- new.env()
e$abc <- letters
e$one <- 1L
e$lst <- list(W = 1:10, X = runif(10), Y = rnorm(10), Z = LETTERS[1:10])
ls(e)

# 'normal' copy
f <- e
ls(f)
f$one
f$one <- 2L
f$one
e$one ## uh oh, e has changed!

# deep copy
e$one <- 1L
g <- Copy(e)
ls(g)
g$one
g$one <- 3L
g$one
f$one
e$one
## To create a new deep copy method, use the following template
## setMethod("Copy", signature = "the class", # where = specify here if not in a package,
##           definition = function(object, filebackendDir, ...) {
##           # write deep copy code here
##           })

}
\seealso{
\code{\link[=.robustDigest]{.robustDigest()}}, \code{\link[=Filenames]{Filenames()}}
}
\author{
Eliot McIntire
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcessTo.R
\name{gdalProject}
\alias{gdalProject}
\alias{gdalResample}
\alias{gdalMask}
\title{3-Step postProcess sequence for SpatRasters using \code{gdalwarp}}
\usage{
gdalProject(
  fromRas,
  toRas,
  filenameDest,
  verbose = getOption("reproducible.verbose"),
  ...
)

gdalResample(
  fromRas,
  toRas,
  filenameDest,
  verbose = getOption("reproducible.verbose"),
  ...
)

gdalMask(
  fromRas,
  maskToVect,
  writeTo = NULL,
  verbose = getOption("reproducible.verbose"),
  ...
)
}
\arguments{
\item{fromRas}{see \code{from} argument from \code{\link[=postProcessTo]{postProcessTo()}}, but can only be a  \code{SpatRaster}.}

\item{toRas}{see \code{to} argument from \code{\link[=postProcessTo]{postProcessTo()}}, but can only be a  \code{SpatRaster}.}

\item{filenameDest}{A filename with an appropriate extension (e.g., \code{.tif}) for
\code{gdal} to write the output to. Since this function is conceived to be part of a
chain, and not the final step, this function does not use \code{writeTo}, which is
reserved for the final step in the chain.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{For \code{gdalProject}, this can be \code{method}. For \code{gdalMask} can be \code{destinationPath} and \code{touches}.
For all \verb{gdal*}, this can also be and \code{datatype}.}

\item{maskToVect}{see \code{maskTo} argument from \code{\link[=maskTo]{maskTo()}}, but can only be a \code{SpatVector}}

\item{writeTo}{Optional character string of a filename to use \code{writeRaster} to save the final
object. Default is \code{NULL}, which means there is no \code{writeRaster}}
}
\description{
DEFUNCT: Please use the \code{postProcessTo} functions.

\code{gdalResample} is a thin wrapper around \code{sf::gdal_utils('gdalwarp', ...)} with specific options
set, notably, \verb{"-r", "near"}, \code{-te}, \code{-te_srs}, \code{tr}, \code{-dstnodata = NA}, \code{-overwrite}.

\code{gdalMask} is a thin wrapper around \code{sf::gdal_utils('gdalwarp', ...)} with specific options
set, notably, \code{-cutline}, \code{-dstnodata = NA}, and \code{-overwrite}.
}
\details{
\code{gdalProject} is a thin wrapper around \code{sf::gdal_utils('gdalwarp', ...)} with specific options
set, notably, \code{-r} to \code{method} (in the ...), \code{-t_srs} to the crs of the \code{toRas},
\code{-te} to the extent of the \code{toRas}, \code{-te_srs} to the \code{crs} of the \code{toRas},
\code{-dstnodata = NA}, and \code{-overwrite}.

These three functions are used within \code{postProcessTo}, in the sequence:
\code{gdalProject}, \code{gdalResample} and \code{gdalMask}, when \code{from} and \code{projectTo} are \code{SpatRaster} and
\code{maskTo} is a \code{SpatVector}, but only if \code{options(reproducible.gdalwarp = TRUE)} is set.

This sequence is a slightly different order than the sequence when \code{gdalwarp = FALSE} or
the arguments do not match the above. This sequence was determined to be faster and
more accurate than any other sequence, including running all three steps in one
\code{gdalwarp} call (which \code{gdalwarp} can do). Using one-step \code{gdalwarp} resulted in
very coarse pixelation when converting from a coarse resolution to fine resolution, which
visually was inappropriate in test cases.
}
\examples{
if (require("terra", quietly = TRUE)) {
  # prepare dummy data -- 3 SpatRasters, 2 SpatVectors
  # need 2 SpatRaster
  rf <- system.file("ex/elev.tif", package = "terra")
  elev1 <- terra::rast(rf)

  # a polygon vector
  f <- system.file("ex/lux.shp", package = "terra")
  vOrig <- terra::vect(f)
  v <- vOrig[1:2, ]

  # utm <- terra::crs("epsg:23028") # $wkt
  utm <- "+proj=utm +zone=28 +datum=WGS84 +units=m +no_defs"
  vInUTM <- terra::project(vOrig, utm)
  vAsRasInLongLat <- terra::rast(vOrig, resolution = 0.008333333)
  res100 <- 100
  rInUTM <- terra::rast(vInUTM, resolution = res100, vals = 1)
  # crop, reproject, mask, crop a raster with a vector in a different projection
  #  --> gives message about not enough information
  t1 <- postProcessTo(elev1, to = vInUTM)
  # crop, reproject, mask a raster to a different projection, then mask
  t2a <- postProcessTo(elev1, to = vAsRasInLongLat, maskTo = vInUTM)
  t3a <- postProcessTo(elev1, to = rInUTM, maskTo = vInUTM)

}
}
\seealso{
\code{\link[=gdalResample]{gdalResample()}}, and \code{\link[=gdalMask]{gdalMask()}} and the overarching \code{\link[=postProcessTo]{postProcessTo()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportedMethods.R
\name{.wrap}
\alias{.wrap}
\alias{.wrap.list}
\alias{.wrap.environment}
\alias{.wrap.default}
\alias{.unwrap.default}
\alias{.unwrap}
\alias{.unwrap.environment}
\alias{.unwrap.list}
\alias{.unwrap.PackedSpatExtent2}
\alias{.unwrap.PackedSpatVector2}
\alias{.unwrap.data.table}
\alias{.unwrap.PackedSpatVector}
\title{Deal with class for saving to and loading from Cache or Disk}
\usage{
.wrap(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  preDigest,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  outputObjects = NULL,
  cacheId = NULL,
  ...
)

\method{.wrap}{list}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  preDigest,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  outputObjects = NULL,
  cacheId = NULL,
  ...
)

\method{.wrap}{environment}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  preDigest,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  outputObjects = NULL,
  cacheId = NULL,
  ...
)

\method{.wrap}{default}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  preDigest,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose"),
  outputObjects = NULL,
  cacheId = NULL,
  ...
)

\method{.unwrap}{default}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  ...
)

.unwrap(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  ...
)

\method{.unwrap}{environment}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  ...
)

\method{.unwrap}{list}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  ...
)

\method{.unwrap}{PackedSpatExtent2}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  ...
)

\method{.unwrap}{PackedSpatVector2}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  ...
)

\method{.unwrap}{data.table}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  ...
)

\method{.unwrap}{PackedSpatVector}(
  obj,
  cachePath = getOption("reproducible.cachePath"),
  cacheId = NULL,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  ...
)
}
\arguments{
\item{obj}{Any arbitrary R object.}

\item{cachePath}{A repository used for storing cached objects.
This is optional if \code{Cache} is used inside a SpaDES module.}

\item{preDigest}{The list of \code{preDigest} that comes from \code{CacheDigest} of an object}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{outputObjects}{Optional character vector indicating which objects to
return. This is only relevant for list, environment (or similar) objects}

\item{cacheId}{Used strictly for messaging. This should be the cacheId of the object being recovered.
Default is \code{NULL}.}

\item{...}{Arguments passed to methods; default does not use anything in \code{...}.}
}
\value{
Returns an object that can be saved to disk e.g., via \code{saveRDS}.
}
\description{
This generic and some methods will do whatever is required to prepare an object for
saving to disk (or RAM) via e.g., \code{saveRDS}. Some objects (e.g., \code{terra}'s \verb{Spat*})
cannot be saved without first wrapping them. Also, file-backed objects are similar.
}
\examples{
# For SpatExtent
if (requireNamespace("terra")) {
  ex <- terra::ext(c(0, 2, 0, 3))
  exWrapped <- .wrap(ex)
  ex1 <- .unwrap(exWrapped)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cloud.R
\name{cloudDownload}
\alias{cloudDownload}
\title{Download from cloud, if necessary}
\usage{
cloudDownload(
  outputHash,
  newFileName,
  gdriveLs,
  cachePath,
  cloudFolderID,
  drv = getDrv(getOption("reproducible.drv", NULL)),
  conn = getOption("reproducible.conn", NULL),
  verbose = getOption("reproducible.verbose")
)
}
\arguments{
\item{outputHash}{The \code{cacheId} of the object to upload}

\item{newFileName}{The character string of the local filename that the downloaded object will have}

\item{gdriveLs}{The result of \code{googledrive::drive_ls(googledrive::as_id(cloudFolderID), pattern = "outputHash")}}

\item{cachePath}{A repository used for storing cached objects.
This is optional if \code{Cache} is used inside a SpaDES module.}

\item{cloudFolderID}{A googledrive dribble of a folder, e.g., using \code{drive_mkdir()}.
If left as \code{NULL}, the function will create a cloud folder with name from last
two folder levels of the \code{cachePath} path, :
\code{paste0(basename(dirname(cachePath)), "_", basename(cachePath))}.
This \code{cloudFolderID} will be added to \code{options("reproducible.cloudFolderID")},
but this will not persist across sessions. If this is a character string, it will
treat this as a folder name to create or use on GoogleDrive.}

\item{drv}{If using a database backend, \code{drv} must be an object that
inherits from \code{DBIDriver} (e.g., \code{RSQLite::SQLite}).}

\item{conn}{an optional \code{DBIConnection} object, as returned by \code{dbConnect()}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}
}
\description{
Meant for internal use, as there are internal objects as arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/showCacheEtc.R
\name{prepopulateCacheAsync}
\alias{prepopulateCacheAsync}
\title{Pre-populate the in-memory \code{showCache} cache for a given \code{cachePath}}
\usage{
prepopulateCacheAsync(cachePath = getOption("reproducible.cachePath"))
}
\arguments{
\item{cachePath}{A character path. Defaults to
\code{getOption("reproducible.cachePath")}.}
}
\value{
Invisibly returns the spawn job handle, or \code{NULL} if the spawn
was skipped.
}
\description{
Forks a background process that runs \code{showCache()} against \code{cachePath};
subsequent \code{showCache()} / \code{Cache()->showSimilar()} calls in the same R
session can then harvest the result instead of re-scanning the cache
directory synchronously. Useful for very large caches (tens of thousands
of entries) where the cold first scan can take a minute or more.
}
\details{
Idempotent: a second call with the same \code{cachePath} reuses the existing
job. Skipped silently on Windows (forking-based) and when the \code{parallel}
package isn't available.

This helper is called automatically the first time \code{Cache()} or
\code{showCache()} is invoked against a given \code{cachePath}, so most users do
not need to call it explicitly. It is exported for workflows that want
to kick off the spawn early (e.g. inside \code{setupProject()}) so the fork
has more wall-clock time to complete before the first manual
\code{showCache()} call.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcess.R
\name{deprecated}
\alias{deprecated}
\alias{cropInputs}
\alias{cropInputs.default}
\alias{fixErrors}
\alias{fixErrors.default}
\alias{projectInputs}
\alias{projectInputs.default}
\alias{maskInputs}
\alias{writeOutputs}
\alias{writeOutputs.default}
\title{Crop a \verb{Spatial*} or \verb{Raster*} object}
\usage{
cropInputs(
  x,
  studyArea,
  rasterToMatch,
  verbose = getOption("reproducible.verbose", 1),
  ...
)

\method{cropInputs}{default}(x, ...)

fixErrors(
  x,
  objectName,
  attemptErrorFixes = TRUE,
  useCache = getOption("reproducible.useCache", FALSE),
  verbose = getOption("reproducible.verbose", 1),
  testValidity = getOption("reproducible.testValidity", TRUE),
  ...
)

\method{fixErrors}{default}(
  x,
  objectName,
  attemptErrorFixes = TRUE,
  useCache = getOption("reproducible.useCache", FALSE),
  verbose = getOption("reproducible.verbose", 1),
  testValidity = getOption("reproducible.testValidity", TRUE),
  ...
)

projectInputs(
  x,
  targetCRS,
  verbose = getOption("reproducible.verbose", 1),
  ...
)

\method{projectInputs}{default}(x, targetCRS, ...)

maskInputs(x, studyArea, ...)

writeOutputs(
  x,
  ...,
  overwrite = getOption("reproducible.overwrite", NULL),
  verbose = getOption("reproducible.verbose", 1)
)

\method{writeOutputs}{default}(
  x,
  ...,
  overwrite = getOption("reproducible.overwrite", FALSE),
  verbose = getOption("reproducible.verbose", 1)
)
}
\arguments{
\item{x}{The object save to disk i.e., write outputs}

\item{studyArea}{\verb{SpatialPolygons*} object used for masking and possibly cropping
if no \code{rasterToMatch} is provided.
If not in same CRS, then it will be \code{spTransform}ed to
CRS of \code{x} before masking. Currently, this function will not reproject the
\code{x}. Optional in \code{postProcess}.}

\item{rasterToMatch}{Template \verb{Raster*} object passed to the \code{to} argument of
\code{\link[=projectTo]{projectTo()}}, thus will changing the
resolution and projection of \code{x}.
See details in \code{\link[=postProcessTo]{postProcessTo()}}.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{...}{Passed into \code{\link[=writeTo]{writeTo()}}}

\item{objectName}{Optional. This is only for messaging; if provided, then messages relayed
to user will mention this.}

\item{attemptErrorFixes}{Will attempt to fix known errors. Currently only some failures
for \verb{SpatialPolygons*} are attempted.
Notably with \code{terra::buffer(..., width = 0)}.
Default \code{TRUE}, though this may not be the right action for all cases.}

\item{useCache}{Logical, default \code{getOption("reproducible.useCache", FALSE)}, whether
Cache is used on the internal \code{terra::buffer} command.}

\item{testValidity}{Logical. If \code{TRUE}, the a test for validity will happen
before actually running buffering (which is the solution in most
cases). However, sometimes it takes longer to test for validity
than just buffer without testing (there are no consequences of
buffering if everything is valid). If \code{FALSE}, then the
test will be skipped and the buffering will happen regardless.
If \code{NA}, then all testing and buffering will be skipped.}

\item{targetCRS}{The CRS of x at the end  of this function (i.e., the goal)}

\item{overwrite}{Logical. Should file being written overwrite an existing file if it exists.}
}
\value{
A GIS file (e.g., RasterLayer, SpatRaster etc.) that has been
appropriately cropped.

A GIS file (e.g., RasterLayer, SpatRaster etc.) that has been
attempted to be fixed, if it finds errors.

A file of the same type as starting, but with projection (and possibly
other characteristics, including resolution, origin, extent if changed).

A GIS file (e.g., RasterLayer, SpatRaster etc.) that has been
appropriately reprojected.

A GIS file (e.g., RasterLayer, SpatRaster etc.) that has been
appropriately masked.

A GIS file (e.g., SpatRaster etc.) that has been
appropriately written to disk. In the case of vector datasets, this will
be a side effect. In the case of gridded objects (Raster*, SpatRaster), the
object will have a file-backing.
}
\description{
This function is deprecated. Use \code{cropTo}. If used, all arguments will be passed
to \code{cropTo} anyway.

\code{fixErrors} --> \code{fixErrosTerra}

Deprecated. Use \code{\link[=projectTo]{projectTo()}}.

\code{maskInputs} is deprecated. Use \code{\link[=maskTo]{maskTo()}}

See \code{\link[=writeTo]{writeTo()}}
}
\examples{
if (requireNamespace("terra", quietly = TRUE)) {
  r <- terra::rast(terra::ext(0, 100, 0, 100), vals = 1:1e2)

  tf <- tempfile(fileext = ".tif")
  writeOutputs(r, tf)
}
}
\seealso{
\code{\link[=fixErrorsIn]{fixErrorsIn()}}, \code{\link[=postProcessTo]{postProcessTo()}}, \code{\link[=postProcess]{postProcess()}}

\code{\link[=projectTo]{projectTo()}}

\code{\link[=maskTo]{maskTo()}}, \code{\link[=postProcessTo]{postProcessTo()}} for related examples
}
\author{
Eliot McIntire, Jean Marchal, Ian Eddy, and Tati Micheletti

Eliot McIntire and Jean Marchal
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{.prefix}
\alias{.prefix}
\alias{suffix}
\alias{.suffix}
\title{Add a prefix or suffix to the basename part of a file path}
\usage{
.prefix(f, prefix = "")

.suffix(f, suffix = "")
}
\arguments{
\item{f}{A character string giving the name/path of a file.}

\item{prefix}{A character string to prepend to the filename.}

\item{suffix}{A character string to postpend to the filename.}
}
\value{
A character string or vector with the prefix pre-pended or suffix post-pended
on the \code{basename} of the \code{f}, before the file extension.
}
\description{
Prepend (or postpend) a filename with a prefix (or suffix).
If the directory name of the file cannot be ascertained from its path,
it is assumed to be in the current working directory.
}
\examples{
# file's full path is specified (i.e., dirname is known)
myFile <- file.path("~/data", "file.tif")
.prefix(myFile, "small_") ## "/home/username/data/small_file.tif"
.suffix(myFile, "_cropped") ## "/home/username/data/myFile_cropped.shp"

# file's full path is not specified
.prefix("myFile.shp", "small") ## "./small_myFile.shp"
.suffix("myFile.shp", "_cropped") ## "./myFile_cropped.shp"

}
\author{
Jean Marchal and Alex Chubaty
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcessTo.R
\name{keepOrigGeom}
\alias{keepOrigGeom}
\title{Keep original geometries of \code{sf} objects}
\usage{
keepOrigGeom(newObj, origObj)
}
\arguments{
\item{newObj}{The new, derived \code{sf} object}

\item{origObj}{The previous, object whose geometries should be used.}
}
\value{
The original \code{newObj}, but with only the type of geometry that entered
into the function.
}
\description{
When intersections occur, what was originally 2 polygons features can become
\code{LINESTRING} and/or \code{POINT} and any \code{COLLECTIONS} or \verb{MULTI-} versions of these.
This function evaluates what the original geometry was and drops any newly created
\emph{different} geometries. For example, if a \code{POLYGON} becomes a \code{COLLECTION} of
\code{MULTIPOLYGON}, \code{POLYGON} and \code{POINT} geometries, the \code{POINT} geometries will
be dropped. This function is used internally in \code{\link[=postProcessTo]{postProcessTo()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{detectActiveCores}
\alias{detectActiveCores}
\title{Count Active Threads Based on CPU Usage}
\usage{
detectActiveCores(pattern = "", minCPU = 50)
}
\arguments{
\item{pattern}{A character string used to filter process lines. Only
processes whose command line matches this pattern will be considered.
Default is \code{""} (matches all).}

\item{minCPU}{A numeric value specifying the minimum CPU usage (in percent)
for a process to be considered active. Default is \code{50}.}
}
\value{
An integer representing the number of active threads matching the
pattern and exceeding the CPU usage threshold. Returns \code{NULL} with a
message if run on Windows.
}
\description{
This function counts the number of active system processes (threads) that
match a given pattern and exceed a specified minimum CPU usage threshold. It
works on Unix-like systems (e.g., Linux, macOS) and does not support Windows.
}
\note{
This function uses the \code{ps -ef} system command and regular expressions
to parse CPU usage. It may not be portable across all Unix variants.
}
\examples{
\dontrun{
  detectActiveCores(pattern = "R", minCPU = 30)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prepInputs.R
\name{extractFromArchive}
\alias{extractFromArchive}
\title{Extract files from archive}
\usage{
extractFromArchive(
  archive,
  destinationPath = getOption("reproducible.destinationPath", dirname(archive)),
  neededFiles = NULL,
  extractedArchives = NULL,
  checkSums = NULL,
  needChecksums = 0,
  filesExtracted = character(),
  checkSumFilePath = character(),
  quick = FALSE,
  verbose = getOption("reproducible.verbose", 1),
  .tempPath,
  ...
)
}
\arguments{
\item{archive}{Character string giving the path of the archive
containing the \code{file} to be extracted. This path must exist or be \code{NULL}}

\item{destinationPath}{Character string giving the path where \code{neededFiles} will be
extracted. Defaults to the archive directory.}

\item{neededFiles}{Character string giving the name of the file(s) to be extracted.}

\item{extractedArchives}{Used internally to track archives that have been extracted from.}

\item{checkSums}{A checksums file, e.g., created by Checksums(..., write = TRUE)}

\item{needChecksums}{A numeric, with \code{0} indicating do not write a new checksums,
\code{1} write a new one,
\code{2} append new information to existing one.}

\item{filesExtracted}{Used internally to track files that have been extracted.}

\item{checkSumFilePath}{The full path to the checksum.txt file}

\item{quick}{Passed to \code{Checksums}}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{.tempPath}{Optional temporary path for internal file intermediate steps.
Will be cleared \code{on.exit} from this function.}

\item{...}{Passed to \code{unzip} or \code{untar}, e.g., \code{overwrite}}
}
\value{
A character vector listing the paths of the extracted archives.
}
\description{
Extract zip or tar archive files, possibly nested in other zip or tar archives.
}
\author{
Jean Marchal and Eliot McIntire
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{tempdir2}
\alias{tempdir2}
\title{Make a temporary (sub-)directory}
\usage{
tempdir2(
  sub = "",
  tempdir = getOption("reproducible.tempPath", .reproducibleTempPath()),
  create = TRUE
)
}
\arguments{
\item{sub}{Character string, length 1. Can be a result of
\code{file.path("smth", "smth2")} for nested temporary subdirectories.
If the zero length character, then a random sub-directory will be created.}

\item{tempdir}{Optional character string where the temporary
directory should be placed. Defaults to \code{getOption("reproducible.tempPath")}.}

\item{create}{Logical. Should the directory be created. Default \code{TRUE}.}
}
\value{
A character string of a path (that will be created if \code{create = TRUE}) in a
sub-directory of the \code{tempdir()}.
}
\description{
Create a temporary subdirectory in \code{getOption("reproducible.tempPath")}.
}
\seealso{
\link{tempfile2}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/postProcessTo.R
\name{fixErrorsIn}
\alias{fixErrorsIn}
\title{Fix common errors in GIS layers, using \code{terra}}
\usage{
fixErrorsIn(
  x,
  error = NULL,
  verbose = getOption("reproducible.verbose"),
  fromFnName = ""
)
}
\arguments{
\item{x}{The SpatStat or SpatVect object to try to fix.}

\item{error}{The error message, e.g., coming from \code{try(...)}}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{fromFnName}{The function name that produced the error, e.g., \code{maskTo}}
}
\value{
An object of the same class as \code{x}, but with some errors fixed via \code{terra::makeValid()}
}
\description{
Currently, this only tests for validity of a SpatVect file, then if there is a problem,
it will run \code{terra::makeValid}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{basename2}
\alias{basename2}
\title{A version of \code{base::basename} that is \code{NULL} resistant}
\usage{
basename2(x)
}
\arguments{
\item{x}{A character vector of paths}
}
\value{
\code{NULL} if x is \code{NULL}, otherwise, as \code{basename}.

Same as \code{\link[base:basename]{base::basename()}}
}
\description{
A version of \code{base::basename} that is \code{NULL} resistant
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preProcess.R
\name{preProcessParams}
\alias{preProcessParams}
\alias{preProcess}
\title{Download, checksum, extract files}
\usage{
preProcessParams(n = NULL)

preProcess(
  targetFile = NULL,
  url = NULL,
  archive = NULL,
  alsoExtract = NULL,
  destinationPath = getOption("reproducible.destinationPath", "."),
  fun = NULL,
  dlFun = NULL,
  quick = getOption("reproducible.quick"),
  overwrite = getOption("reproducible.overwrite", FALSE),
  purge = FALSE,
  verbose = getOption("reproducible.verbose", 1),
  .tempPath,
  .callingEnv = parent.frame(),
  ...
)
}
\arguments{
\item{n}{Number of non-null arguments passed to \code{preProcess}.
E.g., passing \code{n = 1} returns combinations with only a single non-NULL parameter.
If \code{NULL} (default), all parameter combinations are returned.}

\item{targetFile}{Character string giving the filename (without relative or
absolute path) to the eventual file
(raster, shapefile, csv, etc.) after downloading and extracting from a zip
or tar archive. This is the file \emph{before} it is passed to
\code{postProcess}. The internal checksumming does not checksum
the file after it is \code{postProcess}ed (e.g., cropped/reprojected/masked).
Using \code{Cache} around \code{prepInputs} will do a sufficient job in these cases.
See table in \code{\link[=preProcess]{preProcess()}}.}

\item{url}{Optional character string indicating the URL to download from.
If not specified, then no download will be attempted. If not entry
exists in the \code{CHECKSUMS.txt} (in \code{destinationPath}), an entry
will be created or appended to. This \code{CHECKSUMS.txt} entry will be used
in subsequent calls to
\code{prepInputs} or \code{preProcess}, comparing the file on hand with the ad hoc
\code{CHECKSUMS.txt}. See table in \code{\link[=preProcess]{preProcess()}}.}

\item{archive}{Optional character string giving the path of an archive
containing \code{targetFile}, or a vector giving a set of nested archives
(e.g., \code{c("xxx.tar", "inner.zip", "inner.rar")}). If there is/are (an) inner
archive(s), but they are unknown, the function will try all until it finds
the \code{targetFile}. See table in \code{\link[=preProcess]{preProcess()}}. If it is \code{NA},
then it will \emph{not} attempt to see it as an archive, even if it has archive-like
file extension (e.g., \code{.zip}). This may be useful when an R function
is expecting an archive directly.}

\item{alsoExtract}{Optional character string naming files other than
\code{targetFile} that must be extracted from the \code{archive}. If
\code{NULL}, the default, then it will extract all files. Other options:
\code{"similar"} will extract all files with the same filename without
file extension as \code{targetFile}. \code{NA} will extract nothing other
than \code{targetFile}. A character string of specific file names will cause
only those to be extracted. Each element may also be a regular expression:
if an element does not match any archive member literally (by relative path
or basename), it is passed to \code{grep()} against the archive's file list and
all matching members are extracted. For example,
\code{alsoExtract = "CMD_sm|CMD_sp"} extracts every file whose name contains
\code{CMD_sm} or \code{CMD_sp}. See table in \code{\link[=preProcess]{preProcess()}}.}

\item{destinationPath}{Character string of a directory in which to download
and save the file that comes from \code{url} and is also where the function
will look for \code{archive} or \code{targetFile}. NOTE (still experimental):
To prevent repeated downloads in different locations, the user can also set
\code{options("reproducible.destinationPathShared")} to one or more local file paths to
search for the file before attempting to download. Default for that option is
\code{NULL} meaning do not search locally. The previous name
\code{options("reproducible.inputPaths")} is still accepted as a backwards-compatible
alias.}

\item{fun}{Optional. If specified, this will attempt to load whatever
file was downloaded during \code{preProcess} via \code{dlFun}. This can be either a
function (e.g., sf::st_read), character string (e.g., "base::load"),
NA (for no loading, useful if \code{dlFun} already loaded the file) or
if extra arguments are required
in the function call, it must be a call naming
\code{targetFile} (e.g., \code{sf::st_read(targetFile, quiet = TRUE)})
as the file path to the file to load. See details and examples below.}

\item{dlFun}{Optional "download function" name, such as \code{"raster::getData"}, which does
custom downloading, in addition to loading into R. Still experimental.}

\item{quick}{Logical. This is passed internally to \code{\link[=Checksums]{Checksums()}}
(the quickCheck argument), and to
\code{\link[=Cache]{Cache()}} (the quick argument). This results in faster, though
less robust checking of inputs. See the respective functions.}

\item{overwrite}{Logical. Passed to \code{writeTo} (possibly inside \code{postProcess}) and \code{postProcess}.}

\item{purge}{Logical or Integer. \code{0/FALSE} (default) keeps existing \code{CHECKSUMS.txt} file and
\code{prepInputs} will write or append to it. \code{1/TRUE} will deleted the entire \code{CHECKSUMS.txt} file.
Other options, see details.}

\item{verbose}{Numeric, -1 silent (where possible), 0 being very quiet,
1 showing more messaging, 2 being more messaging, etc.
Default is 1. Above 3 will output much more information about the internals of
Caching, which may help diagnose Caching challenges. Can set globally with an
option, e.g., \verb{options('reproducible.verbose' = 0) to reduce to minimal}}

\item{.tempPath}{Optional temporary path for internal file intermediate steps.
Will be cleared \code{on.exit} from this function.}

\item{.callingEnv}{The environment where the function was called from. Used to find
objects, if necessary.}

\item{...}{Additional arguments passed to
\code{\link[=postProcess]{postProcess()}} and \code{\link[=Cache]{Cache()}}.
Since \code{...} is passed to \code{\link[=postProcess]{postProcess()}}, these will
\code{...} will also be passed into the inner
functions, e.g., \code{\link[=cropInputs]{cropInputs()}}. Possibly useful other arguments include
\code{dlFun} which is passed to \code{preProcess}. See details and examples.}
}
\value{
A list with 5 elements: \code{checkSums} (the result of a \code{Checksums}
after downloading), \code{dots} (cleaned up \code{...}, including deprecated argument checks),
\code{fun} (the function to be used to load the \code{preProcess}ed object from disk),
and \code{targetFilePath} (the fully qualified path to the \code{targetFile}).
}
\description{
This does downloading (via \code{downloadFile}), checksumming (\code{Checksums}),
and extracting from archives (\code{extractFromArchive}), plus cleaning up of input
arguments (e.g., paths, function names).
This is the first stage of three used in \code{prepInputs}.
}
\section{Combinations of \code{targetFile}, \code{url}, \code{archive}, \code{alsoExtract}}{


Use \code{preProcessParams()} for a table describing various parameter combinations and their
outcomes.

\code{*} If the \code{url} is a file on Google Drive, checksumming will work
even without a \code{targetFile} specified because there is an initial attempt
to get the remove file information (e.g., file name). With that, the connection
between the \code{url} and the filename used in the \file{CHECKSUMS.txt} file can be made.
}

\author{
Eliot McIntire
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checksums.R
\name{.digest}
\alias{.digest}
\alias{.digest,character-method}
\title{Calculate the hashes of multiple files}
\usage{
.digest(file, quickCheck = FALSE, ...)

\S4method{.digest}{character}(file, quickCheck = FALSE, algo = "xxhash64", ...)
}
\arguments{
\item{file}{Character vector of file paths.}

\item{quickCheck}{Logical indicating whether to use a fast file size check as a heuristic
for determining changes to a file.}

\item{...}{Additional arguments to \code{\link[digest:digest]{digest::digest()}}.}
}
\value{
A character vector of hashes.
}
\description{
Internal function. Wrapper for \code{\link[digest:digest]{digest::digest()}} using \code{algo = xxhash64}.
}
\author{
Alex Chubaty
}
\keyword{internal}
